// Code generated by "stringer -type=ImageFormat"; DO NOT EDIT.

package imagemeta

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ImageFormatAuto-0]
	_ = x[JPEG-1]
	_ = x[TIFF-2]
	_ = x[PNG-3]
	_ = x[WebP-4]
}

const _ImageFormat_name = "ImageFormatAutoJPEGTIFFPNGWebP"

var _ImageFormat_index = [...]uint8{0, 15, 19, 23, 26, 30}

func (i ImageFormat) String() string {
	if i < 0 || i >= ImageFormat(len(_ImageFormat_index)-1) {
		return "ImageFormat(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ImageFormat_name[_ImageFormat_index[i]:_ImageFormat_index[i+1]]
}
