// This file is autogenerated.
package localescompressed

import "github.com/gohugoio/locales/currency"

var currencies = map[string]currency.Type{
	"ADP": currency.ADP,
	"AED": currency.AED,
	"AFA": currency.AFA,
	"AFN": currency.AFN,
	"ALK": currency.ALK,
	"ALL": currency.ALL,
	"AMD": currency.AMD,
	"ANG": currency.ANG,
	"AOA": currency.AOA,
	"AOK": currency.AOK,
	"AON": currency.AON,
	"AOR": currency.AOR,
	"ARA": currency.ARA,
	"ARL": currency.ARL,
	"ARM": currency.ARM,
	"ARP": currency.ARP,
	"ARS": currency.ARS,
	"ATS": currency.ATS,
	"AUD": currency.AUD,
	"AWG": currency.AWG,
	"AZM": currency.AZM,
	"AZN": currency.AZN,
	"BAD": currency.BAD,
	"BAM": currency.BAM,
	"BAN": currency.BAN,
	"BBD": currency.BBD,
	"BDT": currency.BDT,
	"BEC": currency.BEC,
	"BEF": currency.BEF,
	"BEL": currency.BEL,
	"BGL": currency.BGL,
	"BGM": currency.BGM,
	"BGN": currency.BGN,
	"BGO": currency.BGO,
	"BHD": currency.BHD,
	"BIF": currency.BIF,
	"BMD": currency.BMD,
	"BND": currency.BND,
	"BOB": currency.BOB,
	"BOL": currency.BOL,
	"BOP": currency.BOP,
	"BOV": currency.BOV,
	"BRB": currency.BRB,
	"BRC": currency.BRC,
	"BRE": currency.BRE,
	"BRL": currency.BRL,
	"BRN": currency.BRN,
	"BRR": currency.BRR,
	"BRZ": currency.BRZ,
	"BSD": currency.BSD,
	"BTN": currency.BTN,
	"BUK": currency.BUK,
	"BWP": currency.BWP,
	"BYB": currency.BYB,
	"BYN": currency.BYN,
	"BYR": currency.BYR,
	"BZD": currency.BZD,
	"CAD": currency.CAD,
	"CDF": currency.CDF,
	"CHE": currency.CHE,
	"CHF": currency.CHF,
	"CHW": currency.CHW,
	"CLE": currency.CLE,
	"CLF": currency.CLF,
	"CLP": currency.CLP,
	"CNH": currency.CNH,
	"CNX": currency.CNX,
	"CNY": currency.CNY,
	"COP": currency.COP,
	"COU": currency.COU,
	"CRC": currency.CRC,
	"CSD": currency.CSD,
	"CSK": currency.CSK,
	"CUC": currency.CUC,
	"CUP": currency.CUP,
	"CVE": currency.CVE,
	"CYP": currency.CYP,
	"CZK": currency.CZK,
	"DDM": currency.DDM,
	"DEM": currency.DEM,
	"DJF": currency.DJF,
	"DKK": currency.DKK,
	"DOP": currency.DOP,
	"DZD": currency.DZD,
	"ECS": currency.ECS,
	"ECV": currency.ECV,
	"EEK": currency.EEK,
	"EGP": currency.EGP,
	"ERN": currency.ERN,
	"ESA": currency.ESA,
	"ESB": currency.ESB,
	"ESP": currency.ESP,
	"ETB": currency.ETB,
	"EUR": currency.EUR,
	"FIM": currency.FIM,
	"FJD": currency.FJD,
	"FKP": currency.FKP,
	"FRF": currency.FRF,
	"GBP": currency.GBP,
	"GEK": currency.GEK,
	"GEL": currency.GEL,
	"GHC": currency.GHC,
	"GHS": currency.GHS,
	"GIP": currency.GIP,
	"GMD": currency.GMD,
	"GNF": currency.GNF,
	"GNS": currency.GNS,
	"GQE": currency.GQE,
	"GRD": currency.GRD,
	"GTQ": currency.GTQ,
	"GWE": currency.GWE,
	"GWP": currency.GWP,
	"GYD": currency.GYD,
	"HKD": currency.HKD,
	"HNL": currency.HNL,
	"HRD": currency.HRD,
	"HRK": currency.HRK,
	"HTG": currency.HTG,
	"HUF": currency.HUF,
	"IDR": currency.IDR,
	"IEP": currency.IEP,
	"ILP": currency.ILP,
	"ILR": currency.ILR,
	"ILS": currency.ILS,
	"INR": currency.INR,
	"IQD": currency.IQD,
	"IRR": currency.IRR,
	"ISJ": currency.ISJ,
	"ISK": currency.ISK,
	"ITL": currency.ITL,
	"JMD": currency.JMD,
	"JOD": currency.JOD,
	"JPY": currency.JPY,
	"KES": currency.KES,
	"KGS": currency.KGS,
	"KHR": currency.KHR,
	"KMF": currency.KMF,
	"KPW": currency.KPW,
	"KRH": currency.KRH,
	"KRO": currency.KRO,
	"KRW": currency.KRW,
	"KWD": currency.KWD,
	"KYD": currency.KYD,
	"KZT": currency.KZT,
	"LAK": currency.LAK,
	"LBP": currency.LBP,
	"LKR": currency.LKR,
	"LRD": currency.LRD,
	"LSL": currency.LSL,
	"LTL": currency.LTL,
	"LTT": currency.LTT,
	"LUC": currency.LUC,
	"LUF": currency.LUF,
	"LUL": currency.LUL,
	"LVL": currency.LVL,
	"LVR": currency.LVR,
	"LYD": currency.LYD,
	"MAD": currency.MAD,
	"MAF": currency.MAF,
	"MCF": currency.MCF,
	"MDC": currency.MDC,
	"MDL": currency.MDL,
	"MGA": currency.MGA,
	"MGF": currency.MGF,
	"MKD": currency.MKD,
	"MKN": currency.MKN,
	"MLF": currency.MLF,
	"MMK": currency.MMK,
	"MNT": currency.MNT,
	"MOP": currency.MOP,
	"MRO": currency.MRO,
	"MRU": currency.MRU,
	"MTL": currency.MTL,
	"MTP": currency.MTP,
	"MUR": currency.MUR,
	"MVP": currency.MVP,
	"MVR": currency.MVR,
	"MWK": currency.MWK,
	"MXN": currency.MXN,
	"MXP": currency.MXP,
	"MXV": currency.MXV,
	"MYR": currency.MYR,
	"MZE": currency.MZE,
	"MZM": currency.MZM,
	"MZN": currency.MZN,
	"NAD": currency.NAD,
	"NGN": currency.NGN,
	"NIC": currency.NIC,
	"NIO": currency.NIO,
	"NLG": currency.NLG,
	"NOK": currency.NOK,
	"NPR": currency.NPR,
	"NZD": currency.NZD,
	"OMR": currency.OMR,
	"PAB": currency.PAB,
	"PEI": currency.PEI,
	"PEN": currency.PEN,
	"PES": currency.PES,
	"PGK": currency.PGK,
	"PHP": currency.PHP,
	"PKR": currency.PKR,
	"PLN": currency.PLN,
	"PLZ": currency.PLZ,
	"PTE": currency.PTE,
	"PYG": currency.PYG,
	"QAR": currency.QAR,
	"RHD": currency.RHD,
	"ROL": currency.ROL,
	"RON": currency.RON,
	"RSD": currency.RSD,
	"RUB": currency.RUB,
	"RUR": currency.RUR,
	"RWF": currency.RWF,
	"SAR": currency.SAR,
	"SBD": currency.SBD,
	"SCR": currency.SCR,
	"SDD": currency.SDD,
	"SDG": currency.SDG,
	"SDP": currency.SDP,
	"SEK": currency.SEK,
	"SGD": currency.SGD,
	"SHP": currency.SHP,
	"SIT": currency.SIT,
	"SKK": currency.SKK,
	"SLL": currency.SLL,
	"SOS": currency.SOS,
	"SRD": currency.SRD,
	"SRG": currency.SRG,
	"SSP": currency.SSP,
	"STD": currency.STD,
	"STN": currency.STN,
	"SUR": currency.SUR,
	"SVC": currency.SVC,
	"SYP": currency.SYP,
	"SZL": currency.SZL,
	"THB": currency.THB,
	"TJR": currency.TJR,
	"TJS": currency.TJS,
	"TMM": currency.TMM,
	"TMT": currency.TMT,
	"TND": currency.TND,
	"TOP": currency.TOP,
	"TPE": currency.TPE,
	"TRL": currency.TRL,
	"TRY": currency.TRY,
	"TTD": currency.TTD,
	"TWD": currency.TWD,
	"TZS": currency.TZS,
	"UAH": currency.UAH,
	"UAK": currency.UAK,
	"UGS": currency.UGS,
	"UGX": currency.UGX,
	"USD": currency.USD,
	"USN": currency.USN,
	"USS": currency.USS,
	"UYI": currency.UYI,
	"UYP": currency.UYP,
	"UYU": currency.UYU,
	"UYW": currency.UYW,
	"UZS": currency.UZS,
	"VEB": currency.VEB,
	"VEF": currency.VEF,
	"VES": currency.VES,
	"VND": currency.VND,
	"VNN": currency.VNN,
	"VUV": currency.VUV,
	"WST": currency.WST,
	"XAF": currency.XAF,
	"XAG": currency.XAG,
	"XAU": currency.XAU,
	"XBA": currency.XBA,
	"XBB": currency.XBB,
	"XBC": currency.XBC,
	"XBD": currency.XBD,
	"XCD": currency.XCD,
	"XDR": currency.XDR,
	"XEU": currency.XEU,
	"XFO": currency.XFO,
	"XFU": currency.XFU,
	"XOF": currency.XOF,
	"XPD": currency.XPD,
	"XPF": currency.XPF,
	"XPT": currency.XPT,
	"XRE": currency.XRE,
	"XSU": currency.XSU,
	"XTS": currency.XTS,
	"XUA": currency.XUA,
	"XXX": currency.XXX,
	"YDD": currency.YDD,
	"YER": currency.YER,
	"YUD": currency.YUD,
	"YUM": currency.YUM,
	"YUN": currency.YUN,
	"YUR": currency.YUR,
	"ZAL": currency.ZAL,
	"ZAR": currency.ZAR,
	"ZMK": currency.ZMK,
	"ZMW": currency.ZMW,
	"ZRN": currency.ZRN,
	"ZRZ": currency.ZRZ,
	"ZWD": currency.ZWD,
	"ZWL": currency.ZWL,
	"ZWR": currency.ZWR,
}
