// Code generated by "stringer -type=exifType"; DO NOT EDIT.

package imagemeta

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[exifTypeUnsignedByte1-1]
	_ = x[exifTypeASCIIString1-2]
	_ = x[exifTypeUnsignedShort2-3]
	_ = x[exifTypeUnsignedLong4-4]
	_ = x[exifTypeUnsignedRat8-5]
	_ = x[exifTypeSignedByte1-6]
	_ = x[exifTypeUndef1-7]
	_ = x[exifTypeSignedShort2-8]
	_ = x[exifTypeSignedLong4-9]
	_ = x[exifTypeSignedRat8-10]
	_ = x[exifTypeSignedFloat4-11]
	_ = x[exifTypeSignedDouble8-12]
}

const _exifType_name = "exifTypeUnsignedByteexifTypeUnsignedASCIIexifTypeUnsignedShortexifTypeUnsignedLongexifTypeUnsignedRatexifTypeSignedByteexifTypeUndefexifTypeSignedShortexifTypeSignedLongexifTypeSignedRatexifTypeSignedFloatexifTypeSignedDouble"

var _exifType_index = [...]uint8{0, 20, 41, 62, 82, 101, 119, 132, 151, 169, 186, 205, 225}

func (i exifType) String() string {
	i -= 1
	if i >= exifType(len(_exifType_index)-1) {
		return "exifType(" + strconv.FormatInt(int64(i+1), 10) + ")"
	}
	return _exifType_name[_exifType_index[i]:_exifType_index[i+1]]
}
