// This file is autogenerated.
package localescompressed

import (
	"math"
	"strconv"
	"time"

	"github.com/gohugoio/locales"
	"github.com/gohugoio/locales/currency"
)

var fnLocale_5a217059c67094defae82f249f55f4c5 = func(ln *localen) string {
	return ln.locale
}

var fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e = func(ln *localen) []locales.PluralRule {
	return ln.pluralsCardinal
}

var fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de = func(ln *localen) []locales.PluralRule {
	return ln.pluralsOrdinal
}

var fnPluralsRange_332465bda13eb808b3664fa0500f54ce = func(ln *localen) []locales.PluralRule {
	return ln.pluralsRange
}

var fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb = func(ln *localen, num float64, v uint64) locales.PluralRule {
	return locales.PluralRuleOther
}

var fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	return locales.PluralRuleOther
}

var fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478 = func(ln *localen, month time.Month) string {
	return ln.monthsAbbreviated[month]
}

var fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7 = func(ln *localen) []string {
	return ln.monthsAbbreviated[1:]
}

var fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7 = func(ln *localen, month time.Month) string {
	return ln.monthsNarrow[month]
}

var fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc = func(ln *localen) []string {
	return ln.monthsNarrow[1:]
}

var fnMonthWide_4c946a7e7581c665e7407df1841810b5 = func(ln *localen, month time.Month) string {
	return ln.monthsWide[month]
}

var fnMonthsWide_0ecf36a8b72255e2fee9467af362e989 = func(ln *localen) []string {
	return ln.monthsWide[1:]
}

var fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a = func(ln *localen, weekday time.Weekday) string {
	return ln.daysAbbreviated[weekday]
}

var fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be = func(ln *localen) []string {
	return ln.daysAbbreviated
}

var fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a = func(ln *localen, weekday time.Weekday) string {
	return ln.daysNarrow[weekday]
}

var fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524 = func(ln *localen) []string {
	return ln.daysNarrow
}

var fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2 = func(ln *localen, weekday time.Weekday) string {
	return ln.daysShort[weekday]
}

var fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b = func(ln *localen) []string {
	return ln.daysShort
}

var fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a = func(ln *localen, weekday time.Weekday) string {
	return ln.daysWide[weekday]
}

var fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd = func(ln *localen) []string {
	return ln.daysWide
}

var fnDecimal_a78485e7dd32c95f41af4e18420793bf = func(ln *localen) string {
	return ln.decimal
}

var fnGroup_5c4b80a5ccdee81467241784ab8fcd56 = func(ln *localen) string {
	return ln.group
}

var fnMinus_8b8229905258388dc16a5e0c270b0488 = func(ln *localen) string {
	return ln.minus
}

var fnFmtNumber_0866206624669dc4f03297eb76a545b2 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 2 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 3
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_ee2956d02335482ec6b7e9a423beb0b5 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	}

	return string(b)
}

var fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2d}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_85907dc745b1a23de90d4a44c3222404 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_987bec4676d3a5de4218e415e123649d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_506f643cda0514f66d873a9ea99bf639 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	return locales.PluralRuleUnknown
}

var fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	return locales.PluralRuleUnknown
}

var fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	return locales.PluralRuleUnknown
}

var fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 1 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 1
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_222c809496186523124eb16d6c474592 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_5e450105cccbc820e4514899ef73128f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n >= 0 && n <= 1 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba = func(ln *localen) []string {
	return nil
}

var fnFmtNumber_2816fdab36e719be095a370ff42e599f = func(ln *localen, num float64, v uint64) string {
	return strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
}

var fnFmtPercent_016773d83796aa19d80b7c33bf377883 = func(ln *localen, num float64, v uint64) string {
	return strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
}

var fnFmtCurrency_21670b336e6e64700cc7f949f05c390f = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_ec9761d27ad9ead532f5f112e23983b1 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateShort_93244c9b90b1456ed872d8fc125fd668 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if (i == 0) || (n == 1) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	}

	return locales.PluralRuleOther
}

var fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_6c366a94338df615a7797e45c90cead1 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	}

	return string(b)
}

var fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_f27514af189883ea7cd186156edfe998 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)

	if n == 0 {
		return locales.PluralRuleZero
	} else if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 {
		return locales.PluralRuleTwo
	} else if nMod100 >= 3 && nMod100 <= 10 {
		return locales.PluralRuleFew
	} else if nMod100 >= 11 && nMod100 <= 99 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleZero && end == locales.PluralRuleOne {
		return locales.PluralRuleZero
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleTwo {
		return locales.PluralRuleZero
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleTwo {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleTwo {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 5 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 9
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_f74088692afba81682b73af3d3e7c19a = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe2, 0x80, 0x8f, 0x2f}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xe2, 0x80, 0x8f, 0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe2, 0x80, 0x8f, 0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0xe2, 0x80, 0x8f, 0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_2d3b405215099e683568cb9c052cbd54 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_fb7ade3a9509df0c54be19484c433a01 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 8
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_103caa51aab787e802d47e2174ea5313 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtPercent_f074e0c8dff3d1b13a6f552a71cedb58 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 7
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnOrdinalPluralRule_2d950970290e57f43f401dd367ad94dc = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 || n == 5 || n == 7 || n == 8 || n == 9 || n == 10 {
		return locales.PluralRuleOne
	} else if n == 2 || n == 3 {
		return locales.PluralRuleTwo
	} else if n == 4 {
		return locales.PluralRuleFew
	} else if n == 6 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_9f465d9454a1702118c7b32dfb89b3d1 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_cd503abfb0c9bd938bcf20e65858f8d7 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateShort_48f6d56925de0117a6fe7b1ec232f585 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2d}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2d}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_a234d460b585eebc392634bc08485ebd = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_6c9fb1bcda2b27013e9a8a47c7326805 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_74593f8f5ae9c1a3e4637628d509c179 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_b2fdf6411542ab907630622b3d42a6b0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_3ce7dc5b6b225559fd40d102900d6c81 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if i == 1 && v == 0 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_295d01ff3669c94717ee670511c0e25e = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnOrdinalPluralRule_822e96e7fb113a3156c2e7c532855968 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod10 := i % 10
	iMod100 := i % 100
	iMod1000 := i % 1000

	if (iMod10 == 1 || iMod10 == 2 || iMod10 == 5 || iMod10 == 7 || iMod10 == 8) || (iMod100 == 20 || iMod100 == 50 || iMod100 == 70 || iMod100 == 80) {
		return locales.PluralRuleOne
	} else if (iMod10 == 3 || iMod10 == 4) || (iMod1000 == 100 || iMod1000 == 200 || iMod1000 == 300 || iMod1000 == 400 || iMod1000 == 500 || iMod1000 == 600 || iMod1000 == 700 || iMod1000 == 800 || iMod1000 == 900) {
		return locales.PluralRuleFew
	} else if (i == 0) || (iMod10 == 6) || (iMod100 == 40 || iMod100 == 60 || iMod100 == 90) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_a46571fc772262a1fe652a3412eb1140 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateFull_b94af095d9cfcd967654d0fbd68f2e98 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 3
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percentSuffix...)

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_6dd988d8f1a3fac3718e543933915046 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnCardinalPluralRule_aa3706fccbfeb0fac3cf81d326e3fe96 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if nMod10 == 1 && nMod100 != 11 {
		return locales.PluralRuleOne
	} else if nMod10 >= 2 && nMod10 <= 4 && (nMod100 < 12 || nMod100 > 14) {
		return locales.PluralRuleFew
	} else if (nMod10 == 0) || (nMod10 >= 5 && nMod10 <= 9) || (nMod100 >= 11 && nMod100 <= 14) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_62ca292db3807cd1a19e61d6d993a0d9 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if (nMod10 == 2 || nMod10 == 3) && (nMod100 != 12 && nMod100 != 13) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 5
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percentSuffix...)

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_04d2836e8ee5922750958599d99ccc13 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_a62648328d65f4275e2e521198233607 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_cbdb981782261e5a8b8cc19702400c64 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateMedium_73098a578f62c763b1a46cfb742142e3 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_b673377c923e43f185728d5b63716bee = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb3}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb3}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtTimeFull_d302f6108a9d3f4bb461f3a5111143e5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x2c, 0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	}

	return string(b)
}

var fnFmtCurrency_022949044632e54ed5ae02ba82626546 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtCurrency_88fd8698aec1fdcde88908d426b01f78 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_895833c14256127b155b25fc552f5180 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_724096485badda99a595db97f2fdcc96 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb3}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateMedium_387f50f821f8034478c23ca141a1b991 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb3}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b = func(ln *localen, num float64, v uint64) locales.PluralRule {
	return locales.PluralRuleOther
}

var fnFmtCurrency_ac4758b65e3186b6295da12d015494cf = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_1df869b66b1bebb824de5401f039b55a = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtNumber_74395470c4c29cb378d864498139a946 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 1 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_53723702774dedd2122c38a68dcef3da = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_44f743dae0936af6184f1cad1624e543 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateMedium_6eb6d3d989d88a86345ec9bd5152286a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xe0, 0xbd, 0xa3, 0xe0, 0xbd, 0xbc, 0xe0, 0xbd, 0xa0, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xba, 0xe0, 0xbd, 0xa6, 0xe0, 0xbc, 0x8b}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_7a6cbc22eb124572fde0cc182fbd1d3d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0xa4, 0xe0, 0xbe, 0xb1, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0xa3, 0xe0, 0xbd, 0xbc, 0xe0, 0xbc, 0x8b}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xe0, 0xbd, 0xa0, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xba, 0xe0, 0xbd, 0xa6, 0xe0, 0xbc, 0x8b}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_753ec22cf63aa959d868252cf9690e9d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xe0, 0xbd, 0xa0, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xba, 0xe0, 0xbd, 0xa6, 0xe0, 0xbc, 0x8b}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_3fb30797479a030a43141dadf528c83c = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)
	nMod1000000 := math.Mod(n, 1000000)
	nMod10 := math.Mod(n, 10)

	if nMod10 == 1 && (nMod100 != 11 && nMod100 != 71 && nMod100 != 91) {
		return locales.PluralRuleOne
	} else if nMod10 == 2 && (nMod100 != 12 && nMod100 != 72 && nMod100 != 92) {
		return locales.PluralRuleTwo
	} else if nMod10 >= 3 && nMod10 <= 4 && (nMod10 == 9) && (nMod100 < 10 || nMod100 > 19) || (nMod100 < 70 || nMod100 > 79) || (nMod100 < 90 || nMod100 > 99) {
		return locales.PluralRuleFew
	} else if n != 0 && nMod1000000 == 0 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729 = func(ln *localen) []string {
	return nil
}

var fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_5b5e153ac164177593d63740c82296b3 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)
	iMod10 := i % 10
	iMod100 := i % 100
	fMod10 := f % 10
	fMod100 := f % 100

	if (v == 0 && iMod10 == 1 && iMod100 != 11) || (fMod10 == 1 && fMod100 != 11) {
		return locales.PluralRuleOne
	} else if (v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14)) || (fMod10 >= 2 && fMod10 <= 4 && (fMod100 < 12 || fMod100 > 14)) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_aa123c99f95c3c4073e2b02fd117f0c8 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnCardinalPluralRule_7b6df98bb7a40cd3e9c67c7c7f464286 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)
	iMod100 := i % 100
	iMod10 := i % 10
	fMod10 := f % 10
	fMod100 := f % 100

	if (v == 0 && iMod10 == 1 && iMod100 != 11) || (fMod10 == 1 && fMod100 != 11) {
		return locales.PluralRuleOne
	} else if (v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14)) || (fMod10 >= 2 && fMod10 <= 4 && (fMod100 < 12 || fMod100 > 14)) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateMedium_61defeac982d2749785dbfe5e166419f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnCardinalPluralRule_2fd16cba9f129ab3317e08638d736342 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)
	iMod100 := i % 100
	iMod10 := i % 10
	fMod100 := f % 100
	fMod10 := f % 10

	if (v == 0 && iMod10 == 1 && iMod100 != 11) || (fMod10 == 1 && fMod100 != 11) {
		return locales.PluralRuleOne
	} else if (v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14)) || (fMod10 >= 2 && fMod10 <= 4 && (fMod100 < 12 || fMod100 > 14)) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 || n == 3 {
		return locales.PluralRuleOne
	} else if n == 2 {
		return locales.PluralRuleTwo
	} else if n == 4 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtNumber_81c5002a93b06b7f91045a20d95f77c9 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 1
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_52501d1a8eef906b405639f3c42aa47d = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 2
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_4d9675c6ee668a70f27c2127d8acae4e = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_b870df2d704ca52917b45adf78d139de = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_09f1761f0b60aead9f5012476a47c814 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	return string(b)
}

var fnCardinalPluralRule_c1a7400a6209ae7e8eb28592c58f6657 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)
	iMod10 := i % 10
	fMod10 := f % 10

	if (v == 0 && (i == 1 || i == 2 || i == 3)) || (v == 0 && (iMod10 != 4 && iMod10 != 6 && iMod10 != 9)) || (v != 0 && (fMod10 != 4 && fMod10 != 6 && fMod10 != 9)) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtNumber_96d7d1823041a48b887e3e929259dbdf = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_f5e1c0d5e27b0e309f9e016c63b96767 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 10
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percentSuffix...)

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_d61f2036c9f7cc7a15c2d51b90ffbef9 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 8 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_f7223149b88d711e36c711437b51fa28 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 8 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateLong_335b03b04c77445aeb9ba732c8cd38ef = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xdb, 0x8c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xdb, 0x8c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_935abbc817a75b933eeb5ecc41e85d90 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if i == 1 && v == 0 {
		return locales.PluralRuleOne
	} else if i >= 2 && i <= 4 && v == 0 {
		return locales.PluralRuleFew
	} else if v != 0 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_15dd2e1fe33af45e98be7d9f9b1b6b93 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_23f49589a10a043b9966b4062e75298c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateShort_cc33621dc8ea4b3f7e42320aa653cbdf = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_7d9a7daab943102fabd3a368958b52aa = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0xd0, 0xbb}...)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnCardinalPluralRule_6ce82b971f5322e99200abf4c11a0dbe = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 0 {
		return locales.PluralRuleZero
	} else if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 {
		return locales.PluralRuleTwo
	} else if n == 3 {
		return locales.PluralRuleFew
	} else if n == 6 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_01e534b5c587a0b20e565363a8f2673c = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 0 || n == 7 || n == 8 || n == 9 {
		return locales.PluralRuleZero
	} else if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 {
		return locales.PluralRuleTwo
	} else if n == 3 || n == 4 {
		return locales.PluralRuleFew
	} else if n == 5 || n == 6 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_36aaa84f53bfb88e0054a81c3abba2a4 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleZero && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnCardinalPluralRule_79907f4d6e752731d2282b8c488dfb0e = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	t := locales.T(n, v)

	if (n == 1) || (t != 0 && (i == 0 || i == 1)) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_55a922eee312039812d989655f290012 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_8ce43d140d292ed841f0f368561bdb8e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0x64, 0x65, 0x6e}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_25847504d894858ff8768557462c84d0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_391789992db94778b51fb64c1b45a71a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_7b65395a77761388c242d97c0d4bfe5b = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_204cc4159c06830dc072d7ffc4471b59 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 2 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_4924273cdfaa6d993990fdf2a2bad453 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_a6430b96a17c400006d20ebabdf05e4c = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnCardinalPluralRule_283d29635c231a3ca301081bab098305 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)
	iMod100 := i % 100
	fMod100 := f % 100

	if (v == 0 && iMod100 == 1) || (fMod100 == 1) {
		return locales.PluralRuleOne
	} else if (v == 0 && iMod100 == 2) || (fMod100 == 2) {
		return locales.PluralRuleTwo
	} else if (v == 0 && iMod100 >= 3 && iMod100 <= 4) || (fMod100 >= 3 && fMod100 <= 4) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_e2379225f63633cef88f4847394352a7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateMedium_cea69a8f4293bc29e07e966b34b3ea95 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0xa4, 0xe0, 0xbe, 0xb1, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0xa3, 0xe0, 0xbd, 0xbc, 0xe0, 0xbc, 0x8b}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xe0, 0xbd, 0x9f, 0xe0, 0xbe, 0xb3, 0xe0, 0xbc, 0x8b}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xba, 0xe0, 0xbd, 0xa6, 0xe0, 0xbc, 0x8b}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_86c295425cafc585c7a23c6ee82c54e5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0xa4, 0xe0, 0xbe, 0xb1, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0xa3, 0xe0, 0xbd, 0xbc, 0xe0, 0xbc, 0x8b}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xba, 0xe0, 0xbd, 0xa6, 0xe0, 0xbc, 0x8b, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_c1263930de470ad716d47d79ffa18067 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20, 0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0xa4, 0xe0, 0xbe, 0xb1, 0xe0, 0xbd, 0xb2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0xa3, 0xe0, 0xbd, 0xbc, 0xe0, 0xbc, 0x8b}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xba, 0xe0, 0xbd, 0xa6, 0xe0, 0xbc, 0x8b}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtTimeShort_f47f2039d4428f5cb43433a9d8952827 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0x86, 0xe0, 0xbd, 0xb4, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xbc, 0xe0, 0xbd, 0x91, 0xe0, 0xbc, 0x8b, 0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x20, 0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0x90, 0xe0, 0xbd, 0xa2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x98, 0xe0, 0xbc, 0x8b, 0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeMedium_9e0f9c064fe3f07e66fe8fd5bfc4b612 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0x86, 0xe0, 0xbd, 0xb4, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xbc, 0xe0, 0xbd, 0x91, 0xe0, 0xbc, 0x8b}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeLong_f3348bf1c6f276c7b95877da6ebb7ee0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0x86, 0xe0, 0xbd, 0xb4, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xbc, 0xe0, 0xbd, 0x91, 0xe0, 0xbc, 0x8b, 0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x20, 0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0x90, 0xe0, 0xbd, 0xa2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x98, 0xe0, 0xbc, 0x8b, 0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_a9a2519152d6f243a597cc4bd6b266cc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xe0, 0xbd, 0x86, 0xe0, 0xbd, 0xb4, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x9a, 0xe0, 0xbd, 0xbc, 0xe0, 0xbd, 0x91, 0xe0, 0xbc, 0x8b, 0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0x20, 0xe0, 0xbd, 0xa6, 0xe0, 0xbe, 0x90, 0xe0, 0xbd, 0xa2, 0xe0, 0xbc, 0x8b, 0xe0, 0xbd, 0x98, 0xe0, 0xbc, 0x8b, 0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateMedium_ab00ad983b60af2ef2a78b1bbda56e5b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x6c, 0x69, 0x61}...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_e44be20ec70caf75cec942e788bd6c7e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x6c, 0x69, 0x61}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_e5728fd357bcbdd2617427e16f6e55d2 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x6c, 0x69, 0x61}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_8eb7cfdc1e992030ede9a1abfea545be = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20, 0x67, 0x61}...)
	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_aa77d70e448fa95de06962cfa9baf918 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20, 0x67, 0x61}...)
	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_eac3fe1ec156df0f86227350e36ea607 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20, 0x67, 0x61}...)
	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_93b0a8ad8fd85d3e1c5363d7419c3c1f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20, 0x67, 0x61}...)
	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)
	nMod10 := math.Mod(n, 10)

	if nMod10 == 1 && nMod100 != 11 {
		return locales.PluralRuleOne
	} else if nMod10 == 2 && nMod100 != 12 {
		return locales.PluralRuleTwo
	} else if nMod10 == 3 && nMod100 != 13 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if nMod10 == 1 && nMod100 != 11 {
		return locales.PluralRuleOne
	} else if nMod10 == 2 && nMod100 != 12 {
		return locales.PluralRuleTwo
	} else if nMod10 == 3 && nMod100 != 13 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateFull_588a96656cc1da10a40d88dd26ead028 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2d}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtAccounting_1cec81a7ffd91500b29d60f1aa496f40 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtTimeShort_7341725c8ebd7a50d221ac00ecc6aa99 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_ceaaad7cbbdab7b4a378fe8eff296f4f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_ed4e4ed1e2c01c9ce9c9f6344a272944 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_105c4fe53621c460d37a501cec666cf0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtAccounting_343c016939c767f482f0f2a41239f465 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	}

	return string(b)
}

var fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateMedium_d9e5c72d9a18dbee96ce40275bf098d9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_07bed080e46342275478260b22396b2d = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_d9ac2a39e70db1e3f296933e6431fc50 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtNumber_af287e772ea3855f073fceb6afe2a3de = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 2

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_5fa862915ab8be0ffdb30cbcbd4dcadc = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_a127ece36a88f91b941483802dde95a0 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateShort_933fcf2387ca32382dae956e0bc893a7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateMedium_03498001dbf6972dcc828c46e13ec4d2 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2c}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtNumber_4752830681b621a14386608b6d45d3fb = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 4 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_efbfbc738b43a1167c92b076dba0ef50 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 5
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_f487e5cbe1d950a699f4afb3c0267309 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	return string(b)
}

var fnFmtAccounting_1faa91947b5020e2fec6143cbb8a4554 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	} else {
		b = append(b, ln.currencyPositiveSuffix...)
	}

	return string(b)
}

var fnFmtDateShort_f3b37aea35b3c974130a62cb3d9b1580 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x2d}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2d}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateMedium_438f654701b049881de419adc50805ae = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2d}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_29470171f6adaadee5f349206b90e4b3 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2d}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_3d08635fa599fe67dc92007b3dff4e42 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d, 0x61}...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeFull_8592fa74ce11c45fadae8071a016331d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2d, 0x61}...)
	b = append(b, []byte{0x20, 0x68, 0x6f, 0x72, 0x6f}...)
	b = append(b, []byte{0x20, 0x6b, 0x61, 0x6a}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_884dee6d35c0b3a152fe300562570c35 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeFull_e678928158d95fb12912f844eab5a4da = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x28}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateMedium_0d7c652de418db235f2b6e6504ac834b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2d}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtAccounting_faa0f0273050790b69e9c65ef1bb23a5 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtCurrency_07151969777dd3bbde27b10b4452a66e = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	return string(b)
}

var fnFmtAccounting_2a94e5a57391565d92034abc27eaffd6 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	} else {
		b = append(b, ln.currencyPositiveSuffix...)
	}

	return string(b)
}

var fnFmtCurrency_881a0ebc0b06ab2865571a388b8bfc08 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	return string(b)
}

var fnFmtAccounting_be7b974f27e584be73758ded6170b27f = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.minus[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	} else {
		b = append(b, ln.currencyPositiveSuffix...)
	}

	return string(b)
}

var fnFmtDateLong_536b20d0ef96110ea200943add1efbd4 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_dfeece500544d4aac4424f7155ba90ca = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_e4c5476e381e260e6a551a757542b603 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_7fb5040dc04d16c016c545b9369751e6 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_2aa2afdcdd2a2fb3ad7fdbe2b097c54b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateMedium_85a85042ebcb9523c624cbf546fcb94d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_3b8812d78208fcdb8a36c9b00e4a7784 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_34acd29c5979a61e92f53e9eaa5dcb93 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_a8f4207331b08e428b5c3f7d5c7887ea = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 8 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtPercent_7e5d64b2fa6162d30c9f7af69791fb63 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 7 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true

			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	for j := len(ln.percentPrefix) - 1; j >= 0; j-- {
		b = append(b, ln.percentPrefix[j])
	}

	b = append(b, ln.percent[0])

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 8 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateMedium_b224450320ab0608c36fa5f823a8ec71 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x28, 0x65}...)
	b = append(b, []byte{0x29, 0x6b, 0x6f}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x28, 0x61}...)
	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtDateLong_d0f2638a27c1f5c6700933f0a0be8242 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x28, 0x65}...)
	b = append(b, []byte{0x29, 0x6b, 0x6f}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x72, 0x65, 0x6e}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x28, 0x61}...)
	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtDateFull_f8187d99975235abc8e4a8d4e7b75463 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x28, 0x65}...)
	b = append(b, []byte{0x29, 0x6b, 0x6f}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x72, 0x65, 0x6e}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x28, 0x61}...)
	b = append(b, []byte{0x29, 0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtTimeLong_30191015e11c1c183d8d41d8e0e78b42 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x28}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x28}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtPercent_ae9cd15ae8cd02228932d7cfc0262f2b = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 10
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_1ccdf5c54785959abd27ce458030d981 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 11 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_980dd7146e58d4f3ce6aa6637bc1f3bc = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 11 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2f}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x28}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtCurrency_8a07279a530bf4495d0582798587f7c0 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 10 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_15b09dc3a083df3b5892baf7c777f025 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 10 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if i == 0 || i == 1 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtPercent_08e428b6d93eff202ec014fb83b395cf = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 7
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percentSuffix...)

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_d1d28b42f6cc00a3a8fb2b3af29b2b99 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x63, 0x63, 0x63, 0x63, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnOrdinalPluralRule_89b8067afb87364ac493104461523e40 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtTimeFull_3604e9a5c6e536d483b9ceb5e0908b94 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0x68}...)
	b = append(b, []byte{0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0x6d, 0x69, 0x6e}...)
	b = append(b, []byte{0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x73}...)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtTimeShort_77616a5d5285f57c035b3599c557be3f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0x68}...)
	b = append(b, []byte{0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_5ee755e4617064423cc20f8812f3f20e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0x68}...)
	b = append(b, []byte{0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0x6d, 0x69, 0x6e}...)
	b = append(b, []byte{0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x73}...)

	return string(b)
}

var fnFmtTimeLong_890d4627ee8be38589c6184a42b5e739 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0x68}...)
	b = append(b, []byte{0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0x6d, 0x69, 0x6e}...)
	b = append(b, []byte{0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x73}...)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_08768b0816a0563c1bdec56e3a7c7227 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0x68}...)
	b = append(b, []byte{0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0x6d, 0x69, 0x6e}...)
	b = append(b, []byte{0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x73}...)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtTimeFull_7190bcf20785aafef3dd5520b994587a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x68}...)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_35ef415790b91229102369fdc2485b80 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x69}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x61, 0x6c}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_0987fcfde5397e392daea76f23ab1c65 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x69}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x61, 0x6c}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_5424a4808e45b44daf4dd95a868534c9 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 {
		return locales.PluralRuleTwo
	} else if n >= 3 && n <= 6 {
		return locales.PluralRuleFew
	} else if n >= 7 && n <= 10 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_ba2eea4558b51a6ab55725541be07d3f = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_ad2ba7087d8f6e468ec280be01359ba1 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_20afe596297e5df40957d72da7798243 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	}

	return string(b)
}

var fnCardinalPluralRule_5232ad4c13ce5aa918e9c5bcfff748c4 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 || n == 11 {
		return locales.PluralRuleOne
	} else if n == 2 || n == 12 {
		return locales.PluralRuleTwo
	} else if (n >= 3 && n <= 10) || (n >= 13 && n <= 19) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_f828d7a9eaca874285f7573c8823d785 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 || n == 11 {
		return locales.PluralRuleOne
	} else if n == 2 || n == 12 {
		return locales.PluralRuleTwo
	} else if n == 3 || n == 13 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateLong_1b477f18b273c0ada89ad664b7a114af = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x6d, 0x68}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_127dcd8df9ddb869d241194c5acabeea = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x6d, 0x68}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 4 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 3*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnOrdinalPluralRule_a9f120c4a8c40112dec183cd65a8b424 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 || n == 3 {
		return locales.PluralRuleTwo
	} else if n == 4 {
		return locales.PluralRuleFew
	} else if n == 6 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtTimeShort_11febbd269414b302ab4045dd63f47b7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	if h < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeMedium_9efd56b286449b21c42b6b2ea8c1f34c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	if h < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeLong_28302cc672efd8ad015891c59efdc448 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	if h < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_abd60cef530beeba744de08bba2ad540 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	if h < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnCardinalPluralRule_e63270f01497e6cdf2ee8385ca8e8db3 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod10 := i % 10
	iMod100 := i % 100

	if v == 0 && iMod10 == 1 {
		return locales.PluralRuleOne
	} else if v == 0 && iMod10 == 2 {
		return locales.PluralRuleTwo
	} else if v == 0 && (iMod100 == 0 || iMod100 == 20 || iMod100 == 40 || iMod100 == 60 || iMod100 == 80) {
		return locales.PluralRuleFew
	} else if v != 0 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_d8430d5f441fe322a704e04df73f1229 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	return string(b)
}

var fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	} else {
		b = append(b, ln.currencyPositiveSuffix...)
	}

	return string(b)
}

var fnFmtDateFull_4f1a5e5b794a77b5b18293f768464f0a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_cf4ba4e3714ec1726c70da66a7b66698 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	nMod10 := math.Mod(n, 10)

	if i == 1 && v == 0 {
		return locales.PluralRuleOne
	} else if i == 2 && v == 0 {
		return locales.PluralRuleTwo
	} else if v == 0 && (n < 0 || n > 10) && nMod10 == 0 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_11b8a4f474d210e1d71b265f1f8e62d0 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleTwo {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleMany {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleMany && end == locales.PluralRuleOther {
		return locales.PluralRuleMany
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleTwo {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleMany {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtNumber_8a34c21e88caccc3a8728bf464f33cbd = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 5 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_d2f2a59417e049449b3bb6986a239e7b = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 6
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_dcd95366ecd0baffaf3fa998dfd8ecbc = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_ac8c216759a27ae9294265517cfd4549 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateMedium_2a623cab0baa2e039b73232b6b2d515a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0xd7, 0x91}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_e3113833817389f6b233e285c24e687d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0xd7, 0x91}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_61caddbcb5db4de051fdc4ba779a33da = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0xd7, 0x91}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateShort_682acc4ae60b52f9983d14c85913c784 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateShort_474bf8d71d0392f03d943b2085e8e536 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtTimeShort_5d630ed3a6aab7e460d3797e87a803d8 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0x68, 0x6f, 0x64, 0xc5, 0xba}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnOrdinalPluralRule_ddb6ad1d52dc3051d9aca40302fc33cb = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 || n == 5 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_6955160c4014108c44f154550c94bf4a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateMedium_7234e1a0f15c9c77f42b9ce48acda9a9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateLong_d54b619c689771a59ca2c2bc98c286ad = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_96142c8f0234c0bee88b2aec42a20ee0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtDateMedium_227513acca176d2921237a4d2e01beae = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd5, 0xa9, 0x2e}...)

	return string(b)
}

var fnFmtDateLong_569998488d00a7529115ca67accb195e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd5, 0xa9, 0x2e}...)

	return string(b)
}

var fnFmtDateFull_f88f4a85fdaf35c0c7b8106f2f509933 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd5, 0xa9, 0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2d}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_cef970679517ca70c9c659b02d9823d9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_5a51da803052c6679cb1ae79edfb44b5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0x6c, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtPercent_867dc730c8139deaf3b7a764768ab63b = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 13
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percentSuffix...)

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_190e2886d7df91874dc7265208b76611 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_1284b8838aa00e9c05dfab59723d425b = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnCardinalPluralRule_aee42f293f8724ca3ca1ad3bf1a57537 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	t := locales.T(n, v)
	iMod100 := i % 100
	iMod10 := i % 10

	if (t == 0 && iMod10 == 1 && iMod100 != 11) || (t != 0) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnCardinalPluralRule_e217c4f51dda8c076116bf492a837691 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	t := locales.T(n, v)
	iMod10 := i % 10
	iMod100 := i % 100

	if (t == 0 && iMod10 == 1 && iMod100 != 11) || (t != 0) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_b8b63d0d886b70a436900cf618138a9f = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 11 || n == 8 || n == 80 || n == 800 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_30a6b80fc168a8b0808040051491fc26 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe5, 0xb9, 0xb4}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xe6, 0x9c, 0x88}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe6, 0x97, 0xa5}...)

	return string(b)
}

var fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe5, 0xb9, 0xb4}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xe6, 0x9c, 0x88}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe6, 0x97, 0xa5}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtTimeFull_5f445481e576332f9f6276f28f4cc61a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0xe6, 0x99, 0x82}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0xe5, 0x88, 0x86}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0xe7, 0xa7, 0x92, 0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnOrdinalPluralRule_d9d734a5db809617a5116023efed1940 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod100 := i % 100

	if i == 1 {
		return locales.PluralRuleOne
	} else if (i == 0) || (iMod100 >= 2 && iMod100 <= 20 && (iMod100 == 40 || iMod100 == 60 || iMod100 == 80)) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_e975bcc2318a5d1ed3b85b6356649a29 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOther
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_8f50deadc7a225aae72abf7135535a4e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_4edbf4e11f17243cbd3f53cf42390fdf = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_adf14415bc63ac7fa5223315bd231add = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_a352e45549d51689f730b62d4fd819bf = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtAccounting_368d18aec5667a40e744175ddb192235 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateLong_56933fd536cdccf5e7a0f57849436ee6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x69}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x69}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_6df4b160cd1730b200782b7d4e4cd069 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64, 0x69}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0x64, 0x69}...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_f27eb78b129904a4d98d49f647da3164 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_dd5c984fc6201988feed2698382f9150 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 0 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnOrdinalPluralRule_b14713cd1b1f920531a88cd35823e9c6 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)

	if (nMod10 == 6) || (nMod10 == 9) || (nMod10 == 0 && n != 0) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_615c694e243fa2fa6c571f5fc2c9d718 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb6}...)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)

	return string(b)
}

var fnFmtDateLong_c224d2682917532885a5e8179ac7c5c2 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb6}...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnFmtDateFull_4d0973d19d04f6373346e3410c0d7cff = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb6}...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtCurrency_1c327a5182399c63873a891da808a068 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	return string(b)
}

var fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	} else {
		b = append(b, ln.currencyPositiveSuffix...)
	}

	return string(b)
}

var fnFmtDateShort_d82f27051c9ad611e3ef0bbd19df221b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_311268549b3ecbf935e9311c16ee9a54 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_2ebcd49fb93d5f5e0574b122053067c2 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtCurrency_ef0590f3c464dc40ad41a959adae57e7 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_0722fc77da098723d3a7bcc9b8145055 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 4 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateShort_d881b9f1f4fd8f22e57a48f28a75e30a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateMedium_f69e9157403c463ade0395f3ff9240fe = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateLong_17c2343e73783fd11c1ad64b6b231979 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xeb, 0x85, 0x84, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xec, 0x9b, 0x94, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xec, 0x9d, 0xbc}...)

	return string(b)
}

var fnFmtDateFull_c611a5e966ec03dbdacdf654df3799f0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xeb, 0x85, 0x84, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xec, 0x9b, 0x94, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xec, 0x9d, 0xbc, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_a4e3f96b4a6b77d07b6c318bbc1c74ac = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0xec, 0x8b, 0x9c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0xeb, 0xb6, 0x84, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0xec, 0xb4, 0x88, 0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_bfbb513b6692ade43eb8d3cf5179818c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, []byte{0xec, 0x8b, 0x9c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0xeb, 0xb6, 0x84, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0xec, 0xb4, 0x88, 0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateShort_20b00d66ebb3cefd65ec76fba139eaf9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2d}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtNumber_b796bfb92dd260d7d0890a75e953e03c = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 0
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_aa714d39dff020dd613d3cb5622ff940 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_b96d5354a212ba0fb9e6fdc6738a84f8 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_d34cde82a90714ae7e620d04971d2faf = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnCardinalPluralRule_fed1c76cd90b8905708121045ba1b87c = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 0 {
		return locales.PluralRuleZero
	} else if n == 1 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_7666e72292ce611269924033db7087d6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_e96355a5ea41d1e8a56ddce8cbb58393 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_694de915ae18d4ba6d1ade6e9c08daac = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20, 0x64, 0xc3, 0xa4}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtPercent_1842fa1c80e067ee20bb9d4acb93d0eb = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 3 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true

			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	b = append(b, ln.percent[0])

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnCardinalPluralRule_d9471d7d7c171fa2f8441e898fb73597 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)
	nMod1000 := math.Mod(n, 1000)
	nMod100000 := math.Mod(n, 100000)
	nMod1000000 := math.Mod(n, 1000000)

	if n == 0 {
		return locales.PluralRuleZero
	} else if n == 1 {
		return locales.PluralRuleOne
	} else if (nMod100 == 2 || nMod100 == 22 || nMod100 == 42 || nMod100 == 62 || nMod100 == 82) || (nMod1000 == 0 && nMod100000 >= 1000 && nMod100000 <= 20000 && (nMod100000 == 40000 || nMod100000 == 60000 || nMod100000 == 80000)) || (n != 0 && nMod1000000 == 100000) {
		return locales.PluralRuleTwo
	} else if nMod100 == 3 || nMod100 == 23 || nMod100 == 43 || nMod100 == 63 || nMod100 == 83 {
		return locales.PluralRuleFew
	} else if n != 1 && (nMod100 == 1 || nMod100 == 21 || nMod100 == 41 || nMod100 == 61 || nMod100 == 81) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_80a3a34cd3bb2c41e809f9ce9e28ce2a = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)

	if (n >= 1 && n <= 4) || ((nMod100 >= 1 && nMod100 <= 4) || (nMod100 >= 21 && nMod100 <= 24) || (nMod100 >= 41 && nMod100 <= 44) || (nMod100 >= 61 && nMod100 <= 64) || (nMod100 >= 81 && nMod100 <= 84)) {
		return locales.PluralRuleOne
	} else if (n == 5) || (nMod100 == 5) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_3f222d37600871207c139eeba9ae1a98 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d, 0xd0, 0xb6}...)
	b = append(b, []byte{0x2e, 0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)

	return string(b)
}

var fnFmtDateLong_07de017bbe01706764e3fbed80282c3a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d, 0xd0, 0xb6}...)
	b = append(b, []byte{0x2e, 0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnFmtDateFull_2b8291eec4b5351a149455e035ae0af2 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d, 0xd0, 0xb6}...)
	b = append(b, []byte{0x2e, 0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_7664ec40507fc37b05985577fc0c7d99 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if n == 0 {
		return locales.PluralRuleZero
	} else if (i == 0 || i == 1) && n != 0 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_9b06328888547fc535e862da8ee049d8 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.currencyPositiveSuffix...)

	return string(b)
}

var fnFmtAccounting_d0ef2c4a76852adc59df517a99a6000f = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	} else {
		b = append(b, ln.currencyPositiveSuffix...)
	}

	return string(b)
}

var fnFmtAccounting_0174068e708e390eda0b53f8a0ef057e = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateFull_afb42e88482f5f1e5a494bb738650f23 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0xe0, 0xba, 0x97, 0xe0, 0xba, 0xb5, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() < 0 {
		b = append(b, ln.erasWide[0]...)
	} else {
		b = append(b, ln.erasWide[1]...)
	}

	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeLong_2be9d9219c97d619bb809b28233eb70f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xbb, 0x82, 0xe0, 0xba, 0xa1, 0xe0, 0xba, 0x87, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xba, 0x99, 0xe0, 0xba, 0xb2, 0xe0, 0xba, 0x97, 0xe0, 0xba, 0xb5, 0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xba, 0xa7, 0xe0, 0xba, 0xb4, 0xe0, 0xba, 0x99, 0xe0, 0xba, 0xb2, 0xe0, 0xba, 0x97, 0xe0, 0xba, 0xb5, 0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_dafe47a99f691cddd99c295dd3a80472 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xbb, 0x82, 0xe0, 0xba, 0xa1, 0xe0, 0xba, 0x87, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xba, 0x99, 0xe0, 0xba, 0xb2, 0xe0, 0xba, 0x97, 0xe0, 0xba, 0xb5, 0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xba, 0xa7, 0xe0, 0xba, 0xb4, 0xe0, 0xba, 0x99, 0xe0, 0xba, 0xb2, 0xe0, 0xba, 0x97, 0xe0, 0xba, 0xb5, 0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtNumber_529d4d803e923e8eb15440c678f6c7b8 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 0
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtCurrency_bc1dd05937d2f9e1e78f0e1e26de2ce0 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_c1338c4c0c9820a635090f5e89a65c4b = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 2
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnCardinalPluralRule_1a8ed4aac1cf13bc0d4f980cb943bb5f = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	f := locales.F(n, v)
	nMod100 := math.Mod(n, 100)
	nMod10 := math.Mod(n, 10)

	if nMod10 == 1 && (nMod100 < 11 || nMod100 > 19) {
		return locales.PluralRuleOne
	} else if nMod10 >= 2 && nMod10 <= 9 && (nMod100 < 11 || nMod100 > 19) {
		return locales.PluralRuleFew
	} else if f != 0 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_a3f984963a38c1972f5ae182cb000f06 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2d}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2d}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_6af034e83fbef5852ab70b5e01d66101 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0x6d}...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_a6a7cb3b48a46ecca7801bbef267e737 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0x6d}...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x64}...)
	b = append(b, []byte{0x2e, 0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_c9ff0d41f0522116859535aa9029ca16 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	f := locales.F(n, v)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if nMod10 == 1 && (nMod100 < 11 || nMod100 > 19) {
		return locales.PluralRuleOne
	} else if nMod10 >= 2 && nMod10 <= 9 && (nMod100 < 11 || nMod100 > 19) {
		return locales.PluralRuleFew
	} else if f != 0 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 1 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, symbol...)
	} else {
		b = append(b, symbol...)
	}

	return string(b)
}

var fnCardinalPluralRule_35d374490f3e32fac8828ad5843b9d6e = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	f := locales.F(n, v)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)
	fMod100 := f % 100
	fMod10 := f % 10

	if (nMod10 == 0) || (nMod100 >= 11 && nMod100 <= 19) || (v == 2 && fMod100 >= 11 && fMod100 <= 19) {
		return locales.PluralRuleZero
	} else if (nMod10 == 1 && nMod100 != 11) || (v == 2 && fMod10 == 1 && fMod100 != 11) || (v != 2 && fMod10 == 1) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_52f7b7c1971a0689c24fe53cd3a6bf1d = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleZero && end == locales.PluralRuleZero {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleZero && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleZero {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleZero {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_cefeb908e03bbb483b089e5f8a7e3bb5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20, 0x67, 0x61, 0x64, 0x61}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)

	return string(b)
}

var fnFmtDateLong_b7e0e3e1d7c2a12e272d025e98a6facc = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20, 0x67, 0x61, 0x64, 0x61}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnFmtDateFull_c68531efc0310a76517a9616064247e7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2e, 0x20, 0x67, 0x61, 0x64, 0x61}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnCardinalPluralRule_073a4fc904d56c3f55fe9007d4416cce = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	f := locales.F(n, v)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)
	fMod10 := f % 10
	fMod100 := f % 100

	if (nMod10 == 0) || (nMod100 >= 11 && nMod100 <= 19) || (v == 2 && fMod100 >= 11 && fMod100 <= 19) {
		return locales.PluralRuleZero
	} else if (nMod10 == 1 && nMod100 != 11) || (v == 2 && fMod10 == 1 && fMod100 != 11) || (v != 2 && fMod10 == 1) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnCardinalPluralRule_b96558b039ca4fc63138671b418a67f5 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)
	iMod10 := i % 10
	iMod100 := i % 100
	fMod10 := f % 10
	fMod100 := f % 100

	if (v == 0 && iMod10 == 1 && iMod100 != 11) || (fMod10 == 1 && fMod100 != 11) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_885e23e22aae7e14ea3420fe83f75ccf = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod10 := i % 10
	iMod100 := i % 100

	if iMod10 == 1 && iMod100 != 11 {
		return locales.PluralRuleOne
	} else if iMod10 == 2 && iMod100 != 12 {
		return locales.PluralRuleTwo
	} else if (iMod10 == 7 || iMod10 == 8) && (iMod100 != 17 && iMod100 != 18) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_fdbcfd5a321e6370f61bce682fe65b6e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_ab2127ffd2f4043ae38d2f998ae282a6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_74d5e00c4bf6ea6ef8db9baad7610c73 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtDateMedium_34dcbbdacb45f505155cca4c18ec48c0 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xbe, 0xd0, 0xbd, 0xd1, 0x8b}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0xd1, 0x8b, 0xd0, 0xbd}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_7545f2e6128dde4c84d462009151629b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xbe, 0xd0, 0xbd, 0xd1, 0x8b}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xd1, 0x8b, 0xd0, 0xbd}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_4052efc783402567c61ad2003f35b9bf = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xbe, 0xd0, 0xbd, 0xd1, 0x8b}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xd1, 0x8b, 0xd0, 0xbd}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0xd0, 0xb3, 0xd0, 0xb0, 0xd1, 0x80, 0xd0, 0xb0, 0xd0, 0xb3}...)

	return string(b)
}

var fnOrdinalPluralRule_a2a4fea44e28d3286f808ebd85af3bc6 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 || n == 3 {
		return locales.PluralRuleTwo
	} else if n == 4 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateFull_3c8ce40063641c68bcd08d693c6fa8e8 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_e05d2968783cc0de88c2d1b2682d3007 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)

	if n == 1 {
		return locales.PluralRuleOne
	} else if (n == 0) || (nMod100 >= 2 && nMod100 <= 10) {
		return locales.PluralRuleFew
	} else if nMod100 >= 11 && nMod100 <= 19 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateLong_e370bf773fe0548c5c154f2e1cb30b9f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x74, 0x61}...)
	b = append(b, []byte{0xe2, 0x80, 0x99, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_03998ac4c6437ee7c96bd6694bf576fb = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x74, 0x61}...)
	b = append(b, []byte{0xe2, 0x80, 0x99, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 3 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		b = append(b, ln.currencyNegativePrefix[0])

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
	}

	return string(b)
}

var fnFmtDateMedium_4d2c8f906ead2ec2a8ed4de52aa273cb = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe1, 0x81, 0x8a, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateLong_26f464d351da53af35ea071506237a9d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe1, 0x81, 0x8a, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnFmtDateFull_b91fecf4568a8483e671bad76f4d4425 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe1, 0x81, 0x8a, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe1, 0x81, 0x8a, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtTimeShort_3636d5dfdda3e0b6ca17aef52abf5eb3 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x42, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_5d9a472a4327911b01200af597bf5908 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x42, 0x20}...)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_c108ef0052564f5ac9b30fa59af63081 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x20}...)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeFull_9d22f7542356e9fc28813c7e9ca3671d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x20}...)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtCurrency_8bfebaacbf71a9a4bab2a3bb99e52529 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	return string(append(append([]byte{}, symbol...), s...))
}

var fnFmtAccounting_411c735c7aafad814d0db05f621af17c = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	return string(append(append([]byte{}, symbol...), s...))
}

var fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 {
		return locales.PluralRuleOne
	} else if n == 2 {
		return locales.PluralRuleTwo
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_52160a421770c8d84e356e1637ce196a = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_6501a20bdcbb382f1e5e59af6d704f28 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 6 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateFull_91904f1c4adbb925999bdba7f5796de8 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20, 0x64, 0x65}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_e50f14b6d2405eade89d9149a1efaee9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x4b, 0x6c}...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_d09a7c8ab2ed333a37b86babd4139677 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x4b, 0x6c, 0x6f, 0x63, 0x6b}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_1dc6a22a036902eb11698fc0a589743d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x4b, 0x6c, 0x6f, 0x63, 0x6b}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x28}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnFmtTimeFull_e61d6b84b00325a0cc262f834ec42619 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x4b, 0x6c, 0x6f, 0x63, 0x6b}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x28}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x29}...)

	return string(b)
}

var fnOrdinalPluralRule_7e0820ce2c902179e061cbae21c795d7 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n >= 1 && n <= 4 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_c2e9326f17b4ceaeb5497266d0765dd7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2f}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2f}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeFull_702261155c59dd873e83848c0fe82a23 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x6b, 0x6c}...)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Hour() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateLong_53ac6ba1bddae8f8da33e7321fa48492 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x6c, 0x79, 0xc9, 0x9b}...)
	b = append(b, []byte{0xcc, 0x8c, 0xca, 0xbc, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x6e, 0x61}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_1d5a2ee3108dbf80d60a1f3c5c0aaf8d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0x2c, 0x20, 0x6c, 0x79, 0xc9, 0x9b}...)
	b = append(b, []byte{0xcc, 0x8c, 0xca, 0xbc, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0x6e, 0x61}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeLong_5643f7e25a7c963d538012968ff836de = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnFmtTimeFull_fb1e9e6f7242212344122d3c0e63f9d6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	return string(b)
}

var fnOrdinalPluralRule_f7f41a6ef5c99a912fc7e7451c5c3029 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if n == 1 || n == 5 || n >= 7 && n <= 9 {
		return locales.PluralRuleOne
	} else if n == 2 || n == 3 {
		return locales.PluralRuleTwo
	} else if n == 4 {
		return locales.PluralRuleFew
	} else if n == 6 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_5a5202b511779ecb07983ce846c85742 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb0, 0xd0, 0xb7}...)

	return string(b)
}

var fnFmtDateLong_4874c47573ec41408036722eb0d2c6a6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb0, 0xd0, 0xb7}...)

	return string(b)
}

var fnFmtDateFull_a7dc1bab86baf3ec90633b4422fb9cba = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb0, 0xd0, 0xb7}...)

	return string(b)
}

var fnCardinalPluralRule_4e031e59a33d2df374391ed7199b0c7f = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod100 := i % 100
	iMod10 := i % 10

	if i == 1 && v == 0 {
		return locales.PluralRuleOne
	} else if v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14) {
		return locales.PluralRuleFew
	} else if (v == 0 && i != 1 && iMod10 >= 0 && iMod10 <= 1) || (v == 0 && iMod10 >= 5 && iMod10 <= 9) || (v == 0 && iMod100 >= 12 && iMod100 <= 14) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnCardinalPluralRule_3c2302f5244a65cb0b21db084f77f61b = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod10 := i % 10
	iMod100 := i % 100

	if i == 1 && v == 0 {
		return locales.PluralRuleOne
	} else if v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14) {
		return locales.PluralRuleFew
	} else if (v == 0 && i != 1 && iMod10 >= 0 && iMod10 <= 1) || (v == 0 && iMod10 >= 5 && iMod10 <= 9) || (v == 0 && iMod100 >= 12 && iMod100 <= 14) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_657e1e9a61de200098381fee1994c99c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x20, 0x73, 0x74}...)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_bce1be6f532fe054bea1dce83a0ecd58 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0x6d, 0x65, 0x74, 0x74, 0x61, 0x73}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnFmtDateFull_8f18dd0a7a346f349e234c4eb438c7ba = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0x6d, 0x65, 0x74, 0x74, 0x61, 0x73}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)

	return string(b)
}

var fnFmtDateLong_03b64668e0972b02e0a12fe5cd4b5829 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0xd8, 0xaf, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd8, 0xaf, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_7ea895699a09e2a46b34e71ca5ff6068 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0xd8, 0xaf, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd8, 0xaf, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if i >= 0 && i <= 1 {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateFull_4c2ead12910ff721f96d3aca986ca749 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20, 0x69, 0x6c, 0x73}...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnCardinalPluralRule_a60b3f7c6ddcf0678df3df0fe274201a = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	nMod100 := math.Mod(n, 100)

	if i == 1 && v == 0 {
		return locales.PluralRuleOne
	} else if (v != 0) || (n == 0) || (nMod100 >= 2 && nMod100 <= 19) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_6db355e32f98d7cf05b06cd4b38d6c63 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOne {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod10 := i % 10
	iMod100 := i % 100

	if v == 0 && iMod10 == 1 && iMod100 != 11 {
		return locales.PluralRuleOne
	} else if v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14) {
		return locales.PluralRuleFew
	} else if (v == 0 && iMod10 == 0) || (v == 0 && iMod10 >= 5 && iMod10 <= 9) || (v == 0 && iMod100 >= 11 && iMod100 <= 14) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb3}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_57a372adfbe798d29526efd47816e07a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd1, 0x81, 0xd1, 0x8b, 0xd0, 0xbb}...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0xd0, 0xba, 0xd2, 0xaf, 0xd0, 0xbd, 0xd1, 0x8d}...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_c6f27f6f1d6b31d66116bd13f68bb112 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)

	if (i == 0) || (n == 1) {
		return locales.PluralRuleOne
	} else if n >= 2 && n <= 10 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnCardinalPluralRule_780c1153d0b26cb9e6da19f93ef34fac = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	f := locales.F(n, v)

	if (n == 0 || n == 1) || (i == 0 && f == 1) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_088ba095b00c7f7ec73fa5b6bd628417 = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleOne
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleOther
	}

	return locales.PluralRuleOther
}

var fnCardinalPluralRule_0fef1d0e24fa8383f72b5cbddf29c5f9 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod100 := i % 100

	if v == 0 && iMod100 == 1 {
		return locales.PluralRuleOne
	} else if v == 0 && iMod100 == 2 {
		return locales.PluralRuleTwo
	} else if (v == 0 && iMod100 >= 3 && iMod100 <= 4) || (v != 0) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnRangePluralRule_c14bd9bed5ef22ead1e026671990212f = func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule {
	start := ln.CardinalPluralRule(num1, v1)
	end := ln.CardinalPluralRule(num2, v2)

	if start == locales.PluralRuleOne && end == locales.PluralRuleOne {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOne && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleOne {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleTwo && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOne {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleFew && end == locales.PluralRuleOther {
		return locales.PluralRuleOther
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleOne {
		return locales.PluralRuleFew
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleTwo {
		return locales.PluralRuleTwo
	} else if start == locales.PluralRuleOther && end == locales.PluralRuleFew {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateShort_6782bf6cd88b90ec3c33a5cf6ab98fd7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 9 {
		b = append(b, strconv.Itoa(t.Year())[2:]...)
	} else {
		b = append(b, strconv.Itoa(t.Year())[1:]...)
	}

	return string(b)
}

var fnFmtDateLong_7394f5b140d96d3ada830127b6b00a5f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_8c8c1dc47cf6203ce4b51ec99ff21181 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_56894f892061c70a16fdf4ad41cd67e9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_bc6f92d2d84d639433127ff63687d178 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_789e11f07c8b1dffd2bd03da43d0ea66 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, []byte{0x63, 0x63, 0x63, 0x63, 0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_3c1131a653a5e918c3e9381647860bcf = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnOrdinalPluralRule_8b495a08240ebf51a9ad31ffd7bf6c66 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if n == 1 {
		return locales.PluralRuleOne
	} else if nMod10 == 4 && nMod100 != 14 {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnFmtTimeLong_880840264062db9197a5f9820877cc61 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x2c, 0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_922dbe895fd3d466f77caf876ca0d4af = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x2c, 0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnOrdinalPluralRule_1456aec41e57e34cc80a524a2035cc1f = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod100 := math.Mod(n, 100)
	nMod10 := math.Mod(n, 10)

	if (nMod10 == 1 || nMod10 == 2) && (nMod100 != 11 && nMod100 != 12) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_89ff89febaf8ab083f938260d29e96d0 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if (nMod10 == 1 || nMod10 == 2) && (nMod100 != 11 && nMod100 != 12) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtCurrency_2a7c939a96723e34c09bd13cdb6d28d7 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtAccounting_0a9258144dfcd2f61bc1eaaba53bee40 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		b = append(b, ln.minus[0])

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtTimeLong_473cd7451fd2b0582dbc76fa50f0465a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_6fab0d5c33d51b8c50c1ff9de8d0ca73 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateFull_c892ba68b5e8cfe388377f706293ec57 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtDateLong_4ba622cfb6604f92efa67396648b01e1 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() < 0 {
		b = append(b, ln.erasAbbreviated[0]...)
	} else {
		b = append(b, ln.erasAbbreviated[1]...)
	}

	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_2a2625729a79b75f22677ae0545a3802 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0xe0, 0xb8, 0x97, 0xe0, 0xb8, 0xb5, 0xe0, 0xb9, 0x88, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() < 0 {
		b = append(b, ln.erasWide[0]...)
	} else {
		b = append(b, ln.erasWide[1]...)
	}

	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeLong_9d4d0f53074023f50ec443e7d6cc145e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xb8, 0x99, 0xe0, 0xb8, 0xb2, 0xe0, 0xb8, 0xac, 0xe0, 0xb8, 0xb4, 0xe0, 0xb8, 0x81, 0xe0, 0xb8, 0xb2, 0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xb8, 0x99, 0xe0, 0xb8, 0xb2, 0xe0, 0xb8, 0x97, 0xe0, 0xb8, 0xb5, 0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xb8, 0xa7, 0xe0, 0xb8, 0xb4, 0xe0, 0xb8, 0x99, 0xe0, 0xb8, 0xb2, 0xe0, 0xb8, 0x97, 0xe0, 0xb8, 0xb5, 0x20}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	return string(b)
}

var fnFmtTimeFull_54a1d4bd08c970be76504a2aa980af91 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xb8, 0x99, 0xe0, 0xb8, 0xb2, 0xe0, 0xb8, 0xac, 0xe0, 0xb8, 0xb4, 0xe0, 0xb8, 0x81, 0xe0, 0xb8, 0xb2, 0x20}...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xb8, 0x99, 0xe0, 0xb8, 0xb2, 0xe0, 0xb8, 0x97, 0xe0, 0xb8, 0xb5, 0x20}...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0xe0, 0xb8, 0xa7, 0xe0, 0xb8, 0xb4, 0xe0, 0xb8, 0x99, 0xe0, 0xb8, 0xb2, 0xe0, 0xb8, 0x97, 0xe0, 0xb8, 0xb5, 0x20}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	return string(b)
}

var fnFmtDateFull_b145a441f04d612c2b6631b673b83d20 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0xe1, 0x8d, 0xa3, 0x20}...)

	if t.Day() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20, 0xe1, 0x88, 0x98, 0xe1, 0x8b, 0x93, 0xe1, 0x88, 0x8d, 0xe1, 0x89, 0xb2, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)

	if t.Year() < 0 {
		b = append(b, ln.erasWide[0]...)
	} else {
		b = append(b, ln.erasWide[1]...)
	}

	return string(b)
}

var fnOrdinalPluralRule_96a5502b4d80e3468da0c2115685c65f = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)

	if (nMod10 == 6 || nMod10 == 9) || (n == 10) {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateFull_8c251b28aa9de80970be20b481b38b9b = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtPercent_5bdc6fa582de100c13fd696db4607f17 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 12
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percentSuffix...)

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtDateShort_c0764a5a895e1bd677cbe99232a5e159 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2e}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x2e}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_d6c733b45a5fc6c4b67eabb17c635dfe = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb5, 0xd0, 0xbb}...)

	return string(b)
}

var fnFmtDateLong_6384cc06509d395705e5eb2161388671 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb5, 0xd0, 0xbb}...)

	return string(b)
}

var fnFmtDateFull_a5f5cc8ae427c1221f50697b6908f7a5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd0, 0xb5, 0xd0, 0xbb}...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_92e733a5c3e2c8f5728268e0f4b7afd0 = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)

	if (n >= 0 && n <= 1) || (n >= 11 && n <= 99) {
		return locales.PluralRuleOne
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_bdd5fb2012d879143590e275c7d8ebb5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_c5ff4169cfa62bfc0e853fa211ecf219 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_4e14db85a2854023613be0580e048527 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnCardinalPluralRule_a9860e48306c4532a134648f71c7ec2b = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	i := int64(n)
	iMod100 := i % 100
	iMod10 := i % 10

	if v == 0 && iMod10 == 1 && iMod100 != 11 {
		return locales.PluralRuleOne
	} else if v == 0 && iMod10 >= 2 && iMod10 <= 4 && (iMod100 < 12 || iMod100 > 14) {
		return locales.PluralRuleFew
	} else if (v == 0 && iMod10 == 0) || (v == 0 && iMod10 >= 5 && iMod10 <= 9) || (v == 0 && iMod100 >= 11 && iMod100 <= 14) {
		return locales.PluralRuleMany
	}

	return locales.PluralRuleOther
}

var fnOrdinalPluralRule_d372f132a14400bf15cec11e0608103a = func(ln *localen, num float64, v uint64) locales.PluralRule {
	n := math.Abs(num)
	nMod10 := math.Mod(n, 10)
	nMod100 := math.Mod(n, 100)

	if nMod10 == 3 && nMod100 != 13 {
		return locales.PluralRuleFew
	}

	return locales.PluralRuleOther
}

var fnFmtDateMedium_eab28cd2ee8d9db4a9df96d1f7b01551 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd1, 0x80}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateLong_a29e6b3227718d09eae068aaf8d4873a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd1, 0x80}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtDateFull_0b6f60c7e7c9a2b75087215c987e94eb = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd1, 0x80}...)
	b = append(b, []byte{0x2e}...)

	return string(b)
}

var fnFmtCurrency_a6249ceef016228ba91ed7bf933a5843 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_ae41296c273457c904bdda366fe75145 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				b = append(b, ln.group[0])
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {
		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateMedium_82c71fd71a19831d2abc525aff9a3a0e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_39d6f2c09ffa1cac410fe4d18df66a7c = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_4d9460f3dccf0c21118fbf081ee2fe75 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0xd8, 0x8c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtCurrency_c079b862a25ea3e09b2b82e454b35d84 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	for j := len(symbol) - 1; j >= 0; j-- {
		b = append(b, symbol[j])
	}

	for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
		b = append(b, ln.currencyPositivePrefix[j])
	}

	if num < 0 {
		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtAccounting_838c6cb50ab0f7d4efc4d33ce55bccb5 = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 7 + 1*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	inSecondary := false
	groupThreshold := 3

	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == groupThreshold {
				b = append(b, ln.group[0])
				count = 1

				if !inSecondary {
					inSecondary = true
					groupThreshold = 2
				}
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyNegativePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyNegativePrefix[j])
		}

		for j := len(ln.minus) - 1; j >= 0; j-- {
			b = append(b, ln.minus[j])
		}

	} else {

		for j := len(symbol) - 1; j >= 0; j-- {
			b = append(b, symbol[j])
		}

		for j := len(ln.currencyPositivePrefix) - 1; j >= 0; j-- {
			b = append(b, ln.currencyPositivePrefix[j])
		}

	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	return string(b)
}

var fnFmtDateMedium_8ff89c88bdb727333d4c7b42f495646e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_af1ab393e97e5cd9bd5342a65130666d = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_ae39dfe33f097d30e514b77409da8997 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x2d}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtNumber_91f44359c32923b2c2a980df8ee6ab86 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 3 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	return string(b)
}

var fnFmtPercent_141d2f250af9b9f46295bf60380ce0b2 = func(ln *localen, num float64, v uint64) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	l := len(s) + 5
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			continue
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	b = append(b, ln.percent...)

	return string(b)
}

var fnFmtCurrency_ae69fea052716ee7e5be5d84e205ff0c = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	b = append(b, ln.currencyPositiveSuffix...)

	b = append(b, symbol...)

	return string(b)
}

var fnFmtAccounting_c2fb1e60c1c0fac4b7a6ab96cbc4f79e = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			for j := len(ln.decimal) - 1; j >= 0; j-- {
				b = append(b, ln.decimal[j])
			}
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.minus[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateLong_3097d2288f3d500e25f72a56af6d0f39 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0xd9, 0x86, 0xda, 0x86, 0xdb, 0x8c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_64ae09e94f136aa33d4d0e5627a29872 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20, 0xd9, 0x86, 0xda, 0x86, 0xdb, 0x8c, 0x20, 0xdb, 0x8c, 0xdb, 0x8c, 0xd9, 0x84, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20, 0xd9, 0x86, 0xda, 0x86, 0xdb, 0x8c, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x20, 0xda, 0xa9, 0xd9, 0x88, 0xd9, 0x86, 0xdb, 0x8c}...)

	return string(b)
}

var fnFmtDateMedium_c010b271f06248ba07bbc009fc58011a = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)

	return string(b)
}

var fnFmtDateFull_c73645256fa9ec42586084cc4fafc6f6 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x61}...)
	b = append(b, []byte{0x20, 0x64}...)
	b = append(b, []byte{0x2e, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x69, 0x64}...)

	return string(b)
}

var fnFmtDateFull_4f155c5d948146549d2721e09ed6c3a5 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x2c, 0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtAccounting_1942e97ed17a8864331f347260c0e63c = func(ln *localen, num float64, v uint64, currency currency.Type) string {
	s := strconv.FormatFloat(math.Abs(num), 'f', int(v), 64)
	symbol := ln.currencies[currency]
	l := len(s) + len(symbol) + 5 + 2*len(s[:len(s)-int(v)-1])/3
	count := 0
	inWhole := v == 0
	b := make([]byte, 0, l)

	for i := len(s) - 1; i >= 0; i-- {

		if s[i] == '.' {
			b = append(b, ln.decimal[0])
			inWhole = true
			continue
		}

		if inWhole {
			if count == 3 {
				for j := len(ln.group) - 1; j >= 0; j-- {
					b = append(b, ln.group[j])
				}
				count = 1
			} else {
				count++
			}
		}

		b = append(b, s[i])
	}

	if num < 0 {
		b = append(b, ln.currencyNegativePrefix[0])
	}

	// reverse
	for i, j := 0, len(b)-1; i < j; i, j = i+1, j-1 {
		b[i], b[j] = b[j], b[i]
	}

	if int(v) < 2 {

		if v == 0 {
			b = append(b, ln.decimal...)
		}

		for i := 0; i < 2-int(v); i++ {
			b = append(b, '0')
		}
	}

	if num < 0 {
		b = append(b, ln.currencyNegativeSuffix...)
		b = append(b, symbol...)
	} else {

		b = append(b, ln.currencyPositiveSuffix...)
		b = append(b, symbol...)
	}

	return string(b)
}

var fnFmtDateMedium_49f81df3f568828be8ec7cbfdfca6c87 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xd7, 0x98, 0xd7, 0x9f, 0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_ed7a2ed783c0a089bae0b26415cf1eec = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xd7, 0x98, 0xd7, 0x9f, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_533def3ea33c7c5e7121653c4a4fcdfb = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xd7, 0x98, 0xd7, 0x9f, 0x20}...)
	b = append(b, ln.monthsWide[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateMedium_5ee989ce4f98a493eab4c7b536fc1c06 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)

	if t.Month() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Month()), 10)

	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateLong_39c28d689b0bc45d7cd36c6f85c85cd7 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtDateFull_ea32054fb13476ec6d17746ec03afe87 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = append(b, ln.daysWide[t.Weekday()]...)
	b = append(b, []byte{0x2c, 0x20}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0x20}...)
	b = append(b, ln.monthsAbbreviated[t.Month()]...)
	b = append(b, []byte{0x20}...)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	return string(b)
}

var fnFmtTimeShort_e9e37e884fba601b3fce4f8400a30c6f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_f393bba636806dbcfe262f3337c404b9 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	b = strconv.AppendInt(b, int64(t.Hour()), 10)
	b = append(b, ln.timeSeparator...)
	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)
	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe5, 0xb9, 0xb4}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xe6, 0x9c, 0x88}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe6, 0x97, 0xa5}...)

	return string(b)
}

var fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Year() > 0 {
		b = strconv.AppendInt(b, int64(t.Year()), 10)
	} else {
		b = strconv.AppendInt(b, int64(-t.Year()), 10)
	}

	b = append(b, []byte{0xe5, 0xb9, 0xb4}...)
	b = strconv.AppendInt(b, int64(t.Month()), 10)
	b = append(b, []byte{0xe6, 0x9c, 0x88}...)
	b = strconv.AppendInt(b, int64(t.Day()), 10)
	b = append(b, []byte{0xe6, 0x97, 0xa5, 0x20}...)
	b = append(b, ln.daysWide[t.Weekday()]...)

	return string(b)
}

var fnFmtTimeShort_9f509c771840c0382f2405c11cafea81 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)

	return string(b)
}

var fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x5b}...)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x5d}...)

	return string(b)
}

var fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)
	b = append(b, []byte{0x20, 0x5b}...)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x5d}...)

	return string(b)
}

var fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	tz, _ := t.Zone()
	b = append(b, tz...)

	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

var fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657 = func(ln *localen, t time.Time) string {
	b := make([]byte, 0, 32)

	tz, _ := t.Zone()

	if btz, ok := ln.timezones[tz]; ok {
		b = append(b, btz...)
	} else {
		b = append(b, tz...)
	}

	b = append(b, []byte{0x20}...)

	if t.Hour() < 12 {
		b = append(b, ln.periodsAbbreviated[0]...)
	} else {
		b = append(b, ln.periodsAbbreviated[1]...)
	}

	h := t.Hour()

	if h > 12 {
		h -= 12
	}

	b = strconv.AppendInt(b, int64(h), 10)
	b = append(b, ln.timeSeparator...)

	if t.Minute() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Minute()), 10)
	b = append(b, ln.timeSeparator...)

	if t.Second() < 10 {
		b = append(b, '0')
	}

	b = strconv.AppendInt(b, int64(t.Second()), 10)

	return string(b)
}

type localen struct {
	currencies             []string
	currencyNegativePrefix string
	currencyNegativeSuffix string
	currencyPositivePrefix string
	currencyPositiveSuffix string
	daysAbbreviated        []string
	daysNarrow             []string
	daysShort              []string
	daysWide               []string
	decimal                string
	erasAbbreviated        []string
	erasNarrow             []string
	erasWide               []string
	fnCardinalPluralRule   func(ln *localen, num float64, v uint64) locales.PluralRule
	fnDecimal              func(ln *localen) string
	fnFmtAccounting        func(ln *localen, num float64, v uint64, currency currency.Type) string
	fnFmtCurrency          func(ln *localen, num float64, v uint64, currency currency.Type) string
	fnFmtDateFull          func(ln *localen, t time.Time) string
	fnFmtDateLong          func(ln *localen, t time.Time) string
	fnFmtDateMedium        func(ln *localen, t time.Time) string
	fnFmtDateShort         func(ln *localen, t time.Time) string
	fnFmtNumber            func(ln *localen, num float64, v uint64) string
	fnFmtPercent           func(ln *localen, num float64, v uint64) string
	fnFmtTimeFull          func(ln *localen, t time.Time) string
	fnFmtTimeLong          func(ln *localen, t time.Time) string
	fnFmtTimeMedium        func(ln *localen, t time.Time) string
	fnFmtTimeShort         func(ln *localen, t time.Time) string
	fnGroup                func(ln *localen) string
	fnLocale               func(ln *localen) string
	fnMinus                func(ln *localen) string
	fnMonthAbbreviated     func(ln *localen, month time.Month) string
	fnMonthNarrow          func(ln *localen, month time.Month) string
	fnMonthWide            func(ln *localen, month time.Month) string
	fnMonthsAbbreviated    func(ln *localen) []string
	fnMonthsNarrow         func(ln *localen) []string
	fnMonthsWide           func(ln *localen) []string
	fnOrdinalPluralRule    func(ln *localen, num float64, v uint64) locales.PluralRule
	fnPluralsCardinal      func(ln *localen) []locales.PluralRule
	fnPluralsOrdinal       func(ln *localen) []locales.PluralRule
	fnPluralsRange         func(ln *localen) []locales.PluralRule
	fnRangePluralRule      func(ln *localen, num1 float64, v1 uint64, num2 float64, v2 uint64) locales.PluralRule
	fnWeekdayAbbreviated   func(ln *localen, weekday time.Weekday) string
	fnWeekdayNarrow        func(ln *localen, weekday time.Weekday) string
	fnWeekdayShort         func(ln *localen, weekday time.Weekday) string
	fnWeekdayWide          func(ln *localen, weekday time.Weekday) string
	fnWeekdaysAbbreviated  func(ln *localen) []string
	fnWeekdaysNarrow       func(ln *localen) []string
	fnWeekdaysShort        func(ln *localen) []string
	fnWeekdaysWide         func(ln *localen) []string
	group                  string
	inifinity              string
	locale                 string
	minus                  string
	monthsAbbreviated      []string
	monthsNarrow           []string
	monthsWide             []string
	perMille               string
	percent                string
	percentPrefix          string
	percentSuffix          string
	periodsAbbreviated     []string
	periodsNarrow          []string
	periodsShort           []string
	periodsWide            []string
	pluralsCardinal        []locales.PluralRule
	pluralsOrdinal         []locales.PluralRule
	pluralsRange           []locales.PluralRule
	timeSeparator          string
	timezones              map[string]string
}

func init() {
	translatorFuncs["af"] = func() locales.Translator {
		return &localen{

			locale:                 "af",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "Mrt.", "Apr.", "Mei", "Jun.", "Jul.", "Aug.", "Sep.", "Okt.", "Nov.", "Des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januarie", "Februarie", "Maart", "April", "Mei", "Junie", "Julie", "Augustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:        []string{"So.", "Ma.", "Di.", "Wo.", "Do.", "Vr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "W", "D", "V", "S"},
			daysShort:              []string{"So.", "Ma.", "Di.", "Wo.", "Do.", "Vr.", "Sa."},
			daysWide:               []string{"Sondag", "Maandag", "Dinsdag", "Woensdag", "Donderdag", "Vrydag", "Saterdag"},
			periodsAbbreviated:     []string{"vm.", "nm."},
			periodsNarrow:          []string{"v", "n"},
			periodsWide:            []string{"vm.", "nm."},
			erasAbbreviated:        []string{"v.C.", "n.C."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Sentraal-Australiese dagligtyd", "ACST": "Sentraal-Australiese standaardtyd", "ACWDT": "Sentraal-westelike Australiese dagligtyd", "ACWST": "Sentraal-westelike Australiese standaard-tyd", "ADT": "Atlantiese dagligtyd", "AEDT": "Oostelike Australiese dagligtyd", "AEST": "Oostelike Australiese standaardtyd", "AKDT": "Alaska-dagligtyd", "AKST": "Alaska-standaardtyd", "ARST": "Argentinië-somertyd", "ART": "Argentinië-standaardtyd", "AST": "Atlantiese standaardtyd", "AWDT": "Westelike Australiese dagligtyd", "AWST": "Westelike Australiese standaardtyd", "BOT": "Bolivia-tyd", "BT": "Bhoetan-tyd", "CAT": "Sentraal-Afrika-tyd", "CDT": "Noord-Amerikaanse sentrale dagligtyd", "CHADT": "Chatham-dagligtyd", "CHAST": "Chatham-standaardtyd", "CLST": "Chili-somertyd", "CLT": "Chili-standaardtyd", "COST": "Colombia-somertyd", "COT": "Colombië-standaardtyd", "CST": "Noord-Amerikaanse sentrale standaardtyd", "ChST": "Chamorro-standaardtyd", "EAT": "Oos-Afrika-tyd", "ECT": "Ecuador-tyd", "EDT": "Noord-Amerikaanse oostelike dagligtyd", "EST": "Noord-Amerikaanse oostelike standaardtyd", "GFT": "Frans-Guiana-tyd", "GMT": "Greenwich-tyd", "GST": "Persiese Golf-standaardtyd", "GYT": "Guyana-tyd", "HADT": "Hawaii-Aleoete-dagligtyd", "HAST": "Hawaii-Aleoete-standaardtyd", "HAT": "Newfoundland-dagligtyd", "HECU": "Kuba-dagligtyd", "HEEG": "Oos-Groenland-somertyd", "HENOMX": "Noordwes-Meksiko-dagligtyd", "HEOG": "Wes-Groenland-somertyd", "HEPM": "Sint-Pierre en Miquelon-dagligtyd", "HEPMX": "Meksikaanse Pasifiese dagligtyd", "HKST": "Hongkong-somertyd", "HKT": "Hongkong-standaardtyd", "HNCU": "Kuba-standaardtyd", "HNEG": "Oos-Groenland-standaardtyd", "HNNOMX": "Noordwes-Meksiko-standaardtyd", "HNOG": "Wes-Groenland-standaardtyd", "HNPM": "Sint-Pierre en Miquelon-standaardtyd", "HNPMX": "Meksikaanse Pasifiese standaardtyd", "HNT": "Newfoundland-standaardtyd", "IST": "Indië-standaardtyd", "JDT": "Japan-dagligtyd", "JST": "Japan-standaardtyd", "LHDT": "Lord Howe-dagligtyd", "LHST": "Lord Howe-standaardtyd", "MDT": "Noord-Amerikaanse berg-dagligtyd", "MESZ": "Sentraal-Europese somertyd", "MEZ": "Sentraal-Europese standaardtyd", "MST": "Noord-Amerikaanse berg-standaardtyd", "MYT": "Maleisië-tyd", "NZDT": "Nieu-Seeland-dagligtyd", "NZST": "Nieu-Seeland-standaardtyd", "OESZ": "Oos-Europese somertyd", "OEZ": "Oos-Europese standaardtyd", "PDT": "Pasifiese dagligtyd", "PST": "Pasifiese standaardtyd", "SAST": "Suid-Afrika-standaardtyd", "SGT": "Singapoer-standaardtyd", "SRT": "Suriname-tyd", "TMST": "Turkmenistan-somertyd", "TMT": "Turkmenistan-standaardtyd", "UYST": "Uruguay-somertyd", "UYT": "Uruguay-standaardtyd", "VET": "Venezuela-tyd", "WARST": "Wes-Argentinië-somertyd", "WART": "Wes-Argentinië-standaardtyd", "WAST": "Wes-Afrika-somertyd", "WAT": "Wes-Afrika-standaardtyd", "WESZ": "Wes-Europese somertyd", "WEZ": "Wes-Europese standaardtyd", "WIB": "Wes-Indonesië-tyd", "WIT": "Oos-Indonesië-tyd", "WITA": "Sentraal-Indonesiese tyd", "∅∅∅": "Brasilia-somertyd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ee2956d02335482ec6b7e9a423beb0b5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["af_na"] = func() locales.Translator {
		return &localen{

			locale:                 "af_NA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "$", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "Mrt.", "Apr.", "Mei", "Jun.", "Jul.", "Aug.", "Sep.", "Okt.", "Nov.", "Des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januarie", "Februarie", "Maart", "April", "Mei", "Junie", "Julie", "Augustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:        []string{"So.", "Ma.", "Di.", "Wo.", "Do.", "Vr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "W", "D", "V", "S"},
			daysShort:              []string{"So.", "Ma.", "Di.", "Wo.", "Do.", "Vr.", "Sa."},
			daysWide:               []string{"Sondag", "Maandag", "Dinsdag", "Woensdag", "Donderdag", "Vrydag", "Saterdag"},
			periodsAbbreviated:     []string{"vm.", "nm."},
			periodsNarrow:          []string{"v", "n"},
			periodsWide:            []string{"vm.", "nm."},
			erasAbbreviated:        []string{"v.C.", "n.C."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Sentraal-Australiese dagligtyd", "ACST": "Sentraal-Australiese standaardtyd", "ACWDT": "Sentraal-westelike Australiese dagligtyd", "ACWST": "Sentraal-westelike Australiese standaard-tyd", "ADT": "Atlantiese dagligtyd", "AEDT": "Oostelike Australiese dagligtyd", "AEST": "Oostelike Australiese standaardtyd", "AKDT": "Alaska-dagligtyd", "AKST": "Alaska-standaardtyd", "ARST": "Argentinië-somertyd", "ART": "Argentinië-standaardtyd", "AST": "Atlantiese standaardtyd", "AWDT": "Westelike Australiese dagligtyd", "AWST": "Westelike Australiese standaardtyd", "BOT": "Bolivia-tyd", "BT": "Bhoetan-tyd", "CAT": "Sentraal-Afrika-tyd", "CDT": "Noord-Amerikaanse sentrale dagligtyd", "CHADT": "Chatham-dagligtyd", "CHAST": "Chatham-standaardtyd", "CLST": "Chili-somertyd", "CLT": "Chili-standaardtyd", "COST": "Colombia-somertyd", "COT": "Colombië-standaardtyd", "CST": "Noord-Amerikaanse sentrale standaardtyd", "ChST": "Chamorro-standaardtyd", "EAT": "Oos-Afrika-tyd", "ECT": "Ecuador-tyd", "EDT": "Noord-Amerikaanse oostelike dagligtyd", "EST": "Noord-Amerikaanse oostelike standaardtyd", "GFT": "Frans-Guiana-tyd", "GMT": "Greenwich-tyd", "GST": "Persiese Golf-standaardtyd", "GYT": "Guyana-tyd", "HADT": "Hawaii-Aleoete-dagligtyd", "HAST": "Hawaii-Aleoete-standaardtyd", "HAT": "Newfoundland-dagligtyd", "HECU": "Kuba-dagligtyd", "HEEG": "Oos-Groenland-somertyd", "HENOMX": "Noordwes-Meksiko-dagligtyd", "HEOG": "Wes-Groenland-somertyd", "HEPM": "Sint-Pierre en Miquelon-dagligtyd", "HEPMX": "Meksikaanse Pasifiese dagligtyd", "HKST": "Hongkong-somertyd", "HKT": "Hongkong-standaardtyd", "HNCU": "Kuba-standaardtyd", "HNEG": "Oos-Groenland-standaardtyd", "HNNOMX": "Noordwes-Meksiko-standaardtyd", "HNOG": "Wes-Groenland-standaardtyd", "HNPM": "Sint-Pierre en Miquelon-standaardtyd", "HNPMX": "Meksikaanse Pasifiese standaardtyd", "HNT": "Newfoundland-standaardtyd", "IST": "Indië-standaardtyd", "JDT": "Japan-dagligtyd", "JST": "Japan-standaardtyd", "LHDT": "Lord Howe-dagligtyd", "LHST": "Lord Howe-standaardtyd", "MDT": "Noord-Amerikaanse berg-dagligtyd", "MESZ": "Sentraal-Europese somertyd", "MEZ": "Sentraal-Europese standaardtyd", "MST": "Noord-Amerikaanse berg-standaardtyd", "MYT": "Maleisië-tyd", "NZDT": "Nieu-Seeland-dagligtyd", "NZST": "Nieu-Seeland-standaardtyd", "OESZ": "Oos-Europese somertyd", "OEZ": "Oos-Europese standaardtyd", "PDT": "Pasifiese dagligtyd", "PST": "Pasifiese standaardtyd", "SAST": "Suid-Afrika-standaardtyd", "SGT": "Singapoer-standaardtyd", "SRT": "Suriname-tyd", "TMST": "Turkmenistan-somertyd", "TMT": "Turkmenistan-standaardtyd", "UYST": "Uruguay-somertyd", "UYT": "Uruguay-standaardtyd", "VET": "Venezuela-tyd", "WARST": "Wes-Argentinië-somertyd", "WART": "Wes-Argentinië-standaardtyd", "WAST": "Wes-Afrika-somertyd", "WAT": "Wes-Afrika-standaardtyd", "WESZ": "Wes-Europese somertyd", "WEZ": "Wes-Europese standaardtyd", "WIB": "Wes-Indonesië-tyd", "WIT": "Oos-Indonesië-tyd", "WITA": "Sentraal-Indonesiese tyd", "∅∅∅": "Asore-somertyd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ee2956d02335482ec6b7e9a423beb0b5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["af_za"] = func() locales.Translator {
		return &localen{

			locale:                 "af_ZA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "Mrt.", "Apr.", "Mei", "Jun.", "Jul.", "Aug.", "Sep.", "Okt.", "Nov.", "Des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januarie", "Februarie", "Maart", "April", "Mei", "Junie", "Julie", "Augustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:        []string{"So.", "Ma.", "Di.", "Wo.", "Do.", "Vr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "W", "D", "V", "S"},
			daysShort:              []string{"So.", "Ma.", "Di.", "Wo.", "Do.", "Vr.", "Sa."},
			daysWide:               []string{"Sondag", "Maandag", "Dinsdag", "Woensdag", "Donderdag", "Vrydag", "Saterdag"},
			periodsAbbreviated:     []string{"vm.", "nm."},
			periodsNarrow:          []string{"v", "n"},
			periodsWide:            []string{"vm.", "nm."},
			erasAbbreviated:        []string{"v.C.", "n.C."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Sentraal-Australiese dagligtyd", "ACST": "Sentraal-Australiese standaardtyd", "ACWDT": "Sentraal-westelike Australiese dagligtyd", "ACWST": "Sentraal-westelike Australiese standaard-tyd", "ADT": "Atlantiese dagligtyd", "AEDT": "Oostelike Australiese dagligtyd", "AEST": "Oostelike Australiese standaardtyd", "AKDT": "Alaska-dagligtyd", "AKST": "Alaska-standaardtyd", "ARST": "Argentinië-somertyd", "ART": "Argentinië-standaardtyd", "AST": "Atlantiese standaardtyd", "AWDT": "Westelike Australiese dagligtyd", "AWST": "Westelike Australiese standaardtyd", "BOT": "Bolivia-tyd", "BT": "Bhoetan-tyd", "CAT": "Sentraal-Afrika-tyd", "CDT": "Noord-Amerikaanse sentrale dagligtyd", "CHADT": "Chatham-dagligtyd", "CHAST": "Chatham-standaardtyd", "CLST": "Chili-somertyd", "CLT": "Chili-standaardtyd", "COST": "Colombia-somertyd", "COT": "Colombië-standaardtyd", "CST": "Noord-Amerikaanse sentrale standaardtyd", "ChST": "Chamorro-standaardtyd", "EAT": "Oos-Afrika-tyd", "ECT": "Ecuador-tyd", "EDT": "Noord-Amerikaanse oostelike dagligtyd", "EST": "Noord-Amerikaanse oostelike standaardtyd", "GFT": "Frans-Guiana-tyd", "GMT": "Greenwich-tyd", "GST": "Persiese Golf-standaardtyd", "GYT": "Guyana-tyd", "HADT": "Hawaii-Aleoete-dagligtyd", "HAST": "Hawaii-Aleoete-standaardtyd", "HAT": "Newfoundland-dagligtyd", "HECU": "Kuba-dagligtyd", "HEEG": "Oos-Groenland-somertyd", "HENOMX": "Noordwes-Meksiko-dagligtyd", "HEOG": "Wes-Groenland-somertyd", "HEPM": "Sint-Pierre en Miquelon-dagligtyd", "HEPMX": "Meksikaanse Pasifiese dagligtyd", "HKST": "Hongkong-somertyd", "HKT": "Hongkong-standaardtyd", "HNCU": "Kuba-standaardtyd", "HNEG": "Oos-Groenland-standaardtyd", "HNNOMX": "Noordwes-Meksiko-standaardtyd", "HNOG": "Wes-Groenland-standaardtyd", "HNPM": "Sint-Pierre en Miquelon-standaardtyd", "HNPMX": "Meksikaanse Pasifiese standaardtyd", "HNT": "Newfoundland-standaardtyd", "IST": "Indië-standaardtyd", "JDT": "Japan-dagligtyd", "JST": "Japan-standaardtyd", "LHDT": "Lord Howe-dagligtyd", "LHST": "Lord Howe-standaardtyd", "MDT": "MDT", "MESZ": "Sentraal-Europese somertyd", "MEZ": "Sentraal-Europese standaardtyd", "MST": "MST", "MYT": "Maleisië-tyd", "NZDT": "Nieu-Seeland-dagligtyd", "NZST": "Nieu-Seeland-standaardtyd", "OESZ": "Oos-Europese somertyd", "OEZ": "Oos-Europese standaardtyd", "PDT": "Pasifiese dagligtyd", "PST": "Pasifiese standaardtyd", "SAST": "Suid-Afrika-standaardtyd", "SGT": "Singapoer-standaardtyd", "SRT": "Suriname-tyd", "TMST": "Turkmenistan-somertyd", "TMT": "Turkmenistan-standaardtyd", "UYST": "Uruguay-somertyd", "UYT": "Uruguay-standaardtyd", "VET": "Venezuela-tyd", "WARST": "Wes-Argentinië-somertyd", "WART": "Wes-Argentinië-standaardtyd", "WAST": "Wes-Afrika-somertyd", "WAT": "Wes-Afrika-standaardtyd", "WESZ": "Wes-Europese somertyd", "WEZ": "Wes-Europese standaardtyd", "WIB": "Wes-Indonesië-tyd", "WIT": "Oos-Indonesië-tyd", "WITA": "Sentraal-Indonesiese tyd", "∅∅∅": "Peru-somertyd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ee2956d02335482ec6b7e9a423beb0b5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["agq"] = func() locales.Translator {
		return &localen{

			locale:             "agq",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "nùm", "kɨz", "tɨd", "taa", "see", "nzu", "dum", "fɔe", "dzu", "lɔm", "kaa", "fwo"},
			monthsNarrow:       []string{"", "n", "k", "t", "t", "s", "z", "k", "f", "d", "l", "c", "f"},
			monthsWide:         []string{"", "ndzɔ̀ŋɔ̀nùm", "ndzɔ̀ŋɔ̀kƗ̀zùʔ", "ndzɔ̀ŋɔ̀tƗ̀dʉ̀ghà", "ndzɔ̀ŋɔ̀tǎafʉ̄ghā", "ndzɔ̀ŋèsèe", "ndzɔ̀ŋɔ̀nzùghò", "ndzɔ̀ŋɔ̀dùmlo", "ndzɔ̀ŋɔ̀kwîfɔ̀e", "ndzɔ̀ŋɔ̀tƗ̀fʉ̀ghàdzughù", "ndzɔ̀ŋɔ̀ghǔuwelɔ̀m", "ndzɔ̀ŋɔ̀chwaʔàkaa wo", "ndzɔ̀ŋèfwòo"},
			daysAbbreviated:    []string{"nts", "kpa", "ghɔ", "tɔm", "ume", "ghɨ", "dzk"},
			daysNarrow:         []string{"n", "k", "g", "t", "u", "g", "d"},
			daysWide:           []string{"tsuʔntsɨ", "tsuʔukpà", "tsuʔughɔe", "tsuʔutɔ̀mlò", "tsuʔumè", "tsuʔughɨ̂m", "tsuʔndzɨkɔʔɔ"},
			periodsAbbreviated: []string{"a.g", "a.k"},
			periodsWide:        []string{"a.g", "a.k"},
			erasAbbreviated:    []string{"SK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Sěe Kɨ̀lesto", "Bǎa Kɨ̀lesto"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["agq_cm"] = func() locales.Translator {
		return &localen{

			locale:             "agq_CM",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "nùm", "kɨz", "tɨd", "taa", "see", "nzu", "dum", "fɔe", "dzu", "lɔm", "kaa", "fwo"},
			monthsNarrow:       []string{"", "n", "k", "t", "t", "s", "z", "k", "f", "d", "l", "c", "f"},
			monthsWide:         []string{"", "ndzɔ̀ŋɔ̀nùm", "ndzɔ̀ŋɔ̀kƗ̀zùʔ", "ndzɔ̀ŋɔ̀tƗ̀dʉ̀ghà", "ndzɔ̀ŋɔ̀tǎafʉ̄ghā", "ndzɔ̀ŋèsèe", "ndzɔ̀ŋɔ̀nzùghò", "ndzɔ̀ŋɔ̀dùmlo", "ndzɔ̀ŋɔ̀kwîfɔ̀e", "ndzɔ̀ŋɔ̀tƗ̀fʉ̀ghàdzughù", "ndzɔ̀ŋɔ̀ghǔuwelɔ̀m", "ndzɔ̀ŋɔ̀chwaʔàkaa wo", "ndzɔ̀ŋèfwòo"},
			daysAbbreviated:    []string{"nts", "kpa", "ghɔ", "tɔm", "ume", "ghɨ", "dzk"},
			daysNarrow:         []string{"n", "k", "g", "t", "u", "g", "d"},
			daysWide:           []string{"tsuʔntsɨ", "tsuʔukpà", "tsuʔughɔe", "tsuʔutɔ̀mlò", "tsuʔumè", "tsuʔughɨ̂m", "tsuʔndzɨkɔʔɔ"},
			periodsAbbreviated: []string{"a.g", "a.k"},
			periodsWide:        []string{"a.g", "a.k"},
			erasAbbreviated:    []string{"SK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Sěe Kɨ̀lesto", "Bǎa Kɨ̀lesto"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ak"] = func() locales.Translator {
		return &localen{

			locale:             "ak",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GH₵", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "S-Ɔ", "K-Ɔ", "E-Ɔ", "E-O", "E-K", "O-A", "A-K", "D-Ɔ", "F-Ɛ", "Ɔ-A", "Ɔ-O", "M-Ɔ"},
			monthsWide:         []string{"", "Sanda-Ɔpɛpɔn", "Kwakwar-Ɔgyefuo", "Ebɔw-Ɔbenem", "Ebɔbira-Oforisuo", "Esusow Aketseaba-Kɔtɔnimba", "Obirade-Ayɛwohomumu", "Ayɛwoho-Kitawonsa", "Difuu-Ɔsandaa", "Fankwa-Ɛbɔ", "Ɔbɛsɛ-Ahinime", "Ɔberɛfɛw-Obubuo", "Mumu-Ɔpɛnimba"},
			daysAbbreviated:    []string{"Kwe", "Dwo", "Ben", "Wuk", "Yaw", "Fia", "Mem"},
			daysNarrow:         []string{"K", "D", "B", "W", "Y", "F", "M"},
			daysWide:           []string{"Kwesida", "Dwowda", "Benada", "Wukuda", "Yawda", "Fida", "Memeneda"},
			periodsAbbreviated: []string{"AN", "EW"},
			periodsWide:        []string{"AN", "EW"},
			erasAbbreviated:    []string{"AK", "KE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Ansa Kristo", "Kristo Ekyiri"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ec9761d27ad9ead532f5f112e23983b1,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93244c9b90b1456ed872d8fc125fd668,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ak_gh"] = func() locales.Translator {
		return &localen{

			locale:             "ak_GH",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "S-Ɔ", "K-Ɔ", "E-Ɔ", "E-O", "E-K", "O-A", "A-K", "D-Ɔ", "F-Ɛ", "Ɔ-A", "Ɔ-O", "M-Ɔ"},
			monthsWide:         []string{"", "Sanda-Ɔpɛpɔn", "Kwakwar-Ɔgyefuo", "Ebɔw-Ɔbenem", "Ebɔbira-Oforisuo", "Esusow Aketseaba-Kɔtɔnimba", "Obirade-Ayɛwohomumu", "Ayɛwoho-Kitawonsa", "Difuu-Ɔsandaa", "Fankwa-Ɛbɔ", "Ɔbɛsɛ-Ahinime", "Ɔberɛfɛw-Obubuo", "Mumu-Ɔpɛnimba"},
			daysAbbreviated:    []string{"Kwe", "Dwo", "Ben", "Wuk", "Yaw", "Fia", "Mem"},
			daysNarrow:         []string{"K", "D", "B", "W", "Y", "F", "M"},
			daysWide:           []string{"Kwesida", "Dwowda", "Benada", "Wukuda", "Yawda", "Fida", "Memeneda"},
			periodsAbbreviated: []string{"AN", "EW"},
			periodsWide:        []string{"AN", "EW"},
			erasAbbreviated:    []string{"AK", "KE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Ansa Kristo", "Kristo Ekyiri"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ec9761d27ad9ead532f5f112e23983b1,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93244c9b90b1456ed872d8fc125fd668,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["am"] = func() locales.Translator {
		return &localen{

			locale:                 "am",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "የቻይና ዩዋን", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ብር", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ጃንዩ", "ፌብሩ", "ማርች", "ኤፕሪ", "ሜይ", "ጁን", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክቶ", "ኖቬም", "ዲሴም"},
			monthsNarrow:           []string{"", "ጃ", "ፌ", "ማ", "ኤ", "ሜ", "ጁ", "ጁ", "ኦ", "ሴ", "ኦ", "ኖ", "ዲ"},
			monthsWide:             []string{"", "ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕሪል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክቶበር", "ኖቬምበር", "ዲሴምበር"},
			daysAbbreviated:        []string{"እሑድ", "ሰኞ", "ማክሰ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"},
			daysNarrow:             []string{"እ", "ሰ", "ማ", "ረ", "ሐ", "ዓ", "ቅ"},
			daysShort:              []string{"እ", "ሰ", "ማ", "ረ", "ሐ", "ዓ", "ቅ"},
			daysWide:               []string{"እሑድ", "ሰኞ", "ማክሰኞ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"},
			periodsAbbreviated:     []string{"ጥዋት", "ከሰዓት"},
			periodsNarrow:          []string{"ጠ", "ከ"},
			periodsWide:            []string{"ጥዋት", "ከሰዓት"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"ዓ/ዓ", "ዓ/ም"},
			erasWide:               []string{"ዓመተ ዓለም", "ዓመተ ምሕረት"},
			timezones:              map[string]string{"ACDT": "የአውስትራሊያ መካከለኛ የቀን ሰዓት አቆጣጠር", "ACST": "የአውስትራሊያ መካከለኛ መደበኛ የሰዓት አቆጣጠር", "ACWDT": "የአውስትራሊያ መካከለኛው ምስራቅ የቀን ሰዓት አቆጣጠር", "ACWST": "የአውስትራሊያ መካከለኛ ምስራቃዊ መደበኛ ሰዓት አቆጣጠር", "ADT": "የአትላንቲክ የቀን ሰዓት አቆጣጠር", "AEDT": "የአውስትራሊያ ምዕራባዊ የቀን ሰዓት አቆጣጠር", "AEST": "የአውስትራሊያ ምዕራባዊ መደበኛ የሰዓት አቆጣጠር", "AKDT": "የአላስካ የቀን ሰዓት አቆጣጠር", "AKST": "የአላስካ መደበኛ የሰዓት አቆጣጠር", "ARST": "የአርጀንቲና የበጋ ሰዓት አቆጣጠር", "ART": "የአርጀንቲና መደበኛ ሰዓት አቆጣጠር", "AST": "የአትላንቲክ መደበኛ የሰዓት አቆጣጠር", "AWDT": "የአውስትራሊያ ምስራቃዊ የቀን ሰዓት አቆጣጠር", "AWST": "የአውስትራሊያ ምስራቃዊ መደበኛ ሰዓት አቆጣጠር", "BOT": "የቦሊቪያ ሰዓት", "BT": "የቡታን ሰዓት", "CAT": "የመካከለኛው አፍሪካ ሰዓት", "CDT": "የመካከለኛ የቀን ሰዓት አቆጣጠር", "CHADT": "የቻታም የቀን ብርሃን ሰዓት", "CHAST": "የቻታም መደበኛ ሰዓት", "CLST": "የቺሊ ክረምት ሰዓት", "CLT": "የቺሊ መደበኛ ሰዓት", "COST": "የኮሎምቢያ ክረምት ሰዓት", "COT": "የኮሎምቢያ መደበኛ ሰዓት", "CST": "የሰሜን አሜሪካ የመካከለኛ መደበኛ ሰዓት አቆጣጠር", "ChST": "የቻሞሮ መደበኛ ሰዓት", "EAT": "የምስራቅ አፍሪካ ሰዓት", "ECT": "የኢኳዶር ሰዓት", "EDT": "ምስራቃዊ የቀን ሰዓት አቆጣጠር", "EST": "ምስራቃዊ መደበኛ ሰዓት አቆጣጠር", "GFT": "የፈረንሳይ ጉያና ሰዓት", "GMT": "ግሪንዊች ማዕከላዊ ሰዓት", "GST": "የባህረሰላጤ መደበኛ ሰዓት", "GYT": "የጉያና ሰዓት", "HADT": "የሃዋይ አሌኡት የቀን ሰዓት አቆጣጠር", "HAST": "የሃዋይ አሌኡት መደበኛ ሰዓት አቆጣጠር", "HAT": "የኒውፋውንድላንድ የቀን የሰዓት አቆጣጠር", "HECU": "የኩባ የቀን ብርሃን ሰዓት", "HEEG": "የምስራቅ ግሪንላንድ ክረምት ሰዓት", "HENOMX": "ሰሜናዊ ምእራብ የሜክሲኮ የቀን ሰዓት አቆጣጠር", "HEOG": "የምዕራብ ግሪንላንድ ክረምት ሰዓት", "HEPM": "ቅዱስ የፒዬር እና ሚኴሎን የቀን ብርሃን ሰዓት", "HEPMX": "የሜክሲኮ ፓሲፊክ የቀን ሰዓት አቆጣጠር", "HKST": "የሆንግ ኮንግ ክረምት ሰዓት", "HKT": "የሆንግ ኮንግ መደበኛ ሰዓት", "HNCU": "የኩባ መደበኛ ሰዓት", "HNEG": "የምስራቅ ግሪንላንድ መደበኛ ሰዓት", "HNNOMX": "ሰሜናዊ ምእራብ የሜክሲኮ መደበኛ ሰዓት አቆጣጠር", "HNOG": "የምዕራብ ግሪንላንድ መደበኛ ሰዓት", "HNPM": "ቅዱስ የፒዬር እና ሚኴሎን መደበኛ ሰዓት", "HNPMX": "የሜክሲኮ ፓሲፊክ መደበኛ ሰዓት አቆጣጠር", "HNT": "የኒውፋውንድላንድ መደበኛ የሰዓት አቆጣጠር", "IST": "የህንድ መደበኛ ሰዓት", "JDT": "የጃፓን የቀን ብርሃን ሰዓት", "JST": "የጃፓን መደበኛ ሰዓት", "LHDT": "የሎርድ ሆዌ የቀን ሰዓት አቆጣጠር", "LHST": "የሎርድ ሆዌ መደበኛ የሰዓት አቆጣጠር", "MDT": "የተራራ የቀንሰዓት አቆጣጠር", "MESZ": "የመካከለኛው አውሮፓ ክረምት ሰዓት", "MEZ": "የመካከለኛው አውሮፓ መደበኛ ሰዓት", "MST": "የተራራ መደበኛ የሰዓት አቆጣጠር", "MYT": "የማሌይዢያ ሰዓት", "NZDT": "የኒው ዚላንድ የቀን ብርሃን ሰዓት", "NZST": "የኒው ዚላንድ መደበኛ ሰዓት", "OESZ": "የምስራቃዊ አውሮፓ ክረምት ሰዓት", "OEZ": "የምስራቃዊ አውሮፓ መደበኛ ሰዓት", "PDT": "የፓስፊክ የቀን ሰዓት አቆጣጠር", "PST": "የፓስፊክ መደበኛ ሰዓት አቆጣጠር", "SAST": "የደቡብ አፍሪካ መደበኛ ሰዓት", "SGT": "የሲንጋፒር መደበኛ ሰዓት", "SRT": "የሱሪናም ሰዓት", "TMST": "የቱርክመኒስታን ክረምት ሰዓት", "TMT": "የቱርክመኒስታን መደበኛ ሰዓት", "UYST": "የኡራጓይ ክረምት ሰዓት", "UYT": "የኡራጓይ መደበኛ ሰዓት", "VET": "የቬኔዝዌላ ሰዓት", "WARST": "የአርጀንቲና ምስራቃዊ በጋ ሰዓት አቆጣጠር", "WART": "የምዕራባዊ አርጀንቲና መደበኛ ሰዓት አቆጣጠር", "WAST": "የምዕራብ አፍሪካ ክረምት ሰዓት", "WAT": "የምዕራብ አፍሪካ መደበኛ ሰዓት", "WESZ": "የምዕራባዊ አውሮፓ ክረምት ሰዓት", "WEZ": "የምዕራባዊ አውሮፓ መደበኛ ሰዓት", "WIB": "የምዕራባዊ ኢንዶኔዢያ ሰዓት", "WIT": "የምስራቃዊ ኢንዶኔዢያ ሰዓት", "WITA": "የመካከለኛው ኢንዶኔዢያ ሰዓት", "∅∅∅": "የብራዚላ የበጋ ሰዓት አቆጣጠር"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["am_et"] = func() locales.Translator {
		return &localen{

			locale:                 "am_ET",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ጃንዩ", "ፌብሩ", "ማርች", "ኤፕሪ", "ሜይ", "ጁን", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክቶ", "ኖቬም", "ዲሴም"},
			monthsNarrow:           []string{"", "ጃ", "ፌ", "ማ", "ኤ", "ሜ", "ጁ", "ጁ", "ኦ", "ሴ", "ኦ", "ኖ", "ዲ"},
			monthsWide:             []string{"", "ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕሪል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክቶበር", "ኖቬምበር", "ዲሴምበር"},
			daysAbbreviated:        []string{"እሑድ", "ሰኞ", "ማክሰ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"},
			daysNarrow:             []string{"እ", "ሰ", "ማ", "ረ", "ሐ", "ዓ", "ቅ"},
			daysShort:              []string{"እ", "ሰ", "ማ", "ረ", "ሐ", "ዓ", "ቅ"},
			daysWide:               []string{"እሑድ", "ሰኞ", "ማክሰኞ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"},
			periodsAbbreviated:     []string{"ጥዋት", "ከሰዓት"},
			periodsNarrow:          []string{"ጠ", "ከ"},
			periodsWide:            []string{"ጥዋት", "ከሰዓት"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"ዓ/ዓ", "ዓ/ም"},
			erasWide:               []string{"ዓመተ ዓለም", "ዓመተ ምሕረት"},
			timezones:              map[string]string{"ACDT": "የአውስትራሊያ መካከለኛ የቀን ሰዓት አቆጣጠር", "ACST": "የአውስትራሊያ መካከለኛ መደበኛ የሰዓት አቆጣጠር", "ACWDT": "የአውስትራሊያ መካከለኛው ምስራቅ የቀን ሰዓት አቆጣጠር", "ACWST": "የአውስትራሊያ መካከለኛ ምስራቃዊ መደበኛ ሰዓት አቆጣጠር", "ADT": "የአትላንቲክ የቀን ሰዓት አቆጣጠር", "AEDT": "የአውስትራሊያ ምዕራባዊ የቀን ሰዓት አቆጣጠር", "AEST": "የአውስትራሊያ ምዕራባዊ መደበኛ የሰዓት አቆጣጠር", "AKDT": "የአላስካ የቀን ሰዓት አቆጣጠር", "AKST": "የአላስካ መደበኛ የሰዓት አቆጣጠር", "ARST": "የአርጀንቲና የበጋ ሰዓት አቆጣጠር", "ART": "የአርጀንቲና መደበኛ ሰዓት አቆጣጠር", "AST": "የአትላንቲክ መደበኛ የሰዓት አቆጣጠር", "AWDT": "የአውስትራሊያ ምስራቃዊ የቀን ሰዓት አቆጣጠር", "AWST": "የአውስትራሊያ ምስራቃዊ መደበኛ ሰዓት አቆጣጠር", "BOT": "የቦሊቪያ ሰዓት", "BT": "የቡታን ሰዓት", "CAT": "የመካከለኛው አፍሪካ ሰዓት", "CDT": "የመካከለኛ የቀን ሰዓት አቆጣጠር", "CHADT": "የቻታም የቀን ብርሃን ሰዓት", "CHAST": "የቻታም መደበኛ ሰዓት", "CLST": "የቺሊ ክረምት ሰዓት", "CLT": "የቺሊ መደበኛ ሰዓት", "COST": "የኮሎምቢያ ክረምት ሰዓት", "COT": "የኮሎምቢያ መደበኛ ሰዓት", "CST": "የሰሜን አሜሪካ የመካከለኛ መደበኛ ሰዓት አቆጣጠር", "ChST": "የቻሞሮ መደበኛ ሰዓት", "EAT": "የምስራቅ አፍሪካ ሰዓት", "ECT": "የኢኳዶር ሰዓት", "EDT": "ምስራቃዊ የቀን ሰዓት አቆጣጠር", "EST": "ምስራቃዊ መደበኛ ሰዓት አቆጣጠር", "GFT": "የፈረንሳይ ጉያና ሰዓት", "GMT": "ግሪንዊች ማዕከላዊ ሰዓት", "GST": "የባህረሰላጤ መደበኛ ሰዓት", "GYT": "የጉያና ሰዓት", "HADT": "የሃዋይ አሌኡት የቀን ሰዓት አቆጣጠር", "HAST": "የሃዋይ አሌኡት መደበኛ ሰዓት አቆጣጠር", "HAT": "የኒውፋውንድላንድ የቀን የሰዓት አቆጣጠር", "HECU": "የኩባ የቀን ብርሃን ሰዓት", "HEEG": "የምስራቅ ግሪንላንድ ክረምት ሰዓት", "HENOMX": "ሰሜናዊ ምእራብ የሜክሲኮ የቀን ሰዓት አቆጣጠር", "HEOG": "የምዕራብ ግሪንላንድ ክረምት ሰዓት", "HEPM": "ቅዱስ የፒዬር እና ሚኴሎን የቀን ብርሃን ሰዓት", "HEPMX": "የሜክሲኮ ፓሲፊክ የቀን ሰዓት አቆጣጠር", "HKST": "የሆንግ ኮንግ ክረምት ሰዓት", "HKT": "የሆንግ ኮንግ መደበኛ ሰዓት", "HNCU": "የኩባ መደበኛ ሰዓት", "HNEG": "የምስራቅ ግሪንላንድ መደበኛ ሰዓት", "HNNOMX": "ሰሜናዊ ምእራብ የሜክሲኮ መደበኛ ሰዓት አቆጣጠር", "HNOG": "የምዕራብ ግሪንላንድ መደበኛ ሰዓት", "HNPM": "ቅዱስ የፒዬር እና ሚኴሎን መደበኛ ሰዓት", "HNPMX": "የሜክሲኮ ፓሲፊክ መደበኛ ሰዓት አቆጣጠር", "HNT": "የኒውፋውንድላንድ መደበኛ የሰዓት አቆጣጠር", "IST": "የህንድ መደበኛ ሰዓት", "JDT": "የጃፓን የቀን ብርሃን ሰዓት", "JST": "የጃፓን መደበኛ ሰዓት", "LHDT": "የሎርድ ሆዌ የቀን ሰዓት አቆጣጠር", "LHST": "የሎርድ ሆዌ መደበኛ የሰዓት አቆጣጠር", "MDT": "የተራራ የቀንሰዓት አቆጣጠር", "MESZ": "የመካከለኛው አውሮፓ ክረምት ሰዓት", "MEZ": "የመካከለኛው አውሮፓ መደበኛ ሰዓት", "MST": "የተራራ መደበኛ የሰዓት አቆጣጠር", "MYT": "የማሌይዢያ ሰዓት", "NZDT": "የኒው ዚላንድ የቀን ብርሃን ሰዓት", "NZST": "የኒው ዚላንድ መደበኛ ሰዓት", "OESZ": "የምስራቃዊ አውሮፓ ክረምት ሰዓት", "OEZ": "የምስራቃዊ አውሮፓ መደበኛ ሰዓት", "PDT": "የፓስፊክ የቀን ሰዓት አቆጣጠር", "PST": "የፓስፊክ መደበኛ ሰዓት አቆጣጠር", "SAST": "የደቡብ አፍሪካ መደበኛ ሰዓት", "SGT": "የሲንጋፒር መደበኛ ሰዓት", "SRT": "የሱሪናም ሰዓት", "TMST": "የቱርክመኒስታን ክረምት ሰዓት", "TMT": "የቱርክመኒስታን መደበኛ ሰዓት", "UYST": "የኡራጓይ ክረምት ሰዓት", "UYT": "የኡራጓይ መደበኛ ሰዓት", "VET": "የቬኔዝዌላ ሰዓት", "WARST": "የአርጀንቲና ምስራቃዊ በጋ ሰዓት አቆጣጠር", "WART": "የምዕራባዊ አርጀንቲና መደበኛ ሰዓት አቆጣጠር", "WAST": "የምዕራብ አፍሪካ ክረምት ሰዓት", "WAT": "የምዕራብ አፍሪካ መደበኛ ሰዓት", "WESZ": "የምዕራባዊ አውሮፓ ክረምት ሰዓት", "WEZ": "የምዕራባዊ አውሮፓ መደበኛ ሰዓት", "WIB": "የምዕራባዊ ኢንዶኔዢያ ሰዓት", "WIT": "የምስራቃዊ ኢንዶኔዢያ ሰዓት", "WITA": "የመካከለኛው ኢንዶኔዢያ ሰዓት", "∅∅∅": "የፔሩ ክረምት ሰዓት"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar"] = func() locales.Translator {
		return &localen{

			locale:                 "ar",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "د.إ.\u200f", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "د.ب.\u200f", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "د.ج.\u200f", "ECS", "ECV", "EEK", "ج.م.\u200f", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "UK£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "د.ع.\u200f", "ر.إ.", "ISJ", "ISK", "ITL", "JMD", "د.أ.\u200f", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "د.ك.\u200f", "KYD", "KZT", "LAK", "ل.ل.\u200f", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "د.ل.\u200f", "د.م.\u200f", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "أ.م.", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "ر.ع.\u200f", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "ر.ق.\u200f", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "ر.س.\u200f", "SBD", "SCR", "د.س.\u200f", "ج.س.", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "ل.س.\u200f", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "د.ت.\u200f", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "***", "YDD", "ر.ي.\u200f", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_001"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_001",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_ae"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_AE",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "$", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_bh"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_BH",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_dj"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_DJ",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "Fdj", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_dz"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_DZ",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "جانفي", "فيفري", "أفريل", "ماي", "جوان", "جويلية", "أوت"},
			monthsNarrow:           []string{"", "ج", "م", "ج", "ج", "أ", "أ", "ن"},
			monthsWide:             []string{"", "جانفي", "فيفري", "أفريل", "ماي", "جوان", "جويلية", "أوت"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_fb7ade3a9509df0c54be19484c433a01,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_eg"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_EG",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_eh"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_EH",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت الأمازون الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_er"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_ER",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "Nfk", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "MDT", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "MST", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_il"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_IL",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_iq"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_IQ",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين\u00a0الأول", "تشرين الثاني", "كانون الأول"},
			monthsNarrow:           []string{"", "ك", "ش", "آ", "ن", "أ", "ح", "ت", "آ", "أ", "ت", "ت", "ك"},
			monthsWide:             []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_jo"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_JO",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			monthsNarrow:           []string{"", "ك", "ش", "آ", "ن", "أ", "ح", "ت", "آ", "أ", "ت", "ت", "ك"},
			monthsWide:             []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_km"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_KM",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "CF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_kw"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_KW",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_lb"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_LB",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			monthsNarrow:           []string{"", "ك", "ش", "آ", "ن", "أ", "ح", "ت", "آ", "أ", "ت", "ت", "ك"},
			monthsWide:             []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_fb7ade3a9509df0c54be19484c433a01,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_ly"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_LY",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_fb7ade3a9509df0c54be19484c433a01,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_ma"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_MA",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "أبريل", "ماي", "يوليوز", "غشت", "شتنبر", "نونبر", "دجنبر"},
			monthsNarrow:           []string{"", "أ", "م", "ن", "ل", "ش", "ك", "ب", "د"},
			monthsWide:             []string{"", "أبريل", "ماي", "يوليوز", "غشت", "شتنبر", "نونبر", "دجنبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_fb7ade3a9509df0c54be19484c433a01,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_mr"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_MR",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "إبريل", "مايو", "يونيو", "يوليو", "أغشت", "شتمبر", "أكتوبر", "نوفمبر", "دجمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "إ", "و", "ن", "ل", "غ", "ش", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "إبريل", "مايو", "يونيو", "يوليو", "أغشت", "شتمبر", "أكتوبر", "نوفمبر", "دجمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_fb7ade3a9509df0c54be19484c433a01,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_om"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_OM",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت الأمازون الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_ps"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_PS",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			monthsNarrow:           []string{"", "ك", "ش", "آ", "ن", "أ", "ح", "ت", "آ", "أ", "ت", "ت", "ك"},
			monthsWide:             []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_qa"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_QA",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_sa"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_SA",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_f074e0c8dff3d1b13a6f552a71cedb58,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_sd"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_SD",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_so"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_SO",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "S", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_f074e0c8dff3d1b13a6f552a71cedb58,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_ss"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_SS",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "£", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_sy"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_SY",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			monthsNarrow:           []string{"", "ك", "ش", "آ", "ن", "أ", "ح", "ت", "آ", "أ", "ت", "ت", "ك"},
			monthsWide:             []string{"", "كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_td"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_TD",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت أزورس الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_tn"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_TN",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "جانفي", "فيفري", "أفريل", "ماي", "جوان", "جويلية", "أوت"},
			monthsNarrow:           []string{"", "ج", "م", "ج", "ج", "أ", "أ", "ن"},
			monthsWide:             []string{"", "جانفي", "فيفري", "أفريل", "ماي", "جوان", "جويلية", "أوت"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "MDT", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "MST", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت بيرو الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5455f8fcf723c4be2a5a03fd41298f2,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_fb7ade3a9509df0c54be19484c433a01,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ar_ye"] = func() locales.Translator {
		return &localen{

			locale:                 "ar_YE",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{1, 4, 5, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "؜-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			monthsNarrow:           []string{"", "ي", "ف", "م", "أ", "و", "ن", "ل", "غ", "س", "ك", "ب", "د"},
			monthsWide:             []string{"", "يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"},
			daysAbbreviated:        []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			daysNarrow:             []string{"ح", "ن", "ث", "ر", "خ", "ج", "س"},
			daysShort:              []string{"أحد", "إثنين", "ثلاثاء", "أربعاء", "خميس", "جمعة", "سبت"},
			daysWide:               []string{"الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"},
			periodsAbbreviated:     []string{"ص", "م"},
			periodsNarrow:          []string{"ص", "م"},
			periodsWide:            []string{"ص", "م"},
			erasAbbreviated:        []string{"ق.م", "م"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل الميلاد", "ميلادي"},
			timezones:              map[string]string{"ACDT": "توقيت وسط أستراليا الصيفي", "ACST": "توقيت وسط أستراليا الرسمي", "ACWDT": "توقيت غرب وسط أستراليا الصيفي", "ACWST": "توقيت غرب وسط أستراليا الرسمي", "ADT": "التوقيت الصيفي الأطلسي", "AEDT": "توقيت شرق أستراليا الصيفي", "AEST": "توقيت شرق أستراليا الرسمي", "AKDT": "توقيت ألاسكا الصيفي", "AKST": "التوقيت الرسمي لألاسكا", "ARST": "توقيت الأرجنتين الصيفي", "ART": "توقيت الأرجنتين الرسمي", "AST": "التوقيت الرسمي الأطلسي", "AWDT": "توقيت غرب أستراليا الصيفي", "AWST": "توقيت غرب أستراليا الرسمي", "BOT": "توقيت بوليفيا", "BT": "توقيت بوتان", "CAT": "توقيت وسط أفريقيا", "CDT": "التوقيت الصيفي المركزي لأمريكا الشمالية", "CHADT": "توقيت تشاتام الصيفي", "CHAST": "توقيت تشاتام الرسمي", "CLST": "توقيت تشيلي الصيفي", "CLT": "توقيت تشيلي الرسمي", "COST": "توقيت كولومبيا الصيفي", "COT": "توقيت كولومبيا الرسمي", "CST": "التوقيت الرسمي المركزي لأمريكا الشمالية", "ChST": "توقيت تشامورو", "EAT": "توقيت شرق أفريقيا", "ECT": "توقيت الإكوادور", "EDT": "التوقيت الصيفي الشرقي لأمريكا الشمالية", "EST": "التوقيت الرسمي الشرقي لأمريكا الشمالية", "GFT": "توقيت غويانا الفرنسية", "GMT": "توقيت غرينتش", "GST": "توقيت الخليج", "GYT": "توقيت غيانا", "HADT": "توقيت هاواي ألوتيان الصيفي", "HAST": "توقيت هاواي ألوتيان الرسمي", "HAT": "توقيت نيوفاوندلاند الصيفي", "HECU": "توقيت كوبا الصيفي", "HEEG": "توقيت شرق غرينلاند الصيفي", "HENOMX": "التوقيت الصيفي لشمال غرب المكسيك", "HEOG": "توقيت غرب غرينلاند الصيفي", "HEPM": "توقيت سانت بيير وميكولون الصيفي", "HEPMX": "توقيت المحيط الهادي الصيفي للمكسيك", "HKST": "توقيت هونغ كونغ الصيفي", "HKT": "توقيت هونغ كونغ الرسمي", "HNCU": "توقيت كوبا الرسمي", "HNEG": "توقيت شرق غرينلاند الرسمي", "HNNOMX": "التوقيت الرسمي لشمال غرب المكسيك", "HNOG": "توقيت غرب غرينلاند الرسمي", "HNPM": "توقيت سانت بيير وميكولون الرسمي", "HNPMX": "توقيت المحيط الهادي الرسمي للمكسيك", "HNT": "توقيت نيوفاوندلاند الرسمي", "IST": "توقيت الهند", "JDT": "توقيت اليابان الصيفي", "JST": "توقيت اليابان الرسمي", "LHDT": "التوقيت الصيفي للورد هاو", "LHST": "توقيت لورد هاو الرسمي", "MDT": "التوقيت الجبلي الصيفي لأمريكا الشمالية", "MESZ": "توقيت وسط أوروبا الصيفي", "MEZ": "توقيت وسط أوروبا الرسمي", "MST": "التوقيت الجبلي الرسمي لأمريكا الشمالية", "MYT": "توقيت ماليزيا", "NZDT": "توقيت نيوزيلندا الصيفي", "NZST": "توقيت نيوزيلندا الرسمي", "OESZ": "توقيت شرق أوروبا الصيفي", "OEZ": "توقيت شرق أوروبا الرسمي", "PDT": "توقيت المحيط الهادي الصيفي", "PST": "توقيت المحيط الهادي الرسمي", "SAST": "توقيت جنوب أفريقيا", "SGT": "توقيت سنغافورة", "SRT": "توقيت سورينام", "TMST": "توقيت تركمانستان الصيفي", "TMT": "توقيت تركمانستان الرسمي", "UYST": "توقيت أوروغواي الصيفي", "UYT": "توقيت أوروغواي الرسمي", "VET": "توقيت فنزويلا", "WARST": "توقيت غرب الأرجنتين الصيفي", "WART": "توقيت غرب الأرجنتين الرسمي", "WAST": "توقيت غرب أفريقيا الصيفي", "WAT": "توقيت غرب أفريقيا الرسمي", "WESZ": "توقيت غرب أوروبا الصيفي", "WEZ": "توقيت غرب أوروبا الرسمي", "WIB": "توقيت غرب إندونيسيا", "WIT": "توقيت شرق إندونيسيا", "WITA": "توقيت وسط إندونيسيا", "∅∅∅": "توقيت برازيليا الصيفي"},

			fnCardinalPluralRule:  fnCardinalPluralRule_f27514af189883ea7cd186156edfe998,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f74088692afba81682b73af3d3e7c19a,
			fnFmtCurrency:         fnFmtCurrency_e28d302f22cf4ed4b5009e290f74e359,
			fnFmtDateFull:         fnFmtDateFull_2d3b405215099e683568cb9c052cbd54,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_a55320da438aec0cfbc31bc4d718abc1,
			fnFmtDateShort:        fnFmtDateShort_05c85041b5f9b6fa224b9c4ba9be3590,
			fnFmtNumber:           fnFmtNumber_eef52ddc8fa71bd12171ac9ef3cd3328,
			fnFmtPercent:          fnFmtPercent_60ca2e9d27b8160a5fec0f6694129759,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_47500e2a682fd3a64da69f06f7b93c0a,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["as"] = func() locales.Translator {
		return &localen{

			locale:                 "as",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "জানু", "ফেব্ৰু", "মাৰ্চ", "এপ্ৰিল", "মে’", "জুন", "জুলাই", "আগ", "ছেপ্তে", "অক্টো", "নৱে", "ডিচে"},
			monthsNarrow:           []string{"", "জ", "ফ", "ম", "এ", "ম", "জ", "জ", "আ", "ছ", "অ", "ন", "ড"},
			monthsWide:             []string{"", "জানুৱাৰী", "ফেব্ৰুৱাৰী", "মাৰ্চ", "এপ্ৰিল", "মে’", "জুন", "জুলাই", "আগষ্ট", "ছেপ্তেম্বৰ", "অক্টোবৰ", "নৱেম্বৰ", "ডিচেম্বৰ"},
			daysAbbreviated:        []string{"দেও", "সোম", "মঙ্গল", "বুধ", "বৃহ", "শুক্ৰ", "শনি"},
			daysNarrow:             []string{"দ", "স", "ম", "ব", "ব", "শ", "শ"},
			daysShort:              []string{"দেও", "সোম", "মঙ্গল", "বুধ", "বৃহ", "শুক্ৰ", "শনি"},
			daysWide:               []string{"দেওবাৰ", "সোমবাৰ", "মঙ্গলবাৰ", "বুধবাৰ", "বৃহস্পতিবাৰ", "শুক্ৰবাৰ", "শনিবাৰ"},
			periodsAbbreviated:     []string{"পূৰ্বাহ্ন", "অপৰাহ্ন"},
			periodsNarrow:          []string{"পূৰ্বাহ্ন", "অপৰাহ্ন"},
			periodsWide:            []string{"পূৰ্বাহ্ন", "অপৰাহ্ন"},
			erasAbbreviated:        []string{"খ্ৰীঃ পূঃ", "খ্ৰীঃ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"খ্ৰীষ্টপূৰ্ব", "খ্ৰীষ্টাব্দ"},
			timezones:              map[string]string{"ACDT": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় ডেলাইট সময়", "ACST": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় মান সময়", "ACWDT": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় পাশ্চাত্য ডেলাইট সময়", "ACWST": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় পাশ্চাত্য মান সময়", "ADT": "আটলাণ্টিক ডেলাইট সময়", "AEDT": "অষ্ট্ৰেলিয়াৰ প্ৰাচ্য ডেলাইট সময়", "AEST": "অষ্ট্ৰেলিয়াৰ প্ৰাচ্য মান সময়", "AKDT": "আলাস্কাৰ ডেলাইট সময়", "AKST": "আলাস্কাৰ মান সময়", "ARST": "আৰ্জেণ্টিনাৰ গ্ৰীষ্মকালীন সময়", "ART": "আৰ্জেণ্টিনাৰ মান সময়", "AST": "আটলাণ্টিক মান সময়", "AWDT": "অষ্ট্ৰেলিয়াৰ পাশ্চাত্য ডেলাইট সময়", "AWST": "অষ্ট্ৰেলিয়াৰ পাশ্চাত্য মান সময়", "BOT": "বলিভিয়াৰ সময়", "BT": "ভুটানৰ সময়", "CAT": "মধ্য আফ্ৰিকাৰ সময়", "CDT": "উত্তৰ আমেৰিকাৰ কেন্দ্ৰীয় ডেলাইট সময়", "CHADT": "চাথামৰ ডেলাইট সময়", "CHAST": "চাথামৰ মান সময়", "CLST": "চিলিৰ গ্ৰীষ্মকালীন সময়", "CLT": "চিলিৰ মান সময়", "COST": "কলম্বিয়াৰ গ্ৰীষ্মকালীন সময়", "COT": "কলম্বিয়াৰ মান সময়", "CST": "উত্তৰ আমেৰিকাৰ কেন্দ্ৰীয় মান সময়", "ChST": "চামোৰোৰ মান সময়", "EAT": "পূব আফ্ৰিকাৰ সময়", "ECT": "ইকুৱেডৰৰ সময়", "EDT": "উত্তৰ আমেৰিকাৰ প্ৰাচ্য ডেলাইট সময়", "EST": "উত্তৰ আমেৰিকাৰ প্ৰাচ্য মান সময়", "GFT": "ফ্ৰান্স গয়ানাৰ সময়", "GMT": "গ্ৰীণউইচ মান সময়", "GST": "উপসাগৰীয় মান সময়", "GYT": "গায়ানাৰ সময়", "HADT": "হাৱাই-এলিউশ্বনৰ ডেলাইট সময়", "HAST": "হাৱাই-এলিউশ্বনৰ মান সময়", "HAT": "নিউফাউণ্ডলেণ্ডৰ ডেলাইট সময়", "HECU": "কিউবাৰ ডেলাইট সময়", "HEEG": "পূব গ্ৰীণলেণ্ডৰ গ্ৰীষ্মকালীন সময়", "HENOMX": "উত্তৰ-পশ্চিম মেক্সিকোৰ ডেলাইট সময়", "HEOG": "পশ্চিম গ্ৰীণলেণ্ডৰ গ্ৰীষ্মকালীন সময়", "HEPM": "ছেইণ্ট পিয়েৰে আৰু মিকিউৱেলনৰ ডেলাইট সময়", "HEPMX": "মেক্সিকোৰ প্ৰশান্ত ডেলাইট সময়", "HKST": "হং কঙৰ গ্ৰীষ্মকালীন সময়", "HKT": "হং কঙৰ মান সময়", "HNCU": "কিউবাৰ মান সময়", "HNEG": "পূব গ্ৰীণলেণ্ডৰ মান সময়", "HNNOMX": "উত্তৰ-পশ্চিম মেক্সিকোৰ মান সময়", "HNOG": "পশ্চিম গ্ৰীণলেণ্ডৰ মান সময়", "HNPM": "ছেইণ্ট পিয়েৰে আৰু মিকিউৱেলনৰ মান সময়", "HNPMX": "মেক্সিকোৰ প্ৰশান্ত মান সময়", "HNT": "নিউফাউণ্ডলেণ্ডৰ মান সময়", "IST": "ভাৰতীয় মান সময়", "JDT": "জাপানৰ ডেলাইট সময়", "JST": "জাপানৰ মান সময়", "LHDT": "লৰ্ড হাওৰ ডেলাইট সময়", "LHST": "লৰ্ড হাওৰ মান সময়", "MDT": "উত্তৰ আমেৰিকাৰ পৰ্ব্বতীয় ডেলাইট সময়", "MESZ": "মধ্য ইউৰোপীয় গ্ৰীষ্মকালীন সময়", "MEZ": "মধ্য ইউৰোপীয় মান সময়", "MST": "উত্তৰ আমেৰিকাৰ পৰ্ব্বতীয় মান সময়", "MYT": "মালয়েচিয়াৰ সময়", "NZDT": "নিউজিলেণ্ডৰ ডেলাইট সময়", "NZST": "নিউজিলেণ্ডৰ মান সময়", "OESZ": "প্ৰাচ্য ইউৰোপীয় গ্ৰীষ্মকালীন সময়", "OEZ": "প্ৰাচ্য ইউৰোপীয় মান সময়", "PDT": "উত্তৰ আমেৰিকাৰ ডেলাইট সময়", "PST": "উত্তৰ আমেৰিকাৰ প্ৰশান্ত মান সময়", "SAST": "দক্ষিণ আফ্ৰিকাৰ মান সময়", "SGT": "ছিংগাপুৰৰ মান সময়", "SRT": "ছুৰিনামৰ সময়", "TMST": "তুৰ্কমেনিস্তানৰ গ্ৰীষ্মকালীন সময়", "TMT": "তুৰ্কমেনিস্তানৰ মান সময়", "UYST": "উৰুগুৱেৰ গ্ৰীষ্মকালীন সময়", "UYT": "উৰুগুৱেৰ মান সময়", "VET": "ভেনিজুৱেলাৰ সময়", "WARST": "পাশ্চাত্য আৰ্জেণ্টিনাৰ গ্ৰীষ্মকালীন সময়", "WART": "পাশ্চাত্য আৰ্জেণ্টিনাৰ মান সময়", "WAST": "পশ্চিম আফ্ৰিকাৰ গ্ৰীষ্মকালীন সময়", "WAT": "পশ্চিম আফ্ৰিকাৰ মান সময়", "WESZ": "পাশ্চাত্য ইউৰোপীয় গ্ৰীষ্মকালীন সময়", "WEZ": "পাশ্চাত্য ইউৰোপীয় মান সময়", "WIB": "পাশ্চাত্য ইণ্ডোনেচিয়াৰ সময়", "WIT": "প্ৰাচ্য ইণ্ডোনেচিয়াৰ সময়", "WITA": "মধ্য ইণ্ডোনেচিয়াৰ সময়", "∅∅∅": "ব্ৰাজিলিয়াৰ গ্ৰীষ্মকালীন সময়"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cd503abfb0c9bd938bcf20e65858f8d7,
			fnFmtCurrency:         fnFmtCurrency_9f465d9454a1702118c7b32dfb89b3d1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_48f6d56925de0117a6fe7b1ec232f585,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_3ce7dc5b6b225559fd40d102900d6c81,
			fnFmtTimeLong:         fnFmtTimeLong_b2fdf6411542ab907630622b3d42a6b0,
			fnFmtTimeMedium:       fnFmtTimeMedium_74593f8f5ae9c1a3e4637628d509c179,
			fnFmtTimeShort:        fnFmtTimeShort_6c9fb1bcda2b27013e9a8a47c7326805,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_2d950970290e57f43f401dd367ad94dc,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["as_in"] = func() locales.Translator {
		return &localen{

			locale:                 "as_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "জানু", "ফেব্ৰু", "মাৰ্চ", "এপ্ৰিল", "মে’", "জুন", "জুলাই", "আগ", "ছেপ্তে", "অক্টো", "নৱে", "ডিচে"},
			monthsNarrow:           []string{"", "জ", "ফ", "ম", "এ", "ম", "জ", "জ", "আ", "ছ", "অ", "ন", "ড"},
			monthsWide:             []string{"", "জানুৱাৰী", "ফেব্ৰুৱাৰী", "মাৰ্চ", "এপ্ৰিল", "মে’", "জুন", "জুলাই", "আগষ্ট", "ছেপ্তেম্বৰ", "অক্টোবৰ", "নৱেম্বৰ", "ডিচেম্বৰ"},
			daysAbbreviated:        []string{"দেও", "সোম", "মঙ্গল", "বুধ", "বৃহ", "শুক্ৰ", "শনি"},
			daysNarrow:             []string{"দ", "স", "ম", "ব", "ব", "শ", "শ"},
			daysShort:              []string{"দেও", "সোম", "মঙ্গল", "বুধ", "বৃহ", "শুক্ৰ", "শনি"},
			daysWide:               []string{"দেওবাৰ", "সোমবাৰ", "মঙ্গলবাৰ", "বুধবাৰ", "বৃহস্পতিবাৰ", "শুক্ৰবাৰ", "শনিবাৰ"},
			periodsAbbreviated:     []string{"পূৰ্বাহ্ন", "অপৰাহ্ন"},
			periodsNarrow:          []string{"পূৰ্বাহ্ন", "অপৰাহ্ন"},
			periodsWide:            []string{"পূৰ্বাহ্ন", "অপৰাহ্ন"},
			erasAbbreviated:        []string{"খ্ৰীঃ পূঃ", "খ্ৰীঃ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"খ্ৰীষ্টপূৰ্ব", "খ্ৰীষ্টাব্দ"},
			timezones:              map[string]string{"ACDT": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় ডেলাইট সময়", "ACST": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় মান সময়", "ACWDT": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় পাশ্চাত্য ডেলাইট সময়", "ACWST": "অষ্ট্ৰেলিয়াৰ কেন্দ্ৰীয় পাশ্চাত্য মান সময়", "ADT": "আটলাণ্টিক ডেলাইট সময়", "AEDT": "অষ্ট্ৰেলিয়াৰ প্ৰাচ্য ডেলাইট সময়", "AEST": "অষ্ট্ৰেলিয়াৰ প্ৰাচ্য মান সময়", "AKDT": "আলাস্কাৰ ডেলাইট সময়", "AKST": "আলাস্কাৰ মান সময়", "ARST": "আৰ্জেণ্টিনাৰ গ্ৰীষ্মকালীন সময়", "ART": "আৰ্জেণ্টিনাৰ মান সময়", "AST": "আটলাণ্টিক মান সময়", "AWDT": "অষ্ট্ৰেলিয়াৰ পাশ্চাত্য ডেলাইট সময়", "AWST": "অষ্ট্ৰেলিয়াৰ পাশ্চাত্য মান সময়", "BOT": "বলিভিয়াৰ সময়", "BT": "ভুটানৰ সময়", "CAT": "মধ্য আফ্ৰিকাৰ সময়", "CDT": "উত্তৰ আমেৰিকাৰ কেন্দ্ৰীয় ডেলাইট সময়", "CHADT": "চাথামৰ ডেলাইট সময়", "CHAST": "চাথামৰ মান সময়", "CLST": "চিলিৰ গ্ৰীষ্মকালীন সময়", "CLT": "চিলিৰ মান সময়", "COST": "কলম্বিয়াৰ গ্ৰীষ্মকালীন সময়", "COT": "কলম্বিয়াৰ মান সময়", "CST": "উত্তৰ আমেৰিকাৰ কেন্দ্ৰীয় মান সময়", "ChST": "চামোৰোৰ মান সময়", "EAT": "পূব আফ্ৰিকাৰ সময়", "ECT": "ইকুৱেডৰৰ সময়", "EDT": "উত্তৰ আমেৰিকাৰ প্ৰাচ্য ডেলাইট সময়", "EST": "উত্তৰ আমেৰিকাৰ প্ৰাচ্য মান সময়", "GFT": "ফ্ৰান্স গয়ানাৰ সময়", "GMT": "গ্ৰীণউইচ মান সময়", "GST": "উপসাগৰীয় মান সময়", "GYT": "গায়ানাৰ সময়", "HADT": "হাৱাই-এলিউশ্বনৰ ডেলাইট সময়", "HAST": "হাৱাই-এলিউশ্বনৰ মান সময়", "HAT": "নিউফাউণ্ডলেণ্ডৰ ডেলাইট সময়", "HECU": "কিউবাৰ ডেলাইট সময়", "HEEG": "পূব গ্ৰীণলেণ্ডৰ গ্ৰীষ্মকালীন সময়", "HENOMX": "উত্তৰ-পশ্চিম মেক্সিকোৰ ডেলাইট সময়", "HEOG": "পশ্চিম গ্ৰীণলেণ্ডৰ গ্ৰীষ্মকালীন সময়", "HEPM": "ছেইণ্ট পিয়েৰে আৰু মিকিউৱেলনৰ ডেলাইট সময়", "HEPMX": "মেক্সিকোৰ প্ৰশান্ত ডেলাইট সময়", "HKST": "হং কঙৰ গ্ৰীষ্মকালীন সময়", "HKT": "হং কঙৰ মান সময়", "HNCU": "কিউবাৰ মান সময়", "HNEG": "পূব গ্ৰীণলেণ্ডৰ মান সময়", "HNNOMX": "উত্তৰ-পশ্চিম মেক্সিকোৰ মান সময়", "HNOG": "পশ্চিম গ্ৰীণলেণ্ডৰ মান সময়", "HNPM": "ছেইণ্ট পিয়েৰে আৰু মিকিউৱেলনৰ মান সময়", "HNPMX": "মেক্সিকোৰ প্ৰশান্ত মান সময়", "HNT": "নিউফাউণ্ডলেণ্ডৰ মান সময়", "IST": "ভাৰতীয় মান সময়", "JDT": "জাপানৰ ডেলাইট সময়", "JST": "জাপানৰ মান সময়", "LHDT": "লৰ্ড হাওৰ ডেলাইট সময়", "LHST": "লৰ্ড হাওৰ মান সময়", "MDT": "উত্তৰ আমেৰিকাৰ পৰ্ব্বতীয় ডেলাইট সময়", "MESZ": "মধ্য ইউৰোপীয় গ্ৰীষ্মকালীন সময়", "MEZ": "মধ্য ইউৰোপীয় মান সময়", "MST": "উত্তৰ আমেৰিকাৰ পৰ্ব্বতীয় মান সময়", "MYT": "মালয়েচিয়াৰ সময়", "NZDT": "নিউজিলেণ্ডৰ ডেলাইট সময়", "NZST": "নিউজিলেণ্ডৰ মান সময়", "OESZ": "প্ৰাচ্য ইউৰোপীয় গ্ৰীষ্মকালীন সময়", "OEZ": "প্ৰাচ্য ইউৰোপীয় মান সময়", "PDT": "উত্তৰ আমেৰিকাৰ ডেলাইট সময়", "PST": "উত্তৰ আমেৰিকাৰ প্ৰশান্ত মান সময়", "SAST": "দক্ষিণ আফ্ৰিকাৰ মান সময়", "SGT": "ছিংগাপুৰৰ মান সময়", "SRT": "ছুৰিনামৰ সময়", "TMST": "তুৰ্কমেনিস্তানৰ গ্ৰীষ্মকালীন সময়", "TMT": "তুৰ্কমেনিস্তানৰ মান সময়", "UYST": "উৰুগুৱেৰ গ্ৰীষ্মকালীন সময়", "UYT": "উৰুগুৱেৰ মান সময়", "VET": "ভেনিজুৱেলাৰ সময়", "WARST": "পাশ্চাত্য আৰ্জেণ্টিনাৰ গ্ৰীষ্মকালীন সময়", "WART": "পাশ্চাত্য আৰ্জেণ্টিনাৰ মান সময়", "WAST": "পশ্চিম আফ্ৰিকাৰ গ্ৰীষ্মকালীন সময়", "WAT": "পশ্চিম আফ্ৰিকাৰ মান সময়", "WESZ": "পাশ্চাত্য ইউৰোপীয় গ্ৰীষ্মকালীন সময়", "WEZ": "পাশ্চাত্য ইউৰোপীয় মান সময়", "WIB": "পাশ্চাত্য ইণ্ডোনেচিয়াৰ সময়", "WIT": "প্ৰাচ্য ইণ্ডোনেচিয়াৰ সময়", "WITA": "মধ্য ইণ্ডোনেচিয়াৰ সময়", "∅∅∅": "আজোৰেছৰ গ্ৰীষ্মকালীন সময়"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cd503abfb0c9bd938bcf20e65858f8d7,
			fnFmtCurrency:         fnFmtCurrency_9f465d9454a1702118c7b32dfb89b3d1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_48f6d56925de0117a6fe7b1ec232f585,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_3ce7dc5b6b225559fd40d102900d6c81,
			fnFmtTimeLong:         fnFmtTimeLong_b2fdf6411542ab907630622b3d42a6b0,
			fnFmtTimeMedium:       fnFmtTimeMedium_74593f8f5ae9c1a3e4637628d509c179,
			fnFmtTimeShort:        fnFmtTimeShort_6c9fb1bcda2b27013e9a8a47c7326805,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_2d950970290e57f43f401dd367ad94dc,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["asa"] = func() locales.Translator {
		return &localen{

			locale:                 "asa",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Ijm", "Jmo"},
			daysNarrow:             []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"icheheavo", "ichamthi"},
			periodsWide:            []string{"icheheavo", "ichamthi"},
			erasAbbreviated:        []string{"KM", "BM"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla yakwe Yethu", "Baada yakwe Yethu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["asa_tz"] = func() locales.Translator {
		return &localen{

			locale:                 "asa_TZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Ijm", "Jmo"},
			daysNarrow:             []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"icheheavo", "ichamthi"},
			periodsWide:            []string{"icheheavo", "ichamthi"},
			erasAbbreviated:        []string{"KM", "BM"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla yakwe Yethu", "Baada yakwe Yethu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ast"] = func() locales.Translator {
		return &localen{

			locale:                 "ast",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "xin", "feb", "mar", "abr", "may", "xun", "xnt", "ago", "set", "och", "pay", "avi"},
			monthsNarrow:           []string{"", "X", "F", "M", "A", "M", "X", "X", "A", "S", "O", "P", "A"},
			monthsWide:             []string{"", "de xineru", "de febreru", "de marzu", "d’abril", "de mayu", "de xunu", "de xunetu", "d’agostu", "de setiembre", "d’ochobre", "de payares", "d’avientu"},
			daysAbbreviated:        []string{"dom", "llu", "mar", "mié", "xue", "vie", "sáb"},
			daysNarrow:             []string{"D", "L", "M", "M", "X", "V", "S"},
			daysShort:              []string{"do", "ll", "ma", "mi", "xu", "vi", "sá"},
			daysWide:               []string{"domingu", "llunes", "martes", "miércoles", "xueves", "vienres", "sábadu"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"de la mañana", "de la tarde"},
			erasAbbreviated:        []string{"e.C.", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"enantes de Cristu", "después de Cristu"},
			timezones:              map[string]string{"ACDT": "Hora braniega d’Australia central", "ACST": "Hora estándar d’Australia central", "ACWDT": "Hora braniega d’Australia central del oeste", "ACWST": "Hora estándar d’Australia central del oeste", "ADT": "Hora braniega del Atlánticu", "AEDT": "Hora braniega d’Australia del este", "AEST": "Hora estándar d’Australia del este", "AKDT": "Hora braniega d’Alaska", "AKST": "Hora estándar d’Alaska", "ARST": "Hora braniega d’Arxentina", "ART": "Hora estándar d’Arxentina", "AST": "Hora estándar del Atlánticu", "AWDT": "Hora braniega d’Australia del oeste", "AWST": "Hora estándar d’Australia del oeste", "BOT": "Hora de Bolivia", "BT": "Hora de Bután", "CAT": "Hora d’África central", "CDT": "Hora braniega central norteamericana", "CHADT": "Hora braniega de Chatham", "CHAST": "Hora estándar de Chatham", "CLST": "Hora braniega de Chile", "CLT": "Hora estándar de Chile", "COST": "Hora braniega de Colombia", "COT": "Hora estándar de Colombia", "CST": "Hora estándar central norteamericana", "ChST": "Hora estándar de Chamorro", "EAT": "Hora d’África del este", "ECT": "Hora d’Ecuador", "EDT": "Hora braniega del este norteamericanu", "EST": "Hora estándar del este norteamericanu", "GFT": "Hora de La Guyana Francesa", "GMT": "Hora media de Greenwich", "GST": "Hora estándar del Golfu", "GYT": "Hora de La Guyana", "HADT": "Hora braniega de Hawaii-Aleutianes", "HAST": "Hora estándar de Hawaii-Aleutianes", "HAT": "Hora braniega de Newfoundland", "HECU": "Hora braniega de Cuba", "HEEG": "Hora braniega de Groenlandia oriental", "HENOMX": "Hora braniega del noroeste de Méxicu", "HEOG": "Hora braniega de Groenlandia occidental", "HEPM": "Hora braniega de Saint Pierre y Miquelon", "HEPMX": "Hora braniega del Pacíficu de Méxicu", "HKST": "Hora braniega de Ḥong Kong", "HKT": "Hora estándar de Ḥong Kong", "HNCU": "Hora estándar de Cuba", "HNEG": "Hora estándar de Groenlandia oriental", "HNNOMX": "Hora estándar del noroeste de Méxicu", "HNOG": "Hora estándar de Groenlandia occidental", "HNPM": "Hora estándar de Saint Pierre y Miquelon", "HNPMX": "Hora estándar del Pacíficu de Méxicu", "HNT": "Hora estándar de Newfoundland", "IST": "Hora estándar de la India", "JDT": "Hora braniega de Xapón", "JST": "Hora estándar de Xapón", "LHDT": "Hora braniega de Lord Howe", "LHST": "Hora estándar de Lord Howe", "MDT": "Hora braniega de les montañes norteamericanes", "MESZ": "Hora braniega d’Europa Central", "MEZ": "Hora estándar d’Europa Central", "MST": "Hora estándar de les montañes norteamericanes", "MYT": "Hora de Malasia", "NZDT": "Hora braniega de Nueva Zelanda", "NZST": "Hora estándar de Nueva Zelanda", "OESZ": "Hora braniega d’Europa del Este", "OEZ": "Hora estándar d’Europa del Este", "PDT": "Hora braniega del Pacíficu norteamericanu", "PST": "Hora estándar del Pacíficu norteamericanu", "SAST": "Hora de Sudáfrica", "SGT": "Hora estándar de Singapur", "SRT": "Hora del Surinam", "TMST": "Hora braniega del Turkmenistán", "TMT": "Hora estándar del Turkmenistán", "UYST": "Hora braniega del Uruguái", "UYT": "Hora estándar del Uruguái", "VET": "Hora de Venezuela", "WARST": "Hora braniega occidental d’Arxentina", "WART": "Hora estándar occidental d’Arxentina", "WAST": "Hora braniega d’África del oeste", "WAT": "Hora estándar d’África del oeste", "WESZ": "Hora braniega d’Europa Occidental", "WEZ": "Hora estándar d’Europa Occidental", "WIB": "Hora d’Indonesia del oeste", "WIT": "Hora d’Indonesia del este", "WITA": "Hora d’Indonesia central", "∅∅∅": "Hora braniega de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ast_es"] = func() locales.Translator {
		return &localen{

			locale:                 "ast_ES",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "xin", "feb", "mar", "abr", "may", "xun", "xnt", "ago", "set", "och", "pay", "avi"},
			monthsNarrow:           []string{"", "X", "F", "M", "A", "M", "X", "X", "A", "S", "O", "P", "A"},
			monthsWide:             []string{"", "de xineru", "de febreru", "de marzu", "d’abril", "de mayu", "de xunu", "de xunetu", "d’agostu", "de setiembre", "d’ochobre", "de payares", "d’avientu"},
			daysAbbreviated:        []string{"dom", "llu", "mar", "mié", "xue", "vie", "sáb"},
			daysNarrow:             []string{"D", "L", "M", "M", "X", "V", "S"},
			daysShort:              []string{"do", "ll", "ma", "mi", "xu", "vi", "sá"},
			daysWide:               []string{"domingu", "llunes", "martes", "miércoles", "xueves", "vienres", "sábadu"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"de la mañana", "de la tarde"},
			erasAbbreviated:        []string{"e.C.", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"enantes de Cristu", "después de Cristu"},
			timezones:              map[string]string{"ACDT": "Hora braniega d’Australia central", "ACST": "Hora estándar d’Australia central", "ACWDT": "Hora braniega d’Australia central del oeste", "ACWST": "Hora estándar d’Australia central del oeste", "ADT": "Hora braniega del Atlánticu", "AEDT": "Hora braniega d’Australia del este", "AEST": "Hora estándar d’Australia del este", "AKDT": "Hora braniega d’Alaska", "AKST": "Hora estándar d’Alaska", "ARST": "Hora braniega d’Arxentina", "ART": "Hora estándar d’Arxentina", "AST": "Hora estándar del Atlánticu", "AWDT": "Hora braniega d’Australia del oeste", "AWST": "Hora estándar d’Australia del oeste", "BOT": "Hora de Bolivia", "BT": "Hora de Bután", "CAT": "Hora d’África central", "CDT": "Hora braniega central norteamericana", "CHADT": "Hora braniega de Chatham", "CHAST": "Hora estándar de Chatham", "CLST": "Hora braniega de Chile", "CLT": "Hora estándar de Chile", "COST": "Hora braniega de Colombia", "COT": "Hora estándar de Colombia", "CST": "Hora estándar central norteamericana", "ChST": "Hora estándar de Chamorro", "EAT": "Hora d’África del este", "ECT": "Hora d’Ecuador", "EDT": "Hora braniega del este norteamericanu", "EST": "Hora estándar del este norteamericanu", "GFT": "Hora de La Guyana Francesa", "GMT": "Hora media de Greenwich", "GST": "Hora estándar del Golfu", "GYT": "Hora de La Guyana", "HADT": "Hora braniega de Hawaii-Aleutianes", "HAST": "Hora estándar de Hawaii-Aleutianes", "HAT": "Hora braniega de Newfoundland", "HECU": "Hora braniega de Cuba", "HEEG": "Hora braniega de Groenlandia oriental", "HENOMX": "Hora braniega del noroeste de Méxicu", "HEOG": "Hora braniega de Groenlandia occidental", "HEPM": "Hora braniega de Saint Pierre y Miquelon", "HEPMX": "Hora braniega del Pacíficu de Méxicu", "HKST": "Hora braniega de Ḥong Kong", "HKT": "Hora estándar de Ḥong Kong", "HNCU": "Hora estándar de Cuba", "HNEG": "Hora estándar de Groenlandia oriental", "HNNOMX": "Hora estándar del noroeste de Méxicu", "HNOG": "Hora estándar de Groenlandia occidental", "HNPM": "Hora estándar de Saint Pierre y Miquelon", "HNPMX": "Hora estándar del Pacíficu de Méxicu", "HNT": "Hora estándar de Newfoundland", "IST": "Hora estándar de la India", "JDT": "Hora braniega de Xapón", "JST": "Hora estándar de Xapón", "LHDT": "Hora braniega de Lord Howe", "LHST": "Hora estándar de Lord Howe", "MDT": "Hora braniega de les montañes norteamericanes", "MESZ": "Hora braniega d’Europa Central", "MEZ": "Hora estándar d’Europa Central", "MST": "Hora estándar de les montañes norteamericanes", "MYT": "Hora de Malasia", "NZDT": "Hora braniega de Nueva Zelanda", "NZST": "Hora estándar de Nueva Zelanda", "OESZ": "Hora braniega d’Europa del Este", "OEZ": "Hora estándar d’Europa del Este", "PDT": "Hora braniega del Pacíficu norteamericanu", "PST": "Hora estándar del Pacíficu norteamericanu", "SAST": "Hora de Sudáfrica", "SGT": "Hora estándar de Singapur", "SRT": "Hora del Surinam", "TMST": "Hora braniega del Turkmenistán", "TMT": "Hora estándar del Turkmenistán", "UYST": "Hora braniega del Uruguái", "UYT": "Hora estándar del Uruguái", "VET": "Hora de Venezuela", "WARST": "Hora braniega occidental d’Arxentina", "WART": "Hora estándar occidental d’Arxentina", "WAST": "Hora braniega d’África del oeste", "WAT": "Hora estándar d’África del oeste", "WESZ": "Hora braniega d’Europa Occidental", "WEZ": "Hora estándar d’Europa Occidental", "WIB": "Hora d’Indonesia del oeste", "WIT": "Hora d’Indonesia del este", "WITA": "Hora d’Indonesia central", "∅∅∅": "Hora braniega de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["az"] = func() locales.Translator {
		return &localen{

			locale:                 "az",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "₼", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avq", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"B.", "B.e.", "Ç.a.", "Ç.", "C.a.", "C.", "Ş."},
			daysNarrow:             []string{"7", "1", "2", "3", "4", "5", "6"},
			daysShort:              []string{"B.", "B.E.", "Ç.A.", "Ç.", "C.A.", "C.", "Ş."},
			daysWide:               []string{"bazar", "bazar ertəsi", "çərşənbə axşamı", "çərşənbə", "cümə axşamı", "cümə", "şənbə"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"e.ə.", "y.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"eramızdan əvvəl", "yeni era"},
			timezones:              map[string]string{"ACDT": "Mərkəzi Avstraliya Yay Vaxtı", "ACST": "Mərkəzi Avstraliya Standart Vaxtı", "ACWDT": "Mərkəzi Qərbi Avstraliya Yay Vaxtı", "ACWST": "Mərkəzi Qərbi Avstraliya Standart Vaxtı", "ADT": "Atlantik Yay Vaxtı", "AEDT": "Şərqi Avstraliya Yay Vaxtı", "AEST": "Şərqi Avstraliya Standart Vaxtı", "AKDT": "Alyaska Yay Vaxtı", "AKST": "Alyaska Standart Vaxtı", "ARST": "Argentina Yay Vaxtı", "ART": "Argentina Standart Vaxtı", "AST": "Atlantik Standart Vaxt", "AWDT": "Qərbi Avstraliya Yay Vaxtı", "AWST": "Qərbi Avstraliya Standart Vaxtı", "BOT": "Boliviya Vaxtı", "BT": "Butan Vaxtı", "CAT": "Mərkəzi Afrika Vaxtı", "CDT": "Şimali Mərkəzi Amerika Yay Vaxtı", "CHADT": "Çatham Yay Vaxtı", "CHAST": "Çatham Standart Vaxtı", "CLST": "Çili Yay Vaxtı", "CLT": "Çili Standart Vaxtı", "COST": "Kolumbiya Yay Vaxtı", "COT": "Kolumbiya Standart Vaxtı", "CST": "Şimali Mərkəzi Amerika Standart Vaxtı", "ChST": "Çamorro Vaxtı", "EAT": "Şərqi Afrika Vaxtı", "ECT": "Ekvador Vaxtı", "EDT": "Şimali Şərqi Amerika Yay Vaxtı", "EST": "Şimali Şərqi Amerika Standart Vaxtı", "GFT": "Fransız Qvianası Vaxtı", "GMT": "Qrinviç Orta Vaxtı", "GST": "Körfəz Vaxtı", "GYT": "Qayana Vaxtı", "HADT": "Havay-Aleut Yay Vaxtı", "HAST": "Havay-Aleut Standart Vaxtı", "HAT": "Nyufaundlend Yay Vaxtı", "HECU": "Kuba Yay Vaxtı", "HEEG": "Şərqi Qrenlandiya Yay Vaxtı", "HENOMX": "Şimal-Qərbi Meksika Yay Vaxtı", "HEOG": "Qərbi Qrenlandiya Yay Vaxtı", "HEPM": "Müqəddəs Pyer və Mikelon Yay Vaxtı", "HEPMX": "Meksika Sakit Okean Yay Vaxtı", "HKST": "Honq Konq Yay Vaxtı", "HKT": "Honq Konq Standart Vaxtı", "HNCU": "Kuba Standart Vaxtı", "HNEG": "Şərqi Qrenlandiya Standart Vaxtı", "HNNOMX": "Şimal-Qərbi Meksika Standart Vaxtı", "HNOG": "Qərbi Qrenlandiya Standart Vaxtı", "HNPM": "Müqəddəs Pyer və Mikelon Standart Vaxtı", "HNPMX": "Meksika Sakit Okean Standart Vaxtı", "HNT": "Nyufaundlend Standart Vaxtı", "IST": "Hindistan Vaxtı", "JDT": "Yaponiya Yay Vaxtı", "JST": "Yaponiya Standart Vaxtı", "LHDT": "Lord Hau Yay vaxtı", "LHST": "Lord Hau Standart Vaxtı", "MDT": "Şimali Dağlıq Amerika Yay Vaxtı", "MESZ": "Mərkəzi Avropa Yay Vaxtı", "MEZ": "Mərkəzi Avropa Standart Vaxtı", "MST": "Şimali Dağlıq Amerika Standart Vaxtı", "MYT": "Malayziya Vaxtı", "NZDT": "Yeni Zelandiya Yay Vaxtı", "NZST": "Yeni Zelandiya Standart Vaxtı", "OESZ": "Şərqi Avropa Yay Vaxtı", "OEZ": "Şərqi Avropa Standart Vaxtı", "PDT": "Şimali Amerika Sakit Okean Yay Vaxtı", "PST": "Şimali Amerika Sakit Okean Standart Vaxtı", "SAST": "Cənubi Afrika Vaxtı", "SGT": "Sinqapur Vaxtı", "SRT": "Surinam Vaxtı", "TMST": "Türkmənistan Yay Vaxtı", "TMT": "Türkmənistan Standart Vaxtı", "UYST": "Uruqvay Yay Vaxtı", "UYT": "Uruqvay Standart Vaxtı", "VET": "Venesuela Vaxtı", "WARST": "Qərbi Argentina Yay Vaxtı", "WART": "Qərbi Argentina Standart Vaxtı", "WAST": "Qərbi Afrika Yay Vaxtı", "WAT": "Qərbi Afrika Standart Vaxtı", "WESZ": "Qərbi Avropa Yay Vaxtı", "WEZ": "Qərbi Avropa Standart Vaxtı", "WIB": "Qərbi İndoneziya Vaxtı", "WIT": "Şərqi İndoneziya Vaxtı", "WITA": "Mərkəzi İndoneziya Vaxtı", "∅∅∅": "Braziliya Yay Vaxtı"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_b94af095d9cfcd967654d0fbd68f2e98,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_822e96e7fb113a3156c2e7c532855968,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["az_cyrl"] = func() locales.Translator {
		return &localen{

			locale:                 "az_Cyrl",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "₼", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "јан", "фев", "мар", "апр", "май", "ијн", "ијл", "авг", "сен", "окт", "ној", "дек"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "јанвар", "феврал", "март", "апрел", "май", "ијун", "ијул", "август", "сентјабр", "октјабр", "нојабр", "декабр"},
			daysAbbreviated:        []string{"Б.", "Б.Е.", "Ч.А.", "Ч.", "Ҹ.А.", "Ҹ.", "Ш."},
			daysNarrow:             []string{"7", "1", "2", "3", "4", "5", "6"},
			daysShort:              []string{"Б.", "Б.Е.", "Ч.А.", "Ч.", "Ҹ.А.", "Ҹ.", "Ш."},
			daysWide:               []string{"базар", "базар ертәси", "чәршәнбә ахшамы", "чәршәнбә", "ҹүмә ахшамы", "ҹүмә", "шәнбә"},
			periodsAbbreviated:     []string{"АМ", "ПМ"},
			periodsNarrow:          []string{"а", "п"},
			periodsWide:            []string{"АМ", "ПМ"},
			erasAbbreviated:        []string{"е.ә.", "ј.е."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ерамыздан әввәл", "јени ера"},
			timezones:              map[string]string{"ACDT": "Mərkəzi Avstraliya Yay Vaxtı", "ACST": "Mərkəzi Avstraliya Standart Vaxtı", "ACWDT": "Mərkəzi Qərbi Avstraliya Yay Vaxtı", "ACWST": "Mərkəzi Qərbi Avstraliya Standart Vaxtı", "ADT": "Atlantik Yay Vaxtı", "AEDT": "Şərqi Avstraliya Yay Vaxtı", "AEST": "Şərqi Avstraliya Standart Vaxtı", "AKDT": "Alyaska Yay Vaxtı", "AKST": "Alyaska Standart Vaxtı", "ARST": "Argentina Yay Vaxtı", "ART": "Argentina Standart Vaxtı", "AST": "Atlantik Standart Vaxt", "AWDT": "Qərbi Avstraliya Yay Vaxtı", "AWST": "Qərbi Avstraliya Standart Vaxtı", "BOT": "Boliviya Vaxtı", "BT": "Butan Vaxtı", "CAT": "Mərkəzi Afrika Vaxtı", "CDT": "Şimali Mərkəzi Amerika Yay Vaxtı", "CHADT": "Çatham Yay Vaxtı", "CHAST": "Çatham Standart Vaxtı", "CLST": "Çili Yay Vaxtı", "CLT": "Çili Standart Vaxtı", "COST": "Kolumbiya Yay Vaxtı", "COT": "Kolumbiya Standart Vaxtı", "CST": "Şimali Mərkəzi Amerika Standart Vaxtı", "ChST": "Çamorro Vaxtı", "EAT": "Şərqi Afrika Vaxtı", "ECT": "Ekvador Vaxtı", "EDT": "Şimali Şərqi Amerika Yay Vaxtı", "EST": "Şimali Şərqi Amerika Standart Vaxtı", "GFT": "Fransız Qvianası Vaxtı", "GMT": "Qrinviç Orta Vaxtı", "GST": "Körfəz Vaxtı", "GYT": "Qayana Vaxtı", "HADT": "Havay-Aleut Yay Vaxtı", "HAST": "Havay-Aleut Standart Vaxtı", "HAT": "Nyufaundlend Yay Vaxtı", "HECU": "Kuba Yay Vaxtı", "HEEG": "Şərqi Qrenlandiya Yay Vaxtı", "HENOMX": "Şimal-Qərbi Meksika Yay Vaxtı", "HEOG": "Qərbi Qrenlandiya Yay Vaxtı", "HEPM": "Müqəddəs Pyer və Mikelon Yay Vaxtı", "HEPMX": "Meksika Sakit Okean Yay Vaxtı", "HKST": "Honq Konq Yay Vaxtı", "HKT": "Honq Konq Standart Vaxtı", "HNCU": "Kuba Standart Vaxtı", "HNEG": "Şərqi Qrenlandiya Standart Vaxtı", "HNNOMX": "Şimal-Qərbi Meksika Standart Vaxtı", "HNOG": "Qərbi Qrenlandiya Standart Vaxtı", "HNPM": "Müqəddəs Pyer və Mikelon Standart Vaxtı", "HNPMX": "Meksika Sakit Okean Standart Vaxtı", "HNT": "Nyufaundlend Standart Vaxtı", "IST": "Hindistan Vaxtı", "JDT": "Yaponiya Yay Vaxtı", "JST": "Yaponiya Standart Vaxtı", "LHDT": "Lord Hau Yay vaxtı", "LHST": "Lord Hau Standart Vaxtı", "MDT": "Şimali Dağlıq Amerika Yay Vaxtı", "MESZ": "Mərkəzi Avropa Yay Vaxtı", "MEZ": "Mərkəzi Avropa Standart Vaxtı", "MST": "Şimali Dağlıq Amerika Standart Vaxtı", "MYT": "Malayziya Vaxtı", "NZDT": "Yeni Zelandiya Yay Vaxtı", "NZST": "Yeni Zelandiya Standart Vaxtı", "OESZ": "Şərqi Avropa Yay Vaxtı", "OEZ": "Şərqi Avropa Standart Vaxtı", "PDT": "Şimali Amerika Sakit Okean Yay Vaxtı", "PST": "Şimali Amerika Sakit Okean Standart Vaxtı", "SAST": "Cənubi Afrika Vaxtı", "SGT": "Sinqapur Vaxtı", "SRT": "Surinam Vaxtı", "TMST": "Türkmənistan Yay Vaxtı", "TMT": "Türkmənistan Standart Vaxtı", "UYST": "Uruqvay Yay Vaxtı", "UYT": "Uruqvay Standart Vaxtı", "VET": "Venesuela Vaxtı", "WARST": "Qərbi Argentina Yay Vaxtı", "WART": "Qərbi Argentina Standart Vaxtı", "WAST": "Qərbi Afrika Yay Vaxtı", "WAT": "Qərbi Afrika Standart Vaxtı", "WESZ": "Qərbi Avropa Yay Vaxtı", "WEZ": "Qərbi Avropa Standart Vaxtı", "WIB": "Qərbi İndoneziya Vaxtı", "WIT": "Şərqi İndoneziya Vaxtı", "WITA": "Mərkəzi İndoneziya Vaxtı", "∅∅∅": "Azor Yay Vaxtı"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_b94af095d9cfcd967654d0fbd68f2e98,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_822e96e7fb113a3156c2e7c532855968,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["az_cyrl_az"] = func() locales.Translator {
		return &localen{

			locale:                 "az_Cyrl_AZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avq", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"B.", "B.e.", "Ç.a.", "Ç.", "C.a.", "C.", "Ş."},
			daysNarrow:             []string{"7", "1", "2", "3", "4", "5", "6"},
			daysShort:              []string{"B.", "B.E.", "Ç.A.", "Ç.", "C.A.", "C.", "Ş."},
			daysWide:               []string{"bazar", "bazar ertəsi", "çərşənbə axşamı", "çərşənbə", "cümə axşamı", "cümə", "şənbə"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"e.ə.", "y.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"eramızdan əvvəl", "yeni era"},
			timezones:              map[string]string{"ACDT": "Mərkəzi Avstraliya Yay Vaxtı", "ACST": "Mərkəzi Avstraliya Standart Vaxtı", "ACWDT": "Mərkəzi Qərbi Avstraliya Yay Vaxtı", "ACWST": "Mərkəzi Qərbi Avstraliya Standart Vaxtı", "ADT": "Atlantik Yay Vaxtı", "AEDT": "Şərqi Avstraliya Yay Vaxtı", "AEST": "Şərqi Avstraliya Standart Vaxtı", "AKDT": "Alyaska Yay Vaxtı", "AKST": "Alyaska Standart Vaxtı", "ARST": "Argentina Yay Vaxtı", "ART": "Argentina Standart Vaxtı", "AST": "Atlantik Standart Vaxt", "AWDT": "Qərbi Avstraliya Yay Vaxtı", "AWST": "Qərbi Avstraliya Standart Vaxtı", "BOT": "Boliviya Vaxtı", "BT": "Butan Vaxtı", "CAT": "Mərkəzi Afrika Vaxtı", "CDT": "Şimali Mərkəzi Amerika Yay Vaxtı", "CHADT": "Çatham Yay Vaxtı", "CHAST": "Çatham Standart Vaxtı", "CLST": "Çili Yay Vaxtı", "CLT": "Çili Standart Vaxtı", "COST": "Kolumbiya Yay Vaxtı", "COT": "Kolumbiya Standart Vaxtı", "CST": "Şimali Mərkəzi Amerika Standart Vaxtı", "ChST": "Çamorro Vaxtı", "EAT": "Şərqi Afrika Vaxtı", "ECT": "Ekvador Vaxtı", "EDT": "Şimali Şərqi Amerika Yay Vaxtı", "EST": "Şimali Şərqi Amerika Standart Vaxtı", "GFT": "Fransız Qvianası Vaxtı", "GMT": "Qrinviç Orta Vaxtı", "GST": "Körfəz Vaxtı", "GYT": "Qayana Vaxtı", "HADT": "Havay-Aleut Yay Vaxtı", "HAST": "Havay-Aleut Standart Vaxtı", "HAT": "Nyufaundlend Yay Vaxtı", "HECU": "Kuba Yay Vaxtı", "HEEG": "Şərqi Qrenlandiya Yay Vaxtı", "HENOMX": "Şimal-Qərbi Meksika Yay Vaxtı", "HEOG": "Qərbi Qrenlandiya Yay Vaxtı", "HEPM": "Müqəddəs Pyer və Mikelon Yay Vaxtı", "HEPMX": "Meksika Sakit Okean Yay Vaxtı", "HKST": "Honq Konq Yay Vaxtı", "HKT": "Honq Konq Standart Vaxtı", "HNCU": "Kuba Standart Vaxtı", "HNEG": "Şərqi Qrenlandiya Standart Vaxtı", "HNNOMX": "Şimal-Qərbi Meksika Standart Vaxtı", "HNOG": "Qərbi Qrenlandiya Standart Vaxtı", "HNPM": "Müqəddəs Pyer və Mikelon Standart Vaxtı", "HNPMX": "Meksika Sakit Okean Standart Vaxtı", "HNT": "Nyufaundlend Standart Vaxtı", "IST": "Hindistan Vaxtı", "JDT": "Yaponiya Yay Vaxtı", "JST": "Yaponiya Standart Vaxtı", "LHDT": "Lord Hau Yay vaxtı", "LHST": "Lord Hau Standart Vaxtı", "MDT": "Şimali Dağlıq Amerika Yay Vaxtı", "MESZ": "Mərkəzi Avropa Yay Vaxtı", "MEZ": "Mərkəzi Avropa Standart Vaxtı", "MST": "Şimali Dağlıq Amerika Standart Vaxtı", "MYT": "Malayziya Vaxtı", "NZDT": "Yeni Zelandiya Yay Vaxtı", "NZST": "Yeni Zelandiya Standart Vaxtı", "OESZ": "Şərqi Avropa Yay Vaxtı", "OEZ": "Şərqi Avropa Standart Vaxtı", "PDT": "Şimali Amerika Sakit Okean Yay Vaxtı", "PST": "Şimali Amerika Sakit Okean Standart Vaxtı", "SAST": "Cənubi Afrika Vaxtı", "SGT": "Sinqapur Vaxtı", "SRT": "Surinam Vaxtı", "TMST": "Türkmənistan Yay Vaxtı", "TMT": "Türkmənistan Standart Vaxtı", "UYST": "Uruqvay Yay Vaxtı", "UYT": "Uruqvay Standart Vaxtı", "VET": "Venesuela Vaxtı", "WARST": "Qərbi Argentina Yay Vaxtı", "WART": "Qərbi Argentina Standart Vaxtı", "WAST": "Qərbi Afrika Yay Vaxtı", "WAT": "Qərbi Afrika Standart Vaxtı", "WESZ": "Qərbi Avropa Yay Vaxtı", "WEZ": "Qərbi Avropa Standart Vaxtı", "WIB": "Qərbi İndoneziya Vaxtı", "WIT": "Şərqi İndoneziya Vaxtı", "WITA": "Mərkəzi İndoneziya Vaxtı", "∅∅∅": "Azor Yay Vaxtı"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_b94af095d9cfcd967654d0fbd68f2e98,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_822e96e7fb113a3156c2e7c532855968,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["az_latn"] = func() locales.Translator {
		return &localen{

			locale:                 "az_Latn",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avq", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"B.", "B.e.", "Ç.a.", "Ç.", "C.a.", "C.", "Ş."},
			daysNarrow:             []string{"7", "1", "2", "3", "4", "5", "6"},
			daysShort:              []string{"B.", "B.E.", "Ç.A.", "Ç.", "C.A.", "C.", "Ş."},
			daysWide:               []string{"bazar", "bazar ertəsi", "çərşənbə axşamı", "çərşənbə", "cümə axşamı", "cümə", "şənbə"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"e.ə.", "y.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"eramızdan əvvəl", "yeni era"},
			timezones:              map[string]string{"ACDT": "Mərkəzi Avstraliya Yay Vaxtı", "ACST": "Mərkəzi Avstraliya Standart Vaxtı", "ACWDT": "Mərkəzi Qərbi Avstraliya Yay Vaxtı", "ACWST": "Mərkəzi Qərbi Avstraliya Standart Vaxtı", "ADT": "Atlantik Yay Vaxtı", "AEDT": "Şərqi Avstraliya Yay Vaxtı", "AEST": "Şərqi Avstraliya Standart Vaxtı", "AKDT": "Alyaska Yay Vaxtı", "AKST": "Alyaska Standart Vaxtı", "ARST": "Argentina Yay Vaxtı", "ART": "Argentina Standart Vaxtı", "AST": "Atlantik Standart Vaxt", "AWDT": "Qərbi Avstraliya Yay Vaxtı", "AWST": "Qərbi Avstraliya Standart Vaxtı", "BOT": "Boliviya Vaxtı", "BT": "Butan Vaxtı", "CAT": "Mərkəzi Afrika Vaxtı", "CDT": "Şimali Mərkəzi Amerika Yay Vaxtı", "CHADT": "Çatham Yay Vaxtı", "CHAST": "Çatham Standart Vaxtı", "CLST": "Çili Yay Vaxtı", "CLT": "Çili Standart Vaxtı", "COST": "Kolumbiya Yay Vaxtı", "COT": "Kolumbiya Standart Vaxtı", "CST": "Şimali Mərkəzi Amerika Standart Vaxtı", "ChST": "Çamorro Vaxtı", "EAT": "Şərqi Afrika Vaxtı", "ECT": "Ekvador Vaxtı", "EDT": "Şimali Şərqi Amerika Yay Vaxtı", "EST": "Şimali Şərqi Amerika Standart Vaxtı", "GFT": "Fransız Qvianası Vaxtı", "GMT": "Qrinviç Orta Vaxtı", "GST": "Körfəz Vaxtı", "GYT": "Qayana Vaxtı", "HADT": "Havay-Aleut Yay Vaxtı", "HAST": "Havay-Aleut Standart Vaxtı", "HAT": "Nyufaundlend Yay Vaxtı", "HECU": "Kuba Yay Vaxtı", "HEEG": "Şərqi Qrenlandiya Yay Vaxtı", "HENOMX": "Şimal-Qərbi Meksika Yay Vaxtı", "HEOG": "Qərbi Qrenlandiya Yay Vaxtı", "HEPM": "Müqəddəs Pyer və Mikelon Yay Vaxtı", "HEPMX": "Meksika Sakit Okean Yay Vaxtı", "HKST": "Honq Konq Yay Vaxtı", "HKT": "Honq Konq Standart Vaxtı", "HNCU": "Kuba Standart Vaxtı", "HNEG": "Şərqi Qrenlandiya Standart Vaxtı", "HNNOMX": "Şimal-Qərbi Meksika Standart Vaxtı", "HNOG": "Qərbi Qrenlandiya Standart Vaxtı", "HNPM": "Müqəddəs Pyer və Mikelon Standart Vaxtı", "HNPMX": "Meksika Sakit Okean Standart Vaxtı", "HNT": "Nyufaundlend Standart Vaxtı", "IST": "Hindistan Vaxtı", "JDT": "Yaponiya Yay Vaxtı", "JST": "Yaponiya Standart Vaxtı", "LHDT": "Lord Hau Yay vaxtı", "LHST": "Lord Hau Standart Vaxtı", "MDT": "Şimali Dağlıq Amerika Yay Vaxtı", "MESZ": "Mərkəzi Avropa Yay Vaxtı", "MEZ": "Mərkəzi Avropa Standart Vaxtı", "MST": "Şimali Dağlıq Amerika Standart Vaxtı", "MYT": "Malayziya Vaxtı", "NZDT": "Yeni Zelandiya Yay Vaxtı", "NZST": "Yeni Zelandiya Standart Vaxtı", "OESZ": "Şərqi Avropa Yay Vaxtı", "OEZ": "Şərqi Avropa Standart Vaxtı", "PDT": "Şimali Amerika Sakit Okean Yay Vaxtı", "PST": "Şimali Amerika Sakit Okean Standart Vaxtı", "SAST": "Cənubi Afrika Vaxtı", "SGT": "Sinqapur Vaxtı", "SRT": "Surinam Vaxtı", "TMST": "Türkmənistan Yay Vaxtı", "TMT": "Türkmənistan Standart Vaxtı", "UYST": "Uruqvay Yay Vaxtı", "UYT": "Uruqvay Standart Vaxtı", "VET": "Venesuela Vaxtı", "WARST": "Qərbi Argentina Yay Vaxtı", "WART": "Qərbi Argentina Standart Vaxtı", "WAST": "Qərbi Afrika Yay Vaxtı", "WAT": "Qərbi Afrika Standart Vaxtı", "WESZ": "Qərbi Avropa Yay Vaxtı", "WEZ": "Qərbi Avropa Standart Vaxtı", "WIB": "Qərbi İndoneziya Vaxtı", "WIT": "Şərqi İndoneziya Vaxtı", "WITA": "Mərkəzi İndoneziya Vaxtı", "∅∅∅": "Peru Yay Vaxtı"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_b94af095d9cfcd967654d0fbd68f2e98,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_822e96e7fb113a3156c2e7c532855968,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["az_latn_az"] = func() locales.Translator {
		return &localen{

			locale:                 "az_Latn_AZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avq", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"B.", "B.e.", "Ç.a.", "Ç.", "C.a.", "C.", "Ş."},
			daysNarrow:             []string{"7", "1", "2", "3", "4", "5", "6"},
			daysShort:              []string{"B.", "B.E.", "Ç.A.", "Ç.", "C.A.", "C.", "Ş."},
			daysWide:               []string{"bazar", "bazar ertəsi", "çərşənbə axşamı", "çərşənbə", "cümə axşamı", "cümə", "şənbə"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"e.ə.", "y.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"eramızdan əvvəl", "yeni era"},
			timezones:              map[string]string{"ACDT": "Mərkəzi Avstraliya Yay Vaxtı", "ACST": "Mərkəzi Avstraliya Standart Vaxtı", "ACWDT": "Mərkəzi Qərbi Avstraliya Yay Vaxtı", "ACWST": "Mərkəzi Qərbi Avstraliya Standart Vaxtı", "ADT": "Atlantik Yay Vaxtı", "AEDT": "Şərqi Avstraliya Yay Vaxtı", "AEST": "Şərqi Avstraliya Standart Vaxtı", "AKDT": "Alyaska Yay Vaxtı", "AKST": "Alyaska Standart Vaxtı", "ARST": "Argentina Yay Vaxtı", "ART": "Argentina Standart Vaxtı", "AST": "Atlantik Standart Vaxt", "AWDT": "Qərbi Avstraliya Yay Vaxtı", "AWST": "Qərbi Avstraliya Standart Vaxtı", "BOT": "Boliviya Vaxtı", "BT": "Butan Vaxtı", "CAT": "Mərkəzi Afrika Vaxtı", "CDT": "Şimali Mərkəzi Amerika Yay Vaxtı", "CHADT": "Çatham Yay Vaxtı", "CHAST": "Çatham Standart Vaxtı", "CLST": "Çili Yay Vaxtı", "CLT": "Çili Standart Vaxtı", "COST": "Kolumbiya Yay Vaxtı", "COT": "Kolumbiya Standart Vaxtı", "CST": "Şimali Mərkəzi Amerika Standart Vaxtı", "ChST": "Çamorro Vaxtı", "EAT": "Şərqi Afrika Vaxtı", "ECT": "Ekvador Vaxtı", "EDT": "Şimali Şərqi Amerika Yay Vaxtı", "EST": "Şimali Şərqi Amerika Standart Vaxtı", "GFT": "Fransız Qvianası Vaxtı", "GMT": "Qrinviç Orta Vaxtı", "GST": "Körfəz Vaxtı", "GYT": "Qayana Vaxtı", "HADT": "Havay-Aleut Yay Vaxtı", "HAST": "Havay-Aleut Standart Vaxtı", "HAT": "Nyufaundlend Yay Vaxtı", "HECU": "Kuba Yay Vaxtı", "HEEG": "Şərqi Qrenlandiya Yay Vaxtı", "HENOMX": "Şimal-Qərbi Meksika Yay Vaxtı", "HEOG": "Qərbi Qrenlandiya Yay Vaxtı", "HEPM": "Müqəddəs Pyer və Mikelon Yay Vaxtı", "HEPMX": "Meksika Sakit Okean Yay Vaxtı", "HKST": "Honq Konq Yay Vaxtı", "HKT": "Honq Konq Standart Vaxtı", "HNCU": "Kuba Standart Vaxtı", "HNEG": "Şərqi Qrenlandiya Standart Vaxtı", "HNNOMX": "Şimal-Qərbi Meksika Standart Vaxtı", "HNOG": "Qərbi Qrenlandiya Standart Vaxtı", "HNPM": "Müqəddəs Pyer və Mikelon Standart Vaxtı", "HNPMX": "Meksika Sakit Okean Standart Vaxtı", "HNT": "Nyufaundlend Standart Vaxtı", "IST": "Hindistan Vaxtı", "JDT": "Yaponiya Yay Vaxtı", "JST": "Yaponiya Standart Vaxtı", "LHDT": "Lord Hau Yay vaxtı", "LHST": "Lord Hau Standart Vaxtı", "MDT": "Şimali Dağlıq Amerika Yay Vaxtı", "MESZ": "Mərkəzi Avropa Yay Vaxtı", "MEZ": "Mərkəzi Avropa Standart Vaxtı", "MST": "Şimali Dağlıq Amerika Standart Vaxtı", "MYT": "Malayziya Vaxtı", "NZDT": "Yeni Zelandiya Yay Vaxtı", "NZST": "Yeni Zelandiya Standart Vaxtı", "OESZ": "Şərqi Avropa Yay Vaxtı", "OEZ": "Şərqi Avropa Standart Vaxtı", "PDT": "Şimali Amerika Sakit Okean Yay Vaxtı", "PST": "Şimali Amerika Sakit Okean Standart Vaxtı", "SAST": "Cənubi Afrika Vaxtı", "SGT": "Sinqapur Vaxtı", "SRT": "Surinam Vaxtı", "TMST": "Türkmənistan Yay Vaxtı", "TMT": "Türkmənistan Standart Vaxtı", "UYST": "Uruqvay Yay Vaxtı", "UYT": "Uruqvay Standart Vaxtı", "VET": "Venesuela Vaxtı", "WARST": "Qərbi Argentina Yay Vaxtı", "WART": "Qərbi Argentina Standart Vaxtı", "WAST": "Qərbi Afrika Yay Vaxtı", "WAT": "Qərbi Afrika Standart Vaxtı", "WESZ": "Qərbi Avropa Yay Vaxtı", "WEZ": "Qərbi Avropa Standart Vaxtı", "WIB": "Qərbi İndoneziya Vaxtı", "WIT": "Şərqi İndoneziya Vaxtı", "WITA": "Mərkəzi İndoneziya Vaxtı", "∅∅∅": "Peru Yay Vaxtı"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_b94af095d9cfcd967654d0fbd68f2e98,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_822e96e7fb113a3156c2e7c532855968,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bas"] = func() locales.Translator {
		return &localen{

			locale:                 "bas",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "kɔn", "mac", "mat", "mto", "mpu", "hil", "nje", "hik", "dip", "bio", "may", "liɓ"},
			monthsNarrow:           []string{"", "k", "m", "m", "m", "m", "h", "n", "h", "d", "b", "m", "l"},
			monthsWide:             []string{"", "Kɔndɔŋ", "Màcɛ̂l", "Màtùmb", "Màtop", "M̀puyɛ", "Hìlòndɛ̀", "Njèbà", "Hìkaŋ", "Dìpɔ̀s", "Bìòôm", "Màyɛsèp", "Lìbuy li ńyèe"},
			daysAbbreviated:        []string{"nɔy", "nja", "uum", "ŋge", "mbɔ", "kɔɔ", "jon"},
			daysNarrow:             []string{"n", "n", "u", "ŋ", "m", "k", "j"},
			daysWide:               []string{"ŋgwà nɔ̂y", "ŋgwà njaŋgumba", "ŋgwà ûm", "ŋgwà ŋgê", "ŋgwà mbɔk", "ŋgwà kɔɔ", "ŋgwà jôn"},
			periodsAbbreviated:     []string{"I bikɛ̂glà", "I ɓugajɔp"},
			periodsWide:            []string{"I bikɛ̂glà", "I ɓugajɔp"},
			erasAbbreviated:        []string{"b.Y.K", "m.Y.K"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"bisū bi Yesù Krǐstò", "i mbūs Yesù Krǐstò"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bas_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "bas_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "kɔn", "mac", "mat", "mto", "mpu", "hil", "nje", "hik", "dip", "bio", "may", "liɓ"},
			monthsNarrow:           []string{"", "k", "m", "m", "m", "m", "h", "n", "h", "d", "b", "m", "l"},
			monthsWide:             []string{"", "Kɔndɔŋ", "Màcɛ̂l", "Màtùmb", "Màtop", "M̀puyɛ", "Hìlòndɛ̀", "Njèbà", "Hìkaŋ", "Dìpɔ̀s", "Bìòôm", "Màyɛsèp", "Lìbuy li ńyèe"},
			daysAbbreviated:        []string{"nɔy", "nja", "uum", "ŋge", "mbɔ", "kɔɔ", "jon"},
			daysNarrow:             []string{"n", "n", "u", "ŋ", "m", "k", "j"},
			daysWide:               []string{"ŋgwà nɔ̂y", "ŋgwà njaŋgumba", "ŋgwà ûm", "ŋgwà ŋgê", "ŋgwà mbɔk", "ŋgwà kɔɔ", "ŋgwà jôn"},
			periodsAbbreviated:     []string{"I bikɛ̂glà", "I ɓugajɔp"},
			periodsWide:            []string{"I bikɛ̂glà", "I ɓugajɔp"},
			erasAbbreviated:        []string{"b.Y.K", "m.Y.K"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"bisū bi Yesù Krǐstò", "i mbūs Yesù Krǐstò"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["be"] = func() locales.Translator {
		return &localen{

			locale:                 "be",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{4, 6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "Br", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "сту", "лют", "сак", "кра", "мая", "чэр", "ліп", "жні", "вер", "кас", "ліс", "сне"},
			monthsNarrow:           []string{"", "с", "л", "с", "к", "м", "ч", "л", "ж", "в", "к", "л", "с"},
			monthsWide:             []string{"", "студзеня", "лютага", "сакавіка", "красавіка", "мая", "чэрвеня", "ліпеня", "жніўня", "верасня", "кастрычніка", "лістапада", "снежня"},
			daysAbbreviated:        []string{"нд", "пн", "аў", "ср", "чц", "пт", "сб"},
			daysNarrow:             []string{"н", "п", "а", "с", "ч", "п", "с"},
			daysShort:              []string{"нд", "пн", "аў", "ср", "чц", "пт", "сб"},
			daysWide:               []string{"нядзеля", "панядзелак", "аўторак", "серада", "чацвер", "пятніца", "субота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"да н.э.", "н.э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"да нараджэння Хрыстова", "ад нараджэння Хрыстова"},
			timezones:              map[string]string{"ACDT": "Летні час цэнтральнай Аўстраліі", "ACST": "Стандартны час цэнтральнай Аўстраліі", "ACWDT": "Летні цэнтральна-заходні час Аўстраліі", "ACWST": "Стандартны цэнтральна-заходні час Аўстраліі", "ADT": "Атлантычны летні час", "AEDT": "Летні час усходняй Аўстраліі", "AEST": "Стандартны час усходняй Аўстраліі", "AKDT": "Летні час Аляскі", "AKST": "Стандартны час Аляскі", "ARST": "Аргенцінскі летні час", "ART": "Аргенцінскі стандартны час", "AST": "Атлантычны стандартны час", "AWDT": "Летні час заходняй Аўстраліі", "AWST": "Стандартны час заходняй Аўстраліі", "BOT": "Балівійскі час", "BT": "Час Бутана", "CAT": "Цэнтральнаафрыканскі час", "CDT": "Паўночнаамерыканскі цэнтральны летні час", "CHADT": "Летні час Чатэма", "CHAST": "Стандартны час Чатэма", "CLST": "Чылійскі летні час", "CLT": "Чылійскі стандартны час", "COST": "Калумбійскі летні час", "COT": "Калумбійскі стандартны час", "CST": "Паўночнаамерыканскі цэнтральны стандартны час", "ChST": "Час Чамора", "EAT": "Усходнеафрыканскі час", "ECT": "Эквадорскі час", "EDT": "Паўночнаамерыканскі ўсходні летні час", "EST": "Паўночнаамерыканскі ўсходні стандартны час", "GFT": "Час Французскай Гвіяны", "GMT": "Час па Грынвічы", "GST": "Час Персідскага заліва", "GYT": "Час Гаяны", "HADT": "Гавайска-Алеуцкі летні час", "HAST": "Гавайска-Алеуцкі стандартны час", "HAT": "Ньюфаўндлендскі летні час", "HECU": "Летні час Кубы", "HEEG": "Летні час Усходняй Грэнландыі", "HENOMX": "Паўночна-заходні мексіканскі летні час", "HEOG": "Летні час Заходняй Грэнландыі", "HEPM": "Стандартны летні час Сен-П’ер і Мікелон", "HEPMX": "Мексіканскі ціхаакіянскі летні час", "HKST": "Летні час Ганконга", "HKT": "Стандартны час Ганконга", "HNCU": "Стандартны час Кубы", "HNEG": "Стандартны час Усходняй Грэнландыі", "HNNOMX": "Паўночна-заходні мексіканскі стандартны час", "HNOG": "Стандартны час Заходняй Грэнландыі", "HNPM": "Стандартны час Сен-П’ер і Мікелон", "HNPMX": "Мексіканскі ціхаакіянскі стандатны час", "HNT": "Ньюфаўндлендскі стандартны час", "IST": "Час Індыі", "JDT": "Летні час Японіі", "JST": "Стандартны час Японіі", "LHDT": "Летні час Лорд-Хау", "LHST": "Стандартны час Лорд-Хау", "MDT": "Паўночнаамерыканскі горны летні час", "MESZ": "Цэнтральнаеўрапейскі летні час", "MEZ": "Цэнтральнаеўрапейскі стандартны час", "MST": "Паўночнаамерыканскі горны стандартны час", "MYT": "Час Малайзіі", "NZDT": "Летні час Новай Зеландыі", "NZST": "Стандартны час Новай Зеландыі", "OESZ": "Усходнееўрапейскі летні час", "OEZ": "Усходнееўрапейскі стандартны час", "PDT": "Ціхаакіянскі летні час", "PST": "Ціхаакіянскі стандартны час", "SAST": "Паўднёваафрыканскі час", "SGT": "Сінгапурскі час", "SRT": "Час Сурынама", "TMST": "Летні час Туркменістана", "TMT": "Стандартны час Туркменістана", "UYST": "Уругвайскі летні час", "UYT": "Уругвайскі стандартны час", "VET": "Венесуэльскі час", "WARST": "Летні час Заходняй Аргенціны", "WART": "Стандартны час Заходняй Аргенціны", "WAST": "Заходнеафрыканскі летні час", "WAT": "Заходнеафрыканскі стандартны час", "WESZ": "Заходнееўрапейскі летні час", "WEZ": "Заходнееўрапейскі стандартны час", "WIB": "Заходнеінданезійскі час", "WIT": "Усходнеінданезійскі час", "WITA": "Цэнтральнаінданезійскі час", "∅∅∅": "Бразільскі летні час"},

			fnCardinalPluralRule:  fnCardinalPluralRule_aa3706fccbfeb0fac3cf81d326e3fe96,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_73098a578f62c763b1a46cfb742142e3,
			fnFmtDateShort:        fnFmtDateShort_cbdb981782261e5a8b8cc19702400c64,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_d302f6108a9d3f4bb461f3a5111143e5,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_62ca292db3807cd1a19e61d6d993a0d9,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["be_by"] = func() locales.Translator {
		return &localen{

			locale:                 "be_BY",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{4, 6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "сту", "лют", "сак", "кра", "мая", "чэр", "ліп", "жні", "вер", "кас", "ліс", "сне"},
			monthsNarrow:           []string{"", "с", "л", "с", "к", "м", "ч", "л", "ж", "в", "к", "л", "с"},
			monthsWide:             []string{"", "студзеня", "лютага", "сакавіка", "красавіка", "мая", "чэрвеня", "ліпеня", "жніўня", "верасня", "кастрычніка", "лістапада", "снежня"},
			daysAbbreviated:        []string{"нд", "пн", "аў", "ср", "чц", "пт", "сб"},
			daysNarrow:             []string{"н", "п", "а", "с", "ч", "п", "с"},
			daysShort:              []string{"нд", "пн", "аў", "ср", "чц", "пт", "сб"},
			daysWide:               []string{"нядзеля", "панядзелак", "аўторак", "серада", "чацвер", "пятніца", "субота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"да н.э.", "н.э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"да нараджэння Хрыстова", "ад нараджэння Хрыстова"},
			timezones:              map[string]string{"ACDT": "Летні час цэнтральнай Аўстраліі", "ACST": "Стандартны час цэнтральнай Аўстраліі", "ACWDT": "Летні цэнтральна-заходні час Аўстраліі", "ACWST": "Стандартны цэнтральна-заходні час Аўстраліі", "ADT": "Атлантычны летні час", "AEDT": "Летні час усходняй Аўстраліі", "AEST": "Стандартны час усходняй Аўстраліі", "AKDT": "Летні час Аляскі", "AKST": "Стандартны час Аляскі", "ARST": "Аргенцінскі летні час", "ART": "Аргенцінскі стандартны час", "AST": "Атлантычны стандартны час", "AWDT": "Летні час заходняй Аўстраліі", "AWST": "Стандартны час заходняй Аўстраліі", "BOT": "Балівійскі час", "BT": "Час Бутана", "CAT": "Цэнтральнаафрыканскі час", "CDT": "Паўночнаамерыканскі цэнтральны летні час", "CHADT": "Летні час Чатэма", "CHAST": "Стандартны час Чатэма", "CLST": "Чылійскі летні час", "CLT": "Чылійскі стандартны час", "COST": "Калумбійскі летні час", "COT": "Калумбійскі стандартны час", "CST": "Паўночнаамерыканскі цэнтральны стандартны час", "ChST": "Час Чамора", "EAT": "Усходнеафрыканскі час", "ECT": "Эквадорскі час", "EDT": "Паўночнаамерыканскі ўсходні летні час", "EST": "Паўночнаамерыканскі ўсходні стандартны час", "GFT": "Час Французскай Гвіяны", "GMT": "Час па Грынвічы", "GST": "Час Персідскага заліва", "GYT": "Час Гаяны", "HADT": "Гавайска-Алеуцкі летні час", "HAST": "Гавайска-Алеуцкі стандартны час", "HAT": "Ньюфаўндлендскі летні час", "HECU": "Летні час Кубы", "HEEG": "Летні час Усходняй Грэнландыі", "HENOMX": "Паўночна-заходні мексіканскі летні час", "HEOG": "Летні час Заходняй Грэнландыі", "HEPM": "Стандартны летні час Сен-П’ер і Мікелон", "HEPMX": "Мексіканскі ціхаакіянскі летні час", "HKST": "Летні час Ганконга", "HKT": "Стандартны час Ганконга", "HNCU": "Стандартны час Кубы", "HNEG": "Стандартны час Усходняй Грэнландыі", "HNNOMX": "Паўночна-заходні мексіканскі стандартны час", "HNOG": "Стандартны час Заходняй Грэнландыі", "HNPM": "Стандартны час Сен-П’ер і Мікелон", "HNPMX": "Мексіканскі ціхаакіянскі стандатны час", "HNT": "Ньюфаўндлендскі стандартны час", "IST": "Час Індыі", "JDT": "Летні час Японіі", "JST": "Стандартны час Японіі", "LHDT": "Летні час Лорд-Хау", "LHST": "Стандартны час Лорд-Хау", "MDT": "Паўночнаамерыканскі горны летні час", "MESZ": "Цэнтральнаеўрапейскі летні час", "MEZ": "Цэнтральнаеўрапейскі стандартны час", "MST": "Паўночнаамерыканскі горны стандартны час", "MYT": "Час Малайзіі", "NZDT": "Летні час Новай Зеландыі", "NZST": "Стандартны час Новай Зеландыі", "OESZ": "Усходнееўрапейскі летні час", "OEZ": "Усходнееўрапейскі стандартны час", "PDT": "Ціхаакіянскі летні час", "PST": "Ціхаакіянскі стандартны час", "SAST": "Паўднёваафрыканскі час", "SGT": "Сінгапурскі час", "SRT": "Час Сурынама", "TMST": "Летні час Туркменістана", "TMT": "Стандартны час Туркменістана", "UYST": "Уругвайскі летні час", "UYT": "Уругвайскі стандартны час", "VET": "Венесуэльскі час", "WARST": "Летні час Заходняй Аргенціны", "WART": "Стандартны час Заходняй Аргенціны", "WAST": "Заходнеафрыканскі летні час", "WAT": "Заходнеафрыканскі стандартны час", "WESZ": "Заходнееўрапейскі летні час", "WEZ": "Заходнееўрапейскі стандартны час", "WIB": "Заходнеінданезійскі час", "WIT": "Усходнеінданезійскі час", "WITA": "Цэнтральнаінданезійскі час", "∅∅∅": "Бразільскі летні час"},

			fnCardinalPluralRule:  fnCardinalPluralRule_aa3706fccbfeb0fac3cf81d326e3fe96,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_73098a578f62c763b1a46cfb742142e3,
			fnFmtDateShort:        fnFmtDateShort_cbdb981782261e5a8b8cc19702400c64,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_d302f6108a9d3f4bb461f3a5111143e5,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_62ca292db3807cd1a19e61d6d993a0d9,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bem"] = func() locales.Translator {
		return &localen{

			locale:                 "bem",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "K", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Epr", "Mei", "Jun", "Jul", "Oga", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "E", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Epreo", "Mei", "Juni", "Julai", "Ogasti", "Septemba", "Oktoba", "Novemba", "Disemba"},
			daysWide:               []string{"Pa Mulungu", "Palichimo", "Palichibuli", "Palichitatu", "Palichine", "Palichisano", "Pachibelushi"},
			periodsAbbreviated:     []string{"uluchelo", "akasuba"},
			periodsWide:            []string{"uluchelo", "akasuba"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Before Yesu", "After Yesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bem_zm"] = func() locales.Translator {
		return &localen{

			locale:                 "bem_ZM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Epr", "Mei", "Jun", "Jul", "Oga", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "E", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Epreo", "Mei", "Juni", "Julai", "Ogasti", "Septemba", "Oktoba", "Novemba", "Disemba"},
			daysWide:               []string{"Pa Mulungu", "Palichimo", "Palichibuli", "Palichitatu", "Palichine", "Palichisano", "Pachibelushi"},
			periodsAbbreviated:     []string{"uluchelo", "akasuba"},
			periodsWide:            []string{"uluchelo", "akasuba"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Before Yesu", "After Yesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bez"] = func() locales.Translator {
		return &localen{

			locale:             "bez",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Hut", "Vil", "Dat", "Tai", "Han", "Sit", "Sab", "Nan", "Tis", "Kum", "Kmj", "Kmb"},
			monthsNarrow:       []string{"", "H", "V", "D", "T", "H", "S", "S", "N", "T", "K", "K", "K"},
			monthsWide:         []string{"", "pa mwedzi gwa hutala", "pa mwedzi gwa wuvili", "pa mwedzi gwa wudatu", "pa mwedzi gwa wutai", "pa mwedzi gwa wuhanu", "pa mwedzi gwa sita", "pa mwedzi gwa saba", "pa mwedzi gwa nane", "pa mwedzi gwa tisa", "pa mwedzi gwa kumi", "pa mwedzi gwa kumi na moja", "pa mwedzi gwa kumi na mbili"},
			daysAbbreviated:    []string{"Mul", "Vil", "Hiv", "Hid", "Hit", "Hih", "Lem"},
			daysNarrow:         []string{"M", "J", "H", "H", "H", "W", "J"},
			daysWide:           []string{"pa mulungu", "pa shahuviluha", "pa hivili", "pa hidatu", "pa hitayi", "pa hihanu", "pa shahulembela"},
			periodsAbbreviated: []string{"pamilau", "pamunyi"},
			periodsWide:        []string{"pamilau", "pamunyi"},
			erasAbbreviated:    []string{"KM", "BM"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Mtwaa", "Baada ya Mtwaa"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bez_tz"] = func() locales.Translator {
		return &localen{

			locale:             "bez_TZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Hut", "Vil", "Dat", "Tai", "Han", "Sit", "Sab", "Nan", "Tis", "Kum", "Kmj", "Kmb"},
			monthsNarrow:       []string{"", "H", "V", "D", "T", "H", "S", "S", "N", "T", "K", "K", "K"},
			monthsWide:         []string{"", "pa mwedzi gwa hutala", "pa mwedzi gwa wuvili", "pa mwedzi gwa wudatu", "pa mwedzi gwa wutai", "pa mwedzi gwa wuhanu", "pa mwedzi gwa sita", "pa mwedzi gwa saba", "pa mwedzi gwa nane", "pa mwedzi gwa tisa", "pa mwedzi gwa kumi", "pa mwedzi gwa kumi na moja", "pa mwedzi gwa kumi na mbili"},
			daysAbbreviated:    []string{"Mul", "Vil", "Hiv", "Hid", "Hit", "Hih", "Lem"},
			daysNarrow:         []string{"M", "J", "H", "H", "H", "W", "J"},
			daysWide:           []string{"pa mulungu", "pa shahuviluha", "pa hivili", "pa hidatu", "pa hitayi", "pa hihanu", "pa shahulembela"},
			periodsAbbreviated: []string{"pamilau", "pamunyi"},
			periodsWide:        []string{"pamilau", "pamunyi"},
			erasAbbreviated:    []string{"KM", "BM"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Mtwaa", "Baada ya Mtwaa"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bg"] = func() locales.Translator {
		return &localen{

			locale:                 "bg",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "лв.", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "щ.д.", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "яну", "фев", "март", "апр", "май", "юни", "юли", "авг", "сеп", "окт", "ное", "дек"},
			monthsNarrow:           []string{"", "я", "ф", "м", "а", "м", "ю", "ю", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "януари", "февруари", "март", "април", "май", "юни", "юли", "август", "септември", "октомври", "ноември", "декември"},
			daysAbbreviated:        []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"н", "п", "в", "с", "ч", "п", "с"},
			daysShort:              []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"неделя", "понеделник", "вторник", "сряда", "четвъртък", "петък", "събота"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"пр.об.", "сл.об."},
			erasAbbreviated:        []string{"пр.Хр.", "сл.Хр."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"преди Христа", "след Христа"},
			timezones:              map[string]string{"ACDT": "Централноавстралийско лятно часово време", "ACST": "Централноавстралийско стандартно време", "ACWDT": "Австралия – западно централно лятно часово време", "ACWST": "Австралия – западно централно стандартно време", "ADT": "Северноамериканско атлантическо лятно часово време", "AEDT": "Източноавстралийско лятно часово време", "AEST": "Източноавстралийско стандартно време", "AKDT": "Аляска – лятно часово време", "AKST": "Аляска – стандартно време", "ARST": "Аржентинско лятно часово време", "ART": "Аржентинско стандартно време", "AST": "Северноамериканско атлантическо стандартно време", "AWDT": "Западноавстралийско лятно часово време", "AWST": "Западноавстралийско стандартно време", "BOT": "Боливийско време", "BT": "Бутанско време", "CAT": "Централноафриканско време", "CDT": "Северноамериканско централно лятно часово време", "CHADT": "Чатъмско лятно часово време", "CHAST": "Чатъмско стандартно време", "CLST": "Чилийско лятно часово време", "CLT": "Чилийско стандартно време", "COST": "Колумбийско лятно часово време", "COT": "Колумбийско стандартно време", "CST": "Северноамериканско централно стандартно време", "ChST": "Чаморско време", "EAT": "Източноафриканско време", "ECT": "Еквадорско време", "EDT": "Северноамериканско източно лятно часово време", "EST": "Северноамериканско източно стандартно време", "GFT": "Френска Гвиана", "GMT": "Средно гринуичко време", "GST": "Персийски залив", "GYT": "Гаяна", "HADT": "Хавайско-алеутско лятно часово време", "HAST": "Хавайско-алеутско стандартно време", "HAT": "Нюфаундлендско лятно часово време", "HECU": "Кубинско лятно часово време", "HEEG": "Източногренландско лятно часово време", "HENOMX": "Северозападно лятно часово мексиканско време", "HEOG": "Западногренландско лятно часово време", "HEPM": "Сен Пиер и Микелон – лятно часово време", "HEPMX": "Мексиканско тихоокеанско лятно часово време", "HKST": "Хонконгско лятно часово време", "HKT": "Хонконгско стандартно време", "HNCU": "Кубинско стандартно време", "HNEG": "Източногренландско стандартно време", "HNNOMX": "Северозападно стандартно мексиканско време", "HNOG": "Западногренландско стандартно време", "HNPM": "Сен Пиер и Микелон – стандартно време", "HNPMX": "Мексиканско тихоокеанско стандартно време", "HNT": "Нюфаундлендско стандартно време", "IST": "Индийско време", "JDT": "Японско лятно часово време", "JST": "Японско стандартно време", "LHDT": "Лорд Хау – лятно часово време", "LHST": "Лорд Хау – стандартно време", "MDT": "Северноамериканско планинско лятно часово време", "MESZ": "Централноевропейско лятно часово време", "MEZ": "Централноевропейско стандартно време", "MST": "Северноамериканско планинско стандартно време", "MYT": "Малайзийско време", "NZDT": "Новозеландско лятно часово време", "NZST": "Новозеландско стандартно време", "OESZ": "Източноевропейско лятно часово време", "OEZ": "Източноевропейско стандартно време", "PDT": "Северноамериканско тихоокеанско лятно часово време", "PST": "Северноамериканско тихоокеанско стандартно време", "SAST": "Южноафриканско време", "SGT": "Сингапурско време", "SRT": "Суринамско време", "TMST": "Туркменистанско лятно часово време", "TMT": "Туркменистанско стандартно време", "UYST": "Уругвайско лятно часово време", "UYT": "Уругвайско стандартно време", "VET": "Венецуелско време", "WARST": "Западноаржентинско лятно часово време", "WART": "Западноаржентинско стандартно време", "WAST": "Западноафриканско лятно часово време", "WAT": "Западноафриканско стандартно време", "WESZ": "Западноевропейско лятно време", "WEZ": "Западноевропейско стандартно време", "WIB": "Западноиндонезийско време", "WIT": "Източноиндонезийско време", "WITA": "Централноиндонезийско време", "∅∅∅": "Азорски острови – лятно часово време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_895833c14256127b155b25fc552f5180,
			fnFmtCurrency:         fnFmtCurrency_88fd8698aec1fdcde88908d426b01f78,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_387f50f821f8034478c23ca141a1b991,
			fnFmtDateShort:        fnFmtDateShort_724096485badda99a595db97f2fdcc96,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bg_bg"] = func() locales.Translator {
		return &localen{

			locale:                 "bg_BG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "яну", "фев", "март", "апр", "май", "юни", "юли", "авг", "сеп", "окт", "ное", "дек"},
			monthsNarrow:           []string{"", "я", "ф", "м", "а", "м", "ю", "ю", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "януари", "февруари", "март", "април", "май", "юни", "юли", "август", "септември", "октомври", "ноември", "декември"},
			daysAbbreviated:        []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"н", "п", "в", "с", "ч", "п", "с"},
			daysShort:              []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"неделя", "понеделник", "вторник", "сряда", "четвъртък", "петък", "събота"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"пр.об.", "сл.об."},
			erasAbbreviated:        []string{"пр.Хр.", "сл.Хр."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"преди Христа", "след Христа"},
			timezones:              map[string]string{"ACDT": "Централноавстралийско лятно часово време", "ACST": "Централноавстралийско стандартно време", "ACWDT": "Австралия – западно централно лятно часово време", "ACWST": "Австралия – западно централно стандартно време", "ADT": "Северноамериканско атлантическо лятно часово време", "AEDT": "Източноавстралийско лятно часово време", "AEST": "Източноавстралийско стандартно време", "AKDT": "Аляска – лятно часово време", "AKST": "Аляска – стандартно време", "ARST": "Аржентинско лятно часово време", "ART": "Аржентинско стандартно време", "AST": "Северноамериканско атлантическо стандартно време", "AWDT": "Западноавстралийско лятно часово време", "AWST": "Западноавстралийско стандартно време", "BOT": "Боливийско време", "BT": "Бутанско време", "CAT": "Централноафриканско време", "CDT": "Северноамериканско централно лятно часово време", "CHADT": "Чатъмско лятно часово време", "CHAST": "Чатъмско стандартно време", "CLST": "Чилийско лятно часово време", "CLT": "Чилийско стандартно време", "COST": "Колумбийско лятно часово време", "COT": "Колумбийско стандартно време", "CST": "Северноамериканско централно стандартно време", "ChST": "Чаморско време", "EAT": "Източноафриканско време", "ECT": "Еквадорско време", "EDT": "Северноамериканско източно лятно часово време", "EST": "Северноамериканско източно стандартно време", "GFT": "Френска Гвиана", "GMT": "Средно гринуичко време", "GST": "Персийски залив", "GYT": "Гаяна", "HADT": "Хавайско-алеутско лятно часово време", "HAST": "Хавайско-алеутско стандартно време", "HAT": "Нюфаундлендско лятно часово време", "HECU": "Кубинско лятно часово време", "HEEG": "Източногренландско лятно часово време", "HENOMX": "Северозападно лятно часово мексиканско време", "HEOG": "Западногренландско лятно часово време", "HEPM": "Сен Пиер и Микелон – лятно часово време", "HEPMX": "Мексиканско тихоокеанско лятно часово време", "HKST": "Хонконгско лятно часово време", "HKT": "Хонконгско стандартно време", "HNCU": "Кубинско стандартно време", "HNEG": "Източногренландско стандартно време", "HNNOMX": "Северозападно стандартно мексиканско време", "HNOG": "Западногренландско стандартно време", "HNPM": "Сен Пиер и Микелон – стандартно време", "HNPMX": "Мексиканско тихоокеанско стандартно време", "HNT": "Нюфаундлендско стандартно време", "IST": "Индийско време", "JDT": "Японско лятно часово време", "JST": "Японско стандартно време", "LHDT": "Лорд Хау – лятно часово време", "LHST": "Лорд Хау – стандартно време", "MDT": "Северноамериканско планинско лятно часово време", "MESZ": "Централноевропейско лятно часово време", "MEZ": "Централноевропейско стандартно време", "MST": "Северноамериканско планинско стандартно време", "MYT": "Малайзийско време", "NZDT": "Новозеландско лятно часово време", "NZST": "Новозеландско стандартно време", "OESZ": "Източноевропейско лятно часово време", "OEZ": "Източноевропейско стандартно време", "PDT": "Северноамериканско тихоокеанско лятно часово време", "PST": "Северноамериканско тихоокеанско стандартно време", "SAST": "Южноафриканско време", "SGT": "Сингапурско време", "SRT": "Суринамско време", "TMST": "Туркменистанско лятно часово време", "TMT": "Туркменистанско стандартно време", "UYST": "Уругвайско лятно часово време", "UYT": "Уругвайско стандартно време", "VET": "Венецуелско време", "WARST": "Западноаржентинско лятно часово време", "WART": "Западноаржентинско стандартно време", "WAST": "Западноафриканско лятно часово време", "WAT": "Западноафриканско стандартно време", "WESZ": "Западноевропейско лятно време", "WEZ": "Западноевропейско стандартно време", "WIB": "Западноиндонезийско време", "WIT": "Източноиндонезийско време", "WITA": "Централноиндонезийско време", "∅∅∅": "Бразилско лятно часово време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_895833c14256127b155b25fc552f5180,
			fnFmtCurrency:         fnFmtCurrency_88fd8698aec1fdcde88908d426b01f78,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_387f50f821f8034478c23ca141a1b991,
			fnFmtDateShort:        fnFmtDateShort_724096485badda99a595db97f2fdcc96,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bm"] = func() locales.Translator {
		return &localen{

			locale:                 "bm",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "zan", "feb", "mar", "awi", "mɛ", "zuw", "zul", "uti", "sɛt", "ɔku", "now", "des"},
			monthsNarrow:           []string{"", "Z", "F", "M", "A", "M", "Z", "Z", "U", "S", "Ɔ", "N", "D"},
			monthsWide:             []string{"", "zanwuye", "feburuye", "marisi", "awirili", "mɛ", "zuwɛn", "zuluye", "uti", "sɛtanburu", "ɔkutɔburu", "nowanburu", "desanburu"},
			daysAbbreviated:        []string{"kar", "ntɛ", "tar", "ara", "ala", "jum", "sib"},
			daysNarrow:             []string{"K", "N", "T", "A", "A", "J", "S"},
			daysWide:               []string{"kari", "ntɛnɛ", "tarata", "araba", "alamisa", "juma", "sibiri"},
			erasAbbreviated:        []string{"J.-C. ɲɛ", "ni J.-C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"jezu krisiti ɲɛ", "jezu krisiti minkɛ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bm_ml"] = func() locales.Translator {
		return &localen{

			locale:                 "bm_ML",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "zan", "feb", "mar", "awi", "mɛ", "zuw", "zul", "uti", "sɛt", "ɔku", "now", "des"},
			monthsNarrow:           []string{"", "Z", "F", "M", "A", "M", "Z", "Z", "U", "S", "Ɔ", "N", "D"},
			monthsWide:             []string{"", "zanwuye", "feburuye", "marisi", "awirili", "mɛ", "zuwɛn", "zuluye", "uti", "sɛtanburu", "ɔkutɔburu", "nowanburu", "desanburu"},
			daysAbbreviated:        []string{"kar", "ntɛ", "tar", "ara", "ala", "jum", "sib"},
			daysNarrow:             []string{"K", "N", "T", "A", "A", "J", "S"},
			daysWide:               []string{"kari", "ntɛnɛ", "tarata", "araba", "alamisa", "juma", "sibiri"},
			erasAbbreviated:        []string{"J.-C. ɲɛ", "ni J.-C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"jezu krisiti ɲɛ", "jezu krisiti minkɛ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bn"] = func() locales.Translator {
		return &localen{

			locale:             "bn",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "৳", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "জানু", "ফেব", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"},
			monthsNarrow:       []string{"", "জা", "ফে", "মা", "এ", "মে", "জুন", "জু", "আ", "সে", "অ", "ন", "ডি"},
			monthsWide:         []string{"", "জানুয়ারী", "ফেব্রুয়ারী", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"},
			daysAbbreviated:    []string{"রবি", "সোম", "মঙ্গল", "বুধ", "বৃহস্পতি", "শুক্র", "শনি"},
			daysNarrow:         []string{"র", "সো", "ম", "বু", "বৃ", "শু", "শ"},
			daysShort:          []string{"রঃ", "সোঃ", "মঃ", "বুঃ", "বৃঃ", "শুঃ", "শনি"},
			daysWide:           []string{"রবিবার", "সোমবার", "মঙ্গলবার", "বুধবার", "বৃহস্পতিবার", "শুক্রবার", "শনিবার"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"খ্রিস্টপূর্ব", "খৃষ্টাব্দ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"খ্রিস্টপূর্ব", "খ্রীষ্টাব্দ"},
			timezones:          map[string]string{"ACDT": "অস্ট্রেলীয় কেন্দ্রীয় দিবালোক সময়", "ACST": "অস্ট্রেলীয় কেন্দ্রীয় মানক সময়", "ACWDT": "অস্ট্রেলীয় কেন্দ্রীয় পশ্চিমি দিবালোক সময়", "ACWST": "অস্ট্রেলীয় কেন্দ্রীয় পশ্চিমি মানক সময়", "ADT": "অতলান্তিক দিবালোক সময়", "AEDT": "অস্ট্রেলীয় পূর্ব দিবালোক সময়", "AEST": "অস্ট্রেলীয় পূর্ব মানক সময়", "AKDT": "আলাস্কা দিবালোক সময়", "AKST": "আলাস্কা মানক সময়", "ARST": "আর্জেন্টিনা গ্রীষ্মকালীন সময়", "ART": "আর্জেন্টিনা মানক সময়", "AST": "অতলান্তিক মানক সময়", "AWDT": "অস্ট্রেলীয় পশ্চিমি দিবালোক সময়", "AWST": "অস্ট্রেলীয় পশ্চিমি মানক সময়", "BOT": "বোলিভিয়া সময়", "BT": "ভুটান সময়", "CAT": "মধ্য আফ্রিকা সময়", "CDT": "কেন্দ্রীয় দিবালোক সময়", "CHADT": "চ্যাথাম দিবালোক সময়", "CHAST": "চ্যাথাম মানক সময়", "CLST": "চিলি গ্রীষ্মকালীন সময়", "CLT": "চিলি মানক সময়", "COST": "কোলোম্বিয়া গ্রীষ্মকালীন সময়", "COT": "কোলোম্বিয়া মানক সময়", "CST": "কেন্দ্রীয় মানক সময়", "ChST": "চামেরো মানক সময়", "EAT": "পূর্ব আফ্রিকা সময়", "ECT": "ইকুয়েডর সময়", "EDT": "পূর্বাঞ্চলের দিবালোক সময়", "EST": "পূর্বাঞ্চলের প্রমাণ সময়", "GFT": "ফরাসি গায়ানা সময়", "GMT": "গ্রীনিচ মিন টাইম", "GST": "উপসাগরীয় মানক সময়", "GYT": "গুয়ানা সময়", "HADT": "হাওয়াই-আলেউত দিবালোক সময়", "HAST": "হাওয়াই-আলেউত মানক সময়", "HAT": "নিউফাউন্ডল্যান্ড দিবালোক সময়", "HECU": "কিউবা দিবালোক সময়", "HEEG": "পূর্ব গ্রীনল্যান্ড গ্রীষ্মকালীন সময়", "HENOMX": "উত্তরপশ্চিম মেক্সিকোর দিনের সময়", "HEOG": "পশ্চিম গ্রীনল্যান্ড গ্রীষ্মকালীন সময়", "HEPM": "সেন্ট পিয়ের ও মিকেলন দিবালোক সময়", "HEPMX": "মেক্সিকান প্রশান্ত মহাসাগরীয় দিবালোক সময়", "HKST": "হং কং গ্রীষ্মকালীন সময়", "HKT": "হং কং মানক সময়", "HNCU": "কিউবা মানক সময়", "HNEG": "পূর্ব গ্রীনল্যান্ড মানক সময়", "HNNOMX": "উত্তরপশ্চিম মেক্সিকোর মানক সময়", "HNOG": "পশ্চিম গ্রীনল্যান্ড মানক সময়", "HNPM": "সেন্ট পিয়ের ও মিকেলন মানক সময়", "HNPMX": "মেক্সিকান প্রশান্ত মহসাগরীয় মানক সময়", "HNT": "নিউফাউন্ডল্যান্ড মানক সময়", "IST": "ভারতীয় মানক সময়", "JDT": "জাপান দিবালোক সময়", "JST": "জাপান মানক সময়", "LHDT": "লর্ড হাওয়ে দিবালোক মসয়", "LHST": "লর্ড হাওয়ে মানক মসয়", "MDT": "পার্বত্য অঞ্চলের দিনের সময়", "MESZ": "মধ্য ইউরোপীয় গ্রীষ্মকালীন সময়", "MEZ": "মধ্য ইউরোপীয় মানক সময়", "MST": "পার্বত্য অঞ্চলের প্রমাণ সময়", "MYT": "মালয়েশিয়া সময়", "NZDT": "নিউজিল্যান্ড দিবালোক সময়", "NZST": "নিউজিল্যান্ড মানক সময়", "OESZ": "পূর্ব ইউরোপীয় গ্রীষ্মকালীন সময়", "OEZ": "পূর্ব ইউরোপীয় মানক সময়", "PDT": "প্রশান্ত মহাসাগরীয় অঞ্চলের দিনের সময়", "PST": "প্রশান্ত মহাসাগরীয় অঞ্চলের মানক সময়", "SAST": "দক্ষিণ আফ্রিকা মানক সময়", "SGT": "সিঙ্গাপুর মানক সময়", "SRT": "সুরিনাম সময়", "TMST": "তুর্কমেনিস্তান গ্রীষ্মকালীন সময়", "TMT": "তুর্কমেনিস্তান মানক সময়", "UYST": "উরুগুয়ে গ্রীষ্মকালীন সময়", "UYT": "উরুগুয়ে মানক সময়", "VET": "ভেনেজুয়েলা সময়", "WARST": "পশ্চিমি আর্জেনটিনা গ্রীষ্মকালীন সময়", "WART": "পশ্চিমি আর্জেনটিনার প্রমাণ সময়", "WAST": "পশ্চিম আফ্রিকা গ্রীষ্মকালীন সময়", "WAT": "পশ্চিম আফ্রিকা মানক সময়", "WESZ": "পশ্চিম ইউরোপীয় গ্রীষ্মকালীন সময়", "WEZ": "পশ্চিম ইউরোপীয় মানক সময়", "WIB": "পশ্চিমী ইন্দোনেশিয়া সময়", "WIT": "পূর্ব ইন্দোনেশিয়া সময়", "WITA": "কেন্দ্রীয় ইন্দোনেশিয়া সময়", "∅∅∅": "ব্রাসিলিয়া গ্রীষ্মকালীন সময়"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1df869b66b1bebb824de5401f039b55a,
			fnFmtCurrency:         fnFmtCurrency_ac4758b65e3186b6295da12d015494cf,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_2d950970290e57f43f401dd367ad94dc,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bn_bd"] = func() locales.Translator {
		return &localen{

			locale:             "bn_BD",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "জানু", "ফেব", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"},
			monthsNarrow:       []string{"", "জা", "ফে", "মা", "এ", "মে", "জুন", "জু", "আ", "সে", "অ", "ন", "ডি"},
			monthsWide:         []string{"", "জানুয়ারী", "ফেব্রুয়ারী", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"},
			daysAbbreviated:    []string{"রবি", "সোম", "মঙ্গল", "বুধ", "বৃহস্পতি", "শুক্র", "শনি"},
			daysNarrow:         []string{"র", "সো", "ম", "বু", "বৃ", "শু", "শ"},
			daysShort:          []string{"রঃ", "সোঃ", "মঃ", "বুঃ", "বৃঃ", "শুঃ", "শনি"},
			daysWide:           []string{"রবিবার", "সোমবার", "মঙ্গলবার", "বুধবার", "বৃহস্পতিবার", "শুক্রবার", "শনিবার"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"খ্রিস্টপূর্ব", "খৃষ্টাব্দ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"খ্রিস্টপূর্ব", "খ্রীষ্টাব্দ"},
			timezones:          map[string]string{"ACDT": "অস্ট্রেলীয় কেন্দ্রীয় দিবালোক সময়", "ACST": "অস্ট্রেলীয় কেন্দ্রীয় মানক সময়", "ACWDT": "অস্ট্রেলীয় কেন্দ্রীয় পশ্চিমি দিবালোক সময়", "ACWST": "অস্ট্রেলীয় কেন্দ্রীয় পশ্চিমি মানক সময়", "ADT": "অতলান্তিক দিবালোক সময়", "AEDT": "অস্ট্রেলীয় পূর্ব দিবালোক সময়", "AEST": "অস্ট্রেলীয় পূর্ব মানক সময়", "AKDT": "আলাস্কা দিবালোক সময়", "AKST": "আলাস্কা মানক সময়", "ARST": "আর্জেন্টিনা গ্রীষ্মকালীন সময়", "ART": "আর্জেন্টিনা মানক সময়", "AST": "অতলান্তিক মানক সময়", "AWDT": "অস্ট্রেলীয় পশ্চিমি দিবালোক সময়", "AWST": "অস্ট্রেলীয় পশ্চিমি মানক সময়", "BOT": "বোলিভিয়া সময়", "BT": "ভুটান সময়", "CAT": "মধ্য আফ্রিকা সময়", "CDT": "কেন্দ্রীয় দিবালোক সময়", "CHADT": "চ্যাথাম দিবালোক সময়", "CHAST": "চ্যাথাম মানক সময়", "CLST": "চিলি গ্রীষ্মকালীন সময়", "CLT": "চিলি মানক সময়", "COST": "কোলোম্বিয়া গ্রীষ্মকালীন সময়", "COT": "কোলোম্বিয়া মানক সময়", "CST": "কেন্দ্রীয় মানক সময়", "ChST": "চামেরো মানক সময়", "EAT": "পূর্ব আফ্রিকা সময়", "ECT": "ইকুয়েডর সময়", "EDT": "পূর্বাঞ্চলের দিবালোক সময়", "EST": "পূর্বাঞ্চলের প্রমাণ সময়", "GFT": "ফরাসি গায়ানা সময়", "GMT": "গ্রীনিচ মিন টাইম", "GST": "উপসাগরীয় মানক সময়", "GYT": "গুয়ানা সময়", "HADT": "হাওয়াই-আলেউত দিবালোক সময়", "HAST": "হাওয়াই-আলেউত মানক সময়", "HAT": "নিউফাউন্ডল্যান্ড দিবালোক সময়", "HECU": "কিউবা দিবালোক সময়", "HEEG": "পূর্ব গ্রীনল্যান্ড গ্রীষ্মকালীন সময়", "HENOMX": "উত্তরপশ্চিম মেক্সিকোর দিনের সময়", "HEOG": "পশ্চিম গ্রীনল্যান্ড গ্রীষ্মকালীন সময়", "HEPM": "সেন্ট পিয়ের ও মিকেলন দিবালোক সময়", "HEPMX": "মেক্সিকান প্রশান্ত মহাসাগরীয় দিবালোক সময়", "HKST": "হং কং গ্রীষ্মকালীন সময়", "HKT": "হং কং মানক সময়", "HNCU": "কিউবা মানক সময়", "HNEG": "পূর্ব গ্রীনল্যান্ড মানক সময়", "HNNOMX": "উত্তরপশ্চিম মেক্সিকোর মানক সময়", "HNOG": "পশ্চিম গ্রীনল্যান্ড মানক সময়", "HNPM": "সেন্ট পিয়ের ও মিকেলন মানক সময়", "HNPMX": "মেক্সিকান প্রশান্ত মহসাগরীয় মানক সময়", "HNT": "নিউফাউন্ডল্যান্ড মানক সময়", "IST": "ভারতীয় মানক সময়", "JDT": "জাপান দিবালোক সময়", "JST": "জাপান মানক সময়", "LHDT": "লর্ড হাওয়ে দিবালোক মসয়", "LHST": "লর্ড হাওয়ে মানক মসয়", "MDT": "পার্বত্য অঞ্চলের দিনের সময়", "MESZ": "মধ্য ইউরোপীয় গ্রীষ্মকালীন সময়", "MEZ": "মধ্য ইউরোপীয় মানক সময়", "MST": "পার্বত্য অঞ্চলের প্রমাণ সময়", "MYT": "মালয়েশিয়া সময়", "NZDT": "নিউজিল্যান্ড দিবালোক সময়", "NZST": "নিউজিল্যান্ড মানক সময়", "OESZ": "পূর্ব ইউরোপীয় গ্রীষ্মকালীন সময়", "OEZ": "পূর্ব ইউরোপীয় মানক সময়", "PDT": "প্রশান্ত মহাসাগরীয় অঞ্চলের দিনের সময়", "PST": "প্রশান্ত মহাসাগরীয় অঞ্চলের মানক সময়", "SAST": "দক্ষিণ আফ্রিকা মানক সময়", "SGT": "সিঙ্গাপুর মানক সময়", "SRT": "সুরিনাম সময়", "TMST": "তুর্কমেনিস্তান গ্রীষ্মকালীন সময়", "TMT": "তুর্কমেনিস্তান মানক সময়", "UYST": "উরুগুয়ে গ্রীষ্মকালীন সময়", "UYT": "উরুগুয়ে মানক সময়", "VET": "ভেনেজুয়েলা সময়", "WARST": "পশ্চিমি আর্জেনটিনা গ্রীষ্মকালীন সময়", "WART": "পশ্চিমি আর্জেনটিনার প্রমাণ সময়", "WAST": "পশ্চিম আফ্রিকা গ্রীষ্মকালীন সময়", "WAT": "পশ্চিম আফ্রিকা মানক সময়", "WESZ": "পশ্চিম ইউরোপীয় গ্রীষ্মকালীন সময়", "WEZ": "পশ্চিম ইউরোপীয় মানক সময়", "WIB": "পশ্চিমী ইন্দোনেশিয়া সময়", "WIT": "পূর্ব ইন্দোনেশিয়া সময়", "WITA": "কেন্দ্রীয় ইন্দোনেশিয়া সময়", "∅∅∅": "পেরু গ্রীষ্মকালীন সময়"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1df869b66b1bebb824de5401f039b55a,
			fnFmtCurrency:         fnFmtCurrency_ac4758b65e3186b6295da12d015494cf,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_2d950970290e57f43f401dd367ad94dc,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bn_in"] = func() locales.Translator {
		return &localen{

			locale:             "bn_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "জানু", "ফেব", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"},
			monthsNarrow:       []string{"", "জা", "ফে", "মা", "এ", "মে", "জুন", "জু", "আ", "সে", "অ", "ন", "ডি"},
			monthsWide:         []string{"", "জানুয়ারী", "ফেব্রুয়ারী", "মার্চ", "এপ্রিল", "মে", "জুন", "জুলাই", "আগস্ট", "সেপ্টেম্বর", "অক্টোবর", "নভেম্বর", "ডিসেম্বর"},
			daysAbbreviated:    []string{"রবি", "সোম", "মঙ্গল", "বুধ", "বৃহস্পতি", "শুক্র", "শনি"},
			daysNarrow:         []string{"র", "সো", "ম", "বু", "বৃ", "শু", "শ"},
			daysShort:          []string{"রঃ", "সোঃ", "মঃ", "বুঃ", "বৃঃ", "শুঃ", "শনি"},
			daysWide:           []string{"রবিবার", "সোমবার", "মঙ্গলবার", "বুধবার", "বৃহস্পতিবার", "শুক্রবার", "শনিবার"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"খ্রিস্টপূর্ব", "খৃষ্টাব্দ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"খ্রিস্টপূর্ব", "খ্রীষ্টাব্দ"},
			timezones:          map[string]string{"ACDT": "অস্ট্রেলীয় কেন্দ্রীয় দিবালোক সময়", "ACST": "অস্ট্রেলীয় কেন্দ্রীয় মানক সময়", "ACWDT": "অস্ট্রেলীয় কেন্দ্রীয় পশ্চিমি দিবালোক সময়", "ACWST": "অস্ট্রেলীয় কেন্দ্রীয় পশ্চিমি মানক সময়", "ADT": "অতলান্তিক দিবালোক সময়", "AEDT": "অস্ট্রেলীয় পূর্ব দিবালোক সময়", "AEST": "অস্ট্রেলীয় পূর্ব মানক সময়", "AKDT": "আলাস্কা দিবালোক সময়", "AKST": "আলাস্কা মানক সময়", "ARST": "আর্জেন্টিনা গ্রীষ্মকালীন সময়", "ART": "আর্জেন্টিনা মানক সময়", "AST": "অতলান্তিক মানক সময়", "AWDT": "অস্ট্রেলীয় পশ্চিমি দিবালোক সময়", "AWST": "অস্ট্রেলীয় পশ্চিমি মানক সময়", "BOT": "বোলিভিয়া সময়", "BT": "ভুটান সময়", "CAT": "মধ্য আফ্রিকা সময়", "CDT": "কেন্দ্রীয় দিবালোক সময়", "CHADT": "চ্যাথাম দিবালোক সময়", "CHAST": "চ্যাথাম মানক সময়", "CLST": "চিলি গ্রীষ্মকালীন সময়", "CLT": "চিলি মানক সময়", "COST": "কোলোম্বিয়া গ্রীষ্মকালীন সময়", "COT": "কোলোম্বিয়া মানক সময়", "CST": "কেন্দ্রীয় মানক সময়", "ChST": "চামেরো মানক সময়", "EAT": "পূর্ব আফ্রিকা সময়", "ECT": "ইকুয়েডর সময়", "EDT": "পূর্বাঞ্চলের দিবালোক সময়", "EST": "পূর্বাঞ্চলের প্রমাণ সময়", "GFT": "ফরাসি গায়ানা সময়", "GMT": "গ্রীনিচ মিন টাইম", "GST": "উপসাগরীয় মানক সময়", "GYT": "গুয়ানা সময়", "HADT": "হাওয়াই-আলেউত দিবালোক সময়", "HAST": "হাওয়াই-আলেউত মানক সময়", "HAT": "নিউফাউন্ডল্যান্ড দিবালোক সময়", "HECU": "কিউবা দিবালোক সময়", "HEEG": "পূর্ব গ্রীনল্যান্ড গ্রীষ্মকালীন সময়", "HENOMX": "উত্তরপশ্চিম মেক্সিকোর দিনের সময়", "HEOG": "পশ্চিম গ্রীনল্যান্ড গ্রীষ্মকালীন সময়", "HEPM": "সেন্ট পিয়ের ও মিকেলন দিবালোক সময়", "HEPMX": "মেক্সিকান প্রশান্ত মহাসাগরীয় দিবালোক সময়", "HKST": "হং কং গ্রীষ্মকালীন সময়", "HKT": "হং কং মানক সময়", "HNCU": "কিউবা মানক সময়", "HNEG": "পূর্ব গ্রীনল্যান্ড মানক সময়", "HNNOMX": "উত্তরপশ্চিম মেক্সিকোর মানক সময়", "HNOG": "পশ্চিম গ্রীনল্যান্ড মানক সময়", "HNPM": "সেন্ট পিয়ের ও মিকেলন মানক সময়", "HNPMX": "মেক্সিকান প্রশান্ত মহসাগরীয় মানক সময়", "HNT": "নিউফাউন্ডল্যান্ড মানক সময়", "IST": "ভারতীয় মানক সময়", "JDT": "জাপান দিবালোক সময়", "JST": "জাপান মানক সময়", "LHDT": "লর্ড হাওয়ে দিবালোক মসয়", "LHST": "লর্ড হাওয়ে মানক মসয়", "MDT": "পার্বত্য অঞ্চলের দিনের সময়", "MESZ": "মধ্য ইউরোপীয় গ্রীষ্মকালীন সময়", "MEZ": "মধ্য ইউরোপীয় মানক সময়", "MST": "পার্বত্য অঞ্চলের প্রমাণ সময়", "MYT": "মালয়েশিয়া সময়", "NZDT": "নিউজিল্যান্ড দিবালোক সময়", "NZST": "নিউজিল্যান্ড মানক সময়", "OESZ": "পূর্ব ইউরোপীয় গ্রীষ্মকালীন সময়", "OEZ": "পূর্ব ইউরোপীয় মানক সময়", "PDT": "প্রশান্ত মহাসাগরীয় অঞ্চলের দিনের সময়", "PST": "প্রশান্ত মহাসাগরীয় অঞ্চলের মানক সময়", "SAST": "দক্ষিণ আফ্রিকা মানক সময়", "SGT": "সিঙ্গাপুর মানক সময়", "SRT": "সুরিনাম সময়", "TMST": "তুর্কমেনিস্তান গ্রীষ্মকালীন সময়", "TMT": "তুর্কমেনিস্তান মানক সময়", "UYST": "উরুগুয়ে গ্রীষ্মকালীন সময়", "UYT": "উরুগুয়ে মানক সময়", "VET": "ভেনেজুয়েলা সময়", "WARST": "পশ্চিমি আর্জেনটিনা গ্রীষ্মকালীন সময়", "WART": "পশ্চিমি আর্জেনটিনার প্রমাণ সময়", "WAST": "পশ্চিম আফ্রিকা গ্রীষ্মকালীন সময়", "WAT": "পশ্চিম আফ্রিকা মানক সময়", "WESZ": "পশ্চিম ইউরোপীয় গ্রীষ্মকালীন সময়", "WEZ": "পশ্চিম ইউরোপীয় মানক সময়", "WIB": "পশ্চিমী ইন্দোনেশিয়া সময়", "WIT": "পূর্ব ইন্দোনেশিয়া সময়", "WITA": "কেন্দ্রীয় ইন্দোনেশিয়া সময়", "∅∅∅": "অ্যামাজন গ্রীষ্মকালীন সময়"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1df869b66b1bebb824de5401f039b55a,
			fnFmtCurrency:         fnFmtCurrency_ac4758b65e3186b6295da12d015494cf,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_2d950970290e57f43f401dd367ad94dc,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bo"] = func() locales.Translator {
		return &localen{

			locale:                 "bo",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ཟླ་༡", "ཟླ་༢", "ཟླ་༣", "ཟླ་༤", "ཟླ་༥", "ཟླ་༦", "ཟླ་༧", "ཟླ་༨", "ཟླ་༩", "ཟླ་༡༠", "ཟླ་༡༡", "ཟླ་༡༢"},
			monthsWide:             []string{"", "ཟླ་བ་དང་པོ", "ཟླ་བ་གཉིས་པ", "ཟླ་བ་གསུམ་པ", "ཟླ་བ་བཞི་པ", "ཟླ་བ་ལྔ་པ", "ཟླ་བ་དྲུག་པ", "ཟླ་བ་བདུན་པ", "ཟླ་བ་བརྒྱད་པ", "ཟླ་བ་དགུ་པ", "ཟླ་བ་བཅུ་པ", "ཟླ་བ་བཅུ་གཅིག་པ", "ཟླ་བ་བཅུ་གཉིས་པ"},
			daysAbbreviated:        []string{"ཉི་མ་", "ཟླ་བ་", "མིག་དམར་", "ལྷག་པ་", "ཕུར་བུ་", "པ་སངས་", "སྤེན་པ་"},
			daysNarrow:             []string{"ཉི", "ཟླ", "མིག", "ལྷག", "ཕུར", "སངས", "སྤེན"},
			daysWide:               []string{"གཟའ་ཉི་མ་", "གཟའ་ཟླ་བ་", "གཟའ་མིག་དམར་", "གཟའ་ལྷག་པ་", "གཟའ་ཕུར་བུ་", "གཟའ་པ་སངས་", "གཟའ་སྤེན་པ་"},
			periodsAbbreviated:     []string{"སྔ་དྲོ་", "ཕྱི་དྲོ་"},
			periodsWide:            []string{"སྔ་དྲོ་", "ཕྱི་དྲོ་"},
			erasAbbreviated:        []string{"སྤྱི་ལོ་སྔོན་", "སྤྱི་ལོ་"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_753ec22cf63aa959d868252cf9690e9d,
			fnFmtDateLong:         fnFmtDateLong_7a6cbc22eb124572fde0cc182fbd1d3d,
			fnFmtDateMedium:       fnFmtDateMedium_6eb6d3d989d88a86345ec9bd5152286a,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bo_cn"] = func() locales.Translator {
		return &localen{

			locale:                 "bo_CN",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ཟླ་༡", "ཟླ་༢", "ཟླ་༣", "ཟླ་༤", "ཟླ་༥", "ཟླ་༦", "ཟླ་༧", "ཟླ་༨", "ཟླ་༩", "ཟླ་༡༠", "ཟླ་༡༡", "ཟླ་༡༢"},
			monthsWide:             []string{"", "ཟླ་བ་དང་པོ", "ཟླ་བ་གཉིས་པ", "ཟླ་བ་གསུམ་པ", "ཟླ་བ་བཞི་པ", "ཟླ་བ་ལྔ་པ", "ཟླ་བ་དྲུག་པ", "ཟླ་བ་བདུན་པ", "ཟླ་བ་བརྒྱད་པ", "ཟླ་བ་དགུ་པ", "ཟླ་བ་བཅུ་པ", "ཟླ་བ་བཅུ་གཅིག་པ", "ཟླ་བ་བཅུ་གཉིས་པ"},
			daysAbbreviated:        []string{"ཉི་མ་", "ཟླ་བ་", "མིག་དམར་", "ལྷག་པ་", "ཕུར་བུ་", "པ་སངས་", "སྤེན་པ་"},
			daysNarrow:             []string{"ཉི", "ཟླ", "མིག", "ལྷག", "ཕུར", "སངས", "སྤེན"},
			daysWide:               []string{"གཟའ་ཉི་མ་", "གཟའ་ཟླ་བ་", "གཟའ་མིག་དམར་", "གཟའ་ལྷག་པ་", "གཟའ་ཕུར་བུ་", "གཟའ་པ་སངས་", "གཟའ་སྤེན་པ་"},
			periodsAbbreviated:     []string{"སྔ་དྲོ་", "ཕྱི་དྲོ་"},
			periodsWide:            []string{"སྔ་དྲོ་", "ཕྱི་དྲོ་"},
			erasAbbreviated:        []string{"སྤྱི་ལོ་སྔོན་", "སྤྱི་ལོ་"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_753ec22cf63aa959d868252cf9690e9d,
			fnFmtDateLong:         fnFmtDateLong_7a6cbc22eb124572fde0cc182fbd1d3d,
			fnFmtDateMedium:       fnFmtDateMedium_6eb6d3d989d88a86345ec9bd5152286a,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bo_in"] = func() locales.Translator {
		return &localen{

			locale:                 "bo_IN",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ཟླ་༡", "ཟླ་༢", "ཟླ་༣", "ཟླ་༤", "ཟླ་༥", "ཟླ་༦", "ཟླ་༧", "ཟླ་༨", "ཟླ་༩", "ཟླ་༡༠", "ཟླ་༡༡", "ཟླ་༡༢"},
			monthsWide:             []string{"", "ཟླ་བ་དང་པོ", "ཟླ་བ་གཉིས་པ", "ཟླ་བ་གསུམ་པ", "ཟླ་བ་བཞི་པ", "ཟླ་བ་ལྔ་པ", "ཟླ་བ་དྲུག་པ", "ཟླ་བ་བདུན་པ", "ཟླ་བ་བརྒྱད་པ", "ཟླ་བ་དགུ་པ", "ཟླ་བ་བཅུ་པ", "ཟླ་བ་བཅུ་གཅིག་པ", "ཟླ་བ་བཅུ་གཉིས་པ"},
			daysAbbreviated:        []string{"ཉི་མ་", "ཟླ་བ་", "མིག་དམར་", "ལྷག་པ་", "ཕུར་བུ་", "པ་སངས་", "སྤེན་པ་"},
			daysNarrow:             []string{"ཉི", "ཟླ", "མིག", "ལྷག", "ཕུར", "སངས", "སྤེན"},
			daysWide:               []string{"གཟའ་ཉི་མ་", "གཟའ་ཟླ་བ་", "གཟའ་མིག་དམར་", "གཟའ་ལྷག་པ་", "གཟའ་ཕུར་བུ་", "གཟའ་པ་སངས་", "གཟའ་སྤེན་པ་"},
			periodsAbbreviated:     []string{"སྔ་དྲོ་", "ཕྱི་དྲོ་"},
			periodsWide:            []string{"སྔ་དྲོ་", "ཕྱི་དྲོ་"},
			erasAbbreviated:        []string{"སྤྱི་ལོ་སྔོན་", "སྤྱི་ལོ་"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_753ec22cf63aa959d868252cf9690e9d,
			fnFmtDateLong:         fnFmtDateLong_7a6cbc22eb124572fde0cc182fbd1d3d,
			fnFmtDateMedium:       fnFmtDateMedium_6eb6d3d989d88a86345ec9bd5152286a,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["br"] = func() locales.Translator {
		return &localen{

			locale:                 "br",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$A", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "$CA", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£ RU", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "$ HK", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$ ZN", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$ SU", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Gen.", "Cʼhwe.", "Meur.", "Ebr.", "Mae", "Mezh.", "Goue.", "Eost", "Gwen.", "Here", "Du", "Kzu."},
			monthsNarrow:           []string{"", "01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12"},
			monthsWide:             []string{"", "Genver", "Cʼhwevrer", "Meurzh", "Ebrel", "Mae", "Mezheven", "Gouere", "Eost", "Gwengolo", "Here", "Du", "Kerzu"},
			daysAbbreviated:        []string{"Sul", "Lun", "Meu.", "Mer.", "Yaou", "Gwe.", "Sad."},
			daysNarrow:             []string{"Su", "L", "Mz", "Mc", "Y", "G", "Sa"},
			daysShort:              []string{"Sul", "Lun", "Meu.", "Mer.", "Yaou", "Gwe.", "Sad."},
			daysWide:               []string{"Sul", "Lun", "Meurzh", "Mercʼher", "Yaou", "Gwener", "Sadorn"},
			periodsAbbreviated:     []string{"A.M.", "G.M."},
			periodsNarrow:          []string{"am", "gm"},
			periodsWide:            []string{"A.M.", "G.M."},
			erasAbbreviated:        []string{"a-raok J.K.", "goude J.K."},
			erasNarrow:             []string{"a-raok J.K.", "goude J.K."},
			erasWide:               []string{"a-raok Jezuz-Krist", "goude Jezuz-Krist"},
			timezones:              map[string]string{"ACDT": "eur hañv Kreizaostralia", "ACST": "eur cʼhoañv Kreizaostralia", "ACWDT": "eur hañv Kreizaostralia ar Cʼhornôg", "ACWST": "eur cʼhoañv Kreizaostralia ar Cʼhornôg", "ADT": "eur hañv an Atlantel", "AEDT": "eur hañv Aostralia ar Reter", "AEST": "eur cʼhoañv Aostralia ar Reter", "AKDT": "eur hañv Alaska", "AKST": "eur cʼhoañv Alaska", "ARST": "eur hañv Arcʼhantina", "ART": "eur cʼhoañv Arcʼhantina", "AST": "eur cʼhoañv an Atlantel", "AWDT": "eur hañv Aostralia ar Cʼhornôg", "AWST": "eur cʼhoañv Aostralia ar Cʼhornôg", "BOT": "eur Bolivia", "BT": "eur Bhoutan", "CAT": "eur Kreizafrika", "CDT": "eur hañv ar Cʼhreiz", "CHADT": "eur hañv Chatham", "CHAST": "eur cʼhoañv Chatham", "CLST": "eur hañv Chile", "CLT": "eur cʼhoañv Chile", "COST": "eur hañv Kolombia", "COT": "eur cʼhoañv Kolombia", "CST": "eur cʼhoañv ar Cʼhreiz", "ChST": "eur Chamorro", "EAT": "eur Afrika ar Reter", "ECT": "eur Ecuador", "EDT": "eur hañv ar Reter", "EST": "eur cʼhoañv ar Reter", "GFT": "eur Gwiana cʼhall", "GMT": "Amzer keitat Greenwich (AKG)", "GST": "eur cʼhoañv ar Pleg-mor Arab-ha-Pers", "GYT": "eur Guyana", "HADT": "HADT", "HAST": "HAST", "HAT": "eur hañv Newfoundland", "HECU": "eur hañv Kuba", "HEEG": "eur hañv Greunland ar Reter", "HENOMX": "eur hañv Gwalarn Mecʼhiko", "HEOG": "eur hañv Greunland ar Cʼhornôg", "HEPM": "eur hañv Sant-Pêr-ha-Mikelon", "HEPMX": "HEPMX", "HKST": "eur hañv Hong Kong", "HKT": "eur cʼhoañv Hong Kong", "HNCU": "eur cʼhoañv Kuba", "HNEG": "eur cʼhoañv Greunland ar Reter", "HNNOMX": "eur cʼhoañv Gwalarn Mecʼhiko", "HNOG": "eur cʼhoañv Greunland ar Cʼhornôg", "HNPM": "eur cʼhoañv Sant-Pêr-ha-Mikelon", "HNPMX": "HNPMX", "HNT": "eur cʼhoañv Newfoundland", "IST": "eur cʼhoañv India", "JDT": "eur hañv Japan", "JST": "eur cʼhoañv Japan", "LHDT": "LHDT", "LHST": "LHST", "MDT": "eur hañv ar Menezioù", "MESZ": "eur hañv Kreizeuropa", "MEZ": "eur cʼhoañv Kreizeuropa", "MST": "eur cʼhoañv ar Menezioù", "MYT": "eur Malaysia", "NZDT": "eur hañv Zeland-Nevez", "NZST": "eur cʼhoañv Zeland-Nevez", "OESZ": "eur hañv Europa ar Reter", "OEZ": "eur cʼhoañv Europa ar Reter", "PDT": "eur hañv an Habask", "PST": "eur cʼhoañv an Habask", "SAST": "eur cʼhoañv Suafrika", "SGT": "eur cʼhoañv Singapour", "SRT": "eur Surinam", "TMST": "eur hañv Turkmenistan", "TMT": "eur cʼhoañv Turkmenistan", "UYST": "eur hañv Uruguay", "UYT": "eur cʼhoañv Uruguay", "VET": "eur Venezuela", "WARST": "eur hañv Arcʼhantina ar Cʼhornôg", "WART": "eur cʼhoañv Arcʼhantina ar Cʼhornôg", "WAST": "eur hañv Afrika ar Cʼhornôg", "WAT": "eur cʼhoañv Afrika ar Cʼhornôg", "WESZ": "eur hañv Europa ar Cʼhornôg", "WEZ": "eur cʼhoañv Europa ar Cʼhornôg", "WIB": "eur Indonezia ar Cʼhornôg", "WIT": "eur Indonezia ar Reter", "WITA": "eur Kreiz Indonezia", "∅∅∅": "eur hañv Perou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_3fb30797479a030a43141dadf528c83c,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["br_fr"] = func() locales.Translator {
		return &localen{

			locale:                 "br_FR",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Gen.", "Cʼhwe.", "Meur.", "Ebr.", "Mae", "Mezh.", "Goue.", "Eost", "Gwen.", "Here", "Du", "Kzu."},
			monthsNarrow:           []string{"", "01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12"},
			monthsWide:             []string{"", "Genver", "Cʼhwevrer", "Meurzh", "Ebrel", "Mae", "Mezheven", "Gouere", "Eost", "Gwengolo", "Here", "Du", "Kerzu"},
			daysAbbreviated:        []string{"Sul", "Lun", "Meu.", "Mer.", "Yaou", "Gwe.", "Sad."},
			daysNarrow:             []string{"Su", "L", "Mz", "Mc", "Y", "G", "Sa"},
			daysShort:              []string{"Sul", "Lun", "Meu.", "Mer.", "Yaou", "Gwe.", "Sad."},
			daysWide:               []string{"Sul", "Lun", "Meurzh", "Mercʼher", "Yaou", "Gwener", "Sadorn"},
			periodsAbbreviated:     []string{"A.M.", "G.M."},
			periodsNarrow:          []string{"am", "gm"},
			periodsWide:            []string{"A.M.", "G.M."},
			erasAbbreviated:        []string{"a-raok J.K.", "goude J.K."},
			erasNarrow:             []string{"a-raok J.K.", "goude J.K."},
			erasWide:               []string{"a-raok Jezuz-Krist", "goude Jezuz-Krist"},
			timezones:              map[string]string{"ACDT": "eur hañv Kreizaostralia", "ACST": "eur cʼhoañv Kreizaostralia", "ACWDT": "eur hañv Kreizaostralia ar Cʼhornôg", "ACWST": "eur cʼhoañv Kreizaostralia ar Cʼhornôg", "ADT": "eur hañv an Atlantel", "AEDT": "eur hañv Aostralia ar Reter", "AEST": "eur cʼhoañv Aostralia ar Reter", "AKDT": "eur hañv Alaska", "AKST": "eur cʼhoañv Alaska", "ARST": "eur hañv Arcʼhantina", "ART": "eur cʼhoañv Arcʼhantina", "AST": "eur cʼhoañv an Atlantel", "AWDT": "eur hañv Aostralia ar Cʼhornôg", "AWST": "eur cʼhoañv Aostralia ar Cʼhornôg", "BOT": "eur Bolivia", "BT": "eur Bhoutan", "CAT": "eur Kreizafrika", "CDT": "eur hañv ar Cʼhreiz", "CHADT": "eur hañv Chatham", "CHAST": "eur cʼhoañv Chatham", "CLST": "eur hañv Chile", "CLT": "eur cʼhoañv Chile", "COST": "eur hañv Kolombia", "COT": "eur cʼhoañv Kolombia", "CST": "eur cʼhoañv ar Cʼhreiz", "ChST": "eur Chamorro", "EAT": "eur Afrika ar Reter", "ECT": "eur Ecuador", "EDT": "eur hañv ar Reter", "EST": "eur cʼhoañv ar Reter", "GFT": "eur Gwiana cʼhall", "GMT": "Amzer keitat Greenwich (AKG)", "GST": "eur cʼhoañv ar Pleg-mor Arab-ha-Pers", "GYT": "eur Guyana", "HADT": "HADT", "HAST": "HAST", "HAT": "eur hañv Newfoundland", "HECU": "eur hañv Kuba", "HEEG": "eur hañv Greunland ar Reter", "HENOMX": "eur hañv Gwalarn Mecʼhiko", "HEOG": "eur hañv Greunland ar Cʼhornôg", "HEPM": "eur hañv Sant-Pêr-ha-Mikelon", "HEPMX": "HEPMX", "HKST": "eur hañv Hong Kong", "HKT": "eur cʼhoañv Hong Kong", "HNCU": "eur cʼhoañv Kuba", "HNEG": "eur cʼhoañv Greunland ar Reter", "HNNOMX": "eur cʼhoañv Gwalarn Mecʼhiko", "HNOG": "eur cʼhoañv Greunland ar Cʼhornôg", "HNPM": "eur cʼhoañv Sant-Pêr-ha-Mikelon", "HNPMX": "HNPMX", "HNT": "eur cʼhoañv Newfoundland", "IST": "eur cʼhoañv India", "JDT": "eur hañv Japan", "JST": "eur cʼhoañv Japan", "LHDT": "LHDT", "LHST": "LHST", "MDT": "eur hañv ar Menezioù", "MESZ": "eur hañv Kreizeuropa", "MEZ": "eur cʼhoañv Kreizeuropa", "MST": "eur cʼhoañv ar Menezioù", "MYT": "eur Malaysia", "NZDT": "eur hañv Zeland-Nevez", "NZST": "eur cʼhoañv Zeland-Nevez", "OESZ": "eur hañv Europa ar Reter", "OEZ": "eur cʼhoañv Europa ar Reter", "PDT": "eur hañv an Habask", "PST": "eur cʼhoañv an Habask", "SAST": "eur cʼhoañv Suafrika", "SGT": "eur cʼhoañv Singapour", "SRT": "eur Surinam", "TMST": "eur hañv Turkmenistan", "TMT": "eur cʼhoañv Turkmenistan", "UYST": "eur hañv Uruguay", "UYT": "eur cʼhoañv Uruguay", "VET": "eur Venezuela", "WARST": "eur hañv Arcʼhantina ar Cʼhornôg", "WART": "eur cʼhoañv Arcʼhantina ar Cʼhornôg", "WAST": "eur hañv Afrika ar Cʼhornôg", "WAT": "eur cʼhoañv Afrika ar Cʼhornôg", "WESZ": "eur hañv Europa ar Cʼhornôg", "WEZ": "eur cʼhoañv Europa ar Cʼhornôg", "WIB": "eur Indonezia ar Cʼhornôg", "WIT": "eur Indonezia ar Reter", "WITA": "eur Kreiz Indonezia", "∅∅∅": "eur hañv Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_3fb30797479a030a43141dadf528c83c,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["brx"] = func() locales.Translator {
		return &localen{

			locale:                 "brx",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsNarrow:           []string{"", "ज", "फे", "मा", "ए", "मे", "जु", "जु", "आ", "से", "अ", "न", "दि"},
			monthsWide:             []string{"", "जानुवारी", "फेब्रुवारी", "मार्स", "एफ्रिल", "मे", "जुन", "जुलाइ", "आगस्थ", "सेबथेज्ब़र", "अखथबर", "नबेज्ब़र", "दिसेज्ब़र"},
			daysAbbreviated:        []string{"रबि", "सम", "मंगल", "बुद", "बिसथि", "सुखुर", "सुनि"},
			daysNarrow:             []string{"र", "स", "मं", "बु", "बि", "सु", "सु"},
			daysWide:               []string{"रबिबार", "समबार", "मंगलबार", "बुदबार", "बिसथिबार", "सुखुरबार", "सुनिबार"},
			periodsAbbreviated:     []string{"फुं", "बेलासे"},
			periodsWide:            []string{"फुं", "बेलासे"},
			erasAbbreviated:        []string{"ईसा.पूर्व", "सन"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "मध्य ओस्ट्रेलिया डेलाईट टाईम", "ACST": "मध्य ओस्ट्रेलिया स्टैंडर्ड टाईम", "ACWDT": "मध्य-पश्चीम ओस्ट्रेलिया डेलाईट टाईम", "ACWST": "मध्य-पश्चीम ओस्ट्रेलिया स्टैंडर्ड टाईम", "ADT": "अटलांटीक डेलाईट टाईम", "AEDT": "पूर्वी ओस्ट्रेलिया डेलाईट टाईम", "AEST": "पूर्वी ओस्ट्रेलिया स्टैंडर्ड टाईम", "AKDT": "अलास्का डेलाईट टाईम", "AKST": "अलास्का स्टैंडर्ड टाईम", "ARST": "अर्जण्टिना समर टाईम", "ART": "अर्जनटिना स्टैंडर्ड टाईम", "AST": "अटलांटीक स्टैंडर्ड टाईम", "AWDT": "दक्षिण ओस्ट्रेलिया डेलाईट टाईम", "AWST": "दक्षिण ओस्ट्रेलिया स्टैंडर्ड टाईम", "BOT": "बोलिविया स्टैंडर्ड टाईम", "BT": "भुटान स्टैंडर्ड टाईम", "CAT": "मध्य अफ्रीका स्टैंडर्ड टाईम", "CDT": "सैंट्रल अमरिका डेलाईट टाईम", "CHADT": "चैथम डेलाईट टाईम", "CHAST": "चैथम स्टैंडर्ड टाईम", "CLST": "चीली समर टाईम", "CLT": "चीली स्टैंडर्ड टाईम", "COST": "कोलंबिया समर टाईम", "COT": "कोलंबिया स्टैंडर्ड टाईम", "CST": "सैंट्रल अमरिका स्टैंडर्ड टाईम", "ChST": "चामरो स्टैंडर्ड टाईम", "EAT": "पूर्वी अफ्रीका स्टैंडर्ड टाईम", "ECT": "एक्वाडौर स्टैंडर्ड टाईम", "EDT": "ईस्टर्न अमरिका डेलाईट टाईम", "EST": "ईस्टर्न अमरिका स्टैंडर्ड टाईम", "GFT": "फ्रान्सीसी गुयाना स्टैंडर्ड टाईम", "GMT": "ग्रीनीच स्टैंडर्ड टाईम", "GST": "गल्फ़ स्टैंडर्ड टाईम", "GYT": "गुयाना स्टैंडर्ड टाईम", "HADT": "हवाई आलटन डेलाईट टाईम", "HAST": "हवाई आलटन स्टैंडर्ड टाईम", "HAT": "न्युफाऊंडलैण्ड डेलाईट टाईम", "HECU": "क्युबा डेलाईट टाईम", "HEEG": "ग्रीनलैण्ड ईस्टर्न समर टाईम", "HENOMX": "HENOMX", "HEOG": "ग्रीनलैण्ड वेस्टर्न समर टाईम", "HEPM": "सेँ पीयॅर एवं मीकलों डेलाईट टाईम", "HEPMX": "HEPMX", "HKST": "हाँगकॉंग समर टाईम", "HKT": "हाँगकॉंग स्टैंडर्ड टाईम", "HNCU": "क्युबा स्टैंडर्ड टाईम", "HNEG": "ग्रीनलैण्ड ईस्टर्न स्टैंडर्ड टाईम", "HNNOMX": "HNNOMX", "HNOG": "ग्रीनलैण्ड वेस्टर्न स्टैंडर्ड टाईम", "HNPM": "सेँ पीयॅर एवं मीकलों स्टैंडर्ड टाईम", "HNPMX": "HNPMX", "HNT": "न्युफाऊंडलैण्ड स्टैंडर्ड टाईम", "IST": "भारतीय स्टैंडर्ड टाईम", "JDT": "जपान डेलाईट टाईम", "JST": "जपान स्टैंडर्ड टाईम", "LHDT": "लार्ड़ होव डेलाईट टाईम", "LHST": "लार्ड़ होव स्टैंडर्ड टाईम", "MDT": "अमरिका माऊन्टन डेलाईट टाईम", "MESZ": "मध्य यूरोप समर टाईम", "MEZ": "मध्य यूरोप स्टैंडर्ड टाईम", "MST": "अमरिका माऊन्टन स्टैंडर्ड टाईम", "MYT": "मलेशिया स्टैंडर्ड टाईम", "NZDT": "न्युज़ीलैण्ड डेलाईट टाईम", "NZST": "न्युज़ीलैण्ड स्टैंडर्ड टाईम", "OESZ": "ईस्टर्न यूरोप समर टाईम", "OEZ": "ईस्टर्न यूरोप स्टैंडर्ड टाईम", "PDT": "पैसीफीक डेलाईट टाईम", "PST": "पैसीफीक स्टैंडर्ड टाईम", "SAST": "दक्षिण अफ्रीका स्टैंडर्ड टाईम", "SGT": "सींगापुर स्टैंडर्ड टाईम", "SRT": "सुरीनाम स्टैंडर्ड टाईम", "TMST": "तुर्कमेनीस्तान समर टाईम", "TMT": "तुर्कमेनीस्तान स्टैंडर्ड टाईम", "UYST": "ऊरुगुए समर टाईम", "UYT": "ऊरुगुए स्टैंडर्ड टाईम", "VET": "वेनेज़ुएला स्टैंडर्ड टाईम", "WARST": "पश्चीम अर्जण्टिना समर टाईम", "WART": "पश्चीम अर्जण्टिना स्टैंडर्ड टाईम", "WAST": "पश्चीम अफ्रीका समर टाईम", "WAT": "पश्चीम अफ्रीका स्टैंडर्ड टाईम", "WESZ": "वेस्टर्न यूरोप समर टाईम", "WEZ": "वेस्टर्न यूरोप स्टैंडर्ड टाईम", "WIB": "वेस्टर्न ईंडोनीशिया स्टैंडर्ड टाईम", "WIT": "ईस्टर्न ईंडोनीशिया स्टैंडर्ड टाईम", "WITA": "ईंडोनीशिया स्टैंडर्ड टाईम", "∅∅∅": "ब्राज़ीलिया समर टाईम"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cd503abfb0c9bd938bcf20e65858f8d7,
			fnFmtCurrency:         fnFmtCurrency_9f465d9454a1702118c7b32dfb89b3d1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["brx_in"] = func() locales.Translator {
		return &localen{

			locale:                 "brx_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsNarrow:           []string{"", "ज", "फे", "मा", "ए", "मे", "जु", "जु", "आ", "से", "अ", "न", "दि"},
			monthsWide:             []string{"", "जानुवारी", "फेब्रुवारी", "मार्स", "एफ्रिल", "मे", "जुन", "जुलाइ", "आगस्थ", "सेबथेज्ब़र", "अखथबर", "नबेज्ब़र", "दिसेज्ब़र"},
			daysAbbreviated:        []string{"रबि", "सम", "मंगल", "बुद", "बिसथि", "सुखुर", "सुनि"},
			daysNarrow:             []string{"र", "स", "मं", "बु", "बि", "सु", "सु"},
			daysWide:               []string{"रबिबार", "समबार", "मंगलबार", "बुदबार", "बिसथिबार", "सुखुरबार", "सुनिबार"},
			periodsAbbreviated:     []string{"फुं", "बेलासे"},
			periodsWide:            []string{"फुं", "बेलासे"},
			erasAbbreviated:        []string{"ईसा.पूर्व", "सन"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "मध्य ओस्ट्रेलिया डेलाईट टाईम", "ACST": "मध्य ओस्ट्रेलिया स्टैंडर्ड टाईम", "ACWDT": "मध्य-पश्चीम ओस्ट्रेलिया डेलाईट टाईम", "ACWST": "मध्य-पश्चीम ओस्ट्रेलिया स्टैंडर्ड टाईम", "ADT": "अटलांटीक डेलाईट टाईम", "AEDT": "पूर्वी ओस्ट्रेलिया डेलाईट टाईम", "AEST": "पूर्वी ओस्ट्रेलिया स्टैंडर्ड टाईम", "AKDT": "अलास्का डेलाईट टाईम", "AKST": "अलास्का स्टैंडर्ड टाईम", "ARST": "अर्जण्टिना समर टाईम", "ART": "अर्जनटिना स्टैंडर्ड टाईम", "AST": "अटलांटीक स्टैंडर्ड टाईम", "AWDT": "दक्षिण ओस्ट्रेलिया डेलाईट टाईम", "AWST": "दक्षिण ओस्ट्रेलिया स्टैंडर्ड टाईम", "BOT": "बोलिविया स्टैंडर्ड टाईम", "BT": "भुटान स्टैंडर्ड टाईम", "CAT": "मध्य अफ्रीका स्टैंडर्ड टाईम", "CDT": "सैंट्रल अमरिका डेलाईट टाईम", "CHADT": "चैथम डेलाईट टाईम", "CHAST": "चैथम स्टैंडर्ड टाईम", "CLST": "चीली समर टाईम", "CLT": "चीली स्टैंडर्ड टाईम", "COST": "कोलंबिया समर टाईम", "COT": "कोलंबिया स्टैंडर्ड टाईम", "CST": "सैंट्रल अमरिका स्टैंडर्ड टाईम", "ChST": "चामरो स्टैंडर्ड टाईम", "EAT": "पूर्वी अफ्रीका स्टैंडर्ड टाईम", "ECT": "एक्वाडौर स्टैंडर्ड टाईम", "EDT": "ईस्टर्न अमरिका डेलाईट टाईम", "EST": "ईस्टर्न अमरिका स्टैंडर्ड टाईम", "GFT": "फ्रान्सीसी गुयाना स्टैंडर्ड टाईम", "GMT": "ग्रीनीच स्टैंडर्ड टाईम", "GST": "गल्फ़ स्टैंडर्ड टाईम", "GYT": "गुयाना स्टैंडर्ड टाईम", "HADT": "हवाई आलटन डेलाईट टाईम", "HAST": "हवाई आलटन स्टैंडर्ड टाईम", "HAT": "न्युफाऊंडलैण्ड डेलाईट टाईम", "HECU": "क्युबा डेलाईट टाईम", "HEEG": "ग्रीनलैण्ड ईस्टर्न समर टाईम", "HENOMX": "HENOMX", "HEOG": "ग्रीनलैण्ड वेस्टर्न समर टाईम", "HEPM": "सेँ पीयॅर एवं मीकलों डेलाईट टाईम", "HEPMX": "HEPMX", "HKST": "हाँगकॉंग समर टाईम", "HKT": "हाँगकॉंग स्टैंडर्ड टाईम", "HNCU": "क्युबा स्टैंडर्ड टाईम", "HNEG": "ग्रीनलैण्ड ईस्टर्न स्टैंडर्ड टाईम", "HNNOMX": "HNNOMX", "HNOG": "ग्रीनलैण्ड वेस्टर्न स्टैंडर्ड टाईम", "HNPM": "सेँ पीयॅर एवं मीकलों स्टैंडर्ड टाईम", "HNPMX": "HNPMX", "HNT": "न्युफाऊंडलैण्ड स्टैंडर्ड टाईम", "IST": "भारतीय स्टैंडर्ड टाईम", "JDT": "जपान डेलाईट टाईम", "JST": "जपान स्टैंडर्ड टाईम", "LHDT": "लार्ड़ होव डेलाईट टाईम", "LHST": "लार्ड़ होव स्टैंडर्ड टाईम", "MDT": "अमरिका माऊन्टन डेलाईट टाईम", "MESZ": "मध्य यूरोप समर टाईम", "MEZ": "मध्य यूरोप स्टैंडर्ड टाईम", "MST": "अमरिका माऊन्टन स्टैंडर्ड टाईम", "MYT": "मलेशिया स्टैंडर्ड टाईम", "NZDT": "न्युज़ीलैण्ड डेलाईट टाईम", "NZST": "न्युज़ीलैण्ड स्टैंडर्ड टाईम", "OESZ": "ईस्टर्न यूरोप समर टाईम", "OEZ": "ईस्टर्न यूरोप स्टैंडर्ड टाईम", "PDT": "पैसीफीक डेलाईट टाईम", "PST": "पैसीफीक स्टैंडर्ड टाईम", "SAST": "दक्षिण अफ्रीका स्टैंडर्ड टाईम", "SGT": "सींगापुर स्टैंडर्ड टाईम", "SRT": "सुरीनाम स्टैंडर्ड टाईम", "TMST": "तुर्कमेनीस्तान समर टाईम", "TMT": "तुर्कमेनीस्तान स्टैंडर्ड टाईम", "UYST": "ऊरुगुए समर टाईम", "UYT": "ऊरुगुए स्टैंडर्ड टाईम", "VET": "वेनेज़ुएला स्टैंडर्ड टाईम", "WARST": "पश्चीम अर्जण्टिना समर टाईम", "WART": "पश्चीम अर्जण्टिना स्टैंडर्ड टाईम", "WAST": "पश्चीम अफ्रीका समर टाईम", "WAT": "पश्चीम अफ्रीका स्टैंडर्ड टाईम", "WESZ": "वेस्टर्न यूरोप समर टाईम", "WEZ": "वेस्टर्न यूरोप स्टैंडर्ड टाईम", "WIB": "वेस्टर्न ईंडोनीशिया स्टैंडर्ड टाईम", "WIT": "ईस्टर्न ईंडोनीशिया स्टैंडर्ड टाईम", "WITA": "ईंडोनीशिया स्टैंडर्ड टाईम", "∅∅∅": "ब्राज़ीलिया समर टाईम"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cd503abfb0c9bd938bcf20e65858f8d7,
			fnFmtCurrency:         fnFmtCurrency_9f465d9454a1702118c7b32dfb89b3d1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bs"] = func() locales.Translator {
		return &localen{

			locale:                 "bs",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "KM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "kn", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "din.", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "mart", "april", "maj", "juni", "juli", "august", "septembar", "oktobar", "novembar", "decembar"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"prijepodne", "popodne"},
			periodsWide:            []string{"prijepodne", "popodne"},
			erasAbbreviated:        []string{"p. n. e.", "n. e."},
			erasNarrow:             []string{"p.n.e.", "n.e."},
			erasWide:               []string{"prije nove ere", "nove ere"},
			timezones:              map[string]string{"ACDT": "Centralnoaustralijsko ljetno vrijeme", "ACST": "Centralnoaustralijsko standardno vrijeme", "ACWDT": "Australijsko centralnozapadno ljetno vrijeme", "ACWST": "Australijsko centralnozapadno standardno vrijeme", "ADT": "Sjevernoameričko atlantsko ljetno vrijeme", "AEDT": "Istočnoaustralijsko ljetno vrijeme", "AEST": "Istočnoaustralijsko standardno vrijeme", "AKDT": "Aljaskansko ljetno vrijeme", "AKST": "Aljaskansko standardno vrijeme", "ARST": "Argentinsko ljetno vrijeme", "ART": "Argentinsko standardno vrijeme", "AST": "Sjevernoameričko atlantsko standardno vrijeme", "AWDT": "Zapadnoaustralijsko ljetno vrijeme", "AWST": "Zapadnoaustralijsko standardno vrijeme", "BOT": "Bolivijsko vrijeme", "BT": "Butansko vrijeme", "CAT": "Centralnoafričko vrijeme", "CDT": "Sjevernoameričko centralno ljetno vrijeme", "CHADT": "Čatamsko ljetno vrijeme", "CHAST": "Čatamsko standardno vrijeme", "CLST": "Čileansko ljetno vrijeme", "CLT": "Čileansko standardno vrijeme", "COST": "Kolumbijsko ljetno vrijeme", "COT": "Kolumbijsko standardno vrijeme", "CST": "Sjevernoameričko centralno standardno vrijeme", "ChST": "Čamorsko standardno vrijeme", "EAT": "Istočnoafričko vrijeme", "ECT": "Ekvadorsko vrijeme", "EDT": "Sjevernoameričko istočno ljetno vrijeme", "EST": "Sjevernoameričko istočno standardno vrijeme", "GFT": "Francuskogvajansko vrijeme", "GMT": "Griničko vrijeme", "GST": "Zalivsko standardno vrijeme", "GYT": "Gvajansko vrijeme", "HADT": "Havajsko-aleućansko ljetno vrijeme", "HAST": "Havajsko-aleućansko standardno vrijeme", "HAT": "Njufaundlendsko ljetno vrijeme", "HECU": "Kubansko ljetno vrijeme", "HEEG": "Istočnogrenlandsko ljetno vrijeme", "HENOMX": "Sjeverozapadno meksičko ljetno vrijeme", "HEOG": "Zapadnogrenlandsko ljetno vrijeme", "HEPM": "Ljetno vrijeme na Ostrvima Sveti Petar i Mikelon", "HEPMX": "Meksičko pacifičko ljetno vrijeme", "HKST": "Hongkonško ljetno vrijeme", "HKT": "Hongkonško standardno vrijeme", "HNCU": "Kubansko standardno vrijeme", "HNEG": "Istočnogrenlandsko standardno vrijeme", "HNNOMX": "Sjeverozapadno meksičko standardno vrijeme", "HNOG": "Zapadnogrenlandsko standardno vrijeme", "HNPM": "Standardno vrijeme na Ostrvima Sveti Petar i Mikelon", "HNPMX": "Meksičko pacifičko standardno vrijeme", "HNT": "Njufaundlendsko standardno vrijeme", "IST": "Indijsko standardno vrijeme", "JDT": "Japansko ljetno vrijeme", "JST": "Japansko standardno vrijeme", "LHDT": "Ljetno vrijeme na Ostrvu Lord Hau", "LHST": "Standardno vrijeme na Ostrvu Lord Hau", "MDT": "Sjevernoameričko planinsko ljetno vrijeme", "MESZ": "Centralnoevropsko ljetno vrijeme", "MEZ": "Centralnoevropsko standardno vrijeme", "MST": "Sjevernoameričko planinsko standardno vrijeme", "MYT": "Malezijsko vrijeme", "NZDT": "Novozelandsko ljetno vrijeme", "NZST": "Novozelandsko standardno vrijeme", "OESZ": "Istočnoevropsko ljetno vrijeme", "OEZ": "Istočnoevropsko standardno vrijeme", "PDT": "Sjevernoameričko pacifičko ljetno vrijeme", "PST": "Sjevernoameričko pacifičko standardno vrijeme", "SAST": "Južnoafričko standardno vrijeme", "SGT": "Singapursko standardno vrijeme", "SRT": "Surinamsko vrijeme", "TMST": "Turkmenistansko ljetno vrijeme", "TMT": "Turkmenistansko standardno vrijeme", "UYST": "Urugvajsko ljetno vrijeme", "UYT": "Urugvajsko standardno vrijeme", "VET": "Venecuelansko vrijeme", "WARST": "Zapadnoargentinsko ljetno vrijeme", "WART": "Zapadnoargentinsko standardno vrijeme", "WAST": "Zapadnoafričko ljetno vrijeme", "WAT": "Zapadnoafričko standardno vrijeme", "WESZ": "Zapadnoevropsko ljetno vrijeme", "WEZ": "Zapadnoevropsko standardno vrijeme", "WIB": "Zapadnoindonezijsko vrijeme", "WIT": "Istočnoindonezijsko vrijeme", "WITA": "Centralnoindonezijsko vrijeme", "∅∅∅": "Amazonsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_aa123c99f95c3c4073e2b02fd117f0c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bs_cyrl"] = func() locales.Translator {
		return &localen{

			locale:                 "bs_Cyrl",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "КМ", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "Кч", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "зл", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "дин.", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "Тл", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "ауг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јуни", "јули", "аугуст", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сри", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"недјеља", "понедјељак", "уторак", "сриједа", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"пре подне", "поподне"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"прије подне", "послије подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"прије нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно љетње рачунање времена", "ACST": "Аустралијско централно стандардно вријеме", "ACWDT": "Аустралијско централно западно љетње рачунање времена", "ACWST": "Аустралијско централно западно стандардно вријеме", "ADT": "Атланско љетње рачунање времена", "AEDT": "Аустралијско источно љетње рачунање времена", "AEST": "Аустралијско источно стандардно вријеме", "AKDT": "Аљаска љетње вријеме", "AKST": "Аљаска стандардно вријеме", "ARST": "Аргентина љетње рачунање времена", "ART": "Аргентина стандардно вријеме", "AST": "Атланско стандардно вријеме", "AWDT": "Аустралијско западно љетње рачунање времена", "AWST": "Аустралијско западно стандардно вријеме", "BOT": "Боливија вријеме", "BT": "Бутан вријеме", "CAT": "Централно-афричко вријеме", "CDT": "Централно љетње рачунање времена", "CHADT": "Чатам љетње рачунање времена", "CHAST": "Чатам стандардно вријеме", "CLST": "Чиле љетње рачунање времена", "CLT": "Чиле стандардно вријеме", "COST": "Колумбија љетње рачунање времена", "COT": "Колумбија стандардно вријеме", "CST": "Централно стандардно вријеме", "ChST": "Чаморо вријеме", "EAT": "Источно-афричко вријеме", "ECT": "Еквадор вријеме", "EDT": "Источно љетње рачунање времена", "EST": "Источно стандардно вријеме", "GFT": "Француска Гвајана вријеме", "GMT": "Гриничко средње вријеме", "GST": "Заливско вријеме", "GYT": "Гвајана вријеме", "HADT": "Хавајско-алеутско љетње рачунање времена", "HAST": "Хавајско-алеутско стандардно вријеме", "HAT": "Њуфаундленд љетње рачунање времена", "HECU": "Куба љетње рачунање времена", "HEEG": "Источни Гренланд љетње рачунање времена", "HENOMX": "Сјеверномексичко љетње рачунање времена", "HEOG": "Западни Гренланд љетње рачунање времена", "HEPM": "Сен Пјер и Микелон љетње рачунање вемена", "HEPMX": "Мексичко пацифичко љетње рачунање времена", "HKST": "Хонгконшко љетње рачунање времена", "HKT": "Хонг Конг стандардно вријеме", "HNCU": "Куба стандардно вријеме", "HNEG": "Источни Гренланд стандардно вријеме", "HNNOMX": "Сјеверномексичко стандардно вријеме", "HNOG": "Западни Гренланд стандардно вријеме", "HNPM": "Сен Пјер и Микелон стандардно вријеме", "HNPMX": "Мексичко пацифичко стандардно вријеме", "HNT": "Њуфаундленд стандардно вријеме", "IST": "Индијско стандардно вријеме", "JDT": "Јапанско љетње рачунање времена", "JST": "Јапанско стандардно вријеме", "LHDT": "Лорд Хов љетње рачунање времена", "LHST": "Лорд Хов стандардно вријеме", "MDT": "Планинско љетње рачунање времена", "MESZ": "Средњеевропско љетње рачунање времена", "MEZ": "Средњеевропско стандардно вријеме", "MST": "Планинско стандардно вријеме", "MYT": "Малезија вријеме", "NZDT": "Нови Зеланд љетње рачунање времена", "NZST": "Нови Зеланд стандардно вријеме", "OESZ": "Источноевропско љетње рачунање времена", "OEZ": "Источноевропско стандардно вријеме", "PDT": "Пацифичко љетње рачунање времена", "PST": "Пацифичко стандардно вријеме", "SAST": "Јужно-афричко вријеме", "SGT": "Сингапур стандардно вријеме", "SRT": "Суринам вријеме", "TMST": "Туркменистан љетње рачунање времена", "TMT": "Туркменистан стандардно вријеме", "UYST": "Уругвај љетње рачунање времена", "UYT": "Уругвај стандардно вријеме", "VET": "Венецуела вријеме", "WARST": "Западна Аргентина љетње рачунање времена", "WART": "Западна Аргентина стандардно вријеме", "WAST": "Западно-афричко љетње рачунање времена", "WAT": "Западно-афричко стандардно вријеме", "WESZ": "Западноевропско љетње рачунање времена", "WEZ": "Западноевропско стандардно вријеме", "WIB": "Западно-индонезијско вријеме", "WIT": "Источно-индонезијско вријеме", "WITA": "Централно-индонезијско вријеме", "∅∅∅": "Перу љетње рачунање времена"},

			fnCardinalPluralRule:  fnCardinalPluralRule_7b6df98bb7a40cd3e9c67c7c7f464286,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bs_cyrl_ba"] = func() locales.Translator {
		return &localen{

			locale:                 "bs_Cyrl_BA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "mart", "april", "maj", "juni", "juli", "august", "septembar", "oktobar", "novembar", "decembar"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"prijepodne", "popodne"},
			periodsWide:            []string{"prijepodne", "popodne"},
			erasAbbreviated:        []string{"p. n. e.", "n. e."},
			erasNarrow:             []string{"p.n.e.", "n.e."},
			erasWide:               []string{"prije nove ere", "nove ere"},
			timezones:              map[string]string{"ACDT": "Centralnoaustralijsko ljetno vrijeme", "ACST": "Centralnoaustralijsko standardno vrijeme", "ACWDT": "Australijsko centralnozapadno ljetno vrijeme", "ACWST": "Australijsko centralnozapadno standardno vrijeme", "ADT": "Sjevernoameričko atlantsko ljetno vrijeme", "AEDT": "Istočnoaustralijsko ljetno vrijeme", "AEST": "Istočnoaustralijsko standardno vrijeme", "AKDT": "Aljaskansko ljetno vrijeme", "AKST": "Aljaskansko standardno vrijeme", "ARST": "Argentinsko ljetno vrijeme", "ART": "Argentinsko standardno vrijeme", "AST": "Sjevernoameričko atlantsko standardno vrijeme", "AWDT": "Zapadnoaustralijsko ljetno vrijeme", "AWST": "Zapadnoaustralijsko standardno vrijeme", "BOT": "Bolivijsko vrijeme", "BT": "Butansko vrijeme", "CAT": "Centralnoafričko vrijeme", "CDT": "Sjevernoameričko centralno ljetno vrijeme", "CHADT": "Čatamsko ljetno vrijeme", "CHAST": "Čatamsko standardno vrijeme", "CLST": "Čileansko ljetno vrijeme", "CLT": "Čileansko standardno vrijeme", "COST": "Kolumbijsko ljetno vrijeme", "COT": "Kolumbijsko standardno vrijeme", "CST": "Sjevernoameričko centralno standardno vrijeme", "ChST": "Čamorsko standardno vrijeme", "EAT": "Istočnoafričko vrijeme", "ECT": "Ekvadorsko vrijeme", "EDT": "Sjevernoameričko istočno ljetno vrijeme", "EST": "Sjevernoameričko istočno standardno vrijeme", "GFT": "Francuskogvajansko vrijeme", "GMT": "Griničko vrijeme", "GST": "Zalivsko standardno vrijeme", "GYT": "Gvajansko vrijeme", "HADT": "Havajsko-aleućansko ljetno vrijeme", "HAST": "Havajsko-aleućansko standardno vrijeme", "HAT": "Njufaundlendsko ljetno vrijeme", "HECU": "Kubansko ljetno vrijeme", "HEEG": "Istočnogrenlandsko ljetno vrijeme", "HENOMX": "Sjeverozapadno meksičko ljetno vrijeme", "HEOG": "Zapadnogrenlandsko ljetno vrijeme", "HEPM": "Ljetno vrijeme na Ostrvima Sveti Petar i Mikelon", "HEPMX": "Meksičko pacifičko ljetno vrijeme", "HKST": "Hongkonško ljetno vrijeme", "HKT": "Hongkonško standardno vrijeme", "HNCU": "Kubansko standardno vrijeme", "HNEG": "Istočnogrenlandsko standardno vrijeme", "HNNOMX": "Sjeverozapadno meksičko standardno vrijeme", "HNOG": "Zapadnogrenlandsko standardno vrijeme", "HNPM": "Standardno vrijeme na Ostrvima Sveti Petar i Mikelon", "HNPMX": "Meksičko pacifičko standardno vrijeme", "HNT": "Njufaundlendsko standardno vrijeme", "IST": "Indijsko standardno vrijeme", "JDT": "Japansko ljetno vrijeme", "JST": "Japansko standardno vrijeme", "LHDT": "Ljetno vrijeme na Ostrvu Lord Hau", "LHST": "Standardno vrijeme na Ostrvu Lord Hau", "MDT": "Sjevernoameričko planinsko ljetno vrijeme", "MESZ": "Centralnoevropsko ljetno vrijeme", "MEZ": "Centralnoevropsko standardno vrijeme", "MST": "Sjevernoameričko planinsko standardno vrijeme", "MYT": "Malezijsko vrijeme", "NZDT": "Novozelandsko ljetno vrijeme", "NZST": "Novozelandsko standardno vrijeme", "OESZ": "Istočnoevropsko ljetno vrijeme", "OEZ": "Istočnoevropsko standardno vrijeme", "PDT": "Sjevernoameričko pacifičko ljetno vrijeme", "PST": "Sjevernoameričko pacifičko standardno vrijeme", "SAST": "Južnoafričko standardno vrijeme", "SGT": "Singapursko standardno vrijeme", "SRT": "Surinamsko vrijeme", "TMST": "Turkmenistansko ljetno vrijeme", "TMT": "Turkmenistansko standardno vrijeme", "UYST": "Urugvajsko ljetno vrijeme", "UYT": "Urugvajsko standardno vrijeme", "VET": "Venecuelansko vrijeme", "WARST": "Zapadnoargentinsko ljetno vrijeme", "WART": "Zapadnoargentinsko standardno vrijeme", "WAST": "Zapadnoafričko ljetno vrijeme", "WAT": "Zapadnoafričko standardno vrijeme", "WESZ": "Zapadnoevropsko ljetno vrijeme", "WEZ": "Zapadnoevropsko standardno vrijeme", "WIB": "Zapadnoindonezijsko vrijeme", "WIT": "Istočnoindonezijsko vrijeme", "WITA": "Centralnoindonezijsko vrijeme", "∅∅∅": "Brazilijsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_2fd16cba9f129ab3317e08638d736342,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_aa123c99f95c3c4073e2b02fd117f0c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bs_latn"] = func() locales.Translator {
		return &localen{

			locale:                 "bs_Latn",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "mart", "april", "maj", "juni", "juli", "august", "septembar", "oktobar", "novembar", "decembar"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"prijepodne", "popodne"},
			periodsWide:            []string{"prijepodne", "popodne"},
			erasAbbreviated:        []string{"p. n. e.", "n. e."},
			erasNarrow:             []string{"p.n.e.", "n.e."},
			erasWide:               []string{"prije nove ere", "nove ere"},
			timezones:              map[string]string{"ACDT": "Centralnoaustralijsko ljetno vrijeme", "ACST": "Centralnoaustralijsko standardno vrijeme", "ACWDT": "Australijsko centralnozapadno ljetno vrijeme", "ACWST": "Australijsko centralnozapadno standardno vrijeme", "ADT": "Sjevernoameričko atlantsko ljetno vrijeme", "AEDT": "Istočnoaustralijsko ljetno vrijeme", "AEST": "Istočnoaustralijsko standardno vrijeme", "AKDT": "Aljaskansko ljetno vrijeme", "AKST": "Aljaskansko standardno vrijeme", "ARST": "Argentinsko ljetno vrijeme", "ART": "Argentinsko standardno vrijeme", "AST": "Sjevernoameričko atlantsko standardno vrijeme", "AWDT": "Zapadnoaustralijsko ljetno vrijeme", "AWST": "Zapadnoaustralijsko standardno vrijeme", "BOT": "Bolivijsko vrijeme", "BT": "Butansko vrijeme", "CAT": "Centralnoafričko vrijeme", "CDT": "Sjevernoameričko centralno ljetno vrijeme", "CHADT": "Čatamsko ljetno vrijeme", "CHAST": "Čatamsko standardno vrijeme", "CLST": "Čileansko ljetno vrijeme", "CLT": "Čileansko standardno vrijeme", "COST": "Kolumbijsko ljetno vrijeme", "COT": "Kolumbijsko standardno vrijeme", "CST": "Sjevernoameričko centralno standardno vrijeme", "ChST": "Čamorsko standardno vrijeme", "EAT": "Istočnoafričko vrijeme", "ECT": "Ekvadorsko vrijeme", "EDT": "Sjevernoameričko istočno ljetno vrijeme", "EST": "Sjevernoameričko istočno standardno vrijeme", "GFT": "Francuskogvajansko vrijeme", "GMT": "Griničko vrijeme", "GST": "Zalivsko standardno vrijeme", "GYT": "Gvajansko vrijeme", "HADT": "Havajsko-aleućansko ljetno vrijeme", "HAST": "Havajsko-aleućansko standardno vrijeme", "HAT": "Njufaundlendsko ljetno vrijeme", "HECU": "Kubansko ljetno vrijeme", "HEEG": "Istočnogrenlandsko ljetno vrijeme", "HENOMX": "Sjeverozapadno meksičko ljetno vrijeme", "HEOG": "Zapadnogrenlandsko ljetno vrijeme", "HEPM": "Ljetno vrijeme na Ostrvima Sveti Petar i Mikelon", "HEPMX": "Meksičko pacifičko ljetno vrijeme", "HKST": "Hongkonško ljetno vrijeme", "HKT": "Hongkonško standardno vrijeme", "HNCU": "Kubansko standardno vrijeme", "HNEG": "Istočnogrenlandsko standardno vrijeme", "HNNOMX": "Sjeverozapadno meksičko standardno vrijeme", "HNOG": "Zapadnogrenlandsko standardno vrijeme", "HNPM": "Standardno vrijeme na Ostrvima Sveti Petar i Mikelon", "HNPMX": "Meksičko pacifičko standardno vrijeme", "HNT": "Njufaundlendsko standardno vrijeme", "IST": "Indijsko standardno vrijeme", "JDT": "Japansko ljetno vrijeme", "JST": "Japansko standardno vrijeme", "LHDT": "Ljetno vrijeme na Ostrvu Lord Hau", "LHST": "Standardno vrijeme na Ostrvu Lord Hau", "MDT": "Sjevernoameričko planinsko ljetno vrijeme", "MESZ": "Centralnoevropsko ljetno vrijeme", "MEZ": "Centralnoevropsko standardno vrijeme", "MST": "Sjevernoameričko planinsko standardno vrijeme", "MYT": "Malezijsko vrijeme", "NZDT": "Novozelandsko ljetno vrijeme", "NZST": "Novozelandsko standardno vrijeme", "OESZ": "Istočnoevropsko ljetno vrijeme", "OEZ": "Istočnoevropsko standardno vrijeme", "PDT": "Sjevernoameričko pacifičko ljetno vrijeme", "PST": "Sjevernoameričko pacifičko standardno vrijeme", "SAST": "Južnoafričko standardno vrijeme", "SGT": "Singapursko standardno vrijeme", "SRT": "Surinamsko vrijeme", "TMST": "Turkmenistansko ljetno vrijeme", "TMT": "Turkmenistansko standardno vrijeme", "UYST": "Urugvajsko ljetno vrijeme", "UYT": "Urugvajsko standardno vrijeme", "VET": "Venecuelansko vrijeme", "WARST": "Zapadnoargentinsko ljetno vrijeme", "WART": "Zapadnoargentinsko standardno vrijeme", "WAST": "Zapadnoafričko ljetno vrijeme", "WAT": "Zapadnoafričko standardno vrijeme", "WESZ": "Zapadnoevropsko ljetno vrijeme", "WEZ": "Zapadnoevropsko standardno vrijeme", "WIB": "Zapadnoindonezijsko vrijeme", "WIT": "Istočnoindonezijsko vrijeme", "WITA": "Centralnoindonezijsko vrijeme", "∅∅∅": "Brazilijsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_aa123c99f95c3c4073e2b02fd117f0c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["bs_latn_ba"] = func() locales.Translator {
		return &localen{

			locale:                 "bs_Latn_BA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "mart", "april", "maj", "juni", "juli", "august", "septembar", "oktobar", "novembar", "decembar"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"prijepodne", "popodne"},
			periodsWide:            []string{"prijepodne", "popodne"},
			erasAbbreviated:        []string{"p. n. e.", "n. e."},
			erasNarrow:             []string{"p.n.e.", "n.e."},
			erasWide:               []string{"prije nove ere", "nove ere"},
			timezones:              map[string]string{"ACDT": "Centralnoaustralijsko ljetno vrijeme", "ACST": "Centralnoaustralijsko standardno vrijeme", "ACWDT": "Australijsko centralnozapadno ljetno vrijeme", "ACWST": "Australijsko centralnozapadno standardno vrijeme", "ADT": "Sjevernoameričko atlantsko ljetno vrijeme", "AEDT": "Istočnoaustralijsko ljetno vrijeme", "AEST": "Istočnoaustralijsko standardno vrijeme", "AKDT": "Aljaskansko ljetno vrijeme", "AKST": "Aljaskansko standardno vrijeme", "ARST": "Argentinsko ljetno vrijeme", "ART": "Argentinsko standardno vrijeme", "AST": "Sjevernoameričko atlantsko standardno vrijeme", "AWDT": "Zapadnoaustralijsko ljetno vrijeme", "AWST": "Zapadnoaustralijsko standardno vrijeme", "BOT": "Bolivijsko vrijeme", "BT": "Butansko vrijeme", "CAT": "Centralnoafričko vrijeme", "CDT": "Sjevernoameričko centralno ljetno vrijeme", "CHADT": "Čatamsko ljetno vrijeme", "CHAST": "Čatamsko standardno vrijeme", "CLST": "Čileansko ljetno vrijeme", "CLT": "Čileansko standardno vrijeme", "COST": "Kolumbijsko ljetno vrijeme", "COT": "Kolumbijsko standardno vrijeme", "CST": "Sjevernoameričko centralno standardno vrijeme", "ChST": "Čamorsko standardno vrijeme", "EAT": "Istočnoafričko vrijeme", "ECT": "Ekvadorsko vrijeme", "EDT": "Sjevernoameričko istočno ljetno vrijeme", "EST": "Sjevernoameričko istočno standardno vrijeme", "GFT": "Francuskogvajansko vrijeme", "GMT": "Griničko vrijeme", "GST": "Zalivsko standardno vrijeme", "GYT": "Gvajansko vrijeme", "HADT": "Havajsko-aleućansko ljetno vrijeme", "HAST": "Havajsko-aleućansko standardno vrijeme", "HAT": "Njufaundlendsko ljetno vrijeme", "HECU": "Kubansko ljetno vrijeme", "HEEG": "Istočnogrenlandsko ljetno vrijeme", "HENOMX": "Sjeverozapadno meksičko ljetno vrijeme", "HEOG": "Zapadnogrenlandsko ljetno vrijeme", "HEPM": "Ljetno vrijeme na Ostrvima Sveti Petar i Mikelon", "HEPMX": "Meksičko pacifičko ljetno vrijeme", "HKST": "Hongkonško ljetno vrijeme", "HKT": "Hongkonško standardno vrijeme", "HNCU": "Kubansko standardno vrijeme", "HNEG": "Istočnogrenlandsko standardno vrijeme", "HNNOMX": "Sjeverozapadno meksičko standardno vrijeme", "HNOG": "Zapadnogrenlandsko standardno vrijeme", "HNPM": "Standardno vrijeme na Ostrvima Sveti Petar i Mikelon", "HNPMX": "Meksičko pacifičko standardno vrijeme", "HNT": "Njufaundlendsko standardno vrijeme", "IST": "Indijsko standardno vrijeme", "JDT": "Japansko ljetno vrijeme", "JST": "Japansko standardno vrijeme", "LHDT": "Ljetno vrijeme na Ostrvu Lord Hau", "LHST": "Standardno vrijeme na Ostrvu Lord Hau", "MDT": "Sjevernoameričko planinsko ljetno vrijeme", "MESZ": "Centralnoevropsko ljetno vrijeme", "MEZ": "Centralnoevropsko standardno vrijeme", "MST": "Sjevernoameričko planinsko standardno vrijeme", "MYT": "Malezijsko vrijeme", "NZDT": "Novozelandsko ljetno vrijeme", "NZST": "Novozelandsko standardno vrijeme", "OESZ": "Istočnoevropsko ljetno vrijeme", "OEZ": "Istočnoevropsko standardno vrijeme", "PDT": "Sjevernoameričko pacifičko ljetno vrijeme", "PST": "Sjevernoameričko pacifičko standardno vrijeme", "SAST": "Južnoafričko standardno vrijeme", "SGT": "Singapursko standardno vrijeme", "SRT": "Surinamsko vrijeme", "TMST": "Turkmenistansko ljetno vrijeme", "TMT": "Turkmenistansko standardno vrijeme", "UYST": "Urugvajsko ljetno vrijeme", "UYT": "Urugvajsko standardno vrijeme", "VET": "Venecuelansko vrijeme", "WARST": "Zapadnoargentinsko ljetno vrijeme", "WART": "Zapadnoargentinsko standardno vrijeme", "WAST": "Zapadnoafričko ljetno vrijeme", "WAT": "Zapadnoafričko standardno vrijeme", "WESZ": "Zapadnoevropsko ljetno vrijeme", "WEZ": "Zapadnoevropsko standardno vrijeme", "WIB": "Zapadnoindonezijsko vrijeme", "WIT": "Istočnoindonezijsko vrijeme", "WITA": "Centralnoindonezijsko vrijeme", "∅∅∅": "Azorsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_aa123c99f95c3c4073e2b02fd117f0c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ca"] = func() locales.Translator {
		return &localen{

			locale:                 "ca",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "₧", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."},
			monthsNarrow:           []string{"", "GN", "FB", "MÇ", "AB", "MG", "JN", "JL", "AG", "ST", "OC", "NV", "DS"},
			monthsWide:             []string{"", "de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"},
			daysAbbreviated:        []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysNarrow:             []string{"dg", "dl", "dt", "dc", "dj", "dv", "ds"},
			daysShort:              []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysWide:               []string{"diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"aC", "dC"},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"abans de Crist", "després de Crist"},
			timezones:              map[string]string{"ACDT": "Hora d’estiu d’Austràlia Central", "ACST": "Hora estàndard d’Austràlia Central", "ACWDT": "Hora d’estiu d’Austràlia centre-occidental", "ACWST": "Hora estàndard d’Austràlia centre-occidental", "ADT": "Hora d’estiu de l’Atlàntic", "AEDT": "Hora d’estiu d’Austràlia Oriental", "AEST": "Hora estàndard d’Austràlia Oriental", "AKDT": "Hora d’estiu d’Alaska", "AKST": "Hora estàndard d’Alaska", "ARST": "Hora d’estiu de l’Argentina", "ART": "Hora estàndard de l’Argentina", "AST": "Hora estàndard de l’Atlàntic", "AWDT": "Hora d’estiu d’Austràlia Occidental", "AWST": "Hora estàndard d’Austràlia Occidental", "BOT": "Hora de Bolívia", "BT": "Hora de Bhutan", "CAT": "Hora de l’Àfrica Central", "CDT": "Hora d’estiu central d’Amèrica del Nord", "CHADT": "Hora d’estiu de Chatham", "CHAST": "Hora estàndard de Chatham", "CLST": "Hora d’estiu de Xile", "CLT": "Hora estàndard de Xile", "COST": "Hora d’estiu de Colòmbia", "COT": "Hora estàndard de Colòmbia", "CST": "Hora estàndard central d’Amèrica del Nord", "ChST": "Hora de Chamorro", "EAT": "Hora de l’Àfrica Oriental", "ECT": "Hora de l’Equador", "EDT": "Hora d’estiu oriental d’Amèrica del Nord", "EST": "Hora estàndard oriental d’Amèrica del Nord", "GFT": "Hora de la Guaiana Francesa", "GMT": "Hora del Meridià de Greenwich", "GST": "Hora del Golf", "GYT": "Hora de Guyana", "HADT": "Hora d’estiu de Hawaii-Aleutianes", "HAST": "Hora estàndard de Hawaii-Aleutianes", "HAT": "Hora d’estiu de Terranova", "HECU": "Hora d’estiu de Cuba", "HEEG": "Hora d’estiu de l’Est de Groenlàndia", "HENOMX": "Hora d’estiu del nord-oest de Mèxic", "HEOG": "Hora d’estiu de l’Oest de Groenlàndia", "HEPM": "Hora d’estiu de Saint-Pierre-et-Miquelon", "HEPMX": "Hora d’estiu del Pacífic de Mèxic", "HKST": "Hora d’estiu de Hong Kong", "HKT": "Hora estàndard de Hong Kong", "HNCU": "Hora estàndard de Cuba", "HNEG": "Hora estàndard de l’Est de Groenlàndia", "HNNOMX": "Hora estàndard del nord-oest de Mèxic", "HNOG": "Hora estàndard de l’Oest de Groenlàndia", "HNPM": "Hora estàndard de Saint-Pierre-et-Miquelon", "HNPMX": "Hora estàndard del Pacífic de Mèxic", "HNT": "Hora estàndard de Terranova", "IST": "Hora de l’Índia", "JDT": "Hora d’estiu del Japó", "JST": "Hora estàndard del Japó", "LHDT": "Horari d’estiu de Lord Howe", "LHST": "Hora estàndard de Lord Howe", "MDT": "Hora d’estiu de muntanya d’Amèrica del Nord", "MESZ": "Hora d’estiu del Centre d’Europa", "MEZ": "Hora estàndard del Centre d’Europa", "MST": "Hora estàndard de muntanya d’Amèrica del Nord", "MYT": "Hora de Malàisia", "NZDT": "Hora d’estiu de Nova Zelanda", "NZST": "Hora estàndard de Nova Zelanda", "OESZ": "Hora d’estiu de l’Est d’Europa", "OEZ": "Hora estàndard de l’Est d’Europa", "PDT": "Hora d’estiu del Pacífic d’Amèrica del Nord", "PST": "Hora estàndard del Pacífic d’Amèrica del Nord", "SAST": "Hora estàndard del sud de l’Àfrica", "SGT": "Hora de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora d’estiu del Turkmenistan", "TMT": "Hora estàndard del Turkmenistan", "UYST": "Hora d’estiu de l’Uruguai", "UYT": "Hora estàndard de l’Uruguai", "VET": "Hora de Veneçuela", "WARST": "Hora d’estiu de l’oest de l’Argentina", "WART": "Hora estàndard de l’oest de l’Argentina", "WAST": "Hora d’estiu de l’Àfrica Occidental", "WAT": "Hora estàndard de l’Àfrica Occidental", "WESZ": "Hora d’estiu de l’Oest d’Europa", "WEZ": "Hora estàndard de l’Oest d’Europa", "WIB": "Hora de l’oest d’Indonèsia", "WIT": "Hora de l’est d’Indonèsia", "WITA": "Hora central d’Indonèsia", "∅∅∅": "Hora d’estiu de les Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ca_ad"] = func() locales.Translator {
		return &localen{

			locale:                 "ca_AD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."},
			monthsNarrow:           []string{"", "GN", "FB", "MÇ", "AB", "MG", "JN", "JL", "AG", "ST", "OC", "NV", "DS"},
			monthsWide:             []string{"", "de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"},
			daysAbbreviated:        []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysNarrow:             []string{"dg", "dl", "dt", "dc", "dj", "dv", "ds"},
			daysShort:              []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysWide:               []string{"diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"aC", "dC"},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"abans de Crist", "després de Crist"},
			timezones:              map[string]string{"ACDT": "Hora d’estiu d’Austràlia Central", "ACST": "Hora estàndard d’Austràlia Central", "ACWDT": "Hora d’estiu d’Austràlia centre-occidental", "ACWST": "Hora estàndard d’Austràlia centre-occidental", "ADT": "Hora d’estiu de l’Atlàntic", "AEDT": "Hora d’estiu d’Austràlia Oriental", "AEST": "Hora estàndard d’Austràlia Oriental", "AKDT": "Hora d’estiu d’Alaska", "AKST": "Hora estàndard d’Alaska", "ARST": "Hora d’estiu de l’Argentina", "ART": "Hora estàndard de l’Argentina", "AST": "Hora estàndard de l’Atlàntic", "AWDT": "Hora d’estiu d’Austràlia Occidental", "AWST": "Hora estàndard d’Austràlia Occidental", "BOT": "Hora de Bolívia", "BT": "Hora de Bhutan", "CAT": "Hora de l’Àfrica Central", "CDT": "Hora d’estiu central d’Amèrica del Nord", "CHADT": "Hora d’estiu de Chatham", "CHAST": "Hora estàndard de Chatham", "CLST": "Hora d’estiu de Xile", "CLT": "Hora estàndard de Xile", "COST": "Hora d’estiu de Colòmbia", "COT": "Hora estàndard de Colòmbia", "CST": "Hora estàndard central d’Amèrica del Nord", "ChST": "Hora de Chamorro", "EAT": "Hora de l’Àfrica Oriental", "ECT": "Hora de l’Equador", "EDT": "Hora d’estiu oriental d’Amèrica del Nord", "EST": "Hora estàndard oriental d’Amèrica del Nord", "GFT": "Hora de la Guaiana Francesa", "GMT": "Hora del Meridià de Greenwich", "GST": "Hora del Golf", "GYT": "Hora de Guyana", "HADT": "Hora d’estiu de Hawaii-Aleutianes", "HAST": "Hora estàndard de Hawaii-Aleutianes", "HAT": "Hora d’estiu de Terranova", "HECU": "Hora d’estiu de Cuba", "HEEG": "Hora d’estiu de l’Est de Groenlàndia", "HENOMX": "Hora d’estiu del nord-oest de Mèxic", "HEOG": "Hora d’estiu de l’Oest de Groenlàndia", "HEPM": "Hora d’estiu de Saint-Pierre-et-Miquelon", "HEPMX": "Hora d’estiu del Pacífic de Mèxic", "HKST": "Hora d’estiu de Hong Kong", "HKT": "Hora estàndard de Hong Kong", "HNCU": "Hora estàndard de Cuba", "HNEG": "Hora estàndard de l’Est de Groenlàndia", "HNNOMX": "Hora estàndard del nord-oest de Mèxic", "HNOG": "Hora estàndard de l’Oest de Groenlàndia", "HNPM": "Hora estàndard de Saint-Pierre-et-Miquelon", "HNPMX": "Hora estàndard del Pacífic de Mèxic", "HNT": "Hora estàndard de Terranova", "IST": "Hora de l’Índia", "JDT": "Hora d’estiu del Japó", "JST": "Hora estàndard del Japó", "LHDT": "Horari d’estiu de Lord Howe", "LHST": "Hora estàndard de Lord Howe", "MDT": "Hora d’estiu de muntanya d’Amèrica del Nord", "MESZ": "Hora d’estiu del Centre d’Europa", "MEZ": "Hora estàndard del Centre d’Europa", "MST": "Hora estàndard de muntanya d’Amèrica del Nord", "MYT": "Hora de Malàisia", "NZDT": "Hora d’estiu de Nova Zelanda", "NZST": "Hora estàndard de Nova Zelanda", "OESZ": "Hora d’estiu de l’Est d’Europa", "OEZ": "Hora estàndard de l’Est d’Europa", "PDT": "Hora d’estiu del Pacífic d’Amèrica del Nord", "PST": "Hora estàndard del Pacífic d’Amèrica del Nord", "SAST": "Hora estàndard del sud de l’Àfrica", "SGT": "Hora de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora d’estiu del Turkmenistan", "TMT": "Hora estàndard del Turkmenistan", "UYST": "Hora d’estiu de l’Uruguai", "UYT": "Hora estàndard de l’Uruguai", "VET": "Hora de Veneçuela", "WARST": "Hora d’estiu de l’oest de l’Argentina", "WART": "Hora estàndard de l’oest de l’Argentina", "WAST": "Hora d’estiu de l’Àfrica Occidental", "WAT": "Hora estàndard de l’Àfrica Occidental", "WESZ": "Hora d’estiu de l’Oest d’Europa", "WEZ": "Hora estàndard de l’Oest d’Europa", "WIB": "Hora de l’oest d’Indonèsia", "WIT": "Hora de l’est d’Indonèsia", "WITA": "Hora central d’Indonèsia", "∅∅∅": "Hora d’estiu de les Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ca_es"] = func() locales.Translator {
		return &localen{

			locale:                 "ca_ES",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."},
			monthsNarrow:           []string{"", "GN", "FB", "MÇ", "AB", "MG", "JN", "JL", "AG", "ST", "OC", "NV", "DS"},
			monthsWide:             []string{"", "de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"},
			daysAbbreviated:        []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysNarrow:             []string{"dg", "dl", "dt", "dc", "dj", "dv", "ds"},
			daysShort:              []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysWide:               []string{"diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"aC", "dC"},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"abans de Crist", "després de Crist"},
			timezones:              map[string]string{"ACDT": "Hora d’estiu d’Austràlia Central", "ACST": "Hora estàndard d’Austràlia Central", "ACWDT": "Hora d’estiu d’Austràlia centre-occidental", "ACWST": "Hora estàndard d’Austràlia centre-occidental", "ADT": "Hora d’estiu de l’Atlàntic", "AEDT": "Hora d’estiu d’Austràlia Oriental", "AEST": "Hora estàndard d’Austràlia Oriental", "AKDT": "Hora d’estiu d’Alaska", "AKST": "Hora estàndard d’Alaska", "ARST": "Hora d’estiu de l’Argentina", "ART": "Hora estàndard de l’Argentina", "AST": "Hora estàndard de l’Atlàntic", "AWDT": "Hora d’estiu d’Austràlia Occidental", "AWST": "Hora estàndard d’Austràlia Occidental", "BOT": "Hora de Bolívia", "BT": "Hora de Bhutan", "CAT": "Hora de l’Àfrica Central", "CDT": "Hora d’estiu central d’Amèrica del Nord", "CHADT": "Hora d’estiu de Chatham", "CHAST": "Hora estàndard de Chatham", "CLST": "Hora d’estiu de Xile", "CLT": "Hora estàndard de Xile", "COST": "Hora d’estiu de Colòmbia", "COT": "Hora estàndard de Colòmbia", "CST": "Hora estàndard central d’Amèrica del Nord", "ChST": "Hora de Chamorro", "EAT": "Hora de l’Àfrica Oriental", "ECT": "Hora de l’Equador", "EDT": "Hora d’estiu oriental d’Amèrica del Nord", "EST": "Hora estàndard oriental d’Amèrica del Nord", "GFT": "Hora de la Guaiana Francesa", "GMT": "Hora del Meridià de Greenwich", "GST": "Hora del Golf", "GYT": "Hora de Guyana", "HADT": "Hora d’estiu de Hawaii-Aleutianes", "HAST": "Hora estàndard de Hawaii-Aleutianes", "HAT": "Hora d’estiu de Terranova", "HECU": "Hora d’estiu de Cuba", "HEEG": "Hora d’estiu de l’Est de Groenlàndia", "HENOMX": "Hora d’estiu del nord-oest de Mèxic", "HEOG": "Hora d’estiu de l’Oest de Groenlàndia", "HEPM": "Hora d’estiu de Saint-Pierre-et-Miquelon", "HEPMX": "Hora d’estiu del Pacífic de Mèxic", "HKST": "Hora d’estiu de Hong Kong", "HKT": "Hora estàndard de Hong Kong", "HNCU": "Hora estàndard de Cuba", "HNEG": "Hora estàndard de l’Est de Groenlàndia", "HNNOMX": "Hora estàndard del nord-oest de Mèxic", "HNOG": "Hora estàndard de l’Oest de Groenlàndia", "HNPM": "Hora estàndard de Saint-Pierre-et-Miquelon", "HNPMX": "Hora estàndard del Pacífic de Mèxic", "HNT": "Hora estàndard de Terranova", "IST": "Hora de l’Índia", "JDT": "Hora d’estiu del Japó", "JST": "Hora estàndard del Japó", "LHDT": "Horari d’estiu de Lord Howe", "LHST": "Hora estàndard de Lord Howe", "MDT": "Hora d’estiu de muntanya d’Amèrica del Nord", "MESZ": "Hora d’estiu del Centre d’Europa", "MEZ": "Hora estàndard del Centre d’Europa", "MST": "Hora estàndard de muntanya d’Amèrica del Nord", "MYT": "Hora de Malàisia", "NZDT": "Hora d’estiu de Nova Zelanda", "NZST": "Hora estàndard de Nova Zelanda", "OESZ": "Hora d’estiu de l’Est d’Europa", "OEZ": "Hora estàndard de l’Est d’Europa", "PDT": "Hora d’estiu del Pacífic d’Amèrica del Nord", "PST": "Hora estàndard del Pacífic d’Amèrica del Nord", "SAST": "Hora estàndard del sud de l’Àfrica", "SGT": "Hora de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora d’estiu del Turkmenistan", "TMT": "Hora estàndard del Turkmenistan", "UYST": "Hora d’estiu de l’Uruguai", "UYT": "Hora estàndard de l’Uruguai", "VET": "Hora de Veneçuela", "WARST": "Hora d’estiu de l’oest de l’Argentina", "WART": "Hora estàndard de l’oest de l’Argentina", "WAST": "Hora d’estiu de l’Àfrica Occidental", "WAT": "Hora estàndard de l’Àfrica Occidental", "WESZ": "Hora d’estiu de l’Oest d’Europa", "WEZ": "Hora estàndard de l’Oest d’Europa", "WIB": "Hora de l’oest d’Indonèsia", "WIT": "Hora de l’est d’Indonèsia", "WITA": "Hora central d’Indonèsia", "∅∅∅": "Hora d’estiu de l’Amazones"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ca_es_valencia"] = func() locales.Translator {
		return &localen{

			locale:                 "ca_ES_VALENCIA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."},
			monthsNarrow:           []string{"", "GN", "FB", "MÇ", "AB", "MG", "JN", "JL", "AG", "ST", "OC", "NV", "DS"},
			monthsWide:             []string{"", "de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"},
			daysAbbreviated:        []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysNarrow:             []string{"dg", "dl", "dt", "dc", "dj", "dv", "ds"},
			daysShort:              []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysWide:               []string{"diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"aC", "dC"},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"abans de Crist", "després de Crist"},
			timezones:              map[string]string{"ACDT": "Hora d’estiu d’Austràlia Central", "ACST": "Hora estàndard d’Austràlia Central", "ACWDT": "Hora d’estiu d’Austràlia centre-occidental", "ACWST": "Hora estàndard d’Austràlia centre-occidental", "ADT": "Hora d’estiu de l’Atlàntic", "AEDT": "Hora d’estiu d’Austràlia Oriental", "AEST": "Hora estàndard d’Austràlia Oriental", "AKDT": "Hora d’estiu d’Alaska", "AKST": "Hora estàndard d’Alaska", "ARST": "Hora d’estiu de l’Argentina", "ART": "Hora estàndard de l’Argentina", "AST": "Hora estàndard de l’Atlàntic", "AWDT": "Hora d’estiu d’Austràlia Occidental", "AWST": "Hora estàndard d’Austràlia Occidental", "BOT": "Hora de Bolívia", "BT": "Hora de Bhutan", "CAT": "Hora de l’Àfrica Central", "CDT": "Hora d’estiu central d’Amèrica del Nord", "CHADT": "Hora d’estiu de Chatham", "CHAST": "Hora estàndard de Chatham", "CLST": "Hora d’estiu de Xile", "CLT": "Hora estàndard de Xile", "COST": "Hora d’estiu de Colòmbia", "COT": "Hora estàndard de Colòmbia", "CST": "Hora estàndard central d’Amèrica del Nord", "ChST": "Hora de Chamorro", "EAT": "Hora de l’Àfrica Oriental", "ECT": "Hora de l’Equador", "EDT": "Hora d’estiu oriental d’Amèrica del Nord", "EST": "Hora estàndard oriental d’Amèrica del Nord", "GFT": "Hora de la Guaiana Francesa", "GMT": "Hora del Meridià de Greenwich", "GST": "Hora del Golf", "GYT": "Hora de Guyana", "HADT": "Hora d’estiu de Hawaii-Aleutianes", "HAST": "Hora estàndard de Hawaii-Aleutianes", "HAT": "Hora d’estiu de Terranova", "HECU": "Hora d’estiu de Cuba", "HEEG": "Hora d’estiu de l’Est de Groenlàndia", "HENOMX": "Hora d’estiu del nord-oest de Mèxic", "HEOG": "Hora d’estiu de l’Oest de Groenlàndia", "HEPM": "Hora d’estiu de Saint-Pierre-et-Miquelon", "HEPMX": "Hora d’estiu del Pacífic de Mèxic", "HKST": "Hora d’estiu de Hong Kong", "HKT": "Hora estàndard de Hong Kong", "HNCU": "Hora estàndard de Cuba", "HNEG": "Hora estàndard de l’Est de Groenlàndia", "HNNOMX": "Hora estàndard del nord-oest de Mèxic", "HNOG": "Hora estàndard de l’Oest de Groenlàndia", "HNPM": "Hora estàndard de Saint-Pierre-et-Miquelon", "HNPMX": "Hora estàndard del Pacífic de Mèxic", "HNT": "Hora estàndard de Terranova", "IST": "Hora de l’Índia", "JDT": "Hora d’estiu del Japó", "JST": "Hora estàndard del Japó", "LHDT": "Horari d’estiu de Lord Howe", "LHST": "Hora estàndard de Lord Howe", "MDT": "Hora d’estiu de muntanya d’Amèrica del Nord", "MESZ": "Hora d’estiu del Centre d’Europa", "MEZ": "Hora estàndard del Centre d’Europa", "MST": "Hora estàndard de muntanya d’Amèrica del Nord", "MYT": "Hora de Malàisia", "NZDT": "Hora d’estiu de Nova Zelanda", "NZST": "Hora estàndard de Nova Zelanda", "OESZ": "Hora d’estiu de l’Est d’Europa", "OEZ": "Hora estàndard de l’Est d’Europa", "PDT": "Hora d’estiu del Pacífic d’Amèrica del Nord", "PST": "Hora estàndard del Pacífic d’Amèrica del Nord", "SAST": "Hora estàndard del sud de l’Àfrica", "SGT": "Hora de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora d’estiu del Turkmenistan", "TMT": "Hora estàndard del Turkmenistan", "UYST": "Hora d’estiu de l’Uruguai", "UYT": "Hora estàndard de l’Uruguai", "VET": "Hora de Veneçuela", "WARST": "Hora d’estiu de l’oest de l’Argentina", "WART": "Hora estàndard de l’oest de l’Argentina", "WAST": "Hora d’estiu de l’Àfrica Occidental", "WAT": "Hora estàndard de l’Àfrica Occidental", "WESZ": "Hora d’estiu de l’Oest d’Europa", "WEZ": "Hora estàndard de l’Oest d’Europa", "WIB": "Hora de l’oest d’Indonèsia", "WIT": "Hora de l’est d’Indonèsia", "WITA": "Hora central d’Indonèsia", "∅∅∅": "Hora d’estiu de les Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ca_fr"] = func() locales.Translator {
		return &localen{

			locale:                 "ca_FR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "F", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."},
			monthsNarrow:           []string{"", "GN", "FB", "MÇ", "AB", "MG", "JN", "JL", "AG", "ST", "OC", "NV", "DS"},
			monthsWide:             []string{"", "de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"},
			daysAbbreviated:        []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysNarrow:             []string{"dg", "dl", "dt", "dc", "dj", "dv", "ds"},
			daysShort:              []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysWide:               []string{"diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"aC", "dC"},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"abans de Crist", "després de Crist"},
			timezones:              map[string]string{"ACDT": "Hora d’estiu d’Austràlia Central", "ACST": "Hora estàndard d’Austràlia Central", "ACWDT": "Hora d’estiu d’Austràlia centre-occidental", "ACWST": "Hora estàndard d’Austràlia centre-occidental", "ADT": "Hora d’estiu de l’Atlàntic", "AEDT": "Hora d’estiu d’Austràlia Oriental", "AEST": "Hora estàndard d’Austràlia Oriental", "AKDT": "Hora d’estiu d’Alaska", "AKST": "Hora estàndard d’Alaska", "ARST": "Hora d’estiu de l’Argentina", "ART": "Hora estàndard de l’Argentina", "AST": "Hora estàndard de l’Atlàntic", "AWDT": "Hora d’estiu d’Austràlia Occidental", "AWST": "Hora estàndard d’Austràlia Occidental", "BOT": "Hora de Bolívia", "BT": "Hora de Bhutan", "CAT": "Hora de l’Àfrica Central", "CDT": "Hora d’estiu central d’Amèrica del Nord", "CHADT": "Hora d’estiu de Chatham", "CHAST": "Hora estàndard de Chatham", "CLST": "Hora d’estiu de Xile", "CLT": "Hora estàndard de Xile", "COST": "Hora d’estiu de Colòmbia", "COT": "Hora estàndard de Colòmbia", "CST": "Hora estàndard central d’Amèrica del Nord", "ChST": "Hora de Chamorro", "EAT": "Hora de l’Àfrica Oriental", "ECT": "Hora de l’Equador", "EDT": "Hora d’estiu oriental d’Amèrica del Nord", "EST": "Hora estàndard oriental d’Amèrica del Nord", "GFT": "Hora de la Guaiana Francesa", "GMT": "Hora del Meridià de Greenwich", "GST": "Hora del Golf", "GYT": "Hora de Guyana", "HADT": "Hora d’estiu de Hawaii-Aleutianes", "HAST": "Hora estàndard de Hawaii-Aleutianes", "HAT": "Hora d’estiu de Terranova", "HECU": "Hora d’estiu de Cuba", "HEEG": "Hora d’estiu de l’Est de Groenlàndia", "HENOMX": "Hora d’estiu del nord-oest de Mèxic", "HEOG": "Hora d’estiu de l’Oest de Groenlàndia", "HEPM": "Hora d’estiu de Saint-Pierre-et-Miquelon", "HEPMX": "Hora d’estiu del Pacífic de Mèxic", "HKST": "Hora d’estiu de Hong Kong", "HKT": "Hora estàndard de Hong Kong", "HNCU": "Hora estàndard de Cuba", "HNEG": "Hora estàndard de l’Est de Groenlàndia", "HNNOMX": "Hora estàndard del nord-oest de Mèxic", "HNOG": "Hora estàndard de l’Oest de Groenlàndia", "HNPM": "Hora estàndard de Saint-Pierre-et-Miquelon", "HNPMX": "Hora estàndard del Pacífic de Mèxic", "HNT": "Hora estàndard de Terranova", "IST": "Hora de l’Índia", "JDT": "Hora d’estiu del Japó", "JST": "Hora estàndard del Japó", "LHDT": "Horari d’estiu de Lord Howe", "LHST": "Hora estàndard de Lord Howe", "MDT": "Hora d’estiu de muntanya d’Amèrica del Nord", "MESZ": "Hora d’estiu del Centre d’Europa", "MEZ": "Hora estàndard del Centre d’Europa", "MST": "Hora estàndard de muntanya d’Amèrica del Nord", "MYT": "Hora de Malàisia", "NZDT": "Hora d’estiu de Nova Zelanda", "NZST": "Hora estàndard de Nova Zelanda", "OESZ": "Hora d’estiu de l’Est d’Europa", "OEZ": "Hora estàndard de l’Est d’Europa", "PDT": "Hora d’estiu del Pacífic d’Amèrica del Nord", "PST": "Hora estàndard del Pacífic d’Amèrica del Nord", "SAST": "Hora estàndard del sud de l’Àfrica", "SGT": "Hora de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora d’estiu del Turkmenistan", "TMT": "Hora estàndard del Turkmenistan", "UYST": "Hora d’estiu de l’Uruguai", "UYT": "Hora estàndard de l’Uruguai", "VET": "Hora de Veneçuela", "WARST": "Hora d’estiu de l’oest de l’Argentina", "WART": "Hora estàndard de l’oest de l’Argentina", "WAST": "Hora d’estiu de l’Àfrica Occidental", "WAT": "Hora estàndard de l’Àfrica Occidental", "WESZ": "Hora d’estiu de l’Oest d’Europa", "WEZ": "Hora estàndard de l’Oest d’Europa", "WIB": "Hora de l’oest d’Indonèsia", "WIT": "Hora de l’est d’Indonèsia", "WITA": "Hora central d’Indonèsia", "∅∅∅": "Hora d’estiu del Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ca_it"] = func() locales.Translator {
		return &localen{

			locale:                 "ca_IT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."},
			monthsNarrow:           []string{"", "GN", "FB", "MÇ", "AB", "MG", "JN", "JL", "AG", "ST", "OC", "NV", "DS"},
			monthsWide:             []string{"", "de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"},
			daysAbbreviated:        []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysNarrow:             []string{"dg", "dl", "dt", "dc", "dj", "dv", "ds"},
			daysShort:              []string{"dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."},
			daysWide:               []string{"diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"aC", "dC"},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"abans de Crist", "després de Crist"},
			timezones:              map[string]string{"ACDT": "Hora d’estiu d’Austràlia Central", "ACST": "Hora estàndard d’Austràlia Central", "ACWDT": "Hora d’estiu d’Austràlia centre-occidental", "ACWST": "Hora estàndard d’Austràlia centre-occidental", "ADT": "Hora d’estiu de l’Atlàntic", "AEDT": "Hora d’estiu d’Austràlia Oriental", "AEST": "Hora estàndard d’Austràlia Oriental", "AKDT": "Hora d’estiu d’Alaska", "AKST": "Hora estàndard d’Alaska", "ARST": "Hora d’estiu de l’Argentina", "ART": "Hora estàndard de l’Argentina", "AST": "Hora estàndard de l’Atlàntic", "AWDT": "Hora d’estiu d’Austràlia Occidental", "AWST": "Hora estàndard d’Austràlia Occidental", "BOT": "Hora de Bolívia", "BT": "Hora de Bhutan", "CAT": "Hora de l’Àfrica Central", "CDT": "Hora d’estiu central d’Amèrica del Nord", "CHADT": "Hora d’estiu de Chatham", "CHAST": "Hora estàndard de Chatham", "CLST": "Hora d’estiu de Xile", "CLT": "Hora estàndard de Xile", "COST": "Hora d’estiu de Colòmbia", "COT": "Hora estàndard de Colòmbia", "CST": "Hora estàndard central d’Amèrica del Nord", "ChST": "Hora de Chamorro", "EAT": "Hora de l’Àfrica Oriental", "ECT": "Hora de l’Equador", "EDT": "Hora d’estiu oriental d’Amèrica del Nord", "EST": "Hora estàndard oriental d’Amèrica del Nord", "GFT": "Hora de la Guaiana Francesa", "GMT": "Hora del Meridià de Greenwich", "GST": "Hora del Golf", "GYT": "Hora de Guyana", "HADT": "Hora d’estiu de Hawaii-Aleutianes", "HAST": "Hora estàndard de Hawaii-Aleutianes", "HAT": "Hora d’estiu de Terranova", "HECU": "Hora d’estiu de Cuba", "HEEG": "Hora d’estiu de l’Est de Groenlàndia", "HENOMX": "Hora d’estiu del nord-oest de Mèxic", "HEOG": "Hora d’estiu de l’Oest de Groenlàndia", "HEPM": "Hora d’estiu de Saint-Pierre-et-Miquelon", "HEPMX": "Hora d’estiu del Pacífic de Mèxic", "HKST": "Hora d’estiu de Hong Kong", "HKT": "Hora estàndard de Hong Kong", "HNCU": "Hora estàndard de Cuba", "HNEG": "Hora estàndard de l’Est de Groenlàndia", "HNNOMX": "Hora estàndard del nord-oest de Mèxic", "HNOG": "Hora estàndard de l’Oest de Groenlàndia", "HNPM": "Hora estàndard de Saint-Pierre-et-Miquelon", "HNPMX": "Hora estàndard del Pacífic de Mèxic", "HNT": "Hora estàndard de Terranova", "IST": "Hora de l’Índia", "JDT": "Hora d’estiu del Japó", "JST": "Hora estàndard del Japó", "LHDT": "Horari d’estiu de Lord Howe", "LHST": "Hora estàndard de Lord Howe", "MDT": "Hora d’estiu de muntanya d’Amèrica del Nord", "MESZ": "Hora d’estiu del Centre d’Europa", "MEZ": "Hora estàndard del Centre d’Europa", "MST": "Hora estàndard de muntanya d’Amèrica del Nord", "MYT": "Hora de Malàisia", "NZDT": "Hora d’estiu de Nova Zelanda", "NZST": "Hora estàndard de Nova Zelanda", "OESZ": "Hora d’estiu de l’Est d’Europa", "OEZ": "Hora estàndard de l’Est d’Europa", "PDT": "Hora d’estiu del Pacífic d’Amèrica del Nord", "PST": "Hora estàndard del Pacífic d’Amèrica del Nord", "SAST": "Hora estàndard del sud de l’Àfrica", "SGT": "Hora de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora d’estiu del Turkmenistan", "TMT": "Hora estàndard del Turkmenistan", "UYST": "Hora d’estiu de l’Uruguai", "UYT": "Hora estàndard de l’Uruguai", "VET": "Hora de Veneçuela", "WARST": "Hora d’estiu de l’oest de l’Argentina", "WART": "Hora estàndard de l’oest de l’Argentina", "WAST": "Hora d’estiu de l’Àfrica Occidental", "WAT": "Hora estàndard de l’Àfrica Occidental", "WESZ": "Hora d’estiu de l’Oest d’Europa", "WEZ": "Hora estàndard de l’Oest d’Europa", "WIB": "Hora de l’oest d’Indonèsia", "WIT": "Hora de l’est d’Indonèsia", "WITA": "Hora central d’Indonèsia", "∅∅∅": "Hora d’estiu del Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_6f37ff0e733377c3c49c15d7ec4ea94b,
			fnFmtDateLong:         fnFmtDateLong_b496f47cbe02a0b60f833a85f2f19cbc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_535d124951e9b710ee661bfe8ca4c345,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ccp"] = func() locales.Translator {
		return &localen{

			locale:             "ccp",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "৳", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "𑄎𑄚𑄪", "𑄜𑄬𑄛𑄴", "𑄟𑄢𑄴𑄌𑄧", "𑄃𑄬𑄛𑄳𑄢𑄨𑄣𑄴", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪𑄣𑄭", "𑄃𑄉𑄧𑄌𑄴𑄑𑄴", "𑄥𑄬𑄛𑄴𑄑𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄃𑄧𑄇𑄴𑄑𑄮𑄝𑄧𑄢𑄴", "𑄚𑄧𑄞𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄓𑄨𑄥𑄬𑄟𑄴𑄝𑄢𑄴"},
			monthsNarrow:       []string{"", "𑄎", "𑄜𑄬", "𑄟", "𑄃𑄬", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪", "𑄃", "𑄥𑄬", "𑄃𑄧", "𑄚𑄧", "𑄓𑄨"},
			monthsWide:         []string{"", "𑄎𑄚𑄪𑄠𑄢𑄨", "𑄜𑄬𑄛𑄴𑄝𑄳𑄢𑄪𑄠𑄢𑄨", "𑄟𑄢𑄴𑄌𑄧", "𑄃𑄬𑄛𑄳𑄢𑄨𑄣𑄴", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪𑄣𑄭", "𑄃𑄉𑄧𑄌𑄴𑄑𑄴", "𑄥𑄬𑄛𑄴𑄑𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄃𑄧𑄇𑄴𑄑𑄬𑄝𑄧𑄢𑄴", "𑄚𑄧𑄞𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄓𑄨𑄥𑄬𑄟𑄴𑄝𑄧𑄢𑄴"},
			daysAbbreviated:    []string{"𑄢𑄧𑄝𑄨", "𑄥𑄧𑄟𑄴", "𑄟𑄧𑄁𑄉𑄧𑄣𑄴", "𑄝𑄪𑄖𑄴", "𑄝𑄳𑄢𑄨𑄥𑄪𑄛𑄴", "𑄥𑄪𑄇𑄴𑄇𑄮𑄢𑄴", "𑄥𑄧𑄚𑄨"},
			daysNarrow:         []string{"𑄢𑄧", "𑄥𑄧", "𑄟𑄧", "𑄝𑄪", "𑄝𑄳𑄢𑄨", "𑄥𑄪", "𑄥𑄧"},
			daysWide:           []string{"𑄢𑄧𑄝𑄨𑄝𑄢𑄴", "𑄥𑄧𑄟𑄴𑄝𑄢𑄴", "𑄟𑄧𑄁𑄉𑄧𑄣𑄴𑄝𑄢𑄴", "𑄝𑄪𑄖𑄴𑄝𑄢𑄴", "𑄝𑄳𑄢𑄨𑄥𑄪𑄛𑄴𑄝𑄢𑄴", "𑄥𑄪𑄇𑄴𑄇𑄮𑄢𑄴𑄝𑄢𑄴", "𑄥𑄧𑄚𑄨𑄝𑄢𑄴"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"", ""},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ACST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "𑄃𑄑𑄴𑄣𑄚𑄴𑄖𑄨𑄉𑄮𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AEDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄪𑄉𑄬𑄘𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AEST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄪𑄉𑄬𑄘𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AKDT": "𑄃𑄣𑄌𑄴𑄇 𑄘𑄨𑄚𑄮𑄃𑄣𑄮 𑄃𑄧𑄇𑄴𑄖𑄧", "AKST": "𑄃𑄣𑄌𑄴𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ARST": "𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ART": "𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AST": "𑄃𑄑𑄴𑄣𑄚𑄴𑄖𑄨𑄉𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AWDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AWST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "BOT": "𑄝𑄮𑄣𑄨𑄞𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "BT": "𑄞𑄪𑄑𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CAT": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄜𑄳𑄢𑄨𑄇 𑄃𑄧𑄇𑄴𑄖𑄧", "CDT": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦 𑄘𑄨𑄚𑄮𑄃𑄣𑄮 𑄃𑄧𑄇𑄴𑄖𑄧", "CHADT": "𑄌𑄳𑄠𑄗𑄟𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CHAST": "𑄌𑄳𑄠𑄗𑄟𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CLST": "𑄌𑄨𑄣𑄨 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CLT": "𑄌𑄨𑄣𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "COST": "𑄇𑄧𑄣𑄧𑄟𑄴𑄝𑄨𑄠 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "COT": "𑄇𑄧𑄣𑄧𑄟𑄴𑄝𑄨𑄠 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CST": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ChST": "𑄌𑄟𑄬𑄢𑄮 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EAT": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄃𑄧𑄇𑄴𑄖𑄧", "ECT": "𑄃𑄨𑄇𑄪𑄠𑄬𑄓𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EDT": "𑄛𑄪𑄉𑄮 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EST": "𑄛𑄪𑄉𑄮 𑄞𑄨𑄘𑄬𑄢𑄴 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "GFT": "𑄜𑄧𑄢𑄥𑄨 𑄉𑄠𑄚 𑄃𑄧𑄇𑄴𑄖𑄧", "GMT": "𑄉𑄳𑄢𑄨𑄚𑄨𑄌𑄴 𑄟𑄨𑄚𑄴 𑄑𑄬𑄟𑄴", "GST": "𑄃𑄪𑄛𑄧𑄥𑄉𑄧𑄢𑄨𑄠𑄧 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "GYT": "𑄉𑄪𑄠𑄚 𑄃𑄧𑄇𑄴𑄖𑄧", "HADT": "𑄦𑄧𑄃𑄮𑄠𑄭-𑄃𑄣𑄬𑄃𑄪𑄖𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HAST": "𑄦𑄧𑄃𑄮𑄠𑄭-𑄃𑄣𑄬𑄃𑄪𑄖𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HAT": "𑄚𑄨𑄃𑄪𑄜𑄃𑄪𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HECU": "𑄇𑄨𑄃𑄪𑄝 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEEG": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HENOMX": "𑄃𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄛𑄧𑄏𑄨𑄟𑄴 𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEOG": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEPM": "𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄇𑄬𑄣𑄧𑄚𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEPMX": "𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HKST": "𑄦𑄧𑄁 𑄇𑄧𑄁 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HKT": "𑄦𑄧𑄁 𑄇𑄧𑄁 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNCU": "𑄇𑄨𑄃𑄪𑄝 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNEG": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNNOMX": "𑄃𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄛𑄧𑄏𑄨𑄟𑄴 𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNOG": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNPM": "𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄇𑄬𑄣𑄧𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNPMX": "𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNT": "𑄚𑄨𑄃𑄪𑄜𑄃𑄪𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "IST": "𑄃𑄨𑄚𑄴𑄘𑄨𑄠𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "JDT": "𑄎𑄛𑄚𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "JST": "𑄎𑄛𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "LHDT": "𑄣𑄧𑄢𑄴𑄓𑄴 𑄦𑄤𑄬 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "LHST": "𑄣𑄧𑄢𑄴𑄓𑄴 𑄦𑄤𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MDT": "𑄦𑄨𑄣𑄧𑄧𑄱 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MESZ": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MEZ": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MST": "𑄦𑄨𑄣𑄧𑄧𑄱 𑄞𑄨𑄘𑄬𑄢𑄴 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MYT": "𑄟𑄣𑄴𑄠𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "NZDT": "𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "NZST": "𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "OESZ": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "OEZ": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "PDT": "𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "PST": "𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄞𑄨𑄘𑄬𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SAST": "𑄘𑄧𑄉𑄨𑄚𑄴 𑄃𑄜𑄳𑄢𑄨𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SGT": "𑄥𑄨𑄁𑄉𑄛𑄪𑄢 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SRT": "𑄥𑄪𑄢𑄨𑄚𑄟𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "TMST": "𑄖𑄪𑄢𑄴𑄇𑄴𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "TMT": "𑄖𑄪𑄢𑄴𑄇𑄴𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "UYST": "𑄃𑄪𑄢𑄪𑄉𑄪𑄠𑄬 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "UYT": "𑄃𑄪𑄢𑄪𑄉𑄪𑄠𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "VET": "𑄞𑄬𑄚𑄬𑄎𑄪𑄠𑄬𑄣 𑄃𑄧𑄇𑄴𑄖𑄧", "WARST": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WART": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WAST": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WAT": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WESZ": "𑄛𑄧𑄏𑄬𑄟𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WEZ": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WIB": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "WIT": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "WITA": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄢𑄳𑄦 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "∅∅∅": "𑄝𑄳𑄢𑄥𑄨𑄣𑄨𑄠 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b870df2d704ca52917b45adf78d139de,
			fnFmtCurrency:         fnFmtCurrency_4d9675c6ee668a70f27c2127d8acae4e,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_81c5002a93b06b7f91045a20d95f77c9,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ccp_bd"] = func() locales.Translator {
		return &localen{

			locale:             "ccp_BD",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "𑄎𑄚𑄪", "𑄜𑄬𑄛𑄴", "𑄟𑄢𑄴𑄌𑄧", "𑄃𑄬𑄛𑄳𑄢𑄨𑄣𑄴", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪𑄣𑄭", "𑄃𑄉𑄧𑄌𑄴𑄑𑄴", "𑄥𑄬𑄛𑄴𑄑𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄃𑄧𑄇𑄴𑄑𑄮𑄝𑄧𑄢𑄴", "𑄚𑄧𑄞𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄓𑄨𑄥𑄬𑄟𑄴𑄝𑄢𑄴"},
			monthsNarrow:       []string{"", "𑄎", "𑄜𑄬", "𑄟", "𑄃𑄬", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪", "𑄃", "𑄥𑄬", "𑄃𑄧", "𑄚𑄧", "𑄓𑄨"},
			monthsWide:         []string{"", "𑄎𑄚𑄪𑄠𑄢𑄨", "𑄜𑄬𑄛𑄴𑄝𑄳𑄢𑄪𑄠𑄢𑄨", "𑄟𑄢𑄴𑄌𑄧", "𑄃𑄬𑄛𑄳𑄢𑄨𑄣𑄴", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪𑄣𑄭", "𑄃𑄉𑄧𑄌𑄴𑄑𑄴", "𑄥𑄬𑄛𑄴𑄑𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄃𑄧𑄇𑄴𑄑𑄬𑄝𑄧𑄢𑄴", "𑄚𑄧𑄞𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄓𑄨𑄥𑄬𑄟𑄴𑄝𑄧𑄢𑄴"},
			daysAbbreviated:    []string{"𑄢𑄧𑄝𑄨", "𑄥𑄧𑄟𑄴", "𑄟𑄧𑄁𑄉𑄧𑄣𑄴", "𑄝𑄪𑄖𑄴", "𑄝𑄳𑄢𑄨𑄥𑄪𑄛𑄴", "𑄥𑄪𑄇𑄴𑄇𑄮𑄢𑄴", "𑄥𑄧𑄚𑄨"},
			daysNarrow:         []string{"𑄢𑄧", "𑄥𑄧", "𑄟𑄧", "𑄝𑄪", "𑄝𑄳𑄢𑄨", "𑄥𑄪", "𑄥𑄧"},
			daysWide:           []string{"𑄢𑄧𑄝𑄨𑄝𑄢𑄴", "𑄥𑄧𑄟𑄴𑄝𑄢𑄴", "𑄟𑄧𑄁𑄉𑄧𑄣𑄴𑄝𑄢𑄴", "𑄝𑄪𑄖𑄴𑄝𑄢𑄴", "𑄝𑄳𑄢𑄨𑄥𑄪𑄛𑄴𑄝𑄢𑄴", "𑄥𑄪𑄇𑄴𑄇𑄮𑄢𑄴𑄝𑄢𑄴", "𑄥𑄧𑄚𑄨𑄝𑄢𑄴"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"", ""},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ACST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "𑄃𑄑𑄴𑄣𑄚𑄴𑄖𑄨𑄉𑄮𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AEDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄪𑄉𑄬𑄘𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AEST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄪𑄉𑄬𑄘𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AKDT": "𑄃𑄣𑄌𑄴𑄇 𑄘𑄨𑄚𑄮𑄃𑄣𑄮 𑄃𑄧𑄇𑄴𑄖𑄧", "AKST": "𑄃𑄣𑄌𑄴𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ARST": "𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ART": "𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AST": "𑄃𑄑𑄴𑄣𑄚𑄴𑄖𑄨𑄉𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AWDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AWST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "BOT": "𑄝𑄮𑄣𑄨𑄞𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "BT": "𑄞𑄪𑄑𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CAT": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄜𑄳𑄢𑄨𑄇 𑄃𑄧𑄇𑄴𑄖𑄧", "CDT": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦 𑄘𑄨𑄚𑄮𑄃𑄣𑄮 𑄃𑄧𑄇𑄴𑄖𑄧", "CHADT": "𑄌𑄳𑄠𑄗𑄟𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CHAST": "𑄌𑄳𑄠𑄗𑄟𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CLST": "𑄌𑄨𑄣𑄨 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CLT": "𑄌𑄨𑄣𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "COST": "𑄇𑄧𑄣𑄧𑄟𑄴𑄝𑄨𑄠 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "COT": "𑄇𑄧𑄣𑄧𑄟𑄴𑄝𑄨𑄠 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CST": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ChST": "𑄌𑄟𑄬𑄢𑄮 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EAT": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄃𑄧𑄇𑄴𑄖𑄧", "ECT": "𑄃𑄨𑄇𑄪𑄠𑄬𑄓𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EDT": "𑄛𑄪𑄉𑄮 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EST": "𑄛𑄪𑄉𑄮 𑄞𑄨𑄘𑄬𑄢𑄴 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "GFT": "𑄜𑄧𑄢𑄥𑄨 𑄉𑄠𑄚 𑄃𑄧𑄇𑄴𑄖𑄧", "GMT": "𑄉𑄳𑄢𑄨𑄚𑄨𑄌𑄴 𑄟𑄨𑄚𑄴 𑄑𑄬𑄟𑄴", "GST": "𑄃𑄪𑄛𑄧𑄥𑄉𑄧𑄢𑄨𑄠𑄧 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "GYT": "𑄉𑄪𑄠𑄚 𑄃𑄧𑄇𑄴𑄖𑄧", "HADT": "𑄦𑄧𑄃𑄮𑄠𑄭-𑄃𑄣𑄬𑄃𑄪𑄖𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HAST": "𑄦𑄧𑄃𑄮𑄠𑄭-𑄃𑄣𑄬𑄃𑄪𑄖𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HAT": "𑄚𑄨𑄃𑄪𑄜𑄃𑄪𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HECU": "𑄇𑄨𑄃𑄪𑄝 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEEG": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HENOMX": "𑄃𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄛𑄧𑄏𑄨𑄟𑄴 𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEOG": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEPM": "𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄇𑄬𑄣𑄧𑄚𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEPMX": "𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HKST": "𑄦𑄧𑄁 𑄇𑄧𑄁 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HKT": "𑄦𑄧𑄁 𑄇𑄧𑄁 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNCU": "𑄇𑄨𑄃𑄪𑄝 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNEG": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNNOMX": "𑄃𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄛𑄧𑄏𑄨𑄟𑄴 𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNOG": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNPM": "𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄇𑄬𑄣𑄧𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNPMX": "𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNT": "𑄚𑄨𑄃𑄪𑄜𑄃𑄪𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "IST": "𑄃𑄨𑄚𑄴𑄘𑄨𑄠𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "JDT": "𑄎𑄛𑄚𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "JST": "𑄎𑄛𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "LHDT": "𑄣𑄧𑄢𑄴𑄓𑄴 𑄦𑄤𑄬 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "LHST": "𑄣𑄧𑄢𑄴𑄓𑄴 𑄦𑄤𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MDT": "𑄦𑄨𑄣𑄧𑄧𑄱 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MESZ": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MEZ": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MST": "𑄦𑄨𑄣𑄧𑄧𑄱 𑄞𑄨𑄘𑄬𑄢𑄴 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MYT": "𑄟𑄣𑄴𑄠𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "NZDT": "𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "NZST": "𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "OESZ": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "OEZ": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "PDT": "𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "PST": "𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄞𑄨𑄘𑄬𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SAST": "𑄘𑄧𑄉𑄨𑄚𑄴 𑄃𑄜𑄳𑄢𑄨𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SGT": "𑄥𑄨𑄁𑄉𑄛𑄪𑄢 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SRT": "𑄥𑄪𑄢𑄨𑄚𑄟𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "TMST": "𑄖𑄪𑄢𑄴𑄇𑄴𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "TMT": "𑄖𑄪𑄢𑄴𑄇𑄴𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "UYST": "𑄃𑄪𑄢𑄪𑄉𑄪𑄠𑄬 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "UYT": "𑄃𑄪𑄢𑄪𑄉𑄪𑄠𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "VET": "𑄞𑄬𑄚𑄬𑄎𑄪𑄠𑄬𑄣 𑄃𑄧𑄇𑄴𑄖𑄧", "WARST": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WART": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WAST": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WAT": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WESZ": "𑄛𑄧𑄏𑄬𑄟𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WEZ": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WIB": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "WIT": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "WITA": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄢𑄳𑄦 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "∅∅∅": "𑄃𑄬𑄎𑄮𑄢𑄬𑄌𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b870df2d704ca52917b45adf78d139de,
			fnFmtCurrency:         fnFmtCurrency_4d9675c6ee668a70f27c2127d8acae4e,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_81c5002a93b06b7f91045a20d95f77c9,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ccp_in"] = func() locales.Translator {
		return &localen{

			locale:             "ccp_IN",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "𑄎𑄚𑄪", "𑄜𑄬𑄛𑄴", "𑄟𑄢𑄴𑄌𑄧", "𑄃𑄬𑄛𑄳𑄢𑄨𑄣𑄴", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪𑄣𑄭", "𑄃𑄉𑄧𑄌𑄴𑄑𑄴", "𑄥𑄬𑄛𑄴𑄑𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄃𑄧𑄇𑄴𑄑𑄮𑄝𑄧𑄢𑄴", "𑄚𑄧𑄞𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄓𑄨𑄥𑄬𑄟𑄴𑄝𑄢𑄴"},
			monthsNarrow:       []string{"", "𑄎", "𑄜𑄬", "𑄟", "𑄃𑄬", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪", "𑄃", "𑄥𑄬", "𑄃𑄧", "𑄚𑄧", "𑄓𑄨"},
			monthsWide:         []string{"", "𑄎𑄚𑄪𑄠𑄢𑄨", "𑄜𑄬𑄛𑄴𑄝𑄳𑄢𑄪𑄠𑄢𑄨", "𑄟𑄢𑄴𑄌𑄧", "𑄃𑄬𑄛𑄳𑄢𑄨𑄣𑄴", "𑄟𑄬", "𑄎𑄪𑄚𑄴", "𑄎𑄪𑄣𑄭", "𑄃𑄉𑄧𑄌𑄴𑄑𑄴", "𑄥𑄬𑄛𑄴𑄑𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄃𑄧𑄇𑄴𑄑𑄬𑄝𑄧𑄢𑄴", "𑄚𑄧𑄞𑄬𑄟𑄴𑄝𑄧𑄢𑄴", "𑄓𑄨𑄥𑄬𑄟𑄴𑄝𑄧𑄢𑄴"},
			daysAbbreviated:    []string{"𑄢𑄧𑄝𑄨", "𑄥𑄧𑄟𑄴", "𑄟𑄧𑄁𑄉𑄧𑄣𑄴", "𑄝𑄪𑄖𑄴", "𑄝𑄳𑄢𑄨𑄥𑄪𑄛𑄴", "𑄥𑄪𑄇𑄴𑄇𑄮𑄢𑄴", "𑄥𑄧𑄚𑄨"},
			daysNarrow:         []string{"𑄢𑄧", "𑄥𑄧", "𑄟𑄧", "𑄝𑄪", "𑄝𑄳𑄢𑄨", "𑄥𑄪", "𑄥𑄧"},
			daysWide:           []string{"𑄢𑄧𑄝𑄨𑄝𑄢𑄴", "𑄥𑄧𑄟𑄴𑄝𑄢𑄴", "𑄟𑄧𑄁𑄉𑄧𑄣𑄴𑄝𑄢𑄴", "𑄝𑄪𑄖𑄴𑄝𑄢𑄴", "𑄝𑄳𑄢𑄨𑄥𑄪𑄛𑄴𑄝𑄢𑄴", "𑄥𑄪𑄇𑄴𑄇𑄮𑄢𑄴𑄝𑄢𑄴", "𑄥𑄧𑄚𑄨𑄝𑄢𑄴"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"", ""},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ACST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "𑄃𑄑𑄴𑄣𑄚𑄴𑄖𑄨𑄉𑄮𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AEDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄪𑄉𑄬𑄘𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AEST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄪𑄉𑄬𑄘𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AKDT": "𑄃𑄣𑄌𑄴𑄇 𑄘𑄨𑄚𑄮𑄃𑄣𑄮 𑄃𑄧𑄇𑄴𑄖𑄧", "AKST": "𑄃𑄣𑄌𑄴𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ARST": "𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ART": "𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AST": "𑄃𑄑𑄴𑄣𑄚𑄴𑄖𑄨𑄉𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AWDT": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "AWST": "𑄃𑄧𑄌𑄴𑄑𑄳𑄢𑄬𑄣𑄨𑄠𑄧 𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "BOT": "𑄝𑄮𑄣𑄨𑄞𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "BT": "𑄞𑄪𑄑𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CAT": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄜𑄳𑄢𑄨𑄇 𑄃𑄧𑄇𑄴𑄖𑄧", "CDT": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦 𑄘𑄨𑄚𑄮𑄃𑄣𑄮 𑄃𑄧𑄇𑄴𑄖𑄧", "CHADT": "𑄌𑄳𑄠𑄗𑄟𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CHAST": "𑄌𑄳𑄠𑄗𑄟𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CLST": "𑄌𑄨𑄣𑄨 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CLT": "𑄌𑄨𑄣𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "COST": "𑄇𑄧𑄣𑄧𑄟𑄴𑄝𑄨𑄠 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "COT": "𑄇𑄧𑄣𑄧𑄟𑄴𑄝𑄨𑄠 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "CST": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄧𑄢𑄳𑄦 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "ChST": "𑄌𑄟𑄬𑄢𑄮 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EAT": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄃𑄧𑄇𑄴𑄖𑄧", "ECT": "𑄃𑄨𑄇𑄪𑄠𑄬𑄓𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EDT": "𑄛𑄪𑄉𑄮 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "EST": "𑄛𑄪𑄉𑄮 𑄞𑄨𑄘𑄬𑄢𑄴 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "GFT": "𑄜𑄧𑄢𑄥𑄨 𑄉𑄠𑄚 𑄃𑄧𑄇𑄴𑄖𑄧", "GMT": "𑄉𑄳𑄢𑄨𑄚𑄨𑄌𑄴 𑄟𑄨𑄚𑄴 𑄑𑄬𑄟𑄴", "GST": "𑄃𑄪𑄛𑄧𑄥𑄉𑄧𑄢𑄨𑄠𑄧 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "GYT": "𑄉𑄪𑄠𑄚 𑄃𑄧𑄇𑄴𑄖𑄧", "HADT": "𑄦𑄧𑄃𑄮𑄠𑄭-𑄃𑄣𑄬𑄃𑄪𑄖𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HAST": "𑄦𑄧𑄃𑄮𑄠𑄭-𑄃𑄣𑄬𑄃𑄪𑄖𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HAT": "𑄚𑄨𑄃𑄪𑄜𑄃𑄪𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HECU": "𑄇𑄨𑄃𑄪𑄝 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEEG": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HENOMX": "𑄃𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄛𑄧𑄏𑄨𑄟𑄴 𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEOG": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEPM": "𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄇𑄬𑄣𑄧𑄚𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HEPMX": "𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HKST": "𑄦𑄧𑄁 𑄇𑄧𑄁 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HKT": "𑄦𑄧𑄁 𑄇𑄧𑄁 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNCU": "𑄇𑄨𑄃𑄪𑄝 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNEG": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNNOMX": "𑄃𑄪𑄖𑄴𑄖𑄮𑄢𑄴 𑄛𑄧𑄏𑄨𑄟𑄴 𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNOG": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄉𑄳𑄢𑄨𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNPM": "𑄥𑄬𑄚𑄴𑄑𑄴 𑄛𑄨𑄠𑄬𑄢𑄴 𑄃𑄮 𑄟𑄨𑄇𑄬𑄣𑄧𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNPMX": "𑄟𑄬𑄇𑄴𑄥𑄨𑄇𑄚𑄴 𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "HNT": "𑄚𑄨𑄃𑄪𑄜𑄃𑄪𑄚𑄴𑄣𑄳𑄠𑄚𑄴𑄓𑄨 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "IST": "𑄃𑄨𑄚𑄴𑄘𑄨𑄠𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "JDT": "𑄎𑄛𑄚𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "JST": "𑄎𑄛𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "LHDT": "𑄣𑄧𑄢𑄴𑄓𑄴 𑄦𑄤𑄬 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "LHST": "𑄣𑄧𑄢𑄴𑄓𑄴 𑄦𑄤𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MDT": "𑄦𑄨𑄣𑄧𑄧𑄱 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MESZ": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MEZ": "𑄟𑄧𑄖𑄴𑄙𑄳𑄠 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MST": "𑄦𑄨𑄣𑄧𑄧𑄱 𑄞𑄨𑄘𑄬𑄢𑄴 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "MYT": "𑄟𑄣𑄴𑄠𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "NZDT": "𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄘𑄨𑄚𑄮𑄃𑄣𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "NZST": "𑄚𑄨𑄃𑄪𑄎𑄨𑄣𑄳𑄠𑄚𑄴𑄓𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "OESZ": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "OEZ": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "PDT": "𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄞𑄨𑄘𑄬𑄢𑄴 𑄘𑄨𑄚𑄮𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "PST": "𑄛𑄳𑄢𑄧𑄥𑄚𑄴𑄖𑄧 𑄟𑄧𑄦𑄥𑄉𑄧𑄢𑄧𑄢𑄴 𑄞𑄨𑄘𑄬𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SAST": "𑄘𑄧𑄉𑄨𑄚𑄴 𑄃𑄜𑄳𑄢𑄨𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SGT": "𑄥𑄨𑄁𑄉𑄛𑄪𑄢 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "SRT": "𑄥𑄪𑄢𑄨𑄚𑄟𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "TMST": "𑄖𑄪𑄢𑄴𑄇𑄴𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "TMT": "𑄖𑄪𑄢𑄴𑄇𑄴𑄟𑄬𑄚𑄨𑄌𑄴𑄖𑄚𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "UYST": "𑄃𑄪𑄢𑄪𑄉𑄪𑄠𑄬 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "UYT": "𑄃𑄪𑄢𑄪𑄉𑄪𑄠𑄬 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "VET": "𑄞𑄬𑄚𑄬𑄎𑄪𑄠𑄬𑄣 𑄃𑄧𑄇𑄴𑄖𑄧", "WARST": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WART": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄢𑄴𑄎𑄬𑄚𑄴𑄑𑄨𑄚 𑄛𑄳𑄢𑄧𑄟𑄚𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WAST": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WAT": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄜𑄳𑄢𑄨𑄇 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WESZ": "𑄛𑄧𑄏𑄬𑄟𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WEZ": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄨𑄃𑄪𑄢𑄮𑄝𑄮𑄢𑄴 𑄟𑄚𑄧𑄇𑄴 𑄃𑄧𑄇𑄴𑄖𑄧", "WIB": "𑄛𑄧𑄏𑄨𑄟𑄬𑄘𑄨 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "WIT": "𑄛𑄪𑄉𑄬𑄘𑄨 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "WITA": "𑄃𑄏𑄧𑄣𑄴 𑄉𑄢𑄳𑄦 𑄃𑄨𑄚𑄴𑄘𑄮𑄚𑄬𑄥𑄨𑄠 𑄃𑄧𑄇𑄴𑄖𑄧", "∅∅∅": "𑄝𑄳𑄢𑄥𑄨𑄣𑄨𑄠 𑄉𑄧𑄢𑄧𑄟𑄴𑄇𑄣𑄧𑄢𑄴 𑄃𑄧𑄇𑄴𑄖𑄧"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b870df2d704ca52917b45adf78d139de,
			fnFmtCurrency:         fnFmtCurrency_4d9675c6ee668a70f27c2127d8acae4e,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_81c5002a93b06b7f91045a20d95f77c9,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ce"] = func() locales.Translator {
		return &localen{

			locale:                 "ce",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январь", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"},
			daysAbbreviated:        []string{"кӀи", "ор", "ши", "кха", "еа", "пӀе", "шуо"},
			daysNarrow:             []string{"кӀи", "ор", "ши", "кха", "еа", "пӀе", "шуо"},
			daysShort:              []string{"кӀи", "ор", "ши", "кха", "еа", "пӀе", "шуо"},
			daysWide:               []string{"кӀира", "оршот", "шинара", "кхаара", "еара", "пӀераска", "шуот"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ӏийса пайхамар вина де кхачале", "Ӏийса пайхамар вина дийнахь дуьйна"},
			timezones:              map[string]string{"ACDT": "Юккъера Австрали, аьхкенан хан", "ACST": "Юккъера Австрали, стандартан хан", "ACWDT": "Юккъера Австрали, малхбузен аьхкенан хан", "ACWST": "Юккъера Австрали, малхбузен стандартан хан", "ADT": "Атлантикан аьхкенан хан", "AEDT": "Малхбален Австрали, аьхкенан хан", "AEST": "Малхбален Австрали, стандартан хан", "AKDT": "Аляска, аьхкенан хан", "AKST": "Аляска, стандартан хан", "ARST": "Аргентина, аьхкенан хан", "ART": "Аргентина, стандартан хан", "AST": "Атлантикан стандартан хан", "AWDT": "Малхбузен Австрали, аьхкенан хан", "AWST": "Малхбузен Австрали, стандартан хан", "BOT": "Боливи", "BT": "Бутан", "CAT": "Юккъера Африка", "CDT": "Юккъера Америка, аьхкенан хан", "CHADT": "Чатем, аьхкенан хан", "CHAST": "Чатем, стандартан хан", "CLST": "Чили, аьхкенан хан", "CLT": "Чили, стандартан хан", "COST": "Колумби, аьхкенан хан", "COT": "Колумби, стандартан хан", "CST": "Юккъера Америка, стандартан хан", "ChST": "Чаморро", "EAT": "Малхбален Африка", "ECT": "Эквадор", "EDT": "Малхбален Америка, аьхкенан хан", "EST": "Малхбален Америка, стандартан хан", "GFT": "Французийн Гвиана", "GMT": "Гринвичица юкъара хан", "GST": "ГӀажарийн айма", "GYT": "Гайана", "HADT": "Гавайн-алеутийн аьхкенан хан", "HAST": "Гавайн-алеутийн стандартан хан", "HAT": "Ньюфаундленд, аьхкенан хан", "HECU": "Куба, аьхкенан хан", "HEEG": "Малхбален Гренланди, аьхкенан хан", "HENOMX": "Къилбаседа Американ Мексикан аьхкенан хан", "HEOG": "Малхбузен Гренланди, аьхкенан хан", "HEPM": "Сен-Пьер а, Микелон а, аьхкенан хан", "HEPMX": "Тийна океанан Мексикан аьхкенан хан", "HKST": "Гонконг, аьхкенан хан", "HKT": "Гонконг, стандартан хан", "HNCU": "Куба, стандартан хан", "HNEG": "Малхбален Гренланди, стандартан хан", "HNNOMX": "Къилбаседа Американ Мексикан стандартан хан", "HNOG": "Малхбузен Гренланди, стандартан хан", "HNPM": "Сен-Пьер а, Микелон а, стандартан хан", "HNPMX": "Тийна океанан Мексикан стандартан хан", "HNT": "Ньюфаундленд, стандартан хан", "IST": "ХӀинди", "JDT": "Япон, аьхкенан хан", "JST": "Япон, стандартан хан", "LHDT": "Лорд-Хау, аьхкенан хан", "LHST": "Лорд-Хау, стандартан хан", "MDT": "Лаьмнийн аьхкенан хан (АЦШ)", "MESZ": "Юккъера Европа, аьхкенан хан", "MEZ": "Юккъера Европа, стандартан хан", "MST": "Лаьмнийн стандартан хан (АЦШ)", "MYT": "Малайзи", "NZDT": "Керла Зеланди, аьхкенан хан", "NZST": "Керла Зеланди, стандартан хан", "OESZ": "Малхбален Европа, аьхкенан хан", "OEZ": "Малхбален Европа, стандартан хан", "PDT": "Тийна океанан аьхкенан хан", "PST": "Тийна океанан стандартан хан", "SAST": "Къилба Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменин аьхкенан хан", "TMT": "Туркменин стандартан хан", "UYST": "Уругвай, аьхкенан хан", "UYT": "Уругвай, стандартан хан", "VET": "Венесуэла", "WARST": "Малхбузен Аргентина, аьхкенан хан", "WART": "Малхбузен Аргентина, стандартан хан", "WAST": "Малхбузен Африка, аьхкенан хан", "WAT": "Малхбузен Африка, стандартан хан", "WESZ": "Малхбузен Европа, аьхкенан хан", "WEZ": "Малхбузен Европа, стандартан хан", "WIB": "Малхбузен Индонези", "WIT": "Малхбален Индонези", "WITA": "Юккъера Индонези", "∅∅∅": "Перу, аьхкенан хан"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18,
			fnFmtDateLong:         fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d,
			fnFmtDateMedium:       fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ce_ru"] = func() locales.Translator {
		return &localen{

			locale:                 "ce_RU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январь", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"},
			daysAbbreviated:        []string{"кӀи", "ор", "ши", "кха", "еа", "пӀе", "шуо"},
			daysNarrow:             []string{"кӀи", "ор", "ши", "кха", "еа", "пӀе", "шуо"},
			daysShort:              []string{"кӀи", "ор", "ши", "кха", "еа", "пӀе", "шуо"},
			daysWide:               []string{"кӀира", "оршот", "шинара", "кхаара", "еара", "пӀераска", "шуот"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ӏийса пайхамар вина де кхачале", "Ӏийса пайхамар вина дийнахь дуьйна"},
			timezones:              map[string]string{"ACDT": "Юккъера Австрали, аьхкенан хан", "ACST": "Юккъера Австрали, стандартан хан", "ACWDT": "Юккъера Австрали, малхбузен аьхкенан хан", "ACWST": "Юккъера Австрали, малхбузен стандартан хан", "ADT": "Атлантикан аьхкенан хан", "AEDT": "Малхбален Австрали, аьхкенан хан", "AEST": "Малхбален Австрали, стандартан хан", "AKDT": "Аляска, аьхкенан хан", "AKST": "Аляска, стандартан хан", "ARST": "Аргентина, аьхкенан хан", "ART": "Аргентина, стандартан хан", "AST": "Атлантикан стандартан хан", "AWDT": "Малхбузен Австрали, аьхкенан хан", "AWST": "Малхбузен Австрали, стандартан хан", "BOT": "Боливи", "BT": "Бутан", "CAT": "Юккъера Африка", "CDT": "Юккъера Америка, аьхкенан хан", "CHADT": "Чатем, аьхкенан хан", "CHAST": "Чатем, стандартан хан", "CLST": "Чили, аьхкенан хан", "CLT": "Чили, стандартан хан", "COST": "Колумби, аьхкенан хан", "COT": "Колумби, стандартан хан", "CST": "Юккъера Америка, стандартан хан", "ChST": "Чаморро", "EAT": "Малхбален Африка", "ECT": "Эквадор", "EDT": "Малхбален Америка, аьхкенан хан", "EST": "Малхбален Америка, стандартан хан", "GFT": "Французийн Гвиана", "GMT": "Гринвичица юкъара хан", "GST": "ГӀажарийн айма", "GYT": "Гайана", "HADT": "Гавайн-алеутийн аьхкенан хан", "HAST": "Гавайн-алеутийн стандартан хан", "HAT": "Ньюфаундленд, аьхкенан хан", "HECU": "Куба, аьхкенан хан", "HEEG": "Малхбален Гренланди, аьхкенан хан", "HENOMX": "Къилбаседа Американ Мексикан аьхкенан хан", "HEOG": "Малхбузен Гренланди, аьхкенан хан", "HEPM": "Сен-Пьер а, Микелон а, аьхкенан хан", "HEPMX": "Тийна океанан Мексикан аьхкенан хан", "HKST": "Гонконг, аьхкенан хан", "HKT": "Гонконг, стандартан хан", "HNCU": "Куба, стандартан хан", "HNEG": "Малхбален Гренланди, стандартан хан", "HNNOMX": "Къилбаседа Американ Мексикан стандартан хан", "HNOG": "Малхбузен Гренланди, стандартан хан", "HNPM": "Сен-Пьер а, Микелон а, стандартан хан", "HNPMX": "Тийна океанан Мексикан стандартан хан", "HNT": "Ньюфаундленд, стандартан хан", "IST": "ХӀинди", "JDT": "Япон, аьхкенан хан", "JST": "Япон, стандартан хан", "LHDT": "Лорд-Хау, аьхкенан хан", "LHST": "Лорд-Хау, стандартан хан", "MDT": "Лаьмнийн аьхкенан хан (АЦШ)", "MESZ": "Юккъера Европа, аьхкенан хан", "MEZ": "Юккъера Европа, стандартан хан", "MST": "Лаьмнийн стандартан хан (АЦШ)", "MYT": "Малайзи", "NZDT": "Керла Зеланди, аьхкенан хан", "NZST": "Керла Зеланди, стандартан хан", "OESZ": "Малхбален Европа, аьхкенан хан", "OEZ": "Малхбален Европа, стандартан хан", "PDT": "Тийна океанан аьхкенан хан", "PST": "Тийна океанан стандартан хан", "SAST": "Къилба Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменин аьхкенан хан", "TMT": "Туркменин стандартан хан", "UYST": "Уругвай, аьхкенан хан", "UYT": "Уругвай, стандартан хан", "VET": "Венесуэла", "WARST": "Малхбузен Аргентина, аьхкенан хан", "WART": "Малхбузен Аргентина, стандартан хан", "WAST": "Малхбузен Африка, аьхкенан хан", "WAT": "Малхбузен Африка, стандартан хан", "WESZ": "Малхбузен Европа, аьхкенан хан", "WEZ": "Малхбузен Европа, стандартан хан", "WIB": "Малхбузен Индонези", "WIT": "Малхбален Индонези", "WITA": "Юккъера Индонези", "∅∅∅": "Бразили, аьхкенан хан"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18,
			fnFmtDateLong:         fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d,
			fnFmtDateMedium:       fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ceb"] = func() locales.Translator {
		return &localen{

			locale:                 "ceb",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US $", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "En", "Peb", "Mar", "Apr", "May", "Hun", "Hul", "Ag", "Set", "Okt", "Nob", "Dis"},
			monthsNarrow:           []string{"", "E", "P", "M", "A", "M", "H", "H", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Setyembre", "Oktubre", "Nobyembre", "Disyembre"},
			daysAbbreviated:        []string{"Dom", "Lun", "Mar", "Mks", "Hu", "Bi", "Sa"},
			daysNarrow:             []string{"D", "L", "M", "M", "H", "B", "S"},
			daysShort:              []string{"Dom", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"},
			daysWide:               []string{"Domingo", "Lunes", "Martes", "Miyerkules", "Huwebes", "Biyernes", "Sabado"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Sa Wala Pa ang Common Era", "Common Era"},
			timezones:              map[string]string{"ACDT": "Oras sa Adlawan sa Central Australia", "ACST": "Tamdanang Oras sa Central Australia", "ACWDT": "Oras sa Adlawan sa Central Western Australia", "ACWST": "Tamdanang Oras sa Central Western Australia", "ADT": "Norte Amerika Oras sa Kahayag sa Adlaw sa Atlantiko", "AEDT": "Oras sa Adlawan sa Eastern Australia", "AEST": "Tamdanang Oras sa Eastern Australia", "AKDT": "Oras sa Adlawan sa Alaska", "AKST": "Tamdanang Oras sa Alaska", "ARST": "Oras sa Tag-init sa Argentina", "ART": "Tamdanang Oras sa Argentina", "AST": "Norte Amerika Sukdanang Oras sa Atlantiko", "AWDT": "Oras sa Adlawan sa Western Australia", "AWST": "Tamdanang Oras sa Western Australia", "BOT": "Oras sa Bolivia", "BT": "Oras sa Bhutan", "CAT": "Oras sa Central Africa", "CDT": "Taliwala sa Kahayag sa Adlawng Oras Norte Amerika", "CHADT": "Oras sa Adlawan sa Chatham", "CHAST": "Tamdanang Oras sa Chatham", "CLST": "Oras sa Tag-init sa Chile", "CLT": "Tamdanang Oras sa Chile", "COST": "Oras sa Tag-init sa Colombia", "COT": "Tamdanang Oras sa Colombia", "CST": "Taliwala nga Sukdanang Oras Norte Amerika", "ChST": "Tamdanang Oras sa Chamorro", "EAT": "Oras sa East Africa", "ECT": "Oras sa Ecuador", "EDT": "Norte Amerika Oras sa Sidlakan sa Kahayag sa Adlaw", "EST": "Norte Amerika Sukdanang Oras", "GFT": "Oras sa French Guiana", "GMT": "Kasarangang Oras sa Greenwich", "GST": "Tamdanang Oras sa Gulf", "GYT": "Oras sa Guyana", "HADT": "Oras sa Adlawan sa Hawaii-Aleutian", "HAST": "Tamdanang Oras sa Hawaii-Aleutian", "HAT": "Oras sa Adlawan sa Newfoundland", "HECU": "Oras sa Adlawan sa Cuba", "HEEG": "Oras sa Tag-init sa East Greenland", "HENOMX": "Oras sa Adlawan sa Northwest Mexico", "HEOG": "Oras sa Tag-init sa West Greenland", "HEPM": "Oras sa Adlawan sa St. Pierre & Miquelon", "HEPMX": "Oras sa Adlawan sa Mexican Pacific", "HKST": "Oras sa Tag-init sa Hong Kong", "HKT": "Tamdanang Oras sa Hong Kong", "HNCU": "Tamdanang Oras sa Cuba", "HNEG": "Tamdanang Oras sa East Greenland", "HNNOMX": "Tamdanang Oras sa Northwest Mexico", "HNOG": "Tamdanang Oras sa West Greenland", "HNPM": "Tamdanang Oras sa St. Pierre & Miquelon", "HNPMX": "Tamdanang Oras sa Mexican Pacific", "HNT": "Tamdanang Oras sa Newfoundland", "IST": "Tamdanang Oras sa India", "JDT": "Oras sa Adlawan sa Japan", "JST": "Tamdanang Oras sa Japan", "LHDT": "Oras sa Adlawan sa Lord Howe", "LHST": "Tamdanang Oras sa Lord Howe", "MDT": "Norte Amerika Oras sa Kahayag sa Adlaw sa Kabukiran", "MESZ": "Oras sa Europa Taliwala sa Adlaw sa Tag-init", "MEZ": "Sukdanang Oras sa Europa sa Taliwala sa Adlaw", "MST": "Norte Amerika Sukdanang Oras sa Kabukiran", "MYT": "Oras sa Malaysia", "NZDT": "Oras sa Adlawan sa New Zealand", "NZST": "Tamdanang Oras sa New Zealand", "OESZ": "Oras sa Taga-Oryente sa Europa sa Tag-int", "OEZ": "Sukdanang Oras sa Taga-Oryente sa Europa", "PDT": "Norte Amerika Oras sa Kahayag sa Adlaw sa Pasipiko", "PST": "Norte Amerika Sukdanang Oras sa Pasipiko", "SAST": "Tamdanang Oras sa South Africa", "SGT": "Tamdanang Oras sa Singapore", "SRT": "Oras sa Suriname", "TMST": "Oras sa Tag-init sa Turkmenistan", "TMT": "Tamdanang Oras sa Turkmenistan", "UYST": "Oras sa Tag-init sa Uruguay", "UYT": "Tamdanang Oras sa Uruguay", "VET": "Oras sa Venezuela", "WARST": "Oras sa Tag-init sa Western Argentina", "WART": "Tamdanang Oras sa Western Argentina", "WAST": "Oras sa Tag-init sa West Africa", "WAT": "Tamdanang Oras sa West Africa", "WESZ": "Oras sa Kasadpanon sa Europa sa Tag-init", "WEZ": "Sukadanang Oras sa Kasadpanon sa Europa", "WIB": "Oras sa Western Indonesia", "WIT": "Oras sa Eastern Indonesia", "WITA": "Oras sa Central Indonesia", "∅∅∅": "Oras sa Tag-init sa Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c1a7400a6209ae7e8eb28592c58f6657,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ceb_ph"] = func() locales.Translator {
		return &localen{

			locale:                 "ceb_PH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "En", "Peb", "Mar", "Apr", "May", "Hun", "Hul", "Ag", "Set", "Okt", "Nob", "Dis"},
			monthsNarrow:           []string{"", "E", "P", "M", "A", "M", "H", "H", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Setyembre", "Oktubre", "Nobyembre", "Disyembre"},
			daysAbbreviated:        []string{"Dom", "Lun", "Mar", "Mks", "Hu", "Bi", "Sa"},
			daysNarrow:             []string{"D", "L", "M", "M", "H", "B", "S"},
			daysShort:              []string{"Dom", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"},
			daysWide:               []string{"Domingo", "Lunes", "Martes", "Miyerkules", "Huwebes", "Biyernes", "Sabado"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Sa Wala Pa ang Common Era", "Common Era"},
			timezones:              map[string]string{"ACDT": "Oras sa Adlawan sa Central Australia", "ACST": "Tamdanang Oras sa Central Australia", "ACWDT": "Oras sa Adlawan sa Central Western Australia", "ACWST": "Tamdanang Oras sa Central Western Australia", "ADT": "Norte Amerika Oras sa Kahayag sa Adlaw sa Atlantiko", "AEDT": "Oras sa Adlawan sa Eastern Australia", "AEST": "Tamdanang Oras sa Eastern Australia", "AKDT": "Oras sa Adlawan sa Alaska", "AKST": "Tamdanang Oras sa Alaska", "ARST": "Oras sa Tag-init sa Argentina", "ART": "Tamdanang Oras sa Argentina", "AST": "Norte Amerika Sukdanang Oras sa Atlantiko", "AWDT": "Oras sa Adlawan sa Western Australia", "AWST": "Tamdanang Oras sa Western Australia", "BOT": "Oras sa Bolivia", "BT": "Oras sa Bhutan", "CAT": "Oras sa Central Africa", "CDT": "Taliwala sa Kahayag sa Adlawng Oras Norte Amerika", "CHADT": "Oras sa Adlawan sa Chatham", "CHAST": "Tamdanang Oras sa Chatham", "CLST": "Oras sa Tag-init sa Chile", "CLT": "Tamdanang Oras sa Chile", "COST": "Oras sa Tag-init sa Colombia", "COT": "Tamdanang Oras sa Colombia", "CST": "Taliwala nga Sukdanang Oras Norte Amerika", "ChST": "Tamdanang Oras sa Chamorro", "EAT": "Oras sa East Africa", "ECT": "Oras sa Ecuador", "EDT": "Norte Amerika Oras sa Sidlakan sa Kahayag sa Adlaw", "EST": "Norte Amerika Sukdanang Oras", "GFT": "Oras sa French Guiana", "GMT": "Kasarangang Oras sa Greenwich", "GST": "Tamdanang Oras sa Gulf", "GYT": "Oras sa Guyana", "HADT": "Oras sa Adlawan sa Hawaii-Aleutian", "HAST": "Tamdanang Oras sa Hawaii-Aleutian", "HAT": "Oras sa Adlawan sa Newfoundland", "HECU": "Oras sa Adlawan sa Cuba", "HEEG": "Oras sa Tag-init sa East Greenland", "HENOMX": "Oras sa Adlawan sa Northwest Mexico", "HEOG": "Oras sa Tag-init sa West Greenland", "HEPM": "Oras sa Adlawan sa St. Pierre & Miquelon", "HEPMX": "Oras sa Adlawan sa Mexican Pacific", "HKST": "Oras sa Tag-init sa Hong Kong", "HKT": "Tamdanang Oras sa Hong Kong", "HNCU": "Tamdanang Oras sa Cuba", "HNEG": "Tamdanang Oras sa East Greenland", "HNNOMX": "Tamdanang Oras sa Northwest Mexico", "HNOG": "Tamdanang Oras sa West Greenland", "HNPM": "Tamdanang Oras sa St. Pierre & Miquelon", "HNPMX": "Tamdanang Oras sa Mexican Pacific", "HNT": "Tamdanang Oras sa Newfoundland", "IST": "Tamdanang Oras sa India", "JDT": "Oras sa Adlawan sa Japan", "JST": "Tamdanang Oras sa Japan", "LHDT": "Oras sa Adlawan sa Lord Howe", "LHST": "Tamdanang Oras sa Lord Howe", "MDT": "MDT", "MESZ": "Oras sa Europa Taliwala sa Adlaw sa Tag-init", "MEZ": "Sukdanang Oras sa Europa sa Taliwala sa Adlaw", "MST": "MST", "MYT": "Oras sa Malaysia", "NZDT": "Oras sa Adlawan sa New Zealand", "NZST": "Tamdanang Oras sa New Zealand", "OESZ": "Oras sa Taga-Oryente sa Europa sa Tag-int", "OEZ": "Sukdanang Oras sa Taga-Oryente sa Europa", "PDT": "Norte Amerika Oras sa Kahayag sa Adlaw sa Pasipiko", "PST": "Norte Amerika Sukdanang Oras sa Pasipiko", "SAST": "Tamdanang Oras sa South Africa", "SGT": "Tamdanang Oras sa Singapore", "SRT": "Oras sa Suriname", "TMST": "Oras sa Tag-init sa Turkmenistan", "TMT": "Tamdanang Oras sa Turkmenistan", "UYST": "Oras sa Tag-init sa Uruguay", "UYT": "Tamdanang Oras sa Uruguay", "VET": "Oras sa Venezuela", "WARST": "Oras sa Tag-init sa Western Argentina", "WART": "Tamdanang Oras sa Western Argentina", "WAST": "Oras sa Tag-init sa West Africa", "WAT": "Tamdanang Oras sa West Africa", "WESZ": "Oras sa Kasadpanon sa Europa sa Tag-init", "WEZ": "Sukadanang Oras sa Kasadpanon sa Europa", "WIB": "Oras sa Western Indonesia", "WIT": "Oras sa Eastern Indonesia", "WITA": "Oras sa Central Indonesia", "∅∅∅": "Oras sa Tag-init sa Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c1a7400a6209ae7e8eb28592c58f6657,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cgg"] = func() locales.Translator {
		return &localen{

			locale:            "cgg",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "KBZ", "KBR", "KST", "KKN", "KTN", "KMK", "KMS", "KMN", "KMW", "KKM", "KNK", "KNB"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Okwokubanza", "Okwakabiri", "Okwakashatu", "Okwakana", "Okwakataana", "Okwamukaaga", "Okwamushanju", "Okwamunaana", "Okwamwenda", "Okwaikumi", "Okwaikumi na kumwe", "Okwaikumi na ibiri"},
			daysAbbreviated:   []string{"SAN", "ORK", "OKB", "OKS", "OKN", "OKT", "OMK"},
			daysNarrow:        []string{"S", "K", "R", "S", "N", "T", "M"},
			daysWide:          []string{"Sande", "Orwokubanza", "Orwakabiri", "Orwakashatu", "Orwakana", "Orwakataano", "Orwamukaaga"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Kurisito Atakaijire", "Kurisito Yaijire"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cgg_ug"] = func() locales.Translator {
		return &localen{

			locale:            "cgg_UG",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "KBZ", "KBR", "KST", "KKN", "KTN", "KMK", "KMS", "KMN", "KMW", "KKM", "KNK", "KNB"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Okwokubanza", "Okwakabiri", "Okwakashatu", "Okwakana", "Okwakataana", "Okwamukaaga", "Okwamushanju", "Okwamunaana", "Okwamwenda", "Okwaikumi", "Okwaikumi na kumwe", "Okwaikumi na ibiri"},
			daysAbbreviated:   []string{"SAN", "ORK", "OKB", "OKS", "OKN", "OKT", "OMK"},
			daysNarrow:        []string{"S", "K", "R", "S", "N", "T", "M"},
			daysWide:          []string{"Sande", "Orwokubanza", "Orwakabiri", "Orwakashatu", "Orwakana", "Orwakataano", "Orwamukaaga"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Kurisito Atakaijire", "Kurisito Yaijire"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["chr"] = func() locales.Translator {
		return &localen{

			locale:                 "chr",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ᎤᏃ", "ᎧᎦ", "ᎠᏅ", "ᎧᏬ", "ᎠᏂ", "ᏕᎭ", "ᎫᏰ", "ᎦᎶ", "ᏚᎵ", "ᏚᏂ", "ᏅᏓ", "ᎥᏍ"},
			monthsNarrow:           []string{"", "Ꭴ", "Ꭷ", "Ꭰ", "Ꭷ", "Ꭰ", "Ꮥ", "Ꭻ", "Ꭶ", "Ꮪ", "Ꮪ", "Ꮕ", "Ꭵ"},
			monthsWide:             []string{"", "ᎤᏃᎸᏔᏅ", "ᎧᎦᎵ", "ᎠᏅᏱ", "ᎧᏬᏂ", "ᎠᏂᏍᎬᏘ", "ᏕᎭᎷᏱ", "ᎫᏰᏉᏂ", "ᎦᎶᏂ", "ᏚᎵᏍᏗ", "ᏚᏂᏅᏗ", "ᏅᏓᏕᏆ", "ᎥᏍᎩᏱ"},
			daysAbbreviated:        []string{"ᏆᏍᎬ", "ᏉᏅᎯ", "ᏔᎵᏁ", "ᏦᎢᏁ", "ᏅᎩᏁ", "ᏧᎾᎩ", "ᏈᏕᎾ"},
			daysNarrow:             []string{"Ꮖ", "Ꮙ", "Ꮤ", "Ꮶ", "Ꮕ", "Ꮷ", "Ꭴ"},
			daysShort:              []string{"ᏍᎬ", "ᏅᎯ", "ᏔᎵ", "ᏦᎢ", "ᏅᎩ", "ᏧᎾ", "ᏕᎾ"},
			daysWide:               []string{"ᎤᎾᏙᏓᏆᏍᎬ", "ᎤᎾᏙᏓᏉᏅᎯ", "ᏔᎵᏁᎢᎦ", "ᏦᎢᏁᎢᎦ", "ᏅᎩᏁᎢᎦ", "ᏧᎾᎩᎶᏍᏗ", "ᎤᎾᏙᏓᏈᏕᎾ"},
			periodsAbbreviated:     []string{"ᏌᎾᎴ", "ᏒᎯᏱᎢ"},
			periodsNarrow:          []string{"Ꮜ", "Ꮢ"},
			periodsWide:            []string{"ᏌᎾᎴ", "ᏒᎯᏱᎢᏗᏢ"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ᏧᏓᎷᎸ ᎤᎷᎯᏍᏗ ᎦᎶᏁᏛ", "ᎠᏃ ᏙᎻᏂ"},
			timezones:              map[string]string{"ACDT": "ᎠᏰᏟ ᎡᎳᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "ACST": "ᎠᏰᏟ ᎡᎳᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ACWDT": "ᎠᏰᏟ ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "ACWST": "ᎠᏰᏟ ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ADT": "ᏗᎧᎸᎬ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AEDT": "ᎡᎳᏗᏜ ᏗᎧᎸᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AEST": "ᎡᎳᏗᏜ ᏗᎧᎸᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "AKDT": "ᎠᎳᏍᎦ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AKST": "ᎠᎳᏍᎦ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ARST": "ᎠᏥᏂᏘᏂᎠ ᎪᎩ ᎠᏟᎢᎵᏒ", "ART": "ᎠᏥᏂᏘᏂᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "AST": "ᏗᎧᎸᎬ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "AWDT": "ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AWST": "ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "BOT": "ᏉᎵᏫᎠ ᎠᏟᎢᎵᏒ", "BT": "ᏊᏔᏂ ᎠᏟᎢᎵᏒ", "CAT": "ᎠᏰᏟ ᎬᎿᎨᏍᏛ ᎠᏟᎢᎵᏒ", "CDT": "ᎠᏰᏟ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "CHADT": "ᏣᏝᎻ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "CHAST": "ᏣᏝᎻ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "CLST": "ᏥᎵ ᎪᎩ ᎠᏟᎢᎵᏒ", "CLT": "ᏥᎵ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "COST": "ᎪᎸᎻᏈᎢᎠ ᎪᎩ ᎠᏟᎢᎵᏒ", "COT": "ᎪᎸᎻᏈᎢᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "CST": "ᎠᏰᏟ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ChST": "ᏣᎼᎶ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "EAT": "ᏗᎧᎸᎬ ᎬᎿᎨᏍᏛ ᎠᏟᎢᎵᏒ", "ECT": "ᎡᏆᏙᎵ ᎠᏟᎢᎵᏒ", "EDT": "ᏗᎧᎸᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "EST": "ᏗᎧᎸᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "GFT": "ᎠᏂᎦᎸ ᏈᏯᎾ ᎠᏟᎢᎵᏒ", "GMT": "ᎢᏤ ᎢᏳᏍᏗ ᎠᏟᎢᎵᏒ", "GST": "ᎡᏉᏄᎸᏗ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "GYT": "ᎦᏯᎾ ᎠᏟᎢᎵᏒ", "HADT": "ᎭᏩᏱ-ᎠᎵᏳᏏᎠᏂ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HAST": "ᎭᏩᏱ-ᎠᎵᏳᏏᎠᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HAT": "ᎢᏤᎤᏂᏩᏛᏓᎦᏙᎯ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HECU": "ᎫᏆ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HEEG": "ᏗᎧᎸᎬ ᎢᏤᏍᏛᏱ ᎪᎩ ᎠᏟᎢᎵᏒ", "HENOMX": "ᏧᏴᏢ ᏭᏕᎵᎬ ᎠᏂᏍᏆᏂ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HEOG": "ᏭᏕᎵᎬ ᎢᏤᏍᏛᏱ ᎪᎩ ᎠᏟᎢᎵᏒ", "HEPM": "ᎤᏓᏅᏘ ᏈᏰ & ᎻᏇᎶᏂ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HEPMX": "ᎠᏂᏍᏆᏂ ᏭᏕᎵᎬ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HKST": "ᎰᏂᎩ ᎪᏂᎩ ᎪᎩ ᎠᏟᎢᎵᏒ", "HKT": "ᎰᏂᎩ ᎪᏂᎩ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNCU": "ᎫᏆ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNEG": "ᏗᎧᎸᎬ ᎢᏤᏍᏛᏱ ᎠᏟᎶᏍᏗ ᎠᎵᎢᎵᏒ", "HNNOMX": "ᏧᏴᏢ ᏭᏕᎵᎬ ᎠᏂᏍᏆᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNOG": "ᏭᏕᎵᎬ ᎢᏤᏍᏛᏱ ᎠᏟᎶᏍᏗ ᎠᎵᎢᎵᏒ", "HNPM": "ᎤᏓᏅᏘ ᏈᏰ & ᎻᏇᎶᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNPMX": "ᎠᏂᏍᏆᏂ ᏭᏕᎵᎬ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNT": "ᎢᏤᎤᏂᏩᏛᏓᎦᏙᎯ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "IST": "ᎢᏂᏗᎢᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "JDT": "ᏣᏩᏂᏏ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "JST": "ᏣᏩᏂᏏ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "LHDT": "ᎤᎬᏫᏳᎯ ᎭᏫ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "LHST": "ᎤᎬᏫᏳᎯ ᎭᏫ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "MDT": "ᎣᏓᎸ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "MESZ": "ᎠᏰᏟ ᏳᎳᏈ ᎪᎩ ᎠᏟᎢᎵᏒ", "MEZ": "ᎠᏰᏟ ᏳᎳᏈ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "MST": "ᎣᏓᎸ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "MYT": "ᎹᎴᏏᎢᎠ ᎠᏟᎢᎵᏒ", "NZDT": "ᎢᏤ ᏏᎢᎴᏂᏗ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "NZST": "ᎢᏤ ᏏᎢᎴᏂᏗ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "OESZ": "ᏗᎧᎸᎬ ᏗᏜ ᏳᎳᏈ ᎪᎩ ᎠᏟᎢᎵᏒ", "OEZ": "ᏗᎧᎸᎬ ᏗᏜ ᏳᎳᏈ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "PDT": "ᏭᏕᎵᎬ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "PST": "ᏭᏕᎵᎬ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "SAST": "ᏧᎦᎾᏮ ᎬᎿᎨᏍᏛ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "SGT": "ᏏᏂᎦᏉᎵ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "SRT": "ᏒᎵᎾᎻ ᎠᏟᎢᎵᏒ", "TMST": "ᏛᎵᎩᎺᏂᏍᏔᏂ ᎪᎩ ᎠᏟᎢᎵᏒ", "TMT": "ᏛᎵᎩᎺᏂᏍᏔᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "UYST": "ᏳᎷᏇ ᎪᎩ ᎠᏟᎢᎵᏒ", "UYT": "ᏳᎷᏇ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "VET": "ᏪᏁᏑᏪᎳ ᎠᏟᎢᎵᏒ", "WARST": "ᏭᏕᎵᎬ ᏗᏜ ᎠᏥᏂᏘᏂᎠ ᎪᎩ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WART": "ᏭᏕᎵᎬ ᏗᏜ ᎠᏥᏂᏘᏂᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WAST": "ᏭᏕᎵᎬ ᎬᎿᎨᏍᏛ ᎪᎩ ᎠᏟᎢᎵᏒ", "WAT": "ᏭᏕᎵᎬ ᎬᎿᎨᏍᏛ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WESZ": "ᏭᏕᎵᎬ ᏗᏜ ᏳᎳᏈ ᎪᎩ ᎠᏟᎢᎵᏒ", "WEZ": "ᏭᏕᎵᎬ ᏗᏜ ᏳᎳᏈ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WIB": "ᏭᏕᎵᎬ ᏗᏜ ᎢᏂᏙᏂᏍᏯ ᎠᏟᎢᎵᏒ", "WIT": "ᏗᎧᎸᎬ ᏗᏜ ᎢᏂᏙᏂᏍᏯ ᎠᏟᎢᎵᏒ", "WITA": "ᎠᏰᏟ ᎢᏂᏙᏂᏍᏯ ᎠᏟᎢᎵᏒ", "∅∅∅": "ᎠᏐᎴᏏ ᎪᎩ ᎠᏟᎢᎵᏒ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["chr_us"] = func() locales.Translator {
		return &localen{

			locale:                 "chr_US",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ᎤᏃ", "ᎧᎦ", "ᎠᏅ", "ᎧᏬ", "ᎠᏂ", "ᏕᎭ", "ᎫᏰ", "ᎦᎶ", "ᏚᎵ", "ᏚᏂ", "ᏅᏓ", "ᎥᏍ"},
			monthsNarrow:           []string{"", "Ꭴ", "Ꭷ", "Ꭰ", "Ꭷ", "Ꭰ", "Ꮥ", "Ꭻ", "Ꭶ", "Ꮪ", "Ꮪ", "Ꮕ", "Ꭵ"},
			monthsWide:             []string{"", "ᎤᏃᎸᏔᏅ", "ᎧᎦᎵ", "ᎠᏅᏱ", "ᎧᏬᏂ", "ᎠᏂᏍᎬᏘ", "ᏕᎭᎷᏱ", "ᎫᏰᏉᏂ", "ᎦᎶᏂ", "ᏚᎵᏍᏗ", "ᏚᏂᏅᏗ", "ᏅᏓᏕᏆ", "ᎥᏍᎩᏱ"},
			daysAbbreviated:        []string{"ᏆᏍᎬ", "ᏉᏅᎯ", "ᏔᎵᏁ", "ᏦᎢᏁ", "ᏅᎩᏁ", "ᏧᎾᎩ", "ᏈᏕᎾ"},
			daysNarrow:             []string{"Ꮖ", "Ꮙ", "Ꮤ", "Ꮶ", "Ꮕ", "Ꮷ", "Ꭴ"},
			daysShort:              []string{"ᏍᎬ", "ᏅᎯ", "ᏔᎵ", "ᏦᎢ", "ᏅᎩ", "ᏧᎾ", "ᏕᎾ"},
			daysWide:               []string{"ᎤᎾᏙᏓᏆᏍᎬ", "ᎤᎾᏙᏓᏉᏅᎯ", "ᏔᎵᏁᎢᎦ", "ᏦᎢᏁᎢᎦ", "ᏅᎩᏁᎢᎦ", "ᏧᎾᎩᎶᏍᏗ", "ᎤᎾᏙᏓᏈᏕᎾ"},
			periodsAbbreviated:     []string{"ᏌᎾᎴ", "ᏒᎯᏱᎢ"},
			periodsNarrow:          []string{"Ꮜ", "Ꮢ"},
			periodsWide:            []string{"ᏌᎾᎴ", "ᏒᎯᏱᎢᏗᏢ"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ᏧᏓᎷᎸ ᎤᎷᎯᏍᏗ ᎦᎶᏁᏛ", "ᎠᏃ ᏙᎻᏂ"},
			timezones:              map[string]string{"ACDT": "ᎠᏰᏟ ᎡᎳᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "ACST": "ᎠᏰᏟ ᎡᎳᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ACWDT": "ᎠᏰᏟ ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "ACWST": "ᎠᏰᏟ ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ADT": "ᏗᎧᎸᎬ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AEDT": "ᎡᎳᏗᏜ ᏗᎧᎸᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AEST": "ᎡᎳᏗᏜ ᏗᎧᎸᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "AKDT": "ᎠᎳᏍᎦ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AKST": "ᎠᎳᏍᎦ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ARST": "ᎠᏥᏂᏘᏂᎠ ᎪᎩ ᎠᏟᎢᎵᏒ", "ART": "ᎠᏥᏂᏘᏂᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "AST": "ᏗᎧᎸᎬ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "AWDT": "ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "AWST": "ᎡᎳᏗᏜ ᏭᏕᎵᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "BOT": "ᏉᎵᏫᎠ ᎠᏟᎢᎵᏒ", "BT": "ᏊᏔᏂ ᎠᏟᎢᎵᏒ", "CAT": "ᎠᏰᏟ ᎬᎿᎨᏍᏛ ᎠᏟᎢᎵᏒ", "CDT": "ᎠᏰᏟ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "CHADT": "ᏣᏝᎻ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "CHAST": "ᏣᏝᎻ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "CLST": "ᏥᎵ ᎪᎩ ᎠᏟᎢᎵᏒ", "CLT": "ᏥᎵ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "COST": "ᎪᎸᎻᏈᎢᎠ ᎪᎩ ᎠᏟᎢᎵᏒ", "COT": "ᎪᎸᎻᏈᎢᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "CST": "ᎠᏰᏟ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "ChST": "ᏣᎼᎶ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "EAT": "ᏗᎧᎸᎬ ᎬᎿᎨᏍᏛ ᎠᏟᎢᎵᏒ", "ECT": "ᎡᏆᏙᎵ ᎠᏟᎢᎵᏒ", "EDT": "ᏗᎧᎸᎬ ᏗᏜ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "EST": "ᏗᎧᎸᎬ ᏗᏜ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "GFT": "ᎠᏂᎦᎸ ᏈᏯᎾ ᎠᏟᎢᎵᏒ", "GMT": "ᎢᏤ ᎢᏳᏍᏗ ᎠᏟᎢᎵᏒ", "GST": "ᎡᏉᏄᎸᏗ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "GYT": "ᎦᏯᎾ ᎠᏟᎢᎵᏒ", "HADT": "ᎭᏩᏱ-ᎠᎵᏳᏏᎠᏂ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HAST": "ᎭᏩᏱ-ᎠᎵᏳᏏᎠᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HAT": "ᎢᏤᎤᏂᏩᏛᏓᎦᏙᎯ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HECU": "ᎫᏆ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HEEG": "ᏗᎧᎸᎬ ᎢᏤᏍᏛᏱ ᎪᎩ ᎠᏟᎢᎵᏒ", "HENOMX": "ᏧᏴᏢ ᏭᏕᎵᎬ ᎠᏂᏍᏆᏂ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HEOG": "ᏭᏕᎵᎬ ᎢᏤᏍᏛᏱ ᎪᎩ ᎠᏟᎢᎵᏒ", "HEPM": "ᎤᏓᏅᏘ ᏈᏰ & ᎻᏇᎶᏂ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HEPMX": "ᎠᏂᏍᏆᏂ ᏭᏕᎵᎬ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "HKST": "ᎰᏂᎩ ᎪᏂᎩ ᎪᎩ ᎠᏟᎢᎵᏒ", "HKT": "ᎰᏂᎩ ᎪᏂᎩ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNCU": "ᎫᏆ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNEG": "ᏗᎧᎸᎬ ᎢᏤᏍᏛᏱ ᎠᏟᎶᏍᏗ ᎠᎵᎢᎵᏒ", "HNNOMX": "ᏧᏴᏢ ᏭᏕᎵᎬ ᎠᏂᏍᏆᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNOG": "ᏭᏕᎵᎬ ᎢᏤᏍᏛᏱ ᎠᏟᎶᏍᏗ ᎠᎵᎢᎵᏒ", "HNPM": "ᎤᏓᏅᏘ ᏈᏰ & ᎻᏇᎶᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNPMX": "ᎠᏂᏍᏆᏂ ᏭᏕᎵᎬ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "HNT": "ᎢᏤᎤᏂᏩᏛᏓᎦᏙᎯ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "IST": "ᎢᏂᏗᎢᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "JDT": "ᏣᏩᏂᏏ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "JST": "ᏣᏩᏂᏏ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "LHDT": "ᎤᎬᏫᏳᎯ ᎭᏫ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "LHST": "ᎤᎬᏫᏳᎯ ᎭᏫ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "MDT": "ᎣᏓᎸ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "MESZ": "ᎠᏰᏟ ᏳᎳᏈ ᎪᎩ ᎠᏟᎢᎵᏒ", "MEZ": "ᎠᏰᏟ ᏳᎳᏈ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "MST": "ᎣᏓᎸ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "MYT": "ᎹᎴᏏᎢᎠ ᎠᏟᎢᎵᏒ", "NZDT": "ᎢᏤ ᏏᎢᎴᏂᏗ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "NZST": "ᎢᏤ ᏏᎢᎴᏂᏗ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "OESZ": "ᏗᎧᎸᎬ ᏗᏜ ᏳᎳᏈ ᎪᎩ ᎠᏟᎢᎵᏒ", "OEZ": "ᏗᎧᎸᎬ ᏗᏜ ᏳᎳᏈ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "PDT": "ᏭᏕᎵᎬ ᎪᎯ ᎢᎦ ᎠᏟᎢᎵᏒ", "PST": "ᏭᏕᎵᎬ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "SAST": "ᏧᎦᎾᏮ ᎬᎿᎨᏍᏛ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "SGT": "ᏏᏂᎦᏉᎵ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "SRT": "ᏒᎵᎾᎻ ᎠᏟᎢᎵᏒ", "TMST": "ᏛᎵᎩᎺᏂᏍᏔᏂ ᎪᎩ ᎠᏟᎢᎵᏒ", "TMT": "ᏛᎵᎩᎺᏂᏍᏔᏂ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "UYST": "ᏳᎷᏇ ᎪᎩ ᎠᏟᎢᎵᏒ", "UYT": "ᏳᎷᏇ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "VET": "ᏪᏁᏑᏪᎳ ᎠᏟᎢᎵᏒ", "WARST": "ᏭᏕᎵᎬ ᏗᏜ ᎠᏥᏂᏘᏂᎠ ᎪᎩ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WART": "ᏭᏕᎵᎬ ᏗᏜ ᎠᏥᏂᏘᏂᎠ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WAST": "ᏭᏕᎵᎬ ᎬᎿᎨᏍᏛ ᎪᎩ ᎠᏟᎢᎵᏒ", "WAT": "ᏭᏕᎵᎬ ᎬᎿᎨᏍᏛ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WESZ": "ᏭᏕᎵᎬ ᏗᏜ ᏳᎳᏈ ᎪᎩ ᎠᏟᎢᎵᏒ", "WEZ": "ᏭᏕᎵᎬ ᏗᏜ ᏳᎳᏈ ᎠᏟᎶᏍᏗ ᎠᏟᎢᎵᏒ", "WIB": "ᏭᏕᎵᎬ ᏗᏜ ᎢᏂᏙᏂᏍᏯ ᎠᏟᎢᎵᏒ", "WIT": "ᏗᎧᎸᎬ ᏗᏜ ᎢᏂᏙᏂᏍᏯ ᎠᏟᎢᎵᏒ", "WITA": "ᎠᏰᏟ ᎢᏂᏙᏂᏍᏯ ᎠᏟᎢᎵᏒ", "∅∅∅": "ᎠᏐᎴᏏ ᎪᎩ ᎠᏟᎢᎵᏒ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ckb"] = func() locales.Translator {
		return &localen{

			locale:                 "ckb",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‏-",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "د.ع.\u200f", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "کانوونی دووەم", "شوبات", "ئازار", "نیسان", "ئایار", "حوزەیران", "تەمووز", "ئاب", "ئەیلوول", "تشرینی یەکەم", "تشرینی دووەم", "کانونی یەکەم"},
			monthsNarrow:           []string{"", "ک", "ش", "ئ", "ن", "ئ", "ح", "ت", "ئ", "ئ", "ت", "ت", "ک"},
			monthsWide:             []string{"", "کانوونی دووەم", "شوبات", "ئازار", "نیسان", "ئایار", "حوزەیران", "تەمووز", "ئاب", "ئەیلوول", "تشرینی یەکەم", "تشرینی دووەم", "کانونی یەکەم"},
			daysAbbreviated:        []string{"یەکشەممە", "دووشەممە", "سێشەممە", "چوارشەممە", "پێنجشەممە", "ھەینی", "شەممە"},
			daysNarrow:             []string{"ی", "د", "س", "چ", "پ", "ھ", "ش"},
			daysShort:              []string{"١ش", "٢ش", "٣ش", "٤ش", "٥ش", "ھ", "ش"},
			daysWide:               []string{"یەکشەممە", "دووشەممە", "سێشەممە", "چوارشەممە", "پێنجشەممە", "ھەینی", "شەممە"},
			periodsAbbreviated:     []string{"ب.ن", "د.ن"},
			periodsNarrow:          []string{"ب.ن", "د.ن"},
			periodsWide:            []string{"ب.ن", "د.ن"},
			erasAbbreviated:        []string{"پێش زایین", "زایینی"},
			erasNarrow:             []string{"پ.ن", "ز"},
			erasWide:               []string{"پێش زایین", "زایینی"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f7223149b88d711e36c711437b51fa28,
			fnFmtCurrency:         fnFmtCurrency_d61f2036c9f7cc7a15c2d51b90ffbef9,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_335b03b04c77445aeb9ba732c8cd38ef,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_96d7d1823041a48b887e3e929259dbdf,
			fnFmtPercent:          fnFmtPercent_f5e1c0d5e27b0e309f9e016c63b96767,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ckb_iq"] = func() locales.Translator {
		return &localen{

			locale:                 "ckb_IQ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‏-",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "کانوونی دووەم", "شوبات", "ئازار", "نیسان", "ئایار", "حوزەیران", "تەمووز", "ئاب", "ئەیلوول", "تشرینی یەکەم", "تشرینی دووەم", "کانونی یەکەم"},
			monthsNarrow:           []string{"", "ک", "ش", "ئ", "ن", "ئ", "ح", "ت", "ئ", "ئ", "ت", "ت", "ک"},
			monthsWide:             []string{"", "کانوونی دووەم", "شوبات", "ئازار", "نیسان", "ئایار", "حوزەیران", "تەمووز", "ئاب", "ئەیلوول", "تشرینی یەکەم", "تشرینی دووەم", "کانونی یەکەم"},
			daysAbbreviated:        []string{"یەکشەممە", "دووشەممە", "سێشەممە", "چوارشەممە", "پێنجشەممە", "ھەینی", "شەممە"},
			daysNarrow:             []string{"ی", "د", "س", "چ", "پ", "ھ", "ش"},
			daysShort:              []string{"١ش", "٢ش", "٣ش", "٤ش", "٥ش", "ھ", "ش"},
			daysWide:               []string{"یەکشەممە", "دووشەممە", "سێشەممە", "چوارشەممە", "پێنجشەممە", "ھەینی", "شەممە"},
			periodsAbbreviated:     []string{"ب.ن", "د.ن"},
			periodsNarrow:          []string{"ب.ن", "د.ن"},
			periodsWide:            []string{"ب.ن", "د.ن"},
			erasAbbreviated:        []string{"پێش زایین", "زایینی"},
			erasNarrow:             []string{"پ.ن", "ز"},
			erasWide:               []string{"پێش زایین", "زایینی"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f7223149b88d711e36c711437b51fa28,
			fnFmtCurrency:         fnFmtCurrency_d61f2036c9f7cc7a15c2d51b90ffbef9,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_335b03b04c77445aeb9ba732c8cd38ef,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_96d7d1823041a48b887e3e929259dbdf,
			fnFmtPercent:          fnFmtPercent_f5e1c0d5e27b0e309f9e016c63b96767,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ckb_ir"] = func() locales.Translator {
		return &localen{

			locale:                 "ckb_IR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‏-",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "کانوونی دووەم", "شوبات", "ئازار", "نیسان", "ئایار", "حوزەیران", "تەمووز", "ئاب", "ئەیلوول", "تشرینی یەکەم", "تشرینی دووەم", "کانونی یەکەم"},
			monthsNarrow:           []string{"", "ک", "ش", "ئ", "ن", "ئ", "ح", "ت", "ئ", "ئ", "ت", "ت", "ک"},
			monthsWide:             []string{"", "کانوونی دووەم", "شوبات", "ئازار", "نیسان", "ئایار", "حوزەیران", "تەمووز", "ئاب", "ئەیلوول", "تشرینی یەکەم", "تشرینی دووەم", "کانونی یەکەم"},
			daysAbbreviated:        []string{"یەکشەممە", "دووشەممە", "سێشەممە", "چوارشەممە", "پێنجشەممە", "ھەینی", "شەممە"},
			daysNarrow:             []string{"ی", "د", "س", "چ", "پ", "ھ", "ش"},
			daysShort:              []string{"١ش", "٢ش", "٣ش", "٤ش", "٥ش", "ھ", "ش"},
			daysWide:               []string{"یەکشەممە", "دووشەممە", "سێشەممە", "چوارشەممە", "پێنجشەممە", "ھەینی", "شەممە"},
			periodsAbbreviated:     []string{"ب.ن", "د.ن"},
			periodsNarrow:          []string{"ب.ن", "د.ن"},
			periodsWide:            []string{"ب.ن", "د.ن"},
			erasAbbreviated:        []string{"پێش زایین", "زایینی"},
			erasNarrow:             []string{"پ.ن", "ز"},
			erasWide:               []string{"پێش زایین", "زایینی"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f7223149b88d711e36c711437b51fa28,
			fnFmtCurrency:         fnFmtCurrency_d61f2036c9f7cc7a15c2d51b90ffbef9,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_335b03b04c77445aeb9ba732c8cd38ef,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_96d7d1823041a48b887e3e929259dbdf,
			fnFmtPercent:          fnFmtPercent_f5e1c0d5e27b0e309f9e016c63b96767,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cs"] = func() locales.Translator {
		return &localen{

			locale:                 "cs",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "Kčs", "CUC", "CUP", "CVE", "CYP", "Kč", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "ECU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "led", "úno", "bře", "dub", "kvě", "čvn", "čvc", "srp", "zář", "říj", "lis", "pro"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ledna", "února", "března", "dubna", "května", "června", "července", "srpna", "září", "října", "listopadu", "prosince"},
			daysAbbreviated:        []string{"ne", "po", "út", "st", "čt", "pá", "so"},
			daysNarrow:             []string{"N", "P", "Ú", "S", "Č", "P", "S"},
			daysShort:              []string{"ne", "po", "út", "st", "čt", "pá", "so"},
			daysWide:               []string{"neděle", "pondělí", "úterý", "středa", "čtvrtek", "pátek", "sobota"},
			periodsAbbreviated:     []string{"dop.", "odp."},
			periodsNarrow:          []string{"dop.", "odp."},
			periodsWide:            []string{"dop.", "odp."},
			erasAbbreviated:        []string{"př. n. l.", "n. l."},
			erasNarrow:             []string{"př.n.l.", "n.l."},
			erasWide:               []string{"před naším letopočtem", "našeho letopočtu"},
			timezones:              map[string]string{"ACDT": "Středoaustralský letní čas", "ACST": "Středoaustralský standardní čas", "ACWDT": "Středozápadní australský letní čas", "ACWST": "Středozápadní australský standardní čas", "ADT": "Atlantický letní čas", "AEDT": "Východoaustralský letní čas", "AEST": "Východoaustralský standardní čas", "AKDT": "Aljašský letní čas", "AKST": "Aljašský standardní čas", "ARST": "Argentinský letní čas", "ART": "Argentinský standardní čas", "AST": "Atlantický standardní čas", "AWDT": "Západoaustralský letní čas", "AWST": "Západoaustralský standardní čas", "BOT": "Bolivijský čas", "BT": "Bhútánský čas", "CAT": "Středoafrický čas", "CDT": "Severoamerický centrální letní čas", "CHADT": "Chathamský letní čas", "CHAST": "Chathamský standardní čas", "CLST": "Chilský letní čas", "CLT": "Chilský standardní čas", "COST": "Kolumbijský letní čas", "COT": "Kolumbijský standardní čas", "CST": "Severoamerický centrální standardní čas", "ChST": "Chamorrský čas", "EAT": "Východoafrický čas", "ECT": "Ekvádorský čas", "EDT": "Severoamerický východní letní čas", "EST": "Severoamerický východní standardní čas", "GFT": "Francouzskoguyanský čas", "GMT": "Greenwichský střední čas", "GST": "Standardní čas Perského zálivu", "GYT": "Guyanský čas", "HADT": "Havajsko-aleutský letní čas", "HAST": "Havajsko-aleutský standardní čas", "HAT": "Newfoundlandský letní čas", "HECU": "Kubánský letní čas", "HEEG": "Východogrónský letní čas", "HENOMX": "Severozápadní mexický letní čas", "HEOG": "Západogrónský letní čas", "HEPM": "Pierre-miquelonský letní čas", "HEPMX": "Mexický pacifický letní čas", "HKST": "Hongkongský letní čas", "HKT": "Hongkongský standardní čas", "HNCU": "Kubánský standardní čas", "HNEG": "Východogrónský standardní čas", "HNNOMX": "Severozápadní mexický standardní čas", "HNOG": "Západogrónský standardní čas", "HNPM": "Pierre-miquelonský standardní čas", "HNPMX": "Mexický pacifický standardní čas", "HNT": "Newfoundlandský standardní čas", "IST": "Indický čas", "JDT": "Japonský letní čas", "JST": "Japonský standardní čas", "LHDT": "Letní čas ostrova lorda Howa", "LHST": "Standardní čas ostrova lorda Howa", "MDT": "Severoamerický horský letní čas", "MESZ": "Středoevropský letní čas", "MEZ": "Středoevropský standardní čas", "MST": "Severoamerický horský standardní čas", "MYT": "Malajský čas", "NZDT": "Novozélandský letní čas", "NZST": "Novozélandský standardní čas", "OESZ": "Východoevropský letní čas", "OEZ": "Východoevropský standardní čas", "PDT": "Severoamerický pacifický letní čas", "PST": "Severoamerický pacifický standardní čas", "SAST": "Jihoafrický čas", "SGT": "Singapurský čas", "SRT": "Surinamský čas", "TMST": "Turkmenský letní čas", "TMT": "Turkmenský standardní čas", "UYST": "Uruguayský letní čas", "UYT": "Uruguayský standardní čas", "VET": "Venezuelský čas", "WARST": "Západoargentinský letní čas", "WART": "Západoargentinský standardní čas", "WAST": "Západoafrický letní čas", "WAT": "Západoafrický standardní čas", "WESZ": "Západoevropský letní čas", "WEZ": "Západoevropský standardní čas", "WIB": "Západoindonéský čas", "WIT": "Východoindonéský čas", "WITA": "Středoindonéský čas", "∅∅∅": "Azorský letní čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_935abbc817a75b933eeb5ecc41e85d90,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_15dd2e1fe33af45e98be7d9f9b1b6b93,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cs_cz"] = func() locales.Translator {
		return &localen{

			locale:                 "cs_CZ",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "led", "úno", "bře", "dub", "kvě", "čvn", "čvc", "srp", "zář", "říj", "lis", "pro"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ledna", "února", "března", "dubna", "května", "června", "července", "srpna", "září", "října", "listopadu", "prosince"},
			daysAbbreviated:        []string{"ne", "po", "út", "st", "čt", "pá", "so"},
			daysNarrow:             []string{"N", "P", "Ú", "S", "Č", "P", "S"},
			daysShort:              []string{"ne", "po", "út", "st", "čt", "pá", "so"},
			daysWide:               []string{"neděle", "pondělí", "úterý", "středa", "čtvrtek", "pátek", "sobota"},
			periodsAbbreviated:     []string{"dop.", "odp."},
			periodsNarrow:          []string{"dop.", "odp."},
			periodsWide:            []string{"dop.", "odp."},
			erasAbbreviated:        []string{"př. n. l.", "n. l."},
			erasNarrow:             []string{"př.n.l.", "n.l."},
			erasWide:               []string{"před naším letopočtem", "našeho letopočtu"},
			timezones:              map[string]string{"ACDT": "Středoaustralský letní čas", "ACST": "Středoaustralský standardní čas", "ACWDT": "Středozápadní australský letní čas", "ACWST": "Středozápadní australský standardní čas", "ADT": "Atlantický letní čas", "AEDT": "Východoaustralský letní čas", "AEST": "Východoaustralský standardní čas", "AKDT": "Aljašský letní čas", "AKST": "Aljašský standardní čas", "ARST": "Argentinský letní čas", "ART": "Argentinský standardní čas", "AST": "Atlantický standardní čas", "AWDT": "Západoaustralský letní čas", "AWST": "Západoaustralský standardní čas", "BOT": "Bolivijský čas", "BT": "Bhútánský čas", "CAT": "Středoafrický čas", "CDT": "Severoamerický centrální letní čas", "CHADT": "Chathamský letní čas", "CHAST": "Chathamský standardní čas", "CLST": "Chilský letní čas", "CLT": "Chilský standardní čas", "COST": "Kolumbijský letní čas", "COT": "Kolumbijský standardní čas", "CST": "Severoamerický centrální standardní čas", "ChST": "Chamorrský čas", "EAT": "Východoafrický čas", "ECT": "Ekvádorský čas", "EDT": "Severoamerický východní letní čas", "EST": "Severoamerický východní standardní čas", "GFT": "Francouzskoguyanský čas", "GMT": "Greenwichský střední čas", "GST": "Standardní čas Perského zálivu", "GYT": "Guyanský čas", "HADT": "Havajsko-aleutský letní čas", "HAST": "Havajsko-aleutský standardní čas", "HAT": "Newfoundlandský letní čas", "HECU": "Kubánský letní čas", "HEEG": "Východogrónský letní čas", "HENOMX": "Severozápadní mexický letní čas", "HEOG": "Západogrónský letní čas", "HEPM": "Pierre-miquelonský letní čas", "HEPMX": "Mexický pacifický letní čas", "HKST": "Hongkongský letní čas", "HKT": "Hongkongský standardní čas", "HNCU": "Kubánský standardní čas", "HNEG": "Východogrónský standardní čas", "HNNOMX": "Severozápadní mexický standardní čas", "HNOG": "Západogrónský standardní čas", "HNPM": "Pierre-miquelonský standardní čas", "HNPMX": "Mexický pacifický standardní čas", "HNT": "Newfoundlandský standardní čas", "IST": "Indický čas", "JDT": "Japonský letní čas", "JST": "Japonský standardní čas", "LHDT": "Letní čas ostrova lorda Howa", "LHST": "Standardní čas ostrova lorda Howa", "MDT": "Severoamerický horský letní čas", "MESZ": "Středoevropský letní čas", "MEZ": "Středoevropský standardní čas", "MST": "Severoamerický horský standardní čas", "MYT": "Malajský čas", "NZDT": "Novozélandský letní čas", "NZST": "Novozélandský standardní čas", "OESZ": "Východoevropský letní čas", "OEZ": "Východoevropský standardní čas", "PDT": "Severoamerický pacifický letní čas", "PST": "Severoamerický pacifický standardní čas", "SAST": "Jihoafrický čas", "SGT": "Singapurský čas", "SRT": "Surinamský čas", "TMST": "Turkmenský letní čas", "TMT": "Turkmenský standardní čas", "UYST": "Uruguayský letní čas", "UYT": "Uruguayský standardní čas", "VET": "Venezuelský čas", "WARST": "Západoargentinský letní čas", "WART": "Západoargentinský standardní čas", "WAST": "Západoafrický letní čas", "WAT": "Západoafrický standardní čas", "WESZ": "Západoevropský letní čas", "WEZ": "Západoevropský standardní čas", "WIB": "Západoindonéský čas", "WIT": "Východoindonéský čas", "WITA": "Středoindonéský čas", "∅∅∅": "Brasilijský letní čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_935abbc817a75b933eeb5ecc41e85d90,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_15dd2e1fe33af45e98be7d9f9b1b6b93,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cu"] = func() locales.Translator {
		return &localen{

			locale:                 "cu",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "₸", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "₴", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "і҆аⷩ҇", "феⷡ҇", "маⷬ҇", "а҆пⷬ҇", "маꙵ", "і҆ꙋⷩ҇", "і҆ꙋⷧ҇", "а҆́ѵⷢ҇", "сеⷫ҇", "ѻ҆кⷮ", "ноеⷨ", "деⷦ҇"},
			monthsNarrow:           []string{"", "І҆", "Ф", "М", "А҆", "М", "І҆", "І҆", "А҆", "С", "Ѻ҆", "Н", "Д"},
			monthsWide:             []string{"", "і҆аннꙋа́рїа", "феврꙋа́рїа", "ма́рта", "а҆прі́ллїа", "ма́їа", "і҆ꙋ́нїа", "і҆ꙋ́лїа", "а҆́ѵгꙋста", "септе́мврїа", "ѻ҆ктѡ́врїа", "ное́мврїа", "деке́мврїа"},
			daysAbbreviated:        []string{"ндⷧ҇ѧ", "пнⷣе", "втоⷬ҇", "срⷣе", "чеⷦ҇", "пѧⷦ҇", "сꙋⷠ҇"},
			daysNarrow:             []string{"Н", "П", "В", "С", "Ч", "П", "С"},
			daysShort:              []string{"ндⷧ҇ѧ", "пнⷣе", "втоⷬ҇", "срⷣе", "чеⷦ҇", "пѧⷦ҇", "сꙋⷠ҇"},
			daysWide:               []string{"недѣ́лѧ", "понедѣ́льникъ", "вто́рникъ", "среда̀", "четверто́къ", "пѧто́къ", "сꙋббѡ́та"},
			periodsAbbreviated:     []string{"ДП", "ПП"},
			periodsNarrow:          []string{"ДП", "ПП"},
			periodsWide:            []string{"ДП", "ПП"},
			erasAbbreviated:        []string{"пре́дъ р.\u00a0х.", "ѿ р. х."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"пре́дъ р.\u00a0х.", "по р.\u00a0х."},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "а҆тланті́ческое лѣ́тнее вре́мѧ", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "а҆тланті́ческое зи́мнее вре́мѧ", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "среднеамерїка́нское лѣ́тнее вре́мѧ", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "среднеамерїка́нское зи́мнее вре́мѧ", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "восточноамерїка́нское лѣ́тнее вре́мѧ", "EST": "восточноамерїка́нское зи́мнее вре́мѧ", "GFT": "GFT", "GMT": "сре́днее вре́мѧ по грі́нꙋичꙋ", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "а҆мерїка́нское наго́рнее лѣ́тнее вре́мѧ", "MESZ": "среднеєѵрѡпе́йское лѣ́тнее вре́мѧ", "MEZ": "среднеєѵрѡпе́йское зи́мнее вре́мѧ", "MST": "а҆мерїка́нское наго́рнее зи́мнее вре́мѧ", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "восточноєѵрѡпе́йское лѣ́тнее вре́мѧ", "OEZ": "восточноєѵрѡпе́йское зи́мнее вре́мѧ", "PDT": "тихоѻкеа́нское лѣ́тнее вре́мѧ", "PST": "тихоѻкеа́нское зи́мнее вре́мѧ", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "западноєѵрѡпе́йское лѣ́тнее вре́мѧ", "WEZ": "западноєѵрѡпе́йское зи́мнее вре́мѧ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_7d9a7daab943102fabd3a368958b52aa,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_cc33621dc8ea4b3f7e42320aa653cbdf,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cu_ru"] = func() locales.Translator {
		return &localen{

			locale:                 "cu_RU",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "і҆аⷩ҇", "феⷡ҇", "маⷬ҇", "а҆пⷬ҇", "маꙵ", "і҆ꙋⷩ҇", "і҆ꙋⷧ҇", "а҆́ѵⷢ҇", "сеⷫ҇", "ѻ҆кⷮ", "ноеⷨ", "деⷦ҇"},
			monthsNarrow:           []string{"", "І҆", "Ф", "М", "А҆", "М", "І҆", "І҆", "А҆", "С", "Ѻ҆", "Н", "Д"},
			monthsWide:             []string{"", "і҆аннꙋа́рїа", "феврꙋа́рїа", "ма́рта", "а҆прі́ллїа", "ма́їа", "і҆ꙋ́нїа", "і҆ꙋ́лїа", "а҆́ѵгꙋста", "септе́мврїа", "ѻ҆ктѡ́врїа", "ное́мврїа", "деке́мврїа"},
			daysAbbreviated:        []string{"ндⷧ҇ѧ", "пнⷣе", "втоⷬ҇", "срⷣе", "чеⷦ҇", "пѧⷦ҇", "сꙋⷠ҇"},
			daysNarrow:             []string{"Н", "П", "В", "С", "Ч", "П", "С"},
			daysShort:              []string{"ндⷧ҇ѧ", "пнⷣе", "втоⷬ҇", "срⷣе", "чеⷦ҇", "пѧⷦ҇", "сꙋⷠ҇"},
			daysWide:               []string{"недѣ́лѧ", "понедѣ́льникъ", "вто́рникъ", "среда̀", "четверто́къ", "пѧто́къ", "сꙋббѡ́та"},
			periodsAbbreviated:     []string{"ДП", "ПП"},
			periodsNarrow:          []string{"ДП", "ПП"},
			periodsWide:            []string{"ДП", "ПП"},
			erasAbbreviated:        []string{"пре́дъ р.\u00a0х.", "ѿ р. х."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"пре́дъ р.\u00a0х.", "по р.\u00a0х."},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "а҆тланті́ческое лѣ́тнее вре́мѧ", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "а҆тланті́ческое зи́мнее вре́мѧ", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "среднеамерїка́нское лѣ́тнее вре́мѧ", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "среднеамерїка́нское зи́мнее вре́мѧ", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "восточноамерїка́нское лѣ́тнее вре́мѧ", "EST": "восточноамерїка́нское зи́мнее вре́мѧ", "GFT": "GFT", "GMT": "сре́днее вре́мѧ по грі́нꙋичꙋ", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "а҆мерїка́нское наго́рнее лѣ́тнее вре́мѧ", "MESZ": "среднеєѵрѡпе́йское лѣ́тнее вре́мѧ", "MEZ": "среднеєѵрѡпе́йское зи́мнее вре́мѧ", "MST": "а҆мерїка́нское наго́рнее зи́мнее вре́мѧ", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "восточноєѵрѡпе́йское лѣ́тнее вре́мѧ", "OEZ": "восточноєѵрѡпе́йское зи́мнее вре́мѧ", "PDT": "тихоѻкеа́нское лѣ́тнее вре́мѧ", "PST": "тихоѻкеа́нское зи́мнее вре́мѧ", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "западноєѵрѡпе́йское лѣ́тнее вре́мѧ", "WEZ": "западноєѵрѡпе́йское зи́мнее вре́мѧ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_7d9a7daab943102fabd3a368958b52aa,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_cc33621dc8ea4b3f7e42320aa653cbdf,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cy"] = func() locales.Translator {
		return &localen{

			locale:                 "cy",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 3, 4, 5, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ion", "Chwef", "Maw", "Ebr", "Mai", "Meh", "Gorff", "Awst", "Medi", "Hyd", "Tach", "Rhag"},
			monthsNarrow:           []string{"", "I", "Ch", "M", "E", "M", "M", "G", "A", "M", "H", "T", "Rh"},
			monthsWide:             []string{"", "Ionawr", "Chwefror", "Mawrth", "Ebrill", "Mai", "Mehefin", "Gorffennaf", "Awst", "Medi", "Hydref", "Tachwedd", "Rhagfyr"},
			daysAbbreviated:        []string{"Sul", "Llun", "Maw", "Mer", "Iau", "Gwen", "Sad"},
			daysNarrow:             []string{"S", "Ll", "M", "M", "I", "G", "S"},
			daysShort:              []string{"Su", "Ll", "Ma", "Me", "Ia", "Gw", "Sa"},
			daysWide:               []string{"Dydd Sul", "Dydd Llun", "Dydd Mawrth", "Dydd Mercher", "Dydd Iau", "Dydd Gwener", "Dydd Sadwrn"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"b", "h"},
			periodsWide:            []string{"yb", "yh"},
			erasAbbreviated:        []string{"CC", "OC"},
			erasNarrow:             []string{"C", "O"},
			erasWide:               []string{"Cyn Crist", "Oed Crist"},
			timezones:              map[string]string{"ACDT": "Amser Haf Canolbarth Awstralia", "ACST": "Amser Safonol Canolbarth Awstralia", "ACWDT": "Amser Haf Canolbarth Gorllewin Awstralia", "ACWST": "Amser Safonol Canolbarth Gorllewin Awstralia", "ADT": "Amser Haf Cefnfor yr Iwerydd", "AEDT": "Amser Haf Dwyrain Awstralia", "AEST": "Amser Safonol Dwyrain Awstralia", "AKDT": "Amser Haf Alaska", "AKST": "Amser Safonol Alaska", "ARST": "Amser Haf Ariannin", "ART": "Amser Safonol Ariannin", "AST": "Amser Safonol Cefnfor yr Iwerydd", "AWDT": "Amser Haf Gorllewin Awstralia", "AWST": "Amser Safonol Gorllewin Awstralia", "BOT": "Amser Bolivia", "BT": "Amser Bhutan", "CAT": "Amser Canolbarth Affrica", "CDT": "Amser Haf Canolbarth Gogledd America", "CHADT": "Amser Haf Chatham", "CHAST": "Amser Safonol Chatham", "CLST": "Amser Haf Chile", "CLT": "Amser Safonol Chile", "COST": "Amser Haf Colombia", "COT": "Amser Safonol Colombia", "CST": "Amser Safonol Canolbarth Gogledd America", "ChST": "Amser Chamorro", "EAT": "Amser Dwyrain Affrica", "ECT": "Amser Ecuador", "EDT": "Amser Haf Dwyrain Gogledd America", "EST": "Amser Safonol Dwyrain Gogledd America", "GFT": "Amser Guyane Ffrengig", "GMT": "Amser Safonol Greenwich", "GST": "Amser Safonol y Gwlff", "GYT": "Amser Guyana", "HADT": "Amser Haf Hawaii-Aleutian", "HAST": "Amser Safonol Hawaii-Aleutian", "HAT": "Amser Haf Newfoundland", "HECU": "Amser Haf Cuba", "HEEG": "Amser Haf Dwyrain yr Ynys Las", "HENOMX": "Amser Haf Gogledd Orllewin Mecsico", "HEOG": "Amser Haf Gorllewin yr Ynys Las", "HEPM": "Amser Haf Saint-Pierre-et-Miquelon", "HEPMX": "Amser Haf Pasiffig Mecsico", "HKST": "Amser Haf Hong Kong", "HKT": "Amser Safonol Hong Kong", "HNCU": "Amser Safonol Cuba", "HNEG": "Amser Safonol Dwyrain yr Ynys Las", "HNNOMX": "Amser Safonol Gogledd Orllewin Mecsico", "HNOG": "Amser Safonol Gorllewin yr Ynys Las", "HNPM": "Amser Safonol Saint-Pierre-et-Miquelon", "HNPMX": "Amser Safonol Pasiffig Mecsico", "HNT": "Amser Safonol Newfoundland", "IST": "Amser India", "JDT": "Amser Haf Siapan", "JST": "Amser Safonol Siapan", "LHDT": "Amser Haf yr Arglwydd Howe", "LHST": "Amser Safonol yr Arglwydd Howe", "MDT": "Amser Haf Mynyddoedd Gogledd America", "MESZ": "Amser Haf Canolbarth Ewrop", "MEZ": "Amser Safonol Canolbarth Ewrop", "MST": "Amser Safonol Mynyddoedd Gogledd America", "MYT": "Amser Malaysia", "NZDT": "Amser Haf Seland Newydd", "NZST": "Amser Safonol Seland Newydd", "OESZ": "Amser Haf Dwyrain Ewrop", "OEZ": "Amser Safonol Dwyrain Ewrop", "PDT": "Amser Haf Cefnfor Tawel Gogledd America", "PST": "Amser Safonol Cefnfor Tawel Gogledd America", "SAST": "Amser Safonol De Affrica", "SGT": "Amser Singapore", "SRT": "Amser Suriname", "TMST": "Amser Haf Tyrcmenistan", "TMT": "Amser Safonol Tyrcmenistan", "UYST": "Amser Haf Uruguay", "UYT": "Amser Safonol Uruguay", "VET": "Amser Venezuela", "WARST": "Amser Haf Gorllewin Ariannin", "WART": "Amser Safonol Gorllewin Ariannin", "WAST": "Amser Haf Gorllewin Affrica", "WAT": "Amser Safonol Gorllewin Affrica", "WESZ": "Amser Haf Gorllewin Ewrop", "WEZ": "Amser Safonol Gorllewin Ewrop", "WIB": "Amser Gorllewin Indonesia", "WIT": "Amser Dwyrain Indonesia", "WITA": "Amser Canolbarth Indonesia", "∅∅∅": "Amser Haf Amazonas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6ce82b971f5322e99200abf4c11a0dbe,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_01e534b5c587a0b20e565363a8f2673c,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_36aaa84f53bfb88e0054a81c3abba2a4,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["cy_gb"] = func() locales.Translator {
		return &localen{

			locale:                 "cy_GB",
			pluralsCardinal:        []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 3, 4, 5, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ion", "Chwef", "Maw", "Ebr", "Mai", "Meh", "Gorff", "Awst", "Medi", "Hyd", "Tach", "Rhag"},
			monthsNarrow:           []string{"", "I", "Ch", "M", "E", "M", "M", "G", "A", "M", "H", "T", "Rh"},
			monthsWide:             []string{"", "Ionawr", "Chwefror", "Mawrth", "Ebrill", "Mai", "Mehefin", "Gorffennaf", "Awst", "Medi", "Hydref", "Tachwedd", "Rhagfyr"},
			daysAbbreviated:        []string{"Sul", "Llun", "Maw", "Mer", "Iau", "Gwen", "Sad"},
			daysNarrow:             []string{"S", "Ll", "M", "M", "I", "G", "S"},
			daysShort:              []string{"Su", "Ll", "Ma", "Me", "Ia", "Gw", "Sa"},
			daysWide:               []string{"Dydd Sul", "Dydd Llun", "Dydd Mawrth", "Dydd Mercher", "Dydd Iau", "Dydd Gwener", "Dydd Sadwrn"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"b", "h"},
			periodsWide:            []string{"yb", "yh"},
			erasAbbreviated:        []string{"CC", "OC"},
			erasNarrow:             []string{"C", "O"},
			erasWide:               []string{"Cyn Crist", "Oed Crist"},
			timezones:              map[string]string{"ACDT": "Amser Haf Canolbarth Awstralia", "ACST": "Amser Safonol Canolbarth Awstralia", "ACWDT": "Amser Haf Canolbarth Gorllewin Awstralia", "ACWST": "Amser Safonol Canolbarth Gorllewin Awstralia", "ADT": "Amser Haf Cefnfor yr Iwerydd", "AEDT": "Amser Haf Dwyrain Awstralia", "AEST": "Amser Safonol Dwyrain Awstralia", "AKDT": "Amser Haf Alaska", "AKST": "Amser Safonol Alaska", "ARST": "Amser Haf Ariannin", "ART": "Amser Safonol Ariannin", "AST": "Amser Safonol Cefnfor yr Iwerydd", "AWDT": "Amser Haf Gorllewin Awstralia", "AWST": "Amser Safonol Gorllewin Awstralia", "BOT": "Amser Bolivia", "BT": "Amser Bhutan", "CAT": "Amser Canolbarth Affrica", "CDT": "Amser Haf Canolbarth Gogledd America", "CHADT": "Amser Haf Chatham", "CHAST": "Amser Safonol Chatham", "CLST": "Amser Haf Chile", "CLT": "Amser Safonol Chile", "COST": "Amser Haf Colombia", "COT": "Amser Safonol Colombia", "CST": "Amser Safonol Canolbarth Gogledd America", "ChST": "Amser Chamorro", "EAT": "Amser Dwyrain Affrica", "ECT": "Amser Ecuador", "EDT": "Amser Haf Dwyrain Gogledd America", "EST": "Amser Safonol Dwyrain Gogledd America", "GFT": "Amser Guyane Ffrengig", "GMT": "Amser Safonol Greenwich", "GST": "Amser Safonol y Gwlff", "GYT": "Amser Guyana", "HADT": "Amser Haf Hawaii-Aleutian", "HAST": "Amser Safonol Hawaii-Aleutian", "HAT": "Amser Haf Newfoundland", "HECU": "Amser Haf Cuba", "HEEG": "Amser Haf Dwyrain yr Ynys Las", "HENOMX": "Amser Haf Gogledd Orllewin Mecsico", "HEOG": "Amser Haf Gorllewin yr Ynys Las", "HEPM": "Amser Haf Saint-Pierre-et-Miquelon", "HEPMX": "Amser Haf Pasiffig Mecsico", "HKST": "Amser Haf Hong Kong", "HKT": "Amser Safonol Hong Kong", "HNCU": "Amser Safonol Cuba", "HNEG": "Amser Safonol Dwyrain yr Ynys Las", "HNNOMX": "Amser Safonol Gogledd Orllewin Mecsico", "HNOG": "Amser Safonol Gorllewin yr Ynys Las", "HNPM": "Amser Safonol Saint-Pierre-et-Miquelon", "HNPMX": "Amser Safonol Pasiffig Mecsico", "HNT": "Amser Safonol Newfoundland", "IST": "Amser India", "JDT": "Amser Haf Siapan", "JST": "Amser Safonol Siapan", "LHDT": "Amser Haf yr Arglwydd Howe", "LHST": "Amser Safonol yr Arglwydd Howe", "MDT": "Amser Haf Mynyddoedd Gogledd America", "MESZ": "Amser Haf Canolbarth Ewrop", "MEZ": "Amser Safonol Canolbarth Ewrop", "MST": "Amser Safonol Mynyddoedd Gogledd America", "MYT": "Amser Malaysia", "NZDT": "Amser Haf Seland Newydd", "NZST": "Amser Safonol Seland Newydd", "OESZ": "Amser Haf Dwyrain Ewrop", "OEZ": "Amser Safonol Dwyrain Ewrop", "PDT": "Amser Haf Cefnfor Tawel Gogledd America", "PST": "Amser Safonol Cefnfor Tawel Gogledd America", "SAST": "Amser Safonol De Affrica", "SGT": "Amser Singapore", "SRT": "Amser Suriname", "TMST": "Amser Haf Tyrcmenistan", "TMT": "Amser Safonol Tyrcmenistan", "UYST": "Amser Haf Uruguay", "UYT": "Amser Safonol Uruguay", "VET": "Amser Venezuela", "WARST": "Amser Haf Gorllewin Ariannin", "WART": "Amser Safonol Gorllewin Ariannin", "WAST": "Amser Haf Gorllewin Affrica", "WAT": "Amser Safonol Gorllewin Affrica", "WESZ": "Amser Haf Gorllewin Ewrop", "WEZ": "Amser Safonol Gorllewin Ewrop", "WIB": "Amser Gorllewin Indonesia", "WIT": "Amser Dwyrain Indonesia", "WITA": "Amser Canolbarth Indonesia", "∅∅∅": "Amser Haf Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6ce82b971f5322e99200abf4c11a0dbe,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_01e534b5c587a0b20e565363a8f2673c,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_36aaa84f53bfb88e0054a81c3abba2a4,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["da"] = func() locales.Translator {
		return &localen{

			locale:                 "da",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "kr.", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maj", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "marts", "april", "maj", "juni", "juli", "august", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"søn.", "man.", "tir.", "ons.", "tor.", "fre.", "lør."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø", "ma", "ti", "on", "to", "fr", "lø"},
			daysWide:               []string{"søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"fKr", "eKr"},
			erasWide:               []string{"f.Kr.", "e.Kr."},
			timezones:              map[string]string{"ACDT": "Centralaustralsk sommertid", "ACST": "Centralaustralsk normaltid", "ACWDT": "Vestlig centralaustralsk sommertid", "ACWST": "Vestlig centralaustralsk normaltid", "ADT": "Atlantic-sommertid", "AEDT": "Østaustralsk sommertid", "AEST": "Østaustralsk normaltid", "AKDT": "Alaska-sommertid", "AKST": "Alaska-normaltid", "ARST": "Argentinsk sommertid", "ART": "Argentinsk normaltid", "AST": "Atlantic-normaltid", "AWDT": "Vestaustralsk sommertid", "AWST": "Vestaustralsk normaltid", "BOT": "Boliviansk tid", "BT": "Bhutan-tid", "CAT": "Centralafrikansk tid", "CDT": "Central-sommertid", "CHADT": "Chatham-sommertid", "CHAST": "Chatham-normaltid", "CLST": "Chilensk sommertid", "CLT": "Chilensk normaltid", "COST": "Colombiansk sommertid", "COT": "Colombiansk normaltid", "CST": "Central-normaltid", "ChST": "Chamorro-tid", "EAT": "Østafrikansk tid", "ECT": "Ecuadoriansk tid", "EDT": "Eastern-sommertid", "EST": "Eastern-normaltid", "GFT": "Fransk Guyana-tid", "GMT": "GMT", "GST": "Golflandene-normaltid", "GYT": "Guyana-tid", "HADT": "Hawaii-Aleutian-sommertid", "HAST": "Hawaii-Aleutian-normaltid", "HAT": "Newfoundlandsk sommertid", "HECU": "Cubansk sommertid", "HEEG": "Østgrønlandsk sommertid", "HENOMX": "Nordvestmexicansk sommertid", "HEOG": "Vestgrønlandsk sommertid", "HEPM": "Saint Pierre- og Miquelon-sommertid", "HEPMX": "Mexicansk Pacific-sommertid", "HKST": "Hongkong-sommertid", "HKT": "Hongkong-normaltid", "HNCU": "Cubansk normaltid", "HNEG": "Østgrønlandsk normaltid", "HNNOMX": "Nordvestmexicansk normaltid", "HNOG": "Vestgrønlandsk normaltid", "HNPM": "Saint Pierre- og Miquelon-normaltid", "HNPMX": "Mexicansk Pacific-normaltid", "HNT": "Newfoundlandsk normaltid", "IST": "Indisk normaltid", "JDT": "Japansk sommertid", "JST": "Japansk normaltid", "LHDT": "Lord Howe-sommertid", "LHST": "Lord Howe-normaltid", "MDT": "Mountain-sommertid", "MESZ": "Centraleuropæisk sommertid", "MEZ": "Centraleuropæisk normaltid", "MST": "Mountain-normaltid", "MYT": "Malaysia-tid", "NZDT": "Newzealandsk sommertid", "NZST": "Newzealandsk normaltid", "OESZ": "Østeuropæisk sommertid", "OEZ": "Østeuropæisk normaltid", "PDT": "Pacific-sommertid", "PST": "Pacific-normaltid", "SAST": "Sydafrikansk tid", "SGT": "Singapore-tid", "SRT": "Surinam-tid", "TMST": "Turkmensk sommertid", "TMT": "Turkmensk normaltid", "UYST": "Uruguayansk sommertid", "UYT": "Uruguayansk normaltid", "VET": "Venezuelansk tid", "WARST": "Vestargentinsk sommertid", "WART": "Vestargentinsk normaltid", "WAST": "Vestafrikansk sommertid", "WAT": "Vestafrikansk normaltid", "WESZ": "Vesteuropæisk sommertid", "WEZ": "Vesteuropæisk normaltid", "WIB": "Vestindonesisk tid", "WIT": "Østindonesisk tid", "WITA": "Centralindonesisk tid", "∅∅∅": "Brasiliansk sommertid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_79907f4d6e752731d2282b8c488dfb0e,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_8ce43d140d292ed841f0f368561bdb8e,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["da_dk"] = func() locales.Translator {
		return &localen{

			locale:                 "da_DK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maj", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "marts", "april", "maj", "juni", "juli", "august", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"søn.", "man.", "tir.", "ons.", "tor.", "fre.", "lør."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø", "ma", "ti", "on", "to", "fr", "lø"},
			daysWide:               []string{"søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"fKr", "eKr"},
			erasWide:               []string{"f.Kr.", "e.Kr."},
			timezones:              map[string]string{"ACDT": "Centralaustralsk sommertid", "ACST": "Centralaustralsk normaltid", "ACWDT": "Vestlig centralaustralsk sommertid", "ACWST": "Vestlig centralaustralsk normaltid", "ADT": "Atlantic-sommertid", "AEDT": "Østaustralsk sommertid", "AEST": "Østaustralsk normaltid", "AKDT": "Alaska-sommertid", "AKST": "Alaska-normaltid", "ARST": "Argentinsk sommertid", "ART": "Argentinsk normaltid", "AST": "Atlantic-normaltid", "AWDT": "Vestaustralsk sommertid", "AWST": "Vestaustralsk normaltid", "BOT": "Boliviansk tid", "BT": "Bhutan-tid", "CAT": "Centralafrikansk tid", "CDT": "Central-sommertid", "CHADT": "Chatham-sommertid", "CHAST": "Chatham-normaltid", "CLST": "Chilensk sommertid", "CLT": "Chilensk normaltid", "COST": "Colombiansk sommertid", "COT": "Colombiansk normaltid", "CST": "Central-normaltid", "ChST": "Chamorro-tid", "EAT": "Østafrikansk tid", "ECT": "Ecuadoriansk tid", "EDT": "Eastern-sommertid", "EST": "Eastern-normaltid", "GFT": "Fransk Guyana-tid", "GMT": "GMT", "GST": "Golflandene-normaltid", "GYT": "Guyana-tid", "HADT": "Hawaii-Aleutian-sommertid", "HAST": "Hawaii-Aleutian-normaltid", "HAT": "Newfoundlandsk sommertid", "HECU": "Cubansk sommertid", "HEEG": "Østgrønlandsk sommertid", "HENOMX": "Nordvestmexicansk sommertid", "HEOG": "Vestgrønlandsk sommertid", "HEPM": "Saint Pierre- og Miquelon-sommertid", "HEPMX": "Mexicansk Pacific-sommertid", "HKST": "Hongkong-sommertid", "HKT": "Hongkong-normaltid", "HNCU": "Cubansk normaltid", "HNEG": "Østgrønlandsk normaltid", "HNNOMX": "Nordvestmexicansk normaltid", "HNOG": "Vestgrønlandsk normaltid", "HNPM": "Saint Pierre- og Miquelon-normaltid", "HNPMX": "Mexicansk Pacific-normaltid", "HNT": "Newfoundlandsk normaltid", "IST": "Indisk normaltid", "JDT": "Japansk sommertid", "JST": "Japansk normaltid", "LHDT": "Lord Howe-sommertid", "LHST": "Lord Howe-normaltid", "MDT": "Mountain-sommertid", "MESZ": "Centraleuropæisk sommertid", "MEZ": "Centraleuropæisk normaltid", "MST": "Mountain-normaltid", "MYT": "Malaysia-tid", "NZDT": "Newzealandsk sommertid", "NZST": "Newzealandsk normaltid", "OESZ": "Østeuropæisk sommertid", "OEZ": "Østeuropæisk normaltid", "PDT": "Pacific-sommertid", "PST": "Pacific-normaltid", "SAST": "Sydafrikansk tid", "SGT": "Singapore-tid", "SRT": "Surinam-tid", "TMST": "Turkmensk sommertid", "TMT": "Turkmensk normaltid", "UYST": "Uruguayansk sommertid", "UYT": "Uruguayansk normaltid", "VET": "Venezuelansk tid", "WARST": "Vestargentinsk sommertid", "WART": "Vestargentinsk normaltid", "WAST": "Vestafrikansk sommertid", "WAT": "Vestafrikansk normaltid", "WESZ": "Vesteuropæisk sommertid", "WEZ": "Vesteuropæisk normaltid", "WIB": "Vestindonesisk tid", "WIT": "Østindonesisk tid", "WITA": "Centralindonesisk tid", "∅∅∅": "Brasiliansk sommertid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_79907f4d6e752731d2282b8c488dfb0e,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_8ce43d140d292ed841f0f368561bdb8e,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["da_gl"] = func() locales.Translator {
		return &localen{

			locale:                 "da_GL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maj", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "marts", "april", "maj", "juni", "juli", "august", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"søn.", "man.", "tir.", "ons.", "tor.", "fre.", "lør."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø", "ma", "ti", "on", "to", "fr", "lø"},
			daysWide:               []string{"søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"fKr", "eKr"},
			erasWide:               []string{"f.Kr.", "e.Kr."},
			timezones:              map[string]string{"ACDT": "Centralaustralsk sommertid", "ACST": "Centralaustralsk normaltid", "ACWDT": "Vestlig centralaustralsk sommertid", "ACWST": "Vestlig centralaustralsk normaltid", "ADT": "Atlantic-sommertid", "AEDT": "Østaustralsk sommertid", "AEST": "Østaustralsk normaltid", "AKDT": "Alaska-sommertid", "AKST": "Alaska-normaltid", "ARST": "Argentinsk sommertid", "ART": "Argentinsk normaltid", "AST": "Atlantic-normaltid", "AWDT": "Vestaustralsk sommertid", "AWST": "Vestaustralsk normaltid", "BOT": "Boliviansk tid", "BT": "Bhutan-tid", "CAT": "Centralafrikansk tid", "CDT": "Central-sommertid", "CHADT": "Chatham-sommertid", "CHAST": "Chatham-normaltid", "CLST": "Chilensk sommertid", "CLT": "Chilensk normaltid", "COST": "Colombiansk sommertid", "COT": "Colombiansk normaltid", "CST": "Central-normaltid", "ChST": "Chamorro-tid", "EAT": "Østafrikansk tid", "ECT": "Ecuadoriansk tid", "EDT": "Eastern-sommertid", "EST": "Eastern-normaltid", "GFT": "Fransk Guyana-tid", "GMT": "GMT", "GST": "Golflandene-normaltid", "GYT": "Guyana-tid", "HADT": "Hawaii-Aleutian-sommertid", "HAST": "Hawaii-Aleutian-normaltid", "HAT": "Newfoundlandsk sommertid", "HECU": "Cubansk sommertid", "HEEG": "Østgrønlandsk sommertid", "HENOMX": "Nordvestmexicansk sommertid", "HEOG": "Vestgrønlandsk sommertid", "HEPM": "Saint Pierre- og Miquelon-sommertid", "HEPMX": "Mexicansk Pacific-sommertid", "HKST": "Hongkong-sommertid", "HKT": "Hongkong-normaltid", "HNCU": "Cubansk normaltid", "HNEG": "Østgrønlandsk normaltid", "HNNOMX": "Nordvestmexicansk normaltid", "HNOG": "Vestgrønlandsk normaltid", "HNPM": "Saint Pierre- og Miquelon-normaltid", "HNPMX": "Mexicansk Pacific-normaltid", "HNT": "Newfoundlandsk normaltid", "IST": "Indisk normaltid", "JDT": "Japansk sommertid", "JST": "Japansk normaltid", "LHDT": "Lord Howe-sommertid", "LHST": "Lord Howe-normaltid", "MDT": "Mountain-sommertid", "MESZ": "Centraleuropæisk sommertid", "MEZ": "Centraleuropæisk normaltid", "MST": "Mountain-normaltid", "MYT": "Malaysia-tid", "NZDT": "Newzealandsk sommertid", "NZST": "Newzealandsk normaltid", "OESZ": "Østeuropæisk sommertid", "OEZ": "Østeuropæisk normaltid", "PDT": "Pacific-sommertid", "PST": "Pacific-normaltid", "SAST": "Sydafrikansk tid", "SGT": "Singapore-tid", "SRT": "Surinam-tid", "TMST": "Turkmensk sommertid", "TMT": "Turkmensk normaltid", "UYST": "Uruguayansk sommertid", "UYT": "Uruguayansk normaltid", "VET": "Venezuelansk tid", "WARST": "Vestargentinsk sommertid", "WART": "Vestargentinsk normaltid", "WAST": "Vestafrikansk sommertid", "WAT": "Vestafrikansk normaltid", "WESZ": "Vesteuropæisk sommertid", "WEZ": "Vesteuropæisk normaltid", "WIB": "Vestindonesisk tid", "WIT": "Østindonesisk tid", "WITA": "Centralindonesisk tid", "∅∅∅": "Azorerne-sommertid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_79907f4d6e752731d2282b8c488dfb0e,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_8ce43d140d292ed841f0f368561bdb8e,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dav"] = func() locales.Translator {
		return &localen{

			locale:                 "dav",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Imb", "Kaw", "Kad", "Kan", "Kas", "Kar", "Mfu", "Wun", "Ike", "Iku", "Imw", "Iwi"},
			monthsNarrow:           []string{"", "I", "K", "K", "K", "K", "K", "M", "W", "I", "I", "I", "I"},
			monthsWide:             []string{"", "Mori ghwa imbiri", "Mori ghwa kawi", "Mori ghwa kadadu", "Mori ghwa kana", "Mori ghwa kasanu", "Mori ghwa karandadu", "Mori ghwa mfungade", "Mori ghwa wunyanya", "Mori ghwa ikenda", "Mori ghwa ikumi", "Mori ghwa ikumi na imweri", "Mori ghwa ikumi na iwi"},
			daysAbbreviated:        []string{"Jum", "Jim", "Kaw", "Kad", "Kan", "Kas", "Ngu"},
			daysNarrow:             []string{"J", "J", "K", "K", "K", "K", "N"},
			daysWide:               []string{"Ituku ja jumwa", "Kuramuka jimweri", "Kuramuka kawi", "Kuramuka kadadu", "Kuramuka kana", "Kuramuka kasanu", "Kifula nguwo"},
			periodsAbbreviated:     []string{"Luma lwa K", "luma lwa p"},
			periodsWide:            []string{"Luma lwa K", "luma lwa p"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dav_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "dav_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Imb", "Kaw", "Kad", "Kan", "Kas", "Kar", "Mfu", "Wun", "Ike", "Iku", "Imw", "Iwi"},
			monthsNarrow:           []string{"", "I", "K", "K", "K", "K", "K", "M", "W", "I", "I", "I", "I"},
			monthsWide:             []string{"", "Mori ghwa imbiri", "Mori ghwa kawi", "Mori ghwa kadadu", "Mori ghwa kana", "Mori ghwa kasanu", "Mori ghwa karandadu", "Mori ghwa mfungade", "Mori ghwa wunyanya", "Mori ghwa ikenda", "Mori ghwa ikumi", "Mori ghwa ikumi na imweri", "Mori ghwa ikumi na iwi"},
			daysAbbreviated:        []string{"Jum", "Jim", "Kaw", "Kad", "Kan", "Kas", "Ngu"},
			daysNarrow:             []string{"J", "J", "K", "K", "K", "K", "N"},
			daysWide:               []string{"Ituku ja jumwa", "Kuramuka jimweri", "Kuramuka kawi", "Kuramuka kadadu", "Kuramuka kana", "Kuramuka kasanu", "Kifula nguwo"},
			periodsAbbreviated:     []string{"Luma lwa K", "luma lwa p"},
			periodsWide:            []string{"Luma lwa K", "luma lwa p"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de"] = func() locales.Translator {
		return &localen{

			locale:                 "de",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "öS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGK", "BGN", "BGJ", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sept.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Peruanische Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_at"] = func() locales.Translator {
		return &localen{

			locale:                 "de_AT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jän.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Jänner", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"vm.", "nm."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Brasília-Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_204cc4159c06830dc072d7ffc4471b59,
			fnFmtCurrency:         fnFmtCurrency_7b65395a77761388c242d97c0d4bfe5b,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_be"] = func() locales.Translator {
		return &localen{

			locale:                 "de_BE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sept.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Azoren-Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "de_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sept.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Azoren-Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a6430b96a17c400006d20ebabdf05e4c,
			fnFmtCurrency:         fnFmtCurrency_4924273cdfaa6d993990fdf2a2bad453,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_de"] = func() locales.Translator {
		return &localen{

			locale:                 "de_DE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sept.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Brasília-Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_it"] = func() locales.Translator {
		return &localen{

			locale:                 "de_IT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jän.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Jänner", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Peruanische Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_li"] = func() locales.Translator {
		return &localen{

			locale:                 "de_LI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sept.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"vm.", "nm."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Azoren-Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a6430b96a17c400006d20ebabdf05e4c,
			fnFmtCurrency:         fnFmtCurrency_4924273cdfaa6d993990fdf2a2bad453,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["de_lu"] = func() locales.Translator {
		return &localen{

			locale:                 "de_LU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "F", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sept.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mo.", "Di.", "Mi.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"vorm.", "nachm."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralische Sommerzeit", "ACST": "Zentralaustralische Normalzeit", "ACWDT": "Zentral-/Westaustralische Sommerzeit", "ACWST": "Zentral-/Westaustralische Normalzeit", "ADT": "Atlantik-Sommerzeit", "AEDT": "Ostaustralische Sommerzeit", "AEST": "Ostaustralische Normalzeit", "AKDT": "Alaska-Sommerzeit", "AKST": "Alaska-Normalzeit", "ARST": "Argentinische Sommerzeit", "ART": "Argentinische Normalzeit", "AST": "Atlantik-Normalzeit", "AWDT": "Westaustralische Sommerzeit", "AWST": "Westaustralische Normalzeit", "BOT": "Bolivianische Zeit", "BT": "Bhutan-Zeit", "CAT": "Zentralafrikanische Zeit", "CDT": "Nordamerikanische Inland-Sommerzeit", "CHADT": "Chatham-Sommerzeit", "CHAST": "Chatham-Normalzeit", "CLST": "Chilenische Sommerzeit", "CLT": "Chilenische Normalzeit", "COST": "Kolumbianische Sommerzeit", "COT": "Kolumbianische Normalzeit", "CST": "Nordamerikanische Inland-Normalzeit", "ChST": "Chamorro-Zeit", "EAT": "Ostafrikanische Zeit", "ECT": "Ecuadorianische Zeit", "EDT": "Nordamerikanische Ostküsten-Sommerzeit", "EST": "Nordamerikanische Ostküsten-Normalzeit", "GFT": "Französisch-Guayana-Zeit", "GMT": "Mittlere Greenwich-Zeit", "GST": "Golf-Zeit", "GYT": "Guyana-Zeit", "HADT": "Hawaii-Aleuten-Sommerzeit", "HAST": "Hawaii-Aleuten-Normalzeit", "HAT": "Neufundland-Sommerzeit", "HECU": "Kubanische Sommerzeit", "HEEG": "Ostgrönland-Sommerzeit", "HENOMX": "Mexiko Nordwestliche Zone-Sommerzeit", "HEOG": "Westgrönland-Sommerzeit", "HEPM": "St.-Pierre-und-Miquelon-Sommerzeit", "HEPMX": "Mexiko Pazifikzone-Sommerzeit", "HKST": "Hongkong-Sommerzeit", "HKT": "Hongkong-Normalzeit", "HNCU": "Kubanische Normalzeit", "HNEG": "Ostgrönland-Normalzeit", "HNNOMX": "Mexiko Nordwestliche Zone-Normalzeit", "HNOG": "Westgrönland-Normalzeit", "HNPM": "St.-Pierre-und-Miquelon-Normalzeit", "HNPMX": "Mexiko Pazifikzone-Normalzeit", "HNT": "Neufundland-Normalzeit", "IST": "Indische Zeit", "JDT": "Japanische Sommerzeit", "JST": "Japanische Normalzeit", "LHDT": "Lord-Howe-Sommerzeit", "LHST": "Lord-Howe-Normalzeit", "MDT": "Rocky-Mountain-Sommerzeit", "MESZ": "Mitteleuropäische Sommerzeit", "MEZ": "Mitteleuropäische Normalzeit", "MST": "Rocky Mountain-Normalzeit", "MYT": "Malaysische Zeit", "NZDT": "Neuseeland-Sommerzeit", "NZST": "Neuseeland-Normalzeit", "OESZ": "Osteuropäische Sommerzeit", "OEZ": "Osteuropäische Normalzeit", "PDT": "Nordamerikanische Westküsten-Sommerzeit", "PST": "Nordamerikanische Westküsten-Normalzeit", "SAST": "Südafrikanische Zeit", "SGT": "Singapur-Zeit", "SRT": "Suriname-Zeit", "TMST": "Turkmenistan-Sommerzeit", "TMT": "Turkmenistan-Normalzeit", "UYST": "Uruguayanische Sommerzeit", "UYT": "Uruguyanische Normalzeit", "VET": "Venezuela-Zeit", "WARST": "Westargentinische Sommerzeit", "WART": "Westargentinische Normalzeit", "WAST": "Westafrikanische Sommerzeit", "WAT": "Westafrikanische Normalzeit", "WESZ": "Westeuropäische Sommerzeit", "WEZ": "Westeuropäische Normalzeit", "WIB": "Westindonesische Zeit", "WIT": "Ostindonesische Zeit", "WITA": "Zentralindonesische Zeit", "∅∅∅": "Peruanische Sommerzeit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dje"] = func() locales.Translator {
		return &localen{

			locale:             "dje",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alz", "Asi"},
			daysNarrow:         []string{"H", "T", "T", "L", "M", "Z", "S"},
			daysWide:           []string{"Alhadi", "Atinni", "Atalaata", "Alarba", "Alhamisi", "Alzuma", "Asibti"},
			periodsAbbreviated: []string{"Subbaahi", "Zaarikay b"},
			periodsWide:        []string{"Subbaahi", "Zaarikay b"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa zamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dje_ne"] = func() locales.Translator {
		return &localen{

			locale:             "dje_NE",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alz", "Asi"},
			daysNarrow:         []string{"H", "T", "T", "L", "M", "Z", "S"},
			daysWide:           []string{"Alhadi", "Atinni", "Atalaata", "Alarba", "Alhamisi", "Alzuma", "Asibti"},
			periodsAbbreviated: []string{"Subbaahi", "Zaarikay b"},
			periodsWide:        []string{"Subbaahi", "Zaarikay b"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa zamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dsb"] = func() locales.Translator {
		return &localen{

			locale:                 "dsb",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "zł", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "měr.", "apr.", "maj.", "jun.", "jul.", "awg.", "sep.", "okt.", "now.", "dec."},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januara", "februara", "měrca", "apryla", "maja", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"},
			daysAbbreviated:        []string{"nje", "pón", "wał", "srj", "stw", "pět", "sob"},
			daysNarrow:             []string{"n", "p", "w", "s", "s", "p", "s"},
			daysShort:              []string{"nj", "pó", "wa", "sr", "st", "pě", "so"},
			daysWide:               []string{"njeźela", "pónjeźele", "wałtora", "srjoda", "stwórtk", "pětk", "sobota"},
			periodsAbbreviated:     []string{"dopołdnja", "wótpołdnja"},
			periodsNarrow:          []string{"dop.", "wótp."},
			periodsWide:            []string{"dopołdnja", "wótpołdnja"},
			erasAbbreviated:        []string{"pś.Chr.n.", "pó Chr.n."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"pśed Kristusowym naroźenim", "pó Kristusowem naroźenju"},
			timezones:              map[string]string{"ACDT": "Srjejźoawstralski lěśojski cas", "ACST": "Srjejźoawstralski standardny cas", "ACWDT": "Srjejźopódwjacorny awstralski lěśojski cas", "ACWST": "Srjejźopódwjacorny awstralski standardny cas", "ADT": "Atlantiski lěśojski cas", "AEDT": "Pódzajtšnoawstralski lěśojski cas", "AEST": "Pódzajtšnoawstralski standardny cas", "AKDT": "Alaskojski lěśojski cas", "AKST": "Alaskojski standardny cas", "ARST": "Argentinski lěśojski cas", "ART": "Argentinski standardny cas", "AST": "Atlantiski standardny cas", "AWDT": "Pódwjacornoawstralski lěśojski cas", "AWST": "Pódwjacornoawstralski standardny cas", "BOT": "Boliwiski cas", "BT": "Bhutański cas", "CAT": "Srjejźoafriski cas", "CDT": "Pódpołnocnoameriski centralny lěśojski cas", "CHADT": "Chathamski lěśojski cas", "CHAST": "Chathamski standardny cas", "CLST": "Chilski lěśojski cas", "CLT": "Chilski standardny cas", "COST": "Kolumbiski lěśojski cas", "COT": "Kolumbiski standardny cas", "CST": "Pódpołnocnoameriski centralny standardny cas", "ChST": "Chamorrski cas", "EAT": "Pódzajtšnoafriski cas", "ECT": "Ekuadorski cas", "EDT": "Pódpołnocnoameriski pódzajtšny lěśojski cas", "EST": "Pódpołnocnoameriski pódzajtšny standardny cas", "GFT": "Francojskoguyański cas", "GMT": "Greenwichski cas", "GST": "cas Persiskego golfa", "GYT": "Guyański cas", "HADT": "Hawaiisko-aleutski lěśojski cas", "HAST": "Hawaiisko-aleutski standardny cas", "HAT": "Nowofundlandski lěśojski cas", "HECU": "Kubański lěśojski cas", "HEEG": "Pódzajtšnogrönlandski lěśojski cas", "HENOMX": "Mexiski dłujkowjacorny lěśojski cas", "HEOG": "Pódwjacornogrönlandski lěśojski cas", "HEPM": "St.-Pierre-a-Miqueloński lěśojski cas", "HEPMX": "Mexiski pacifiski lěśojski cas", "HKST": "Hongkongski lěśojski cas", "HKT": "Hongkongski standardny cas", "HNCU": "Kubański standardny cas", "HNEG": "Pódzajtšnogrönlandski standardny cas", "HNNOMX": "Mexiski dłujkowjacorny standardny cas", "HNOG": "Pódwjacornogrönlandski standardny cas", "HNPM": "St.-Pierre-a-Miqueloński standardny cas", "HNPMX": "Mexiski pacifiski standardny cas", "HNT": "Nowofundlandski standardny cas", "IST": "Indiski cas", "JDT": "Japański lěśojski cas", "JST": "Japański standardny cas", "LHDT": "lěśojski cas kupy Lord-Howe", "LHST": "Standardny cas kupy Lord-Howe", "MDT": "Pódpołnocnoameriski górski lěśojski cas", "MESZ": "Srjejźoeuropski lěśojski cas", "MEZ": "Srjejźoeuropski standardny cas", "MST": "Pódpołnocnoameriski górski standardny cas", "MYT": "Malajziski cas", "NZDT": "Nowoseelandski lěśojski cas", "NZST": "Nowoseelandski standardny cas", "OESZ": "Pódzajtšnoeuropski lěśojski cas", "OEZ": "Pódzajtšnoeuropski standardny cas", "PDT": "Pódpołnocnoameriski pacifiski lěśojski cas", "PST": "Pódpołnocnoameriski pacifiski standardny cas", "SAST": "Pódpołdnjowoafriski cas", "SGT": "Singapurski cas", "SRT": "Surinamski cas", "TMST": "Turkmeniski lěśojski cas", "TMT": "Turkmeniski standardny cas", "UYST": "Uruguayski lěśojski cas", "UYT": "Uruguayski standardny cas", "VET": "Venezuelski cas", "WARST": "Pódwjacornoargentinski lěśojski cas", "WART": "Pódwjacornoargentinski standardny cas", "WAST": "Pódwjacornoafriski lěśojski cas", "WAT": "Pódwjacornoafriski standardny cas", "WESZ": "Pódwjacornoeuropski lěśojski cas", "WEZ": "Pódwjacornoeuropski standardny cas", "WIB": "Pódwjacornoindoneski cas", "WIT": "Pódzajtšnoindoneski", "WITA": "Srjejźoindoneski cas", "∅∅∅": "Acorski lěśojski cas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_283d29635c231a3ca301081bab098305,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dsb_de"] = func() locales.Translator {
		return &localen{

			locale:                 "dsb_DE",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "měr.", "apr.", "maj.", "jun.", "jul.", "awg.", "sep.", "okt.", "now.", "dec."},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januara", "februara", "měrca", "apryla", "maja", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"},
			daysAbbreviated:        []string{"nje", "pón", "wał", "srj", "stw", "pět", "sob"},
			daysNarrow:             []string{"n", "p", "w", "s", "s", "p", "s"},
			daysShort:              []string{"nj", "pó", "wa", "sr", "st", "pě", "so"},
			daysWide:               []string{"njeźela", "pónjeźele", "wałtora", "srjoda", "stwórtk", "pětk", "sobota"},
			periodsAbbreviated:     []string{"dopołdnja", "wótpołdnja"},
			periodsNarrow:          []string{"dop.", "wótp."},
			periodsWide:            []string{"dopołdnja", "wótpołdnja"},
			erasAbbreviated:        []string{"pś.Chr.n.", "pó Chr.n."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"pśed Kristusowym naroźenim", "pó Kristusowem naroźenju"},
			timezones:              map[string]string{"ACDT": "Srjejźoawstralski lěśojski cas", "ACST": "Srjejźoawstralski standardny cas", "ACWDT": "Srjejźopódwjacorny awstralski lěśojski cas", "ACWST": "Srjejźopódwjacorny awstralski standardny cas", "ADT": "Atlantiski lěśojski cas", "AEDT": "Pódzajtšnoawstralski lěśojski cas", "AEST": "Pódzajtšnoawstralski standardny cas", "AKDT": "Alaskojski lěśojski cas", "AKST": "Alaskojski standardny cas", "ARST": "Argentinski lěśojski cas", "ART": "Argentinski standardny cas", "AST": "Atlantiski standardny cas", "AWDT": "Pódwjacornoawstralski lěśojski cas", "AWST": "Pódwjacornoawstralski standardny cas", "BOT": "Boliwiski cas", "BT": "Bhutański cas", "CAT": "Srjejźoafriski cas", "CDT": "Pódpołnocnoameriski centralny lěśojski cas", "CHADT": "Chathamski lěśojski cas", "CHAST": "Chathamski standardny cas", "CLST": "Chilski lěśojski cas", "CLT": "Chilski standardny cas", "COST": "Kolumbiski lěśojski cas", "COT": "Kolumbiski standardny cas", "CST": "Pódpołnocnoameriski centralny standardny cas", "ChST": "Chamorrski cas", "EAT": "Pódzajtšnoafriski cas", "ECT": "Ekuadorski cas", "EDT": "Pódpołnocnoameriski pódzajtšny lěśojski cas", "EST": "Pódpołnocnoameriski pódzajtšny standardny cas", "GFT": "Francojskoguyański cas", "GMT": "Greenwichski cas", "GST": "cas Persiskego golfa", "GYT": "Guyański cas", "HADT": "Hawaiisko-aleutski lěśojski cas", "HAST": "Hawaiisko-aleutski standardny cas", "HAT": "Nowofundlandski lěśojski cas", "HECU": "Kubański lěśojski cas", "HEEG": "Pódzajtšnogrönlandski lěśojski cas", "HENOMX": "Mexiski dłujkowjacorny lěśojski cas", "HEOG": "Pódwjacornogrönlandski lěśojski cas", "HEPM": "St.-Pierre-a-Miqueloński lěśojski cas", "HEPMX": "Mexiski pacifiski lěśojski cas", "HKST": "Hongkongski lěśojski cas", "HKT": "Hongkongski standardny cas", "HNCU": "Kubański standardny cas", "HNEG": "Pódzajtšnogrönlandski standardny cas", "HNNOMX": "Mexiski dłujkowjacorny standardny cas", "HNOG": "Pódwjacornogrönlandski standardny cas", "HNPM": "St.-Pierre-a-Miqueloński standardny cas", "HNPMX": "Mexiski pacifiski standardny cas", "HNT": "Nowofundlandski standardny cas", "IST": "Indiski cas", "JDT": "Japański lěśojski cas", "JST": "Japański standardny cas", "LHDT": "lěśojski cas kupy Lord-Howe", "LHST": "Standardny cas kupy Lord-Howe", "MDT": "Pódpołnocnoameriski górski lěśojski cas", "MESZ": "Srjejźoeuropski lěśojski cas", "MEZ": "Srjejźoeuropski standardny cas", "MST": "Pódpołnocnoameriski górski standardny cas", "MYT": "Malajziski cas", "NZDT": "Nowoseelandski lěśojski cas", "NZST": "Nowoseelandski standardny cas", "OESZ": "Pódzajtšnoeuropski lěśojski cas", "OEZ": "Pódzajtšnoeuropski standardny cas", "PDT": "Pódpołnocnoameriski pacifiski lěśojski cas", "PST": "Pódpołnocnoameriski pacifiski standardny cas", "SAST": "Pódpołdnjowoafriski cas", "SGT": "Singapurski cas", "SRT": "Surinamski cas", "TMST": "Turkmeniski lěśojski cas", "TMT": "Turkmeniski standardny cas", "UYST": "Uruguayski lěśojski cas", "UYT": "Uruguayski standardny cas", "VET": "Venezuelski cas", "WARST": "Pódwjacornoargentinski lěśojski cas", "WART": "Pódwjacornoargentinski standardny cas", "WAST": "Pódwjacornoafriski lěśojski cas", "WAT": "Pódwjacornoafriski standardny cas", "WESZ": "Pódwjacornoeuropski lěśojski cas", "WEZ": "Pódwjacornoeuropski standardny cas", "WIB": "Pódwjacornoindoneski cas", "WIT": "Pódzajtšnoindoneski", "WITA": "Srjejźoindoneski cas", "∅∅∅": "Acorski lěśojski cas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_283d29635c231a3ca301081bab098305,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dua"] = func() locales.Translator {
		return &localen{

			locale:                 "dua",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "di", "ŋgɔn", "sɔŋ", "diɓ", "emi", "esɔ", "mad", "diŋ", "nyɛt", "may", "tin", "elá"},
			monthsNarrow:           []string{"", "d", "ŋ", "s", "d", "e", "e", "m", "d", "n", "m", "t", "e"},
			monthsWide:             []string{"", "dimɔ́di", "ŋgɔndɛ", "sɔŋɛ", "diɓáɓá", "emiasele", "esɔpɛsɔpɛ", "madiɓɛ́díɓɛ́", "diŋgindi", "nyɛtɛki", "mayésɛ́", "tiníní", "eláŋgɛ́"},
			daysAbbreviated:        []string{"ét", "mɔ́s", "kwa", "muk", "ŋgi", "ɗón", "esa"},
			daysNarrow:             []string{"e", "m", "k", "m", "ŋ", "ɗ", "e"},
			daysWide:               []string{"éti", "mɔ́sú", "kwasú", "mukɔ́sú", "ŋgisú", "ɗónɛsú", "esaɓasú"},
			periodsAbbreviated:     []string{"idiɓa", "ebyámu"},
			periodsWide:            []string{"idiɓa", "ebyámu"},
			erasAbbreviated:        []string{"ɓ.Ys", "mb.Ys"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ɓoso ɓwá yáɓe lá", "mbúsa kwédi a Yés"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dua_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "dua_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "di", "ŋgɔn", "sɔŋ", "diɓ", "emi", "esɔ", "mad", "diŋ", "nyɛt", "may", "tin", "elá"},
			monthsNarrow:           []string{"", "d", "ŋ", "s", "d", "e", "e", "m", "d", "n", "m", "t", "e"},
			monthsWide:             []string{"", "dimɔ́di", "ŋgɔndɛ", "sɔŋɛ", "diɓáɓá", "emiasele", "esɔpɛsɔpɛ", "madiɓɛ́díɓɛ́", "diŋgindi", "nyɛtɛki", "mayésɛ́", "tiníní", "eláŋgɛ́"},
			daysAbbreviated:        []string{"ét", "mɔ́s", "kwa", "muk", "ŋgi", "ɗón", "esa"},
			daysNarrow:             []string{"e", "m", "k", "m", "ŋ", "ɗ", "e"},
			daysWide:               []string{"éti", "mɔ́sú", "kwasú", "mukɔ́sú", "ŋgisú", "ɗónɛsú", "esaɓasú"},
			periodsAbbreviated:     []string{"idiɓa", "ebyámu"},
			periodsWide:            []string{"idiɓa", "ebyámu"},
			erasAbbreviated:        []string{"ɓ.Ys", "mb.Ys"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ɓoso ɓwá yáɓe lá", "mbúsa kwédi a Yés"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dyo"] = func() locales.Translator {
		return &localen{

			locale:                 "dyo",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Sa", "Fe", "Ma", "Ab", "Me", "Su", "Sú", "Ut", "Se", "Ok", "No", "De"},
			monthsNarrow:           []string{"", "S", "F", "M", "A", "M", "S", "S", "U", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Sanvie", "Fébirie", "Mars", "Aburil", "Mee", "Sueŋ", "Súuyee", "Ut", "Settembar", "Oktobar", "Novembar", "Disambar"},
			daysAbbreviated:        []string{"Dim", "Ten", "Tal", "Ala", "Ara", "Arj", "Sib"},
			daysNarrow:             []string{"D", "T", "T", "A", "A", "A", "S"},
			daysWide:               []string{"Dimas", "Teneŋ", "Talata", "Alarbay", "Aramisay", "Arjuma", "Sibiti"},
			erasAbbreviated:        []string{"ArY", "AtY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ariŋuu Yeesu", "Atooŋe Yeesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dyo_sn"] = func() locales.Translator {
		return &localen{

			locale:                 "dyo_SN",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Sa", "Fe", "Ma", "Ab", "Me", "Su", "Sú", "Ut", "Se", "Ok", "No", "De"},
			monthsNarrow:           []string{"", "S", "F", "M", "A", "M", "S", "S", "U", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Sanvie", "Fébirie", "Mars", "Aburil", "Mee", "Sueŋ", "Súuyee", "Ut", "Settembar", "Oktobar", "Novembar", "Disambar"},
			daysAbbreviated:        []string{"Dim", "Ten", "Tal", "Ala", "Ara", "Arj", "Sib"},
			daysNarrow:             []string{"D", "T", "T", "A", "A", "A", "S"},
			daysWide:               []string{"Dimas", "Teneŋ", "Talata", "Alarbay", "Aramisay", "Arjuma", "Sibiti"},
			erasAbbreviated:        []string{"ArY", "AtY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ariŋuu Yeesu", "Atooŋe Yeesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dz"] = func() locales.Translator {
		return &localen{

			locale:             "dz",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "གྲངས་མེད",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "Nu.", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KR₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "TH฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "༡", "༢", "༣", "༤", "༥", "༦", "༧", "༨", "༩", "༡༠", "༡༡", "12"},
			monthsNarrow:       []string{"", "༡", "༢", "༣", "4", "༥", "༦", "༧", "༨", "9", "༡༠", "༡༡", "༡༢"},
			monthsWide:         []string{"", "ཟླ་དངཔ་", "ཟླ་གཉིས་པ་", "ཟླ་གསུམ་པ་", "ཟླ་བཞི་པ་", "ཟླ་ལྔ་པ་", "ཟླ་དྲུག་པ", "ཟླ་བདུན་པ་", "ཟླ་བརྒྱད་པ་", "ཟླ་དགུ་པ་", "ཟླ་བཅུ་པ་", "ཟླ་བཅུ་གཅིག་པ་", "ཟླ་བཅུ་གཉིས་པ་"},
			daysAbbreviated:    []string{"ཟླ་", "མིར་", "ལྷག་", "ཕུར་", "སངས་", "སྤེན་", "ཉི་"},
			daysNarrow:         []string{"ཟླ", "མིར", "ལྷག", "ཕུར", "སངྶ", "སྤེན", "ཉི"},
			daysShort:          []string{"ཟླ་", "མིར་", "ལྷག་", "ཕུར་", "སངས་", "སྤེན་", "ཉི་"},
			daysWide:           []string{"གཟའ་ཟླ་བ་", "གཟའ་མིག་དམར་", "གཟའ་ལྷག་པ་", "གཟའ་ཕུར་བུ་", "གཟའ་པ་སངས་", "གཟའ་སྤེན་པ་", "གཟའ་ཉི་མ་"},
			periodsAbbreviated: []string{"སྔ་ཆ་", "ཕྱི་ཆ་"},
			periodsNarrow:      []string{"སྔ་ཆ་", "ཕྱི་ཆ་"},
			periodsWide:        []string{"སྔ་ཆ་", "ཕྱི་ཆ་"},
			erasAbbreviated:    []string{"BCE", "CE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "དབུས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "ACST": "དབུས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "ACWDT": "དབུས་ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "ACWST": "དབུས་ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "ADT": "ཨེཊ་ལེན་ཊིཀ་ཉིན་སྲུང་ཆུ་ཚོད", "AEDT": "ཤར་ཕྱོགས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "AEST": "ཤར་ཕྱོགས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "AKDT": "ཨ་ལསི་ཀ་ཉིན་སྲུང་ཆུ་ཚོད", "AKST": "ཨ་ལསི་ཀ་ཚད་ལྡན་ཆུ་ཚོད", "ARST": "ཨར་ཇེན་ཊི་ན་བྱཱར་དུས་ཆུ་ཚོད", "ART": "ཨར་ཇེན་ཊི་ན་ཚད་ལྡན་ཆུ་ཚོད", "AST": "ཨེཊ་ལེན་ཊིཀ་ཚད་ལྡན་ཆུ་ཚོད", "AWDT": "ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "AWST": "ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "BOT": "བྷོ་ལི་བི་ཡ་ཆུ་ཚོད", "BT": "འབྲུག་ཡུལ་ཆུ་ཚོད", "CAT": "དབུས་ཕྱོགས་ཨཕ་རི་ཀཱ་ཆུ་ཚོད", "CDT": "བྱང་ཨ་མི་རི་ཀ་དབུས་ཕྱོགས་ཉིན་སྲུང་ཆུ་ཚོད", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "ཅི་ལི་བྱཱར་དུས་ཆུ་ཚོད", "CLT": "ཅི་ལི་ཚད་ལྡན་ཆུ་ཚོད", "COST": "ཀོ་ལོམ་བྷི་ཡ་བྱཱར་དུས་ཆུ་ཚོད", "COT": "ཀོ་ལོམ་བྷི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "CST": "བྱང་ཨ་མི་རི་ཀ་དབུས་ཕྱོགས་ཚད་ལྡན་ཆུ་ཚོད", "ChST": "ChST", "EAT": "ཤར་ཕྱོགས་ཨཕ་རི་ཀཱ་ཆུ་ཚོད", "ECT": "ཨེ་ཀུ་ཌཽ་ཆུ་ཚོད", "EDT": "བྱང་ཨ་མི་རི་ཀ་ཤར་ཕྱོགས་ཉིན་སྲུང་ཆུ་ཚོད", "EST": "བྱང་ཨ་མི་རི་ཀ་ཤར་ཕྱོགས་ཚད་ལྡན་ཆུ་ཚོད", "GFT": "ཕིརེནཅ་གི་ཡ་ན་ཆུ་ཚོད", "GMT": "གིརིན་ཝིཆ་ལུ་ཡོད་པའི་ཆུ་ཚོད", "GST": "གཱལཕི་ཆུ་ཚོད", "GYT": "གུ་ཡ་ན་ཆུ་ཚོད", "HADT": "ཧ་ཝའི་-ཨེ་ལིའུ་ཤེན་ཉིན་སྲུང་ཆུ་ཚོད", "HAST": "ཧ་ཝའི་-ཨེ་ལིའུ་ཤེན་ཚད་ལྡན་ཆུ་ཚོད", "HAT": "ནིའུ་ཕའུནཌ་ལེནཌ་ཉིན་སྲུང་ཆུ་ཚོད", "HECU": "ཀིའུ་བྷ་ཉིན་སྲུང་ཆུ་ཚོད", "HEEG": "ཤར་ཕྱོགས་གིརིན་ལེནཌ་བྱཱར་དུས་ཆུ་ཚོད", "HENOMX": "HENOMX", "HEOG": "ནུབ་ཕྱོགས་གིརིན་ལེནཌ་བྱཱར་དུས་ཆུ་ཚོད", "HEPM": "པའི་རི་དང་མི་ཀི་ལཱོན་ཉིན་སྲུང་ཆུ་ཚོད", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "ཀིའུ་བྷ་ཚད་ལྡན་ཆུ་ཚོད", "HNEG": "ཤར་ཕྱོགས་གིརིན་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "HNNOMX": "HNNOMX", "HNOG": "ནུབ་ཕྱོགས་གིརིན་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "HNPM": "པའི་རི་དང་མི་ཀི་ལཱོན་ཚད་ལྡན་ཆུ་ཚོད", "HNPMX": "HNPMX", "HNT": "ནིའུ་ཕའུནཌ་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "IST": "རྒྱ་གར་ཆུ་ཚོད", "JDT": "ཇ་པཱན་ཉིན་སྲུང་ཆུ་ཚོད", "JST": "ཇ་པཱན་ཚད་ལྡན་ཆུ་ཚོད", "LHDT": "LHDT", "LHST": "LHST", "MDT": "བྱང་ཨ་མི་རི་ཀ་མའུ་ཊེན་ཉིན་སྲུང་ཆུ་ཚོད", "MESZ": "དབུས་ཕྱོགས་ཡུ་རོ་པེན་བྱཱར་དུས་ཆུ་ཚོད", "MEZ": "དབུས་ཕྱོགས་ཡུ་རོ་པེན་ཚད་ལྡན་ཆུ་ཚོད", "MST": "བྱང་ཨ་མི་རི་ཀ་མའུ་ཊེན་ཚད་ལྡན་ཆུ་ཚོད", "MYT": "MYT", "NZDT": "ནིའུ་ཛི་ལེནཌ་ཉིན་སྲུང་ཆུ་ཚོད", "NZST": "ནིའུ་ཛི་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "OESZ": "ཤར་ཕྱོགས་ཡུ་རོ་པེན་བྱཱར་དུས་ཆུ་ཚོད", "OEZ": "ཤར་ཕྱོགས་ཡུ་རོ་པེན་ཚད་ལྡན་ཆུ་ཚོད", "PDT": "བྱང་ཨ་མི་རི་ཀ་པེ་སི་ཕིག་ཉིན་སྲུང་ཆུ་ཚོད", "PST": "བྱང་ཨ་མི་རི་ཀ་པེ་སི་ཕིག་ཚད་ལྡན་ཆུ་ཚོད", "SAST": "ལྷོ་ཕྱོགས་ཨཕ་རི་ཀཱ་ཆུ་ཚོད", "SGT": "SGT", "SRT": "སུ་རི་ནཱམ་ཆུ་ཚོད", "TMST": "TMST", "TMT": "TMT", "UYST": "ཡུ་རུ་གུ་ཝཱའི་བྱཱར་དུས་ཆུ་ཚོད", "UYT": "ཡུ་རུ་གུ་ཝཱའི་ཚད་ལྡན་ཆུ་ཚོད", "VET": "བེ་ནི་ཛུ་ཝེ་ལ་ཆུ་ཚོད", "WARST": "ནུབ་ཕྱོགས་ཨར་ཇེན་ཊི་ན་བྱཱར་དུས་ཆུ་ཚོད", "WART": "ནུབ་ཕྱོགས་ཨར་ཇེན་ཊི་ན་ཚད་ལྡན་ཆུ་ཚོད", "WAST": "ནུབ་ཕྱོགས་ཨཕ་རི་ཀཱ་བྱཱར་དུས་ཆུ་ཚོད", "WAT": "ནུབ་ཕྱོགས་ཨཕ་རི་ཀཱ་ཚད་ལྡན་ཆུ་ཚོད", "WESZ": "ནུབ་ཕྱོགས་ཡུ་རོ་པེན་བྱཱར་དུས་ཆུ་ཚོད", "WEZ": "ནུབ་ཕྱོགས་ཡུ་རོ་པེན་ཚད་ལྡན་ཆུ་ཚོད", "WIB": "ནུབ་ཕྱོགས་ཨིན་ཌོ་ནེ་ཤི་ཡ་ཆུ་ཚོད", "WIT": "ཤར་ཕྱོགས་ཨིན་ཌོ་ནེ་ཤི་ཡ་ཆུ་ཚོད", "WITA": "དབུས་ཕྱོགས་ཨིན་ཌོ་ནེ་ཤི་ཡ་ཆུ་ཚོད", "∅∅∅": "བྲ་ཛི་ལི་ཡ་བྱཱར་དུས་ཆུ་ཚོད"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c1263930de470ad716d47d79ffa18067,
			fnFmtDateLong:         fnFmtDateLong_86c295425cafc585c7a23c6ee82c54e5,
			fnFmtDateMedium:       fnFmtDateMedium_cea69a8f4293bc29e07e966b34b3ea95,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_a9a2519152d6f243a597cc4bd6b266cc,
			fnFmtTimeLong:         fnFmtTimeLong_f3348bf1c6f276c7b95877da6ebb7ee0,
			fnFmtTimeMedium:       fnFmtTimeMedium_9e0f9c064fe3f07e66fe8fd5bfc4b612,
			fnFmtTimeShort:        fnFmtTimeShort_f47f2039d4428f5cb43433a9d8952827,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["dz_bt"] = func() locales.Translator {
		return &localen{

			locale:             "dz_BT",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "གྲངས་མེད",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "༡", "༢", "༣", "༤", "༥", "༦", "༧", "༨", "༩", "༡༠", "༡༡", "12"},
			monthsNarrow:       []string{"", "༡", "༢", "༣", "4", "༥", "༦", "༧", "༨", "9", "༡༠", "༡༡", "༡༢"},
			monthsWide:         []string{"", "ཟླ་དངཔ་", "ཟླ་གཉིས་པ་", "ཟླ་གསུམ་པ་", "ཟླ་བཞི་པ་", "ཟླ་ལྔ་པ་", "ཟླ་དྲུག་པ", "ཟླ་བདུན་པ་", "ཟླ་བརྒྱད་པ་", "ཟླ་དགུ་པ་", "ཟླ་བཅུ་པ་", "ཟླ་བཅུ་གཅིག་པ་", "ཟླ་བཅུ་གཉིས་པ་"},
			daysAbbreviated:    []string{"ཟླ་", "མིར་", "ལྷག་", "ཕུར་", "སངས་", "སྤེན་", "ཉི་"},
			daysNarrow:         []string{"ཟླ", "མིར", "ལྷག", "ཕུར", "སངྶ", "སྤེན", "ཉི"},
			daysShort:          []string{"ཟླ་", "མིར་", "ལྷག་", "ཕུར་", "སངས་", "སྤེན་", "ཉི་"},
			daysWide:           []string{"གཟའ་ཟླ་བ་", "གཟའ་མིག་དམར་", "གཟའ་ལྷག་པ་", "གཟའ་ཕུར་བུ་", "གཟའ་པ་སངས་", "གཟའ་སྤེན་པ་", "གཟའ་ཉི་མ་"},
			periodsAbbreviated: []string{"སྔ་ཆ་", "ཕྱི་ཆ་"},
			periodsNarrow:      []string{"སྔ་ཆ་", "ཕྱི་ཆ་"},
			periodsWide:        []string{"སྔ་ཆ་", "ཕྱི་ཆ་"},
			erasAbbreviated:    []string{"BCE", "CE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "དབུས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "ACST": "དབུས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "ACWDT": "དབུས་ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "ACWST": "དབུས་ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "ADT": "ཨེཊ་ལེན་ཊིཀ་ཉིན་སྲུང་ཆུ་ཚོད", "AEDT": "ཤར་ཕྱོགས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "AEST": "ཤར་ཕྱོགས་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "AKDT": "ཨ་ལསི་ཀ་ཉིན་སྲུང་ཆུ་ཚོད", "AKST": "ཨ་ལསི་ཀ་ཚད་ལྡན་ཆུ་ཚོད", "ARST": "ཨར་ཇེན་ཊི་ན་བྱཱར་དུས་ཆུ་ཚོད", "ART": "ཨར་ཇེན་ཊི་ན་ཚད་ལྡན་ཆུ་ཚོད", "AST": "ཨེཊ་ལེན་ཊིཀ་ཚད་ལྡན་ཆུ་ཚོད", "AWDT": "ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཉིན་སྲུང་ཆུ་ཚོད", "AWST": "ནུབ་ཕྱོགས་ཨཱོས་ཊྲེལ་ལི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "BOT": "བྷོ་ལི་བི་ཡ་ཆུ་ཚོད", "BT": "འབྲུག་ཡུལ་ཆུ་ཚོད", "CAT": "དབུས་ཕྱོགས་ཨཕ་རི་ཀཱ་ཆུ་ཚོད", "CDT": "བྱང་ཨ་མི་རི་ཀ་དབུས་ཕྱོགས་ཉིན་སྲུང་ཆུ་ཚོད", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "ཅི་ལི་བྱཱར་དུས་ཆུ་ཚོད", "CLT": "ཅི་ལི་ཚད་ལྡན་ཆུ་ཚོད", "COST": "ཀོ་ལོམ་བྷི་ཡ་བྱཱར་དུས་ཆུ་ཚོད", "COT": "ཀོ་ལོམ་བྷི་ཡ་ཚད་ལྡན་ཆུ་ཚོད", "CST": "བྱང་ཨ་མི་རི་ཀ་དབུས་ཕྱོགས་ཚད་ལྡན་ཆུ་ཚོད", "ChST": "ChST", "EAT": "ཤར་ཕྱོགས་ཨཕ་རི་ཀཱ་ཆུ་ཚོད", "ECT": "ཨེ་ཀུ་ཌཽ་ཆུ་ཚོད", "EDT": "བྱང་ཨ་མི་རི་ཀ་ཤར་ཕྱོགས་ཉིན་སྲུང་ཆུ་ཚོད", "EST": "བྱང་ཨ་མི་རི་ཀ་ཤར་ཕྱོགས་ཚད་ལྡན་ཆུ་ཚོད", "GFT": "ཕིརེནཅ་གི་ཡ་ན་ཆུ་ཚོད", "GMT": "གིརིན་ཝིཆ་ལུ་ཡོད་པའི་ཆུ་ཚོད", "GST": "གཱལཕི་ཆུ་ཚོད", "GYT": "གུ་ཡ་ན་ཆུ་ཚོད", "HADT": "ཧ་ཝའི་-ཨེ་ལིའུ་ཤེན་ཉིན་སྲུང་ཆུ་ཚོད", "HAST": "ཧ་ཝའི་-ཨེ་ལིའུ་ཤེན་ཚད་ལྡན་ཆུ་ཚོད", "HAT": "ནིའུ་ཕའུནཌ་ལེནཌ་ཉིན་སྲུང་ཆུ་ཚོད", "HECU": "ཀིའུ་བྷ་ཉིན་སྲུང་ཆུ་ཚོད", "HEEG": "ཤར་ཕྱོགས་གིརིན་ལེནཌ་བྱཱར་དུས་ཆུ་ཚོད", "HENOMX": "HENOMX", "HEOG": "ནུབ་ཕྱོགས་གིརིན་ལེནཌ་བྱཱར་དུས་ཆུ་ཚོད", "HEPM": "པའི་རི་དང་མི་ཀི་ལཱོན་ཉིན་སྲུང་ཆུ་ཚོད", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "ཀིའུ་བྷ་ཚད་ལྡན་ཆུ་ཚོད", "HNEG": "ཤར་ཕྱོགས་གིརིན་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "HNNOMX": "HNNOMX", "HNOG": "ནུབ་ཕྱོགས་གིརིན་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "HNPM": "པའི་རི་དང་མི་ཀི་ལཱོན་ཚད་ལྡན་ཆུ་ཚོད", "HNPMX": "HNPMX", "HNT": "ནིའུ་ཕའུནཌ་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "IST": "རྒྱ་གར་ཆུ་ཚོད", "JDT": "ཇ་པཱན་ཉིན་སྲུང་ཆུ་ཚོད", "JST": "ཇ་པཱན་ཚད་ལྡན་ཆུ་ཚོད", "LHDT": "LHDT", "LHST": "LHST", "MDT": "བྱང་ཨ་མི་རི་ཀ་མའུ་ཊེན་ཉིན་སྲུང་ཆུ་ཚོད", "MESZ": "དབུས་ཕྱོགས་ཡུ་རོ་པེན་བྱཱར་དུས་ཆུ་ཚོད", "MEZ": "དབུས་ཕྱོགས་ཡུ་རོ་པེན་ཚད་ལྡན་ཆུ་ཚོད", "MST": "བྱང་ཨ་མི་རི་ཀ་མའུ་ཊེན་ཚད་ལྡན་ཆུ་ཚོད", "MYT": "MYT", "NZDT": "ནིའུ་ཛི་ལེནཌ་ཉིན་སྲུང་ཆུ་ཚོད", "NZST": "ནིའུ་ཛི་ལེནཌ་ཚད་ལྡན་ཆུ་ཚོད", "OESZ": "ཤར་ཕྱོགས་ཡུ་རོ་པེན་བྱཱར་དུས་ཆུ་ཚོད", "OEZ": "ཤར་ཕྱོགས་ཡུ་རོ་པེན་ཚད་ལྡན་ཆུ་ཚོད", "PDT": "བྱང་ཨ་མི་རི་ཀ་པེ་སི་ཕིག་ཉིན་སྲུང་ཆུ་ཚོད", "PST": "བྱང་ཨ་མི་རི་ཀ་པེ་སི་ཕིག་ཚད་ལྡན་ཆུ་ཚོད", "SAST": "ལྷོ་ཕྱོགས་ཨཕ་རི་ཀཱ་ཆུ་ཚོད", "SGT": "SGT", "SRT": "སུ་རི་ནཱམ་ཆུ་ཚོད", "TMST": "TMST", "TMT": "TMT", "UYST": "ཡུ་རུ་གུ་ཝཱའི་བྱཱར་དུས་ཆུ་ཚོད", "UYT": "ཡུ་རུ་གུ་ཝཱའི་ཚད་ལྡན་ཆུ་ཚོད", "VET": "བེ་ནི་ཛུ་ཝེ་ལ་ཆུ་ཚོད", "WARST": "ནུབ་ཕྱོགས་ཨར་ཇེན་ཊི་ན་བྱཱར་དུས་ཆུ་ཚོད", "WART": "ནུབ་ཕྱོགས་ཨར་ཇེན་ཊི་ན་ཚད་ལྡན་ཆུ་ཚོད", "WAST": "ནུབ་ཕྱོགས་ཨཕ་རི་ཀཱ་བྱཱར་དུས་ཆུ་ཚོད", "WAT": "ནུབ་ཕྱོགས་ཨཕ་རི་ཀཱ་ཚད་ལྡན་ཆུ་ཚོད", "WESZ": "ནུབ་ཕྱོགས་ཡུ་རོ་པེན་བྱཱར་དུས་ཆུ་ཚོད", "WEZ": "ནུབ་ཕྱོགས་ཡུ་རོ་པེན་ཚད་ལྡན་ཆུ་ཚོད", "WIB": "ནུབ་ཕྱོགས་ཨིན་ཌོ་ནེ་ཤི་ཡ་ཆུ་ཚོད", "WIT": "ཤར་ཕྱོགས་ཨིན་ཌོ་ནེ་ཤི་ཡ་ཆུ་ཚོད", "WITA": "དབུས་ཕྱོགས་ཨིན་ཌོ་ནེ་ཤི་ཡ་ཆུ་ཚོད", "∅∅∅": "ཨེ་ཛོརས་བྱཱར་དུས་ཆུ་ཚོད"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c1263930de470ad716d47d79ffa18067,
			fnFmtDateLong:         fnFmtDateLong_86c295425cafc585c7a23c6ee82c54e5,
			fnFmtDateMedium:       fnFmtDateMedium_cea69a8f4293bc29e07e966b34b3ea95,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_a9a2519152d6f243a597cc4bd6b266cc,
			fnFmtTimeLong:         fnFmtTimeLong_f3348bf1c6f276c7b95877da6ebb7ee0,
			fnFmtTimeMedium:       fnFmtTimeMedium_9e0f9c064fe3f07e66fe8fd5bfc4b612,
			fnFmtTimeShort:        fnFmtTimeShort_f47f2039d4428f5cb43433a9d8952827,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ebu"] = func() locales.Translator {
		return &localen{

			locale:                 "ebu",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Mbe", "Kai", "Kat", "Kan", "Gat", "Gan", "Mug", "Knn", "Ken", "Iku", "Imw", "Igi"},
			monthsNarrow:           []string{"", "M", "K", "K", "K", "G", "G", "M", "K", "K", "I", "I", "I"},
			monthsWide:             []string{"", "Mweri wa mbere", "Mweri wa kaĩri", "Mweri wa kathatũ", "Mweri wa kana", "Mweri wa gatano", "Mweri wa gatantatũ", "Mweri wa mũgwanja", "Mweri wa kanana", "Mweri wa kenda", "Mweri wa ikũmi", "Mweri wa ikũmi na ũmwe", "Mweri wa ikũmi na Kaĩrĩ"},
			daysAbbreviated:        []string{"Kma", "Tat", "Ine", "Tan", "Arm", "Maa", "NMM"},
			daysNarrow:             []string{"K", "N", "N", "N", "A", "M", "N"},
			daysWide:               []string{"Kiumia", "Njumatatu", "Njumaine", "Njumatano", "Aramithi", "Njumaa", "NJumamothii"},
			periodsAbbreviated:     []string{"KI", "UT"},
			periodsWide:            []string{"KI", "UT"},
			erasAbbreviated:        []string{"MK", "TK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbere ya Kristo", "Thutha wa Kristo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ebu_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "ebu_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Mbe", "Kai", "Kat", "Kan", "Gat", "Gan", "Mug", "Knn", "Ken", "Iku", "Imw", "Igi"},
			monthsNarrow:           []string{"", "M", "K", "K", "K", "G", "G", "M", "K", "K", "I", "I", "I"},
			monthsWide:             []string{"", "Mweri wa mbere", "Mweri wa kaĩri", "Mweri wa kathatũ", "Mweri wa kana", "Mweri wa gatano", "Mweri wa gatantatũ", "Mweri wa mũgwanja", "Mweri wa kanana", "Mweri wa kenda", "Mweri wa ikũmi", "Mweri wa ikũmi na ũmwe", "Mweri wa ikũmi na Kaĩrĩ"},
			daysAbbreviated:        []string{"Kma", "Tat", "Ine", "Tan", "Arm", "Maa", "NMM"},
			daysNarrow:             []string{"K", "N", "N", "N", "A", "M", "N"},
			daysWide:               []string{"Kiumia", "Njumatatu", "Njumaine", "Njumatano", "Aramithi", "Njumaa", "NJumamothii"},
			periodsAbbreviated:     []string{"KI", "UT"},
			periodsWide:            []string{"KI", "UT"},
			erasAbbreviated:        []string{"MK", "TK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbere ya Kristo", "Thutha wa Kristo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ee"] = func() locales.Translator {
		return &localen{

			locale:                 "ee",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GH₵", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "dzv", "dzd", "ted", "afɔ", "dam", "mas", "sia", "dea", "any", "kel", "ade", "dzm"},
			monthsNarrow:           []string{"", "d", "d", "t", "a", "d", "m", "s", "d", "a", "k", "a", "d"},
			monthsWide:             []string{"", "dzove", "dzodze", "tedoxe", "afɔfĩe", "dama", "masa", "siamlɔm", "deasiamime", "anyɔnyɔ", "kele", "adeɛmekpɔxe", "dzome"},
			daysAbbreviated:        []string{"kɔs", "dzo", "bla", "kuɖ", "yaw", "fiɖ", "mem"},
			daysNarrow:             []string{"k", "d", "b", "k", "y", "f", "m"},
			daysShort:              []string{"kɔs", "dzo", "bla", "kuɖ", "yaw", "fiɖ", "mem"},
			daysWide:               []string{"kɔsiɖa", "dzoɖa", "blaɖa", "kuɖa", "yawoɖa", "fiɖa", "memleɖa"},
			periodsAbbreviated:     []string{"ŋdi", "ɣetrɔ"},
			periodsNarrow:          []string{"ŋ", "ɣ"},
			periodsWide:            []string{"ŋdi", "ɣetrɔ"},
			erasAbbreviated:        []string{"HYV", "Yŋ"},
			erasNarrow:             []string{"hY", "Yŋ"},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central dzomeli gaƒoƒo me", "ACST": "Australian Central nutome gaƒoƒo me", "ACWDT": "Australian Central Western kele gaƒoƒo me", "ACWST": "Australian Central Western nutome gaƒoƒo me", "ADT": "Atlantic kele gaƒoƒome", "AEDT": "Australian Eastern kele gaƒoƒo me", "AEST": "Australian Eastern nutome gaƒoƒo me", "AKDT": "Alaska kele gaƒoƒo me", "AKST": "Alaska nutome gaƒoƒo me", "ARST": "Argentina dzomeŋɔli gaƒoƒo me", "ART": "Argentina nutome gaƒoƒo me", "AST": "Atlantic nutome gaƒoƒome", "AWDT": "Australian Western kele gaƒoƒo me", "AWST": "Australian Western nutome gaƒoƒo me", "BOT": "Bolivia gaƒoƒo me", "BT": "Bhutan gaƒoƒo me", "CAT": "Central Africa gaƒoƒo me", "CDT": "Titina America kele gaƒoƒo me", "CHADT": "Chatham kele gaƒoƒo me", "CHAST": "Chatham nutome gaƒoƒo me", "CLST": "Chile dzomeŋɔli gaƒoƒo me", "CLT": "Chile nutome gaƒoƒo me", "COST": "Colombia dzomeŋɔli gaƒoƒo me", "COT": "Colombia nutome gaƒoƒo me", "CST": "Titina America nutome gaƒoƒo me", "ChST": "Chamorro gaƒoƒo me", "EAT": "East Africa gaƒoƒo me", "ECT": "Ecuador gaƒoƒo me", "EDT": "Eastern America kele gaƒoƒo me", "EST": "Eastern America nutome gaƒoƒo me", "GFT": "French Guiana gaƒoƒo me", "GMT": "Greenwich gaƒoƒo me", "GST": "Gulf nutome gaƒoƒo me", "GYT": "Guyana gaƒoƒo me", "HADT": "Hawaii-Aleutia kele gaƒoƒo me", "HAST": "Hawaii-Aleutia nutome gaƒoƒo me", "HAT": "Newfoundland kele gaƒoƒome", "HECU": "Cuba kele gaƒoƒome", "HEEG": "East Greenland dzomeŋɔli gaƒoƒo me", "HENOMX": "Northwest Mexico kele gaƒoƒo me", "HEOG": "West Greenland kele gaƒoƒo me", "HEPM": "St. Pierre & Miquelon kele gaƒoƒome", "HEPMX": "Mexican Pacific kele gaƒoƒome", "HKST": "Hong Kong dzomeŋɔli gaƒoƒo me", "HKT": "Hong Kong nutome gaƒoƒo me", "HNCU": "Cuba nutome gaƒoƒome", "HNEG": "East Greenland nutome gaƒoƒo me", "HNNOMX": "Northwest Mexico nutome gaƒoƒo me", "HNOG": "West Greenland nutome gaƒoƒo me", "HNPM": "St. Pierre & Miquelon nutome gaƒoƒome", "HNPMX": "Mexican Pacific nutome gaƒoƒo me", "HNT": "Newfoundland nutome gaƒoƒome", "IST": "India gaƒoƒo me", "JDT": "Japan dzomeŋɔli gaƒoƒo me", "JST": "Japan nutome gaƒoƒo me", "LHDT": "Lord Howe kele gaƒoƒo me", "LHST": "Lord Howe nutome gaƒoƒo me", "MDT": "Mountain kele gaƒoƒo me", "MESZ": "Central Europe dzomeŋɔli gaƒoƒo me", "MEZ": "Central Europe nutome gaƒoƒo me", "MST": "Mountain nutome gaƒoƒo me", "MYT": "Malaysia gaƒoƒo me", "NZDT": "New Zealand kele gaƒoƒo me", "NZST": "New Zealand nutome gaƒoƒo me", "OESZ": "Ɣedzeƒe Europe ŋkekeme gaƒoƒome", "OEZ": "Ɣedzeƒe Europe gaƒoƒoɖoanyime", "PDT": "Pacific kele gaƒoƒo me", "PST": "Pacific nutome gaƒoƒo me", "SAST": "South Africa nutome gaƒoƒo me", "SGT": "Singapore nutome gaƒoƒo me", "SRT": "Suriname gaƒoƒome", "TMST": "Turkmenistan dzomeŋɔli gaƒoƒo me", "TMT": "Turkmenistan nutome gaƒoƒo me", "UYST": "Uruguay dzomeŋɔli gaƒoƒo me", "UYT": "Uruguay nutome gaƒoƒo me", "VET": "Venezuela gaƒoƒo me", "WARST": "Ɣetoɖoƒe Argentina dzomeŋɔli gaƒoƒo me", "WART": "Ɣetoɖoƒe Argentina nutome gaƒoƒo me", "WAST": "West Africa dzomeŋɔli gaƒoƒo me", "WAT": "West Africa nutome gaƒoƒo me", "WESZ": "Western Europe dzomeŋɔli gaƒoƒo me", "WEZ": "Western Europe nutome gaƒoƒo me", "WIB": "Western Indonesia gaƒoƒo me", "WIT": "Eastern Indonesia gaƒoƒo me", "WITA": "Central Indonesia gaƒoƒo me", "∅∅∅": "Peru dzomeŋɔli gaƒoƒome"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5728fd357bcbdd2617427e16f6e55d2,
			fnFmtDateLong:         fnFmtDateLong_e44be20ec70caf75cec942e788bd6c7e,
			fnFmtDateMedium:       fnFmtDateMedium_ab00ad983b60af2ef2a78b1bbda56e5b,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_93b0a8ad8fd85d3e1c5363d7419c3c1f,
			fnFmtTimeLong:         fnFmtTimeLong_eac3fe1ec156df0f86227350e36ea607,
			fnFmtTimeMedium:       fnFmtTimeMedium_aa77d70e448fa95de06962cfa9baf918,
			fnFmtTimeShort:        fnFmtTimeShort_8eb7cfdc1e992030ede9a1abfea545be,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ee_gh"] = func() locales.Translator {
		return &localen{

			locale:                 "ee_GH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "dzv", "dzd", "ted", "afɔ", "dam", "mas", "sia", "dea", "any", "kel", "ade", "dzm"},
			monthsNarrow:           []string{"", "d", "d", "t", "a", "d", "m", "s", "d", "a", "k", "a", "d"},
			monthsWide:             []string{"", "dzove", "dzodze", "tedoxe", "afɔfĩe", "dama", "masa", "siamlɔm", "deasiamime", "anyɔnyɔ", "kele", "adeɛmekpɔxe", "dzome"},
			daysAbbreviated:        []string{"kɔs", "dzo", "bla", "kuɖ", "yaw", "fiɖ", "mem"},
			daysNarrow:             []string{"k", "d", "b", "k", "y", "f", "m"},
			daysShort:              []string{"kɔs", "dzo", "bla", "kuɖ", "yaw", "fiɖ", "mem"},
			daysWide:               []string{"kɔsiɖa", "dzoɖa", "blaɖa", "kuɖa", "yawoɖa", "fiɖa", "memleɖa"},
			periodsAbbreviated:     []string{"ŋdi", "ɣetrɔ"},
			periodsNarrow:          []string{"ŋ", "ɣ"},
			periodsWide:            []string{"ŋdi", "ɣetrɔ"},
			erasAbbreviated:        []string{"HYV", "Yŋ"},
			erasNarrow:             []string{"hY", "Yŋ"},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central dzomeli gaƒoƒo me", "ACST": "Australian Central nutome gaƒoƒo me", "ACWDT": "Australian Central Western kele gaƒoƒo me", "ACWST": "Australian Central Western nutome gaƒoƒo me", "ADT": "Atlantic kele gaƒoƒome", "AEDT": "Australian Eastern kele gaƒoƒo me", "AEST": "Australian Eastern nutome gaƒoƒo me", "AKDT": "Alaska kele gaƒoƒo me", "AKST": "Alaska nutome gaƒoƒo me", "ARST": "Argentina dzomeŋɔli gaƒoƒo me", "ART": "Argentina nutome gaƒoƒo me", "AST": "Atlantic nutome gaƒoƒome", "AWDT": "Australian Western kele gaƒoƒo me", "AWST": "Australian Western nutome gaƒoƒo me", "BOT": "Bolivia gaƒoƒo me", "BT": "Bhutan gaƒoƒo me", "CAT": "Central Africa gaƒoƒo me", "CDT": "Titina America kele gaƒoƒo me", "CHADT": "Chatham kele gaƒoƒo me", "CHAST": "Chatham nutome gaƒoƒo me", "CLST": "Chile dzomeŋɔli gaƒoƒo me", "CLT": "Chile nutome gaƒoƒo me", "COST": "Colombia dzomeŋɔli gaƒoƒo me", "COT": "Colombia nutome gaƒoƒo me", "CST": "Titina America nutome gaƒoƒo me", "ChST": "Chamorro gaƒoƒo me", "EAT": "East Africa gaƒoƒo me", "ECT": "Ecuador gaƒoƒo me", "EDT": "Eastern America kele gaƒoƒo me", "EST": "Eastern America nutome gaƒoƒo me", "GFT": "French Guiana gaƒoƒo me", "GMT": "Greenwich gaƒoƒo me", "GST": "Gulf nutome gaƒoƒo me", "GYT": "Guyana gaƒoƒo me", "HADT": "Hawaii-Aleutia kele gaƒoƒo me", "HAST": "Hawaii-Aleutia nutome gaƒoƒo me", "HAT": "Newfoundland kele gaƒoƒome", "HECU": "Cuba kele gaƒoƒome", "HEEG": "East Greenland dzomeŋɔli gaƒoƒo me", "HENOMX": "Northwest Mexico kele gaƒoƒo me", "HEOG": "West Greenland kele gaƒoƒo me", "HEPM": "St. Pierre & Miquelon kele gaƒoƒome", "HEPMX": "Mexican Pacific kele gaƒoƒome", "HKST": "Hong Kong dzomeŋɔli gaƒoƒo me", "HKT": "Hong Kong nutome gaƒoƒo me", "HNCU": "Cuba nutome gaƒoƒome", "HNEG": "East Greenland nutome gaƒoƒo me", "HNNOMX": "Northwest Mexico nutome gaƒoƒo me", "HNOG": "West Greenland nutome gaƒoƒo me", "HNPM": "St. Pierre & Miquelon nutome gaƒoƒome", "HNPMX": "Mexican Pacific nutome gaƒoƒo me", "HNT": "Newfoundland nutome gaƒoƒome", "IST": "India gaƒoƒo me", "JDT": "Japan dzomeŋɔli gaƒoƒo me", "JST": "Japan nutome gaƒoƒo me", "LHDT": "Lord Howe kele gaƒoƒo me", "LHST": "Lord Howe nutome gaƒoƒo me", "MDT": "Mountain kele gaƒoƒo me", "MESZ": "Central Europe dzomeŋɔli gaƒoƒo me", "MEZ": "Central Europe nutome gaƒoƒo me", "MST": "Mountain nutome gaƒoƒo me", "MYT": "Malaysia gaƒoƒo me", "NZDT": "New Zealand kele gaƒoƒo me", "NZST": "New Zealand nutome gaƒoƒo me", "OESZ": "Ɣedzeƒe Europe ŋkekeme gaƒoƒome", "OEZ": "Ɣedzeƒe Europe gaƒoƒoɖoanyime", "PDT": "Pacific kele gaƒoƒo me", "PST": "Pacific nutome gaƒoƒo me", "SAST": "South Africa nutome gaƒoƒo me", "SGT": "Singapore nutome gaƒoƒo me", "SRT": "Suriname gaƒoƒome", "TMST": "Turkmenistan dzomeŋɔli gaƒoƒo me", "TMT": "Turkmenistan nutome gaƒoƒo me", "UYST": "Uruguay dzomeŋɔli gaƒoƒo me", "UYT": "Uruguay nutome gaƒoƒo me", "VET": "Venezuela gaƒoƒo me", "WARST": "Ɣetoɖoƒe Argentina dzomeŋɔli gaƒoƒo me", "WART": "Ɣetoɖoƒe Argentina nutome gaƒoƒo me", "WAST": "West Africa dzomeŋɔli gaƒoƒo me", "WAT": "West Africa nutome gaƒoƒo me", "WESZ": "Western Europe dzomeŋɔli gaƒoƒo me", "WEZ": "Western Europe nutome gaƒoƒo me", "WIB": "Western Indonesia gaƒoƒo me", "WIT": "Eastern Indonesia gaƒoƒo me", "WITA": "Central Indonesia gaƒoƒo me", "∅∅∅": "Azores dzomeŋɔli gaƒoƒo me"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5728fd357bcbdd2617427e16f6e55d2,
			fnFmtDateLong:         fnFmtDateLong_e44be20ec70caf75cec942e788bd6c7e,
			fnFmtDateMedium:       fnFmtDateMedium_ab00ad983b60af2ef2a78b1bbda56e5b,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_93b0a8ad8fd85d3e1c5363d7419c3c1f,
			fnFmtTimeLong:         fnFmtTimeLong_eac3fe1ec156df0f86227350e36ea607,
			fnFmtTimeMedium:       fnFmtTimeMedium_aa77d70e448fa95de06962cfa9baf918,
			fnFmtTimeShort:        fnFmtTimeShort_8eb7cfdc1e992030ede9a1abfea545be,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ee_tg"] = func() locales.Translator {
		return &localen{

			locale:                 "ee_TG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "dzv", "dzd", "ted", "afɔ", "dam", "mas", "sia", "dea", "any", "kel", "ade", "dzm"},
			monthsNarrow:           []string{"", "d", "d", "t", "a", "d", "m", "s", "d", "a", "k", "a", "d"},
			monthsWide:             []string{"", "dzove", "dzodze", "tedoxe", "afɔfĩe", "dama", "masa", "siamlɔm", "deasiamime", "anyɔnyɔ", "kele", "adeɛmekpɔxe", "dzome"},
			daysAbbreviated:        []string{"kɔs", "dzo", "bla", "kuɖ", "yaw", "fiɖ", "mem"},
			daysNarrow:             []string{"k", "d", "b", "k", "y", "f", "m"},
			daysShort:              []string{"kɔs", "dzo", "bla", "kuɖ", "yaw", "fiɖ", "mem"},
			daysWide:               []string{"kɔsiɖa", "dzoɖa", "blaɖa", "kuɖa", "yawoɖa", "fiɖa", "memleɖa"},
			periodsAbbreviated:     []string{"ŋdi", "ɣetrɔ"},
			periodsNarrow:          []string{"ŋ", "ɣ"},
			periodsWide:            []string{"ŋdi", "ɣetrɔ"},
			erasAbbreviated:        []string{"HYV", "Yŋ"},
			erasNarrow:             []string{"hY", "Yŋ"},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central dzomeli gaƒoƒo me", "ACST": "Australian Central nutome gaƒoƒo me", "ACWDT": "Australian Central Western kele gaƒoƒo me", "ACWST": "Australian Central Western nutome gaƒoƒo me", "ADT": "Atlantic kele gaƒoƒome", "AEDT": "Australian Eastern kele gaƒoƒo me", "AEST": "Australian Eastern nutome gaƒoƒo me", "AKDT": "Alaska kele gaƒoƒo me", "AKST": "Alaska nutome gaƒoƒo me", "ARST": "Argentina dzomeŋɔli gaƒoƒo me", "ART": "Argentina nutome gaƒoƒo me", "AST": "Atlantic nutome gaƒoƒome", "AWDT": "Australian Western kele gaƒoƒo me", "AWST": "Australian Western nutome gaƒoƒo me", "BOT": "Bolivia gaƒoƒo me", "BT": "Bhutan gaƒoƒo me", "CAT": "Central Africa gaƒoƒo me", "CDT": "Titina America kele gaƒoƒo me", "CHADT": "Chatham kele gaƒoƒo me", "CHAST": "Chatham nutome gaƒoƒo me", "CLST": "Chile dzomeŋɔli gaƒoƒo me", "CLT": "Chile nutome gaƒoƒo me", "COST": "Colombia dzomeŋɔli gaƒoƒo me", "COT": "Colombia nutome gaƒoƒo me", "CST": "Titina America nutome gaƒoƒo me", "ChST": "Chamorro gaƒoƒo me", "EAT": "East Africa gaƒoƒo me", "ECT": "Ecuador gaƒoƒo me", "EDT": "Eastern America kele gaƒoƒo me", "EST": "Eastern America nutome gaƒoƒo me", "GFT": "French Guiana gaƒoƒo me", "GMT": "Greenwich gaƒoƒo me", "GST": "Gulf nutome gaƒoƒo me", "GYT": "Guyana gaƒoƒo me", "HADT": "Hawaii-Aleutia kele gaƒoƒo me", "HAST": "Hawaii-Aleutia nutome gaƒoƒo me", "HAT": "Newfoundland kele gaƒoƒome", "HECU": "Cuba kele gaƒoƒome", "HEEG": "East Greenland dzomeŋɔli gaƒoƒo me", "HENOMX": "Northwest Mexico kele gaƒoƒo me", "HEOG": "West Greenland kele gaƒoƒo me", "HEPM": "St. Pierre & Miquelon kele gaƒoƒome", "HEPMX": "Mexican Pacific kele gaƒoƒome", "HKST": "Hong Kong dzomeŋɔli gaƒoƒo me", "HKT": "Hong Kong nutome gaƒoƒo me", "HNCU": "Cuba nutome gaƒoƒome", "HNEG": "East Greenland nutome gaƒoƒo me", "HNNOMX": "Northwest Mexico nutome gaƒoƒo me", "HNOG": "West Greenland nutome gaƒoƒo me", "HNPM": "St. Pierre & Miquelon nutome gaƒoƒome", "HNPMX": "Mexican Pacific nutome gaƒoƒo me", "HNT": "Newfoundland nutome gaƒoƒome", "IST": "India gaƒoƒo me", "JDT": "Japan dzomeŋɔli gaƒoƒo me", "JST": "Japan nutome gaƒoƒo me", "LHDT": "Lord Howe kele gaƒoƒo me", "LHST": "Lord Howe nutome gaƒoƒo me", "MDT": "Mountain kele gaƒoƒo me", "MESZ": "Central Europe dzomeŋɔli gaƒoƒo me", "MEZ": "Central Europe nutome gaƒoƒo me", "MST": "Mountain nutome gaƒoƒo me", "MYT": "Malaysia gaƒoƒo me", "NZDT": "New Zealand kele gaƒoƒo me", "NZST": "New Zealand nutome gaƒoƒo me", "OESZ": "Ɣedzeƒe Europe ŋkekeme gaƒoƒome", "OEZ": "Ɣedzeƒe Europe gaƒoƒoɖoanyime", "PDT": "Pacific kele gaƒoƒo me", "PST": "Pacific nutome gaƒoƒo me", "SAST": "South Africa nutome gaƒoƒo me", "SGT": "Singapore nutome gaƒoƒo me", "SRT": "Suriname gaƒoƒome", "TMST": "Turkmenistan dzomeŋɔli gaƒoƒo me", "TMT": "Turkmenistan nutome gaƒoƒo me", "UYST": "Uruguay dzomeŋɔli gaƒoƒo me", "UYT": "Uruguay nutome gaƒoƒo me", "VET": "Venezuela gaƒoƒo me", "WARST": "Ɣetoɖoƒe Argentina dzomeŋɔli gaƒoƒo me", "WART": "Ɣetoɖoƒe Argentina nutome gaƒoƒo me", "WAST": "West Africa dzomeŋɔli gaƒoƒo me", "WAT": "West Africa nutome gaƒoƒo me", "WESZ": "Western Europe dzomeŋɔli gaƒoƒo me", "WEZ": "Western Europe nutome gaƒoƒo me", "WIB": "Western Indonesia gaƒoƒo me", "WIT": "Eastern Indonesia gaƒoƒo me", "WITA": "Central Indonesia gaƒoƒo me", "∅∅∅": "Azores dzomeŋɔli gaƒoƒo me"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5728fd357bcbdd2617427e16f6e55d2,
			fnFmtDateLong:         fnFmtDateLong_e44be20ec70caf75cec942e788bd6c7e,
			fnFmtDateMedium:       fnFmtDateMedium_ab00ad983b60af2ef2a78b1bbda56e5b,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["el"] = func() locales.Translator {
		return &localen{

			locale:                 "el",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "Δρχ", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"},
			monthsNarrow:           []string{"", "Ι", "Φ", "Μ", "Α", "Μ", "Ι", "Ι", "Α", "Σ", "Ο", "Ν", "Δ"},
			monthsWide:             []string{"", "Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"},
			daysAbbreviated:        []string{"Κυρ", "Δευ", "Τρί", "Τετ", "Πέμ", "Παρ", "Σάβ"},
			daysNarrow:             []string{"Κ", "Δ", "Τ", "Τ", "Π", "Π", "Σ"},
			daysShort:              []string{"Κυ", "Δε", "Τρ", "Τε", "Πέ", "Πα", "Σά"},
			daysWide:               []string{"Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"},
			periodsAbbreviated:     []string{"π.μ.", "μ.μ."},
			periodsNarrow:          []string{"πμ", "μμ"},
			periodsWide:            []string{"π.μ.", "μ.μ."},
			erasAbbreviated:        []string{"π.Χ.", "μ.Χ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"προ Χριστού", "μετά Χριστόν"},
			timezones:              map[string]string{"ACDT": "Θερινή ώρα Κεντρικής Αυστραλίας", "ACST": "Χειμερινή ώρα Κεντρικής Αυστραλίας", "ACWDT": "Θερινή ώρα Κεντροδυτικής Αυστραλίας", "ACWST": "Χειμερινή ώρα Κεντροδυτικής Αυστραλίας", "ADT": "Θερινή ώρα Ατλαντικού", "AEDT": "Θερινή ώρα Ανατολικής Αυστραλίας", "AEST": "Χειμερινή ώρα Ανατολικής Αυστραλίας", "AKDT": "Θερινή ώρα Αλάσκας", "AKST": "Χειμερινή ώρα Αλάσκας", "ARST": "Θερινή ώρα Αργεντινής", "ART": "Χειμερινή ώρα Αργεντινής", "AST": "Χειμερινή ώρα Ατλαντικού", "AWDT": "Θερινή ώρα Δυτικής Αυστραλίας", "AWST": "Χειμερινή ώρα Δυτικής Αυστραλίας", "BOT": "Ώρα Βολιβίας", "BT": "Ώρα Μπουτάν", "CAT": "Ώρα Κεντρικής Αφρικής", "CDT": "Κεντρική θερινή ώρα Βόρειας Αμερικής", "CHADT": "Θερινή ώρα Τσάταμ", "CHAST": "Χειμερινή ώρα Τσάταμ", "CLST": "Θερινή ώρα Χιλής", "CLT": "Χειμερινή ώρα Χιλής", "COST": "Θερινή ώρα Κολομβίας", "COT": "Χειμερινή ώρα Κολομβίας", "CST": "Κεντρική χειμερινή ώρα Βόρειας Αμερικής", "ChST": "Ώρα Τσαμόρο", "EAT": "Ώρα Ανατολικής Αφρικής", "ECT": "Ώρα Ισημερινού", "EDT": "Ανατολική θερινή ώρα Βόρειας Αμερικής", "EST": "Ανατολική χειμερινή ώρα Βόρειας Αμερικής", "GFT": "Ώρα Γαλλικής Γουιάνας", "GMT": "Μέση ώρα Γκρίνουιτς", "GST": "Ώρα Κόλπου", "GYT": "Ώρα Γουιάνας", "HADT": "Θερινή ώρα Χαβάης-Αλεούτιων Νήσων", "HAST": "Χειμερινή ώρα Χαβάης-Αλεούτιων Νήσων", "HAT": "Θερινή ώρα Νέας Γης", "HECU": "Θερινή ώρα Κούβας", "HEEG": "Θερινή ώρα Ανατολικής Γροιλανδίας", "HENOMX": "Θερινή ώρα Βορειοδυτικού Μεξικού", "HEOG": "Θερινή ώρα Δυτικής Γροιλανδίας", "HEPM": "Θερινή ώρα Σεν Πιερ και Μικελόν", "HEPMX": "Θερινή ώρα Ειρηνικού Μεξικού", "HKST": "Θερινή ώρα Χονγκ Κονγκ", "HKT": "Χειμερινή ώρα Χονγκ Κονγκ", "HNCU": "Χειμερινή ώρα Κούβας", "HNEG": "Χειμερινή ώρα Ανατολικής Γροιλανδίας", "HNNOMX": "Χειμερινή ώρα Βορειοδυτικού Μεξικού", "HNOG": "Χειμερινή ώρα Δυτικής Γροιλανδίας", "HNPM": "Χειμερινή ώρα Σεν Πιερ και Μικελόν", "HNPMX": "Χειμερινή ώρα Ειρηνικού Μεξικού", "HNT": "Χειμερινή ώρα Νέας Γης", "IST": "Ώρα Ινδίας", "JDT": "Θερινή ώρα Ιαπωνίας", "JST": "Χειμερινή ώρα Ιαπωνίας", "LHDT": "Θερινή ώρα Λορντ Χάου", "LHST": "Χειμερινή ώρα Λορντ Χάου", "MDT": "Ορεινή θερινή ώρα Βόρειας Αμερικής", "MESZ": "Θερινή ώρα Κεντρικής Ευρώπης", "MEZ": "Χειμερινή ώρα Κεντρικής Ευρώπης", "MST": "Ορεινή χειμερινή ώρα Βόρειας Αμερικής", "MYT": "Ώρα Μαλαισίας", "NZDT": "Θερινή ώρα Νέας Ζηλανδίας", "NZST": "Χειμερινή ώρα Νέας Ζηλανδίας", "OESZ": "Θερινή ώρα Ανατολικής Ευρώπης", "OEZ": "Χειμερινή ώρα Ανατολικής Ευρώπης", "PDT": "Θερινή ώρα Ειρηνικού", "PST": "Χειμερινή ώρα Ειρηνικού", "SAST": "Χειμερινή ώρα Νότιας Αφρικής", "SGT": "Ώρα Σιγκαπούρης", "SRT": "Ώρα Σουρινάμ", "TMST": "Θερινή ώρα Τουρκμενιστάν", "TMT": "Χειμερινή ώρα Τουρκμενιστάν", "UYST": "Θερινή ώρα Ουρουγουάης", "UYT": "Χειμερινή ώρα Ουρουγουάης", "VET": "Ώρα Βενεζουέλας", "WARST": "Θερινή ώρα Δυτικής Αργεντινής", "WART": "Χειμερινή ώρα Δυτικής Αργεντινής", "WAST": "Θερινή ώρα Δυτικής Αφρικής", "WAT": "Χειμερινή ώρα Δυτικής Αφρικής", "WESZ": "Θερινή ώρα Δυτικής Ευρώπης", "WEZ": "Χειμερινή ώρα Δυτικής Ευρώπης", "WIB": "Ώρα Δυτικής Ινδονησίας", "WIT": "Ώρα Ανατολικής Ινδονησίας", "WITA": "Ώρα Κεντρικής Ινδονησίας", "∅∅∅": "Θερινή ώρα Μπραζίλιας"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["el_cy"] = func() locales.Translator {
		return &localen{

			locale:                 "el_CY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"},
			monthsNarrow:           []string{"", "Ι", "Φ", "Μ", "Α", "Μ", "Ι", "Ι", "Α", "Σ", "Ο", "Ν", "Δ"},
			monthsWide:             []string{"", "Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"},
			daysAbbreviated:        []string{"Κυρ", "Δευ", "Τρί", "Τετ", "Πέμ", "Παρ", "Σάβ"},
			daysNarrow:             []string{"Κ", "Δ", "Τ", "Τ", "Π", "Π", "Σ"},
			daysShort:              []string{"Κυ", "Δε", "Τρ", "Τε", "Πέ", "Πα", "Σά"},
			daysWide:               []string{"Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"},
			periodsAbbreviated:     []string{"π.μ.", "μ.μ."},
			periodsNarrow:          []string{"πμ", "μμ"},
			periodsWide:            []string{"π.μ.", "μ.μ."},
			erasAbbreviated:        []string{"π.Χ.", "μ.Χ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"προ Χριστού", "μετά Χριστόν"},
			timezones:              map[string]string{"ACDT": "Θερινή ώρα Κεντρικής Αυστραλίας", "ACST": "Χειμερινή ώρα Κεντρικής Αυστραλίας", "ACWDT": "Θερινή ώρα Κεντροδυτικής Αυστραλίας", "ACWST": "Χειμερινή ώρα Κεντροδυτικής Αυστραλίας", "ADT": "Θερινή ώρα Ατλαντικού", "AEDT": "Θερινή ώρα Ανατολικής Αυστραλίας", "AEST": "Χειμερινή ώρα Ανατολικής Αυστραλίας", "AKDT": "Θερινή ώρα Αλάσκας", "AKST": "Χειμερινή ώρα Αλάσκας", "ARST": "Θερινή ώρα Αργεντινής", "ART": "Χειμερινή ώρα Αργεντινής", "AST": "Χειμερινή ώρα Ατλαντικού", "AWDT": "Θερινή ώρα Δυτικής Αυστραλίας", "AWST": "Χειμερινή ώρα Δυτικής Αυστραλίας", "BOT": "Ώρα Βολιβίας", "BT": "Ώρα Μπουτάν", "CAT": "Ώρα Κεντρικής Αφρικής", "CDT": "Κεντρική θερινή ώρα Βόρειας Αμερικής", "CHADT": "Θερινή ώρα Τσάταμ", "CHAST": "Χειμερινή ώρα Τσάταμ", "CLST": "Θερινή ώρα Χιλής", "CLT": "Χειμερινή ώρα Χιλής", "COST": "Θερινή ώρα Κολομβίας", "COT": "Χειμερινή ώρα Κολομβίας", "CST": "Κεντρική χειμερινή ώρα Βόρειας Αμερικής", "ChST": "Ώρα Τσαμόρο", "EAT": "Ώρα Ανατολικής Αφρικής", "ECT": "Ώρα Ισημερινού", "EDT": "Ανατολική θερινή ώρα Βόρειας Αμερικής", "EST": "Ανατολική χειμερινή ώρα Βόρειας Αμερικής", "GFT": "Ώρα Γαλλικής Γουιάνας", "GMT": "Μέση ώρα Γκρίνουιτς", "GST": "Ώρα Κόλπου", "GYT": "Ώρα Γουιάνας", "HADT": "Θερινή ώρα Χαβάης-Αλεούτιων Νήσων", "HAST": "Χειμερινή ώρα Χαβάης-Αλεούτιων Νήσων", "HAT": "Θερινή ώρα Νέας Γης", "HECU": "Θερινή ώρα Κούβας", "HEEG": "Θερινή ώρα Ανατολικής Γροιλανδίας", "HENOMX": "Θερινή ώρα Βορειοδυτικού Μεξικού", "HEOG": "Θερινή ώρα Δυτικής Γροιλανδίας", "HEPM": "Θερινή ώρα Σεν Πιερ και Μικελόν", "HEPMX": "Θερινή ώρα Ειρηνικού Μεξικού", "HKST": "Θερινή ώρα Χονγκ Κονγκ", "HKT": "Χειμερινή ώρα Χονγκ Κονγκ", "HNCU": "Χειμερινή ώρα Κούβας", "HNEG": "Χειμερινή ώρα Ανατολικής Γροιλανδίας", "HNNOMX": "Χειμερινή ώρα Βορειοδυτικού Μεξικού", "HNOG": "Χειμερινή ώρα Δυτικής Γροιλανδίας", "HNPM": "Χειμερινή ώρα Σεν Πιερ και Μικελόν", "HNPMX": "Χειμερινή ώρα Ειρηνικού Μεξικού", "HNT": "Χειμερινή ώρα Νέας Γης", "IST": "Ώρα Ινδίας", "JDT": "Θερινή ώρα Ιαπωνίας", "JST": "Χειμερινή ώρα Ιαπωνίας", "LHDT": "Θερινή ώρα Λορντ Χάου", "LHST": "Χειμερινή ώρα Λορντ Χάου", "MDT": "Ορεινή θερινή ώρα Βόρειας Αμερικής", "MESZ": "Θερινή ώρα Κεντρικής Ευρώπης", "MEZ": "Χειμερινή ώρα Κεντρικής Ευρώπης", "MST": "Ορεινή χειμερινή ώρα Βόρειας Αμερικής", "MYT": "Ώρα Μαλαισίας", "NZDT": "Θερινή ώρα Νέας Ζηλανδίας", "NZST": "Χειμερινή ώρα Νέας Ζηλανδίας", "OESZ": "Θερινή ώρα Ανατολικής Ευρώπης", "OEZ": "Χειμερινή ώρα Ανατολικής Ευρώπης", "PDT": "Θερινή ώρα Ειρηνικού", "PST": "Χειμερινή ώρα Ειρηνικού", "SAST": "Χειμερινή ώρα Νότιας Αφρικής", "SGT": "Ώρα Σιγκαπούρης", "SRT": "Ώρα Σουρινάμ", "TMST": "Θερινή ώρα Τουρκμενιστάν", "TMT": "Χειμερινή ώρα Τουρκμενιστάν", "UYST": "Θερινή ώρα Ουρουγουάης", "UYT": "Χειμερινή ώρα Ουρουγουάης", "VET": "Ώρα Βενεζουέλας", "WARST": "Θερινή ώρα Δυτικής Αργεντινής", "WART": "Χειμερινή ώρα Δυτικής Αργεντινής", "WAST": "Θερινή ώρα Δυτικής Αφρικής", "WAT": "Χειμερινή ώρα Δυτικής Αφρικής", "WESZ": "Θερινή ώρα Δυτικής Ευρώπης", "WEZ": "Χειμερινή ώρα Δυτικής Ευρώπης", "WIB": "Ώρα Δυτικής Ινδονησίας", "WIT": "Ώρα Ανατολικής Ινδονησίας", "WITA": "Ώρα Κεντρικής Ινδονησίας", "∅∅∅": "Θερινή ώρα Περού"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["el_gr"] = func() locales.Translator {
		return &localen{

			locale:                 "el_GR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"},
			monthsNarrow:           []string{"", "Ι", "Φ", "Μ", "Α", "Μ", "Ι", "Ι", "Α", "Σ", "Ο", "Ν", "Δ"},
			monthsWide:             []string{"", "Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"},
			daysAbbreviated:        []string{"Κυρ", "Δευ", "Τρί", "Τετ", "Πέμ", "Παρ", "Σάβ"},
			daysNarrow:             []string{"Κ", "Δ", "Τ", "Τ", "Π", "Π", "Σ"},
			daysShort:              []string{"Κυ", "Δε", "Τρ", "Τε", "Πέ", "Πα", "Σά"},
			daysWide:               []string{"Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"},
			periodsAbbreviated:     []string{"π.μ.", "μ.μ."},
			periodsNarrow:          []string{"πμ", "μμ"},
			periodsWide:            []string{"π.μ.", "μ.μ."},
			erasAbbreviated:        []string{"π.Χ.", "μ.Χ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"προ Χριστού", "μετά Χριστόν"},
			timezones:              map[string]string{"ACDT": "Θερινή ώρα Κεντρικής Αυστραλίας", "ACST": "Χειμερινή ώρα Κεντρικής Αυστραλίας", "ACWDT": "Θερινή ώρα Κεντροδυτικής Αυστραλίας", "ACWST": "Χειμερινή ώρα Κεντροδυτικής Αυστραλίας", "ADT": "Θερινή ώρα Ατλαντικού", "AEDT": "Θερινή ώρα Ανατολικής Αυστραλίας", "AEST": "Χειμερινή ώρα Ανατολικής Αυστραλίας", "AKDT": "Θερινή ώρα Αλάσκας", "AKST": "Χειμερινή ώρα Αλάσκας", "ARST": "Θερινή ώρα Αργεντινής", "ART": "Χειμερινή ώρα Αργεντινής", "AST": "Χειμερινή ώρα Ατλαντικού", "AWDT": "Θερινή ώρα Δυτικής Αυστραλίας", "AWST": "Χειμερινή ώρα Δυτικής Αυστραλίας", "BOT": "Ώρα Βολιβίας", "BT": "Ώρα Μπουτάν", "CAT": "Ώρα Κεντρικής Αφρικής", "CDT": "Κεντρική θερινή ώρα Βόρειας Αμερικής", "CHADT": "Θερινή ώρα Τσάταμ", "CHAST": "Χειμερινή ώρα Τσάταμ", "CLST": "Θερινή ώρα Χιλής", "CLT": "Χειμερινή ώρα Χιλής", "COST": "Θερινή ώρα Κολομβίας", "COT": "Χειμερινή ώρα Κολομβίας", "CST": "Κεντρική χειμερινή ώρα Βόρειας Αμερικής", "ChST": "Ώρα Τσαμόρο", "EAT": "Ώρα Ανατολικής Αφρικής", "ECT": "Ώρα Ισημερινού", "EDT": "Ανατολική θερινή ώρα Βόρειας Αμερικής", "EST": "Ανατολική χειμερινή ώρα Βόρειας Αμερικής", "GFT": "Ώρα Γαλλικής Γουιάνας", "GMT": "Μέση ώρα Γκρίνουιτς", "GST": "Ώρα Κόλπου", "GYT": "Ώρα Γουιάνας", "HADT": "Θερινή ώρα Χαβάης-Αλεούτιων Νήσων", "HAST": "Χειμερινή ώρα Χαβάης-Αλεούτιων Νήσων", "HAT": "Θερινή ώρα Νέας Γης", "HECU": "Θερινή ώρα Κούβας", "HEEG": "Θερινή ώρα Ανατολικής Γροιλανδίας", "HENOMX": "Θερινή ώρα Βορειοδυτικού Μεξικού", "HEOG": "Θερινή ώρα Δυτικής Γροιλανδίας", "HEPM": "Θερινή ώρα Σεν Πιερ και Μικελόν", "HEPMX": "Θερινή ώρα Ειρηνικού Μεξικού", "HKST": "Θερινή ώρα Χονγκ Κονγκ", "HKT": "Χειμερινή ώρα Χονγκ Κονγκ", "HNCU": "Χειμερινή ώρα Κούβας", "HNEG": "Χειμερινή ώρα Ανατολικής Γροιλανδίας", "HNNOMX": "Χειμερινή ώρα Βορειοδυτικού Μεξικού", "HNOG": "Χειμερινή ώρα Δυτικής Γροιλανδίας", "HNPM": "Χειμερινή ώρα Σεν Πιερ και Μικελόν", "HNPMX": "Χειμερινή ώρα Ειρηνικού Μεξικού", "HNT": "Χειμερινή ώρα Νέας Γης", "IST": "Ώρα Ινδίας", "JDT": "Θερινή ώρα Ιαπωνίας", "JST": "Χειμερινή ώρα Ιαπωνίας", "LHDT": "Θερινή ώρα Λορντ Χάου", "LHST": "Χειμερινή ώρα Λορντ Χάου", "MDT": "Ορεινή θερινή ώρα Βόρειας Αμερικής", "MESZ": "Θερινή ώρα Κεντρικής Ευρώπης", "MEZ": "Χειμερινή ώρα Κεντρικής Ευρώπης", "MST": "Ορεινή χειμερινή ώρα Βόρειας Αμερικής", "MYT": "Ώρα Μαλαισίας", "NZDT": "Θερινή ώρα Νέας Ζηλανδίας", "NZST": "Χειμερινή ώρα Νέας Ζηλανδίας", "OESZ": "Θερινή ώρα Ανατολικής Ευρώπης", "OEZ": "Χειμερινή ώρα Ανατολικής Ευρώπης", "PDT": "Θερινή ώρα Ειρηνικού", "PST": "Χειμερινή ώρα Ειρηνικού", "SAST": "Χειμερινή ώρα Νότιας Αφρικής", "SGT": "Ώρα Σιγκαπούρης", "SRT": "Ώρα Σουρινάμ", "TMST": "Θερινή ώρα Τουρκμενιστάν", "TMT": "Χειμερινή ώρα Τουρκμενιστάν", "UYST": "Θερινή ώρα Ουρουγουάης", "UYT": "Χειμερινή ώρα Ουρουγουάης", "VET": "Ώρα Βενεζουέλας", "WARST": "Θερινή ώρα Δυτικής Αργεντινής", "WART": "Χειμερινή ώρα Δυτικής Αργεντινής", "WAST": "Θερινή ώρα Δυτικής Αφρικής", "WAT": "Χειμερινή ώρα Δυτικής Αφρικής", "WESZ": "Θερινή ώρα Δυτικής Ευρώπης", "WEZ": "Χειμερινή ώρα Δυτικής Ευρώπης", "WIB": "Ώρα Δυτικής Ινδονησίας", "WIT": "Ώρα Ανατολικής Ινδονησίας", "WITA": "Ώρα Κεντρικής Ινδονησίας", "∅∅∅": "Θερινή ώρα Μπραζίλιας"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en"] = func() locales.Translator {
		return &localen{

			locale:                 "en",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_001"] = func() locales.Translator {
		return &localen{

			locale:                 "en_001",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_150"] = func() locales.Translator {
		return &localen{

			locale:                 "en_150",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ae"] = func() locales.Translator {
		return &localen{

			locale:                 "en_AE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ag"] = func() locales.Translator {
		return &localen{

			locale:                 "en_AG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ai"] = func() locales.Translator {
		return &localen{

			locale:                 "en_AI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_as"] = func() locales.Translator {
		return &localen{

			locale:                 "en_AS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_at"] = func() locales.Translator {
		return &localen{

			locale:                 "en_AT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_au"] = func() locales.Translator {
		return &localen{

			locale:                 "en_AU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "₱", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "Rs", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "CFP", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"Su.", "M.", "Tu.", "W.", "Th.", "F.", "Sa."},
			daysShort:              []string{"Su", "Mon", "Tu", "Wed", "Th", "Fri", "Sat"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "Eastern Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Summer Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_bb"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BB",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "$", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_be"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_bi"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "FBu", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_bm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "$", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_bs"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "$", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_bw"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "P", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_bz"] = func() locales.Translator {
		return &localen{

			locale:                 "en_BZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "$", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ca"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "Mar.", "Apr.", "May", "Jun.", "Jul.", "Aug.", "Sep.", "Oct.", "Nov.", "Dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun.", "Mon.", "Tue.", "Wed.", "Thu.", "Fri.", "Sat."},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_cc"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  "’",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "-",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1cec81a7ffd91500b29d60f1aa496f40,
			fnFmtCurrency:         fnFmtCurrency_4924273cdfaa6d993990fdf2a2bad453,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ck"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_cx"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CX",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_cy"] = func() locales.Translator {
		return &localen{

			locale:                 "en_CY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_de"] = func() locales.Translator {
		return &localen{

			locale:                 "en_DE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_dg"] = func() locales.Translator {
		return &localen{

			locale:                 "en_DG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_dk"] = func() locales.Translator {
		return &localen{

			locale:                 "en_DK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "kr.", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_dm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_DM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_er"] = func() locales.Translator {
		return &localen{

			locale:                 "en_ER",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "Nfk", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_fi"] = func() locales.Translator {
		return &localen{

			locale:                 "en_FI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_105c4fe53621c460d37a501cec666cf0,
			fnFmtTimeLong:         fnFmtTimeLong_ed4e4ed1e2c01c9ce9c9f6344a272944,
			fnFmtTimeMedium:       fnFmtTimeMedium_ceaaad7cbbdab7b4a378fe8eff296f4f,
			fnFmtTimeShort:        fnFmtTimeShort_7341725c8ebd7a50d221ac00ecc6aa99,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_fj"] = func() locales.Translator {
		return &localen{

			locale:                 "en_FJ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "$", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_fk"] = func() locales.Translator {
		return &localen{

			locale:                 "en_FK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "£", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_fm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_FM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gb"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GB",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gd"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gg"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gh"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GH₵", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gi"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "£", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "D", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gu"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_gy"] = func() locales.Translator {
		return &localen{

			locale:                 "en_GY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "$", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_hk"] = func() locales.Translator {
		return &localen{

			locale:                 "en_HK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ie"] = func() locales.Translator {
		return &localen{

			locale:                 "en_IE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_il"] = func() locales.Translator {
		return &localen{

			locale:                 "en_IL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_im"] = func() locales.Translator {
		return &localen{

			locale:                 "en_IM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_in"] = func() locales.Translator {
		return &localen{

			locale:             "en_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:    []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:           []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated: []string{"am", "pm"},
			periodsNarrow:      []string{"", "pm"},
			periodsWide:        []string{"am", "pm"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"B", "A"},
			erasWide:           []string{"Before Christ", "Anno Domini"},
			timezones:          map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_io"] = func() locales.Translator {
		return &localen{

			locale:                 "en_IO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_je"] = func() locales.Translator {
		return &localen{

			locale:                 "en_JE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_jm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_JM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "$", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "en_KE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ki"] = func() locales.Translator {
		return &localen{

			locale:                 "en_KI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_kn"] = func() locales.Translator {
		return &localen{

			locale:                 "en_KN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ky"] = func() locales.Translator {
		return &localen{

			locale:                 "en_KY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "$", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Macao Summer Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Macao Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_lc"] = func() locales.Translator {
		return &localen{

			locale:                 "en_LC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_lr"] = func() locales.Translator {
		return &localen{

			locale:                 "en_LR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "$", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ls"] = func() locales.Translator {
		return &localen{

			locale:                 "en_LS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mg"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "Ar", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mh"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mo"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP$", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mp"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MP",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ms"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mt"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mu"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "Rs", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_mw"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_my"] = func() locales.Translator {
		return &localen{

			locale:                 "en_MY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Macao Summer Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Macao Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_na"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "$", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_nf"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ng"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "₦", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_nl"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_nr"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_nu"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_nz"] = func() locales.Translator {
		return &localen{

			locale:                 "en_NZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_d9e5c72d9a18dbee96ce40275bf098d9,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_pg"] = func() locales.Translator {
		return &localen{

			locale:                 "en_PG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "K", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ph"] = func() locales.Translator {
		return &localen{

			locale:                 "en_PH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_pk"] = func() locales.Translator {
		return &localen{

			locale:                 "en_PK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "Rs", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_pn"] = func() locales.Translator {
		return &localen{

			locale:                 "en_PN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_pr"] = func() locales.Translator {
		return &localen{

			locale:                 "en_PR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_pw"] = func() locales.Translator {
		return &localen{

			locale:                 "en_PW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_rw"] = func() locales.Translator {
		return &localen{

			locale:                 "en_RW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sb"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SB",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "$", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sc"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Macao Summer Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Macao Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sd"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_se"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "kr", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sg"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "$", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sh"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "£", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_si"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " )",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d9ac2a39e70db1e3f296933e6431fc50,
			fnFmtCurrency:         fnFmtCurrency_07bed080e46342275478260b22396b2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sl"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "Le", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ss"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "£", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sx"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SX",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "NAf.", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Macao Summer Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Macao Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_sz"] = func() locales.Translator {
		return &localen{

			locale:                 "en_SZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "E", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_tc"] = func() locales.Translator {
		return &localen{

			locale:                 "en_TC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_tk"] = func() locales.Translator {
		return &localen{

			locale:                 "en_TK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_to"] = func() locales.Translator {
		return &localen{

			locale:                 "en_TO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "T$", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_tt"] = func() locales.Translator {
		return &localen{

			locale:                 "en_TT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "$", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_tv"] = func() locales.Translator {
		return &localen{

			locale:                 "en_TV",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_tz"] = func() locales.Translator {
		return &localen{

			locale:                 "en_TZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ug"] = func() locales.Translator {
		return &localen{

			locale:                 "en_UG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_um"] = func() locales.Translator {
		return &localen{

			locale:                 "en_UM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_us"] = func() locales.Translator {
		return &localen{

			locale:                 "en_US",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_us_posix"] = func() locales.Translator {
		return &localen{

			locale:                 "en_US_POSIX",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "0/00",
			timeSeparator:          ":",
			inifinity:              "INF",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Amazon Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a127ece36a88f91b941483802dde95a0,
			fnFmtCurrency:         fnFmtCurrency_5fa862915ab8be0ffdb30cbcbd4dcadc,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_af287e772ea3855f073fceb6afe2a3de,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_vc"] = func() locales.Translator {
		return &localen{

			locale:                 "en_VC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Macao Summer Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Macao Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_vg"] = func() locales.Translator {
		return &localen{

			locale:                 "en_VG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cd651b8fbb843044a30f45f56d776cdd,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_vi"] = func() locales.Translator {
		return &localen{

			locale:                 "en_VI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St. Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St. Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_vu"] = func() locales.Translator {
		return &localen{

			locale:                 "en_VU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VT", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_ws"] = func() locales.Translator {
		return &localen{

			locale:                 "en_WS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WS$", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_za"] = func() locales.Translator {
		return &localen{

			locale:                 "en_ZA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ee2956d02335482ec6b7e9a423beb0b5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_933fcf2387ca32382dae956e0bc893a7,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_zm"] = func() locales.Translator {
		return &localen{

			locale:                 "en_ZM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "K", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["en_zw"] = func() locales.Translator {
		return &localen{

			locale:                 "en_ZW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"},
			daysWide:               []string{"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsNarrow:          []string{"", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"B", "A"},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Atlantic Daylight Time", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Atlantic Standard Time", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Central Daylight Time", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "Central Standard Time", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Eastern Daylight Time", "EST": "Eastern Standard Time", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time", "GYT": "Guyana Time", "HADT": "Hawaii-Aleutian Daylight Time", "HAST": "Hawaii-Aleutian Standard Time", "HAT": "Newfoundland Daylight Time", "HECU": "Cuba Daylight Time", "HEEG": "East Greenland Summer Time", "HENOMX": "Northwest Mexico Daylight Time", "HEOG": "West Greenland Summer Time", "HEPM": "St Pierre & Miquelon Daylight Time", "HEPMX": "Mexican Pacific Daylight Time", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Cuba Standard Time", "HNEG": "East Greenland Standard Time", "HNNOMX": "Northwest Mexico Standard Time", "HNOG": "West Greenland Standard Time", "HNPM": "St Pierre & Miquelon Standard Time", "HNPMX": "Mexican Pacific Standard Time", "HNT": "Newfoundland Standard Time", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Mountain Daylight Time", "MESZ": "Central European Summer Time", "MEZ": "Central European Standard Time", "MST": "Mountain Standard Time", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Eastern European Summer Time", "OEZ": "Eastern European Standard Time", "PDT": "Pacific Daylight Time", "PST": "Pacific Standard Time", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Western European Summer Time", "WEZ": "Western European Standard Time", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_03498001dbf6972dcc828c46e13ec4d2,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_cffd7bd3da07e13d4f77ebcac21d7f1e,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["eo"] = func() locales.Translator {
		return &localen{

			locale:                 "eo",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "₺", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "aŭg", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januaro", "februaro", "marto", "aprilo", "majo", "junio", "julio", "aŭgusto", "septembro", "oktobro", "novembro", "decembro"},
			daysAbbreviated:        []string{"di", "lu", "ma", "me", "ĵa", "ve", "sa"},
			daysNarrow:             []string{"D", "L", "M", "M", "Ĵ", "V", "S"},
			daysWide:               []string{"dimanĉo", "lundo", "mardo", "merkredo", "ĵaŭdo", "vendredo", "sabato"},
			periodsAbbreviated:     []string{"atm", "ptm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"atm", "ptm"},
			erasAbbreviated:        []string{"aK", "pK"},
			erasNarrow:             []string{"aK", "pK"},
			erasWide:               []string{"aK", "pK"},
			timezones:              map[string]string{"ACDT": "centra aŭstralia somera tempo", "ACST": "centra aŭstralia norma tempo", "ACWDT": "centrokcidenta aŭstralia somera tempo", "ACWST": "centrokcidenta aŭstralia norma tempo", "ADT": "atlantika nord-amerika somera tempo", "AEDT": "orienta aŭstralia somera tempo", "AEST": "orienta aŭstralia norma tempo", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "atlantika nord-amerika norma tempo", "AWDT": "okcidenta aŭstralia somera tempo", "AWST": "okcidenta aŭstralia norma tempo", "BOT": "BOT", "BT": "BT", "CAT": "centra afrika tempo", "CDT": "centra nord-amerika somera tempo", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "centra nord-amerika norma tempo", "ChST": "ChST", "EAT": "orienta afrika tempo", "ECT": "ECT", "EDT": "orienta nord-amerika somera tempo", "EST": "orienta nord-amerika norma tempo", "GFT": "GFT", "GMT": "universala tempo kunordigita", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "barata tempo", "JDT": "japana somera tempo", "JST": "japana norma tempo", "LHDT": "LHDT", "LHST": "LHST", "MDT": "monta nord-amerika somera tempo", "MESZ": "centra eŭropa somera tempo", "MEZ": "centra eŭropa norma tempo", "MST": "monta nord-amerika norma tempo", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "orienta eŭropa somera tempo", "OEZ": "orienta eŭropa norma tempo", "PDT": "pacifika nord-amerika somera tempo", "PST": "pacifika nord-amerika norma tempo", "SAST": "suda afrika tempo", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "okcidenta afrika somera tempo", "WAT": "okcidenta afrika norma tempo", "WESZ": "okcidenta eŭropa somera tempo", "WEZ": "okcidenta eŭropa norma tempo", "WIB": "okcidenta indonezia tempo", "WIT": "orienta indonezia tempo", "WITA": "centra indonezia tempo", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1faa91947b5020e2fec6143cbb8a4554,
			fnFmtCurrency:         fnFmtCurrency_f487e5cbe1d950a699f4afb3c0267309,
			fnFmtDateFull:         fnFmtDateFull_3d08635fa599fe67dc92007b3dff4e42,
			fnFmtDateLong:         fnFmtDateLong_29470171f6adaadee5f349206b90e4b3,
			fnFmtDateMedium:       fnFmtDateMedium_438f654701b049881de419adc50805ae,
			fnFmtDateShort:        fnFmtDateShort_f3b37aea35b3c974130a62cb3d9b1580,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_efbfbc738b43a1167c92b076dba0ef50,
			fnFmtTimeFull:         fnFmtTimeFull_8592fa74ce11c45fadae8071a016331d,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["eo_001"] = func() locales.Translator {
		return &localen{

			locale:                 "eo_001",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "aŭg", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januaro", "februaro", "marto", "aprilo", "majo", "junio", "julio", "aŭgusto", "septembro", "oktobro", "novembro", "decembro"},
			daysAbbreviated:        []string{"di", "lu", "ma", "me", "ĵa", "ve", "sa"},
			daysNarrow:             []string{"D", "L", "M", "M", "Ĵ", "V", "S"},
			daysWide:               []string{"dimanĉo", "lundo", "mardo", "merkredo", "ĵaŭdo", "vendredo", "sabato"},
			periodsAbbreviated:     []string{"atm", "ptm"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"atm", "ptm"},
			erasAbbreviated:        []string{"aK", "pK"},
			erasNarrow:             []string{"aK", "pK"},
			erasWide:               []string{"aK", "pK"},
			timezones:              map[string]string{"ACDT": "centra aŭstralia somera tempo", "ACST": "centra aŭstralia norma tempo", "ACWDT": "centrokcidenta aŭstralia somera tempo", "ACWST": "centrokcidenta aŭstralia norma tempo", "ADT": "atlantika nord-amerika somera tempo", "AEDT": "orienta aŭstralia somera tempo", "AEST": "orienta aŭstralia norma tempo", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "atlantika nord-amerika norma tempo", "AWDT": "okcidenta aŭstralia somera tempo", "AWST": "okcidenta aŭstralia norma tempo", "BOT": "BOT", "BT": "BT", "CAT": "centra afrika tempo", "CDT": "centra nord-amerika somera tempo", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "centra nord-amerika norma tempo", "ChST": "ChST", "EAT": "orienta afrika tempo", "ECT": "ECT", "EDT": "orienta nord-amerika somera tempo", "EST": "orienta nord-amerika norma tempo", "GFT": "GFT", "GMT": "universala tempo kunordigita", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "barata tempo", "JDT": "japana somera tempo", "JST": "japana norma tempo", "LHDT": "LHDT", "LHST": "LHST", "MDT": "monta nord-amerika somera tempo", "MESZ": "centra eŭropa somera tempo", "MEZ": "centra eŭropa norma tempo", "MST": "monta nord-amerika norma tempo", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "orienta eŭropa somera tempo", "OEZ": "orienta eŭropa norma tempo", "PDT": "pacifika nord-amerika somera tempo", "PST": "pacifika nord-amerika norma tempo", "SAST": "suda afrika tempo", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "okcidenta afrika somera tempo", "WAT": "okcidenta afrika norma tempo", "WESZ": "okcidenta eŭropa somera tempo", "WEZ": "okcidenta eŭropa norma tempo", "WIB": "okcidenta indonezia tempo", "WIT": "orienta indonezia tempo", "WITA": "centra indonezia tempo", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1faa91947b5020e2fec6143cbb8a4554,
			fnFmtCurrency:         fnFmtCurrency_f487e5cbe1d950a699f4afb3c0267309,
			fnFmtDateFull:         fnFmtDateFull_3d08635fa599fe67dc92007b3dff4e42,
			fnFmtDateLong:         fnFmtDateLong_29470171f6adaadee5f349206b90e4b3,
			fnFmtDateMedium:       fnFmtDateMedium_438f654701b049881de419adc50805ae,
			fnFmtDateShort:        fnFmtDateShort_f3b37aea35b3c974130a62cb3d9b1580,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_efbfbc738b43a1167c92b076dba0ef50,
			fnFmtTimeFull:         fnFmtTimeFull_8592fa74ce11c45fadae8071a016331d,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es"] = func() locales.Translator {
		return &localen{

			locale:                 "es",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "₧", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora estándar de la India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "Hora de verano de Macao", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "Hora estándar de Macao", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_419"] = func() locales.Translator {
		return &localen{

			locale:             "es_419",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "E£", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FK£", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SD£", "STD", "STN", "SUR", "SVC", "S£", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "BsF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ar"] = func() locales.Translator {
		return &localen{

			locale:                 "es_AR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_bo"] = func() locales.Translator {
		return &localen{

			locale:             "es_BO",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "Bs", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_0d7c652de418db235f2b6e6504ac834b,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_br"] = func() locales.Translator {
		return &localen{

			locale:             "es_BR",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de las Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_bz"] = func() locales.Translator {
		return &localen{

			locale:             "es_BZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "$", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_cl"] = func() locales.Translator {
		return &localen{

			locale:             "es_CL",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "$", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"do", "lu", "ma", "mi", "ju", "vi", "sá"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_co"] = func() locales.Translator {
		return &localen{

			locale:                 "es_CO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "$", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_d9e5c72d9a18dbee96ce40275bf098d9,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_cr"] = func() locales.Translator {
		return &localen{

			locale:             "es_CR",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "₡", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_faa0f0273050790b69e9c65ef1bb23a5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_cu"] = func() locales.Translator {
		return &localen{

			locale:             "es_CU",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "$", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_do"] = func() locales.Translator {
		return &localen{

			locale:                 "es_DO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "RD$", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: ")",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de la Era Común", "Era Común"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_2a94e5a57391565d92034abc27eaffd6,
			fnFmtCurrency:         fnFmtCurrency_07151969777dd3bbde27b10b4452a66e,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ea"] = func() locales.Translator {
		return &localen{

			locale:                 "es_EA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora estándar de la India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de las Montañas Rocosas", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de las Montañas Rocosas", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de las Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ec"] = func() locales.Translator {
		return &localen{

			locale:             "es_EC",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_es"] = func() locales.Translator {
		return &localen{

			locale:                 "es_ES",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora estándar de la India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "Hora de verano de Macao", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "Hora estándar de Macao", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_gq"] = func() locales.Translator {
		return &localen{

			locale:             "es_GQ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora estándar de la India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de las Montañas Rocosas", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de las Montañas Rocosas", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_gt"] = func() locales.Translator {
		return &localen{

			locale:                 "es_GT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "Q", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: "K",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_be7b974f27e584be73758ded6170b27f,
			fnFmtCurrency:         fnFmtCurrency_881a0ebc0b06ab2865571a388b8bfc08,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_d9e5c72d9a18dbee96ce40275bf098d9,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_hn"] = func() locales.Translator {
		return &localen{

			locale:             "es_HN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "L", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_dfeece500544d4aac4424f7155ba90ca,
			fnFmtDateLong:         fnFmtDateLong_536b20d0ef96110ea200943add1efbd4,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ic"] = func() locales.Translator {
		return &localen{

			locale:                 "es_IC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora estándar de la India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de las Montañas Rocosas", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de las Montañas Rocosas", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_mx"] = func() locales.Translator {
		return &localen{

			locale:                 "es_MX",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "p.", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "£", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRU", "UM", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "lei", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "£", "STD", "STN", "SUR", "SVC", "£", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "Bs", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " k",
			currencyNegativeSuffix: " k",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7fb5040dc04d16c016c545b9369751e6,
			fnFmtCurrency:         fnFmtCurrency_e4c5476e381e260e6a551a757542b603,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ni"] = func() locales.Translator {
		return &localen{

			locale:             "es_NI",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "C$", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_pa"] = func() locales.Translator {
		return &localen{

			locale:             "es_PA",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "B/.", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_85a85042ebcb9523c624cbf546fcb94d,
			fnFmtDateShort:        fnFmtDateShort_2aa2afdcdd2a2fb3ad7fdbe2b097c54b,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_pe"] = func() locales.Translator {
		return &localen{

			locale:                 "es_PE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "S/", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "set.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "setiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Perú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ph"] = func() locales.Translator {
		return &localen{

			locale:                 "es_PH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora estándar de la India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de las Montañas Rocosas", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de las Montañas Rocosas", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_pr"] = func() locales.Translator {
		return &localen{

			locale:             "es_PR",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de las Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_85a85042ebcb9523c624cbf546fcb94d,
			fnFmtDateShort:        fnFmtDateShort_2aa2afdcdd2a2fb3ad7fdbe2b097c54b,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_py"] = func() locales.Translator {
		return &localen{

			locale:                 "es_PY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "Gs.", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:              []string{"do", "lu", "ma", "mi", "ju", "vi", "sa"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_sv"] = func() locales.Translator {
		return &localen{

			locale:             "es_SV",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_us"] = func() locales.Translator {
		return &localen{

			locale:                 "es_US",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "p.", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "£", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "lei", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "£", "STD", "STN", "SUR", "SVC", "£", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "Bs", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " K ",
			currencyNegativeSuffix: " K ",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sep.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de las Montañas Rocosas", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de las Montañas Rocosas", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa oriental", "OEZ": "hora estándar de Europa oriental", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa occidental", "WEZ": "hora estándar de Europa occidental", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_34acd29c5979a61e92f53e9eaa5dcb93,
			fnFmtCurrency:         fnFmtCurrency_3b8812d78208fcdb8a36c9b00e4a7784,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_uy"] = func() locales.Translator {
		return &localen{

			locale:                 "es_UY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "$", "UP", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "set.", "oct.", "nov.", "dic."},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "setiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:        []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:             []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:              []string{"DO", "LU", "MA", "MI", "JU", "VI", "SA"},
			daysWide:               []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated:     []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:          []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsWide:            []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:        []string{"a. C.", "d. C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "después de Cristo"},
			timezones:              map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de las Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["es_ve"] = func() locales.Translator {
		return &localen{

			locale:             "es_VE",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "Bs.", "Bs.S", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ene.", "feb.", "mar.", "abr.", "may.", "jun.", "jul.", "ago.", "sept.", "oct.", "nov.", "dic."},
			monthsNarrow:       []string{"", "E", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"},
			daysAbbreviated:    []string{"dom.", "lun.", "mar.", "mié.", "jue.", "vie.", "sáb."},
			daysNarrow:         []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:          []string{"Do", "Lu", "Ma", "Mi", "Ju", "Vi", "Sa"},
			daysWide:           []string{"domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"},
			periodsAbbreviated: []string{"a.\u00a0m.", "p.\u00a0m."},
			periodsNarrow:      []string{"", ""},
			periodsWide:        []string{"a.\u00a0m.", "p.\u00a0m."},
			erasAbbreviated:    []string{"a. C.", "d. C."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"antes de Cristo", "después de Cristo"},
			timezones:          map[string]string{"ACDT": "hora de verano de Australia central", "ACST": "hora estándar de Australia central", "ACWDT": "hora de verano de Australia centroccidental", "ACWST": "hora estándar de Australia centroccidental", "ADT": "hora de verano del Atlántico", "AEDT": "hora de verano de Australia oriental", "AEST": "hora estándar de Australia oriental", "AKDT": "hora de verano de Alaska", "AKST": "hora estándar de Alaska", "ARST": "hora de verano de Argentina", "ART": "hora estándar de Argentina", "AST": "hora estándar del Atlántico", "AWDT": "hora de verano de Australia occidental", "AWST": "hora estándar de Australia occidental", "BOT": "hora de Bolivia", "BT": "hora de Bután", "CAT": "hora de África central", "CDT": "hora de verano central", "CHADT": "hora de verano de Chatham", "CHAST": "hora estándar de Chatham", "CLST": "hora de verano de Chile", "CLT": "hora estándar de Chile", "COST": "hora de verano de Colombia", "COT": "hora estándar de Colombia", "CST": "hora estándar central", "ChST": "hora estándar de Chamorro", "EAT": "hora de África oriental", "ECT": "hora de Ecuador", "EDT": "hora de verano oriental", "EST": "hora estándar oriental", "GFT": "hora de la Guayana Francesa", "GMT": "hora del meridiano de Greenwich", "GST": "hora estándar del Golfo", "GYT": "hora de Guyana", "HADT": "hora de verano de Hawái-Aleutianas", "HAST": "hora estándar de Hawái-Aleutianas", "HAT": "hora de verano de Terranova", "HECU": "hora de verano de Cuba", "HEEG": "hora de verano de Groenlandia oriental", "HENOMX": "hora de verano del noroeste de México", "HEOG": "hora de verano de Groenlandia occidental", "HEPM": "hora de verano de San Pedro y Miquelón", "HEPMX": "hora de verano del Pacífico de México", "HKST": "hora de verano de Hong Kong", "HKT": "hora estándar de Hong Kong", "HNCU": "hora estándar de Cuba", "HNEG": "hora estándar de Groenlandia oriental", "HNNOMX": "hora estándar del noroeste de México", "HNOG": "hora estándar de Groenlandia occidental", "HNPM": "hora estándar de San Pedro y Miquelón", "HNPMX": "hora estándar del Pacífico de México", "HNT": "hora estándar de Terranova", "IST": "hora de India", "JDT": "hora de verano de Japón", "JST": "hora estándar de Japón", "LHDT": "hora de verano de Lord Howe", "LHST": "hora estándar de Lord Howe", "MDT": "hora de verano de la montaña", "MESZ": "hora de verano de Europa central", "MEZ": "hora estándar de Europa central", "MST": "hora estándar de la montaña", "MYT": "hora de Malasia", "NZDT": "hora de verano de Nueva Zelanda", "NZST": "hora estándar de Nueva Zelanda", "OESZ": "hora de verano de Europa del Este", "OEZ": "hora estándar de Europa del Este", "PDT": "hora de verano del Pacífico", "PST": "hora estándar del Pacífico", "SAST": "hora de Sudáfrica", "SGT": "hora de Singapur", "SRT": "hora de Surinam", "TMST": "hora de verano de Turkmenistán", "TMT": "hora estándar de Turkmenistán", "UYST": "hora de verano de Uruguay", "UYT": "hora estándar de Uruguay", "VET": "hora de Venezuela", "WARST": "hora de verano de Argentina occidental", "WART": "hora estándar de Argentina occidental", "WAST": "hora de verano de África occidental", "WAT": "hora estándar de África occidental", "WESZ": "hora de verano de Europa del Oeste", "WEZ": "hora estándar de Europa del Oeste", "WIB": "hora de Indonesia occidental", "WIT": "hora de Indonesia oriental", "WITA": "hora de Indonesia central", "∅∅∅": "hora de verano de las Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["et"] = func() locales.Translator {
		return &localen{

			locale:                 "et",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "kr", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jaan", "veebr", "märts", "apr", "mai", "juuni", "juuli", "aug", "sept", "okt", "nov", "dets"},
			monthsNarrow:           []string{"", "J", "V", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "jaanuar", "veebruar", "märts", "aprill", "mai", "juuni", "juuli", "august", "september", "oktoober", "november", "detsember"},
			daysAbbreviated:        []string{"P", "E", "T", "K", "N", "R", "L"},
			daysNarrow:             []string{"P", "E", "T", "K", "N", "R", "L"},
			daysShort:              []string{"P", "E", "T", "K", "N", "R", "L"},
			daysWide:               []string{"pühapäev", "esmaspäev", "teisipäev", "kolmapäev", "neljapäev", "reede", "laupäev"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"eKr", "pKr"},
			erasNarrow:             []string{"eKr", "pKr"},
			erasWide:               []string{"enne Kristust", "pärast Kristust"},
			timezones:              map[string]string{"ACDT": "Kesk-Austraalia suveaeg", "ACST": "Kesk-Austraalia standardaeg", "ACWDT": "Austraalia Kesk-Lääne suveaeg", "ACWST": "Austraalia Kesk-Lääne standardaeg", "ADT": "Atlandi suveaeg", "AEDT": "Ida-Austraalia suveaeg", "AEST": "Ida-Austraalia standardaeg", "AKDT": "Alaska suveaeg", "AKST": "Alaska standardaeg", "ARST": "Argentina suveaeg", "ART": "Argentina standardaeg", "AST": "Atlandi standardaeg", "AWDT": "Lääne-Austraalia suveaeg", "AWST": "Lääne-Austraalia standardaeg", "BOT": "Boliivia aeg", "BT": "Bhutani aeg", "CAT": "Kesk-Aafrika aeg", "CDT": "Kesk-Ameerika suveaeg", "CHADT": "Chathami suveaeg", "CHAST": "Chathami standardaeg", "CLST": "Tšiili suveaeg", "CLT": "Tšiili standardaeg", "COST": "Colombia suveaeg", "COT": "Colombia standardaeg", "CST": "Kesk-Ameerika standardaeg", "ChST": "Tšamorro standardaeg", "EAT": "Ida-Aafrika aeg", "ECT": "Ecuadori aeg", "EDT": "Idaranniku suveaeg", "EST": "Idaranniku standardaeg", "GFT": "Prantsuse Guajaana aeg", "GMT": "Greenwichi aeg", "GST": "Pärsia lahe standardaeg", "GYT": "Guyana aeg", "HADT": "Hawaii-Aleuudi suveaeg", "HAST": "Hawaii-Aleuudi standardaeg", "HAT": "Newfoundlandi suveaeg", "HECU": "Kuuba suveaeg", "HEEG": "Ida-Gröönimaa suveaeg", "HENOMX": "Loode-Mehhiko suveaeg", "HEOG": "Lääne-Gröönimaa suveaeg", "HEPM": "Saint-Pierre’i ja Miqueloni suveaeg", "HEPMX": "Mehhiko Vaikse ookeani suveaeg", "HKST": "Hongkongi suveaeg", "HKT": "Hongkongi standardaeg", "HNCU": "Kuuba standardaeg", "HNEG": "Ida-Gröönimaa standardaeg", "HNNOMX": "Loode-Mehhiko standardaeg", "HNOG": "Lääne-Gröönimaa standardaeg", "HNPM": "Saint-Pierre’i ja Miqueloni standardaeg", "HNPMX": "Mehhiko Vaikse ookeani standardaeg", "HNT": "Newfoundlandi standardaeg", "IST": "India aeg", "JDT": "Jaapani suveaeg", "JST": "Jaapani standardaeg", "LHDT": "Lord Howe’i suveaeg", "LHST": "Lord Howe’i standardaeg", "MDT": "Mäestikuvööndi suveaeg", "MESZ": "Kesk-Euroopa suveaeg", "MEZ": "Kesk-Euroopa standardaeg", "MST": "Mäestikuvööndi standardaeg", "MYT": "Malaisia aeg", "NZDT": "Uus-Meremaa suveaeg", "NZST": "Uus-Meremaa standardaeg", "OESZ": "Ida-Euroopa suveaeg", "OEZ": "Ida-Euroopa standardaeg", "PDT": "Vaikse ookeani suveaeg", "PST": "Vaikse ookeani standardaeg", "SAST": "Lõuna-Aafrika standardaeg", "SGT": "Singapuri standardaeg", "SRT": "Suriname aeg", "TMST": "Türkmenistani suveaeg", "TMT": "Türkmenistani standardaeg", "UYST": "Uruguay suveaeg", "UYT": "Uruguay standardaeg", "VET": "Venezuela aeg", "WARST": "Lääne-Argentina suveaeg", "WART": "Lääne-Argentina standardaeg", "WAST": "Lääne-Aafrika suveaeg", "WAT": "Lääne-Aafrika standardaeg", "WESZ": "Lääne-Euroopa suveaeg", "WEZ": "Lääne-Euroopa standardaeg", "WIB": "Lääne-Indoneesia aeg", "WIT": "Ida-Indoneesia aeg", "WITA": "Kesk-Indoneesia aeg", "∅∅∅": "Peruu suveaeg"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a8f4207331b08e428b5c3f7d5c7887ea,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_efbfbc738b43a1167c92b076dba0ef50,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["et_ee"] = func() locales.Translator {
		return &localen{

			locale:                 "et_EE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jaan", "veebr", "märts", "apr", "mai", "juuni", "juuli", "aug", "sept", "okt", "nov", "dets"},
			monthsNarrow:           []string{"", "J", "V", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "jaanuar", "veebruar", "märts", "aprill", "mai", "juuni", "juuli", "august", "september", "oktoober", "november", "detsember"},
			daysAbbreviated:        []string{"P", "E", "T", "K", "N", "R", "L"},
			daysNarrow:             []string{"P", "E", "T", "K", "N", "R", "L"},
			daysShort:              []string{"P", "E", "T", "K", "N", "R", "L"},
			daysWide:               []string{"pühapäev", "esmaspäev", "teisipäev", "kolmapäev", "neljapäev", "reede", "laupäev"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"eKr", "pKr"},
			erasNarrow:             []string{"eKr", "pKr"},
			erasWide:               []string{"enne Kristust", "pärast Kristust"},
			timezones:              map[string]string{"ACDT": "Kesk-Austraalia suveaeg", "ACST": "Kesk-Austraalia standardaeg", "ACWDT": "Austraalia Kesk-Lääne suveaeg", "ACWST": "Austraalia Kesk-Lääne standardaeg", "ADT": "Atlandi suveaeg", "AEDT": "Ida-Austraalia suveaeg", "AEST": "Ida-Austraalia standardaeg", "AKDT": "Alaska suveaeg", "AKST": "Alaska standardaeg", "ARST": "Argentina suveaeg", "ART": "Argentina standardaeg", "AST": "Atlandi standardaeg", "AWDT": "Lääne-Austraalia suveaeg", "AWST": "Lääne-Austraalia standardaeg", "BOT": "Boliivia aeg", "BT": "Bhutani aeg", "CAT": "Kesk-Aafrika aeg", "CDT": "Kesk-Ameerika suveaeg", "CHADT": "Chathami suveaeg", "CHAST": "Chathami standardaeg", "CLST": "Tšiili suveaeg", "CLT": "Tšiili standardaeg", "COST": "Colombia suveaeg", "COT": "Colombia standardaeg", "CST": "Kesk-Ameerika standardaeg", "ChST": "Tšamorro standardaeg", "EAT": "Ida-Aafrika aeg", "ECT": "Ecuadori aeg", "EDT": "Idaranniku suveaeg", "EST": "Idaranniku standardaeg", "GFT": "Prantsuse Guajaana aeg", "GMT": "Greenwichi aeg", "GST": "Pärsia lahe standardaeg", "GYT": "Guyana aeg", "HADT": "Hawaii-Aleuudi suveaeg", "HAST": "Hawaii-Aleuudi standardaeg", "HAT": "Newfoundlandi suveaeg", "HECU": "Kuuba suveaeg", "HEEG": "Ida-Gröönimaa suveaeg", "HENOMX": "Loode-Mehhiko suveaeg", "HEOG": "Lääne-Gröönimaa suveaeg", "HEPM": "Saint-Pierre’i ja Miqueloni suveaeg", "HEPMX": "Mehhiko Vaikse ookeani suveaeg", "HKST": "Hongkongi suveaeg", "HKT": "Hongkongi standardaeg", "HNCU": "Kuuba standardaeg", "HNEG": "Ida-Gröönimaa standardaeg", "HNNOMX": "Loode-Mehhiko standardaeg", "HNOG": "Lääne-Gröönimaa standardaeg", "HNPM": "Saint-Pierre’i ja Miqueloni standardaeg", "HNPMX": "Mehhiko Vaikse ookeani standardaeg", "HNT": "Newfoundlandi standardaeg", "IST": "India aeg", "JDT": "Jaapani suveaeg", "JST": "Jaapani standardaeg", "LHDT": "Lord Howe’i suveaeg", "LHST": "Lord Howe’i standardaeg", "MDT": "Mäestikuvööndi suveaeg", "MESZ": "Kesk-Euroopa suveaeg", "MEZ": "Kesk-Euroopa standardaeg", "MST": "Mäestikuvööndi standardaeg", "MYT": "Malaisia aeg", "NZDT": "Uus-Meremaa suveaeg", "NZST": "Uus-Meremaa standardaeg", "OESZ": "Ida-Euroopa suveaeg", "OEZ": "Ida-Euroopa standardaeg", "PDT": "Vaikse ookeani suveaeg", "PST": "Vaikse ookeani standardaeg", "SAST": "Lõuna-Aafrika standardaeg", "SGT": "Singapuri standardaeg", "SRT": "Suriname aeg", "TMST": "Türkmenistani suveaeg", "TMT": "Türkmenistani standardaeg", "UYST": "Uruguay suveaeg", "UYT": "Uruguay standardaeg", "VET": "Venezuela aeg", "WARST": "Lääne-Argentina suveaeg", "WART": "Lääne-Argentina standardaeg", "WAST": "Lääne-Aafrika suveaeg", "WAT": "Lääne-Aafrika standardaeg", "WESZ": "Lääne-Euroopa suveaeg", "WEZ": "Lääne-Euroopa standardaeg", "WIB": "Lääne-Indoneesia aeg", "WIT": "Ida-Indoneesia aeg", "WITA": "Kesk-Indoneesia aeg", "∅∅∅": "Brasiilia suveaeg"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a8f4207331b08e428b5c3f7d5c7887ea,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_efbfbc738b43a1167c92b076dba0ef50,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["eu"] = func() locales.Translator {
		return &localen{

			locale:                 "eu",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "₧", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentPrefix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "urt.", "ots.", "mar.", "api.", "mai.", "eka.", "uzt.", "abu.", "ira.", "urr.", "aza.", "abe."},
			monthsNarrow:           []string{"", "U", "O", "M", "A", "M", "E", "U", "A", "I", "U", "A", "A"},
			monthsWide:             []string{"", "urtarrila", "otsaila", "martxoa", "apirila", "maiatza", "ekaina", "uztaila", "abuztua", "iraila", "urria", "azaroa", "abendua"},
			daysAbbreviated:        []string{"ig.", "al.", "ar.", "az.", "og.", "or.", "lr."},
			daysNarrow:             []string{"I", "A", "A", "A", "O", "O", "L"},
			daysShort:              []string{"ig.", "al.", "ar.", "az.", "og.", "or.", "lr."},
			daysWide:               []string{"igandea", "astelehena", "asteartea", "asteazkena", "osteguna", "ostirala", "larunbata"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"g", "a"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"K.a.", "K.o."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"K.a.", "Kristo ondoren"},
			timezones:              map[string]string{"ACDT": "Australiako erdialdeko udako ordua", "ACST": "Australiako erdialdeko ordu estandarra", "ACWDT": "Australiako erdi-mendebaldeko udako ordua", "ACWST": "Australiako erdi-mendebaldeko ordu estandarra", "ADT": "Ipar Amerikako Atlantikoko udako ordua", "AEDT": "Australiako ekialdeko udako ordua", "AEST": "Australiako ekialdeko ordu estandarra", "AKDT": "Alaskako udako ordua", "AKST": "Alaskako ordu estandarra", "ARST": "Argentinako udako ordua", "ART": "Argentinako ordu estandarra", "AST": "Ipar Amerikako Atlantikoko ordu estandarra", "AWDT": "Australiako mendebaldeko udako ordua", "AWST": "Australiako mendebaldeko ordu estandarra", "BOT": "Boliviako ordua", "BT": "Bhutango ordua", "CAT": "Afrikako erdialdeko ordua", "CDT": "Ipar Amerikako erdialdeko udako ordua", "CHADT": "Chathamgo udako ordua", "CHAST": "Chathamgo ordu estandarra", "CLST": "Txileko udako ordua", "CLT": "Txileko ordu estandarra", "COST": "Kolonbiako udako ordua", "COT": "Kolonbiako ordu estandarra", "CST": "Ipar Amerikako erdialdeko ordu estandarra", "ChST": "Chamorroko ordu estandarra", "EAT": "Afrikako ekialdeko ordua", "ECT": "Ekuadorreko ordua", "EDT": "Ipar Amerikako ekialdeko udako ordua", "EST": "Ipar Amerikako ekialdeko ordu estandarra", "GFT": "Guyana Frantseseko ordua", "GMT": "Greenwichko meridianoaren ordua", "GST": "Golkoko ordu estandarra", "GYT": "Guyanako ordua", "HADT": "Hawaii-Aleutiar uharteetako udako ordua", "HAST": "Hawaii-Aleutiar uharteetako ordu estandarra", "HAT": "Ternuako udako ordua", "HECU": "Kubako udako ordua", "HEEG": "Groenlandiako ekialdeko udako ordua", "HENOMX": "Mexikoko ipar-ekialdeko udako ordua", "HEOG": "Groenlandiako mendebaldeko udako ordua", "HEPM": "Saint-Pierre eta Mikeluneko udako ordua", "HEPMX": "Mexikoko Pazifikoko udako ordua", "HKST": "Hong Kongo udako ordua", "HKT": "Hong Kongo ordu estandarra", "HNCU": "Kubako ordu estandarra", "HNEG": "Groenlandiako ekialdeko ordu estandarra", "HNNOMX": "Mexikoko ipar-ekialdeko ordu estandarra", "HNOG": "Groenlandiako mendebaldeko ordu estandarra", "HNPM": "Saint-Pierre eta Mikeluneko ordu estandarra", "HNPMX": "Mexikoko Pazifikoko ordu estandarra", "HNT": "Ternuako ordu estandarra", "IST": "Indiako ordua", "JDT": "Japoniako udako ordua", "JST": "Japoniako ordu estandarra", "LHDT": "Lord Howeko udako ordua", "LHST": "Lord Howeko ordu estandarra", "MDT": "Ipar Amerikako mendialdeko udako ordua", "MESZ": "Europako erdialdeko udako ordua", "MEZ": "Europako erdialdeko ordu estandarra", "MST": "Ipar Amerikako mendialdeko ordu estandarra", "MYT": "Malaysiako ordua", "NZDT": "Zeelanda Berriko udako ordua", "NZST": "Zeelanda Berriko ordu estandarra", "OESZ": "Europako ekialdeko udako ordua", "OEZ": "Europako ekialdeko ordu estandarra", "PDT": "Ipar Amerikako Pazifikoko udako ordua", "PST": "Ipar Amerikako Pazifikoko ordu estandarra", "SAST": "Afrikako hegoaldeko ordua", "SGT": "Singapurreko ordu estandarra", "SRT": "Surinamgo ordua", "TMST": "Turkmenistango udako ordua", "TMT": "Turkmenistango ordu estandarra", "UYST": "Uruguaiko udako ordua", "UYT": "Uruguaiko ordu estandarra", "VET": "Venezuelako ordua", "WARST": "Argentina mendebaldeko udako ordua", "WART": "Argentina mendebaldeko ordu estandarra", "WAST": "Afrikako mendebaldeko udako ordua", "WAT": "Afrikako mendebaldeko ordu estandarra", "WESZ": "Europako mendebaldeko udako ordua", "WEZ": "Europako mendebaldeko ordu estandarra", "WIB": "Indonesiako mendebaldeko ordua", "WIT": "Indonesiako ekialdeko ordua", "WITA": "Indonesiako erdialdeko ordua", "∅∅∅": "Azoreetako udako ordua"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_f8187d99975235abc8e4a8d4e7b75463,
			fnFmtDateLong:         fnFmtDateLong_d0f2638a27c1f5c6700933f0a0be8242,
			fnFmtDateMedium:       fnFmtDateMedium_b224450320ab0608c36fa5f823a8ec71,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_7e5d64b2fa6162d30c9f7af69791fb63,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_30191015e11c1c183d8d41d8e0e78b42,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["eu_es"] = func() locales.Translator {
		return &localen{

			locale:                 "eu_ES",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentPrefix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "urt.", "ots.", "mar.", "api.", "mai.", "eka.", "uzt.", "abu.", "ira.", "urr.", "aza.", "abe."},
			monthsNarrow:           []string{"", "U", "O", "M", "A", "M", "E", "U", "A", "I", "U", "A", "A"},
			monthsWide:             []string{"", "urtarrila", "otsaila", "martxoa", "apirila", "maiatza", "ekaina", "uztaila", "abuztua", "iraila", "urria", "azaroa", "abendua"},
			daysAbbreviated:        []string{"ig.", "al.", "ar.", "az.", "og.", "or.", "lr."},
			daysNarrow:             []string{"I", "A", "A", "A", "O", "O", "L"},
			daysShort:              []string{"ig.", "al.", "ar.", "az.", "og.", "or.", "lr."},
			daysWide:               []string{"igandea", "astelehena", "asteartea", "asteazkena", "osteguna", "ostirala", "larunbata"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"g", "a"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"K.a.", "K.o."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"K.a.", "Kristo ondoren"},
			timezones:              map[string]string{"ACDT": "Australiako erdialdeko udako ordua", "ACST": "Australiako erdialdeko ordu estandarra", "ACWDT": "Australiako erdi-mendebaldeko udako ordua", "ACWST": "Australiako erdi-mendebaldeko ordu estandarra", "ADT": "Ipar Amerikako Atlantikoko udako ordua", "AEDT": "Australiako ekialdeko udako ordua", "AEST": "Australiako ekialdeko ordu estandarra", "AKDT": "Alaskako udako ordua", "AKST": "Alaskako ordu estandarra", "ARST": "Argentinako udako ordua", "ART": "Argentinako ordu estandarra", "AST": "Ipar Amerikako Atlantikoko ordu estandarra", "AWDT": "Australiako mendebaldeko udako ordua", "AWST": "Australiako mendebaldeko ordu estandarra", "BOT": "Boliviako ordua", "BT": "Bhutango ordua", "CAT": "Afrikako erdialdeko ordua", "CDT": "Ipar Amerikako erdialdeko udako ordua", "CHADT": "Chathamgo udako ordua", "CHAST": "Chathamgo ordu estandarra", "CLST": "Txileko udako ordua", "CLT": "Txileko ordu estandarra", "COST": "Kolonbiako udako ordua", "COT": "Kolonbiako ordu estandarra", "CST": "Ipar Amerikako erdialdeko ordu estandarra", "ChST": "Chamorroko ordu estandarra", "EAT": "Afrikako ekialdeko ordua", "ECT": "Ekuadorreko ordua", "EDT": "Ipar Amerikako ekialdeko udako ordua", "EST": "Ipar Amerikako ekialdeko ordu estandarra", "GFT": "Guyana Frantseseko ordua", "GMT": "Greenwichko meridianoaren ordua", "GST": "Golkoko ordu estandarra", "GYT": "Guyanako ordua", "HADT": "Hawaii-Aleutiar uharteetako udako ordua", "HAST": "Hawaii-Aleutiar uharteetako ordu estandarra", "HAT": "Ternuako udako ordua", "HECU": "Kubako udako ordua", "HEEG": "Groenlandiako ekialdeko udako ordua", "HENOMX": "Mexikoko ipar-ekialdeko udako ordua", "HEOG": "Groenlandiako mendebaldeko udako ordua", "HEPM": "Saint-Pierre eta Mikeluneko udako ordua", "HEPMX": "Mexikoko Pazifikoko udako ordua", "HKST": "Hong Kongo udako ordua", "HKT": "Hong Kongo ordu estandarra", "HNCU": "Kubako ordu estandarra", "HNEG": "Groenlandiako ekialdeko ordu estandarra", "HNNOMX": "Mexikoko ipar-ekialdeko ordu estandarra", "HNOG": "Groenlandiako mendebaldeko ordu estandarra", "HNPM": "Saint-Pierre eta Mikeluneko ordu estandarra", "HNPMX": "Mexikoko Pazifikoko ordu estandarra", "HNT": "Ternuako ordu estandarra", "IST": "Indiako ordua", "JDT": "Japoniako udako ordua", "JST": "Japoniako ordu estandarra", "LHDT": "Lord Howeko udako ordua", "LHST": "Lord Howeko ordu estandarra", "MDT": "MDT", "MESZ": "Europako erdialdeko udako ordua", "MEZ": "Europako erdialdeko ordu estandarra", "MST": "MST", "MYT": "Malaysiako ordua", "NZDT": "Zeelanda Berriko udako ordua", "NZST": "Zeelanda Berriko ordu estandarra", "OESZ": "Europako ekialdeko udako ordua", "OEZ": "Europako ekialdeko ordu estandarra", "PDT": "Ipar Amerikako Pazifikoko udako ordua", "PST": "Ipar Amerikako Pazifikoko ordu estandarra", "SAST": "Afrikako hegoaldeko ordua", "SGT": "Singapurreko ordu estandarra", "SRT": "Surinamgo ordua", "TMST": "Turkmenistango udako ordua", "TMT": "Turkmenistango ordu estandarra", "UYST": "Uruguaiko udako ordua", "UYT": "Uruguaiko ordu estandarra", "VET": "Venezuelako ordua", "WARST": "Argentina mendebaldeko udako ordua", "WART": "Argentina mendebaldeko ordu estandarra", "WAST": "Afrikako mendebaldeko udako ordua", "WAT": "Afrikako mendebaldeko ordu estandarra", "WESZ": "Europako mendebaldeko udako ordua", "WEZ": "Europako mendebaldeko ordu estandarra", "WIB": "Indonesiako mendebaldeko ordua", "WIT": "Indonesiako ekialdeko ordua", "WITA": "Indonesiako erdialdeko ordua", "∅∅∅": "Peruko udako ordua"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_f8187d99975235abc8e4a8d4e7b75463,
			fnFmtDateLong:         fnFmtDateLong_d0f2638a27c1f5c6700933f0a0be8242,
			fnFmtDateMedium:       fnFmtDateMedium_b224450320ab0608c36fa5f823a8ec71,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_7e5d64b2fa6162d30c9f7af69791fb63,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_30191015e11c1c183d8d41d8e0e78b42,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ewo"] = func() locales.Translator {
		return &localen{

			locale:                 "ewo",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ngo", "ngb", "ngl", "ngn", "ngt", "ngs", "ngz", "ngm", "nge", "nga", "ngad", "ngab"},
			monthsNarrow:           []string{"", "o", "b", "l", "n", "t", "s", "z", "m", "e", "a", "d", "b"},
			monthsWide:             []string{"", "ngɔn osú", "ngɔn bɛ̌", "ngɔn lála", "ngɔn nyina", "ngɔn tána", "ngɔn saməna", "ngɔn zamgbála", "ngɔn mwom", "ngɔn ebulú", "ngɔn awóm", "ngɔn awóm ai dziá", "ngɔn awóm ai bɛ̌"},
			daysAbbreviated:        []string{"sɔ́n", "mɔ́n", "smb", "sml", "smn", "fúl", "sér"},
			daysNarrow:             []string{"s", "m", "s", "s", "s", "f", "s"},
			daysWide:               []string{"sɔ́ndɔ", "mɔ́ndi", "sɔ́ndɔ məlú mə́bɛ̌", "sɔ́ndɔ məlú mə́lɛ́", "sɔ́ndɔ məlú mə́nyi", "fúladé", "séradé"},
			periodsAbbreviated:     []string{"kíkíríg", "ngəgógəle"},
			periodsWide:            []string{"kíkíríg", "ngəgógəle"},
			erasAbbreviated:        []string{"oyk", "ayk"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"osúsúa Yésus kiri", "ámvus Yésus Kirís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ewo_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "ewo_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ngo", "ngb", "ngl", "ngn", "ngt", "ngs", "ngz", "ngm", "nge", "nga", "ngad", "ngab"},
			monthsNarrow:           []string{"", "o", "b", "l", "n", "t", "s", "z", "m", "e", "a", "d", "b"},
			monthsWide:             []string{"", "ngɔn osú", "ngɔn bɛ̌", "ngɔn lála", "ngɔn nyina", "ngɔn tána", "ngɔn saməna", "ngɔn zamgbála", "ngɔn mwom", "ngɔn ebulú", "ngɔn awóm", "ngɔn awóm ai dziá", "ngɔn awóm ai bɛ̌"},
			daysAbbreviated:        []string{"sɔ́n", "mɔ́n", "smb", "sml", "smn", "fúl", "sér"},
			daysNarrow:             []string{"s", "m", "s", "s", "s", "f", "s"},
			daysWide:               []string{"sɔ́ndɔ", "mɔ́ndi", "sɔ́ndɔ məlú mə́bɛ̌", "sɔ́ndɔ məlú mə́lɛ́", "sɔ́ndɔ məlú mə́nyi", "fúladé", "séradé"},
			periodsAbbreviated:     []string{"kíkíríg", "ngəgógəle"},
			periodsWide:            []string{"kíkíríg", "ngəgógəle"},
			erasAbbreviated:        []string{"oyk", "ayk"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"osúsúa Yésus kiri", "ámvus Yésus Kirís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fa"] = func() locales.Translator {
		return &localen{

			locale:                 "fa",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‎−",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "؋", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "$CA", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "¥CN", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "$HK", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "ریال", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "$MX", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$NZ", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$EC", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: "‎",
			currencyNegativePrefix: "‎",
			monthsAbbreviated:      []string{"", "ژانویهٔ", "فوریهٔ", "مارس", "آوریل", "مهٔ", "ژوئن", "ژوئیهٔ", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			monthsNarrow:           []string{"", "ژ", "ف", "م", "آ", "م", "ژ", "ژ", "ا", "س", "ا", "ن", "د"},
			monthsWide:             []string{"", "ژانویهٔ", "فوریهٔ", "مارس", "آوریل", "مهٔ", "ژوئن", "ژوئیهٔ", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			daysAbbreviated:        []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			daysNarrow:             []string{"ی", "د", "س", "چ", "پ", "ج", "ش"},
			daysShort:              []string{"۱ش", "۲ش", "۳ش", "۴ش", "۵ش", "ج", "ش"},
			daysWide:               []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			periodsAbbreviated:     []string{"ق.ظ.", "ب.ظ."},
			periodsNarrow:          []string{"ق", "ب"},
			periodsWide:            []string{"قبل\u200cازظهر", "بعدازظهر"},
			erasAbbreviated:        []string{"ق.م.", "م."},
			erasNarrow:             []string{"ق", "م"},
			erasWide:               []string{"قبل از میلاد", "میلادی"},
			timezones:              map[string]string{"ACDT": "وقت تابستانی مرکز استرالیا", "ACST": "وقت عادی مرکز استرالیا", "ACWDT": "وقت تابستانی مرکز-غرب استرالیا", "ACWST": "وقت عادی مرکز-غرب استرالیا", "ADT": "وقت تابستانی آتلانتیک", "AEDT": "وقت تابستانی شرق استرالیا", "AEST": "وقت عادی شرق استرالیا", "AKDT": "وقت تابستانی آلاسکا", "AKST": "وقت عادی آلاسکا", "ARST": "وقت تابستانی آرژانتین", "ART": "وقت عادی آرژانتین", "AST": "وقت عادی آتلانتیک", "AWDT": "وقت تابستانی غرب استرالیا", "AWST": "وقت عادی غرب استرالیا", "BOT": "وقت بولیوی", "BT": "وقت بوتان", "CAT": "وقت مرکز افریقا", "CDT": "وقت تابستانی مرکز امریکا", "CHADT": "وقت تابستانی چت\u200cهام", "CHAST": "وقت عادی چت\u200cهام", "CLST": "وقت تابستانی شیلی", "CLT": "وقت عادی شیلی", "COST": "وقت تابستانی کلمبیا", "COT": "وقت عادی کلمبیا", "CST": "وقت عادی مرکز امریکا", "ChST": "وقت عادی چامورو", "EAT": "وقت شرق افریقا", "ECT": "وقت اکوادور", "EDT": "وقت تابستانی شرق امریکا", "EST": "وقت عادی شرق امریکا", "GFT": "وقت گویان فرانسه", "GMT": "وقت گرینویچ", "GST": "وقت عادی خلیج فارس", "GYT": "وقت گویان", "HADT": "وقت تابستانی هاوایی‐الوشن", "HAST": "وقت عادی هاوایی‐الوشن", "HAT": "وقت تابستانی نیوفاندلند", "HECU": "وقت تابستانی کوبا", "HEEG": "وقت تابستانی شرق گرینلند", "HENOMX": "وقت تابستانی شمال غرب مکزیک", "HEOG": "وقت تابستانی غرب گرینلند", "HEPM": "وقت تابستانی سنت\u200cپیر و میکلون", "HEPMX": "وقت تابستانی شرق مکزیک", "HKST": "وقت تابستانی هنگ\u200cکنگ", "HKT": "وقت عادی هنگ\u200cکنگ", "HNCU": "وقت عادی کوبا", "HNEG": "وقت عادی شرق گرینلند", "HNNOMX": "وقت عادی شمال غرب مکزیک", "HNOG": "وقت عادی غرب گرینلند", "HNPM": "وقت عادی سنت\u200cپیر و میکلون", "HNPMX": "وقت عادی شرق مکزیک", "HNT": "وقت عادی نیوفاندلند", "IST": "وقت هند", "JDT": "وقت تابستانی ژاپن", "JST": "وقت عادی ژاپن", "LHDT": "وقت تابستانی لردهو", "LHST": "وقت عادی لردهو", "MDT": "وقت تابستانی کوهستانی امریکا", "MESZ": "وقت تابستانی مرکز اروپا", "MEZ": "وقت عادی مرکز اروپا", "MST": "وقت عادی کوهستانی امریکا", "MYT": "وقت مالزی", "NZDT": "وقت تابستانی زلاند نو", "NZST": "وقت عادی زلاند نو", "OESZ": "وقت تابستانی شرق اروپا", "OEZ": "وقت عادی شرق اروپا", "PDT": "وقت تابستانی غرب امریکا", "PST": "وقت عادی غرب امریکا", "SAST": "وقت عادی جنوب افریقا", "SGT": "وقت سنگاپور", "SRT": "وقت سورینام", "TMST": "وقت تابستانی ترکمنستان", "TMT": "وقت عادی ترکمنستان", "UYST": "وقت تابستانی اروگوئه", "UYT": "وقت عادی اروگوئه", "VET": "وقت ونزوئلا", "WARST": "وقت تابستانی غرب آرژانتین", "WART": "وقت عادی غرب آرژانتین", "WAST": "وقت تابستانی غرب افریقا", "WAT": "وقت عادی غرب افریقا", "WESZ": "وقت تابستانی غرب اروپا", "WEZ": "وقت عادی غرب اروپا", "WIB": "وقت غرب اندونزی", "WIT": "وقت شرق اندونزی", "WITA": "وقت مرکز اندونزی", "∅∅∅": "وقت تابستانی پرو"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_980dd7146e58d4f3ce6aa6637bc1f3bc,
			fnFmtCurrency:         fnFmtCurrency_1ccdf5c54785959abd27ce458030d981,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_ae9cd15ae8cd02228932d7cfc0262f2b,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fa_af"] = func() locales.Translator {
		return &localen{

			locale:                 "fa_AF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‎−",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جنو", "فبروری", "مارچ", "اپریل", "می", "جون", "جول", "اگست", "سپتمبر", "اکتوبر", "نومبر", "دسم"},
			monthsNarrow:           []string{"", "ج", "ف", "م", "ا", "م", "ج", "ج", "ا", "س", "ا", "ن", "د"},
			monthsWide:             []string{"", "جنوری", "فبروری", "مارچ", "اپریل", "می", "جون", "جولای", "اگست", "سپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			daysNarrow:             []string{"ی", "د", "س", "چ", "پ", "ج", "ش"},
			daysShort:              []string{"۱ش", "۲ش", "۳ش", "۴ش", "۵ش", "ج", "ش"},
			daysWide:               []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"ق.م.", "م."},
			erasNarrow:             []string{"ق", "م"},
			erasWide:               []string{"قبل از میلاد", "میلادی"},
			timezones:              map[string]string{"ACDT": "وقت تابستانی مرکز استرالیا", "ACST": "وقت عادی مرکز استرالیا", "ACWDT": "وقت تابستانی مرکز-غرب استرالیا", "ACWST": "وقت عادی مرکز-غرب استرالیا", "ADT": "وقت تابستانی آتلانتیک", "AEDT": "وقت تابستانی شرق استرالیا", "AEST": "وقت عادی شرق استرالیا", "AKDT": "وقت تابستانی آلاسکا", "AKST": "وقت عادی آلاسکا", "ARST": "وقت تابستانی آرژانتین", "ART": "وقت عادی آرژانتین", "AST": "وقت عادی آتلانتیک", "AWDT": "وقت تابستانی غرب استرالیا", "AWST": "وقت عادی غرب استرالیا", "BOT": "وقت بولیوی", "BT": "وقت بوتان", "CAT": "وقت مرکز افریقا", "CDT": "وقت تابستانی مرکز امریکا", "CHADT": "وقت تابستانی چت\u200cهام", "CHAST": "وقت عادی چت\u200cهام", "CLST": "وقت تابستانی شیلی", "CLT": "وقت عادی شیلی", "COST": "وقت تابستانی کلمبیا", "COT": "وقت عادی کلمبیا", "CST": "وقت عادی مرکز امریکا", "ChST": "وقت عادی چامورو", "EAT": "وقت شرق افریقا", "ECT": "وقت اکوادور", "EDT": "وقت تابستانی شرق امریکا", "EST": "وقت عادی شرق امریکا", "GFT": "وقت گویان فرانسه", "GMT": "وقت گرینویچ", "GST": "وقت عادی خلیج فارس", "GYT": "وقت گویان", "HADT": "وقت تابستانی هاوایی‐الوشن", "HAST": "وقت عادی هاوایی‐الوشن", "HAT": "وقت تابستانی نیوفاندلند", "HECU": "وقت تابستانی کوبا", "HEEG": "وقت تابستانی شرق گرینلند", "HENOMX": "وقت تابستانی شمال غرب مکزیک", "HEOG": "وقت تابستانی غرب گرینلند", "HEPM": "وقت تابستانی سنت\u200cپیر و میکلون", "HEPMX": "وقت تابستانی شرق مکزیک", "HKST": "وقت تابستانی هنگ\u200cکنگ", "HKT": "وقت عادی هنگ\u200cکنگ", "HNCU": "وقت عادی کوبا", "HNEG": "وقت عادی شرق گرینلند", "HNNOMX": "وقت عادی شمال غرب مکزیک", "HNOG": "وقت عادی غرب گرینلند", "HNPM": "وقت عادی سنت\u200cپیر و میکلون", "HNPMX": "وقت عادی شرق مکزیک", "HNT": "وقت عادی نیوفاندلند", "IST": "وقت هند", "JDT": "وقت تابستانی ژاپن", "JST": "وقت عادی ژاپن", "LHDT": "وقت تابستانی لردهو", "LHST": "وقت عادی لردهو", "MDT": "وقت تابستانی کوهستانی امریکا", "MESZ": "وقت تابستانی مرکز اروپا", "MEZ": "وقت عادی مرکز اروپا", "MST": "وقت عادی کوهستانی امریکا", "MYT": "وقت مالزی", "NZDT": "وقت تابستانی زلاند نو", "NZST": "وقت عادی زلاند نو", "OESZ": "وقت تابستانی شرق اروپا", "OEZ": "وقت عادی شرق اروپا", "PDT": "وقت تابستانی غرب امریکا", "PST": "وقت عادی غرب امریکا", "SAST": "وقت عادی جنوب افریقا", "SGT": "وقت سنگاپور", "SRT": "وقت سورینام", "TMST": "وقت تابستانی ترکمنستان", "TMT": "وقت عادی ترکمنستان", "UYST": "وقت تابستانی اروگوئه", "UYT": "وقت عادی اروگوئه", "VET": "وقت ونزوئلا", "WARST": "وقت تابستانی غرب آرژانتین", "WART": "وقت عادی غرب آرژانتین", "WAST": "وقت تابستانی غرب افریقا", "WAT": "وقت عادی غرب افریقا", "WESZ": "وقت تابستانی غرب اروپا", "WEZ": "وقت عادی غرب اروپا", "WIB": "وقت غرب اندونزی", "WIT": "وقت شرق اندونزی", "WITA": "وقت مرکز اندونزی", "∅∅∅": "وقت تابستانی برازیلیا"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_15b09dc3a083df3b5892baf7c777f025,
			fnFmtCurrency:         fnFmtCurrency_8a07279a530bf4495d0582798587f7c0,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_ae9cd15ae8cd02228932d7cfc0262f2b,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fa_ir"] = func() locales.Translator {
		return &localen{

			locale:                 "fa_IR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‎−",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: "‎",
			currencyNegativePrefix: "‎",
			monthsAbbreviated:      []string{"", "ژانویهٔ", "فوریهٔ", "مارس", "آوریل", "مهٔ", "ژوئن", "ژوئیهٔ", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			monthsNarrow:           []string{"", "ژ", "ف", "م", "آ", "م", "ژ", "ژ", "ا", "س", "ا", "ن", "د"},
			monthsWide:             []string{"", "ژانویهٔ", "فوریهٔ", "مارس", "آوریل", "مهٔ", "ژوئن", "ژوئیهٔ", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			daysAbbreviated:        []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			daysNarrow:             []string{"ی", "د", "س", "چ", "پ", "ج", "ش"},
			daysShort:              []string{"۱ش", "۲ش", "۳ش", "۴ش", "۵ش", "ج", "ش"},
			daysWide:               []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			periodsAbbreviated:     []string{"ق.ظ.", "ب.ظ."},
			periodsNarrow:          []string{"ق", "ب"},
			periodsWide:            []string{"قبل\u200cازظهر", "بعدازظهر"},
			erasAbbreviated:        []string{"ق.م.", "م."},
			erasNarrow:             []string{"ق", "م"},
			erasWide:               []string{"قبل از میلاد", "میلادی"},
			timezones:              map[string]string{"ACDT": "وقت تابستانی مرکز استرالیا", "ACST": "وقت عادی مرکز استرالیا", "ACWDT": "وقت تابستانی مرکز-غرب استرالیا", "ACWST": "وقت عادی مرکز-غرب استرالیا", "ADT": "وقت تابستانی آتلانتیک", "AEDT": "وقت تابستانی شرق استرالیا", "AEST": "وقت عادی شرق استرالیا", "AKDT": "وقت تابستانی آلاسکا", "AKST": "وقت عادی آلاسکا", "ARST": "وقت تابستانی آرژانتین", "ART": "وقت عادی آرژانتین", "AST": "وقت عادی آتلانتیک", "AWDT": "وقت تابستانی غرب استرالیا", "AWST": "وقت عادی غرب استرالیا", "BOT": "وقت بولیوی", "BT": "وقت بوتان", "CAT": "وقت مرکز افریقا", "CDT": "وقت تابستانی مرکز امریکا", "CHADT": "وقت تابستانی چت\u200cهام", "CHAST": "وقت عادی چت\u200cهام", "CLST": "وقت تابستانی شیلی", "CLT": "وقت عادی شیلی", "COST": "وقت تابستانی کلمبیا", "COT": "وقت عادی کلمبیا", "CST": "وقت عادی مرکز امریکا", "ChST": "وقت عادی چامورو", "EAT": "وقت شرق افریقا", "ECT": "وقت اکوادور", "EDT": "وقت تابستانی شرق امریکا", "EST": "وقت عادی شرق امریکا", "GFT": "وقت گویان فرانسه", "GMT": "وقت گرینویچ", "GST": "وقت عادی خلیج فارس", "GYT": "وقت گویان", "HADT": "وقت تابستانی هاوایی‐الوشن", "HAST": "وقت عادی هاوایی‐الوشن", "HAT": "وقت تابستانی نیوفاندلند", "HECU": "وقت تابستانی کوبا", "HEEG": "وقت تابستانی شرق گرینلند", "HENOMX": "وقت تابستانی شمال غرب مکزیک", "HEOG": "وقت تابستانی غرب گرینلند", "HEPM": "وقت تابستانی سنت\u200cپیر و میکلون", "HEPMX": "وقت تابستانی شرق مکزیک", "HKST": "وقت تابستانی هنگ\u200cکنگ", "HKT": "وقت عادی هنگ\u200cکنگ", "HNCU": "وقت عادی کوبا", "HNEG": "وقت عادی شرق گرینلند", "HNNOMX": "وقت عادی شمال غرب مکزیک", "HNOG": "وقت عادی غرب گرینلند", "HNPM": "وقت عادی سنت\u200cپیر و میکلون", "HNPMX": "وقت عادی شرق مکزیک", "HNT": "وقت عادی نیوفاندلند", "IST": "وقت هند", "JDT": "وقت تابستانی ژاپن", "JST": "وقت عادی ژاپن", "LHDT": "وقت تابستانی لردهو", "LHST": "وقت عادی لردهو", "MDT": "وقت تابستانی کوهستانی امریکا", "MESZ": "وقت تابستانی مرکز اروپا", "MEZ": "وقت عادی مرکز اروپا", "MST": "وقت عادی کوهستانی امریکا", "MYT": "وقت مالزی", "NZDT": "وقت تابستانی زلاند نو", "NZST": "وقت عادی زلاند نو", "OESZ": "وقت تابستانی شرق اروپا", "OEZ": "وقت عادی شرق اروپا", "PDT": "وقت تابستانی غرب امریکا", "PST": "وقت عادی غرب امریکا", "SAST": "وقت عادی جنوب افریقا", "SGT": "وقت سنگاپور", "SRT": "وقت سورینام", "TMST": "وقت تابستانی ترکمنستان", "TMT": "وقت عادی ترکمنستان", "UYST": "وقت تابستانی اروگوئه", "UYT": "وقت عادی اروگوئه", "VET": "وقت ونزوئلا", "WARST": "وقت تابستانی غرب آرژانتین", "WART": "وقت عادی غرب آرژانتین", "WAST": "وقت تابستانی غرب افریقا", "WAT": "وقت عادی غرب افریقا", "WESZ": "وقت تابستانی غرب اروپا", "WEZ": "وقت عادی غرب اروپا", "WIB": "وقت غرب اندونزی", "WIT": "وقت شرق اندونزی", "WITA": "وقت مرکز اندونزی", "∅∅∅": "وقت تابستانی برازیلیا"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_980dd7146e58d4f3ce6aa6637bc1f3bc,
			fnFmtCurrency:         fnFmtCurrency_1ccdf5c54785959abd27ce458030d981,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_ae9cd15ae8cd02228932d7cfc0262f2b,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff"] = func() locales.Translator {
		return &localen{

			locale:                 "ff",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UZS", "VEB", "VEF", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"MEZ": "MEZ", "WARST": "WARST", "EAT": "EAT", "TMT": "TMT", "GFT": "GFT", "SGT": "SGT", "MDT": "MDT", "CHADT": "CHADT", "WIB": "WIB", "NZST": "NZST", "EST": "EST", "SRT": "SRT", "HADT": "HADT", "LHDT": "LHDT", "VET": "VET", "HNNOMX": "HNNOMX", "COST": "COST", "AEDT": "AEDT", "ADT": "ADT", "AST": "AST", "HKT": "HKT", "WITA": "WITA", "UYT": "UYT", "HECU": "HECU", "PDT": "PDT", "AWDT": "AWDT", "CST": "CST", "NZDT": "NZDT", "HNOG": "HNOG", "∅∅∅": "∅∅∅", "HENOMX": "HENOMX", "HAST": "HAST", "AEST": "AEST", "SAST": "SAST", "WART": "WART", "HNT": "HNT", "CLT": "CLT", "ARST": "ARST", "ACST": "ACST", "COT": "COT", "ChST": "ChST", "AKDT": "AKDT", "LHST": "LHST", "CLST": "CLST", "CAT": "CAT", "ART": "ART", "GYT": "GYT", "WEZ": "WEZ", "BT": "BT", "MESZ": "MESZ", "OESZ": "OESZ", "WAT": "WAT", "HEOG": "HEOG", "TMST": "TMST", "CHAST": "CHAST", "PST": "PST", "CDT": "CDT", "ACDT": "ACDT", "HEEG": "HEEG", "EDT": "EDT", "HEPM": "HEPM", "WAST": "WAST", "BOT": "BOT", "HKST": "HKST", "GMT": "GMT", "HNCU": "HNCU", "HEPMX": "HEPMX", "ACWDT": "ACWDT", "IST": "IST", "HAT": "HAT", "MST": "MST", "WIT": "WIT", "OEZ": "OEZ", "MYT": "MYT", "AKST": "AKST", "HNEG": "HNEG", "AWST": "AWST", "WESZ": "WESZ", "JST": "JST", "JDT": "JDT", "ECT": "ECT", "ACWST": "ACWST", "HNPM": "HNPM", "UYST": "UYST", "HNPMX": "HNPMX"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_gn"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_GN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "FG", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UZS", "VEB", "VEF", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"MEZ": "MEZ", "SRT": "SRT", "PST": "PST", "HEPMX": "HEPMX", "WEZ": "WEZ", "GFT": "GFT", "MYT": "MYT", "AKDT": "AKDT", "HEOG": "HEOG", "HKT": "HKT", "∅∅∅": "∅∅∅", "WIT": "WIT", "ChST": "ChST", "PDT": "PDT", "WAT": "WAT", "HENOMX": "HENOMX", "OESZ": "OESZ", "ACDT": "ACDT", "LHDT": "LHDT", "CHADT": "CHADT", "NZST": "NZST", "HEEG": "HEEG", "COST": "COST", "GMT": "GMT", "WAST": "WAST", "WESZ": "WESZ", "JST": "JST", "LHST": "LHST", "TMT": "TMT", "EAT": "EAT", "HNPMX": "HNPMX", "JDT": "JDT", "BOT": "BOT", "SGT": "SGT", "HAST": "HAST", "UYST": "UYST", "BT": "BT", "ECT": "ECT", "HNOG": "HNOG", "ACST": "ACST", "OEZ": "OEZ", "COT": "COT", "AWST": "AWST", "ADT": "ADT", "SAST": "SAST", "HKST": "HKST", "WITA": "WITA", "HADT": "HADT", "CHAST": "CHAST", "NZDT": "NZDT", "WARST": "WARST", "TMST": "TMST", "WIB": "WIB", "ACWST": "ACWST", "ART": "ART", "HNCU": "HNCU", "EST": "EST", "VET": "VET", "MST": "MST", "HECU": "HECU", "AWDT": "AWDT", "AEST": "AEST", "AKST": "AKST", "HEPM": "HEPM", "CLT": "CLT", "GYT": "GYT", "AST": "AST", "AEDT": "AEDT", "CLST": "CLST", "UYT": "UYT", "CST": "CST", "CDT": "CDT", "ACWDT": "ACWDT", "HNEG": "HNEG", "IST": "IST", "WART": "WART", "HAT": "HAT", "CAT": "CAT", "ARST": "ARST", "EDT": "EDT", "MESZ": "MESZ", "HNT": "HNT", "HNPM": "HNPM", "HNNOMX": "HNNOMX", "MDT": "MDT"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_bf"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_BF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_gh"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_GH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GH₵", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_gm"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_GM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "D", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_gn"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_GN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "FG", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_gw"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_GW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_lr"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_LR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "$", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_mr"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_MR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "UM", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_ne"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_NE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_ng"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_NG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "₦", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_sl"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_SL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "Le", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_latn_sn"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_Latn_SN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_mr"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_MR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "UM", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UZS", "VEB", "VEF", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"HNPM": "HNPM", "MST": "MST", "CAT": "CAT", "TMST": "TMST", "COT": "COT", "PST": "PST", "HNNOMX": "HNNOMX", "TMT": "TMT", "ART": "ART", "HNPMX": "HNPMX", "SAST": "SAST", "HEOG": "HEOG", "HAT": "HAT", "MDT": "MDT", "EAT": "EAT", "ChST": "ChST", "CDT": "CDT", "ACST": "ACST", "OEZ": "OEZ", "AEST": "AEST", "EDT": "EDT", "HADT": "HADT", "CST": "CST", "AST": "AST", "VET": "VET", "HNCU": "HNCU", "NZDT": "NZDT", "AKST": "AKST", "HAST": "HAST", "GMT": "GMT", "GYT": "GYT", "AWST": "AWST", "BOT": "BOT", "JST": "JST", "JDT": "JDT", "AKDT": "AKDT", "ECT": "ECT", "EST": "EST", "ACDT": "ACDT", "HNT": "HNT", "WIT": "WIT", "AWDT": "AWDT", "HKT": "HKT", "WART": "WART", "MEZ": "MEZ", "LHDT": "LHDT", "WITA": "WITA", "UYST": "UYST", "WEZ": "WEZ", "ACWDT": "ACWDT", "IST": "IST", "CHAST": "CHAST", "HECU": "HECU", "ADT": "ADT", "WAT": "WAT", "BT": "BT", "NZST": "NZST", "HNEG": "HNEG", "LHST": "LHST", "SRT": "SRT", "OESZ": "OESZ", "ARST": "ARST", "UYT": "UYT", "CHADT": "CHADT", "PDT": "PDT", "HEEG": "HEEG", "HENOMX": "HENOMX", "CLT": "CLT", "CLST": "CLST", "∅∅∅": "∅∅∅", "WAST": "WAST", "ACWST": "ACWST", "HEPM": "HEPM", "HEPMX": "HEPMX", "AEDT": "AEDT", "HNOG": "HNOG", "MESZ": "MESZ", "COST": "COST", "WESZ": "WESZ", "WIB": "WIB", "MYT": "MYT", "GFT": "GFT", "SGT": "SGT", "HKST": "HKST", "WARST": "WARST"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ff_sn"] = func() locales.Translator {
		return &localen{

			locale:                 "ff_SN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UZS", "VEB", "VEF", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"},
			monthsNarrow:           []string{"", "s", "c", "m", "s", "d", "k", "m", "j", "s", "y", "j", "b"},
			monthsWide:             []string{"", "siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"},
			daysAbbreviated:        []string{"dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"},
			daysNarrow:             []string{"d", "a", "m", "n", "n", "m", "h"},
			daysWide:               []string{"dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"},
			periodsAbbreviated:     []string{"subaka", "kikiiɗe"},
			periodsWide:            []string{"subaka", "kikiiɗe"},
			erasAbbreviated:        []string{"H-I", "C-I"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hade Iisa", "Caggal Iisa"},
			timezones:              map[string]string{"ChST": "ChST", "WAT": "WAT", "AKDT": "AKDT", "TMT": "TMT", "UYT": "UYT", "WESZ": "WESZ", "SAST": "SAST", "GFT": "GFT", "NZST": "NZST", "OEZ": "OEZ", "HNPM": "HNPM", "WITA": "WITA", "HAT": "HAT", "MDT": "MDT", "SRT": "SRT", "GMT": "GMT", "HEPMX": "HEPMX", "MEZ": "MEZ", "JDT": "JDT", "GYT": "GYT", "HNPMX": "HNPMX", "CST": "CST", "MYT": "MYT", "HNOG": "HNOG", "EDT": "EDT", "CAT": "CAT", "AWDT": "AWDT", "CDT": "CDT", "AST": "AST", "WEZ": "WEZ", "ACWST": "ACWST", "IST": "IST", "LHST": "LHST", "WIB": "WIB", "HEOG": "HEOG", "CLST": "CLST", "WIT": "WIT", "CHAST": "CHAST", "HECU": "HECU", "AWST": "AWST", "BOT": "BOT", "SGT": "SGT", "HENOMX": "HENOMX", "MESZ": "MESZ", "AEDT": "AEDT", "EST": "EST", "ART": "ART", "PDT": "PDT", "NZDT": "NZDT", "ACST": "ACST", "HNNOMX": "HNNOMX", "COT": "COT", "HAST": "HAST", "HADT": "HADT", "JST": "JST", "ECT": "ECT", "ACDT": "ACDT", "OESZ": "OESZ", "PST": "PST", "WAST": "WAST", "WARST": "WARST", "HNEG": "HNEG", "HKST": "HKST", "CLT": "CLT", "TMST": "TMST", "EAT": "EAT", "ACWDT": "ACWDT", "BT": "BT", "HNCU": "HNCU", "MST": "MST", "∅∅∅": "∅∅∅", "HKT": "HKT", "WART": "WART", "LHDT": "LHDT", "VET": "VET", "COST": "COST", "UYST": "UYST", "HEEG": "HEEG", "AKST": "AKST", "HNT": "HNT", "ARST": "ARST", "CHADT": "CHADT", "ADT": "ADT", "AEST": "AEST", "HEPM": "HEPM"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fi"] = func() locales.Translator {
		return &localen{

			locale:                 "fi",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "mk", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "tammik.", "helmik.", "maalisk.", "huhtik.", "toukok.", "kesäk.", "heinäk.", "elok.", "syysk.", "lokak.", "marrask.", "jouluk."},
			monthsNarrow:           []string{"", "T", "H", "M", "H", "T", "K", "H", "E", "S", "L", "M", "J"},
			monthsWide:             []string{"", "tammikuuta", "helmikuuta", "maaliskuuta", "huhtikuuta", "toukokuuta", "kesäkuuta", "heinäkuuta", "elokuuta", "syyskuuta", "lokakuuta", "marraskuuta", "joulukuuta"},
			daysAbbreviated:        []string{"su", "ma", "ti", "ke", "to", "pe", "la"},
			daysNarrow:             []string{"S", "M", "T", "K", "T", "P", "L"},
			daysShort:              []string{"su", "ma", "ti", "ke", "to", "pe", "la"},
			daysWide:               []string{"sunnuntaina", "maanantaina", "tiistaina", "keskiviikkona", "torstaina", "perjantaina", "lauantaina"},
			periodsAbbreviated:     []string{"ap.", "ip."},
			periodsNarrow:          []string{"ap.", "ip."},
			periodsWide:            []string{"ap.", "ip."},
			erasAbbreviated:        []string{"eKr.", "jKr."},
			erasNarrow:             []string{"eKr", "jKr"},
			erasWide:               []string{"ennen Kristuksen syntymää", "jälkeen Kristuksen syntymän"},
			timezones:              map[string]string{"ACDT": "Keski-Australian kesäaika", "ACST": "Keski-Australian normaaliaika", "ACWDT": "Läntisen Keski-Australian kesäaika", "ACWST": "Läntisen Keski-Australian normaaliaika", "ADT": "Kanadan Atlantin kesäaika", "AEDT": "Itä-Australian kesäaika", "AEST": "Itä-Australian normaaliaika", "AKDT": "Alaskan kesäaika", "AKST": "Alaskan normaaliaika", "ARST": "Argentiinan kesäaika", "ART": "Argentiinan normaaliaika", "AST": "Kanadan Atlantin normaaliaika", "AWDT": "Länsi-Australian kesäaika", "AWST": "Länsi-Australian normaaliaika", "BOT": "Bolivian aika", "BT": "Bhutanin aika", "CAT": "Keski-Afrikan aika", "CDT": "Yhdysvaltain keskinen kesäaika", "CHADT": "Chathamin kesäaika", "CHAST": "Chathamin normaaliaika", "CLST": "Chilen kesäaika", "CLT": "Chilen normaaliaika", "COST": "Kolumbian kesäaika", "COT": "Kolumbian normaaliaika", "CST": "Yhdysvaltain keskinen normaaliaika", "ChST": "Tšamorron aika", "EAT": "Itä-Afrikan aika", "ECT": "Ecuadorin aika", "EDT": "Yhdysvaltain itäinen kesäaika", "EST": "Yhdysvaltain itäinen normaaliaika", "GFT": "Ranskan Guayanan aika", "GMT": "Greenwichin normaaliaika", "GST": "Arabiemiirikuntien normaaliaika", "GYT": "Guyanan aika", "HADT": "Havaijin-Aleuttien kesäaika", "HAST": "Havaijin-Aleuttien normaaliaika", "HAT": "Newfoundlandin kesäaika", "HECU": "Kuuban kesäaika", "HEEG": "Itä-Grönlannin kesäaika", "HENOMX": "Luoteis-Meksikon kesäaika", "HEOG": "Länsi-Grönlannin kesäaika", "HEPM": "Saint-Pierren ja Miquelonin kesäaika", "HEPMX": "Meksikon Tyynenmeren kesäaika", "HKST": "Hongkongin kesäaika", "HKT": "Hongkongin normaaliaika", "HNCU": "Kuuban normaaliaika", "HNEG": "Itä-Grönlannin normaaliaika", "HNNOMX": "Luoteis-Meksikon normaaliaika", "HNOG": "Länsi-Grönlannin normaaliaika", "HNPM": "Saint-Pierren ja Miquelonin normaaliaika", "HNPMX": "Meksikon Tyynenmeren normaaliaika", "HNT": "Newfoundlandin normaaliaika", "IST": "Intian aika", "JDT": "Japanin kesäaika", "JST": "Japanin normaaliaika", "LHDT": "Lord Howen kesäaika", "LHST": "Lord Howen normaaliaika", "MDT": "Kalliovuorten kesäaika", "MESZ": "Keski-Euroopan kesäaika", "MEZ": "Keski-Euroopan normaaliaika", "MST": "Kalliovuorten normaaliaika", "MYT": "Malesian aika", "NZDT": "Uuden-Seelannin kesäaika", "NZST": "Uuden-Seelannin normaaliaika", "OESZ": "Itä-Euroopan kesäaika", "OEZ": "Itä-Euroopan normaaliaika", "PDT": "Yhdysvaltain Tyynenmeren kesäaika", "PST": "Yhdysvaltain Tyynenmeren normaaliaika", "SAST": "Etelä-Afrikan aika", "SGT": "Singaporen aika", "SRT": "Surinamen aika", "TMST": "Turkmenistanin kesäaika", "TMT": "Turkmenistanin normaaliaika", "UYST": "Uruguayn kesäaika", "UYT": "Uruguayn normaaliaika", "VET": "Venezuelan aika", "WARST": "Länsi-Argentiinan kesäaika", "WART": "Länsi-Argentiinan normaaliaika", "WAST": "Länsi-Afrikan kesäaika", "WAT": "Länsi-Afrikan normaaliaika", "WESZ": "Länsi-Euroopan kesäaika", "WEZ": "Länsi-Euroopan normaaliaika", "WIB": "Länsi-Indonesian aika", "WIT": "Itä-Indonesian aika", "WITA": "Keski-Indonesian aika", "∅∅∅": "Brasilian kesäaika"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_d1d28b42f6cc00a3a8fb2b3af29b2b99,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_105c4fe53621c460d37a501cec666cf0,
			fnFmtTimeLong:         fnFmtTimeLong_ed4e4ed1e2c01c9ce9c9f6344a272944,
			fnFmtTimeMedium:       fnFmtTimeMedium_ceaaad7cbbdab7b4a378fe8eff296f4f,
			fnFmtTimeShort:        fnFmtTimeShort_7341725c8ebd7a50d221ac00ecc6aa99,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fi_fi"] = func() locales.Translator {
		return &localen{

			locale:                 "fi_FI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "tammik.", "helmik.", "maalisk.", "huhtik.", "toukok.", "kesäk.", "heinäk.", "elok.", "syysk.", "lokak.", "marrask.", "jouluk."},
			monthsNarrow:           []string{"", "T", "H", "M", "H", "T", "K", "H", "E", "S", "L", "M", "J"},
			monthsWide:             []string{"", "tammikuuta", "helmikuuta", "maaliskuuta", "huhtikuuta", "toukokuuta", "kesäkuuta", "heinäkuuta", "elokuuta", "syyskuuta", "lokakuuta", "marraskuuta", "joulukuuta"},
			daysAbbreviated:        []string{"su", "ma", "ti", "ke", "to", "pe", "la"},
			daysNarrow:             []string{"S", "M", "T", "K", "T", "P", "L"},
			daysShort:              []string{"su", "ma", "ti", "ke", "to", "pe", "la"},
			daysWide:               []string{"sunnuntaina", "maanantaina", "tiistaina", "keskiviikkona", "torstaina", "perjantaina", "lauantaina"},
			periodsAbbreviated:     []string{"ap.", "ip."},
			periodsNarrow:          []string{"ap.", "ip."},
			periodsWide:            []string{"ap.", "ip."},
			erasAbbreviated:        []string{"eKr.", "jKr."},
			erasNarrow:             []string{"eKr", "jKr"},
			erasWide:               []string{"ennen Kristuksen syntymää", "jälkeen Kristuksen syntymän"},
			timezones:              map[string]string{"ACDT": "Keski-Australian kesäaika", "ACST": "Keski-Australian normaaliaika", "ACWDT": "Läntisen Keski-Australian kesäaika", "ACWST": "Läntisen Keski-Australian normaaliaika", "ADT": "Kanadan Atlantin kesäaika", "AEDT": "Itä-Australian kesäaika", "AEST": "Itä-Australian normaaliaika", "AKDT": "Alaskan kesäaika", "AKST": "Alaskan normaaliaika", "ARST": "Argentiinan kesäaika", "ART": "Argentiinan normaaliaika", "AST": "Kanadan Atlantin normaaliaika", "AWDT": "Länsi-Australian kesäaika", "AWST": "Länsi-Australian normaaliaika", "BOT": "Bolivian aika", "BT": "Bhutanin aika", "CAT": "Keski-Afrikan aika", "CDT": "Yhdysvaltain keskinen kesäaika", "CHADT": "Chathamin kesäaika", "CHAST": "Chathamin normaaliaika", "CLST": "Chilen kesäaika", "CLT": "Chilen normaaliaika", "COST": "Kolumbian kesäaika", "COT": "Kolumbian normaaliaika", "CST": "Yhdysvaltain keskinen normaaliaika", "ChST": "Tšamorron aika", "EAT": "Itä-Afrikan aika", "ECT": "Ecuadorin aika", "EDT": "Yhdysvaltain itäinen kesäaika", "EST": "Yhdysvaltain itäinen normaaliaika", "GFT": "Ranskan Guayanan aika", "GMT": "Greenwichin normaaliaika", "GST": "Arabiemiirikuntien normaaliaika", "GYT": "Guyanan aika", "HADT": "Havaijin-Aleuttien kesäaika", "HAST": "Havaijin-Aleuttien normaaliaika", "HAT": "Newfoundlandin kesäaika", "HECU": "Kuuban kesäaika", "HEEG": "Itä-Grönlannin kesäaika", "HENOMX": "Luoteis-Meksikon kesäaika", "HEOG": "Länsi-Grönlannin kesäaika", "HEPM": "Saint-Pierren ja Miquelonin kesäaika", "HEPMX": "Meksikon Tyynenmeren kesäaika", "HKST": "Hongkongin kesäaika", "HKT": "Hongkongin normaaliaika", "HNCU": "Kuuban normaaliaika", "HNEG": "Itä-Grönlannin normaaliaika", "HNNOMX": "Luoteis-Meksikon normaaliaika", "HNOG": "Länsi-Grönlannin normaaliaika", "HNPM": "Saint-Pierren ja Miquelonin normaaliaika", "HNPMX": "Meksikon Tyynenmeren normaaliaika", "HNT": "Newfoundlandin normaaliaika", "IST": "Intian aika", "JDT": "Japanin kesäaika", "JST": "Japanin normaaliaika", "LHDT": "Lord Howen kesäaika", "LHST": "Lord Howen normaaliaika", "MDT": "Kalliovuorten kesäaika", "MESZ": "Keski-Euroopan kesäaika", "MEZ": "Keski-Euroopan normaaliaika", "MST": "Kalliovuorten normaaliaika", "MYT": "Malesian aika", "NZDT": "Uuden-Seelannin kesäaika", "NZST": "Uuden-Seelannin normaaliaika", "OESZ": "Itä-Euroopan kesäaika", "OEZ": "Itä-Euroopan normaaliaika", "PDT": "Yhdysvaltain Tyynenmeren kesäaika", "PST": "Yhdysvaltain Tyynenmeren normaaliaika", "SAST": "Etelä-Afrikan aika", "SGT": "Singaporen aika", "SRT": "Surinamen aika", "TMST": "Turkmenistanin kesäaika", "TMT": "Turkmenistanin normaaliaika", "UYST": "Uruguayn kesäaika", "UYT": "Uruguayn normaaliaika", "VET": "Venezuelan aika", "WARST": "Länsi-Argentiinan kesäaika", "WART": "Länsi-Argentiinan normaaliaika", "WAST": "Länsi-Afrikan kesäaika", "WAT": "Länsi-Afrikan normaaliaika", "WESZ": "Länsi-Euroopan kesäaika", "WEZ": "Länsi-Euroopan normaaliaika", "WIB": "Länsi-Indonesian aika", "WIT": "Itä-Indonesian aika", "WITA": "Keski-Indonesian aika", "∅∅∅": "Brasilian kesäaika"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_d1d28b42f6cc00a3a8fb2b3af29b2b99,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_105c4fe53621c460d37a501cec666cf0,
			fnFmtTimeLong:         fnFmtTimeLong_ed4e4ed1e2c01c9ce9c9f6344a272944,
			fnFmtTimeMedium:       fnFmtTimeMedium_ceaaad7cbbdab7b4a378fe8eff296f4f,
			fnFmtTimeShort:        fnFmtTimeShort_7341725c8ebd7a50d221ac00ecc6aa99,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fil"] = func() locales.Translator {
		return &localen{

			locale:                 "fil",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Set", "Okt", "Nob", "Dis"},
			monthsNarrow:           []string{"", "Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Set", "Okt", "Nob", "Dis"},
			monthsWide:             []string{"", "Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Setyembre", "Oktubre", "Nobyembre", "Disyembre"},
			daysAbbreviated:        []string{"Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"},
			daysNarrow:             []string{"Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"},
			daysShort:              []string{"Li", "Lu", "Ma", "Mi", "Hu", "Bi", "Sa"},
			daysWide:               []string{"Linggo", "Lunes", "Martes", "Miyerkules", "Huwebes", "Biyernes", "Sabado"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Daylight Time sa Gitnang Australya", "ACST": "Standard na Oras sa Gitnang Australya", "ACWDT": "Daylight Time sa Gitnang Kanlurang Australya", "ACWST": "Standard Time ng Gitnang Kanluran ng Australya", "ADT": "Daylight Time sa Atlantiko", "AEDT": "Daylight Time sa Silangang Australya", "AEST": "Standard na Oras sa Silangang Australya", "AKDT": "Daylight Time sa Alaska", "AKST": "Standard na Oras sa Alaska", "ARST": "Oras sa Tag-init ng Argentina", "ART": "Standard na Oras sa Argentina", "AST": "Standard na Oras sa Atlantiko", "AWDT": "Daylight Time sa Kanlurang Australya", "AWST": "Standard na Oras sa Kanlurang Australya", "BOT": "Oras sa Bolivia", "BT": "Oras sa Bhutan", "CAT": "Oras sa Gitnang Africa", "CDT": "Sentral na Daylight Time", "CHADT": "Daylight Time sa Chatham", "CHAST": "Standard na Oras sa Chatham", "CLST": "Oras sa Tag-init ng Chile", "CLT": "Standard na Oras sa Chile", "COST": "Oras sa Tag-init ng Colombia", "COT": "Standard na Oras sa Colombia", "CST": "Sentral na Karaniwang Oras", "ChST": "Standard na Oras sa Chamorro", "EAT": "Oras sa Silangang Africa", "ECT": "Oras sa Ecuador", "EDT": "Eastern Daylight Time", "EST": "Eastern na Standard na Oras", "GFT": "Oras sa French Guiana", "GMT": "Greenwich Mean Time", "GST": "Oras sa Gulf", "GYT": "Oras sa Guyana", "HADT": "Oras sa Tag-init ng Hawaii-Aleutian", "HAST": "Standard na Oras sa Hawaii-Aleutian", "HAT": "Daylight Time sa Newfoundland", "HECU": "Daylight Time sa Cuba", "HEEG": "Oras sa Tag-init ng Silangang Greenland", "HENOMX": "Daylight Time sa Hilagang-kanlurang Mexico", "HEOG": "Oras sa Tag-init ng Kanlurang Greenland", "HEPM": "Daylight Time sa Saint Pierre & Miquelon", "HEPMX": "Daylight Time sa Pasipiko ng Mexico", "HKST": "Oras sa Tag-init ng Hong Kong", "HKT": "Standard na Oras sa Hong Kong", "HNCU": "Standard na Oras sa Cuba", "HNEG": "Standard na Oras sa Silangang Greenland", "HNNOMX": "Standard na Oras sa Hilagang-kanlurang Mexico", "HNOG": "Standard na Oras sa Kanlurang Greenland", "HNPM": "Standard na Oras sa Saint Pierre & Miquelon", "HNPMX": "Standard na Oras sa Pasipiko ng Mexico", "HNT": "Standard na Oras sa Newfoundland", "IST": "Standard na Oras sa Bhutan", "JDT": "Daylight Time sa Japan", "JST": "Standard na Oras sa Japan", "LHDT": "Daylight Time sa Lorde Howe", "LHST": "Standard na Oras sa Lord Howe", "MDT": "Daylight Time sa Bundok", "MESZ": "Oras sa Tag-init ng Gitnang Europe", "MEZ": "Standard na Oras sa Gitnang Europe", "MST": "Standard na Oras sa Bundok", "MYT": "Oras sa Malaysia", "NZDT": "Daylight Time sa New Zealand", "NZST": "Standard na Oras sa New Zealand", "OESZ": "Oras sa Tag-init ng Silangang Europe", "OEZ": "Standard na Oras sa Silangang Europe", "PDT": "Daylight Time sa Pasipiko", "PST": "Standard na Oras sa Pasipiko", "SAST": "Oras sa Timog Africa", "SGT": "Standard na Oras sa Singapore", "SRT": "Oras sa Suriname", "TMST": "Oras sa Tag-init ng Turkmenistan", "TMT": "Standard na Oras sa Turkmenistan", "UYST": "Oras sa Tag-init ng Uruguay", "UYT": "Standard na Oras sa Uruguay", "VET": "Oras sa Venezuela", "WARST": "Oras sa Tag-init ng Kanlurang Argentina", "WART": "Standard na Oras sa Kanlurang Argentina", "WAST": "Oras sa Tag-init ng Kanlurang Africa", "WAT": "Standard na Oras sa Kanlurang Africa", "WESZ": "Oras sa Tag-init ng Kanlurang Europe", "WEZ": "Standard na Oras sa Kanlurang Europe", "WIB": "Oras sa Kanlurang Indonesia", "WIT": "Oras sa Silangang Indonesia", "WITA": "Oras sa Gitnang Indonesia", "∅∅∅": "Oras sa Tag-init ng Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c1a7400a6209ae7e8eb28592c58f6657,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fil_ph"] = func() locales.Translator {
		return &localen{

			locale:                 "fil_PH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Set", "Okt", "Nob", "Dis"},
			monthsNarrow:           []string{"", "Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Set", "Okt", "Nob", "Dis"},
			monthsWide:             []string{"", "Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Setyembre", "Oktubre", "Nobyembre", "Disyembre"},
			daysAbbreviated:        []string{"Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"},
			daysNarrow:             []string{"Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"},
			daysShort:              []string{"Li", "Lu", "Ma", "Mi", "Hu", "Bi", "Sa"},
			daysWide:               []string{"Linggo", "Lunes", "Martes", "Miyerkules", "Huwebes", "Biyernes", "Sabado"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Before Christ", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Daylight Time sa Gitnang Australya", "ACST": "Standard na Oras sa Gitnang Australya", "ACWDT": "Daylight Time sa Gitnang Kanlurang Australya", "ACWST": "Standard Time ng Gitnang Kanluran ng Australya", "ADT": "Daylight Time sa Atlantiko", "AEDT": "Daylight Time sa Silangang Australya", "AEST": "Standard na Oras sa Silangang Australya", "AKDT": "Daylight Time sa Alaska", "AKST": "Standard na Oras sa Alaska", "ARST": "Oras sa Tag-init ng Argentina", "ART": "Standard na Oras sa Argentina", "AST": "Standard na Oras sa Atlantiko", "AWDT": "Daylight Time sa Kanlurang Australya", "AWST": "Standard na Oras sa Kanlurang Australya", "BOT": "Oras sa Bolivia", "BT": "Oras sa Bhutan", "CAT": "Oras sa Gitnang Africa", "CDT": "Sentral na Daylight Time", "CHADT": "Daylight Time sa Chatham", "CHAST": "Standard na Oras sa Chatham", "CLST": "Oras sa Tag-init ng Chile", "CLT": "Standard na Oras sa Chile", "COST": "Oras sa Tag-init ng Colombia", "COT": "Standard na Oras sa Colombia", "CST": "Sentral na Karaniwang Oras", "ChST": "Standard na Oras sa Chamorro", "EAT": "Oras sa Silangang Africa", "ECT": "Oras sa Ecuador", "EDT": "Eastern Daylight Time", "EST": "Eastern na Standard na Oras", "GFT": "Oras sa French Guiana", "GMT": "Greenwich Mean Time", "GST": "Oras sa Gulf", "GYT": "Oras sa Guyana", "HADT": "Oras sa Tag-init ng Hawaii-Aleutian", "HAST": "Standard na Oras sa Hawaii-Aleutian", "HAT": "Daylight Time sa Newfoundland", "HECU": "Daylight Time sa Cuba", "HEEG": "Oras sa Tag-init ng Silangang Greenland", "HENOMX": "Daylight Time sa Hilagang-kanlurang Mexico", "HEOG": "Oras sa Tag-init ng Kanlurang Greenland", "HEPM": "Daylight Time sa Saint Pierre & Miquelon", "HEPMX": "Daylight Time sa Pasipiko ng Mexico", "HKST": "Oras sa Tag-init ng Hong Kong", "HKT": "Standard na Oras sa Hong Kong", "HNCU": "Standard na Oras sa Cuba", "HNEG": "Standard na Oras sa Silangang Greenland", "HNNOMX": "Standard na Oras sa Hilagang-kanlurang Mexico", "HNOG": "Standard na Oras sa Kanlurang Greenland", "HNPM": "Standard na Oras sa Saint Pierre & Miquelon", "HNPMX": "Standard na Oras sa Pasipiko ng Mexico", "HNT": "Standard na Oras sa Newfoundland", "IST": "Standard na Oras sa Bhutan", "JDT": "Daylight Time sa Japan", "JST": "Standard na Oras sa Japan", "LHDT": "Daylight Time sa Lorde Howe", "LHST": "Standard na Oras sa Lord Howe", "MDT": "Daylight Time sa Bundok", "MESZ": "Oras sa Tag-init ng Gitnang Europe", "MEZ": "Standard na Oras sa Gitnang Europe", "MST": "Standard na Oras sa Bundok", "MYT": "Oras sa Malaysia", "NZDT": "Daylight Time sa New Zealand", "NZST": "Standard na Oras sa New Zealand", "OESZ": "Oras sa Tag-init ng Silangang Europe", "OEZ": "Standard na Oras sa Silangang Europe", "PDT": "Daylight Time sa Pasipiko", "PST": "Standard na Oras sa Pasipiko", "SAST": "Oras sa Timog Africa", "SGT": "Standard na Oras sa Singapore", "SRT": "Oras sa Suriname", "TMST": "Oras sa Tag-init ng Turkmenistan", "TMT": "Standard na Oras sa Turkmenistan", "UYST": "Oras sa Tag-init ng Uruguay", "UYT": "Standard na Oras sa Uruguay", "VET": "Oras sa Venezuela", "WARST": "Oras sa Tag-init ng Kanlurang Argentina", "WART": "Standard na Oras sa Kanlurang Argentina", "WAST": "Oras sa Tag-init ng Kanlurang Africa", "WAT": "Standard na Oras sa Kanlurang Africa", "WESZ": "Oras sa Tag-init ng Kanlurang Europe", "WEZ": "Standard na Oras sa Kanlurang Europe", "WIB": "Oras sa Kanlurang Indonesia", "WIT": "Oras sa Silangang Indonesia", "WITA": "Oras sa Gitnang Indonesia", "∅∅∅": "Oras sa Tag-init ng Amazon"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c1a7400a6209ae7e8eb28592c58f6657,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fo"] = func() locales.Translator {
		return &localen{

			locale:                 "fo",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "kr", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "mai", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "apríl", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"sun.", "mán.", "týs.", "mik.", "hós.", "frí.", "ley."},
			daysNarrow:             []string{"S", "M", "T", "M", "H", "F", "L"},
			daysShort:              []string{"su.", "má.", "tý.", "mi.", "hó.", "fr.", "le."},
			daysWide:               []string{"sunnudagur", "mánadagur", "týsdagur", "mikudagur", "hósdagur", "fríggjadagur", "leygardagur"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"fKr", "eKr"},
			erasWide:               []string{"fyri Krist", "eftir Krist"},
			timezones:              map[string]string{"ACDT": "mið Avstralia summartíð", "ACST": "mið Avstralia vanlig tíð", "ACWDT": "miðvestur Avstralia summartíð", "ACWST": "miðvestur Avstralia vanlig tíð", "ADT": "Atlantic summartíð", "AEDT": "eystur Avstralia summartíð", "AEST": "eystur Avstralia vanlig tíð", "AKDT": "Alaska summartíð", "AKST": "Alaska vanlig tíð", "ARST": "Argentina summartíð", "ART": "Argentina vanlig tíð", "AST": "Atlantic vanlig tíð", "AWDT": "vestur Avstralia summartíð", "AWST": "vestur Avstralia vanlig tíð", "BOT": "Bolivia tíð", "BT": "Butan tíð", "CAT": "Miðafrika tíð", "CDT": "Central summartíð", "CHADT": "Chatham summartíð", "CHAST": "Chatham vanlig tíð", "CLST": "Kili summartíð", "CLT": "Kili vanlig tíð", "COST": "Kolombia summartíð", "COT": "Kolombia vanlig tíð", "CST": "Central vanlig tíð", "ChST": "Chamorro vanlig tíð", "EAT": "Eysturafrika tíð", "ECT": "Ekvador tíð", "EDT": "Eastern summartíð", "EST": "Eastern vanlig tíð", "GFT": "Franska Gujana tíð", "GMT": "Greenwich Mean tíð", "GST": "Gulf vanlig tíð", "GYT": "Gujana tíð", "HADT": "Hawaii-Aleutian summartíð", "HAST": "Hawaii-Aleutian vanlig tíð", "HAT": "Newfoundland summartíð", "HECU": "Cuba summartíð", "HEEG": "Eystur grønlendsk summartíð", "HENOMX": "Northwest Mexico summartíð", "HEOG": "Vestur grønlendsk summartíð", "HEPM": "St. Pierre & Miquelon summartíð", "HEPMX": "Mexican Pacific summartíð", "HKST": "Hong Kong summartíð", "HKT": "Hong Kong vanlig tíð", "HNCU": "Cuba vanlig tíð", "HNEG": "Eystur grønlendsk vanlig tíð", "HNNOMX": "Northwest Mexico vanlig tíð", "HNOG": "Vestur grønlendsk vanlig tíð", "HNPM": "St. Pierre & Miquelon vanlig tíð", "HNPMX": "Mexican Pacific vanlig tíð", "HNT": "Newfoundland vanlig tíð", "IST": "India tíð", "JDT": "Japan summartíð", "JST": "Japan vanlig tíð", "LHDT": "Lord Howe summartíð", "LHST": "Lord Howe vanlig tíð", "MDT": "Mountain summartíð", "MESZ": "Miðevropa summartíð", "MEZ": "Miðevropa vanlig tíð", "MST": "Mountain vanlig tíð", "MYT": "Malaisia tíð", "NZDT": "Nýsæland summartíð", "NZST": "Nýsæland vanlig tíð", "OESZ": "Eysturevropa summartíð", "OEZ": "Eysturevropa vanlig tíð", "PDT": "Pacific summartíð", "PST": "Pacific vanlig tíð", "SAST": "Suðurafrika vanlig tíð", "SGT": "Singapor tíð", "SRT": "Surinam tíð", "TMST": "Turkmenistan summartíð", "TMT": "Turkmenistan vanlig tíð", "UYST": "Uruguai summartíð", "UYT": "Uruguai vanlig tíð", "VET": "Venesuela tíð", "WARST": "Vestur Argentina summartíð", "WART": "Vestur Argentina vanlig tíð", "WAST": "Vesturafrika summartíð", "WAT": "Vesturafrika vanlig tíð", "WESZ": "Vesturevropa summartíð", "WEZ": "Vesturevropa vanlig tíð", "WIB": "Vestur Indonesia tíð", "WIT": "Eystur Indonesia tíð", "WITA": "Mið Indonesia tíð", "∅∅∅": "Amasona summartíð"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fo_dk"] = func() locales.Translator {
		return &localen{

			locale:                 "fo_DK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "kr.", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "mai", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "apríl", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"sun.", "mán.", "týs.", "mik.", "hós.", "frí.", "ley."},
			daysNarrow:             []string{"S", "M", "T", "M", "H", "F", "L"},
			daysShort:              []string{"su.", "má.", "tý.", "mi.", "hó.", "fr.", "le."},
			daysWide:               []string{"sunnudagur", "mánadagur", "týsdagur", "mikudagur", "hósdagur", "fríggjadagur", "leygardagur"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"fKr", "eKr"},
			erasWide:               []string{"fyri Krist", "eftir Krist"},
			timezones:              map[string]string{"ACDT": "mið Avstralia summartíð", "ACST": "mið Avstralia vanlig tíð", "ACWDT": "miðvestur Avstralia summartíð", "ACWST": "miðvestur Avstralia vanlig tíð", "ADT": "Atlantic summartíð", "AEDT": "eystur Avstralia summartíð", "AEST": "eystur Avstralia vanlig tíð", "AKDT": "Alaska summartíð", "AKST": "Alaska vanlig tíð", "ARST": "Argentina summartíð", "ART": "Argentina vanlig tíð", "AST": "Atlantic vanlig tíð", "AWDT": "vestur Avstralia summartíð", "AWST": "vestur Avstralia vanlig tíð", "BOT": "Bolivia tíð", "BT": "Butan tíð", "CAT": "Miðafrika tíð", "CDT": "Central summartíð", "CHADT": "Chatham summartíð", "CHAST": "Chatham vanlig tíð", "CLST": "Kili summartíð", "CLT": "Kili vanlig tíð", "COST": "Kolombia summartíð", "COT": "Kolombia vanlig tíð", "CST": "Central vanlig tíð", "ChST": "Chamorro vanlig tíð", "EAT": "Eysturafrika tíð", "ECT": "Ekvador tíð", "EDT": "Eastern summartíð", "EST": "Eastern vanlig tíð", "GFT": "Franska Gujana tíð", "GMT": "Greenwich Mean tíð", "GST": "Gulf vanlig tíð", "GYT": "Gujana tíð", "HADT": "Hawaii-Aleutian summartíð", "HAST": "Hawaii-Aleutian vanlig tíð", "HAT": "Newfoundland summartíð", "HECU": "Cuba summartíð", "HEEG": "Eystur grønlendsk summartíð", "HENOMX": "Northwest Mexico summartíð", "HEOG": "Vestur grønlendsk summartíð", "HEPM": "St. Pierre & Miquelon summartíð", "HEPMX": "Mexican Pacific summartíð", "HKST": "Hong Kong summartíð", "HKT": "Hong Kong vanlig tíð", "HNCU": "Cuba vanlig tíð", "HNEG": "Eystur grønlendsk vanlig tíð", "HNNOMX": "Northwest Mexico vanlig tíð", "HNOG": "Vestur grønlendsk vanlig tíð", "HNPM": "St. Pierre & Miquelon vanlig tíð", "HNPMX": "Mexican Pacific vanlig tíð", "HNT": "Newfoundland vanlig tíð", "IST": "India tíð", "JDT": "Japan summartíð", "JST": "Japan vanlig tíð", "LHDT": "Lord Howe summartíð", "LHST": "Lord Howe vanlig tíð", "MDT": "Mountain summartíð", "MESZ": "Miðevropa summartíð", "MEZ": "Miðevropa vanlig tíð", "MST": "Mountain vanlig tíð", "MYT": "Malaisia tíð", "NZDT": "Nýsæland summartíð", "NZST": "Nýsæland vanlig tíð", "OESZ": "Eysturevropa summartíð", "OEZ": "Eysturevropa vanlig tíð", "PDT": "Pacific summartíð", "PST": "Pacific vanlig tíð", "SAST": "Suðurafrika vanlig tíð", "SGT": "Singapor tíð", "SRT": "Surinam tíð", "TMST": "Turkmenistan summartíð", "TMT": "Turkmenistan vanlig tíð", "UYST": "Uruguai summartíð", "UYT": "Uruguai vanlig tíð", "VET": "Venesuela tíð", "WARST": "Vestur Argentina summartíð", "WART": "Vestur Argentina vanlig tíð", "WAST": "Vesturafrika summartíð", "WAT": "Vesturafrika vanlig tíð", "WESZ": "Vesturevropa summartíð", "WEZ": "Vesturevropa vanlig tíð", "WIB": "Vestur Indonesia tíð", "WIT": "Eystur Indonesia tíð", "WITA": "Mið Indonesia tíð", "∅∅∅": "Brasilia summartíð"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fo_fo"] = func() locales.Translator {
		return &localen{

			locale:                 "fo_FO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "mai", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "apríl", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"sun.", "mán.", "týs.", "mik.", "hós.", "frí.", "ley."},
			daysNarrow:             []string{"S", "M", "T", "M", "H", "F", "L"},
			daysShort:              []string{"su.", "má.", "tý.", "mi.", "hó.", "fr.", "le."},
			daysWide:               []string{"sunnudagur", "mánadagur", "týsdagur", "mikudagur", "hósdagur", "fríggjadagur", "leygardagur"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"fKr", "eKr"},
			erasWide:               []string{"fyri Krist", "eftir Krist"},
			timezones:              map[string]string{"ACDT": "mið Avstralia summartíð", "ACST": "mið Avstralia vanlig tíð", "ACWDT": "miðvestur Avstralia summartíð", "ACWST": "miðvestur Avstralia vanlig tíð", "ADT": "Atlantic summartíð", "AEDT": "eystur Avstralia summartíð", "AEST": "eystur Avstralia vanlig tíð", "AKDT": "Alaska summartíð", "AKST": "Alaska vanlig tíð", "ARST": "Argentina summartíð", "ART": "Argentina vanlig tíð", "AST": "Atlantic vanlig tíð", "AWDT": "vestur Avstralia summartíð", "AWST": "vestur Avstralia vanlig tíð", "BOT": "Bolivia tíð", "BT": "Butan tíð", "CAT": "Miðafrika tíð", "CDT": "Central summartíð", "CHADT": "Chatham summartíð", "CHAST": "Chatham vanlig tíð", "CLST": "Kili summartíð", "CLT": "Kili vanlig tíð", "COST": "Kolombia summartíð", "COT": "Kolombia vanlig tíð", "CST": "Central vanlig tíð", "ChST": "Chamorro vanlig tíð", "EAT": "Eysturafrika tíð", "ECT": "Ekvador tíð", "EDT": "Eastern summartíð", "EST": "Eastern vanlig tíð", "GFT": "Franska Gujana tíð", "GMT": "Greenwich Mean tíð", "GST": "Gulf vanlig tíð", "GYT": "Gujana tíð", "HADT": "Hawaii-Aleutian summartíð", "HAST": "Hawaii-Aleutian vanlig tíð", "HAT": "Newfoundland summartíð", "HECU": "Cuba summartíð", "HEEG": "Eystur grønlendsk summartíð", "HENOMX": "Northwest Mexico summartíð", "HEOG": "Vestur grønlendsk summartíð", "HEPM": "St. Pierre & Miquelon summartíð", "HEPMX": "Mexican Pacific summartíð", "HKST": "Hong Kong summartíð", "HKT": "Hong Kong vanlig tíð", "HNCU": "Cuba vanlig tíð", "HNEG": "Eystur grønlendsk vanlig tíð", "HNNOMX": "Northwest Mexico vanlig tíð", "HNOG": "Vestur grønlendsk vanlig tíð", "HNPM": "St. Pierre & Miquelon vanlig tíð", "HNPMX": "Mexican Pacific vanlig tíð", "HNT": "Newfoundland vanlig tíð", "IST": "India tíð", "JDT": "Japan summartíð", "JST": "Japan vanlig tíð", "LHDT": "Lord Howe summartíð", "LHST": "Lord Howe vanlig tíð", "MDT": "Mountain summartíð", "MESZ": "Miðevropa summartíð", "MEZ": "Miðevropa vanlig tíð", "MST": "Mountain vanlig tíð", "MYT": "Malaisia tíð", "NZDT": "Nýsæland summartíð", "NZST": "Nýsæland vanlig tíð", "OESZ": "Eysturevropa summartíð", "OEZ": "Eysturevropa vanlig tíð", "PDT": "Pacific summartíð", "PST": "Pacific vanlig tíð", "SAST": "Suðurafrika vanlig tíð", "SGT": "Singapor tíð", "SRT": "Surinam tíð", "TMST": "Turkmenistan summartíð", "TMT": "Turkmenistan vanlig tíð", "UYST": "Uruguai summartíð", "UYT": "Uruguai vanlig tíð", "VET": "Venesuela tíð", "WARST": "Vestur Argentina summartíð", "WART": "Vestur Argentina vanlig tíð", "WAST": "Vesturafrika summartíð", "WAT": "Vesturafrika vanlig tíð", "WESZ": "Vesturevropa summartíð", "WEZ": "Vesturevropa vanlig tíð", "WIB": "Vestur Indonesia tíð", "WIT": "Eystur Indonesia tíð", "WITA": "Mið Indonesia tíð", "∅∅∅": "Brasilia summartíð"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr"] = func() locales.Translator {
		return &localen{

			locale:                 "fr",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$AR", "ATS", "$AU", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "FB", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "$BM", "$BN", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "$BZ", "$CA", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "$CL", "CNH", "CNX", "CNY", "$CO", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "£CY", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "$FJ", "£FK", "F", "£GB", "GEK", "GEL", "GHC", "GHS", "£GI", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "£IE", "£IL", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "₤IT", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "£LB", "LKR", "LRD", "lLS", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "fMA", "MCF", "MDC", "MDL", "MGA", "Fmg", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "£MT", "MUR", "MVP", "MVR", "MWK", "$MX", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "$NA", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$NZ", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "$RH", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "$SB", "SCR", "SDD", "SDG", "SDP", "SEK", "$SG", "SHP", "SIT", "SKK", "SLL", "SOS", "$SR", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "$TT", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$US", "USN", "USS", "UYI", "UYP", "$UY", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "$WS", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "DTS", "XEU", "XFO", "XFU", "CFA", "XPD", "FCFP", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_be"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_BE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_3604e9a5c6e536d483b9ceb5e0908b94,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_bf"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_BF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_bi"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_BI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "FBu", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_bj"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_BJ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_bl"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_BL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ca"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$\u00a0AU", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "Br", "BYR", "BZD", "$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "$\u00a0HK", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "CF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "C$", "NLG", "NOK", "NPR", "$\u00a0NZ", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "$\u00a0SG", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TL", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$\u00a0US", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZK", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " k",
			currencyNegativeSuffix: " k",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juill.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"avant l’ère chrétienne", "de l’ère chrétienne"},
			timezones:              map[string]string{"ACDT": "heure avancée du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure avancée du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure avancée de l’Atlantique", "AEDT": "heure avancée de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure avancée de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure avancée de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure avancée de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure d’Afrique centrale", "CDT": "heure avancée du Centre", "CHADT": "heure avancée des Îles Chatham", "CHAST": "heure normale des Îles Chatham", "CLST": "heure avancée du Chili", "CLT": "heure normale du Chili", "COST": "heure avancée de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du Centre", "ChST": "heure des Chamorro", "EAT": "heure d’Afrique orientale", "ECT": "heure de l’Équateur", "EDT": "heure avancée de l’Est", "EST": "heure normale de l’Est", "GFT": "heure de Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure avancée d’Hawaï-Aléoutiennes", "HAST": "heure normale d’Hawaï-Aléoutiennes", "HAT": "heure avancée de Terre-Neuve", "HECU": "heure avancée de Cuba", "HEEG": "heure avancée de l’Est du Groenland", "HENOMX": "heure avancée du Nord-Ouest du Mexique", "HEOG": "heure avancée de l’Ouest du Groenland", "HEPM": "heure avancée de Saint-Pierre-et-Miquelon", "HEPMX": "heure avancée du Pacifique mexicain", "HKST": "heure avancée de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure avancée du Japon", "JST": "heure normale du Japon", "LHDT": "heure avancée de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure avancée des Rocheuses", "MESZ": "heure avancée d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure avancée de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure avancée d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure avancée du Pacifique", "PST": "heure normale du Pacifique", "SAST": "heure normale d’Afrique du Sud", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure avancée du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure avancée de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure avancée de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure avancée d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure avancée d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure avancée de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7fb5040dc04d16c016c545b9369751e6,
			fnFmtCurrency:         fnFmtCurrency_e4c5476e381e260e6a551a757542b603,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_08768b0816a0563c1bdec56e3a7c7227,
			fnFmtTimeLong:         fnFmtTimeLong_890d4627ee8be38589c6184a42b5e739,
			fnFmtTimeMedium:       fnFmtTimeMedium_5ee755e4617064423cc20f8812f3f20e,
			fnFmtTimeShort:        fnFmtTimeShort_77616a5d5285f57c035b3599c557be3f,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_cd"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "FC", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_cf"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_cg"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_7190bcf20785aafef3dd5520b994587a,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ci"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"mat.", "soir"},
			periodsNarrow:          []string{"mat.", "soir"},
			periodsWide:            []string{"matin", "soir"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_dj"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_DJ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "Fdj", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_dz"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_DZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DA", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_fr"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_FR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ga"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_GA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_gf"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_GF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_gn"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_GN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "FG", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_gp"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_GP",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_gq"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_GQ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ht"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_HT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "G", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_km"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_KM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "CF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_lu"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_LU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "F", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été de Macao", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale de Macao", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ma"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fév.", "mar.", "avr.", "mai", "jui.", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_mc"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_mf"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été de Macao", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale de Macao", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_mg"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "Ar", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ml"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_ML",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_mq"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MQ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_mr"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "UM", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_mu"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_MU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "Rs", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_nc"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_NC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_ne"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_NE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_pf"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_PF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de l’Amazonie"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_pm"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_PM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_re"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_RE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_rw"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_RW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_sc"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_SC",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_sn"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_SN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_sy"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_SY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "LS", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été de Macao", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale de Macao", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été du Pérou"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_td"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_TD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_tg"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_TG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_tn"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_TN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "DT", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été des Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_vu"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_VU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VT", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_wf"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_WF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fr_yt"] = func() locales.Translator {
		return &localen{

			locale:                 "fr_YT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "janv.", "févr.", "mars", "avr.", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"},
			daysAbbreviated:        []string{"dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"di", "lu", "ma", "me", "je", "ve", "sa"},
			daysWide:               []string{"dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. J.-C.", "ap. J.-C."},
			erasNarrow:             []string{"av. J.-C.", "ap. J.-C."},
			erasWide:               []string{"avant Jésus-Christ", "après Jésus-Christ"},
			timezones:              map[string]string{"ACDT": "heure d’été du centre de l’Australie", "ACST": "heure normale du centre de l’Australie", "ACWDT": "heure d’été du centre-ouest de l’Australie", "ACWST": "heure normale du centre-ouest de l’Australie", "ADT": "heure d’été de l’Atlantique", "AEDT": "heure d’été de l’Est de l’Australie", "AEST": "heure normale de l’Est de l’Australie", "AKDT": "heure d’été de l’Alaska", "AKST": "heure normale de l’Alaska", "ARST": "heure d’été de l’Argentine", "ART": "heure normale d’Argentine", "AST": "heure normale de l’Atlantique", "AWDT": "heure d’été de l’Ouest de l’Australie", "AWST": "heure normale de l’Ouest de l’Australie", "BOT": "heure de Bolivie", "BT": "heure du Bhoutan", "CAT": "heure normale d’Afrique centrale", "CDT": "heure d’été du Centre", "CHADT": "heure d’été des îles Chatham", "CHAST": "heure normale des îles Chatham", "CLST": "heure d’été du Chili", "CLT": "heure normale du Chili", "COST": "heure d’été de Colombie", "COT": "heure normale de Colombie", "CST": "heure normale du centre nord-américain", "ChST": "heure des Chamorro", "EAT": "heure normale d’Afrique de l’Est", "ECT": "heure de l’Équateur", "EDT": "heure d’été de l’Est", "EST": "heure normale de l’Est nord-américain", "GFT": "heure de la Guyane française", "GMT": "heure moyenne de Greenwich", "GST": "heure du Golfe", "GYT": "heure du Guyana", "HADT": "heure d’été d’Hawaii - Aléoutiennes", "HAST": "heure normale d’Hawaii - Aléoutiennes", "HAT": "heure d’été de Terre-Neuve", "HECU": "heure d’été de Cuba", "HEEG": "heure d’été de l’Est du Groenland", "HENOMX": "heure d’été du Nord-Ouest du Mexique", "HEOG": "heure d’été de l’Ouest du Groenland", "HEPM": "heure d’été de Saint-Pierre-et-Miquelon", "HEPMX": "heure d’été du Pacifique mexicain", "HKST": "heure d’été de Hong Kong", "HKT": "heure normale de Hong Kong", "HNCU": "heure normale de Cuba", "HNEG": "heure normale de l’Est du Groenland", "HNNOMX": "heure normale du Nord-Ouest du Mexique", "HNOG": "heure normale de l’Ouest du Groenland", "HNPM": "heure normale de Saint-Pierre-et-Miquelon", "HNPMX": "heure normale du Pacifique mexicain", "HNT": "heure normale de Terre-Neuve", "IST": "heure de l’Inde", "JDT": "heure d’été du Japon", "JST": "heure normale du Japon", "LHDT": "heure d’été de Lord Howe", "LHST": "heure normale de Lord Howe", "MDT": "heure d’été des Rocheuses", "MESZ": "heure d’été d’Europe centrale", "MEZ": "heure normale d’Europe centrale", "MST": "heure normale des Rocheuses", "MYT": "heure de la Malaisie", "NZDT": "heure d’été de la Nouvelle-Zélande", "NZST": "heure normale de la Nouvelle-Zélande", "OESZ": "heure d’été d’Europe de l’Est", "OEZ": "heure normale d’Europe de l’Est", "PDT": "heure d’été du Pacifique", "PST": "heure normale du Pacifique nord-américain", "SAST": "heure normale d’Afrique méridionale", "SGT": "heure de Singapour", "SRT": "heure du Suriname", "TMST": "heure d’été du Turkménistan", "TMT": "heure normale du Turkménistan", "UYST": "heure d’été de l’Uruguay", "UYT": "heure normale de l’Uruguay", "VET": "heure du Venezuela", "WARST": "heure d’été de l’Ouest argentin", "WART": "heure normale de l’Ouest argentin", "WAST": "heure d’été d’Afrique de l’Ouest", "WAT": "heure normale d’Afrique de l’Ouest", "WESZ": "heure d’été d’Europe de l’Ouest", "WEZ": "heure normale d’Europe de l’Ouest", "WIB": "heure de l’Ouest indonésien", "WIT": "heure de l’Est indonésien", "WITA": "heure du Centre indonésien", "∅∅∅": "heure d’été de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_cf57cb3d1c6f3cc11c398367995ecf75,
			fnFmtCurrency:         fnFmtCurrency_781a083ba875d0db385fcba596c8fe2f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fur"] = func() locales.Translator {
		return &localen{

			locale:                 "fur",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Zen", "Fev", "Mar", "Avr", "Mai", "Jug", "Lui", "Avo", "Set", "Otu", "Nov", "Dic"},
			monthsNarrow:           []string{"", "Z", "F", "M", "A", "M", "J", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Zenâr", "Fevrâr", "Març", "Avrîl", "Mai", "Jugn", "Lui", "Avost", "Setembar", "Otubar", "Novembar", "Dicembar"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mie", "joi", "vin", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysWide:               []string{"domenie", "lunis", "martars", "miercus", "joibe", "vinars", "sabide"},
			periodsAbbreviated:     []string{"a.", "p."},
			periodsWide:            []string{"a.", "p."},
			erasAbbreviated:        []string{"pdC", "ddC"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Ore estive de Europe centrâl", "MEZ": "Ore standard de Europe centrâl", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Ore estive de Europe orientâl", "OEZ": "Ore standard de Europe orientâl", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Ore estive de Europe ocidentâl", "WEZ": "Ore standard de Europe ocidentâl", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_0987fcfde5397e392daea76f23ab1c65,
			fnFmtDateLong:         fnFmtDateLong_35ef415790b91229102369fdc2485b80,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fur_it"] = func() locales.Translator {
		return &localen{

			locale:                 "fur_IT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Zen", "Fev", "Mar", "Avr", "Mai", "Jug", "Lui", "Avo", "Set", "Otu", "Nov", "Dic"},
			monthsNarrow:           []string{"", "Z", "F", "M", "A", "M", "J", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Zenâr", "Fevrâr", "Març", "Avrîl", "Mai", "Jugn", "Lui", "Avost", "Setembar", "Otubar", "Novembar", "Dicembar"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mie", "joi", "vin", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysWide:               []string{"domenie", "lunis", "martars", "miercus", "joibe", "vinars", "sabide"},
			periodsAbbreviated:     []string{"a.", "p."},
			periodsWide:            []string{"a.", "p."},
			erasAbbreviated:        []string{"pdC", "ddC"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Ore estive de Europe centrâl", "MEZ": "Ore standard de Europe centrâl", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Ore estive de Europe orientâl", "OEZ": "Ore standard de Europe orientâl", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Ore estive de Europe ocidentâl", "WEZ": "Ore standard de Europe ocidentâl", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_0987fcfde5397e392daea76f23ab1c65,
			fnFmtDateLong:         fnFmtDateLong_35ef415790b91229102369fdc2485b80,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fy"] = func() locales.Translator {
		return &localen{

			locale:                 "fy",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "C$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJ$", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SI$", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mrt", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Jannewaris", "Febrewaris", "Maart", "April", "Maaie", "Juny", "July", "Augustus", "Septimber", "Oktober", "Novimber", "Desimber"},
			daysAbbreviated:        []string{"si", "mo", "ti", "wo", "to", "fr", "so"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"si", "mo", "ti", "wo", "to", "fr", "so"},
			daysWide:               []string{"snein", "moandei", "tiisdei", "woansdei", "tongersdei", "freed", "sneon"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "n.Kr."},
			erasNarrow:             []string{"f.K.", "n.K."},
			erasWide:               []string{"Foar Kristus", "nei Kristus"},
			timezones:              map[string]string{"ACDT": "Midden-Australyske simmertiid", "ACST": "Midden-Australyske standerttiid", "ACWDT": "Midden-Australyske westelijke simmertiid", "ACWST": "Midden-Australyske westelijke standerttiid", "ADT": "Atlantic-simmertiid", "AEDT": "East-Australyske simmertiid", "AEST": "East-Australyske standerttiid", "AKDT": "Alaska-simmertiid", "AKST": "Alaska-standerttiid", "ARST": "Argentynske simmertiid", "ART": "Argentynske standerttiid", "AST": "Atlantic-standerttiid", "AWDT": "West-Australyske simmertiid", "AWST": "West-Australyske standerttiid", "BOT": "Boliviaanske tiid", "BT": "Bhutaanske tiid", "CAT": "Sintraal-Afrikaanske tiid", "CDT": "Central-simmertiid", "CHADT": "Chatham simmertiid", "CHAST": "Chatham standerttiid", "CLST": "Sileenske simmertiid", "CLT": "Sileenske standerttiid", "COST": "Kolombiaanske simmertiid", "COT": "Kolombiaanske standerttiid", "CST": "Central-standerttiid", "ChST": "Chamorro-tiid", "EAT": "East-Afrikaanske tiid", "ECT": "Ecuadoraanske tiid", "EDT": "Eastern-simmertiid", "EST": "Eastern-standerttiid", "GFT": "Frâns-Guyaanske tiid", "GMT": "Greenwich Mean Time", "GST": "Golf standerttiid", "GYT": "Guyaanske tiid", "HADT": "Hawaii-Aleoetyske simmertiid", "HAST": "Hawaii-Aleoetyske standerttiid", "HAT": "Newfoundlânske-simmertiid", "HECU": "Kubaanske simmertiid", "HEEG": "East-Groenlânske simmertiid", "HENOMX": "HENOMX", "HEOG": "West-Groenlânske simmertiid", "HEPM": "Saint Pierre en Miquelon-simmertiid", "HEPMX": "HEPMX", "HKST": "Hongkongse simmertiid", "HKT": "Hongkongse standerttiid", "HNCU": "Kubaanske standerttiid", "HNEG": "East-Groenlânske standerttiid", "HNNOMX": "HNNOMX", "HNOG": "West-Groenlânske standerttiid", "HNPM": "Saint Pierre en Miquelon-standerttiid", "HNPMX": "HNPMX", "HNT": "Newfoundlânske-standerttiid", "IST": "Yndiaaske tiid", "JDT": "Japanske simmertiid", "JST": "Japanske standerttiid", "LHDT": "Lord Howe-eilânske simmertiid", "LHST": "Lord Howe-eilânske standerttiid", "MDT": "Mountain-simmertiid", "MESZ": "Midden-Europeeske simmertiid", "MEZ": "Midden-Europeeske standerttiid", "MST": "Mountain-standerttiid", "MYT": "Maleisyske tiid", "NZDT": "Nij-Seelânske simmertiid", "NZST": "Nij-Seelânske standerttiid", "OESZ": "East-Europeeske simmertiid", "OEZ": "East-Europeeske standerttiid", "PDT": "Pasifik-simmertiid", "PST": "Pasifik-standerttiid", "SAST": "Sûd-Afrikaanske tiid", "SGT": "Singaporese standerttiid", "SRT": "Surinaamske tiid", "TMST": "Turkmeense simmertiid", "TMT": "Turkmeense standerttiid", "UYST": "Uruguayaanske simmertiid", "UYT": "Uruguayaanske standerttiid", "VET": "Fenezolaanske tiid", "WARST": "West-Argentynske simmertiid", "WART": "West-Argentynske standerttiid", "WAST": "West-Afrikaanske simmertiid", "WAT": "West-Afrikaanske standerttiid", "WESZ": "West-Europeeske simmertiid", "WEZ": "West-Europeeske standerttiid", "WIB": "West-Yndonezyske tiid", "WIT": "East-Yndonezyske tiid", "WITA": "Sintraal-Yndonezyske tiid", "∅∅∅": "Brazyljaanske simmertiid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["fy_nl"] = func() locales.Translator {
		return &localen{

			locale:                 "fy_NL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mrt", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Jannewaris", "Febrewaris", "Maart", "April", "Maaie", "Juny", "July", "Augustus", "Septimber", "Oktober", "Novimber", "Desimber"},
			daysAbbreviated:        []string{"si", "mo", "ti", "wo", "to", "fr", "so"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"si", "mo", "ti", "wo", "to", "fr", "so"},
			daysWide:               []string{"snein", "moandei", "tiisdei", "woansdei", "tongersdei", "freed", "sneon"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"f.Kr.", "n.Kr."},
			erasNarrow:             []string{"f.K.", "n.K."},
			erasWide:               []string{"Foar Kristus", "nei Kristus"},
			timezones:              map[string]string{"ACDT": "Midden-Australyske simmertiid", "ACST": "Midden-Australyske standerttiid", "ACWDT": "Midden-Australyske westelijke simmertiid", "ACWST": "Midden-Australyske westelijke standerttiid", "ADT": "Atlantic-simmertiid", "AEDT": "East-Australyske simmertiid", "AEST": "East-Australyske standerttiid", "AKDT": "Alaska-simmertiid", "AKST": "Alaska-standerttiid", "ARST": "Argentynske simmertiid", "ART": "Argentynske standerttiid", "AST": "Atlantic-standerttiid", "AWDT": "West-Australyske simmertiid", "AWST": "West-Australyske standerttiid", "BOT": "Boliviaanske tiid", "BT": "Bhutaanske tiid", "CAT": "Sintraal-Afrikaanske tiid", "CDT": "Central-simmertiid", "CHADT": "Chatham simmertiid", "CHAST": "Chatham standerttiid", "CLST": "Sileenske simmertiid", "CLT": "Sileenske standerttiid", "COST": "Kolombiaanske simmertiid", "COT": "Kolombiaanske standerttiid", "CST": "Central-standerttiid", "ChST": "Chamorro-tiid", "EAT": "East-Afrikaanske tiid", "ECT": "Ecuadoraanske tiid", "EDT": "Eastern-simmertiid", "EST": "Eastern-standerttiid", "GFT": "Frâns-Guyaanske tiid", "GMT": "Greenwich Mean Time", "GST": "Golf standerttiid", "GYT": "Guyaanske tiid", "HADT": "Hawaii-Aleoetyske simmertiid", "HAST": "Hawaii-Aleoetyske standerttiid", "HAT": "Newfoundlânske-simmertiid", "HECU": "Kubaanske simmertiid", "HEEG": "East-Groenlânske simmertiid", "HENOMX": "HENOMX", "HEOG": "West-Groenlânske simmertiid", "HEPM": "Saint Pierre en Miquelon-simmertiid", "HEPMX": "HEPMX", "HKST": "Hongkongse simmertiid", "HKT": "Hongkongse standerttiid", "HNCU": "Kubaanske standerttiid", "HNEG": "East-Groenlânske standerttiid", "HNNOMX": "HNNOMX", "HNOG": "West-Groenlânske standerttiid", "HNPM": "Saint Pierre en Miquelon-standerttiid", "HNPMX": "HNPMX", "HNT": "Newfoundlânske-standerttiid", "IST": "Yndiaaske tiid", "JDT": "Japanske simmertiid", "JST": "Japanske standerttiid", "LHDT": "Lord Howe-eilânske simmertiid", "LHST": "Lord Howe-eilânske standerttiid", "MDT": "Macause simmertiid", "MESZ": "Midden-Europeeske simmertiid", "MEZ": "Midden-Europeeske standerttiid", "MST": "Macause standerttiid", "MYT": "Maleisyske tiid", "NZDT": "Nij-Seelânske simmertiid", "NZST": "Nij-Seelânske standerttiid", "OESZ": "East-Europeeske simmertiid", "OEZ": "East-Europeeske standerttiid", "PDT": "Pasifik-simmertiid", "PST": "Pasifik-standerttiid", "SAST": "Sûd-Afrikaanske tiid", "SGT": "Singaporese standerttiid", "SRT": "Surinaamske tiid", "TMST": "Turkmeense simmertiid", "TMT": "Turkmeense standerttiid", "UYST": "Uruguayaanske simmertiid", "UYT": "Uruguayaanske standerttiid", "VET": "Fenezolaanske tiid", "WARST": "West-Argentynske simmertiid", "WART": "West-Argentynske standerttiid", "WAST": "West-Afrikaanske simmertiid", "WAT": "West-Afrikaanske standerttiid", "WESZ": "West-Europeeske simmertiid", "WEZ": "West-Europeeske standerttiid", "WIB": "West-Yndonezyske tiid", "WIT": "East-Yndonezyske tiid", "WITA": "Sintraal-Yndonezyske tiid", "∅∅∅": "Peruaanske simmertiid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ga"] = func() locales.Translator {
		return &localen{

			locale:                 "ga",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 3, 4, 5, 6},
			group:                  ",",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "Kz", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "KM", "BAN", "$", "৳", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "$", "$", "Bs", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "$", "BTN", "BUK", "P", "BYB", "BYN", "BYR", "$", "$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "$", "CNH", "CNX", "¥", "$", "COU", "₡", "CSD", "CSK", "$", "$", "CVE", "CYP", "Kč", "DDM", "DEM", "DJF", "kr", "$", "DZD", "ECS", "ECV", "EEK", "E£", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "$", "£", "FRF", "£", "GEK", "₾", "GHC", "GHS", "£", "GMD", "FG", "GNS", "GQE", "GRD", "Q", "GWE", "GWP", "$", "$", "L", "HRD", "kn", "HTG", "Ft", "Rp", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "kr", "ITL", "$", "JOD", "¥", "KES", "KGS", "៛", "CF", "₩", "KRH", "KRO", "₩", "KWD", "$", "₸", "₭", "L£", "Rs", "$", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "Ar", "MGF", "MKD", "MKN", "MLF", "K", "₮", "MOP", "MRO", "MRU", "MTL", "MTP", "Rs", "MVP", "MVR", "MWK", "$", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "$", "₦", "NIC", "C$", "NLG", "kr", "Rs", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "Rs", "zł", "PLZ", "PTE", "₲", "QAR", "RHD", "ROL", "lei", "RSD", "₽", "RUR", "RF", "SAR", "$", "SCR", "SDD", "SDG", "SDP", "kr", "$", "£", "SIT", "SKK", "SLL", "SOS", "$", "SRG", "£", "STD", "Db", "SUR", "SVC", "£", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "T$", "TPE", "TRL", "₺", "$", "NT$", "TZS", "₴", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "$", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZK", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ean", "Feabh", "Márta", "Aib", "Beal", "Meith", "Iúil", "Lún", "MFómh", "DFómh", "Samh", "Noll"},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "B", "M", "I", "L", "M", "D", "S", "N"},
			monthsWide:             []string{"", "Eanáir", "Feabhra", "Márta", "Aibreán", "Bealtaine", "Meitheamh", "Iúil", "Lúnasa", "Meán Fómhair", "Deireadh Fómhair", "Samhain", "Nollaig"},
			daysAbbreviated:        []string{"Domh", "Luan", "Máirt", "Céad", "Déar", "Aoine", "Sath"},
			daysNarrow:             []string{"D", "L", "M", "C", "D", "A", "S"},
			daysShort:              []string{"Do", "Lu", "Má", "Cé", "Dé", "Ao", "Sa"},
			daysWide:               []string{"Dé Domhnaigh", "Dé Luain", "Dé Máirt", "Dé Céadaoin", "Déardaoin", "Dé hAoine", "Dé Sathairn"},
			periodsAbbreviated:     []string{"r.n.", "i.n."},
			periodsNarrow:          []string{"r.n.", "i.n."},
			periodsWide:            []string{"r.n.", "i.n."},
			erasAbbreviated:        []string{"RC", "AD"},
			erasNarrow:             []string{"RC", "AD"},
			erasWide:               []string{"Roimh Chríost", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Am Samhraidh Lár na hAstráile", "ACST": "Am Caighdeánach Lár na hAstráile", "ACWDT": "Am Samhraidh Mheániarthar na hAstráile", "ACWST": "Am Caighdeánach Mheániarthar na hAstráile", "ADT": "Am Samhraidh an Atlantaigh", "AEDT": "Am Samhraidh Oirthear na hAstráile", "AEST": "Am Caighdeánach Oirthear na hAstráile", "AKDT": "Am Samhraidh Alasca", "AKST": "Am Caighdeánach Alasca", "ARST": "Am Samhraidh na hAirgintíne", "ART": "Am Caighdeánach na hAirgintíne", "AST": "Am Caighdeánach an Atlantaigh", "AWDT": "Am Samhraidh Iarthar na hAstráile", "AWST": "Am Caighdeánach Iarthar na hAstráile", "BOT": "Am na Bolaive", "BT": "Am na Bútáine", "CAT": "Am Lár na hAfraice", "CDT": "Am Samhraidh Lárnach", "CHADT": "Am Samhraidh Chatham", "CHAST": "Am Caighdeánach Chatham", "CLST": "Am Samhraidh na Sile", "CLT": "Am Caighdeánach na Sile", "COST": "Am Samhraidh na Colóime", "COT": "Am Caighdeánach na Colóime", "CST": "Am Caighdeánach Lárnach", "ChST": "Am Caighdeánach Seamórach", "EAT": "Am Oirthear na hAfraice", "ECT": "Am Eacuadór", "EDT": "Am Samhraidh an Oirthir", "EST": "Am Caighdeánach an Oirthir", "GFT": "Am Ghuáin na Fraince", "GMT": "Meán-Am Greenwich", "GST": "Am Caighdeánach na Murascaille", "GYT": "Am na Guáine", "HADT": "Am Samhraidh Haváí-Ailiúit", "HAST": "Am Caighdeánach Haváí-Ailiúit", "HAT": "Am Samhraidh Thalamh an Éisc", "HECU": "Am Samhraidh Chúba", "HEEG": "Am Samhraidh Oirthear na Graonlainne", "HENOMX": "Am Samhraidh Iarthuaisceart Mheicsiceo", "HEOG": "Am Samhraidh Iarthar na Graonlainne", "HEPM": "Am Samhraidh Saint-Pierre-et-Miquelon", "HEPMX": "Am Samhraidh Meicsiceach an Aigéin Chiúin", "HKST": "Am Samhraidh Hong Cong", "HKT": "Am Caighdeánach Hong Cong", "HNCU": "Am Caighdeánach Chúba", "HNEG": "Am Caighdeánach Oirthear na Graonlainne", "HNNOMX": "Am Caighdeánach Iarthuaisceart Mheicsiceo", "HNOG": "Am Caighdeánach Iarthar na Graonlainne", "HNPM": "Am Caighdeánach Saint-Pierre-et-Miquelon", "HNPMX": "Am Caighdeánach Meicsiceach an Aigéin Chiúin", "HNT": "Am Caighdeánach Thalamh an Éisc", "IST": "Am Caighdeánach na hIndia", "JDT": "Am Samhraidh na Seapáine", "JST": "Am Caighdeánach na Seapáine", "LHDT": "Am Samhraidh Lord Howe", "LHST": "Am Caighdeánach Lord Howe", "MDT": "Am Samhraidh na Sléibhte", "MESZ": "Am Samhraidh Lár na hEorpa", "MEZ": "Am Caighdeánach Lár na hEorpa", "MST": "Am Caighdeánach na Sléibhte", "MYT": "Am na Malaeisia", "NZDT": "Am Samhraidh na Nua-Shéalainne", "NZST": "Am Caighdeánach na Nua-Shéalainne", "OESZ": "Am Samhraidh Oirthear na hEorpa", "OEZ": "Am Caighdeánach Oirthear na hEorpa", "PDT": "Am Samhraidh an Aigéin Chiúin", "PST": "Am Caighdeánach an Aigéin Chiúin", "SAST": "Am Caighdeánach na hAfraice Theas", "SGT": "Am Caighdeánach Shingeapór", "SRT": "Am Shuranam", "TMST": "Am Samhraidh na Tuircméanastáine", "TMT": "Am Caighdeánach na Tuircméanastáine", "UYST": "Am Samhraidh Uragua", "UYT": "Am Caighdeánach Uragua", "VET": "Am Veiniséala", "WARST": "Am Samhraidh Iarthar na hAirgintíne", "WART": "Am Caighdeánach Iarthar na hAirgintíne", "WAST": "Am Samhraidh Iarthar na hAfraice", "WAT": "Am Caighdeánach Iarthar na hAfraice", "WESZ": "Am Samhraidh Iarthar na hEorpa", "WEZ": "Am Caighdeánach Iarthar na hEorpa", "WIB": "Am Iarthar na hIndinéise", "WIT": "Am Oirthear na hIndinéise", "WITA": "Am Lár na hIndinéise", "∅∅∅": "Am Samhraidh Pheiriú"},

			fnCardinalPluralRule:  fnCardinalPluralRule_5424a4808e45b44daf4dd95a868534c9,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_20afe596297e5df40957d72da7798243,
			fnFmtCurrency:         fnFmtCurrency_ad2ba7087d8f6e468ec280be01359ba1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_ba2eea4558b51a6ab55725541be07d3f,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ga_gb"] = func() locales.Translator {
		return &localen{

			locale:                 "ga_GB",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 3, 4, 5, 6},
			group:                  ",",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ean", "Feabh", "Márta", "Aib", "Beal", "Meith", "Iúil", "Lún", "MFómh", "DFómh", "Samh", "Noll"},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "B", "M", "I", "L", "M", "D", "S", "N"},
			monthsWide:             []string{"", "Eanáir", "Feabhra", "Márta", "Aibreán", "Bealtaine", "Meitheamh", "Iúil", "Lúnasa", "Meán Fómhair", "Deireadh Fómhair", "Samhain", "Nollaig"},
			daysAbbreviated:        []string{"Domh", "Luan", "Máirt", "Céad", "Déar", "Aoine", "Sath"},
			daysNarrow:             []string{"D", "L", "M", "C", "D", "A", "S"},
			daysShort:              []string{"Do", "Lu", "Má", "Cé", "Dé", "Ao", "Sa"},
			daysWide:               []string{"Dé Domhnaigh", "Dé Luain", "Dé Máirt", "Dé Céadaoin", "Déardaoin", "Dé hAoine", "Dé Sathairn"},
			periodsAbbreviated:     []string{"r.n.", "i.n."},
			periodsNarrow:          []string{"r.n.", "i.n."},
			periodsWide:            []string{"r.n.", "i.n."},
			erasAbbreviated:        []string{"RC", "AD"},
			erasNarrow:             []string{"RC", "AD"},
			erasWide:               []string{"Roimh Chríost", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Am Samhraidh Lár na hAstráile", "ACST": "Am Caighdeánach Lár na hAstráile", "ACWDT": "Am Samhraidh Mheániarthar na hAstráile", "ACWST": "Am Caighdeánach Mheániarthar na hAstráile", "ADT": "Am Samhraidh an Atlantaigh", "AEDT": "Am Samhraidh Oirthear na hAstráile", "AEST": "Am Caighdeánach Oirthear na hAstráile", "AKDT": "Am Samhraidh Alasca", "AKST": "Am Caighdeánach Alasca", "ARST": "Am Samhraidh na hAirgintíne", "ART": "Am Caighdeánach na hAirgintíne", "AST": "Am Caighdeánach an Atlantaigh", "AWDT": "Am Samhraidh Iarthar na hAstráile", "AWST": "Am Caighdeánach Iarthar na hAstráile", "BOT": "Am na Bolaive", "BT": "Am na Bútáine", "CAT": "Am Lár na hAfraice", "CDT": "Am Samhraidh Lárnach", "CHADT": "Am Samhraidh Chatham", "CHAST": "Am Caighdeánach Chatham", "CLST": "Am Samhraidh na Sile", "CLT": "Am Caighdeánach na Sile", "COST": "Am Samhraidh na Colóime", "COT": "Am Caighdeánach na Colóime", "CST": "Am Caighdeánach Lárnach", "ChST": "Am Caighdeánach Seamórach", "EAT": "Am Oirthear na hAfraice", "ECT": "Am Eacuadór", "EDT": "Am Samhraidh an Oirthir", "EST": "Am Caighdeánach an Oirthir", "GFT": "Am Ghuáin na Fraince", "GMT": "Meán-Am Greenwich", "GST": "Am Caighdeánach na Murascaille", "GYT": "Am na Guáine", "HADT": "Am Samhraidh Haváí-Ailiúit", "HAST": "Am Caighdeánach Haváí-Ailiúit", "HAT": "Am Samhraidh Thalamh an Éisc", "HECU": "Am Samhraidh Chúba", "HEEG": "Am Samhraidh Oirthear na Graonlainne", "HENOMX": "Am Samhraidh Iarthuaisceart Mheicsiceo", "HEOG": "Am Samhraidh Iarthar na Graonlainne", "HEPM": "Am Samhraidh Saint-Pierre-et-Miquelon", "HEPMX": "Am Samhraidh Meicsiceach an Aigéin Chiúin", "HKST": "Am Samhraidh Hong Cong", "HKT": "Am Caighdeánach Hong Cong", "HNCU": "Am Caighdeánach Chúba", "HNEG": "Am Caighdeánach Oirthear na Graonlainne", "HNNOMX": "Am Caighdeánach Iarthuaisceart Mheicsiceo", "HNOG": "Am Caighdeánach Iarthar na Graonlainne", "HNPM": "Am Caighdeánach Saint-Pierre-et-Miquelon", "HNPMX": "Am Caighdeánach Meicsiceach an Aigéin Chiúin", "HNT": "Am Caighdeánach Thalamh an Éisc", "IST": "Am Caighdeánach na hIndia", "JDT": "Am Samhraidh na Seapáine", "JST": "Am Caighdeánach na Seapáine", "LHDT": "Am Samhraidh Lord Howe", "LHST": "Am Caighdeánach Lord Howe", "MDT": "Am Samhraidh na Sléibhte", "MESZ": "Am Samhraidh Lár na hEorpa", "MEZ": "Am Caighdeánach Lár na hEorpa", "MST": "Am Caighdeánach na Sléibhte", "MYT": "Am na Malaeisia", "NZDT": "Am Samhraidh na Nua-Shéalainne", "NZST": "Am Caighdeánach na Nua-Shéalainne", "OESZ": "Am Samhraidh Oirthear na hEorpa", "OEZ": "Am Caighdeánach Oirthear na hEorpa", "PDT": "Am Samhraidh an Aigéin Chiúin", "PST": "Am Caighdeánach an Aigéin Chiúin", "SAST": "Am Caighdeánach na hAfraice Theas", "SGT": "Am Caighdeánach Shingeapór", "SRT": "Am Shuranam", "TMST": "Am Samhraidh na Tuircméanastáine", "TMT": "Am Caighdeánach na Tuircméanastáine", "UYST": "Am Samhraidh Uragua", "UYT": "Am Caighdeánach Uragua", "VET": "Am Veiniséala", "WARST": "Am Samhraidh Iarthar na hAirgintíne", "WART": "Am Caighdeánach Iarthar na hAirgintíne", "WAST": "Am Samhraidh Iarthar na hAfraice", "WAT": "Am Caighdeánach Iarthar na hAfraice", "WESZ": "Am Samhraidh Iarthar na hEorpa", "WEZ": "Am Caighdeánach Iarthar na hEorpa", "WIB": "Am Iarthar na hIndinéise", "WIT": "Am Oirthear na hIndinéise", "WITA": "Am Lár na hIndinéise", "∅∅∅": "Am Samhraidh Bhrasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_5424a4808e45b44daf4dd95a868534c9,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_20afe596297e5df40957d72da7798243,
			fnFmtCurrency:         fnFmtCurrency_ad2ba7087d8f6e468ec280be01359ba1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_ba2eea4558b51a6ab55725541be07d3f,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ga_ie"] = func() locales.Translator {
		return &localen{

			locale:                 "ga_IE",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 3, 4, 5, 6},
			group:                  ",",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ean", "Feabh", "Márta", "Aib", "Beal", "Meith", "Iúil", "Lún", "MFómh", "DFómh", "Samh", "Noll"},
			monthsNarrow:           []string{"", "E", "F", "M", "A", "B", "M", "I", "L", "M", "D", "S", "N"},
			monthsWide:             []string{"", "Eanáir", "Feabhra", "Márta", "Aibreán", "Bealtaine", "Meitheamh", "Iúil", "Lúnasa", "Meán Fómhair", "Deireadh Fómhair", "Samhain", "Nollaig"},
			daysAbbreviated:        []string{"Domh", "Luan", "Máirt", "Céad", "Déar", "Aoine", "Sath"},
			daysNarrow:             []string{"D", "L", "M", "C", "D", "A", "S"},
			daysShort:              []string{"Do", "Lu", "Má", "Cé", "Dé", "Ao", "Sa"},
			daysWide:               []string{"Dé Domhnaigh", "Dé Luain", "Dé Máirt", "Dé Céadaoin", "Déardaoin", "Dé hAoine", "Dé Sathairn"},
			periodsAbbreviated:     []string{"r.n.", "i.n."},
			periodsNarrow:          []string{"r.n.", "i.n."},
			periodsWide:            []string{"r.n.", "i.n."},
			erasAbbreviated:        []string{"RC", "AD"},
			erasNarrow:             []string{"RC", "AD"},
			erasWide:               []string{"Roimh Chríost", "Anno Domini"},
			timezones:              map[string]string{"ACDT": "Am Samhraidh Lár na hAstráile", "ACST": "Am Caighdeánach Lár na hAstráile", "ACWDT": "Am Samhraidh Mheániarthar na hAstráile", "ACWST": "Am Caighdeánach Mheániarthar na hAstráile", "ADT": "Am Samhraidh an Atlantaigh", "AEDT": "Am Samhraidh Oirthear na hAstráile", "AEST": "Am Caighdeánach Oirthear na hAstráile", "AKDT": "Am Samhraidh Alasca", "AKST": "Am Caighdeánach Alasca", "ARST": "Am Samhraidh na hAirgintíne", "ART": "Am Caighdeánach na hAirgintíne", "AST": "Am Caighdeánach an Atlantaigh", "AWDT": "Am Samhraidh Iarthar na hAstráile", "AWST": "Am Caighdeánach Iarthar na hAstráile", "BOT": "Am na Bolaive", "BT": "Am na Bútáine", "CAT": "Am Lár na hAfraice", "CDT": "Am Samhraidh Lárnach", "CHADT": "Am Samhraidh Chatham", "CHAST": "Am Caighdeánach Chatham", "CLST": "Am Samhraidh na Sile", "CLT": "Am Caighdeánach na Sile", "COST": "Am Samhraidh na Colóime", "COT": "Am Caighdeánach na Colóime", "CST": "Am Caighdeánach Lárnach", "ChST": "Am Caighdeánach Seamórach", "EAT": "Am Oirthear na hAfraice", "ECT": "Am Eacuadór", "EDT": "Am Samhraidh an Oirthir", "EST": "Am Caighdeánach an Oirthir", "GFT": "Am Ghuáin na Fraince", "GMT": "Meán-Am Greenwich", "GST": "Am Caighdeánach na Murascaille", "GYT": "Am na Guáine", "HADT": "Am Samhraidh Haváí-Ailiúit", "HAST": "Am Caighdeánach Haváí-Ailiúit", "HAT": "Am Samhraidh Thalamh an Éisc", "HECU": "Am Samhraidh Chúba", "HEEG": "Am Samhraidh Oirthear na Graonlainne", "HENOMX": "Am Samhraidh Iarthuaisceart Mheicsiceo", "HEOG": "Am Samhraidh Iarthar na Graonlainne", "HEPM": "Am Samhraidh Saint-Pierre-et-Miquelon", "HEPMX": "Am Samhraidh Meicsiceach an Aigéin Chiúin", "HKST": "Am Samhraidh Hong Cong", "HKT": "Am Caighdeánach Hong Cong", "HNCU": "Am Caighdeánach Chúba", "HNEG": "Am Caighdeánach Oirthear na Graonlainne", "HNNOMX": "Am Caighdeánach Iarthuaisceart Mheicsiceo", "HNOG": "Am Caighdeánach Iarthar na Graonlainne", "HNPM": "Am Caighdeánach Saint-Pierre-et-Miquelon", "HNPMX": "Am Caighdeánach Meicsiceach an Aigéin Chiúin", "HNT": "Am Caighdeánach Thalamh an Éisc", "IST": "Am Caighdeánach na hIndia", "JDT": "Am Samhraidh na Seapáine", "JST": "Am Caighdeánach na Seapáine", "LHDT": "Am Samhraidh Lord Howe", "LHST": "Am Caighdeánach Lord Howe", "MDT": "Am Samhraidh na Sléibhte", "MESZ": "Am Samhraidh Lár na hEorpa", "MEZ": "Am Caighdeánach Lár na hEorpa", "MST": "Am Caighdeánach na Sléibhte", "MYT": "Am na Malaeisia", "NZDT": "Am Samhraidh na Nua-Shéalainne", "NZST": "Am Caighdeánach na Nua-Shéalainne", "OESZ": "Am Samhraidh Oirthear na hEorpa", "OEZ": "Am Caighdeánach Oirthear na hEorpa", "PDT": "Am Samhraidh an Aigéin Chiúin", "PST": "Am Caighdeánach an Aigéin Chiúin", "SAST": "Am Caighdeánach na hAfraice Theas", "SGT": "Am Caighdeánach Shingeapór", "SRT": "Am Shuranam", "TMST": "Am Samhraidh na Tuircméanastáine", "TMT": "Am Caighdeánach na Tuircméanastáine", "UYST": "Am Samhraidh Uragua", "UYT": "Am Caighdeánach Uragua", "VET": "Am Veiniséala", "WARST": "Am Samhraidh Iarthar na hAirgintíne", "WART": "Am Caighdeánach Iarthar na hAirgintíne", "WAST": "Am Samhraidh Iarthar na hAfraice", "WAT": "Am Caighdeánach Iarthar na hAfraice", "WESZ": "Am Samhraidh Iarthar na hEorpa", "WEZ": "Am Caighdeánach Iarthar na hEorpa", "WIB": "Am Iarthar na hIndinéise", "WIT": "Am Oirthear na hIndinéise", "WITA": "Am Lár na hIndinéise", "∅∅∅": "Am Samhraidh Bhrasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_5424a4808e45b44daf4dd95a868534c9,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_20afe596297e5df40957d72da7798243,
			fnFmtCurrency:         fnFmtCurrency_ad2ba7087d8f6e468ec280be01359ba1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_ba2eea4558b51a6ab55725541be07d3f,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gd"] = func() locales.Translator {
		return &localen{

			locale:                 "gd",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Faoi", "Gearr", "Màrt", "Gibl", "Cèit", "Ògmh", "Iuch", "Lùna", "Sult", "Dàmh", "Samh", "Dùbh"},
			monthsNarrow:           []string{"", "F", "G", "M", "G", "C", "Ò", "I", "L", "S", "D", "S", "D"},
			monthsWide:             []string{"", "dhen Fhaoilleach", "dhen Ghearran", "dhen Mhàrt", "dhen Ghiblean", "dhen Chèitean", "dhen Ògmhios", "dhen Iuchar", "dhen Lùnastal", "dhen t-Sultain", "dhen Dàmhair", "dhen t-Samhain", "dhen Dùbhlachd"},
			daysAbbreviated:        []string{"DiD", "DiL", "DiM", "DiC", "Dia", "Dih", "DiS"},
			daysNarrow:             []string{"D", "L", "M", "C", "A", "H", "S"},
			daysShort:              []string{"Dò", "Lu", "Mà", "Ci", "Da", "hA", "Sa"},
			daysWide:               []string{"DiDòmhnaich", "DiLuain", "DiMàirt", "DiCiadain", "DiarDaoin", "DihAoine", "DiSathairne"},
			periodsAbbreviated:     []string{"m", "f"},
			periodsNarrow:          []string{"m", "f"},
			periodsWide:            []string{"m", "f"},
			erasAbbreviated:        []string{"RC", "AD"},
			erasNarrow:             []string{"R", "A"},
			erasWide:               []string{"Ro Chrìosta", "An dèidh Chrìosta"},
			timezones:              map[string]string{"ACDT": "Tìde samhraidh Meadhan Astràilia", "ACST": "Bun-àm Meadhan Astràilia", "ACWDT": "Tìde samhraidh Meadhan Astràilia an Iar", "ACWST": "Bun-àm Meadhan Astràilia an Iar", "ADT": "Tìde samhraidh a’ Chuain Siar", "AEDT": "Tìde samhraidh Astràilia an Ear", "AEST": "Bun-àm Astràilia an Ear", "AKDT": "Tìde samhraidh Alaska", "AKST": "Bun-àm Alaska", "ARST": "Tìde samhraidh na h-Argantaine", "ART": "Bun-àm na h-Argantaine", "AST": "Bun-àm a’ Chuain Siar", "AWDT": "Tìde samhraidh Astràilia an Iar", "AWST": "Bun-àm Astràilia an Iar", "BOT": "Àm Boilibhia", "BT": "Àm Butàin", "CAT": "Àm Meadhan Afraga", "CDT": "Tìde samhraidh Meadhan Aimeireaga a Tuath", "CHADT": "Tìde samhraidh Chatham", "CHAST": "Bun-àm Chatham", "CLST": "Tìde samhraidh na Sile", "CLT": "Bun-àm na Sile", "COST": "Tìde samhraidh Coloimbia", "COT": "Bun-àm Coloimbia", "CST": "Bun-àm Meadhan Aimeireaga a Tuath", "ChST": "Àm Chamorro", "EAT": "Àm Afraga an Ear", "ECT": "Àm Eacuadoir", "EDT": "Tìde samhraidh Aimeireaga a Tuath an Ear", "EST": "Bun-àm Aimeireaga a Tuath an Ear", "GFT": "Àm Guidheàna na Frainge", "GMT": "Greenwich Mean Time", "GST": "Àm a’ Chamais", "GYT": "Àm Guidheàna", "HADT": "Tìde Samhraidh nan Eileanan Hawai’i ’s Aleutach", "HAST": "Bun-àm nan Eileanan Hawai’i ’s Aleutach", "HAT": "Tìde samhraidh Talamh an Èisg", "HECU": "Tìde samhraidh Cùba", "HEEG": "Tìde samhraidh na Graonlainn an Ear", "HENOMX": "Tìde samhraidh Mheagsago an Iar-thuath", "HEOG": "Tìde samhraidh na Graonlainn an Iar", "HEPM": "Tìde Samhraidh Saint Pierre agus Miquelon", "HEPMX": "Tìde samhraidh a’ Chuain Sèimh Mheagsago", "HKST": "Tìde samhraidh Hong Kong", "HKT": "Bun-àm Hong Kong", "HNCU": "Bun-àm Cùba", "HNEG": "Bun-àm na Graonlainn an Ear", "HNNOMX": "Bun-àm Mheagsago an Iar-thuath", "HNOG": "Bun-àm na Graonlainn an Iar", "HNPM": "Bun-àm Saint Pierre agus Miquelon", "HNPMX": "Bun-àm a’ Chuain Sèimh Mheagsago", "HNT": "Bun-àm Talamh an Èisg", "IST": "Àm nan Innseachan", "JDT": "Tìde samhraidh na Seapaine", "JST": "Bun-àm na Seapaine", "LHDT": "Tìde samhraidh Lord Howe", "LHST": "Bun-àm Lord Howe", "MDT": "Tìde samhraidh Monadh Aimeireaga a Tuath", "MESZ": "Tìde samhraidh Meadhan na Roinn-Eòrpa", "MEZ": "Bun-àm Meadhan na Roinn-Eòrpa", "MST": "Bun-àm Monadh Aimeireaga a Tuath", "MYT": "Àm Mhalaidhsea", "NZDT": "Tìde samhraidh Shealainn Nuaidh", "NZST": "Bun-àm Shealainn Nuaidh", "OESZ": "Tìde samhraidh na Roinn-Eòrpa an Ear", "OEZ": "Bun-àm na Roinn-Eòrpa an Ear", "PDT": "Tìde samhraidh a’ Chuain Sèimh", "PST": "Bun-àm a’ Chuain Sèimh", "SAST": "Àm Afraga a Deas", "SGT": "Àm Singeapòr", "SRT": "Àm Suranaim", "TMST": "Tìde samhraidh Turcmanastàin", "TMT": "Bun-àm Turcmanastàin", "UYST": "Tìde samhraidh Uruguaidh", "UYT": "Bun-àm Uruguaidh", "VET": "Àm na Bheiniseala", "WARST": "Tìde samhraidh na h-Argantaine Siaraich", "WART": "Bun-àm na h-Argantaine Siaraich", "WAST": "Tìde Samhraidh Afraga an Iar", "WAT": "Bun-àm Afraga an Iar", "WESZ": "Tìde samhraidh na Roinn-Eòrpa an Iar", "WEZ": "Bun-àm na Roinn-Eòrpa an Iar", "WIB": "Àm nan Innd-Innse an Iar", "WIT": "Àm nan Innd-Innse an Ear", "WITA": "Àm Meadhan nan Innd-Innse", "∅∅∅": "Tìde samhraidh Amasoin"},

			fnCardinalPluralRule:  fnCardinalPluralRule_5232ad4c13ce5aa918e9c5bcfff748c4,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_127dcd8df9ddb869d241194c5acabeea,
			fnFmtDateLong:         fnFmtDateLong_1b477f18b273c0ada89ad664b7a114af,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_f828d7a9eaca874285f7573c8823d785,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gd_gb"] = func() locales.Translator {
		return &localen{

			locale:                 "gd_GB",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Faoi", "Gearr", "Màrt", "Gibl", "Cèit", "Ògmh", "Iuch", "Lùna", "Sult", "Dàmh", "Samh", "Dùbh"},
			monthsNarrow:           []string{"", "F", "G", "M", "G", "C", "Ò", "I", "L", "S", "D", "S", "D"},
			monthsWide:             []string{"", "dhen Fhaoilleach", "dhen Ghearran", "dhen Mhàrt", "dhen Ghiblean", "dhen Chèitean", "dhen Ògmhios", "dhen Iuchar", "dhen Lùnastal", "dhen t-Sultain", "dhen Dàmhair", "dhen t-Samhain", "dhen Dùbhlachd"},
			daysAbbreviated:        []string{"DiD", "DiL", "DiM", "DiC", "Dia", "Dih", "DiS"},
			daysNarrow:             []string{"D", "L", "M", "C", "A", "H", "S"},
			daysShort:              []string{"Dò", "Lu", "Mà", "Ci", "Da", "hA", "Sa"},
			daysWide:               []string{"DiDòmhnaich", "DiLuain", "DiMàirt", "DiCiadain", "DiarDaoin", "DihAoine", "DiSathairne"},
			periodsAbbreviated:     []string{"m", "f"},
			periodsNarrow:          []string{"m", "f"},
			periodsWide:            []string{"m", "f"},
			erasAbbreviated:        []string{"RC", "AD"},
			erasNarrow:             []string{"R", "A"},
			erasWide:               []string{"Ro Chrìosta", "An dèidh Chrìosta"},
			timezones:              map[string]string{"ACDT": "Tìde samhraidh Meadhan Astràilia", "ACST": "Bun-àm Meadhan Astràilia", "ACWDT": "Tìde samhraidh Meadhan Astràilia an Iar", "ACWST": "Bun-àm Meadhan Astràilia an Iar", "ADT": "Tìde samhraidh a’ Chuain Siar", "AEDT": "Tìde samhraidh Astràilia an Ear", "AEST": "Bun-àm Astràilia an Ear", "AKDT": "Tìde samhraidh Alaska", "AKST": "Bun-àm Alaska", "ARST": "Tìde samhraidh na h-Argantaine", "ART": "Bun-àm na h-Argantaine", "AST": "Bun-àm a’ Chuain Siar", "AWDT": "Tìde samhraidh Astràilia an Iar", "AWST": "Bun-àm Astràilia an Iar", "BOT": "Àm Boilibhia", "BT": "Àm Butàin", "CAT": "Àm Meadhan Afraga", "CDT": "Tìde samhraidh Meadhan Aimeireaga a Tuath", "CHADT": "Tìde samhraidh Chatham", "CHAST": "Bun-àm Chatham", "CLST": "Tìde samhraidh na Sile", "CLT": "Bun-àm na Sile", "COST": "Tìde samhraidh Coloimbia", "COT": "Bun-àm Coloimbia", "CST": "Bun-àm Meadhan Aimeireaga a Tuath", "ChST": "Àm Chamorro", "EAT": "Àm Afraga an Ear", "ECT": "Àm Eacuadoir", "EDT": "Tìde samhraidh Aimeireaga a Tuath an Ear", "EST": "Bun-àm Aimeireaga a Tuath an Ear", "GFT": "Àm Guidheàna na Frainge", "GMT": "Greenwich Mean Time", "GST": "Àm a’ Chamais", "GYT": "Àm Guidheàna", "HADT": "Tìde Samhraidh nan Eileanan Hawai’i ’s Aleutach", "HAST": "Bun-àm nan Eileanan Hawai’i ’s Aleutach", "HAT": "Tìde samhraidh Talamh an Èisg", "HECU": "Tìde samhraidh Cùba", "HEEG": "Tìde samhraidh na Graonlainn an Ear", "HENOMX": "Tìde samhraidh Mheagsago an Iar-thuath", "HEOG": "Tìde samhraidh na Graonlainn an Iar", "HEPM": "Tìde Samhraidh Saint Pierre agus Miquelon", "HEPMX": "Tìde samhraidh a’ Chuain Sèimh Mheagsago", "HKST": "Tìde samhraidh Hong Kong", "HKT": "Bun-àm Hong Kong", "HNCU": "Bun-àm Cùba", "HNEG": "Bun-àm na Graonlainn an Ear", "HNNOMX": "Bun-àm Mheagsago an Iar-thuath", "HNOG": "Bun-àm na Graonlainn an Iar", "HNPM": "Bun-àm Saint Pierre agus Miquelon", "HNPMX": "Bun-àm a’ Chuain Sèimh Mheagsago", "HNT": "Bun-àm Talamh an Èisg", "IST": "Àm nan Innseachan", "JDT": "Tìde samhraidh na Seapaine", "JST": "Bun-àm na Seapaine", "LHDT": "Tìde samhraidh Lord Howe", "LHST": "Bun-àm Lord Howe", "MDT": "Tìde samhraidh Monadh Aimeireaga a Tuath", "MESZ": "Tìde samhraidh Meadhan na Roinn-Eòrpa", "MEZ": "Bun-àm Meadhan na Roinn-Eòrpa", "MST": "Bun-àm Monadh Aimeireaga a Tuath", "MYT": "Àm Mhalaidhsea", "NZDT": "Tìde samhraidh Shealainn Nuaidh", "NZST": "Bun-àm Shealainn Nuaidh", "OESZ": "Tìde samhraidh na Roinn-Eòrpa an Ear", "OEZ": "Bun-àm na Roinn-Eòrpa an Ear", "PDT": "Tìde samhraidh a’ Chuain Sèimh", "PST": "Bun-àm a’ Chuain Sèimh", "SAST": "Àm Afraga a Deas", "SGT": "Àm Singeapòr", "SRT": "Àm Suranaim", "TMST": "Tìde samhraidh Turcmanastàin", "TMT": "Bun-àm Turcmanastàin", "UYST": "Tìde samhraidh Uruguaidh", "UYT": "Bun-àm Uruguaidh", "VET": "Àm na Bheiniseala", "WARST": "Tìde samhraidh na h-Argantaine Siaraich", "WART": "Bun-àm na h-Argantaine Siaraich", "WAST": "Tìde Samhraidh Afraga an Iar", "WAT": "Bun-àm Afraga an Iar", "WESZ": "Tìde samhraidh na Roinn-Eòrpa an Iar", "WEZ": "Bun-àm na Roinn-Eòrpa an Iar", "WIB": "Àm nan Innd-Innse an Iar", "WIT": "Àm nan Innd-Innse an Ear", "WITA": "Àm Meadhan nan Innd-Innse", "∅∅∅": "Tìde samhraidh Bhrasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_5232ad4c13ce5aa918e9c5bcfff748c4,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_127dcd8df9ddb869d241194c5acabeea,
			fnFmtDateLong:         fnFmtDateLong_1b477f18b273c0ada89ad664b7a114af,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_f828d7a9eaca874285f7573c8823d785,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gl"] = func() locales.Translator {
		return &localen{

			locale:                 "gl",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "₧", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "$MX", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "xan.", "feb.", "mar.", "abr.", "maio", "xuño", "xul.", "ago.", "set.", "out.", "nov.", "dec."},
			monthsNarrow:           []string{"", "x.", "f.", "m.", "a.", "m.", "x.", "x.", "a.", "s.", "o.", "n.", "d."},
			monthsWide:             []string{"", "xaneiro", "febreiro", "marzo", "abril", "maio", "xuño", "xullo", "agosto", "setembro", "outubro", "novembro", "decembro"},
			daysAbbreviated:        []string{"dom.", "luns", "mar.", "mér.", "xov.", "ven.", "sáb."},
			daysNarrow:             []string{"d.", "l.", "m.", "m.", "x.", "v.", "s."},
			daysShort:              []string{"do.", "lu.", "ma.", "mé.", "xo.", "ve.", "sá."},
			daysWide:               []string{"domingo", "luns", "martes", "mércores", "xoves", "venres", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "despois de Cristo"},
			timezones:              map[string]string{"ACDT": "Horario de verán de Australia Central", "ACST": "Horario estándar de Australia Central", "ACWDT": "Horario de verán de Australia Occidental Central", "ACWST": "Horario estándar de Australia Occidental Central", "ADT": "Horario de verán do Atlántico", "AEDT": "Horario de verán de Australia Oriental", "AEST": "Horario estándar de Australia Oriental", "AKDT": "Horario de verán de Alasca", "AKST": "Horario estándar de Alasca", "ARST": "Horario de verán da Arxentina", "ART": "Horario estándar da Arxentina", "AST": "Horario estándar do Atlántico", "AWDT": "Horario de verán de Australia Occidental", "AWST": "Horario estándar de Australia Occidental", "BOT": "Horario de Bolivia", "BT": "Horario de Bután", "CAT": "Horario de África Central", "CDT": "Horario de verán central, Norteamérica", "CHADT": "Horario de verán de Chatham", "CHAST": "Horario estándar de Chatham", "CLST": "Horario de verán de Chile", "CLT": "Horario estándar de Chile", "COST": "Horario de verán de Colombia", "COT": "Horario estándar de Colombia", "CST": "Horario estándar central, Norteamérica", "ChST": "Horario estándar chamorro", "EAT": "Horario de África Oriental", "ECT": "Horario de Ecuador", "EDT": "Horario de verán do leste, América do Norte", "EST": "Horario estándar do leste, América do Norte", "GFT": "Horario da Güiana Francesa", "GMT": "Horario do meridiano de Greenwich", "GST": "Horario do Golfo", "GYT": "Horario da Güiana", "HADT": "Horario de verán de Hawai-illas Aleutianas", "HAST": "Horario estándar de Hawai-illas Aleutianas", "HAT": "Horario de verán de Terra Nova", "HECU": "Horario de verán de Cuba", "HEEG": "Horario de verán de Groenlandia Oriental", "HENOMX": "Horario de verán do noroeste de México", "HEOG": "Horario de verán de Groenlandia Occidental", "HEPM": "Horario de verán de Saint Pierre et Miquelon", "HEPMX": "Horario de verán do Pacífico mexicano", "HKST": "Horario de verán de Hong Kong", "HKT": "Horario estándar de Hong Kong", "HNCU": "Horario estándar de Cuba", "HNEG": "Horario estándar de Groenlandia Oriental", "HNNOMX": "Horario estándar do noroeste de México", "HNOG": "Horario estándar de Groenlandia Occidental", "HNPM": "Horario estándar de Saint Pierre et Miquelon", "HNPMX": "Horario estándar do Pacífico mexicano", "HNT": "Horario estándar de Terra Nova", "IST": "Horario da India", "JDT": "Horario de verán do Xapón", "JST": "Horario estándar do Xapón", "LHDT": "Horario de verán de Lord Howe", "LHST": "Horario estándar de Lord Howe", "MDT": "Horario de verán da montaña, América do Norte", "MESZ": "Horario de verán de Europa Central", "MEZ": "Horario estándar de Europa Central", "MST": "Horario estándar da montaña, América do Norte", "MYT": "Horario de Malaisia", "NZDT": "Horario de verán de Nova Zelandia", "NZST": "Horario estándar de Nova Zelandia", "OESZ": "Horario de verán de Europa Oriental", "OEZ": "Horario estándar de Europa Oriental", "PDT": "Horario de verán do Pacífico, América do Norte", "PST": "Horario estándar do Pacífico, América do Norte", "SAST": "Horario de África Meridional", "SGT": "Horario de Singapur", "SRT": "Horario de Suriname", "TMST": "Horario de verán de Turkmenistán", "TMT": "Horario estándar de Turkmenistán", "UYST": "Horario de verán do Uruguai", "UYT": "Horario estándar do Uruguai", "VET": "Horario de Venezuela", "WARST": "Horario de verán da Arxentina Occidental", "WART": "Horario estándar da Arxentina Occidental", "WAST": "Horario de verán de África Occidental", "WAT": "Horario estándar de África Occidental", "WESZ": "Horario de verán de Europa Occidental", "WEZ": "Horario estándar de Europa Occidental", "WIB": "Horario de Indonesia Occidental", "WIT": "Horario de Indonesia Oriental", "WITA": "Horario de Indonesia Central", "∅∅∅": "Horario de verán dos Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gl_es"] = func() locales.Translator {
		return &localen{

			locale:                 "gl_ES",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "xan.", "feb.", "mar.", "abr.", "maio", "xuño", "xul.", "ago.", "set.", "out.", "nov.", "dec."},
			monthsNarrow:           []string{"", "x.", "f.", "m.", "a.", "m.", "x.", "x.", "a.", "s.", "o.", "n.", "d."},
			monthsWide:             []string{"", "xaneiro", "febreiro", "marzo", "abril", "maio", "xuño", "xullo", "agosto", "setembro", "outubro", "novembro", "decembro"},
			daysAbbreviated:        []string{"dom.", "luns", "mar.", "mér.", "xov.", "ven.", "sáb."},
			daysNarrow:             []string{"d.", "l.", "m.", "m.", "x.", "v.", "s."},
			daysShort:              []string{"do.", "lu.", "ma.", "mé.", "xo.", "ve.", "sá."},
			daysWide:               []string{"domingo", "luns", "martes", "mércores", "xoves", "venres", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "despois de Cristo"},
			timezones:              map[string]string{"ACDT": "Horario de verán de Australia Central", "ACST": "Horario estándar de Australia Central", "ACWDT": "Horario de verán de Australia Occidental Central", "ACWST": "Horario estándar de Australia Occidental Central", "ADT": "Horario de verán do Atlántico", "AEDT": "Horario de verán de Australia Oriental", "AEST": "Horario estándar de Australia Oriental", "AKDT": "Horario de verán de Alasca", "AKST": "Horario estándar de Alasca", "ARST": "Horario de verán da Arxentina", "ART": "Horario estándar da Arxentina", "AST": "Horario estándar do Atlántico", "AWDT": "Horario de verán de Australia Occidental", "AWST": "Horario estándar de Australia Occidental", "BOT": "Horario de Bolivia", "BT": "Horario de Bután", "CAT": "Horario de África Central", "CDT": "Horario de verán central, Norteamérica", "CHADT": "Horario de verán de Chatham", "CHAST": "Horario estándar de Chatham", "CLST": "Horario de verán de Chile", "CLT": "Horario estándar de Chile", "COST": "Horario de verán de Colombia", "COT": "Horario estándar de Colombia", "CST": "Horario estándar central, Norteamérica", "ChST": "Horario estándar chamorro", "EAT": "Horario de África Oriental", "ECT": "Horario de Ecuador", "EDT": "Horario de verán do leste, América do Norte", "EST": "Horario estándar do leste, América do Norte", "GFT": "Horario da Güiana Francesa", "GMT": "Horario do meridiano de Greenwich", "GST": "Horario do Golfo", "GYT": "Horario da Güiana", "HADT": "Horario de verán de Hawai-illas Aleutianas", "HAST": "Horario estándar de Hawai-illas Aleutianas", "HAT": "Horario de verán de Terra Nova", "HECU": "Horario de verán de Cuba", "HEEG": "Horario de verán de Groenlandia Oriental", "HENOMX": "Horario de verán do noroeste de México", "HEOG": "Horario de verán de Groenlandia Occidental", "HEPM": "Horario de verán de Saint Pierre et Miquelon", "HEPMX": "Horario de verán do Pacífico mexicano", "HKST": "Horario de verán de Hong Kong", "HKT": "Horario estándar de Hong Kong", "HNCU": "Horario estándar de Cuba", "HNEG": "Horario estándar de Groenlandia Oriental", "HNNOMX": "Horario estándar do noroeste de México", "HNOG": "Horario estándar de Groenlandia Occidental", "HNPM": "Horario estándar de Saint Pierre et Miquelon", "HNPMX": "Horario estándar do Pacífico mexicano", "HNT": "Horario estándar de Terra Nova", "IST": "Horario da India", "JDT": "Horario de verán do Xapón", "JST": "Horario estándar do Xapón", "LHDT": "Horario de verán de Lord Howe", "LHST": "Horario estándar de Lord Howe", "MDT": "Horario de verán da montaña, América do Norte", "MESZ": "Horario de verán de Europa Central", "MEZ": "Horario estándar de Europa Central", "MST": "Horario estándar da montaña, América do Norte", "MYT": "Horario de Malaisia", "NZDT": "Horario de verán de Nova Zelandia", "NZST": "Horario estándar de Nova Zelandia", "OESZ": "Horario de verán de Europa Oriental", "OEZ": "Horario estándar de Europa Oriental", "PDT": "Horario de verán do Pacífico, América do Norte", "PST": "Horario estándar do Pacífico, América do Norte", "SAST": "Horario de África Meridional", "SGT": "Horario de Singapur", "SRT": "Horario de Suriname", "TMST": "Horario de verán de Turkmenistán", "TMT": "Horario estándar de Turkmenistán", "UYST": "Horario de verán do Uruguai", "UYT": "Horario estándar do Uruguai", "VET": "Horario de Venezuela", "WARST": "Horario de verán da Arxentina Occidental", "WART": "Horario estándar da Arxentina Occidental", "WAST": "Horario de verán de África Occidental", "WAT": "Horario estándar de África Occidental", "WESZ": "Horario de verán de Europa Occidental", "WEZ": "Horario estándar de Europa Occidental", "WIB": "Horario de Indonesia Occidental", "WIT": "Horario de Indonesia Oriental", "WITA": "Horario de Indonesia Central", "∅∅∅": "Horario de verán de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gsw"] = func() locales.Translator {
		return &localen{

			locale:                 "gsw",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "öS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "Auguscht", "Septämber", "Oktoober", "Novämber", "Dezämber"},
			daysAbbreviated:        []string{"Su.", "Mä.", "Zi.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sunntig", "Määntig", "Ziischtig", "Mittwuch", "Dunschtig", "Friitig", "Samschtig"},
			periodsAbbreviated:     []string{"vorm.", "nam."},
			periodsWide:            []string{"am Vormittag", "am Namittag"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "Alaska-Summerziit", "AKST": "Alaska-Schtandardziit", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Zentralafrikanischi Ziit", "CDT": "Amerika-Zentraal Summerziit", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Amerika-Zentraal Schtandardziit", "ChST": "ChST", "EAT": "Oschtafrikanischi Ziit", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Mitteleuropäischi Summerziit", "MEZ": "Mitteleuropäischi Schtandardziit", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oschteuropäischi Summerziit", "OEZ": "Oschteuropäischi Schtandardziit", "PDT": "PDT", "PST": "PST", "SAST": "Süüdafrikanischi ziit", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Weschtafrikanischi Summerziit", "WAT": "Weschtafrikanischi Schtandardziit", "WESZ": "Weschteuropäischi Summerziit", "WEZ": "Weschteuropäischi Schtandardziit", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74,
			fnFmtCurrency:         fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gsw_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "gsw_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "Auguscht", "Septämber", "Oktoober", "Novämber", "Dezämber"},
			daysAbbreviated:        []string{"Su.", "Mä.", "Zi.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sunntig", "Määntig", "Ziischtig", "Mittwuch", "Dunschtig", "Friitig", "Samschtig"},
			periodsAbbreviated:     []string{"vorm.", "nam."},
			periodsWide:            []string{"am Vormittag", "am Namittag"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "Alaska-Summerziit", "AKST": "Alaska-Schtandardziit", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Zentralafrikanischi Ziit", "CDT": "Amerika-Zentraal Summerziit", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Amerika-Zentraal Schtandardziit", "ChST": "ChST", "EAT": "Oschtafrikanischi Ziit", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Mitteleuropäischi Summerziit", "MEZ": "Mitteleuropäischi Schtandardziit", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oschteuropäischi Summerziit", "OEZ": "Oschteuropäischi Schtandardziit", "PDT": "PDT", "PST": "PST", "SAST": "Süüdafrikanischi ziit", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Weschtafrikanischi Summerziit", "WAT": "Weschtafrikanischi Schtandardziit", "WESZ": "Weschteuropäischi Summerziit", "WEZ": "Weschteuropäischi Schtandardziit", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74,
			fnFmtCurrency:         fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gsw_fr"] = func() locales.Translator {
		return &localen{

			locale:                 "gsw_FR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "Auguscht", "Septämber", "Oktoober", "Novämber", "Dezämber"},
			daysAbbreviated:        []string{"Su.", "Mä.", "Zi.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sunntig", "Määntig", "Ziischtig", "Mittwuch", "Dunschtig", "Friitig", "Samschtig"},
			periodsAbbreviated:     []string{"vorm.", "nam."},
			periodsWide:            []string{"am Vormittag", "am Namittag"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "Alaska-Summerziit", "AKST": "Alaska-Schtandardziit", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Zentralafrikanischi Ziit", "CDT": "Amerika-Zentraal Summerziit", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Amerika-Zentraal Schtandardziit", "ChST": "ChST", "EAT": "Oschtafrikanischi Ziit", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Mitteleuropäischi Summerziit", "MEZ": "Mitteleuropäischi Schtandardziit", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oschteuropäischi Summerziit", "OEZ": "Oschteuropäischi Schtandardziit", "PDT": "PDT", "PST": "PST", "SAST": "Süüdafrikanischi ziit", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Weschtafrikanischi Summerziit", "WAT": "Weschtafrikanischi Schtandardziit", "WESZ": "Weschteuropäischi Summerziit", "WEZ": "Weschteuropäischi Schtandardziit", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74,
			fnFmtCurrency:         fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gsw_li"] = func() locales.Translator {
		return &localen{

			locale:                 "gsw_LI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "Auguscht", "Septämber", "Oktoober", "Novämber", "Dezämber"},
			daysAbbreviated:        []string{"Su.", "Mä.", "Zi.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sunntig", "Määntig", "Ziischtig", "Mittwuch", "Dunschtig", "Friitig", "Samschtig"},
			periodsAbbreviated:     []string{"vorm.", "nam."},
			periodsWide:            []string{"am Vormittag", "am Namittag"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"v. Chr.", "n. Chr."},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "Alaska-Summerziit", "AKST": "Alaska-Schtandardziit", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Zentralafrikanischi Ziit", "CDT": "Amerika-Zentraal Summerziit", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Amerika-Zentraal Schtandardziit", "ChST": "ChST", "EAT": "Oschtafrikanischi Ziit", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Mitteleuropäischi Summerziit", "MEZ": "Mitteleuropäischi Schtandardziit", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oschteuropäischi Summerziit", "OEZ": "Oschteuropäischi Schtandardziit", "PDT": "PDT", "PST": "PST", "SAST": "Süüdafrikanischi ziit", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Weschtafrikanischi Summerziit", "WAT": "Weschtafrikanischi Schtandardziit", "WESZ": "Weschteuropäischi Summerziit", "WEZ": "Weschteuropäischi Schtandardziit", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74,
			fnFmtCurrency:         fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_13bf223568676db2ecc2cdf26633b243,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gu"] = func() locales.Translator {
		return &localen{

			locale:             "gu",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "જાન્યુ", "ફેબ્રુ", "માર્ચ", "એપ્રિલ", "મે", "જૂન", "જુલાઈ", "ઑગસ્ટ", "સપ્ટે", "ઑક્ટો", "નવે", "ડિસે"},
			monthsNarrow:       []string{"", "જા", "ફે", "મા", "એ", "મે", "જૂ", "જુ", "ઑ", "સ", "ઑ", "ન", "ડિ"},
			monthsWide:         []string{"", "જાન્યુઆરી", "ફેબ્રુઆરી", "માર્ચ", "એપ્રિલ", "મે", "જૂન", "જુલાઈ", "ઑગસ્ટ", "સપ્ટેમ્બર", "ઑક્ટોબર", "નવેમ્બર", "ડિસેમ્બર"},
			daysAbbreviated:    []string{"રવિ", "સોમ", "મંગળ", "બુધ", "ગુરુ", "શુક્ર", "શનિ"},
			daysNarrow:         []string{"ર", "સો", "મં", "બુ", "ગુ", "શુ", "શ"},
			daysShort:          []string{"ર", "સો", "મં", "બુ", "ગુ", "શુ", "શ"},
			daysWide:           []string{"રવિવાર", "સોમવાર", "મંગળવાર", "બુધવાર", "ગુરુવાર", "શુક્રવાર", "શનિવાર"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"ઈ.સ.પૂર્વે", "ઈ.સ."},
			erasNarrow:         []string{"ઇ સ પુ", "ઇસ"},
			erasWide:           []string{"ઈસવીસન પૂર્વે", "ઇસવીસન"},
			timezones:          map[string]string{"ACDT": "ઓસ્ટ્રેલિયન કેન્દ્રીય દિવસ સમય", "ACST": "ઑસ્ટ્રેલિયન કેન્દ્રીય માનક સમય", "ACWDT": "ઓસ્ટ્રેલિયન કેન્દ્રીય પશ્ચિમી દિવસ સમય", "ACWST": "ઑસ્ટ્રેલિયન કેન્દ્રીય પશ્ચિમી માનક સમય", "ADT": "એટલાન્ટિક દિવસ સમય", "AEDT": "ઓસ્ટ્રેલિયન પૂર્વીય દિવસ સમય", "AEST": "ઑસ્ટ્રેલિયન પૂર્વીય માનક સમય", "AKDT": "અલાસ્કા દિવસ સમય", "AKST": "અલાસ્કા માનક સમય", "ARST": "આર્જેન્ટીના ગ્રીષ્મ સમય", "ART": "આર્જેન્ટીના માનક સમય", "AST": "એટલાન્ટિક માનક સમય", "AWDT": "ઓસ્ટ્રેલિયન પશ્ચિમી દિવસ સમય", "AWST": "ઑસ્ટ્રેલિયન પશ્ચિમી માનક સમય", "BOT": "બોલિવિયા સમય", "BT": "ભૂટાન સમય", "CAT": "મધ્ય આફ્રિકા સમય", "CDT": "ઉત્તર અમેરિકન કેન્દ્રીય દિવસ સમય", "CHADT": "ચેતહામ દિવસ સમય", "CHAST": "ચેતહામ માનક સમય", "CLST": "ચિલી ગ્રીષ્મ સમય", "CLT": "ચિલી માનક સમય", "COST": "કોલમ્બિયા ગ્રીષ્મ સમય", "COT": "કોલમ્બિયા માનક સમય", "CST": "ઉત્તર અમેરિકન કેન્દ્રીય માનક સમય", "ChST": "કેમોરો માનક સમય", "EAT": "પૂર્વ આફ્રિકા સમય", "ECT": "એક્વાડોર સમય", "EDT": "ઉત્તર અમેરિકન પૂર્વી દિવસ સમય", "EST": "ઉત્તર અમેરિકન પૂર્વી માનક સમય", "GFT": "ફ્રેંચ ગયાના સમય", "GMT": "ગ્રીનવિચ મધ્યમ સમય", "GST": "ગલ્ફ માનક સમય", "GYT": "ગયાના સમય", "HADT": "હવાઇ-એલ્યુશિઅન દિવસ સમય", "HAST": "હવાઇ-એલ્યુશિઅન માનક સમય", "HAT": "ન્યૂફાઉન્ડલેન્ડ દિવસ સમય", "HECU": "ક્યુબા દિવસ સમય", "HEEG": "પૂર્વ ગ્રીનલેન્ડ ગ્રીષ્મ સમય", "HENOMX": "ઉત્તરપશ્ચિમ મેક્સિકો દિવસ સમય", "HEOG": "પશ્ચિમ ગ્રીનલેન્ડ ગ્રીષ્મ સમય", "HEPM": "સેંટ પીએરી અને મિક્યુલોન દિવસ સમય", "HEPMX": "મેક્સિકન પેસિફિક દિવસ સમય", "HKST": "હોંગ કોંગ ગ્રીષ્મ સમય", "HKT": "હોંગ કોંગ માનક સમય", "HNCU": "ક્યુબા માનક સમય", "HNEG": "પૂર્વ ગ્રીનલેન્ડ માનક સમય", "HNNOMX": "ઉત્તરપશ્ચિમ મેક્સિકો માનક સમય", "HNOG": "પશ્ચિમ ગ્રીનલેન્ડ માનક સમય", "HNPM": "સેંટ પીએરી અને મિક્યુલોન માનક સમય", "HNPMX": "મેક્સિકન પેસિફિક માનક સમય", "HNT": "ન્યૂફાઉન્ડલેન્ડ માનક સમય", "IST": "ભારતીય માનક સમય", "JDT": "જાપાન દિવસ સમય", "JST": "જાપાન માનક સમય", "LHDT": "લોર્ડ હોવ દિવસ સમય", "LHST": "લોર્ડ હોવ માનક સમય", "MDT": "ઉત્તર અમેરિકન માઉન્ટન દિવસ સમય", "MESZ": "મધ્ય યુરોપિયન ગ્રીષ્મ સમય", "MEZ": "મધ્ય યુરોપિયન માનક સમય", "MST": "ઉત્તર અમેરિકન માઉન્ટન માનક સમય", "MYT": "મલેશિયા સમય", "NZDT": "ન્યુઝીલેન્ડ દિવસ સમય", "NZST": "ન્યુઝીલેન્ડ માનક સમય", "OESZ": "પૂર્વી યુરોપિયન ગ્રીષ્મ સમય", "OEZ": "પૂર્વી યુરોપિયન માનક સમય", "PDT": "ઉત્તર અમેરિકન પેસિફિક દિવસ સમય", "PST": "ઉત્તર અમેરિકન પેસિફિક માનક સમય", "SAST": "દક્ષિણ આફ્રિકા માનક સમય", "SGT": "સિંગાપુર માનક સમય", "SRT": "સુરીનામ સમય", "TMST": "તુર્કમેનિસ્તાન ગ્રીષ્મ સમય", "TMT": "તુર્કમેનિસ્તાન માનક સમય", "UYST": "ઉરુગ્વે ગ્રીષ્મ સમય", "UYT": "ઉરુગ્વે માનક સમય", "VET": "વેનેઝુએલા સમય", "WARST": "પશ્ચિમી આર્જેન્ટીના ગ્રીષ્મ સમય", "WART": "પશ્ચિમી આર્જેન્ટીના માનક સમય", "WAST": "પશ્ચિમ આફ્રિકા ગ્રીષ્મ સમય", "WAT": "પશ્ચિમ આફ્રિકા માનક સમય", "WESZ": "પશ્ચિમી યુરોપિયન ગ્રીષ્મ સમય", "WEZ": "પશ્ચિમી યુરોપિયન માનક સમય", "WIB": "પશ્ચિમી ઇન્ડોનેશિયા સમય", "WIT": "પૂર્વીય ઇન્ડોનેશિયા સમય", "WITA": "મધ્ય ઇન્ડોનેશિયા સમય", "∅∅∅": "એમેઝોન ગ્રીષ્મ સમય"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_abd60cef530beeba744de08bba2ad540,
			fnFmtTimeLong:         fnFmtTimeLong_28302cc672efd8ad015891c59efdc448,
			fnFmtTimeMedium:       fnFmtTimeMedium_9efd56b286449b21c42b6b2ea8c1f34c,
			fnFmtTimeShort:        fnFmtTimeShort_11febbd269414b302ab4045dd63f47b7,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a9f120c4a8c40112dec183cd65a8b424,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gu_in"] = func() locales.Translator {
		return &localen{

			locale:             "gu_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "જાન્યુ", "ફેબ્રુ", "માર્ચ", "એપ્રિલ", "મે", "જૂન", "જુલાઈ", "ઑગસ્ટ", "સપ્ટે", "ઑક્ટો", "નવે", "ડિસે"},
			monthsNarrow:       []string{"", "જા", "ફે", "મા", "એ", "મે", "જૂ", "જુ", "ઑ", "સ", "ઑ", "ન", "ડિ"},
			monthsWide:         []string{"", "જાન્યુઆરી", "ફેબ્રુઆરી", "માર્ચ", "એપ્રિલ", "મે", "જૂન", "જુલાઈ", "ઑગસ્ટ", "સપ્ટેમ્બર", "ઑક્ટોબર", "નવેમ્બર", "ડિસેમ્બર"},
			daysAbbreviated:    []string{"રવિ", "સોમ", "મંગળ", "બુધ", "ગુરુ", "શુક્ર", "શનિ"},
			daysNarrow:         []string{"ર", "સો", "મં", "બુ", "ગુ", "શુ", "શ"},
			daysShort:          []string{"ર", "સો", "મં", "બુ", "ગુ", "શુ", "શ"},
			daysWide:           []string{"રવિવાર", "સોમવાર", "મંગળવાર", "બુધવાર", "ગુરુવાર", "શુક્રવાર", "શનિવાર"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"ઈ.સ.પૂર્વે", "ઈ.સ."},
			erasNarrow:         []string{"ઇ સ પુ", "ઇસ"},
			erasWide:           []string{"ઈસવીસન પૂર્વે", "ઇસવીસન"},
			timezones:          map[string]string{"ACDT": "ઓસ્ટ્રેલિયન કેન્દ્રીય દિવસ સમય", "ACST": "ઑસ્ટ્રેલિયન કેન્દ્રીય માનક સમય", "ACWDT": "ઓસ્ટ્રેલિયન કેન્દ્રીય પશ્ચિમી દિવસ સમય", "ACWST": "ઑસ્ટ્રેલિયન કેન્દ્રીય પશ્ચિમી માનક સમય", "ADT": "એટલાન્ટિક દિવસ સમય", "AEDT": "ઓસ્ટ્રેલિયન પૂર્વીય દિવસ સમય", "AEST": "ઑસ્ટ્રેલિયન પૂર્વીય માનક સમય", "AKDT": "અલાસ્કા દિવસ સમય", "AKST": "અલાસ્કા માનક સમય", "ARST": "આર્જેન્ટીના ગ્રીષ્મ સમય", "ART": "આર્જેન્ટીના માનક સમય", "AST": "એટલાન્ટિક માનક સમય", "AWDT": "ઓસ્ટ્રેલિયન પશ્ચિમી દિવસ સમય", "AWST": "ઑસ્ટ્રેલિયન પશ્ચિમી માનક સમય", "BOT": "બોલિવિયા સમય", "BT": "ભૂટાન સમય", "CAT": "મધ્ય આફ્રિકા સમય", "CDT": "ઉત્તર અમેરિકન કેન્દ્રીય દિવસ સમય", "CHADT": "ચેતહામ દિવસ સમય", "CHAST": "ચેતહામ માનક સમય", "CLST": "ચિલી ગ્રીષ્મ સમય", "CLT": "ચિલી માનક સમય", "COST": "કોલમ્બિયા ગ્રીષ્મ સમય", "COT": "કોલમ્બિયા માનક સમય", "CST": "ઉત્તર અમેરિકન કેન્દ્રીય માનક સમય", "ChST": "કેમોરો માનક સમય", "EAT": "પૂર્વ આફ્રિકા સમય", "ECT": "એક્વાડોર સમય", "EDT": "ઉત્તર અમેરિકન પૂર્વી દિવસ સમય", "EST": "ઉત્તર અમેરિકન પૂર્વી માનક સમય", "GFT": "ફ્રેંચ ગયાના સમય", "GMT": "ગ્રીનવિચ મધ્યમ સમય", "GST": "ગલ્ફ માનક સમય", "GYT": "ગયાના સમય", "HADT": "હવાઇ-એલ્યુશિઅન દિવસ સમય", "HAST": "હવાઇ-એલ્યુશિઅન માનક સમય", "HAT": "ન્યૂફાઉન્ડલેન્ડ દિવસ સમય", "HECU": "ક્યુબા દિવસ સમય", "HEEG": "પૂર્વ ગ્રીનલેન્ડ ગ્રીષ્મ સમય", "HENOMX": "ઉત્તરપશ્ચિમ મેક્સિકો દિવસ સમય", "HEOG": "પશ્ચિમ ગ્રીનલેન્ડ ગ્રીષ્મ સમય", "HEPM": "સેંટ પીએરી અને મિક્યુલોન દિવસ સમય", "HEPMX": "મેક્સિકન પેસિફિક દિવસ સમય", "HKST": "હોંગ કોંગ ગ્રીષ્મ સમય", "HKT": "હોંગ કોંગ માનક સમય", "HNCU": "ક્યુબા માનક સમય", "HNEG": "પૂર્વ ગ્રીનલેન્ડ માનક સમય", "HNNOMX": "ઉત્તરપશ્ચિમ મેક્સિકો માનક સમય", "HNOG": "પશ્ચિમ ગ્રીનલેન્ડ માનક સમય", "HNPM": "સેંટ પીએરી અને મિક્યુલોન માનક સમય", "HNPMX": "મેક્સિકન પેસિફિક માનક સમય", "HNT": "ન્યૂફાઉન્ડલેન્ડ માનક સમય", "IST": "ભારતીય માનક સમય", "JDT": "જાપાન દિવસ સમય", "JST": "જાપાન માનક સમય", "LHDT": "લોર્ડ હોવ દિવસ સમય", "LHST": "લોર્ડ હોવ માનક સમય", "MDT": "ઉત્તર અમેરિકન માઉન્ટન દિવસ સમય", "MESZ": "મધ્ય યુરોપિયન ગ્રીષ્મ સમય", "MEZ": "મધ્ય યુરોપિયન માનક સમય", "MST": "ઉત્તર અમેરિકન માઉન્ટન માનક સમય", "MYT": "મલેશિયા સમય", "NZDT": "ન્યુઝીલેન્ડ દિવસ સમય", "NZST": "ન્યુઝીલેન્ડ માનક સમય", "OESZ": "પૂર્વી યુરોપિયન ગ્રીષ્મ સમય", "OEZ": "પૂર્વી યુરોપિયન માનક સમય", "PDT": "ઉત્તર અમેરિકન પેસિફિક દિવસ સમય", "PST": "ઉત્તર અમેરિકન પેસિફિક માનક સમય", "SAST": "દક્ષિણ આફ્રિકા માનક સમય", "SGT": "સિંગાપુર માનક સમય", "SRT": "સુરીનામ સમય", "TMST": "તુર્કમેનિસ્તાન ગ્રીષ્મ સમય", "TMT": "તુર્કમેનિસ્તાન માનક સમય", "UYST": "ઉરુગ્વે ગ્રીષ્મ સમય", "UYT": "ઉરુગ્વે માનક સમય", "VET": "વેનેઝુએલા સમય", "WARST": "પશ્ચિમી આર્જેન્ટીના ગ્રીષ્મ સમય", "WART": "પશ્ચિમી આર્જેન્ટીના માનક સમય", "WAST": "પશ્ચિમ આફ્રિકા ગ્રીષ્મ સમય", "WAT": "પશ્ચિમ આફ્રિકા માનક સમય", "WESZ": "પશ્ચિમી યુરોપિયન ગ્રીષ્મ સમય", "WEZ": "પશ્ચિમી યુરોપિયન માનક સમય", "WIB": "પશ્ચિમી ઇન્ડોનેશિયા સમય", "WIT": "પૂર્વીય ઇન્ડોનેશિયા સમય", "WITA": "મધ્ય ઇન્ડોનેશિયા સમય", "∅∅∅": "એઝોર્સ ગ્રીષ્મ સમય"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_abd60cef530beeba744de08bba2ad540,
			fnFmtTimeLong:         fnFmtTimeLong_28302cc672efd8ad015891c59efdc448,
			fnFmtTimeMedium:       fnFmtTimeMedium_9efd56b286449b21c42b6b2ea8c1f34c,
			fnFmtTimeShort:        fnFmtTimeShort_11febbd269414b302ab4045dd63f47b7,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a9f120c4a8c40112dec183cd65a8b424,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["guz"] = func() locales.Translator {
		return &localen{

			locale:                 "guz",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Can", "Feb", "Mac", "Apr", "Mei", "Jun", "Cul", "Agt", "Sep", "Okt", "Nob", "Dis"},
			monthsNarrow:           []string{"", "C", "F", "M", "A", "M", "J", "C", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Chanuari", "Feburari", "Machi", "Apiriri", "Mei", "Juni", "Chulai", "Agosti", "Septemba", "Okitoba", "Nobemba", "Disemba"},
			daysAbbreviated:        []string{"Cpr", "Ctt", "Cmn", "Cmt", "Ars", "Icm", "Est"},
			daysNarrow:             []string{"C", "C", "C", "C", "A", "I", "E"},
			daysWide:               []string{"Chumapiri", "Chumatato", "Chumaine", "Chumatano", "Aramisi", "Ichuma", "Esabato"},
			periodsAbbreviated:     []string{"Ma", "Mo"},
			periodsWide:            []string{"Mambia", "Mog"},
			erasAbbreviated:        []string{"YA", "YK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yeso ataiborwa", "Yeso kaiboirwe"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["guz_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "guz_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Can", "Feb", "Mac", "Apr", "Mei", "Jun", "Cul", "Agt", "Sep", "Okt", "Nob", "Dis"},
			monthsNarrow:           []string{"", "C", "F", "M", "A", "M", "J", "C", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Chanuari", "Feburari", "Machi", "Apiriri", "Mei", "Juni", "Chulai", "Agosti", "Septemba", "Okitoba", "Nobemba", "Disemba"},
			daysAbbreviated:        []string{"Cpr", "Ctt", "Cmn", "Cmt", "Ars", "Icm", "Est"},
			daysNarrow:             []string{"C", "C", "C", "C", "A", "I", "E"},
			daysWide:               []string{"Chumapiri", "Chumatato", "Chumaine", "Chumatano", "Aramisi", "Ichuma", "Esabato"},
			periodsAbbreviated:     []string{"Ma", "Mo"},
			periodsWide:            []string{"Mambia", "Mog"},
			erasAbbreviated:        []string{"YA", "YK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yeso ataiborwa", "Yeso kaiboirwe"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gv"] = func() locales.Translator {
		return &localen{

			locale:             "gv",
			pluralsCardinal:    []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "J-guer", "T-arree", "Mayrnt", "Avrril", "Boaldyn", "M-souree", "J-souree", "Luanistyn", "M-fouyir", "J-fouyir", "M-Houney", "M-Nollick"},
			monthsWide:         []string{"", "Jerrey-geuree", "Toshiaght-arree", "Mayrnt", "Averil", "Boaldyn", "Mean-souree", "Jerrey-souree", "Luanistyn", "Mean-fouyir", "Jerrey-fouyir", "Mee Houney", "Mee ny Nollick"},
			daysAbbreviated:    []string{"Jed", "Jel", "Jem", "Jerc", "Jerd", "Jeh", "Jes"},
			daysWide:           []string{"Jedoonee", "Jelhein", "Jemayrt", "Jercean", "Jerdein", "Jeheiney", "Jesarn"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"RC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_e63270f01497e6cdf2ee8385ca8e8db3,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_d8430d5f441fe322a704e04df73f1229,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["gv_im"] = func() locales.Translator {
		return &localen{

			locale:             "gv_IM",
			pluralsCardinal:    []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "J-guer", "T-arree", "Mayrnt", "Avrril", "Boaldyn", "M-souree", "J-souree", "Luanistyn", "M-fouyir", "J-fouyir", "M-Houney", "M-Nollick"},
			monthsWide:         []string{"", "Jerrey-geuree", "Toshiaght-arree", "Mayrnt", "Averil", "Boaldyn", "Mean-souree", "Jerrey-souree", "Luanistyn", "Mean-fouyir", "Jerrey-fouyir", "Mee Houney", "Mee ny Nollick"},
			daysAbbreviated:    []string{"Jed", "Jel", "Jem", "Jerc", "Jerd", "Jeh", "Jes"},
			daysWide:           []string{"Jedoonee", "Jelhein", "Jemayrt", "Jercean", "Jerdein", "Jeheiney", "Jesarn"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"RC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_e63270f01497e6cdf2ee8385ca8e8db3,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_d8430d5f441fe322a704e04df73f1229,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ha"] = func() locales.Translator {
		return &localen{

			locale:                 "ha",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "₦", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "D",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "D",
			monthsAbbreviated:      []string{"", "Jan", "Fab", "Mar", "Afi", "May", "Yun", "Yul", "Agu", "Sat", "Okt", "Nuw", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "Y", "Y", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janairu", "Faburairu", "Maris", "Afirilu", "Mayu", "Yuni", "Yuli", "Agusta", "Satumba", "Oktoba", "Nuwamba", "Disamba"},
			daysAbbreviated:        []string{"Lah", "Lit", "Tal", "Lar", "Alh", "Jum", "Asa"},
			daysNarrow:             []string{"L", "L", "T", "L", "A", "J", "A"},
			daysShort:              []string{"Lh", "Li", "Ta", "Lr", "Al", "Ju", "As"},
			daysWide:               []string{"Lahadi", "Litinin", "Talata", "Laraba", "Alhamis", "Jummaʼa", "Asabar"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"Safiya", "Yamma"},
			erasAbbreviated:        []string{"K.H", "BHAI"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Lokacin Rana na Kanada, Puerto Rico da Virgin Islands", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Lokacin Rana na Alaska", "AKST": "Tsayayyen Lokacin Alaska", "ARST": "ARST", "ART": "ART", "AST": "Lokaci Tsayayye na Kanada, Puerto Rico da Virgin Islands", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "BOT", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Lokacin Rana dake Arewacin Amurika ta Tsakiya", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Tsayayyen Lokaci dake Arewacin Amurika ta Tsakiya", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "ECT", "EDT": "Lokacin Rana ta Gabas dake Arewacin Amurika", "EST": "Tsayayyen Lokacin Gabas dake Arewacin Amurika", "GFT": "GFT", "GMT": "Lokacin Greenwhich a London", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "GYT", "HADT": "Lokaci rana ta Hawaii-Aleutian", "HAST": "Ida Lokaci ta Hawaii-Aleutian", "HAT": "Lokaci rana ta Newfoundland", "HECU": "Lokaci rana ta Kuba", "HEEG": "Lokaci rana a gabas ta Greeland", "HENOMX": "Lokacin rana na arewa maso gabashin Mesiko", "HEOG": "Lokacin rana a yammacin Greeland", "HEPM": "Lokaci rana ta St. Pierre da Miquelon", "HEPMX": "Lokaci na rana na Mesiko Pacific", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Lokaci Tsayayye na Kuba", "HNEG": "Lokaci Tsayayye a gabashin ta Greenland", "HNNOMX": "Lokaci Tsayayye na arewa maso gabashin Mesiko", "HNOG": "Lokaci Tsayayye a yammacin Greeland", "HNPM": "Lokaci tsayayye St. Pierre da Miquelon", "HNPMX": "Lokaci Tsayayye na Mesiko Pacific", "HNT": "Lokaci Tsayayye ta Newfoundland", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Lokaci rana tsauni a arewacin da Amirka", "MESZ": "Tsakiyar bazara a lokaci turai", "MEZ": "Ida Tsakiyar a Lokaci Turai", "MST": "Lokaci tsayayye na tsauni a Arewacin Amurica", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Gabas a lokaci turai da bazara", "OEZ": "Ida lokaci a turai gabas", "PDT": "Lokaci da rana a Arewacin Amurika", "PST": "Lokaci Tsayayye na Arewacin Amurika", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "SRT", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Ida lokaci ta yammacin turai da bazara", "WEZ": "Ida lokaci ta yammacin turai", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_4f1a5e5b794a77b5b18293f768464f0a,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ha_gh"] = func() locales.Translator {
		return &localen{

			locale:                 "ha_GH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GH₵", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "D",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "D",
			monthsAbbreviated:      []string{"", "Jan", "Fab", "Mar", "Afi", "May", "Yun", "Yul", "Agu", "Sat", "Okt", "Nuw", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "Y", "Y", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janairu", "Faburairu", "Maris", "Afirilu", "Mayu", "Yuni", "Yuli", "Agusta", "Satumba", "Oktoba", "Nuwamba", "Disamba"},
			daysAbbreviated:        []string{"Lah", "Lit", "Tal", "Lar", "Alh", "Jum", "Asa"},
			daysNarrow:             []string{"L", "L", "T", "L", "A", "J", "A"},
			daysShort:              []string{"Lh", "Li", "Ta", "Lr", "Al", "Ju", "As"},
			daysWide:               []string{"Lahadi", "Litinin", "Talata", "Laraba", "Alhamis", "Jummaʼa", "Asabar"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"Safiya", "Yamma"},
			erasAbbreviated:        []string{"K.H", "BHAI"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Lokacin Rana na Kanada, Puerto Rico da Virgin Islands", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Lokacin Rana na Alaska", "AKST": "Tsayayyen Lokacin Alaska", "ARST": "ARST", "ART": "ART", "AST": "Lokaci Tsayayye na Kanada, Puerto Rico da Virgin Islands", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "BOT", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Lokacin Rana dake Arewacin Amurika ta Tsakiya", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Tsayayyen Lokaci dake Arewacin Amurika ta Tsakiya", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "ECT", "EDT": "Lokacin Rana ta Gabas dake Arewacin Amurika", "EST": "Tsayayyen Lokacin Gabas dake Arewacin Amurika", "GFT": "GFT", "GMT": "Lokacin Greenwhich a London", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "GYT", "HADT": "Lokaci rana ta Hawaii-Aleutian", "HAST": "Ida Lokaci ta Hawaii-Aleutian", "HAT": "Lokaci rana ta Newfoundland", "HECU": "Lokaci rana ta Kuba", "HEEG": "Lokaci rana a gabas ta Greeland", "HENOMX": "Lokacin rana na arewa maso gabashin Mesiko", "HEOG": "Lokacin rana a yammacin Greeland", "HEPM": "Lokaci rana ta St. Pierre da Miquelon", "HEPMX": "Lokaci na rana na Mesiko Pacific", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Lokaci Tsayayye na Kuba", "HNEG": "Lokaci Tsayayye a gabashin ta Greenland", "HNNOMX": "Lokaci Tsayayye na arewa maso gabashin Mesiko", "HNOG": "Lokaci Tsayayye a yammacin Greeland", "HNPM": "Lokaci tsayayye St. Pierre da Miquelon", "HNPMX": "Lokaci Tsayayye na Mesiko Pacific", "HNT": "Lokaci Tsayayye ta Newfoundland", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Lokaci rana tsauni a arewacin da Amirka", "MESZ": "Tsakiyar bazara a lokaci turai", "MEZ": "Ida Tsakiyar a Lokaci Turai", "MST": "Lokaci tsayayye na tsauni a Arewacin Amurica", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Gabas a lokaci turai da bazara", "OEZ": "Ida lokaci a turai gabas", "PDT": "Lokaci da rana a Arewacin Amurika", "PST": "Lokaci Tsayayye na Arewacin Amurika", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "SRT", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Ida lokaci ta yammacin turai da bazara", "WEZ": "Ida lokaci ta yammacin turai", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_4f1a5e5b794a77b5b18293f768464f0a,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ha_ne"] = func() locales.Translator {
		return &localen{

			locale:                 "ha_NE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "D",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "D",
			monthsAbbreviated:      []string{"", "Jan", "Fab", "Mar", "Afi", "May", "Yun", "Yul", "Agu", "Sat", "Okt", "Nuw", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "Y", "Y", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janairu", "Faburairu", "Maris", "Afirilu", "Mayu", "Yuni", "Yuli", "Agusta", "Satumba", "Oktoba", "Nuwamba", "Disamba"},
			daysAbbreviated:        []string{"Lah", "Lit", "Tal", "Lar", "Alh", "Jum", "Asa"},
			daysNarrow:             []string{"L", "L", "T", "L", "A", "J", "A"},
			daysShort:              []string{"Lh", "Li", "Ta", "Lr", "Al", "Ju", "As"},
			daysWide:               []string{"Lahadi", "Litinin", "Talata", "Laraba", "Alhamis", "Jummaʼa", "Asabar"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"Safiya", "Yamma"},
			erasAbbreviated:        []string{"K.H", "BHAI"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Lokacin Rana na Kanada, Puerto Rico da Virgin Islands", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Lokacin Rana na Alaska", "AKST": "Tsayayyen Lokacin Alaska", "ARST": "ARST", "ART": "ART", "AST": "Lokaci Tsayayye na Kanada, Puerto Rico da Virgin Islands", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "BOT", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Lokacin Rana dake Arewacin Amurika ta Tsakiya", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Tsayayyen Lokaci dake Arewacin Amurika ta Tsakiya", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "ECT", "EDT": "Lokacin Rana ta Gabas dake Arewacin Amurika", "EST": "Tsayayyen Lokacin Gabas dake Arewacin Amurika", "GFT": "GFT", "GMT": "Lokacin Greenwhich a London", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "GYT", "HADT": "Lokaci rana ta Hawaii-Aleutian", "HAST": "Ida Lokaci ta Hawaii-Aleutian", "HAT": "Lokaci rana ta Newfoundland", "HECU": "Lokaci rana ta Kuba", "HEEG": "Lokaci rana a gabas ta Greeland", "HENOMX": "Lokacin rana na arewa maso gabashin Mesiko", "HEOG": "Lokacin rana a yammacin Greeland", "HEPM": "Lokaci rana ta St. Pierre da Miquelon", "HEPMX": "Lokaci na rana na Mesiko Pacific", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Lokaci Tsayayye na Kuba", "HNEG": "Lokaci Tsayayye a gabashin ta Greenland", "HNNOMX": "Lokaci Tsayayye na arewa maso gabashin Mesiko", "HNOG": "Lokaci Tsayayye a yammacin Greeland", "HNPM": "Lokaci tsayayye St. Pierre da Miquelon", "HNPMX": "Lokaci Tsayayye na Mesiko Pacific", "HNT": "Lokaci Tsayayye ta Newfoundland", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Lokaci rana tsauni a arewacin da Amirka", "MESZ": "Tsakiyar bazara a lokaci turai", "MEZ": "Ida Tsakiyar a Lokaci Turai", "MST": "Lokaci tsayayye na tsauni a Arewacin Amurica", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Gabas a lokaci turai da bazara", "OEZ": "Ida lokaci a turai gabas", "PDT": "Lokaci da rana a Arewacin Amurika", "PST": "Lokaci Tsayayye na Arewacin Amurika", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "SRT", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Ida lokaci ta yammacin turai da bazara", "WEZ": "Ida lokaci ta yammacin turai", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_4f1a5e5b794a77b5b18293f768464f0a,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ha_ng"] = func() locales.Translator {
		return &localen{

			locale:                 "ha_NG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "D",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "D",
			monthsAbbreviated:      []string{"", "Jan", "Fab", "Mar", "Afi", "May", "Yun", "Yul", "Agu", "Sat", "Okt", "Nuw", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "Y", "Y", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janairu", "Faburairu", "Maris", "Afirilu", "Mayu", "Yuni", "Yuli", "Agusta", "Satumba", "Oktoba", "Nuwamba", "Disamba"},
			daysAbbreviated:        []string{"Lah", "Lit", "Tal", "Lar", "Alh", "Jum", "Asa"},
			daysNarrow:             []string{"L", "L", "T", "L", "A", "J", "A"},
			daysShort:              []string{"Lh", "Li", "Ta", "Lr", "Al", "Ju", "As"},
			daysWide:               []string{"Lahadi", "Litinin", "Talata", "Laraba", "Alhamis", "Jummaʼa", "Asabar"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"Safiya", "Yamma"},
			erasAbbreviated:        []string{"K.H", "BHAI"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Lokacin Rana na Kanada, Puerto Rico da Virgin Islands", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Lokacin Rana na Alaska", "AKST": "Tsayayyen Lokacin Alaska", "ARST": "ARST", "ART": "ART", "AST": "Lokaci Tsayayye na Kanada, Puerto Rico da Virgin Islands", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "BOT", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Lokacin Rana dake Arewacin Amurika ta Tsakiya", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Tsayayyen Lokaci dake Arewacin Amurika ta Tsakiya", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "ECT", "EDT": "Lokacin Rana ta Gabas dake Arewacin Amurika", "EST": "Tsayayyen Lokacin Gabas dake Arewacin Amurika", "GFT": "GFT", "GMT": "Lokacin Greenwhich a London", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "GYT", "HADT": "Lokaci rana ta Hawaii-Aleutian", "HAST": "Ida Lokaci ta Hawaii-Aleutian", "HAT": "Lokaci rana ta Newfoundland", "HECU": "Lokaci rana ta Kuba", "HEEG": "Lokaci rana a gabas ta Greeland", "HENOMX": "Lokacin rana na arewa maso gabashin Mesiko", "HEOG": "Lokacin rana a yammacin Greeland", "HEPM": "Lokaci rana ta St. Pierre da Miquelon", "HEPMX": "Lokaci na rana na Mesiko Pacific", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "Lokaci Tsayayye na Kuba", "HNEG": "Lokaci Tsayayye a gabashin ta Greenland", "HNNOMX": "Lokaci Tsayayye na arewa maso gabashin Mesiko", "HNOG": "Lokaci Tsayayye a yammacin Greeland", "HNPM": "Lokaci tsayayye St. Pierre da Miquelon", "HNPMX": "Lokaci Tsayayye na Mesiko Pacific", "HNT": "Lokaci Tsayayye ta Newfoundland", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Lokaci rana tsauni a arewacin da Amirka", "MESZ": "Tsakiyar bazara a lokaci turai", "MEZ": "Ida Tsakiyar a Lokaci Turai", "MST": "Lokaci tsayayye na tsauni a Arewacin Amurica", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Gabas a lokaci turai da bazara", "OEZ": "Ida lokaci a turai gabas", "PDT": "Lokaci da rana a Arewacin Amurika", "PST": "Lokaci Tsayayye na Arewacin Amurika", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "SRT", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Ida lokaci ta yammacin turai da bazara", "WEZ": "Ida lokaci ta yammacin turai", "WIB": "Western Indonesia Time", "WIT": "Eastern Indonesia Time", "WITA": "Central Indonesia Time", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_4f1a5e5b794a77b5b18293f768464f0a,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["haw"] = func() locales.Translator {
		return &localen{

			locale:                 "haw",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ian.", "Pep.", "Mal.", "ʻAp.", "Mei", "Iun.", "Iul.", "ʻAu.", "Kep.", "ʻOk.", "Now.", "Kek."},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Ianuali", "Pepeluali", "Malaki", "ʻApelila", "Mei", "Iune", "Iulai", "ʻAukake", "Kepakemapa", "ʻOkakopa", "Nowemapa", "Kekemapa"},
			daysAbbreviated:        []string{"LP", "P1", "P2", "P3", "P4", "P5", "P6"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"LP", "P1", "P2", "P3", "P4", "P5", "P6"},
			daysWide:               []string{"Lāpule", "Poʻakahi", "Poʻalua", "Poʻakolu", "Poʻahā", "Poʻalima", "Poʻaono"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "CE"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["haw_us"] = func() locales.Translator {
		return &localen{

			locale:                 "haw_US",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ian.", "Pep.", "Mal.", "ʻAp.", "Mei", "Iun.", "Iul.", "ʻAu.", "Kep.", "ʻOk.", "Now.", "Kek."},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Ianuali", "Pepeluali", "Malaki", "ʻApelila", "Mei", "Iune", "Iulai", "ʻAukake", "Kepakemapa", "ʻOkakopa", "Nowemapa", "Kekemapa"},
			daysAbbreviated:        []string{"LP", "P1", "P2", "P3", "P4", "P5", "P6"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"LP", "P1", "P2", "P3", "P4", "P5", "P6"},
			daysWide:               []string{"Lāpule", "Poʻakahi", "Poʻalua", "Poʻakolu", "Poʻahā", "Poʻalima", "Poʻaono"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "CE"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["he"] = func() locales.Translator {
		return &localen{

			locale:                 "he",
			pluralsCardinal:        []locales.PluralRule{2, 3, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{5, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "‎-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "\u200eCN¥\u200e", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ל״י", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ינו׳", "פבר׳", "מרץ", "אפר׳", "מאי", "יוני", "יולי", "אוג׳", "ספט׳", "אוק׳", "נוב׳", "דצמ׳"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ינואר", "פברואר", "מרץ", "אפריל", "מאי", "יוני", "יולי", "אוגוסט", "ספטמבר", "אוקטובר", "נובמבר", "דצמבר"},
			daysAbbreviated:        []string{"יום א׳", "יום ב׳", "יום ג׳", "יום ד׳", "יום ה׳", "יום ו׳", "שבת"},
			daysNarrow:             []string{"א׳", "ב׳", "ג׳", "ד׳", "ה׳", "ו׳", "ש׳"},
			daysShort:              []string{"א׳", "ב׳", "ג׳", "ד׳", "ה׳", "ו׳", "ש׳"},
			daysWide:               []string{"יום ראשון", "יום שני", "יום שלישי", "יום רביעי", "יום חמישי", "יום שישי", "יום שבת"},
			periodsAbbreviated:     []string{"לפנה״צ", "אחה״צ"},
			periodsNarrow:          []string{"לפנה״צ", "אחה״צ"},
			periodsWide:            []string{"לפנה״צ", "אחה״צ"},
			erasAbbreviated:        []string{"לפנה״ס", "לספירה"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"לפני הספירה", "לספירה"},
			timezones:              map[string]string{"ACDT": "שעון מרכז אוסטרליה (קיץ)", "ACST": "שעון מרכז אוסטרליה (חורף)", "ACWDT": "שעון מרכז-מערב אוסטרליה (קיץ)", "ACWST": "שעון מרכז-מערב אוסטרליה (חורף)", "ADT": "שעון האוקיינוס האטלנטי (קיץ)", "AEDT": "שעון מזרח אוסטרליה (קיץ)", "AEST": "שעון מזרח אוסטרליה (חורף)", "AKDT": "שעון אלסקה (קיץ)", "AKST": "שעון אלסקה (חורף)", "ARST": "שעון ארגנטינה (קיץ)", "ART": "שעון ארגנטינה (חורף)", "AST": "שעון האוקיינוס האטלנטי (חורף)", "AWDT": "שעון מערב אוסטרליה (קיץ)", "AWST": "שעון מערב אוסטרליה (חורף)", "BOT": "שעון בוליביה", "BT": "שעון בהוטן", "CAT": "שעון מרכז אפריקה", "CDT": "שעון מרכז ארה״ב (קיץ)", "CHADT": "שעון צ׳טהאם (קיץ)", "CHAST": "שעון צ׳טהאם (חורף)", "CLST": "שעון צ׳ילה (קיץ)", "CLT": "שעון צ׳ילה (חורף)", "COST": "שעון קולומביה (קיץ)", "COT": "שעון קולומביה (חורף)", "CST": "שעון מרכז ארה״ב (חורף)", "ChST": "שעון צ׳אמורו", "EAT": "שעון מזרח אפריקה", "ECT": "שעון אקוודור", "EDT": "שעון החוף המזרחי (קיץ)", "EST": "שעון החוף המזרחי (חורף)", "GFT": "שעון גיאנה הצרפתית", "GMT": "שעון גריניץ׳\u200f", "GST": "שעון מדינות המפרץ", "GYT": "שעון גיאנה", "HADT": "שעון האיים האלאוטיים הוואי (קיץ)", "HAST": "שעון האיים האלאוטיים הוואי (חורף)", "HAT": "שעון ניופאונדלנד (קיץ)", "HECU": "שעון קובה (קיץ)", "HEEG": "שעון מזרח גרינלנד (קיץ)", "HENOMX": "שעון צפון-מערב מקסיקו (קיץ)", "HEOG": "שעון מערב גרינלנד (קיץ)", "HEPM": "שעון סנט פייר ומיקלון (קיץ)", "HEPMX": "שעון מערב מקסיקו (קיץ)", "HKST": "שעון הונג קונג (קיץ)", "HKT": "שעון הונג קונג (חורף)", "HNCU": "שעון קובה (חורף)", "HNEG": "שעון מזרח גרינלנד (חורף)", "HNNOMX": "שעון צפון-מערב מקסיקו (חורף)", "HNOG": "שעון מערב גרינלנד (חורף)", "HNPM": "שעון סנט פייר ומיקלון (חורף)", "HNPMX": "שעון מערב מקסיקו (חורף)", "HNT": "שעון ניופאונדלנד (חורף)", "IST": "שעון הודו", "JDT": "שעון יפן (קיץ)", "JST": "שעון יפן (חורף)", "LHDT": "שעון אי הלורד האו (קיץ)", "LHST": "שעון אי הלורד האו (חורף)", "MDT": "שעון קיץ מקאו", "MESZ": "שעון מרכז אירופה (קיץ)", "MEZ": "שעון מרכז אירופה (חורף)", "MST": "שעון חורף מקאו", "MYT": "שעון מלזיה", "NZDT": "שעון ניו זילנד (קיץ)", "NZST": "שעון ניו זילנד (חורף)", "OESZ": "שעון מזרח אירופה (קיץ)", "OEZ": "שעון מזרח אירופה (חורף)", "PDT": "שעון מערב ארה״ב (קיץ)", "PST": "שעון מערב ארה״ב (חורף)", "SAST": "שעון דרום אפריקה", "SGT": "שעון סינגפור", "SRT": "שעון סורינאם", "TMST": "שעון טורקמניסטן (קיץ)", "TMT": "שעון טורקמניסטן (חורף)", "UYST": "שעון אורוגוואי (קיץ)", "UYT": "שעון אורוגוואי (חורף)", "VET": "שעון ונצואלה", "WARST": "שעון מערב ארגנטינה (קיץ)", "WART": "שעון מערב ארגנטינה (חורף)", "WAST": "שעון מערב אפריקה (קיץ)", "WAT": "שעון מערב אפריקה (חורף)", "WESZ": "שעון מערב אירופה (קיץ)", "WEZ": "שעון מערב אירופה (חורף)", "WIB": "שעון מערב אינדונזיה", "WIT": "שעון מזרח אינדונזיה", "WITA": "שעון מרכז אינדונזיה", "∅∅∅": "שעון פרו (קיץ)"},

			fnCardinalPluralRule:  fnCardinalPluralRule_cf4ba4e3714ec1726c70da66a7b66698,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ac8c216759a27ae9294265517cfd4549,
			fnFmtCurrency:         fnFmtCurrency_dcd95366ecd0baffaf3fa998dfd8ecbc,
			fnFmtDateFull:         fnFmtDateFull_61caddbcb5db4de051fdc4ba779a33da,
			fnFmtDateLong:         fnFmtDateLong_e3113833817389f6b233e285c24e687d,
			fnFmtDateMedium:       fnFmtDateMedium_2a623cab0baa2e039b73232b6b2d515a,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_8a34c21e88caccc3a8728bf464f33cbd,
			fnFmtPercent:          fnFmtPercent_d2f2a59417e049449b3bb6986a239e7b,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_11b8a4f474d210e1d71b265f1f8e62d0,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["he_il"] = func() locales.Translator {
		return &localen{

			locale:                 "he_IL",
			pluralsCardinal:        []locales.PluralRule{2, 3, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{5, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "‎-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ינו׳", "פבר׳", "מרץ", "אפר׳", "מאי", "יוני", "יולי", "אוג׳", "ספט׳", "אוק׳", "נוב׳", "דצמ׳"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ינואר", "פברואר", "מרץ", "אפריל", "מאי", "יוני", "יולי", "אוגוסט", "ספטמבר", "אוקטובר", "נובמבר", "דצמבר"},
			daysAbbreviated:        []string{"יום א׳", "יום ב׳", "יום ג׳", "יום ד׳", "יום ה׳", "יום ו׳", "שבת"},
			daysNarrow:             []string{"א׳", "ב׳", "ג׳", "ד׳", "ה׳", "ו׳", "ש׳"},
			daysShort:              []string{"א׳", "ב׳", "ג׳", "ד׳", "ה׳", "ו׳", "ש׳"},
			daysWide:               []string{"יום ראשון", "יום שני", "יום שלישי", "יום רביעי", "יום חמישי", "יום שישי", "יום שבת"},
			periodsAbbreviated:     []string{"לפנה״צ", "אחה״צ"},
			periodsNarrow:          []string{"לפנה״צ", "אחה״צ"},
			periodsWide:            []string{"לפנה״צ", "אחה״צ"},
			erasAbbreviated:        []string{"לפנה״ס", "לספירה"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"לפני הספירה", "לספירה"},
			timezones:              map[string]string{"ACDT": "שעון מרכז אוסטרליה (קיץ)", "ACST": "שעון מרכז אוסטרליה (חורף)", "ACWDT": "שעון מרכז-מערב אוסטרליה (קיץ)", "ACWST": "שעון מרכז-מערב אוסטרליה (חורף)", "ADT": "שעון האוקיינוס האטלנטי (קיץ)", "AEDT": "שעון מזרח אוסטרליה (קיץ)", "AEST": "שעון מזרח אוסטרליה (חורף)", "AKDT": "שעון אלסקה (קיץ)", "AKST": "שעון אלסקה (חורף)", "ARST": "שעון ארגנטינה (קיץ)", "ART": "שעון ארגנטינה (חורף)", "AST": "שעון האוקיינוס האטלנטי (חורף)", "AWDT": "שעון מערב אוסטרליה (קיץ)", "AWST": "שעון מערב אוסטרליה (חורף)", "BOT": "שעון בוליביה", "BT": "שעון בהוטן", "CAT": "שעון מרכז אפריקה", "CDT": "שעון מרכז ארה״ב (קיץ)", "CHADT": "שעון צ׳טהאם (קיץ)", "CHAST": "שעון צ׳טהאם (חורף)", "CLST": "שעון צ׳ילה (קיץ)", "CLT": "שעון צ׳ילה (חורף)", "COST": "שעון קולומביה (קיץ)", "COT": "שעון קולומביה (חורף)", "CST": "שעון מרכז ארה״ב (חורף)", "ChST": "שעון צ׳אמורו", "EAT": "שעון מזרח אפריקה", "ECT": "שעון אקוודור", "EDT": "שעון החוף המזרחי (קיץ)", "EST": "שעון החוף המזרחי (חורף)", "GFT": "שעון גיאנה הצרפתית", "GMT": "שעון גריניץ׳\u200f", "GST": "שעון מדינות המפרץ", "GYT": "שעון גיאנה", "HADT": "שעון האיים האלאוטיים הוואי (קיץ)", "HAST": "שעון האיים האלאוטיים הוואי (חורף)", "HAT": "שעון ניופאונדלנד (קיץ)", "HECU": "שעון קובה (קיץ)", "HEEG": "שעון מזרח גרינלנד (קיץ)", "HENOMX": "שעון צפון-מערב מקסיקו (קיץ)", "HEOG": "שעון מערב גרינלנד (קיץ)", "HEPM": "שעון סנט פייר ומיקלון (קיץ)", "HEPMX": "שעון מערב מקסיקו (קיץ)", "HKST": "שעון הונג קונג (קיץ)", "HKT": "שעון הונג קונג (חורף)", "HNCU": "שעון קובה (חורף)", "HNEG": "שעון מזרח גרינלנד (חורף)", "HNNOMX": "שעון צפון-מערב מקסיקו (חורף)", "HNOG": "שעון מערב גרינלנד (חורף)", "HNPM": "שעון סנט פייר ומיקלון (חורף)", "HNPMX": "שעון מערב מקסיקו (חורף)", "HNT": "שעון ניופאונדלנד (חורף)", "IST": "שעון הודו", "JDT": "שעון יפן (קיץ)", "JST": "שעון יפן (חורף)", "LHDT": "שעון אי הלורד האו (קיץ)", "LHST": "שעון אי הלורד האו (חורף)", "MDT": "שעון אזור ההרים בארה״ב (קיץ)", "MESZ": "שעון מרכז אירופה (קיץ)", "MEZ": "שעון מרכז אירופה (חורף)", "MST": "שעון אזור ההרים בארה״ב (חורף)", "MYT": "שעון מלזיה", "NZDT": "שעון ניו זילנד (קיץ)", "NZST": "שעון ניו זילנד (חורף)", "OESZ": "שעון מזרח אירופה (קיץ)", "OEZ": "שעון מזרח אירופה (חורף)", "PDT": "שעון מערב ארה״ב (קיץ)", "PST": "שעון מערב ארה״ב (חורף)", "SAST": "שעון דרום אפריקה", "SGT": "שעון סינגפור", "SRT": "שעון סורינאם", "TMST": "שעון טורקמניסטן (קיץ)", "TMT": "שעון טורקמניסטן (חורף)", "UYST": "שעון אורוגוואי (קיץ)", "UYT": "שעון אורוגוואי (חורף)", "VET": "שעון ונצואלה", "WARST": "שעון מערב ארגנטינה (קיץ)", "WART": "שעון מערב ארגנטינה (חורף)", "WAST": "שעון מערב אפריקה (קיץ)", "WAT": "שעון מערב אפריקה (חורף)", "WESZ": "שעון מערב אירופה (קיץ)", "WEZ": "שעון מערב אירופה (חורף)", "WIB": "שעון מערב אינדונזיה", "WIT": "שעון מזרח אינדונזיה", "WITA": "שעון מרכז אינדונזיה", "∅∅∅": "שעון ברזיליה (קיץ)"},

			fnCardinalPluralRule:  fnCardinalPluralRule_cf4ba4e3714ec1726c70da66a7b66698,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ac8c216759a27ae9294265517cfd4549,
			fnFmtCurrency:         fnFmtCurrency_dcd95366ecd0baffaf3fa998dfd8ecbc,
			fnFmtDateFull:         fnFmtDateFull_61caddbcb5db4de051fdc4ba779a33da,
			fnFmtDateLong:         fnFmtDateLong_e3113833817389f6b233e285c24e687d,
			fnFmtDateMedium:       fnFmtDateMedium_2a623cab0baa2e039b73232b6b2d515a,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_8a34c21e88caccc3a8728bf464f33cbd,
			fnFmtPercent:          fnFmtPercent_d2f2a59417e049449b3bb6986a239e7b,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_11b8a4f474d210e1d71b265f1f8e62d0,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hi"] = func() locales.Translator {
		return &localen{

			locale:             "hi",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "जन॰", "फ़र॰", "मार्च", "अप्रैल", "मई", "जून", "जुल॰", "अग॰", "सित॰", "अक्तू॰", "नव॰", "दिस॰"},
			monthsNarrow:       []string{"", "ज", "फ़", "मा", "अ", "म", "जू", "जु", "अ", "सि", "अ", "न", "दि"},
			monthsWide:         []string{"", "जनवरी", "फ़रवरी", "मार्च", "अप्रैल", "मई", "जून", "जुलाई", "अगस्त", "सितंबर", "अक्तूबर", "नवंबर", "दिसंबर"},
			daysAbbreviated:    []string{"रवि", "सोम", "मंगल", "बुध", "गुरु", "शुक्र", "शनि"},
			daysNarrow:         []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysShort:          []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysWide:           []string{"रविवार", "सोमवार", "मंगलवार", "बुधवार", "गुरुवार", "शुक्रवार", "शनिवार"},
			periodsAbbreviated: []string{"am", "pm"},
			periodsNarrow:      []string{"am", "pm"},
			periodsWide:        []string{"am", "pm"},
			erasAbbreviated:    []string{"ईसा-पूर्व", "ईस्वी"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ईसा-पूर्व", "ईसवी सन"},
			timezones:          map[string]string{"ACDT": "ऑस्\u200dट्रेलियाई केंद्रीय डेलाइट समय", "ACST": "ऑस्\u200dट्रेलियाई केंद्रीय मानक समय", "ACWDT": "ऑस्\u200dट्रेलियाई केंद्रीय पश्चिमी डेलाइट समय", "ACWST": "ऑस्\u200dट्रेलियाई केंद्रीय पश्चिमी मानक समय", "ADT": "अटलांटिक डेलाइट समय", "AEDT": "ऑस्\u200dट्रेलियाई पूर्वी डेलाइट समय", "AEST": "ऑस्\u200dट्रेलियाई पूर्वी मानक समय", "AKDT": "अलास्\u200dका डेलाइट समय", "AKST": "अलास्\u200dका मानक समय", "ARST": "अर्जेंटीना ग्रीष्मकालीन समय", "ART": "अर्जेंटीना मानक समय", "AST": "अटलांटिक मानक समय", "AWDT": "ऑस्ट्रेलियाई पश्चिमी डेलाइट समय", "AWST": "ऑस्ट्रेलियाई पश्चिमी मानक समय", "BOT": "बोलीविया समय", "BT": "भूटान समय", "CAT": "मध्य अफ़्रीका समय", "CDT": "उत्तरी अमेरिकी केंद्रीय डेलाइट समय", "CHADT": "चैथम डेलाइट समय", "CHAST": "चैथम मानक समय", "CLST": "चिली ग्रीष्मकालीन समय", "CLT": "चिली मानक समय", "COST": "कोलंबिया ग्रीष्मकालीन समय", "COT": "कोलंबिया मानक समय", "CST": "उत्तरी अमेरिकी केंद्रीय मानक समय", "ChST": "चामोरो मानक समय", "EAT": "पूर्वी अफ़्रीका समय", "ECT": "इक्वाडोर समय", "EDT": "उत्तरी अमेरिकी पूर्वी डेलाइट समय", "EST": "उत्तरी अमेरिकी पूर्वी मानक समय", "GFT": "फ़्रेंच गुयाना समय", "GMT": "ग्रीनविच मीन टाइम", "GST": "खाड़ी मानक समय", "GYT": "गुयाना समय", "HADT": "हवाई–आल्यूशन डेलाइट समय", "HAST": "हवाई–आल्यूशन मानक समय", "HAT": "न्यूफ़ाउंडलैंड डेलाइट समय", "HECU": "क्यूबा डेलाइट समय", "HEEG": "पूर्वी ग्रीनलैंड ग्रीष्मकालीन समय", "HENOMX": "उत्तर पश्चिमी मेक्सिको डेलाइट समय", "HEOG": "पश्चिमी ग्रीनलैंड ग्रीष्मकालीन समय", "HEPM": "सेंट पिएरे और मिक्वेलान डेलाइट समय", "HEPMX": "मेक्सिकन प्रशांत डेलाइट समय", "HKST": "हाँग काँग ग्रीष्मकालीन समय", "HKT": "हाँग काँग मानक समय", "HNCU": "क्यूबा मानक समय", "HNEG": "पूर्वी ग्रीनलैंड मानक समय", "HNNOMX": "उत्तर पश्चिमी मेक्सिको मानक समय", "HNOG": "पश्चिमी ग्रीनलैंड मानक समय", "HNPM": "सेंट पिएरे और मिक्वेलान मानक समय", "HNPMX": "मेक्सिकन प्रशांत मानक समय", "HNT": "न्यूफ़ाउंडलैंड मानक समय", "IST": "भारतीय मानक समय", "JDT": "जापान डेलाइट समय", "JST": "जापान मानक समय", "LHDT": "लॉर्ड होवे डेलाइट समय", "LHST": "लॉर्ड होवे मानक समय", "MDT": "उत्तरी अमेरिकी माउंटेन डेलाइट समय", "MESZ": "मध्\u200dय यूरोपीय ग्रीष्\u200dमकालीन समय", "MEZ": "मध्य यूरोपीय मानक समय", "MST": "उत्तरी अमेरिकी माउंटेन मानक समय", "MYT": "मलेशिया समय", "NZDT": "न्यूज़ीलैंड डेलाइट समय", "NZST": "न्यूज़ीलैंड मानक समय", "OESZ": "पूर्वी यूरोपीय ग्रीष्मकालीन समय", "OEZ": "पूर्वी यूरोपीय मानक समय", "PDT": "उत्तरी अमेरिकी प्रशांत डेलाइट समय", "PST": "उत्तरी अमेरिकी प्रशांत मानक समय", "SAST": "दक्षिण अफ़्रीका मानक समय", "SGT": "सिंगापुर समय", "SRT": "सूरीनाम समय", "TMST": "तुर्कमेनिस्तान ग्रीष्मकालीन समय", "TMT": "तुर्कमेनिस्तान मानक समय", "UYST": "उरुग्वे ग्रीष्मकालीन समय", "UYT": "उरुग्वे मानक समय", "VET": "वेनेज़ुएला समय", "WARST": "पश्चिमी अर्जेंटीना ग्रीष्मकालीन समय", "WART": "पश्चिमी अर्जेंटीना मानक समय", "WAST": "पश्चिम अफ़्रीका ग्रीष्मकालीन समय", "WAT": "पश्चिम अफ़्रीका मानक समय", "WESZ": "पश्चिमी यूरोपीय ग्रीष्\u200dमकालीन समय", "WEZ": "पश्चिमी यूरोपीय मानक समय", "WIB": "पश्चिमी इंडोनेशिया समय", "WIT": "पूर्वी इंडोनेशिया समय", "WITA": "मध्य इंडोनेशिया समय", "∅∅∅": "ब्राज़ीलिया ग्रीष्मकालीन समय"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a9f120c4a8c40112dec183cd65a8b424,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hi_in"] = func() locales.Translator {
		return &localen{

			locale:             "hi_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "जन॰", "फ़र॰", "मार्च", "अप्रैल", "मई", "जून", "जुल॰", "अग॰", "सित॰", "अक्तू॰", "नव॰", "दिस॰"},
			monthsNarrow:       []string{"", "ज", "फ़", "मा", "अ", "म", "जू", "जु", "अ", "सि", "अ", "न", "दि"},
			monthsWide:         []string{"", "जनवरी", "फ़रवरी", "मार्च", "अप्रैल", "मई", "जून", "जुलाई", "अगस्त", "सितंबर", "अक्तूबर", "नवंबर", "दिसंबर"},
			daysAbbreviated:    []string{"रवि", "सोम", "मंगल", "बुध", "गुरु", "शुक्र", "शनि"},
			daysNarrow:         []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysShort:          []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysWide:           []string{"रविवार", "सोमवार", "मंगलवार", "बुधवार", "गुरुवार", "शुक्रवार", "शनिवार"},
			periodsAbbreviated: []string{"am", "pm"},
			periodsNarrow:      []string{"am", "pm"},
			periodsWide:        []string{"am", "pm"},
			erasAbbreviated:    []string{"ईसा-पूर्व", "ईस्वी"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ईसा-पूर्व", "ईसवी सन"},
			timezones:          map[string]string{"ACDT": "ऑस्\u200dट्रेलियाई केंद्रीय डेलाइट समय", "ACST": "ऑस्\u200dट्रेलियाई केंद्रीय मानक समय", "ACWDT": "ऑस्\u200dट्रेलियाई केंद्रीय पश्चिमी डेलाइट समय", "ACWST": "ऑस्\u200dट्रेलियाई केंद्रीय पश्चिमी मानक समय", "ADT": "अटलांटिक डेलाइट समय", "AEDT": "ऑस्\u200dट्रेलियाई पूर्वी डेलाइट समय", "AEST": "ऑस्\u200dट्रेलियाई पूर्वी मानक समय", "AKDT": "अलास्\u200dका डेलाइट समय", "AKST": "अलास्\u200dका मानक समय", "ARST": "अर्जेंटीना ग्रीष्मकालीन समय", "ART": "अर्जेंटीना मानक समय", "AST": "अटलांटिक मानक समय", "AWDT": "ऑस्ट्रेलियाई पश्चिमी डेलाइट समय", "AWST": "ऑस्ट्रेलियाई पश्चिमी मानक समय", "BOT": "बोलीविया समय", "BT": "भूटान समय", "CAT": "मध्य अफ़्रीका समय", "CDT": "उत्तरी अमेरिकी केंद्रीय डेलाइट समय", "CHADT": "चैथम डेलाइट समय", "CHAST": "चैथम मानक समय", "CLST": "चिली ग्रीष्मकालीन समय", "CLT": "चिली मानक समय", "COST": "कोलंबिया ग्रीष्मकालीन समय", "COT": "कोलंबिया मानक समय", "CST": "उत्तरी अमेरिकी केंद्रीय मानक समय", "ChST": "चामोरो मानक समय", "EAT": "पूर्वी अफ़्रीका समय", "ECT": "इक्वाडोर समय", "EDT": "उत्तरी अमेरिकी पूर्वी डेलाइट समय", "EST": "उत्तरी अमेरिकी पूर्वी मानक समय", "GFT": "फ़्रेंच गुयाना समय", "GMT": "ग्रीनविच मीन टाइम", "GST": "खाड़ी मानक समय", "GYT": "गुयाना समय", "HADT": "हवाई–आल्यूशन डेलाइट समय", "HAST": "हवाई–आल्यूशन मानक समय", "HAT": "न्यूफ़ाउंडलैंड डेलाइट समय", "HECU": "क्यूबा डेलाइट समय", "HEEG": "पूर्वी ग्रीनलैंड ग्रीष्मकालीन समय", "HENOMX": "उत्तर पश्चिमी मेक्सिको डेलाइट समय", "HEOG": "पश्चिमी ग्रीनलैंड ग्रीष्मकालीन समय", "HEPM": "सेंट पिएरे और मिक्वेलान डेलाइट समय", "HEPMX": "मेक्सिकन प्रशांत डेलाइट समय", "HKST": "हाँग काँग ग्रीष्मकालीन समय", "HKT": "हाँग काँग मानक समय", "HNCU": "क्यूबा मानक समय", "HNEG": "पूर्वी ग्रीनलैंड मानक समय", "HNNOMX": "उत्तर पश्चिमी मेक्सिको मानक समय", "HNOG": "पश्चिमी ग्रीनलैंड मानक समय", "HNPM": "सेंट पिएरे और मिक्वेलान मानक समय", "HNPMX": "मेक्सिकन प्रशांत मानक समय", "HNT": "न्यूफ़ाउंडलैंड मानक समय", "IST": "भारतीय मानक समय", "JDT": "जापान डेलाइट समय", "JST": "जापान मानक समय", "LHDT": "लॉर्ड होवे डेलाइट समय", "LHST": "लॉर्ड होवे मानक समय", "MDT": "उत्तरी अमेरिकी माउंटेन डेलाइट समय", "MESZ": "मध्\u200dय यूरोपीय ग्रीष्\u200dमकालीन समय", "MEZ": "मध्य यूरोपीय मानक समय", "MST": "उत्तरी अमेरिकी माउंटेन मानक समय", "MYT": "मलेशिया समय", "NZDT": "न्यूज़ीलैंड डेलाइट समय", "NZST": "न्यूज़ीलैंड मानक समय", "OESZ": "पूर्वी यूरोपीय ग्रीष्मकालीन समय", "OEZ": "पूर्वी यूरोपीय मानक समय", "PDT": "उत्तरी अमेरिकी प्रशांत डेलाइट समय", "PST": "उत्तरी अमेरिकी प्रशांत मानक समय", "SAST": "दक्षिण अफ़्रीका मानक समय", "SGT": "सिंगापुर समय", "SRT": "सूरीनाम समय", "TMST": "तुर्कमेनिस्तान ग्रीष्मकालीन समय", "TMT": "तुर्कमेनिस्तान मानक समय", "UYST": "उरुग्वे ग्रीष्मकालीन समय", "UYT": "उरुग्वे मानक समय", "VET": "वेनेज़ुएला समय", "WARST": "पश्चिमी अर्जेंटीना ग्रीष्मकालीन समय", "WART": "पश्चिमी अर्जेंटीना मानक समय", "WAST": "पश्चिम अफ़्रीका ग्रीष्मकालीन समय", "WAT": "पश्चिम अफ़्रीका मानक समय", "WESZ": "पश्चिमी यूरोपीय ग्रीष्\u200dमकालीन समय", "WEZ": "पश्चिमी यूरोपीय मानक समय", "WIB": "पश्चिमी इंडोनेशिया समय", "WIT": "पूर्वी इंडोनेशिया समय", "WITA": "मध्य इंडोनेशिया समय", "∅∅∅": "ब्राज़ीलिया ग्रीष्मकालीन समय"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a9f120c4a8c40112dec183cd65a8b424,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hr"] = func() locales.Translator {
		return &localen{

			locale:                 "hr",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sij", "velj", "ožu", "tra", "svi", "lip", "srp", "kol", "ruj", "lis", "stu", "pro"},
			monthsNarrow:           []string{"", "1.", "2.", "3.", "4.", "5.", "6.", "7.", "8.", "9.", "10.", "11.", "12."},
			monthsWide:             []string{"", "siječnja", "veljače", "ožujka", "travnja", "svibnja", "lipnja", "srpnja", "kolovoza", "rujna", "listopada", "studenoga", "prosinca"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"pr. Kr.", "po. Kr."},
			erasNarrow:             []string{"pr.n.e.", "AD"},
			erasWide:               []string{"prije Krista", "poslije Krista"},
			timezones:              map[string]string{"ACDT": "srednjoaustralsko ljetno vrijeme", "ACST": "srednjoaustralsko standardno vrijeme", "ACWDT": "australsko središnje zapadno ljetno vrijeme", "ACWST": "australsko središnje zapadno standardno vrijeme", "ADT": "atlantsko ljetno vrijeme", "AEDT": "istočnoaustralsko ljetno vrijeme", "AEST": "istočnoaustralsko standardno vrijeme", "AKDT": "aljaško ljetno vrijeme", "AKST": "aljaško standardno vrijeme", "ARST": "argentinsko ljetno vrijeme", "ART": "argentinsko standardno vrijeme", "AST": "atlantsko standardno vrijeme", "AWDT": "zapadnoaustralsko ljetno vrijeme", "AWST": "zapadnoaustralsko standardno vrijeme", "BOT": "bolivijsko vrijeme", "BT": "butansko vrijeme", "CAT": "srednjoafričko vrijeme", "CDT": "središnje ljetno vrijeme", "CHADT": "ljetno vrijeme Chathama", "CHAST": "standardno vrijeme Chathama", "CLST": "čileansko ljetno vrijeme", "CLT": "čileansko standardno vrijeme", "COST": "kolumbijsko ljetno vrijeme", "COT": "kolumbijsko standardno vrijeme", "CST": "središnje standardno vrijeme", "ChST": "standardno vrijeme Chamorra", "EAT": "istočnoafričko vrijeme", "ECT": "ekvadorsko vrijeme", "EDT": "istočno ljetno vrijeme", "EST": "istočno standardno vrijeme", "GFT": "vrijeme Francuske Gijane", "GMT": "univerzalno vrijeme", "GST": "zaljevsko standardno vrijeme", "GYT": "gvajansko vrijeme", "HADT": "havajsko-aleutsko ljetno vrijeme", "HAST": "havajsko-aleutsko standardno vrijeme", "HAT": "newfoundlandsko ljetno vrijeme", "HECU": "kubansko ljetno vrijeme", "HEEG": "istočnogrenlandsko ljetno vrijeme", "HENOMX": "sjeverozapadno meksičko ljetno vrijeme", "HEOG": "zapadnogrenlandsko ljetno vrijeme", "HEPM": "ljetno vrijeme za Sveti Petar i Mikelon", "HEPMX": "meksičko pacifičko ljetno vrijeme", "HKST": "hongkonško ljetno vrijeme", "HKT": "hongkonško standardno vrijeme", "HNCU": "kubansko standardno vrijeme", "HNEG": "istočnogrenlandsko standardno vrijeme", "HNNOMX": "sjeverozapadno meksičko standardno vrijeme", "HNOG": "zapadnogrenlandsko standardno vrijeme", "HNPM": "standardno vrijeme za Sveti Petar i Mikelon", "HNPMX": "meksičko pacifičko standardno vrijeme", "HNT": "newfoundlandsko standardno vrijeme", "IST": "indijsko vrijeme", "JDT": "japansko ljetno vrijeme", "JST": "japansko standardno vrijeme", "LHDT": "ljetno vrijeme otoka Lord Howe", "LHST": "standardno vrijeme otoka Lord Howe", "MDT": "planinsko ljetno vrijeme", "MESZ": "srednjoeuropsko ljetno vrijeme", "MEZ": "srednjoeuropsko standardno vrijeme", "MST": "planinsko standardno vrijeme", "MYT": "malezijsko vrijeme", "NZDT": "novozelandsko ljetno vrijeme", "NZST": "novozelandsko standardno vrijeme", "OESZ": "istočnoeuropsko ljetno vrijeme", "OEZ": "istočnoeuropsko standardno vrijeme", "PDT": "pacifičko ljetno vrijeme", "PST": "pacifičko standardno vrijeme", "SAST": "južnoafričko vrijeme", "SGT": "singapursko vrijeme", "SRT": "surinamsko vrijeme", "TMST": "turkmenistansko ljetno vrijeme", "TMT": "turkmenistansko standardno vrijeme", "UYST": "urugvajsko ljetno vrijeme", "UYT": "urugvajsko standardno vrijeme", "VET": "venezuelsko vrijeme", "WARST": "zapadnoargentinsko ljetno vrijeme", "WART": "zapadnoargentinsko standardno vrijeme", "WAST": "zapadnoafričko ljetno vrijeme", "WAT": "zapadnoafričko standardno vrijeme", "WESZ": "zapadnoeuropsko ljetno vrijeme", "WEZ": "zapadnoeuropsko standardno vrijeme", "WIB": "zapadnoindonezijsko vrijeme", "WIT": "istočnoindonezijsko vrijeme", "WITA": "srednjoindonezijsko vrijeme", "∅∅∅": "brazilijsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_682acc4ae60b52f9983d14c85913c784,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hr_ba"] = func() locales.Translator {
		return &localen{

			locale:                 "hr_BA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "KM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sij", "velj", "ožu", "tra", "svi", "lip", "srp", "kol", "ruj", "lis", "stu", "pro"},
			monthsNarrow:           []string{"", "1.", "2.", "3.", "4.", "5.", "6.", "7.", "8.", "9.", "10.", "11.", "12."},
			monthsWide:             []string{"", "siječnja", "veljače", "ožujka", "travnja", "svibnja", "lipnja", "srpnja", "kolovoza", "rujna", "listopada", "studenoga", "prosinca"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"pr. Kr.", "po. Kr."},
			erasNarrow:             []string{"pr.n.e.", "AD"},
			erasWide:               []string{"prije Krista", "poslije Krista"},
			timezones:              map[string]string{"ACDT": "srednjoaustralsko ljetno vrijeme", "ACST": "srednjoaustralsko standardno vrijeme", "ACWDT": "australsko središnje zapadno ljetno vrijeme", "ACWST": "australsko središnje zapadno standardno vrijeme", "ADT": "atlantsko ljetno vrijeme", "AEDT": "istočnoaustralsko ljetno vrijeme", "AEST": "istočnoaustralsko standardno vrijeme", "AKDT": "aljaško ljetno vrijeme", "AKST": "aljaško standardno vrijeme", "ARST": "argentinsko ljetno vrijeme", "ART": "argentinsko standardno vrijeme", "AST": "atlantsko standardno vrijeme", "AWDT": "zapadnoaustralsko ljetno vrijeme", "AWST": "zapadnoaustralsko standardno vrijeme", "BOT": "bolivijsko vrijeme", "BT": "butansko vrijeme", "CAT": "srednjoafričko vrijeme", "CDT": "središnje ljetno vrijeme", "CHADT": "ljetno vrijeme Chathama", "CHAST": "standardno vrijeme Chathama", "CLST": "čileansko ljetno vrijeme", "CLT": "čileansko standardno vrijeme", "COST": "kolumbijsko ljetno vrijeme", "COT": "kolumbijsko standardno vrijeme", "CST": "središnje standardno vrijeme", "ChST": "standardno vrijeme Chamorra", "EAT": "istočnoafričko vrijeme", "ECT": "ekvadorsko vrijeme", "EDT": "istočno ljetno vrijeme", "EST": "istočno standardno vrijeme", "GFT": "vrijeme Francuske Gijane", "GMT": "univerzalno vrijeme", "GST": "zaljevsko standardno vrijeme", "GYT": "gvajansko vrijeme", "HADT": "havajsko-aleutsko ljetno vrijeme", "HAST": "havajsko-aleutsko standardno vrijeme", "HAT": "newfoundlandsko ljetno vrijeme", "HECU": "kubansko ljetno vrijeme", "HEEG": "istočnogrenlandsko ljetno vrijeme", "HENOMX": "sjeverozapadno meksičko ljetno vrijeme", "HEOG": "zapadnogrenlandsko ljetno vrijeme", "HEPM": "ljetno vrijeme za Sveti Petar i Mikelon", "HEPMX": "meksičko pacifičko ljetno vrijeme", "HKST": "hongkonško ljetno vrijeme", "HKT": "hongkonško standardno vrijeme", "HNCU": "kubansko standardno vrijeme", "HNEG": "istočnogrenlandsko standardno vrijeme", "HNNOMX": "sjeverozapadno meksičko standardno vrijeme", "HNOG": "zapadnogrenlandsko standardno vrijeme", "HNPM": "standardno vrijeme za Sveti Petar i Mikelon", "HNPMX": "meksičko pacifičko standardno vrijeme", "HNT": "newfoundlandsko standardno vrijeme", "IST": "indijsko vrijeme", "JDT": "japansko ljetno vrijeme", "JST": "japansko standardno vrijeme", "LHDT": "ljetno vrijeme otoka Lord Howe", "LHST": "standardno vrijeme otoka Lord Howe", "MDT": "planinsko ljetno vrijeme", "MESZ": "srednjoeuropsko ljetno vrijeme", "MEZ": "srednjoeuropsko standardno vrijeme", "MST": "planinsko standardno vrijeme", "MYT": "malezijsko vrijeme", "NZDT": "novozelandsko ljetno vrijeme", "NZST": "novozelandsko standardno vrijeme", "OESZ": "istočnoeuropsko ljetno vrijeme", "OEZ": "istočnoeuropsko standardno vrijeme", "PDT": "pacifičko ljetno vrijeme", "PST": "pacifičko standardno vrijeme", "SAST": "južnoafričko vrijeme", "SGT": "singapursko vrijeme", "SRT": "surinamsko vrijeme", "TMST": "turkmenistansko ljetno vrijeme", "TMT": "turkmenistansko standardno vrijeme", "UYST": "urugvajsko ljetno vrijeme", "UYT": "urugvajsko standardno vrijeme", "VET": "venezuelsko vrijeme", "WARST": "zapadnoargentinsko ljetno vrijeme", "WART": "zapadnoargentinsko standardno vrijeme", "WAST": "zapadnoafričko ljetno vrijeme", "WAT": "zapadnoafričko standardno vrijeme", "WESZ": "zapadnoeuropsko ljetno vrijeme", "WEZ": "zapadnoeuropsko standardno vrijeme", "WIB": "zapadnoindonezijsko vrijeme", "WIT": "istočnoindonezijsko vrijeme", "WITA": "srednjoindonezijsko vrijeme", "∅∅∅": "brazilijsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_474bf8d71d0392f03d943b2085e8e536,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hr_hr"] = func() locales.Translator {
		return &localen{

			locale:                 "hr_HR",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "sij", "velj", "ožu", "tra", "svi", "lip", "srp", "kol", "ruj", "lis", "stu", "pro"},
			monthsNarrow:           []string{"", "1.", "2.", "3.", "4.", "5.", "6.", "7.", "8.", "9.", "10.", "11.", "12."},
			monthsWide:             []string{"", "siječnja", "veljače", "ožujka", "travnja", "svibnja", "lipnja", "srpnja", "kolovoza", "rujna", "listopada", "studenoga", "prosinca"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysNarrow:             []string{"N", "P", "U", "S", "Č", "P", "S"},
			daysShort:              []string{"ned", "pon", "uto", "sri", "čet", "pet", "sub"},
			daysWide:               []string{"nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"pr. Kr.", "po. Kr."},
			erasNarrow:             []string{"pr.n.e.", "AD"},
			erasWide:               []string{"prije Krista", "poslije Krista"},
			timezones:              map[string]string{"ACDT": "srednjoaustralsko ljetno vrijeme", "ACST": "srednjoaustralsko standardno vrijeme", "ACWDT": "australsko središnje zapadno ljetno vrijeme", "ACWST": "australsko središnje zapadno standardno vrijeme", "ADT": "atlantsko ljetno vrijeme", "AEDT": "istočnoaustralsko ljetno vrijeme", "AEST": "istočnoaustralsko standardno vrijeme", "AKDT": "aljaško ljetno vrijeme", "AKST": "aljaško standardno vrijeme", "ARST": "argentinsko ljetno vrijeme", "ART": "argentinsko standardno vrijeme", "AST": "atlantsko standardno vrijeme", "AWDT": "zapadnoaustralsko ljetno vrijeme", "AWST": "zapadnoaustralsko standardno vrijeme", "BOT": "bolivijsko vrijeme", "BT": "butansko vrijeme", "CAT": "srednjoafričko vrijeme", "CDT": "središnje ljetno vrijeme", "CHADT": "ljetno vrijeme Chathama", "CHAST": "standardno vrijeme Chathama", "CLST": "čileansko ljetno vrijeme", "CLT": "čileansko standardno vrijeme", "COST": "kolumbijsko ljetno vrijeme", "COT": "kolumbijsko standardno vrijeme", "CST": "središnje standardno vrijeme", "ChST": "standardno vrijeme Chamorra", "EAT": "istočnoafričko vrijeme", "ECT": "ekvadorsko vrijeme", "EDT": "istočno ljetno vrijeme", "EST": "istočno standardno vrijeme", "GFT": "vrijeme Francuske Gijane", "GMT": "univerzalno vrijeme", "GST": "zaljevsko standardno vrijeme", "GYT": "gvajansko vrijeme", "HADT": "havajsko-aleutsko ljetno vrijeme", "HAST": "havajsko-aleutsko standardno vrijeme", "HAT": "newfoundlandsko ljetno vrijeme", "HECU": "kubansko ljetno vrijeme", "HEEG": "istočnogrenlandsko ljetno vrijeme", "HENOMX": "sjeverozapadno meksičko ljetno vrijeme", "HEOG": "zapadnogrenlandsko ljetno vrijeme", "HEPM": "ljetno vrijeme za Sveti Petar i Mikelon", "HEPMX": "meksičko pacifičko ljetno vrijeme", "HKST": "hongkonško ljetno vrijeme", "HKT": "hongkonško standardno vrijeme", "HNCU": "kubansko standardno vrijeme", "HNEG": "istočnogrenlandsko standardno vrijeme", "HNNOMX": "sjeverozapadno meksičko standardno vrijeme", "HNOG": "zapadnogrenlandsko standardno vrijeme", "HNPM": "standardno vrijeme za Sveti Petar i Mikelon", "HNPMX": "meksičko pacifičko standardno vrijeme", "HNT": "newfoundlandsko standardno vrijeme", "IST": "indijsko vrijeme", "JDT": "japansko ljetno vrijeme", "JST": "japansko standardno vrijeme", "LHDT": "ljetno vrijeme otoka Lord Howe", "LHST": "standardno vrijeme otoka Lord Howe", "MDT": "planinsko ljetno vrijeme", "MESZ": "srednjoeuropsko ljetno vrijeme", "MEZ": "srednjoeuropsko standardno vrijeme", "MST": "planinsko standardno vrijeme", "MYT": "malezijsko vrijeme", "NZDT": "novozelandsko ljetno vrijeme", "NZST": "novozelandsko standardno vrijeme", "OESZ": "istočnoeuropsko ljetno vrijeme", "OEZ": "istočnoeuropsko standardno vrijeme", "PDT": "pacifičko ljetno vrijeme", "PST": "pacifičko standardno vrijeme", "SAST": "južnoafričko vrijeme", "SGT": "singapursko vrijeme", "SRT": "surinamsko vrijeme", "TMST": "turkmenistansko ljetno vrijeme", "TMT": "turkmenistansko standardno vrijeme", "UYST": "urugvajsko ljetno vrijeme", "UYT": "urugvajsko standardno vrijeme", "VET": "venezuelsko vrijeme", "WARST": "zapadnoargentinsko ljetno vrijeme", "WART": "zapadnoargentinsko standardno vrijeme", "WAST": "zapadnoafričko ljetno vrijeme", "WAT": "zapadnoafričko standardno vrijeme", "WESZ": "zapadnoeuropsko ljetno vrijeme", "WEZ": "zapadnoeuropsko standardno vrijeme", "WIB": "zapadnoindonezijsko vrijeme", "WIT": "istočnoindonezijsko vrijeme", "WITA": "srednjoindonezijsko vrijeme", "∅∅∅": "azorsko ljetno vrijeme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_5aef9dacdb26bf87063b666a034b424e,
			fnFmtDateLong:         fnFmtDateLong_65f6b60e7b56b6fcb221b95f5d7324ae,
			fnFmtDateMedium:       fnFmtDateMedium_a05abc83947bad9ca5156c8f2aa83cdc,
			fnFmtDateShort:        fnFmtDateShort_682acc4ae60b52f9983d14c85913c784,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hsb"] = func() locales.Translator {
		return &localen{

			locale:                 "hsb",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "zł", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "měr.", "apr.", "mej.", "jun.", "jul.", "awg.", "sep.", "okt.", "now.", "dec."},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januara", "februara", "měrca", "apryla", "meje", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"},
			daysAbbreviated:        []string{"nje", "pón", "wut", "srj", "štw", "pja", "sob"},
			daysNarrow:             []string{"n", "p", "w", "s", "š", "p", "s"},
			daysShort:              []string{"nj", "pó", "wu", "sr", "št", "pj", "so"},
			daysWide:               []string{"njedźela", "póndźela", "wutora", "srjeda", "štwórtk", "pjatk", "sobota"},
			periodsAbbreviated:     []string{"dopołdnja", "popołdnju"},
			periodsNarrow:          []string{"dop.", "pop."},
			periodsWide:            []string{"dopołdnja", "popołdnju"},
			erasAbbreviated:        []string{"př.Chr.n.", "po Chr.n."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"před Chrystowym narodźenjom", "po Chrystowym narodźenju"},
			timezones:              map[string]string{"ACDT": "srjedźoawstralski lětni čas", "ACST": "srjedźoawstralski standardny čas", "ACWDT": "sjedźozapadny awstralski lětni čas", "ACWST": "srjedźozapadny awstralski standardny čas", "ADT": "atlantiski lětni čas", "AEDT": "wuchodoawstralski lětni čas", "AEST": "wuchodoawstralski standardny čas", "AKDT": "alaskaski lětni čas", "AKST": "alaskaski standardny čas", "ARST": "argentinski lětni čas", "ART": "argentinski standardny čas", "AST": "atlantiski standardny čas", "AWDT": "zapadoawstralski lětni čas", "AWST": "zapadoawstralski standardny čas", "BOT": "boliwiski čas", "BT": "bhutanski čas", "CAT": "centralnoafriski čas", "CDT": "sewjeroameriski centralny lětni čas", "CHADT": "chathamski lětni čas", "CHAST": "chathamski standardny čas", "CLST": "chilski lětni čas", "CLT": "chilski standardny čas", "COST": "kolumbiski lětni čas", "COT": "kolumbiski standardny čas", "CST": "sewjeroameriski centralny standardny čas", "ChST": "chamorroski čas", "EAT": "wuchodoafriski čas", "ECT": "ekwadorski čas", "EDT": "sewjeroameriski wuchodny lětni čas", "EST": "sewjeroameriski wuchodny standardny čas", "GFT": "francoskoguyanski čas", "GMT": "Greenwichski čas", "GST": "čas Persiskeho golfa", "GYT": "guyanski čas", "HADT": "hawaiisko-aleutski lětni čas", "HAST": "hawaiisko-aleutski standardny čas", "HAT": "nowofundlandski lětni čas", "HECU": "kubaski lětni čas", "HEEG": "wuchodogrönlandski lětni čas", "HENOMX": "mexiski sewjerozapadny lětni čas", "HEOG": "zapadogrönlandski lětni čas", "HEPM": "lětni čas kupow St. Pierre a Miquelon", "HEPMX": "mexiski pacifiski lětni čas", "HKST": "Hongkongski lětni čas", "HKT": "Hongkongski standardny čas", "HNCU": "kubaski standardny čas", "HNEG": "wuchodogrönlandski standardny čas", "HNNOMX": "mexiski sewjerozapadny standardny čas", "HNOG": "zapadogrönlandski standardny čas", "HNPM": "standardny čas kupow St. Pierre a Miquelon", "HNPMX": "mexiski pacifiski standardny čas", "HNT": "nowofundlandski standardny čas", "IST": "indiski čas", "JDT": "japanski lětni čas", "JST": "japanski standardny čas", "LHDT": "lětni čas kupy Lord-Howe", "LHST": "standardny čas kupy Lord-Howe", "MDT": "sewjeroameriski hórski lětni čas", "MESZ": "srjedźoeuropski lětni čas", "MEZ": "srjedźoeuropski standardny čas", "MST": "sewjeroameriski hórski standardny čas", "MYT": "malajziski čas", "NZDT": "nowoseelandski lětni čas", "NZST": "nowoseelandski standardny čas", "OESZ": "wuchodoeuropski lětni čas", "OEZ": "wuchodoeuropski standardny čas", "PDT": "sewjeroameriski pacifiski lětni čas", "PST": "sewjeroameriski pacifiski standardny čas", "SAST": "južnoafriski čas", "SGT": "Singapurski čas", "SRT": "surinamski čas", "TMST": "turkmenski lětni čas", "TMT": "turkmenski standardny čas", "UYST": "uruguayski lětni čas", "UYT": "uruguayski standardny čas", "VET": "venezuelski čas", "WARST": "zapadoargentinski lětni čas", "WART": "zapadoargentinski standardny čas", "WAST": "zapadoafriski lětni čas", "WAT": "zapadoafriski standardny čas", "WESZ": "zapadoeuropski lětni čas", "WEZ": "zapadoeuropski standardny čas", "WIB": "zapadoindoneski čas", "WIT": "wuchodoindoneski", "WITA": "srjedźoindoneski čas", "∅∅∅": "Brasiliski lětni čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_283d29635c231a3ca301081bab098305,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_5d630ed3a6aab7e460d3797e87a803d8,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hsb_de"] = func() locales.Translator {
		return &localen{

			locale:                 "hsb_DE",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "měr.", "apr.", "mej.", "jun.", "jul.", "awg.", "sep.", "okt.", "now.", "dec."},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januara", "februara", "měrca", "apryla", "meje", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"},
			daysAbbreviated:        []string{"nje", "pón", "wut", "srj", "štw", "pja", "sob"},
			daysNarrow:             []string{"n", "p", "w", "s", "š", "p", "s"},
			daysShort:              []string{"nj", "pó", "wu", "sr", "št", "pj", "so"},
			daysWide:               []string{"njedźela", "póndźela", "wutora", "srjeda", "štwórtk", "pjatk", "sobota"},
			periodsAbbreviated:     []string{"dopołdnja", "popołdnju"},
			periodsNarrow:          []string{"dop.", "pop."},
			periodsWide:            []string{"dopołdnja", "popołdnju"},
			erasAbbreviated:        []string{"př.Chr.n.", "po Chr.n."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"před Chrystowym narodźenjom", "po Chrystowym narodźenju"},
			timezones:              map[string]string{"ACDT": "srjedźoawstralski lětni čas", "ACST": "srjedźoawstralski standardny čas", "ACWDT": "sjedźozapadny awstralski lětni čas", "ACWST": "srjedźozapadny awstralski standardny čas", "ADT": "atlantiski lětni čas", "AEDT": "wuchodoawstralski lětni čas", "AEST": "wuchodoawstralski standardny čas", "AKDT": "alaskaski lětni čas", "AKST": "alaskaski standardny čas", "ARST": "argentinski lětni čas", "ART": "argentinski standardny čas", "AST": "atlantiski standardny čas", "AWDT": "zapadoawstralski lětni čas", "AWST": "zapadoawstralski standardny čas", "BOT": "boliwiski čas", "BT": "bhutanski čas", "CAT": "centralnoafriski čas", "CDT": "sewjeroameriski centralny lětni čas", "CHADT": "chathamski lětni čas", "CHAST": "chathamski standardny čas", "CLST": "chilski lětni čas", "CLT": "chilski standardny čas", "COST": "kolumbiski lětni čas", "COT": "kolumbiski standardny čas", "CST": "sewjeroameriski centralny standardny čas", "ChST": "chamorroski čas", "EAT": "wuchodoafriski čas", "ECT": "ekwadorski čas", "EDT": "sewjeroameriski wuchodny lětni čas", "EST": "sewjeroameriski wuchodny standardny čas", "GFT": "francoskoguyanski čas", "GMT": "Greenwichski čas", "GST": "čas Persiskeho golfa", "GYT": "guyanski čas", "HADT": "hawaiisko-aleutski lětni čas", "HAST": "hawaiisko-aleutski standardny čas", "HAT": "nowofundlandski lětni čas", "HECU": "kubaski lětni čas", "HEEG": "wuchodogrönlandski lětni čas", "HENOMX": "mexiski sewjerozapadny lětni čas", "HEOG": "zapadogrönlandski lětni čas", "HEPM": "lětni čas kupow St. Pierre a Miquelon", "HEPMX": "mexiski pacifiski lětni čas", "HKST": "Hongkongski lětni čas", "HKT": "Hongkongski standardny čas", "HNCU": "kubaski standardny čas", "HNEG": "wuchodogrönlandski standardny čas", "HNNOMX": "mexiski sewjerozapadny standardny čas", "HNOG": "zapadogrönlandski standardny čas", "HNPM": "standardny čas kupow St. Pierre a Miquelon", "HNPMX": "mexiski pacifiski standardny čas", "HNT": "nowofundlandski standardny čas", "IST": "indiski čas", "JDT": "japanski lětni čas", "JST": "japanski standardny čas", "LHDT": "lětni čas kupy Lord-Howe", "LHST": "standardny čas kupy Lord-Howe", "MDT": "sewjeroameriski hórski lětni čas", "MESZ": "srjedźoeuropski lětni čas", "MEZ": "srjedźoeuropski standardny čas", "MST": "sewjeroameriski hórski standardny čas", "MYT": "malajziski čas", "NZDT": "nowoseelandski lětni čas", "NZST": "nowoseelandski standardny čas", "OESZ": "wuchodoeuropski lětni čas", "OEZ": "wuchodoeuropski standardny čas", "PDT": "sewjeroameriski pacifiski lětni čas", "PST": "sewjeroameriski pacifiski standardny čas", "SAST": "južnoafriski čas", "SGT": "Singapurski čas", "SRT": "surinamski čas", "TMST": "turkmenski lětni čas", "TMT": "turkmenski standardny čas", "UYST": "uruguayski lětni čas", "UYT": "uruguayski standardny čas", "VET": "venezuelski čas", "WARST": "zapadoargentinski lětni čas", "WART": "zapadoargentinski standardny čas", "WAST": "zapadoafriski lětni čas", "WAT": "zapadoafriski standardny čas", "WESZ": "zapadoeuropski lětni čas", "WEZ": "zapadoeuropski standardny čas", "WIB": "zapadoindoneski čas", "WIT": "wuchodoindoneski", "WITA": "srjedźoindoneski čas", "∅∅∅": "peruski lětni čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_283d29635c231a3ca301081bab098305,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_5d630ed3a6aab7e460d3797e87a803d8,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hu"] = func() locales.Translator {
		return &localen{

			locale:                 "hu",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "Ft", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "febr.", "márc.", "ápr.", "máj.", "jún.", "júl.", "aug.", "szept.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "Á", "M", "J", "J", "A", "Sz", "O", "N", "D"},
			monthsWide:             []string{"", "január", "február", "március", "április", "május", "június", "július", "augusztus", "szeptember", "október", "november", "december"},
			daysAbbreviated:        []string{"V", "H", "K", "Sze", "Cs", "P", "Szo"},
			daysNarrow:             []string{"V", "H", "K", "Sz", "Cs", "P", "Sz"},
			daysShort:              []string{"V", "H", "K", "Sze", "Cs", "P", "Szo"},
			daysWide:               []string{"vasárnap", "hétfő", "kedd", "szerda", "csütörtök", "péntek", "szombat"},
			periodsAbbreviated:     []string{"de.", "du."},
			periodsNarrow:          []string{"de.", "du."},
			periodsWide:            []string{"de.", "du."},
			erasAbbreviated:        []string{"i. e.", "i. sz."},
			erasNarrow:             []string{"ie.", "isz."},
			erasWide:               []string{"Krisztus előtt", "időszámításunk szerint"},
			timezones:              map[string]string{"ACDT": "közép-ausztráliai nyári idő", "ACST": "közép-ausztráliai téli idő", "ACWDT": "közép-nyugat-ausztráliai nyári idő", "ACWST": "közép-nyugat-ausztráliai téli idő", "ADT": "atlanti-óceáni nyári idő", "AEDT": "kelet-ausztráliai nyári idő", "AEST": "kelet-ausztráliai téli idő", "AKDT": "alaszkai nyári idő", "AKST": "alaszkai zónaidő", "ARST": "argentínai nyári idő", "ART": "argentínai téli idő", "AST": "atlanti-óceáni zónaidő", "AWDT": "nyugat-ausztráliai nyári idő", "AWST": "nyugat-ausztráliai téli idő", "BOT": "bolíviai téli idő", "BT": "butáni idő", "CAT": "közép-afrikai téli idő", "CDT": "középső államokbeli nyári idő", "CHADT": "chathami nyári idő", "CHAST": "chathami téli idő", "CLST": "chilei nyári idő", "CLT": "chilei téli idő", "COST": "kolumbiai nyári idő", "COT": "kolumbiai téli idő", "CST": "középső államokbeli zónaidő", "ChST": "chamorrói téli idő", "EAT": "kelet-afrikai téli idő", "ECT": "ecuadori téli idő", "EDT": "keleti államokbeli nyári idő", "EST": "keleti államokbeli zónaidő", "GFT": "francia-guyanai idő", "GMT": "greenwichi középidő, téli idő", "GST": "öbölbeli téli idő", "GYT": "guyanai téli idő", "HADT": "hawaii-aleuti nyári idő", "HAST": "hawaii-aleuti téli idő", "HAT": "új-fundlandi nyári idő", "HECU": "kubai nyári idő", "HEEG": "kelet-grönlandi nyári idő", "HENOMX": "északnyugat-mexikói nyári idő", "HEOG": "nyugat-grönlandi nyári idő", "HEPM": "Saint-Pierre és Miquelon-i nyári idő", "HEPMX": "mexikói csendes-óceáni nyári idő", "HKST": "hongkongi nyári idő", "HKT": "hongkongi téli idő", "HNCU": "kubai téli idő", "HNEG": "kelet-grönlandi téli idő", "HNNOMX": "északnyugat-mexikói zónaidő", "HNOG": "nyugat-grönlandi téli idő", "HNPM": "Saint-Pierre és Miquelon-i zónaidő", "HNPMX": "mexikói csendes-óceáni zónaidő", "HNT": "új-fundlandi zónaidő", "IST": "indiai téli idő", "JDT": "japán nyári idő", "JST": "japán téli idő", "LHDT": "Lord Howe-szigeti nyári idő", "LHST": "Lord Howe-szigeti téli idő", "MDT": "hegyvidéki nyári idő", "MESZ": "közép-európai nyári idő", "MEZ": "közép-európai téli idő", "MST": "hegyvidéki zónaidő", "MYT": "malajziai idő", "NZDT": "új-zélandi nyári idő", "NZST": "új-zélandi téli idő", "OESZ": "kelet-európai nyári idő", "OEZ": "kelet-európai téli idő", "PDT": "csendes-óceáni nyári idő", "PST": "csendes-óceáni zónaidő", "SAST": "dél-afrikai téli idő", "SGT": "szingapúri téli idő", "SRT": "szurinámi idő", "TMST": "türkmenisztáni nyári idő", "TMT": "türkmenisztáni téli idő", "UYST": "uruguayi nyári idő", "UYT": "uruguayi téli idő", "VET": "venezuelai idő", "WARST": "nyugat-argentínai nyári idő", "WART": "nyugat-argentínai téli idő", "WAST": "nyugat-afrikai nyári idő", "WAT": "nyugat-afrikai téli idő", "WESZ": "nyugat-európai nyári idő", "WEZ": "nyugat-európai téli idő", "WIB": "nyugat-indonéziai téli idő", "WIT": "kelet-indonéziai idő", "WITA": "közép-indonéziai idő", "∅∅∅": "azori nyári idő"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_96142c8f0234c0bee88b2aec42a20ee0,
			fnFmtDateLong:         fnFmtDateLong_d54b619c689771a59ca2c2bc98c286ad,
			fnFmtDateMedium:       fnFmtDateMedium_7234e1a0f15c9c77f42b9ce48acda9a9,
			fnFmtDateShort:        fnFmtDateShort_6955160c4014108c44f154550c94bf4a,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_ddb6ad1d52dc3051d9aca40302fc33cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hu_hu"] = func() locales.Translator {
		return &localen{

			locale:                 "hu_HU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "febr.", "márc.", "ápr.", "máj.", "jún.", "júl.", "aug.", "szept.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "Á", "M", "J", "J", "A", "Sz", "O", "N", "D"},
			monthsWide:             []string{"", "január", "február", "március", "április", "május", "június", "július", "augusztus", "szeptember", "október", "november", "december"},
			daysAbbreviated:        []string{"V", "H", "K", "Sze", "Cs", "P", "Szo"},
			daysNarrow:             []string{"V", "H", "K", "Sz", "Cs", "P", "Sz"},
			daysShort:              []string{"V", "H", "K", "Sze", "Cs", "P", "Szo"},
			daysWide:               []string{"vasárnap", "hétfő", "kedd", "szerda", "csütörtök", "péntek", "szombat"},
			periodsAbbreviated:     []string{"de.", "du."},
			periodsNarrow:          []string{"de.", "du."},
			periodsWide:            []string{"de.", "du."},
			erasAbbreviated:        []string{"i. e.", "i. sz."},
			erasNarrow:             []string{"ie.", "isz."},
			erasWide:               []string{"Krisztus előtt", "időszámításunk szerint"},
			timezones:              map[string]string{"ACDT": "közép-ausztráliai nyári idő", "ACST": "közép-ausztráliai téli idő", "ACWDT": "közép-nyugat-ausztráliai nyári idő", "ACWST": "közép-nyugat-ausztráliai téli idő", "ADT": "atlanti-óceáni nyári idő", "AEDT": "kelet-ausztráliai nyári idő", "AEST": "kelet-ausztráliai téli idő", "AKDT": "alaszkai nyári idő", "AKST": "alaszkai zónaidő", "ARST": "argentínai nyári idő", "ART": "argentínai téli idő", "AST": "atlanti-óceáni zónaidő", "AWDT": "nyugat-ausztráliai nyári idő", "AWST": "nyugat-ausztráliai téli idő", "BOT": "bolíviai téli idő", "BT": "butáni idő", "CAT": "közép-afrikai téli idő", "CDT": "középső államokbeli nyári idő", "CHADT": "chathami nyári idő", "CHAST": "chathami téli idő", "CLST": "chilei nyári idő", "CLT": "chilei téli idő", "COST": "kolumbiai nyári idő", "COT": "kolumbiai téli idő", "CST": "középső államokbeli zónaidő", "ChST": "chamorrói téli idő", "EAT": "kelet-afrikai téli idő", "ECT": "ecuadori téli idő", "EDT": "keleti államokbeli nyári idő", "EST": "keleti államokbeli zónaidő", "GFT": "francia-guyanai idő", "GMT": "greenwichi középidő, téli idő", "GST": "öbölbeli téli idő", "GYT": "guyanai téli idő", "HADT": "hawaii-aleuti nyári idő", "HAST": "hawaii-aleuti téli idő", "HAT": "új-fundlandi nyári idő", "HECU": "kubai nyári idő", "HEEG": "kelet-grönlandi nyári idő", "HENOMX": "északnyugat-mexikói nyári idő", "HEOG": "nyugat-grönlandi nyári idő", "HEPM": "Saint-Pierre és Miquelon-i nyári idő", "HEPMX": "mexikói csendes-óceáni nyári idő", "HKST": "hongkongi nyári idő", "HKT": "hongkongi téli idő", "HNCU": "kubai téli idő", "HNEG": "kelet-grönlandi téli idő", "HNNOMX": "északnyugat-mexikói zónaidő", "HNOG": "nyugat-grönlandi téli idő", "HNPM": "Saint-Pierre és Miquelon-i zónaidő", "HNPMX": "mexikói csendes-óceáni zónaidő", "HNT": "új-fundlandi zónaidő", "IST": "indiai téli idő", "JDT": "japán nyári idő", "JST": "japán téli idő", "LHDT": "Lord Howe-szigeti nyári idő", "LHST": "Lord Howe-szigeti téli idő", "MDT": "hegyvidéki nyári idő", "MESZ": "közép-európai nyári idő", "MEZ": "közép-európai téli idő", "MST": "hegyvidéki zónaidő", "MYT": "malajziai idő", "NZDT": "új-zélandi nyári idő", "NZST": "új-zélandi téli idő", "OESZ": "kelet-európai nyári idő", "OEZ": "kelet-európai téli idő", "PDT": "csendes-óceáni nyári idő", "PST": "csendes-óceáni zónaidő", "SAST": "dél-afrikai téli idő", "SGT": "szingapúri téli idő", "SRT": "szurinámi idő", "TMST": "türkmenisztáni nyári idő", "TMT": "türkmenisztáni téli idő", "UYST": "uruguayi nyári idő", "UYT": "uruguayi téli idő", "VET": "venezuelai idő", "WARST": "nyugat-argentínai nyári idő", "WART": "nyugat-argentínai téli idő", "WAST": "nyugat-afrikai nyári idő", "WAT": "nyugat-afrikai téli idő", "WESZ": "nyugat-európai nyári idő", "WEZ": "nyugat-európai téli idő", "WIB": "nyugat-indonéziai téli idő", "WIT": "kelet-indonéziai idő", "WITA": "közép-indonéziai idő", "∅∅∅": "azori nyári idő"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_96142c8f0234c0bee88b2aec42a20ee0,
			fnFmtDateLong:         fnFmtDateLong_d54b619c689771a59ca2c2bc98c286ad,
			fnFmtDateMedium:       fnFmtDateMedium_7234e1a0f15c9c77f42b9ce48acda9a9,
			fnFmtDateShort:        fnFmtDateShort_6955160c4014108c44f154550c94bf4a,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_ddb6ad1d52dc3051d9aca40302fc33cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hy"] = func() locales.Translator {
		return &localen{

			locale:                 "hy",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "֏", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "հնվ", "փտվ", "մրտ", "ապր", "մյս", "հնս", "հլս", "օգս", "սեպ", "հոկ", "նոյ", "դեկ"},
			monthsNarrow:           []string{"", "Հ", "Փ", "Մ", "Ա", "Մ", "Հ", "Հ", "Օ", "Ս", "Հ", "Ն", "Դ"},
			monthsWide:             []string{"", "հունվարի", "փետրվարի", "մարտի", "ապրիլի", "մայիսի", "հունիսի", "հուլիսի", "օգոստոսի", "սեպտեմբերի", "հոկտեմբերի", "նոյեմբերի", "դեկտեմբերի"},
			daysAbbreviated:        []string{"կիր", "երկ", "երք", "չրք", "հնգ", "ուր", "շբթ"},
			daysNarrow:             []string{"Կ", "Ե", "Ե", "Չ", "Հ", "Ո", "Շ"},
			daysShort:              []string{"կր", "եկ", "եք", "չք", "հգ", "ու", "շբ"},
			daysWide:               []string{"կիրակի", "երկուշաբթի", "երեքշաբթի", "չորեքշաբթի", "հինգշաբթի", "ուրբաթ", "շաբաթ"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"ա", "հ"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"մ.թ.ա.", "մ.թ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Քրիստոսից առաջ", "Քրիստոսից հետո"},
			timezones:              map[string]string{"ACDT": "Կենտրոնական Ավստրալիայի ամառային ժամանակ", "ACST": "Կենտրոնական Ավստրալիայի ստանդարտ ժամանակ", "ACWDT": "Կենտրոնական Ավստրալիայի արևմտյան ամառային ժամանակ", "ACWST": "Կենտրոնական Ավստրալիայի արևմտյան ստանդարտ ժամանակ", "ADT": "Ատլանտյան ամառային ժամանակ", "AEDT": "Արևելյան Ավստրալիայի ամառային ժամանակ", "AEST": "Արևելյան Ավստրալիայի ստանդարտ ժամանակ", "AKDT": "Ալյասկայի ամառային ժամանակ", "AKST": "Ալյասկայի ստանդարտ ժամանակ", "ARST": "Արգենտինայի ամառային ժամանակ", "ART": "Արգենտինայի ստնադարտ ժամանակ", "AST": "Ատլանտյան ստանդարտ ժամանակ", "AWDT": "Արևմտյան Ավստրալիայի ամառային ժամանակ", "AWST": "Արևմտյան Ավստրալիայի ստանդարտ ժամանակ", "BOT": "Բոլիվիայի ժամանակ", "BT": "Բութանի ժամանակ", "CAT": "Կենտրոնական Աֆրիկայի ժամանակ", "CDT": "Կենտրոնական Ամերիկայի ամառային ժամանակ", "CHADT": "Չաթեմ կղզու ամառային ժամանակ", "CHAST": "Չաթեմ կղզու ստանդարտ ժամանակ", "CLST": "Չիլիի ամառային ժամանակ", "CLT": "Չիլիի ստանդարտ ժամանակ", "COST": "Կոլումբիայի ամառային ժամանակ", "COT": "Կոլումբիայի ստանդարտ ժամանակ", "CST": "Կենտրոնական Ամերիկայի ստանդարտ ժամանակ", "ChST": "Չամոռոյի ժամանակ", "EAT": "Արևելյան Աֆրիկայի ժամանակ", "ECT": "Էկվադորի ժամանակ", "EDT": "Արևելյան Ամերիկայի ամառային ժամանակ", "EST": "Արևելյան Ամերիկայի ստանդարտ ժամանակ", "GFT": "Ֆրանսիական Գվիանայի ժամանակ", "GMT": "Գրինվիչի ժամանակ", "GST": "Պարսից ծոցի ստանդարտ ժամանակ", "GYT": "Գայանայի ժամանակ", "HADT": "Հավայան-ալեության ամառային ժամանակ", "HAST": "Հավայան-ալեության ստանդարտ ժամանակ", "HAT": "Նյուֆաունդլենդի ամառային ժամանակ", "HECU": "Կուբայի ամառային ժամանակ", "HEEG": "Արևելյան Գրենլանդիայի ամառային ժամանակ", "HENOMX": "Հյուսիսարևմտյան Մեքսիկայի ամառային ժամանակ", "HEOG": "Արևմտյան Գրենլանդիայի ամառային ժամանակ", "HEPM": "Սեն Պիեռ և Միքելոնի ամառային ժամանակ", "HEPMX": "Մեքսիկայի խաղաղօվկիանոսյան ամառային ժամանակ", "HKST": "Հոնկոնգի ամառային ժամանակ", "HKT": "Հոնկոնգի ստանդարտ ժամանակ", "HNCU": "Կուբայի ստանդարտ ժամանակ", "HNEG": "Արևելյան Գրենլանդիայի ստանդարտ ժամանակ", "HNNOMX": "Հյուսիսարևմտյան Մեքսիկայի ստանդարտ ժամանակ", "HNOG": "Արևմտյան Գրենլանդիայի ստանդարտ ժամանակ", "HNPM": "Սեն Պիեռ և Միքելոնի ստանդարտ ժամանակ", "HNPMX": "Մեքսիկայի խաղաղօվկիանոսյան ստանդարտ ժամանակ", "HNT": "Նյուֆաունդլենդի ստանդարտ ժամանակ", "IST": "Հնդկաստանի ստանդարտ ժամանակ", "JDT": "Ճապոնիայի ամառային ժամանակ", "JST": "Ճապոնիայի ստանդարտ ժամանակ", "LHDT": "Լորդ Հաուի ամառային ժամանակ", "LHST": "Լորդ Հաուի ստանդարտ ժամանակ", "MDT": "Լեռնային ամառային ժամանակ (ԱՄՆ)", "MESZ": "Կենտրոնական Եվրոպայի ամառային ժամանակ", "MEZ": "Կենտրոնական Եվրոպայի ստանդարտ ժամանակ", "MST": "Լեռնային ստանդարտ ժամանակ (ԱՄՆ)", "MYT": "Մալայզիայի ժամանակ", "NZDT": "Նոր Զելանդիայի ամառային ժամանակ", "NZST": "Նոր Զելանդիայի ստանդարտ ժամանակ", "OESZ": "Արևելյան Եվրոպայի ամառային ժամանակ", "OEZ": "Արևելյան Եվրոպայի ստանդարտ ժամանակ", "PDT": "Խաղաղօվկիանոսյան ամառային ժամանակ", "PST": "Խաղաղօվկիանոսյան ստանդարտ ժամանակ", "SAST": "Հարավային Աֆրիկայի ժամանակ", "SGT": "Սինգապուրի ժամանակ", "SRT": "Սուրինամի ժամանակ", "TMST": "Թուրքմենստանի ամառային ժամանակ", "TMT": "Թուրքմենստանի ստանդարտ ժամանակ", "UYST": "Ուրուգվայի ամառային ժամանակ", "UYT": "Ուրուգվայի ստանդարտ ժամանակ", "VET": "Վենեսուելայի ժամանակ", "WARST": "Արևմտյան Արգենտինայի ամառային ժամանակ", "WART": "Արևմտյան Արգենտինայի ստնադարտ ժամանակ", "WAST": "Արևմտյան Աֆրիկայի ամառային ժամանակ", "WAT": "Արևմտյան Աֆրիկայի ստանդարտ ժամանակ", "WESZ": "Արևմտյան Եվրոպայի ամառային ժամանակ", "WEZ": "Արևմտյան Եվրոպայի ստանդարտ ժամանակ", "WIB": "Արևմտյան Ինդոնեզիայի ժամանակ", "WIT": "Արևելյան Ինդոնեզիայի ժամանակ", "WITA": "Կենտրոնական Ինդոնեզիայի ժամանակ", "∅∅∅": "Ամազոնյան ամառային ժամանակ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_f88f4a85fdaf35c0c7b8106f2f509933,
			fnFmtDateLong:         fnFmtDateLong_569998488d00a7529115ca67accb195e,
			fnFmtDateMedium:       fnFmtDateMedium_227513acca176d2921237a4d2e01beae,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["hy_am"] = func() locales.Translator {
		return &localen{

			locale:                 "hy_AM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "հնվ", "փտվ", "մրտ", "ապր", "մյս", "հնս", "հլս", "օգս", "սեպ", "հոկ", "նոյ", "դեկ"},
			monthsNarrow:           []string{"", "Հ", "Փ", "Մ", "Ա", "Մ", "Հ", "Հ", "Օ", "Ս", "Հ", "Ն", "Դ"},
			monthsWide:             []string{"", "հունվարի", "փետրվարի", "մարտի", "ապրիլի", "մայիսի", "հունիսի", "հուլիսի", "օգոստոսի", "սեպտեմբերի", "հոկտեմբերի", "նոյեմբերի", "դեկտեմբերի"},
			daysAbbreviated:        []string{"կիր", "երկ", "երք", "չրք", "հնգ", "ուր", "շբթ"},
			daysNarrow:             []string{"Կ", "Ե", "Ե", "Չ", "Հ", "Ո", "Շ"},
			daysShort:              []string{"կր", "եկ", "եք", "չք", "հգ", "ու", "շբ"},
			daysWide:               []string{"կիրակի", "երկուշաբթի", "երեքշաբթի", "չորեքշաբթի", "հինգշաբթի", "ուրբաթ", "շաբաթ"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"ա", "հ"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"մ.թ.ա.", "մ.թ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Քրիստոսից առաջ", "Քրիստոսից հետո"},
			timezones:              map[string]string{"ACDT": "Կենտրոնական Ավստրալիայի ամառային ժամանակ", "ACST": "Կենտրոնական Ավստրալիայի ստանդարտ ժամանակ", "ACWDT": "Կենտրոնական Ավստրալիայի արևմտյան ամառային ժամանակ", "ACWST": "Կենտրոնական Ավստրալիայի արևմտյան ստանդարտ ժամանակ", "ADT": "Ատլանտյան ամառային ժամանակ", "AEDT": "Արևելյան Ավստրալիայի ամառային ժամանակ", "AEST": "Արևելյան Ավստրալիայի ստանդարտ ժամանակ", "AKDT": "Ալյասկայի ամառային ժամանակ", "AKST": "Ալյասկայի ստանդարտ ժամանակ", "ARST": "Արգենտինայի ամառային ժամանակ", "ART": "Արգենտինայի ստնադարտ ժամանակ", "AST": "Ատլանտյան ստանդարտ ժամանակ", "AWDT": "Արևմտյան Ավստրալիայի ամառային ժամանակ", "AWST": "Արևմտյան Ավստրալիայի ստանդարտ ժամանակ", "BOT": "Բոլիվիայի ժամանակ", "BT": "Բութանի ժամանակ", "CAT": "Կենտրոնական Աֆրիկայի ժամանակ", "CDT": "Կենտրոնական Ամերիկայի ամառային ժամանակ", "CHADT": "Չաթեմ կղզու ամառային ժամանակ", "CHAST": "Չաթեմ կղզու ստանդարտ ժամանակ", "CLST": "Չիլիի ամառային ժամանակ", "CLT": "Չիլիի ստանդարտ ժամանակ", "COST": "Կոլումբիայի ամառային ժամանակ", "COT": "Կոլումբիայի ստանդարտ ժամանակ", "CST": "Կենտրոնական Ամերիկայի ստանդարտ ժամանակ", "ChST": "Չամոռոյի ժամանակ", "EAT": "Արևելյան Աֆրիկայի ժամանակ", "ECT": "Էկվադորի ժամանակ", "EDT": "Արևելյան Ամերիկայի ամառային ժամանակ", "EST": "Արևելյան Ամերիկայի ստանդարտ ժամանակ", "GFT": "Ֆրանսիական Գվիանայի ժամանակ", "GMT": "Գրինվիչի ժամանակ", "GST": "Պարսից ծոցի ստանդարտ ժամանակ", "GYT": "Գայանայի ժամանակ", "HADT": "Հավայան-ալեության ամառային ժամանակ", "HAST": "Հավայան-ալեության ստանդարտ ժամանակ", "HAT": "Նյուֆաունդլենդի ամառային ժամանակ", "HECU": "Կուբայի ամառային ժամանակ", "HEEG": "Արևելյան Գրենլանդիայի ամառային ժամանակ", "HENOMX": "Հյուսիսարևմտյան Մեքսիկայի ամառային ժամանակ", "HEOG": "Արևմտյան Գրենլանդիայի ամառային ժամանակ", "HEPM": "Սեն Պիեռ և Միքելոնի ամառային ժամանակ", "HEPMX": "Մեքսիկայի խաղաղօվկիանոսյան ամառային ժամանակ", "HKST": "Հոնկոնգի ամառային ժամանակ", "HKT": "Հոնկոնգի ստանդարտ ժամանակ", "HNCU": "Կուբայի ստանդարտ ժամանակ", "HNEG": "Արևելյան Գրենլանդիայի ստանդարտ ժամանակ", "HNNOMX": "Հյուսիսարևմտյան Մեքսիկայի ստանդարտ ժամանակ", "HNOG": "Արևմտյան Գրենլանդիայի ստանդարտ ժամանակ", "HNPM": "Սեն Պիեռ և Միքելոնի ստանդարտ ժամանակ", "HNPMX": "Մեքսիկայի խաղաղօվկիանոսյան ստանդարտ ժամանակ", "HNT": "Նյուֆաունդլենդի ստանդարտ ժամանակ", "IST": "Հնդկաստանի ստանդարտ ժամանակ", "JDT": "Ճապոնիայի ամառային ժամանակ", "JST": "Ճապոնիայի ստանդարտ ժամանակ", "LHDT": "Լորդ Հաուի ամառային ժամանակ", "LHST": "Լորդ Հաուի ստանդարտ ժամանակ", "MDT": "Լեռնային ամառային ժամանակ (ԱՄՆ)", "MESZ": "Կենտրոնական Եվրոպայի ամառային ժամանակ", "MEZ": "Կենտրոնական Եվրոպայի ստանդարտ ժամանակ", "MST": "Լեռնային ստանդարտ ժամանակ (ԱՄՆ)", "MYT": "Մալայզիայի ժամանակ", "NZDT": "Նոր Զելանդիայի ամառային ժամանակ", "NZST": "Նոր Զելանդիայի ստանդարտ ժամանակ", "OESZ": "Արևելյան Եվրոպայի ամառային ժամանակ", "OEZ": "Արևելյան Եվրոպայի ստանդարտ ժամանակ", "PDT": "Խաղաղօվկիանոսյան ամառային ժամանակ", "PST": "Խաղաղօվկիանոսյան ստանդարտ ժամանակ", "SAST": "Հարավային Աֆրիկայի ժամանակ", "SGT": "Սինգապուրի ժամանակ", "SRT": "Սուրինամի ժամանակ", "TMST": "Թուրքմենստանի ամառային ժամանակ", "TMT": "Թուրքմենստանի ստանդարտ ժամանակ", "UYST": "Ուրուգվայի ամառային ժամանակ", "UYT": "Ուրուգվայի ստանդարտ ժամանակ", "VET": "Վենեսուելայի ժամանակ", "WARST": "Արևմտյան Արգենտինայի ամառային ժամանակ", "WART": "Արևմտյան Արգենտինայի ստնադարտ ժամանակ", "WAST": "Արևմտյան Աֆրիկայի ամառային ժամանակ", "WAT": "Արևմտյան Աֆրիկայի ստանդարտ ժամանակ", "WESZ": "Արևմտյան Եվրոպայի ամառային ժամանակ", "WEZ": "Արևմտյան Եվրոպայի ստանդարտ ժամանակ", "WIB": "Արևմտյան Ինդոնեզիայի ժամանակ", "WIT": "Արևելյան Ինդոնեզիայի ժամանակ", "WITA": "Կենտրոնական Ինդոնեզիայի ժամանակ", "∅∅∅": "Բրազիլիայի ամառային ժամանակ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_f88f4a85fdaf35c0c7b8106f2f509933,
			fnFmtDateLong:         fnFmtDateLong_569998488d00a7529115ca67accb195e,
			fnFmtDateMedium:       fnFmtDateMedium_227513acca176d2921237a4d2e01beae,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ia"] = func() locales.Translator {
		return &localen{

			locale:                 "ia",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "ƒ", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "mai", "jun", "jul", "aug", "sep", "oct", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januario", "februario", "martio", "april", "maio", "junio", "julio", "augusto", "septembre", "octobre", "novembre", "decembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "jov", "ven", "sab"},
			daysNarrow:             []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:              []string{"do", "lu", "ma", "me", "jo", "ve", "sa"},
			daysWide:               []string{"dominica", "lunedi", "martedi", "mercuridi", "jovedi", "venerdi", "sabbato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.Chr.", "p.Chr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ante Christo", "post Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "hora estive atlantic", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "hora estive de Alaska", "AKST": "hora normal de Alaska", "ARST": "ARST", "ART": "ART", "AST": "hora normal atlantic", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "hora estive central", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "hora normal central", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "hora estive del est", "EST": "hora normal del est", "GFT": "GFT", "GMT": "hora medie de Greenwich", "GST": "GST", "GYT": "GYT", "HADT": "hora estive de Hawaii-Aleutianas", "HAST": "hora normal de Hawaii-Aleutianas", "HAT": "hora estive de Terranova", "HECU": "hora estive de Cuba", "HEEG": "hora estive de Groenlandia oriental", "HENOMX": "hora estive del nordwest de Mexico", "HEOG": "hora estive de Groenlandia occidental", "HEPM": "hora estive de Saint-Pierre e Miquelon", "HEPMX": "hora estive del Pacifico mexican", "HKST": "HKST", "HKT": "HKT", "HNCU": "hora normal de Cuba", "HNEG": "hora normal de Groenlandia oriental", "HNNOMX": "hora normal del nordwest de Mexico", "HNOG": "hora normal de Groenlandia occidental", "HNPM": "hora normal de Saint-Pierre e Miquelon", "HNPMX": "hora normal del Pacifico mexican", "HNT": "hora normal de Terranova", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "hora estive del montanias", "MESZ": "hora estive de Europa central", "MEZ": "hora normal de Europa central", "MST": "hora normal del montanias", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "hora estive de Europa oriental", "OEZ": "hora normal de Europa oriental", "PDT": "hora estive pacific", "PST": "hora normal pacific", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "hora estive de Europa occidental", "WEZ": "hora normal de Europa occidental", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_5a51da803052c6679cb1ae79edfb44b5,
			fnFmtDateLong:         fnFmtDateLong_cef970679517ca70c9c659b02d9823d9,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ia_001"] = func() locales.Translator {
		return &localen{

			locale:                 "ia_001",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: "( ",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "mai", "jun", "jul", "aug", "sep", "oct", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januario", "februario", "martio", "april", "maio", "junio", "julio", "augusto", "septembre", "octobre", "novembre", "decembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "jov", "ven", "sab"},
			daysNarrow:             []string{"d", "l", "m", "m", "j", "v", "s"},
			daysShort:              []string{"do", "lu", "ma", "me", "jo", "ve", "sa"},
			daysWide:               []string{"dominica", "lunedi", "martedi", "mercuridi", "jovedi", "venerdi", "sabbato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.Chr.", "p.Chr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ante Christo", "post Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "hora estive atlantic", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "hora estive de Alaska", "AKST": "hora normal de Alaska", "ARST": "ARST", "ART": "ART", "AST": "hora normal atlantic", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "hora estive central", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "hora normal central", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "hora estive del est", "EST": "hora normal del est", "GFT": "GFT", "GMT": "hora medie de Greenwich", "GST": "GST", "GYT": "GYT", "HADT": "hora estive de Hawaii-Aleutianas", "HAST": "hora normal de Hawaii-Aleutianas", "HAT": "hora estive de Terranova", "HECU": "hora estive de Cuba", "HEEG": "hora estive de Groenlandia oriental", "HENOMX": "hora estive del nordwest de Mexico", "HEOG": "hora estive de Groenlandia occidental", "HEPM": "hora estive de Saint-Pierre e Miquelon", "HEPMX": "hora estive del Pacifico mexican", "HKST": "HKST", "HKT": "HKT", "HNCU": "hora normal de Cuba", "HNEG": "hora normal de Groenlandia oriental", "HNNOMX": "hora normal del nordwest de Mexico", "HNOG": "hora normal de Groenlandia occidental", "HNPM": "hora normal de Saint-Pierre e Miquelon", "HNPMX": "hora normal del Pacifico mexican", "HNT": "hora normal de Terranova", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "hora estive del montanias", "MESZ": "hora estive de Europa central", "MEZ": "hora normal de Europa central", "MST": "hora normal del montanias", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "hora estive de Europa oriental", "OEZ": "hora normal de Europa oriental", "PDT": "hora estive pacific", "PST": "hora normal pacific", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "hora estive de Europa occidental", "WEZ": "hora normal de Europa occidental", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_343c016939c767f482f0f2a41239f465,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_5a51da803052c6679cb1ae79edfb44b5,
			fnFmtDateLong:         fnFmtDateLong_cef970679517ca70c9c659b02d9823d9,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["id"] = func() locales.Translator {
		return &localen{

			locale:             "id",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ".",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "Rp", "IEP", "ILP", "ILR", "₪", "Rs", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Agu", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Maret", "April", "Mei", "Juni", "Juli", "Agustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:    []string{"Min", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysNarrow:         []string{"M", "S", "S", "R", "K", "J", "S"},
			daysShort:          []string{"Min", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysWide:           []string{"Minggu", "Senin", "Selasa", "Rabu", "Kamis", "Jumat", "Sabtu"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"SM", "M"},
			erasNarrow:         []string{"SM", "M"},
			erasWide:           []string{"Sebelum Masehi", "Masehi"},
			timezones:          map[string]string{"ACDT": "Waktu Musim Panas Tengah Australia", "ACST": "Waktu Standar Tengah Australia", "ACWDT": "Waktu Musim Panas Barat Tengah Australia", "ACWST": "Waktu Standar Barat Tengah Australia", "ADT": "Waktu Musim Panas Atlantik", "AEDT": "Waktu Musim Panas Timur Australia", "AEST": "Waktu Standar Timur Australia", "AKDT": "Waktu Musim Panas Alaska", "AKST": "Waktu Standar Alaska", "ARST": "Waktu Musim Panas Argentina", "ART": "Waktu Standar Argentina", "AST": "Waktu Standar Atlantik", "AWDT": "Waktu Musim Panas Barat Australia", "AWST": "Waktu Standar Barat Australia", "BOT": "Waktu Bolivia", "BT": "Waktu Bhutan", "CAT": "Waktu Afrika Tengah", "CDT": "Waktu Musim Panas Tengah", "CHADT": "Waktu Musim Panas Chatham", "CHAST": "Waktu Standar Chatham", "CLST": "Waktu Musim Panas Cile", "CLT": "Waktu Standar Cile", "COST": "Waktu Musim Panas Kolombia", "COT": "Waktu Standar Kolombia", "CST": "Waktu Standar Tengah", "ChST": "Waktu Standar Chamorro", "EAT": "Waktu Afrika Timur", "ECT": "Waktu Ekuador", "EDT": "Waktu Musim Panas Timur", "EST": "Waktu Standar Timur", "GFT": "Waktu Guyana Prancis", "GMT": "Greenwich Mean Time", "GST": "Waktu Standar Teluk", "GYT": "Waktu Guyana", "HADT": "Waktu Musim Panas Hawaii-Aleutian", "HAST": "Waktu Standar Hawaii-Aleutian", "HAT": "Waktu Musim Panas Newfoundland", "HECU": "Waktu Musim Panas Kuba", "HEEG": "Waktu Musim Panas Greenland Timur", "HENOMX": "Waktu Musim Panas Meksiko Barat Laut", "HEOG": "Waktu Musim Panas Greenland Barat", "HEPM": "Waktu Musim Panas Saint Pierre dan Miquelon", "HEPMX": "Waktu Musim Panas Pasifik Meksiko", "HKST": "Waktu Musim Panas Hong Kong", "HKT": "Waktu Standar Hong Kong", "HNCU": "Waktu Standar Kuba", "HNEG": "Waktu Standar Greenland Timur", "HNNOMX": "Waktu Standar Meksiko Barat Laut", "HNOG": "Waktu Standar Greenland Barat", "HNPM": "Waktu Standar Saint Pierre dan Miquelon", "HNPMX": "Waktu Standar Pasifik Meksiko", "HNT": "Waktu Standar Newfoundland", "IST": "Waktu India", "JDT": "Waktu Musim Panas Jepang", "JST": "Waktu Standar Jepang", "LHDT": "Waktu Musim Panas Lord Howe", "LHST": "Waktu Standar Lord Howe", "MDT": "Waktu Musim Panas Pegunungan", "MESZ": "Waktu Musim Panas Eropa Tengah", "MEZ": "Waktu Standar Eropa Tengah", "MST": "Waktu Standar Pegunungan", "MYT": "Waktu Malaysia", "NZDT": "Waktu Musim Panas Selandia Baru", "NZST": "Waktu Standar Selandia Baru", "OESZ": "Waktu Musim Panas Eropa Timur", "OEZ": "Waktu Standar Eropa Timur", "PDT": "Waktu Musim Panas Pasifik", "PST": "Waktu Standar Pasifik", "SAST": "Waktu Standar Afrika Selatan", "SGT": "Waktu Standar Singapura", "SRT": "Waktu Suriname", "TMST": "Waktu Musim Panas Turkmenistan", "TMT": "Waktu Standar Turkmenistan", "UYST": "Waktu Musim Panas Uruguay", "UYT": "Waktu Standar Uruguay", "VET": "Waktu Venezuela", "WARST": "Waktu Musim Panas Argentina Bagian Barat", "WART": "Waktu Standar Argentina Bagian Barat", "WAST": "Waktu Musim Panas Afrika Barat", "WAT": "Waktu Standar Afrika Barat", "WESZ": "Waktu Musim Panas Eropa Barat", "WEZ": "Waktu Standar Eropa Barat", "WIB": "Waktu Indonesia Barat", "WIT": "Waktu Indonesia Timur", "WITA": "Waktu Indonesia Tengah", "∅∅∅": "Waktu Musim Panas Brasil"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["id_id"] = func() locales.Translator {
		return &localen{

			locale:             "id_ID",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ",",
			group:              ".",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ".",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Agu", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Maret", "April", "Mei", "Juni", "Juli", "Agustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:    []string{"Min", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysNarrow:         []string{"M", "S", "S", "R", "K", "J", "S"},
			daysShort:          []string{"Min", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysWide:           []string{"Minggu", "Senin", "Selasa", "Rabu", "Kamis", "Jumat", "Sabtu"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"SM", "M"},
			erasNarrow:         []string{"SM", "M"},
			erasWide:           []string{"Sebelum Masehi", "Masehi"},
			timezones:          map[string]string{"ACDT": "Waktu Musim Panas Tengah Australia", "ACST": "Waktu Standar Tengah Australia", "ACWDT": "Waktu Musim Panas Barat Tengah Australia", "ACWST": "Waktu Standar Barat Tengah Australia", "ADT": "Waktu Musim Panas Atlantik", "AEDT": "Waktu Musim Panas Timur Australia", "AEST": "Waktu Standar Timur Australia", "AKDT": "Waktu Musim Panas Alaska", "AKST": "Waktu Standar Alaska", "ARST": "Waktu Musim Panas Argentina", "ART": "Waktu Standar Argentina", "AST": "Waktu Standar Atlantik", "AWDT": "Waktu Musim Panas Barat Australia", "AWST": "Waktu Standar Barat Australia", "BOT": "Waktu Bolivia", "BT": "Waktu Bhutan", "CAT": "Waktu Afrika Tengah", "CDT": "Waktu Musim Panas Tengah", "CHADT": "Waktu Musim Panas Chatham", "CHAST": "Waktu Standar Chatham", "CLST": "Waktu Musim Panas Cile", "CLT": "Waktu Standar Cile", "COST": "Waktu Musim Panas Kolombia", "COT": "Waktu Standar Kolombia", "CST": "Waktu Standar Tengah", "ChST": "Waktu Standar Chamorro", "EAT": "Waktu Afrika Timur", "ECT": "Waktu Ekuador", "EDT": "Waktu Musim Panas Timur", "EST": "Waktu Standar Timur", "GFT": "Waktu Guyana Prancis", "GMT": "Greenwich Mean Time", "GST": "Waktu Standar Teluk", "GYT": "Waktu Guyana", "HADT": "Waktu Musim Panas Hawaii-Aleutian", "HAST": "Waktu Standar Hawaii-Aleutian", "HAT": "Waktu Musim Panas Newfoundland", "HECU": "Waktu Musim Panas Kuba", "HEEG": "Waktu Musim Panas Greenland Timur", "HENOMX": "Waktu Musim Panas Meksiko Barat Laut", "HEOG": "Waktu Musim Panas Greenland Barat", "HEPM": "Waktu Musim Panas Saint Pierre dan Miquelon", "HEPMX": "Waktu Musim Panas Pasifik Meksiko", "HKST": "Waktu Musim Panas Hong Kong", "HKT": "Waktu Standar Hong Kong", "HNCU": "Waktu Standar Kuba", "HNEG": "Waktu Standar Greenland Timur", "HNNOMX": "Waktu Standar Meksiko Barat Laut", "HNOG": "Waktu Standar Greenland Barat", "HNPM": "Waktu Standar Saint Pierre dan Miquelon", "HNPMX": "Waktu Standar Pasifik Meksiko", "HNT": "Waktu Standar Newfoundland", "IST": "Waktu India", "JDT": "Waktu Musim Panas Jepang", "JST": "Waktu Standar Jepang", "LHDT": "Waktu Musim Panas Lord Howe", "LHST": "Waktu Standar Lord Howe", "MDT": "Waktu Musim Panas Pegunungan", "MESZ": "Waktu Musim Panas Eropa Tengah", "MEZ": "Waktu Standar Eropa Tengah", "MST": "Waktu Standar Pegunungan", "MYT": "Waktu Malaysia", "NZDT": "Waktu Musim Panas Selandia Baru", "NZST": "Waktu Standar Selandia Baru", "OESZ": "Waktu Musim Panas Eropa Timur", "OEZ": "Waktu Standar Eropa Timur", "PDT": "Waktu Musim Panas Pasifik", "PST": "Waktu Standar Pasifik", "SAST": "Waktu Standar Afrika Selatan", "SGT": "Waktu Standar Singapura", "SRT": "Waktu Suriname", "TMST": "Waktu Musim Panas Turkmenistan", "TMT": "Waktu Standar Turkmenistan", "UYST": "Waktu Musim Panas Uruguay", "UYT": "Waktu Standar Uruguay", "VET": "Waktu Venezuela", "WARST": "Waktu Musim Panas Argentina Bagian Barat", "WART": "Waktu Standar Argentina Bagian Barat", "WAST": "Waktu Musim Panas Afrika Barat", "WAT": "Waktu Standar Afrika Barat", "WESZ": "Waktu Musim Panas Eropa Barat", "WEZ": "Waktu Standar Eropa Barat", "WIB": "Waktu Indonesia Barat", "WIT": "Waktu Indonesia Timur", "WITA": "Waktu Indonesia Tengah", "∅∅∅": "Waktu Musim Panas Amazon"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ig"] = func() locales.Translator {
		return &localen{

			locale:             "ig",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            "٫",
			group:              "٬",
			minus:              "‏-",
			percent:            "٪‏",
			perMille:           "؉",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "₦", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "Jen", "Feb", "Maa", "Epr", "Mee", "Juu", "Jul", "Ọgọ", "Sep", "Ọkt", "Nov", "Dis"},
			monthsNarrow:       []string{"", "J", "F", "M", "E", "M", "J", "J", "Ọ", "S", "Ọ", "N", "D"},
			monthsWide:         []string{"", "Jenụwarị", "Febrụwarị", "Maachị", "Epreel", "Mee", "Juun", "Julaị", "Ọgọọst", "Septemba", "Ọktoba", "Novemba", "Disemba"},
			daysAbbreviated:    []string{"Ụka", "Mọn", "Tiu", "Wen", "Tọọ", "Fraị", "Sat"},
			daysShort:          []string{"Sọn", "Mọn", "Tiu", "Wen", "Tọọ", "Fraị", "Sat"},
			daysWide:           []string{"Sọndee", "Mọnde", "Tiuzdee", "Wenezdee", "Tọọzdee", "Fraịdee", "Satọdee"},
			periodsAbbreviated: []string{"A.M.", "P.M."},
			periodsNarrow:      []string{"A.M.", "P.M."},
			periodsWide:        []string{"N’ụtụtụ", "N’abali"},
			erasAbbreviated:    []string{"T.K.", "A.K."},
			erasNarrow:         []string{"T.K.", "A.K."},
			erasWide:           []string{"Tupu Kraist", "Afọ Kraịst"},
			timezones:          map[string]string{"ACDT": "Oge Ihe Etiti Australia", "ACST": "Oge Izugbe Etiti Australia", "ACWDT": "Oge Ihe Mpaghara Ọdịda Anyanwụ Etiti Australia", "ACWST": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Etiti Australia", "ADT": "Oge Ihe Mpaghara Atlantic", "AEDT": "Oge Ihe Mpaghara Ọwụwa Anyanwụ Australia", "AEST": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ Australia", "AKDT": "Oge Ihe Alaska", "AKST": "Oge Izugbe Alaska", "ARST": "Oge Okpomọkụ Argentina", "ART": "Oge Izugbe Argentina", "AST": "Oge Izugbe Mpaghara Atlantic", "AWDT": "Oge Ihe Mpaghara Ọdịda Anyanwụ Australia", "AWST": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Australia", "BOT": "Oge Bolivia", "BT": "Oge Bhutan", "CAT": "Oge Etiti Afrịka", "CDT": "Oge Ihe Mpaghara Etiti", "CHADT": "Oge Ihe Chatham", "CHAST": "Oge Izugbe Chatham", "CLST": "Oge Okpomọkụ Chile", "CLT": "Oge Izugbe Chile", "COST": "Oge Okpomọkụ Columbia", "COT": "Oge Izugbe Columbia", "CST": "Oge Izugbe Mpaghara Etiti", "ChST": "Oge Izugbe Chamorro", "EAT": "Oge Mpaghara Ọwụwa Anyanwụ Afrịka", "ECT": "Oge Ecuador", "EDT": "Oge Ihe Mpaghara Ọwụwa Anyanwụ", "EST": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ", "GFT": "Oge French Guiana", "GMT": "Oge Mpaghara Greemwich Mean", "GST": "Oge Izugbe Gulf", "GYT": "Oge Guyana", "HADT": "Oge Ihe Hawaii-Aleutian", "HAST": "Oge Izugbe Hawaii-Aleutian", "HAT": "Oge Ihe Newfoundland", "HECU": "Oge Ihe Mpaghara Cuba", "HEEG": "Oge Okpomọkụ Mpaghara Ọwụwa Anyanwụ Greenland", "HENOMX": "Oge Ihe Northwest Mexico", "HEOG": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Greenland", "HEPM": "Oge Ihe St. Pierre & Miquelon", "HEPMX": "Oge Ihe Mexican Pacific", "HKST": "Oge Okpomọkụ Hong Kong", "HKT": "Oge Izugbe Hong Kong", "HNCU": "Oge Izugbe Cuba", "HNEG": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ Greenland", "HNNOMX": "Oge Izugbe Northwest Mexico", "HNOG": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Greenland", "HNPM": "Oge Izugbe St. Pierre & Miquelon", "HNPMX": "Oge Izugbe Mexican Pacific", "HNT": "Oge Izugbe Newfoundland", "IST": "Oge Izugbe India", "JDT": "Oge Ihe Japan", "JST": "Oge Izugbe Japan", "LHDT": "Oge Ihe Lord Howe", "LHST": "Oge Izugbe Lord Howe", "MDT": "Oge Ihe Mpaghara Ugwu", "MESZ": "Oge Okpomọkụ Mpaghara Etiti Europe", "MEZ": "Oge Izugbe Mpaghara Etiti Europe", "MST": "Oge Izugbe Mpaghara Ugwu", "MYT": "Oge Malaysia", "NZDT": "Oge Ihe New Zealand", "NZST": "Oge Izugbe New Zealand", "OESZ": "Oge Okpomọkụ Mpaghara Ọwụwa Anyanwụ Europe", "OEZ": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ Europe", "PDT": "Oge Ihe Mpaghara Pacific", "PST": "Oge Izugbe Mpaghara Pacific", "SAST": "Oge Izugbe Mpaghara Mgbada Ugwu Afrịka", "SGT": "Oge Izugbe Singapore", "SRT": "Oge Suriname", "TMST": "Oge Okpomọkụ Turkmenist", "TMT": "Oge Izugbe Turkmenist", "UYST": "Oge Okpomọkụ Uruguay", "UYT": "Oge Izugbe Uruguay", "VET": "Oge Venezuela", "WARST": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Argentina", "WART": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Argentina", "WAST": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Afrịka", "WAT": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Afrịka", "WESZ": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Europe", "WEZ": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Europe", "WIB": "Oge Mpaghara Ọdịda Anyanwụ Indonesia", "WIT": "Oge Mpaghara Ọwụwa Anyanwụ Indonesia", "WITA": "Oge Etiti Indonesia", "∅∅∅": "Oge Okpomọkụ Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1284b8838aa00e9c05dfab59723d425b,
			fnFmtCurrency:         fnFmtCurrency_190e2886d7df91874dc7265208b76611,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_867dc730c8139deaf3b7a764768ab63b,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ig_ng"] = func() locales.Translator {
		return &localen{

			locale:             "ig_NG",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            "٫",
			group:              "٬",
			minus:              "‏-",
			percent:            "٪‏",
			perMille:           "؉",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "Jen", "Feb", "Maa", "Epr", "Mee", "Juu", "Jul", "Ọgọ", "Sep", "Ọkt", "Nov", "Dis"},
			monthsNarrow:       []string{"", "J", "F", "M", "E", "M", "J", "J", "Ọ", "S", "Ọ", "N", "D"},
			monthsWide:         []string{"", "Jenụwarị", "Febrụwarị", "Maachị", "Epreel", "Mee", "Juun", "Julaị", "Ọgọọst", "Septemba", "Ọktoba", "Novemba", "Disemba"},
			daysAbbreviated:    []string{"Ụka", "Mọn", "Tiu", "Wen", "Tọọ", "Fraị", "Sat"},
			daysShort:          []string{"Sọn", "Mọn", "Tiu", "Wen", "Tọọ", "Fraị", "Sat"},
			daysWide:           []string{"Sọndee", "Mọnde", "Tiuzdee", "Wenezdee", "Tọọzdee", "Fraịdee", "Satọdee"},
			periodsAbbreviated: []string{"A.M.", "P.M."},
			periodsNarrow:      []string{"A.M.", "P.M."},
			periodsWide:        []string{"N’ụtụtụ", "N’abali"},
			erasAbbreviated:    []string{"T.K.", "A.K."},
			erasNarrow:         []string{"T.K.", "A.K."},
			erasWide:           []string{"Tupu Kraist", "Afọ Kraịst"},
			timezones:          map[string]string{"ACDT": "Oge Ihe Etiti Australia", "ACST": "Oge Izugbe Etiti Australia", "ACWDT": "Oge Ihe Mpaghara Ọdịda Anyanwụ Etiti Australia", "ACWST": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Etiti Australia", "ADT": "Oge Ihe Mpaghara Atlantic", "AEDT": "Oge Ihe Mpaghara Ọwụwa Anyanwụ Australia", "AEST": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ Australia", "AKDT": "Oge Ihe Alaska", "AKST": "Oge Izugbe Alaska", "ARST": "Oge Okpomọkụ Argentina", "ART": "Oge Izugbe Argentina", "AST": "Oge Izugbe Mpaghara Atlantic", "AWDT": "Oge Ihe Mpaghara Ọdịda Anyanwụ Australia", "AWST": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Australia", "BOT": "Oge Bolivia", "BT": "Oge Bhutan", "CAT": "Oge Etiti Afrịka", "CDT": "Oge Ihe Mpaghara Etiti", "CHADT": "Oge Ihe Chatham", "CHAST": "Oge Izugbe Chatham", "CLST": "Oge Okpomọkụ Chile", "CLT": "Oge Izugbe Chile", "COST": "Oge Okpomọkụ Columbia", "COT": "Oge Izugbe Columbia", "CST": "Oge Izugbe Mpaghara Etiti", "ChST": "Oge Izugbe Chamorro", "EAT": "Oge Mpaghara Ọwụwa Anyanwụ Afrịka", "ECT": "Oge Ecuador", "EDT": "Oge Ihe Mpaghara Ọwụwa Anyanwụ", "EST": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ", "GFT": "Oge French Guiana", "GMT": "Oge Mpaghara Greemwich Mean", "GST": "Oge Izugbe Gulf", "GYT": "Oge Guyana", "HADT": "Oge Ihe Hawaii-Aleutian", "HAST": "Oge Izugbe Hawaii-Aleutian", "HAT": "Oge Ihe Newfoundland", "HECU": "Oge Ihe Mpaghara Cuba", "HEEG": "Oge Okpomọkụ Mpaghara Ọwụwa Anyanwụ Greenland", "HENOMX": "Oge Ihe Northwest Mexico", "HEOG": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Greenland", "HEPM": "Oge Ihe St. Pierre & Miquelon", "HEPMX": "Oge Ihe Mexican Pacific", "HKST": "Oge Okpomọkụ Hong Kong", "HKT": "Oge Izugbe Hong Kong", "HNCU": "Oge Izugbe Cuba", "HNEG": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ Greenland", "HNNOMX": "Oge Izugbe Northwest Mexico", "HNOG": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Greenland", "HNPM": "Oge Izugbe St. Pierre & Miquelon", "HNPMX": "Oge Izugbe Mexican Pacific", "HNT": "Oge Izugbe Newfoundland", "IST": "Oge Izugbe India", "JDT": "Oge Ihe Japan", "JST": "Oge Izugbe Japan", "LHDT": "Oge Ihe Lord Howe", "LHST": "Oge Izugbe Lord Howe", "MDT": "Oge Ihe Mpaghara Ugwu", "MESZ": "Oge Okpomọkụ Mpaghara Etiti Europe", "MEZ": "Oge Izugbe Mpaghara Etiti Europe", "MST": "Oge Izugbe Mpaghara Ugwu", "MYT": "Oge Malaysia", "NZDT": "Oge Ihe New Zealand", "NZST": "Oge Izugbe New Zealand", "OESZ": "Oge Okpomọkụ Mpaghara Ọwụwa Anyanwụ Europe", "OEZ": "Oge Izugbe Mpaghara Ọwụwa Anyanwụ Europe", "PDT": "Oge Ihe Mpaghara Pacific", "PST": "Oge Izugbe Mpaghara Pacific", "SAST": "Oge Izugbe Mpaghara Mgbada Ugwu Afrịka", "SGT": "Oge Izugbe Singapore", "SRT": "Oge Suriname", "TMST": "Oge Okpomọkụ Turkmenist", "TMT": "Oge Izugbe Turkmenist", "UYST": "Oge Okpomọkụ Uruguay", "UYT": "Oge Izugbe Uruguay", "VET": "Oge Venezuela", "WARST": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Argentina", "WART": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Argentina", "WAST": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Afrịka", "WAT": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Afrịka", "WESZ": "Oge Okpomọkụ Mpaghara Ọdịda Anyanwụ Europe", "WEZ": "Oge Izugbe Mpaghara Ọdịda Anyanwụ Europe", "WIB": "Oge Mpaghara Ọdịda Anyanwụ Indonesia", "WIT": "Oge Mpaghara Ọwụwa Anyanwụ Indonesia", "WITA": "Oge Etiti Indonesia", "∅∅∅": "Oge Okpomọkụ Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1284b8838aa00e9c05dfab59723d425b,
			fnFmtCurrency:         fnFmtCurrency_190e2886d7df91874dc7265208b76611,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_867dc730c8139deaf3b7a764768ab63b,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ii"] = func() locales.Translator {
		return &localen{

			locale:                 "ii",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ꋍꆪ", "ꑍꆪ", "ꌕꆪ", "ꇖꆪ", "ꉬꆪ", "ꃘꆪ", "ꏃꆪ", "ꉆꆪ", "ꈬꆪ", "ꊰꆪ", "ꊰꊪꆪ", "ꊰꑋꆪ"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ꋍꆪ", "ꑍꆪ", "ꌕꆪ", "ꇖꆪ", "ꉬꆪ", "ꃘꆪ", "ꏃꆪ", "ꉆꆪ", "ꈬꆪ", "ꊰꆪ", "ꊰꊪꆪ", "ꊰꑋꆪ"},
			daysAbbreviated:        []string{"ꑭꆏ", "ꆏꋍ", "ꆏꑍ", "ꆏꌕ", "ꆏꇖ", "ꆏꉬ", "ꆏꃘ"},
			daysNarrow:             []string{"ꆏ", "ꋍ", "ꑍ", "ꌕ", "ꇖ", "ꉬ", "ꃘ"},
			daysWide:               []string{"ꑭꆏꑍ", "ꆏꊂꋍ", "ꆏꊂꑍ", "ꆏꊂꌕ", "ꆏꊂꇖ", "ꆏꊂꉬ", "ꆏꊂꃘ"},
			periodsAbbreviated:     []string{"ꎸꄑ", "ꁯꋒ"},
			periodsWide:            []string{"ꎸꄑ", "ꁯꋒ"},
			erasAbbreviated:        []string{"ꃅꋊꂿ", "ꃅꋊꊂ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ii_cn"] = func() locales.Translator {
		return &localen{

			locale:                 "ii_CN",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ꋍꆪ", "ꑍꆪ", "ꌕꆪ", "ꇖꆪ", "ꉬꆪ", "ꃘꆪ", "ꏃꆪ", "ꉆꆪ", "ꈬꆪ", "ꊰꆪ", "ꊰꊪꆪ", "ꊰꑋꆪ"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ꋍꆪ", "ꑍꆪ", "ꌕꆪ", "ꇖꆪ", "ꉬꆪ", "ꃘꆪ", "ꏃꆪ", "ꉆꆪ", "ꈬꆪ", "ꊰꆪ", "ꊰꊪꆪ", "ꊰꑋꆪ"},
			daysAbbreviated:        []string{"ꑭꆏ", "ꆏꋍ", "ꆏꑍ", "ꆏꌕ", "ꆏꇖ", "ꆏꉬ", "ꆏꃘ"},
			daysNarrow:             []string{"ꆏ", "ꋍ", "ꑍ", "ꌕ", "ꇖ", "ꉬ", "ꃘ"},
			daysWide:               []string{"ꑭꆏꑍ", "ꆏꊂꋍ", "ꆏꊂꑍ", "ꆏꊂꌕ", "ꆏꊂꇖ", "ꆏꊂꉬ", "ꆏꊂꃘ"},
			periodsAbbreviated:     []string{"ꎸꄑ", "ꁯꋒ"},
			periodsWide:            []string{"ꎸꄑ", "ꁯꋒ"},
			erasAbbreviated:        []string{"ꃅꋊꂿ", "ꃅꋊꊂ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["is"] = func() locales.Translator {
		return &localen{

			locale:                 "is",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maí", "jún.", "júl.", "ágú.", "sep.", "okt.", "nóv.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "Á", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janúar", "febrúar", "mars", "apríl", "maí", "júní", "júlí", "ágúst", "september", "október", "nóvember", "desember"},
			daysAbbreviated:        []string{"sun.", "mán.", "þri.", "mið.", "fim.", "fös.", "lau."},
			daysNarrow:             []string{"S", "M", "Þ", "M", "F", "F", "L"},
			daysShort:              []string{"su.", "má.", "þr.", "mi.", "fi.", "fö.", "la."},
			daysWide:               []string{"sunnudagur", "mánudagur", "þriðjudagur", "miðvikudagur", "fimmtudagur", "föstudagur", "laugardagur"},
			periodsAbbreviated:     []string{"f.h.", "e.h."},
			periodsNarrow:          []string{"f.", "e."},
			periodsWide:            []string{"f.h.", "e.h."},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.k.", "e.k."},
			erasWide:               []string{"fyrir Krist", "eftir Krist"},
			timezones:              map[string]string{"ACDT": "Sumartími í Mið-Ástralíu", "ACST": "Staðaltími í Mið-Ástralíu", "ACWDT": "Sumartími í miðvesturhluta Ástralíu", "ACWST": "Staðaltími í miðvesturhluta Ástralíu", "ADT": "Sumartími á Atlantshafssvæðinu", "AEDT": "Sumartími í Austur-Ástralíu", "AEST": "Staðaltími í Austur-Ástralíu", "AKDT": "Sumartími í Alaska", "AKST": "Staðaltími í Alaska", "ARST": "Sumartími í Argentínu", "ART": "Staðaltími í Argentínu", "AST": "Staðaltími á Atlantshafssvæðinu", "AWDT": "Sumartími í Vestur-Ástralíu", "AWST": "Staðaltími í Vestur-Ástralíu", "BOT": "Bólivíutími", "BT": "Bútantími", "CAT": "Mið-Afríkutími", "CDT": "Sumartími í miðhluta Bandaríkjanna og Kanada", "CHADT": "Sumartími í Chatham", "CHAST": "Staðaltími í Chatham", "CLST": "Sumartími í Síle", "CLT": "Staðaltími í Síle", "COST": "Sumartími í Kólumbíu", "COT": "Staðaltími í Kólumbíu", "CST": "Staðaltími í miðhluta Bandaríkjanna og Kanada", "ChST": "Chamorro-staðaltími", "EAT": "Austur-Afríkutími", "ECT": "Ekvadortími", "EDT": "Sumartími í austurhluta Bandaríkjanna og Kanada", "EST": "Staðaltími í austurhluta Bandaríkjanna og Kanada", "GFT": "Tími í Frönsku Gvæjana", "GMT": "Greenwich-staðaltími", "GST": "Staðaltími við Persaflóa", "GYT": "Gvæjanatími", "HADT": "Sumartími á Havaí og Aleúta", "HAST": "Staðaltími á Havaí og Aleúta", "HAT": "Sumartími á Nýfundnalandi", "HECU": "Sumartími á Kúbu", "HEEG": "Sumartími á Austur-Grænlandi", "HENOMX": "Sumartími í Norðvestur-Mexíkó", "HEOG": "Sumartími á Vestur-Grænlandi", "HEPM": "Sumartími á Sankti Pierre og Miquelon", "HEPMX": "Sumartími í Mexíkó á Kyrrahafssvæðinu", "HKST": "Sumartími í Hong Kong", "HKT": "Staðaltími í Hong Kong", "HNCU": "Staðaltími á Kúbu", "HNEG": "Staðaltími á Austur-Grænlandi", "HNNOMX": "Staðaltími í Norðvestur-Mexíkó", "HNOG": "Staðaltími á Vestur-Grænlandi", "HNPM": "Staðaltími á Sankti Pierre og Miquelon", "HNPMX": "Staðaltími í Mexíkó á Kyrrahafssvæðinu", "HNT": "Staðaltími á Nýfundnalandi", "IST": "Indlandstími", "JDT": "Sumartími í Japan", "JST": "Staðaltími í Japan", "LHDT": "Sumartími á Lord Howe-eyju", "LHST": "Staðaltími á Lord Howe-eyju", "MDT": "Sumartími í Klettafjöllum", "MESZ": "Sumartími í Mið-Evrópu", "MEZ": "Staðaltími í Mið-Evrópu", "MST": "Staðaltími í Klettafjöllum", "MYT": "Malasíutími", "NZDT": "Sumartími á Nýja-Sjálandi", "NZST": "Staðaltími á Nýja-Sjálandi", "OESZ": "Sumartími í Austur-Evrópu", "OEZ": "Staðaltími í Austur-Evrópu", "PDT": "Sumartími á Kyrrahafssvæðinu", "PST": "Staðaltími á Kyrrahafssvæðinu", "SAST": "Suður-Afríkutími", "SGT": "Singapúrtími", "SRT": "Súrinamtími", "TMST": "Sumartími í Túrkmenistan", "TMT": "Staðaltími í Túrkmenistan", "UYST": "Sumartími í Úrúgvæ", "UYT": "Staðaltími í Úrúgvæ", "VET": "Venesúelatími", "WARST": "Sumartími í Vestur-Argentínu", "WART": "Staðaltími í Vestur-Argentínu", "WAST": "Sumartími í Vestur-Afríku", "WAT": "Staðaltími í Vestur-Afríku", "WESZ": "Sumartími í Vestur-Evrópu", "WEZ": "Staðaltími í Vestur-Evrópu", "WIB": "Vestur-Indónesíutími", "WIT": "Austur-Indónesíutími", "WITA": "Mið-Indónesíutími", "∅∅∅": "Sumartími í Brasilíu"},

			fnCardinalPluralRule:  fnCardinalPluralRule_aee42f293f8724ca3ca1ad3bf1a57537,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["is_is"] = func() locales.Translator {
		return &localen{

			locale:                 "is_IS",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maí", "jún.", "júl.", "ágú.", "sep.", "okt.", "nóv.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "Á", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janúar", "febrúar", "mars", "apríl", "maí", "júní", "júlí", "ágúst", "september", "október", "nóvember", "desember"},
			daysAbbreviated:        []string{"sun.", "mán.", "þri.", "mið.", "fim.", "fös.", "lau."},
			daysNarrow:             []string{"S", "M", "Þ", "M", "F", "F", "L"},
			daysShort:              []string{"su.", "má.", "þr.", "mi.", "fi.", "fö.", "la."},
			daysWide:               []string{"sunnudagur", "mánudagur", "þriðjudagur", "miðvikudagur", "fimmtudagur", "föstudagur", "laugardagur"},
			periodsAbbreviated:     []string{"f.h.", "e.h."},
			periodsNarrow:          []string{"f.", "e."},
			periodsWide:            []string{"f.h.", "e.h."},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.k.", "e.k."},
			erasWide:               []string{"fyrir Krist", "eftir Krist"},
			timezones:              map[string]string{"ACDT": "Sumartími í Mið-Ástralíu", "ACST": "Staðaltími í Mið-Ástralíu", "ACWDT": "Sumartími í miðvesturhluta Ástralíu", "ACWST": "Staðaltími í miðvesturhluta Ástralíu", "ADT": "Sumartími á Atlantshafssvæðinu", "AEDT": "Sumartími í Austur-Ástralíu", "AEST": "Staðaltími í Austur-Ástralíu", "AKDT": "Sumartími í Alaska", "AKST": "Staðaltími í Alaska", "ARST": "Sumartími í Argentínu", "ART": "Staðaltími í Argentínu", "AST": "Staðaltími á Atlantshafssvæðinu", "AWDT": "Sumartími í Vestur-Ástralíu", "AWST": "Staðaltími í Vestur-Ástralíu", "BOT": "Bólivíutími", "BT": "Bútantími", "CAT": "Mið-Afríkutími", "CDT": "Sumartími í miðhluta Bandaríkjanna og Kanada", "CHADT": "Sumartími í Chatham", "CHAST": "Staðaltími í Chatham", "CLST": "Sumartími í Síle", "CLT": "Staðaltími í Síle", "COST": "Sumartími í Kólumbíu", "COT": "Staðaltími í Kólumbíu", "CST": "Staðaltími í miðhluta Bandaríkjanna og Kanada", "ChST": "Chamorro-staðaltími", "EAT": "Austur-Afríkutími", "ECT": "Ekvadortími", "EDT": "Sumartími í austurhluta Bandaríkjanna og Kanada", "EST": "Staðaltími í austurhluta Bandaríkjanna og Kanada", "GFT": "Tími í Frönsku Gvæjana", "GMT": "Greenwich-staðaltími", "GST": "Staðaltími við Persaflóa", "GYT": "Gvæjanatími", "HADT": "Sumartími á Havaí og Aleúta", "HAST": "Staðaltími á Havaí og Aleúta", "HAT": "Sumartími á Nýfundnalandi", "HECU": "Sumartími á Kúbu", "HEEG": "Sumartími á Austur-Grænlandi", "HENOMX": "Sumartími í Norðvestur-Mexíkó", "HEOG": "Sumartími á Vestur-Grænlandi", "HEPM": "Sumartími á Sankti Pierre og Miquelon", "HEPMX": "Sumartími í Mexíkó á Kyrrahafssvæðinu", "HKST": "Sumartími í Hong Kong", "HKT": "Staðaltími í Hong Kong", "HNCU": "Staðaltími á Kúbu", "HNEG": "Staðaltími á Austur-Grænlandi", "HNNOMX": "Staðaltími í Norðvestur-Mexíkó", "HNOG": "Staðaltími á Vestur-Grænlandi", "HNPM": "Staðaltími á Sankti Pierre og Miquelon", "HNPMX": "Staðaltími í Mexíkó á Kyrrahafssvæðinu", "HNT": "Staðaltími á Nýfundnalandi", "IST": "Indlandstími", "JDT": "Sumartími í Japan", "JST": "Staðaltími í Japan", "LHDT": "Sumartími á Lord Howe-eyju", "LHST": "Staðaltími á Lord Howe-eyju", "MDT": "Sumartími í Klettafjöllum", "MESZ": "Sumartími í Mið-Evrópu", "MEZ": "Staðaltími í Mið-Evrópu", "MST": "Staðaltími í Klettafjöllum", "MYT": "Malasíutími", "NZDT": "Sumartími á Nýja-Sjálandi", "NZST": "Staðaltími á Nýja-Sjálandi", "OESZ": "Sumartími í Austur-Evrópu", "OEZ": "Staðaltími í Austur-Evrópu", "PDT": "Sumartími á Kyrrahafssvæðinu", "PST": "Staðaltími á Kyrrahafssvæðinu", "SAST": "Suður-Afríkutími", "SGT": "Singapúrtími", "SRT": "Súrinamtími", "TMST": "Sumartími í Túrkmenistan", "TMT": "Staðaltími í Túrkmenistan", "UYST": "Sumartími í Úrúgvæ", "UYT": "Staðaltími í Úrúgvæ", "VET": "Venesúelatími", "WARST": "Sumartími í Vestur-Argentínu", "WART": "Staðaltími í Vestur-Argentínu", "WAST": "Sumartími í Vestur-Afríku", "WAT": "Staðaltími í Vestur-Afríku", "WESZ": "Sumartími í Vestur-Evrópu", "WEZ": "Staðaltími í Vestur-Evrópu", "WIB": "Vestur-Indónesíutími", "WIT": "Austur-Indónesíutími", "WITA": "Mið-Indónesíutími", "∅∅∅": "Sumartími á Amasónsvæðinu"},

			fnCardinalPluralRule:  fnCardinalPluralRule_e217c4f51dda8c076116bf492a837691,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["it"] = func() locales.Translator {
		return &localen{

			locale:                 "it",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "Kz", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "KM", "BAN", "$", "৳", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "$", "$", "Bs", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "$", "BTN", "BUK", "P", "BYB", "Br", "BYR", "$", "$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "$", "CNH", "CNX", "¥", "$", "COU", "₡", "CSD", "CSK", "$", "$", "CVE", "CYP", "Kč", "DDM", "DEM", "DJF", "kr", "$", "DZD", "ECS", "ECV", "EEK", "£E", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "$", "£", "FRF", "£", "GEK", "₾", "GHC", "GHS", "£", "GMD", "FG", "GNS", "GQE", "GRD", "Q", "GWE", "GWP", "$", "HKD", "L", "HRD", "kn", "HTG", "Ft", "Rp", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "kr", "ITL", "$", "JOD", "JPY", "KES", "KGS", "៛", "CF", "₩", "KRH", "KRO", "KRW", "KWD", "$", "₸", "₭", "L£", "Rs", "$", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "Ar", "MGF", "MKD", "MKN", "MLF", "K", "₮", "MOP", "MRO", "MRU", "MTL", "MTP", "Rs", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "$", "NGN", "NIC", "C$", "NLG", "NKr", "Rs", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "Rs", "PLN", "PLZ", "PTE", "₲", "QAR", "RHD", "ROL", "lei", "RSD", "₽", "RUR", "RF", "SAR", "$", "SCR", "SDD", "SDG", "SDP", "kr", "$", "£", "SIT", "SKK", "SLL", "SOS", "$", "SRG", "£", "STD", "STN", "SUR", "SVC", "£", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "T$", "TPE", "TRL", "₺", "$", "TWD", "TZS", "₴", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "$", "UYW", "UZS", "VEB", "Bs", "VES", "₫", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZK", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"},
			monthsNarrow:           []string{"", "G", "F", "M", "A", "M", "G", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "G", "V", "S"},
			daysShort:              []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysWide:               []string{"domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"m.", "p."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"avanti Cristo", "dopo Cristo"},
			timezones:              map[string]string{"ACDT": "Ora legale dell’Australia centrale", "ACST": "Ora standard dell’Australia centrale", "ACWDT": "Ora legale dell’Australia centroccidentale", "ACWST": "Ora standard dell’Australia centroccidentale", "ADT": "Ora legale dell’Atlantico", "AEDT": "Ora legale dell’Australia orientale", "AEST": "Ora standard dell’Australia orientale", "AKDT": "Ora legale dell’Alaska", "AKST": "Ora standard dell’Alaska", "ARST": "Ora legale dell’Argentina", "ART": "Ora standard dell’Argentina", "AST": "Ora standard dell’Atlantico", "AWDT": "Ora legale dell’Australia occidentale", "AWST": "Ora standard dell’Australia occidentale", "BOT": "Ora della Bolivia", "BT": "Ora del Bhutan", "CAT": "Ora dell’Africa centrale", "CDT": "Ora legale centrale USA", "CHADT": "Ora legale delle Chatham", "CHAST": "Ora standard delle Chatham", "CLST": "Ora legale del Cile", "CLT": "Ora standard del Cile", "COST": "Ora legale della Colombia", "COT": "Ora standard della Colombia", "CST": "Ora standard centrale USA", "ChST": "Ora di Chamorro", "EAT": "Ora dell’Africa orientale", "ECT": "Ora dell’Ecuador", "EDT": "Ora legale orientale USA", "EST": "Ora standard orientale USA", "GFT": "Ora della Guiana francese", "GMT": "Ora del meridiano di Greenwich", "GST": "Ora del Golfo", "GYT": "Ora della Guyana", "HADT": "Ora legale delle Isole Hawaii-Aleutine", "HAST": "Ora standard delle Isole Hawaii-Aleutine", "HAT": "Ora legale di Terranova", "HECU": "Ora legale di Cuba", "HEEG": "Ora legale della Groenlandia orientale", "HENOMX": "Ora legale del Messico nord-occidentale", "HEOG": "Ora legale della Groenlandia occidentale", "HEPM": "Ora legale di Saint-Pierre e Miquelon", "HEPMX": "Ora legale del Pacifico (Messico)", "HKST": "Ora legale di Hong Kong", "HKT": "Ora standard di Hong Kong", "HNCU": "Ora standard di Cuba", "HNEG": "Ora standard della Groenlandia orientale", "HNNOMX": "Ora standard del Messico nord-occidentale", "HNOG": "Ora standard della Groenlandia occidentale", "HNPM": "Ora standard di Saint-Pierre e Miquelon", "HNPMX": "Ora standard del Pacifico (Messico)", "HNT": "Ora standard di Terranova", "IST": "Ora standard dell’India", "JDT": "Ora legale del Giappone", "JST": "Ora standard del Giappone", "LHDT": "Ora legale di Lord Howe", "LHST": "Ora standard di Lord Howe", "MDT": "Ora legale Montagne Rocciose USA", "MESZ": "Ora legale dell’Europa centrale", "MEZ": "Ora standard dell’Europa centrale", "MST": "Ora standard Montagne Rocciose USA", "MYT": "Ora della Malesia", "NZDT": "Ora legale della Nuova Zelanda", "NZST": "Ora standard della Nuova Zelanda", "OESZ": "Ora legale dell’Europa orientale", "OEZ": "Ora standard dell’Europa orientale", "PDT": "Ora legale del Pacifico USA", "PST": "Ora standard del Pacifico USA", "SAST": "Ora dell’Africa meridionale", "SGT": "Ora di Singapore", "SRT": "Ora del Suriname", "TMST": "Ora legale del Turkmenistan", "TMT": "Ora standard del Turkmenistan", "UYST": "Ora legale dell’Uruguay", "UYT": "Ora standard dell’Uruguay", "VET": "Ora del Venezuela", "WARST": "Ora legale dell’Argentina occidentale", "WART": "Ora standard dell’Argentina occidentale", "WAST": "Ora legale dell’Africa occidentale", "WAT": "Ora standard dell’Africa occidentale", "WESZ": "Ora legale dell’Europa occidentale", "WEZ": "Ora standard dell’Europa occidentale", "WIB": "Ora dell’Indonesia occidentale", "WIT": "Ora dell’Indonesia orientale", "WITA": "Ora dell’Indonesia centrale", "∅∅∅": "Ora legale di Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b8b63d0d886b70a436900cf618138a9f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["it_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "it_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  "’",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"},
			monthsNarrow:           []string{"", "G", "F", "M", "A", "M", "G", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "G", "V", "S"},
			daysShort:              []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysWide:               []string{"domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"m.", "p."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"avanti Cristo", "dopo Cristo"},
			timezones:              map[string]string{"ACDT": "Ora legale dell’Australia centrale", "ACST": "Ora standard dell’Australia centrale", "ACWDT": "Ora legale dell’Australia centroccidentale", "ACWST": "Ora standard dell’Australia centroccidentale", "ADT": "Ora legale dell’Atlantico", "AEDT": "Ora legale dell’Australia orientale", "AEST": "Ora standard dell’Australia orientale", "AKDT": "Ora legale dell’Alaska", "AKST": "Ora standard dell’Alaska", "ARST": "Ora legale dell’Argentina", "ART": "Ora standard dell’Argentina", "AST": "Ora standard dell’Atlantico", "AWDT": "Ora legale dell’Australia occidentale", "AWST": "Ora standard dell’Australia occidentale", "BOT": "Ora della Bolivia", "BT": "Ora del Bhutan", "CAT": "Ora dell’Africa centrale", "CDT": "Ora legale centrale USA", "CHADT": "Ora legale delle Chatham", "CHAST": "Ora standard delle Chatham", "CLST": "Ora legale del Cile", "CLT": "Ora standard del Cile", "COST": "Ora legale della Colombia", "COT": "Ora standard della Colombia", "CST": "Ora standard centrale USA", "ChST": "Ora di Chamorro", "EAT": "Ora dell’Africa orientale", "ECT": "Ora dell’Ecuador", "EDT": "Ora legale orientale USA", "EST": "Ora standard orientale USA", "GFT": "Ora della Guiana francese", "GMT": "Ora del meridiano di Greenwich", "GST": "Ora del Golfo", "GYT": "Ora della Guyana", "HADT": "Ora legale delle Isole Hawaii-Aleutine", "HAST": "Ora standard delle Isole Hawaii-Aleutine", "HAT": "Ora legale di Terranova", "HECU": "Ora legale di Cuba", "HEEG": "Ora legale della Groenlandia orientale", "HENOMX": "Ora legale del Messico nord-occidentale", "HEOG": "Ora legale della Groenlandia occidentale", "HEPM": "Ora legale di Saint-Pierre e Miquelon", "HEPMX": "Ora legale del Pacifico (Messico)", "HKST": "Ora legale di Hong Kong", "HKT": "Ora standard di Hong Kong", "HNCU": "Ora standard di Cuba", "HNEG": "Ora standard della Groenlandia orientale", "HNNOMX": "Ora standard del Messico nord-occidentale", "HNOG": "Ora standard della Groenlandia occidentale", "HNPM": "Ora standard di Saint-Pierre e Miquelon", "HNPMX": "Ora standard del Pacifico (Messico)", "HNT": "Ora standard di Terranova", "IST": "Ora standard dell’India", "JDT": "Ora legale del Giappone", "JST": "Ora standard del Giappone", "LHDT": "Ora legale di Lord Howe", "LHST": "Ora standard di Lord Howe", "MDT": "Ora legale Montagne Rocciose USA", "MESZ": "Ora legale dell’Europa centrale", "MEZ": "Ora standard dell’Europa centrale", "MST": "Ora standard Montagne Rocciose USA", "MYT": "Ora della Malesia", "NZDT": "Ora legale della Nuova Zelanda", "NZST": "Ora standard della Nuova Zelanda", "OESZ": "Ora legale dell’Europa orientale", "OEZ": "Ora standard dell’Europa orientale", "PDT": "Ora legale del Pacifico USA", "PST": "Ora standard del Pacifico USA", "SAST": "Ora dell’Africa meridionale", "SGT": "Ora di Singapore", "SRT": "Ora del Suriname", "TMST": "Ora legale del Turkmenistan", "TMT": "Ora standard del Turkmenistan", "UYST": "Ora legale dell’Uruguay", "UYT": "Ora standard dell’Uruguay", "VET": "Ora del Venezuela", "WARST": "Ora legale dell’Argentina occidentale", "WART": "Ora standard dell’Argentina occidentale", "WAST": "Ora legale dell’Africa occidentale", "WAT": "Ora standard dell’Africa occidentale", "WESZ": "Ora legale dell’Europa occidentale", "WEZ": "Ora standard dell’Europa occidentale", "WIB": "Ora dell’Indonesia occidentale", "WIT": "Ora dell’Indonesia orientale", "WITA": "Ora dell’Indonesia centrale", "∅∅∅": "Ora legale dell’Amazzonia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a6430b96a17c400006d20ebabdf05e4c,
			fnFmtCurrency:         fnFmtCurrency_4924273cdfaa6d993990fdf2a2bad453,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_86454cdda5e540e71bc886ed08f9c66b,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b8b63d0d886b70a436900cf618138a9f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["it_it"] = func() locales.Translator {
		return &localen{

			locale:                 "it_IT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"},
			monthsNarrow:           []string{"", "G", "F", "M", "A", "M", "G", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "G", "V", "S"},
			daysShort:              []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysWide:               []string{"domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"m.", "p."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"avanti Cristo", "dopo Cristo"},
			timezones:              map[string]string{"ACDT": "Ora legale dell’Australia centrale", "ACST": "Ora standard dell’Australia centrale", "ACWDT": "Ora legale dell’Australia centroccidentale", "ACWST": "Ora standard dell’Australia centroccidentale", "ADT": "Ora legale dell’Atlantico", "AEDT": "Ora legale dell’Australia orientale", "AEST": "Ora standard dell’Australia orientale", "AKDT": "Ora legale dell’Alaska", "AKST": "Ora standard dell’Alaska", "ARST": "Ora legale dell’Argentina", "ART": "Ora standard dell’Argentina", "AST": "Ora standard dell’Atlantico", "AWDT": "Ora legale dell’Australia occidentale", "AWST": "Ora standard dell’Australia occidentale", "BOT": "Ora della Bolivia", "BT": "Ora del Bhutan", "CAT": "Ora dell’Africa centrale", "CDT": "Ora legale centrale USA", "CHADT": "Ora legale delle Chatham", "CHAST": "Ora standard delle Chatham", "CLST": "Ora legale del Cile", "CLT": "Ora standard del Cile", "COST": "Ora legale della Colombia", "COT": "Ora standard della Colombia", "CST": "Ora standard centrale USA", "ChST": "Ora di Chamorro", "EAT": "Ora dell’Africa orientale", "ECT": "Ora dell’Ecuador", "EDT": "Ora legale orientale USA", "EST": "Ora standard orientale USA", "GFT": "Ora della Guiana francese", "GMT": "Ora del meridiano di Greenwich", "GST": "Ora del Golfo", "GYT": "Ora della Guyana", "HADT": "Ora legale delle Isole Hawaii-Aleutine", "HAST": "Ora standard delle Isole Hawaii-Aleutine", "HAT": "Ora legale di Terranova", "HECU": "Ora legale di Cuba", "HEEG": "Ora legale della Groenlandia orientale", "HENOMX": "Ora legale del Messico nord-occidentale", "HEOG": "Ora legale della Groenlandia occidentale", "HEPM": "Ora legale di Saint-Pierre e Miquelon", "HEPMX": "Ora legale del Pacifico (Messico)", "HKST": "Ora legale di Hong Kong", "HKT": "Ora standard di Hong Kong", "HNCU": "Ora standard di Cuba", "HNEG": "Ora standard della Groenlandia orientale", "HNNOMX": "Ora standard del Messico nord-occidentale", "HNOG": "Ora standard della Groenlandia occidentale", "HNPM": "Ora standard di Saint-Pierre e Miquelon", "HNPMX": "Ora standard del Pacifico (Messico)", "HNT": "Ora standard di Terranova", "IST": "Ora standard dell’India", "JDT": "Ora legale del Giappone", "JST": "Ora standard del Giappone", "LHDT": "Ora legale di Lord Howe", "LHST": "Ora standard di Lord Howe", "MDT": "Ora legale Montagne Rocciose USA", "MESZ": "Ora legale dell’Europa centrale", "MEZ": "Ora standard dell’Europa centrale", "MST": "Ora standard Montagne Rocciose USA", "MYT": "Ora della Malesia", "NZDT": "Ora legale della Nuova Zelanda", "NZST": "Ora standard della Nuova Zelanda", "OESZ": "Ora legale dell’Europa orientale", "OEZ": "Ora standard dell’Europa orientale", "PDT": "Ora legale del Pacifico USA", "PST": "Ora standard del Pacifico USA", "SAST": "Ora dell’Africa meridionale", "SGT": "Ora di Singapore", "SRT": "Ora del Suriname", "TMST": "Ora legale del Turkmenistan", "TMT": "Ora standard del Turkmenistan", "UYST": "Ora legale dell’Uruguay", "UYT": "Ora standard dell’Uruguay", "VET": "Ora del Venezuela", "WARST": "Ora legale dell’Argentina occidentale", "WART": "Ora standard dell’Argentina occidentale", "WAST": "Ora legale dell’Africa occidentale", "WAT": "Ora standard dell’Africa occidentale", "WESZ": "Ora legale dell’Europa occidentale", "WEZ": "Ora standard dell’Europa occidentale", "WIB": "Ora dell’Indonesia occidentale", "WIT": "Ora dell’Indonesia orientale", "WITA": "Ora dell’Indonesia centrale", "∅∅∅": "Ora legale di Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b8b63d0d886b70a436900cf618138a9f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["it_sm"] = func() locales.Translator {
		return &localen{

			locale:                 "it_SM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"},
			monthsNarrow:           []string{"", "G", "F", "M", "A", "M", "G", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "G", "V", "S"},
			daysShort:              []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysWide:               []string{"domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"m.", "p."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"avanti Cristo", "dopo Cristo"},
			timezones:              map[string]string{"ACDT": "Ora legale dell’Australia centrale", "ACST": "Ora standard dell’Australia centrale", "ACWDT": "Ora legale dell’Australia centroccidentale", "ACWST": "Ora standard dell’Australia centroccidentale", "ADT": "Ora legale dell’Atlantico", "AEDT": "Ora legale dell’Australia orientale", "AEST": "Ora standard dell’Australia orientale", "AKDT": "Ora legale dell’Alaska", "AKST": "Ora standard dell’Alaska", "ARST": "Ora legale dell’Argentina", "ART": "Ora standard dell’Argentina", "AST": "Ora standard dell’Atlantico", "AWDT": "Ora legale dell’Australia occidentale", "AWST": "Ora standard dell’Australia occidentale", "BOT": "Ora della Bolivia", "BT": "Ora del Bhutan", "CAT": "Ora dell’Africa centrale", "CDT": "Ora legale centrale USA", "CHADT": "Ora legale delle Chatham", "CHAST": "Ora standard delle Chatham", "CLST": "Ora legale del Cile", "CLT": "Ora standard del Cile", "COST": "Ora legale della Colombia", "COT": "Ora standard della Colombia", "CST": "Ora standard centrale USA", "ChST": "Ora di Chamorro", "EAT": "Ora dell’Africa orientale", "ECT": "Ora dell’Ecuador", "EDT": "Ora legale orientale USA", "EST": "Ora standard orientale USA", "GFT": "Ora della Guiana francese", "GMT": "Ora del meridiano di Greenwich", "GST": "Ora del Golfo", "GYT": "Ora della Guyana", "HADT": "Ora legale delle Isole Hawaii-Aleutine", "HAST": "Ora standard delle Isole Hawaii-Aleutine", "HAT": "Ora legale di Terranova", "HECU": "Ora legale di Cuba", "HEEG": "Ora legale della Groenlandia orientale", "HENOMX": "Ora legale del Messico nord-occidentale", "HEOG": "Ora legale della Groenlandia occidentale", "HEPM": "Ora legale di Saint-Pierre e Miquelon", "HEPMX": "Ora legale del Pacifico (Messico)", "HKST": "Ora legale di Hong Kong", "HKT": "Ora standard di Hong Kong", "HNCU": "Ora standard di Cuba", "HNEG": "Ora standard della Groenlandia orientale", "HNNOMX": "Ora standard del Messico nord-occidentale", "HNOG": "Ora standard della Groenlandia occidentale", "HNPM": "Ora standard di Saint-Pierre e Miquelon", "HNPMX": "Ora standard del Pacifico (Messico)", "HNT": "Ora standard di Terranova", "IST": "Ora standard dell’India", "JDT": "Ora legale del Giappone", "JST": "Ora standard del Giappone", "LHDT": "Ora legale di Lord Howe", "LHST": "Ora standard di Lord Howe", "MDT": "MDT", "MESZ": "Ora legale dell’Europa centrale", "MEZ": "Ora standard dell’Europa centrale", "MST": "MST", "MYT": "Ora della Malesia", "NZDT": "Ora legale della Nuova Zelanda", "NZST": "Ora standard della Nuova Zelanda", "OESZ": "Ora legale dell’Europa orientale", "OEZ": "Ora standard dell’Europa orientale", "PDT": "Ora legale del Pacifico USA", "PST": "Ora standard del Pacifico USA", "SAST": "Ora dell’Africa meridionale", "SGT": "Ora di Singapore", "SRT": "Ora del Suriname", "TMST": "Ora legale del Turkmenistan", "TMT": "Ora standard del Turkmenistan", "UYST": "Ora legale dell’Uruguay", "UYT": "Ora standard dell’Uruguay", "VET": "Ora del Venezuela", "WARST": "Ora legale dell’Argentina occidentale", "WART": "Ora standard dell’Argentina occidentale", "WAST": "Ora legale dell’Africa occidentale", "WAT": "Ora standard dell’Africa occidentale", "WESZ": "Ora legale dell’Europa occidentale", "WEZ": "Ora standard dell’Europa occidentale", "WIB": "Ora dell’Indonesia occidentale", "WIT": "Ora dell’Indonesia orientale", "WITA": "Ora dell’Indonesia centrale", "∅∅∅": "Ora legale del Perù"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b8b63d0d886b70a436900cf618138a9f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["it_va"] = func() locales.Translator {
		return &localen{

			locale:                 "it_VA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"},
			monthsNarrow:           []string{"", "G", "F", "M", "A", "M", "G", "L", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"},
			daysAbbreviated:        []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysNarrow:             []string{"D", "L", "M", "M", "G", "V", "S"},
			daysShort:              []string{"dom", "lun", "mar", "mer", "gio", "ven", "sab"},
			daysWide:               []string{"domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"m.", "p."},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"aC", "dC"},
			erasWide:               []string{"avanti Cristo", "dopo Cristo"},
			timezones:              map[string]string{"ACDT": "Ora legale dell’Australia centrale", "ACST": "Ora standard dell’Australia centrale", "ACWDT": "Ora legale dell’Australia centroccidentale", "ACWST": "Ora standard dell’Australia centroccidentale", "ADT": "Ora legale dell’Atlantico", "AEDT": "Ora legale dell’Australia orientale", "AEST": "Ora standard dell’Australia orientale", "AKDT": "Ora legale dell’Alaska", "AKST": "Ora standard dell’Alaska", "ARST": "Ora legale dell’Argentina", "ART": "Ora standard dell’Argentina", "AST": "Ora standard dell’Atlantico", "AWDT": "Ora legale dell’Australia occidentale", "AWST": "Ora standard dell’Australia occidentale", "BOT": "Ora della Bolivia", "BT": "Ora del Bhutan", "CAT": "Ora dell’Africa centrale", "CDT": "Ora legale centrale USA", "CHADT": "Ora legale delle Chatham", "CHAST": "Ora standard delle Chatham", "CLST": "Ora legale del Cile", "CLT": "Ora standard del Cile", "COST": "Ora legale della Colombia", "COT": "Ora standard della Colombia", "CST": "Ora standard centrale USA", "ChST": "Ora di Chamorro", "EAT": "Ora dell’Africa orientale", "ECT": "Ora dell’Ecuador", "EDT": "Ora legale orientale USA", "EST": "Ora standard orientale USA", "GFT": "Ora della Guiana francese", "GMT": "Ora del meridiano di Greenwich", "GST": "Ora del Golfo", "GYT": "Ora della Guyana", "HADT": "Ora legale delle Isole Hawaii-Aleutine", "HAST": "Ora standard delle Isole Hawaii-Aleutine", "HAT": "Ora legale di Terranova", "HECU": "Ora legale di Cuba", "HEEG": "Ora legale della Groenlandia orientale", "HENOMX": "Ora legale del Messico nord-occidentale", "HEOG": "Ora legale della Groenlandia occidentale", "HEPM": "Ora legale di Saint-Pierre e Miquelon", "HEPMX": "Ora legale del Pacifico (Messico)", "HKST": "Ora legale di Hong Kong", "HKT": "Ora standard di Hong Kong", "HNCU": "Ora standard di Cuba", "HNEG": "Ora standard della Groenlandia orientale", "HNNOMX": "Ora standard del Messico nord-occidentale", "HNOG": "Ora standard della Groenlandia occidentale", "HNPM": "Ora standard di Saint-Pierre e Miquelon", "HNPMX": "Ora standard del Pacifico (Messico)", "HNT": "Ora standard di Terranova", "IST": "Ora standard dell’India", "JDT": "Ora legale del Giappone", "JST": "Ora standard del Giappone", "LHDT": "Ora legale di Lord Howe", "LHST": "Ora standard di Lord Howe", "MDT": "Ora legale Montagne Rocciose USA", "MESZ": "Ora legale dell’Europa centrale", "MEZ": "Ora standard dell’Europa centrale", "MST": "Ora standard Montagne Rocciose USA", "MYT": "Ora della Malesia", "NZDT": "Ora legale della Nuova Zelanda", "NZST": "Ora standard della Nuova Zelanda", "OESZ": "Ora legale dell’Europa orientale", "OEZ": "Ora standard dell’Europa orientale", "PDT": "Ora legale del Pacifico USA", "PST": "Ora standard del Pacifico USA", "SAST": "Ora dell’Africa meridionale", "SGT": "Ora di Singapore", "SRT": "Ora del Suriname", "TMST": "Ora legale del Turkmenistan", "TMT": "Ora standard del Turkmenistan", "UYST": "Ora legale dell’Uruguay", "UYT": "Ora standard dell’Uruguay", "VET": "Ora del Venezuela", "WARST": "Ora legale dell’Argentina occidentale", "WART": "Ora standard dell’Argentina occidentale", "WAST": "Ora legale dell’Africa occidentale", "WAT": "Ora standard dell’Africa occidentale", "WESZ": "Ora legale dell’Europa occidentale", "WEZ": "Ora standard dell’Europa occidentale", "WIB": "Ora dell’Indonesia occidentale", "WIT": "Ora dell’Indonesia orientale", "WITA": "Ora dell’Indonesia centrale", "∅∅∅": "Ora legale delle Azzorre"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b8b63d0d886b70a436900cf618138a9f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ja"] = func() locales.Translator {
		return &localen{

			locale:                 "ja",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "元", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "￥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:        []string{"日", "月", "火", "水", "木", "金", "土"},
			daysNarrow:             []string{"日", "月", "火", "水", "木", "金", "土"},
			daysShort:              []string{"日", "月", "火", "水", "木", "金", "土"},
			daysWide:               []string{"日曜日", "月曜日", "火曜日", "水曜日", "木曜日", "金曜日", "土曜日"},
			periodsAbbreviated:     []string{"午前", "午後"},
			periodsNarrow:          []string{"午前", "午後"},
			periodsWide:            []string{"午前", "午後"},
			erasAbbreviated:        []string{"紀元前", "西暦"},
			erasNarrow:             []string{"BC", "AD"},
			erasWide:               []string{"紀元前", "西暦"},
			timezones:              map[string]string{"ACDT": "オーストラリア中部夏時間", "ACST": "オーストラリア中部標準時", "ACWDT": "オーストラリア中西部夏時間", "ACWST": "オーストラリア中西部標準時", "ADT": "大西洋夏時間", "AEDT": "オーストラリア東部夏時間", "AEST": "オーストラリア東部標準時", "AKDT": "アラスカ夏時間", "AKST": "アラスカ標準時", "ARST": "アルゼンチン夏時間", "ART": "アルゼンチン標準時", "AST": "大西洋標準時", "AWDT": "オーストラリア西部夏時間", "AWST": "オーストラリア西部標準時", "BOT": "ボリビア時間", "BT": "ブータン時間", "CAT": "中央アフリカ時間", "CDT": "アメリカ中部夏時間", "CHADT": "チャタム夏時間", "CHAST": "チャタム標準時", "CLST": "チリ夏時間", "CLT": "チリ標準時", "COST": "コロンビア夏時間", "COT": "コロンビア標準時", "CST": "アメリカ中部標準時", "ChST": "チャモロ時間", "EAT": "東アフリカ時間", "ECT": "エクアドル時間", "EDT": "アメリカ東部夏時間", "EST": "アメリカ東部標準時", "GFT": "仏領ギアナ時間", "GMT": "グリニッジ標準時", "GST": "湾岸標準時", "GYT": "ガイアナ時間", "HADT": "ハワイ・アリューシャン夏時間", "HAST": "ハワイ・アリューシャン標準時", "HAT": "ニューファンドランド夏時間", "HECU": "キューバ夏時間", "HEEG": "グリーンランド東部夏時間", "HENOMX": "メキシコ北西部夏時間", "HEOG": "グリーンランド西部夏時間", "HEPM": "サンピエール島・ミクロン島夏時間", "HEPMX": "メキシコ太平洋夏時間", "HKST": "香港夏時間", "HKT": "香港標準時", "HNCU": "キューバ標準時", "HNEG": "グリーンランド東部標準時", "HNNOMX": "メキシコ北西部標準時", "HNOG": "グリーンランド西部標準時", "HNPM": "サンピエール島・ミクロン島標準時", "HNPMX": "メキシコ太平洋標準時", "HNT": "ニューファンドランド標準時", "IST": "インド標準時", "JDT": "日本夏時間", "JST": "日本標準時", "LHDT": "ロードハウ夏時間", "LHST": "ロードハウ標準時", "MDT": "アメリカ山地夏時間", "MESZ": "中央ヨーロッパ夏時間", "MEZ": "中央ヨーロッパ標準時", "MST": "アメリカ山地標準時", "MYT": "マレーシア時間", "NZDT": "ニュージーランド夏時間", "NZST": "ニュージーランド標準時", "OESZ": "東ヨーロッパ夏時間", "OEZ": "東ヨーロッパ標準時", "PDT": "アメリカ太平洋夏時間", "PST": "アメリカ太平洋標準時", "SAST": "南アフリカ標準時", "SGT": "シンガポール標準時", "SRT": "スリナム時間", "TMST": "トルクメニスタン夏時間", "TMT": "トルクメニスタン標準時", "UYST": "ウルグアイ夏時間", "UYT": "ウルグアイ標準時", "VET": "ベネズエラ時間", "WARST": "西部アルゼンチン夏時間", "WART": "西部アルゼンチン標準時", "WAST": "西アフリカ夏時間", "WAT": "西アフリカ標準時", "WESZ": "西ヨーロッパ夏時間", "WEZ": "西ヨーロッパ標準時", "WIB": "インドネシア西部時間", "WIT": "インドネシア東部時間", "WITA": "インドネシア中部時間", "∅∅∅": "ペルー夏時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_30a6b80fc168a8b0808040051491fc26,
			fnFmtDateShort:        fnFmtDateShort_933fcf2387ca32382dae956e0bc893a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5f445481e576332f9f6276f28f4cc61a,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ja_jp"] = func() locales.Translator {
		return &localen{

			locale:                 "ja_JP",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:        []string{"日", "月", "火", "水", "木", "金", "土"},
			daysNarrow:             []string{"日", "月", "火", "水", "木", "金", "土"},
			daysShort:              []string{"日", "月", "火", "水", "木", "金", "土"},
			daysWide:               []string{"日曜日", "月曜日", "火曜日", "水曜日", "木曜日", "金曜日", "土曜日"},
			periodsAbbreviated:     []string{"午前", "午後"},
			periodsNarrow:          []string{"午前", "午後"},
			periodsWide:            []string{"午前", "午後"},
			erasAbbreviated:        []string{"紀元前", "西暦"},
			erasNarrow:             []string{"BC", "AD"},
			erasWide:               []string{"紀元前", "西暦"},
			timezones:              map[string]string{"ACDT": "オーストラリア中部夏時間", "ACST": "オーストラリア中部標準時", "ACWDT": "オーストラリア中西部夏時間", "ACWST": "オーストラリア中西部標準時", "ADT": "大西洋夏時間", "AEDT": "オーストラリア東部夏時間", "AEST": "オーストラリア東部標準時", "AKDT": "アラスカ夏時間", "AKST": "アラスカ標準時", "ARST": "アルゼンチン夏時間", "ART": "アルゼンチン標準時", "AST": "大西洋標準時", "AWDT": "オーストラリア西部夏時間", "AWST": "オーストラリア西部標準時", "BOT": "ボリビア時間", "BT": "ブータン時間", "CAT": "中央アフリカ時間", "CDT": "アメリカ中部夏時間", "CHADT": "チャタム夏時間", "CHAST": "チャタム標準時", "CLST": "チリ夏時間", "CLT": "チリ標準時", "COST": "コロンビア夏時間", "COT": "コロンビア標準時", "CST": "アメリカ中部標準時", "ChST": "チャモロ時間", "EAT": "東アフリカ時間", "ECT": "エクアドル時間", "EDT": "アメリカ東部夏時間", "EST": "アメリカ東部標準時", "GFT": "仏領ギアナ時間", "GMT": "グリニッジ標準時", "GST": "湾岸標準時", "GYT": "ガイアナ時間", "HADT": "ハワイ・アリューシャン夏時間", "HAST": "ハワイ・アリューシャン標準時", "HAT": "ニューファンドランド夏時間", "HECU": "キューバ夏時間", "HEEG": "グリーンランド東部夏時間", "HENOMX": "メキシコ北西部夏時間", "HEOG": "グリーンランド西部夏時間", "HEPM": "サンピエール島・ミクロン島夏時間", "HEPMX": "メキシコ太平洋夏時間", "HKST": "香港夏時間", "HKT": "香港標準時", "HNCU": "キューバ標準時", "HNEG": "グリーンランド東部標準時", "HNNOMX": "メキシコ北西部標準時", "HNOG": "グリーンランド西部標準時", "HNPM": "サンピエール島・ミクロン島標準時", "HNPMX": "メキシコ太平洋標準時", "HNT": "ニューファンドランド標準時", "IST": "インド標準時", "JDT": "日本夏時間", "JST": "日本標準時", "LHDT": "ロードハウ夏時間", "LHST": "ロードハウ標準時", "MDT": "アメリカ山地夏時間", "MESZ": "中央ヨーロッパ夏時間", "MEZ": "中央ヨーロッパ標準時", "MST": "アメリカ山地標準時", "MYT": "マレーシア時間", "NZDT": "ニュージーランド夏時間", "NZST": "ニュージーランド標準時", "OESZ": "東ヨーロッパ夏時間", "OEZ": "東ヨーロッパ標準時", "PDT": "アメリカ太平洋夏時間", "PST": "アメリカ太平洋標準時", "SAST": "南アフリカ標準時", "SGT": "シンガポール標準時", "SRT": "スリナム時間", "TMST": "トルクメニスタン夏時間", "TMT": "トルクメニスタン標準時", "UYST": "ウルグアイ夏時間", "UYT": "ウルグアイ標準時", "VET": "ベネズエラ時間", "WARST": "西部アルゼンチン夏時間", "WART": "西部アルゼンチン標準時", "WAST": "西アフリカ夏時間", "WAT": "西アフリカ標準時", "WESZ": "西ヨーロッパ夏時間", "WEZ": "西ヨーロッパ標準時", "WIB": "インドネシア西部時間", "WIT": "インドネシア東部時間", "WITA": "インドネシア中部時間", "∅∅∅": "ペルー夏時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_30a6b80fc168a8b0808040051491fc26,
			fnFmtDateShort:        fnFmtDateShort_933fcf2387ca32382dae956e0bc893a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5f445481e576332f9f6276f28f4cc61a,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["jgo"] = func() locales.Translator {
		return &localen{

			locale:                 "jgo",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Nduŋmbi Saŋ", "Pɛsaŋ Pɛ́pá", "Pɛsaŋ Pɛ́tát", "Pɛsaŋ Pɛ́nɛ́kwa", "Pɛsaŋ Pataa", "Pɛsaŋ Pɛ́nɛ́ntúkú", "Pɛsaŋ Saambá", "Pɛsaŋ Pɛ́nɛ́fɔm", "Pɛsaŋ Pɛ́nɛ́pfúꞋú", "Pɛsaŋ Nɛgɛ́m", "Pɛsaŋ Ntsɔ̌pmɔ́", "Pɛsaŋ Ntsɔ̌ppá"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Nduŋmbi Saŋ", "Pɛsaŋ Pɛ́pá", "Pɛsaŋ Pɛ́tát", "Pɛsaŋ Pɛ́nɛ́kwa", "Pɛsaŋ Pataa", "Pɛsaŋ Pɛ́nɛ́ntúkú", "Pɛsaŋ Saambá", "Pɛsaŋ Pɛ́nɛ́fɔm", "Pɛsaŋ Pɛ́nɛ́pfúꞋú", "Pɛsaŋ Nɛgɛ́m", "Pɛsaŋ Ntsɔ̌pmɔ́", "Pɛsaŋ Ntsɔ̌ppá"},
			daysAbbreviated:        []string{"Sɔ́ndi", "Mɔ́ndi", "Ápta Mɔ́ndi", "Wɛ́nɛsɛdɛ", "Tɔ́sɛdɛ", "Fɛlâyɛdɛ", "Sásidɛ"},
			daysNarrow:             []string{"Sɔ́", "Mɔ́", "ÁM", "Wɛ́", "Tɔ́", "Fɛ", "Sá"},
			daysWide:               []string{"Sɔ́ndi", "Mɔ́ndi", "Ápta Mɔ́ndi", "Wɛ́nɛsɛdɛ", "Tɔ́sɛdɛ", "Fɛlâyɛdɛ", "Sásidɛ"},
			periodsAbbreviated:     []string{"mbaꞌmbaꞌ", "ŋka mbɔ́t nji"},
			periodsWide:            []string{"mbaꞌmbaꞌ", "ŋka mbɔ́t nji"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"tsɛttsɛt mɛŋguꞌ mi ɛ́ lɛɛnɛ Kɛlísɛtɔ gɔ ńɔ́", "tsɛttsɛt mɛŋguꞌ mi ɛ́ fúnɛ Kɛlísɛtɔ tɔ́ mɔ́"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["jgo_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "jgo_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Nduŋmbi Saŋ", "Pɛsaŋ Pɛ́pá", "Pɛsaŋ Pɛ́tát", "Pɛsaŋ Pɛ́nɛ́kwa", "Pɛsaŋ Pataa", "Pɛsaŋ Pɛ́nɛ́ntúkú", "Pɛsaŋ Saambá", "Pɛsaŋ Pɛ́nɛ́fɔm", "Pɛsaŋ Pɛ́nɛ́pfúꞋú", "Pɛsaŋ Nɛgɛ́m", "Pɛsaŋ Ntsɔ̌pmɔ́", "Pɛsaŋ Ntsɔ̌ppá"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Nduŋmbi Saŋ", "Pɛsaŋ Pɛ́pá", "Pɛsaŋ Pɛ́tát", "Pɛsaŋ Pɛ́nɛ́kwa", "Pɛsaŋ Pataa", "Pɛsaŋ Pɛ́nɛ́ntúkú", "Pɛsaŋ Saambá", "Pɛsaŋ Pɛ́nɛ́fɔm", "Pɛsaŋ Pɛ́nɛ́pfúꞋú", "Pɛsaŋ Nɛgɛ́m", "Pɛsaŋ Ntsɔ̌pmɔ́", "Pɛsaŋ Ntsɔ̌ppá"},
			daysAbbreviated:        []string{"Sɔ́ndi", "Mɔ́ndi", "Ápta Mɔ́ndi", "Wɛ́nɛsɛdɛ", "Tɔ́sɛdɛ", "Fɛlâyɛdɛ", "Sásidɛ"},
			daysNarrow:             []string{"Sɔ́", "Mɔ́", "ÁM", "Wɛ́", "Tɔ́", "Fɛ", "Sá"},
			daysWide:               []string{"Sɔ́ndi", "Mɔ́ndi", "Ápta Mɔ́ndi", "Wɛ́nɛsɛdɛ", "Tɔ́sɛdɛ", "Fɛlâyɛdɛ", "Sásidɛ"},
			periodsAbbreviated:     []string{"mbaꞌmbaꞌ", "ŋka mbɔ́t nji"},
			periodsWide:            []string{"mbaꞌmbaꞌ", "ŋka mbɔ́t nji"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"tsɛttsɛt mɛŋguꞌ mi ɛ́ lɛɛnɛ Kɛlísɛtɔ gɔ ńɔ́", "tsɛttsɛt mɛŋguꞌ mi ɛ́ fúnɛ Kɛlísɛtɔ tɔ́ mɔ́"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["jmc"] = func() locales.Translator {
		return &localen{

			locale:             "jmc",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprilyi", "Mei", "Junyi", "Julyai", "Agusti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Jumapilyi", "Jumatatuu", "Jumanne", "Jumatanu", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"utuko", "kyiukonyi"},
			periodsWide:        []string{"utuko", "kyiukonyi"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Kristu", "Baada ya Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["jmc_tz"] = func() locales.Translator {
		return &localen{

			locale:             "jmc_TZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprilyi", "Mei", "Junyi", "Julyai", "Agusti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Jumapilyi", "Jumatatuu", "Jumanne", "Jumatanu", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"utuko", "kyiukonyi"},
			periodsWide:        []string{"utuko", "kyiukonyi"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Kristu", "Baada ya Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["jv"] = func() locales.Translator {
		return &localen{

			locale:                 "jv",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "Rp", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Agt", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Maret", "April", "Mei", "Juni", "Juli", "Agustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:        []string{"Ahad", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysNarrow:             []string{"A", "S", "S", "R", "K", "J", "S"},
			daysShort:              []string{"Ahad", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysWide:               []string{"Ahad", "Senin", "Selasa", "Rabu", "Kamis", "Jumat", "Sabtu"},
			periodsAbbreviated:     []string{"Isuk", "Wengi"},
			periodsNarrow:          []string{"Isuk", "Wengi"},
			periodsWide:            []string{"Isuk", "Wengi"},
			erasAbbreviated:        []string{"SM", "M"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Sakdurunge Masehi", "Masehi"},
			timezones:              map[string]string{"ACDT": "Wektu Ketigo Australia Tengah", "ACST": "Wektu Standar Australia Tengah", "ACWDT": "Wektu Ketigo Australia Tengah sisih Kulon", "ACWST": "Wektu Standar Australia Tengah sisih Kulon", "ADT": "Wektu Ketigo Atlantik", "AEDT": "Wektu Ketigo Australia sisih Wetan", "AEST": "Wektu Standar Australia sisih Wetan", "AKDT": "Wektu Ketigo Alaska", "AKST": "Wektu Standar Alaska", "ARST": "Wektu Ketigo Argentina", "ART": "Wektu Standar Argentina", "AST": "Wektu Standar Atlantik", "AWDT": "Wektu Ketigo Australia sisih Kulon", "AWST": "Wektu Standar Australia sisih Kulon", "BOT": "Wektu Bolivia", "BT": "Wektu Bhutan", "CAT": "Wektu Afrika Tengah", "CDT": "Wektu Ketigo Tengah", "CHADT": "Wektu Ketigo Chatham", "CHAST": "Wektu Standar Chatham", "CLST": "Wektu Ketigo Chili", "CLT": "Wektu Standar Chili", "COST": "Wektu Ketigo Kolombia", "COT": "Wektu Standar Kolombia", "CST": "Wektu Standar Tengah", "ChST": "Wektu Chamorro", "EAT": "Wektu Afrika Wetan", "ECT": "Wektu Ekuador", "EDT": "Wektu Ketigo sisih Wetah", "EST": "Wektu Standar sisih Wetan", "GFT": "Wektu Guiana Prancis", "GMT": "Wektu Rerata Greenwich", "GST": "Wektu Standar Teluk", "GYT": "Wektu Guyana", "HADT": "Wektu Ketigo Hawaii-Aleutian", "HAST": "Wektu Standar Hawaii-Aleutian", "HAT": "Wektu Ketigo Newfoundland", "HECU": "Wektu Ketigo Kuba", "HEEG": "Wektu Ketigo Grinland Wetan", "HENOMX": "Wektu Ketigo Meksiko Lor-Kulon", "HEOG": "Wektu Ketigo Grinland Kulon", "HEPM": "Wektu Ketigo Santa Pierre lan Miquelon", "HEPMX": "Wektu Ketigo Pasifik Meksiko", "HKST": "Wektu Ketigo Hong Kong", "HKT": "Wektu Standar Hong Kong", "HNCU": "Wektu Standar Kuba", "HNEG": "Wektu Standar Grinland Wetan", "HNNOMX": "Wektu Standar Meksiko Lor-Kulon", "HNOG": "Wektu Standar Grinland Kulon", "HNPM": "Wektu Standar Santa Pierre lan Miquelon", "HNPMX": "Wektu Standar Pasifik Meksiko", "HNT": "Wektu Standar Newfoundland", "IST": "Wektu Standar India", "JDT": "Wektu Ketigo Jepang", "JST": "Wektu Standar Jepang", "LHDT": "Wektu Ketigo Lord Howe", "LHST": "Wektu Standar Lord Howe", "MDT": "Wektu Ketigo Giri", "MESZ": "Wektu Ketigo Eropa Tengah", "MEZ": "Wektu Standar Eropa Tengah", "MST": "Wektu Standar Giri", "MYT": "Wektu Malaysia", "NZDT": "Wektu Ketigo Selandia Anyar", "NZST": "Wektu Standar Selandia Anyar", "OESZ": "Wektu Ketigo Eropa sisih Wetan", "OEZ": "Wektu Standar Eropa sisih Wetan", "PDT": "Wektu Ketigo Pasifik", "PST": "Wektu Standar Pasifik", "SAST": "Wektu Standar Afrika Kidul", "SGT": "Wektu Singapura", "SRT": "Wektu Suriname", "TMST": "Wektu Ketigo Turkmenistan", "TMT": "Wektu Standar Turkmenistan", "UYST": "Wektu Ketigo Uruguay", "UYT": "Wektu Standar Uruguay", "VET": "Wektu Venezuela", "WARST": "Wektu Ketigo Argentina sisih Kulon", "WART": "Wektu Standar Argentina sisih Kulon", "WAST": "Wektu Ketigo Afrika Kulon", "WAT": "Wektu Standar Afrika Kulon", "WESZ": "Wektu Ketigo Eropa sisih Kulon", "WEZ": "Wektu Standar Eropa sisih Kulon", "WIB": "Wektu Indonesia sisih Kulon", "WIT": "Wektu Indonesia sisih Wetan", "WITA": "Wektu Indonesia Tengah", "∅∅∅": "Wektu Ketigo Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["jv_id"] = func() locales.Translator {
		return &localen{

			locale:                 "jv_ID",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Agt", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Maret", "April", "Mei", "Juni", "Juli", "Agustus", "September", "Oktober", "November", "Desember"},
			daysAbbreviated:        []string{"Ahad", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysNarrow:             []string{"A", "S", "S", "R", "K", "J", "S"},
			daysShort:              []string{"Ahad", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"},
			daysWide:               []string{"Ahad", "Senin", "Selasa", "Rabu", "Kamis", "Jumat", "Sabtu"},
			periodsAbbreviated:     []string{"Isuk", "Wengi"},
			periodsNarrow:          []string{"Isuk", "Wengi"},
			periodsWide:            []string{"Isuk", "Wengi"},
			erasAbbreviated:        []string{"SM", "M"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Sakdurunge Masehi", "Masehi"},
			timezones:              map[string]string{"ACDT": "Wektu Ketigo Australia Tengah", "ACST": "Wektu Standar Australia Tengah", "ACWDT": "Wektu Ketigo Australia Tengah sisih Kulon", "ACWST": "Wektu Standar Australia Tengah sisih Kulon", "ADT": "Wektu Ketigo Atlantik", "AEDT": "Wektu Ketigo Australia sisih Wetan", "AEST": "Wektu Standar Australia sisih Wetan", "AKDT": "Wektu Ketigo Alaska", "AKST": "Wektu Standar Alaska", "ARST": "Wektu Ketigo Argentina", "ART": "Wektu Standar Argentina", "AST": "Wektu Standar Atlantik", "AWDT": "Wektu Ketigo Australia sisih Kulon", "AWST": "Wektu Standar Australia sisih Kulon", "BOT": "Wektu Bolivia", "BT": "Wektu Bhutan", "CAT": "Wektu Afrika Tengah", "CDT": "Wektu Ketigo Tengah", "CHADT": "Wektu Ketigo Chatham", "CHAST": "Wektu Standar Chatham", "CLST": "Wektu Ketigo Chili", "CLT": "Wektu Standar Chili", "COST": "Wektu Ketigo Kolombia", "COT": "Wektu Standar Kolombia", "CST": "Wektu Standar Tengah", "ChST": "Wektu Chamorro", "EAT": "Wektu Afrika Wetan", "ECT": "Wektu Ekuador", "EDT": "Wektu Ketigo sisih Wetah", "EST": "Wektu Standar sisih Wetan", "GFT": "Wektu Guiana Prancis", "GMT": "Wektu Rerata Greenwich", "GST": "Wektu Standar Teluk", "GYT": "Wektu Guyana", "HADT": "Wektu Ketigo Hawaii-Aleutian", "HAST": "Wektu Standar Hawaii-Aleutian", "HAT": "Wektu Ketigo Newfoundland", "HECU": "Wektu Ketigo Kuba", "HEEG": "Wektu Ketigo Grinland Wetan", "HENOMX": "Wektu Ketigo Meksiko Lor-Kulon", "HEOG": "Wektu Ketigo Grinland Kulon", "HEPM": "Wektu Ketigo Santa Pierre lan Miquelon", "HEPMX": "Wektu Ketigo Pasifik Meksiko", "HKST": "Wektu Ketigo Hong Kong", "HKT": "Wektu Standar Hong Kong", "HNCU": "Wektu Standar Kuba", "HNEG": "Wektu Standar Grinland Wetan", "HNNOMX": "Wektu Standar Meksiko Lor-Kulon", "HNOG": "Wektu Standar Grinland Kulon", "HNPM": "Wektu Standar Santa Pierre lan Miquelon", "HNPMX": "Wektu Standar Pasifik Meksiko", "HNT": "Wektu Standar Newfoundland", "IST": "Wektu Standar India", "JDT": "Wektu Ketigo Jepang", "JST": "Wektu Standar Jepang", "LHDT": "Wektu Ketigo Lord Howe", "LHST": "Wektu Standar Lord Howe", "MDT": "Wektu Ketigo Giri", "MESZ": "Wektu Ketigo Eropa Tengah", "MEZ": "Wektu Standar Eropa Tengah", "MST": "Wektu Standar Giri", "MYT": "Wektu Malaysia", "NZDT": "Wektu Ketigo Selandia Anyar", "NZST": "Wektu Standar Selandia Anyar", "OESZ": "Wektu Ketigo Eropa sisih Wetan", "OEZ": "Wektu Standar Eropa sisih Wetan", "PDT": "Wektu Ketigo Pasifik", "PST": "Wektu Standar Pasifik", "SAST": "Wektu Standar Afrika Kidul", "SGT": "Wektu Singapura", "SRT": "Wektu Suriname", "TMST": "Wektu Ketigo Turkmenistan", "TMT": "Wektu Standar Turkmenistan", "UYST": "Wektu Ketigo Uruguay", "UYT": "Wektu Standar Uruguay", "VET": "Wektu Venezuela", "WARST": "Wektu Ketigo Argentina sisih Kulon", "WART": "Wektu Standar Argentina sisih Kulon", "WAST": "Wektu Ketigo Afrika Kulon", "WAT": "Wektu Standar Afrika Kulon", "WESZ": "Wektu Ketigo Eropa sisih Kulon", "WEZ": "Wektu Standar Eropa sisih Kulon", "WIB": "Wektu Indonesia sisih Kulon", "WIT": "Wektu Indonesia sisih Wetan", "WITA": "Wektu Indonesia Tengah", "∅∅∅": "Wektu Ketigo Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ka"] = func() locales.Translator {
		return &localen{

			locale:                 "ka",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "₾", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "იან", "თებ", "მარ", "აპრ", "მაი", "ივნ", "ივლ", "აგვ", "სექ", "ოქტ", "ნოე", "დეკ"},
			monthsNarrow:           []string{"", "ი", "თ", "მ", "ა", "მ", "ი", "ი", "ა", "ს", "ო", "ნ", "დ"},
			monthsWide:             []string{"", "იანვარი", "თებერვალი", "მარტი", "აპრილი", "მაისი", "ივნისი", "ივლისი", "აგვისტო", "სექტემბერი", "ოქტომბერი", "ნოემბერი", "დეკემბერი"},
			daysAbbreviated:        []string{"კვი", "ორშ", "სამ", "ოთხ", "ხუთ", "პარ", "შაბ"},
			daysNarrow:             []string{"კ", "ო", "ს", "ო", "ხ", "პ", "შ"},
			daysShort:              []string{"კვ", "ორ", "სმ", "ოთ", "ხთ", "პრ", "შბ"},
			daysWide:               []string{"კვირა", "ორშაბათი", "სამშაბათი", "ოთხშაბათი", "ხუთშაბათი", "პარასკევი", "შაბათი"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"ძვ. წ.", "ახ. წ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ძველი წელთაღრიცხვით", "ახალი წელთაღრიცხვით"},
			timezones:              map[string]string{"ACDT": "ავსტრალიის ცენტრალური ზაფხულის დრო", "ACST": "ავსტრალიის ცენტრალური სტანდარტული დრო", "ACWDT": "ცენტრალური და დასავლეთ ავსტრალიის ზაფხულის დრო", "ACWST": "ცენტრალური და დასავლეთ ავსტრალიის სტანდარტული დრო", "ADT": "ატლანტიკის ოკეანის ზაფხულის დრო", "AEDT": "აღმოსავლეთ ავსტრალიის ზაფხულის დრო", "AEST": "აღმოსავლეთ ავსტრალიის სტანდარტული დრო", "AKDT": "ალასკის ზაფხულის დრო", "AKST": "ალასკის სტანდარტული დრო", "ARST": "არგენტინის ზაფხულის დრო", "ART": "არგენტინის სტანდარტული დრო", "AST": "ატლანტიკის ოკეანის სტანდარტული დრო", "AWDT": "დასავლეთ ავსტრალიის ზაფხულის დრო", "AWST": "დასავლეთ ავსტრალიის სტანდარტული დრო", "BOT": "ბოლივიის დრო", "BT": "ბუტანის დრო", "CAT": "ცენტრალური აფრიკის დრო", "CDT": "ჩრდილოეთ ამერიკის ცენტრალური ზაფხულის დრო", "CHADT": "ჩატემის ზაფხულის დრო", "CHAST": "ჩატემის სტანდარტული დრო", "CLST": "ჩილეს ზაფხულის დრო", "CLT": "ჩილეს სტანდარტული დრო", "COST": "კოლუმბიის ზაფხულის დრო", "COT": "კოლუმბიის სტანდარტული დრო", "CST": "ჩრდილოეთ ამერიკის ცენტრალური სტანდარტული დრო", "ChST": "ჩამოროს დრო", "EAT": "აღმოსავლეთ აფრიკის დრო", "ECT": "ეკვადორის დრო", "EDT": "ჩრდილოეთ ამერიკის აღმოსავლეთის ზაფხულის დრო", "EST": "ჩრდილოეთ ამერიკის აღმოსავლეთის სტანდარტული დრო", "GFT": "საფრანგეთის გვიანის დრო", "GMT": "გრინვიჩის საშუალო დრო", "GST": "სპარსეთის ყურის სტანდარტული დრო", "GYT": "გაიანის დრო", "HADT": "ჰავაისა და ალეუტის ზაფხულის დრო", "HAST": "ჰავაისა და ალეუტის სტანდარტული დრო", "HAT": "ნიუფაუნდლენდის ზაფხულის დრო", "HECU": "კუბის ზაფხულის დრო", "HEEG": "აღმოსავლეთ გრენლანდიის ზაფხულის დრო", "HENOMX": "ჩრდილო-დასავლეთ მექსიკის ზაფხულის დრო", "HEOG": "დასავლეთ გრენლანდიის ზაფხულის დრო", "HEPM": "სენ-პიერის და მიკელონის ზაფხულის დრო", "HEPMX": "მექსიკის წყნარი ოკეანის ზაფხულის დრო", "HKST": "ჰონკონგის ზაფხულის დრო", "HKT": "ჰონკონგის სტანდარტული დრო", "HNCU": "კუბის სტანდარტული დრო", "HNEG": "აღმოსავლეთ გრენლანდიის სტანდარტული დრო", "HNNOMX": "ჩრდილო-დასავლეთ მექსიკის დრო", "HNOG": "დასავლეთ გრენლანდიის სტანდარტული დრო", "HNPM": "სენ-პიერის და მიკელონის სტანდარტული დრო", "HNPMX": "მექსიკის წყნარი ოკეანის სტანდარტული დრო", "HNT": "ნიუფაუნდლენდის სტანდარტული დრო", "IST": "ინდოეთის დრო", "JDT": "იაპონიის ზაფხულის დრო", "JST": "იაპონიის სტანდარტული დრო", "LHDT": "ლორდ-ჰაუს ზაფხულის დრო", "LHST": "ლორდ-ჰაუს სტანდარტული დრო", "MDT": "ჩრდილოეთ ამერიკის მაუნთინის ზაფხულის დრო", "MESZ": "ცენტრალური ევროპის ზაფხულის დრო", "MEZ": "ცენტრალური ევროპის სტანდარტული დრო", "MST": "ჩრდილოეთ ამერიკის მაუნთინის სტანდარტული დრო", "MYT": "მალაიზიის დრო", "NZDT": "ახალი ზელანდიის ზაფხულის დრო", "NZST": "ახალი ზელანდიის სტანდარტული დრო", "OESZ": "აღმოსავლეთ ევროპის ზაფხულის დრო", "OEZ": "აღმოსავლეთ ევროპის სტანდარტული დრო", "PDT": "ჩრდილოეთ ამერიკის წყნარი ოკეანის ზაფხულის დრო", "PST": "ჩრდილოეთ ამერიკის წყნარი ოკეანის სტანდარტული დრო", "SAST": "სამხრეთ აფრიკის დრო", "SGT": "სინგაპურის დრო", "SRT": "სურინამის დრო", "TMST": "თურქმენეთის ზაფხულის დრო", "TMT": "თურქმენეთის სტანდარტული დრო", "UYST": "ურუგვაის ზაფხულის დრო", "UYT": "ურუგვაის სტანდარტული დრო", "VET": "ვენესუელის დრო", "WARST": "დასავლეთ არგენტინის ზაფხულის დრო", "WART": "დასავლეთ არგენტინის სტანდარტული დრო", "WAST": "დასავლეთ აფრიკის ზაფხულის დრო", "WAT": "დასავლეთ აფრიკის სტანდარტული დრო", "WESZ": "დასავლეთ ევროპის ზაფხულის დრო", "WEZ": "დასავლეთ ევროპის სტანდარტული დრო", "WIB": "დასავლეთ ინდონეზიის დრო", "WIT": "აღმოსავლეთ ინდონეზიის დრო", "WITA": "ცენტრალური ინდონეზიის დრო", "∅∅∅": "პერუს ზაფხულის დრო"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_4edbf4e11f17243cbd3f53cf42390fdf,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_8f50deadc7a225aae72abf7135535a4e,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_d9d734a5db809617a5116023efed1940,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e975bcc2318a5d1ed3b85b6356649a29,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ka_ge"] = func() locales.Translator {
		return &localen{

			locale:                 "ka_GE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "იან", "თებ", "მარ", "აპრ", "მაი", "ივნ", "ივლ", "აგვ", "სექ", "ოქტ", "ნოე", "დეკ"},
			monthsNarrow:           []string{"", "ი", "თ", "მ", "ა", "მ", "ი", "ი", "ა", "ს", "ო", "ნ", "დ"},
			monthsWide:             []string{"", "იანვარი", "თებერვალი", "მარტი", "აპრილი", "მაისი", "ივნისი", "ივლისი", "აგვისტო", "სექტემბერი", "ოქტომბერი", "ნოემბერი", "დეკემბერი"},
			daysAbbreviated:        []string{"კვი", "ორშ", "სამ", "ოთხ", "ხუთ", "პარ", "შაბ"},
			daysNarrow:             []string{"კ", "ო", "ს", "ო", "ხ", "პ", "შ"},
			daysShort:              []string{"კვ", "ორ", "სმ", "ოთ", "ხთ", "პრ", "შბ"},
			daysWide:               []string{"კვირა", "ორშაბათი", "სამშაბათი", "ოთხშაბათი", "ხუთშაბათი", "პარასკევი", "შაბათი"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"ძვ. წ.", "ახ. წ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ძველი წელთაღრიცხვით", "ახალი წელთაღრიცხვით"},
			timezones:              map[string]string{"ACDT": "ავსტრალიის ცენტრალური ზაფხულის დრო", "ACST": "ავსტრალიის ცენტრალური სტანდარტული დრო", "ACWDT": "ცენტრალური და დასავლეთ ავსტრალიის ზაფხულის დრო", "ACWST": "ცენტრალური და დასავლეთ ავსტრალიის სტანდარტული დრო", "ADT": "ატლანტიკის ოკეანის ზაფხულის დრო", "AEDT": "აღმოსავლეთ ავსტრალიის ზაფხულის დრო", "AEST": "აღმოსავლეთ ავსტრალიის სტანდარტული დრო", "AKDT": "ალასკის ზაფხულის დრო", "AKST": "ალასკის სტანდარტული დრო", "ARST": "არგენტინის ზაფხულის დრო", "ART": "არგენტინის სტანდარტული დრო", "AST": "ატლანტიკის ოკეანის სტანდარტული დრო", "AWDT": "დასავლეთ ავსტრალიის ზაფხულის დრო", "AWST": "დასავლეთ ავსტრალიის სტანდარტული დრო", "BOT": "ბოლივიის დრო", "BT": "ბუტანის დრო", "CAT": "ცენტრალური აფრიკის დრო", "CDT": "ჩრდილოეთ ამერიკის ცენტრალური ზაფხულის დრო", "CHADT": "ჩატემის ზაფხულის დრო", "CHAST": "ჩატემის სტანდარტული დრო", "CLST": "ჩილეს ზაფხულის დრო", "CLT": "ჩილეს სტანდარტული დრო", "COST": "კოლუმბიის ზაფხულის დრო", "COT": "კოლუმბიის სტანდარტული დრო", "CST": "ჩრდილოეთ ამერიკის ცენტრალური სტანდარტული დრო", "ChST": "ჩამოროს დრო", "EAT": "აღმოსავლეთ აფრიკის დრო", "ECT": "ეკვადორის დრო", "EDT": "ჩრდილოეთ ამერიკის აღმოსავლეთის ზაფხულის დრო", "EST": "ჩრდილოეთ ამერიკის აღმოსავლეთის სტანდარტული დრო", "GFT": "საფრანგეთის გვიანის დრო", "GMT": "გრინვიჩის საშუალო დრო", "GST": "სპარსეთის ყურის სტანდარტული დრო", "GYT": "გაიანის დრო", "HADT": "ჰავაისა და ალეუტის ზაფხულის დრო", "HAST": "ჰავაისა და ალეუტის სტანდარტული დრო", "HAT": "ნიუფაუნდლენდის ზაფხულის დრო", "HECU": "კუბის ზაფხულის დრო", "HEEG": "აღმოსავლეთ გრენლანდიის ზაფხულის დრო", "HENOMX": "ჩრდილო-დასავლეთ მექსიკის ზაფხულის დრო", "HEOG": "დასავლეთ გრენლანდიის ზაფხულის დრო", "HEPM": "სენ-პიერის და მიკელონის ზაფხულის დრო", "HEPMX": "მექსიკის წყნარი ოკეანის ზაფხულის დრო", "HKST": "ჰონკონგის ზაფხულის დრო", "HKT": "ჰონკონგის სტანდარტული დრო", "HNCU": "კუბის სტანდარტული დრო", "HNEG": "აღმოსავლეთ გრენლანდიის სტანდარტული დრო", "HNNOMX": "ჩრდილო-დასავლეთ მექსიკის დრო", "HNOG": "დასავლეთ გრენლანდიის სტანდარტული დრო", "HNPM": "სენ-პიერის და მიკელონის სტანდარტული დრო", "HNPMX": "მექსიკის წყნარი ოკეანის სტანდარტული დრო", "HNT": "ნიუფაუნდლენდის სტანდარტული დრო", "IST": "ინდოეთის დრო", "JDT": "იაპონიის ზაფხულის დრო", "JST": "იაპონიის სტანდარტული დრო", "LHDT": "ლორდ-ჰაუს ზაფხულის დრო", "LHST": "ლორდ-ჰაუს სტანდარტული დრო", "MDT": "ჩრდილოეთ ამერიკის მაუნთინის ზაფხულის დრო", "MESZ": "ცენტრალური ევროპის ზაფხულის დრო", "MEZ": "ცენტრალური ევროპის სტანდარტული დრო", "MST": "ჩრდილოეთ ამერიკის მაუნთინის სტანდარტული დრო", "MYT": "მალაიზიის დრო", "NZDT": "ახალი ზელანდიის ზაფხულის დრო", "NZST": "ახალი ზელანდიის სტანდარტული დრო", "OESZ": "აღმოსავლეთ ევროპის ზაფხულის დრო", "OEZ": "აღმოსავლეთ ევროპის სტანდარტული დრო", "PDT": "ჩრდილოეთ ამერიკის წყნარი ოკეანის ზაფხულის დრო", "PST": "ჩრდილოეთ ამერიკის წყნარი ოკეანის სტანდარტული დრო", "SAST": "სამხრეთ აფრიკის დრო", "SGT": "სინგაპურის დრო", "SRT": "სურინამის დრო", "TMST": "თურქმენეთის ზაფხულის დრო", "TMT": "თურქმენეთის სტანდარტული დრო", "UYST": "ურუგვაის ზაფხულის დრო", "UYT": "ურუგვაის სტანდარტული დრო", "VET": "ვენესუელის დრო", "WARST": "დასავლეთ არგენტინის ზაფხულის დრო", "WART": "დასავლეთ არგენტინის სტანდარტული დრო", "WAST": "დასავლეთ აფრიკის ზაფხულის დრო", "WAT": "დასავლეთ აფრიკის სტანდარტული დრო", "WESZ": "დასავლეთ ევროპის ზაფხულის დრო", "WEZ": "დასავლეთ ევროპის სტანდარტული დრო", "WIB": "დასავლეთ ინდონეზიის დრო", "WIT": "აღმოსავლეთ ინდონეზიის დრო", "WITA": "ცენტრალური ინდონეზიის დრო", "∅∅∅": "აზორის კუნძულების ზაფხულის დრო"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_4edbf4e11f17243cbd3f53cf42390fdf,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_8f50deadc7a225aae72abf7135535a4e,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_d9d734a5db809617a5116023efed1940,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e975bcc2318a5d1ed3b85b6356649a29,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kab"] = func() locales.Translator {
		return &localen{

			locale:             "kab",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$AR", "ATS", "$AU", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "FB", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "$BM", "$BN", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "$BS", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "$BZ", "$CA", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "$CL", "CNH", "CNX", "CNY", "$CO", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "£CY", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DA", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "$FJ", "£FK", "F", "£GB", "GEK", "GEL", "GHC", "GHS", "£GI", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "£IE", "£IL", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "₤IT", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "£LB", "LKR", "LRD", "lLS", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "fMA", "MCF", "MDC", "MDL", "MGA", "Fmg", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "£MT", "MUR", "MVP", "MVR", "MWK", "$MX", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "$NA", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$NZ", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "$RH", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "$SB", "SCR", "SDD", "SDG", "SDP", "SEK", "$SG", "SHP", "SIT", "SKK", "SLL", "SOS", "$SR", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "$TT", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$US", "USN", "USS", "UYI", "UYP", "$UY", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WS$", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "DTS", "XEU", "XFO", "XFU", "CFA", "XPD", "FCFP", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Yen", "Fur", "Meɣ", "Yeb", "May", "Yun", "Yul", "Ɣuc", "Cte", "Tub", "Nun", "Duǧ"},
			monthsNarrow:       []string{"", "Y", "F", "Ɣ", "B", "M", "N", "L", "C", "T", "R", "W", "D"},
			monthsWide:         []string{"", "Yennayer", "Fuṛar", "Meɣres", "Yebrir", "Mayyu", "Yunyu", "Yulyu", "Ɣuct", "Ctembeṛ", "Tubeṛ", "Nunembeṛ", "Duǧembeṛ"},
			daysAbbreviated:    []string{"Yan", "San", "Kraḍ", "Kuẓ", "Sam", "Sḍis", "Say"},
			daysNarrow:         []string{"C", "R", "A", "H", "M", "S", "D"},
			daysShort:          []string{"Cr", "Ri", "Ra", "Hd", "Mh", "Sm", "Sd"},
			daysWide:           []string{"Yanass", "Sanass", "Kraḍass", "Kuẓass", "Samass", "Sḍisass", "Sayass"},
			periodsAbbreviated: []string{"n tufat", "n tmeddit"},
			periodsNarrow:      []string{"f", "m"},
			periodsWide:        []string{"n tufat", "n tmeddit"},
			erasAbbreviated:    []string{"snd. T.Ɛ", "sld. T.Ɛ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"send talalit n Ɛisa", "seld talalit n Ɛisa"},
			timezones:          map[string]string{"ACDT": "Akud n Unebdu n Ustralya Talemmast", "ACST": "Akud Amagnu n Ustralya Talemmast", "ACWDT": "Akud n Unebdu n Tlemmast n Umalu n Ustṛalya", "ACWST": "Akud Amagnu n Tlemmast n Umalu n Ustṛalya", "ADT": "Akud Aṭlasan n Unebdu", "AEDT": "Akud n Unebdu n Ustṛalya n Usammar", "AEST": "Akud Amagnu n Ustṛalya n Usammar", "AKDT": "Akud n Unebdu n Alaska", "AKST": "Akud Amagnu n Alaska", "ARST": "Akud n Unebdu n Arjuntin", "ART": "Akud Amagnu n Arjuntin", "AST": "Akud Amagnu Aṭlasan", "AWDT": "Akud n Unebdu Ustṛalya n Umalu", "AWST": "Akud Amagnu n Ustṛalya n Umalu", "BOT": "Akud n Bulivi", "BT": "Akud n Butan", "CAT": "Akud n tefriqt talemmast", "CDT": "Akud n Unebdu n Tlemmast n Marikan", "CHADT": "Akud n Unebdu Catham", "CHAST": "Akud Amagnu n Catham", "CLST": "Akud n Unebdu n Cili", "CLT": "Akud Amagnu n Cili", "COST": "Akud n Unebdu n Kulumbya", "COT": "Akud Amagnu n Kulumbya", "CST": "Akud Amagnu n Tlemmast n Marikan", "ChST": "Akud Amagnu n Camuṛṛu", "EAT": "Akud n tefriqt n usammar", "ECT": "Akud n Ikwaṭur", "EDT": "Akud n Unebdu n Usammar Agafa n Marikan", "EST": "Akud Amagnu n Usammar Agafa n Marikan", "GFT": "Akud n Gwiyan Tafṛansist", "GMT": "Akud alemmas n Greenwich", "GST": "Akud Amagnu n Gulf", "GYT": "Akud n Gwiyan", "HADT": "Akud n Unebu n Haway-Aliwsyan", "HAST": "Akud Amagnu n Haway-Aliwsyan", "HAT": "Akud n Unebdu n Wakal Amaynut", "HECU": "Akud n Unebdu n Kuba", "HEEG": "Akud n Unebdu n Grinland n Usammar", "HENOMX": "Akud n Unebdu n Ugafa Amalu n Miksik", "HEOG": "Akud n Unebdu n Grinland n Umalu", "HEPM": "Akud n Unebdu n San Pyir & Miklun", "HEPMX": "Akud Amelwi n Unebdu n Miksik", "HKST": "Akud n Unebdu n Hung Kung", "HKT": "Akud Amagnu n Hung Kung", "HNCU": "Akud Amagnu n Kuba", "HNEG": "Akud Amagnu n Grinland n Usammar", "HNNOMX": "Akud Amagnu n Ugafa Amalu n Miksik", "HNOG": "Akud Amagnu n Grinland n Umalu", "HNPM": "Akud Amagnu n San Pyir & Miklun", "HNPMX": "Akud amagnu Amelwi n Miksik", "HNT": "Akud Amagnu n Wakal Amaynut", "IST": "Akud Amagnu n Hend", "JDT": "Akud n Unebdu n Japun", "JST": "Akud Amagnu n Japun", "LHDT": "Akud n Unebdu n Lord Howe", "LHST": "Akud Amagnu n Lord Howe", "MDT": "Akud n Unebdu n Idurar n Marikan", "MESZ": "Akud n unebdu n Turuft Talemmast", "MEZ": "Akud amagnu n Turuft Talemmast", "MST": "Akud Amagnu n Idurar n Marikan", "MYT": "Akud n Malizya", "NZDT": "Akud n Unebdu Ziland Tamaynut", "NZST": "Akud Amagnu n Ziland Tamaynut", "OESZ": "Akud n unebdu n Turuft n Usammar", "OEZ": "Akud amagnu n Turuft n Usammar", "PDT": "Akud Amelwi n Unebdu n Marikan n Ugafa", "PST": "Akud Amelwi Amagnu n Marikan n Ugafa", "SAST": "Akud amagnu n tefriqt n unẓul", "SGT": "Akud Amagnu n Sangapur", "SRT": "Akud n Surinam", "TMST": "Akud n Unebdu n Ṭurkmanistan", "TMT": "Akud Amagnu n Ṭurkmanistan", "UYST": "Akud n Unebdu n Urugway", "UYT": "Akud amagnu n Urugway", "VET": "Akud n Vinizwila", "WARST": "Akud n Unebdu n Arjuntin n Usammar", "WART": "Akud Amagnu n Arjuntin n Usammar", "WAST": "Akud n unebdu n tefriqt n umalu", "WAT": "Akud amagnu n tefriqt n umalu", "WESZ": "Akud n unebdu turuft n umalu", "WEZ": "Akud amagnu n turuft n umalu", "WIB": "Akud n Umalu n Indunisya", "WIT": "Akud n Usammar n Indunisya", "WITA": "Akud n Tlemmast n Indunisya", "∅∅∅": "Akud n Unebdu n Bṛazilya"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kab_dz"] = func() locales.Translator {
		return &localen{

			locale:             "kab_DZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Yen", "Fur", "Meɣ", "Yeb", "May", "Yun", "Yul", "Ɣuc", "Cte", "Tub", "Nun", "Duǧ"},
			monthsNarrow:       []string{"", "Y", "F", "Ɣ", "B", "M", "N", "L", "C", "T", "R", "W", "D"},
			monthsWide:         []string{"", "Yennayer", "Fuṛar", "Meɣres", "Yebrir", "Mayyu", "Yunyu", "Yulyu", "Ɣuct", "Ctembeṛ", "Tubeṛ", "Nunembeṛ", "Duǧembeṛ"},
			daysAbbreviated:    []string{"Yan", "San", "Kraḍ", "Kuẓ", "Sam", "Sḍis", "Say"},
			daysNarrow:         []string{"C", "R", "A", "H", "M", "S", "D"},
			daysShort:          []string{"Cr", "Ri", "Ra", "Hd", "Mh", "Sm", "Sd"},
			daysWide:           []string{"Yanass", "Sanass", "Kraḍass", "Kuẓass", "Samass", "Sḍisass", "Sayass"},
			periodsAbbreviated: []string{"n tufat", "n tmeddit"},
			periodsNarrow:      []string{"f", "m"},
			periodsWide:        []string{"n tufat", "n tmeddit"},
			erasAbbreviated:    []string{"snd. T.Ɛ", "sld. T.Ɛ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"send talalit n Ɛisa", "seld talalit n Ɛisa"},
			timezones:          map[string]string{"ACDT": "Akud n Unebdu n Ustralya Talemmast", "ACST": "Akud Amagnu n Ustralya Talemmast", "ACWDT": "Akud n Unebdu n Tlemmast n Umalu n Ustṛalya", "ACWST": "Akud Amagnu n Tlemmast n Umalu n Ustṛalya", "ADT": "Akud Aṭlasan n Unebdu", "AEDT": "Akud n Unebdu n Ustṛalya n Usammar", "AEST": "Akud Amagnu n Ustṛalya n Usammar", "AKDT": "Akud n Unebdu n Alaska", "AKST": "Akud Amagnu n Alaska", "ARST": "Akud n Unebdu n Arjuntin", "ART": "Akud Amagnu n Arjuntin", "AST": "Akud Amagnu Aṭlasan", "AWDT": "Akud n Unebdu Ustṛalya n Umalu", "AWST": "Akud Amagnu n Ustṛalya n Umalu", "BOT": "Akud n Bulivi", "BT": "Akud n Butan", "CAT": "Akud n tefriqt talemmast", "CDT": "Akud n Unebdu n Tlemmast n Marikan", "CHADT": "Akud n Unebdu Catham", "CHAST": "Akud Amagnu n Catham", "CLST": "Akud n Unebdu n Cili", "CLT": "Akud Amagnu n Cili", "COST": "Akud n Unebdu n Kulumbya", "COT": "Akud Amagnu n Kulumbya", "CST": "Akud Amagnu n Tlemmast n Marikan", "ChST": "Akud Amagnu n Camuṛṛu", "EAT": "Akud n tefriqt n usammar", "ECT": "Akud n Ikwaṭur", "EDT": "Akud n Unebdu n Usammar Agafa n Marikan", "EST": "Akud Amagnu n Usammar Agafa n Marikan", "GFT": "Akud n Gwiyan Tafṛansist", "GMT": "Akud alemmas n Greenwich", "GST": "Akud Amagnu n Gulf", "GYT": "Akud n Gwiyan", "HADT": "Akud n Unebu n Haway-Aliwsyan", "HAST": "Akud Amagnu n Haway-Aliwsyan", "HAT": "Akud n Unebdu n Wakal Amaynut", "HECU": "Akud n Unebdu n Kuba", "HEEG": "Akud n Unebdu n Grinland n Usammar", "HENOMX": "Akud n Unebdu n Ugafa Amalu n Miksik", "HEOG": "Akud n Unebdu n Grinland n Umalu", "HEPM": "Akud n Unebdu n San Pyir & Miklun", "HEPMX": "Akud Amelwi n Unebdu n Miksik", "HKST": "Akud n Unebdu n Hung Kung", "HKT": "Akud Amagnu n Hung Kung", "HNCU": "Akud Amagnu n Kuba", "HNEG": "Akud Amagnu n Grinland n Usammar", "HNNOMX": "Akud Amagnu n Ugafa Amalu n Miksik", "HNOG": "Akud Amagnu n Grinland n Umalu", "HNPM": "Akud Amagnu n San Pyir & Miklun", "HNPMX": "Akud amagnu Amelwi n Miksik", "HNT": "Akud Amagnu n Wakal Amaynut", "IST": "Akud Amagnu n Hend", "JDT": "Akud n Unebdu n Japun", "JST": "Akud Amagnu n Japun", "LHDT": "Akud n Unebdu n Lord Howe", "LHST": "Akud Amagnu n Lord Howe", "MDT": "Akud n Unebdu n Idurar n Marikan", "MESZ": "Akud n unebdu n Turuft Talemmast", "MEZ": "Akud amagnu n Turuft Talemmast", "MST": "Akud Amagnu n Idurar n Marikan", "MYT": "Akud n Malizya", "NZDT": "Akud n Unebdu Ziland Tamaynut", "NZST": "Akud Amagnu n Ziland Tamaynut", "OESZ": "Akud n unebdu n Turuft n Usammar", "OEZ": "Akud amagnu n Turuft n Usammar", "PDT": "Akud Amelwi n Unebdu n Marikan n Ugafa", "PST": "Akud Amelwi Amagnu n Marikan n Ugafa", "SAST": "Akud amagnu n tefriqt n unẓul", "SGT": "Akud Amagnu n Sangapur", "SRT": "Akud n Surinam", "TMST": "Akud n Unebdu n Ṭurkmanistan", "TMT": "Akud Amagnu n Ṭurkmanistan", "UYST": "Akud n Unebdu n Urugway", "UYT": "Akud amagnu n Urugway", "VET": "Akud n Vinizwila", "WARST": "Akud n Unebdu n Arjuntin n Usammar", "WART": "Akud Amagnu n Arjuntin n Usammar", "WAST": "Akud n unebdu n tefriqt n umalu", "WAT": "Akud amagnu n tefriqt n umalu", "WESZ": "Akud n unebdu turuft n umalu", "WEZ": "Akud amagnu n turuft n umalu", "WIB": "Akud n Umalu n Indunisya", "WIT": "Akud n Usammar n Indunisya", "WITA": "Akud n Tlemmast n Indunisya", "∅∅∅": "Akud n Unebdu n Bṛazilya"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1412ead9823f992d32fe783f7305e41f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kam"] = func() locales.Translator {
		return &localen{

			locale:                 "kam",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Mbe", "Kel", "Ktũ", "Kan", "Ktn", "Tha", "Moo", "Nya", "Knd", "Ĩku", "Ĩkm", "Ĩkl"},
			monthsNarrow:           []string{"", "M", "K", "K", "K", "K", "T", "M", "N", "K", "Ĩ", "Ĩ", "Ĩ"},
			monthsWide:             []string{"", "Mwai wa mbee", "Mwai wa kelĩ", "Mwai wa katatũ", "Mwai wa kana", "Mwai wa katano", "Mwai wa thanthatũ", "Mwai wa muonza", "Mwai wa nyaanya", "Mwai wa kenda", "Mwai wa ĩkumi", "Mwai wa ĩkumi na ĩmwe", "Mwai wa ĩkumi na ilĩ"},
			daysAbbreviated:        []string{"Wky", "Wkw", "Wkl", "Wtũ", "Wkn", "Wtn", "Wth"},
			daysNarrow:             []string{"Y", "W", "E", "A", "A", "A", "A"},
			daysWide:               []string{"Wa kyumwa", "Wa kwambĩlĩlya", "Wa kelĩ", "Wa katatũ", "Wa kana", "Wa katano", "Wa thanthatũ"},
			periodsAbbreviated:     []string{"Ĩyakwakya", "Ĩyawĩoo"},
			periodsWide:            []string{"Ĩyakwakya", "Ĩyawĩoo"},
			erasAbbreviated:        []string{"MY", "IY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbee wa Yesũ", "Ĩtina wa Yesũ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kam_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "kam_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Mbe", "Kel", "Ktũ", "Kan", "Ktn", "Tha", "Moo", "Nya", "Knd", "Ĩku", "Ĩkm", "Ĩkl"},
			monthsNarrow:           []string{"", "M", "K", "K", "K", "K", "T", "M", "N", "K", "Ĩ", "Ĩ", "Ĩ"},
			monthsWide:             []string{"", "Mwai wa mbee", "Mwai wa kelĩ", "Mwai wa katatũ", "Mwai wa kana", "Mwai wa katano", "Mwai wa thanthatũ", "Mwai wa muonza", "Mwai wa nyaanya", "Mwai wa kenda", "Mwai wa ĩkumi", "Mwai wa ĩkumi na ĩmwe", "Mwai wa ĩkumi na ilĩ"},
			daysAbbreviated:        []string{"Wky", "Wkw", "Wkl", "Wtũ", "Wkn", "Wtn", "Wth"},
			daysNarrow:             []string{"Y", "W", "E", "A", "A", "A", "A"},
			daysWide:               []string{"Wa kyumwa", "Wa kwambĩlĩlya", "Wa kelĩ", "Wa katatũ", "Wa kana", "Wa katano", "Wa thanthatũ"},
			periodsAbbreviated:     []string{"Ĩyakwakya", "Ĩyawĩoo"},
			periodsWide:            []string{"Ĩyakwakya", "Ĩyawĩoo"},
			erasAbbreviated:        []string{"MY", "IY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbee wa Yesũ", "Ĩtina wa Yesũ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kde"] = func() locales.Translator {
		return &localen{

			locale:                 "kde",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Mwedi Ntandi", "Mwedi wa Pili", "Mwedi wa Tatu", "Mwedi wa Nchechi", "Mwedi wa Nnyano", "Mwedi wa Nnyano na Umo", "Mwedi wa Nnyano na Mivili", "Mwedi wa Nnyano na Mitatu", "Mwedi wa Nnyano na Nchechi", "Mwedi wa Nnyano na Nnyano", "Mwedi wa Nnyano na Nnyano na U", "Mwedi wa Nnyano na Nnyano na M"},
			daysAbbreviated:        []string{"Ll2", "Ll3", "Ll4", "Ll5", "Ll6", "Ll7", "Ll1"},
			daysNarrow:             []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:               []string{"Liduva lyapili", "Liduva lyatatu", "Liduva lyanchechi", "Liduva lyannyano", "Liduva lyannyano na linji", "Liduva lyannyano na mavili", "Liduva litandi"},
			periodsAbbreviated:     []string{"Muhi", "Chilo"},
			periodsWide:            []string{"Muhi", "Chilo"},
			erasAbbreviated:        []string{"AY", "NY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Akanapawa Yesu", "Nankuida Yesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kde_tz"] = func() locales.Translator {
		return &localen{

			locale:                 "kde_TZ",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Mwedi Ntandi", "Mwedi wa Pili", "Mwedi wa Tatu", "Mwedi wa Nchechi", "Mwedi wa Nnyano", "Mwedi wa Nnyano na Umo", "Mwedi wa Nnyano na Mivili", "Mwedi wa Nnyano na Mitatu", "Mwedi wa Nnyano na Nchechi", "Mwedi wa Nnyano na Nnyano", "Mwedi wa Nnyano na Nnyano na U", "Mwedi wa Nnyano na Nnyano na M"},
			daysAbbreviated:        []string{"Ll2", "Ll3", "Ll4", "Ll5", "Ll6", "Ll7", "Ll1"},
			daysNarrow:             []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:               []string{"Liduva lyapili", "Liduva lyatatu", "Liduva lyanchechi", "Liduva lyannyano", "Liduva lyannyano na linji", "Liduva lyannyano na mavili", "Liduva litandi"},
			periodsAbbreviated:     []string{"Muhi", "Chilo"},
			periodsWide:            []string{"Muhi", "Chilo"},
			erasAbbreviated:        []string{"AY", "NY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Akanapawa Yesu", "Nankuida Yesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kea"] = func() locales.Translator {
		return &localen{

			locale:                 "kea",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "৳", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "$", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "\u200b", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "$", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "៛", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "₸", "₭", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "₮", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "$", "SCR", "SDD", "SDG", "SDP", "SEK", "$", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "₺", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "Mai", "Jun", "Jul", "Ago", "Set", "Otu", "Nuv", "Diz"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janeru", "Febreru", "Marsu", "Abril", "Maiu", "Junhu", "Julhu", "Agostu", "Setenbru", "Otubru", "Nuvenbru", "Dizenbru"},
			daysAbbreviated:        []string{"dum", "sig", "ter", "kua", "kin", "ses", "sab"},
			daysNarrow:             []string{"D", "S", "T", "K", "K", "S", "S"},
			daysShort:              []string{"du", "si", "te", "ku", "ki", "se", "sa"},
			daysWide:               []string{"dumingu", "sigunda-fera", "tersa-fera", "kuarta-fera", "kinta-fera", "sesta-fera", "sabadu"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"AK", "DK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antis di Kristu", "dispos di Kristu"},
			timezones:              map[string]string{"ACDT": "Ora di Veron di Austrália Sentral", "ACST": "Ora Padron di Austrália Sentral", "ACWDT": "Ora di Veron di Austrália Sentru-Osidental", "ACWST": "Ora Padron di Austrália Sentru-Osidental", "ADT": "Ora di Veron di Atlántiku", "AEDT": "Ora di Veron di Austrália Oriental", "AEST": "Ora Padron di Austrália Oriental", "AKDT": "Ora di Veron di Alaska", "AKST": "Ora Padron di Alaska", "ARST": "ARST", "ART": "ART", "AST": "Ora Padron di Atlántiku", "AWDT": "Ora di Veron di Austrália Osidental", "AWST": "Ora Padron di Austrália Osidental", "BOT": "BOT", "BT": "BT", "CAT": "Ora di Áfrika Sentral", "CDT": "Ora di Veron Sentral", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Ora Padron Sentral", "ChST": "ChST", "EAT": "Ora di Áfrika Oriental", "ECT": "ECT", "EDT": "Ora di Veron Oriental", "EST": "Ora Padron Oriental", "GFT": "GFT", "GMT": "Ora Médiu di Greenwich", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "Ora di Veron di Kuba", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "Ora Padron di Kuba", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Ora di Veron di Montanha", "MESZ": "Ora di Veron di Europa Sentral", "MEZ": "Ora Padron di Europa Sentral", "MST": "Ora Padron di Montanha", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Ora di Veron di Europa Oriental", "OEZ": "Ora Padron di Europa Oriental", "PDT": "Ora di Veron di Pasífiku", "PST": "Ora Padron di Pasífiku", "SAST": "Ora di Sul di Áfrika", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Ora di Veron di Áfrika Osidental", "WAT": "Ora Padron di Áfrika Osidental", "WESZ": "Ora di Veron di Europa Osidental", "WEZ": "Ora Padron di Europa Osidental", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_6df4b160cd1730b200782b7d4e4cd069,
			fnFmtDateLong:         fnFmtDateLong_56933fd536cdccf5e7a0f57849436ee6,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kea_cv"] = func() locales.Translator {
		return &localen{

			locale:                 "kea_CV",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "Mai", "Jun", "Jul", "Ago", "Set", "Otu", "Nuv", "Diz"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janeru", "Febreru", "Marsu", "Abril", "Maiu", "Junhu", "Julhu", "Agostu", "Setenbru", "Otubru", "Nuvenbru", "Dizenbru"},
			daysAbbreviated:        []string{"dum", "sig", "ter", "kua", "kin", "ses", "sab"},
			daysNarrow:             []string{"D", "S", "T", "K", "K", "S", "S"},
			daysShort:              []string{"du", "si", "te", "ku", "ki", "se", "sa"},
			daysWide:               []string{"dumingu", "sigunda-fera", "tersa-fera", "kuarta-fera", "kinta-fera", "sesta-fera", "sabadu"},
			periodsAbbreviated:     []string{"am", "pm"},
			periodsWide:            []string{"am", "pm"},
			erasAbbreviated:        []string{"AK", "DK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antis di Kristu", "dispos di Kristu"},
			timezones:              map[string]string{"ACDT": "Ora di Veron di Austrália Sentral", "ACST": "Ora Padron di Austrália Sentral", "ACWDT": "Ora di Veron di Austrália Sentru-Osidental", "ACWST": "Ora Padron di Austrália Sentru-Osidental", "ADT": "Ora di Veron di Atlántiku", "AEDT": "Ora di Veron di Austrália Oriental", "AEST": "Ora Padron di Austrália Oriental", "AKDT": "Ora di Veron di Alaska", "AKST": "Ora Padron di Alaska", "ARST": "ARST", "ART": "ART", "AST": "Ora Padron di Atlántiku", "AWDT": "Ora di Veron di Austrália Osidental", "AWST": "Ora Padron di Austrália Osidental", "BOT": "BOT", "BT": "BT", "CAT": "Ora di Áfrika Sentral", "CDT": "Ora di Veron Sentral", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Ora Padron Sentral", "ChST": "ChST", "EAT": "Ora di Áfrika Oriental", "ECT": "ECT", "EDT": "Ora di Veron Oriental", "EST": "Ora Padron Oriental", "GFT": "GFT", "GMT": "Ora Médiu di Greenwich", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "Ora di Veron di Kuba", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "Ora Padron di Kuba", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Ora di Veron di Montanha", "MESZ": "Ora di Veron di Europa Sentral", "MEZ": "Ora Padron di Europa Sentral", "MST": "Ora Padron di Montanha", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Ora di Veron di Europa Oriental", "OEZ": "Ora Padron di Europa Oriental", "PDT": "Ora di Veron di Pasífiku", "PST": "Ora Padron di Pasífiku", "SAST": "Ora di Sul di Áfrika", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Ora di Veron di Áfrika Osidental", "WAT": "Ora Padron di Áfrika Osidental", "WESZ": "Ora di Veron di Europa Osidental", "WEZ": "Ora Padron di Europa Osidental", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_6df4b160cd1730b200782b7d4e4cd069,
			fnFmtDateLong:         fnFmtDateLong_56933fd536cdccf5e7a0f57849436ee6,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["khq"] = func() locales.Translator {
		return &localen{

			locale:             "khq",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alj", "Ass"},
			daysNarrow:         []string{"H", "T", "T", "L", "L", "L", "S"},
			daysWide:           []string{"Alhadi", "Atini", "Atalata", "Alarba", "Alhamiisa", "Aljuma", "Assabdu"},
			periodsAbbreviated: []string{"Adduha", "Aluula"},
			periodsWide:        []string{"Adduha", "Aluula"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa jamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_dd5c984fc6201988feed2698382f9150,
			fnFmtCurrency:         fnFmtCurrency_f27eb78b129904a4d98d49f647da3164,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["khq_ml"] = func() locales.Translator {
		return &localen{

			locale:             "khq_ML",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alj", "Ass"},
			daysNarrow:         []string{"H", "T", "T", "L", "L", "L", "S"},
			daysWide:           []string{"Alhadi", "Atini", "Atalata", "Alarba", "Alhamiisa", "Aljuma", "Assabdu"},
			periodsAbbreviated: []string{"Adduha", "Aluula"},
			periodsWide:        []string{"Adduha", "Aluula"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa jamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_dd5c984fc6201988feed2698382f9150,
			fnFmtCurrency:         fnFmtCurrency_f27eb78b129904a4d98d49f647da3164,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ki"] = func() locales.Translator {
		return &localen{

			locale:                 "ki",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "JEN", "WKR", "WGT", "WKN", "WTN", "WTD", "WMJ", "WNN", "WKD", "WIK", "WMW", "DIT"},
			monthsNarrow:           []string{"", "J", "K", "G", "K", "G", "G", "M", "K", "K", "I", "I", "D"},
			monthsWide:             []string{"", "Njenuarĩ", "Mwere wa kerĩ", "Mwere wa gatatũ", "Mwere wa kana", "Mwere wa gatano", "Mwere wa gatandatũ", "Mwere wa mũgwanja", "Mwere wa kanana", "Mwere wa kenda", "Mwere wa ikũmi", "Mwere wa ikũmi na ũmwe", "Ndithemba"},
			daysAbbreviated:        []string{"KMA", "NTT", "NMN", "NMT", "ART", "NMA", "NMM"},
			daysNarrow:             []string{"K", "N", "N", "N", "A", "N", "N"},
			daysWide:               []string{"Kiumia", "Njumatatũ", "Njumaine", "Njumatana", "Aramithi", "Njumaa", "Njumamothi"},
			periodsAbbreviated:     []string{"Kiroko", "Hwaĩ-inĩ"},
			periodsWide:            []string{"Kiroko", "Hwaĩ-inĩ"},
			erasAbbreviated:        []string{"MK", "TK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbere ya Kristo", "Thutha wa Kristo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ki_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "ki_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "JEN", "WKR", "WGT", "WKN", "WTN", "WTD", "WMJ", "WNN", "WKD", "WIK", "WMW", "DIT"},
			monthsNarrow:           []string{"", "J", "K", "G", "K", "G", "G", "M", "K", "K", "I", "I", "D"},
			monthsWide:             []string{"", "Njenuarĩ", "Mwere wa kerĩ", "Mwere wa gatatũ", "Mwere wa kana", "Mwere wa gatano", "Mwere wa gatandatũ", "Mwere wa mũgwanja", "Mwere wa kanana", "Mwere wa kenda", "Mwere wa ikũmi", "Mwere wa ikũmi na ũmwe", "Ndithemba"},
			daysAbbreviated:        []string{"KMA", "NTT", "NMN", "NMT", "ART", "NMA", "NMM"},
			daysNarrow:             []string{"K", "N", "N", "N", "A", "N", "N"},
			daysWide:               []string{"Kiumia", "Njumatatũ", "Njumaine", "Njumatana", "Aramithi", "Njumaa", "Njumamothi"},
			periodsAbbreviated:     []string{"Kiroko", "Hwaĩ-inĩ"},
			periodsWide:            []string{"Kiroko", "Hwaĩ-inĩ"},
			erasAbbreviated:        []string{"MK", "TK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbere ya Kristo", "Thutha wa Kristo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kk"] = func() locales.Translator {
		return &localen{

			locale:                 "kk",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "₸", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "қаң.", "ақп.", "нау.", "сәу.", "мам.", "мау.", "шіл.", "там.", "қыр.", "қаз.", "қар.", "жел."},
			monthsNarrow:           []string{"", "Қ", "А", "Н", "С", "М", "М", "Ш", "Т", "Қ", "Қ", "Қ", "Ж"},
			monthsWide:             []string{"", "қаңтар", "ақпан", "наурыз", "сәуір", "мамыр", "маусым", "шілде", "тамыз", "қыркүйек", "қазан", "қараша", "желтоқсан"},
			daysAbbreviated:        []string{"жс", "дс", "сс", "ср", "бс", "жм", "сб"},
			daysNarrow:             []string{"Ж", "Д", "С", "С", "Б", "Ж", "С"},
			daysShort:              []string{"жс", "дс", "сс", "ср", "бс", "жм", "сб"},
			daysWide:               []string{"жексенбі", "дүйсенбі", "сейсенбі", "сәрсенбі", "бейсенбі", "жұма", "сенбі"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"б.з.д.", "б.з."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Біздің заманымызға дейін", "біздің заманымыз"},
			timezones:              map[string]string{"ACDT": "Австралия жазғы орталық уақыты", "ACST": "Австралия стандартты орталық уақыты", "ACWDT": "Австралия жазғы орталық-батыс уақыты", "ACWST": "Австралия стандартты орталық-батыс уақыты", "ADT": "Атлантика жазғы уақыты", "AEDT": "Австралия жазғы шығыс уақыты", "AEST": "Австралия стандартты шығыс уақыты", "AKDT": "Аляска жазғы уақыты", "AKST": "Аляска стандартты уақыты", "ARST": "Аргентина жазғы уақыты", "ART": "Аргентина стандартты уақыты", "AST": "Атлантика стандартты уақыты", "AWDT": "Австралия жазғы батыс уақыты", "AWST": "Австралия стандартты батыс уақыты", "BOT": "Боливия уақыты", "BT": "Бутан уақыты", "CAT": "Орталық Африка уақыты", "CDT": "Солтүстік Америка жазғы орталық уақыты", "CHADT": "Чатем жазғы уақыты", "CHAST": "Чатем стандартты уақыты", "CLST": "Чили жазғы уақыты", "CLT": "Чили стандартты уақыты", "COST": "Колумбия жазғы уақыты", "COT": "Колумбия стандартты уақыты", "CST": "Солтүстік Америка стандартты орталық уақыты", "ChST": "Чаморро стандартты уақыты", "EAT": "Шығыс Африка уақыты", "ECT": "Эквадор уақыты", "EDT": "Солтүстік Америка жазғы шығыс уақыты", "EST": "Солтүстік Америка стандартты шығыс уақыты", "GFT": "Француз Гвианасы уақыты", "GMT": "Гринвич уақыты", "GST": "Парсы шығанағы уақыты", "GYT": "Гайана уақыты", "HADT": "Гавай және Алеут аралдары жазғы уақыты", "HAST": "Гавай және Алеут аралдары стандартты уақыты", "HAT": "Ньюфаундленд жазғы уақыты", "HECU": "Куба жазғы уақыты", "HEEG": "Шығыс Гренландия жазғы уақыты", "HENOMX": "Солтүстік-батыс Мексика жазғы уақыты", "HEOG": "Батыс Гренландия жазғы уақыты", "HEPM": "Сен-Пьер және Микелон жазғы уақыты", "HEPMX": "Мексика жазғы Тынық мұхит уақыты", "HKST": "Гонконг жазғы уақыты", "HKT": "Гонконг стандартты уақыты", "HNCU": "Куба стандартты уақыты", "HNEG": "Шығыс Гренландия стандартты уақыты", "HNNOMX": "Солтүстік-батыс Мексика стандартты уақыты", "HNOG": "Батыс Гренландия стандартты уақыты", "HNPM": "Сен-Пьер және Микелон стандартты уақыты", "HNPMX": "Мексика стандартты Тынық мұхит уақыты", "HNT": "Ньюфаундленд стандартты уақыты", "IST": "Үндістан стандартты уақыты", "JDT": "Жапония жазғы уақыты", "JST": "Жапония стандартты уақыты", "LHDT": "Лорд-Хау жазғы уақыты", "LHST": "Лорд-Хау стандартты уақыты", "MDT": "Солтүстік Америка жазғы тау уақыты", "MESZ": "Орталық Еуропа жазғы уақыты", "MEZ": "Орталық Еуропа стандартты уақыты", "MST": "Солтүстік Америка стандартты тау уақыты", "MYT": "Малайзия уақыты", "NZDT": "Жаңа Зеландия жазғы уақыты", "NZST": "Жаңа Зеландия стандартты уақыты", "OESZ": "Шығыс Еуропа жазғы уақыты", "OEZ": "Шығыс Еуропа стандартты уақыты", "PDT": "Солтүстік Америка жазғы Тынық мұхиты уақыты", "PST": "Солтүстік Америка стандартты Тынық мұхиты уақыты", "SAST": "Оңтүстік Африка уақыты", "SGT": "Сингапур стандартты уақыты", "SRT": "Суринам уақыты", "TMST": "Түрікменстан жазғы уақыты", "TMT": "Түрікменстан стандартты уақыты", "UYST": "Уругвай жазғы уақыты", "UYT": "Уругвай стандартты уақыты", "VET": "Венесуэла уақыты", "WARST": "Батыс Аргентина жазғы уақыты", "WART": "Батыс Аргентина стандартты уақыты", "WAST": "Батыс Африка жазғы уақыты", "WAT": "Батыс Африка стандартты уақыты", "WESZ": "Батыс Еуропа жазғы уақыты", "WEZ": "Батыс Еуропа стандартты уақыты", "WIB": "Батыс Индонезия уақыты", "WIT": "Шығыс Индонезия уақыты", "WITA": "Орталық Индонезия уақыты", "∅∅∅": "Бразилия жазғы уақыты"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_4d0973d19d04f6373346e3410c0d7cff,
			fnFmtDateLong:         fnFmtDateLong_c224d2682917532885a5e8179ac7c5c2,
			fnFmtDateMedium:       fnFmtDateMedium_615c694e243fa2fa6c571f5fc2c9d718,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b14713cd1b1f920531a88cd35823e9c6,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kk_kz"] = func() locales.Translator {
		return &localen{

			locale:                 "kk_KZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "қаң.", "ақп.", "нау.", "сәу.", "мам.", "мау.", "шіл.", "там.", "қыр.", "қаз.", "қар.", "жел."},
			monthsNarrow:           []string{"", "Қ", "А", "Н", "С", "М", "М", "Ш", "Т", "Қ", "Қ", "Қ", "Ж"},
			monthsWide:             []string{"", "қаңтар", "ақпан", "наурыз", "сәуір", "мамыр", "маусым", "шілде", "тамыз", "қыркүйек", "қазан", "қараша", "желтоқсан"},
			daysAbbreviated:        []string{"жс", "дс", "сс", "ср", "бс", "жм", "сб"},
			daysNarrow:             []string{"Ж", "Д", "С", "С", "Б", "Ж", "С"},
			daysShort:              []string{"жс", "дс", "сс", "ср", "бс", "жм", "сб"},
			daysWide:               []string{"жексенбі", "дүйсенбі", "сейсенбі", "сәрсенбі", "бейсенбі", "жұма", "сенбі"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"б.з.д.", "б.з."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Біздің заманымызға дейін", "біздің заманымыз"},
			timezones:              map[string]string{"ACDT": "Австралия жазғы орталық уақыты", "ACST": "Австралия стандартты орталық уақыты", "ACWDT": "Австралия жазғы орталық-батыс уақыты", "ACWST": "Австралия стандартты орталық-батыс уақыты", "ADT": "Атлантика жазғы уақыты", "AEDT": "Австралия жазғы шығыс уақыты", "AEST": "Австралия стандартты шығыс уақыты", "AKDT": "Аляска жазғы уақыты", "AKST": "Аляска стандартты уақыты", "ARST": "Аргентина жазғы уақыты", "ART": "Аргентина стандартты уақыты", "AST": "Атлантика стандартты уақыты", "AWDT": "Австралия жазғы батыс уақыты", "AWST": "Австралия стандартты батыс уақыты", "BOT": "Боливия уақыты", "BT": "Бутан уақыты", "CAT": "Орталық Африка уақыты", "CDT": "Солтүстік Америка жазғы орталық уақыты", "CHADT": "Чатем жазғы уақыты", "CHAST": "Чатем стандартты уақыты", "CLST": "Чили жазғы уақыты", "CLT": "Чили стандартты уақыты", "COST": "Колумбия жазғы уақыты", "COT": "Колумбия стандартты уақыты", "CST": "Солтүстік Америка стандартты орталық уақыты", "ChST": "Чаморро стандартты уақыты", "EAT": "Шығыс Африка уақыты", "ECT": "Эквадор уақыты", "EDT": "Солтүстік Америка жазғы шығыс уақыты", "EST": "Солтүстік Америка стандартты шығыс уақыты", "GFT": "Француз Гвианасы уақыты", "GMT": "Гринвич уақыты", "GST": "Парсы шығанағы уақыты", "GYT": "Гайана уақыты", "HADT": "Гавай және Алеут аралдары жазғы уақыты", "HAST": "Гавай және Алеут аралдары стандартты уақыты", "HAT": "Ньюфаундленд жазғы уақыты", "HECU": "Куба жазғы уақыты", "HEEG": "Шығыс Гренландия жазғы уақыты", "HENOMX": "Солтүстік-батыс Мексика жазғы уақыты", "HEOG": "Батыс Гренландия жазғы уақыты", "HEPM": "Сен-Пьер және Микелон жазғы уақыты", "HEPMX": "Мексика жазғы Тынық мұхит уақыты", "HKST": "Гонконг жазғы уақыты", "HKT": "Гонконг стандартты уақыты", "HNCU": "Куба стандартты уақыты", "HNEG": "Шығыс Гренландия стандартты уақыты", "HNNOMX": "Солтүстік-батыс Мексика стандартты уақыты", "HNOG": "Батыс Гренландия стандартты уақыты", "HNPM": "Сен-Пьер және Микелон стандартты уақыты", "HNPMX": "Мексика стандартты Тынық мұхит уақыты", "HNT": "Ньюфаундленд стандартты уақыты", "IST": "Үндістан стандартты уақыты", "JDT": "Жапония жазғы уақыты", "JST": "Жапония стандартты уақыты", "LHDT": "Лорд-Хау жазғы уақыты", "LHST": "Лорд-Хау стандартты уақыты", "MDT": "Солтүстік Америка жазғы тау уақыты", "MESZ": "Орталық Еуропа жазғы уақыты", "MEZ": "Орталық Еуропа стандартты уақыты", "MST": "Солтүстік Америка стандартты тау уақыты", "MYT": "Малайзия уақыты", "NZDT": "Жаңа Зеландия жазғы уақыты", "NZST": "Жаңа Зеландия стандартты уақыты", "OESZ": "Шығыс Еуропа жазғы уақыты", "OEZ": "Шығыс Еуропа стандартты уақыты", "PDT": "Солтүстік Америка жазғы Тынық мұхиты уақыты", "PST": "Солтүстік Америка стандартты Тынық мұхиты уақыты", "SAST": "Оңтүстік Африка уақыты", "SGT": "Сингапур стандартты уақыты", "SRT": "Суринам уақыты", "TMST": "Түрікменстан жазғы уақыты", "TMT": "Түрікменстан стандартты уақыты", "UYST": "Уругвай жазғы уақыты", "UYT": "Уругвай стандартты уақыты", "VET": "Венесуэла уақыты", "WARST": "Батыс Аргентина жазғы уақыты", "WART": "Батыс Аргентина стандартты уақыты", "WAST": "Батыс Африка жазғы уақыты", "WAT": "Батыс Африка стандартты уақыты", "WESZ": "Батыс Еуропа жазғы уақыты", "WEZ": "Батыс Еуропа стандартты уақыты", "WIB": "Батыс Индонезия уақыты", "WIT": "Шығыс Индонезия уақыты", "WITA": "Орталық Индонезия уақыты", "∅∅∅": "Перу жазғы уақыты"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_4d0973d19d04f6373346e3410c0d7cff,
			fnFmtDateLong:         fnFmtDateLong_c224d2682917532885a5e8179ac7c5c2,
			fnFmtDateMedium:       fnFmtDateMedium_615c694e243fa2fa6c571f5fc2c9d718,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_b14713cd1b1f920531a88cd35823e9c6,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kkj"] = func() locales.Translator {
		return &localen{

			locale:                 "kkj",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsWide:             []string{"", "pamba", "wanja", "mbiyɔ mɛndoŋgɔ", "Nyɔlɔmbɔŋgɔ", "Mɔnɔ ŋgbanja", "Nyaŋgwɛ ŋgbanja", "kuŋgwɛ", "fɛ", "njapi", "nyukul", "11", "ɓulɓusɛ"},
			daysAbbreviated:        []string{"sɔndi", "lundi", "mardi", "mɛrkɛrɛdi", "yedi", "vaŋdɛrɛdi", "mɔnɔ sɔndi"},
			daysNarrow:             []string{"so", "lu", "ma", "mɛ", "ye", "va", "ms"},
			daysShort:              []string{"sɔndi", "lundi", "mardi", "mɛrkɛrɛdi", "yedi", "vaŋdɛrɛdi", "mɔnɔ sɔndi"},
			daysWide:               []string{"sɔndi", "lundi", "mardi", "mɛrkɛrɛdi", "yedi", "vaŋdɛrɛdi", "mɔnɔ sɔndi"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06,
			fnFmtCurrency:         fnFmtCurrency_1c327a5182399c63873a891da808a068,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_d82f27051c9ad611e3ef0bbd19df221b,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kkj_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "kkj_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsWide:             []string{"", "pamba", "wanja", "mbiyɔ mɛndoŋgɔ", "Nyɔlɔmbɔŋgɔ", "Mɔnɔ ŋgbanja", "Nyaŋgwɛ ŋgbanja", "kuŋgwɛ", "fɛ", "njapi", "nyukul", "11", "ɓulɓusɛ"},
			daysAbbreviated:        []string{"sɔndi", "lundi", "mardi", "mɛrkɛrɛdi", "yedi", "vaŋdɛrɛdi", "mɔnɔ sɔndi"},
			daysNarrow:             []string{"so", "lu", "ma", "mɛ", "ye", "va", "ms"},
			daysShort:              []string{"sɔndi", "lundi", "mardi", "mɛrkɛrɛdi", "yedi", "vaŋdɛrɛdi", "mɔnɔ sɔndi"},
			daysWide:               []string{"sɔndi", "lundi", "mardi", "mɛrkɛrɛdi", "yedi", "vaŋdɛrɛdi", "mɔnɔ sɔndi"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06,
			fnFmtCurrency:         fnFmtCurrency_1c327a5182399c63873a891da808a068,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_d82f27051c9ad611e3ef0bbd19df221b,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kl"] = func() locales.Translator {
		return &localen{

			locale:             "kl",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			minus:              "−",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "kr.", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "Nkr", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "Skr", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "jan", "febr", "mar", "apr", "maj", "jun", "jul", "aug", "sept", "okt", "nov", "dec"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "januaarip", "februaarip", "marsip", "apriilip", "maajip", "juunip", "juulip", "aggustip", "septembarip", "oktobarip", "novembarip", "decembarip"},
			daysAbbreviated:    []string{"sap", "ata", "mar", "pin", "sis", "tal", "arf"},
			daysNarrow:         []string{"S", "A", "M", "P", "S", "T", "A"},
			daysShort:          []string{"sap", "ata", "mar", "pin", "sis", "tal", "arf"},
			daysWide:           []string{"sapaat", "ataasinngorneq", "marlunngorneq", "pingasunngorneq", "sisamanngorneq", "tallimanngorneq", "arfininngorneq"},
			periodsAbbreviated: []string{"u.t.", "u.k."},
			periodsWide:        []string{"ulloqeqqata-tungaa", "ulloqeqqata-kingorna"},
			erasAbbreviated:    []string{"Kr.in.si.", "Kr.in.king."},
			erasNarrow:         []string{"Kr.s.", "Kr.k."},
			erasWide:           []string{"Kristusip inunngornerata siornagut", "Kristusip inunngornerata kingornagut"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_2ebcd49fb93d5f5e0574b122053067c2,
			fnFmtCurrency:         fnFmtCurrency_311268549b3ecbf935e9311c16ee9a54,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kl_gl"] = func() locales.Translator {
		return &localen{

			locale:             "kl_GL",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			minus:              "−",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "jan", "febr", "mar", "apr", "maj", "jun", "jul", "aug", "sept", "okt", "nov", "dec"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "januaarip", "februaarip", "marsip", "apriilip", "maajip", "juunip", "juulip", "aggustip", "septembarip", "oktobarip", "novembarip", "decembarip"},
			daysAbbreviated:    []string{"sap", "ata", "mar", "pin", "sis", "tal", "arf"},
			daysNarrow:         []string{"S", "A", "M", "P", "S", "T", "A"},
			daysShort:          []string{"sap", "ata", "mar", "pin", "sis", "tal", "arf"},
			daysWide:           []string{"sapaat", "ataasinngorneq", "marlunngorneq", "pingasunngorneq", "sisamanngorneq", "tallimanngorneq", "arfininngorneq"},
			periodsAbbreviated: []string{"u.t.", "u.k."},
			periodsWide:        []string{"ulloqeqqata-tungaa", "ulloqeqqata-kingorna"},
			erasAbbreviated:    []string{"Kr.in.si.", "Kr.in.king."},
			erasNarrow:         []string{"Kr.s.", "Kr.k."},
			erasWide:           []string{"Kristusip inunngornerata siornagut", "Kristusip inunngornerata kingornagut"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_2ebcd49fb93d5f5e0574b122053067c2,
			fnFmtCurrency:         fnFmtCurrency_311268549b3ecbf935e9311c16ee9a54,
			fnFmtDateFull:         fnFmtDateFull_aa89072e939f6d411d6d93fa47eed117,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kln"] = func() locales.Translator {
		return &localen{

			locale:                 "kln",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Mul", "Ngat", "Taa", "Iwo", "Mam", "Paa", "Nge", "Roo", "Bur", "Epe", "Kpt", "Kpa"},
			monthsNarrow:           []string{"", "M", "N", "T", "I", "M", "P", "N", "R", "B", "E", "K", "K"},
			monthsWide:             []string{"", "Mulgul", "Ng’atyaato", "Kiptaamo", "Iwootkuut", "Mamuut", "Paagi", "Ng’eiyeet", "Rooptui", "Bureet", "Epeeso", "Kipsuunde ne taai", "Kipsuunde nebo aeng’"},
			daysAbbreviated:        []string{"Kts", "Kot", "Koo", "Kos", "Koa", "Kom", "Kol"},
			daysNarrow:             []string{"T", "T", "O", "S", "A", "M", "L"},
			daysWide:               []string{"Kotisap", "Kotaai", "Koaeng’", "Kosomok", "Koang’wan", "Komuut", "Kolo"},
			periodsAbbreviated:     []string{"krn", "koosk"},
			periodsWide:            []string{"karoon", "kooskoliny"},
			erasAbbreviated:        []string{"AM", "KO"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Amait kesich Jesu", "Kokakesich Jesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kln_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "kln_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Mul", "Ngat", "Taa", "Iwo", "Mam", "Paa", "Nge", "Roo", "Bur", "Epe", "Kpt", "Kpa"},
			monthsNarrow:           []string{"", "M", "N", "T", "I", "M", "P", "N", "R", "B", "E", "K", "K"},
			monthsWide:             []string{"", "Mulgul", "Ng’atyaato", "Kiptaamo", "Iwootkuut", "Mamuut", "Paagi", "Ng’eiyeet", "Rooptui", "Bureet", "Epeeso", "Kipsuunde ne taai", "Kipsuunde nebo aeng’"},
			daysAbbreviated:        []string{"Kts", "Kot", "Koo", "Kos", "Koa", "Kom", "Kol"},
			daysNarrow:             []string{"T", "T", "O", "S", "A", "M", "L"},
			daysWide:               []string{"Kotisap", "Kotaai", "Koaeng’", "Kosomok", "Koang’wan", "Komuut", "Kolo"},
			periodsAbbreviated:     []string{"krn", "koosk"},
			periodsWide:            []string{"karoon", "kooskoliny"},
			erasAbbreviated:        []string{"AM", "KO"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Amait kesich Jesu", "Kokakesich Jesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["km"] = func() locales.Translator {
		return &localen{

			locale:                 "km",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "៛", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "មករា", "កុម្ភៈ", "មីនា", "មេសា", "ឧសភា", "មិថុនា", "កក្កដា", "សីហា", "កញ្ញា", "តុលា", "វិច្ឆិកា", "ធ្នូ"},
			monthsNarrow:           []string{"", "ម", "ក", "ម", "ម", "ឧ", "ម", "ក", "ស", "ក", "ត", "វ", "ធ"},
			monthsWide:             []string{"", "មករា", "កុម្ភៈ", "មីនា", "មេសា", "ឧសភា", "មិថុនា", "កក្កដា", "សីហា", "កញ្ញា", "តុលា", "វិច្ឆិកា", "ធ្នូ"},
			daysAbbreviated:        []string{"អាទិត្យ", "ចន្ទ", "អង្គារ", "ពុធ", "ព្រហ", "សុក្រ", "សៅរ៍"},
			daysNarrow:             []string{"អ", "ច", "អ", "ព", "ព", "ស", "ស"},
			daysShort:              []string{"អា", "ច", "អ", "ពុ", "ព្រ", "សុ", "ស"},
			daysWide:               []string{"អាទិត្យ", "ច័ន្ទ", "អង្គារ", "ពុធ", "ព្រហស្បតិ៍", "សុក្រ", "សៅរ៍"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"មុន គ.ស.", "គ.ស."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"មុន\u200bគ្រិស្តសករាជ", "គ្រិស្តសករាជ"},
			timezones:              map[string]string{"ACDT": "ម៉ោង\u200bពេលថ្ងៃ\u200b\u200b\u200b\u200bនៅ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ACST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ACWDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200b\u200bភាគ\u200bខាង\u200bលិច\u200bនៃ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ACWST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bភាគ\u200bខាង\u200bលិច\u200bនៃ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ADT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bអាត្លង់ទិក", "AEDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bកើត", "AEST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bកើត", "AKDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200b\u200bអាឡាស្កា", "AKST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអាឡាស្កា", "ARST": "ម៉ោង\u200bនៅ\u200bអាហ្សង់ទីននារដូវក្តៅ", "ART": "ម៉ោងស្តង់ដារ\u200bនៅ\u200bអាហ្សង់ទីន", "AST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអាត្លង់ទិក", "AWDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bលិច", "AWST": "ម៉ោង\u200b\u200bស្តង់ដារ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bលិច", "BOT": "ម៉ោង\u200bនៅ\u200bបូលីវី", "BT": "ម៉ោងនៅប៊ូតង់", "CAT": "ម៉ោង\u200bនៅ\u200bអាហ្វ្រិក\u200bកណ្ដាល", "CDT": "ម៉ោង\u200b\u200bពេលថ្ងៃនៅ\u200bទ្វីបអាមេរិក\u200bខាង\u200bជើងភាគកណ្តាល", "CHADT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bចាថាំ", "CHAST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bចាថាំ", "CLST": "ម៉ោងនៅស៊ីលីនារដូវក្តៅ", "CLT": "ម៉ោងស្តង់ដារនៅស៊ីលី", "COST": "ម៉ោង\u200bនៅ\u200bកូឡុំប៊ីនា\u200bរដូវ\u200bក្ដៅ", "COT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bកូឡុំប៊ី", "CST": "ម៉ោង\u200b\u200bស្តង់ដារនៅ\u200bទ្វីបអាមេរិក\u200bខាង\u200bជើងភាគកណ្តាល", "ChST": "ម៉ោង\u200bស្តង់ដារនៅ\u200bចាំម៉ូរ៉ូ", "EAT": "ម៉ោង\u200bនៅ\u200bអាហ្វ្រិក\u200bខាង\u200bកើត", "ECT": "ម៉ោង\u200bនៅ\u200bអេក្វាទ័រ", "EDT": "ម៉ោងពេលថ្ងៃនៅទ្វីបអាមេរិកខាងជើងភាគខាងកើត", "EST": "ម៉ោងស្តង់ដារនៅទ្វីបអាមេរិកខាងជើងភាគខាងកើត", "GFT": "ម៉ោង\u200bនៅ\u200bហ្គីយ៉ាន\u200bបារាំង", "GMT": "ម៉ោងនៅគ្រីនវិច", "GST": "ម៉ោង\u200bស្តង់ដា\u200bនៅ\u200bកាល់", "GYT": "ម៉ោង\u200bនៅ\u200bហ្គីយ៉ាន", "HADT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bហាវៃ-អាល់ដ្យូសិន", "HAST": "ម៉ោង\u200bស្តង់ដារ\u200b\u200bនៅ\u200bហាវៃ-អាល់ដ្យូសិន", "HAT": "ម៉ោង\u200bពេលថ្ងៃ\u200bនៅ\u200bញូវហ្វោនឡែន", "HECU": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bគុយបា", "HEEG": "ម៉ោង\u200bនៅ\u200bហ្គ្រីនលែនខាង\u200bកើតនា\u200bរដូវ\u200bក្ដៅ", "HENOMX": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bម៉ិកស៊ិកភាគពាយព្យ", "HEOG": "ម៉ោងនៅហ្គ្រីនលែនខាងលិចនារដូវក្តៅ", "HEPM": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅសង់\u200bព្យែរ និង\u200bមីគុយឡុង", "HEPMX": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bប៉ាស៊ីហ្វិក\u200bម៉ិកស៊ិក", "HKST": "ម៉ោងនៅ\u200bហុងកុងនា\u200bរដូវ\u200bក្ដៅ\u200b", "HKT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bហុងកុង", "HNCU": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bគុយបា", "HNEG": "ម៉ោង\u200b\u200b\u200bស្តង់ដារ\u200bនៅ\u200b\u200bហ្គ្រីនលែន\u200bខាង\u200bកើត", "HNNOMX": "ម៉ោង\u200bស្តង់ដារនៅ\u200bម៉ិកស៊ិកភាគពាយព្យ", "HNOG": "ម៉ោងស្តង់ដារនៅហ្គ្រីនលែនខាងលិច", "HNPM": "ម៉ោង\u200bស្តង់ដារ\u200bនៅសង់\u200bព្យែរ និង\u200bមីគុយឡុង", "HNPMX": "ម៉ោង\u200bស្តង់ដា\u200bនៅ\u200bប៉ាស៊ីហ្វិក\u200bម៉ិកស៊ិក", "HNT": "ម៉ោង\u200b\u200bស្តង់ដារ\u200b\u200bនៅ\u200bញូវហ្វោនឡែន", "IST": "ម៉ោង\u200bស្តង់ដារនៅ\u200bឥណ្ឌា", "JDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅជប៉ុន", "JST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bជប៉ុន", "LHDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bឡតហៅ", "LHST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bឡត\u200bហៅ", "MDT": "MDT", "MESZ": "ម៉ោង\u200bនៅ\u200bអឺរ៉ុប\u200bកណ្ដាលនា\u200bរដូវ\u200bក្ដៅ", "MEZ": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអឺរ៉ុប\u200bកណ្ដាល", "MST": "MST", "MYT": "ម៉ោង\u200bនៅ\u200bម៉ាឡេស៊ី", "NZDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bនូវែលសេឡង់", "NZST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bនូវែលសេឡង់", "OESZ": "ម៉ោង\u200bនៅ\u200bអឺរ៉ុប\u200b\u200bខាង\u200bកើត\u200bនា\u200bរដូវ\u200bក្ដៅ", "OEZ": "ម៉ោង\u200bស្តង់ដារ\u200b\u200bនៅ\u200bអឺរ៉ុប\u200b\u200bខាង\u200bកើត\u200b", "PDT": "ម៉ោង\u200b\u200bពេលថ្ងៃនៅទ្វីប\u200bអាមរិក\u200bខាង\u200bជើងភាគខាងលិច", "PST": "ម៉ោងស្តង់ដារ\u200b\u200b\u200bនៅទ្វីប\u200bអាមរិក\u200bខាង\u200bជើងភាគខាងលិច", "SAST": "ម៉ោង\u200bនៅ\u200bអាហ្វ្រិក\u200bខាង\u200bត្បូង", "SGT": "ម៉ោង\u200bនៅ\u200bសិង្ហបូរី", "SRT": "ម៉ោង\u200bនៅ\u200bសូរីណាម", "TMST": "ម៉ោង\u200bរដូវ\u200bក្ដៅ\u200bនៅ\u200bតួកម៉េនីស្ថាន\u200b", "TMT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅតួកម៉េនីស្ថាន", "UYST": "ម៉ោង\u200bនៅ\u200bអ៊ុយរូហ្គាយនា\u200b\u200bរដូវ\u200bក្ដៅ", "UYT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអ៊ុយរូហ្គាយ", "VET": "ម៉ោង\u200bនៅ\u200bវ៉េណេស៊ុយអេឡា", "WARST": "ម៉ោង\u200bនៅ\u200bអាហ្សង់ទីនភាគខាងលិចនារដូវក្តៅ", "WART": "ម៉ោងស្តង់ដារ\u200bនៅ\u200bអាហ្សង់ទីនភាគខាងលិច", "WAST": "ម៉ោងនៅ\u200bអាហ្វ្រិក\u200b\u200b\u200bខាងលិច\u200b\u200bនារដូវ\u200bក្ដៅ\u200b", "WAT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអាហ្វ្រិក\u200bខាង\u200bលិច", "WESZ": "ម៉ោង\u200bនៅ\u200bអឺរ៉ុប\u200bខាង\u200bលិចនារដូវ\u200bក្ដៅ\u200b", "WEZ": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអឺរ៉ុប\u200bខាង\u200bលិច", "WIB": "ម៉ោង\u200bនៅ\u200bឥណ្ឌូណេស៊ី\u200b\u200bខាង\u200bលិច", "WIT": "ម៉ោង\u200bនៅ\u200bឥណ្ឌូណេស៊ី\u200b\u200bខាង\u200bកើត", "WITA": "ម៉ោង\u200bនៅ\u200bឥណ្ឌូណេស៊ី\u200b\u200b\u200bកណ្ដាល", "∅∅∅": "ម៉ោង\u200b\u200bនៅ\u200bប៉េរូនារដូវ\u200bក្ដៅ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_0722fc77da098723d3a7bcc9b8145055,
			fnFmtCurrency:         fnFmtCurrency_ef0590f3c464dc40ad41a959adae57e7,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["km_kh"] = func() locales.Translator {
		return &localen{

			locale:                 "km_KH",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "មករា", "កុម្ភៈ", "មីនា", "មេសា", "ឧសភា", "មិថុនា", "កក្កដា", "សីហា", "កញ្ញា", "តុលា", "វិច្ឆិកា", "ធ្នូ"},
			monthsNarrow:           []string{"", "ម", "ក", "ម", "ម", "ឧ", "ម", "ក", "ស", "ក", "ត", "វ", "ធ"},
			monthsWide:             []string{"", "មករា", "កុម្ភៈ", "មីនា", "មេសា", "ឧសភា", "មិថុនា", "កក្កដា", "សីហា", "កញ្ញា", "តុលា", "វិច្ឆិកា", "ធ្នូ"},
			daysAbbreviated:        []string{"អាទិត្យ", "ចន្ទ", "អង្គារ", "ពុធ", "ព្រហ", "សុក្រ", "សៅរ៍"},
			daysNarrow:             []string{"អ", "ច", "អ", "ព", "ព", "ស", "ស"},
			daysShort:              []string{"អា", "ច", "អ", "ពុ", "ព្រ", "សុ", "ស"},
			daysWide:               []string{"អាទិត្យ", "ច័ន្ទ", "អង្គារ", "ពុធ", "ព្រហស្បតិ៍", "សុក្រ", "សៅរ៍"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"មុន គ.ស.", "គ.ស."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"មុន\u200bគ្រិស្តសករាជ", "គ្រិស្តសករាជ"},
			timezones:              map[string]string{"ACDT": "ម៉ោង\u200bពេលថ្ងៃ\u200b\u200b\u200b\u200bនៅ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ACST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ACWDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200b\u200bភាគ\u200bខាង\u200bលិច\u200bនៃ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ACWST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bភាគ\u200bខាង\u200bលិច\u200bនៃ\u200bអូស្ត្រាលី\u200bកណ្ដាល", "ADT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bអាត្លង់ទិក", "AEDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bកើត", "AEST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bកើត", "AKDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200b\u200bអាឡាស្កា", "AKST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអាឡាស្កា", "ARST": "ម៉ោង\u200bនៅ\u200bអាហ្សង់ទីននារដូវក្តៅ", "ART": "ម៉ោងស្តង់ដារ\u200bនៅ\u200bអាហ្សង់ទីន", "AST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអាត្លង់ទិក", "AWDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bលិច", "AWST": "ម៉ោង\u200b\u200bស្តង់ដារ\u200bនៅ\u200bអូស្ត្រាលី\u200bខាង\u200bលិច", "BOT": "ម៉ោង\u200bនៅ\u200bបូលីវី", "BT": "ម៉ោងនៅប៊ូតង់", "CAT": "ម៉ោង\u200bនៅ\u200bអាហ្វ្រិក\u200bកណ្ដាល", "CDT": "ម៉ោង\u200b\u200bពេលថ្ងៃនៅ\u200bទ្វីបអាមេរិក\u200bខាង\u200bជើងភាគកណ្តាល", "CHADT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bចាថាំ", "CHAST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bចាថាំ", "CLST": "ម៉ោងនៅស៊ីលីនារដូវក្តៅ", "CLT": "ម៉ោងស្តង់ដារនៅស៊ីលី", "COST": "ម៉ោង\u200bនៅ\u200bកូឡុំប៊ីនា\u200bរដូវ\u200bក្ដៅ", "COT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bកូឡុំប៊ី", "CST": "ម៉ោង\u200b\u200bស្តង់ដារនៅ\u200bទ្វីបអាមេរិក\u200bខាង\u200bជើងភាគកណ្តាល", "ChST": "ម៉ោង\u200bស្តង់ដារនៅ\u200bចាំម៉ូរ៉ូ", "EAT": "ម៉ោង\u200bនៅ\u200bអាហ្វ្រិក\u200bខាង\u200bកើត", "ECT": "ម៉ោង\u200bនៅ\u200bអេក្វាទ័រ", "EDT": "ម៉ោងពេលថ្ងៃនៅទ្វីបអាមេរិកខាងជើងភាគខាងកើត", "EST": "ម៉ោងស្តង់ដារនៅទ្វីបអាមេរិកខាងជើងភាគខាងកើត", "GFT": "ម៉ោង\u200bនៅ\u200bហ្គីយ៉ាន\u200bបារាំង", "GMT": "ម៉ោងនៅគ្រីនវិច", "GST": "ម៉ោង\u200bស្តង់ដា\u200bនៅ\u200bកាល់", "GYT": "ម៉ោង\u200bនៅ\u200bហ្គីយ៉ាន", "HADT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bហាវៃ-អាល់ដ្យូសិន", "HAST": "ម៉ោង\u200bស្តង់ដារ\u200b\u200bនៅ\u200bហាវៃ-អាល់ដ្យូសិន", "HAT": "ម៉ោង\u200bពេលថ្ងៃ\u200bនៅ\u200bញូវហ្វោនឡែន", "HECU": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bគុយបា", "HEEG": "ម៉ោង\u200bនៅ\u200bហ្គ្រីនលែនខាង\u200bកើតនា\u200bរដូវ\u200bក្ដៅ", "HENOMX": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bម៉ិកស៊ិកភាគពាយព្យ", "HEOG": "ម៉ោងនៅហ្គ្រីនលែនខាងលិចនារដូវក្តៅ", "HEPM": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅសង់\u200bព្យែរ និង\u200bមីគុយឡុង", "HEPMX": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bប៉ាស៊ីហ្វិក\u200bម៉ិកស៊ិក", "HKST": "ម៉ោងនៅ\u200bហុងកុងនា\u200bរដូវ\u200bក្ដៅ\u200b", "HKT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bហុងកុង", "HNCU": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bគុយបា", "HNEG": "ម៉ោង\u200b\u200b\u200bស្តង់ដារ\u200bនៅ\u200b\u200bហ្គ្រីនលែន\u200bខាង\u200bកើត", "HNNOMX": "ម៉ោង\u200bស្តង់ដារនៅ\u200bម៉ិកស៊ិកភាគពាយព្យ", "HNOG": "ម៉ោងស្តង់ដារនៅហ្គ្រីនលែនខាងលិច", "HNPM": "ម៉ោង\u200bស្តង់ដារ\u200bនៅសង់\u200bព្យែរ និង\u200bមីគុយឡុង", "HNPMX": "ម៉ោង\u200bស្តង់ដា\u200bនៅ\u200bប៉ាស៊ីហ្វិក\u200bម៉ិកស៊ិក", "HNT": "ម៉ោង\u200b\u200bស្តង់ដារ\u200b\u200bនៅ\u200bញូវហ្វោនឡែន", "IST": "ម៉ោង\u200bស្តង់ដារនៅ\u200bឥណ្ឌា", "JDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅជប៉ុន", "JST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bជប៉ុន", "LHDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bឡតហៅ", "LHST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bឡត\u200bហៅ", "MDT": "ម៉ោង\u200bពេលថ្ងៃនៅតំបន់ភ្នំនៃទ្វីប\u200bអាមេរិក\u200b\u200b\u200bខាង\u200bជើង", "MESZ": "ម៉ោង\u200bនៅ\u200bអឺរ៉ុប\u200bកណ្ដាលនា\u200bរដូវ\u200bក្ដៅ", "MEZ": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអឺរ៉ុប\u200bកណ្ដាល", "MST": "ម៉ោងស្តង់ដារ\u200bនៅតំបន់ភ្នំនៃទ្វីប\u200bអាមេរិក\u200b\u200b\u200bខាង\u200bជើង", "MYT": "ម៉ោង\u200bនៅ\u200bម៉ាឡេស៊ី", "NZDT": "ម៉ោង\u200bពេល\u200bថ្ងៃ\u200bនៅ\u200bនូវែលសេឡង់", "NZST": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bនូវែលសេឡង់", "OESZ": "ម៉ោង\u200bនៅ\u200bអឺរ៉ុប\u200b\u200bខាង\u200bកើត\u200bនា\u200bរដូវ\u200bក្ដៅ", "OEZ": "ម៉ោង\u200bស្តង់ដារ\u200b\u200bនៅ\u200bអឺរ៉ុប\u200b\u200bខាង\u200bកើត\u200b", "PDT": "ម៉ោង\u200b\u200bពេលថ្ងៃនៅទ្វីប\u200bអាមរិក\u200bខាង\u200bជើងភាគខាងលិច", "PST": "ម៉ោងស្តង់ដារ\u200b\u200b\u200bនៅទ្វីប\u200bអាមរិក\u200bខាង\u200bជើងភាគខាងលិច", "SAST": "ម៉ោង\u200bនៅ\u200bអាហ្វ្រិក\u200bខាង\u200bត្បូង", "SGT": "ម៉ោង\u200bនៅ\u200bសិង្ហបូរី", "SRT": "ម៉ោង\u200bនៅ\u200bសូរីណាម", "TMST": "ម៉ោង\u200bរដូវ\u200bក្ដៅ\u200bនៅ\u200bតួកម៉េនីស្ថាន\u200b", "TMT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅតួកម៉េនីស្ថាន", "UYST": "ម៉ោង\u200bនៅ\u200bអ៊ុយរូហ្គាយនា\u200b\u200bរដូវ\u200bក្ដៅ", "UYT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអ៊ុយរូហ្គាយ", "VET": "ម៉ោង\u200bនៅ\u200bវ៉េណេស៊ុយអេឡា", "WARST": "ម៉ោង\u200bនៅ\u200bអាហ្សង់ទីនភាគខាងលិចនារដូវក្តៅ", "WART": "ម៉ោងស្តង់ដារ\u200bនៅ\u200bអាហ្សង់ទីនភាគខាងលិច", "WAST": "ម៉ោងនៅ\u200bអាហ្វ្រិក\u200b\u200b\u200bខាងលិច\u200b\u200bនារដូវ\u200bក្ដៅ\u200b", "WAT": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអាហ្វ្រិក\u200bខាង\u200bលិច", "WESZ": "ម៉ោង\u200bនៅ\u200bអឺរ៉ុប\u200bខាង\u200bលិចនារដូវ\u200bក្ដៅ\u200b", "WEZ": "ម៉ោង\u200bស្តង់ដារ\u200bនៅ\u200bអឺរ៉ុប\u200bខាង\u200bលិច", "WIB": "ម៉ោង\u200bនៅ\u200bឥណ្ឌូណេស៊ី\u200b\u200bខាង\u200bលិច", "WIT": "ម៉ោង\u200bនៅ\u200bឥណ្ឌូណេស៊ី\u200b\u200bខាង\u200bកើត", "WITA": "ម៉ោង\u200bនៅ\u200bឥណ្ឌូណេស៊ី\u200b\u200b\u200bកណ្ដាល", "∅∅∅": "ម៉ោង\u200bនៅ\u200bប្រាស៊ីលីយ៉ានា\u200b\u200bរដូវ\u200bក្ដៅ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_0722fc77da098723d3a7bcc9b8145055,
			fnFmtCurrency:         fnFmtCurrency_ef0590f3c464dc40ad41a959adae57e7,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kn"] = func() locales.Translator {
		return &localen{

			locale:             "kn",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ಜನವರಿ", "ಫೆಬ್ರವರಿ", "ಮಾರ್ಚ್", "ಏಪ್ರಿ", "ಮೇ", "ಜೂನ್", "ಜುಲೈ", "ಆಗ", "ಸೆಪ್ಟೆಂ", "ಅಕ್ಟೋ", "ನವೆಂ", "ಡಿಸೆಂ"},
			monthsNarrow:       []string{"", "ಜ", "ಫೆ", "ಮಾ", "ಏ", "ಮೇ", "ಜೂ", "ಜು", "ಆ", "ಸೆ", "ಅ", "ನ", "ಡಿ"},
			monthsWide:         []string{"", "ಜನವರಿ", "ಫೆಬ್ರವರಿ", "ಮಾರ್ಚ್", "ಏಪ್ರಿಲ್", "ಮೇ", "ಜೂನ್", "ಜುಲೈ", "ಆಗಸ್ಟ್", "ಸೆಪ್ಟೆಂಬರ್", "ಅಕ್ಟೋಬರ್", "ನವೆಂಬರ್", "ಡಿಸೆಂಬರ್"},
			daysAbbreviated:    []string{"ಭಾನು", "ಸೋಮ", "ಮಂಗಳ", "ಬುಧ", "ಗುರು", "ಶುಕ್ರ", "ಶನಿ"},
			daysNarrow:         []string{"ಭಾ", "ಸೋ", "ಮಂ", "ಬು", "ಗು", "ಶು", "ಶ"},
			daysShort:          []string{"ಭಾನು", "ಸೋಮ", "ಮಂಗಳ", "ಬುಧ", "ಗುರು", "ಶುಕ್ರ", "ಶನಿ"},
			daysWide:           []string{"ಭಾನುವಾರ", "ಸೋಮವಾರ", "ಮಂಗಳವಾರ", "ಬುಧವಾರ", "ಗುರುವಾರ", "ಶುಕ್ರವಾರ", "ಶನಿವಾರ"},
			periodsAbbreviated: []string{"ಪೂರ್ವಾಹ್ನ", "ಅಪರಾಹ್ನ"},
			periodsNarrow:      []string{"ಪೂ", "ಅ"},
			periodsWide:        []string{"ಪೂರ್ವಾಹ್ನ", "ಅಪರಾಹ್ನ"},
			erasAbbreviated:    []string{"ಕ್ರಿ.ಪೂ", "ಕ್ರಿ.ಶ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ಕ್ರಿಸ್ತ ಪೂರ್ವ", "ಕ್ರಿಸ್ತ ಶಕ"},
			timezones:          map[string]string{"ACDT": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಹಗಲು ಸಮಯ", "ACST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಪ್ರಮಾಣಿತ ಸಮಯ", "ACWDT": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಪಶ್ಚಿಮ ಹಗಲು ಸಮಯ", "ACWST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಪಶ್ಚಿಮ ಪ್ರಮಾಣಿತ ಸಮಯ", "ADT": "ಅಟ್ಲಾಂಟಿಕ್ ದಿನದ ಸಮಯ", "AEDT": "ಪೂರ್ವ ಆಸ್ಟ್ರೇಲಿಯಾದ ಹಗಲು ಸಮಯ", "AEST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಪೂರ್ವ ಪ್ರಮಾಣಿತ ಸಮಯ", "AKDT": "\u200cಅಲಾಸ್ಕಾ ಹಗಲು ಸಮಯ", "AKST": "ಅಲಸ್ಕಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "ARST": "ಅರ್ಜೆಂಟಿನಾ ಬೇಸಿಗೆ ಸಮಯ", "ART": "ಅರ್ಜೆಂಟೀನಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "AST": "ಅಟ್ಲಾಂಟಿಕ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "AWDT": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಪಶ್ಚಿಮ ಹಗಲು ಸಮಯ", "AWST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಪಶ್ಚಿಮ ಪ್ರಮಾಣಿತ ಸಮಯ", "BOT": "ಬೊಲಿವಿಯಾ ಸಮಯ", "BT": "ಭೂತಾನ್ ಸಮಯ", "CAT": "ಮಧ್ಯ ಆಫ್ರಿಕಾ ಸಮಯ", "CDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಕೇಂದ್ರೀಯ ದಿನದ ಸಮಯ", "CHADT": "ಚಥಾಮ್ ಹಗಲು ಸಮಯ", "CHAST": "ಚಥಾಮ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "CLST": "ಚಿಲಿ ಬೇಸಿಗೆ ಸಮಯ", "CLT": "ಚಿಲಿ ಪ್ರಮಾಣಿತ ಸಮಯ", "COST": "ಕೊಲಂಬಿಯಾ ಬೇಸಿಗೆ ಸಮಯ", "COT": "ಕೊಲಂಬಿಯಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "CST": "ಉತ್ತರ ಅಮೆರಿಕದ ಕೇಂದ್ರ ಪ್ರಮಾಣಿತ ಸಮಯ", "ChST": "ಚಮೋರೋ ಪ್ರಮಾಣಿತ ಸಮಯ", "EAT": "ಪೂರ್ವ ಆಫ್ರಿಕಾ ಸಮಯ", "ECT": "ಈಕ್ವೆಡಾರ್ ಸಮಯ", "EDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೂರ್ವದ ದಿನದ ಸಮಯ", "EST": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೂರ್ವದ ಪ್ರಮಾಣಿತ ಸಮಯ", "GFT": "ಫ್ರೆಂಚ್ ಗಯಾನಾ ಸಮಯ", "GMT": "ಗ್ರೀನ್\u200cವಿಚ್ ಸರಾಸರಿ ಕಾಲಮಾನ", "GST": "ಗಲ್ಫ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "GYT": "ಗಯಾನಾ ಸಮಯ", "HADT": "ಹವಾಯಿ-ಅಲ್ಯುಟಿಯನ್ ಹಗಲು ಸಮಯ", "HAST": "ಹವಾಯಿ-ಅಲ್ಯುಟಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HAT": "ನ್ಯೂಫೌಂಡ್\u200cಲ್ಯಾಂಡ್ ದಿನದ ಸಮಯ", "HECU": "ಕ್ಯೂಬಾ ದಿನದ ಸಮಯ", "HEEG": "ಪೂರ್ವ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಬೇಸಿಗೆ ಸಮಯ", "HENOMX": "ವಾಯವ್ಯ ಮೆಕ್ಸಿಕೊ ಹಗಲು ಸಮಯ", "HEOG": "ಪಶ್ಚಿಮ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಬೇಸಿಗೆ ಸಮಯ", "HEPM": "ಸೇಂಟ್ ಪಿಯರ್ ಮತ್ತು ಮಿಕ್ವೆಲನ್ ಹಗಲು ಸಮಯ", "HEPMX": "ಮೆಕ್ಸಿಕನ್ ಪೆಸಿಫಿಕ್ ಹಗಲು ಸಮಯ", "HKST": "ಹಾಂಗ್ ಕಾಂಗ್ ಬೇಸಿಗೆ ಸಮಯ", "HKT": "ಹಾಂಗ್ ಕಾಂಗ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNCU": "ಕ್ಯೂಬಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNEG": "ಪೂರ್ವ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNNOMX": "ವಾಯವ್ಯ ಮೆಕ್ಸಿಕೊ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNOG": "ಪಶ್ಚಿಮ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNPM": "ಸೇಂಟ್ ಪಿಯರ್ ಮತ್ತು ಮಿಕ್ವೆಲನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNPMX": "ಮೆಕ್ಸಿಕನ್ ಪೆಸಿಫಿಕ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNT": "ನ್ಯೂಫೌಂಡ್\u200cಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "IST": "ಭಾರತೀಯ ಪ್ರಮಾಣಿತ ಸಮಯ", "JDT": "ಜಪಾನ್ ಹಗಲು ಸಮಯ", "JST": "ಜಪಾನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "LHDT": "ಲಾರ್ಡ್ ಹೋವ್ ಬೆಳಗಿನ ಸಮಯ", "LHST": "ಲಾರ್ಡ್ ಹೋವ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "MDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಪರ್ವತ ದಿನದ ಸಮಯ", "MESZ": "ಮಧ್ಯ ಯುರೋಪಿಯನ್ ಬೇಸಿಗೆ ಸಮಯ", "MEZ": "ಮಧ್ಯ ಯುರೋಪಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "MST": "ಉತ್ತರ ಅಮೆರಿಕದ ಪರ್ವತ ಪ್ರಮಾಣಿತ ಸಮಯ", "MYT": "ಮಲೇಷಿಯಾ ಸಮಯ", "NZDT": "ನ್ಯೂಜಿಲ್ಯಾಂಡ್ ಹಗಲು ಸಮಯ", "NZST": "ನ್ಯೂಜಿಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "OESZ": "ಪೂರ್ವ ಯುರೋಪಿಯನ್ ಬೇಸಿಗೆ ಸಮಯ", "OEZ": "ಪೂರ್ವ ಯುರೋಪಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "PDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೆಸಿಫಿಕ್ ದಿನದ ಸಮಯ", "PST": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೆಸಿಫಿಕ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "SAST": "ದಕ್ಷಿಣ ಆಫ್ರಿಕಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "SGT": "ಸಿಂಗಪುರ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "SRT": "ಸುರಿನೇಮ್ ಸಮಯ", "TMST": "ತುರ್ಕ್\u200cಮೇನಿಸ್ತಾನ್ ಬೇಸಿಗೆ ಸಮಯ", "TMT": "ತುರ್ಕ್\u200cಮೇನಿಸ್ತಾನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "UYST": "ಉರುಗ್ವೇ ಬೇಸಿಗೆ ಸಮಯ", "UYT": "ಉರುಗ್ವೇ ಪ್ರಮಾಣಿತ ಸಮಯ", "VET": "ವೆನಿಜುವೆಲಾ ಸಮಯ", "WARST": "ಪಶ್ಚಿಮ ಅರ್ಜೆಂಟೀನಾ ಬೇಸಿಗೆ ಸಮಯ", "WART": "ಪಶ್ಚಿಮ ಅರ್ಜೆಂಟೀನಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "WAST": "ಪಶ್ಚಿಮ ಆಫ್ರಿಕಾ ಬೇಸಿಗೆ ಸಮಯ", "WAT": "ಪಶ್ಚಿಮ ಆಫ್ರಿಕಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "WESZ": "ಪಶ್ಚಿಮ ಯುರೋಪಿಯನ್ ಬೇಸಿಗೆ ಸಮಯ", "WEZ": "ಪಶ್ಚಿಮ ಯುರೋಪಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "WIB": "ಪಶ್ಚಿಮ ಇಂಡೋನೇಷಿಯ ಸಮಯ", "WIT": "ಪೂರ್ವ ಇಂಡೋನೇಷಿಯಾ ಸಮಯ", "WITA": "ಮಧ್ಯ ಇಂಡೋನೇಷಿಯಾ ಸಮಯ", "∅∅∅": "ಬ್ರೆಸಿಲಿಯಾ ಬೇಸಿಗೆ ಸಮಯ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_abd60cef530beeba744de08bba2ad540,
			fnFmtTimeLong:         fnFmtTimeLong_28302cc672efd8ad015891c59efdc448,
			fnFmtTimeMedium:       fnFmtTimeMedium_9efd56b286449b21c42b6b2ea8c1f34c,
			fnFmtTimeShort:        fnFmtTimeShort_11febbd269414b302ab4045dd63f47b7,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kn_in"] = func() locales.Translator {
		return &localen{

			locale:             "kn_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ಜನವರಿ", "ಫೆಬ್ರವರಿ", "ಮಾರ್ಚ್", "ಏಪ್ರಿ", "ಮೇ", "ಜೂನ್", "ಜುಲೈ", "ಆಗ", "ಸೆಪ್ಟೆಂ", "ಅಕ್ಟೋ", "ನವೆಂ", "ಡಿಸೆಂ"},
			monthsNarrow:       []string{"", "ಜ", "ಫೆ", "ಮಾ", "ಏ", "ಮೇ", "ಜೂ", "ಜು", "ಆ", "ಸೆ", "ಅ", "ನ", "ಡಿ"},
			monthsWide:         []string{"", "ಜನವರಿ", "ಫೆಬ್ರವರಿ", "ಮಾರ್ಚ್", "ಏಪ್ರಿಲ್", "ಮೇ", "ಜೂನ್", "ಜುಲೈ", "ಆಗಸ್ಟ್", "ಸೆಪ್ಟೆಂಬರ್", "ಅಕ್ಟೋಬರ್", "ನವೆಂಬರ್", "ಡಿಸೆಂಬರ್"},
			daysAbbreviated:    []string{"ಭಾನು", "ಸೋಮ", "ಮಂಗಳ", "ಬುಧ", "ಗುರು", "ಶುಕ್ರ", "ಶನಿ"},
			daysNarrow:         []string{"ಭಾ", "ಸೋ", "ಮಂ", "ಬು", "ಗು", "ಶು", "ಶ"},
			daysShort:          []string{"ಭಾನು", "ಸೋಮ", "ಮಂಗಳ", "ಬುಧ", "ಗುರು", "ಶುಕ್ರ", "ಶನಿ"},
			daysWide:           []string{"ಭಾನುವಾರ", "ಸೋಮವಾರ", "ಮಂಗಳವಾರ", "ಬುಧವಾರ", "ಗುರುವಾರ", "ಶುಕ್ರವಾರ", "ಶನಿವಾರ"},
			periodsAbbreviated: []string{"ಪೂರ್ವಾಹ್ನ", "ಅಪರಾಹ್ನ"},
			periodsNarrow:      []string{"ಪೂ", "ಅ"},
			periodsWide:        []string{"ಪೂರ್ವಾಹ್ನ", "ಅಪರಾಹ್ನ"},
			erasAbbreviated:    []string{"ಕ್ರಿ.ಪೂ", "ಕ್ರಿ.ಶ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ಕ್ರಿಸ್ತ ಪೂರ್ವ", "ಕ್ರಿಸ್ತ ಶಕ"},
			timezones:          map[string]string{"ACDT": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಹಗಲು ಸಮಯ", "ACST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಪ್ರಮಾಣಿತ ಸಮಯ", "ACWDT": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಪಶ್ಚಿಮ ಹಗಲು ಸಮಯ", "ACWST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಕೇಂದ್ರ ಪಶ್ಚಿಮ ಪ್ರಮಾಣಿತ ಸಮಯ", "ADT": "ಅಟ್ಲಾಂಟಿಕ್ ದಿನದ ಸಮಯ", "AEDT": "ಪೂರ್ವ ಆಸ್ಟ್ರೇಲಿಯಾದ ಹಗಲು ಸಮಯ", "AEST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಪೂರ್ವ ಪ್ರಮಾಣಿತ ಸಮಯ", "AKDT": "\u200cಅಲಾಸ್ಕಾ ಹಗಲು ಸಮಯ", "AKST": "ಅಲಸ್ಕಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "ARST": "ಅರ್ಜೆಂಟಿನಾ ಬೇಸಿಗೆ ಸಮಯ", "ART": "ಅರ್ಜೆಂಟೀನಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "AST": "ಅಟ್ಲಾಂಟಿಕ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "AWDT": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಪಶ್ಚಿಮ ಹಗಲು ಸಮಯ", "AWST": "ಆಸ್ಟ್ರೇಲಿಯಾದ ಪಶ್ಚಿಮ ಪ್ರಮಾಣಿತ ಸಮಯ", "BOT": "ಬೊಲಿವಿಯಾ ಸಮಯ", "BT": "ಭೂತಾನ್ ಸಮಯ", "CAT": "ಮಧ್ಯ ಆಫ್ರಿಕಾ ಸಮಯ", "CDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಕೇಂದ್ರೀಯ ದಿನದ ಸಮಯ", "CHADT": "ಚಥಾಮ್ ಹಗಲು ಸಮಯ", "CHAST": "ಚಥಾಮ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "CLST": "ಚಿಲಿ ಬೇಸಿಗೆ ಸಮಯ", "CLT": "ಚಿಲಿ ಪ್ರಮಾಣಿತ ಸಮಯ", "COST": "ಕೊಲಂಬಿಯಾ ಬೇಸಿಗೆ ಸಮಯ", "COT": "ಕೊಲಂಬಿಯಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "CST": "ಉತ್ತರ ಅಮೆರಿಕದ ಕೇಂದ್ರ ಪ್ರಮಾಣಿತ ಸಮಯ", "ChST": "ಚಮೋರೋ ಪ್ರಮಾಣಿತ ಸಮಯ", "EAT": "ಪೂರ್ವ ಆಫ್ರಿಕಾ ಸಮಯ", "ECT": "ಈಕ್ವೆಡಾರ್ ಸಮಯ", "EDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೂರ್ವದ ದಿನದ ಸಮಯ", "EST": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೂರ್ವದ ಪ್ರಮಾಣಿತ ಸಮಯ", "GFT": "ಫ್ರೆಂಚ್ ಗಯಾನಾ ಸಮಯ", "GMT": "ಗ್ರೀನ್\u200cವಿಚ್ ಸರಾಸರಿ ಕಾಲಮಾನ", "GST": "ಗಲ್ಫ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "GYT": "ಗಯಾನಾ ಸಮಯ", "HADT": "ಹವಾಯಿ-ಅಲ್ಯುಟಿಯನ್ ಹಗಲು ಸಮಯ", "HAST": "ಹವಾಯಿ-ಅಲ್ಯುಟಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HAT": "ನ್ಯೂಫೌಂಡ್\u200cಲ್ಯಾಂಡ್ ದಿನದ ಸಮಯ", "HECU": "ಕ್ಯೂಬಾ ದಿನದ ಸಮಯ", "HEEG": "ಪೂರ್ವ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಬೇಸಿಗೆ ಸಮಯ", "HENOMX": "ವಾಯವ್ಯ ಮೆಕ್ಸಿಕೊ ಹಗಲು ಸಮಯ", "HEOG": "ಪಶ್ಚಿಮ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಬೇಸಿಗೆ ಸಮಯ", "HEPM": "ಸೇಂಟ್ ಪಿಯರ್ ಮತ್ತು ಮಿಕ್ವೆಲನ್ ಹಗಲು ಸಮಯ", "HEPMX": "ಮೆಕ್ಸಿಕನ್ ಪೆಸಿಫಿಕ್ ಹಗಲು ಸಮಯ", "HKST": "ಹಾಂಗ್ ಕಾಂಗ್ ಬೇಸಿಗೆ ಸಮಯ", "HKT": "ಹಾಂಗ್ ಕಾಂಗ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNCU": "ಕ್ಯೂಬಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNEG": "ಪೂರ್ವ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNNOMX": "ವಾಯವ್ಯ ಮೆಕ್ಸಿಕೊ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNOG": "ಪಶ್ಚಿಮ ಗ್ರೀನ್\u200cಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNPM": "ಸೇಂಟ್ ಪಿಯರ್ ಮತ್ತು ಮಿಕ್ವೆಲನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNPMX": "ಮೆಕ್ಸಿಕನ್ ಪೆಸಿಫಿಕ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "HNT": "ನ್ಯೂಫೌಂಡ್\u200cಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "IST": "ಭಾರತೀಯ ಪ್ರಮಾಣಿತ ಸಮಯ", "JDT": "ಜಪಾನ್ ಹಗಲು ಸಮಯ", "JST": "ಜಪಾನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "LHDT": "ಲಾರ್ಡ್ ಹೋವ್ ಬೆಳಗಿನ ಸಮಯ", "LHST": "ಲಾರ್ಡ್ ಹೋವ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "MDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಪರ್ವತ ದಿನದ ಸಮಯ", "MESZ": "ಮಧ್ಯ ಯುರೋಪಿಯನ್ ಬೇಸಿಗೆ ಸಮಯ", "MEZ": "ಮಧ್ಯ ಯುರೋಪಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "MST": "ಉತ್ತರ ಅಮೆರಿಕದ ಪರ್ವತ ಪ್ರಮಾಣಿತ ಸಮಯ", "MYT": "ಮಲೇಷಿಯಾ ಸಮಯ", "NZDT": "ನ್ಯೂಜಿಲ್ಯಾಂಡ್ ಹಗಲು ಸಮಯ", "NZST": "ನ್ಯೂಜಿಲ್ಯಾಂಡ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "OESZ": "ಪೂರ್ವ ಯುರೋಪಿಯನ್ ಬೇಸಿಗೆ ಸಮಯ", "OEZ": "ಪೂರ್ವ ಯುರೋಪಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "PDT": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೆಸಿಫಿಕ್ ದಿನದ ಸಮಯ", "PST": "ಉತ್ತರ ಅಮೆರಿಕದ ಪೆಸಿಫಿಕ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "SAST": "ದಕ್ಷಿಣ ಆಫ್ರಿಕಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "SGT": "ಸಿಂಗಪುರ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "SRT": "ಸುರಿನೇಮ್ ಸಮಯ", "TMST": "ತುರ್ಕ್\u200cಮೇನಿಸ್ತಾನ್ ಬೇಸಿಗೆ ಸಮಯ", "TMT": "ತುರ್ಕ್\u200cಮೇನಿಸ್ತಾನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "UYST": "ಉರುಗ್ವೇ ಬೇಸಿಗೆ ಸಮಯ", "UYT": "ಉರುಗ್ವೇ ಪ್ರಮಾಣಿತ ಸಮಯ", "VET": "ವೆನಿಜುವೆಲಾ ಸಮಯ", "WARST": "ಪಶ್ಚಿಮ ಅರ್ಜೆಂಟೀನಾ ಬೇಸಿಗೆ ಸಮಯ", "WART": "ಪಶ್ಚಿಮ ಅರ್ಜೆಂಟೀನಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "WAST": "ಪಶ್ಚಿಮ ಆಫ್ರಿಕಾ ಬೇಸಿಗೆ ಸಮಯ", "WAT": "ಪಶ್ಚಿಮ ಆಫ್ರಿಕಾ ಪ್ರಮಾಣಿತ ಸಮಯ", "WESZ": "ಪಶ್ಚಿಮ ಯುರೋಪಿಯನ್ ಬೇಸಿಗೆ ಸಮಯ", "WEZ": "ಪಶ್ಚಿಮ ಯುರೋಪಿಯನ್ ಪ್ರಮಾಣಿತ ಸಮಯ", "WIB": "ಪಶ್ಚಿಮ ಇಂಡೋನೇಷಿಯ ಸಮಯ", "WIT": "ಪೂರ್ವ ಇಂಡೋನೇಷಿಯಾ ಸಮಯ", "WITA": "ಮಧ್ಯ ಇಂಡೋನೇಷಿಯಾ ಸಮಯ", "∅∅∅": "ಬ್ರೆಸಿಲಿಯಾ ಬೇಸಿಗೆ ಸಮಯ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_abd60cef530beeba744de08bba2ad540,
			fnFmtTimeLong:         fnFmtTimeLong_28302cc672efd8ad015891c59efdc448,
			fnFmtTimeMedium:       fnFmtTimeMedium_9efd56b286449b21c42b6b2ea8c1f34c,
			fnFmtTimeShort:        fnFmtTimeShort_11febbd269414b302ab4045dd63f47b7,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ko"] = func() locales.Translator {
		return &localen{

			locale:                 "ko",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			monthsNarrow:           []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			monthsWide:             []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			daysAbbreviated:        []string{"일", "월", "화", "수", "목", "금", "토"},
			daysNarrow:             []string{"일", "월", "화", "수", "목", "금", "토"},
			daysShort:              []string{"일", "월", "화", "수", "목", "금", "토"},
			daysWide:               []string{"일요일", "월요일", "화요일", "수요일", "목요일", "금요일", "토요일"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"오전", "오후"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"기원전", "서기"},
			timezones:              map[string]string{"ACDT": "오스트레일리아 중부 하계 표준시", "ACST": "오스트레일리아 중부 표준시", "ACWDT": "오스트레일리아 중서부 하계 표준시", "ACWST": "오스트레일리아 중서부 표준시", "ADT": "대서양 하계 표준시", "AEDT": "오스트레일리아 동부 하계 표준시", "AEST": "오스트레일리아 동부 표준시", "AKDT": "알래스카 하계 표준시", "AKST": "알래스카 표준시", "ARST": "아르헨티나 하계 표준시", "ART": "아르헨티나 표준시", "AST": "대서양 표준시", "AWDT": "오스트레일리아 서부 하계 표준시", "AWST": "오스트레일리아 서부 표준시", "BOT": "볼리비아 시간", "BT": "부탄 시간", "CAT": "중앙아프리카 시간", "CDT": "미 중부 하계 표준시", "CHADT": "채텀 하계 표준시", "CHAST": "채텀 표준시", "CLST": "칠레 하계 표준시", "CLT": "칠레 표준시", "COST": "콜롬비아 하계 표준시", "COT": "콜롬비아 표준시", "CST": "미 중부 표준시", "ChST": "차모로 시간", "EAT": "동아프리카 시간", "ECT": "에콰도르 시간", "EDT": "미 동부 하계 표준시", "EST": "미 동부 표준시", "GFT": "프랑스령 가이아나 시간", "GMT": "그리니치 표준시", "GST": "걸프만 표준시", "GYT": "가이아나 시간", "HADT": "하와이 알류샨 하계 표준시", "HAST": "하와이 알류샨 표준시", "HAT": "뉴펀들랜드 하계 표준시", "HECU": "쿠바 하계 표준시", "HEEG": "그린란드 동부 하계 표준시", "HENOMX": "멕시코 북서부 하계 표준시", "HEOG": "그린란드 서부 하계 표준시", "HEPM": "세인트피에르 미클롱 하계 표준시", "HEPMX": "멕시코 태평양 하계 표준시", "HKST": "홍콩 하계 표준시", "HKT": "홍콩 표준시", "HNCU": "쿠바 표준시", "HNEG": "그린란드 동부 표준시", "HNNOMX": "멕시코 북서부 표준시", "HNOG": "그린란드 서부 표준시", "HNPM": "세인트피에르 미클롱 표준시", "HNPMX": "멕시코 태평양 표준시", "HNT": "뉴펀들랜드 표준시", "IST": "인도 표준시", "JDT": "일본 하계 표준시", "JST": "일본 표준시", "LHDT": "로드 하우 하계 표준시", "LHST": "로드 하우 표준시", "MDT": "미 산지 하계 표준시", "MESZ": "중부 유럽 하계 표준시", "MEZ": "중부 유럽 표준시", "MST": "미 산악 표준시", "MYT": "말레이시아 시간", "NZDT": "뉴질랜드 하계 표준시", "NZST": "뉴질랜드 표준시", "OESZ": "동유럽 하계 표준시", "OEZ": "동유럽 표준시", "PDT": "미 태평양 하계 표준시", "PST": "미 태평양 표준시", "SAST": "남아프리카 시간", "SGT": "싱가포르 표준시", "SRT": "수리남 시간", "TMST": "투르크메니스탄 하계 표준시", "TMT": "투르크메니스탄 표준시", "UYST": "우루과이 하계 표준시", "UYT": "우루과이 표준시", "VET": "베네수엘라 시간", "WARST": "아르헨티나 서부 하계 표준시", "WART": "아르헨티나 서부 표준시", "WAST": "서아프리카 하계 표준시", "WAT": "서아프리카 표준시", "WESZ": "서유럽 하계 표준시", "WEZ": "서유럽 표준시", "WIB": "서부 인도네시아 시간", "WIT": "동부 인도네시아 시간", "WITA": "중부 인도네시아 시간", "∅∅∅": "브라질리아 하계 표준시"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c611a5e966ec03dbdacdf654df3799f0,
			fnFmtDateLong:         fnFmtDateLong_17c2343e73783fd11c1ad64b6b231979,
			fnFmtDateMedium:       fnFmtDateMedium_f69e9157403c463ade0395f3ff9240fe,
			fnFmtDateShort:        fnFmtDateShort_d881b9f1f4fd8f22e57a48f28a75e30a,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_bfbb513b6692ade43eb8d3cf5179818c,
			fnFmtTimeLong:         fnFmtTimeLong_a4e3f96b4a6b77d07b6c318bbc1c74ac,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ko_kp"] = func() locales.Translator {
		return &localen{

			locale:                 "ko_KP",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			monthsNarrow:           []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			monthsWide:             []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			daysAbbreviated:        []string{"일", "월", "화", "수", "목", "금", "토"},
			daysNarrow:             []string{"일", "월", "화", "수", "목", "금", "토"},
			daysShort:              []string{"일", "월", "화", "수", "목", "금", "토"},
			daysWide:               []string{"일요일", "월요일", "화요일", "수요일", "목요일", "금요일", "토요일"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"오전", "오후"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"기원전", "서기"},
			timezones:              map[string]string{"ACDT": "오스트레일리아 중부 하계 표준시", "ACST": "오스트레일리아 중부 표준시", "ACWDT": "오스트레일리아 중서부 하계 표준시", "ACWST": "오스트레일리아 중서부 표준시", "ADT": "대서양 하계 표준시", "AEDT": "오스트레일리아 동부 하계 표준시", "AEST": "오스트레일리아 동부 표준시", "AKDT": "알래스카 하계 표준시", "AKST": "알래스카 표준시", "ARST": "아르헨티나 하계 표준시", "ART": "아르헨티나 표준시", "AST": "대서양 표준시", "AWDT": "오스트레일리아 서부 하계 표준시", "AWST": "오스트레일리아 서부 표준시", "BOT": "볼리비아 시간", "BT": "부탄 시간", "CAT": "중앙아프리카 시간", "CDT": "미 중부 하계 표준시", "CHADT": "채텀 하계 표준시", "CHAST": "채텀 표준시", "CLST": "칠레 하계 표준시", "CLT": "칠레 표준시", "COST": "콜롬비아 하계 표준시", "COT": "콜롬비아 표준시", "CST": "미 중부 표준시", "ChST": "차모로 시간", "EAT": "동아프리카 시간", "ECT": "에콰도르 시간", "EDT": "미 동부 하계 표준시", "EST": "미 동부 표준시", "GFT": "프랑스령 가이아나 시간", "GMT": "그리니치 표준시", "GST": "걸프만 표준시", "GYT": "가이아나 시간", "HADT": "하와이 알류샨 하계 표준시", "HAST": "하와이 알류샨 표준시", "HAT": "뉴펀들랜드 하계 표준시", "HECU": "쿠바 하계 표준시", "HEEG": "그린란드 동부 하계 표준시", "HENOMX": "멕시코 북서부 하계 표준시", "HEOG": "그린란드 서부 하계 표준시", "HEPM": "세인트피에르 미클롱 하계 표준시", "HEPMX": "멕시코 태평양 하계 표준시", "HKST": "홍콩 하계 표준시", "HKT": "홍콩 표준시", "HNCU": "쿠바 표준시", "HNEG": "그린란드 동부 표준시", "HNNOMX": "멕시코 북서부 표준시", "HNOG": "그린란드 서부 표준시", "HNPM": "세인트피에르 미클롱 표준시", "HNPMX": "멕시코 태평양 표준시", "HNT": "뉴펀들랜드 표준시", "IST": "인도 표준시", "JDT": "일본 하계 표준시", "JST": "일본 표준시", "LHDT": "로드 하우 하계 표준시", "LHST": "로드 하우 표준시", "MDT": "마카오 하계 표준시", "MESZ": "중부 유럽 하계 표준시", "MEZ": "중부 유럽 표준시", "MST": "마카오 표준 시간", "MYT": "말레이시아 시간", "NZDT": "뉴질랜드 하계 표준시", "NZST": "뉴질랜드 표준시", "OESZ": "동유럽 하계 표준시", "OEZ": "동유럽 표준시", "PDT": "미 태평양 하계 표준시", "PST": "미 태평양 표준시", "SAST": "남아프리카 시간", "SGT": "싱가포르 표준시", "SRT": "수리남 시간", "TMST": "투르크메니스탄 하계 표준시", "TMT": "투르크메니스탄 표준시", "UYST": "우루과이 하계 표준시", "UYT": "우루과이 표준시", "VET": "베네수엘라 시간", "WARST": "아르헨티나 서부 하계 표준시", "WART": "아르헨티나 서부 표준시", "WAST": "서아프리카 하계 표준시", "WAT": "서아프리카 표준시", "WESZ": "서유럽 하계 표준시", "WEZ": "서유럽 표준시", "WIB": "서부 인도네시아 시간", "WIT": "동부 인도네시아 시간", "WITA": "중부 인도네시아 시간", "∅∅∅": "페루 하계 표준시"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c611a5e966ec03dbdacdf654df3799f0,
			fnFmtDateLong:         fnFmtDateLong_17c2343e73783fd11c1ad64b6b231979,
			fnFmtDateMedium:       fnFmtDateMedium_f69e9157403c463ade0395f3ff9240fe,
			fnFmtDateShort:        fnFmtDateShort_d881b9f1f4fd8f22e57a48f28a75e30a,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_bfbb513b6692ade43eb8d3cf5179818c,
			fnFmtTimeLong:         fnFmtTimeLong_a4e3f96b4a6b77d07b6c318bbc1c74ac,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ko_kr"] = func() locales.Translator {
		return &localen{

			locale:                 "ko_KR",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			monthsNarrow:           []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			monthsWide:             []string{"", "1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"},
			daysAbbreviated:        []string{"일", "월", "화", "수", "목", "금", "토"},
			daysNarrow:             []string{"일", "월", "화", "수", "목", "금", "토"},
			daysShort:              []string{"일", "월", "화", "수", "목", "금", "토"},
			daysWide:               []string{"일요일", "월요일", "화요일", "수요일", "목요일", "금요일", "토요일"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"오전", "오후"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"기원전", "서기"},
			timezones:              map[string]string{"ACDT": "오스트레일리아 중부 하계 표준시", "ACST": "오스트레일리아 중부 표준시", "ACWDT": "오스트레일리아 중서부 하계 표준시", "ACWST": "오스트레일리아 중서부 표준시", "ADT": "대서양 하계 표준시", "AEDT": "오스트레일리아 동부 하계 표준시", "AEST": "오스트레일리아 동부 표준시", "AKDT": "알래스카 하계 표준시", "AKST": "알래스카 표준시", "ARST": "아르헨티나 하계 표준시", "ART": "아르헨티나 표준시", "AST": "대서양 표준시", "AWDT": "오스트레일리아 서부 하계 표준시", "AWST": "오스트레일리아 서부 표준시", "BOT": "볼리비아 시간", "BT": "부탄 시간", "CAT": "중앙아프리카 시간", "CDT": "미 중부 하계 표준시", "CHADT": "채텀 하계 표준시", "CHAST": "채텀 표준시", "CLST": "칠레 하계 표준시", "CLT": "칠레 표준시", "COST": "콜롬비아 하계 표준시", "COT": "콜롬비아 표준시", "CST": "미 중부 표준시", "ChST": "차모로 시간", "EAT": "동아프리카 시간", "ECT": "에콰도르 시간", "EDT": "미 동부 하계 표준시", "EST": "미 동부 표준시", "GFT": "프랑스령 가이아나 시간", "GMT": "그리니치 표준시", "GST": "걸프만 표준시", "GYT": "가이아나 시간", "HADT": "하와이 알류샨 하계 표준시", "HAST": "하와이 알류샨 표준시", "HAT": "뉴펀들랜드 하계 표준시", "HECU": "쿠바 하계 표준시", "HEEG": "그린란드 동부 하계 표준시", "HENOMX": "멕시코 북서부 하계 표준시", "HEOG": "그린란드 서부 하계 표준시", "HEPM": "세인트피에르 미클롱 하계 표준시", "HEPMX": "멕시코 태평양 하계 표준시", "HKST": "홍콩 하계 표준시", "HKT": "홍콩 표준시", "HNCU": "쿠바 표준시", "HNEG": "그린란드 동부 표준시", "HNNOMX": "멕시코 북서부 표준시", "HNOG": "그린란드 서부 표준시", "HNPM": "세인트피에르 미클롱 표준시", "HNPMX": "멕시코 태평양 표준시", "HNT": "뉴펀들랜드 표준시", "IST": "인도 표준시", "JDT": "일본 하계 표준시", "JST": "일본 표준시", "LHDT": "로드 하우 하계 표준시", "LHST": "로드 하우 표준시", "MDT": "미 산지 하계 표준시", "MESZ": "중부 유럽 하계 표준시", "MEZ": "중부 유럽 표준시", "MST": "미 산악 표준시", "MYT": "말레이시아 시간", "NZDT": "뉴질랜드 하계 표준시", "NZST": "뉴질랜드 표준시", "OESZ": "동유럽 하계 표준시", "OEZ": "동유럽 표준시", "PDT": "미 태평양 하계 표준시", "PST": "미 태평양 표준시", "SAST": "남아프리카 시간", "SGT": "싱가포르 표준시", "SRT": "수리남 시간", "TMST": "투르크메니스탄 하계 표준시", "TMT": "투르크메니스탄 표준시", "UYST": "우루과이 하계 표준시", "UYT": "우루과이 표준시", "VET": "베네수엘라 시간", "WARST": "아르헨티나 서부 하계 표준시", "WART": "아르헨티나 서부 표준시", "WAST": "서아프리카 하계 표준시", "WAT": "서아프리카 표준시", "WESZ": "서유럽 하계 표준시", "WEZ": "서유럽 표준시", "WIB": "서부 인도네시아 시간", "WIT": "동부 인도네시아 시간", "WITA": "중부 인도네시아 시간", "∅∅∅": "브라질리아 하계 표준시"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c611a5e966ec03dbdacdf654df3799f0,
			fnFmtDateLong:         fnFmtDateLong_17c2343e73783fd11c1ad64b6b231979,
			fnFmtDateMedium:       fnFmtDateMedium_f69e9157403c463ade0395f3ff9240fe,
			fnFmtDateShort:        fnFmtDateShort_d881b9f1f4fd8f22e57a48f28a75e30a,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_bfbb513b6692ade43eb8d3cf5179818c,
			fnFmtTimeLong:         fnFmtTimeLong_a4e3f96b4a6b77d07b6c318bbc1c74ac,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kok"] = func() locales.Translator {
		return &localen{

			locale:             "kok",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "रॉन", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "जानेवारी", "फेब्रुवारी", "मार्च", "एप्रिल", "मे", "जून", "जुलाय", "आगोस्त", "सप्टेंबर", "ऑक्टोबर", "नोव्हेंबर", "डिसेंबर"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "जानेवारी", "फेब्रुवारी", "मार्च", "एप्रिल", "मे", "जून", "जुलाय", "आगोस्त", "सप्टेंबर", "ऑक्टोबर", "नोव्हेंबर", "डिसेंबर"},
			daysAbbreviated:    []string{"आयतार", "सोमार", "मंगळार", "बुधवार", "गुरुवार", "शुक्रार", "शेनवार"},
			daysNarrow:         []string{"आ", "सो", "मं", "बु", "गु", "शु", "शे"},
			daysShort:          []string{"आय", "सोम", "मंगळ", "बुध", "गुरु", "शुक्र", "शेन"},
			daysWide:           []string{"आयतार", "सोमार", "मंगळार", "बुधवार", "गुरुवार", "शुक्रार", "शेनवार"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"a", "p"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"क्रिस्तपूर्व", "क्रिस्तशखा"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"क्रिस्तपूर्व", "क्रिस्तशखा"},
			timezones:          map[string]string{"ACDT": "ऑस्ट्रेलीयन मध्य डेलायट वेळ", "ACST": "ऑस्ट्रेलीयन मध्य प्रमाणित वेळ", "ACWDT": "ऑस्ट्रेलीयन मध्य अस्तंत डेलायट वेळ", "ACWST": "ऑस्ट्रेलीयन मध्य अस्तंत प्रमाणित वेळ", "ADT": "अटलांटीक डेलायट वेळ", "AEDT": "ऑस्ट्रेलीयन उदेंत डेलायट वेळ", "AEST": "ऑस्ट्रेलीयन उदेंत प्रमाणित वेळ", "AKDT": "अलास्का डेलायट वेळ", "AKST": "अलास्का प्रमाणित वेळ", "ARST": "अर्जेंटिना ग्रीष्म वेळ", "ART": "अर्जेंटिना प्रमाणित वेळ", "AST": "अटलांटीक प्रमाणित वेळ", "AWDT": "ऑस्ट्रेलीयन अस्तंत डेलायट वेळ", "AWST": "ऑस्ट्रेलीयन अस्तंत प्रमाणित वेळ", "BOT": "बोलिव्हिया वेळ", "BT": "भूतान", "CAT": "मध्य आफ्रिका वेळ", "CDT": "मध्य डेलायट वेळ", "CHADT": "चॅथम डेलायट वेळ", "CHAST": "चॅथम प्रमाणित वेळ", "CLST": "चिली ग्रीष्म वेळ", "CLT": "चिली प्रमाणित वेळ", "COST": "कोलंबिया ग्रीष्म वेळ", "COT": "कोलंबिया प्रमाणित वेळ", "CST": "मध्य प्रमाणित वेळ", "ChST": "कॅमोरा प्रमाणित वेळ", "EAT": "उदेंत आफ्रिका वेळ", "ECT": "इक्वेडोर वेळ", "EDT": "उदेंत डेलायट वेळ", "EST": "उदेंत प्रमाणित वेळ", "GFT": "फ्रेंच गयाना वेळ", "GMT": "ग्रीनविच मध्यc वेळ", "GST": "गल्फ प्रमाणित वेळ", "GYT": "गुयाना वेळ", "HADT": "हवाई-अलेयुशिन डेलायट वेळ", "HAST": "हवाई-अलेयुशिन प्रमाणित वेळ", "HAT": "न्युफावंडलँड डेलायट वेळ", "HECU": "क्युबा डेलायट वेळ", "HEEG": "उदेंत ग्रीनलँड ग्रीष्म वेळ", "HENOMX": "वायव्य मेक्सिको डेलायट वेळ", "HEOG": "अस्तंत ग्रीनलँड ग्रीष्म वेळ", "HEPM": "सेंट पियर आनी मिकलान डेलायट वेळ", "HEPMX": "मेक्सिकन प्रशांत डेलायट वेळ", "HKST": "हाँग काँग ग्रीष्म वेळ", "HKT": "हाँग काँग प्रमाणित वेळ", "HNCU": "क्युबा प्रमाणित वेळ", "HNEG": "उदेंत ग्रीनलँड प्रमाणित वेळ", "HNNOMX": "वायव्य मेक्सिको प्रमाणित वेळ", "HNOG": "अस्तंत ग्रीनलँड प्रमाणित वेळ", "HNPM": "सेंट पियर आनी मिकलान प्रमाणित वेळ", "HNPMX": "मेक्सिकन प्रशांत प्रमाणित वेळ", "HNT": "न्युफावंडलँड प्रमाणित वेळ", "IST": "भारतीय प्रमाणित वेळ", "JDT": "जपान डेलायट वेळ", "JST": "जपान प्रमाणित वेळ", "LHDT": "लॉर्ड होवे डेलायट वेळ", "LHST": "लॉर्ड होवे प्रमाणित वेळ", "MDT": "पर्वतीय डेलायट वेळ", "MESZ": "मध्य युवरोपियन ग्रीष्म वेळ", "MEZ": "मध्य युरोपियन प्रमाणित वेळ", "MST": "पर्वतीय प्रमाणित वेळ", "MYT": "मलेशिया वेळ", "NZDT": "न्युझीलॅन्ड डेलायट वेळ", "NZST": "न्युझीलॅन्ड प्रमाणित वेळ", "OESZ": "उदेंत युरोपियन ग्रीष्म वेळ", "OEZ": "उदेंत युरोपियन प्रमाणित वेळ", "PDT": "प्रशांत डेलायट वेळ", "PST": "प्रशांत प्रमाणित वेळ", "SAST": "दक्षिण आफ्रिका प्रमाणित वेळ", "SGT": "सिंगापूर प्रमाणित वेळ", "SRT": "सुरिनाम वेळ", "TMST": "तुर्कमेनिस्तान ग्रीष्म वेळ", "TMT": "तुर्कमेनिस्तान प्रमाणित वेळ", "UYST": "उरुग्वे ग्रीष्म वेळ", "UYT": "उरुग्वे प्रमाणित वेळ", "VET": "वेनेझुएला वेळ", "WARST": "अस्तंत अर्जेंटिना ग्रीष्म वेळ", "WART": "अस्तंत अर्जेंटिना प्रमाणित वेळ", "WAST": "अस्तंत आफ्रिका ग्रीष्म वेळ", "WAT": "अस्तंत आफ्रिका प्रमाणित वेळ", "WESZ": "अस्तंत युरोपियन ग्रीष्म वेळ", "WEZ": "अस्तंत युरोपियन प्रमाणित वेळ", "WIB": "अस्तंत इंडोनेशिया वेळ", "WIT": "उदेंत इंडोनेशिया वेळ", "WITA": "मध्य इंडोनेशिया वेळ", "∅∅∅": "अझोरेस ग्रीष्म वेळ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_20b00d66ebb3cefd65ec76fba139eaf9,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kok_in"] = func() locales.Translator {
		return &localen{

			locale:             "kok_IN",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "जानेवारी", "फेब्रुवारी", "मार्च", "एप्रिल", "मे", "जून", "जुलाय", "आगोस्त", "सप्टेंबर", "ऑक्टोबर", "नोव्हेंबर", "डिसेंबर"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "जानेवारी", "फेब्रुवारी", "मार्च", "एप्रिल", "मे", "जून", "जुलाय", "आगोस्त", "सप्टेंबर", "ऑक्टोबर", "नोव्हेंबर", "डिसेंबर"},
			daysAbbreviated:    []string{"आयतार", "सोमार", "मंगळार", "बुधवार", "गुरुवार", "शुक्रार", "शेनवार"},
			daysNarrow:         []string{"आ", "सो", "मं", "बु", "गु", "शु", "शे"},
			daysShort:          []string{"आय", "सोम", "मंगळ", "बुध", "गुरु", "शुक्र", "शेन"},
			daysWide:           []string{"आयतार", "सोमार", "मंगळार", "बुधवार", "गुरुवार", "शुक्रार", "शेनवार"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"a", "p"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"क्रिस्तपूर्व", "क्रिस्तशखा"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"क्रिस्तपूर्व", "क्रिस्तशखा"},
			timezones:          map[string]string{"ACDT": "ऑस्ट्रेलीयन मध्य डेलायट वेळ", "ACST": "ऑस्ट्रेलीयन मध्य प्रमाणित वेळ", "ACWDT": "ऑस्ट्रेलीयन मध्य अस्तंत डेलायट वेळ", "ACWST": "ऑस्ट्रेलीयन मध्य अस्तंत प्रमाणित वेळ", "ADT": "अटलांटीक डेलायट वेळ", "AEDT": "ऑस्ट्रेलीयन उदेंत डेलायट वेळ", "AEST": "ऑस्ट्रेलीयन उदेंत प्रमाणित वेळ", "AKDT": "अलास्का डेलायट वेळ", "AKST": "अलास्का प्रमाणित वेळ", "ARST": "अर्जेंटिना ग्रीष्म वेळ", "ART": "अर्जेंटिना प्रमाणित वेळ", "AST": "अटलांटीक प्रमाणित वेळ", "AWDT": "ऑस्ट्रेलीयन अस्तंत डेलायट वेळ", "AWST": "ऑस्ट्रेलीयन अस्तंत प्रमाणित वेळ", "BOT": "बोलिव्हिया वेळ", "BT": "भूतान", "CAT": "मध्य आफ्रिका वेळ", "CDT": "मध्य डेलायट वेळ", "CHADT": "चॅथम डेलायट वेळ", "CHAST": "चॅथम प्रमाणित वेळ", "CLST": "चिली ग्रीष्म वेळ", "CLT": "चिली प्रमाणित वेळ", "COST": "कोलंबिया ग्रीष्म वेळ", "COT": "कोलंबिया प्रमाणित वेळ", "CST": "मध्य प्रमाणित वेळ", "ChST": "कॅमोरा प्रमाणित वेळ", "EAT": "उदेंत आफ्रिका वेळ", "ECT": "इक्वेडोर वेळ", "EDT": "उदेंत डेलायट वेळ", "EST": "उदेंत प्रमाणित वेळ", "GFT": "फ्रेंच गयाना वेळ", "GMT": "ग्रीनविच मध्यc वेळ", "GST": "गल्फ प्रमाणित वेळ", "GYT": "गुयाना वेळ", "HADT": "हवाई-अलेयुशिन डेलायट वेळ", "HAST": "हवाई-अलेयुशिन प्रमाणित वेळ", "HAT": "न्युफावंडलँड डेलायट वेळ", "HECU": "क्युबा डेलायट वेळ", "HEEG": "उदेंत ग्रीनलँड ग्रीष्म वेळ", "HENOMX": "वायव्य मेक्सिको डेलायट वेळ", "HEOG": "अस्तंत ग्रीनलँड ग्रीष्म वेळ", "HEPM": "सेंट पियर आनी मिकलान डेलायट वेळ", "HEPMX": "मेक्सिकन प्रशांत डेलायट वेळ", "HKST": "हाँग काँग ग्रीष्म वेळ", "HKT": "हाँग काँग प्रमाणित वेळ", "HNCU": "क्युबा प्रमाणित वेळ", "HNEG": "उदेंत ग्रीनलँड प्रमाणित वेळ", "HNNOMX": "वायव्य मेक्सिको प्रमाणित वेळ", "HNOG": "अस्तंत ग्रीनलँड प्रमाणित वेळ", "HNPM": "सेंट पियर आनी मिकलान प्रमाणित वेळ", "HNPMX": "मेक्सिकन प्रशांत प्रमाणित वेळ", "HNT": "न्युफावंडलँड प्रमाणित वेळ", "IST": "भारतीय प्रमाणित वेळ", "JDT": "जपान डेलायट वेळ", "JST": "जपान प्रमाणित वेळ", "LHDT": "लॉर्ड होवे डेलायट वेळ", "LHST": "लॉर्ड होवे प्रमाणित वेळ", "MDT": "पर्वतीय डेलायट वेळ", "MESZ": "मध्य युवरोपियन ग्रीष्म वेळ", "MEZ": "मध्य युरोपियन प्रमाणित वेळ", "MST": "पर्वतीय प्रमाणित वेळ", "MYT": "मलेशिया वेळ", "NZDT": "न्युझीलॅन्ड डेलायट वेळ", "NZST": "न्युझीलॅन्ड प्रमाणित वेळ", "OESZ": "उदेंत युरोपियन ग्रीष्म वेळ", "OEZ": "उदेंत युरोपियन प्रमाणित वेळ", "PDT": "प्रशांत डेलायट वेळ", "PST": "प्रशांत प्रमाणित वेळ", "SAST": "दक्षिण आफ्रिका प्रमाणित वेळ", "SGT": "सिंगापूर प्रमाणित वेळ", "SRT": "सुरिनाम वेळ", "TMST": "तुर्कमेनिस्तान ग्रीष्म वेळ", "TMT": "तुर्कमेनिस्तान प्रमाणित वेळ", "UYST": "उरुग्वे ग्रीष्म वेळ", "UYT": "उरुग्वे प्रमाणित वेळ", "VET": "वेनेझुएला वेळ", "WARST": "अस्तंत अर्जेंटिना ग्रीष्म वेळ", "WART": "अस्तंत अर्जेंटिना प्रमाणित वेळ", "WAST": "अस्तंत आफ्रिका ग्रीष्म वेळ", "WAT": "अस्तंत आफ्रिका प्रमाणित वेळ", "WESZ": "अस्तंत युरोपियन ग्रीष्म वेळ", "WEZ": "अस्तंत युरोपियन प्रमाणित वेळ", "WIB": "अस्तंत इंडोनेशिया वेळ", "WIT": "उदेंत इंडोनेशिया वेळ", "WITA": "मध्य इंडोनेशिया वेळ", "∅∅∅": "अझोरेस ग्रीष्म वेळ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_20b00d66ebb3cefd65ec76fba139eaf9,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ks"] = func() locales.Translator {
		return &localen{

			locale:                 "ks",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsNarrow:           []string{"", "ج", "ف", "م", "ا", "م", "ج", "ج", "ا", "س", "س", "ا", "ن"},
			monthsWide:             []string{"", "جنؤری", "فرؤری", "مارٕچ", "اپریل", "میٔ", "جوٗن", "جوٗلایی", "اگست", "ستمبر", "اکتوٗبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"آتھوار", "ژٔندٕروار", "بۆموار", "بودوار", "برؠسوار", "جُمہ", "بٹوار"},
			daysNarrow:             []string{"ا", "ژ", "ب", "ب", "ب", "ج", "ب"},
			daysWide:               []string{"اَتھوار", "ژٔندرٕروار", "بۆموار", "بودوار", "برؠسوار", "جُمہ", "بٹوار"},
			erasAbbreviated:        []string{"بی سی", "اے ڈی"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبٕل مسیٖح", "عیٖسوی سنہٕ"},
			timezones:              map[string]string{"ACDT": "آسٹریلِیَن مرکزی ڈےلایِٔٹ ٹایِم", "ACST": "آسٹریلِیَن مرکزی سٹینڑاڑ ٹایِم", "ACWDT": "آسٹریلِیَن مرکزی مغربی ڈےلایِٔٹ ٹایِم", "ACWST": "آسٹریلِیَن مرکزی مغربی سٹینڑاڑ ٹایِم", "ADT": "اؠٹلانٹِک ڈےلایِٔٹ ٹایِم", "AEDT": "آسٹریلِیَن مشرقی ڈےلایِٔٹ ٹایِم", "AEST": "آسٹریلِیَن مشرقی سٹینڑاڑ ٹایِم", "AKDT": "اؠلاسکا ڈےلایِٔٹ ٹایِم", "AKST": "اؠلاسکا سٹینڑاڑ ٹایِم", "ARST": "ارجؠنٹیٖنا سَمَر ٹایِم", "ART": "ارجؠنٹیٖنا سٹینڑاڑ ٹایِم", "AST": "اؠٹلانٹِک سٹینڑاڑ ٹایِم", "AWDT": "آسٹریلِیَن مغرِبیٖ ڈےلایٔٹ ٹایِم", "AWST": "آسٹریلِیَن مغرِبی سٹینڑاڑ ٹایِم", "BOT": "بولِوِیا ٹایِم", "BT": "بوٗٹان ٹایِم", "CAT": "مرکزی افریٖقا ٹایِم", "CDT": "مرکزی ڈےلایِٔٹ ٹایِم", "CHADT": "چؠتھَم سَمَر ٹایِم", "CHAST": "کؠتھَم سٹینڑاڑ ٹایِم", "CLST": "چِلی سَمَر ٹایِم", "CLT": "چِلی سٹینڑاڑ ٹایِم", "COST": "کولومبِیا سَمَر ٹایِم", "COT": "کولومبِیا سٹینڑاڑ ٹایِم", "CST": "مرکزی سٹینڑاڑ ٹایِم", "ChST": "کؠمورو سٹینڑاڑ ٹایِم", "EAT": "مشرقی افریٖقا ٹایِم", "ECT": "اِکویڑَر ٹایِم", "EDT": "مشرقی ڈےلایِٔٹ ٹایِم", "EST": "مشرقی سٹینڑاڑ ٹایِم", "GFT": "فرؠنچ گیوٗؠنا ٹایِم", "GMT": "گریٖن وِچ میٖن ٹایِم", "GST": "گَلف سٹینڑاڑ ٹایِم", "GYT": "گُیَنا ٹایِم", "HADT": "حَواے اؠلیوٗٹِیَن سَمَر ٹایِم", "HAST": "حَواے اؠلیوٗٹِیَن سٹینڑاڑ ٹایِم", "HAT": "نیوٗ فاونڑ لینڑ ڑےلایِٔٹ ٹایِم", "HECU": "کیوٗبا ڈےلایِٔٹ ٹایِم", "HEEG": "مشرِقی گریٖن لینڑُک سَمَر ٹایِم", "HENOMX": "HENOMX", "HEOG": "مغرِبی گریٖن لینڑُک سَمَر ٹایِم", "HEPM": "سینٹ پَیری مِقیوٗلَن ڑےلایِٔٹ ٹایِم", "HEPMX": "HEPMX", "HKST": "حانگ کانگ سَمَر ٹایِم", "HKT": "حانگ کانگ سٹینڑاڑ ٹایِم", "HNCU": "کیوٗبا سٹینڑاڑ ٹایِم", "HNEG": "مشرِقی گریٖن لینڑُک سٹینڑاڑ ٹایِم", "HNNOMX": "HNNOMX", "HNOG": "مغرِبی گریٖن لینڑُک سٹینڑاڑ ٹایِم", "HNPM": "سینٹ پَیری مِقیوٗلَن سٹینڑاڑ ٹایِم", "HNPMX": "HNPMX", "HNT": "نیوٗ فاونڑلینڑ سٹینڑاڑ ٹایِم", "IST": "ہِندوستان", "JDT": "جاپٲنۍ ڑےلایِٔٹ ٹایِم", "JST": "جاپٲنۍ سٹینڑاڑ ٹایِم", "LHDT": "لعاڑ ڑےلایٔٹ ٹایِم", "LHST": "لعاڑ حووے سٹینڑاڑ ٹایِم", "MDT": "مَکَعوٗ سَمَر ٹایِم", "MESZ": "مرکزی یوٗرپی سَمَر ٹایِم", "MEZ": "مرکزی یوٗرپی سٹینڑاڑ ٹایِم", "MST": "مَکَعوٗ سٹینڑاڑ ٹایِم", "MYT": "مَلیشِیا ٹایِم", "NZDT": "نِوزِلینڑ ڑےلایٔٹ ٹایِم", "NZST": "نِوزِلینڑ سٹینڑاڑ ٹایِم", "OESZ": "مشرقی یوٗرپی سَمَر ٹایِم", "OEZ": "مشرقی یوٗرپی سٹینڑاڑ ٹایِم", "PDT": "پیسِفِک ڈےلایِٔٹ ٹایِم", "PST": "پیسِفِک سٹینڑاڑ ٹایِم", "SAST": "جنوٗبی افریقا ٹایِم", "SGT": "سِنگاپوٗر ٹایِم", "SRT": "سُرِنام ٹایِم", "TMST": "تُرکمؠنِستان سَمَر ٹایِم", "TMT": "تُرکمؠنِستان سٹینڑاڑ ٹایِم", "UYST": "یوٗرؠگوَے سَمَر ٹایِم", "UYT": "یوٗرؠگوَے سٹینڑاڑ ٹایِم", "VET": "وؠنؠزیوٗلا ٹایِم", "WARST": "مغربی ارجؠنٹیٖنا سَمَر ٹایِم", "WART": "مغربی ارجؠنٹیٖنا سٹینڑاڑ ٹایِم", "WAST": "مغربی افریٖقا سَمَر ٹایِم", "WAT": "مغربی افریٖقا سٹینڑاڑ ٹایِم", "WESZ": "مغرِبی یوٗرِپی سَمَر ٹایِم", "WEZ": "مغرِبی یوٗرپی سٹینڑاڑ ٹایِم", "WIB": "مغرِبی اِنڑونیشِیا ٹایِم", "WIT": "مشرِقی اِنڑونیشِیا ٹایِم", "WITA": "مرکزی اِنڑونیشِیا ٹایِم", "∅∅∅": "پٔروٗ سَمَر ٹایِم"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d34cde82a90714ae7e620d04971d2faf,
			fnFmtCurrency:         fnFmtCurrency_b96d5354a212ba0fb9e6fdc6738a84f8,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_b796bfb92dd260d7d0890a75e953e03c,
			fnFmtPercent:          fnFmtPercent_aa714d39dff020dd613d3cb5622ff940,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ks_in"] = func() locales.Translator {
		return &localen{

			locale:                 "ks_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsNarrow:           []string{"", "ج", "ف", "م", "ا", "م", "ج", "ج", "ا", "س", "س", "ا", "ن"},
			monthsWide:             []string{"", "جنؤری", "فرؤری", "مارٕچ", "اپریل", "میٔ", "جوٗن", "جوٗلایی", "اگست", "ستمبر", "اکتوٗبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"آتھوار", "ژٔندٕروار", "بۆموار", "بودوار", "برؠسوار", "جُمہ", "بٹوار"},
			daysNarrow:             []string{"ا", "ژ", "ب", "ب", "ب", "ج", "ب"},
			daysWide:               []string{"اَتھوار", "ژٔندرٕروار", "بۆموار", "بودوار", "برؠسوار", "جُمہ", "بٹوار"},
			erasAbbreviated:        []string{"بی سی", "اے ڈی"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبٕل مسیٖح", "عیٖسوی سنہٕ"},
			timezones:              map[string]string{"ACDT": "آسٹریلِیَن مرکزی ڈےلایِٔٹ ٹایِم", "ACST": "آسٹریلِیَن مرکزی سٹینڑاڑ ٹایِم", "ACWDT": "آسٹریلِیَن مرکزی مغربی ڈےلایِٔٹ ٹایِم", "ACWST": "آسٹریلِیَن مرکزی مغربی سٹینڑاڑ ٹایِم", "ADT": "اؠٹلانٹِک ڈےلایِٔٹ ٹایِم", "AEDT": "آسٹریلِیَن مشرقی ڈےلایِٔٹ ٹایِم", "AEST": "آسٹریلِیَن مشرقی سٹینڑاڑ ٹایِم", "AKDT": "اؠلاسکا ڈےلایِٔٹ ٹایِم", "AKST": "اؠلاسکا سٹینڑاڑ ٹایِم", "ARST": "ارجؠنٹیٖنا سَمَر ٹایِم", "ART": "ارجؠنٹیٖنا سٹینڑاڑ ٹایِم", "AST": "اؠٹلانٹِک سٹینڑاڑ ٹایِم", "AWDT": "آسٹریلِیَن مغرِبیٖ ڈےلایٔٹ ٹایِم", "AWST": "آسٹریلِیَن مغرِبی سٹینڑاڑ ٹایِم", "BOT": "بولِوِیا ٹایِم", "BT": "بوٗٹان ٹایِم", "CAT": "مرکزی افریٖقا ٹایِم", "CDT": "مرکزی ڈےلایِٔٹ ٹایِم", "CHADT": "چؠتھَم سَمَر ٹایِم", "CHAST": "کؠتھَم سٹینڑاڑ ٹایِم", "CLST": "چِلی سَمَر ٹایِم", "CLT": "چِلی سٹینڑاڑ ٹایِم", "COST": "کولومبِیا سَمَر ٹایِم", "COT": "کولومبِیا سٹینڑاڑ ٹایِم", "CST": "مرکزی سٹینڑاڑ ٹایِم", "ChST": "کؠمورو سٹینڑاڑ ٹایِم", "EAT": "مشرقی افریٖقا ٹایِم", "ECT": "اِکویڑَر ٹایِم", "EDT": "مشرقی ڈےلایِٔٹ ٹایِم", "EST": "مشرقی سٹینڑاڑ ٹایِم", "GFT": "فرؠنچ گیوٗؠنا ٹایِم", "GMT": "گریٖن وِچ میٖن ٹایِم", "GST": "گَلف سٹینڑاڑ ٹایِم", "GYT": "گُیَنا ٹایِم", "HADT": "حَواے اؠلیوٗٹِیَن سَمَر ٹایِم", "HAST": "حَواے اؠلیوٗٹِیَن سٹینڑاڑ ٹایِم", "HAT": "نیوٗ فاونڑ لینڑ ڑےلایِٔٹ ٹایِم", "HECU": "کیوٗبا ڈےلایِٔٹ ٹایِم", "HEEG": "مشرِقی گریٖن لینڑُک سَمَر ٹایِم", "HENOMX": "HENOMX", "HEOG": "مغرِبی گریٖن لینڑُک سَمَر ٹایِم", "HEPM": "سینٹ پَیری مِقیوٗلَن ڑےلایِٔٹ ٹایِم", "HEPMX": "HEPMX", "HKST": "حانگ کانگ سَمَر ٹایِم", "HKT": "حانگ کانگ سٹینڑاڑ ٹایِم", "HNCU": "کیوٗبا سٹینڑاڑ ٹایِم", "HNEG": "مشرِقی گریٖن لینڑُک سٹینڑاڑ ٹایِم", "HNNOMX": "HNNOMX", "HNOG": "مغرِبی گریٖن لینڑُک سٹینڑاڑ ٹایِم", "HNPM": "سینٹ پَیری مِقیوٗلَن سٹینڑاڑ ٹایِم", "HNPMX": "HNPMX", "HNT": "نیوٗ فاونڑلینڑ سٹینڑاڑ ٹایِم", "IST": "ہِندوستان", "JDT": "جاپٲنۍ ڑےلایِٔٹ ٹایِم", "JST": "جاپٲنۍ سٹینڑاڑ ٹایِم", "LHDT": "لعاڑ ڑےلایٔٹ ٹایِم", "LHST": "لعاڑ حووے سٹینڑاڑ ٹایِم", "MDT": "ماونٹین ڈےلایِٔٹ ٹایِم", "MESZ": "مرکزی یوٗرپی سَمَر ٹایِم", "MEZ": "مرکزی یوٗرپی سٹینڑاڑ ٹایِم", "MST": "ماونٹین سٹینڑاڑ ٹایِم", "MYT": "مَلیشِیا ٹایِم", "NZDT": "نِوزِلینڑ ڑےلایٔٹ ٹایِم", "NZST": "نِوزِلینڑ سٹینڑاڑ ٹایِم", "OESZ": "مشرقی یوٗرپی سَمَر ٹایِم", "OEZ": "مشرقی یوٗرپی سٹینڑاڑ ٹایِم", "PDT": "پیسِفِک ڈےلایِٔٹ ٹایِم", "PST": "پیسِفِک سٹینڑاڑ ٹایِم", "SAST": "جنوٗبی افریقا ٹایِم", "SGT": "سِنگاپوٗر ٹایِم", "SRT": "سُرِنام ٹایِم", "TMST": "تُرکمؠنِستان سَمَر ٹایِم", "TMT": "تُرکمؠنِستان سٹینڑاڑ ٹایِم", "UYST": "یوٗرؠگوَے سَمَر ٹایِم", "UYT": "یوٗرؠگوَے سٹینڑاڑ ٹایِم", "VET": "وؠنؠزیوٗلا ٹایِم", "WARST": "مغربی ارجؠنٹیٖنا سَمَر ٹایِم", "WART": "مغربی ارجؠنٹیٖنا سٹینڑاڑ ٹایِم", "WAST": "مغربی افریٖقا سَمَر ٹایِم", "WAT": "مغربی افریٖقا سٹینڑاڑ ٹایِم", "WESZ": "مغرِبی یوٗرِپی سَمَر ٹایِم", "WEZ": "مغرِبی یوٗرپی سٹینڑاڑ ٹایِم", "WIB": "مغرِبی اِنڑونیشِیا ٹایِم", "WIT": "مشرِقی اِنڑونیشِیا ٹایِم", "WITA": "مرکزی اِنڑونیشِیا ٹایِم", "∅∅∅": "اؠزورٕس سَمَر ٹ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d34cde82a90714ae7e620d04971d2faf,
			fnFmtCurrency:         fnFmtCurrency_b96d5354a212ba0fb9e6fdc6738a84f8,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_b796bfb92dd260d7d0890a75e953e03c,
			fnFmtPercent:          fnFmtPercent_aa714d39dff020dd613d3cb5622ff940,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ksb"] = func() locales.Translator {
		return &localen{

			locale:             "ksb",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januali", "Febluali", "Machi", "Aplili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jmn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"2", "3", "4", "5", "A", "I", "1"},
			daysWide:           []string{"Jumaapii", "Jumaatatu", "Jumaane", "Jumaatano", "Alhamisi", "Ijumaa", "Jumaamosi"},
			periodsAbbreviated: []string{"makeo", "nyiaghuo"},
			periodsWide:        []string{"makeo", "nyiaghuo"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Klisto", "Baada ya Klisto"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ksb_tz"] = func() locales.Translator {
		return &localen{

			locale:             "ksb_TZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januali", "Febluali", "Machi", "Aplili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jmn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"2", "3", "4", "5", "A", "I", "1"},
			daysWide:           []string{"Jumaapii", "Jumaatatu", "Jumaane", "Jumaatano", "Alhamisi", "Ijumaa", "Jumaamosi"},
			periodsAbbreviated: []string{"makeo", "nyiaghuo"},
			periodsWide:        []string{"makeo", "nyiaghuo"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Klisto", "Baada ya Klisto"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ksf"] = func() locales.Translator {
		return &localen{

			locale:                 "ksf",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ŋ1", "ŋ2", "ŋ3", "ŋ4", "ŋ5", "ŋ6", "ŋ7", "ŋ8", "ŋ9", "ŋ10", "ŋ11", "ŋ12"},
			monthsWide:             []string{"", "ŋwíí a ntɔ́ntɔ", "ŋwíí akǝ bɛ́ɛ", "ŋwíí akǝ ráá", "ŋwíí akǝ nin", "ŋwíí akǝ táan", "ŋwíí akǝ táafɔk", "ŋwíí akǝ táabɛɛ", "ŋwíí akǝ táaraa", "ŋwíí akǝ táanin", "ŋwíí akǝ ntɛk", "ŋwíí akǝ ntɛk di bɔ́k", "ŋwíí akǝ ntɛk di bɛ́ɛ"},
			daysAbbreviated:        []string{"sɔ́n", "lǝn", "maa", "mɛk", "jǝǝ", "júm", "sam"},
			daysNarrow:             []string{"s", "l", "m", "m", "j", "j", "s"},
			daysWide:               []string{"sɔ́ndǝ", "lǝndí", "maadí", "mɛkrɛdí", "jǝǝdí", "júmbá", "samdí"},
			periodsAbbreviated:     []string{"sárúwá", "cɛɛ́nko"},
			periodsWide:            []string{"sárúwá", "cɛɛ́nko"},
			erasAbbreviated:        []string{"d.Y.", "k.Y."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"di Yɛ́sus aká yálɛ", "cámɛɛn kǝ kǝbɔpka Y"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ksf_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "ksf_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ŋ1", "ŋ2", "ŋ3", "ŋ4", "ŋ5", "ŋ6", "ŋ7", "ŋ8", "ŋ9", "ŋ10", "ŋ11", "ŋ12"},
			monthsWide:             []string{"", "ŋwíí a ntɔ́ntɔ", "ŋwíí akǝ bɛ́ɛ", "ŋwíí akǝ ráá", "ŋwíí akǝ nin", "ŋwíí akǝ táan", "ŋwíí akǝ táafɔk", "ŋwíí akǝ táabɛɛ", "ŋwíí akǝ táaraa", "ŋwíí akǝ táanin", "ŋwíí akǝ ntɛk", "ŋwíí akǝ ntɛk di bɔ́k", "ŋwíí akǝ ntɛk di bɛ́ɛ"},
			daysAbbreviated:        []string{"sɔ́n", "lǝn", "maa", "mɛk", "jǝǝ", "júm", "sam"},
			daysNarrow:             []string{"s", "l", "m", "m", "j", "j", "s"},
			daysWide:               []string{"sɔ́ndǝ", "lǝndí", "maadí", "mɛkrɛdí", "jǝǝdí", "júmbá", "samdí"},
			periodsAbbreviated:     []string{"sárúwá", "cɛɛ́nko"},
			periodsWide:            []string{"sárúwá", "cɛɛ́nko"},
			erasAbbreviated:        []string{"d.Y.", "k.Y."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"di Yɛ́sus aká yálɛ", "cámɛɛn kǝ kǝbɔpka Y"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ksh"] = func() locales.Translator {
		return &localen{

			locale:                 "ksh",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Fäb", "Mäz", "Apr", "Mai", "Jun", "Jul", "Ouj", "Säp", "Okt", "Nov", "Dez"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Jannewa", "Fäbrowa", "Määz", "Aprell", "Mai", "Juuni", "Juuli", "Oujoß", "Septämber", "Oktohber", "Novämber", "Dezämber"},
			daysAbbreviated:        []string{"Su.", "Mo.", "Di.", "Me.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Di", "Me", "Du", "Fr", "Sa"},
			daysWide:               []string{"Sunndaach", "Mohndaach", "Dinnsdaach", "Metwoch", "Dunnersdaach", "Friidaach", "Samsdaach"},
			periodsAbbreviated:     []string{"v.M.", "n.M."},
			periodsWide:            []string{"Uhr vörmiddaachs", "Uhr nommendaachs"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"vC", "nC"},
			erasWide:               []string{"vür Krestos", "noh Krestos"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Zentraal-Affrekaanesche Zigg", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Oß-Affrekaanesche Zigg", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Greenwich sing Standat-Zick", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Meddel-Europpa sing Summerzick", "MEZ": "Meddel-Europpa sing jewöhnlijje Zick", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oß-Europpa sing Summerzick", "OEZ": "Oß-Europpa sing jewöhnlijje Zick", "PDT": "PDT", "PST": "PST", "SAST": "Söd-Affrekaanesche Zigg", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Wäß-Affrekaanesche Sommerzigg", "WAT": "Jewöhnlijje Wäß-Affrekaanesche Zigg", "WESZ": "Weß-Europpa sing Summerzick", "WEZ": "Weß-Europpa sing jewöhnlijje Zick", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_fed1c76cd90b8905708121045ba1b87c,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_694de915ae18d4ba6d1ade6e9c08daac,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_e96355a5ea41d1e8a56ddce8cbb58393,
			fnFmtDateShort:        fnFmtDateShort_7666e72292ce611269924033db7087d6,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ksh_de"] = func() locales.Translator {
		return &localen{

			locale:                 "ksh_DE",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Fäb", "Mäz", "Apr", "Mai", "Jun", "Jul", "Ouj", "Säp", "Okt", "Nov", "Dez"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Jannewa", "Fäbrowa", "Määz", "Aprell", "Mai", "Juuni", "Juuli", "Oujoß", "Septämber", "Oktohber", "Novämber", "Dezämber"},
			daysAbbreviated:        []string{"Su.", "Mo.", "Di.", "Me.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"Su", "Mo", "Di", "Me", "Du", "Fr", "Sa"},
			daysWide:               []string{"Sunndaach", "Mohndaach", "Dinnsdaach", "Metwoch", "Dunnersdaach", "Friidaach", "Samsdaach"},
			periodsAbbreviated:     []string{"v.M.", "n.M."},
			periodsWide:            []string{"Uhr vörmiddaachs", "Uhr nommendaachs"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"vC", "nC"},
			erasWide:               []string{"vür Krestos", "noh Krestos"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Zentraal-Affrekaanesche Zigg", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Oß-Affrekaanesche Zigg", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Greenwich sing Standat-Zick", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Meddel-Europpa sing Summerzick", "MEZ": "Meddel-Europpa sing jewöhnlijje Zick", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oß-Europpa sing Summerzick", "OEZ": "Oß-Europpa sing jewöhnlijje Zick", "PDT": "PDT", "PST": "PST", "SAST": "Söd-Affrekaanesche Zigg", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Wäß-Affrekaanesche Sommerzigg", "WAT": "Jewöhnlijje Wäß-Affrekaanesche Zigg", "WESZ": "Weß-Europpa sing Summerzick", "WEZ": "Weß-Europpa sing jewöhnlijje Zick", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "de Azore ier Summerzick"},

			fnCardinalPluralRule:  fnCardinalPluralRule_fed1c76cd90b8905708121045ba1b87c,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_694de915ae18d4ba6d1ade6e9c08daac,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_e96355a5ea41d1e8a56ddce8cbb58393,
			fnFmtDateShort:        fnFmtDateShort_7666e72292ce611269924033db7087d6,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ku"] = func() locales.Translator {
		return &localen{

			locale:                 "ku",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "₺", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "rêb", "reş", "ada", "avr", "gul", "pûş", "tîr", "gel", "rez", "kew", "ser", "ber"},
			monthsNarrow:           []string{"", "R", "R", "A", "A", "G", "P", "T", "G", "R", "K", "S", "B"},
			monthsWide:             []string{"", "rêbendanê", "reşemiyê", "adarê", "avrêlê", "gulanê", "pûşperê", "tîrmehê", "gelawêjê", "rezberê", "kewçêrê", "sermawezê", "berfanbarê"},
			daysAbbreviated:        []string{"yş", "dş", "sş", "çş", "pş", "în", "ş"},
			daysNarrow:             []string{"Y", "D", "S", "Ç", "P", "Î", "Ş"},
			daysShort:              []string{"yş", "dş", "sş", "çş", "pş", "în", "ş"},
			daysWide:               []string{"yekşem", "duşem", "sêşem", "çarşem", "pêncşem", "în", "şemî"},
			periodsAbbreviated:     []string{"BN", "PN"},
			periodsWide:            []string{"BN", "PN"},
			erasAbbreviated:        []string{"BZ", "PZ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"berî zayînê", "piştî zayînê"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18,
			fnFmtDateLong:         fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d,
			fnFmtDateMedium:       fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1842fa1c80e067ee20bb9d4acb93d0eb,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ku_tr"] = func() locales.Translator {
		return &localen{

			locale:                 "ku_TR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "rêb", "reş", "ada", "avr", "gul", "pûş", "tîr", "gel", "rez", "kew", "ser", "ber"},
			monthsNarrow:           []string{"", "R", "R", "A", "A", "G", "P", "T", "G", "R", "K", "S", "B"},
			monthsWide:             []string{"", "rêbendanê", "reşemiyê", "adarê", "avrêlê", "gulanê", "pûşperê", "tîrmehê", "gelawêjê", "rezberê", "kewçêrê", "sermawezê", "berfanbarê"},
			daysAbbreviated:        []string{"yş", "dş", "sş", "çş", "pş", "în", "ş"},
			daysNarrow:             []string{"Y", "D", "S", "Ç", "P", "Î", "Ş"},
			daysShort:              []string{"yş", "dş", "sş", "çş", "pş", "în", "ş"},
			daysWide:               []string{"yekşem", "duşem", "sêşem", "çarşem", "pêncşem", "în", "şemî"},
			periodsAbbreviated:     []string{"BN", "PN"},
			periodsWide:            []string{"BN", "PN"},
			erasAbbreviated:        []string{"BZ", "PZ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"berî zayînê", "piştî zayînê"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18,
			fnFmtDateLong:         fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d,
			fnFmtDateMedium:       fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1842fa1c80e067ee20bb9d4acb93d0eb,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kw"] = func() locales.Translator {
		return &localen{

			locale:             "kw",
			pluralsCardinal:    []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 5, 6},
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Gen", "Hwe", "Meu", "Ebr", "Me", "Met", "Gor", "Est", "Gwn", "Hed", "Du", "Kev"},
			monthsWide:         []string{"", "mis Genver", "mis Hwevrer", "mis Meurth", "mis Ebrel", "mis Me", "mis Metheven", "mis Gortheren", "mis Est", "mis Gwynngala", "mis Hedra", "mis Du", "mis Kevardhu"},
			daysAbbreviated:    []string{"Sul", "Lun", "Mth", "Mhr", "Yow", "Gwe", "Sad"},
			daysWide:           []string{"dy Sul", "dy Lun", "dy Meurth", "dy Merher", "dy Yow", "dy Gwener", "dy Sadorn"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"RC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d9471d7d7c171fa2f8441e898fb73597,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_80a3a34cd3bb2c41e809f9ce9e28ce2a,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["kw_gb"] = func() locales.Translator {
		return &localen{

			locale:             "kw_GB",
			pluralsCardinal:    []locales.PluralRule{1, 2, 3, 4, 5, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 5, 6},
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Gen", "Hwe", "Meu", "Ebr", "Me", "Met", "Gor", "Est", "Gwn", "Hed", "Du", "Kev"},
			monthsWide:         []string{"", "mis Genver", "mis Hwevrer", "mis Meurth", "mis Ebrel", "mis Me", "mis Metheven", "mis Gortheren", "mis Est", "mis Gwynngala", "mis Hedra", "mis Du", "mis Kevardhu"},
			daysAbbreviated:    []string{"Sul", "Lun", "Mth", "Mhr", "Yow", "Gwe", "Sad"},
			daysWide:           []string{"dy Sul", "dy Lun", "dy Meurth", "dy Merher", "dy Yow", "dy Gwener", "dy Sadorn"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"RC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d9471d7d7c171fa2f8441e898fb73597,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_80a3a34cd3bb2c41e809f9ce9e28ce2a,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ky"] = func() locales.Translator {
		return &localen{

			locale:                 "ky",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "сом", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "фев.", "мар.", "апр.", "май", "июн.", "июл.", "авг.", "сен.", "окт.", "ноя.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январь", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"},
			daysAbbreviated:        []string{"жек.", "дүй.", "шейш.", "шарш.", "бейш.", "жума", "ишм."},
			daysNarrow:             []string{"Ж", "Д", "Ш", "Ш", "Б", "Ж", "И"},
			daysShort:              []string{"жш.", "дш.", "шш.", "шр.", "бш.", "жм.", "иш."},
			daysWide:               []string{"жекшемби", "дүйшөмбү", "шейшемби", "шаршемби", "бейшемби", "жума", "ишемби"},
			periodsAbbreviated:     []string{"тң", "тк"},
			periodsNarrow:          []string{"тң", "тк"},
			periodsWide:            []string{"таңкы", "түштөн кийинки"},
			erasAbbreviated:        []string{"б.з.ч.", "б.з."},
			erasNarrow:             []string{"б.з.ч.", "б.з."},
			erasWide:               []string{"биздин заманга чейин", "биздин заман"},
			timezones:              map[string]string{"ACDT": "Австралия борбордук жайкы убактысы", "ACST": "Австралия борбордук кышкы убакыты", "ACWDT": "Австралия борбордук чыгыш жайкы убактысы", "ACWST": "Австралия борбордук батыш кышкы убакыты", "ADT": "Атлантика жайкы убактысы", "AEDT": "Австралия чыгыш жайкы убактысы", "AEST": "Австралия чыгыш кышкы убакыты", "AKDT": "Аляска жайкы убактысы", "AKST": "Аляска кышкы убактысы", "ARST": "Аргентина жайкы убактысы", "ART": "Аргентина кышкы убактысы", "AST": "Атлантика кышкы убактысы", "AWDT": "Австралия батыш жайкы убактысы", "AWST": "Австралия батыш кышкы убакыты", "BOT": "Боливия убактысы", "BT": "Бутан убактысы", "CAT": "Борбордук Африка убактысы", "CDT": "Түндүк Америка, борбордук жайкы убакыт", "CHADT": "Чатам жайкы убактысы", "CHAST": "Чатам кышкы убакыт", "CLST": "Чили жайкы убактысы", "CLT": "Чили кышкы убактысы", "COST": "Колумбия жайкы убактысы", "COT": "Колумбия кышкы убактысы", "CST": "Түндүк Америка, борбордук кышкы убактысы", "ChST": "Чаморро убактысы", "EAT": "Чыгыш Африка убактысы", "ECT": "Экуадор убактысы", "EDT": "Түндүк Америка, чыгыш жайкы убактысы", "EST": "Түндүк Америка, чыгыш кышкы убактысы", "GFT": "Француз Гвиана убактысы", "GMT": "Гринвич боюнча орточо убакыт", "GST": "Булуңдун стандарттык убакыты", "GYT": "Гвиана убактысы", "HADT": "Гавайи-Алеут жайкы убактысы", "HAST": "Гавайи-Алеут кышкы убактысы", "HAT": "Нюфаундлэнд жайкы убактысы", "HECU": "Куба жайкы убактысы", "HEEG": "Чыгыш Гренландия жайкы убактысы", "HENOMX": "Түндүк-чыгыш Мексика жайкы убактысы", "HEOG": "Батыш Гренландия жайкы убактысы", "HEPM": "Сен Пьер жана Микелон жайкы убактысы", "HEPMX": "Мексика, Тынч океан жайкы убактысы", "HKST": "Гонконг жайкы убактысы", "HKT": "Гонконг кышкы убакыты", "HNCU": "Куба кышкы убактысы", "HNEG": "Чыгыш Гренландия кышкы убактысы", "HNNOMX": "Түндүк-чыгыш Мексика кышкы убактысы", "HNOG": "Батыш Гренландия кышкы убактысы", "HNPM": "Сен Пьер жана Микелон кышкы убактысы", "HNPMX": "Мексика, Тынч океан кышкы убактысы", "HNT": "Нюфаундлэнд кышкы убактысы", "IST": "Индия убактысы", "JDT": "Жапон жайкы убактысы", "JST": "Жапон кышкы убакыты", "LHDT": "Лорд Хау жайкы убактысы", "LHST": "Лорд Хау кышкы убакыты", "MDT": "Түндүк Америка, тоо жайкы убактысы", "MESZ": "Борбордук Европа жайкы убактысы", "MEZ": "Борбордук Европа кышкы убакыты", "MST": "Түндүк Америка, тоо кышкы убактысы", "MYT": "Малайзия убактысы", "NZDT": "Жаңы Зеландия жайкы убакыты", "NZST": "Жаңы Зеландия кышкы убакыты", "OESZ": "Чыгыш Европа жайкы убактысы", "OEZ": "Чыгыш Европа кышкы убакыты", "PDT": "Түндүк Америка, Тынч океан жайкы убактысы", "PST": "Түндүк Америка, Тынч океан кышкы убактысы", "SAST": "Түштүк Африка убактысы", "SGT": "Сингапур убактысы", "SRT": "Суринаме убактысы", "TMST": "Түркмөнстан жайкы убактысы", "TMT": "Түркмөнстан кышкы убакыты", "UYST": "Уругвай жайкы убактысы", "UYT": "Уругвай кышкы убактысы", "VET": "Венесуэла убактысы", "WARST": "Батыш Аргентина жайкы убактысы", "WART": "Батыш Аргентина кышкы убактысы", "WAST": "Батыш Африка жайкы убактысы", "WAT": "Батыш Африка кышкы убакыты", "WESZ": "Батыш Европа жайкы убактысы", "WEZ": "Батыш Европа кышкы убакыты", "WIB": "Батыш Индонезия убактысы", "WIT": "Чыгыш Индонезия убактысы", "WITA": "Борбордук Индонезия убактысы", "∅∅∅": "Перу жайкы убактысы"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_2b8291eec4b5351a149455e035ae0af2,
			fnFmtDateLong:         fnFmtDateLong_07de017bbe01706764e3fbed80282c3a,
			fnFmtDateMedium:       fnFmtDateMedium_3f222d37600871207c139eeba9ae1a98,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ky_kg"] = func() locales.Translator {
		return &localen{

			locale:                 "ky_KG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "фев.", "мар.", "апр.", "май", "июн.", "июл.", "авг.", "сен.", "окт.", "ноя.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январь", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"},
			daysAbbreviated:        []string{"жек.", "дүй.", "шейш.", "шарш.", "бейш.", "жума", "ишм."},
			daysNarrow:             []string{"Ж", "Д", "Ш", "Ш", "Б", "Ж", "И"},
			daysShort:              []string{"жш.", "дш.", "шш.", "шр.", "бш.", "жм.", "иш."},
			daysWide:               []string{"жекшемби", "дүйшөмбү", "шейшемби", "шаршемби", "бейшемби", "жума", "ишемби"},
			periodsAbbreviated:     []string{"тң", "тк"},
			periodsNarrow:          []string{"тң", "тк"},
			periodsWide:            []string{"таңкы", "түштөн кийинки"},
			erasAbbreviated:        []string{"б.з.ч.", "б.з."},
			erasNarrow:             []string{"б.з.ч.", "б.з."},
			erasWide:               []string{"биздин заманга чейин", "биздин заман"},
			timezones:              map[string]string{"ACDT": "Австралия борбордук жайкы убактысы", "ACST": "Австралия борбордук кышкы убакыты", "ACWDT": "Австралия борбордук чыгыш жайкы убактысы", "ACWST": "Австралия борбордук батыш кышкы убакыты", "ADT": "Атлантика жайкы убактысы", "AEDT": "Австралия чыгыш жайкы убактысы", "AEST": "Австралия чыгыш кышкы убакыты", "AKDT": "Аляска жайкы убактысы", "AKST": "Аляска кышкы убактысы", "ARST": "Аргентина жайкы убактысы", "ART": "Аргентина кышкы убактысы", "AST": "Атлантика кышкы убактысы", "AWDT": "Австралия батыш жайкы убактысы", "AWST": "Австралия батыш кышкы убакыты", "BOT": "Боливия убактысы", "BT": "Бутан убактысы", "CAT": "Борбордук Африка убактысы", "CDT": "Түндүк Америка, борбордук жайкы убакыт", "CHADT": "Чатам жайкы убактысы", "CHAST": "Чатам кышкы убакыт", "CLST": "Чили жайкы убактысы", "CLT": "Чили кышкы убактысы", "COST": "Колумбия жайкы убактысы", "COT": "Колумбия кышкы убактысы", "CST": "Түндүк Америка, борбордук кышкы убактысы", "ChST": "Чаморро убактысы", "EAT": "Чыгыш Африка убактысы", "ECT": "Экуадор убактысы", "EDT": "Түндүк Америка, чыгыш жайкы убактысы", "EST": "Түндүк Америка, чыгыш кышкы убактысы", "GFT": "Француз Гвиана убактысы", "GMT": "Гринвич боюнча орточо убакыт", "GST": "Булуңдун стандарттык убакыты", "GYT": "Гвиана убактысы", "HADT": "Гавайи-Алеут жайкы убактысы", "HAST": "Гавайи-Алеут кышкы убактысы", "HAT": "Нюфаундлэнд жайкы убактысы", "HECU": "Куба жайкы убактысы", "HEEG": "Чыгыш Гренландия жайкы убактысы", "HENOMX": "Түндүк-чыгыш Мексика жайкы убактысы", "HEOG": "Батыш Гренландия жайкы убактысы", "HEPM": "Сен Пьер жана Микелон жайкы убактысы", "HEPMX": "Мексика, Тынч океан жайкы убактысы", "HKST": "Гонконг жайкы убактысы", "HKT": "Гонконг кышкы убакыты", "HNCU": "Куба кышкы убактысы", "HNEG": "Чыгыш Гренландия кышкы убактысы", "HNNOMX": "Түндүк-чыгыш Мексика кышкы убактысы", "HNOG": "Батыш Гренландия кышкы убактысы", "HNPM": "Сен Пьер жана Микелон кышкы убактысы", "HNPMX": "Мексика, Тынч океан кышкы убактысы", "HNT": "Нюфаундлэнд кышкы убактысы", "IST": "Индия убактысы", "JDT": "Жапон жайкы убактысы", "JST": "Жапон кышкы убакыты", "LHDT": "Лорд Хау жайкы убактысы", "LHST": "Лорд Хау кышкы убакыты", "MDT": "Түндүк Америка, тоо жайкы убактысы", "MESZ": "Борбордук Европа жайкы убактысы", "MEZ": "Борбордук Европа кышкы убакыты", "MST": "Түндүк Америка, тоо кышкы убактысы", "MYT": "Малайзия убактысы", "NZDT": "Жаңы Зеландия жайкы убакыты", "NZST": "Жаңы Зеландия кышкы убакыты", "OESZ": "Чыгыш Европа жайкы убактысы", "OEZ": "Чыгыш Европа кышкы убакыты", "PDT": "Түндүк Америка, Тынч океан жайкы убактысы", "PST": "Түндүк Америка, Тынч океан кышкы убактысы", "SAST": "Түштүк Африка убактысы", "SGT": "Сингапур убактысы", "SRT": "Суринаме убактысы", "TMST": "Түркмөнстан жайкы убактысы", "TMT": "Түркмөнстан кышкы убакыты", "UYST": "Уругвай жайкы убактысы", "UYT": "Уругвай кышкы убактысы", "VET": "Венесуэла убактысы", "WARST": "Батыш Аргентина жайкы убактысы", "WART": "Батыш Аргентина кышкы убактысы", "WAST": "Батыш Африка жайкы убактысы", "WAT": "Батыш Африка кышкы убакыты", "WESZ": "Батыш Европа жайкы убактысы", "WEZ": "Батыш Европа кышкы убакыты", "WIB": "Батыш Индонезия убактысы", "WIT": "Чыгыш Индонезия убактысы", "WITA": "Борбордук Индонезия убактысы", "∅∅∅": "Перу жайкы убактысы"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_2b8291eec4b5351a149455e035ae0af2,
			fnFmtDateLong:         fnFmtDateLong_07de017bbe01706764e3fbed80282c3a,
			fnFmtDateMedium:       fnFmtDateMedium_3f222d37600871207c139eeba9ae1a98,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lag"] = func() locales.Translator {
		return &localen{

			locale:                 "lag",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "Fúngatɨ", "Naanɨ", "Keenda", "Ikúmi", "Inyambala", "Idwaata", "Mʉʉnchɨ", "Vɨɨrɨ", "Saatʉ", "Inyi", "Saano", "Sasatʉ"},
			monthsNarrow:           []string{"", "F", "N", "K", "I", "I", "I", "M", "V", "S", "I", "S", "S"},
			monthsWide:             []string{"", "Kʉfúngatɨ", "Kʉnaanɨ", "Kʉkeenda", "Kwiikumi", "Kwiinyambála", "Kwiidwaata", "Kʉmʉʉnchɨ", "Kʉvɨɨrɨ", "Kʉsaatʉ", "Kwiinyi", "Kʉsaano", "Kʉsasatʉ"},
			daysAbbreviated:        []string{"Píili", "Táatu", "Íne", "Táano", "Alh", "Ijm", "Móosi"},
			daysNarrow:             []string{"P", "T", "E", "O", "A", "I", "M"},
			daysWide:               []string{"Jumapíiri", "Jumatátu", "Jumaíne", "Jumatáano", "Alamíisi", "Ijumáa", "Jumamóosi"},
			periodsAbbreviated:     []string{"TOO", "MUU"},
			periodsWide:            []string{"TOO", "MUU"},
			erasAbbreviated:        []string{"KSA", "KA"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kɨrɨsitʉ sɨ anavyaal", "Kɨrɨsitʉ akavyaalwe"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_7664ec40507fc37b05985577fc0c7d99,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lag_tz"] = func() locales.Translator {
		return &localen{

			locale:                 "lag_TZ",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "Fúngatɨ", "Naanɨ", "Keenda", "Ikúmi", "Inyambala", "Idwaata", "Mʉʉnchɨ", "Vɨɨrɨ", "Saatʉ", "Inyi", "Saano", "Sasatʉ"},
			monthsNarrow:           []string{"", "F", "N", "K", "I", "I", "I", "M", "V", "S", "I", "S", "S"},
			monthsWide:             []string{"", "Kʉfúngatɨ", "Kʉnaanɨ", "Kʉkeenda", "Kwiikumi", "Kwiinyambála", "Kwiidwaata", "Kʉmʉʉnchɨ", "Kʉvɨɨrɨ", "Kʉsaatʉ", "Kwiinyi", "Kʉsaano", "Kʉsasatʉ"},
			daysAbbreviated:        []string{"Píili", "Táatu", "Íne", "Táano", "Alh", "Ijm", "Móosi"},
			daysNarrow:             []string{"P", "T", "E", "O", "A", "I", "M"},
			daysWide:               []string{"Jumapíiri", "Jumatátu", "Jumaíne", "Jumatáano", "Alamíisi", "Ijumáa", "Jumamóosi"},
			periodsAbbreviated:     []string{"TOO", "MUU"},
			periodsWide:            []string{"TOO", "MUU"},
			erasAbbreviated:        []string{"KSA", "KA"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kɨrɨsitʉ sɨ anavyaal", "Kɨrɨsitʉ akavyaalwe"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_7664ec40507fc37b05985577fc0c7d99,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lb"] = func() locales.Translator {
		return &localen{

			locale:                 "lb",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "öS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "Mäe.", "Abr.", "Mee", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "Mäerz", "Abrëll", "Mee", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"Son.", "Méi.", "Dën.", "Mët.", "Don.", "Fre.", "Sam."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mé.", "Dë.", "Më.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonndeg", "Méindeg", "Dënschdeg", "Mëttwoch", "Donneschdeg", "Freideg", "Samschdeg"},
			periodsAbbreviated:     []string{"moies", "nomëttes"},
			periodsNarrow:          []string{"mo.", "nomë."},
			periodsWide:            []string{"moies", "nomëttes"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralesch Summerzäit", "ACST": "Zentralaustralesch Normalzäit", "ACWDT": "Zentral-/Westaustralesch Summerzäit", "ACWST": "Zentral-/Westaustralesch Normalzäit", "ADT": "Atlantik-Summerzäit", "AEDT": "Ostaustralesch Summerzäit", "AEST": "Ostaustralesch Normalzäit", "AKDT": "Alaska-Summerzäit", "AKST": "Alaska-Normalzäit", "ARST": "Argentinesch Summerzäit", "ART": "Argentinesch Normalzäit", "AST": "Atlantik-Normalzäit", "AWDT": "Westaustralesch Summerzäit", "AWST": "Westaustralesch Normalzäit", "BOT": "Bolivianesch Zäit", "BT": "Bhutan-Zäit", "CAT": "Zentralafrikanesch Zäit", "CDT": "Nordamerikanesch Inland-Summerzäit", "CHADT": "Chatham-Summerzäit", "CHAST": "Chatham-Normalzäit", "CLST": "Chilenesch Summerzäit", "CLT": "Chilenesch Normalzäit", "COST": "Kolumbianesch Summerzäit", "COT": "Kolumbianesch Normalzäit", "CST": "Nordamerikanesch Inland-Normalzäit", "ChST": "Chamorro-Zäit", "EAT": "Ostafrikanesch Zäit", "ECT": "Ecuadorianesch Zäit", "EDT": "Nordamerikanesch Ostküsten-Summerzäit", "EST": "Nordamerikanesch Ostküsten-Normalzäit", "GFT": "Franséisch-Guayane-Zäit", "GMT": "Mëttler Greenwich-Zäit", "GST": "Golf-Zäit", "GYT": "Guyana-Zäit", "HADT": "Hawaii-Aleuten-Summerzäit", "HAST": "Hawaii-Aleuten-Normalzäit", "HAT": "Neifundland-Summerzäit", "HECU": "Kubanesch Summerzäit", "HEEG": "Ostgrönland-Summerzäit", "HENOMX": "Nordwest-Mexiko-Summerzäit", "HEOG": "Westgrönland-Summerzäit", "HEPM": "Saint-Pierre-a-Miquelon-Summerzäit", "HEPMX": "Mexikanesch Pazifik-Summerzäit", "HKST": "Hong-Kong-Summerzäit", "HKT": "Hong-Kong-Normalzäit", "HNCU": "Kubanesch Normalzäit", "HNEG": "Ostgrönland-Normalzäit", "HNNOMX": "Nordwest-Mexiko-Normalzäit", "HNOG": "Westgrönland-Normalzäit", "HNPM": "Saint-Pierre-a-Miquelon-Normalzäit", "HNPMX": "Mexikanesch Pazifik-Normalzäit", "HNT": "Neifundland-Normalzäit", "IST": "Indesch Zäit", "JDT": "Japanesch Summerzäit", "JST": "Japanesch Normalzäit", "LHDT": "Lord-Howe-Summerzäit", "LHST": "Lord-Howe-Normalzäit", "MDT": "Rocky-Mountain-Summerzäit", "MESZ": "Mëtteleuropäesch Summerzäit", "MEZ": "Mëtteleuropäesch Normalzäit", "MST": "Rocky-Mountain-Normalzäit", "MYT": "Malaysesch Zäit", "NZDT": "Neiséiland-Summerzäit", "NZST": "Neiséiland-Normalzäit", "OESZ": "Osteuropäesch Summerzäit", "OEZ": "Osteuropäesch Normalzäit", "PDT": "Nordamerikanesch Westküsten-Summerzäit", "PST": "Nordamerikanesch Westküsten-Normalzäit", "SAST": "Südafrikanesch Zäit", "SGT": "Singapur-Standardzäit", "SRT": "Suriname-Zäit", "TMST": "Turkmenistan-Summerzäit", "TMT": "Turkmenistan-Normalzäit", "UYST": "Uruguayanesch Summerzäit", "UYT": "Uruguyanesch Normalzäit", "VET": "Venezuela-Zäit", "WARST": "Westargentinesch Summerzäit", "WART": "Westargentinesch Normalzäit", "WAST": "Westafrikanesch Summerzäit", "WAT": "Westafrikanesch Normalzäit", "WESZ": "Westeuropäesch Summerzäit", "WEZ": "Westeuropäesch Normalzäit", "WIB": "Westindonesesch Zäit", "WIT": "Ostindonesesch Zäit", "WITA": "Zentralindonesesch Zäit", "∅∅∅": "Brasília-Summerzäit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lb_lu"] = func() locales.Translator {
		return &localen{

			locale:                 "lb_LU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "Mäe.", "Abr.", "Mee", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januar", "Februar", "Mäerz", "Abrëll", "Mee", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"},
			daysAbbreviated:        []string{"Son.", "Méi.", "Dën.", "Mët.", "Don.", "Fre.", "Sam."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysShort:              []string{"So.", "Mé.", "Dë.", "Më.", "Do.", "Fr.", "Sa."},
			daysWide:               []string{"Sonndeg", "Méindeg", "Dënschdeg", "Mëttwoch", "Donneschdeg", "Freideg", "Samschdeg"},
			periodsAbbreviated:     []string{"moies", "nomëttes"},
			periodsNarrow:          []string{"mo.", "nomë."},
			periodsWide:            []string{"moies", "nomëttes"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"v. Chr.", "n. Chr."},
			timezones:              map[string]string{"ACDT": "Zentralaustralesch Summerzäit", "ACST": "Zentralaustralesch Normalzäit", "ACWDT": "Zentral-/Westaustralesch Summerzäit", "ACWST": "Zentral-/Westaustralesch Normalzäit", "ADT": "Atlantik-Summerzäit", "AEDT": "Ostaustralesch Summerzäit", "AEST": "Ostaustralesch Normalzäit", "AKDT": "Alaska-Summerzäit", "AKST": "Alaska-Normalzäit", "ARST": "Argentinesch Summerzäit", "ART": "Argentinesch Normalzäit", "AST": "Atlantik-Normalzäit", "AWDT": "Westaustralesch Summerzäit", "AWST": "Westaustralesch Normalzäit", "BOT": "Bolivianesch Zäit", "BT": "Bhutan-Zäit", "CAT": "Zentralafrikanesch Zäit", "CDT": "Nordamerikanesch Inland-Summerzäit", "CHADT": "Chatham-Summerzäit", "CHAST": "Chatham-Normalzäit", "CLST": "Chilenesch Summerzäit", "CLT": "Chilenesch Normalzäit", "COST": "Kolumbianesch Summerzäit", "COT": "Kolumbianesch Normalzäit", "CST": "Nordamerikanesch Inland-Normalzäit", "ChST": "Chamorro-Zäit", "EAT": "Ostafrikanesch Zäit", "ECT": "Ecuadorianesch Zäit", "EDT": "Nordamerikanesch Ostküsten-Summerzäit", "EST": "Nordamerikanesch Ostküsten-Normalzäit", "GFT": "Franséisch-Guayane-Zäit", "GMT": "Mëttler Greenwich-Zäit", "GST": "Golf-Zäit", "GYT": "Guyana-Zäit", "HADT": "Hawaii-Aleuten-Summerzäit", "HAST": "Hawaii-Aleuten-Normalzäit", "HAT": "Neifundland-Summerzäit", "HECU": "Kubanesch Summerzäit", "HEEG": "Ostgrönland-Summerzäit", "HENOMX": "Nordwest-Mexiko-Summerzäit", "HEOG": "Westgrönland-Summerzäit", "HEPM": "Saint-Pierre-a-Miquelon-Summerzäit", "HEPMX": "Mexikanesch Pazifik-Summerzäit", "HKST": "Hong-Kong-Summerzäit", "HKT": "Hong-Kong-Normalzäit", "HNCU": "Kubanesch Normalzäit", "HNEG": "Ostgrönland-Normalzäit", "HNNOMX": "Nordwest-Mexiko-Normalzäit", "HNOG": "Westgrönland-Normalzäit", "HNPM": "Saint-Pierre-a-Miquelon-Normalzäit", "HNPMX": "Mexikanesch Pazifik-Normalzäit", "HNT": "Neifundland-Normalzäit", "IST": "Indesch Zäit", "JDT": "Japanesch Summerzäit", "JST": "Japanesch Normalzäit", "LHDT": "Lord-Howe-Summerzäit", "LHST": "Lord-Howe-Normalzäit", "MDT": "Rocky-Mountain-Summerzäit", "MESZ": "Mëtteleuropäesch Summerzäit", "MEZ": "Mëtteleuropäesch Normalzäit", "MST": "Rocky-Mountain-Normalzäit", "MYT": "Malaysesch Zäit", "NZDT": "Neiséiland-Summerzäit", "NZST": "Neiséiland-Normalzäit", "OESZ": "Osteuropäesch Summerzäit", "OEZ": "Osteuropäesch Normalzäit", "PDT": "Nordamerikanesch Westküsten-Summerzäit", "PST": "Nordamerikanesch Westküsten-Normalzäit", "SAST": "Südafrikanesch Zäit", "SGT": "Singapur-Standardzäit", "SRT": "Suriname-Zäit", "TMST": "Turkmenistan-Summerzäit", "TMT": "Turkmenistan-Normalzäit", "UYST": "Uruguayanesch Summerzäit", "UYT": "Uruguyanesch Normalzäit", "VET": "Venezuela-Zäit", "WARST": "Westargentinesch Summerzäit", "WART": "Westargentinesch Normalzäit", "WAST": "Westafrikanesch Summerzäit", "WAT": "Westafrikanesch Normalzäit", "WESZ": "Westeuropäesch Summerzäit", "WEZ": "Westeuropäesch Normalzäit", "WIB": "Westindonesesch Zäit", "WIT": "Ostindonesesch Zäit", "WITA": "Zentralindonesesch Zäit", "∅∅∅": "Azoren-Summerzäit"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lg"] = func() locales.Translator {
		return &localen{

			locale:            "lg",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "Jan", "Feb", "Mar", "Apu", "Maa", "Juu", "Jul", "Agu", "Seb", "Oki", "Nov", "Des"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Janwaliyo", "Febwaliyo", "Marisi", "Apuli", "Maayi", "Juuni", "Julaayi", "Agusito", "Sebuttemba", "Okitobba", "Novemba", "Desemba"},
			daysAbbreviated:   []string{"Sab", "Bal", "Lw2", "Lw3", "Lw4", "Lw5", "Lw6"},
			daysNarrow:        []string{"S", "B", "L", "L", "L", "L", "L"},
			daysWide:          []string{"Sabbiiti", "Balaza", "Lwakubiri", "Lwakusatu", "Lwakuna", "Lwakutaano", "Lwamukaaga"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Kulisito nga tannaza", "Bukya Kulisito Azaal"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lg_ug"] = func() locales.Translator {
		return &localen{

			locale:            "lg_UG",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "Jan", "Feb", "Mar", "Apu", "Maa", "Juu", "Jul", "Agu", "Seb", "Oki", "Nov", "Des"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Janwaliyo", "Febwaliyo", "Marisi", "Apuli", "Maayi", "Juuni", "Julaayi", "Agusito", "Sebuttemba", "Okitobba", "Novemba", "Desemba"},
			daysAbbreviated:   []string{"Sab", "Bal", "Lw2", "Lw3", "Lw4", "Lw5", "Lw6"},
			daysNarrow:        []string{"S", "B", "L", "L", "L", "L", "L"},
			daysWide:          []string{"Sabbiiti", "Balaza", "Lwakubiri", "Lwakusatu", "Lwakuna", "Lwakutaano", "Lwamukaaga"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Kulisito nga tannaza", "Bukya Kulisito Azaal"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lkt"] = func() locales.Translator {
		return &localen{

			locale:                 "lkt",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsWide:             []string{"", "Wiótheȟika Wí", "Thiyóȟeyuŋka Wí", "Ištáwičhayazaŋ Wí", "Pȟežítȟo Wí", "Čhaŋwápetȟo Wí", "Wípazukȟa-wašté Wí", "Čhaŋpȟásapa Wí", "Wasútȟuŋ Wí", "Čhaŋwápeǧi Wí", "Čhaŋwápe-kasná Wí", "Waníyetu Wí", "Tȟahékapšuŋ Wí"},
			daysNarrow:             []string{"A", "W", "N", "Y", "T", "Z", "O"},
			daysWide:               []string{"Aŋpétuwakȟaŋ", "Aŋpétuwaŋži", "Aŋpétunuŋpa", "Aŋpétuyamni", "Aŋpétutopa", "Aŋpétuzaptaŋ", "Owáŋgyužažapi"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d0ef2c4a76852adc59df517a99a6000f,
			fnFmtCurrency:         fnFmtCurrency_9b06328888547fc535e862da8ee049d8,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lkt_us"] = func() locales.Translator {
		return &localen{

			locale:                 "lkt_US",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsWide:             []string{"", "Wiótheȟika Wí", "Thiyóȟeyuŋka Wí", "Ištáwičhayazaŋ Wí", "Pȟežítȟo Wí", "Čhaŋwápetȟo Wí", "Wípazukȟa-wašté Wí", "Čhaŋpȟásapa Wí", "Wasútȟuŋ Wí", "Čhaŋwápeǧi Wí", "Čhaŋwápe-kasná Wí", "Waníyetu Wí", "Tȟahékapšuŋ Wí"},
			daysNarrow:             []string{"A", "W", "N", "Y", "T", "Z", "O"},
			daysWide:               []string{"Aŋpétuwakȟaŋ", "Aŋpétuwaŋži", "Aŋpétunuŋpa", "Aŋpétuyamni", "Aŋpétutopa", "Aŋpétuzaptaŋ", "Owáŋgyužažapi"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d0ef2c4a76852adc59df517a99a6000f,
			fnFmtCurrency:         fnFmtCurrency_9b06328888547fc535e862da8ee049d8,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ln"] = func() locales.Translator {
		return &localen{

			locale:                 "ln",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "Kz", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "FC", "CHE", "Fr.", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "T$", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "F CFP", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"},
			monthsNarrow:           []string{"", "y", "f", "m", "a", "m", "y", "y", "a", "s", "ɔ", "n", "d"},
			monthsWide:             []string{"", "sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ̌kɔ́", "sánzá ya zómi na míbalé"},
			daysAbbreviated:        []string{"eye", "ybo", "mbl", "mst", "min", "mtn", "mps"},
			daysNarrow:             []string{"e", "y", "m", "m", "m", "m", "p"},
			daysWide:               []string{"eyenga", "mokɔlɔ mwa yambo", "mokɔlɔ mwa míbalé", "mokɔlɔ mwa mísáto", "mokɔlɔ ya mínéi", "mokɔlɔ ya mítáno", "mpɔ́sɔ"},
			periodsAbbreviated:     []string{"ntɔ́ngɔ́", "mpókwa"},
			periodsWide:            []string{"ntɔ́ngɔ́", "mpókwa"},
			erasAbbreviated:        []string{"libóso ya", "nsima ya Y"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yambo ya Yézu Krís", "Nsima ya Yézu Krís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Ntángo ya Lubumbashi", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Ntángo ya Afríka ya Ɛ́sita", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Ntángo ya Londoni", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "Ntángo ya Afríka ya Sidi", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ln_ao"] = func() locales.Translator {
		return &localen{

			locale:                 "ln_AO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "Kz", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"},
			monthsNarrow:           []string{"", "y", "f", "m", "a", "m", "y", "y", "a", "s", "ɔ", "n", "d"},
			monthsWide:             []string{"", "sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ̌kɔ́", "sánzá ya zómi na míbalé"},
			daysAbbreviated:        []string{"eye", "ybo", "mbl", "mst", "min", "mtn", "mps"},
			daysNarrow:             []string{"e", "y", "m", "m", "m", "m", "p"},
			daysWide:               []string{"eyenga", "mokɔlɔ mwa yambo", "mokɔlɔ mwa míbalé", "mokɔlɔ mwa mísáto", "mokɔlɔ ya mínéi", "mokɔlɔ ya mítáno", "mpɔ́sɔ"},
			periodsAbbreviated:     []string{"ntɔ́ngɔ́", "mpókwa"},
			periodsWide:            []string{"ntɔ́ngɔ́", "mpókwa"},
			erasAbbreviated:        []string{"libóso ya", "nsima ya Y"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yambo ya Yézu Krís", "Nsima ya Yézu Krís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Ntángo ya Lubumbashi", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Ntángo ya Afríka ya Ɛ́sita", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Ntángo ya Londoni", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "Ntángo ya Afríka ya Sidi", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ln_cd"] = func() locales.Translator {
		return &localen{

			locale:                 "ln_CD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"},
			monthsNarrow:           []string{"", "y", "f", "m", "a", "m", "y", "y", "a", "s", "ɔ", "n", "d"},
			monthsWide:             []string{"", "sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ̌kɔ́", "sánzá ya zómi na míbalé"},
			daysAbbreviated:        []string{"eye", "ybo", "mbl", "mst", "min", "mtn", "mps"},
			daysNarrow:             []string{"e", "y", "m", "m", "m", "m", "p"},
			daysWide:               []string{"eyenga", "mokɔlɔ mwa yambo", "mokɔlɔ mwa míbalé", "mokɔlɔ mwa mísáto", "mokɔlɔ ya mínéi", "mokɔlɔ ya mítáno", "mpɔ́sɔ"},
			periodsAbbreviated:     []string{"ntɔ́ngɔ́", "mpókwa"},
			periodsWide:            []string{"ntɔ́ngɔ́", "mpókwa"},
			erasAbbreviated:        []string{"libóso ya", "nsima ya Y"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yambo ya Yézu Krís", "Nsima ya Yézu Krís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Ntángo ya Lubumbashi", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Ntángo ya Afríka ya Ɛ́sita", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Ntángo ya Londoni", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "Ntángo ya Afríka ya Sidi", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ln_cf"] = func() locales.Translator {
		return &localen{

			locale:                 "ln_CF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"},
			monthsNarrow:           []string{"", "y", "f", "m", "a", "m", "y", "y", "a", "s", "ɔ", "n", "d"},
			monthsWide:             []string{"", "sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ̌kɔ́", "sánzá ya zómi na míbalé"},
			daysAbbreviated:        []string{"eye", "ybo", "mbl", "mst", "min", "mtn", "mps"},
			daysNarrow:             []string{"e", "y", "m", "m", "m", "m", "p"},
			daysWide:               []string{"eyenga", "mokɔlɔ mwa yambo", "mokɔlɔ mwa míbalé", "mokɔlɔ mwa mísáto", "mokɔlɔ ya mínéi", "mokɔlɔ ya mítáno", "mpɔ́sɔ"},
			periodsAbbreviated:     []string{"ntɔ́ngɔ́", "mpókwa"},
			periodsWide:            []string{"ntɔ́ngɔ́", "mpókwa"},
			erasAbbreviated:        []string{"libóso ya", "nsima ya Y"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yambo ya Yézu Krís", "Nsima ya Yézu Krís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Ntángo ya Lubumbashi", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Ntángo ya Afríka ya Ɛ́sita", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Ntángo ya Londoni", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "Ntángo ya Afríka ya Sidi", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ln_cg"] = func() locales.Translator {
		return &localen{

			locale:                 "ln_CG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"},
			monthsNarrow:           []string{"", "y", "f", "m", "a", "m", "y", "y", "a", "s", "ɔ", "n", "d"},
			monthsWide:             []string{"", "sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ̌kɔ́", "sánzá ya zómi na míbalé"},
			daysAbbreviated:        []string{"eye", "ybo", "mbl", "mst", "min", "mtn", "mps"},
			daysNarrow:             []string{"e", "y", "m", "m", "m", "m", "p"},
			daysWide:               []string{"eyenga", "mokɔlɔ mwa yambo", "mokɔlɔ mwa míbalé", "mokɔlɔ mwa mísáto", "mokɔlɔ ya mínéi", "mokɔlɔ ya mítáno", "mpɔ́sɔ"},
			periodsAbbreviated:     []string{"ntɔ́ngɔ́", "mpókwa"},
			periodsWide:            []string{"ntɔ́ngɔ́", "mpókwa"},
			erasAbbreviated:        []string{"libóso ya", "nsima ya Y"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Yambo ya Yézu Krís", "Nsima ya Yézu Krís"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "Ntángo ya Lubumbashi", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "Ntángo ya Afríka ya Ɛ́sita", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Ntángo ya Londoni", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "Ntángo ya Afríka ya Sidi", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lo"] = func() locales.Translator {
		return &localen{

			locale:                 "lo",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "₭", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "-",
			monthsAbbreviated:      []string{"", "ມ.ກ.", "ກ.ພ.", "ມ.ນ.", "ມ.ສ.", "ພ.ພ.", "ມິ.ຖ.", "ກ.ລ.", "ສ.ຫ.", "ກ.ຍ.", "ຕ.ລ.", "ພ.ຈ.", "ທ.ວ."},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ມັງກອນ", "ກຸມພາ", "ມີນາ", "ເມສາ", "ພຶດສະພາ", "ມິຖຸນາ", "ກໍລະກົດ", "ສິງຫາ", "ກັນຍາ", "ຕຸລາ", "ພະຈິກ", "ທັນວາ"},
			daysAbbreviated:        []string{"ອາທິດ", "ຈັນ", "ອັງຄານ", "ພຸດ", "ພະຫັດ", "ສຸກ", "ເສົາ"},
			daysNarrow:             []string{"ອາ", "ຈ", "ອ", "ພ", "ພຫ", "ສຸ", "ສ"},
			daysShort:              []string{"ອາ.", "ຈ.", "ອ.", "ພ.", "ພຫ.", "ສຸ.", "ສ."},
			daysWide:               []string{"ວັນອາທິດ", "ວັນຈັນ", "ວັນອັງຄານ", "ວັນພຸດ", "ວັນພະຫັດ", "ວັນສຸກ", "ວັນເສົາ"},
			periodsAbbreviated:     []string{"ກ່ອນທ່ຽງ", "ຫຼັງທ່ຽງ"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"ກ່ອນທ່ຽງ", "ຫຼັງທ່ຽງ"},
			erasAbbreviated:        []string{"ກ່ອນ ຄ.ສ.", "ຄ.ສ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ກ່ອນຄຣິດສັກກະລາດ", "ຄຣິດສັກກະລາດ"},
			timezones:              map[string]string{"ACDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bເລຍ\u200bກາງ", "ACST": "ເວ\u200bລາມາດ\u200bຕະ\u200bຖານອອ\u200bສ\u200bເຕຣ\u200bເລຍ\u200bກ\u200bາງ", "ACWDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bກາງ\u200bຕາ\u200bເວັນ\u200bຕົກ", "ACWST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານອອສ\u200bເຕຣ\u200bລຽນ\u200bກາງ\u200bຕາ\u200bເວັນ\u200bຕົກ", "ADT": "ເວລາກາງເວັນຂອງອາແລນຕິກ", "AEDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bອອກ", "AEST": "ເວ\u200bລາ\u200bມາດຕະຖານ\u200b\u200b\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bອອກ", "AKDT": "ເວລາກາງເວັນອະລັສກາ", "AKST": "ເວລາມາດຕະຖານອະລັສກາ", "ARST": "\u200bເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bອາ\u200bເຈນ\u200bທິ\u200bນາ", "ART": "\u200bເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານອາ\u200bເຈນ\u200bທິ\u200bນາ", "AST": "ເວລາມາດຕະຖານຂອງອາແລນຕິກ", "AWDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bຕົກ", "AWST": "ເວ\u200bລາ\u200bມາ\u200bດ\u200bຕະ\u200bຖານອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bຕົກ", "BOT": "ເວ\u200bລາ\u200bໂບ\u200bລິ\u200bເວຍ", "BT": "ເວ\u200bລາ\u200bພູ\u200bຖານ", "CAT": "ເວ\u200bລາ\u200bອາ\u200bຟຣິ\u200bກາ\u200bກາງ", "CDT": "ເວລາກາງເວັນກາງ", "CHADT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bຊາ\u200bທາມ", "CHAST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຊາ\u200bທາມ", "CLST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນຊິ\u200bລີ", "CLT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານຊິ\u200bລີ", "COST": "ເວລາລະດູຮ້ອນໂຄລໍາເບຍ", "COT": "ເວລາມາດຕະຖານໂຄລຳເບຍ", "CST": "ເວລາມາດຕະຖານກາງ", "ChST": "ເວ\u200bລາ\u200bຈາ\u200bໂມ\u200bໂຣ", "EAT": "ເວ\u200bລາ\u200bອາ\u200bຟຣິ\u200bກາ\u200bຕາ\u200bເວັນ\u200bອອກ", "ECT": "ເວ\u200bລາ\u200bເອ\u200bກົວ\u200bດໍ", "EDT": "ເວລາກາງເວັນຕາເວັນອອກ", "EST": "ເວລາມາດຕະຖານຕາເວັນອອກ", "GFT": "ເວ\u200bລາ\u200bເຟ\u200bຣນ\u200bຊ໌\u200bເກຍ\u200bນາ", "GMT": "ເວ\u200bລາກຣີນ\u200bວິ\u200bຊ", "GST": "ເວ\u200bລາ\u200bກູ\u200bລ\u200b໌ຟ", "GYT": "ເວລາກາຍອານາ", "HADT": "ເວລາຕອນທ່ຽງຮາວາຍ-ເອລູທຽນ", "HAST": "ເວລາມາດຕະຖານຮາວາຍ-ເອລູທຽນ", "HAT": "ເວລາກາງເວັນນິວຟາວແລນ", "HECU": "ເວລາກາງເວັນຄິວບາ", "HEEG": "ເວລາລະດູຮ້ອນກຣີນແລນຕາເວັນອອກ", "HENOMX": "ເວລາກາງເວັນເມັກຊິກັນນອດເວສ", "HEOG": "ເວລາຕອນທ່ຽງກຣີນແລນຕາເວັນຕົກ", "HEPM": "\u200bເວ\u200bລາຕອນ\u200bທ່ຽງເຊນ\u200bປີ\u200bແອ ແລະ\u200bມິ\u200bກົວ\u200bລອນ", "HEPMX": "ເວລາກາງເວັນແປຊິຟິກເມັກຊິກັນ", "HKST": "\u200bເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bຮອງ\u200bກົງ", "HKT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຮອງ\u200bກົງ", "HNCU": "ເວລາມາດຕະຖານຂອງຄິວບາ", "HNEG": "ເວລາມາດຕະຖານຕາເວັນອອກກຣີນແລນ", "HNNOMX": "\u200bເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານນອດ\u200bເວ\u200bສ\u200bເມັກ\u200bຊິ\u200bໂກ", "HNOG": "ເວລາມາດຕະຖານກຣີນແລນຕາເວັນຕົກ", "HNPM": "\u200bເວ\u200bລາມາດ\u200bຕະ\u200bຖານເຊນ\u200bປີ\u200bແອ ແລະ\u200bມິ\u200bກົວ\u200bລອນ", "HNPMX": "ເວລາມາດຕະຖານແປຊິຟິກເມັກຊິກັນ", "HNT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bນິວ\u200bຟາວ\u200bແລນ", "IST": "ເວລາ ອິນເດຍ", "JDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bຍີ່\u200bປຸ່ນ", "JST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຍີ່\u200bປຸ່ນ", "LHDT": "\u200bເວ\u200bລ\u200bສາ\u200bຕອນ\u200b\u200bທ່ຽງ\u200bລອດ\u200bເຮົາ\u200b", "LHST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bລອດ\u200bເຮົາ", "MDT": "ເວລາກາງເວັນແຖບພູເຂົາ", "MESZ": "\u200bເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bຢູ\u200bໂຣບ\u200bກາງ", "MEZ": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຢູ\u200bໂຣບກາງ", "MST": "ເວລາມາດຕະຖານແຖບພູເຂົາ", "MYT": "ເວ\u200bລາ\u200bມາ\u200bເລ\u200bເຊຍ", "NZDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bນິວ\u200bຊີ\u200bແລນ", "NZST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານນິວ\u200bຊີ\u200bແລນ", "OESZ": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bອອກ", "OEZ": "ເວ\u200bລາ\u200bມາ\u200bດ\u200bຕະ\u200bຖານ\u200bຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bອອກ", "PDT": "ເວລາກາງເວັນແປຊິຟິກ", "PST": "ເວລາມາດຕະຖານແປຊິຟິກ", "SAST": "ເວ\u200bລາ\u200bອາ\u200bຟຣິ\u200bກາ\u200bໃຕ້", "SGT": "ເວ\u200bລາ\u200bສິງ\u200bກະ\u200bໂປ", "SRT": "ເວ\u200bລາ\u200bຊຸ\u200bຣິ\u200bນາມ", "TMST": "ເວລາລະດູຮ້ອນຕວກເມນິສຖານ", "TMT": "ເວລາມາດຕະຖານຕວກເມນິສຖານ", "UYST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bອູ\u200bຣູ\u200bກວຍ", "UYT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bອູ\u200bຣູ\u200bກວຍ", "VET": "ເວ\u200bລາ\u200bເວ\u200bເນ\u200bຊູ\u200bເອ\u200bລາ", "WARST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນເວ\u200bສ\u200bເທິນອາ\u200bເຈນ\u200bທິ\u200bນາ", "WART": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານເວ\u200bສ\u200bເທິນອາ\u200bເຈນ\u200bທິ\u200bນາ", "WAST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bອາ\u200bຟຣິ\u200bກາ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WAT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bອາ\u200bຟຣິ\u200bກາ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WESZ": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WEZ": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WIB": "ເວ\u200bລາ\u200bອິນ\u200bໂດ\u200bເນ\u200bເຊຍ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WIT": "ເວ\u200bລາ\u200bອິນ\u200bໂດ\u200bເນ\u200bເຊຍ\u200bຕາ\u200bເວັນ\u200bອອກ", "WITA": "ເວ\u200bລາ\u200bອິນ\u200bໂດ\u200bເນ\u200bເຊຍ\u200bກາງ", "∅∅∅": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນອາ\u200bເມ\u200bຊອນ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_0174068e708e390eda0b53f8a0ef057e,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_afb42e88482f5f1e5a494bb738650f23,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_dafe47a99f691cddd99c295dd3a80472,
			fnFmtTimeLong:         fnFmtTimeLong_2be9d9219c97d619bb809b28233eb70f,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lo_la"] = func() locales.Translator {
		return &localen{

			locale:                 "lo_LA",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "-",
			monthsAbbreviated:      []string{"", "ມ.ກ.", "ກ.ພ.", "ມ.ນ.", "ມ.ສ.", "ພ.ພ.", "ມິ.ຖ.", "ກ.ລ.", "ສ.ຫ.", "ກ.ຍ.", "ຕ.ລ.", "ພ.ຈ.", "ທ.ວ."},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "ມັງກອນ", "ກຸມພາ", "ມີນາ", "ເມສາ", "ພຶດສະພາ", "ມິຖຸນາ", "ກໍລະກົດ", "ສິງຫາ", "ກັນຍາ", "ຕຸລາ", "ພະຈິກ", "ທັນວາ"},
			daysAbbreviated:        []string{"ອາທິດ", "ຈັນ", "ອັງຄານ", "ພຸດ", "ພະຫັດ", "ສຸກ", "ເສົາ"},
			daysNarrow:             []string{"ອາ", "ຈ", "ອ", "ພ", "ພຫ", "ສຸ", "ສ"},
			daysShort:              []string{"ອາ.", "ຈ.", "ອ.", "ພ.", "ພຫ.", "ສຸ.", "ສ."},
			daysWide:               []string{"ວັນອາທິດ", "ວັນຈັນ", "ວັນອັງຄານ", "ວັນພຸດ", "ວັນພະຫັດ", "ວັນສຸກ", "ວັນເສົາ"},
			periodsAbbreviated:     []string{"ກ່ອນທ່ຽງ", "ຫຼັງທ່ຽງ"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"ກ່ອນທ່ຽງ", "ຫຼັງທ່ຽງ"},
			erasAbbreviated:        []string{"ກ່ອນ ຄ.ສ.", "ຄ.ສ."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ກ່ອນຄຣິດສັກກະລາດ", "ຄຣິດສັກກະລາດ"},
			timezones:              map[string]string{"ACDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bເລຍ\u200bກາງ", "ACST": "ເວ\u200bລາມາດ\u200bຕະ\u200bຖານອອ\u200bສ\u200bເຕຣ\u200bເລຍ\u200bກ\u200bາງ", "ACWDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bກາງ\u200bຕາ\u200bເວັນ\u200bຕົກ", "ACWST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານອອສ\u200bເຕຣ\u200bລຽນ\u200bກາງ\u200bຕາ\u200bເວັນ\u200bຕົກ", "ADT": "ເວລາກາງເວັນຂອງອາແລນຕິກ", "AEDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bອອກ", "AEST": "ເວ\u200bລາ\u200bມາດຕະຖານ\u200b\u200b\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bອອກ", "AKDT": "ເວລາກາງເວັນອະລັສກາ", "AKST": "ເວລາມາດຕະຖານອະລັສກາ", "ARST": "\u200bເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bອາ\u200bເຈນ\u200bທິ\u200bນາ", "ART": "\u200bເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານອາ\u200bເຈນ\u200bທິ\u200bນາ", "AST": "ເວລາມາດຕະຖານຂອງອາແລນຕິກ", "AWDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bຕົກ", "AWST": "ເວ\u200bລາ\u200bມາ\u200bດ\u200bຕະ\u200bຖານອອສ\u200bເຕຣ\u200bລຽນ\u200bຕາ\u200bເວັນ\u200bຕົກ", "BOT": "ເວ\u200bລາ\u200bໂບ\u200bລິ\u200bເວຍ", "BT": "ເວ\u200bລາ\u200bພູ\u200bຖານ", "CAT": "ເວ\u200bລາ\u200bອາ\u200bຟຣິ\u200bກາ\u200bກາງ", "CDT": "ເວລາກາງເວັນກາງ", "CHADT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bຊາ\u200bທາມ", "CHAST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຊາ\u200bທາມ", "CLST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນຊິ\u200bລີ", "CLT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານຊິ\u200bລີ", "COST": "ເວລາລະດູຮ້ອນໂຄລໍາເບຍ", "COT": "ເວລາມາດຕະຖານໂຄລຳເບຍ", "CST": "ເວລາມາດຕະຖານກາງ", "ChST": "ເວ\u200bລາ\u200bຈາ\u200bໂມ\u200bໂຣ", "EAT": "ເວ\u200bລາ\u200bອາ\u200bຟຣິ\u200bກາ\u200bຕາ\u200bເວັນ\u200bອອກ", "ECT": "ເວ\u200bລາ\u200bເອ\u200bກົວ\u200bດໍ", "EDT": "ເວລາກາງເວັນຕາເວັນອອກ", "EST": "ເວລາມາດຕະຖານຕາເວັນອອກ", "GFT": "ເວ\u200bລາ\u200bເຟ\u200bຣນ\u200bຊ໌\u200bເກຍ\u200bນາ", "GMT": "ເວ\u200bລາກຣີນ\u200bວິ\u200bຊ", "GST": "ເວ\u200bລາ\u200bກູ\u200bລ\u200b໌ຟ", "GYT": "ເວລາກາຍອານາ", "HADT": "ເວລາຕອນທ່ຽງຮາວາຍ-ເອລູທຽນ", "HAST": "ເວລາມາດຕະຖານຮາວາຍ-ເອລູທຽນ", "HAT": "ເວລາກາງເວັນນິວຟາວແລນ", "HECU": "ເວລາກາງເວັນຄິວບາ", "HEEG": "ເວລາລະດູຮ້ອນກຣີນແລນຕາເວັນອອກ", "HENOMX": "ເວລາກາງເວັນເມັກຊິກັນນອດເວສ", "HEOG": "ເວລາຕອນທ່ຽງກຣີນແລນຕາເວັນຕົກ", "HEPM": "\u200bເວ\u200bລາຕອນ\u200bທ່ຽງເຊນ\u200bປີ\u200bແອ ແລະ\u200bມິ\u200bກົວ\u200bລອນ", "HEPMX": "ເວລາກາງເວັນແປຊິຟິກເມັກຊິກັນ", "HKST": "\u200bເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bຮອງ\u200bກົງ", "HKT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຮອງ\u200bກົງ", "HNCU": "ເວລາມາດຕະຖານຂອງຄິວບາ", "HNEG": "ເວລາມາດຕະຖານຕາເວັນອອກກຣີນແລນ", "HNNOMX": "\u200bເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານນອດ\u200bເວ\u200bສ\u200bເມັກ\u200bຊິ\u200bໂກ", "HNOG": "ເວລາມາດຕະຖານກຣີນແລນຕາເວັນຕົກ", "HNPM": "\u200bເວ\u200bລາມາດ\u200bຕະ\u200bຖານເຊນ\u200bປີ\u200bແອ ແລະ\u200bມິ\u200bກົວ\u200bລອນ", "HNPMX": "ເວລາມາດຕະຖານແປຊິຟິກເມັກຊິກັນ", "HNT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bນິວ\u200bຟາວ\u200bແລນ", "IST": "ເວລາ ອິນເດຍ", "JDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bຍີ່\u200bປຸ່ນ", "JST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຍີ່\u200bປຸ່ນ", "LHDT": "\u200bເວ\u200bລ\u200bສາ\u200bຕອນ\u200b\u200bທ່ຽງ\u200bລອດ\u200bເຮົາ\u200b", "LHST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bລອດ\u200bເຮົາ", "MDT": "ເວລາກາງເວັນແຖບພູເຂົາ", "MESZ": "\u200bເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bຢູ\u200bໂຣບ\u200bກາງ", "MEZ": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bຢູ\u200bໂຣບກາງ", "MST": "ເວລາມາດຕະຖານແຖບພູເຂົາ", "MYT": "ເວ\u200bລາ\u200bມາ\u200bເລ\u200bເຊຍ", "NZDT": "ເວ\u200bລາ\u200bຕອນ\u200bທ່ຽງ\u200bນິວ\u200bຊີ\u200bແລນ", "NZST": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານນິວ\u200bຊີ\u200bແລນ", "OESZ": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bອອກ", "OEZ": "ເວ\u200bລາ\u200bມາ\u200bດ\u200bຕະ\u200bຖານ\u200bຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bອອກ", "PDT": "ເວລາກາງເວັນແປຊິຟິກ", "PST": "ເວລາມາດຕະຖານແປຊິຟິກ", "SAST": "ເວ\u200bລາ\u200bອາ\u200bຟຣິ\u200bກາ\u200bໃຕ້", "SGT": "ເວ\u200bລາ\u200bສິງ\u200bກະ\u200bໂປ", "SRT": "ເວ\u200bລາ\u200bຊຸ\u200bຣິ\u200bນາມ", "TMST": "ເວລາລະດູຮ້ອນຕວກເມນິສຖານ", "TMT": "ເວລາມາດຕະຖານຕວກເມນິສຖານ", "UYST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bອູ\u200bຣູ\u200bກວຍ", "UYT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bອູ\u200bຣູ\u200bກວຍ", "VET": "ເວ\u200bລາ\u200bເວ\u200bເນ\u200bຊູ\u200bເອ\u200bລາ", "WARST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນເວ\u200bສ\u200bເທິນອາ\u200bເຈນ\u200bທິ\u200bນາ", "WART": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານເວ\u200bສ\u200bເທິນອາ\u200bເຈນ\u200bທິ\u200bນາ", "WAST": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນ\u200bອາ\u200bຟຣິ\u200bກາ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WAT": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານ\u200bອາ\u200bຟຣິ\u200bກາ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WESZ": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WEZ": "ເວ\u200bລາ\u200bມາດ\u200bຕະ\u200bຖານຢູ\u200bໂຣບ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WIB": "ເວ\u200bລາ\u200bອິນ\u200bໂດ\u200bເນ\u200bເຊຍ\u200bຕາ\u200bເວັນ\u200bຕົກ", "WIT": "ເວ\u200bລາ\u200bອິນ\u200bໂດ\u200bເນ\u200bເຊຍ\u200bຕາ\u200bເວັນ\u200bອອກ", "WITA": "ເວ\u200bລາ\u200bອິນ\u200bໂດ\u200bເນ\u200bເຊຍ\u200bກາງ", "∅∅∅": "ເວ\u200bລາ\u200bລະ\u200bດູ\u200bຮ້ອນອາ\u200bໂຊ\u200bເຣ\u200bສ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_0174068e708e390eda0b53f8a0ef057e,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_afb42e88482f5f1e5a494bb738650f23,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_dafe47a99f691cddd99c295dd3a80472,
			fnFmtTimeLong:         fnFmtTimeLong_2be9d9219c97d619bb809b28233eb70f,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lrc"] = func() locales.Translator {
		return &localen{

			locale:                 "lrc",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "د.ع.\u200f", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جانڤیە", "فئڤریە", "مارس", "آڤریل", "مئی", "جوٙأن", "جوٙلا", "آگوست", "سئپتامر", "ئوکتوڤر", "نوڤامر", "دئسامر"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "جانڤیە", "فئڤریە", "مارس", "آڤریل", "مئی", "جوٙأن", "جوٙلا", "آگوست", "سئپتامر", "ئوکتوڤر", "نوڤامر", "دئسامر"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:               []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "روٙشنایی نئهادار روٙز", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "گاٛت مینجاٛیی ئستاٛنداٛرد", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c1338c4c0c9820a635090f5e89a65c4b,
			fnFmtCurrency:         fnFmtCurrency_bc1dd05937d2f9e1e78f0e1e26de2ce0,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_529d4d803e923e8eb15440c678f6c7b8,
			fnFmtPercent:          fnFmtPercent_aa714d39dff020dd613d3cb5622ff940,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lrc_iq"] = func() locales.Translator {
		return &localen{

			locale:                 "lrc_IQ",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جانڤیە", "فئڤریە", "مارس", "آڤریل", "مئی", "جوٙأن", "جوٙلا", "آگوست", "سئپتامر", "ئوکتوڤر", "نوڤامر", "دئسامر"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "جانڤیە", "فئڤریە", "مارس", "آڤریل", "مئی", "جوٙأن", "جوٙلا", "آگوست", "سئپتامر", "ئوکتوڤر", "نوڤامر", "دئسامر"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:               []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "روٙشنایی نئهادار روٙز", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "گاٛت مینجاٛیی ئستاٛنداٛرد", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c1338c4c0c9820a635090f5e89a65c4b,
			fnFmtCurrency:         fnFmtCurrency_bc1dd05937d2f9e1e78f0e1e26de2ce0,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_529d4d803e923e8eb15440c678f6c7b8,
			fnFmtPercent:          fnFmtPercent_aa714d39dff020dd613d3cb5622ff940,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lrc_ir"] = func() locales.Translator {
		return &localen{

			locale:                 "lrc_IR",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جانڤیە", "فئڤریە", "مارس", "آڤریل", "مئی", "جوٙأن", "جوٙلا", "آگوست", "سئپتامر", "ئوکتوڤر", "نوڤامر", "دئسامر"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "جانڤیە", "فئڤریە", "مارس", "آڤریل", "مئی", "جوٙأن", "جوٙلا", "آگوست", "سئپتامر", "ئوکتوڤر", "نوڤامر", "دئسامر"},
			daysAbbreviated:        []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:               []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "روٙشنایی نئهادار روٙز", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "گاٛت مینجاٛیی ئستاٛنداٛرد", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c1338c4c0c9820a635090f5e89a65c4b,
			fnFmtCurrency:         fnFmtCurrency_bc1dd05937d2f9e1e78f0e1e26de2ce0,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_529d4d803e923e8eb15440c678f6c7b8,
			fnFmtPercent:          fnFmtPercent_aa714d39dff020dd613d3cb5622ff940,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lt"] = func() locales.Translator {
		return &localen{

			locale:                 "lt",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "saus.", "vas.", "kov.", "bal.", "geg.", "birž.", "liep.", "rugp.", "rugs.", "spal.", "lapkr.", "gruod."},
			monthsNarrow:           []string{"", "S", "V", "K", "B", "G", "B", "L", "R", "R", "S", "L", "G"},
			monthsWide:             []string{"", "sausio", "vasario", "kovo", "balandžio", "gegužės", "birželio", "liepos", "rugpjūčio", "rugsėjo", "spalio", "lapkričio", "gruodžio"},
			daysAbbreviated:        []string{"sk", "pr", "an", "tr", "kt", "pn", "št"},
			daysNarrow:             []string{"S", "P", "A", "T", "K", "P", "Š"},
			daysShort:              []string{"Sk", "Pr", "An", "Tr", "Kt", "Pn", "Št"},
			daysWide:               []string{"sekmadienis", "pirmadienis", "antradienis", "trečiadienis", "ketvirtadienis", "penktadienis", "šeštadienis"},
			periodsAbbreviated:     []string{"priešpiet", "popiet"},
			periodsNarrow:          []string{"pr.\u00a0p.", "pop."},
			periodsWide:            []string{"priešpiet", "popiet"},
			erasAbbreviated:        []string{"pr. Kr.", "po Kr."},
			erasNarrow:             []string{"pr. Kr.", "po Kr."},
			erasWide:               []string{"prieš Kristų", "po Kristaus"},
			timezones:              map[string]string{"ACDT": "Centrinės Australijos vasaros laikas", "ACST": "Centrinės Australijos žiemos laikas", "ACWDT": "Centrinės vakarų Australijos vasaros laikas", "ACWST": "Centrinės vakarų Australijos žiemos laikas", "ADT": "Atlanto vasaros laikas", "AEDT": "Rytų Australijos vasaros laikas", "AEST": "Rytų Australijos žiemos laikas", "AKDT": "Aliaskos vasaros laikas", "AKST": "Aliaskos žiemos laikas", "ARST": "Argentinos vasaros laikas", "ART": "Argentinos žiemos laikas", "AST": "Atlanto žiemos laikas", "AWDT": "Vakarų Australijos vasaros laikas", "AWST": "Vakarų Australijos žiemos laikas", "BOT": "Bolivijos laikas", "BT": "Butano laikas", "CAT": "Centrinės Afrikos laikas", "CDT": "Šiaurės Amerikos centro vasaros laikas", "CHADT": "Čatamo vasaros laikas", "CHAST": "Čatamo žiemos laikas", "CLST": "Čilės vasaros laikas", "CLT": "Čilės žiemos laikas", "COST": "Kolumbijos vasaros laikas", "COT": "Kolumbijos žiemos laikas", "CST": "Šiaurės Amerikos centro žiemos laikas", "ChST": "Čamoro laikas", "EAT": "Rytų Afrikos laikas", "ECT": "Ekvadoro laikas", "EDT": "Šiaurės Amerikos rytų vasaros laikas", "EST": "Šiaurės Amerikos rytų žiemos laikas", "GFT": "Prancūzijos Gvianos laikas", "GMT": "Grinvičo laikas", "GST": "Persijos įlankos laikas", "GYT": "Gajanos laikas", "HADT": "Havajų–Aleutų vasaros laikas", "HAST": "Havajų–Aleutų žiemos laikas", "HAT": "Niufaundlendo vasaros laikas", "HECU": "Kubos vasaros laikas", "HEEG": "Grenlandijos rytų vasaros laikas", "HENOMX": "Šiaurės Vakarų Meksikos vasaros laikas", "HEOG": "Grenlandijos vakarų vasaros laikas", "HEPM": "Sen Pjero ir Mikelono vasaros laikas", "HEPMX": "Meksikos Ramiojo vandenyno vasaros laikas", "HKST": "Honkongo vasaros laikas", "HKT": "Honkongo žiemos laikas", "HNCU": "Kubos žiemos laikas", "HNEG": "Grenlandijos rytų žiemos laikas", "HNNOMX": "Šiaurės Vakarų Meksikos žiemos laikas", "HNOG": "Grenlandijos vakarų žiemos laikas", "HNPM": "Sen Pjero ir Mikelono žiemos laikas", "HNPMX": "Meksikos Ramiojo vandenyno žiemos laikas", "HNT": "Niufaundlendo žiemos laikas", "IST": "Indijos laikas", "JDT": "Japonijos vasaros laikas", "JST": "Japonijos žiemos laikas", "LHDT": "Lordo Hau vasaros laikas", "LHST": "Lordo Hau žiemos laikas", "MDT": "Šiaurės Amerikos kalnų vasaros laikas", "MESZ": "Vidurio Europos vasaros laikas", "MEZ": "Vidurio Europos žiemos laikas", "MST": "Šiaurės Amerikos kalnų žiemos laikas", "MYT": "Malaizijos laikas", "NZDT": "Naujosios Zelandijos vasaros laikas", "NZST": "Naujosios Zelandijos žiemos laikas", "OESZ": "Rytų Europos vasaros laikas", "OEZ": "Rytų Europos žiemos laikas", "PDT": "Šiaurės Amerikos Ramiojo vandenyno vasaros laikas", "PST": "Šiaurės Amerikos Ramiojo vandenyno žiemos laikas", "SAST": "Pietų Afrikos laikas", "SGT": "Singapūro laikas", "SRT": "Surinamo laikas", "TMST": "Turkmėnistano vasaros laikas", "TMT": "Turkmėnistano žiemos laikas", "UYST": "Urugvajaus vasaros laikas", "UYT": "Urugvajaus žiemos laikas", "VET": "Venesuelos laikas", "WARST": "Vakarų Argentinos vasaros laikas", "WART": "Vakarų Argentinos žiemos laikas", "WAST": "Vakarų Afrikos vasaros laikas", "WAT": "Vakarų Afrikos žiemos laikas", "WESZ": "Vakarų Europos vasaros laikas", "WEZ": "Vakarų Europos žiemos laikas", "WIB": "Vakarų Indonezijos laikas", "WIT": "Rytų Indonezijos laikas", "WITA": "Centrinės Indonezijos laikas", "∅∅∅": "Peru vasaros laikas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a8ed4aac1cf13bc0d4f980cb943bb5f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_a6a7cb3b48a46ecca7801bbef267e737,
			fnFmtDateLong:         fnFmtDateLong_6af034e83fbef5852ab70b5e01d66101,
			fnFmtDateMedium:       fnFmtDateMedium_a3f984963a38c1972f5ae182cb000f06,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lt_lt"] = func() locales.Translator {
		return &localen{

			locale:                 "lt_LT",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "saus.", "vas.", "kov.", "bal.", "geg.", "birž.", "liep.", "rugp.", "rugs.", "spal.", "lapkr.", "gruod."},
			monthsNarrow:           []string{"", "S", "V", "K", "B", "G", "B", "L", "R", "R", "S", "L", "G"},
			monthsWide:             []string{"", "sausio", "vasario", "kovo", "balandžio", "gegužės", "birželio", "liepos", "rugpjūčio", "rugsėjo", "spalio", "lapkričio", "gruodžio"},
			daysAbbreviated:        []string{"sk", "pr", "an", "tr", "kt", "pn", "št"},
			daysNarrow:             []string{"S", "P", "A", "T", "K", "P", "Š"},
			daysShort:              []string{"Sk", "Pr", "An", "Tr", "Kt", "Pn", "Št"},
			daysWide:               []string{"sekmadienis", "pirmadienis", "antradienis", "trečiadienis", "ketvirtadienis", "penktadienis", "šeštadienis"},
			periodsAbbreviated:     []string{"priešpiet", "popiet"},
			periodsNarrow:          []string{"pr.\u00a0p.", "pop."},
			periodsWide:            []string{"priešpiet", "popiet"},
			erasAbbreviated:        []string{"pr. Kr.", "po Kr."},
			erasNarrow:             []string{"pr. Kr.", "po Kr."},
			erasWide:               []string{"prieš Kristų", "po Kristaus"},
			timezones:              map[string]string{"ACDT": "Centrinės Australijos vasaros laikas", "ACST": "Centrinės Australijos žiemos laikas", "ACWDT": "Centrinės vakarų Australijos vasaros laikas", "ACWST": "Centrinės vakarų Australijos žiemos laikas", "ADT": "Atlanto vasaros laikas", "AEDT": "Rytų Australijos vasaros laikas", "AEST": "Rytų Australijos žiemos laikas", "AKDT": "Aliaskos vasaros laikas", "AKST": "Aliaskos žiemos laikas", "ARST": "Argentinos vasaros laikas", "ART": "Argentinos žiemos laikas", "AST": "Atlanto žiemos laikas", "AWDT": "Vakarų Australijos vasaros laikas", "AWST": "Vakarų Australijos žiemos laikas", "BOT": "Bolivijos laikas", "BT": "Butano laikas", "CAT": "Centrinės Afrikos laikas", "CDT": "Šiaurės Amerikos centro vasaros laikas", "CHADT": "Čatamo vasaros laikas", "CHAST": "Čatamo žiemos laikas", "CLST": "Čilės vasaros laikas", "CLT": "Čilės žiemos laikas", "COST": "Kolumbijos vasaros laikas", "COT": "Kolumbijos žiemos laikas", "CST": "Šiaurės Amerikos centro žiemos laikas", "ChST": "Čamoro laikas", "EAT": "Rytų Afrikos laikas", "ECT": "Ekvadoro laikas", "EDT": "Šiaurės Amerikos rytų vasaros laikas", "EST": "Šiaurės Amerikos rytų žiemos laikas", "GFT": "Prancūzijos Gvianos laikas", "GMT": "Grinvičo laikas", "GST": "Persijos įlankos laikas", "GYT": "Gajanos laikas", "HADT": "Havajų–Aleutų vasaros laikas", "HAST": "Havajų–Aleutų žiemos laikas", "HAT": "Niufaundlendo vasaros laikas", "HECU": "Kubos vasaros laikas", "HEEG": "Grenlandijos rytų vasaros laikas", "HENOMX": "Šiaurės Vakarų Meksikos vasaros laikas", "HEOG": "Grenlandijos vakarų vasaros laikas", "HEPM": "Sen Pjero ir Mikelono vasaros laikas", "HEPMX": "Meksikos Ramiojo vandenyno vasaros laikas", "HKST": "Honkongo vasaros laikas", "HKT": "Honkongo žiemos laikas", "HNCU": "Kubos žiemos laikas", "HNEG": "Grenlandijos rytų žiemos laikas", "HNNOMX": "Šiaurės Vakarų Meksikos žiemos laikas", "HNOG": "Grenlandijos vakarų žiemos laikas", "HNPM": "Sen Pjero ir Mikelono žiemos laikas", "HNPMX": "Meksikos Ramiojo vandenyno žiemos laikas", "HNT": "Niufaundlendo žiemos laikas", "IST": "Indijos laikas", "JDT": "Japonijos vasaros laikas", "JST": "Japonijos žiemos laikas", "LHDT": "Lordo Hau vasaros laikas", "LHST": "Lordo Hau žiemos laikas", "MDT": "Šiaurės Amerikos kalnų vasaros laikas", "MESZ": "Vidurio Europos vasaros laikas", "MEZ": "Vidurio Europos žiemos laikas", "MST": "Šiaurės Amerikos kalnų žiemos laikas", "MYT": "Malaizijos laikas", "NZDT": "Naujosios Zelandijos vasaros laikas", "NZST": "Naujosios Zelandijos žiemos laikas", "OESZ": "Rytų Europos vasaros laikas", "OEZ": "Rytų Europos žiemos laikas", "PDT": "Šiaurės Amerikos Ramiojo vandenyno vasaros laikas", "PST": "Šiaurės Amerikos Ramiojo vandenyno žiemos laikas", "SAST": "Pietų Afrikos laikas", "SGT": "Singapūro laikas", "SRT": "Surinamo laikas", "TMST": "Turkmėnistano vasaros laikas", "TMT": "Turkmėnistano žiemos laikas", "UYST": "Urugvajaus vasaros laikas", "UYT": "Urugvajaus žiemos laikas", "VET": "Venesuelos laikas", "WARST": "Vakarų Argentinos vasaros laikas", "WART": "Vakarų Argentinos žiemos laikas", "WAST": "Vakarų Afrikos vasaros laikas", "WAT": "Vakarų Afrikos žiemos laikas", "WESZ": "Vakarų Europos vasaros laikas", "WEZ": "Vakarų Europos žiemos laikas", "WIB": "Vakarų Indonezijos laikas", "WIT": "Rytų Indonezijos laikas", "WITA": "Centrinės Indonezijos laikas", "∅∅∅": "Brazilijos vasaros laikas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c9ff0d41f0522116859535aa9029ca16,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_a6a7cb3b48a46ecca7801bbef267e737,
			fnFmtDateLong:         fnFmtDateLong_6af034e83fbef5852ab70b5e01d66101,
			fnFmtDateMedium:       fnFmtDateMedium_a3f984963a38c1972f5ae182cb000f06,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lu"] = func() locales.Translator {
		return &localen{

			locale:             "lu",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "FC", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Cio", "Lui", "Lus", "Muu", "Lum", "Luf", "Kab", "Lush", "Lut", "Lun", "Kas", "Cis"},
			monthsNarrow:       []string{"", "C", "L", "L", "M", "L", "L", "K", "L", "L", "L", "K", "C"},
			monthsWide:         []string{"", "Ciongo", "Lùishi", "Lusòlo", "Mùuyà", "Lumùngùlù", "Lufuimi", "Kabàlàshìpù", "Lùshìkà", "Lutongolo", "Lungùdi", "Kaswèkèsè", "Ciswà"},
			daysAbbreviated:    []string{"Lum", "Nko", "Ndy", "Ndg", "Njw", "Ngv", "Lub"},
			daysNarrow:         []string{"L", "N", "N", "N", "N", "N", "L"},
			daysWide:           []string{"Lumingu", "Nkodya", "Ndàayà", "Ndangù", "Njòwa", "Ngòvya", "Lubingu"},
			periodsAbbreviated: []string{"Dinda", "Dilolo"},
			periodsWide:        []string{"Dinda", "Dilolo"},
			erasAbbreviated:    []string{"kmp. Y.K.", "kny. Y. K."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kumpala kwa Yezu Kli", "Kunyima kwa Yezu Kli"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lu_cd"] = func() locales.Translator {
		return &localen{

			locale:             "lu_CD",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Cio", "Lui", "Lus", "Muu", "Lum", "Luf", "Kab", "Lush", "Lut", "Lun", "Kas", "Cis"},
			monthsNarrow:       []string{"", "C", "L", "L", "M", "L", "L", "K", "L", "L", "L", "K", "C"},
			monthsWide:         []string{"", "Ciongo", "Lùishi", "Lusòlo", "Mùuyà", "Lumùngùlù", "Lufuimi", "Kabàlàshìpù", "Lùshìkà", "Lutongolo", "Lungùdi", "Kaswèkèsè", "Ciswà"},
			daysAbbreviated:    []string{"Lum", "Nko", "Ndy", "Ndg", "Njw", "Ngv", "Lub"},
			daysNarrow:         []string{"L", "N", "N", "N", "N", "N", "L"},
			daysWide:           []string{"Lumingu", "Nkodya", "Ndàayà", "Ndangù", "Njòwa", "Ngòvya", "Lubingu"},
			periodsAbbreviated: []string{"Dinda", "Dilolo"},
			periodsWide:        []string{"Dinda", "Dilolo"},
			erasAbbreviated:    []string{"kmp. Y.K.", "kny. Y. K."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kumpala kwa Yezu Kli", "Kunyima kwa Yezu Kli"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["luo"] = func() locales.Translator {
		return &localen{

			locale:             "luo",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "DAC", "DAR", "DAD", "DAN", "DAH", "DAU", "DAO", "DAB", "DOC", "DAP", "DGI", "DAG"},
			monthsNarrow:       []string{"", "C", "R", "D", "N", "B", "U", "B", "B", "C", "P", "C", "P"},
			monthsWide:         []string{"", "Dwe mar Achiel", "Dwe mar Ariyo", "Dwe mar Adek", "Dwe mar Ang’wen", "Dwe mar Abich", "Dwe mar Auchiel", "Dwe mar Abiriyo", "Dwe mar Aboro", "Dwe mar Ochiko", "Dwe mar Apar", "Dwe mar gi achiel", "Dwe mar Apar gi ariyo"},
			daysAbbreviated:    []string{"JMP", "WUT", "TAR", "TAD", "TAN", "TAB", "NGS"},
			daysNarrow:         []string{"J", "W", "T", "T", "T", "T", "N"},
			daysWide:           []string{"Jumapil", "Wuok Tich", "Tich Ariyo", "Tich Adek", "Tich Ang’wen", "Tich Abich", "Ngeso"},
			periodsAbbreviated: []string{"OD", "OT"},
			periodsWide:        []string{"OD", "OT"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kapok Kristo obiro", "Ka Kristo osebiro"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["luo_ke"] = func() locales.Translator {
		return &localen{

			locale:             "luo_KE",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "DAC", "DAR", "DAD", "DAN", "DAH", "DAU", "DAO", "DAB", "DOC", "DAP", "DGI", "DAG"},
			monthsNarrow:       []string{"", "C", "R", "D", "N", "B", "U", "B", "B", "C", "P", "C", "P"},
			monthsWide:         []string{"", "Dwe mar Achiel", "Dwe mar Ariyo", "Dwe mar Adek", "Dwe mar Ang’wen", "Dwe mar Abich", "Dwe mar Auchiel", "Dwe mar Abiriyo", "Dwe mar Aboro", "Dwe mar Ochiko", "Dwe mar Apar", "Dwe mar gi achiel", "Dwe mar Apar gi ariyo"},
			daysAbbreviated:    []string{"JMP", "WUT", "TAR", "TAD", "TAN", "TAB", "NGS"},
			daysNarrow:         []string{"J", "W", "T", "T", "T", "T", "N"},
			daysWide:           []string{"Jumapil", "Wuok Tich", "Tich Ariyo", "Tich Adek", "Tich Ang’wen", "Tich Abich", "Ngeso"},
			periodsAbbreviated: []string{"OD", "OT"},
			periodsWide:        []string{"OD", "OT"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kapok Kristo obiro", "Ka Kristo osebiro"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["luy"] = func() locales.Translator {
		return &localen{

			locale:             "luy",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"J2", "J3", "J4", "J5", "Al", "Ij", "J1"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:           []string{"Jumapiri", "Jumatatu", "Jumanne", "Jumatano", "Murwa wa Kanne", "Murwa wa Katano", "Jumamosi"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Imberi ya Kuuza Kwa", "Muhiga Kuvita Kuuza"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["luy_ke"] = func() locales.Translator {
		return &localen{

			locale:             "luy_KE",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"J2", "J3", "J4", "J5", "Al", "Ij", "J1"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:           []string{"Jumapiri", "Jumatatu", "Jumanne", "Jumatano", "Murwa wa Kanne", "Murwa wa Katano", "Jumamosi"},
			periodsAbbreviated: []string{"a.m.", "p.m."},
			periodsWide:        []string{"a.m.", "p.m."},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Imberi ya Kuuza Kwa", "Muhiga Kuvita Kuuza"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lv"] = func() locales.Translator {
		return &localen{

			locale:                 "lv",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "Ls", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "janv.", "febr.", "marts", "apr.", "maijs", "jūn.", "jūl.", "aug.", "sept.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvāris", "februāris", "marts", "aprīlis", "maijs", "jūnijs", "jūlijs", "augusts", "septembris", "oktobris", "novembris", "decembris"},
			daysAbbreviated:        []string{"svētd.", "pirmd.", "otrd.", "trešd.", "ceturtd.", "piektd.", "sestd."},
			daysNarrow:             []string{"S", "P", "O", "T", "C", "P", "S"},
			daysShort:              []string{"Sv", "Pr", "Ot", "Tr", "Ce", "Pk", "Se"},
			daysWide:               []string{"svētdiena", "pirmdiena", "otrdiena", "trešdiena", "ceturtdiena", "piektdiena", "sestdiena"},
			periodsAbbreviated:     []string{"priekšp.", "pēcp."},
			periodsNarrow:          []string{"priekšp.", "pēcp."},
			periodsWide:            []string{"priekšpusdienā", "pēcpusdienā"},
			erasAbbreviated:        []string{"p.m.ē.", "m.ē."},
			erasNarrow:             []string{"p.m.ē.", "m.ē."},
			erasWide:               []string{"pirms mūsu ēras", "mūsu ērā"},
			timezones:              map[string]string{"ACDT": "Austrālijas centrālais vasaras laiks", "ACST": "Austrālijas centrālais ziemas laiks", "ACWDT": "Austrālijas centrālais rietumu vasaras laiks", "ACWST": "Austrālijas centrālais rietumu ziemas laiks", "ADT": "Atlantijas vasaras laiks", "AEDT": "Austrālijas austrumu vasaras laiks", "AEST": "Austrālijas austrumu ziemas laiks", "AKDT": "Aļaskas vasaras laiks", "AKST": "Aļaskas ziemas laiks", "ARST": "Argentīnas vasaras laiks", "ART": "Argentīnas ziemas laiks", "AST": "Atlantijas ziemas laiks", "AWDT": "Austrālijas rietumu vasaras laiks", "AWST": "Austrālijas rietumu ziemas laiks", "BOT": "Bolīvijas laiks", "BT": "Butānas laiks", "CAT": "Centrālāfrikas laiks", "CDT": "Centrālais vasaras laiks", "CHADT": "Četemas vasaras laiks", "CHAST": "Četemas ziemas laiks", "CLST": "Čīles vasaras laiks", "CLT": "Čīles ziemas laiks", "COST": "Kolumbijas vasaras laiks", "COT": "Kolumbijas ziemas laiks", "CST": "Centrālais ziemas laiks", "ChST": "Čamorra ziemas laiks", "EAT": "Austrumāfrikas laiks", "ECT": "Ekvadoras laiks", "EDT": "Austrumu vasaras laiks", "EST": "Austrumu ziemas laiks", "GFT": "Francijas Gviānas laiks", "GMT": "Griničas laiks", "GST": "Persijas līča laiks", "GYT": "Gajānas laiks", "HADT": "Havaju–Aleutu vasaras laiks", "HAST": "Havaju–Aleutu ziemas laiks", "HAT": "Ņūfaundlendas vasaras laiks", "HECU": "Kubas vasaras laiks", "HEEG": "Austrumgrenlandes vasaras laiks", "HENOMX": "Ziemeļrietumu Meksikas vasaras laiks", "HEOG": "Rietumgrenlandes vasaras laiks", "HEPM": "Senpjēras un Mikelonas vasaras laiks", "HEPMX": "Meksikas Klusā okeāna piekrastes vasaras laiks", "HKST": "Honkongas vasaras laiks", "HKT": "Honkongas ziemas laiks", "HNCU": "Kubas ziemas laiks", "HNEG": "Austrumgrenlandes ziemas laiks", "HNNOMX": "Ziemeļrietumu Meksikas ziemas laiks", "HNOG": "Rietumgrenlandes ziemas laiks", "HNPM": "Senpjēras un Mikelonas ziemas laiks", "HNPMX": "Meksikas Klusā okeāna piekrastes ziemas laiks", "HNT": "Ņūfaundlendas ziemas laiks", "IST": "Indijas ziemas laiks", "JDT": "Japānas vasaras laiks", "JST": "Japānas ziemas laiks", "LHDT": "Lorda Hava salas vasaras laiks", "LHST": "Lorda Hava salas ziemas laiks", "MDT": "Kalnu vasaras laiks", "MESZ": "Centrāleiropas vasaras laiks", "MEZ": "Centrāleiropas ziemas laiks", "MST": "Kalnu ziemas laiks", "MYT": "Malaizijas laiks", "NZDT": "Jaunzēlandes vasaras laiks", "NZST": "Jaunzēlandes ziemas laiks", "OESZ": "Austrumeiropas vasaras laiks", "OEZ": "Austrumeiropas ziemas laiks", "PDT": "Klusā okeāna vasaras laiks", "PST": "Klusā okeāna ziemas laiks", "SAST": "Dienvidāfrikas ziemas laiks", "SGT": "Singapūras laiks", "SRT": "Surinamas laiks", "TMST": "Turkmenistānas vasaras laiks", "TMT": "Turkmenistānas ziemas laiks", "UYST": "Urugvajas vasaras laiks", "UYT": "Urugvajas ziemas laiks", "VET": "Venecuēlas laiks", "WARST": "Rietumargentīnas vasaras laiks", "WART": "Rietumargentīnas ziemas laiks", "WAST": "Rietumāfrikas vasaras laiks", "WAT": "Rietumāfrikas ziemas laiks", "WESZ": "Rietumeiropas vasaras laiks", "WEZ": "Rietumeiropas ziemas laiks", "WIB": "Rietumindonēzijas laiks", "WIT": "Austrumindonēzijas laiks", "WITA": "Centrālindonēzijas laiks", "∅∅∅": "Azoru salu vasaras laiks"},

			fnCardinalPluralRule:  fnCardinalPluralRule_35d374490f3e32fac8828ad5843b9d6e,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c68531efc0310a76517a9616064247e7,
			fnFmtDateLong:         fnFmtDateLong_b7e0e3e1d7c2a12e272d025e98a6facc,
			fnFmtDateMedium:       fnFmtDateMedium_cefeb908e03bbb483b089e5f8a7e3bb5,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52f7b7c1971a0689c24fe53cd3a6bf1d,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["lv_lv"] = func() locales.Translator {
		return &localen{

			locale:                 "lv_LV",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "janv.", "febr.", "marts", "apr.", "maijs", "jūn.", "jūl.", "aug.", "sept.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janvāris", "februāris", "marts", "aprīlis", "maijs", "jūnijs", "jūlijs", "augusts", "septembris", "oktobris", "novembris", "decembris"},
			daysAbbreviated:        []string{"svētd.", "pirmd.", "otrd.", "trešd.", "ceturtd.", "piektd.", "sestd."},
			daysNarrow:             []string{"S", "P", "O", "T", "C", "P", "S"},
			daysShort:              []string{"Sv", "Pr", "Ot", "Tr", "Ce", "Pk", "Se"},
			daysWide:               []string{"svētdiena", "pirmdiena", "otrdiena", "trešdiena", "ceturtdiena", "piektdiena", "sestdiena"},
			periodsAbbreviated:     []string{"priekšp.", "pēcp."},
			periodsNarrow:          []string{"priekšp.", "pēcp."},
			periodsWide:            []string{"priekšpusdienā", "pēcpusdienā"},
			erasAbbreviated:        []string{"p.m.ē.", "m.ē."},
			erasNarrow:             []string{"p.m.ē.", "m.ē."},
			erasWide:               []string{"pirms mūsu ēras", "mūsu ērā"},
			timezones:              map[string]string{"ACDT": "Austrālijas centrālais vasaras laiks", "ACST": "Austrālijas centrālais ziemas laiks", "ACWDT": "Austrālijas centrālais rietumu vasaras laiks", "ACWST": "Austrālijas centrālais rietumu ziemas laiks", "ADT": "Atlantijas vasaras laiks", "AEDT": "Austrālijas austrumu vasaras laiks", "AEST": "Austrālijas austrumu ziemas laiks", "AKDT": "Aļaskas vasaras laiks", "AKST": "Aļaskas ziemas laiks", "ARST": "Argentīnas vasaras laiks", "ART": "Argentīnas ziemas laiks", "AST": "Atlantijas ziemas laiks", "AWDT": "Austrālijas rietumu vasaras laiks", "AWST": "Austrālijas rietumu ziemas laiks", "BOT": "Bolīvijas laiks", "BT": "Butānas laiks", "CAT": "Centrālāfrikas laiks", "CDT": "Centrālais vasaras laiks", "CHADT": "Četemas vasaras laiks", "CHAST": "Četemas ziemas laiks", "CLST": "Čīles vasaras laiks", "CLT": "Čīles ziemas laiks", "COST": "Kolumbijas vasaras laiks", "COT": "Kolumbijas ziemas laiks", "CST": "Centrālais ziemas laiks", "ChST": "Čamorra ziemas laiks", "EAT": "Austrumāfrikas laiks", "ECT": "Ekvadoras laiks", "EDT": "Austrumu vasaras laiks", "EST": "Austrumu ziemas laiks", "GFT": "Francijas Gviānas laiks", "GMT": "Griničas laiks", "GST": "Persijas līča laiks", "GYT": "Gajānas laiks", "HADT": "Havaju–Aleutu vasaras laiks", "HAST": "Havaju–Aleutu ziemas laiks", "HAT": "Ņūfaundlendas vasaras laiks", "HECU": "Kubas vasaras laiks", "HEEG": "Austrumgrenlandes vasaras laiks", "HENOMX": "Ziemeļrietumu Meksikas vasaras laiks", "HEOG": "Rietumgrenlandes vasaras laiks", "HEPM": "Senpjēras un Mikelonas vasaras laiks", "HEPMX": "Meksikas Klusā okeāna piekrastes vasaras laiks", "HKST": "Honkongas vasaras laiks", "HKT": "Honkongas ziemas laiks", "HNCU": "Kubas ziemas laiks", "HNEG": "Austrumgrenlandes ziemas laiks", "HNNOMX": "Ziemeļrietumu Meksikas ziemas laiks", "HNOG": "Rietumgrenlandes ziemas laiks", "HNPM": "Senpjēras un Mikelonas ziemas laiks", "HNPMX": "Meksikas Klusā okeāna piekrastes ziemas laiks", "HNT": "Ņūfaundlendas ziemas laiks", "IST": "Indijas ziemas laiks", "JDT": "Japānas vasaras laiks", "JST": "Japānas ziemas laiks", "LHDT": "Lorda Hava salas vasaras laiks", "LHST": "Lorda Hava salas ziemas laiks", "MDT": "Kalnu vasaras laiks", "MESZ": "Centrāleiropas vasaras laiks", "MEZ": "Centrāleiropas ziemas laiks", "MST": "Kalnu ziemas laiks", "MYT": "Malaizijas laiks", "NZDT": "Jaunzēlandes vasaras laiks", "NZST": "Jaunzēlandes ziemas laiks", "OESZ": "Austrumeiropas vasaras laiks", "OEZ": "Austrumeiropas ziemas laiks", "PDT": "Klusā okeāna vasaras laiks", "PST": "Klusā okeāna ziemas laiks", "SAST": "Dienvidāfrikas ziemas laiks", "SGT": "Singapūras laiks", "SRT": "Surinamas laiks", "TMST": "Turkmenistānas vasaras laiks", "TMT": "Turkmenistānas ziemas laiks", "UYST": "Urugvajas vasaras laiks", "UYT": "Urugvajas ziemas laiks", "VET": "Venecuēlas laiks", "WARST": "Rietumargentīnas vasaras laiks", "WART": "Rietumargentīnas ziemas laiks", "WAST": "Rietumāfrikas vasaras laiks", "WAT": "Rietumāfrikas ziemas laiks", "WESZ": "Rietumeiropas vasaras laiks", "WEZ": "Rietumeiropas ziemas laiks", "WIB": "Rietumindonēzijas laiks", "WIT": "Austrumindonēzijas laiks", "WITA": "Centrālindonēzijas laiks", "∅∅∅": "Brazīlijas vasaras laiks"},

			fnCardinalPluralRule:  fnCardinalPluralRule_073a4fc904d56c3f55fe9007d4416cce,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c68531efc0310a76517a9616064247e7,
			fnFmtDateLong:         fnFmtDateLong_b7e0e3e1d7c2a12e272d025e98a6facc,
			fnFmtDateMedium:       fnFmtDateMedium_cefeb908e03bbb483b089e5f8a7e3bb5,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52f7b7c1971a0689c24fe53cd3a6bf1d,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mas"] = func() locales.Translator {
		return &localen{

			locale:                 "mas",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Dal", "Ará", "Ɔɛn", "Doy", "Lép", "Rok", "Sás", "Bɔ́r", "Kús", "Gís", "Shʉ́", "Ntʉ́"},
			monthsWide:             []string{"", "Oladalʉ́", "Arát", "Ɔɛnɨ́ɔɨŋɔk", "Olodoyíóríê inkókúâ", "Oloilépūnyīē inkókúâ", "Kújúɔrɔk", "Mórusásin", "Ɔlɔ́ɨ́bɔ́rárɛ", "Kúshîn", "Olgísan", "Pʉshʉ́ka", "Ntʉ́ŋʉ́s"},
			daysAbbreviated:        []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:             []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:               []string{"Jumapílí", "Jumatátu", "Jumane", "Jumatánɔ", "Alaámisi", "Jumáa", "Jumamósi"},
			periodsAbbreviated:     []string{"Ɛnkakɛnyá", "Ɛndámâ"},
			periodsWide:            []string{"Ɛnkakɛnyá", "Ɛndámâ"},
			erasAbbreviated:        []string{"MY", "EY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Meínō Yɛ́sʉ", "Eínō Yɛ́sʉ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mas_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "mas_KE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Dal", "Ará", "Ɔɛn", "Doy", "Lép", "Rok", "Sás", "Bɔ́r", "Kús", "Gís", "Shʉ́", "Ntʉ́"},
			monthsWide:             []string{"", "Oladalʉ́", "Arát", "Ɔɛnɨ́ɔɨŋɔk", "Olodoyíóríê inkókúâ", "Oloilépūnyīē inkókúâ", "Kújúɔrɔk", "Mórusásin", "Ɔlɔ́ɨ́bɔ́rárɛ", "Kúshîn", "Olgísan", "Pʉshʉ́ka", "Ntʉ́ŋʉ́s"},
			daysAbbreviated:        []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:             []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:               []string{"Jumapílí", "Jumatátu", "Jumane", "Jumatánɔ", "Alaámisi", "Jumáa", "Jumamósi"},
			periodsAbbreviated:     []string{"Ɛnkakɛnyá", "Ɛndámâ"},
			periodsWide:            []string{"Ɛnkakɛnyá", "Ɛndámâ"},
			erasAbbreviated:        []string{"MY", "EY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Meínō Yɛ́sʉ", "Eínō Yɛ́sʉ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mas_tz"] = func() locales.Translator {
		return &localen{

			locale:                 "mas_TZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Dal", "Ará", "Ɔɛn", "Doy", "Lép", "Rok", "Sás", "Bɔ́r", "Kús", "Gís", "Shʉ́", "Ntʉ́"},
			monthsWide:             []string{"", "Oladalʉ́", "Arát", "Ɔɛnɨ́ɔɨŋɔk", "Olodoyíóríê inkókúâ", "Oloilépūnyīē inkókúâ", "Kújúɔrɔk", "Mórusásin", "Ɔlɔ́ɨ́bɔ́rárɛ", "Kúshîn", "Olgísan", "Pʉshʉ́ka", "Ntʉ́ŋʉ́s"},
			daysAbbreviated:        []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:             []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:               []string{"Jumapílí", "Jumatátu", "Jumane", "Jumatánɔ", "Alaámisi", "Jumáa", "Jumamósi"},
			periodsAbbreviated:     []string{"Ɛnkakɛnyá", "Ɛndámâ"},
			periodsWide:            []string{"Ɛnkakɛnyá", "Ɛndámâ"},
			erasAbbreviated:        []string{"MY", "EY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Meínō Yɛ́sʉ", "Eínō Yɛ́sʉ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mer"] = func() locales.Translator {
		return &localen{

			locale:                 "mer",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "JAN", "FEB", "MAC", "ĨPU", "MĨĨ", "NJU", "NJR", "AGA", "SPT", "OKT", "NOV", "DEC"},
			monthsNarrow:           []string{"", "J", "F", "M", "Ĩ", "M", "N", "N", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januarĩ", "Feburuarĩ", "Machi", "Ĩpurũ", "Mĩĩ", "Njuni", "Njuraĩ", "Agasti", "Septemba", "Oktũba", "Novemba", "Dicemba"},
			daysAbbreviated:        []string{"KIU", "MRA", "WAI", "WET", "WEN", "WTN", "JUM"},
			daysNarrow:             []string{"K", "M", "W", "W", "W", "W", "J"},
			daysWide:               []string{"Kiumia", "Muramuko", "Wairi", "Wethatu", "Wena", "Wetano", "Jumamosi"},
			periodsAbbreviated:     []string{"RŨ", "ŨG"},
			periodsWide:            []string{"RŨ", "ŨG"},
			erasAbbreviated:        []string{"MK", "NK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbere ya Kristũ", "Nyuma ya Kristũ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mer_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "mer_KE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "JAN", "FEB", "MAC", "ĨPU", "MĨĨ", "NJU", "NJR", "AGA", "SPT", "OKT", "NOV", "DEC"},
			monthsNarrow:           []string{"", "J", "F", "M", "Ĩ", "M", "N", "N", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januarĩ", "Feburuarĩ", "Machi", "Ĩpurũ", "Mĩĩ", "Njuni", "Njuraĩ", "Agasti", "Septemba", "Oktũba", "Novemba", "Dicemba"},
			daysAbbreviated:        []string{"KIU", "MRA", "WAI", "WET", "WEN", "WTN", "JUM"},
			daysNarrow:             []string{"K", "M", "W", "W", "W", "W", "J"},
			daysWide:               []string{"Kiumia", "Muramuko", "Wairi", "Wethatu", "Wena", "Wetano", "Jumamosi"},
			periodsAbbreviated:     []string{"RŨ", "ŨG"},
			periodsWide:            []string{"RŨ", "ŨG"},
			erasAbbreviated:        []string{"MK", "NK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Mbere ya Kristũ", "Nyuma ya Kristũ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mfe"] = func() locales.Translator {
		return &localen{

			locale:                 "mfe",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "Rs", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "zan", "fev", "mar", "avr", "me", "zin", "zil", "out", "sep", "okt", "nov", "des"},
			monthsNarrow:           []string{"", "z", "f", "m", "a", "m", "z", "z", "o", "s", "o", "n", "d"},
			monthsWide:             []string{"", "zanvie", "fevriye", "mars", "avril", "me", "zin", "zilye", "out", "septam", "oktob", "novam", "desam"},
			daysAbbreviated:        []string{"dim", "lin", "mar", "mer", "ze", "van", "sam"},
			daysNarrow:             []string{"d", "l", "m", "m", "z", "v", "s"},
			daysWide:               []string{"dimans", "lindi", "mardi", "merkredi", "zedi", "vandredi", "samdi"},
			erasAbbreviated:        []string{"av. Z-K", "ap. Z-K"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"avan Zezi-Krist", "apre Zezi-Krist"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mfe_mu"] = func() locales.Translator {
		return &localen{

			locale:                 "mfe_MU",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "zan", "fev", "mar", "avr", "me", "zin", "zil", "out", "sep", "okt", "nov", "des"},
			monthsNarrow:           []string{"", "z", "f", "m", "a", "m", "z", "z", "o", "s", "o", "n", "d"},
			monthsWide:             []string{"", "zanvie", "fevriye", "mars", "avril", "me", "zin", "zilye", "out", "septam", "oktob", "novam", "desam"},
			daysAbbreviated:        []string{"dim", "lin", "mar", "mer", "ze", "van", "sam"},
			daysNarrow:             []string{"d", "l", "m", "m", "z", "v", "s"},
			daysWide:               []string{"dimans", "lindi", "mardi", "merkredi", "zedi", "vandredi", "samdi"},
			erasAbbreviated:        []string{"av. Z-K", "ap. Z-K"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"avan Zezi-Krist", "apre Zezi-Krist"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_997e5abdc6c3d9acf1269cf6ebc154e1,
			fnFmtCurrency:         fnFmtCurrency_10729cc1c61d4b51865a5b0ad149a301,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mg"] = func() locales.Translator {
		return &localen{

			locale:             "mg",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "Ar", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "Mey", "Jon", "Jol", "Aog", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Janoary", "Febroary", "Martsa", "Aprily", "Mey", "Jona", "Jolay", "Aogositra", "Septambra", "Oktobra", "Novambra", "Desambra"},
			daysAbbreviated:    []string{"Alah", "Alats", "Tal", "Alar", "Alak", "Zom", "Asab"},
			daysNarrow:         []string{"A", "A", "T", "A", "A", "Z", "A"},
			daysShort:          []string{"Alah", "Alats", "Tal", "Alar", "Alak", "Zom", "Asab"},
			daysWide:           []string{"Alahady", "Alatsinainy", "Talata", "Alarobia", "Alakamisy", "Zoma", "Asabotsy"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Alohan’i JK", "Aorian’i JK"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mg_mg"] = func() locales.Translator {
		return &localen{

			locale:             "mg_MG",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "Mey", "Jon", "Jol", "Aog", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Janoary", "Febroary", "Martsa", "Aprily", "Mey", "Jona", "Jolay", "Aogositra", "Septambra", "Oktobra", "Novambra", "Desambra"},
			daysAbbreviated:    []string{"Alah", "Alats", "Tal", "Alar", "Alak", "Zom", "Asab"},
			daysNarrow:         []string{"A", "A", "T", "A", "A", "Z", "A"},
			daysShort:          []string{"Alah", "Alats", "Tal", "Alar", "Alak", "Zom", "Asab"},
			daysWide:           []string{"Alahady", "Alatsinainy", "Talata", "Alarobia", "Alakamisy", "Zoma", "Asabotsy"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Alohan’i JK", "Aorian’i JK"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mgh"] = func() locales.Translator {
		return &localen{

			locale:                 "mgh",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MTn", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "Kwa", "Una", "Rar", "Che", "Tha", "Moc", "Sab", "Nan", "Tis", "Kum", "Moj", "Yel"},
			monthsNarrow:           []string{"", "K", "U", "R", "C", "T", "M", "S", "N", "T", "K", "M", "Y"},
			monthsWide:             []string{"", "Mweri wo kwanza", "Mweri wo unayeli", "Mweri wo uneraru", "Mweri wo unecheshe", "Mweri wo unethanu", "Mweri wo thanu na mocha", "Mweri wo saba", "Mweri wo nane", "Mweri wo tisa", "Mweri wo kumi", "Mweri wo kumi na moja", "Mweri wo kumi na yel’li"},
			daysAbbreviated:        []string{"Sab", "Jtt", "Jnn", "Jtn", "Ara", "Iju", "Jmo"},
			daysNarrow:             []string{"S", "J", "J", "J", "A", "I", "J"},
			daysWide:               []string{"Sabato", "Jumatatu", "Jumanne", "Jumatano", "Arahamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"wichishu", "mchochil’l"},
			periodsWide:            []string{"wichishu", "mchochil’l"},
			erasAbbreviated:        []string{"HY", "YY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hinapiya yesu", "Yopia yesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06,
			fnFmtCurrency:         fnFmtCurrency_1c327a5182399c63873a891da808a068,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mgh_mz"] = func() locales.Translator {
		return &localen{

			locale:                 "mgh_MZ",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "Kwa", "Una", "Rar", "Che", "Tha", "Moc", "Sab", "Nan", "Tis", "Kum", "Moj", "Yel"},
			monthsNarrow:           []string{"", "K", "U", "R", "C", "T", "M", "S", "N", "T", "K", "M", "Y"},
			monthsWide:             []string{"", "Mweri wo kwanza", "Mweri wo unayeli", "Mweri wo uneraru", "Mweri wo unecheshe", "Mweri wo unethanu", "Mweri wo thanu na mocha", "Mweri wo saba", "Mweri wo nane", "Mweri wo tisa", "Mweri wo kumi", "Mweri wo kumi na moja", "Mweri wo kumi na yel’li"},
			daysAbbreviated:        []string{"Sab", "Jtt", "Jnn", "Jtn", "Ara", "Iju", "Jmo"},
			daysNarrow:             []string{"S", "J", "J", "J", "A", "I", "J"},
			daysWide:               []string{"Sabato", "Jumatatu", "Jumanne", "Jumatano", "Arahamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"wichishu", "mchochil’l"},
			periodsWide:            []string{"wichishu", "mchochil’l"},
			erasAbbreviated:        []string{"HY", "YY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Hinapiya yesu", "Yopia yesu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06,
			fnFmtCurrency:         fnFmtCurrency_1c327a5182399c63873a891da808a068,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mgo"] = func() locales.Translator {
		return &localen{

			locale:                 "mgo",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "mbegtug", "imeg àbùbì", "imeg mbəŋchubi", "iməg ngwə̀t", "iməg fog", "iməg ichiibɔd", "iməg àdùmbə̀ŋ", "iməg ichika", "iməg kud", "iməg tèsiʼe", "iməg zò", "iməg krizmed"},
			monthsNarrow:           []string{"", "M1", "A2", "M3", "N4", "F5", "I6", "A7", "I8", "K9", "10", "11", "12"},
			monthsWide:             []string{"", "iməg mbegtug", "imeg àbùbì", "imeg mbəŋchubi", "iməg ngwə̀t", "iməg fog", "iməg ichiibɔd", "iməg àdùmbə̀ŋ", "iməg ichika", "iməg kud", "iməg tèsiʼe", "iməg zò", "iməg krizmed"},
			daysAbbreviated:        []string{"Aneg 1", "Aneg 2", "Aneg 3", "Aneg 4", "Aneg 5", "Aneg 6", "Aneg 7"},
			daysNarrow:             []string{"A1", "A2", "A3", "A4", "A5", "A6", "A7"},
			daysShort:              []string{"1", "2", "3", "4", "5", "6", "7"},
			daysWide:               []string{"Aneg 1", "Aneg 2", "Aneg 3", "Aneg 4", "Aneg 5", "Aneg 6", "Aneg 7"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mgo_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "mgo_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "mbegtug", "imeg àbùbì", "imeg mbəŋchubi", "iməg ngwə̀t", "iməg fog", "iməg ichiibɔd", "iməg àdùmbə̀ŋ", "iməg ichika", "iməg kud", "iməg tèsiʼe", "iməg zò", "iməg krizmed"},
			monthsNarrow:           []string{"", "M1", "A2", "M3", "N4", "F5", "I6", "A7", "I8", "K9", "10", "11", "12"},
			monthsWide:             []string{"", "iməg mbegtug", "imeg àbùbì", "imeg mbəŋchubi", "iməg ngwə̀t", "iməg fog", "iməg ichiibɔd", "iməg àdùmbə̀ŋ", "iməg ichika", "iməg kud", "iməg tèsiʼe", "iməg zò", "iməg krizmed"},
			daysAbbreviated:        []string{"Aneg 1", "Aneg 2", "Aneg 3", "Aneg 4", "Aneg 5", "Aneg 6", "Aneg 7"},
			daysNarrow:             []string{"A1", "A2", "A3", "A4", "A5", "A6", "A7"},
			daysShort:              []string{"1", "2", "3", "4", "5", "6", "7"},
			daysWide:               []string{"Aneg 1", "Aneg 2", "Aneg 3", "Aneg 4", "Aneg 5", "Aneg 6", "Aneg 7"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_9d575487ce5a187f6efc22e6d3f8a058,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mi"] = func() locales.Translator {
		return &localen{

			locale:                 "mi",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Kohi", "Hui", "Pou", "Pae", "Hara", "Pipi", "Hōngo", "Here", "Mahu", "Nuku", "Rangi", "Haki"},
			monthsNarrow:           []string{"", "K", "H", "P", "P", "H", "P", "H", "H", "M", "N", "R", "H"},
			monthsWide:             []string{"", "Kohitātea", "Huitanguru", "Poutūterangi", "Paengawhāwhā", "Haratua", "Pipiri", "Hōngongoi", "Hereturikōkā", "Mahuru", "Whiringa-ā-nuku", "Whiringa-ā-rangi", "Hakihea"},
			daysAbbreviated:        []string{"Tap", "Hin", "Tū", "Apa", "Par", "Mer", "Hor"},
			daysNarrow:             []string{"T", "H", "T", "A", "P", "M", "H"},
			daysShort:              []string{"Tap", "Hin", "Tū", "Apa", "Par", "Mer", "Hor"},
			daysWide:               []string{"Rātapu", "Rāhina", "Rātū", "Rāapa", "Rāpare", "Rāmere", "Rāhoroi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "CE"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Wā Awatea Ranatiki", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Wā Arowhānui Ranatiki", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Wā Awatea Waenga", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Wā Arowhānui Waenga", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Wā Awatea Rāwhiti", "EST": "Wā Arowhānui Rāwhiti", "GFT": "GFT", "GMT": "Wā Toharite Greenwich", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Wā Awatea Maunga", "MESZ": "Wā Raumati Uropi Waenga", "MEZ": "Wā Arowhānui Uropi Waenga", "MST": "Wā Arowhānui Maunga", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Wā Raumati Uropi Rāwhiti", "OEZ": "Wā Arowhānui Uropi Rāwhiti", "PDT": "Wā Awatea Kiwa", "PST": "Wā Arowhānui Kiwa", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Wā Raumati Uropi Uru", "WEZ": "Wā Arowhānui Uropi Uru", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mi_nz"] = func() locales.Translator {
		return &localen{

			locale:                 "mi_NZ",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Kohi", "Hui", "Pou", "Pae", "Hara", "Pipi", "Hōngo", "Here", "Mahu", "Nuku", "Rangi", "Haki"},
			monthsNarrow:           []string{"", "K", "H", "P", "P", "H", "P", "H", "H", "M", "N", "R", "H"},
			monthsWide:             []string{"", "Kohitātea", "Huitanguru", "Poutūterangi", "Paengawhāwhā", "Haratua", "Pipiri", "Hōngongoi", "Hereturikōkā", "Mahuru", "Whiringa-ā-nuku", "Whiringa-ā-rangi", "Hakihea"},
			daysAbbreviated:        []string{"Tap", "Hin", "Tū", "Apa", "Par", "Mer", "Hor"},
			daysNarrow:             []string{"T", "H", "T", "A", "P", "M", "H"},
			daysShort:              []string{"Tap", "Hin", "Tū", "Apa", "Par", "Mer", "Hor"},
			daysWide:               []string{"Rātapu", "Rāhina", "Rātū", "Rāapa", "Rāpare", "Rāmere", "Rāhoroi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "CE"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Wā Awatea Ranatiki", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Wā Arowhānui Ranatiki", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Wā Awatea Waenga", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Wā Arowhānui Waenga", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Wā Awatea Rāwhiti", "EST": "Wā Arowhānui Rāwhiti", "GFT": "GFT", "GMT": "Wā Toharite Greenwich", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Wā Awatea Maunga", "MESZ": "Wā Raumati Uropi Waenga", "MEZ": "Wā Arowhānui Uropi Waenga", "MST": "Wā Arowhānui Maunga", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Wā Raumati Uropi Rāwhiti", "OEZ": "Wā Arowhānui Uropi Rāwhiti", "PDT": "Wā Awatea Kiwa", "PST": "Wā Arowhānui Kiwa", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Wā Raumati Uropi Uru", "WEZ": "Wā Arowhānui Uropi Uru", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mk"] = func() locales.Translator {
		return &localen{

			locale:                 "mk",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 5, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "ден.", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "јан.", "фев.", "мар.", "апр.", "мај", "јун.", "јул.", "авг.", "септ.", "окт.", "ноем.", "дек."},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануари", "февруари", "март", "април", "мај", "јуни", "јули", "август", "септември", "октомври", "ноември", "декември"},
			daysAbbreviated:        []string{"нед.", "пон.", "вт.", "сре.", "чет.", "пет.", "саб."},
			daysNarrow:             []string{"н", "п", "в", "с", "ч", "п", "с"},
			daysShort:              []string{"нед.", "пон.", "вто.", "сре.", "чет.", "пет.", "саб."},
			daysWide:               []string{"недела", "понеделник", "вторник", "среда", "четврток", "петок", "сабота"},
			periodsAbbreviated:     []string{"претпл.", "попл."},
			periodsNarrow:          []string{"претпл.", "попл."},
			periodsWide:            []string{"претпладне", "попладне"},
			erasAbbreviated:        []string{"пр.н.е.", "н.е."},
			erasNarrow:             []string{"пр.н.е.", "н.е."},
			erasWide:               []string{"пред нашата ера", "од нашата ера"},
			timezones:              map[string]string{"ACDT": "Летно сметање на времето во Централна Австралија", "ACST": "Стандардно време во Централна Австралија", "ACWDT": "Летно сметање на времето во Централна и Западна Австралија", "ACWST": "Стандардно време во Централна и Западна Австралија", "ADT": "Атлантско летно сметање на времето", "AEDT": "Летно сметање на времето во Источна Австралија", "AEST": "Стандардно време во Источна Австралија", "AKDT": "Летно сметање на времето во Алјаска", "AKST": "Стандардно време во Алјаска", "ARST": "Летно сметање на времето во Аргентина", "ART": "Стандардно време во Аргентина", "AST": "Атлантско стандардно време", "AWDT": "Летно сметање на времето во Западна Австралија", "AWST": "Стандардно време во Западна Австралија", "BOT": "Време во Боливија", "BT": "Време во Бутан", "CAT": "Средноафриканско време", "CDT": "Централно летно сметање на времето", "CHADT": "Летно сметање на времето во Чатам", "CHAST": "Стандардно време во Чатам", "CLST": "Летно сметање на времето во Чиле", "CLT": "Стандардно време во Чиле", "COST": "Летно сметање на времето во Колумбија", "COT": "Стандардно време во Колумбија", "CST": "Централно стандардно време во Северна Америка", "ChST": "Време во Чаморо", "EAT": "Источноафриканско време", "ECT": "Време во Еквадор", "EDT": "Источно летно сметање на времето", "EST": "Источно стандардно време", "GFT": "Време во Француска Гвајана", "GMT": "Средно време по Гринич", "GST": "Време на Мексиканскиот Залив", "GYT": "Време во Гвајана", "HADT": "Летно сметање на времето во Хаваи - Алеутски острови", "HAST": "Стандардно време во Хаваи - Алеутски острови", "HAT": "Летно сметање на времето на Њуфаундленд", "HECU": "Летно сметање на времето во Куба", "HEEG": "Летно сметање на времето во Источен Гренланд", "HENOMX": "Летно сметање на времето во северозападно Мексико", "HEOG": "Летно сметање на времето во Западен Гренланд", "HEPM": "Летно сметање на времето на Сент Пјер и Микелан", "HEPMX": "Летно пацифичко време во Мексико", "HKST": "Летно време во Хонг Конг", "HKT": "Стандардно време во Хонг Конг", "HNCU": "Стандардно време во Куба", "HNEG": "Стандардно време во Источен Гренланд", "HNNOMX": "Стандардно време во северозападно Мексико", "HNOG": "Стандардно време во Западен Гренланд", "HNPM": "Стандардно време на Сент Пјер и Микелан", "HNPMX": "Стандардно пацифичко време во Мексико", "HNT": "Стандардно време на Њуфаундленд", "IST": "Време во Индија", "JDT": "Летно сметање на времето во Јапонија", "JST": "Стандардно време во Јапонија", "LHDT": "Летно сметање на времето во Лорд Хау", "LHST": "Стандардно време во Лорд Хау", "MDT": "Планинско летно сметање на времето", "MESZ": "Средноевропско летно време", "MEZ": "Средноевропско стандардно време", "MST": "Планинско стандардно време", "MYT": "Време во Малезија", "NZDT": "Летно сметање на времето во Нов Зеланд", "NZST": "Стандардно време во Нов Зеланд", "OESZ": "Источноевропско летно време", "OEZ": "Источноевропско стандардно време", "PDT": "Пацифичко летно сметање на времето", "PST": "Пацифичко стандардно време", "SAST": "Време во Јужноафриканска Република", "SGT": "Време во Сингапур", "SRT": "Време во Суринам", "TMST": "Летно време во Туркменистан", "TMT": "Стандардно време во Туркменистан", "UYST": "Летно сметање на времето во Уругвај", "UYT": "Стандардно време во Уругвај", "VET": "Време во Венецуела", "WARST": "Летно сметање на времето во западна Аргентина", "WART": "Стандардно време во западна Аргентина", "WAST": "Западноафриканско летно сметање на времето", "WAT": "Западноафриканско стандардно време", "WESZ": "Западноевропско летно време", "WEZ": "Западноевропско стандардно време", "WIB": "Време во Западна Индонезија", "WIT": "Време во Источна Индонезија", "WITA": "Време во Централна Индонезија", "∅∅∅": "Летно време на Азорските Острови"},

			fnCardinalPluralRule:  fnCardinalPluralRule_b96558b039ca4fc63138671b418a67f5,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_885e23e22aae7e14ea3420fe83f75ccf,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mk_mk"] = func() locales.Translator {
		return &localen{

			locale:                 "mk_MK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 5, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "јан.", "фев.", "мар.", "апр.", "мај", "јун.", "јул.", "авг.", "септ.", "окт.", "ноем.", "дек."},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануари", "февруари", "март", "април", "мај", "јуни", "јули", "август", "септември", "октомври", "ноември", "декември"},
			daysAbbreviated:        []string{"нед.", "пон.", "вт.", "сре.", "чет.", "пет.", "саб."},
			daysNarrow:             []string{"н", "п", "в", "с", "ч", "п", "с"},
			daysShort:              []string{"нед.", "пон.", "вто.", "сре.", "чет.", "пет.", "саб."},
			daysWide:               []string{"недела", "понеделник", "вторник", "среда", "четврток", "петок", "сабота"},
			periodsAbbreviated:     []string{"претпл.", "попл."},
			periodsNarrow:          []string{"претпл.", "попл."},
			periodsWide:            []string{"претпладне", "попладне"},
			erasAbbreviated:        []string{"пр.н.е.", "н.е."},
			erasNarrow:             []string{"пр.н.е.", "н.е."},
			erasWide:               []string{"пред нашата ера", "од нашата ера"},
			timezones:              map[string]string{"ACDT": "Летно сметање на времето во Централна Австралија", "ACST": "Стандардно време во Централна Австралија", "ACWDT": "Летно сметање на времето во Централна и Западна Австралија", "ACWST": "Стандардно време во Централна и Западна Австралија", "ADT": "Атлантско летно сметање на времето", "AEDT": "Летно сметање на времето во Источна Австралија", "AEST": "Стандардно време во Источна Австралија", "AKDT": "Летно сметање на времето во Алјаска", "AKST": "Стандардно време во Алјаска", "ARST": "Летно сметање на времето во Аргентина", "ART": "Стандардно време во Аргентина", "AST": "Атлантско стандардно време", "AWDT": "Летно сметање на времето во Западна Австралија", "AWST": "Стандардно време во Западна Австралија", "BOT": "Време во Боливија", "BT": "Време во Бутан", "CAT": "Средноафриканско време", "CDT": "Централно летно сметање на времето", "CHADT": "Летно сметање на времето во Чатам", "CHAST": "Стандардно време во Чатам", "CLST": "Летно сметање на времето во Чиле", "CLT": "Стандардно време во Чиле", "COST": "Летно сметање на времето во Колумбија", "COT": "Стандардно време во Колумбија", "CST": "Централно стандардно време во Северна Америка", "ChST": "Време во Чаморо", "EAT": "Источноафриканско време", "ECT": "Време во Еквадор", "EDT": "Источно летно сметање на времето", "EST": "Источно стандардно време", "GFT": "Време во Француска Гвајана", "GMT": "Средно време по Гринич", "GST": "Време на Мексиканскиот Залив", "GYT": "Време во Гвајана", "HADT": "Летно сметање на времето во Хаваи - Алеутски острови", "HAST": "Стандардно време во Хаваи - Алеутски острови", "HAT": "Летно сметање на времето на Њуфаундленд", "HECU": "Летно сметање на времето во Куба", "HEEG": "Летно сметање на времето во Источен Гренланд", "HENOMX": "Летно сметање на времето во северозападно Мексико", "HEOG": "Летно сметање на времето во Западен Гренланд", "HEPM": "Летно сметање на времето на Сент Пјер и Микелан", "HEPMX": "Летно пацифичко време во Мексико", "HKST": "Летно време во Хонг Конг", "HKT": "Стандардно време во Хонг Конг", "HNCU": "Стандардно време во Куба", "HNEG": "Стандардно време во Источен Гренланд", "HNNOMX": "Стандардно време во северозападно Мексико", "HNOG": "Стандардно време во Западен Гренланд", "HNPM": "Стандардно време на Сент Пјер и Микелан", "HNPMX": "Стандардно пацифичко време во Мексико", "HNT": "Стандардно време на Њуфаундленд", "IST": "Време во Индија", "JDT": "Летно сметање на времето во Јапонија", "JST": "Стандардно време во Јапонија", "LHDT": "Летно сметање на времето во Лорд Хау", "LHST": "Стандардно време во Лорд Хау", "MDT": "Планинско летно сметање на времето", "MESZ": "Средноевропско летно време", "MEZ": "Средноевропско стандардно време", "MST": "Планинско стандардно време", "MYT": "Време во Малезија", "NZDT": "Летно сметање на времето во Нов Зеланд", "NZST": "Стандардно време во Нов Зеланд", "OESZ": "Источноевропско летно време", "OEZ": "Источноевропско стандардно време", "PDT": "Пацифичко летно сметање на времето", "PST": "Пацифичко стандардно време", "SAST": "Време во Јужноафриканска Република", "SGT": "Време во Сингапур", "SRT": "Време во Суринам", "TMST": "Летно време во Туркменистан", "TMT": "Стандардно време во Туркменистан", "UYST": "Летно сметање на времето во Уругвај", "UYT": "Стандардно време во Уругвај", "VET": "Време во Венецуела", "WARST": "Летно сметање на времето во западна Аргентина", "WART": "Стандардно време во западна Аргентина", "WAST": "Западноафриканско летно сметање на времето", "WAT": "Западноафриканско стандардно време", "WESZ": "Западноевропско летно време", "WEZ": "Западноевропско стандардно време", "WIB": "Време во Западна Индонезија", "WIT": "Време во Источна Индонезија", "WITA": "Време во Централна Индонезија", "∅∅∅": "Летно сметање на времето во Бразилија"},

			fnCardinalPluralRule:  fnCardinalPluralRule_b96558b039ca4fc63138671b418a67f5,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_ef6a0d569c19898d4747be489ddaa092,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_885e23e22aae7e14ea3420fe83f75ccf,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ml"] = func() locales.Translator {
		return &localen{

			locale:                 "ml",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ജനു", "ഫെബ്രു", "മാർ", "ഏപ്രി", "മേയ്", "ജൂൺ", "ജൂലൈ", "ഓഗ", "സെപ്റ്റം", "ഒക്ടോ", "നവം", "ഡിസം"},
			monthsNarrow:           []string{"", "ജ", "ഫെ", "മാ", "ഏ", "മെ", "ജൂൺ", "ജൂ", "ഓ", "സെ", "ഒ", "ന", "ഡി"},
			monthsWide:             []string{"", "ജനുവരി", "ഫെബ്രുവരി", "മാർച്ച്", "ഏപ്രിൽ", "മേയ്", "ജൂൺ", "ജൂലൈ", "ഓഗസ്റ്റ്", "സെപ്റ്റംബർ", "ഒക്\u200cടോബർ", "നവംബർ", "ഡിസംബർ"},
			daysAbbreviated:        []string{"ഞായർ", "തിങ്കൾ", "ചൊവ്വ", "ബുധൻ", "വ്യാഴം", "വെള്ളി", "ശനി"},
			daysNarrow:             []string{"ഞ", "തി", "ചൊ", "ബു", "വ്യാ", "വെ", "ശ"},
			daysShort:              []string{"ഞാ", "തി", "ചൊ", "ബു", "വ്യാ", "വെ", "ശ"},
			daysWide:               []string{"ഞായറാഴ്\u200cച", "തിങ്കളാഴ്\u200cച", "ചൊവ്വാഴ്ച", "ബുധനാഴ്\u200cച", "വ്യാഴാഴ്\u200cച", "വെള്ളിയാഴ്\u200cച", "ശനിയാഴ്\u200cച"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"ക്രി.മു.", "എഡി"},
			erasNarrow:             []string{"ക്രി.മു.", "എഡി"},
			erasWide:               []string{"ക്രിസ്\u200cതുവിന് മുമ്പ്", "ആന്നോ ഡൊമിനി"},
			timezones:              map[string]string{"ACDT": "ഓസ്ട്രേലിയൻ സെൻട്രൽ ഡേലൈറ്റ് സമയം", "ACST": "ഓസ്ട്രേലിയൻ സെൻട്രൽ സ്റ്റാൻഡേർഡ് സമയം", "ACWDT": "ഓസ്ട്രേലിയൻ സെൻട്രൽ പടിഞ്ഞാറൻ ഡേലൈറ്റ് സമയം", "ACWST": "ഓസ്ട്രേലിയൻ സെൻട്രൽ പടിഞ്ഞാറൻ സ്റ്റാൻഡേർഡ് സമയം", "ADT": "അറ്റ്\u200cലാന്റിക് ഡേലൈറ്റ് സമയം", "AEDT": "ഓസ്\u200cട്രേലിയൻ കിഴക്കൻ ഡേലൈറ്റ് സമയം", "AEST": "ഓസ്\u200cട്രേലിയൻ കിഴക്കൻ സ്റ്റാൻഡേർഡ് സമയം", "AKDT": "അലാസ്\u200cക ഡേലൈറ്റ് സമയം", "AKST": "അലാസ്ക സ്റ്റാൻഡേർഡ് സമയം", "ARST": "അർജന്റീന ഗ്രീഷ്\u200cമകാല സമയം", "ART": "അർജന്റീന സ്റ്റാൻഡേർഡ് സമയം", "AST": "അറ്റ്\u200cലാന്റിക് സ്റ്റാൻഡേർഡ് സമയം", "AWDT": "ഓസ്\u200cട്രേലിയൻ പടിഞ്ഞാറൻ ഡേലൈറ്റ് സമയം", "AWST": "ഓസ്\u200cട്രേലിയൻ പടിഞ്ഞാറൻ സ്റ്റാൻഡേർഡ് സമയം", "BOT": "ബൊളീവിയ സമയം", "BT": "ഭൂട്ടാൻ സമയം", "CAT": "മധ്യ ആഫ്രിക്ക സമയം", "CDT": "വടക്കെ അമേരിക്കൻ സെൻട്രൽ ഡേലൈറ്റ് സമയം", "CHADT": "ചാത്തം ഗ്രീഷ്\u200cമകാല സമയം", "CHAST": "ചാത്തം സ്റ്റാൻഡേർഡ് സമയം", "CLST": "ചിലി ഗ്രീഷ്\u200cമകാല സമയം", "CLT": "ചിലി സ്റ്റാൻഡേർഡ് സമയം", "COST": "കൊളംബിയ ഗ്രീഷ്\u200cമകാല സമയം", "COT": "കൊളംബിയ സ്റ്റാൻഡേർഡ് സമയം", "CST": "വടക്കെ അമേരിക്കൻ സെൻട്രൽ സ്റ്റാൻഡേർഡ് സമയം", "ChST": "ചമോറോ സ്റ്റാൻഡേർഡ് സമയം", "EAT": "കിഴക്കൻ ആഫ്രിക്ക സമയം", "ECT": "ഇക്വഡോർ സമയം", "EDT": "വടക്കെ അമേരിക്കൻ കിഴക്കൻ ഡേലൈറ്റ് സമയം", "EST": "വടക്കെ അമേരിക്കൻ കിഴക്കൻ സ്റ്റാൻഡേർഡ് സമയം", "GFT": "ഫ്രഞ്ച് ഗയാന സമയം", "GMT": "ഗ്രീൻവിച്ച് മീൻ സമയം", "GST": "ഗൾഫ് സ്റ്റാൻഡേർഡ് സമയം", "GYT": "ഗയാന സമയം", "HADT": "ഹവായ്-അലൂഷ്യൻ ഡേലൈറ്റ് സമയം", "HAST": "ഹവായ്-അലൂഷ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "HAT": "ന്യൂഫൗണ്ട്\u200cലാന്റ് ഡേലൈറ്റ് സമയം", "HECU": "ക്യൂബ ഡേലൈറ്റ് സമയം", "HEEG": "കിഴക്കൻ ഗ്രീൻലാൻഡ് ഗ്രീഷ്\u200cമകാല സമയം", "HENOMX": "വടക്കുപടിഞ്ഞാറൻ മെക്സിക്കൻ ഡേലൈറ്റ് സമയം", "HEOG": "പടിഞ്ഞാറൻ ഗ്രീൻലാൻഡ് ഗ്രീഷ്\u200cമകാല സമയം", "HEPM": "സെന്റ് പിയറി ആൻഡ് മിക്വലൻ ഡേലൈറ്റ് സമയം", "HEPMX": "മെക്സിക്കൻ പസഫിക് ഡേലൈറ്റ് സമയം", "HKST": "ഹോങ്കോങ്ങ് ഗ്രീഷ്\u200cമകാല സമയം", "HKT": "ഹോങ്കോങ്ങ് സ്റ്റാൻഡേർഡ് സമയം", "HNCU": "ക്യൂബ സ്റ്റാൻഡേർഡ് സമയം", "HNEG": "കിഴക്കൻ ഗ്രീൻലാൻഡ് സ്റ്റാൻഡേർഡ് സമയം", "HNNOMX": "വടക്കുപടിഞ്ഞാറൻ മെക്\u200cസിക്കൻ സ്റ്റാൻഡേർഡ് സമയം", "HNOG": "പടിഞ്ഞാറൻ ഗ്രീൻലാൻഡ് സ്റ്റാൻഡേർഡ് സമയം", "HNPM": "സെന്റ് പിയറി ആൻഡ് മിക്വലൻ സ്റ്റാൻഡേർഡ് സമയം", "HNPMX": "മെക്\u200cസിക്കൻ പസഫിക് സ്റ്റാൻഡേർഡ് സമയം", "HNT": "ന്യൂഫൗണ്ട്\u200cലാന്റ് സ്റ്റാൻഡേർഡ് സമയം", "IST": "ഇന്ത്യൻ സ്റ്റാൻഡേർഡ് സമയം", "JDT": "ജപ്പാൻ ഡേലൈറ്റ് സമയം", "JST": "ജപ്പാൻ സ്റ്റാൻഡേർഡ് സമയം", "LHDT": "ലോർഡ് ഹോവ് ഡേലൈറ്റ് സമയം", "LHST": "ലോർഡ് ഹോവ് സ്റ്റാൻഡേർഡ് സമയം", "MDT": "വടക്കെ അമേരിക്കൻ മൗണ്ടൻ ഡേലൈറ്റ് സമയം", "MESZ": "സെൻട്രൽ യൂറോപ്യൻ ഗ്രീഷ്മകാല സമയം", "MEZ": "സെൻട്രൽ യൂറോപ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "MST": "വടക്കെ അമേരിക്കൻ മൗണ്ടൻ സ്റ്റാൻഡേർഡ് സമയം", "MYT": "മലേഷ്യ സമയം", "NZDT": "ന്യൂസിലാൻഡ് ഡേലൈറ്റ് സമയം", "NZST": "ന്യൂസിലാൻഡ് സ്റ്റാൻഡേർഡ് സമയം", "OESZ": "കിഴക്കൻ യൂറോപ്യൻ ഗ്രീഷ്മകാല സമയം", "OEZ": "കിഴക്കൻ യൂറോപ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "PDT": "വടക്കെ അമേരിക്കൻ പസഫിക് ഡേലൈറ്റ് സമയം", "PST": "വടക്കെ അമേരിക്കൻ പസഫിക് സ്റ്റാൻഡേർഡ് സമയം", "SAST": "ദക്ഷിണാഫ്രിക്ക സ്റ്റാൻഡേർഡ് സമയം", "SGT": "സിംഗപ്പൂർ സ്റ്റാൻഡേർഡ് സമയം", "SRT": "സുരിനെയിം സമയം", "TMST": "തുർക്ക്\u200cമെനിസ്ഥാൻ ഗ്രീഷ്\u200cമകാല സമയം", "TMT": "തുർക്ക്\u200cമെനിസ്ഥാൻ സ്റ്റാൻഡേർഡ് സമയം", "UYST": "ഉറുഗ്വേ ഗ്രീഷ്\u200cമകാല സമയം", "UYT": "ഉറുഗ്വേ സ്റ്റാൻഡേർഡ് സമയം", "VET": "വെനിസ്വേല സമയം", "WARST": "പടിഞ്ഞാറൻ അർജന്റീന ഗ്രീഷ്\u200cമകാല സമയം", "WART": "പടിഞ്ഞാറൻ അർജന്റീന സ്റ്റാൻഡേർഡ് സമയം", "WAST": "പടിഞ്ഞാറൻ ആഫ്രിക്ക ഗ്രീഷ്\u200cമകാല സമയം", "WAT": "പടിഞ്ഞാറൻ ആഫ്രിക്ക സ്റ്റാൻഡേർഡ് സമയം", "WESZ": "പടിഞ്ഞാറൻ യൂറോപ്യൻ ഗ്രീഷ്\u200cമകാല സമയം", "WEZ": "പടിഞ്ഞാറൻ യൂറോപ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "WIB": "പടിഞ്ഞാറൻ ഇന്തോനേഷ്യ സമയം", "WIT": "കിഴക്കൻ ഇന്തോനേഷ്യ സമയം", "WITA": "മധ്യ ഇന്തോനേഷ്യ സമയം", "∅∅∅": "ബ്രസീലിയ ഗ്രീഷ്\u200cമകാല സമയം"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_74d5e00c4bf6ea6ef8db9baad7610c73,
			fnFmtDateLong:         fnFmtDateLong_ab2127ffd2f4043ae38d2f998ae282a6,
			fnFmtDateMedium:       fnFmtDateMedium_fdbcfd5a321e6370f61bce682fe65b6e,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ml_in"] = func() locales.Translator {
		return &localen{

			locale:                 "ml_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ജനു", "ഫെബ്രു", "മാർ", "ഏപ്രി", "മേയ്", "ജൂൺ", "ജൂലൈ", "ഓഗ", "സെപ്റ്റം", "ഒക്ടോ", "നവം", "ഡിസം"},
			monthsNarrow:           []string{"", "ജ", "ഫെ", "മാ", "ഏ", "മെ", "ജൂൺ", "ജൂ", "ഓ", "സെ", "ഒ", "ന", "ഡി"},
			monthsWide:             []string{"", "ജനുവരി", "ഫെബ്രുവരി", "മാർച്ച്", "ഏപ്രിൽ", "മേയ്", "ജൂൺ", "ജൂലൈ", "ഓഗസ്റ്റ്", "സെപ്റ്റംബർ", "ഒക്\u200cടോബർ", "നവംബർ", "ഡിസംബർ"},
			daysAbbreviated:        []string{"ഞായർ", "തിങ്കൾ", "ചൊവ്വ", "ബുധൻ", "വ്യാഴം", "വെള്ളി", "ശനി"},
			daysNarrow:             []string{"ഞ", "തി", "ചൊ", "ബു", "വ്യാ", "വെ", "ശ"},
			daysShort:              []string{"ഞാ", "തി", "ചൊ", "ബു", "വ്യാ", "വെ", "ശ"},
			daysWide:               []string{"ഞായറാഴ്\u200cച", "തിങ്കളാഴ്\u200cച", "ചൊവ്വാഴ്ച", "ബുധനാഴ്\u200cച", "വ്യാഴാഴ്\u200cച", "വെള്ളിയാഴ്\u200cച", "ശനിയാഴ്\u200cച"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"ക്രി.മു.", "എഡി"},
			erasNarrow:             []string{"ക്രി.മു.", "എഡി"},
			erasWide:               []string{"ക്രിസ്\u200cതുവിന് മുമ്പ്", "ആന്നോ ഡൊമിനി"},
			timezones:              map[string]string{"ACDT": "ഓസ്ട്രേലിയൻ സെൻട്രൽ ഡേലൈറ്റ് സമയം", "ACST": "ഓസ്ട്രേലിയൻ സെൻട്രൽ സ്റ്റാൻഡേർഡ് സമയം", "ACWDT": "ഓസ്ട്രേലിയൻ സെൻട്രൽ പടിഞ്ഞാറൻ ഡേലൈറ്റ് സമയം", "ACWST": "ഓസ്ട്രേലിയൻ സെൻട്രൽ പടിഞ്ഞാറൻ സ്റ്റാൻഡേർഡ് സമയം", "ADT": "അറ്റ്\u200cലാന്റിക് ഡേലൈറ്റ് സമയം", "AEDT": "ഓസ്\u200cട്രേലിയൻ കിഴക്കൻ ഡേലൈറ്റ് സമയം", "AEST": "ഓസ്\u200cട്രേലിയൻ കിഴക്കൻ സ്റ്റാൻഡേർഡ് സമയം", "AKDT": "അലാസ്\u200cക ഡേലൈറ്റ് സമയം", "AKST": "അലാസ്ക സ്റ്റാൻഡേർഡ് സമയം", "ARST": "അർജന്റീന ഗ്രീഷ്\u200cമകാല സമയം", "ART": "അർജന്റീന സ്റ്റാൻഡേർഡ് സമയം", "AST": "അറ്റ്\u200cലാന്റിക് സ്റ്റാൻഡേർഡ് സമയം", "AWDT": "ഓസ്\u200cട്രേലിയൻ പടിഞ്ഞാറൻ ഡേലൈറ്റ് സമയം", "AWST": "ഓസ്\u200cട്രേലിയൻ പടിഞ്ഞാറൻ സ്റ്റാൻഡേർഡ് സമയം", "BOT": "ബൊളീവിയ സമയം", "BT": "ഭൂട്ടാൻ സമയം", "CAT": "മധ്യ ആഫ്രിക്ക സമയം", "CDT": "വടക്കെ അമേരിക്കൻ സെൻട്രൽ ഡേലൈറ്റ് സമയം", "CHADT": "ചാത്തം ഗ്രീഷ്\u200cമകാല സമയം", "CHAST": "ചാത്തം സ്റ്റാൻഡേർഡ് സമയം", "CLST": "ചിലി ഗ്രീഷ്\u200cമകാല സമയം", "CLT": "ചിലി സ്റ്റാൻഡേർഡ് സമയം", "COST": "കൊളംബിയ ഗ്രീഷ്\u200cമകാല സമയം", "COT": "കൊളംബിയ സ്റ്റാൻഡേർഡ് സമയം", "CST": "വടക്കെ അമേരിക്കൻ സെൻട്രൽ സ്റ്റാൻഡേർഡ് സമയം", "ChST": "ചമോറോ സ്റ്റാൻഡേർഡ് സമയം", "EAT": "കിഴക്കൻ ആഫ്രിക്ക സമയം", "ECT": "ഇക്വഡോർ സമയം", "EDT": "വടക്കെ അമേരിക്കൻ കിഴക്കൻ ഡേലൈറ്റ് സമയം", "EST": "വടക്കെ അമേരിക്കൻ കിഴക്കൻ സ്റ്റാൻഡേർഡ് സമയം", "GFT": "ഫ്രഞ്ച് ഗയാന സമയം", "GMT": "ഗ്രീൻവിച്ച് മീൻ സമയം", "GST": "ഗൾഫ് സ്റ്റാൻഡേർഡ് സമയം", "GYT": "ഗയാന സമയം", "HADT": "ഹവായ്-അലൂഷ്യൻ ഡേലൈറ്റ് സമയം", "HAST": "ഹവായ്-അലൂഷ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "HAT": "ന്യൂഫൗണ്ട്\u200cലാന്റ് ഡേലൈറ്റ് സമയം", "HECU": "ക്യൂബ ഡേലൈറ്റ് സമയം", "HEEG": "കിഴക്കൻ ഗ്രീൻലാൻഡ് ഗ്രീഷ്\u200cമകാല സമയം", "HENOMX": "വടക്കുപടിഞ്ഞാറൻ മെക്സിക്കൻ ഡേലൈറ്റ് സമയം", "HEOG": "പടിഞ്ഞാറൻ ഗ്രീൻലാൻഡ് ഗ്രീഷ്\u200cമകാല സമയം", "HEPM": "സെന്റ് പിയറി ആൻഡ് മിക്വലൻ ഡേലൈറ്റ് സമയം", "HEPMX": "മെക്സിക്കൻ പസഫിക് ഡേലൈറ്റ് സമയം", "HKST": "ഹോങ്കോങ്ങ് ഗ്രീഷ്\u200cമകാല സമയം", "HKT": "ഹോങ്കോങ്ങ് സ്റ്റാൻഡേർഡ് സമയം", "HNCU": "ക്യൂബ സ്റ്റാൻഡേർഡ് സമയം", "HNEG": "കിഴക്കൻ ഗ്രീൻലാൻഡ് സ്റ്റാൻഡേർഡ് സമയം", "HNNOMX": "വടക്കുപടിഞ്ഞാറൻ മെക്\u200cസിക്കൻ സ്റ്റാൻഡേർഡ് സമയം", "HNOG": "പടിഞ്ഞാറൻ ഗ്രീൻലാൻഡ് സ്റ്റാൻഡേർഡ് സമയം", "HNPM": "സെന്റ് പിയറി ആൻഡ് മിക്വലൻ സ്റ്റാൻഡേർഡ് സമയം", "HNPMX": "മെക്\u200cസിക്കൻ പസഫിക് സ്റ്റാൻഡേർഡ് സമയം", "HNT": "ന്യൂഫൗണ്ട്\u200cലാന്റ് സ്റ്റാൻഡേർഡ് സമയം", "IST": "ഇന്ത്യൻ സ്റ്റാൻഡേർഡ് സമയം", "JDT": "ജപ്പാൻ ഡേലൈറ്റ് സമയം", "JST": "ജപ്പാൻ സ്റ്റാൻഡേർഡ് സമയം", "LHDT": "ലോർഡ് ഹോവ് ഡേലൈറ്റ് സമയം", "LHST": "ലോർഡ് ഹോവ് സ്റ്റാൻഡേർഡ് സമയം", "MDT": "വടക്കെ അമേരിക്കൻ മൗണ്ടൻ ഡേലൈറ്റ് സമയം", "MESZ": "സെൻട്രൽ യൂറോപ്യൻ ഗ്രീഷ്മകാല സമയം", "MEZ": "സെൻട്രൽ യൂറോപ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "MST": "വടക്കെ അമേരിക്കൻ മൗണ്ടൻ സ്റ്റാൻഡേർഡ് സമയം", "MYT": "മലേഷ്യ സമയം", "NZDT": "ന്യൂസിലാൻഡ് ഡേലൈറ്റ് സമയം", "NZST": "ന്യൂസിലാൻഡ് സ്റ്റാൻഡേർഡ് സമയം", "OESZ": "കിഴക്കൻ യൂറോപ്യൻ ഗ്രീഷ്മകാല സമയം", "OEZ": "കിഴക്കൻ യൂറോപ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "PDT": "വടക്കെ അമേരിക്കൻ പസഫിക് ഡേലൈറ്റ് സമയം", "PST": "വടക്കെ അമേരിക്കൻ പസഫിക് സ്റ്റാൻഡേർഡ് സമയം", "SAST": "ദക്ഷിണാഫ്രിക്ക സ്റ്റാൻഡേർഡ് സമയം", "SGT": "സിംഗപ്പൂർ സ്റ്റാൻഡേർഡ് സമയം", "SRT": "സുരിനെയിം സമയം", "TMST": "തുർക്ക്\u200cമെനിസ്ഥാൻ ഗ്രീഷ്\u200cമകാല സമയം", "TMT": "തുർക്ക്\u200cമെനിസ്ഥാൻ സ്റ്റാൻഡേർഡ് സമയം", "UYST": "ഉറുഗ്വേ ഗ്രീഷ്\u200cമകാല സമയം", "UYT": "ഉറുഗ്വേ സ്റ്റാൻഡേർഡ് സമയം", "VET": "വെനിസ്വേല സമയം", "WARST": "പടിഞ്ഞാറൻ അർജന്റീന ഗ്രീഷ്\u200cമകാല സമയം", "WART": "പടിഞ്ഞാറൻ അർജന്റീന സ്റ്റാൻഡേർഡ് സമയം", "WAST": "പടിഞ്ഞാറൻ ആഫ്രിക്ക ഗ്രീഷ്\u200cമകാല സമയം", "WAT": "പടിഞ്ഞാറൻ ആഫ്രിക്ക സ്റ്റാൻഡേർഡ് സമയം", "WESZ": "പടിഞ്ഞാറൻ യൂറോപ്യൻ ഗ്രീഷ്\u200cമകാല സമയം", "WEZ": "പടിഞ്ഞാറൻ യൂറോപ്യൻ സ്റ്റാൻഡേർഡ് സമയം", "WIB": "പടിഞ്ഞാറൻ ഇന്തോനേഷ്യ സമയം", "WIT": "കിഴക്കൻ ഇന്തോനേഷ്യ സമയം", "WITA": "മധ്യ ഇന്തോനേഷ്യ സമയം", "∅∅∅": "ബ്രസീലിയ ഗ്രീഷ്\u200cമകാല സമയം"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_74d5e00c4bf6ea6ef8db9baad7610c73,
			fnFmtDateLong:         fnFmtDateLong_ab2127ffd2f4043ae38d2f998ae282a6,
			fnFmtDateMedium:       fnFmtDateMedium_fdbcfd5a321e6370f61bce682fe65b6e,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mn"] = func() locales.Translator {
		return &localen{

			locale:                 "mn",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "₮", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "1-р сар", "2-р сар", "3-р сар", "4-р сар", "5-р сар", "6-р сар", "7-р сар", "8-р сар", "9-р сар", "10-р сар", "11-р сар", "12-р сар"},
			monthsNarrow:           []string{"", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X", "XI", "XII"},
			monthsWide:             []string{"", "нэгдүгээр сар", "хоёрдугаар сар", "гуравдугаар сар", "дөрөвдүгээр сар", "тавдугаар сар", "зургаадугаар сар", "долоодугаар сар", "наймдугаар сар", "есдүгээр сар", "аравдугаар сар", "арван нэгдүгээр сар", "арван хоёрдугаар сар"},
			daysAbbreviated:        []string{"Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"},
			daysNarrow:             []string{"Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"},
			daysShort:              []string{"Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"},
			daysWide:               []string{"ням", "даваа", "мягмар", "лхагва", "пүрэв", "баасан", "бямба"},
			periodsAbbreviated:     []string{"ү.ө.", "ү.х."},
			periodsNarrow:          []string{"ү.ө.", "ү.х."},
			periodsWide:            []string{"ү.ө.", "ү.х."},
			erasAbbreviated:        []string{"МЭӨ", "МЭ"},
			erasNarrow:             []string{"МЭӨ", "МЭ"},
			erasWide:               []string{"манай эриний өмнөх", "манай эриний"},
			timezones:              map[string]string{"ACDT": "Төв Австралийн зуны цаг", "ACST": "Төв Австралийн стандарт цаг", "ACWDT": "Австралийн төв баруун эргийн зуны цаг", "ACWST": "Австралийн төв баруун эргийн стандарт цаг", "ADT": "Атлантын зуны цаг", "AEDT": "Австралийн зүүн эргийн зуны цаг", "AEST": "Австралийн зүүн эргийн стандарт цаг", "AKDT": "Аляскийн зуны цаг", "AKST": "Аляскийн стандарт цаг", "ARST": "Аргентины зуны цаг", "ART": "Аргентины стандарт цаг", "AST": "Атлантын стандарт цаг", "AWDT": "Австралийн баруун эргийн зуны цаг", "AWST": "Австралийн баруун эргийн стандарт цаг", "BOT": "Боливийн цаг", "BT": "Бутаны цаг", "CAT": "Төв Африкийн цаг", "CDT": "Төв зуны цаг", "CHADT": "Чатемын зуны цаг", "CHAST": "Чатемын стандарт цаг", "CLST": "Чилийн зуны цаг", "CLT": "Чилийн стандарт цаг", "COST": "Колумбын зуны цаг", "COT": "Колумбын стандарт цаг", "CST": "Төв стандарт цаг", "ChST": "Чаморрогийн цаг", "EAT": "Зүүн Африкийн цаг", "ECT": "Эквадорын цаг", "EDT": "Зүүн эргийн зуны цаг", "EST": "Зүүн эргийн стандарт цаг", "GFT": "Францын Гвианагийн цаг", "GMT": "Гринвичийн цаг", "GST": "Персийн булангийн цаг", "GYT": "Гайанагийн цаг", "HADT": "Хавай-Алеутын зуны цаг", "HAST": "Хавай-Алеутын стандарт цаг", "HAT": "Нью-Фаундлендын зуны цаг", "HECU": "Кубын зуны цаг", "HEEG": "Зүүн Гренландын зуны цаг", "HENOMX": "Баруун хойд Мексикийн зуны цаг", "HEOG": "Баруун Гренландын зуны цаг", "HEPM": "Сент-Пьер ба Микелоны зуны цаг", "HEPMX": "Мексик-Номхон далайн зуны цаг", "HKST": "Хонг Конгийн зуны цаг", "HKT": "Хонг Конгийн стандарт цаг", "HNCU": "Кубын стандарт цаг", "HNEG": "Зүүн Гренландын стандарт цаг", "HNNOMX": "Баруун хойд Мексикийн стандарт цаг", "HNOG": "Баруун Гренландын стандарт цаг", "HNPM": "Сент-Пьер ба Микелоны стандарт цаг", "HNPMX": "Мексик-Номхон далайн стандарт цаг", "HNT": "Нью-Фаундлендын стандарт цаг", "IST": "Энэтхэгийн цаг", "JDT": "Японы зуны цаг", "JST": "Японы стандарт цаг", "LHDT": "Лорд Хоугийн зуны цаг", "LHST": "Лорд Хоугийн стандарт цаг", "MDT": "Уулын зуны цаг", "MESZ": "Төв Европын зуны цаг", "MEZ": "Төв Европын стандарт цаг", "MST": "Уулын стандарт цаг", "MYT": "Малайзын цаг", "NZDT": "Шинэ Зеландын зуны цаг", "NZST": "Шинэ Зеландын стандарт цаг", "OESZ": "Зүүн Европын зуны цаг", "OEZ": "Зүүн Европын стандарт цаг", "PDT": "Номхон далайн зуны цаг", "PST": "Номхон далайн стандарт цаг", "SAST": "Өмнөд Африкийн стандарт цаг", "SGT": "Сингапурын цаг", "SRT": "Суринамын цаг", "TMST": "Туркменистаны зуны цаг", "TMT": "Туркменистаны стандарт цаг", "UYST": "Уругвайн зуны цаг", "UYT": "Уругвайн стандарт цаг", "VET": "Венесуэлийн цаг", "WARST": "Баруун Аргентины зуны цаг", "WART": "Баруун Аргентины стандарт цаг", "WAST": "Баруун Африкийн зуны цаг", "WAT": "Баруун Африкийн стандарт цаг", "WESZ": "Баруун Европын зуны цаг", "WEZ": "Баруун Европын стандарт цаг", "WIB": "Баруун Индонезийн цаг", "WIT": "Зүүн Индонезийн цаг", "WITA": "Төв Индонезийн цаг", "∅∅∅": "Бразилийн зуны цаг"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_4052efc783402567c61ad2003f35b9bf,
			fnFmtDateLong:         fnFmtDateLong_7545f2e6128dde4c84d462009151629b,
			fnFmtDateMedium:       fnFmtDateMedium_34dcbbdacb45f505155cca4c18ec48c0,
			fnFmtDateShort:        fnFmtDateShort_cc33621dc8ea4b3f7e42320aa653cbdf,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_30191015e11c1c183d8d41d8e0e78b42,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mn_mn"] = func() locales.Translator {
		return &localen{

			locale:                 "mn_MN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "1-р сар", "2-р сар", "3-р сар", "4-р сар", "5-р сар", "6-р сар", "7-р сар", "8-р сар", "9-р сар", "10-р сар", "11-р сар", "12-р сар"},
			monthsNarrow:           []string{"", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X", "XI", "XII"},
			monthsWide:             []string{"", "нэгдүгээр сар", "хоёрдугаар сар", "гуравдугаар сар", "дөрөвдүгээр сар", "тавдугаар сар", "зургаадугаар сар", "долоодугаар сар", "наймдугаар сар", "есдүгээр сар", "аравдугаар сар", "арван нэгдүгээр сар", "арван хоёрдугаар сар"},
			daysAbbreviated:        []string{"Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"},
			daysNarrow:             []string{"Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"},
			daysShort:              []string{"Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"},
			daysWide:               []string{"ням", "даваа", "мягмар", "лхагва", "пүрэв", "баасан", "бямба"},
			periodsAbbreviated:     []string{"ү.ө.", "ү.х."},
			periodsNarrow:          []string{"ү.ө.", "ү.х."},
			periodsWide:            []string{"ү.ө.", "ү.х."},
			erasAbbreviated:        []string{"МЭӨ", "МЭ"},
			erasNarrow:             []string{"МЭӨ", "МЭ"},
			erasWide:               []string{"манай эриний өмнөх", "манай эриний"},
			timezones:              map[string]string{"ACDT": "Төв Австралийн зуны цаг", "ACST": "Төв Австралийн стандарт цаг", "ACWDT": "Австралийн төв баруун эргийн зуны цаг", "ACWST": "Австралийн төв баруун эргийн стандарт цаг", "ADT": "Атлантын зуны цаг", "AEDT": "Австралийн зүүн эргийн зуны цаг", "AEST": "Австралийн зүүн эргийн стандарт цаг", "AKDT": "Аляскийн зуны цаг", "AKST": "Аляскийн стандарт цаг", "ARST": "Аргентины зуны цаг", "ART": "Аргентины стандарт цаг", "AST": "Атлантын стандарт цаг", "AWDT": "Австралийн баруун эргийн зуны цаг", "AWST": "Австралийн баруун эргийн стандарт цаг", "BOT": "Боливийн цаг", "BT": "Бутаны цаг", "CAT": "Төв Африкийн цаг", "CDT": "Төв зуны цаг", "CHADT": "Чатемын зуны цаг", "CHAST": "Чатемын стандарт цаг", "CLST": "Чилийн зуны цаг", "CLT": "Чилийн стандарт цаг", "COST": "Колумбын зуны цаг", "COT": "Колумбын стандарт цаг", "CST": "Төв стандарт цаг", "ChST": "Чаморрогийн цаг", "EAT": "Зүүн Африкийн цаг", "ECT": "Эквадорын цаг", "EDT": "Зүүн эргийн зуны цаг", "EST": "Зүүн эргийн стандарт цаг", "GFT": "Францын Гвианагийн цаг", "GMT": "Гринвичийн цаг", "GST": "Персийн булангийн цаг", "GYT": "Гайанагийн цаг", "HADT": "Хавай-Алеутын зуны цаг", "HAST": "Хавай-Алеутын стандарт цаг", "HAT": "Нью-Фаундлендын зуны цаг", "HECU": "Кубын зуны цаг", "HEEG": "Зүүн Гренландын зуны цаг", "HENOMX": "Баруун хойд Мексикийн зуны цаг", "HEOG": "Баруун Гренландын зуны цаг", "HEPM": "Сент-Пьер ба Микелоны зуны цаг", "HEPMX": "Мексик-Номхон далайн зуны цаг", "HKST": "Хонг Конгийн зуны цаг", "HKT": "Хонг Конгийн стандарт цаг", "HNCU": "Кубын стандарт цаг", "HNEG": "Зүүн Гренландын стандарт цаг", "HNNOMX": "Баруун хойд Мексикийн стандарт цаг", "HNOG": "Баруун Гренландын стандарт цаг", "HNPM": "Сент-Пьер ба Микелоны стандарт цаг", "HNPMX": "Мексик-Номхон далайн стандарт цаг", "HNT": "Нью-Фаундлендын стандарт цаг", "IST": "Энэтхэгийн цаг", "JDT": "Японы зуны цаг", "JST": "Японы стандарт цаг", "LHDT": "Лорд Хоугийн зуны цаг", "LHST": "Лорд Хоугийн стандарт цаг", "MDT": "Уулын зуны цаг", "MESZ": "Төв Европын зуны цаг", "MEZ": "Төв Европын стандарт цаг", "MST": "Уулын стандарт цаг", "MYT": "Малайзын цаг", "NZDT": "Шинэ Зеландын зуны цаг", "NZST": "Шинэ Зеландын стандарт цаг", "OESZ": "Зүүн Европын зуны цаг", "OEZ": "Зүүн Европын стандарт цаг", "PDT": "Номхон далайн зуны цаг", "PST": "Номхон далайн стандарт цаг", "SAST": "Өмнөд Африкийн стандарт цаг", "SGT": "Сингапурын цаг", "SRT": "Суринамын цаг", "TMST": "Туркменистаны зуны цаг", "TMT": "Туркменистаны стандарт цаг", "UYST": "Уругвайн зуны цаг", "UYT": "Уругвайн стандарт цаг", "VET": "Венесуэлийн цаг", "WARST": "Баруун Аргентины зуны цаг", "WART": "Баруун Аргентины стандарт цаг", "WAST": "Баруун Африкийн зуны цаг", "WAT": "Баруун Африкийн стандарт цаг", "WESZ": "Баруун Европын зуны цаг", "WEZ": "Баруун Европын стандарт цаг", "WIB": "Баруун Индонезийн цаг", "WIT": "Зүүн Индонезийн цаг", "WITA": "Төв Индонезийн цаг", "∅∅∅": "Бразилийн зуны цаг"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_4052efc783402567c61ad2003f35b9bf,
			fnFmtDateLong:         fnFmtDateLong_7545f2e6128dde4c84d462009151629b,
			fnFmtDateMedium:       fnFmtDateMedium_34dcbbdacb45f505155cca4c18ec48c0,
			fnFmtDateShort:        fnFmtDateShort_cc33621dc8ea4b3f7e42320aa653cbdf,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_30191015e11c1c183d8d41d8e0e78b42,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mr"] = func() locales.Translator {
		return &localen{

			locale:             "mr",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "जाने", "फेब्रु", "मार्च", "एप्रि", "मे", "जून", "जुलै", "ऑग", "सप्टें", "ऑक्टो", "नोव्हें", "डिसें"},
			monthsNarrow:       []string{"", "जा", "फे", "मा", "ए", "मे", "जू", "जु", "ऑ", "स", "ऑ", "नो", "डि"},
			monthsWide:         []string{"", "जानेवारी", "फेब्रुवारी", "मार्च", "एप्रिल", "मे", "जून", "जुलै", "ऑगस्ट", "सप्टेंबर", "ऑक्टोबर", "नोव्हेंबर", "डिसेंबर"},
			daysAbbreviated:    []string{"रवि", "सोम", "मंगळ", "बुध", "गुरु", "शुक्र", "शनि"},
			daysNarrow:         []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysShort:          []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysWide:           []string{"रविवार", "सोमवार", "मंगळवार", "बुधवार", "गुरुवार", "शुक्रवार", "शनिवार"},
			periodsAbbreviated: []string{"म.पू.", "म.उ."},
			periodsNarrow:      []string{"स", "सं"},
			periodsWide:        []string{"म.पू.", "म.उ."},
			erasAbbreviated:    []string{"इ. स. पू.", "इ. स."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ईसवीसनपूर्व", "ईसवीसन"},
			timezones:          map[string]string{"ACDT": "ऑस्ट्रेलियन मध्य सूर्यप्रकाश वेळ", "ACST": "ऑस्ट्रेलियन मध्य प्रमाण वेळ", "ACWDT": "ऑस्ट्रेलियन मध्य-पश्चिम सूर्यप्रकाश वेळ", "ACWST": "ऑस्ट्रेलियन मध्य-पश्चिम प्रमाण वेळ", "ADT": "अटलांटिक सूर्यप्रकाश वेळ", "AEDT": "ऑस्ट्रेलियन पूर्व सूर्यप्रकाश वेळ", "AEST": "ऑस्ट्रेलियन पूर्व प्रमाण वेळ", "AKDT": "अलास्का सूर्यप्रकाश वेळ", "AKST": "अलास्का प्रमाण वेळ", "ARST": "अर्जेंटिना उन्हाळी वेळ", "ART": "अर्जेंटिना प्रमाण वेळ", "AST": "अटलांटिक प्रमाण वेळ", "AWDT": "ऑस्ट्रेलियन पश्चिम सूर्यप्रकाश वेळ", "AWST": "ऑस्ट्रेलियन पश्चिम प्रमाण वेळ", "BOT": "बोलिव्हिया वेळ", "BT": "भूतान वेळ", "CAT": "मध्\u200dय आफ्रिका वेळ", "CDT": "केंद्रीय सूर्यप्रकाश वेळ", "CHADT": "चॅथम सूर्यप्रकाश वेळ", "CHAST": "चॅथम प्रमाण वेळ", "CLST": "चिली उन्हाळी वेळ", "CLT": "चिली प्रमाण वेळ", "COST": "कोलंबिया उन्हाळी वेळ", "COT": "कोलंबिया प्रमाण वेळ", "CST": "केंद्रीय प्रमाण वेळ", "ChST": "चामोरो प्रमाण वेळ", "EAT": "पूर्व आफ्रिका वेळ", "ECT": "इक्वेडोर वेळ", "EDT": "पौर्वात्य सूर्यप्रकाश वेळ", "EST": "पौर्वात्य प्रमाण वेळ", "GFT": "फ्रेंच गयाना वेळ", "GMT": "ग्रीनिच प्रमाण वेळ", "GST": "खाडी प्रमाण वेळ", "GYT": "गयाना वेळ", "HADT": "हवाई-अलूशन सूर्यप्रकाश वेळ", "HAST": "हवाई-अलूशन प्रमाण वेळ", "HAT": "न्यू फाउंडलंड सूर्यप्रकाश वेळ", "HECU": "क्यूबा सूर्यप्रकाश वेळ", "HEEG": "पूर्व ग्रीनलँड उन्हाळी वेळ", "HENOMX": "वायव्य मेक्सिको सूर्यप्रकाश वेळ", "HEOG": "पश्चिम ग्रीनलँड उन्हाळी वेळ", "HEPM": "सेंट पियरे आणि मिक्वेलोन सूर्यप्रकाश वेळ", "HEPMX": "मेक्सिको पॅसिफिक सूर्यप्रकाश वेळ", "HKST": "हाँग काँग उन्हाळी वेळ", "HKT": "हाँग काँग प्रमाण वेळ", "HNCU": "क्यूबा प्रमाण वेळ", "HNEG": "पूर्व ग्रीनलँड प्रमाण वेळ", "HNNOMX": "वायव्य मेक्सिको प्रमाण वेळ", "HNOG": "पश्चिम ग्रीनलँड प्रमाण वेळ", "HNPM": "सेंट पियरे आणि मिक्वेलोन प्रमाण वेळ", "HNPMX": "मेक्सिको पॅसिफिक प्रमाण वेळ", "HNT": "न्यू फाउंडलंड प्रमाण वेळ", "IST": "भारतीय प्रमाण वेळ", "JDT": "जपान सूर्यप्रकाश वेळ", "JST": "जपान प्रमाण वेळ", "LHDT": "लॉर्ड होवे सूर्यप्रकाश वेळ", "LHST": "लॉर्ड होवे प्रमाण वेळ", "MDT": "पर्वतीय सूर्यप्रकाश वेळ", "MESZ": "मध्\u200dय युरोपियन उन्हाळी वेळ", "MEZ": "मध्\u200dय युरोपियन प्रमाण वेळ", "MST": "पर्वतीय प्रमाण वेळ", "MYT": "मलेशिया वेळ", "NZDT": "न्यूझीलंड सूर्यप्रकाश वेळ", "NZST": "न्यूझीलंड प्रमाण वेळ", "OESZ": "पूर्व युरोपियन उन्हाळी वेळ", "OEZ": "पूर्व युरोपियन प्रमाण वेळ", "PDT": "पॅसिफिक सूर्यप्रकाश वेळ", "PST": "पॅसिफिक प्रमाण वेळ", "SAST": "दक्षिण आफ्रिका प्रमाण वेळ", "SGT": "सिंगापूर प्रमाण वेळ", "SRT": "सुरिनाम वेळ", "TMST": "तुर्कमेनिस्तान उन्हाळी वेळ", "TMT": "तुर्कमेनिस्तान प्रमाण वेळ", "UYST": "उरुग्वे उन्हाळी वेळ", "UYT": "उरुग्वे प्रमाण वेळ", "VET": "व्हेनेझुएला वेळ", "WARST": "पश्चिमी अर्जेंटिना उन्हाळी वेळ", "WART": "पश्चिमी अर्जेंटिना प्रमाण वेळ", "WAST": "पश्चिम आफ्रिका उन्हाळी वेळ", "WAT": "पश्चिम आफ्रिका प्रमाण वेळ", "WESZ": "पश्चिम युरोपियन उन्हाळी वेळ", "WEZ": "पश्चिम युरोपियन प्रमाण वेळ", "WIB": "पश्चिमी इंडोनेशिया वेळ", "WIT": "पौर्वात्य इंडोनेशिया वेळ", "WITA": "मध्\u200dय इंडोनेशिया वेळ", "∅∅∅": "अ\u200dॅझोरेस उन्हाळी वेळ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a2a4fea44e28d3286f808ebd85af3bc6,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mr_in"] = func() locales.Translator {
		return &localen{

			locale:             "mr_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{2, 3, 4, 6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "जाने", "फेब्रु", "मार्च", "एप्रि", "मे", "जून", "जुलै", "ऑग", "सप्टें", "ऑक्टो", "नोव्हें", "डिसें"},
			monthsNarrow:       []string{"", "जा", "फे", "मा", "ए", "मे", "जू", "जु", "ऑ", "स", "ऑ", "नो", "डि"},
			monthsWide:         []string{"", "जानेवारी", "फेब्रुवारी", "मार्च", "एप्रिल", "मे", "जून", "जुलै", "ऑगस्ट", "सप्टेंबर", "ऑक्टोबर", "नोव्हेंबर", "डिसेंबर"},
			daysAbbreviated:    []string{"रवि", "सोम", "मंगळ", "बुध", "गुरु", "शुक्र", "शनि"},
			daysNarrow:         []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysShort:          []string{"र", "सो", "मं", "बु", "गु", "शु", "श"},
			daysWide:           []string{"रविवार", "सोमवार", "मंगळवार", "बुधवार", "गुरुवार", "शुक्रवार", "शनिवार"},
			periodsAbbreviated: []string{"म.पू.", "म.उ."},
			periodsNarrow:      []string{"स", "सं"},
			periodsWide:        []string{"म.पू.", "म.उ."},
			erasAbbreviated:    []string{"इ. स. पू.", "इ. स."},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ईसवीसनपूर्व", "ईसवीसन"},
			timezones:          map[string]string{"ACDT": "ऑस्ट्रेलियन मध्य सूर्यप्रकाश वेळ", "ACST": "ऑस्ट्रेलियन मध्य प्रमाण वेळ", "ACWDT": "ऑस्ट्रेलियन मध्य-पश्चिम सूर्यप्रकाश वेळ", "ACWST": "ऑस्ट्रेलियन मध्य-पश्चिम प्रमाण वेळ", "ADT": "अटलांटिक सूर्यप्रकाश वेळ", "AEDT": "ऑस्ट्रेलियन पूर्व सूर्यप्रकाश वेळ", "AEST": "ऑस्ट्रेलियन पूर्व प्रमाण वेळ", "AKDT": "अलास्का सूर्यप्रकाश वेळ", "AKST": "अलास्का प्रमाण वेळ", "ARST": "अर्जेंटिना उन्हाळी वेळ", "ART": "अर्जेंटिना प्रमाण वेळ", "AST": "अटलांटिक प्रमाण वेळ", "AWDT": "ऑस्ट्रेलियन पश्चिम सूर्यप्रकाश वेळ", "AWST": "ऑस्ट्रेलियन पश्चिम प्रमाण वेळ", "BOT": "बोलिव्हिया वेळ", "BT": "भूतान वेळ", "CAT": "मध्\u200dय आफ्रिका वेळ", "CDT": "केंद्रीय सूर्यप्रकाश वेळ", "CHADT": "चॅथम सूर्यप्रकाश वेळ", "CHAST": "चॅथम प्रमाण वेळ", "CLST": "चिली उन्हाळी वेळ", "CLT": "चिली प्रमाण वेळ", "COST": "कोलंबिया उन्हाळी वेळ", "COT": "कोलंबिया प्रमाण वेळ", "CST": "केंद्रीय प्रमाण वेळ", "ChST": "चामोरो प्रमाण वेळ", "EAT": "पूर्व आफ्रिका वेळ", "ECT": "इक्वेडोर वेळ", "EDT": "पौर्वात्य सूर्यप्रकाश वेळ", "EST": "पौर्वात्य प्रमाण वेळ", "GFT": "फ्रेंच गयाना वेळ", "GMT": "ग्रीनिच प्रमाण वेळ", "GST": "खाडी प्रमाण वेळ", "GYT": "गयाना वेळ", "HADT": "हवाई-अलूशन सूर्यप्रकाश वेळ", "HAST": "हवाई-अलूशन प्रमाण वेळ", "HAT": "न्यू फाउंडलंड सूर्यप्रकाश वेळ", "HECU": "क्यूबा सूर्यप्रकाश वेळ", "HEEG": "पूर्व ग्रीनलँड उन्हाळी वेळ", "HENOMX": "वायव्य मेक्सिको सूर्यप्रकाश वेळ", "HEOG": "पश्चिम ग्रीनलँड उन्हाळी वेळ", "HEPM": "सेंट पियरे आणि मिक्वेलोन सूर्यप्रकाश वेळ", "HEPMX": "मेक्सिको पॅसिफिक सूर्यप्रकाश वेळ", "HKST": "हाँग काँग उन्हाळी वेळ", "HKT": "हाँग काँग प्रमाण वेळ", "HNCU": "क्यूबा प्रमाण वेळ", "HNEG": "पूर्व ग्रीनलँड प्रमाण वेळ", "HNNOMX": "वायव्य मेक्सिको प्रमाण वेळ", "HNOG": "पश्चिम ग्रीनलँड प्रमाण वेळ", "HNPM": "सेंट पियरे आणि मिक्वेलोन प्रमाण वेळ", "HNPMX": "मेक्सिको पॅसिफिक प्रमाण वेळ", "HNT": "न्यू फाउंडलंड प्रमाण वेळ", "IST": "भारतीय प्रमाण वेळ", "JDT": "जपान सूर्यप्रकाश वेळ", "JST": "जपान प्रमाण वेळ", "LHDT": "लॉर्ड होवे सूर्यप्रकाश वेळ", "LHST": "लॉर्ड होवे प्रमाण वेळ", "MDT": "पर्वतीय सूर्यप्रकाश वेळ", "MESZ": "मध्\u200dय युरोपियन उन्हाळी वेळ", "MEZ": "मध्\u200dय युरोपियन प्रमाण वेळ", "MST": "पर्वतीय प्रमाण वेळ", "MYT": "मलेशिया वेळ", "NZDT": "न्यूझीलंड सूर्यप्रकाश वेळ", "NZST": "न्यूझीलंड प्रमाण वेळ", "OESZ": "पूर्व युरोपियन उन्हाळी वेळ", "OEZ": "पूर्व युरोपियन प्रमाण वेळ", "PDT": "पॅसिफिक सूर्यप्रकाश वेळ", "PST": "पॅसिफिक प्रमाण वेळ", "SAST": "दक्षिण आफ्रिका प्रमाण वेळ", "SGT": "सिंगापूर प्रमाण वेळ", "SRT": "सुरिनाम वेळ", "TMST": "तुर्कमेनिस्तान उन्हाळी वेळ", "TMT": "तुर्कमेनिस्तान प्रमाण वेळ", "UYST": "उरुग्वे उन्हाळी वेळ", "UYT": "उरुग्वे प्रमाण वेळ", "VET": "व्हेनेझुएला वेळ", "WARST": "पश्चिमी अर्जेंटिना उन्हाळी वेळ", "WART": "पश्चिमी अर्जेंटिना प्रमाण वेळ", "WAST": "पश्चिम आफ्रिका उन्हाळी वेळ", "WAT": "पश्चिम आफ्रिका प्रमाण वेळ", "WESZ": "पश्चिम युरोपियन उन्हाळी वेळ", "WEZ": "पश्चिम युरोपियन प्रमाण वेळ", "WIB": "पश्चिमी इंडोनेशिया वेळ", "WIT": "पौर्वात्य इंडोनेशिया वेळ", "WITA": "मध्\u200dय इंडोनेशिया वेळ", "∅∅∅": "अ\u200dॅझोरेस उन्हाळी वेळ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a2a4fea44e28d3286f808ebd85af3bc6,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ms"] = func() locales.Translator {
		return &localen{

			locale:                 "ms",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ogo", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Mac", "April", "Mei", "Jun", "Julai", "Ogos", "September", "Oktober", "November", "Disember"},
			daysAbbreviated:        []string{"Ahd", "Isn", "Sel", "Rab", "Kha", "Jum", "Sab"},
			daysNarrow:             []string{"A", "I", "S", "R", "K", "J", "S"},
			daysShort:              []string{"Ah", "Is", "Se", "Ra", "Kh", "Ju", "Sa"},
			daysWide:               []string{"Ahad", "Isnin", "Selasa", "Rabu", "Khamis", "Jumaat", "Sabtu"},
			periodsAbbreviated:     []string{"PG", "PTG"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"PG", "PTG"},
			erasAbbreviated:        []string{"S.M.", "TM"},
			erasNarrow:             []string{"S.M.", "TM"},
			erasWide:               []string{"S.M.", "TM"},
			timezones:              map[string]string{"ACDT": "Waktu Siang Australia Tengah", "ACST": "Waktu Piawai Australia Tengah", "ACWDT": "Waktu Siang Barat Tengah Australia", "ACWST": "Waktu Piawai Barat Tengah Australia", "ADT": "Waktu Siang Atlantik", "AEDT": "Waktu Siang Australia Timur", "AEST": "Waktu Piawai Timur Australia", "AKDT": "Waktu Siang Alaska", "AKST": "Waktu Piawai Alaska", "ARST": "Waktu Musim Panas Argentina", "ART": "Waktu Piawai Argentina", "AST": "Waktu Piawai Atlantik", "AWDT": "Waktu Siang Australia Barat", "AWST": "Waktu Piawai Australia Barat", "BOT": "Waktu Bolivia", "BT": "Waktu Bhutan", "CAT": "Waktu Afrika Tengah", "CDT": "Waktu Siang Tengah", "CHADT": "Waktu Siang Chatham", "CHAST": "Waktu Piawai Chatham", "CLST": "Waktu Musim Panas Chile", "CLT": "Waktu Piawai Chile", "COST": "Waktu Musim Panas Colombia", "COT": "Waktu Piawai Colombia", "CST": "Waktu Piawai Pusat", "ChST": "Waktu Piawai Chamorro", "EAT": "Waktu Afrika Timur", "ECT": "Waktu Ecuador", "EDT": "Waktu Siang Timur", "EST": "Waktu Piawai Timur", "GFT": "Waktu Guyana Perancis", "GMT": "Waktu Min Greenwich", "GST": "Waktu Piawai Teluk", "GYT": "Waktu Guyana", "HADT": "Waktu Siang Hawaii-Aleutian", "HAST": "Waktu Piawai Hawaii-Aleutian", "HAT": "Waktu Siang Newfoundland", "HECU": "Waktu Siang Cuba", "HEEG": "Waktu Musim Panas Greenland Timur", "HENOMX": "Waktu Siang Barat Laut Mexico", "HEOG": "Waktu Musim Panas Greenland Barat", "HEPM": "Waktu Siang Saint Pierre dan Miquelon", "HEPMX": "Waktu Siang Pasifik Mexico", "HKST": "Waktu Musim Panas Hong Kong", "HKT": "Waktu Piawai Hong Kong", "HNCU": "Waktu Piawai Cuba", "HNEG": "Waktu Piawai Greenland Timur", "HNNOMX": "Waktu Piawai Barat Laut Mexico", "HNOG": "Waktu Piawai Greenland Barat", "HNPM": "Waktu Piawai Saint Pierre dan Miquelon", "HNPMX": "Waktu Piawai Pasifik Mexico", "HNT": "Waktu Piawai Newfoundland", "IST": "Waktu Piawai India", "JDT": "Waktu Siang Jepun", "JST": "Waktu Piawai Jepun", "LHDT": "Waktu Siang Lord Howe", "LHST": "Waktu Piawai Lord Howe", "MDT": "Waktu Siang Pergunungan", "MESZ": "Waktu Musim Panas Eropah Tengah", "MEZ": "Waktu Piawai Eropah Tengah", "MST": "Waktu Piawai Pergunungan", "MYT": "Waktu Malaysia", "NZDT": "Waktu Siang New Zealand", "NZST": "Waktu Piawai New Zealand", "OESZ": "Waktu Musim Panas Eropah Timur", "OEZ": "Waktu Piawai Eropah Timur", "PDT": "Waktu Siang Pasifik", "PST": "Waktu Piawai Pasifik", "SAST": "Waktu Piawai Afrika Selatan", "SGT": "Waktu Piawai Singapura", "SRT": "Waktu Suriname", "TMST": "Waktu Musim Panas Turkmenistan", "TMT": "Waktu Piawai Turkmenistan", "UYST": "Waktu Musim Panas Uruguay", "UYT": "Waktu Piawai Uruguay", "VET": "Waktu Venezuela", "WARST": "Waktu Musim Panas Argentina Barat", "WART": "Waktu Piawai Argentina Barat", "WAST": "Waktu Musim Panas Afrika Barat", "WAT": "Waktu Piawai Afrika Barat", "WESZ": "Waktu Musim Panas Eropah Barat", "WEZ": "Waktu Piawai Eropah Barat", "WIB": "Waktu Indonesia Barat", "WIT": "Waktu Indonesia Timur", "WITA": "Waktu Indonesia Tengah", "∅∅∅": "Waktu Musim Panas Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ms_bn"] = func() locales.Translator {
		return &localen{

			locale:                 "ms_BN",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "$", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ogo", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Mac", "April", "Mei", "Jun", "Julai", "Ogos", "September", "Oktober", "November", "Disember"},
			daysAbbreviated:        []string{"Ahd", "Isn", "Sel", "Rab", "Kha", "Jum", "Sab"},
			daysNarrow:             []string{"A", "I", "S", "R", "K", "J", "S"},
			daysShort:              []string{"Ah", "Is", "Se", "Ra", "Kh", "Ju", "Sa"},
			daysWide:               []string{"Ahad", "Isnin", "Selasa", "Rabu", "Khamis", "Jumaat", "Sabtu"},
			periodsAbbreviated:     []string{"PG", "PTG"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"PG", "PTG"},
			erasAbbreviated:        []string{"S.M.", "TM"},
			erasNarrow:             []string{"S.M.", "TM"},
			erasWide:               []string{"S.M.", "TM"},
			timezones:              map[string]string{"ACDT": "Waktu Siang Australia Tengah", "ACST": "Waktu Piawai Australia Tengah", "ACWDT": "Waktu Siang Barat Tengah Australia", "ACWST": "Waktu Piawai Barat Tengah Australia", "ADT": "Waktu Siang Atlantik", "AEDT": "Waktu Siang Australia Timur", "AEST": "Waktu Piawai Timur Australia", "AKDT": "Waktu Siang Alaska", "AKST": "Waktu Piawai Alaska", "ARST": "Waktu Musim Panas Argentina", "ART": "Waktu Piawai Argentina", "AST": "Waktu Piawai Atlantik", "AWDT": "Waktu Siang Australia Barat", "AWST": "Waktu Piawai Australia Barat", "BOT": "Waktu Bolivia", "BT": "Waktu Bhutan", "CAT": "Waktu Afrika Tengah", "CDT": "Waktu Siang Tengah", "CHADT": "Waktu Siang Chatham", "CHAST": "Waktu Piawai Chatham", "CLST": "Waktu Musim Panas Chile", "CLT": "Waktu Piawai Chile", "COST": "Waktu Musim Panas Colombia", "COT": "Waktu Piawai Colombia", "CST": "Waktu Piawai Pusat", "ChST": "Waktu Piawai Chamorro", "EAT": "Waktu Afrika Timur", "ECT": "Waktu Ecuador", "EDT": "Waktu Siang Timur", "EST": "Waktu Piawai Timur", "GFT": "Waktu Guyana Perancis", "GMT": "Waktu Min Greenwich", "GST": "Waktu Piawai Teluk", "GYT": "Waktu Guyana", "HADT": "Waktu Siang Hawaii-Aleutian", "HAST": "Waktu Piawai Hawaii-Aleutian", "HAT": "Waktu Siang Newfoundland", "HECU": "Waktu Siang Cuba", "HEEG": "Waktu Musim Panas Greenland Timur", "HENOMX": "Waktu Siang Barat Laut Mexico", "HEOG": "Waktu Musim Panas Greenland Barat", "HEPM": "Waktu Siang Saint Pierre dan Miquelon", "HEPMX": "Waktu Siang Pasifik Mexico", "HKST": "Waktu Musim Panas Hong Kong", "HKT": "Waktu Piawai Hong Kong", "HNCU": "Waktu Piawai Cuba", "HNEG": "Waktu Piawai Greenland Timur", "HNNOMX": "Waktu Piawai Barat Laut Mexico", "HNOG": "Waktu Piawai Greenland Barat", "HNPM": "Waktu Piawai Saint Pierre dan Miquelon", "HNPMX": "Waktu Piawai Pasifik Mexico", "HNT": "Waktu Piawai Newfoundland", "IST": "Waktu Piawai India", "JDT": "Waktu Siang Jepun", "JST": "Waktu Piawai Jepun", "LHDT": "Waktu Siang Lord Howe", "LHST": "Waktu Piawai Lord Howe", "MDT": "Waktu Siang Pergunungan", "MESZ": "Waktu Musim Panas Eropah Tengah", "MEZ": "Waktu Piawai Eropah Tengah", "MST": "Waktu Piawai Pergunungan", "MYT": "Waktu Malaysia", "NZDT": "Waktu Siang New Zealand", "NZST": "Waktu Piawai New Zealand", "OESZ": "Waktu Musim Panas Eropah Timur", "OEZ": "Waktu Piawai Eropah Timur", "PDT": "Waktu Siang Pasifik", "PST": "Waktu Piawai Pasifik", "SAST": "Waktu Piawai Afrika Selatan", "SGT": "Waktu Piawai Singapura", "SRT": "Waktu Suriname", "TMST": "Waktu Musim Panas Turkmenistan", "TMT": "Waktu Piawai Turkmenistan", "UYST": "Waktu Musim Panas Uruguay", "UYT": "Waktu Piawai Uruguay", "VET": "Waktu Venezuela", "WARST": "Waktu Musim Panas Argentina Barat", "WART": "Waktu Piawai Argentina Barat", "WAST": "Waktu Musim Panas Afrika Barat", "WAT": "Waktu Piawai Afrika Barat", "WESZ": "Waktu Musim Panas Eropah Barat", "WEZ": "Waktu Piawai Eropah Barat", "WIB": "Waktu Indonesia Barat", "WIT": "Waktu Indonesia Timur", "WITA": "Waktu Indonesia Tengah", "∅∅∅": "Waktu Musim Panas Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3c8ce40063641c68bcd08d693c6fa8e8,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ms_my"] = func() locales.Translator {
		return &localen{

			locale:                 "ms_MY",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ogo", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Mac", "April", "Mei", "Jun", "Julai", "Ogos", "September", "Oktober", "November", "Disember"},
			daysAbbreviated:        []string{"Ahd", "Isn", "Sel", "Rab", "Kha", "Jum", "Sab"},
			daysNarrow:             []string{"A", "I", "S", "R", "K", "J", "S"},
			daysShort:              []string{"Ah", "Is", "Se", "Ra", "Kh", "Ju", "Sa"},
			daysWide:               []string{"Ahad", "Isnin", "Selasa", "Rabu", "Khamis", "Jumaat", "Sabtu"},
			periodsAbbreviated:     []string{"PG", "PTG"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"PG", "PTG"},
			erasAbbreviated:        []string{"S.M.", "TM"},
			erasNarrow:             []string{"S.M.", "TM"},
			erasWide:               []string{"S.M.", "TM"},
			timezones:              map[string]string{"ACDT": "Waktu Siang Australia Tengah", "ACST": "Waktu Piawai Australia Tengah", "ACWDT": "Waktu Siang Barat Tengah Australia", "ACWST": "Waktu Piawai Barat Tengah Australia", "ADT": "Waktu Siang Atlantik", "AEDT": "Waktu Siang Australia Timur", "AEST": "Waktu Piawai Timur Australia", "AKDT": "Waktu Siang Alaska", "AKST": "Waktu Piawai Alaska", "ARST": "Waktu Musim Panas Argentina", "ART": "Waktu Piawai Argentina", "AST": "Waktu Piawai Atlantik", "AWDT": "Waktu Siang Australia Barat", "AWST": "Waktu Piawai Australia Barat", "BOT": "Waktu Bolivia", "BT": "Waktu Bhutan", "CAT": "Waktu Afrika Tengah", "CDT": "Waktu Siang Tengah", "CHADT": "Waktu Siang Chatham", "CHAST": "Waktu Piawai Chatham", "CLST": "Waktu Musim Panas Chile", "CLT": "Waktu Piawai Chile", "COST": "Waktu Musim Panas Colombia", "COT": "Waktu Piawai Colombia", "CST": "Waktu Piawai Pusat", "ChST": "Waktu Piawai Chamorro", "EAT": "Waktu Afrika Timur", "ECT": "Waktu Ecuador", "EDT": "Waktu Siang Timur", "EST": "Waktu Piawai Timur", "GFT": "Waktu Guyana Perancis", "GMT": "Waktu Min Greenwich", "GST": "Waktu Piawai Teluk", "GYT": "Waktu Guyana", "HADT": "Waktu Siang Hawaii-Aleutian", "HAST": "Waktu Piawai Hawaii-Aleutian", "HAT": "Waktu Siang Newfoundland", "HECU": "Waktu Siang Cuba", "HEEG": "Waktu Musim Panas Greenland Timur", "HENOMX": "Waktu Siang Barat Laut Mexico", "HEOG": "Waktu Musim Panas Greenland Barat", "HEPM": "Waktu Siang Saint Pierre dan Miquelon", "HEPMX": "Waktu Siang Pasifik Mexico", "HKST": "Waktu Musim Panas Hong Kong", "HKT": "Waktu Piawai Hong Kong", "HNCU": "Waktu Piawai Cuba", "HNEG": "Waktu Piawai Greenland Timur", "HNNOMX": "Waktu Piawai Barat Laut Mexico", "HNOG": "Waktu Piawai Greenland Barat", "HNPM": "Waktu Piawai Saint Pierre dan Miquelon", "HNPMX": "Waktu Piawai Pasifik Mexico", "HNT": "Waktu Piawai Newfoundland", "IST": "Waktu Piawai India", "JDT": "Waktu Siang Jepun", "JST": "Waktu Piawai Jepun", "LHDT": "Waktu Siang Lord Howe", "LHST": "Waktu Piawai Lord Howe", "MDT": "Waktu Siang Pergunungan", "MESZ": "Waktu Musim Panas Eropah Tengah", "MEZ": "Waktu Piawai Eropah Tengah", "MST": "Waktu Piawai Pergunungan", "MYT": "Waktu Malaysia", "NZDT": "Waktu Siang New Zealand", "NZST": "Waktu Piawai New Zealand", "OESZ": "Waktu Musim Panas Eropah Timur", "OEZ": "Waktu Piawai Eropah Timur", "PDT": "Waktu Siang Pasifik", "PST": "Waktu Piawai Pasifik", "SAST": "Waktu Piawai Afrika Selatan", "SGT": "Waktu Piawai Singapura", "SRT": "Waktu Suriname", "TMST": "Waktu Musim Panas Turkmenistan", "TMT": "Waktu Piawai Turkmenistan", "UYST": "Waktu Musim Panas Uruguay", "UYT": "Waktu Piawai Uruguay", "VET": "Waktu Venezuela", "WARST": "Waktu Musim Panas Argentina Barat", "WART": "Waktu Piawai Argentina Barat", "WAST": "Waktu Musim Panas Afrika Barat", "WAT": "Waktu Piawai Afrika Barat", "WESZ": "Waktu Musim Panas Eropah Barat", "WEZ": "Waktu Piawai Eropah Barat", "WIB": "Waktu Indonesia Barat", "WIT": "Waktu Indonesia Timur", "WITA": "Waktu Indonesia Tengah", "∅∅∅": "Waktu Musim Panas Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ms_sg"] = func() locales.Translator {
		return &localen{

			locale:                 "ms_SG",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "$", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ogo", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Mac", "April", "Mei", "Jun", "Julai", "Ogos", "September", "Oktober", "November", "Disember"},
			daysAbbreviated:        []string{"Ahd", "Isn", "Sel", "Rab", "Kha", "Jum", "Sab"},
			daysNarrow:             []string{"A", "I", "S", "R", "K", "J", "S"},
			daysShort:              []string{"Ah", "Is", "Se", "Ra", "Kh", "Ju", "Sa"},
			daysWide:               []string{"Ahad", "Isnin", "Selasa", "Rabu", "Khamis", "Jumaat", "Sabtu"},
			periodsAbbreviated:     []string{"PG", "PTG"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"PG", "PTG"},
			erasAbbreviated:        []string{"S.M.", "TM"},
			erasNarrow:             []string{"S.M.", "TM"},
			erasWide:               []string{"S.M.", "TM"},
			timezones:              map[string]string{"ACDT": "Waktu Siang Australia Tengah", "ACST": "Waktu Piawai Australia Tengah", "ACWDT": "Waktu Siang Barat Tengah Australia", "ACWST": "Waktu Piawai Barat Tengah Australia", "ADT": "Waktu Siang Atlantik", "AEDT": "Waktu Siang Australia Timur", "AEST": "Waktu Piawai Timur Australia", "AKDT": "Waktu Siang Alaska", "AKST": "Waktu Piawai Alaska", "ARST": "Waktu Musim Panas Argentina", "ART": "Waktu Piawai Argentina", "AST": "Waktu Piawai Atlantik", "AWDT": "Waktu Siang Australia Barat", "AWST": "Waktu Piawai Australia Barat", "BOT": "Waktu Bolivia", "BT": "Waktu Bhutan", "CAT": "Waktu Afrika Tengah", "CDT": "Waktu Siang Tengah", "CHADT": "Waktu Siang Chatham", "CHAST": "Waktu Piawai Chatham", "CLST": "Waktu Musim Panas Chile", "CLT": "Waktu Piawai Chile", "COST": "Waktu Musim Panas Colombia", "COT": "Waktu Piawai Colombia", "CST": "Waktu Piawai Pusat", "ChST": "Waktu Piawai Chamorro", "EAT": "Waktu Afrika Timur", "ECT": "Waktu Ecuador", "EDT": "Waktu Siang Timur", "EST": "Waktu Piawai Timur", "GFT": "Waktu Guyana Perancis", "GMT": "Waktu Min Greenwich", "GST": "Waktu Piawai Teluk", "GYT": "Waktu Guyana", "HADT": "Waktu Siang Hawaii-Aleutian", "HAST": "Waktu Piawai Hawaii-Aleutian", "HAT": "Waktu Siang Newfoundland", "HECU": "Waktu Siang Cuba", "HEEG": "Waktu Musim Panas Greenland Timur", "HENOMX": "Waktu Siang Barat Laut Mexico", "HEOG": "Waktu Musim Panas Greenland Barat", "HEPM": "Waktu Siang Saint Pierre dan Miquelon", "HEPMX": "Waktu Siang Pasifik Mexico", "HKST": "Waktu Musim Panas Hong Kong", "HKT": "Waktu Piawai Hong Kong", "HNCU": "Waktu Piawai Cuba", "HNEG": "Waktu Piawai Greenland Timur", "HNNOMX": "Waktu Piawai Barat Laut Mexico", "HNOG": "Waktu Piawai Greenland Barat", "HNPM": "Waktu Piawai Saint Pierre dan Miquelon", "HNPMX": "Waktu Piawai Pasifik Mexico", "HNT": "Waktu Piawai Newfoundland", "IST": "Waktu Piawai India", "JDT": "Waktu Siang Jepun", "JST": "Waktu Piawai Jepun", "LHDT": "Waktu Siang Lord Howe", "LHST": "Waktu Piawai Lord Howe", "MDT": "Waktu Siang Pergunungan", "MESZ": "Waktu Musim Panas Eropah Tengah", "MEZ": "Waktu Piawai Eropah Tengah", "MST": "Waktu Piawai Pergunungan", "MYT": "Waktu Malaysia", "NZDT": "Waktu Siang New Zealand", "NZST": "Waktu Piawai New Zealand", "OESZ": "Waktu Musim Panas Eropah Timur", "OEZ": "Waktu Piawai Eropah Timur", "PDT": "Waktu Siang Pasifik", "PST": "Waktu Piawai Pasifik", "SAST": "Waktu Piawai Afrika Selatan", "SGT": "Waktu Piawai Singapura", "SRT": "Waktu Suriname", "TMST": "Waktu Musim Panas Turkmenistan", "TMT": "Waktu Piawai Turkmenistan", "UYST": "Waktu Musim Panas Uruguay", "UYT": "Waktu Piawai Uruguay", "VET": "Waktu Venezuela", "WARST": "Waktu Musim Panas Argentina Barat", "WART": "Waktu Piawai Argentina Barat", "WAST": "Waktu Musim Panas Afrika Barat", "WAT": "Waktu Piawai Afrika Barat", "WESZ": "Waktu Musim Panas Eropah Barat", "WEZ": "Waktu Piawai Eropah Barat", "WIB": "Waktu Indonesia Barat", "WIT": "Waktu Indonesia Timur", "WITA": "Waktu Indonesia Tengah", "∅∅∅": "Waktu Musim Panas Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_fed047b78adc0ecb9a1387099de9feec,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mt"] = func() locales.Translator {
		return &localen{

			locale:             "mt",
			pluralsCardinal:    []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "Rs", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Fra", "Mar", "Apr", "Mej", "Ġun", "Lul", "Aww", "Set", "Ott", "Nov", "Diċ"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "Ġ", "L", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Jannar", "Frar", "Marzu", "April", "Mejju", "Ġunju", "Lulju", "Awwissu", "Settembru", "Ottubru", "Novembru", "Diċembru"},
			daysAbbreviated:    []string{"Ħad", "Tne", "Tli", "Erb", "Ħam", "Ġim", "Sib"},
			daysNarrow:         []string{"Ħd", "T", "Tl", "Er", "Ħm", "Ġm", "Sb"},
			daysShort:          []string{"Ħad", "Tne", "Tli", "Erb", "Ħam", "Ġim", "Sib"},
			daysWide:           []string{"Il-Ħadd", "It-Tnejn", "It-Tlieta", "L-Erbgħa", "Il-Ħamis", "Il-Ġimgħa", "Is-Sibt"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"am", "pm"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"QK", "WK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Qabel Kristu", "Wara Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Ħin Ċentrali Ewropew tas-Sajf", "MEZ": "Ħin Ċentrali Ewropew Standard", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_e05d2968783cc0de88c2d1b2682d3007,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_03998ac4c6437ee7c96bd6694bf576fb,
			fnFmtDateLong:         fnFmtDateLong_e370bf773fe0548c5c154f2e1cb30b9f,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mt_mt"] = func() locales.Translator {
		return &localen{

			locale:             "mt_MT",
			pluralsCardinal:    []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Fra", "Mar", "Apr", "Mej", "Ġun", "Lul", "Aww", "Set", "Ott", "Nov", "Diċ"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "Ġ", "L", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Jannar", "Frar", "Marzu", "April", "Mejju", "Ġunju", "Lulju", "Awwissu", "Settembru", "Ottubru", "Novembru", "Diċembru"},
			daysAbbreviated:    []string{"Ħad", "Tne", "Tli", "Erb", "Ħam", "Ġim", "Sib"},
			daysNarrow:         []string{"Ħd", "T", "Tl", "Er", "Ħm", "Ġm", "Sb"},
			daysShort:          []string{"Ħad", "Tne", "Tli", "Erb", "Ħam", "Ġim", "Sib"},
			daysWide:           []string{"Il-Ħadd", "It-Tnejn", "It-Tlieta", "L-Erbgħa", "Il-Ħamis", "Il-Ġimgħa", "Is-Sibt"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"am", "pm"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"QK", "WK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Qabel Kristu", "Wara Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Ħin Ċentrali Ewropew tas-Sajf", "MEZ": "Ħin Ċentrali Ewropew Standard", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_e05d2968783cc0de88c2d1b2682d3007,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_03998ac4c6437ee7c96bd6694bf576fb,
			fnFmtDateLong:         fnFmtDateLong_e370bf773fe0548c5c154f2e1cb30b9f,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mua"] = func() locales.Translator {
		return &localen{

			locale:                 "mua",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "FLO", "CLA", "CKI", "FMF", "MAD", "MBI", "MLI", "MAM", "FDE", "FMU", "FGW", "FYU"},
			monthsNarrow:           []string{"", "O", "A", "I", "F", "D", "B", "L", "M", "E", "U", "W", "Y"},
			monthsWide:             []string{"", "Fĩi Loo", "Cokcwaklaŋne", "Cokcwaklii", "Fĩi Marfoo", "Madǝǝuutǝbijaŋ", "Mamǝŋgwãafahbii", "Mamǝŋgwãalii", "Madǝmbii", "Fĩi Dǝɓlii", "Fĩi Mundaŋ", "Fĩi Gwahlle", "Fĩi Yuru"},
			daysAbbreviated:        []string{"Cya", "Cla", "Czi", "Cko", "Cka", "Cga", "Cze"},
			daysNarrow:             []string{"Y", "L", "Z", "O", "A", "G", "E"},
			daysWide:               []string{"Com’yakke", "Comlaaɗii", "Comzyiiɗii", "Comkolle", "Comkaldǝɓlii", "Comgaisuu", "Comzyeɓsuu"},
			periodsAbbreviated:     []string{"comme", "lilli"},
			periodsWide:            []string{"comme", "lilli"},
			erasAbbreviated:        []string{"KK", "PK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"KǝPel Kristu", "Pel Kristu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mua_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "mua_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "FLO", "CLA", "CKI", "FMF", "MAD", "MBI", "MLI", "MAM", "FDE", "FMU", "FGW", "FYU"},
			monthsNarrow:           []string{"", "O", "A", "I", "F", "D", "B", "L", "M", "E", "U", "W", "Y"},
			monthsWide:             []string{"", "Fĩi Loo", "Cokcwaklaŋne", "Cokcwaklii", "Fĩi Marfoo", "Madǝǝuutǝbijaŋ", "Mamǝŋgwãafahbii", "Mamǝŋgwãalii", "Madǝmbii", "Fĩi Dǝɓlii", "Fĩi Mundaŋ", "Fĩi Gwahlle", "Fĩi Yuru"},
			daysAbbreviated:        []string{"Cya", "Cla", "Czi", "Cko", "Cka", "Cga", "Cze"},
			daysNarrow:             []string{"Y", "L", "Z", "O", "A", "G", "E"},
			daysWide:               []string{"Com’yakke", "Comlaaɗii", "Comzyiiɗii", "Comkolle", "Comkaldǝɓlii", "Comgaisuu", "Comzyeɓsuu"},
			periodsAbbreviated:     []string{"comme", "lilli"},
			periodsWide:            []string{"comme", "lilli"},
			erasAbbreviated:        []string{"KK", "PK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"KǝPel Kristu", "Pel Kristu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["my"] = func() locales.Translator {
		return &localen{

			locale:                 "my",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "NAf", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "Afl", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "G", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "K", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "B/.", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TT$", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ဇန်", "ဖေ", "မတ်", "ဧ", "မေ", "ဇွန်", "ဇူ", "ဩ", "စက်", "အောက်", "နို", "ဒီ"},
			monthsNarrow:           []string{"", "ဇ", "ဖ", "မ", "ဧ", "မ", "ဇ", "ဇ", "ဩ", "စ", "အ", "န", "ဒ"},
			monthsWide:             []string{"", "ဇန်နဝါရီ", "ဖေဖော်ဝါရီ", "မတ်", "ဧပြီ", "မေ", "ဇွန်", "ဇူလိုင်", "ဩဂုတ်", "စက်တင်ဘာ", "အောက်တိုဘာ", "နိုဝင်ဘာ", "ဒီဇင်ဘာ"},
			daysAbbreviated:        []string{"တနင်္ဂနွေ", "တနင်္လာ", "အင်္ဂါ", "ဗုဒ္ဓဟူး", "ကြာသပတေး", "သောကြာ", "စနေ"},
			daysNarrow:             []string{"တ", "တ", "အ", "ဗ", "က", "သ", "စ"},
			daysShort:              []string{"တနင်္ဂနွေ", "တနင်္လာ", "အင်္ဂါ", "ဗုဒ္ဓဟူး", "ကြာသပတေး", "သောကြာ", "စနေ"},
			daysWide:               []string{"တနင်္ဂနွေ", "တနင်္လာ", "အင်္ဂါ", "ဗုဒ္ဓဟူး", "ကြာသပတေး", "သောကြာ", "စနေ"},
			periodsAbbreviated:     []string{"နံနက်", "ညနေ"},
			periodsNarrow:          []string{"နံနက်", "ညနေ"},
			periodsWide:            []string{"နံနက်", "ညနေ"},
			erasAbbreviated:        []string{"ဘီစီ", "အဒေီ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ခရစ်တော် မပေါ်မီနှစ်", "ခရစ်နှစ်"},
			timezones:              map[string]string{"ACDT": "ဩစတြေးလျ အလယ်ပိုင်း နွေရာသီ အချိန်", "ACST": "ဩစတြေးလျ အလယ်ပိုင်း စံတော်ချိန်", "ACWDT": "သြစတြေးလျား အနောက်အလယ်ပိုင်း နွေရာသီ အချိန်", "ACWST": "သြစတြေးလျား အနောက်အလယ်ပိုင်း စံတော်ချိန်", "ADT": "အတ္တလန်တစ် နွေရာသီ စံတော်ချိန်", "AEDT": "အရှေ့ဩစတြေးလျ နွေရာသီ အချိန်", "AEST": "အရှေ့ဩစတြေးလျ စံတော်ချိန်", "AKDT": "အလာစကာ နွေရာသီ စံတော်ချိန်", "AKST": "အလာစကာ စံတော်ချိန်", "ARST": "အာဂျင်တီးနား နွေရာသီအချိန်", "ART": "အာဂျင်တီးနား စံတော်ချိန်", "AST": "အတ္တလန်တစ် စံတော်ချိန်", "AWDT": "ဩစတြေးလျ နွေရာသီ အချိန်", "AWST": "အနောက်ဩစတြေးလျ စံတော်ချိန်", "BOT": "ဘိုလီးဘီးယား အချိန်", "BT": "ဘူတန် အချိန်", "CAT": "အလယ်အာဖရိက အချိန်", "CDT": "မြောက်အမေရိက အလယ်ပိုင်း နွေရာသီစံတော်ချိန်", "CHADT": "ချာသမ် နွေရာသီ အချိန်", "CHAST": "ချားသမ်စံတော်ချိန်", "CLST": "ချီလီ နွေရာသီ အချိန်", "CLT": "ချီလီ စံတော်ချိန်", "COST": "ကိုလံဘီယာ နွေရာသီ အချိန်", "COT": "ကိုလံဘီယာ စံတော်ချိန်", "CST": "မြောက်အမေရိက အလယ်ပိုင်းစံတော်ချိန်", "ChST": "ချာမိုရို အချိန်", "EAT": "အရှေ့အာဖရိက အချိန်", "ECT": "အီကွေဒေါ အချိန်", "EDT": "မြောက်အမေရိက အရှေ့ပိုင်း နွေရာသီစံတော်ချိန်", "EST": "မြောက်အမေရိက အရှေ့ပိုင်းစံတော်ချိန်", "GFT": "ပြင်သစ် ဂီအားနား အချိန်", "GMT": "ဂရင်းနစ် စံတော်ချိန်", "GST": "ပင်လယ်ကွေ့ အချိန်", "GYT": "ဂိုင်ယာနာ အချိန်", "HADT": "ဟာဝိုင်ယီ အယ်လူးရှန်း နွေရာသီ စံတော်ချိန်", "HAST": "ဟာဝိုင်ယီ အယ်လူးရှန်း စံတော်ချိန်", "HAT": "နယူးဖောင်လန် နွေရာသီ စံတော်ချိန်", "HECU": "ကျူးဘား နွေရာသီ စံတော်ချိန်", "HEEG": "အရှေ့ဂရင်းလန် နွေရာသီ စံတော်ချိန်", "HENOMX": "အနောက်တောင် မက္ကစီကို နွေရာသီ စံတော်ချိန်", "HEOG": "အနောက် ဂရင်းလန် နွေရာသီ စံတော်ချိန်", "HEPM": "စိန့်ပီအဲနှင့် မီခွီလွန် နွေရာသီ စံတော်ချိန်", "HEPMX": "မက္ကစီကန် ပစိဖိတ် နွေရာသီ စံတော်ချိန်", "HKST": "ဟောင်ကောင် နွေရာသီ အချိန်", "HKT": "ဟောင်ကောင် စံတော်ချိန်", "HNCU": "ကျူးဘား စံတော်ချိန်", "HNEG": "အရှေ့ဂရင်းလန်း စံတော်ချိန်", "HNNOMX": "အနောက်တောင် မက္ကဆီကို စံတော်ချိန်", "HNOG": "အနောက် ဂရင်းလန်း စံတော်ချိန်", "HNPM": "စိန့်ပီအဲနှင့်မီခွီလွန်စံတော်ချိန်", "HNPMX": "မက္ကဆီကန် ပစိဖိတ် စံတော်ချိန်", "HNT": "နယူးဖောင်လန် စံတော်ချိန်", "IST": "အိန္ဒိယ စံတော်ချိန်", "JDT": "ဂျပန် နွေရာသီ အချိန်", "JST": "ဂျပန် စံတော်ချိန်", "LHDT": "လော့ဒ်ဟောင် နွေရာသီ အချိန်", "LHST": "လော့ဒ်ဟောင် စံတော်ချိန်", "MDT": "မြောက်အမေရိက တောင်တန်းဒေသ နွေရာသီစံတော်ချိန်", "MESZ": "ဥရောပအလယ်ပိုင်း နွေရာသီ အချိန်", "MEZ": "ဥရောပအလယ်ပိုင်း စံတော်ချိန်", "MST": "မြောက်အမေရိက တောင်တန်းဒေသစံတော်ချိန်", "MYT": "မလေးရှား အချိန်", "NZDT": "နယူးဇီလန် နွေရာသီ အချိန်", "NZST": "နယူးဇီလန် စံတော်ချိန်", "OESZ": "အရှေ့ဥရောပ နွေရာသီ အချိန်", "OEZ": "အရှေ့ဥရောပ စံတော်ချိန်", "PDT": "မြောက်အမေရိက ပစိဖိတ် နွေရာသီစံတော်ချိန်", "PST": "မြောက်အမေရိက ပစိဖိတ်စံတော်ချိန်", "SAST": "တောင်အာဖရိက အချိန်", "SGT": "စင်္ကာပူ အချိန်", "SRT": "စူးရီနာမ်အချိန်", "TMST": "တာ့ခ်မင်နစ္စတန် နွေရာသီ အချိန်", "TMT": "တာ့ခ်မင်နစ္စတန် စံတော်ချိန်", "UYST": "ဥရုဂွေး နွေရာသီ အချိန်", "UYT": "ဥရုဂွေး စံတော်ချိန်", "VET": "ဗင်နီဇွဲလား အချိန်", "WARST": "အနောက် အာဂျင်တီးနား နွေရာသီ အချိန်", "WART": "အနောက် အာဂျင်တီးနား စံတော်ချိန်", "WAST": "အနောက်အာဖရိက နွေရာသီ အချိန်", "WAT": "အနောက်အာဖရိက စံတော်ချိန်", "WESZ": "အနောက်ဥရောပ နွေရာသီ အချိန်", "WEZ": "အနောက်ဥရောပ စံတော်ချိန်", "WIB": "အနောက်ပိုင်း အင်ဒိုနီးရှား အချိန်", "WIT": "အရှေ့ပိုင်း အင်ဒိုနီးရှား အချိန်", "WITA": "အလယ်ပိုင်း အင်ဒိုနီးရှား အချိန်", "∅∅∅": "ဘရာဇီး နွေရာသီ အချိန်"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_b91fecf4568a8483e671bad76f4d4425,
			fnFmtDateLong:         fnFmtDateLong_26f464d351da53af35ea071506237a9d,
			fnFmtDateMedium:       fnFmtDateMedium_4d2c8f906ead2ec2a8ed4de52aa273cb,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_9d22f7542356e9fc28813c7e9ca3671d,
			fnFmtTimeLong:         fnFmtTimeLong_c108ef0052564f5ac9b30fa59af63081,
			fnFmtTimeMedium:       fnFmtTimeMedium_5d9a472a4327911b01200af597bf5908,
			fnFmtTimeShort:        fnFmtTimeShort_3636d5dfdda3e0b6ca17aef52abf5eb3,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["my_mm"] = func() locales.Translator {
		return &localen{

			locale:                 "my_MM",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ဇန်", "ဖေ", "မတ်", "ဧ", "မေ", "ဇွန်", "ဇူ", "ဩ", "စက်", "အောက်", "နို", "ဒီ"},
			monthsNarrow:           []string{"", "ဇ", "ဖ", "မ", "ဧ", "မ", "ဇ", "ဇ", "ဩ", "စ", "အ", "န", "ဒ"},
			monthsWide:             []string{"", "ဇန်နဝါရီ", "ဖေဖော်ဝါရီ", "မတ်", "ဧပြီ", "မေ", "ဇွန်", "ဇူလိုင်", "ဩဂုတ်", "စက်တင်ဘာ", "အောက်တိုဘာ", "နိုဝင်ဘာ", "ဒီဇင်ဘာ"},
			daysAbbreviated:        []string{"တနင်္ဂနွေ", "တနင်္လာ", "အင်္ဂါ", "ဗုဒ္ဓဟူး", "ကြာသပတေး", "သောကြာ", "စနေ"},
			daysNarrow:             []string{"တ", "တ", "အ", "ဗ", "က", "သ", "စ"},
			daysShort:              []string{"တနင်္ဂနွေ", "တနင်္လာ", "အင်္ဂါ", "ဗုဒ္ဓဟူး", "ကြာသပတေး", "သောကြာ", "စနေ"},
			daysWide:               []string{"တနင်္ဂနွေ", "တနင်္လာ", "အင်္ဂါ", "ဗုဒ္ဓဟူး", "ကြာသပတေး", "သောကြာ", "စနေ"},
			periodsAbbreviated:     []string{"နံနက်", "ညနေ"},
			periodsNarrow:          []string{"နံနက်", "ညနေ"},
			periodsWide:            []string{"နံနက်", "ညနေ"},
			erasAbbreviated:        []string{"ဘီစီ", "အဒေီ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ခရစ်တော် မပေါ်မီနှစ်", "ခရစ်နှစ်"},
			timezones:              map[string]string{"ACDT": "ဩစတြေးလျ အလယ်ပိုင်း နွေရာသီ အချိန်", "ACST": "ဩစတြေးလျ အလယ်ပိုင်း စံတော်ချိန်", "ACWDT": "သြစတြေးလျား အနောက်အလယ်ပိုင်း နွေရာသီ အချိန်", "ACWST": "သြစတြေးလျား အနောက်အလယ်ပိုင်း စံတော်ချိန်", "ADT": "အတ္တလန်တစ် နွေရာသီ စံတော်ချိန်", "AEDT": "အရှေ့ဩစတြေးလျ နွေရာသီ အချိန်", "AEST": "အရှေ့ဩစတြေးလျ စံတော်ချိန်", "AKDT": "အလာစကာ နွေရာသီ စံတော်ချိန်", "AKST": "အလာစကာ စံတော်ချိန်", "ARST": "အာဂျင်တီးနား နွေရာသီအချိန်", "ART": "အာဂျင်တီးနား စံတော်ချိန်", "AST": "အတ္တလန်တစ် စံတော်ချိန်", "AWDT": "ဩစတြေးလျ နွေရာသီ အချိန်", "AWST": "အနောက်ဩစတြေးလျ စံတော်ချိန်", "BOT": "ဘိုလီးဘီးယား အချိန်", "BT": "ဘူတန် အချိန်", "CAT": "အလယ်အာဖရိက အချိန်", "CDT": "မြောက်အမေရိက အလယ်ပိုင်း နွေရာသီစံတော်ချိန်", "CHADT": "ချာသမ် နွေရာသီ အချိန်", "CHAST": "ချားသမ်စံတော်ချိန်", "CLST": "ချီလီ နွေရာသီ အချိန်", "CLT": "ချီလီ စံတော်ချိန်", "COST": "ကိုလံဘီယာ နွေရာသီ အချိန်", "COT": "ကိုလံဘီယာ စံတော်ချိန်", "CST": "မြောက်အမေရိက အလယ်ပိုင်းစံတော်ချိန်", "ChST": "ချာမိုရို အချိန်", "EAT": "အရှေ့အာဖရိက အချိန်", "ECT": "အီကွေဒေါ အချိန်", "EDT": "မြောက်အမေရိက အရှေ့ပိုင်း နွေရာသီစံတော်ချိန်", "EST": "မြောက်အမေရိက အရှေ့ပိုင်းစံတော်ချိန်", "GFT": "ပြင်သစ် ဂီအားနား အချိန်", "GMT": "ဂရင်းနစ် စံတော်ချိန်", "GST": "ပင်လယ်ကွေ့ အချိန်", "GYT": "ဂိုင်ယာနာ အချိန်", "HADT": "ဟာဝိုင်ယီ အယ်လူးရှန်း နွေရာသီ စံတော်ချိန်", "HAST": "ဟာဝိုင်ယီ အယ်လူးရှန်း စံတော်ချိန်", "HAT": "နယူးဖောင်လန် နွေရာသီ စံတော်ချိန်", "HECU": "ကျူးဘား နွေရာသီ စံတော်ချိန်", "HEEG": "အရှေ့ဂရင်းလန် နွေရာသီ စံတော်ချိန်", "HENOMX": "အနောက်တောင် မက္ကစီကို နွေရာသီ စံတော်ချိန်", "HEOG": "အနောက် ဂရင်းလန် နွေရာသီ စံတော်ချိန်", "HEPM": "စိန့်ပီအဲနှင့် မီခွီလွန် နွေရာသီ စံတော်ချိန်", "HEPMX": "မက္ကစီကန် ပစိဖိတ် နွေရာသီ စံတော်ချိန်", "HKST": "ဟောင်ကောင် နွေရာသီ အချိန်", "HKT": "ဟောင်ကောင် စံတော်ချိန်", "HNCU": "ကျူးဘား စံတော်ချိန်", "HNEG": "အရှေ့ဂရင်းလန်း စံတော်ချိန်", "HNNOMX": "အနောက်တောင် မက္ကဆီကို စံတော်ချိန်", "HNOG": "အနောက် ဂရင်းလန်း စံတော်ချိန်", "HNPM": "စိန့်ပီအဲနှင့်မီခွီလွန်စံတော်ချိန်", "HNPMX": "မက္ကဆီကန် ပစိဖိတ် စံတော်ချိန်", "HNT": "နယူးဖောင်လန် စံတော်ချိန်", "IST": "အိန္ဒိယ စံတော်ချိန်", "JDT": "ဂျပန် နွေရာသီ အချိန်", "JST": "ဂျပန် စံတော်ချိန်", "LHDT": "လော့ဒ်ဟောင် နွေရာသီ အချိန်", "LHST": "လော့ဒ်ဟောင် စံတော်ချိန်", "MDT": "မြောက်အမေရိက တောင်တန်းဒေသ နွေရာသီစံတော်ချိန်", "MESZ": "ဥရောပအလယ်ပိုင်း နွေရာသီ အချိန်", "MEZ": "ဥရောပအလယ်ပိုင်း စံတော်ချိန်", "MST": "မြောက်အမေရိက တောင်တန်းဒေသစံတော်ချိန်", "MYT": "မလေးရှား အချိန်", "NZDT": "နယူးဇီလန် နွေရာသီ အချိန်", "NZST": "နယူးဇီလန် စံတော်ချိန်", "OESZ": "အရှေ့ဥရောပ နွေရာသီ အချိန်", "OEZ": "အရှေ့ဥရောပ စံတော်ချိန်", "PDT": "မြောက်အမေရိက ပစိဖိတ် နွေရာသီစံတော်ချိန်", "PST": "မြောက်အမေရိက ပစိဖိတ်စံတော်ချိန်", "SAST": "တောင်အာဖရိက အချိန်", "SGT": "စင်္ကာပူ အချိန်", "SRT": "စူးရီနာမ်အချိန်", "TMST": "တာ့ခ်မင်နစ္စတန် နွေရာသီ အချိန်", "TMT": "တာ့ခ်မင်နစ္စတန် စံတော်ချိန်", "UYST": "ဥရုဂွေး နွေရာသီ အချိန်", "UYT": "ဥရုဂွေး စံတော်ချိန်", "VET": "ဗင်နီဇွဲလား အချိန်", "WARST": "အနောက် အာဂျင်တီးနား နွေရာသီ အချိန်", "WART": "အနောက် အာဂျင်တီးနား စံတော်ချိန်", "WAST": "အနောက်အာဖရိက နွေရာသီ အချိန်", "WAT": "အနောက်အာဖရိက စံတော်ချိန်", "WESZ": "အနောက်ဥရောပ နွေရာသီ အချိန်", "WEZ": "အနောက်ဥရောပ စံတော်ချိန်", "WIB": "အနောက်ပိုင်း အင်ဒိုနီးရှား အချိန်", "WIT": "အရှေ့ပိုင်း အင်ဒိုနီးရှား အချိန်", "WITA": "အလယ်ပိုင်း အင်ဒိုနီးရှား အချိန်", "∅∅∅": "ပီရူး နွေရာသီ အချိန်"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_b91fecf4568a8483e671bad76f4d4425,
			fnFmtDateLong:         fnFmtDateLong_26f464d351da53af35ea071506237a9d,
			fnFmtDateMedium:       fnFmtDateMedium_4d2c8f906ead2ec2a8ed4de52aa273cb,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_9d22f7542356e9fc28813c7e9ca3671d,
			fnFmtTimeLong:         fnFmtTimeLong_c108ef0052564f5ac9b30fa59af63081,
			fnFmtTimeMedium:       fnFmtTimeMedium_5d9a472a4327911b01200af597bf5908,
			fnFmtTimeShort:        fnFmtTimeShort_3636d5dfdda3e0b6ca17aef52abf5eb3,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mzn"] = func() locales.Translator {
		return &localen{

			locale:            "mzn",
			pluralsCardinal:   nil,
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			monthsWide:        []string{"", "ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			erasAbbreviated:   []string{"پ.م", "م."},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"قبل میلاد", "بعد میلاد"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_411c735c7aafad814d0db05f621af17c,
			fnFmtCurrency:         fnFmtCurrency_8bfebaacbf71a9a4bab2a3bb99e52529,
			fnFmtDateFull:         fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18,
			fnFmtDateLong:         fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d,
			fnFmtDateMedium:       fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["mzn_ir"] = func() locales.Translator {
		return &localen{

			locale:            "mzn_IR",
			pluralsCardinal:   nil,
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			monthsWide:        []string{"", "ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اکتبر", "نوامبر", "دسامبر"},
			erasAbbreviated:   []string{"پ.م", "م."},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"قبل میلاد", "بعد میلاد"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_411c735c7aafad814d0db05f621af17c,
			fnFmtCurrency:         fnFmtCurrency_8bfebaacbf71a9a4bab2a3bb99e52529,
			fnFmtDateFull:         fnFmtDateFull_89e5603fd6c03ae17109c6eb729a3e18,
			fnFmtDateLong:         fnFmtDateLong_bfdb64849bf0a513d30bbef60108376d,
			fnFmtDateMedium:       fnFmtDateMedium_9f7a8e1d4465e6a8ede358df4b71943e,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["naq"] = func() locales.Translator {
		return &localen{

			locale:             "naq",
			pluralsCardinal:    []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "$", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "ǃKhanni", "ǃKhanǀgôab", "ǀKhuuǁkhâb", "ǃHôaǂkhaib", "ǃKhaitsâb", "Gamaǀaeb", "ǂKhoesaob", "Aoǁkhuumûǁkhâb", "Taraǀkhuumûǁkhâb", "ǂNûǁnâiseb", "ǀHooǂgaeb", "Hôasoreǁkhâb"},
			daysAbbreviated:    []string{"Son", "Ma", "De", "Wu", "Do", "Fr", "Sat"},
			daysNarrow:         []string{"S", "M", "E", "W", "D", "F", "A"},
			daysWide:           []string{"Sontaxtsees", "Mantaxtsees", "Denstaxtsees", "Wunstaxtsees", "Dondertaxtsees", "Fraitaxtsees", "Satertaxtsees"},
			periodsAbbreviated: []string{"ǁgoagas", "ǃuias"},
			periodsWide:        []string{"ǁgoagas", "ǃuias"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Xristub aiǃâ", "Xristub khaoǃgâ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["naq_na"] = func() locales.Translator {
		return &localen{

			locale:             "naq_NA",
			pluralsCardinal:    []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "ǃKhanni", "ǃKhanǀgôab", "ǀKhuuǁkhâb", "ǃHôaǂkhaib", "ǃKhaitsâb", "Gamaǀaeb", "ǂKhoesaob", "Aoǁkhuumûǁkhâb", "Taraǀkhuumûǁkhâb", "ǂNûǁnâiseb", "ǀHooǂgaeb", "Hôasoreǁkhâb"},
			daysAbbreviated:    []string{"Son", "Ma", "De", "Wu", "Do", "Fr", "Sat"},
			daysNarrow:         []string{"S", "M", "E", "W", "D", "F", "A"},
			daysWide:           []string{"Sontaxtsees", "Mantaxtsees", "Denstaxtsees", "Wunstaxtsees", "Dondertaxtsees", "Fraitaxtsees", "Satertaxtsees"},
			periodsAbbreviated: []string{"ǁgoagas", "ǃuias"},
			periodsWide:        []string{"ǁgoagas", "ǃuias"},
			erasAbbreviated:    []string{"BC", "AD"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Xristub aiǃâ", "Xristub khaoǃgâ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nb"] = func() locales.Translator {
		return &localen{

			locale:                 "nb",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "kr", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "mai", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"søn.", "man.", "tir.", "ons.", "tor.", "fre.", "lør."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø.", "ma.", "ti.", "on.", "to.", "fr.", "lø."},
			daysWide:               []string{"søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"før Kristus", "etter Kristus"},
			timezones:              map[string]string{"ACDT": "sentralaustralsk sommertid", "ACST": "sentralaustralsk normaltid", "ACWDT": "vest-sentralaustralsk sommertid", "ACWST": "vest-sentralaustralsk normaltid", "ADT": "sommertid for den nordamerikanske atlanterhavskysten", "AEDT": "østaustralsk sommertid", "AEST": "østaustralsk normaltid", "AKDT": "alaskisk sommertid", "AKST": "alaskisk normaltid", "ARST": "argentinsk sommertid", "ART": "argentinsk normaltid", "AST": "normaltid for den nordamerikanske atlanterhavskysten", "AWDT": "vestaustralsk sommertid", "AWST": "vestaustralsk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "sentralafrikansk tid", "CDT": "sommertid for det sentrale Nord-Amerika", "CHADT": "sommertid for Chatham", "CHAST": "normaltid for Chatham", "CLST": "chilensk sommertid", "CLT": "chilensk normaltid", "COST": "colombiansk sommertid", "COT": "colombiansk normaltid", "CST": "normaltid for det sentrale Nord-Amerika", "ChST": "tidssone for Chamorro", "EAT": "østafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "sommertid for den nordamerikanske østkysten", "EST": "normaltid for den nordamerikanske østkysten", "GFT": "tidssone for Fransk Guyana", "GMT": "Greenwich middeltid", "GST": "tidssone for Persiabukta", "GYT": "guyansk tid", "HADT": "sommertid for Hawaii og Aleutene", "HAST": "normaltid for Hawaii og Aleutene", "HAT": "sommertid for Newfoundland", "HECU": "cubansk sommertid", "HEEG": "østgrønlandsk sommertid", "HENOMX": "sommertid for nordvestlige Mexico", "HEOG": "vestgrønlandsk sommertid", "HEPM": "sommertid for Saint-Pierre-et-Miquelon", "HEPMX": "sommertid for den meksikanske Stillehavskysten", "HKST": "sommertid for Hongkong", "HKT": "normaltid for Hongkong", "HNCU": "cubansk normaltid", "HNEG": "østgrønlandsk normaltid", "HNNOMX": "normaltid for nordvestlige Mexico", "HNOG": "vestgrønlandsk normaltid", "HNPM": "normaltid for Saint-Pierre-et-Miquelon", "HNPMX": "normaltid for den meksikanske Stillehavskysten", "HNT": "normaltid for Newfoundland", "IST": "indisk tid", "JDT": "japansk sommertid", "JST": "japansk normaltid", "LHDT": "sommertid for Lord Howe-øya", "LHST": "normaltid for Lord Howe-øya", "MDT": "sommertid for Rocky Mountains (USA)", "MESZ": "sentraleuropeisk sommertid", "MEZ": "sentraleuropeisk normaltid", "MST": "normaltid for Rocky Mountains (USA)", "MYT": "malaysisk tid", "NZDT": "newzealandsk sommertid", "NZST": "newzealandsk normaltid", "OESZ": "østeuropeisk sommertid", "OEZ": "østeuropeisk normaltid", "PDT": "sommertid for den nordamerikanske Stillehavskysten", "PST": "normaltid for den nordamerikanske Stillehavskysten", "SAST": "sørafrikansk tid", "SGT": "singaporsk tid", "SRT": "surinamsk tid", "TMST": "turkmensk sommertid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommertid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "vestargentinsk sommertid", "WART": "vestargentinsk normaltid", "WAST": "vestafrikansk sommertid", "WAT": "vestafrikansk normaltid", "WESZ": "vesteuropeisk sommertid", "WEZ": "vesteuropeisk normaltid", "WIB": "vestindonesisk tid", "WIT": "østindonesisk tid", "WITA": "sentralindonesisk tid", "∅∅∅": "sommertid for Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6501a20bdcbb382f1e5e59af6d704f28,
			fnFmtCurrency:         fnFmtCurrency_52160a421770c8d84e356e1637ce196a,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nb_no"] = func() locales.Translator {
		return &localen{

			locale:                 "nb_NO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "mai", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"søn.", "man.", "tir.", "ons.", "tor.", "fre.", "lør."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø.", "ma.", "ti.", "on.", "to.", "fr.", "lø."},
			daysWide:               []string{"søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"før Kristus", "etter Kristus"},
			timezones:              map[string]string{"ACDT": "sentralaustralsk sommertid", "ACST": "sentralaustralsk normaltid", "ACWDT": "vest-sentralaustralsk sommertid", "ACWST": "vest-sentralaustralsk normaltid", "ADT": "sommertid for den nordamerikanske atlanterhavskysten", "AEDT": "østaustralsk sommertid", "AEST": "østaustralsk normaltid", "AKDT": "alaskisk sommertid", "AKST": "alaskisk normaltid", "ARST": "argentinsk sommertid", "ART": "argentinsk normaltid", "AST": "normaltid for den nordamerikanske atlanterhavskysten", "AWDT": "vestaustralsk sommertid", "AWST": "vestaustralsk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "sentralafrikansk tid", "CDT": "sommertid for det sentrale Nord-Amerika", "CHADT": "sommertid for Chatham", "CHAST": "normaltid for Chatham", "CLST": "chilensk sommertid", "CLT": "chilensk normaltid", "COST": "colombiansk sommertid", "COT": "colombiansk normaltid", "CST": "normaltid for det sentrale Nord-Amerika", "ChST": "tidssone for Chamorro", "EAT": "østafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "sommertid for den nordamerikanske østkysten", "EST": "normaltid for den nordamerikanske østkysten", "GFT": "tidssone for Fransk Guyana", "GMT": "Greenwich middeltid", "GST": "tidssone for Persiabukta", "GYT": "guyansk tid", "HADT": "sommertid for Hawaii og Aleutene", "HAST": "normaltid for Hawaii og Aleutene", "HAT": "sommertid for Newfoundland", "HECU": "cubansk sommertid", "HEEG": "østgrønlandsk sommertid", "HENOMX": "sommertid for nordvestlige Mexico", "HEOG": "vestgrønlandsk sommertid", "HEPM": "sommertid for Saint-Pierre-et-Miquelon", "HEPMX": "sommertid for den meksikanske Stillehavskysten", "HKST": "sommertid for Hongkong", "HKT": "normaltid for Hongkong", "HNCU": "cubansk normaltid", "HNEG": "østgrønlandsk normaltid", "HNNOMX": "normaltid for nordvestlige Mexico", "HNOG": "vestgrønlandsk normaltid", "HNPM": "normaltid for Saint-Pierre-et-Miquelon", "HNPMX": "normaltid for den meksikanske Stillehavskysten", "HNT": "normaltid for Newfoundland", "IST": "indisk tid", "JDT": "japansk sommertid", "JST": "japansk normaltid", "LHDT": "sommertid for Lord Howe-øya", "LHST": "normaltid for Lord Howe-øya", "MDT": "sommertid for Rocky Mountains (USA)", "MESZ": "sentraleuropeisk sommertid", "MEZ": "sentraleuropeisk normaltid", "MST": "normaltid for Rocky Mountains (USA)", "MYT": "malaysisk tid", "NZDT": "newzealandsk sommertid", "NZST": "newzealandsk normaltid", "OESZ": "østeuropeisk sommertid", "OEZ": "østeuropeisk normaltid", "PDT": "sommertid for den nordamerikanske Stillehavskysten", "PST": "normaltid for den nordamerikanske Stillehavskysten", "SAST": "sørafrikansk tid", "SGT": "singaporsk tid", "SRT": "surinamsk tid", "TMST": "turkmensk sommertid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommertid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "vestargentinsk sommertid", "WART": "vestargentinsk normaltid", "WAST": "vestafrikansk sommertid", "WAT": "vestafrikansk normaltid", "WESZ": "vesteuropeisk sommertid", "WEZ": "vesteuropeisk normaltid", "WIB": "vestindonesisk tid", "WIT": "østindonesisk tid", "WITA": "sentralindonesisk tid", "∅∅∅": "peruansk sommertid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6501a20bdcbb382f1e5e59af6d704f28,
			fnFmtCurrency:         fnFmtCurrency_52160a421770c8d84e356e1637ce196a,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nb_sj"] = func() locales.Translator {
		return &localen{

			locale:                 "nb_SJ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "mai", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"søn.", "man.", "tir.", "ons.", "tor.", "fre.", "lør."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø.", "ma.", "ti.", "on.", "to.", "fr.", "lø."},
			daysWide:               []string{"søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"før Kristus", "etter Kristus"},
			timezones:              map[string]string{"ACDT": "sentralaustralsk sommertid", "ACST": "sentralaustralsk normaltid", "ACWDT": "vest-sentralaustralsk sommertid", "ACWST": "vest-sentralaustralsk normaltid", "ADT": "sommertid for den nordamerikanske atlanterhavskysten", "AEDT": "østaustralsk sommertid", "AEST": "østaustralsk normaltid", "AKDT": "alaskisk sommertid", "AKST": "alaskisk normaltid", "ARST": "argentinsk sommertid", "ART": "argentinsk normaltid", "AST": "normaltid for den nordamerikanske atlanterhavskysten", "AWDT": "vestaustralsk sommertid", "AWST": "vestaustralsk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "sentralafrikansk tid", "CDT": "sommertid for det sentrale Nord-Amerika", "CHADT": "sommertid for Chatham", "CHAST": "normaltid for Chatham", "CLST": "chilensk sommertid", "CLT": "chilensk normaltid", "COST": "colombiansk sommertid", "COT": "colombiansk normaltid", "CST": "normaltid for det sentrale Nord-Amerika", "ChST": "tidssone for Chamorro", "EAT": "østafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "sommertid for den nordamerikanske østkysten", "EST": "normaltid for den nordamerikanske østkysten", "GFT": "tidssone for Fransk Guyana", "GMT": "Greenwich middeltid", "GST": "tidssone for Persiabukta", "GYT": "guyansk tid", "HADT": "sommertid for Hawaii og Aleutene", "HAST": "normaltid for Hawaii og Aleutene", "HAT": "sommertid for Newfoundland", "HECU": "cubansk sommertid", "HEEG": "østgrønlandsk sommertid", "HENOMX": "sommertid for nordvestlige Mexico", "HEOG": "vestgrønlandsk sommertid", "HEPM": "sommertid for Saint-Pierre-et-Miquelon", "HEPMX": "sommertid for den meksikanske Stillehavskysten", "HKST": "sommertid for Hongkong", "HKT": "normaltid for Hongkong", "HNCU": "cubansk normaltid", "HNEG": "østgrønlandsk normaltid", "HNNOMX": "normaltid for nordvestlige Mexico", "HNOG": "vestgrønlandsk normaltid", "HNPM": "normaltid for Saint-Pierre-et-Miquelon", "HNPMX": "normaltid for den meksikanske Stillehavskysten", "HNT": "normaltid for Newfoundland", "IST": "indisk tid", "JDT": "japansk sommertid", "JST": "japansk normaltid", "LHDT": "sommertid for Lord Howe-øya", "LHST": "normaltid for Lord Howe-øya", "MDT": "sommertid for Rocky Mountains (USA)", "MESZ": "sentraleuropeisk sommertid", "MEZ": "sentraleuropeisk normaltid", "MST": "normaltid for Rocky Mountains (USA)", "MYT": "malaysisk tid", "NZDT": "newzealandsk sommertid", "NZST": "newzealandsk normaltid", "OESZ": "østeuropeisk sommertid", "OEZ": "østeuropeisk normaltid", "PDT": "sommertid for den nordamerikanske Stillehavskysten", "PST": "normaltid for den nordamerikanske Stillehavskysten", "SAST": "sørafrikansk tid", "SGT": "singaporsk tid", "SRT": "surinamsk tid", "TMST": "turkmensk sommertid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommertid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "vestargentinsk sommertid", "WART": "vestargentinsk normaltid", "WAST": "vestafrikansk sommertid", "WAT": "vestafrikansk normaltid", "WESZ": "vesteuropeisk sommertid", "WEZ": "vesteuropeisk normaltid", "WIB": "vestindonesisk tid", "WIT": "østindonesisk tid", "WITA": "sentralindonesisk tid", "∅∅∅": "sommertid for Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6501a20bdcbb382f1e5e59af6d704f28,
			fnFmtCurrency:         fnFmtCurrency_52160a421770c8d84e356e1637ce196a,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nd"] = func() locales.Translator {
		return &localen{

			locale:                 "nd",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Zib", "Nhlo", "Mbi", "Mab", "Nkw", "Nhla", "Ntu", "Ncw", "Mpan", "Mfu", "Lwe", "Mpal"},
			monthsNarrow:           []string{"", "Z", "N", "M", "M", "N", "N", "N", "N", "M", "M", "L", "M"},
			monthsWide:             []string{"", "Zibandlela", "Nhlolanja", "Mbimbitho", "Mabasa", "Nkwenkwezi", "Nhlangula", "Ntulikazi", "Ncwabakazi", "Mpandula", "Mfumfu", "Lwezi", "Mpalakazi"},
			daysAbbreviated:        []string{"Son", "Mvu", "Sib", "Sit", "Sin", "Sih", "Mgq"},
			daysNarrow:             []string{"S", "M", "S", "S", "S", "S", "M"},
			daysWide:               []string{"Sonto", "Mvulo", "Sibili", "Sithathu", "Sine", "Sihlanu", "Mgqibelo"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"UKristo angakabuyi", "Ukristo ebuyile"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nd_zw"] = func() locales.Translator {
		return &localen{

			locale:                 "nd_ZW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Zib", "Nhlo", "Mbi", "Mab", "Nkw", "Nhla", "Ntu", "Ncw", "Mpan", "Mfu", "Lwe", "Mpal"},
			monthsNarrow:           []string{"", "Z", "N", "M", "M", "N", "N", "N", "N", "M", "M", "L", "M"},
			monthsWide:             []string{"", "Zibandlela", "Nhlolanja", "Mbimbitho", "Mabasa", "Nkwenkwezi", "Nhlangula", "Ntulikazi", "Ncwabakazi", "Mpandula", "Mfumfu", "Lwezi", "Mpalakazi"},
			daysAbbreviated:        []string{"Son", "Mvu", "Sib", "Sit", "Sin", "Sih", "Mgq"},
			daysNarrow:             []string{"S", "M", "S", "S", "S", "S", "M"},
			daysWide:               []string{"Sonto", "Mvulo", "Sibili", "Sithathu", "Sine", "Sihlanu", "Mgqibelo"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"UKristo angakabuyi", "Ukristo ebuyile"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nds"] = func() locales.Translator {
		return &localen{

			locale:                 "nds",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januaar", "Februaar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktover", "November", "Dezember"},
			daysAbbreviated:        []string{"Sü.", "Ma.", "Di.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sünndag", "Maandag", "Dingsdag", "Middeweken", "Dunnersdag", "Freedag", "Sünnavend"},
			periodsAbbreviated:     []string{"vm", "nm"},
			periodsWide:            []string{"vm", "nm"},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"vC", "nC"},
			erasWide:               []string{"vör Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Zentraalaustraalsch Summertiet", "ACST": "Zentraalaustraalsch Standardtiet", "ACWDT": "Westzentraalaustraalsch Summertiet", "ACWST": "Westzentraalaustraalsch Standardtiet", "ADT": "Noordamerikaansch Atlantik-Summertiet", "AEDT": "Oostaustraalsch Summertiet", "AEST": "Oostaustraalsch Standardtiet", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Noordamerikaansch Atlantik-Standardtiet", "AWDT": "Westaustraalsch Summertiet", "AWST": "Westaustraalsch Standardtiet", "BOT": "BOT", "BT": "BT", "CAT": "Zentraalafrikaansch Tiet", "CDT": "Noordamerikaansch zentraal Summertiet", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Noordamerikaansch zentraal Standardtiet", "ChST": "ChST", "EAT": "Oostafrikaansch Tiet", "ECT": "ECT", "EDT": "Noordamerikaansch oosten Summertiet", "EST": "Noordamerikaansch oosten Standardtiet", "GFT": "GFT", "GMT": "Gröönwisch-Welttiet", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "Indien-Tiet", "JDT": "Japaansch Summertiet", "JST": "Japaansch Standardtiet", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Noordamerikaansch Barg-Summertiet", "MESZ": "Zentraaleuropääsch Summertiet", "MEZ": "Zentraaleuropääsch Standardtiet", "MST": "Noordamerikaansch Barg-Standardtiet", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oosteuropääsch Summertiet", "OEZ": "Oosteuropääsch Standardtiet", "PDT": "Noordamerikaansch Pazifik-Summertiet", "PST": "Noordamerikaansch Pazifik-Standardtiet", "SAST": "Söödafrikaansch Tiet", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Westafrikaansch Summertiet", "WAT": "Westafrikaansch Standardtiet", "WESZ": "Westeuropääsch Summertiet", "WEZ": "Westeuropääsch Standardtiet", "WIB": "Westindoneesch Tiet", "WIT": "Oostindoneesch Tiet", "WITA": "Indoneesch Zentraaltiet", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_91904f1c4adbb925999bdba7f5796de8,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_cbdb981782261e5a8b8cc19702400c64,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e61d6b84b00325a0cc262f834ec42619,
			fnFmtTimeLong:         fnFmtTimeLong_1dc6a22a036902eb11698fc0a589743d,
			fnFmtTimeMedium:       fnFmtTimeMedium_d09a7c8ab2ed333a37b86babd4139677,
			fnFmtTimeShort:        fnFmtTimeShort_e50f14b6d2405eade89d9149a1efaee9,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nds_de"] = func() locales.Translator {
		return &localen{

			locale:                 "nds_DE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januaar", "Februaar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktover", "November", "Dezember"},
			daysAbbreviated:        []string{"Sü.", "Ma.", "Di.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sünndag", "Maandag", "Dingsdag", "Middeweken", "Dunnersdag", "Freedag", "Sünnavend"},
			periodsAbbreviated:     []string{"vm", "nm"},
			periodsWide:            []string{"vm", "nm"},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"vC", "nC"},
			erasWide:               []string{"vör Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Zentraalaustraalsch Summertiet", "ACST": "Zentraalaustraalsch Standardtiet", "ACWDT": "Westzentraalaustraalsch Summertiet", "ACWST": "Westzentraalaustraalsch Standardtiet", "ADT": "Noordamerikaansch Atlantik-Summertiet", "AEDT": "Oostaustraalsch Summertiet", "AEST": "Oostaustraalsch Standardtiet", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Noordamerikaansch Atlantik-Standardtiet", "AWDT": "Westaustraalsch Summertiet", "AWST": "Westaustraalsch Standardtiet", "BOT": "BOT", "BT": "BT", "CAT": "Zentraalafrikaansch Tiet", "CDT": "Noordamerikaansch zentraal Summertiet", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Noordamerikaansch zentraal Standardtiet", "ChST": "ChST", "EAT": "Oostafrikaansch Tiet", "ECT": "ECT", "EDT": "Noordamerikaansch oosten Summertiet", "EST": "Noordamerikaansch oosten Standardtiet", "GFT": "GFT", "GMT": "Gröönwisch-Welttiet", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "Indien-Tiet", "JDT": "Japaansch Summertiet", "JST": "Japaansch Standardtiet", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Noordamerikaansch Barg-Summertiet", "MESZ": "Zentraaleuropääsch Summertiet", "MEZ": "Zentraaleuropääsch Standardtiet", "MST": "Noordamerikaansch Barg-Standardtiet", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oosteuropääsch Summertiet", "OEZ": "Oosteuropääsch Standardtiet", "PDT": "Noordamerikaansch Pazifik-Summertiet", "PST": "Noordamerikaansch Pazifik-Standardtiet", "SAST": "Söödafrikaansch Tiet", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Westafrikaansch Summertiet", "WAT": "Westafrikaansch Standardtiet", "WESZ": "Westeuropääsch Summertiet", "WEZ": "Westeuropääsch Standardtiet", "WIB": "Westindoneesch Tiet", "WIT": "Oostindoneesch Tiet", "WITA": "Indoneesch Zentraaltiet", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_91904f1c4adbb925999bdba7f5796de8,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_cbdb981782261e5a8b8cc19702400c64,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e61d6b84b00325a0cc262f834ec42619,
			fnFmtTimeLong:         fnFmtTimeLong_1dc6a22a036902eb11698fc0a589743d,
			fnFmtTimeMedium:       fnFmtTimeMedium_d09a7c8ab2ed333a37b86babd4139677,
			fnFmtTimeShort:        fnFmtTimeShort_e50f14b6d2405eade89d9149a1efaee9,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nds_nl"] = func() locales.Translator {
		return &localen{

			locale:                 "nds_NL",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan.", "Feb.", "März", "Apr.", "Mai", "Juni", "Juli", "Aug.", "Sep.", "Okt.", "Nov.", "Dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januaar", "Februaar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktover", "November", "Dezember"},
			daysAbbreviated:        []string{"Sü.", "Ma.", "Di.", "Mi.", "Du.", "Fr.", "Sa."},
			daysNarrow:             []string{"S", "M", "D", "M", "D", "F", "S"},
			daysWide:               []string{"Sünndag", "Maandag", "Dingsdag", "Middeweken", "Dunnersdag", "Freedag", "Sünnavend"},
			periodsAbbreviated:     []string{"vm", "nm"},
			periodsWide:            []string{"vm", "nm"},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"vC", "nC"},
			erasWide:               []string{"vör Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Zentraalaustraalsch Summertiet", "ACST": "Zentraalaustraalsch Standardtiet", "ACWDT": "Westzentraalaustraalsch Summertiet", "ACWST": "Westzentraalaustraalsch Standardtiet", "ADT": "Noordamerikaansch Atlantik-Summertiet", "AEDT": "Oostaustraalsch Summertiet", "AEST": "Oostaustraalsch Standardtiet", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Noordamerikaansch Atlantik-Standardtiet", "AWDT": "Westaustraalsch Summertiet", "AWST": "Westaustraalsch Standardtiet", "BOT": "BOT", "BT": "BT", "CAT": "Zentraalafrikaansch Tiet", "CDT": "Noordamerikaansch zentraal Summertiet", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Noordamerikaansch zentraal Standardtiet", "ChST": "ChST", "EAT": "Oostafrikaansch Tiet", "ECT": "ECT", "EDT": "Noordamerikaansch oosten Summertiet", "EST": "Noordamerikaansch oosten Standardtiet", "GFT": "GFT", "GMT": "Gröönwisch-Welttiet", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "Indien-Tiet", "JDT": "Japaansch Summertiet", "JST": "Japaansch Standardtiet", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Noordamerikaansch Barg-Summertiet", "MESZ": "Zentraaleuropääsch Summertiet", "MEZ": "Zentraaleuropääsch Standardtiet", "MST": "Noordamerikaansch Barg-Standardtiet", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Oosteuropääsch Summertiet", "OEZ": "Oosteuropääsch Standardtiet", "PDT": "Noordamerikaansch Pazifik-Summertiet", "PST": "Noordamerikaansch Pazifik-Standardtiet", "SAST": "Söödafrikaansch Tiet", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "Westafrikaansch Summertiet", "WAT": "Westafrikaansch Standardtiet", "WESZ": "Westeuropääsch Summertiet", "WEZ": "Westeuropääsch Standardtiet", "WIB": "Westindoneesch Tiet", "WIT": "Oostindoneesch Tiet", "WITA": "Indoneesch Zentraaltiet", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_91904f1c4adbb925999bdba7f5796de8,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_cbdb981782261e5a8b8cc19702400c64,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_e61d6b84b00325a0cc262f834ec42619,
			fnFmtTimeLong:         fnFmtTimeLong_1dc6a22a036902eb11698fc0a589743d,
			fnFmtTimeMedium:       fnFmtTimeMedium_d09a7c8ab2ed333a37b86babd4139677,
			fnFmtTimeShort:        fnFmtTimeShort_e50f14b6d2405eade89d9149a1efaee9,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ne"] = func() locales.Translator {
		return &localen{

			locale:                 "ne",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "नेरू", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "जनवरी", "फेब्रुअरी", "मार्च", "अप्रिल", "मे", "जुन", "जुलाई", "अगस्ट", "सेप्टेम्बर", "अक्टोबर", "नोभेम्बर", "डिसेम्बर"},
			monthsNarrow:           []string{"", "जन", "फेब", "मार्च", "अप्र", "मे", "जुन", "जुल", "अग", "सेप", "अक्टो", "नोभे", "डिसे"},
			monthsWide:             []string{"", "जनवरी", "फेब्रुअरी", "मार्च", "अप्रिल", "मे", "जुन", "जुलाई", "अगस्ट", "सेप्टेम्बर", "अक्टोबर", "नोभेम्बर", "डिसेम्बर"},
			daysAbbreviated:        []string{"आइत", "सोम", "मङ्गल", "बुध", "बिहि", "शुक्र", "शनि"},
			daysNarrow:             []string{"आ", "सो", "म", "बु", "बि", "शु", "श"},
			daysShort:              []string{"आइत", "सोम", "मङ्गल", "बुध", "बिहि", "शुक्र", "शनि"},
			daysWide:               []string{"आइतबार", "सोमबार", "मङ्गलबार", "बुधबार", "बिहिबार", "शुक्रबार", "शनिबार"},
			periodsAbbreviated:     []string{"पूर्वाह्न", "अपराह्न"},
			periodsNarrow:          []string{"पूर्वाह्न", "अपराह्न"},
			periodsWide:            []string{"पूर्वाह्न", "अपराह्न"},
			erasAbbreviated:        []string{"ईसा पूर्व", "सन्"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ईसा पूर्व", "सन्"},
			timezones:              map[string]string{"ACDT": "केन्द्रीय अस्ट्रेलिया दिवा समय", "ACST": "केन्द्रीय अस्ट्रेलिया मानक समय", "ACWDT": "केन्द्रीय पश्चिमी अस्ट्रेलिया दिवा समय", "ACWST": "केन्द्रीय पश्चिमी अस्ट्रेलिया मानक समय", "ADT": "एट्लान्टिक दिवा समय", "AEDT": "पूर्वी अस्ट्रेलिया दिवा समय", "AEST": "पूर्वी अस्ट्रेलिया मानक समय", "AKDT": "अलस्काको दिवा समय", "AKST": "अलस्काको मानक समय", "ARST": "अर्जेनटिनी ग्रीष्मकालीन समय", "ART": "अर्जेनटिनी मानक समय", "AST": "एट्लान्टिक मानक समय", "AWDT": "पश्चिमी अस्ट्रेलिया दिवा समय", "AWST": "पश्चिमी अस्ट्रेलिया मानक समय", "BOT": "बोलिभिया समय", "BT": "भुटानी समय", "CAT": "केन्द्रीय अफ्रिकी समय", "CDT": "केन्द्रीय दिवा समय", "CHADT": "चाथाम दिवा समय", "CHAST": "चाथाम मानक समय", "CLST": "चिली ग्रीष्मकालीन समय", "CLT": "चिली मानक समय", "COST": "कोलम्बियाली ग्रीष्मकालीन समय", "COT": "कोलम्बियाली मानक समय", "CST": "केन्द्रीय मानक समय", "ChST": "चामोर्रो मानक समय", "EAT": "पूर्वी अफ्रिकी समय", "ECT": "ईक्वोडोर समय", "EDT": "पूर्वी दिवा समय", "EST": "पूर्वी मानक समय", "GFT": "फ्रेन्च ग्वाना समय", "GMT": "ग्रीनविच मिन समय", "GST": "खाडी मानक समय", "GYT": "गुयाना समय", "HADT": "हवाई-एलुटियन दिवा समय", "HAST": "हवाई-एलुटियन मानक समय", "HAT": "न्यूफाउनल्यान्डको दिवा समय", "HECU": "क्यूबाको दिवा समय", "HEEG": "पूर्वी ग्रीनल्यान्डको ग्रीष्मकालीन समय", "HENOMX": "उत्तर पश्चिम मेक्सिकोको दिवा समय", "HEOG": "पश्चिमी ग्रीनल्यान्डको ग्रीष्मकालीन समय", "HEPM": "सेन्ट पियर्रे र मिक्युलोनको दिवा समय", "HEPMX": "मेक्सिकन प्यासिफिक दिवा समय", "HKST": "हङकङ ग्रीष्मकालीन समय", "HKT": "हङकङ मानक समय", "HNCU": "क्यूबाको मानक समय", "HNEG": "पूर्वी ग्रीनल्यान्डको मानक समय", "HNNOMX": "उत्तर पश्चिम मेक्सिकोको मानक समय", "HNOG": "पश्चिमी ग्रीनल्यान्डको मानक समय", "HNPM": "सेन्ट पियर्रे र मिक्युलोनको मानक समय", "HNPMX": "मेक्सिकन प्यासिफिक मानक समय", "HNT": "न्यूफाउनडल्यान्डको मानक समय", "IST": "भारतीय मानक समय", "JDT": "जापान दिवा समय", "JST": "जापान मानक समय", "LHDT": "लर्ड हावे दिवा समय", "LHST": "लर्ड हावे मानक समय", "MDT": "हिमाली दिवा समय", "MESZ": "केन्द्रीय युरोपेली ग्रीष्मकालीन समय", "MEZ": "केन्द्रीय युरोपेली मानक समय", "MST": "हिमाली मानक समय", "MYT": "मलेसिया समय", "NZDT": "न्यूजिल्यान्ड दिवा समय", "NZST": "न्यूजिल्यान्ड मानक समय", "OESZ": "पूर्वी युरोपेली ग्रीष्मकालीन समय", "OEZ": "पूर्वी युरोपेली मानक समय", "PDT": "प्यासिफिक दिवा समय", "PST": "प्यासिफिक मानक समय", "SAST": "दक्षिण अफ्रिकी समय", "SGT": "सिंगापुर मानक समय", "SRT": "सुरिनामा समय", "TMST": "तुर्कमेनिस्तान ग्रीष्मकालीन मानक समय", "TMT": "तुर्कमेनिस्तान मानक समय", "UYST": "उरुग्वे ग्रीष्मकालीन समय", "UYT": "उरूग्वे मानक समय", "VET": "भेनेज्युएला समय", "WARST": "पश्चिमी अर्जेनटिनी ग्रीष्मकालीन समय", "WART": "पश्चिमी अर्जेनटिनी मानक समय", "WAST": "पश्चिम अफ्रिकी ग्रीष्मकालीन समय", "WAT": "पश्चिम अफ्रिकी मानक समय", "WESZ": "युरोपेली ग्रीष्मकालीन समय", "WEZ": "पश्चिमी युरोपेली मानक समय", "WIB": "पश्चिमी इन्डोनेशिया समय", "WIT": "पूर्वी इन्डोनेशिया समय", "WITA": "केन्द्रीय इन्डोनेशिया समय", "∅∅∅": "ब्राजिलीया ग्रीष्मकालीन समय"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_7e0820ce2c902179e061cbae21c795d7,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ne_in"] = func() locales.Translator {
		return &localen{

			locale:                 "ne_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "जनवरी", "फेब्रुअरी", "मार्च", "अप्रिल", "मे", "जुन", "जुलाई", "अगस्ट", "सेप्टेम्बर", "अक्टोबर", "नोभेम्बर", "डिसेम्बर"},
			monthsNarrow:           []string{"", "जन", "फेब", "मार्च", "अप्र", "मे", "जुन", "जुल", "अग", "सेप", "अक्टो", "नोभे", "डिसे"},
			monthsWide:             []string{"", "जनवरी", "फेब्रुअरी", "मार्च", "अप्रिल", "मे", "जुन", "जुलाई", "अगस्ट", "सेप्टेम्बर", "अक्टोबर", "नोभेम्बर", "डिसेम्बर"},
			daysAbbreviated:        []string{"आइत", "सोम", "मङ्गल", "बुध", "बिहि", "शुक्र", "शनि"},
			daysNarrow:             []string{"आ", "सो", "म", "बु", "बि", "शु", "श"},
			daysShort:              []string{"आइत", "सोम", "मङ्गल", "बुध", "बिहि", "शुक्र", "शनि"},
			daysWide:               []string{"आइतबार", "सोमबार", "मङ्गलबार", "बुधबार", "बिहिबार", "शुक्रबार", "शनिबार"},
			periodsAbbreviated:     []string{"पूर्वाह्न", "अपराह्न"},
			periodsNarrow:          []string{"पूर्वाह्न", "अपराह्न"},
			periodsWide:            []string{"पूर्वाह्न", "अपराह्न"},
			erasAbbreviated:        []string{"ईसा पूर्व", "सन्"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ईसा पूर्व", "सन्"},
			timezones:              map[string]string{"ACDT": "केन्द्रीय अस्ट्रेलिया दिवा समय", "ACST": "केन्द्रीय अस्ट्रेलिया मानक समय", "ACWDT": "केन्द्रीय पश्चिमी अस्ट्रेलिया दिवा समय", "ACWST": "केन्द्रीय पश्चिमी अस्ट्रेलिया मानक समय", "ADT": "एट्लान्टिक दिवा समय", "AEDT": "पूर्वी अस्ट्रेलिया दिवा समय", "AEST": "पूर्वी अस्ट्रेलिया मानक समय", "AKDT": "अलस्काको दिवा समय", "AKST": "अलस्काको मानक समय", "ARST": "अर्जेनटिनी ग्रीष्मकालीन समय", "ART": "अर्जेनटिनी मानक समय", "AST": "एट्लान्टिक मानक समय", "AWDT": "पश्चिमी अस्ट्रेलिया दिवा समय", "AWST": "पश्चिमी अस्ट्रेलिया मानक समय", "BOT": "बोलिभिया समय", "BT": "भुटानी समय", "CAT": "केन्द्रीय अफ्रिकी समय", "CDT": "केन्द्रीय दिवा समय", "CHADT": "चाथाम दिवा समय", "CHAST": "चाथाम मानक समय", "CLST": "चिली ग्रीष्मकालीन समय", "CLT": "चिली मानक समय", "COST": "कोलम्बियाली ग्रीष्मकालीन समय", "COT": "कोलम्बियाली मानक समय", "CST": "केन्द्रीय मानक समय", "ChST": "चामोर्रो मानक समय", "EAT": "पूर्वी अफ्रिकी समय", "ECT": "ईक्वोडोर समय", "EDT": "पूर्वी दिवा समय", "EST": "पूर्वी मानक समय", "GFT": "फ्रेन्च ग्वाना समय", "GMT": "ग्रीनविच मिन समय", "GST": "खाडी मानक समय", "GYT": "गुयाना समय", "HADT": "हवाई-एलुटियन दिवा समय", "HAST": "हवाई-एलुटियन मानक समय", "HAT": "न्यूफाउनल्यान्डको दिवा समय", "HECU": "क्यूबाको दिवा समय", "HEEG": "पूर्वी ग्रीनल्यान्डको ग्रीष्मकालीन समय", "HENOMX": "उत्तर पश्चिम मेक्सिकोको दिवा समय", "HEOG": "पश्चिमी ग्रीनल्यान्डको ग्रीष्मकालीन समय", "HEPM": "सेन्ट पियर्रे र मिक्युलोनको दिवा समय", "HEPMX": "मेक्सिकन प्यासिफिक दिवा समय", "HKST": "हङकङ ग्रीष्मकालीन समय", "HKT": "हङकङ मानक समय", "HNCU": "क्यूबाको मानक समय", "HNEG": "पूर्वी ग्रीनल्यान्डको मानक समय", "HNNOMX": "उत्तर पश्चिम मेक्सिकोको मानक समय", "HNOG": "पश्चिमी ग्रीनल्यान्डको मानक समय", "HNPM": "सेन्ट पियर्रे र मिक्युलोनको मानक समय", "HNPMX": "मेक्सिकन प्यासिफिक मानक समय", "HNT": "न्यूफाउनडल्यान्डको मानक समय", "IST": "भारतीय मानक समय", "JDT": "जापान दिवा समय", "JST": "जापान मानक समय", "LHDT": "लर्ड हावे दिवा समय", "LHST": "लर्ड हावे मानक समय", "MDT": "हिमाली दिवा समय", "MESZ": "केन्द्रीय युरोपेली ग्रीष्मकालीन समय", "MEZ": "केन्द्रीय युरोपेली मानक समय", "MST": "हिमाली मानक समय", "MYT": "मलेसिया समय", "NZDT": "न्यूजिल्यान्ड दिवा समय", "NZST": "न्यूजिल्यान्ड मानक समय", "OESZ": "पूर्वी युरोपेली ग्रीष्मकालीन समय", "OEZ": "पूर्वी युरोपेली मानक समय", "PDT": "प्यासिफिक दिवा समय", "PST": "प्यासिफिक मानक समय", "SAST": "दक्षिण अफ्रिकी समय", "SGT": "सिंगापुर मानक समय", "SRT": "सुरिनामा समय", "TMST": "तुर्कमेनिस्तान ग्रीष्मकालीन मानक समय", "TMT": "तुर्कमेनिस्तान मानक समय", "UYST": "उरुग्वे ग्रीष्मकालीन समय", "UYT": "उरूग्वे मानक समय", "VET": "भेनेज्युएला समय", "WARST": "पश्चिमी अर्जेनटिनी ग्रीष्मकालीन समय", "WART": "पश्चिमी अर्जेनटिनी मानक समय", "WAST": "पश्चिम अफ्रिकी ग्रीष्मकालीन समय", "WAT": "पश्चिम अफ्रिकी मानक समय", "WESZ": "युरोपेली ग्रीष्मकालीन समय", "WEZ": "पश्चिमी युरोपेली मानक समय", "WIB": "पश्चिमी इन्डोनेशिया समय", "WIT": "पूर्वी इन्डोनेशिया समय", "WITA": "केन्द्रीय इन्डोनेशिया समय", "∅∅∅": "ब्राजिलीया ग्रीष्मकालीन समय"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_7e0820ce2c902179e061cbae21c795d7,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ne_np"] = func() locales.Translator {
		return &localen{

			locale:                 "ne_NP",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "जनवरी", "फेब्रुअरी", "मार्च", "अप्रिल", "मे", "जुन", "जुलाई", "अगस्ट", "सेप्टेम्बर", "अक्टोबर", "नोभेम्बर", "डिसेम्बर"},
			monthsNarrow:           []string{"", "जन", "फेब", "मार्च", "अप्र", "मे", "जुन", "जुल", "अग", "सेप", "अक्टो", "नोभे", "डिसे"},
			monthsWide:             []string{"", "जनवरी", "फेब्रुअरी", "मार्च", "अप्रिल", "मे", "जुन", "जुलाई", "अगस्ट", "सेप्टेम्बर", "अक्टोबर", "नोभेम्बर", "डिसेम्बर"},
			daysAbbreviated:        []string{"आइत", "सोम", "मङ्गल", "बुध", "बिहि", "शुक्र", "शनि"},
			daysNarrow:             []string{"आ", "सो", "म", "बु", "बि", "शु", "श"},
			daysShort:              []string{"आइत", "सोम", "मङ्गल", "बुध", "बिहि", "शुक्र", "शनि"},
			daysWide:               []string{"आइतबार", "सोमबार", "मङ्गलबार", "बुधबार", "बिहिबार", "शुक्रबार", "शनिबार"},
			periodsAbbreviated:     []string{"पूर्वाह्न", "अपराह्न"},
			periodsNarrow:          []string{"पूर्वाह्न", "अपराह्न"},
			periodsWide:            []string{"पूर्वाह्न", "अपराह्न"},
			erasAbbreviated:        []string{"ईसा पूर्व", "सन्"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ईसा पूर्व", "सन्"},
			timezones:              map[string]string{"ACDT": "केन्द्रीय अस्ट्रेलिया दिवा समय", "ACST": "केन्द्रीय अस्ट्रेलिया मानक समय", "ACWDT": "केन्द्रीय पश्चिमी अस्ट्रेलिया दिवा समय", "ACWST": "केन्द्रीय पश्चिमी अस्ट्रेलिया मानक समय", "ADT": "एट्लान्टिक दिवा समय", "AEDT": "पूर्वी अस्ट्रेलिया दिवा समय", "AEST": "पूर्वी अस्ट्रेलिया मानक समय", "AKDT": "अलस्काको दिवा समय", "AKST": "अलस्काको मानक समय", "ARST": "अर्जेनटिनी ग्रीष्मकालीन समय", "ART": "अर्जेनटिनी मानक समय", "AST": "एट्लान्टिक मानक समय", "AWDT": "पश्चिमी अस्ट्रेलिया दिवा समय", "AWST": "पश्चिमी अस्ट्रेलिया मानक समय", "BOT": "बोलिभिया समय", "BT": "भुटानी समय", "CAT": "केन्द्रीय अफ्रिकी समय", "CDT": "केन्द्रीय दिवा समय", "CHADT": "चाथाम दिवा समय", "CHAST": "चाथाम मानक समय", "CLST": "चिली ग्रीष्मकालीन समय", "CLT": "चिली मानक समय", "COST": "कोलम्बियाली ग्रीष्मकालीन समय", "COT": "कोलम्बियाली मानक समय", "CST": "केन्द्रीय मानक समय", "ChST": "चामोर्रो मानक समय", "EAT": "पूर्वी अफ्रिकी समय", "ECT": "ईक्वोडोर समय", "EDT": "पूर्वी दिवा समय", "EST": "पूर्वी मानक समय", "GFT": "फ्रेन्च ग्वाना समय", "GMT": "ग्रीनविच मिन समय", "GST": "खाडी मानक समय", "GYT": "गुयाना समय", "HADT": "हवाई-एलुटियन दिवा समय", "HAST": "हवाई-एलुटियन मानक समय", "HAT": "न्यूफाउनल्यान्डको दिवा समय", "HECU": "क्यूबाको दिवा समय", "HEEG": "पूर्वी ग्रीनल्यान्डको ग्रीष्मकालीन समय", "HENOMX": "उत्तर पश्चिम मेक्सिकोको दिवा समय", "HEOG": "पश्चिमी ग्रीनल्यान्डको ग्रीष्मकालीन समय", "HEPM": "सेन्ट पियर्रे र मिक्युलोनको दिवा समय", "HEPMX": "मेक्सिकन प्यासिफिक दिवा समय", "HKST": "हङकङ ग्रीष्मकालीन समय", "HKT": "हङकङ मानक समय", "HNCU": "क्यूबाको मानक समय", "HNEG": "पूर्वी ग्रीनल्यान्डको मानक समय", "HNNOMX": "उत्तर पश्चिम मेक्सिकोको मानक समय", "HNOG": "पश्चिमी ग्रीनल्यान्डको मानक समय", "HNPM": "सेन्ट पियर्रे र मिक्युलोनको मानक समय", "HNPMX": "मेक्सिकन प्यासिफिक मानक समय", "HNT": "न्यूफाउनडल्यान्डको मानक समय", "IST": "भारतीय मानक समय", "JDT": "जापान दिवा समय", "JST": "जापान मानक समय", "LHDT": "लर्ड हावे दिवा समय", "LHST": "लर्ड हावे मानक समय", "MDT": "हिमाली दिवा समय", "MESZ": "केन्द्रीय युरोपेली ग्रीष्मकालीन समय", "MEZ": "केन्द्रीय युरोपेली मानक समय", "MST": "हिमाली मानक समय", "MYT": "मलेसिया समय", "NZDT": "न्यूजिल्यान्ड दिवा समय", "NZST": "न्यूजिल्यान्ड मानक समय", "OESZ": "पूर्वी युरोपेली ग्रीष्मकालीन समय", "OEZ": "पूर्वी युरोपेली मानक समय", "PDT": "प्यासिफिक दिवा समय", "PST": "प्यासिफिक मानक समय", "SAST": "दक्षिण अफ्रिकी समय", "SGT": "सिंगापुर मानक समय", "SRT": "सुरिनामा समय", "TMST": "तुर्कमेनिस्तान ग्रीष्मकालीन मानक समय", "TMT": "तुर्कमेनिस्तान मानक समय", "UYST": "उरुग्वे ग्रीष्मकालीन समय", "UYT": "उरूग्वे मानक समय", "VET": "भेनेज्युएला समय", "WARST": "पश्चिमी अर्जेनटिनी ग्रीष्मकालीन समय", "WART": "पश्चिमी अर्जेनटिनी मानक समय", "WAST": "पश्चिम अफ्रिकी ग्रीष्मकालीन समय", "WAT": "पश्चिम अफ्रिकी मानक समय", "WESZ": "युरोपेली ग्रीष्मकालीन समय", "WEZ": "पश्चिमी युरोपेली मानक समय", "WIB": "पश्चिमी इन्डोनेशिया समय", "WIT": "पूर्वी इन्डोनेशिया समय", "WITA": "केन्द्रीय इन्डोनेशिया समय", "∅∅∅": "एजोरेस् ग्रीष्मकालीन समय"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_7e0820ce2c902179e061cbae21c795d7,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl"] = func() locales.Translator {
		return &localen{

			locale:                 "nl",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "C$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJ$", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SI$", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Braziliaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_aw"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_AW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "Afl.", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Braziliaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_be"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_BE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Braziliaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c2e9326f17b4ceaeb5497266d0765dd7,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_bq"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_BQ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Azoren-zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_cw"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_CW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "NAf.", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Peruaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_nl"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_NL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Macause zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Macause standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Peruaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_sr"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_SR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "$", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Peruaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nl_sx"] = func() locales.Translator {
		return &localen{

			locale:                 "nl_SX",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "NAf.", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mrt.", "apr.", "mei", "jun.", "jul.", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysNarrow:             []string{"Z", "M", "D", "W", "D", "V", "Z"},
			daysShort:              []string{"zo", "ma", "di", "wo", "do", "vr", "za"},
			daysWide:               []string{"zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"v.Chr.", "n.Chr."},
			erasNarrow:             []string{"v.C.", "n.C."},
			erasWide:               []string{"voor Christus", "na Christus"},
			timezones:              map[string]string{"ACDT": "Midden-Australische zomertijd", "ACST": "Midden-Australische standaardtijd", "ACWDT": "Midden-Australische westelijke zomertijd", "ACWST": "Midden-Australische westelijke standaardtijd", "ADT": "Atlantic-zomertijd", "AEDT": "Oost-Australische zomertijd", "AEST": "Oost-Australische standaardtijd", "AKDT": "Alaska-zomertijd", "AKST": "Alaska-standaardtijd", "ARST": "Argentijnse zomertijd", "ART": "Argentijnse standaardtijd", "AST": "Atlantic-standaardtijd", "AWDT": "West-Australische zomertijd", "AWST": "West-Australische standaardtijd", "BOT": "Boliviaanse tijd", "BT": "Bhutaanse tijd", "CAT": "Centraal-Afrikaanse tijd", "CDT": "Central-zomertijd", "CHADT": "Chatham-zomertijd", "CHAST": "Chatham-standaardtijd", "CLST": "Chileense zomertijd", "CLT": "Chileense standaardtijd", "COST": "Colombiaanse zomertijd", "COT": "Colombiaanse standaardtijd", "CST": "Central-standaardtijd", "ChST": "Chamorro-tijd", "EAT": "Oost-Afrikaanse tijd", "ECT": "Ecuadoraanse tijd", "EDT": "Eastern-zomertijd", "EST": "Eastern-standaardtijd", "GFT": "Frans-Guyaanse tijd", "GMT": "Greenwich Mean Time", "GST": "Golf-standaardtijd", "GYT": "Guyaanse tijd", "HADT": "Hawaii-Aleoetische zomertijd", "HAST": "Hawaii-Aleoetische standaardtijd", "HAT": "Newfoundland-zomertijd", "HECU": "Cubaanse zomertijd", "HEEG": "Oost-Groenlandse zomertijd", "HENOMX": "Noordwest-Mexicaanse zomertijd", "HEOG": "West-Groenlandse zomertijd", "HEPM": "Saint Pierre en Miquelon-zomertijd", "HEPMX": "Mexicaanse Pacific-zomertijd", "HKST": "Hongkongse zomertijd", "HKT": "Hongkongse standaardtijd", "HNCU": "Cubaanse standaardtijd", "HNEG": "Oost-Groenlandse standaardtijd", "HNNOMX": "Noordwest-Mexicaanse standaardtijd", "HNOG": "West-Groenlandse standaardtijd", "HNPM": "Saint Pierre en Miquelon-standaardtijd", "HNPMX": "Mexicaanse Pacific-standaardtijd", "HNT": "Newfoundland-standaardtijd", "IST": "Indiase tijd", "JDT": "Japanse zomertijd", "JST": "Japanse standaardtijd", "LHDT": "Lord Howe-eilandse zomertijd", "LHST": "Lord Howe-eilandse standaardtijd", "MDT": "Mountain-zomertijd", "MESZ": "Midden-Europese zomertijd", "MEZ": "Midden-Europese standaardtijd", "MST": "Mountain-standaardtijd", "MYT": "Maleisische tijd", "NZDT": "Nieuw-Zeelandse zomertijd", "NZST": "Nieuw-Zeelandse standaardtijd", "OESZ": "Oost-Europese zomertijd", "OEZ": "Oost-Europese standaardtijd", "PDT": "Pacific-zomertijd", "PST": "Pacific-standaardtijd", "SAST": "Zuid-Afrikaanse tijd", "SGT": "Singaporese standaardtijd", "SRT": "Surinaamse tijd", "TMST": "Turkmeense zomertijd", "TMT": "Turkmeense standaardtijd", "UYST": "Uruguayaanse zomertijd", "UYT": "Uruguayaanse standaardtijd", "VET": "Venezolaanse tijd", "WARST": "West-Argentijnse zomertijd", "WART": "West-Argentijnse standaardtijd", "WAST": "West-Afrikaanse zomertijd", "WAT": "West-Afrikaanse standaardtijd", "WESZ": "West-Europese zomertijd", "WEZ": "West-Europese standaardtijd", "WIB": "West-Indonesische tijd", "WIT": "Oost-Indonesische tijd", "WITA": "Centraal-Indonesische tijd", "∅∅∅": "Braziliaanse zomertijd"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nmg"] = func() locales.Translator {
		return &localen{

			locale:                 "nmg",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ng1", "ng2", "ng3", "ng4", "ng5", "ng6", "ng7", "ng8", "ng9", "ng10", "ng11", "kris"},
			monthsWide:             []string{"", "ngwɛn matáhra", "ngwɛn ńmba", "ngwɛn ńlal", "ngwɛn ńna", "ngwɛn ńtan", "ngwɛn ńtuó", "ngwɛn hɛmbuɛrí", "ngwɛn lɔmbi", "ngwɛn rɛbvuâ", "ngwɛn wum", "ngwɛn wum navǔr", "krísimin"},
			daysAbbreviated:        []string{"sɔ́n", "mɔ́n", "smb", "sml", "smn", "mbs", "sas"},
			daysNarrow:             []string{"s", "m", "s", "s", "s", "m", "s"},
			daysWide:               []string{"sɔ́ndɔ", "mɔ́ndɔ", "sɔ́ndɔ mafú mába", "sɔ́ndɔ mafú málal", "sɔ́ndɔ mafú mána", "mabágá má sukul", "sásadi"},
			periodsAbbreviated:     []string{"maná", "kugú"},
			periodsWide:            []string{"maná", "kugú"},
			erasAbbreviated:        []string{"BL", "PB"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Bó Lahlɛ̄", "Pfiɛ Burī"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nmg_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "nmg_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ng1", "ng2", "ng3", "ng4", "ng5", "ng6", "ng7", "ng8", "ng9", "ng10", "ng11", "kris"},
			monthsWide:             []string{"", "ngwɛn matáhra", "ngwɛn ńmba", "ngwɛn ńlal", "ngwɛn ńna", "ngwɛn ńtan", "ngwɛn ńtuó", "ngwɛn hɛmbuɛrí", "ngwɛn lɔmbi", "ngwɛn rɛbvuâ", "ngwɛn wum", "ngwɛn wum navǔr", "krísimin"},
			daysAbbreviated:        []string{"sɔ́n", "mɔ́n", "smb", "sml", "smn", "mbs", "sas"},
			daysNarrow:             []string{"s", "m", "s", "s", "s", "m", "s"},
			daysWide:               []string{"sɔ́ndɔ", "mɔ́ndɔ", "sɔ́ndɔ mafú mába", "sɔ́ndɔ mafú málal", "sɔ́ndɔ mafú mána", "mabágá má sukul", "sásadi"},
			periodsAbbreviated:     []string{"maná", "kugú"},
			periodsWide:            []string{"maná", "kugú"},
			erasAbbreviated:        []string{"BL", "PB"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Bó Lahlɛ̄", "Pfiɛ Burī"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nn"] = func() locales.Translator {
		return &localen{

			locale:                 "nn",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "kr", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mars", "apr.", "mai", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"sø.", "må.", "ty.", "on.", "to.", "fr.", "la."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø.", "må.", "ty.", "on.", "to.", "fr.", "la."},
			daysWide:               []string{"søndag", "måndag", "tysdag", "onsdag", "torsdag", "fredag", "laurdag"},
			periodsAbbreviated:     []string{"f.m.", "e.m."},
			periodsNarrow:          []string{"f.m.", "e.m."},
			periodsWide:            []string{"formiddag", "ettermiddag"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"f.Kr.", "e.Kr."},
			timezones:              map[string]string{"ACDT": "sentralaustralsk sommartid", "ACST": "sentralaustralsk standardtid", "ACWDT": "vest-sentralaustralsk sommartid", "ACWST": "vest-sentralaustralsk standardtid", "ADT": "sumartid for den nordamerikanske atlanterhavskysten", "AEDT": "austaustralsk sommartid", "AEST": "austaustralsk standardtid", "AKDT": "alaskisk sumartid", "AKST": "alaskisk normaltid", "ARST": "argentinsk sumartid", "ART": "argentinsk normaltid", "AST": "normaltid for den nordamerikanske atlanterhavskysten", "AWDT": "vestaustralsk sommartid", "AWST": "vestaustralsk standardtid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "sentralafrikansk tid", "CDT": "sumartid for sentrale Nord-Amerika", "CHADT": "sumartid for Chatham", "CHAST": "normaltid for Chatham", "CLST": "chilensk sumartid", "CLT": "chilensk normaltid", "COST": "kolombiansk sumartid", "COT": "kolombiansk normaltid", "CST": "normaltid for sentrale Nord-Amerika", "ChST": "tidssone for Chamorro", "EAT": "austafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "sumartid for den nordamerikansk austkysten", "EST": "normaltid for den nordamerikansk austkysten", "GFT": "tidssone for Fransk Guyana", "GMT": "Greenwich middeltid", "GST": "tidssone for Persiabukta", "GYT": "guyansk tid", "HADT": "sumartid for Hawaii og Aleutene", "HAST": "normaltid for Hawaii og Aleutene", "HAT": "sumartid for Newfoundland", "HECU": "kubansk sumartid", "HEEG": "austgrønlandsk sumartid", "HENOMX": "sumartid for nordvestlege Mexico", "HEOG": "vestgrønlandsk sumartid", "HEPM": "sumartid for Saint-Pierre-et-Miquelon", "HEPMX": "sumartid for den meksikanske stillehavskysten", "HKST": "hongkongkinesisk sumartid", "HKT": "hongkongkinesisk normaltid", "HNCU": "kubansk normaltid", "HNEG": "austgrønlandsk normaltid", "HNNOMX": "normaltid for nordvestlege Mexico", "HNOG": "vestgrønlandsk normaltid", "HNPM": "normaltid for Saint-Pierre-et-Miquelon", "HNPMX": "normaltid for den meksikanske stillehavskysten", "HNT": "normaltid for Newfoundland", "IST": "indisk tid", "JDT": "japansk sumartid", "JST": "japansk normaltid", "LHDT": "sumartid for Lord Howe-øya", "LHST": "normaltid for Lord Howe-øya", "MDT": "sumartid for Rocky Mountains (USA)", "MESZ": "sentraleuropeisk sommartid", "MEZ": "sentraleuropeisk standardtid", "MST": "normaltid for Rocky Mountains (USA)", "MYT": "malaysisk tid", "NZDT": "nyzealandsk sumartid", "NZST": "nyzealandsk normaltid", "OESZ": "austeuropeisk sommartid", "OEZ": "austeuropeisk standardtid", "PDT": "sumartid for den nordamerikanske stillehavskysten", "PST": "normaltid for den nordamerikanske stillehavskysten", "SAST": "sørafrikansk tid", "SGT": "singaporsk tid", "SRT": "surinamsk tid", "TMST": "turkmensk sumartid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sumartid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "vestargentinsk sumartid", "WART": "vestargentinsk normaltid", "WAST": "vestafrikansk sommartid", "WAT": "vestafrikansk standardtid", "WESZ": "vesteuropeisk sommartid", "WEZ": "vesteuropeisk standardtid", "WIB": "vestindonesisk tid", "WIT": "austindonesisk tid", "WITA": "sentralindonesisk tid", "∅∅∅": "peruansk sumartid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_702261155c59dd873e83848c0fe82a23,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nn_no"] = func() locales.Translator {
		return &localen{

			locale:                 "nn_NO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mars", "apr.", "mai", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "des."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"},
			daysAbbreviated:        []string{"sø.", "må.", "ty.", "on.", "to.", "fr.", "la."},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sø.", "må.", "ty.", "on.", "to.", "fr.", "la."},
			daysWide:               []string{"søndag", "måndag", "tysdag", "onsdag", "torsdag", "fredag", "laurdag"},
			periodsAbbreviated:     []string{"f.m.", "e.m."},
			periodsNarrow:          []string{"f.m.", "e.m."},
			periodsWide:            []string{"formiddag", "ettermiddag"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"f.Kr.", "e.Kr."},
			timezones:              map[string]string{"ACDT": "sentralaustralsk sommartid", "ACST": "sentralaustralsk standardtid", "ACWDT": "vest-sentralaustralsk sommartid", "ACWST": "vest-sentralaustralsk standardtid", "ADT": "sumartid for den nordamerikanske atlanterhavskysten", "AEDT": "austaustralsk sommartid", "AEST": "austaustralsk standardtid", "AKDT": "alaskisk sumartid", "AKST": "alaskisk normaltid", "ARST": "argentinsk sumartid", "ART": "argentinsk normaltid", "AST": "normaltid for den nordamerikanske atlanterhavskysten", "AWDT": "vestaustralsk sommartid", "AWST": "vestaustralsk standardtid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "sentralafrikansk tid", "CDT": "sumartid for sentrale Nord-Amerika", "CHADT": "sumartid for Chatham", "CHAST": "normaltid for Chatham", "CLST": "chilensk sumartid", "CLT": "chilensk normaltid", "COST": "kolombiansk sumartid", "COT": "kolombiansk normaltid", "CST": "normaltid for sentrale Nord-Amerika", "ChST": "tidssone for Chamorro", "EAT": "austafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "sumartid for den nordamerikansk austkysten", "EST": "normaltid for den nordamerikansk austkysten", "GFT": "tidssone for Fransk Guyana", "GMT": "Greenwich middeltid", "GST": "tidssone for Persiabukta", "GYT": "guyansk tid", "HADT": "sumartid for Hawaii og Aleutene", "HAST": "normaltid for Hawaii og Aleutene", "HAT": "sumartid for Newfoundland", "HECU": "kubansk sumartid", "HEEG": "austgrønlandsk sumartid", "HENOMX": "sumartid for nordvestlege Mexico", "HEOG": "vestgrønlandsk sumartid", "HEPM": "sumartid for Saint-Pierre-et-Miquelon", "HEPMX": "sumartid for den meksikanske stillehavskysten", "HKST": "hongkongkinesisk sumartid", "HKT": "hongkongkinesisk normaltid", "HNCU": "kubansk normaltid", "HNEG": "austgrønlandsk normaltid", "HNNOMX": "normaltid for nordvestlege Mexico", "HNOG": "vestgrønlandsk normaltid", "HNPM": "normaltid for Saint-Pierre-et-Miquelon", "HNPMX": "normaltid for den meksikanske stillehavskysten", "HNT": "normaltid for Newfoundland", "IST": "indisk tid", "JDT": "japansk sumartid", "JST": "japansk normaltid", "LHDT": "sumartid for Lord Howe-øya", "LHST": "normaltid for Lord Howe-øya", "MDT": "sumartid for Rocky Mountains (USA)", "MESZ": "sentraleuropeisk sommartid", "MEZ": "sentraleuropeisk standardtid", "MST": "normaltid for Rocky Mountains (USA)", "MYT": "malaysisk tid", "NZDT": "nyzealandsk sumartid", "NZST": "nyzealandsk normaltid", "OESZ": "austeuropeisk sommartid", "OEZ": "austeuropeisk standardtid", "PDT": "sumartid for den nordamerikanske stillehavskysten", "PST": "normaltid for den nordamerikanske stillehavskysten", "SAST": "sørafrikansk tid", "SGT": "singaporsk tid", "SRT": "surinamsk tid", "TMST": "turkmensk sumartid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sumartid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "vestargentinsk sumartid", "WART": "vestargentinsk normaltid", "WAST": "vestafrikansk sommartid", "WAT": "vestafrikansk standardtid", "WESZ": "vesteuropeisk sommartid", "WEZ": "vesteuropeisk standardtid", "WIB": "vestindonesisk tid", "WIT": "austindonesisk tid", "WITA": "sentralindonesisk tid", "∅∅∅": "sumartid for Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_702261155c59dd873e83848c0fe82a23,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nnh"] = func() locales.Translator {
		return &localen{

			locale:                 "nnh",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "saŋ tsetsɛ̀ɛ lùm", "saŋ kàg ngwóŋ", "saŋ lepyè shúm", "saŋ cÿó", "saŋ tsɛ̀ɛ cÿó", "saŋ njÿoláʼ", "saŋ tyɛ̀b tyɛ̀b mbʉ̀ŋ", "saŋ mbʉ̀ŋ", "saŋ ngwɔ̀ʼ mbÿɛ", "saŋ tàŋa tsetsáʼ", "saŋ mejwoŋó", "saŋ lùm"},
			monthsWide:             []string{"", "saŋ tsetsɛ̀ɛ lùm", "saŋ kàg ngwóŋ", "saŋ lepyè shúm", "saŋ cÿó", "saŋ tsɛ̀ɛ cÿó", "saŋ njÿoláʼ", "saŋ tyɛ̀b tyɛ̀b mbʉ̀ŋ", "saŋ mbʉ̀ŋ", "saŋ ngwɔ̀ʼ mbÿɛ", "saŋ tàŋa tsetsáʼ", "saŋ mejwoŋó", "saŋ lùm"},
			daysAbbreviated:        []string{"lyɛʼɛ́ sẅíŋtè", "mvfò lyɛ̌ʼ", "mbɔ́ɔntè mvfò lyɛ̌ʼ", "tsètsɛ̀ɛ lyɛ̌ʼ", "mbɔ́ɔntè tsetsɛ̀ɛ lyɛ̌ʼ", "mvfò màga lyɛ̌ʼ", "màga lyɛ̌ʼ"},
			daysShort:              []string{"lyɛʼɛ́ sẅíŋtè", "mvfò lyɛ̌ʼ", "mbɔ́ɔntè mvfò lyɛ̌ʼ", "tsètsɛ̀ɛ lyɛ̌ʼ", "mbɔ́ɔntè tsetsɛ̀ɛ lyɛ̌ʼ", "mvfò màga lyɛ̌ʼ", "màga lyɛ̌ʼ"},
			daysWide:               []string{"lyɛʼɛ́ sẅíŋtè", "mvfò lyɛ̌ʼ", "mbɔ́ɔntè mvfò lyɛ̌ʼ", "tsètsɛ̀ɛ lyɛ̌ʼ", "mbɔ́ɔntè tsetsɛ̀ɛ lyɛ̌ʼ", "mvfò màga lyɛ̌ʼ", "màga lyɛ̌ʼ"},
			periodsAbbreviated:     []string{"mbaʼámbaʼ", "ncwònzém"},
			periodsWide:            []string{"mbaʼámbaʼ", "ncwònzém"},
			erasAbbreviated:        []string{"m.z.Y.", "m.g.n.Y."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"mé zyé Yěsô", "mé gÿo ńzyé Yěsô"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_1d5a2ee3108dbf80d60a1f3c5c0aaf8d,
			fnFmtDateLong:         fnFmtDateLong_53ac6ba1bddae8f8da33e7321fa48492,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nnh_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "nnh_CM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "saŋ tsetsɛ̀ɛ lùm", "saŋ kàg ngwóŋ", "saŋ lepyè shúm", "saŋ cÿó", "saŋ tsɛ̀ɛ cÿó", "saŋ njÿoláʼ", "saŋ tyɛ̀b tyɛ̀b mbʉ̀ŋ", "saŋ mbʉ̀ŋ", "saŋ ngwɔ̀ʼ mbÿɛ", "saŋ tàŋa tsetsáʼ", "saŋ mejwoŋó", "saŋ lùm"},
			monthsWide:             []string{"", "saŋ tsetsɛ̀ɛ lùm", "saŋ kàg ngwóŋ", "saŋ lepyè shúm", "saŋ cÿó", "saŋ tsɛ̀ɛ cÿó", "saŋ njÿoláʼ", "saŋ tyɛ̀b tyɛ̀b mbʉ̀ŋ", "saŋ mbʉ̀ŋ", "saŋ ngwɔ̀ʼ mbÿɛ", "saŋ tàŋa tsetsáʼ", "saŋ mejwoŋó", "saŋ lùm"},
			daysAbbreviated:        []string{"lyɛʼɛ́ sẅíŋtè", "mvfò lyɛ̌ʼ", "mbɔ́ɔntè mvfò lyɛ̌ʼ", "tsètsɛ̀ɛ lyɛ̌ʼ", "mbɔ́ɔntè tsetsɛ̀ɛ lyɛ̌ʼ", "mvfò màga lyɛ̌ʼ", "màga lyɛ̌ʼ"},
			daysShort:              []string{"lyɛʼɛ́ sẅíŋtè", "mvfò lyɛ̌ʼ", "mbɔ́ɔntè mvfò lyɛ̌ʼ", "tsètsɛ̀ɛ lyɛ̌ʼ", "mbɔ́ɔntè tsetsɛ̀ɛ lyɛ̌ʼ", "mvfò màga lyɛ̌ʼ", "màga lyɛ̌ʼ"},
			daysWide:               []string{"lyɛʼɛ́ sẅíŋtè", "mvfò lyɛ̌ʼ", "mbɔ́ɔntè mvfò lyɛ̌ʼ", "tsètsɛ̀ɛ lyɛ̌ʼ", "mbɔ́ɔntè tsetsɛ̀ɛ lyɛ̌ʼ", "mvfò màga lyɛ̌ʼ", "màga lyɛ̌ʼ"},
			periodsAbbreviated:     []string{"mbaʼámbaʼ", "ncwònzém"},
			periodsWide:            []string{"mbaʼámbaʼ", "ncwònzém"},
			erasAbbreviated:        []string{"m.z.Y.", "m.g.n.Y."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"mé zyé Yěsô", "mé gÿo ńzyé Yěsô"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_44f743dae0936af6184f1cad1624e543,
			fnFmtCurrency:         fnFmtCurrency_53723702774dedd2122c38a68dcef3da,
			fnFmtDateFull:         fnFmtDateFull_1d5a2ee3108dbf80d60a1f3c5c0aaf8d,
			fnFmtDateLong:         fnFmtDateLong_53ac6ba1bddae8f8da33e7321fa48492,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nus"] = func() locales.Translator {
		return &localen{

			locale:                 "nus",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "£", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Tiop", "Pɛt", "Duɔ̱ɔ̱", "Guak", "Duä", "Kor", "Pay", "Thoo", "Tɛɛ", "Laa", "Kur", "Tid"},
			monthsNarrow:           []string{"", "T", "P", "D", "G", "D", "K", "P", "T", "T", "L", "K", "T"},
			monthsWide:             []string{"", "Tiop thar pɛt", "Pɛt", "Duɔ̱ɔ̱ŋ", "Guak", "Duät", "Kornyoot", "Pay yie̱tni", "Tho̱o̱r", "Tɛɛr", "Laath", "Kur", "Tio̱p in di̱i̱t"},
			daysAbbreviated:        []string{"Cäŋ", "Jiec", "Rɛw", "Diɔ̱k", "Ŋuaan", "Dhieec", "Bäkɛl"},
			daysNarrow:             []string{"C", "J", "R", "D", "Ŋ", "D", "B"},
			daysWide:               []string{"Cäŋ kuɔth", "Jiec la̱t", "Rɛw lätni", "Diɔ̱k lätni", "Ŋuaan lätni", "Dhieec lätni", "Bäkɛl lätni"},
			periodsAbbreviated:     []string{"RW", "TŊ"},
			periodsWide:            []string{"RW", "TŊ"},
			erasAbbreviated:        []string{"AY", "ƐY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"A ka̱n Yecu ni dap", "Ɛ ca Yecu dap"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c2e9326f17b4ceaeb5497266d0765dd7,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_fb1e9e6f7242212344122d3c0e63f9d6,
			fnFmtTimeLong:         fnFmtTimeLong_5643f7e25a7c963d538012968ff836de,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nus_ss"] = func() locales.Translator {
		return &localen{

			locale:                 "nus_SS",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Tiop", "Pɛt", "Duɔ̱ɔ̱", "Guak", "Duä", "Kor", "Pay", "Thoo", "Tɛɛ", "Laa", "Kur", "Tid"},
			monthsNarrow:           []string{"", "T", "P", "D", "G", "D", "K", "P", "T", "T", "L", "K", "T"},
			monthsWide:             []string{"", "Tiop thar pɛt", "Pɛt", "Duɔ̱ɔ̱ŋ", "Guak", "Duät", "Kornyoot", "Pay yie̱tni", "Tho̱o̱r", "Tɛɛr", "Laath", "Kur", "Tio̱p in di̱i̱t"},
			daysAbbreviated:        []string{"Cäŋ", "Jiec", "Rɛw", "Diɔ̱k", "Ŋuaan", "Dhieec", "Bäkɛl"},
			daysNarrow:             []string{"C", "J", "R", "D", "Ŋ", "D", "B"},
			daysWide:               []string{"Cäŋ kuɔth", "Jiec la̱t", "Rɛw lätni", "Diɔ̱k lätni", "Ŋuaan lätni", "Dhieec lätni", "Bäkɛl lätni"},
			periodsAbbreviated:     []string{"RW", "TŊ"},
			periodsWide:            []string{"RW", "TŊ"},
			erasAbbreviated:        []string{"AY", "ƐY"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"A ka̱n Yecu ni dap", "Ɛ ca Yecu dap"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c2e9326f17b4ceaeb5497266d0765dd7,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_fb1e9e6f7242212344122d3c0e63f9d6,
			fnFmtTimeLong:         fnFmtTimeLong_5643f7e25a7c963d538012968ff836de,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nyn"] = func() locales.Translator {
		return &localen{

			locale:            "nyn",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "KBZ", "KBR", "KST", "KKN", "KTN", "KMK", "KMS", "KMN", "KMW", "KKM", "KNK", "KNB"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Okwokubanza", "Okwakabiri", "Okwakashatu", "Okwakana", "Okwakataana", "Okwamukaaga", "Okwamushanju", "Okwamunaana", "Okwamwenda", "Okwaikumi", "Okwaikumi na kumwe", "Okwaikumi na ibiri"},
			daysAbbreviated:   []string{"SAN", "ORK", "OKB", "OKS", "OKN", "OKT", "OMK"},
			daysNarrow:        []string{"S", "K", "R", "S", "N", "T", "M"},
			daysWide:          []string{"Sande", "Orwokubanza", "Orwakabiri", "Orwakashatu", "Orwakana", "Orwakataano", "Orwamukaaga"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Kurisito Atakaijire", "Kurisito Yaijire"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["nyn_ug"] = func() locales.Translator {
		return &localen{

			locale:            "nyn_UG",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "KBZ", "KBR", "KST", "KKN", "KTN", "KMK", "KMS", "KMN", "KMW", "KKM", "KNK", "KNB"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Okwokubanza", "Okwakabiri", "Okwakashatu", "Okwakana", "Okwakataana", "Okwamukaaga", "Okwamushanju", "Okwamunaana", "Okwamwenda", "Okwaikumi", "Okwaikumi na kumwe", "Okwaikumi na ibiri"},
			daysAbbreviated:   []string{"SAN", "ORK", "OKB", "OKS", "OKN", "OKT", "OMK"},
			daysNarrow:        []string{"S", "K", "R", "S", "N", "T", "M"},
			daysWide:          []string{"Sande", "Orwokubanza", "Orwakabiri", "Orwakashatu", "Orwakana", "Orwakataano", "Orwamukaaga"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Kurisito Atakaijire", "Kurisito Yaijire"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["om"] = func() locales.Translator {
		return &localen{

			locale:             "om",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "Br", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Ama", "Gur", "Bit", "Elb", "Cam", "Wax", "Ado", "Hag", "Ful", "Onk", "Sad", "Mud"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Amajjii", "Guraandhala", "Bitooteessa", "Elba", "Caamsa", "Waxabajjii", "Adooleessa", "Hagayya", "Fuulbana", "Onkololeessa", "Sadaasa", "Muddee"},
			daysAbbreviated:    []string{"Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"},
			daysWide:           []string{"Dilbata", "Wiixata", "Qibxata", "Roobii", "Kamiisa", "Jimaata", "Sanbata"},
			periodsAbbreviated: []string{"WD", "WB"},
			periodsNarrow:      []string{"WD", "WB"},
			periodsWide:        []string{"WD", "WB"},
			erasAbbreviated:    []string{"BCE", "CE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["om_et"] = func() locales.Translator {
		return &localen{

			locale:             "om_ET",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Ama", "Gur", "Bit", "Elb", "Cam", "Wax", "Ado", "Hag", "Ful", "Onk", "Sad", "Mud"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Amajjii", "Guraandhala", "Bitooteessa", "Elba", "Caamsa", "Waxabajjii", "Adooleessa", "Hagayya", "Fuulbana", "Onkololeessa", "Sadaasa", "Muddee"},
			daysAbbreviated:    []string{"Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"},
			daysWide:           []string{"Dilbata", "Wiixata", "Qibxata", "Roobii", "Kamiisa", "Jimaata", "Sanbata"},
			periodsAbbreviated: []string{"WD", "WB"},
			periodsNarrow:      []string{"WD", "WB"},
			periodsWide:        []string{"WD", "WB"},
			erasAbbreviated:    []string{"BCE", "CE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["om_ke"] = func() locales.Translator {
		return &localen{

			locale:             "om_KE",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Ama", "Gur", "Bit", "Elb", "Cam", "Wax", "Ado", "Hag", "Ful", "Onk", "Sad", "Mud"},
			monthsNarrow:       []string{"", "A", "G", "B", "E", "C", "W", "A", "H", "F", "O", "S", "M"},
			monthsWide:         []string{"", "Amajjii", "Guraandhala", "Bitooteessa", "Elba", "Caamsa", "Waxabajjii", "Adooleessa", "Hagayya", "Fuulbana", "Onkololeessa", "Sadaasa", "Muddee"},
			daysAbbreviated:    []string{"Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"},
			daysNarrow:         []string{"D", "W", "Q", "R", "K", "J", "S"},
			daysShort:          []string{"Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"},
			daysWide:           []string{"Dilbata", "Wiixata", "Qibxata", "Roobii", "Kamiisa", "Jimaata", "Sanbata"},
			periodsAbbreviated: []string{"WD", "WB"},
			periodsNarrow:      []string{"WD", "WB"},
			periodsWide:        []string{"WD", "WB"},
			erasAbbreviated:    []string{"", ""},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["or"] = func() locales.Translator {
		return &localen{

			locale:                 "or",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ଜାନୁଆରୀ", "ଫେବୃଆରୀ", "ମାର୍ଚ୍ଚ", "ଅପ୍ରେଲ", "ମଇ", "ଜୁନ", "ଜୁଲାଇ", "ଅଗଷ୍ଟ", "ସେପ୍ଟେମ୍ବର", "ଅକ୍ଟୋବର", "ନଭେମ୍ବର", "ଡିସେମ୍ବର"},
			monthsNarrow:           []string{"", "ଜା", "ଫେ", "ମା", "ଅ", "ମଇ", "ଜୁ", "ଜୁ", "ଅ", "ସେ", "ଅ", "ନ", "ଡି"},
			monthsWide:             []string{"", "ଜାନୁଆରୀ", "ଫେବୃଆରୀ", "ମାର୍ଚ୍ଚ", "ଅପ୍ରେଲ", "ମଇ", "ଜୁନ", "ଜୁଲାଇ", "ଅଗଷ୍ଟ", "ସେପ୍ଟେମ୍ବର", "ଅକ୍ଟୋବର", "ନଭେମ୍ବର", "ଡିସେମ୍ବର"},
			daysAbbreviated:        []string{"ରବି", "ସୋମ", "ମଙ୍ଗଳ", "ବୁଧ", "ଗୁରୁ", "ଶୁକ୍ର", "ଶନି"},
			daysNarrow:             []string{"ର", "ସୋ", "ମ", "ବୁ", "ଗୁ", "ଶୁ", "ଶ"},
			daysShort:              []string{"ରବି", "ସୋମ", "ମଙ୍ଗଳ", "ବୁଧ", "ଗୁରୁ", "ଶୁକ୍ର", "ଶନି"},
			daysWide:               []string{"ରବିବାର", "ସୋମବାର", "ମଙ୍ଗଳବାର", "ବୁଧବାର", "ଗୁରୁବାର", "ଶୁକ୍ରବାର", "ଶନିବାର"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"ପୂ", "ଅ"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ଖ୍ରୀଷ୍ଟପୂର୍ବ", "ଖ୍ରୀଷ୍ଟାବ୍ଦ"},
			timezones:              map[string]string{"ACDT": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ଦିବାଲୋକ ସମୟ", "ACST": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ମାନାଙ୍କ ସମୟ", "ACWDT": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ପଶ୍ଚିମ ଦିବାଲୋକ ସମୟ", "ACWST": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ପଶ୍ଚିମ ମାନାଙ୍କ ସମୟ", "ADT": "ଆଟଲାଣ୍ଟିକ୍ ଦିବାଲୋକ ସମୟ", "AEDT": "ଅଷ୍ଟ୍ରେଲିୟ ପୂର୍ବ ଦିବାଲୋକ ସମୟ", "AEST": "ଅଷ୍ଟ୍ରେଲିୟ ପୂର୍ବ ମାନାଙ୍କ ସମୟ", "AKDT": "ଆଲାସ୍କା ଦିବାଲୋକ ସମୟ", "AKST": "ଆଲାସ୍କା ମାନାଙ୍କ ସମୟ", "ARST": "ଆର୍ଜେଣ୍ଟିନା ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "ART": "ଆର୍ଜେଣ୍ଟିନା ମାନାଙ୍କ ସମୟ", "AST": "ଆଟଲାଣ୍ଟିକ୍ ମାନାଙ୍କ ସମୟ", "AWDT": "ଅଷ୍ଟ୍ରେଲିୟ ପଶ୍ଚିମ ଦିବାଲୋକ ସମୟ", "AWST": "ଅଷ୍ଟ୍ରେଲିୟ ପଶ୍ଚିମ ମାନାଙ୍କ ସମୟ", "BOT": "ବଲିଭିଆ ସମୟ", "BT": "ଭୁଟାନ ସମୟ", "CAT": "ମଧ୍ୟ ଆଫ୍ରିକା ସମୟ", "CDT": "କେନ୍ଦ୍ରୀୟ ଦିବାଲୋକ ସମୟ", "CHADT": "ଚାଥାମ୍\u200c ଦିବାଲୋକ ସମୟ", "CHAST": "ଚାଥାମ୍\u200c ମାନାଙ୍କ ସମୟ", "CLST": "ଚିଲି ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "CLT": "ଚିଲି ମାନାଙ୍କ ସମୟ", "COST": "କଲମ୍ବିଆ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "COT": "କଲମ୍ବିଆ ମାନାଙ୍କ ସମୟ", "CST": "କେନ୍ଦ୍ରୀୟ ମାନାଙ୍କ ସମୟ", "ChST": "ଚାମୋରୋ ମାନାଙ୍କ ସମୟ", "EAT": "ପୂର୍ବ ଆଫ୍ରିକା ସମୟ", "ECT": "ଇକ୍ୱେଡର ସମୟ", "EDT": "ପୂର୍ବାଞ୍ଚଳ ଦିବାଲୋକ ସମୟ", "EST": "ପୂର୍ବାଞ୍ଚଳ ମାନାଙ୍କ ସମୟ", "GFT": "ଫ୍ରେଞ୍ଚ ଗୁଆନା ସମୟ", "GMT": "ଗ୍ରୀନୱିଚ୍ ମିନ୍ ସମୟ", "GST": "ଗଲ୍ଫ ମାନାଙ୍କ ସମୟ", "GYT": "ଗୁଏନା ସମୟ", "HADT": "ହୱାଇ-ଆଲେଉଟିୟ ଦିବାଲୋକ ସମୟ", "HAST": "ହୱାଇ-ଆଲେଉଟିୟ ମାନାଙ୍କ ସମୟ", "HAT": "ନ୍ୟୁଫାଉଣ୍ଡଲ୍ୟାଣ୍ଡ୍ ଦିବାଲୋକ ସମୟ", "HECU": "କ୍ୟୁବା ଦିବାଲୋକ ସମୟ", "HEEG": "ପୂର୍ବ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "HENOMX": "ଉତ୍ତରପଶ୍ଚିମ ମେକ୍ସିକୋ ଦିବାଲୋକ ସମୟ", "HEOG": "ପଶ୍ଚିମ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ଗ୍ରୀଷ୍ମ ସମୟ", "HEPM": "ସେଣ୍ଟ. ପିଏରେ ଏବଂ ମିକ୍ୟୁଲୋନ୍ ଦିବାଲୋକ ସମୟ", "HEPMX": "ମେକ୍ସିକାନ୍ ପାସିଫିକ୍ ଦିବାଲୋକ ସମୟ", "HKST": "ହଂ କଂ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "HKT": "ହଂ କଂ ମାନାଙ୍କ ସମୟ", "HNCU": "କ୍ୟୁବା ମାନାଙ୍କ ସମୟ", "HNEG": "ପୂର୍ବ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ମାନାଙ୍କ ସମୟ", "HNNOMX": "ଉତ୍ତରପଶ୍ଚିମ ମେକ୍ସିକୋ ମାନାଙ୍କ ସମୟ", "HNOG": "ପଶ୍ଚିମ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ମାନାଙ୍କ ସମୟ", "HNPM": "ସେଣ୍ଟ. ପିଏରେ ଏବଂ ମିକ୍ୟୁଲୋନ୍ ମାନାଙ୍କ ସମୟ", "HNPMX": "ମେକ୍ସିକାନ୍ ପାସିଫିକ୍ ମାନାଙ୍କ ସମୟ", "HNT": "ନ୍ୟୁଫାଉଣ୍ଡଲ୍ୟାଣ୍ଡ୍ ମାନାଙ୍କ ସମୟ", "IST": "ଭାରତ ମାନାଙ୍କ ସମୟ", "JDT": "ଜାପାନ ଦିବାଲୋକ ସମୟ", "JST": "ଜାପାନ ମାନାଙ୍କ ସମୟ", "LHDT": "ଲର୍ଡ ହୋୱେ ଦିବାଲୋକ ସମୟ", "LHST": "ଲର୍ଡ ହୋୱେ ମାନାଙ୍କ ସମୟ", "MDT": "ପାର୍ବତ୍ୟ ଦିବାଲୋକ ସମୟ", "MESZ": "କେନ୍ଦ୍ରୀୟ ୟୁରୋପୀୟ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "MEZ": "କେନ୍ଦ୍ରୀୟ ୟୁରୋପୀୟ ମାନାଙ୍କ ସମୟ", "MST": "ପାର୍ବତ୍ୟ ମାନାଙ୍କ ସମୟ", "MYT": "ମାଲେସିଆ ସମୟ", "NZDT": "ନ୍ୟୁଜିଲାଣ୍ଡ ଦିବାଲୋକ ସମୟ", "NZST": "ନ୍ୟୁଜିଲାଣ୍ଡ ମାନାଙ୍କ ସମୟ", "OESZ": "ପୂର୍ବାଞ୍ଚଳ ୟୁରୋପୀୟ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "OEZ": "ପୂର୍ବାଞ୍ଚଳ ୟୁରୋପୀୟ ମାନାଙ୍କ ସମୟ", "PDT": "ପାସିଫିକ୍ ଦିବାଲୋକ ସମୟ", "PST": "ପାସିଫିକ୍ ମାନାଙ୍କ ସମୟ", "SAST": "ଦକ୍ଷିଣ ଆଫ୍ରିକା ମାନାଙ୍କ ସମୟ", "SGT": "ସିଙ୍ଗାପୁର୍\u200c ମାନାଙ୍କ ସମୟ", "SRT": "ସୁରିନେମ୍\u200c ସମୟ", "TMST": "ତୁର୍କମେନିସ୍ତାନ ଖରାଦିନ ସମୟ", "TMT": "ତୁର୍କମେନିସ୍ତାନ ମାନାଙ୍କ ସମୟ", "UYST": "ଉରୁଗୁଏ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "UYT": "ଉରୁଗୁଏ ମାନାଙ୍କ ସମୟ", "VET": "ଭେନିଜୁଏଲା ସମୟ", "WARST": "ପଶ୍ଚିମ ଆର୍ଜେଣ୍ଟିନା ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "WART": "ପଶ୍ଚିମ ଆର୍ଜେଣ୍ଟିନା ମାନାଙ୍କ ସମୟ", "WAST": "ପଶ୍ଚିମ ଆଫ୍ରିକା ଖରାଦିନ ସମୟ", "WAT": "ପଶ୍ଚିମ ଆଫ୍ରିକା ମାନାଙ୍କ ସମୟ", "WESZ": "ପଶ୍ଚିମାଞ୍ଚଳ ୟୁରୋପୀୟ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "WEZ": "ପଶ୍ଚିମାଞ୍ଚଳ ୟୁରୋପୀୟ ମାନାଙ୍କ ସମୟ", "WIB": "ପଶ୍ଚିମ ଇଣ୍ଡୋନେସିଆ ସମୟ", "WIT": "ପୂର୍ବ ଇଣ୍ଡୋନେସିଆ ସମୟ", "WITA": "ମଧ୍ୟ ଇଣ୍ଡୋନେସିଆ ସମୟ", "∅∅∅": "ଆଜୋରେସ୍ ଗ୍ରୀଷ୍ମକାଳ ସମୟ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_f7f41a6ef5c99a912fc7e7451c5c3029,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["or_in"] = func() locales.Translator {
		return &localen{

			locale:                 "or_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 3, 4, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ଜାନୁଆରୀ", "ଫେବୃଆରୀ", "ମାର୍ଚ୍ଚ", "ଅପ୍ରେଲ", "ମଇ", "ଜୁନ", "ଜୁଲାଇ", "ଅଗଷ୍ଟ", "ସେପ୍ଟେମ୍ବର", "ଅକ୍ଟୋବର", "ନଭେମ୍ବର", "ଡିସେମ୍ବର"},
			monthsNarrow:           []string{"", "ଜା", "ଫେ", "ମା", "ଅ", "ମଇ", "ଜୁ", "ଜୁ", "ଅ", "ସେ", "ଅ", "ନ", "ଡି"},
			monthsWide:             []string{"", "ଜାନୁଆରୀ", "ଫେବୃଆରୀ", "ମାର୍ଚ୍ଚ", "ଅପ୍ରେଲ", "ମଇ", "ଜୁନ", "ଜୁଲାଇ", "ଅଗଷ୍ଟ", "ସେପ୍ଟେମ୍ବର", "ଅକ୍ଟୋବର", "ନଭେମ୍ବର", "ଡିସେମ୍ବର"},
			daysAbbreviated:        []string{"ରବି", "ସୋମ", "ମଙ୍ଗଳ", "ବୁଧ", "ଗୁରୁ", "ଶୁକ୍ର", "ଶନି"},
			daysNarrow:             []string{"ର", "ସୋ", "ମ", "ବୁ", "ଗୁ", "ଶୁ", "ଶ"},
			daysShort:              []string{"ରବି", "ସୋମ", "ମଙ୍ଗଳ", "ବୁଧ", "ଗୁରୁ", "ଶୁକ୍ର", "ଶନି"},
			daysWide:               []string{"ରବିବାର", "ସୋମବାର", "ମଙ୍ଗଳବାର", "ବୁଧବାର", "ଗୁରୁବାର", "ଶୁକ୍ରବାର", "ଶନିବାର"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"ପୂ", "ଅ"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ଖ୍ରୀଷ୍ଟପୂର୍ବ", "ଖ୍ରୀଷ୍ଟାବ୍ଦ"},
			timezones:              map[string]string{"ACDT": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ଦିବାଲୋକ ସମୟ", "ACST": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ମାନାଙ୍କ ସମୟ", "ACWDT": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ପଶ୍ଚିମ ଦିବାଲୋକ ସମୟ", "ACWST": "ଅଷ୍ଟ୍ରେଲିୟ ମଧ୍ୟ ପଶ୍ଚିମ ମାନାଙ୍କ ସମୟ", "ADT": "ଆଟଲାଣ୍ଟିକ୍ ଦିବାଲୋକ ସମୟ", "AEDT": "ଅଷ୍ଟ୍ରେଲିୟ ପୂର୍ବ ଦିବାଲୋକ ସମୟ", "AEST": "ଅଷ୍ଟ୍ରେଲିୟ ପୂର୍ବ ମାନାଙ୍କ ସମୟ", "AKDT": "ଆଲାସ୍କା ଦିବାଲୋକ ସମୟ", "AKST": "ଆଲାସ୍କା ମାନାଙ୍କ ସମୟ", "ARST": "ଆର୍ଜେଣ୍ଟିନା ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "ART": "ଆର୍ଜେଣ୍ଟିନା ମାନାଙ୍କ ସମୟ", "AST": "ଆଟଲାଣ୍ଟିକ୍ ମାନାଙ୍କ ସମୟ", "AWDT": "ଅଷ୍ଟ୍ରେଲିୟ ପଶ୍ଚିମ ଦିବାଲୋକ ସମୟ", "AWST": "ଅଷ୍ଟ୍ରେଲିୟ ପଶ୍ଚିମ ମାନାଙ୍କ ସମୟ", "BOT": "ବଲିଭିଆ ସମୟ", "BT": "ଭୁଟାନ ସମୟ", "CAT": "ମଧ୍ୟ ଆଫ୍ରିକା ସମୟ", "CDT": "କେନ୍ଦ୍ରୀୟ ଦିବାଲୋକ ସମୟ", "CHADT": "ଚାଥାମ୍\u200c ଦିବାଲୋକ ସମୟ", "CHAST": "ଚାଥାମ୍\u200c ମାନାଙ୍କ ସମୟ", "CLST": "ଚିଲି ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "CLT": "ଚିଲି ମାନାଙ୍କ ସମୟ", "COST": "କଲମ୍ବିଆ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "COT": "କଲମ୍ବିଆ ମାନାଙ୍କ ସମୟ", "CST": "କେନ୍ଦ୍ରୀୟ ମାନାଙ୍କ ସମୟ", "ChST": "ଚାମୋରୋ ମାନାଙ୍କ ସମୟ", "EAT": "ପୂର୍ବ ଆଫ୍ରିକା ସମୟ", "ECT": "ଇକ୍ୱେଡର ସମୟ", "EDT": "ପୂର୍ବାଞ୍ଚଳ ଦିବାଲୋକ ସମୟ", "EST": "ପୂର୍ବାଞ୍ଚଳ ମାନାଙ୍କ ସମୟ", "GFT": "ଫ୍ରେଞ୍ଚ ଗୁଆନା ସମୟ", "GMT": "ଗ୍ରୀନୱିଚ୍ ମିନ୍ ସମୟ", "GST": "ଗଲ୍ଫ ମାନାଙ୍କ ସମୟ", "GYT": "ଗୁଏନା ସମୟ", "HADT": "ହୱାଇ-ଆଲେଉଟିୟ ଦିବାଲୋକ ସମୟ", "HAST": "ହୱାଇ-ଆଲେଉଟିୟ ମାନାଙ୍କ ସମୟ", "HAT": "ନ୍ୟୁଫାଉଣ୍ଡଲ୍ୟାଣ୍ଡ୍ ଦିବାଲୋକ ସମୟ", "HECU": "କ୍ୟୁବା ଦିବାଲୋକ ସମୟ", "HEEG": "ପୂର୍ବ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "HENOMX": "ଉତ୍ତରପଶ୍ଚିମ ମେକ୍ସିକୋ ଦିବାଲୋକ ସମୟ", "HEOG": "ପଶ୍ଚିମ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ଗ୍ରୀଷ୍ମ ସମୟ", "HEPM": "ସେଣ୍ଟ. ପିଏରେ ଏବଂ ମିକ୍ୟୁଲୋନ୍ ଦିବାଲୋକ ସମୟ", "HEPMX": "ମେକ୍ସିକାନ୍ ପାସିଫିକ୍ ଦିବାଲୋକ ସମୟ", "HKST": "ହଂ କଂ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "HKT": "ହଂ କଂ ମାନାଙ୍କ ସମୟ", "HNCU": "କ୍ୟୁବା ମାନାଙ୍କ ସମୟ", "HNEG": "ପୂର୍ବ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ମାନାଙ୍କ ସମୟ", "HNNOMX": "ଉତ୍ତରପଶ୍ଚିମ ମେକ୍ସିକୋ ମାନାଙ୍କ ସମୟ", "HNOG": "ପଶ୍ଚିମ ଗ୍ରୀନଲ୍ୟାଣ୍ଡ୍ ମାନାଙ୍କ ସମୟ", "HNPM": "ସେଣ୍ଟ. ପିଏରେ ଏବଂ ମିକ୍ୟୁଲୋନ୍ ମାନାଙ୍କ ସମୟ", "HNPMX": "ମେକ୍ସିକାନ୍ ପାସିଫିକ୍ ମାନାଙ୍କ ସମୟ", "HNT": "ନ୍ୟୁଫାଉଣ୍ଡଲ୍ୟାଣ୍ଡ୍ ମାନାଙ୍କ ସମୟ", "IST": "ଭାରତ ମାନାଙ୍କ ସମୟ", "JDT": "ଜାପାନ ଦିବାଲୋକ ସମୟ", "JST": "ଜାପାନ ମାନାଙ୍କ ସମୟ", "LHDT": "ଲର୍ଡ ହୋୱେ ଦିବାଲୋକ ସମୟ", "LHST": "ଲର୍ଡ ହୋୱେ ମାନାଙ୍କ ସମୟ", "MDT": "ପାର୍ବତ୍ୟ ଦିବାଲୋକ ସମୟ", "MESZ": "କେନ୍ଦ୍ରୀୟ ୟୁରୋପୀୟ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "MEZ": "କେନ୍ଦ୍ରୀୟ ୟୁରୋପୀୟ ମାନାଙ୍କ ସମୟ", "MST": "ପାର୍ବତ୍ୟ ମାନାଙ୍କ ସମୟ", "MYT": "ମାଲେସିଆ ସମୟ", "NZDT": "ନ୍ୟୁଜିଲାଣ୍ଡ ଦିବାଲୋକ ସମୟ", "NZST": "ନ୍ୟୁଜିଲାଣ୍ଡ ମାନାଙ୍କ ସମୟ", "OESZ": "ପୂର୍ବାଞ୍ଚଳ ୟୁରୋପୀୟ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "OEZ": "ପୂର୍ବାଞ୍ଚଳ ୟୁରୋପୀୟ ମାନାଙ୍କ ସମୟ", "PDT": "ପାସିଫିକ୍ ଦିବାଲୋକ ସମୟ", "PST": "ପାସିଫିକ୍ ମାନାଙ୍କ ସମୟ", "SAST": "ଦକ୍ଷିଣ ଆଫ୍ରିକା ମାନାଙ୍କ ସମୟ", "SGT": "ସିଙ୍ଗାପୁର୍\u200c ମାନାଙ୍କ ସମୟ", "SRT": "ସୁରିନେମ୍\u200c ସମୟ", "TMST": "ତୁର୍କମେନିସ୍ତାନ ଖରାଦିନ ସମୟ", "TMT": "ତୁର୍କମେନିସ୍ତାନ ମାନାଙ୍କ ସମୟ", "UYST": "ଉରୁଗୁଏ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "UYT": "ଉରୁଗୁଏ ମାନାଙ୍କ ସମୟ", "VET": "ଭେନିଜୁଏଲା ସମୟ", "WARST": "ପଶ୍ଚିମ ଆର୍ଜେଣ୍ଟିନା ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "WART": "ପଶ୍ଚିମ ଆର୍ଜେଣ୍ଟିନା ମାନାଙ୍କ ସମୟ", "WAST": "ପଶ୍ଚିମ ଆଫ୍ରିକା ଖରାଦିନ ସମୟ", "WAT": "ପଶ୍ଚିମ ଆଫ୍ରିକା ମାନାଙ୍କ ସମୟ", "WESZ": "ପଶ୍ଚିମାଞ୍ଚଳ ୟୁରୋପୀୟ ଗ୍ରୀଷ୍ମକାଳ ସମୟ", "WEZ": "ପଶ୍ଚିମାଞ୍ଚଳ ୟୁରୋପୀୟ ମାନାଙ୍କ ସମୟ", "WIB": "ପଶ୍ଚିମ ଇଣ୍ଡୋନେସିଆ ସମୟ", "WIT": "ପୂର୍ବ ଇଣ୍ଡୋନେସିଆ ସମୟ", "WITA": "ମଧ୍ୟ ଇଣ୍ଡୋନେସିଆ ସମୟ", "∅∅∅": "ବ୍ରାସିଲିଆ ଗ୍ରୀଷ୍ମକାଳ ସମୟ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_f7f41a6ef5c99a912fc7e7451c5c3029,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["os"] = func() locales.Translator {
		return &localen{

			locale:                 "os",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "₾", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "янв.", "фев.", "мар.", "апр.", "майы", "июны", "июлы", "авг.", "сен.", "окт.", "ноя.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январы", "февралы", "мартъийы", "апрелы", "майы", "июны", "июлы", "августы", "сентябры", "октябры", "ноябры", "декабры"},
			daysAbbreviated:        []string{"хцб", "крс", "дцг", "ӕрт", "цпр", "мрб", "сбт"},
			daysNarrow:             []string{"Х", "К", "Д", "Ӕ", "Ц", "М", "С"},
			daysWide:               []string{"хуыцаубон", "къуырисӕр", "дыццӕг", "ӕртыццӕг", "цыппӕрӕм", "майрӕмбон", "сабат"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"ӕмбисбоны размӕ", "ӕмбисбоны фӕстӕ"},
			erasAbbreviated:        []string{"н.д.а.", "н.д."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Гринвичы рӕстӕмбис рӕстӕг", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Астӕуккаг Европӕйаг сӕрдыгон рӕстӕг", "MEZ": "Астӕуккаг Европӕйаг стандартон рӕстӕг", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Скӕсӕн Европӕйаг сӕрдыгон рӕстӕг", "OEZ": "Скӕсӕн Европӕйаг стандартон рӕстӕг", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Ныгъуылӕн Европӕйаг сӕрдыгон рӕстӕг", "WEZ": "Ныгъуылӕн Европӕйаг стандартон рӕстӕг", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_204cc4159c06830dc072d7ffc4471b59,
			fnFmtCurrency:         fnFmtCurrency_7b65395a77761388c242d97c0d4bfe5b,
			fnFmtDateFull:         fnFmtDateFull_a7dc1bab86baf3ec90633b4422fb9cba,
			fnFmtDateLong:         fnFmtDateLong_4874c47573ec41408036722eb0d2c6a6,
			fnFmtDateMedium:       fnFmtDateMedium_5a5202b511779ecb07983ce846c85742,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["os_ge"] = func() locales.Translator {
		return &localen{

			locale:                 "os_GE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "янв.", "фев.", "мар.", "апр.", "майы", "июны", "июлы", "авг.", "сен.", "окт.", "ноя.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январы", "февралы", "мартъийы", "апрелы", "майы", "июны", "июлы", "августы", "сентябры", "октябры", "ноябры", "декабры"},
			daysAbbreviated:        []string{"хцб", "крс", "дцг", "ӕрт", "цпр", "мрб", "сбт"},
			daysNarrow:             []string{"Х", "К", "Д", "Ӕ", "Ц", "М", "С"},
			daysWide:               []string{"хуыцаубон", "къуырисӕр", "дыццӕг", "ӕртыццӕг", "цыппӕрӕм", "майрӕмбон", "сабат"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"ӕмбисбоны размӕ", "ӕмбисбоны фӕстӕ"},
			erasAbbreviated:        []string{"н.д.а.", "н.д."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Гринвичы рӕстӕмбис рӕстӕг", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Астӕуккаг Европӕйаг сӕрдыгон рӕстӕг", "MEZ": "Астӕуккаг Европӕйаг стандартон рӕстӕг", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Скӕсӕн Европӕйаг сӕрдыгон рӕстӕг", "OEZ": "Скӕсӕн Европӕйаг стандартон рӕстӕг", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Ныгъуылӕн Европӕйаг сӕрдыгон рӕстӕг", "WEZ": "Ныгъуылӕн Европӕйаг стандартон рӕстӕг", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_204cc4159c06830dc072d7ffc4471b59,
			fnFmtCurrency:         fnFmtCurrency_7b65395a77761388c242d97c0d4bfe5b,
			fnFmtDateFull:         fnFmtDateFull_a7dc1bab86baf3ec90633b4422fb9cba,
			fnFmtDateLong:         fnFmtDateLong_4874c47573ec41408036722eb0d2c6a6,
			fnFmtDateMedium:       fnFmtDateMedium_5a5202b511779ecb07983ce846c85742,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["os_ru"] = func() locales.Translator {
		return &localen{

			locale:                 "os_RU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "янв.", "фев.", "мар.", "апр.", "майы", "июны", "июлы", "авг.", "сен.", "окт.", "ноя.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январы", "февралы", "мартъийы", "апрелы", "майы", "июны", "июлы", "августы", "сентябры", "октябры", "ноябры", "декабры"},
			daysAbbreviated:        []string{"хцб", "крс", "дцг", "ӕрт", "цпр", "мрб", "сбт"},
			daysNarrow:             []string{"Х", "К", "Д", "Ӕ", "Ц", "М", "С"},
			daysWide:               []string{"хуыцаубон", "къуырисӕр", "дыццӕг", "ӕртыццӕг", "цыппӕрӕм", "майрӕмбон", "сабат"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"ӕмбисбоны размӕ", "ӕмбисбоны фӕстӕ"},
			erasAbbreviated:        []string{"н.д.а.", "н.д."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Гринвичы рӕстӕмбис рӕстӕг", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Астӕуккаг Европӕйаг сӕрдыгон рӕстӕг", "MEZ": "Астӕуккаг Европӕйаг стандартон рӕстӕг", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Скӕсӕн Европӕйаг сӕрдыгон рӕстӕг", "OEZ": "Скӕсӕн Европӕйаг стандартон рӕстӕг", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Ныгъуылӕн Европӕйаг сӕрдыгон рӕстӕг", "WEZ": "Ныгъуылӕн Европӕйаг стандартон рӕстӕг", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_204cc4159c06830dc072d7ffc4471b59,
			fnFmtCurrency:         fnFmtCurrency_7b65395a77761388c242d97c0d4bfe5b,
			fnFmtDateFull:         fnFmtDateFull_a7dc1bab86baf3ec90633b4422fb9cba,
			fnFmtDateLong:         fnFmtDateLong_4874c47573ec41408036722eb0d2c6a6,
			fnFmtDateMedium:       fnFmtDateMedium_5a5202b511779ecb07983ce846c85742,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pa"] = func() locales.Translator {
		return &localen{

			locale:             "pa",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "р.", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ਜਨ", "ਫ਼ਰ", "ਮਾਰਚ", "ਅਪ੍ਰੈ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾ", "ਅਗ", "ਸਤੰ", "ਅਕਤੂ", "ਨਵੰ", "ਦਸੰ"},
			monthsNarrow:       []string{"", "ਜ", "ਫ਼", "ਮਾ", "ਅ", "ਮ", "ਜੂ", "ਜੁ", "ਅ", "ਸ", "ਅ", "ਨ", "ਦ"},
			monthsWide:         []string{"", "ਜਨਵਰੀ", "ਫ਼ਰਵਰੀ", "ਮਾਰਚ", "ਅਪ੍ਰੈਲ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾਈ", "ਅਗਸਤ", "ਸਤੰਬਰ", "ਅਕਤੂਬਰ", "ਨਵੰਬਰ", "ਦਸੰਬਰ"},
			daysAbbreviated:    []string{"ਐਤ", "ਸੋਮ", "ਮੰਗਲ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕਰ", "ਸ਼ਨਿੱਚਰ"},
			daysNarrow:         []string{"ਐ", "ਸੋ", "ਮੰ", "ਬੁੱ", "ਵੀ", "ਸ਼ੁੱ", "ਸ਼"},
			daysShort:          []string{"ਐਤ", "ਸੋਮ", "ਮੰਗ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕ", "ਸ਼ਨਿੱ"},
			daysWide:           []string{"ਐਤਵਾਰ", "ਸੋਮਵਾਰ", "ਮੰਗਲਵਾਰ", "ਬੁੱਧਵਾਰ", "ਵੀਰਵਾਰ", "ਸ਼ੁੱਕਰਵਾਰ", "ਸ਼ਨਿੱਚਰਵਾਰ"},
			periodsAbbreviated: []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			periodsNarrow:      []string{"ਸ.", "ਸ਼."},
			periodsWide:        []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			erasAbbreviated:    []string{"ਈ. ਪੂ.", "ਸੰਨ"},
			erasNarrow:         []string{"ਈ.ਪੂ.", "ਸੰਨ"},
			erasWide:           []string{"ਈਸਵੀ ਪੂਰਵ", "ਈਸਵੀ ਸੰਨ"},
			timezones:          map[string]string{"ACDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ACWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "ADT": "ਅਟਲਾਂਟਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "AKDT": "ਅਲਾਸਕਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AKST": "ਅਲਾਸਕਾ ਮਿਆਰੀ ਵੇਲਾ", "ARST": "ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "ART": "ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "AST": "ਅਟਲਾਂਟਿਕ ਮਿਆਰੀ ਵੇਲਾ", "AWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "BOT": "ਬੋਲੀਵੀਆ ਵੇਲਾ", "BT": "ਭੂਟਾਨ ਵੇਲਾ", "CAT": "ਕੇਂਦਰੀ ਅਫਰੀਕਾ ਵੇਲਾ", "CDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHADT": "ਚੈਥਮ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHAST": "ਚੈਥਮ ਮਿਆਰੀ ਵੇਲਾ", "CLST": "ਚਿਲੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "CLT": "ਚਿਲੀ ਮਿਆਰੀ ਵੇਲਾ", "COST": "ਕੋਲੰਬੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "COT": "ਕੋਲੰਬੀਆ ਮਿਆਰੀ ਵੇਲਾ", "CST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ChST": "ਚਾਮੋਰੋ ਮਿਆਰੀ ਵੇਲਾ", "EAT": "ਪੂਰਬੀ ਅਫਰੀਕਾ ਵੇਲਾ", "ECT": "ਇਕਵੇਡੋਰ ਵੇਲਾ", "EDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "EST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "GFT": "ਫ੍ਰੈਂਚ ਗੁਏਨਾ ਵੇਲਾ", "GMT": "ਗ੍ਰੀਨਵਿਚ ਮੀਨ ਵੇਲਾ", "GST": "ਖਾੜੀ ਮਿਆਰੀ ਵੇਲਾ", "GYT": "ਗੁਯਾਨਾ ਵੇਲਾ", "HADT": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HAST": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਮਿਆਰੀ ਵੇਲਾ", "HAT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HECU": "ਕਿਊਬਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HENOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HEPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HKST": "ਹਾਂਗ ਕਾਂਗ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HKT": "ਹਾਂਗ ਕਾਂਗ ਮਿਆਰੀ ਵੇਲਾ", "HNCU": "ਕਿਊਬਾ ਮਿਆਰੀ ਵੇਲਾ", "HNEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNNOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਮਿਆਰੀ ਵੇਲਾ", "HNOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਮਿਆਰੀ ਵੇਲਾ", "HNPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "HNT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "IST": "ਭਾਰਤੀ ਮਿਆਰੀ ਵੇਲਾ", "JDT": "ਜਪਾਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "JST": "ਜਪਾਨ ਮਿਆਰੀ ਵੇਲਾ", "LHDT": "ਲੌਰਡ ਹੋਵੇ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "LHST": "ਲੌਰਡ ਹੋਵੇ ਮਿਆਰੀ ਵੇਲਾ", "MDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "MESZ": "ਮੱਧ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "MEZ": "ਮੱਧ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "MST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਮਿਆਰੀ ਵੇਲਾ", "MYT": "ਮਲੇਸ਼ੀਆ ਵੇਲਾ", "NZDT": "ਨਿਊਜ਼ੀਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "NZST": "ਨਿਊਜ਼ੀਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "OESZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "OEZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "PDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "PST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "SAST": "ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "SGT": "ਸਿੰਗਾਪੁਰ ਮਿਆਰੀ ਵੇਲਾ", "SRT": "ਸੂਰੀਨਾਮ ਵੇਲਾ", "TMST": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "TMT": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਮਿਆਰੀ ਵੇਲਾ", "UYST": "ਉਰੂਗਵੇ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "UYT": "ਉਰੂਗਵੇ ਮਿਆਰੀ ਵੇਲਾ", "VET": "ਵੈਨੇਜ਼ੂਏਲਾ ਵੇਲਾ", "WARST": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WART": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "WAST": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WAT": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "WESZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WEZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "WIB": "ਪੱਛਮੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WIT": "ਪੂਰਬੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WITA": "ਮੱਧ ਇੰਡੋਨੇਸ਼ੀਆਈ ਵੇਲਾ", "∅∅∅": "ਬ੍ਰਾਜ਼ੀਲੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pa_arab"] = func() locales.Translator {
		return &localen{

			locale:                 "pa_Arab",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "ر", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "ਜਨ", "ਫ਼ਰ", "ਮਾਰਚ", "ਅਪ੍ਰੈ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾ", "ਅਗ", "ਸਤੰ", "ਅਕਤੂ", "ਨਵੰ", "ਦਸੰ"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئ", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"ਐਤ", "ਸੋਮ", "ਮੰਗਲ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕਰ", "ਸ਼ਨਿੱਚਰ"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"ਐਤ", "ਸੋਮ", "ਮੰਗ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕ", "ਸ਼ਨਿੱ"},
			daysWide:               []string{"اتوار", "پیر", "منگل", "بُدھ", "جمعرات", "جمعہ", "ہفتہ"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"ايساپورو", "سں"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ايساپورو", "سں"},
			timezones:              map[string]string{"ACDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ACWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "ADT": "ਅਟਲਾਂਟਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "AKDT": "ਅਲਾਸਕਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AKST": "ਅਲਾਸਕਾ ਮਿਆਰੀ ਵੇਲਾ", "ARST": "ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "ART": "ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "AST": "ਅਟਲਾਂਟਿਕ ਮਿਆਰੀ ਵੇਲਾ", "AWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "BOT": "ਬੋਲੀਵੀਆ ਵੇਲਾ", "BT": "ਭੂਟਾਨ ਵੇਲਾ", "CAT": "ਕੇਂਦਰੀ ਅਫਰੀਕਾ ਵੇਲਾ", "CDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHADT": "ਚੈਥਮ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHAST": "ਚੈਥਮ ਮਿਆਰੀ ਵੇਲਾ", "CLST": "ਚਿਲੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "CLT": "ਚਿਲੀ ਮਿਆਰੀ ਵੇਲਾ", "COST": "ਕੋਲੰਬੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "COT": "ਕੋਲੰਬੀਆ ਮਿਆਰੀ ਵੇਲਾ", "CST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ChST": "ਚਾਮੋਰੋ ਮਿਆਰੀ ਵੇਲਾ", "EAT": "ਪੂਰਬੀ ਅਫਰੀਕਾ ਵੇਲਾ", "ECT": "ਇਕਵੇਡੋਰ ਵੇਲਾ", "EDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "EST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "GFT": "ਫ੍ਰੈਂਚ ਗੁਏਨਾ ਵੇਲਾ", "GMT": "ਗ੍ਰੀਨਵਿਚ ਮੀਨ ਵੇਲਾ", "GST": "ਖਾੜੀ ਮਿਆਰੀ ਵੇਲਾ", "GYT": "ਗੁਯਾਨਾ ਵੇਲਾ", "HADT": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HAST": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਮਿਆਰੀ ਵੇਲਾ", "HAT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HECU": "ਕਿਊਬਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HENOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HEPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HKST": "ਹਾਂਗ ਕਾਂਗ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HKT": "ਹਾਂਗ ਕਾਂਗ ਮਿਆਰੀ ਵੇਲਾ", "HNCU": "ਕਿਊਬਾ ਮਿਆਰੀ ਵੇਲਾ", "HNEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNNOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਮਿਆਰੀ ਵੇਲਾ", "HNOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਮਿਆਰੀ ਵੇਲਾ", "HNPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "HNT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "IST": "ਭਾਰਤੀ ਮਿਆਰੀ ਵੇਲਾ", "JDT": "ਜਪਾਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "JST": "ਜਪਾਨ ਮਿਆਰੀ ਵੇਲਾ", "LHDT": "ਲੌਰਡ ਹੋਵੇ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "LHST": "ਲੌਰਡ ਹੋਵੇ ਮਿਆਰੀ ਵੇਲਾ", "MDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "MESZ": "ਮੱਧ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "MEZ": "ਮੱਧ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "MST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਮਿਆਰੀ ਵੇਲਾ", "MYT": "ਮਲੇਸ਼ੀਆ ਵੇਲਾ", "NZDT": "ਨਿਊਜ਼ੀਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "NZST": "ਨਿਊਜ਼ੀਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "OESZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "OEZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "PDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "PST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "SAST": "ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "SGT": "ਸਿੰਗਾਪੁਰ ਮਿਆਰੀ ਵੇਲਾ", "SRT": "ਸੂਰੀਨਾਮ ਵੇਲਾ", "TMST": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "TMT": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਮਿਆਰੀ ਵੇਲਾ", "UYST": "ਉਰੂਗਵੇ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "UYT": "ਉਰੂਗਵੇ ਮਿਆਰੀ ਵੇਲਾ", "VET": "ਵੈਨੇਜ਼ੂਏਲਾ ਵੇਲਾ", "WARST": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WART": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "WAST": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WAT": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "WESZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WEZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "WIB": "ਪੱਛਮੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WIT": "ਪੂਰਬੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WITA": "ਮੱਧ ਇੰਡੋਨੇਸ਼ੀਆਈ ਵੇਲਾ", "∅∅∅": "ਅਜੋਰੇਸ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d0ef2c4a76852adc59df517a99a6000f,
			fnFmtCurrency:         fnFmtCurrency_9b06328888547fc535e862da8ee049d8,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pa_arab_pk"] = func() locales.Translator {
		return &localen{

			locale:             "pa_Arab_PK",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ਜਨ", "ਫ਼ਰ", "ਮਾਰਚ", "ਅਪ੍ਰੈ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾ", "ਅਗ", "ਸਤੰ", "ਅਕਤੂ", "ਨਵੰ", "ਦਸੰ"},
			monthsNarrow:       []string{"", "ਜ", "ਫ਼", "ਮਾ", "ਅ", "ਮ", "ਜੂ", "ਜੁ", "ਅ", "ਸ", "ਅ", "ਨ", "ਦ"},
			monthsWide:         []string{"", "ਜਨਵਰੀ", "ਫ਼ਰਵਰੀ", "ਮਾਰਚ", "ਅਪ੍ਰੈਲ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾਈ", "ਅਗਸਤ", "ਸਤੰਬਰ", "ਅਕਤੂਬਰ", "ਨਵੰਬਰ", "ਦਸੰਬਰ"},
			daysAbbreviated:    []string{"ਐਤ", "ਸੋਮ", "ਮੰਗਲ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕਰ", "ਸ਼ਨਿੱਚਰ"},
			daysNarrow:         []string{"ਐ", "ਸੋ", "ਮੰ", "ਬੁੱ", "ਵੀ", "ਸ਼ੁੱ", "ਸ਼"},
			daysShort:          []string{"ਐਤ", "ਸੋਮ", "ਮੰਗ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕ", "ਸ਼ਨਿੱ"},
			daysWide:           []string{"ਐਤਵਾਰ", "ਸੋਮਵਾਰ", "ਮੰਗਲਵਾਰ", "ਬੁੱਧਵਾਰ", "ਵੀਰਵਾਰ", "ਸ਼ੁੱਕਰਵਾਰ", "ਸ਼ਨਿੱਚਰਵਾਰ"},
			periodsAbbreviated: []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			periodsNarrow:      []string{"ਸ.", "ਸ਼."},
			periodsWide:        []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			erasAbbreviated:    []string{"ਈ. ਪੂ.", "ਸੰਨ"},
			erasNarrow:         []string{"ਈ.ਪੂ.", "ਸੰਨ"},
			erasWide:           []string{"ਈਸਵੀ ਪੂਰਵ", "ਈਸਵੀ ਸੰਨ"},
			timezones:          map[string]string{"ACDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ACWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "ADT": "ਅਟਲਾਂਟਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "AKDT": "ਅਲਾਸਕਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AKST": "ਅਲਾਸਕਾ ਮਿਆਰੀ ਵੇਲਾ", "ARST": "ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "ART": "ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "AST": "ਅਟਲਾਂਟਿਕ ਮਿਆਰੀ ਵੇਲਾ", "AWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "BOT": "ਬੋਲੀਵੀਆ ਵੇਲਾ", "BT": "ਭੂਟਾਨ ਵੇਲਾ", "CAT": "ਕੇਂਦਰੀ ਅਫਰੀਕਾ ਵੇਲਾ", "CDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHADT": "ਚੈਥਮ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHAST": "ਚੈਥਮ ਮਿਆਰੀ ਵੇਲਾ", "CLST": "ਚਿਲੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "CLT": "ਚਿਲੀ ਮਿਆਰੀ ਵੇਲਾ", "COST": "ਕੋਲੰਬੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "COT": "ਕੋਲੰਬੀਆ ਮਿਆਰੀ ਵੇਲਾ", "CST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ChST": "ਚਾਮੋਰੋ ਮਿਆਰੀ ਵੇਲਾ", "EAT": "ਪੂਰਬੀ ਅਫਰੀਕਾ ਵੇਲਾ", "ECT": "ਇਕਵੇਡੋਰ ਵੇਲਾ", "EDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "EST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "GFT": "ਫ੍ਰੈਂਚ ਗੁਏਨਾ ਵੇਲਾ", "GMT": "ਗ੍ਰੀਨਵਿਚ ਮੀਨ ਵੇਲਾ", "GST": "ਖਾੜੀ ਮਿਆਰੀ ਵੇਲਾ", "GYT": "ਗੁਯਾਨਾ ਵੇਲਾ", "HADT": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HAST": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਮਿਆਰੀ ਵੇਲਾ", "HAT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HECU": "ਕਿਊਬਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HENOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HEPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HKST": "ਹਾਂਗ ਕਾਂਗ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HKT": "ਹਾਂਗ ਕਾਂਗ ਮਿਆਰੀ ਵੇਲਾ", "HNCU": "ਕਿਊਬਾ ਮਿਆਰੀ ਵੇਲਾ", "HNEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNNOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਮਿਆਰੀ ਵੇਲਾ", "HNOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਮਿਆਰੀ ਵੇਲਾ", "HNPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "HNT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "IST": "ਭਾਰਤੀ ਮਿਆਰੀ ਵੇਲਾ", "JDT": "ਜਪਾਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "JST": "ਜਪਾਨ ਮਿਆਰੀ ਵੇਲਾ", "LHDT": "ਲੌਰਡ ਹੋਵੇ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "LHST": "ਲੌਰਡ ਹੋਵੇ ਮਿਆਰੀ ਵੇਲਾ", "MDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "MESZ": "ਮੱਧ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "MEZ": "ਮੱਧ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "MST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਮਿਆਰੀ ਵੇਲਾ", "MYT": "ਮਲੇਸ਼ੀਆ ਵੇਲਾ", "NZDT": "ਨਿਊਜ਼ੀਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "NZST": "ਨਿਊਜ਼ੀਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "OESZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "OEZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "PDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "PST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "SAST": "ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "SGT": "ਸਿੰਗਾਪੁਰ ਮਿਆਰੀ ਵੇਲਾ", "SRT": "ਸੂਰੀਨਾਮ ਵੇਲਾ", "TMST": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "TMT": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਮਿਆਰੀ ਵੇਲਾ", "UYST": "ਉਰੂਗਵੇ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "UYT": "ਉਰੂਗਵੇ ਮਿਆਰੀ ਵੇਲਾ", "VET": "ਵੈਨੇਜ਼ੂਏਲਾ ਵੇਲਾ", "WARST": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WART": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "WAST": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WAT": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "WESZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WEZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "WIB": "ਪੱਛਮੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WIT": "ਪੂਰਬੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WITA": "ਮੱਧ ਇੰਡੋਨੇਸ਼ੀਆਈ ਵੇਲਾ", "∅∅∅": "ਅਜੋਰੇਸ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pa_guru"] = func() locales.Translator {
		return &localen{

			locale:             "pa_Guru",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ਜਨ", "ਫ਼ਰ", "ਮਾਰਚ", "ਅਪ੍ਰੈ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾ", "ਅਗ", "ਸਤੰ", "ਅਕਤੂ", "ਨਵੰ", "ਦਸੰ"},
			monthsNarrow:       []string{"", "ਜ", "ਫ਼", "ਮਾ", "ਅ", "ਮ", "ਜੂ", "ਜੁ", "ਅ", "ਸ", "ਅ", "ਨ", "ਦ"},
			monthsWide:         []string{"", "ਜਨਵਰੀ", "ਫ਼ਰਵਰੀ", "ਮਾਰਚ", "ਅਪ੍ਰੈਲ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾਈ", "ਅਗਸਤ", "ਸਤੰਬਰ", "ਅਕਤੂਬਰ", "ਨਵੰਬਰ", "ਦਸੰਬਰ"},
			daysAbbreviated:    []string{"ਐਤ", "ਸੋਮ", "ਮੰਗਲ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕਰ", "ਸ਼ਨਿੱਚਰ"},
			daysNarrow:         []string{"ਐ", "ਸੋ", "ਮੰ", "ਬੁੱ", "ਵੀ", "ਸ਼ੁੱ", "ਸ਼"},
			daysShort:          []string{"ਐਤ", "ਸੋਮ", "ਮੰਗ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕ", "ਸ਼ਨਿੱ"},
			daysWide:           []string{"ਐਤਵਾਰ", "ਸੋਮਵਾਰ", "ਮੰਗਲਵਾਰ", "ਬੁੱਧਵਾਰ", "ਵੀਰਵਾਰ", "ਸ਼ੁੱਕਰਵਾਰ", "ਸ਼ਨਿੱਚਰਵਾਰ"},
			periodsAbbreviated: []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			periodsNarrow:      []string{"ਸ.", "ਸ਼."},
			periodsWide:        []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			erasAbbreviated:    []string{"ਈ. ਪੂ.", "ਸੰਨ"},
			erasNarrow:         []string{"ਈ.ਪੂ.", "ਸੰਨ"},
			erasWide:           []string{"ਈਸਵੀ ਪੂਰਵ", "ਈਸਵੀ ਸੰਨ"},
			timezones:          map[string]string{"ACDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ACWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "ADT": "ਅਟਲਾਂਟਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "AKDT": "ਅਲਾਸਕਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AKST": "ਅਲਾਸਕਾ ਮਿਆਰੀ ਵੇਲਾ", "ARST": "ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "ART": "ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "AST": "ਅਟਲਾਂਟਿਕ ਮਿਆਰੀ ਵੇਲਾ", "AWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "BOT": "ਬੋਲੀਵੀਆ ਵੇਲਾ", "BT": "ਭੂਟਾਨ ਵੇਲਾ", "CAT": "ਕੇਂਦਰੀ ਅਫਰੀਕਾ ਵੇਲਾ", "CDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHADT": "ਚੈਥਮ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHAST": "ਚੈਥਮ ਮਿਆਰੀ ਵੇਲਾ", "CLST": "ਚਿਲੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "CLT": "ਚਿਲੀ ਮਿਆਰੀ ਵੇਲਾ", "COST": "ਕੋਲੰਬੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "COT": "ਕੋਲੰਬੀਆ ਮਿਆਰੀ ਵੇਲਾ", "CST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ChST": "ਚਾਮੋਰੋ ਮਿਆਰੀ ਵੇਲਾ", "EAT": "ਪੂਰਬੀ ਅਫਰੀਕਾ ਵੇਲਾ", "ECT": "ਇਕਵੇਡੋਰ ਵੇਲਾ", "EDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "EST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "GFT": "ਫ੍ਰੈਂਚ ਗੁਏਨਾ ਵੇਲਾ", "GMT": "ਗ੍ਰੀਨਵਿਚ ਮੀਨ ਵੇਲਾ", "GST": "ਖਾੜੀ ਮਿਆਰੀ ਵੇਲਾ", "GYT": "ਗੁਯਾਨਾ ਵੇਲਾ", "HADT": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HAST": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਮਿਆਰੀ ਵੇਲਾ", "HAT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HECU": "ਕਿਊਬਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HENOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HEPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HKST": "ਹਾਂਗ ਕਾਂਗ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HKT": "ਹਾਂਗ ਕਾਂਗ ਮਿਆਰੀ ਵੇਲਾ", "HNCU": "ਕਿਊਬਾ ਮਿਆਰੀ ਵੇਲਾ", "HNEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNNOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਮਿਆਰੀ ਵੇਲਾ", "HNOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਮਿਆਰੀ ਵੇਲਾ", "HNPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "HNT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "IST": "ਭਾਰਤੀ ਮਿਆਰੀ ਵੇਲਾ", "JDT": "ਜਪਾਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "JST": "ਜਪਾਨ ਮਿਆਰੀ ਵੇਲਾ", "LHDT": "ਲੌਰਡ ਹੋਵੇ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "LHST": "ਲੌਰਡ ਹੋਵੇ ਮਿਆਰੀ ਵੇਲਾ", "MDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "MESZ": "ਮੱਧ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "MEZ": "ਮੱਧ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "MST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਮਿਆਰੀ ਵੇਲਾ", "MYT": "ਮਲੇਸ਼ੀਆ ਵੇਲਾ", "NZDT": "ਨਿਊਜ਼ੀਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "NZST": "ਨਿਊਜ਼ੀਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "OESZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "OEZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "PDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "PST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "SAST": "ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "SGT": "ਸਿੰਗਾਪੁਰ ਮਿਆਰੀ ਵੇਲਾ", "SRT": "ਸੂਰੀਨਾਮ ਵੇਲਾ", "TMST": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "TMT": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਮਿਆਰੀ ਵੇਲਾ", "UYST": "ਉਰੂਗਵੇ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "UYT": "ਉਰੂਗਵੇ ਮਿਆਰੀ ਵੇਲਾ", "VET": "ਵੈਨੇਜ਼ੂਏਲਾ ਵੇਲਾ", "WARST": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WART": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "WAST": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WAT": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "WESZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WEZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "WIB": "ਪੱਛਮੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WIT": "ਪੂਰਬੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WITA": "ਮੱਧ ਇੰਡੋਨੇਸ਼ੀਆਈ ਵੇਲਾ", "∅∅∅": "ਬ੍ਰਾਜ਼ੀਲੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pa_guru_in"] = func() locales.Translator {
		return &localen{

			locale:             "pa_Guru_IN",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{2, 6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ਜਨ", "ਫ਼ਰ", "ਮਾਰਚ", "ਅਪ੍ਰੈ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾ", "ਅਗ", "ਸਤੰ", "ਅਕਤੂ", "ਨਵੰ", "ਦਸੰ"},
			monthsNarrow:       []string{"", "ਜ", "ਫ਼", "ਮਾ", "ਅ", "ਮ", "ਜੂ", "ਜੁ", "ਅ", "ਸ", "ਅ", "ਨ", "ਦ"},
			monthsWide:         []string{"", "ਜਨਵਰੀ", "ਫ਼ਰਵਰੀ", "ਮਾਰਚ", "ਅਪ੍ਰੈਲ", "ਮਈ", "ਜੂਨ", "ਜੁਲਾਈ", "ਅਗਸਤ", "ਸਤੰਬਰ", "ਅਕਤੂਬਰ", "ਨਵੰਬਰ", "ਦਸੰਬਰ"},
			daysAbbreviated:    []string{"ਐਤ", "ਸੋਮ", "ਮੰਗਲ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕਰ", "ਸ਼ਨਿੱਚਰ"},
			daysNarrow:         []string{"ਐ", "ਸੋ", "ਮੰ", "ਬੁੱ", "ਵੀ", "ਸ਼ੁੱ", "ਸ਼"},
			daysShort:          []string{"ਐਤ", "ਸੋਮ", "ਮੰਗ", "ਬੁੱਧ", "ਵੀਰ", "ਸ਼ੁੱਕ", "ਸ਼ਨਿੱ"},
			daysWide:           []string{"ਐਤਵਾਰ", "ਸੋਮਵਾਰ", "ਮੰਗਲਵਾਰ", "ਬੁੱਧਵਾਰ", "ਵੀਰਵਾਰ", "ਸ਼ੁੱਕਰਵਾਰ", "ਸ਼ਨਿੱਚਰਵਾਰ"},
			periodsAbbreviated: []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			periodsNarrow:      []string{"ਸ.", "ਸ਼."},
			periodsWide:        []string{"ਪੂ.ਦੁ.", "ਬਾ.ਦੁ."},
			erasAbbreviated:    []string{"ਈ. ਪੂ.", "ਸੰਨ"},
			erasNarrow:         []string{"ਈ.ਪੂ.", "ਸੰਨ"},
			erasWide:           []string{"ਈਸਵੀ ਪੂਰਵ", "ਈਸਵੀ ਸੰਨ"},
			timezones:          map[string]string{"ACDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ACWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "ACWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਕੇਂਦਰੀ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "ADT": "ਅਟਲਾਂਟਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AEST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "AKDT": "ਅਲਾਸਕਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AKST": "ਅਲਾਸਕਾ ਮਿਆਰੀ ਵੇਲਾ", "ARST": "ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "ART": "ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "AST": "ਅਟਲਾਂਟਿਕ ਮਿਆਰੀ ਵੇਲਾ", "AWDT": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "AWST": "ਆਸਟ੍ਰੇਲੀਆਈ ਪੱਛਮੀ ਮਿਆਰੀ ਵੇਲਾ", "BOT": "ਬੋਲੀਵੀਆ ਵੇਲਾ", "BT": "ਭੂਟਾਨ ਵੇਲਾ", "CAT": "ਕੇਂਦਰੀ ਅਫਰੀਕਾ ਵੇਲਾ", "CDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHADT": "ਚੈਥਮ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "CHAST": "ਚੈਥਮ ਮਿਆਰੀ ਵੇਲਾ", "CLST": "ਚਿਲੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "CLT": "ਚਿਲੀ ਮਿਆਰੀ ਵੇਲਾ", "COST": "ਕੋਲੰਬੀਆ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "COT": "ਕੋਲੰਬੀਆ ਮਿਆਰੀ ਵੇਲਾ", "CST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਕੇਂਦਰੀ ਮਿਆਰੀ ਵੇਲਾ", "ChST": "ਚਾਮੋਰੋ ਮਿਆਰੀ ਵੇਲਾ", "EAT": "ਪੂਰਬੀ ਅਫਰੀਕਾ ਵੇਲਾ", "ECT": "ਇਕਵੇਡੋਰ ਵੇਲਾ", "EDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "EST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੂਰਬੀ ਮਿਆਰੀ ਵੇਲਾ", "GFT": "ਫ੍ਰੈਂਚ ਗੁਏਨਾ ਵੇਲਾ", "GMT": "ਗ੍ਰੀਨਵਿਚ ਮੀਨ ਵੇਲਾ", "GST": "ਖਾੜੀ ਮਿਆਰੀ ਵੇਲਾ", "GYT": "ਗੁਯਾਨਾ ਵੇਲਾ", "HADT": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HAST": "ਹਵਾਈ-ਅਲੇਯੂਸ਼ਿਅਨ ਮਿਆਰੀ ਵੇਲਾ", "HAT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HECU": "ਕਿਊਬਾ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HENOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HEPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HEPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "HKST": "ਹਾਂਗ ਕਾਂਗ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "HKT": "ਹਾਂਗ ਕਾਂਗ ਮਿਆਰੀ ਵੇਲਾ", "HNCU": "ਕਿਊਬਾ ਮਿਆਰੀ ਵੇਲਾ", "HNEG": "ਪੂਰਬੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNNOMX": "ਉੱਤਰ ਪੱਛਮੀ ਮੈਕਸੀਕੋ ਮਿਆਰੀ ਵੇਲਾ", "HNOG": "ਪੱਛਮੀ ਗ੍ਰੀਨਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "HNPM": "ਸੈਂਟ ਪੀਅਰੇ ਅਤੇ ਮਿਕੇਲਨ ਮਿਆਰੀ ਵੇਲਾ", "HNPMX": "ਮੈਕਸੀਕਨ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "HNT": "ਨਿਊਫਾਉਂਡਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "IST": "ਭਾਰਤੀ ਮਿਆਰੀ ਵੇਲਾ", "JDT": "ਜਪਾਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "JST": "ਜਪਾਨ ਮਿਆਰੀ ਵੇਲਾ", "LHDT": "ਲੌਰਡ ਹੋਵੇ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "LHST": "ਲੌਰਡ ਹੋਵੇ ਮਿਆਰੀ ਵੇਲਾ", "MDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "MESZ": "ਮੱਧ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "MEZ": "ਮੱਧ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "MST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਮਾਉਂਟੇਨ ਮਿਆਰੀ ਵੇਲਾ", "MYT": "ਮਲੇਸ਼ੀਆ ਵੇਲਾ", "NZDT": "ਨਿਊਜ਼ੀਲੈਂਡ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "NZST": "ਨਿਊਜ਼ੀਲੈਂਡ ਮਿਆਰੀ ਵੇਲਾ", "OESZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "OEZ": "ਪੂਰਬੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "PDT": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਪ੍ਰਕਾਸ਼ ਵੇਲਾ", "PST": "ਉੱਤਰੀ ਅਮਰੀਕੀ ਪੈਸਿਫਿਕ ਮਿਆਰੀ ਵੇਲਾ", "SAST": "ਦੱਖਣੀ ਅਫ਼ਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "SGT": "ਸਿੰਗਾਪੁਰ ਮਿਆਰੀ ਵੇਲਾ", "SRT": "ਸੂਰੀਨਾਮ ਵੇਲਾ", "TMST": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "TMT": "ਤੁਰਕਮੇਨਿਸਤਾਨ ਮਿਆਰੀ ਵੇਲਾ", "UYST": "ਉਰੂਗਵੇ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "UYT": "ਉਰੂਗਵੇ ਮਿਆਰੀ ਵੇਲਾ", "VET": "ਵੈਨੇਜ਼ੂਏਲਾ ਵੇਲਾ", "WARST": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WART": "ਪੱਛਮੀ ਅਰਜਨਟੀਨਾ ਮਿਆਰੀ ਵੇਲਾ", "WAST": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WAT": "ਪੱਛਮੀ ਅਫਰੀਕਾ ਮਿਆਰੀ ਵੇਲਾ", "WESZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ", "WEZ": "ਪੱਛਮੀ ਯੂਰਪੀ ਮਿਆਰੀ ਵੇਲਾ", "WIB": "ਪੱਛਮੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WIT": "ਪੂਰਬੀ ਇੰਡੋਨੇਸ਼ੀਆ ਵੇਲਾ", "WITA": "ਮੱਧ ਇੰਡੋਨੇਸ਼ੀਆਈ ਵੇਲਾ", "∅∅∅": "ਅਮੇਜ਼ਨ ਗਰਮੀਆਂ ਦਾ ਵੇਲਾ"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_802721cab60b1e4dbb9ad232ecdbf1f8,
			fnFmtCurrency:         fnFmtCurrency_aef7b1a5f0fa6e5e0b90141eafbbc2bc,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6ac6faa53655f798bb439bf2ef50aa19,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pl"] = func() locales.Translator {
		return &localen{

			locale:                 "pl",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "zł", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "sty", "lut", "mar", "kwi", "maj", "cze", "lip", "sie", "wrz", "paź", "lis", "gru"},
			monthsNarrow:           []string{"", "s", "l", "m", "k", "m", "c", "l", "s", "w", "p", "l", "g"},
			monthsWide:             []string{"", "stycznia", "lutego", "marca", "kwietnia", "maja", "czerwca", "lipca", "sierpnia", "września", "października", "listopada", "grudnia"},
			daysAbbreviated:        []string{"niedz.", "pon.", "wt.", "śr.", "czw.", "pt.", "sob."},
			daysNarrow:             []string{"n", "p", "w", "ś", "c", "p", "s"},
			daysShort:              []string{"nie", "pon", "wto", "śro", "czw", "pią", "sob"},
			daysWide:               []string{"niedziela", "poniedziałek", "wtorek", "środa", "czwartek", "piątek", "sobota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"p.n.e.", "n.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"przed naszą erą", "naszej ery"},
			timezones:              map[string]string{"ACDT": "czas środkowoaustralijski letni", "ACST": "czas środkowoaustralijski standardowy", "ACWDT": "czas środkowo-zachodnioaustralijski letni", "ACWST": "czas środkowo-zachodnioaustralijski standardowy", "ADT": "czas atlantycki letni", "AEDT": "czas wschodnioaustralijski letni", "AEST": "czas wschodnioaustralijski standardowy", "AKDT": "Alaska (czas letni)", "AKST": "Alaska (czas standardowy)", "ARST": "Argentyna (czas letni)", "ART": "Argentyna (czas standardowy)", "AST": "czas atlantycki standardowy", "AWDT": "czas zachodnioaustralijski letni", "AWST": "czas zachodnioaustralijski standardowy", "BOT": "Boliwia", "BT": "Bhutan", "CAT": "czas środkowoafrykański", "CDT": "czas środkowoamerykański letni", "CHADT": "Chatham (czas letni)", "CHAST": "Chatham (czas standardowy)", "CLST": "Chile (czas letni)", "CLT": "Chile (czas standardowy)", "COST": "Kolumbia (czas letni)", "COT": "Kolumbia (czas standardowy)", "CST": "czas środkowoamerykański standardowy", "ChST": "Czamorro", "EAT": "czas wschodnioafrykański", "ECT": "Ekwador", "EDT": "czas wschodnioamerykański letni", "EST": "czas wschodnioamerykański standardowy", "GFT": "Gujana Francuska", "GMT": "czas uniwersalny", "GST": "Zatoka Perska", "GYT": "Gujana", "HADT": "Hawaje-Aleuty (czas letni)", "HAST": "Hawaje-Aleuty (czas standardowy)", "HAT": "Nowa Fundlandia (czas letni)", "HECU": "Kuba (czas letni)", "HEEG": "Grenlandia Wschodnia (czas letni)", "HENOMX": "Meksyk Północno-Zachodni (czas letni)", "HEOG": "Grenlandia Zachodnia (czas letni)", "HEPM": "Saint-Pierre i Miquelon (czas letni)", "HEPMX": "Meksyk (czas pacyficzny letni)", "HKST": "Hongkong (czas letni)", "HKT": "Hongkong (czas standardowy)", "HNCU": "Kuba (czas standardowy)", "HNEG": "Grenlandia Wschodnia (czas standardowy)", "HNNOMX": "Meksyk Północno-Zachodni (czas standardowy)", "HNOG": "Grenlandia Zachodnia (czas standardowy)", "HNPM": "Saint-Pierre i Miquelon (czas standardowy)", "HNPMX": "Meksyk (czas pacyficzny standardowy)", "HNT": "Nowa Fundlandia (czas standardowy)", "IST": "czas indyjski standardowy", "JDT": "Japonia (czas letni)", "JST": "Japonia (czas standardowy)", "LHDT": "Lord Howe (czas letni)", "LHST": "Lord Howe (czas standardowy)", "MDT": "czas górski letni", "MESZ": "czas środkowoeuropejski letni", "MEZ": "czas środkowoeuropejski standardowy", "MST": "czas górski standardowy", "MYT": "Malezja", "NZDT": "Nowa Zelandia (czas letni)", "NZST": "Nowa Zelandia (czas standardowy)", "OESZ": "czas wschodnioeuropejski letni", "OEZ": "czas wschodnioeuropejski standardowy", "PDT": "czas pacyficzny letni", "PST": "czas pacyficzny standardowy", "SAST": "czas południowoafrykański", "SGT": "Singapur", "SRT": "Surinam", "TMST": "Turkmenistan (czas letni)", "TMT": "Turkmenistan (czas standardowy)", "UYST": "Urugwaj (czas letni)", "UYT": "Urugwaj (czas standardowy)", "VET": "Wenezuela", "WARST": "Argentyna Zachodnia (czas letni)", "WART": "Argentyna Zachodnia (czas standardowy)", "WAST": "czas zachodnioafrykański letni", "WAT": "czas zachodnioafrykański standardowy", "WESZ": "czas zachodnioeuropejski letni", "WEZ": "czas zachodnioeuropejski standardowy", "WIB": "Indonezja Zachodnia", "WIT": "Indonezja Wschodnia", "WITA": "Indonezja Środkowa", "∅∅∅": "Azory (czas letni)"},

			fnCardinalPluralRule:  fnCardinalPluralRule_4e031e59a33d2df374391ed7199b0c7f,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pl_pl"] = func() locales.Translator {
		return &localen{

			locale:                 "pl_PL",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "sty", "lut", "mar", "kwi", "maj", "cze", "lip", "sie", "wrz", "paź", "lis", "gru"},
			monthsNarrow:           []string{"", "s", "l", "m", "k", "m", "c", "l", "s", "w", "p", "l", "g"},
			monthsWide:             []string{"", "stycznia", "lutego", "marca", "kwietnia", "maja", "czerwca", "lipca", "sierpnia", "września", "października", "listopada", "grudnia"},
			daysAbbreviated:        []string{"niedz.", "pon.", "wt.", "śr.", "czw.", "pt.", "sob."},
			daysNarrow:             []string{"n", "p", "w", "ś", "c", "p", "s"},
			daysShort:              []string{"nie", "pon", "wto", "śro", "czw", "pią", "sob"},
			daysWide:               []string{"niedziela", "poniedziałek", "wtorek", "środa", "czwartek", "piątek", "sobota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"p.n.e.", "n.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"przed naszą erą", "naszej ery"},
			timezones:              map[string]string{"ACDT": "czas środkowoaustralijski letni", "ACST": "czas środkowoaustralijski standardowy", "ACWDT": "czas środkowo-zachodnioaustralijski letni", "ACWST": "czas środkowo-zachodnioaustralijski standardowy", "ADT": "czas atlantycki letni", "AEDT": "czas wschodnioaustralijski letni", "AEST": "czas wschodnioaustralijski standardowy", "AKDT": "Alaska (czas letni)", "AKST": "Alaska (czas standardowy)", "ARST": "Argentyna (czas letni)", "ART": "Argentyna (czas standardowy)", "AST": "czas atlantycki standardowy", "AWDT": "czas zachodnioaustralijski letni", "AWST": "czas zachodnioaustralijski standardowy", "BOT": "Boliwia", "BT": "Bhutan", "CAT": "czas środkowoafrykański", "CDT": "czas środkowoamerykański letni", "CHADT": "Chatham (czas letni)", "CHAST": "Chatham (czas standardowy)", "CLST": "Chile (czas letni)", "CLT": "Chile (czas standardowy)", "COST": "Kolumbia (czas letni)", "COT": "Kolumbia (czas standardowy)", "CST": "czas środkowoamerykański standardowy", "ChST": "Czamorro", "EAT": "czas wschodnioafrykański", "ECT": "Ekwador", "EDT": "czas wschodnioamerykański letni", "EST": "czas wschodnioamerykański standardowy", "GFT": "Gujana Francuska", "GMT": "czas uniwersalny", "GST": "Zatoka Perska", "GYT": "Gujana", "HADT": "Hawaje-Aleuty (czas letni)", "HAST": "Hawaje-Aleuty (czas standardowy)", "HAT": "Nowa Fundlandia (czas letni)", "HECU": "Kuba (czas letni)", "HEEG": "Grenlandia Wschodnia (czas letni)", "HENOMX": "Meksyk Północno-Zachodni (czas letni)", "HEOG": "Grenlandia Zachodnia (czas letni)", "HEPM": "Saint-Pierre i Miquelon (czas letni)", "HEPMX": "Meksyk (czas pacyficzny letni)", "HKST": "Hongkong (czas letni)", "HKT": "Hongkong (czas standardowy)", "HNCU": "Kuba (czas standardowy)", "HNEG": "Grenlandia Wschodnia (czas standardowy)", "HNNOMX": "Meksyk Północno-Zachodni (czas standardowy)", "HNOG": "Grenlandia Zachodnia (czas standardowy)", "HNPM": "Saint-Pierre i Miquelon (czas standardowy)", "HNPMX": "Meksyk (czas pacyficzny standardowy)", "HNT": "Nowa Fundlandia (czas standardowy)", "IST": "czas indyjski standardowy", "JDT": "Japonia (czas letni)", "JST": "Japonia (czas standardowy)", "LHDT": "Lord Howe (czas letni)", "LHST": "Lord Howe (czas standardowy)", "MDT": "czas górski letni", "MESZ": "czas środkowoeuropejski letni", "MEZ": "czas środkowoeuropejski standardowy", "MST": "czas górski standardowy", "MYT": "Malezja", "NZDT": "Nowa Zelandia (czas letni)", "NZST": "Nowa Zelandia (czas standardowy)", "OESZ": "czas wschodnioeuropejski letni", "OEZ": "czas wschodnioeuropejski standardowy", "PDT": "czas pacyficzny letni", "PST": "czas pacyficzny standardowy", "SAST": "czas południowoafrykański", "SGT": "Singapur", "SRT": "Surinam", "TMST": "Turkmenistan (czas letni)", "TMT": "Turkmenistan (czas standardowy)", "UYST": "Urugwaj (czas letni)", "UYT": "Urugwaj (czas standardowy)", "VET": "Wenezuela", "WARST": "Argentyna Zachodnia (czas letni)", "WART": "Argentyna Zachodnia (czas standardowy)", "WAST": "czas zachodnioafrykański letni", "WAT": "czas zachodnioafrykański standardowy", "WESZ": "czas zachodnioeuropejski letni", "WEZ": "czas zachodnioeuropejski standardowy", "WIB": "Indonezja Zachodnia", "WIT": "Indonezja Wschodnia", "WITA": "Indonezja Środkowa", "∅∅∅": "Brasília (czas letni)"},

			fnCardinalPluralRule:  fnCardinalPluralRule_3c2302f5244a65cb0b21db084f77f61b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["prg"] = func() locales.Translator {
		return &localen{

			locale:                 "prg",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "rag", "was", "pūl", "sak", "zal", "sīm", "līp", "dag", "sil", "spa", "lap", "sal"},
			monthsNarrow:           []string{"", "R", "W", "P", "S", "Z", "S", "L", "D", "S", "S", "L", "S"},
			monthsWide:             []string{"", "rags", "wassarins", "pūlis", "sakkis", "zallaws", "sīmenis", "līpa", "daggis", "sillins", "spallins", "lapkrūtis", "sallaws"},
			daysAbbreviated:        []string{"nad", "pan", "wis", "pus", "ket", "pēn", "sab"},
			daysNarrow:             []string{"N", "P", "W", "P", "K", "P", "S"},
			daysWide:               []string{"nadīli", "panadīli", "wisasīdis", "pussisawaiti", "ketwirtiks", "pēntniks", "sabattika"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"ankstāinan", "pa pussideinan"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Atlāntiska daggas kerdā", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Atlāntiska zēimas kerdā", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Centrālas Amērikas daggas kerdā", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Centrālas Amērikas zēimas kerdā", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Dēiniskas Amērikas daggas kerdā", "EST": "Dēiniskas Amērikas zēimas kerdā", "GFT": "GFT", "GMT": "Greenwich kerdā", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Amērikas gārban daggas kerdā", "MESZ": "Centrālas Eurōpas daggas kerdā", "MEZ": "Centrālas Eurōpas zēimas kerdā", "MST": "Amērikas gārban zēimas kerdā", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Dēiniskas Eurōpas daggas kerdā", "OEZ": "Dēiniskas Eurōpas zēimas kerdā", "PDT": "Pacīfiskas Amērikas daggas kerdā", "PST": "Pacīfiskas Amērikas zēimas kerdā", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Wakkariskas Eurōpas daggas kerdā", "WEZ": "Wakkariskas Eurōpas zēimas kerdā", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_35d374490f3e32fac8828ad5843b9d6e,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_8f18dd0a7a346f349e234c4eb438c7ba,
			fnFmtDateLong:         fnFmtDateLong_bce1be6f532fe054bea1dce83a0ecd58,
			fnFmtDateMedium:       fnFmtDateMedium_657e1e9a61de200098381fee1994c99c,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["prg_001"] = func() locales.Translator {
		return &localen{

			locale:                 "prg_001",
			pluralsCardinal:        []locales.PluralRule{1, 2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "rag", "was", "pūl", "sak", "zal", "sīm", "līp", "dag", "sil", "spa", "lap", "sal"},
			monthsNarrow:           []string{"", "R", "W", "P", "S", "Z", "S", "L", "D", "S", "S", "L", "S"},
			monthsWide:             []string{"", "rags", "wassarins", "pūlis", "sakkis", "zallaws", "sīmenis", "līpa", "daggis", "sillins", "spallins", "lapkrūtis", "sallaws"},
			daysAbbreviated:        []string{"nad", "pan", "wis", "pus", "ket", "pēn", "sab"},
			daysNarrow:             []string{"N", "P", "W", "P", "K", "P", "S"},
			daysWide:               []string{"nadīli", "panadīli", "wisasīdis", "pussisawaiti", "ketwirtiks", "pēntniks", "sabattika"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsWide:            []string{"ankstāinan", "pa pussideinan"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Atlāntiska daggas kerdā", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Atlāntiska zēimas kerdā", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Centrālas Amērikas daggas kerdā", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Centrālas Amērikas zēimas kerdā", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Dēiniskas Amērikas daggas kerdā", "EST": "Dēiniskas Amērikas zēimas kerdā", "GFT": "GFT", "GMT": "Greenwich kerdā", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Amērikas gārban daggas kerdā", "MESZ": "Centrālas Eurōpas daggas kerdā", "MEZ": "Centrālas Eurōpas zēimas kerdā", "MST": "Amērikas gārban zēimas kerdā", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Dēiniskas Eurōpas daggas kerdā", "OEZ": "Dēiniskas Eurōpas zēimas kerdā", "PDT": "Pacīfiskas Amērikas daggas kerdā", "PST": "Pacīfiskas Amērikas zēimas kerdā", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Wakkariskas Eurōpas daggas kerdā", "WEZ": "Wakkariskas Eurōpas zēimas kerdā", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_35d374490f3e32fac8828ad5843b9d6e,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_8f18dd0a7a346f349e234c4eb438c7ba,
			fnFmtDateLong:         fnFmtDateLong_bce1be6f532fe054bea1dce83a0ecd58,
			fnFmtDateMedium:       fnFmtDateMedium_657e1e9a61de200098381fee1994c99c,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ps"] = func() locales.Translator {
		return &localen{

			locale:                 "ps",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "؋", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "р.", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "C$", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سېپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			monthsNarrow:           []string{"", "ج", "ف", "م", "ا", "م", "ج", "ج", "ا", "س", "ا", "ن", "د"},
			monthsWide:             []string{"", "جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سېپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			daysWide:               []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			periodsAbbreviated:     []string{"غ.م.", "غ.و."},
			periodsNarrow:          []string{"غ.م.", "غ.و."},
			periodsWide:            []string{"غ.م.", "غ.و."},
			erasAbbreviated:        []string{"له میلاد وړاندې", "م."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"له میلاد څخه وړاندې", "له میلاد څخه وروسته"},
			timezones:              map[string]string{"ACDT": "آسترالوي مرکزي د ورځې روښانه وخت", "ACST": "آسترالوي مرکزي معياري وخت", "ACWDT": "آسترالوي مرکزي لوېديځ د ورځې روښانه وخت", "ACWST": "آسترالوي مرکزي لوېديځ معياري وخت", "ADT": "اتلانتیک د رڼا ورځې وخت", "AEDT": "آسترالوي ختيځ د ورځې روښانه وخت", "AEST": "آسترالوي ختيځ معياري وخت", "AKDT": "الاسکا د ورځې روښانه وخت", "AKST": "الاسکا معياري وخت", "ARST": "ارجنټاین اوړي وخت", "ART": "ارجنټاین معیاری وخت", "AST": "اتلانتیک معياري وخت", "AWDT": "آسترالوي لوېديځ د ورځې روښانه وخت", "AWST": "آسترالوي لوېديځ معياري وخت", "BOT": "بولیویا وخت", "BT": "بهوټان وخت", "CAT": "منځنی افريقا وخت", "CDT": "مرکزي د ورځې روښانه وخت", "CHADT": "چاتام د ورځې روښانه وخت", "CHAST": "چاتام معياري وخت", "CLST": "چلی اوړي وخت", "CLT": "چلی معیاری وخت", "COST": "کولمبیا اوړي وخت", "COT": "کولمبیا معیاری وخت", "CST": "مرکزي معياري وخت", "ChST": "چمارو معياري وخت", "EAT": "ختيځ افريقا وخت", "ECT": "د اکوادور وخت", "EDT": "ختيځ د رڼا ورځې وخت", "EST": "ختيځ معياري وخت", "GFT": "د فرانسوي ګانا وخت", "GMT": "ګرينويچ معياري وخت", "GST": "خلیج معياري وخت", "GYT": "د ګوانانا وخت", "HADT": "هوایی الیوتین رڼا ورځې وخت", "HAST": "هوایی الیوتین معیاری وخت", "HAT": "نيو فاونډلېنډ د ورځې روښانه وخت", "HECU": "کیوبا د رڼا ورځې وخت", "HEEG": "د ختیځ ګرینلینډ اوړي وخت", "HENOMX": "د شمال لویدیځ مکسیکو رڼا ورځې وخت", "HEOG": "لویدیځ ګرینلینډ اوړي وخت", "HEPM": "سینټ پییرا و ميکلين رڼا ورځې وخت", "HEPMX": "مکسیکن پیسفک رڼا ورځې وخت", "HKST": "هانګ کانګ اوړي وخت", "HKT": "هانګ کانګ معياري وخت", "HNCU": "کیوبا معياري وخت", "HNEG": "د ختیځ ګرینلینډ معياري وخت", "HNNOMX": "د شمال لویدیځ مکسیکو معیاري وخت", "HNOG": "لویدیځ ګرینلینډ معياري وخت", "HNPM": "سینټ پییرا و ميکلين معیاری وخت", "HNPMX": "مکسیکن پیسفک معیاری وخت", "HNT": "د نوي فیلډلینډ معیاری وخت", "IST": "هند معیاري وخت", "JDT": "جاپان د ورځې روښانه وخت", "JST": "جاپان معياري وخت", "LHDT": "لارډ هوي د ورځې روښانه وخت", "LHST": "لارډ هوي معياري وخت", "MDT": "د غره د رڼا ورځې وخت", "MESZ": "وسطي اروپايي د اوړي وخت", "MEZ": "د مرکزي اروپا معیاري وخت", "MST": "د غره معياري وخت", "MYT": "ملائیشیا وخت", "NZDT": "نيوزي لېنډ د ورځې روښانه وخت", "NZST": "نيوزي لېنډ معياري وخت", "OESZ": "ختيځ اروپايي اوړي وخت", "OEZ": "ختيځ اروپايي معياري وخت", "PDT": "پیسفک د رڼا ورځې وخت", "PST": "د پیسفک معياري وخت", "SAST": "جنوبي افريقا معياري وخت", "SGT": "سنګاپور معیاري وخت", "SRT": "سورینام وخت", "TMST": "ترکمنستان اوړي وخت", "TMT": "ترکمنستان معياري وخت", "UYST": "یوروګوای اوړي وخت", "UYT": "یوروګوای معياري وخت", "VET": "وینزویلا وخت", "WARST": "لوېديځ ارجنټاين اوړي وخت", "WART": "لوېديځ ارجنټاين معياري وخت", "WAST": "د افریقا افریقا لویدیځ وخت", "WAT": "لویدیځ افریقایي معیاري وخت", "WESZ": "لوېديځ اروپايي اوړي وخت", "WEZ": "لوېديځ اروپايي معياري وخت", "WIB": "لویدیځ اندونیزیا وخت", "WIT": "اندونیزیا وخت", "WITA": "مرکزي ادونيزيا وخت", "∅∅∅": "ايزورس اوړي وخت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_7ea895699a09e2a46b34e71ca5ff6068,
			fnFmtDateLong:         fnFmtDateLong_03b64668e0972b02e0a12fe5cd4b5829,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ps_af"] = func() locales.Translator {
		return &localen{

			locale:                 "ps_AF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سېپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			monthsNarrow:           []string{"", "ج", "ف", "م", "ا", "م", "ج", "ج", "ا", "س", "ا", "ن", "د"},
			monthsWide:             []string{"", "جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سېپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			daysWide:               []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			periodsAbbreviated:     []string{"غ.م.", "غ.و."},
			periodsNarrow:          []string{"غ.م.", "غ.و."},
			periodsWide:            []string{"غ.م.", "غ.و."},
			erasAbbreviated:        []string{"له میلاد وړاندې", "م."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"له میلاد څخه وړاندې", "له میلاد څخه وروسته"},
			timezones:              map[string]string{"ACDT": "آسترالوي مرکزي د ورځې روښانه وخت", "ACST": "آسترالوي مرکزي معياري وخت", "ACWDT": "آسترالوي مرکزي لوېديځ د ورځې روښانه وخت", "ACWST": "آسترالوي مرکزي لوېديځ معياري وخت", "ADT": "اتلانتیک د رڼا ورځې وخت", "AEDT": "آسترالوي ختيځ د ورځې روښانه وخت", "AEST": "آسترالوي ختيځ معياري وخت", "AKDT": "الاسکا د ورځې روښانه وخت", "AKST": "الاسکا معياري وخت", "ARST": "ارجنټاین اوړي وخت", "ART": "ارجنټاین معیاری وخت", "AST": "اتلانتیک معياري وخت", "AWDT": "آسترالوي لوېديځ د ورځې روښانه وخت", "AWST": "آسترالوي لوېديځ معياري وخت", "BOT": "بولیویا وخت", "BT": "بهوټان وخت", "CAT": "منځنی افريقا وخت", "CDT": "مرکزي د ورځې روښانه وخت", "CHADT": "چاتام د ورځې روښانه وخت", "CHAST": "چاتام معياري وخت", "CLST": "چلی اوړي وخت", "CLT": "چلی معیاری وخت", "COST": "کولمبیا اوړي وخت", "COT": "کولمبیا معیاری وخت", "CST": "مرکزي معياري وخت", "ChST": "چمارو معياري وخت", "EAT": "ختيځ افريقا وخت", "ECT": "د اکوادور وخت", "EDT": "ختيځ د رڼا ورځې وخت", "EST": "ختيځ معياري وخت", "GFT": "د فرانسوي ګانا وخت", "GMT": "ګرينويچ معياري وخت", "GST": "خلیج معياري وخت", "GYT": "د ګوانانا وخت", "HADT": "هوایی الیوتین رڼا ورځې وخت", "HAST": "هوایی الیوتین معیاری وخت", "HAT": "نيو فاونډلېنډ د ورځې روښانه وخت", "HECU": "کیوبا د رڼا ورځې وخت", "HEEG": "د ختیځ ګرینلینډ اوړي وخت", "HENOMX": "د شمال لویدیځ مکسیکو رڼا ورځې وخت", "HEOG": "لویدیځ ګرینلینډ اوړي وخت", "HEPM": "سینټ پییرا و ميکلين رڼا ورځې وخت", "HEPMX": "مکسیکن پیسفک رڼا ورځې وخت", "HKST": "هانګ کانګ اوړي وخت", "HKT": "هانګ کانګ معياري وخت", "HNCU": "کیوبا معياري وخت", "HNEG": "د ختیځ ګرینلینډ معياري وخت", "HNNOMX": "د شمال لویدیځ مکسیکو معیاري وخت", "HNOG": "لویدیځ ګرینلینډ معياري وخت", "HNPM": "سینټ پییرا و ميکلين معیاری وخت", "HNPMX": "مکسیکن پیسفک معیاری وخت", "HNT": "د نوي فیلډلینډ معیاری وخت", "IST": "هند معیاري وخت", "JDT": "جاپان د ورځې روښانه وخت", "JST": "جاپان معياري وخت", "LHDT": "لارډ هوي د ورځې روښانه وخت", "LHST": "لارډ هوي معياري وخت", "MDT": "د غره د رڼا ورځې وخت", "MESZ": "وسطي اروپايي د اوړي وخت", "MEZ": "د مرکزي اروپا معیاري وخت", "MST": "د غره معياري وخت", "MYT": "ملائیشیا وخت", "NZDT": "نيوزي لېنډ د ورځې روښانه وخت", "NZST": "نيوزي لېنډ معياري وخت", "OESZ": "ختيځ اروپايي اوړي وخت", "OEZ": "ختيځ اروپايي معياري وخت", "PDT": "پیسفک د رڼا ورځې وخت", "PST": "د پیسفک معياري وخت", "SAST": "جنوبي افريقا معياري وخت", "SGT": "سنګاپور معیاري وخت", "SRT": "سورینام وخت", "TMST": "ترکمنستان اوړي وخت", "TMT": "ترکمنستان معياري وخت", "UYST": "یوروګوای اوړي وخت", "UYT": "یوروګوای معياري وخت", "VET": "وینزویلا وخت", "WARST": "لوېديځ ارجنټاين اوړي وخت", "WART": "لوېديځ ارجنټاين معياري وخت", "WAST": "د افریقا افریقا لویدیځ وخت", "WAT": "لویدیځ افریقایي معیاري وخت", "WESZ": "لوېديځ اروپايي اوړي وخت", "WEZ": "لوېديځ اروپايي معياري وخت", "WIB": "لویدیځ اندونیزیا وخت", "WIT": "اندونیزیا وخت", "WITA": "مرکزي ادونيزيا وخت", "∅∅∅": "برسلیا اوړي وخت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_7ea895699a09e2a46b34e71ca5ff6068,
			fnFmtDateLong:         fnFmtDateLong_03b64668e0972b02e0a12fe5cd4b5829,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ps_pk"] = func() locales.Translator {
		return &localen{

			locale:                 "ps_PK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "Rs", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سېپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			monthsNarrow:           []string{"", "ج", "ف", "م", "ا", "م", "ج", "ج", "ا", "س", "ا", "ن", "د"},
			monthsWide:             []string{"", "جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سېپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			daysWide:               []string{"يونۍ", "دونۍ", "درېنۍ", "څلرنۍ", "پينځنۍ", "جمعه", "اونۍ"},
			periodsAbbreviated:     []string{"غ.م.", "غ.و."},
			periodsNarrow:          []string{"غ.م.", "غ.و."},
			periodsWide:            []string{"غ.م.", "غ.و."},
			erasAbbreviated:        []string{"له میلاد وړاندې", "م."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"له میلاد څخه وړاندې", "له میلاد څخه وروسته"},
			timezones:              map[string]string{"ACDT": "آسترالوي مرکزي د ورځې روښانه وخت", "ACST": "آسترالوي مرکزي معياري وخت", "ACWDT": "آسترالوي مرکزي لوېديځ د ورځې روښانه وخت", "ACWST": "آسترالوي مرکزي لوېديځ معياري وخت", "ADT": "اتلانتیک د رڼا ورځے وخت", "AEDT": "آسترالوي ختيځ د ورځې روښانه وخت", "AEST": "آسترالوي ختيځ معياري وخت", "AKDT": "د الاسکا د ورځے روښانه کول", "AKST": "", "ARST": "ارجنټاین اوړي وخت", "ART": "ارجنټاین معیاری وخت", "AST": "", "AWDT": "د اسټرالیا لویدیځ د ورځے وخت", "AWST": "", "BOT": "بولیویا وخت", "BT": "بهوټان وخت", "CAT": "منځنی افريقا وخت", "CDT": "مرکزي رڼا ورځے وخت", "CHADT": "چاتام د ورځې روښانه وخت", "CHAST": "چاتام معياري وخت", "CLST": "چلی اوړي وخت", "CLT": "چلی معیاری وخت", "COST": "کولمبیا اوړي وخت", "COT": "کولمبیا معیاری وخت", "CST": "", "ChST": "چمارو معياري وخت", "EAT": "ختيځ افريقا وخت", "ECT": "د اکوادور وخت", "EDT": "ختيځ د رڼا ورځے وخت", "EST": "", "GFT": "د فرانسوي ګانا وخت", "GMT": "ګرينويچ معياري وخت", "GST": "خلیج معياري وخت", "GYT": "د ګوانانا وخت", "HADT": "هوایی الیوتین رڼا ورځے وخت", "HAST": "", "HAT": "د نوي فیلډلینډ رڼا ورځے وخت", "HECU": "کیوبا د رڼا ورځے وخت", "HEEG": "د ختیځ ګرینلینډ اوړي وخت", "HENOMX": "د شمال لویدیځ مکسیکو رڼا ورځے وخت", "HEOG": "لویدیځ ګرینلینډ اوړي وخت", "HEPM": "سینټ پییرا و ميکلين رڼا ورځے وخت", "HEPMX": "مکسیکن پیسفک رڼا ورځے وخت", "HKST": "هانګ کانګ اوړي وخت", "HKT": "هانګ کانګ معياري وخت", "HNCU": "", "HNEG": "د ختیځ ګرینلینډ معياري وخت", "HNNOMX": "", "HNOG": "لویدیځ ګرینلینډ معياري وخت", "HNPM": "", "HNPMX": "", "HNT": "", "IST": "هند معیاري وخت", "JDT": "جاپان د رڼا ورځے وخت", "JST": "", "LHDT": "رب هاو د ورځے د رڼا وخت", "LHST": "", "MDT": "د غره د رڼا ورځے وخت", "MESZ": "وسطي اروپايي د اوړي وخت", "MEZ": "د مرکزي اروپا معیاري وخت", "MST": "", "MYT": "ملائیشیا وخت", "NZDT": "د نیوزی لینڈ د ورځے د رڼا وخت", "NZST": "", "OESZ": "ختيځ اروپايي اوړي وخت", "OEZ": "ختيځ اروپايي معياري وخت", "PDT": "پیسفک د رڼا ورځے وخت", "PST": "", "SAST": "جنوبي افريقا معياري وخت", "SGT": "سنګاپور معیاري وخت", "SRT": "سورینام وخت", "TMST": "ترکمنستان اوړي وخت", "TMT": "ترکمنستان معياري وخت", "UYST": "یوروګوای اوړي وخت", "UYT": "یوروګوای معياري وخت", "VET": "وینزویلا وخت", "WARST": "لوېديځ ارجنټاين اوړي وخت", "WART": "لوېديځ ارجنټاين معياري وخت", "WAST": "د افریقا افریقا لویدیځ وخت", "WAT": "لویدیځ افریقایي معیاري وخت", "WESZ": "د لودیځے اورپا د اوړي وخت", "WEZ": "د لودیځے اروپا معیاري وخت", "WIB": "لویدیځ اندونیزیا وخت", "WIT": "اندونیزیا وخت", "WITA": "مرکزي ادونيزيا وخت", "∅∅∅": "ايزورس اوړي وخت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_7ea895699a09e2a46b34e71ca5ff6068,
			fnFmtDateLong:         fnFmtDateLong_03b64668e0972b02e0a12fe5cd4b5829,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_52501d1a8eef906b405639f3c42aa47d,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt"] = func() locales.Translator {
		return &localen{

			locale:                 "pt",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "Esc.", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"dom.", "seg.", "ter.", "qua.", "qui.", "sex.", "sáb."},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "depois de Cristo"},
			timezones:              map[string]string{"ACDT": "Horário de Verão da Austrália Central", "ACST": "Horário Padrão da Austrália Central", "ACWDT": "Horário de Verão da Austrália Centro-Ocidental", "ACWST": "Horário Padrão da Austrália Centro-Ocidental", "ADT": "Horário de Verão do Atlântico", "AEDT": "Horário de Verão da Austrália Oriental", "AEST": "Horário Padrão da Austrália Oriental", "AKDT": "Horário de Verão do Alasca", "AKST": "Horário Padrão do Alasca", "ARST": "Horário de Verão da Argentina", "ART": "Horário Padrão da Argentina", "AST": "Horário Padrão do Atlântico", "AWDT": "Horário de Verão da Austrália Ocidental", "AWST": "Horário Padrão da Austrália Ocidental", "BOT": "Horário da Bolívia", "BT": "Horário do Butão", "CAT": "Horário da África Central", "CDT": "Horário de Verão Central", "CHADT": "Horário de Verão de Chatham", "CHAST": "Horário Padrão de Chatham", "CLST": "Horário de Verão do Chile", "CLT": "Horário Padrão do Chile", "COST": "Horário de Verão da Colômbia", "COT": "Horário Padrão da Colômbia", "CST": "Horário Padrão Central", "ChST": "Horário de Chamorro", "EAT": "Horário da África Oriental", "ECT": "Horário do Equador", "EDT": "Horário de Verão do Leste", "EST": "Horário Padrão do Leste", "GFT": "Horário da Guiana Francesa", "GMT": "Horário do Meridiano de Greenwich", "GST": "Horário do Golfo", "GYT": "Horário da Guiana", "HADT": "Horário de Verão do Havaí e Ilhas Aleutas", "HAST": "Horário Padrão do Havaí e Ilhas Aleutas", "HAT": "Horário de Verão da Terra Nova", "HECU": "Horário de Verão de Cuba", "HEEG": "Horário de Verão da Groelândia Oriental", "HENOMX": "Horário de Verão do Noroeste do México", "HEOG": "Horário de Verão da Groenlândia Ocidental", "HEPM": "Horário Verão de São Pedro e Miquelão", "HEPMX": "Horário de Verão do Pacífico Mexicano", "HKST": "Horário de Verão de Hong Kong", "HKT": "Horário Padrão de Hong Kong", "HNCU": "Horário Padrão de Cuba", "HNEG": "Horário Padrão da Groelândia Oriental", "HNNOMX": "Horário Padrão do Noroeste do México", "HNOG": "Horário Padrão da Groenlândia Ocidental", "HNPM": "Horário Padrão de São Pedro e Miquelão", "HNPMX": "Horário Padrão do Pacífico Mexicano", "HNT": "Horário Padrão da Terra Nova", "IST": "Horário Padrão da Índia", "JDT": "Horário de Verão do Japão", "JST": "Horário Padrão do Japão", "LHDT": "Horário de Verão de Lord Howe", "LHST": "Horário Padrão de Lord Howe", "MDT": "Horário de Verão de Macau", "MESZ": "Horário de Verão da Europa Central", "MEZ": "Horário Padrão da Europa Central", "MST": "Horário Padrão de Macau", "MYT": "Horário da Malásia", "NZDT": "Horário de Verão da Nova Zelândia", "NZST": "Horário Padrão da Nova Zelândia", "OESZ": "Horário de Verão da Europa Oriental", "OEZ": "Horário Padrão da Europa Oriental", "PDT": "Horário de Verão do Pacífico", "PST": "Horário Padrão do Pacífico", "SAST": "Horário da África do Sul", "SGT": "Horário Padrão de Cingapura", "SRT": "Horário do Suriname", "TMST": "Horário de Verão do Turcomenistão", "TMT": "Horário Padrão do Turcomenistão", "UYST": "Horário de Verão do Uruguai", "UYT": "Horário Padrão do Uruguai", "VET": "Horário da Venezuela", "WARST": "Horário de Verão da Argentina Ocidental", "WART": "Horário Padrão da Argentina Ocidental", "WAST": "Horário de Verão da África Ocidental", "WAT": "Horário Padrão da África Ocidental", "WESZ": "Horário de Verão da Europa Ocidental", "WEZ": "Horário Padrão da Europa Ocidental", "WIB": "Horário da Indonésia Ocidental", "WIT": "Horário da Indonésia Oriental", "WITA": "Horário da Indonésia Central", "∅∅∅": "Horário de Verão do Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_ao"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_AO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "Kz", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de Macau", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de Macau", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão do Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_br"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_BR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"dom.", "seg.", "ter.", "qua.", "qui.", "sex.", "sáb."},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"a.C.", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"antes de Cristo", "depois de Cristo"},
			timezones:              map[string]string{"ACDT": "Horário de Verão da Austrália Central", "ACST": "Horário Padrão da Austrália Central", "ACWDT": "Horário de Verão da Austrália Centro-Ocidental", "ACWST": "Horário Padrão da Austrália Centro-Ocidental", "ADT": "Horário de Verão do Atlântico", "AEDT": "Horário de Verão da Austrália Oriental", "AEST": "Horário Padrão da Austrália Oriental", "AKDT": "Horário de Verão do Alasca", "AKST": "Horário Padrão do Alasca", "ARST": "Horário de Verão da Argentina", "ART": "Horário Padrão da Argentina", "AST": "Horário Padrão do Atlântico", "AWDT": "Horário de Verão da Austrália Ocidental", "AWST": "Horário Padrão da Austrália Ocidental", "BOT": "Horário da Bolívia", "BT": "Horário do Butão", "CAT": "Horário da África Central", "CDT": "Horário de Verão Central", "CHADT": "Horário de Verão de Chatham", "CHAST": "Horário Padrão de Chatham", "CLST": "Horário de Verão do Chile", "CLT": "Horário Padrão do Chile", "COST": "Horário de Verão da Colômbia", "COT": "Horário Padrão da Colômbia", "CST": "Horário Padrão Central", "ChST": "Horário de Chamorro", "EAT": "Horário da África Oriental", "ECT": "Horário do Equador", "EDT": "Horário de Verão do Leste", "EST": "Horário Padrão do Leste", "GFT": "Horário da Guiana Francesa", "GMT": "Horário do Meridiano de Greenwich", "GST": "Horário do Golfo", "GYT": "Horário da Guiana", "HADT": "Horário de Verão do Havaí e Ilhas Aleutas", "HAST": "Horário Padrão do Havaí e Ilhas Aleutas", "HAT": "Horário de Verão da Terra Nova", "HECU": "Horário de Verão de Cuba", "HEEG": "Horário de Verão da Groelândia Oriental", "HENOMX": "Horário de Verão do Noroeste do México", "HEOG": "Horário de Verão da Groenlândia Ocidental", "HEPM": "Horário Verão de São Pedro e Miquelão", "HEPMX": "Horário de Verão do Pacífico Mexicano", "HKST": "Horário de Verão de Hong Kong", "HKT": "Horário Padrão de Hong Kong", "HNCU": "Horário Padrão de Cuba", "HNEG": "Horário Padrão da Groelândia Oriental", "HNNOMX": "Horário Padrão do Noroeste do México", "HNOG": "Horário Padrão da Groenlândia Ocidental", "HNPM": "Horário Padrão de São Pedro e Miquelão", "HNPMX": "Horário Padrão do Pacífico Mexicano", "HNT": "Horário Padrão da Terra Nova", "IST": "Horário Padrão da Índia", "JDT": "Horário de Verão do Japão", "JST": "Horário Padrão do Japão", "LHDT": "Horário de Verão de Lord Howe", "LHST": "Horário Padrão de Lord Howe", "MDT": "Horário de Verão das Montanhas", "MESZ": "Horário de Verão da Europa Central", "MEZ": "Horário Padrão da Europa Central", "MST": "Horário Padrão das Montanhas", "MYT": "Horário da Malásia", "NZDT": "Horário de Verão da Nova Zelândia", "NZST": "Horário Padrão da Nova Zelândia", "OESZ": "Horário de Verão da Europa Oriental", "OEZ": "Horário Padrão da Europa Oriental", "PDT": "Horário de Verão do Pacífico", "PST": "Horário Padrão do Pacífico", "SAST": "Horário da África do Sul", "SGT": "Horário Padrão de Cingapura", "SRT": "Horário do Suriname", "TMST": "Horário de Verão do Turcomenistão", "TMT": "Horário Padrão do Turcomenistão", "UYST": "Horário de Verão do Uruguai", "UYT": "Horário Padrão do Uruguai", "VET": "Horário da Venezuela", "WARST": "Horário de Verão da Argentina Ocidental", "WART": "Horário Padrão da Argentina Ocidental", "WAST": "Horário de Verão da África Ocidental", "WAT": "Horário Padrão da África Ocidental", "WESZ": "Horário de Verão da Europa Ocidental", "WEZ": "Horário Padrão da Europa Ocidental", "WIB": "Horário da Indonésia Ocidental", "WIT": "Horário da Indonésia Oriental", "WITA": "Horário da Indonésia Central", "∅∅∅": "Horário de Verão de Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão de Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_cv"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_CV",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "\u200b", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "\u200bPTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão de Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_gq"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_GQ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão dos Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_gw"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_GW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão dos Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_lu"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_LU",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "F", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão do Amazonas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_mo"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_MO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP$", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão de Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_mz"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_MZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MTn", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão dos Açores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_pt"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_PT",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "\u200b", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "£", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão do Amazonas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_st"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_ST",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão de Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["pt_tl"] = func() locales.Translator {
		return &localen{

			locale:                 "pt_TL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "fev.", "mar.", "abr.", "mai.", "jun.", "jul.", "ago.", "set.", "out.", "nov.", "dez."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"},
			daysAbbreviated:        []string{"domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"},
			daysNarrow:             []string{"D", "S", "T", "Q", "Q", "S", "S"},
			daysWide:               []string{"domingo", "segunda-feira", "terça-feira", "quarta-feira", "quinta-feira", "sexta-feira", "sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"da manhã", "da tarde"},
			erasAbbreviated:        []string{"a.E.C.", "E.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Hora de verão da Austrália Central", "ACST": "Hora padrão da Austrália Central", "ACWDT": "Hora de verão da Austrália Central Ocidental", "ACWST": "Hora padrão da Austrália Central Ocidental", "ADT": "Hora de verão do Atlântico", "AEDT": "Hora de verão da Austrália Oriental", "AEST": "Hora padrão da Austrália Oriental", "AKDT": "Hora de verão do Alasca", "AKST": "Hora padrão do Alasca", "ARST": "Hora de verão da Argentina", "ART": "Hora padrão da Argentina", "AST": "Hora padrão do Atlântico", "AWDT": "Hora de verão da Austrália Ocidental", "AWST": "Hora padrão da Austrália Ocidental", "BOT": "Hora da Bolívia", "BT": "Hora do Butão", "CAT": "Hora da África Central", "CDT": "Hora de verão central norte-americana", "CHADT": "Hora de verão de Chatham", "CHAST": "Hora padrão de Chatham", "CLST": "Hora de verão do Chile", "CLT": "Hora padrão do Chile", "COST": "Hora de verão da Colômbia", "COT": "Hora padrão da Colômbia", "CST": "Hora padrão central norte-americana", "ChST": "Hora padrão de Chamorro", "EAT": "Hora da África Oriental", "ECT": "Hora do Equador", "EDT": "Hora de verão oriental norte-americana", "EST": "Hora padrão oriental norte-americana", "GFT": "Hora da Guiana Francesa", "GMT": "Hora de Greenwich", "GST": "Hora padrão do Golfo", "GYT": "Hora da Guiana", "HADT": "Hora de verão do Havai e Aleutas", "HAST": "Hora padrão do Havai e Aleutas", "HAT": "Hora de verão da Terra Nova", "HECU": "Hora de verão de Cuba", "HEEG": "Hora de verão da Gronelândia Oriental", "HENOMX": "Hora de verão do Noroeste do México", "HEOG": "Hora de verão da Gronelândia Ocidental", "HEPM": "Hora de verão de São Pedro e Miquelão", "HEPMX": "Hora de verão do Pacífico Mexicano", "HKST": "Hora de verão de Hong Kong", "HKT": "Hora padrão de Hong Kong", "HNCU": "Hora padrão de Cuba", "HNEG": "Hora padrão da Gronelândia Oriental", "HNNOMX": "Hora padrão do Noroeste do México", "HNOG": "Hora padrão da Gronelândia Ocidental", "HNPM": "Hora padrão de São Pedro e Miquelão", "HNPMX": "Hora padrão do Pacífico Mexicano", "HNT": "Hora padrão da Terra Nova", "IST": "Hora padrão da Índia", "JDT": "Hora de verão do Japão", "JST": "Hora padrão do Japão", "LHDT": "Hora de verão de Lord Howe", "LHST": "Hora padrão de Lord Howe", "MDT": "Hora de verão de montanha norte-americana", "MESZ": "Hora de verão da Europa Central", "MEZ": "Hora padrão da Europa Central", "MST": "Hora padrão de montanha norte-americana", "MYT": "Hora da Malásia", "NZDT": "Hora de verão da Nova Zelândia", "NZST": "Hora padrão da Nova Zelândia", "OESZ": "Hora de verão da Europa Oriental", "OEZ": "Hora padrão da Europa Oriental", "PDT": "Hora de verão do Pacífico norte-americana", "PST": "Hora padrão do Pacífico norte-americana", "SAST": "Hora da África do Sul", "SGT": "Hora padrão de Singapura", "SRT": "Hora do Suriname", "TMST": "Hora de verão do Turquemenistão", "TMT": "Hora padrão do Turquemenistão", "UYST": "Hora de verão do Uruguai", "UYT": "Hora padrão do Uruguai", "VET": "Hora da Venezuela", "WARST": "Hora de verão da Argentina Ocidental", "WART": "Hora padrão da Argentina Ocidental", "WAST": "Hora de verão da África Ocidental", "WAT": "Hora padrão da África Ocidental", "WESZ": "Hora de verão da Europa Ocidental", "WEZ": "Hora padrão da Europa Ocidental", "WIB": "Hora da Indonésia Ocidental", "WIT": "Hora da Indonésia Oriental", "WITA": "Hora da Indonésia Central", "∅∅∅": "Hora de verão de Brasília"},

			fnCardinalPluralRule:  fnCardinalPluralRule_72e0d02c4132f59ce64c852514edc801,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_259dde974c52d4b5dc45a53d5ed154c6,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["qu"] = func() locales.Translator {
		return &localen{

			locale:                 "qu",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBG", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "DBM", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "DBZ", "$CA", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "S/", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$US", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Ene", "Feb", "Mar", "Abr", "May", "Jun", "Jul", "Ago", "Set", "Oct", "Nov", "Dic"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Enero", "Febrero", "Marzo", "Abril", "Mayo", "Junio", "Julio", "Agosto", "Setiembre", "Octubre", "Noviembre", "Diciembre"},
			daysAbbreviated:        []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysWide:               []string{"Domingo", "Lunes", "Martes", "Miércoles", "Jueves", "Viernes", "Sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"BCE", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "d.C."},
			timezones:              map[string]string{"ACDT": "Hora de Verano de Australia Central", "ACST": "Hora Estandar de Australia Central", "ACWDT": "Hora de Verano de Australia Central Occidental", "ACWST": "Hora Estandar de Australia Central Occidental", "ADT": "Hora De Verano del Atlántico", "AEDT": "Hora de Verano de Australia Oriental", "AEST": "Hora Estandar de Australia Oriental", "AKDT": "Hora de Verano de Alaska", "AKST": "Hora Estandar de Alaska", "ARST": "Hora de Verano de Argentina", "ART": "Hora Estandar de Argentina", "AST": "Hora Estandar del Atlántico", "AWDT": "Hora de Verano de Australia Occidental", "AWST": "Hora Estandar de Australia Occidental", "BOT": "Bolivia Time", "BT": "Hora de Bután", "CAT": "Hora de Africa Central", "CDT": "Hora Central de Verano", "CHADT": "Hora de Verano de Chatham", "CHAST": "Hora Estandar de Chatham", "CLST": "Hora de Verano de Chile", "CLT": "Hora Estandar de Chile", "COST": "Hora de Verano de Colombia", "COT": "Hora Estandar de Colombia", "CST": "Estandard Hora Central", "ChST": "Hora Estandar de Chamorro", "EAT": "Hora de Africa Oriental", "ECT": "Hora de Ecuador", "EDT": "Hora de Verano del Este", "EST": "Hora Estandar del Este", "GFT": "Hora de Guayana Francesa", "GMT": "Hora del Meridiano de Greenwich", "GST": "Hora Estandar del Golfo", "GYT": "Hora de Guyana", "HADT": "Hora de Verano de Hawai-Aleutiano", "HAST": "Hora Estandar de Hawai-Aleutiano", "HAT": "Hora de Verano de Terranova", "HECU": "Hora de Verano de Cuba", "HEEG": "Hora de Verano de Groenlandia", "HENOMX": "Hora de Verano del Noroeste de México", "HEOG": "Hora de Verano de Groenlandia Occidental", "HEPM": "Hora de Verano de San Pedro y Miquelón", "HEPMX": "Hora de Verano del Pacífico Mexicano", "HKST": "Hora de Verano de Hong Kong", "HKT": "Hora Estandar de Hong Kong", "HNCU": "Hora Estandar de Cuba", "HNEG": "Hora Estandar de Groenlandia", "HNNOMX": "Hora Estandar del Noroeste de México", "HNOG": "Hora Estandar de Groenlandia Occidental", "HNPM": "Hora Estandar de San Pedro y Miquelón", "HNPMX": "Hora Estandar del Pacífico Mexicano", "HNT": "Hora Estandar de Terranova", "IST": "Hora Estandar de India", "JDT": "Hora de Verano de Japón", "JST": "Hora Estandar de Japón", "LHDT": "Hora de Verano de Lord Howe", "LHST": "Hora Estandar de Lord Howe", "MDT": "Hora de Verano de la Montaña", "MESZ": "Hora de Verano de Europa Central", "MEZ": "Hora Estandar de Europa Central", "MST": "Hora Estandar de la Montaña", "MYT": "Hora de Malasia", "NZDT": "Hora de Verano de Nueva Zelanda", "NZST": "Hora Estandar de Nueva Zelanda", "OESZ": "Hora de Verano de Europa Oriental", "OEZ": "Hora Estandar de Europa Oriental", "PDT": "Hora de Verano del Pacífico", "PST": "Hora Estandar del Pacífico", "SAST": "Hora de Sudafrica", "SGT": "Hora Estandar de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora de Verano de Turkmenistán", "TMT": "Hora Estandar de Turkmenistán", "UYST": "Hora de Verano de Uruguay", "UYT": "Hora Estandar de Uruguay", "VET": "Hora de Venezuela", "WARST": "Hora de Verano del Oeste de Argentina", "WART": "Hora Estandar del Oeste de Argentina", "WAST": "Hora Estandar de Verano de Africa Occidental", "WAT": "Hora Estandar de Africa Occidental", "WESZ": "Hora de Verano de Europa Occidental", "WEZ": "Hora Estandar de Europa Occidental", "WIB": "Hora de Indonesia Occidental", "WIT": "Hora de Indonesia Oriental", "WITA": "Hora de Indonesia Central", "∅∅∅": "Hora de Verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["qu_bo"] = func() locales.Translator {
		return &localen{

			locale:                 "qu_BO",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "Bs", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Ene", "Feb", "Mar", "Abr", "May", "Jun", "Jul", "Ago", "Set", "Oct", "Nov", "Dic"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Enero", "Febrero", "Marzo", "Abril", "Mayo", "Junio", "Julio", "Agosto", "Setiembre", "Octubre", "Noviembre", "Diciembre"},
			daysAbbreviated:        []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysWide:               []string{"Domingo", "Lunes", "Martes", "Miércoles", "Jueves", "Viernes", "Sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"BCE", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "d.C."},
			timezones:              map[string]string{"ACDT": "Hora de Verano de Australia Central", "ACST": "Hora Estandar de Australia Central", "ACWDT": "Hora de Verano de Australia Central Occidental", "ACWST": "Hora Estandar de Australia Central Occidental", "ADT": "Hora De Verano del Atlántico", "AEDT": "Hora de Verano de Australia Oriental", "AEST": "Hora Estandar de Australia Oriental", "AKDT": "Hora de Verano de Alaska", "AKST": "Hora Estandar de Alaska", "ARST": "Hora de Verano de Argentina", "ART": "Hora Estandar de Argentina", "AST": "Hora Estandar del Atlántico", "AWDT": "Hora de Verano de Australia Occidental", "AWST": "Hora Estandar de Australia Occidental", "BOT": "Bolivia Time", "BT": "Hora de Bután", "CAT": "Hora de Africa Central", "CDT": "Hora Central de Verano", "CHADT": "Hora de Verano de Chatham", "CHAST": "Hora Estandar de Chatham", "CLST": "Hora de Verano de Chile", "CLT": "Hora Estandar de Chile", "COST": "Hora de Verano de Colombia", "COT": "Hora Estandar de Colombia", "CST": "Estandard Hora Central", "ChST": "Hora Estandar de Chamorro", "EAT": "Hora de Africa Oriental", "ECT": "Hora de Ecuador", "EDT": "Hora de Verano del Este", "EST": "Hora Estandar del Este", "GFT": "Hora de Guayana Francesa", "GMT": "Hora del Meridiano de Greenwich", "GST": "Hora Estandar del Golfo", "GYT": "Hora de Guyana", "HADT": "Hora de Verano de Hawai-Aleutiano", "HAST": "Hora Estandar de Hawai-Aleutiano", "HAT": "Hora de Verano de Terranova", "HECU": "Hora de Verano de Cuba", "HEEG": "Hora de Verano de Groenlandia", "HENOMX": "Hora de Verano del Noroeste de México", "HEOG": "Hora de Verano de Groenlandia Occidental", "HEPM": "Hora de Verano de San Pedro y Miquelón", "HEPMX": "Hora de Verano del Pacífico Mexicano", "HKST": "Hora de Verano de Hong Kong", "HKT": "Hora Estandar de Hong Kong", "HNCU": "Hora Estandar de Cuba", "HNEG": "Hora Estandar de Groenlandia", "HNNOMX": "Hora Estandar del Noroeste de México", "HNOG": "Hora Estandar de Groenlandia Occidental", "HNPM": "Hora Estandar de San Pedro y Miquelón", "HNPMX": "Hora Estandar del Pacífico Mexicano", "HNT": "Hora Estandar de Terranova", "IST": "Hora Estandar de India", "JDT": "Hora de Verano de Japón", "JST": "Hora Estandar de Japón", "LHDT": "Hora de Verano de Lord Howe", "LHST": "Hora Estandar de Lord Howe", "MDT": "Hora de Verano de la Montaña", "MESZ": "Hora de Verano de Europa Central", "MEZ": "Hora Estandar de Europa Central", "MST": "Hora Estandar de la Montaña", "MYT": "Hora de Malasia", "NZDT": "Hora de Verano de Nueva Zelanda", "NZST": "Hora Estandar de Nueva Zelanda", "OESZ": "Hora de Verano de Europa Oriental", "OEZ": "Hora Estandar de Europa Oriental", "PDT": "Hora de Verano del Pacífico", "PST": "Hora Estandar del Pacífico", "SAST": "Hora de Sudafrica", "SGT": "Hora Estandar de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora de Verano de Turkmenistán", "TMT": "Hora Estandar de Turkmenistán", "UYST": "Hora de Verano de Uruguay", "UYT": "Hora Estandar de Uruguay", "VET": "Hora de Venezuela", "WARST": "Hora de Verano del Oeste de Argentina", "WART": "Hora Estandar del Oeste de Argentina", "WAST": "Hora Estandar de Verano de Africa Occidental", "WAT": "Hora Estandar de Africa Occidental", "WESZ": "Hora de Verano de Europa Occidental", "WEZ": "Hora Estandar de Europa Occidental", "WIB": "Hora de Indonesia Occidental", "WIT": "Hora de Indonesia Oriental", "WITA": "Hora de Indonesia Central", "∅∅∅": "Hora de Verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["qu_ec"] = func() locales.Translator {
		return &localen{

			locale:                 "qu_EC",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Ene", "Feb", "Mar", "Abr", "May", "Jun", "Jul", "Ago", "Set", "Oct", "Nov", "Dic"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Enero", "Febrero", "Marzo", "Abril", "Mayo", "Junio", "Julio", "Agosto", "Setiembre", "Octubre", "Noviembre", "Diciembre"},
			daysAbbreviated:        []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysWide:               []string{"Domingo", "Lunes", "Martes", "Miércoles", "Jueves", "Viernes", "Sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"BCE", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "d.C."},
			timezones:              map[string]string{"ACDT": "Hora de Verano de Australia Central", "ACST": "Hora Estandar de Australia Central", "ACWDT": "Hora de Verano de Australia Central Occidental", "ACWST": "Hora Estandar de Australia Central Occidental", "ADT": "Hora De Verano del Atlántico", "AEDT": "Hora de Verano de Australia Oriental", "AEST": "Hora Estandar de Australia Oriental", "AKDT": "Hora de Verano de Alaska", "AKST": "Hora Estandar de Alaska", "ARST": "Hora de Verano de Argentina", "ART": "Hora Estandar de Argentina", "AST": "Hora Estandar del Atlántico", "AWDT": "Hora de Verano de Australia Occidental", "AWST": "Hora Estandar de Australia Occidental", "BOT": "Bolivia Time", "BT": "Hora de Bután", "CAT": "Hora de Africa Central", "CDT": "Hora Central de Verano", "CHADT": "Hora de Verano de Chatham", "CHAST": "Hora Estandar de Chatham", "CLST": "Hora de Verano de Chile", "CLT": "Hora Estandar de Chile", "COST": "Hora de Verano de Colombia", "COT": "Hora Estandar de Colombia", "CST": "Estandard Hora Central", "ChST": "Hora Estandar de Chamorro", "EAT": "Hora de Africa Oriental", "ECT": "Hora de Ecuador", "EDT": "Hora de Verano del Este", "EST": "Hora Estandar del Este", "GFT": "Hora de Guayana Francesa", "GMT": "Hora del Meridiano de Greenwich", "GST": "Hora Estandar del Golfo", "GYT": "Hora de Guyana", "HADT": "Hora de Verano de Hawai-Aleutiano", "HAST": "Hora Estandar de Hawai-Aleutiano", "HAT": "Hora de Verano de Terranova", "HECU": "Hora de Verano de Cuba", "HEEG": "Hora de Verano de Groenlandia", "HENOMX": "Hora de Verano del Noroeste de México", "HEOG": "Hora de Verano de Groenlandia Occidental", "HEPM": "Hora de Verano de San Pedro y Miquelón", "HEPMX": "Hora de Verano del Pacífico Mexicano", "HKST": "Hora de Verano de Hong Kong", "HKT": "Hora Estandar de Hong Kong", "HNCU": "Hora Estandar de Cuba", "HNEG": "Hora Estandar de Groenlandia", "HNNOMX": "Hora Estandar del Noroeste de México", "HNOG": "Hora Estandar de Groenlandia Occidental", "HNPM": "Hora Estandar de San Pedro y Miquelón", "HNPMX": "Hora Estandar del Pacífico Mexicano", "HNT": "Hora Estandar de Terranova", "IST": "Hora Estandar de India", "JDT": "Hora de Verano de Japón", "JST": "Hora Estandar de Japón", "LHDT": "Hora de Verano de Lord Howe", "LHST": "Hora Estandar de Lord Howe", "MDT": "Hora de Verano de la Montaña", "MESZ": "Hora de Verano de Europa Central", "MEZ": "Hora Estandar de Europa Central", "MST": "Hora Estandar de la Montaña", "MYT": "Hora de Malasia", "NZDT": "Hora de Verano de Nueva Zelanda", "NZST": "Hora Estandar de Nueva Zelanda", "OESZ": "Hora de Verano de Europa Oriental", "OEZ": "Hora Estandar de Europa Oriental", "PDT": "Hora de Verano del Pacífico", "PST": "Hora Estandar del Pacífico", "SAST": "Hora de Sudafrica", "SGT": "Hora Estandar de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora de Verano de Turkmenistán", "TMT": "Hora Estandar de Turkmenistán", "UYST": "Hora de Verano de Uruguay", "UYT": "Hora Estandar de Uruguay", "VET": "Hora de Venezuela", "WARST": "Hora de Verano del Oeste de Argentina", "WART": "Hora Estandar del Oeste de Argentina", "WAST": "Hora Estandar de Verano de Africa Occidental", "WAT": "Hora Estandar de Africa Occidental", "WESZ": "Hora de Verano de Europa Occidental", "WEZ": "Hora Estandar de Europa Occidental", "WIB": "Hora de Indonesia Occidental", "WIT": "Hora de Indonesia Oriental", "WITA": "Hora de Indonesia Central", "∅∅∅": "Hora de Verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["qu_pe"] = func() locales.Translator {
		return &localen{

			locale:                 "qu_PE",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Ene", "Feb", "Mar", "Abr", "May", "Jun", "Jul", "Ago", "Set", "Oct", "Nov", "Dic"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Enero", "Febrero", "Marzo", "Abril", "Mayo", "Junio", "Julio", "Agosto", "Setiembre", "Octubre", "Noviembre", "Diciembre"},
			daysAbbreviated:        []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysNarrow:             []string{"D", "L", "M", "X", "J", "V", "S"},
			daysShort:              []string{"Dom", "Lun", "Mar", "Mié", "Jue", "Vie", "Sab"},
			daysWide:               []string{"Domingo", "Lunes", "Martes", "Miércoles", "Jueves", "Viernes", "Sábado"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"BCE", "d.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "d.C."},
			timezones:              map[string]string{"ACDT": "Hora de Verano de Australia Central", "ACST": "Hora Estandar de Australia Central", "ACWDT": "Hora de Verano de Australia Central Occidental", "ACWST": "Hora Estandar de Australia Central Occidental", "ADT": "Hora De Verano del Atlántico", "AEDT": "Hora de Verano de Australia Oriental", "AEST": "Hora Estandar de Australia Oriental", "AKDT": "Hora de Verano de Alaska", "AKST": "Hora Estandar de Alaska", "ARST": "Hora de Verano de Argentina", "ART": "Hora Estandar de Argentina", "AST": "Hora Estandar del Atlántico", "AWDT": "Hora de Verano de Australia Occidental", "AWST": "Hora Estandar de Australia Occidental", "BOT": "Bolivia Time", "BT": "Hora de Bután", "CAT": "Hora de Africa Central", "CDT": "Hora Central de Verano", "CHADT": "Hora de Verano de Chatham", "CHAST": "Hora Estandar de Chatham", "CLST": "Hora de Verano de Chile", "CLT": "Hora Estandar de Chile", "COST": "Hora de Verano de Colombia", "COT": "Hora Estandar de Colombia", "CST": "Estandard Hora Central", "ChST": "Hora Estandar de Chamorro", "EAT": "Hora de Africa Oriental", "ECT": "Hora de Ecuador", "EDT": "Hora de Verano del Este", "EST": "Hora Estandar del Este", "GFT": "Hora de Guayana Francesa", "GMT": "Hora del Meridiano de Greenwich", "GST": "Hora Estandar del Golfo", "GYT": "Hora de Guyana", "HADT": "Hora de Verano de Hawai-Aleutiano", "HAST": "Hora Estandar de Hawai-Aleutiano", "HAT": "Hora de Verano de Terranova", "HECU": "Hora de Verano de Cuba", "HEEG": "Hora de Verano de Groenlandia", "HENOMX": "Hora de Verano del Noroeste de México", "HEOG": "Hora de Verano de Groenlandia Occidental", "HEPM": "Hora de Verano de San Pedro y Miquelón", "HEPMX": "Hora de Verano del Pacífico Mexicano", "HKST": "Hora de Verano de Hong Kong", "HKT": "Hora Estandar de Hong Kong", "HNCU": "Hora Estandar de Cuba", "HNEG": "Hora Estandar de Groenlandia", "HNNOMX": "Hora Estandar del Noroeste de México", "HNOG": "Hora Estandar de Groenlandia Occidental", "HNPM": "Hora Estandar de San Pedro y Miquelón", "HNPMX": "Hora Estandar del Pacífico Mexicano", "HNT": "Hora Estandar de Terranova", "IST": "Hora Estandar de India", "JDT": "Hora de Verano de Japón", "JST": "Hora Estandar de Japón", "LHDT": "Hora de Verano de Lord Howe", "LHST": "Hora Estandar de Lord Howe", "MDT": "Hora de Verano de la Montaña", "MESZ": "Hora de Verano de Europa Central", "MEZ": "Hora Estandar de Europa Central", "MST": "Hora Estandar de la Montaña", "MYT": "Hora de Malasia", "NZDT": "Hora de Verano de Nueva Zelanda", "NZST": "Hora Estandar de Nueva Zelanda", "OESZ": "Hora de Verano de Europa Oriental", "OEZ": "Hora Estandar de Europa Oriental", "PDT": "Hora de Verano del Pacífico", "PST": "Hora Estandar del Pacífico", "SAST": "Hora de Sudafrica", "SGT": "Hora Estandar de Singapur", "SRT": "Hora de Surinam", "TMST": "Hora de Verano de Turkmenistán", "TMT": "Hora Estandar de Turkmenistán", "UYST": "Hora de Verano de Uruguay", "UYT": "Hora Estandar de Uruguay", "VET": "Hora de Venezuela", "WARST": "Hora de Verano del Oeste de Argentina", "WART": "Hora Estandar del Oeste de Argentina", "WAST": "Hora Estandar de Verano de Africa Occidental", "WAT": "Hora Estandar de Africa Occidental", "WESZ": "Hora de Verano de Europa Occidental", "WEZ": "Hora Estandar de Europa Occidental", "WIB": "Hora de Indonesia Occidental", "WIT": "Hora de Indonesia Oriental", "WITA": "Hora de Indonesia Central", "∅∅∅": "Hora de Verano de Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rm"] = func() locales.Translator {
		return &localen{

			locale:                 "rm",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  "’",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "schan.", "favr.", "mars", "avr.", "matg", "zercl.", "fan.", "avust", "sett.", "oct.", "nov.", "dec."},
			monthsNarrow:           []string{"", "S", "F", "M", "A", "M", "Z", "F", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "da schaner", "da favrer", "da mars", "d’avrigl", "da matg", "da zercladur", "da fanadur", "d’avust", "da settember", "d’october", "da november", "da december"},
			daysAbbreviated:        []string{"du", "gli", "ma", "me", "gie", "ve", "so"},
			daysNarrow:             []string{"D", "G", "M", "M", "G", "V", "S"},
			daysShort:              []string{"du", "gli", "ma", "me", "gie", "ve", "so"},
			daysWide:               []string{"dumengia", "glindesdi", "mardi", "mesemna", "gievgia", "venderdi", "sonda"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. Cr.", "s. Cr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"avant Cristus", "suenter Cristus"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74,
			fnFmtCurrency:         fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204,
			fnFmtDateFull:         fnFmtDateFull_4c2ead12910ff721f96d3aca986ca749,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rm_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "rm_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  "’",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "schan.", "favr.", "mars", "avr.", "matg", "zercl.", "fan.", "avust", "sett.", "oct.", "nov.", "dec."},
			monthsNarrow:           []string{"", "S", "F", "M", "A", "M", "Z", "F", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "da schaner", "da favrer", "da mars", "d’avrigl", "da matg", "da zercladur", "da fanadur", "d’avust", "da settember", "d’october", "da november", "da december"},
			daysAbbreviated:        []string{"du", "gli", "ma", "me", "gie", "ve", "so"},
			daysNarrow:             []string{"D", "G", "M", "M", "G", "V", "S"},
			daysShort:              []string{"du", "gli", "ma", "me", "gie", "ve", "so"},
			daysWide:               []string{"dumengia", "glindesdi", "mardi", "mesemna", "gievgia", "venderdi", "sonda"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"av. Cr.", "s. Cr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"avant Cristus", "suenter Cristus"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_54cd50d7a0db8f91c56e54eeea38fa74,
			fnFmtCurrency:         fnFmtCurrency_a81f72a39548e95f8f6af89799fb3204,
			fnFmtDateFull:         fnFmtDateFull_4c2ead12910ff721f96d3aca986ca749,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_7b5970ebef47b6071b7a9bff570eba75,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_0a6b715b13c31edcb79152f44bd8401c,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rn"] = func() locales.Translator {
		return &localen{

			locale:             "rn",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "FBu", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "Mut.", "Gas.", "Wer.", "Mat.", "Gic.", "Kam.", "Nya.", "Kan.", "Nze.", "Ukw.", "Ugu.", "Uku."},
			monthsWide:         []string{"", "Nzero", "Ruhuhuma", "Ntwarante", "Ndamukiza", "Rusama", "Ruheshi", "Mukakaro", "Nyandagaro", "Nyakanga", "Gitugutu", "Munyonyo", "Kigarama"},
			daysAbbreviated:    []string{"cu.", "mbe.", "kab.", "gtu.", "kan.", "gnu.", "gnd."},
			daysWide:           []string{"Ku w’indwi", "Ku wa mbere", "Ku wa kabiri", "Ku wa gatatu", "Ku wa kane", "Ku wa gatanu", "Ku wa gatandatu"},
			periodsAbbreviated: []string{"Z.MU.", "Z.MW."},
			periodsWide:        []string{"Z.MU.", "Z.MW."},
			erasAbbreviated:    []string{"Mb.Y.", "Ny.Y"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Mbere ya Yezu", "Nyuma ya Yezu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rn_bi"] = func() locales.Translator {
		return &localen{

			locale:             "rn_BI",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "Mut.", "Gas.", "Wer.", "Mat.", "Gic.", "Kam.", "Nya.", "Kan.", "Nze.", "Ukw.", "Ugu.", "Uku."},
			monthsWide:         []string{"", "Nzero", "Ruhuhuma", "Ntwarante", "Ndamukiza", "Rusama", "Ruheshi", "Mukakaro", "Nyandagaro", "Nyakanga", "Gitugutu", "Munyonyo", "Kigarama"},
			daysAbbreviated:    []string{"cu.", "mbe.", "kab.", "gtu.", "kan.", "gnu.", "gnd."},
			daysWide:           []string{"Ku w’indwi", "Ku wa mbere", "Ku wa kabiri", "Ku wa gatatu", "Ku wa kane", "Ku wa gatanu", "Ku wa gatandatu"},
			periodsAbbreviated: []string{"Z.MU.", "Z.MW."},
			periodsWide:        []string{"Z.MU.", "Z.MW."},
			erasAbbreviated:    []string{"Mb.Y.", "Ny.Y"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Mbere ya Yezu", "Nyuma ya Yezu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_e53d7ec2cf4741b3601391772e4c83b0,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ro"] = func() locales.Translator {
		return &localen{

			locale:                 "ro",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "ian.", "feb.", "mar.", "apr.", "mai", "iun.", "iul.", "aug.", "sept.", "oct.", "nov.", "dec."},
			monthsNarrow:           []string{"", "I", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "ianuarie", "februarie", "martie", "aprilie", "mai", "iunie", "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie"},
			daysAbbreviated:        []string{"dum.", "lun.", "mar.", "mie.", "joi", "vin.", "sâm."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"du.", "lu.", "ma.", "mi.", "joi", "vi.", "sâ."},
			daysWide:               []string{"duminică", "luni", "marți", "miercuri", "joi", "vineri", "sâmbătă"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"î.Hr.", "d.Hr."},
			erasNarrow:             []string{"î.Hr.", "d.Hr."},
			erasWide:               []string{"înainte de Hristos", "după Hristos"},
			timezones:              map[string]string{"ACDT": "Ora de vară a Australiei Centrale", "ACST": "Ora standard a Australiei Centrale", "ACWDT": "Ora de vară a Australiei Central Occidentale", "ACWST": "Ora standard a Australiei Central Occidentale", "ADT": "Ora de vară în zona Atlantic nord-americană", "AEDT": "Ora de vară a Australiei Orientale", "AEST": "Ora standard a Australiei Orientale", "AKDT": "Ora de vară din Alaska", "AKST": "Ora standard din Alaska", "ARST": "Ora de vară a Argentinei", "ART": "Ora standard a Argentinei", "AST": "Ora standard în zona Atlantic nord-americană", "AWDT": "Ora de vară a Australiei Occidentale", "AWST": "Ora standard a Australiei Occidentale", "BOT": "Ora Boliviei", "BT": "Ora Bhutanului", "CAT": "Ora Africii Centrale", "CDT": "Ora de vară centrală nord-americană", "CHADT": "Ora de vară din Chatham", "CHAST": "Ora standard din Chatham", "CLST": "Ora de vară din Chile", "CLT": "Ora standard din Chile", "COST": "Ora de vară a Columbiei", "COT": "Ora standard a Columbiei", "CST": "Ora standard centrală nord-americană", "ChST": "Ora din Chamorro", "EAT": "Ora Africii Orientale", "ECT": "Ora Ecuadorului", "EDT": "Ora de vară orientală nord-americană", "EST": "Ora standard orientală nord-americană", "GFT": "Ora din Guyana Franceză", "GMT": "Ora de Greenwhich", "GST": "Ora standard a Golfului", "GYT": "Ora din Guyana", "HADT": "Ora de vară din Hawaii-Aleutine", "HAST": "Ora standard din Hawaii-Aleutine", "HAT": "Ora de vară din Newfoundland", "HECU": "Ora de vară a Cubei", "HEEG": "Ora de vară a Groenlandei orientale", "HENOMX": "Ora de vară a Mexicului de nord-vest", "HEOG": "Ora de vară a Groenlandei occidentale", "HEPM": "Ora de vară din Saint-Pierre și Miquelon", "HEPMX": "Ora de vară a zonei Pacific mexicane", "HKST": "Ora de vară din Hong Kong", "HKT": "Ora standard din Hong Kong", "HNCU": "Ora standard a Cubei", "HNEG": "Ora standard a Groenlandei orientale", "HNNOMX": "Ora standard a Mexicului de nord-vest", "HNOG": "Ora standard a Groenlandei occidentale", "HNPM": "Ora standard din Saint-Pierre și Miquelon", "HNPMX": "Ora standard a zonei Pacific mexicane", "HNT": "Ora standard din Newfoundland", "IST": "Ora Indiei", "JDT": "Ora de vară a Japoniei", "JST": "Ora standard a Japoniei", "LHDT": "Ora de vară din Lord Howe", "LHST": "Ora standard din Lord Howe", "MDT": "Ora de vară în zona montană nord-americană", "MESZ": "Ora de vară a Europei Centrale", "MEZ": "Ora standard a Europei Centrale", "MST": "Ora standard în zona montană nord-americană", "MYT": "Ora din Malaysia", "NZDT": "Ora de vară a Noii Zeelande", "NZST": "Ora standard a Noii Zeelande", "OESZ": "Ora de vară a Europei de Est", "OEZ": "Ora standard a Europei de Est", "PDT": "Ora de vară în zona Pacific nord-americană", "PST": "Ora standard în zona Pacific nord-americană", "SAST": "Ora Africii Meridionale", "SGT": "Ora din Singapore", "SRT": "Ora Surinamului", "TMST": "Ora de vară din Turkmenistan", "TMT": "Ora standard din Turkmenistan", "UYST": "Ora de vară a Uruguayului", "UYT": "Ora standard a Uruguayului", "VET": "Ora Venezuelei", "WARST": "Ora de vară a Argentinei Occidentale", "WART": "Ora standard a Argentinei Occidentale", "WAST": "Ora de vară a Africii Occidentale", "WAT": "Ora standard a Africii Occidentale", "WESZ": "Ora de vară a Europei de Vest", "WEZ": "Ora standard a Europei de Vest", "WIB": "Ora Indoneziei de Vest", "WIT": "Ora Indoneziei de Est", "WITA": "Ora Indoneziei Centrale", "∅∅∅": "Ora de vară din Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_a60b3f7c6ddcf0678df3df0fe274201a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6db355e32f98d7cf05b06cd4b38d6c63,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ro_md"] = func() locales.Translator {
		return &localen{

			locale:                 "ro_MD",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "L", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "ian.", "feb.", "mar.", "apr.", "mai", "iun.", "iul.", "aug.", "sept.", "oct.", "nov.", "dec."},
			monthsNarrow:           []string{"", "I", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "ianuarie", "februarie", "martie", "aprilie", "mai", "iunie", "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie"},
			daysAbbreviated:        []string{"Dum", "Lun", "Mar", "Mie", "Joi", "Vin", "Sâm"},
			daysNarrow:             []string{"D", "L", "Ma", "Mi", "J", "V", "S"},
			daysShort:              []string{"Du", "Lu", "Ma", "Mi", "Jo", "Vi", "Sâ"},
			daysWide:               []string{"duminică", "luni", "marți", "miercuri", "joi", "vineri", "sâmbătă"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Ora de vară a Australiei Centrale", "ACST": "Ora standard a Australiei Centrale", "ACWDT": "Ora de vară a Australiei Central Occidentale", "ACWST": "Ora standard a Australiei Central Occidentale", "ADT": "Ora de vară în zona Atlantic nord-americană", "AEDT": "Ora de vară a Australiei Orientale", "AEST": "Ora standard a Australiei Orientale", "AKDT": "Ora de vară din Alaska", "AKST": "Ora standard din Alaska", "ARST": "Ora de vară a Argentinei", "ART": "Ora standard a Argentinei", "AST": "Ora standard în zona Atlantic nord-americană", "AWDT": "Ora de vară a Australiei Occidentale", "AWST": "Ora standard a Australiei Occidentale", "BOT": "Ora Boliviei", "BT": "Ora Bhutanului", "CAT": "Ora Africii Centrale", "CDT": "Ora de vară centrală nord-americană", "CHADT": "Ora de vară din Chatham", "CHAST": "Ora standard din Chatham", "CLST": "Ora de vară din Chile", "CLT": "Ora standard din Chile", "COST": "Ora de vară a Columbiei", "COT": "Ora standard a Columbiei", "CST": "Ora standard centrală nord-americană", "ChST": "Ora din Chamorro", "EAT": "Ora Africii Orientale", "ECT": "Ora Ecuadorului", "EDT": "Ora de vară orientală nord-americană", "EST": "Ora standard orientală nord-americană", "GFT": "Ora din Guyana Franceză", "GMT": "Ora de Greenwhich", "GST": "Ora standard a Golfului", "GYT": "Ora din Guyana", "HADT": "Ora de vară din Hawaii-Aleutine", "HAST": "Ora standard din Hawaii-Aleutine", "HAT": "Ora de vară din Newfoundland", "HECU": "Ora de vară a Cubei", "HEEG": "Ora de vară a Groenlandei orientale", "HENOMX": "Ora de vară a Mexicului de nord-vest", "HEOG": "Ora de vară a Groenlandei occidentale", "HEPM": "Ora de vară din Saint-Pierre și Miquelon", "HEPMX": "Ora de vară a zonei Pacific mexicane", "HKST": "Ora de vară din Hong Kong", "HKT": "Ora standard din Hong Kong", "HNCU": "Ora standard a Cubei", "HNEG": "Ora standard a Groenlandei orientale", "HNNOMX": "Ora standard a Mexicului de nord-vest", "HNOG": "Ora standard a Groenlandei occidentale", "HNPM": "Ora standard din Saint-Pierre și Miquelon", "HNPMX": "Ora standard a zonei Pacific mexicane", "HNT": "Ora standard din Newfoundland", "IST": "Ora Indiei", "JDT": "Ora de vară a Japoniei", "JST": "Ora standard a Japoniei", "LHDT": "Ora de vară din Lord Howe", "LHST": "Ora standard din Lord Howe", "MDT": "Ora de vară în zona montană nord-americană", "MESZ": "Ora de vară a Europei Centrale", "MEZ": "Ora standard a Europei Centrale", "MST": "Ora standard în zona montană nord-americană", "MYT": "Ora din Malaysia", "NZDT": "Ora de vară a Noii Zeelande", "NZST": "Ora standard a Noii Zeelande", "OESZ": "Ora de vară a Europei de Est", "OEZ": "Ora standard a Europei de Est", "PDT": "Ora de vară în zona Pacific nord-americană", "PST": "Ora standard în zona Pacific nord-americană", "SAST": "Ora Africii Meridionale", "SGT": "Ora din Singapore", "SRT": "Ora Surinamului", "TMST": "Ora de vară din Turkmenistan", "TMT": "Ora standard din Turkmenistan", "UYST": "Ora de vară a Uruguayului", "UYT": "Ora standard a Uruguayului", "VET": "Ora Venezuelei", "WARST": "Ora de vară a Argentinei Occidentale", "WART": "Ora standard a Argentinei Occidentale", "WAST": "Ora de vară a Africii Occidentale", "WAT": "Ora standard a Africii Occidentale", "WESZ": "Ora de vară a Europei de Vest", "WEZ": "Ora standard a Europei de Vest", "WIB": "Ora Indoneziei de Vest", "WIT": "Ora Indoneziei de Est", "WITA": "Ora Indoneziei Centrale", "∅∅∅": "Ora de vară a Brasiliei"},

			fnCardinalPluralRule:  fnCardinalPluralRule_a60b3f7c6ddcf0678df3df0fe274201a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6db355e32f98d7cf05b06cd4b38d6c63,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ro_ro"] = func() locales.Translator {
		return &localen{

			locale:                 "ro_RO",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "ian.", "feb.", "mar.", "apr.", "mai", "iun.", "iul.", "aug.", "sept.", "oct.", "nov.", "dec."},
			monthsNarrow:           []string{"", "I", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "ianuarie", "februarie", "martie", "aprilie", "mai", "iunie", "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie"},
			daysAbbreviated:        []string{"dum.", "lun.", "mar.", "mie.", "joi", "vin.", "sâm."},
			daysNarrow:             []string{"D", "L", "M", "M", "J", "V", "S"},
			daysShort:              []string{"du.", "lu.", "ma.", "mi.", "joi", "vi.", "sâ."},
			daysWide:               []string{"duminică", "luni", "marți", "miercuri", "joi", "vineri", "sâmbătă"},
			periodsAbbreviated:     []string{"a.m.", "p.m."},
			periodsNarrow:          []string{"a.m.", "p.m."},
			periodsWide:            []string{"a.m.", "p.m."},
			erasAbbreviated:        []string{"î.Hr.", "d.Hr."},
			erasNarrow:             []string{"î.Hr.", "d.Hr."},
			erasWide:               []string{"înainte de Hristos", "după Hristos"},
			timezones:              map[string]string{"ACDT": "Ora de vară a Australiei Centrale", "ACST": "Ora standard a Australiei Centrale", "ACWDT": "Ora de vară a Australiei Central Occidentale", "ACWST": "Ora standard a Australiei Central Occidentale", "ADT": "Ora de vară în zona Atlantic nord-americană", "AEDT": "Ora de vară a Australiei Orientale", "AEST": "Ora standard a Australiei Orientale", "AKDT": "Ora de vară din Alaska", "AKST": "Ora standard din Alaska", "ARST": "Ora de vară a Argentinei", "ART": "Ora standard a Argentinei", "AST": "Ora standard în zona Atlantic nord-americană", "AWDT": "Ora de vară a Australiei Occidentale", "AWST": "Ora standard a Australiei Occidentale", "BOT": "Ora Boliviei", "BT": "Ora Bhutanului", "CAT": "Ora Africii Centrale", "CDT": "Ora de vară centrală nord-americană", "CHADT": "Ora de vară din Chatham", "CHAST": "Ora standard din Chatham", "CLST": "Ora de vară din Chile", "CLT": "Ora standard din Chile", "COST": "Ora de vară a Columbiei", "COT": "Ora standard a Columbiei", "CST": "Ora standard centrală nord-americană", "ChST": "Ora din Chamorro", "EAT": "Ora Africii Orientale", "ECT": "Ora Ecuadorului", "EDT": "Ora de vară orientală nord-americană", "EST": "Ora standard orientală nord-americană", "GFT": "Ora din Guyana Franceză", "GMT": "Ora de Greenwhich", "GST": "Ora standard a Golfului", "GYT": "Ora din Guyana", "HADT": "Ora de vară din Hawaii-Aleutine", "HAST": "Ora standard din Hawaii-Aleutine", "HAT": "Ora de vară din Newfoundland", "HECU": "Ora de vară a Cubei", "HEEG": "Ora de vară a Groenlandei orientale", "HENOMX": "Ora de vară a Mexicului de nord-vest", "HEOG": "Ora de vară a Groenlandei occidentale", "HEPM": "Ora de vară din Saint-Pierre și Miquelon", "HEPMX": "Ora de vară a zonei Pacific mexicane", "HKST": "Ora de vară din Hong Kong", "HKT": "Ora standard din Hong Kong", "HNCU": "Ora standard a Cubei", "HNEG": "Ora standard a Groenlandei orientale", "HNNOMX": "Ora standard a Mexicului de nord-vest", "HNOG": "Ora standard a Groenlandei occidentale", "HNPM": "Ora standard din Saint-Pierre și Miquelon", "HNPMX": "Ora standard a zonei Pacific mexicane", "HNT": "Ora standard din Newfoundland", "IST": "Ora Indiei", "JDT": "Ora de vară a Japoniei", "JST": "Ora standard a Japoniei", "LHDT": "Ora de vară din Lord Howe", "LHST": "Ora standard din Lord Howe", "MDT": "Ora de vară în zona montană nord-americană", "MESZ": "Ora de vară a Europei Centrale", "MEZ": "Ora standard a Europei Centrale", "MST": "Ora standard în zona montană nord-americană", "MYT": "Ora din Malaysia", "NZDT": "Ora de vară a Noii Zeelande", "NZST": "Ora standard a Noii Zeelande", "OESZ": "Ora de vară a Europei de Est", "OEZ": "Ora standard a Europei de Est", "PDT": "Ora de vară în zona Pacific nord-americană", "PST": "Ora standard în zona Pacific nord-americană", "SAST": "Ora Africii Meridionale", "SGT": "Ora din Singapore", "SRT": "Ora Surinamului", "TMST": "Ora de vară din Turkmenistan", "TMT": "Ora standard din Turkmenistan", "UYST": "Ora de vară a Uruguayului", "UYT": "Ora standard a Uruguayului", "VET": "Ora Venezuelei", "WARST": "Ora de vară a Argentinei Occidentale", "WART": "Ora standard a Argentinei Occidentale", "WAST": "Ora de vară a Africii Occidentale", "WAT": "Ora standard a Africii Occidentale", "WESZ": "Ora de vară a Europei de Vest", "WEZ": "Ora standard a Europei de Vest", "WIB": "Ora Indoneziei de Vest", "WIT": "Ora Indoneziei de Est", "WITA": "Ora Indoneziei Centrale", "∅∅∅": "Ora de vară din Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_a60b3f7c6ddcf0678df3df0fe274201a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_6db355e32f98d7cf05b06cd4b38d6c63,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rof"] = func() locales.Translator {
		return &localen{

			locale:             "rof",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "M1", "M2", "M3", "M4", "M5", "M6", "M7", "M8", "M9", "M10", "M11", "M12"},
			monthsNarrow:       []string{"", "K", "K", "K", "K", "T", "S", "S", "N", "T", "I", "I", "I"},
			monthsWide:         []string{"", "Mweri wa kwanza", "Mweri wa kaili", "Mweri wa katatu", "Mweri wa kaana", "Mweri wa tanu", "Mweri wa sita", "Mweri wa saba", "Mweri wa nane", "Mweri wa tisa", "Mweri wa ikumi", "Mweri wa ikumi na moja", "Mweri wa ikumi na mbili"},
			daysAbbreviated:    []string{"Ijp", "Ijt", "Ijn", "Ijtn", "Alh", "Iju", "Ijm"},
			daysNarrow:         []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:           []string{"Ijumapili", "Ijumatatu", "Ijumanne", "Ijumatano", "Alhamisi", "Ijumaa", "Ijumamosi"},
			periodsAbbreviated: []string{"kang’ama", "kingoto"},
			periodsWide:        []string{"kang’ama", "kingoto"},
			erasAbbreviated:    []string{"KM", "BM"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Mayesu", "Baada ya Mayesu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rof_tz"] = func() locales.Translator {
		return &localen{

			locale:             "rof_TZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "M1", "M2", "M3", "M4", "M5", "M6", "M7", "M8", "M9", "M10", "M11", "M12"},
			monthsNarrow:       []string{"", "K", "K", "K", "K", "T", "S", "S", "N", "T", "I", "I", "I"},
			monthsWide:         []string{"", "Mweri wa kwanza", "Mweri wa kaili", "Mweri wa katatu", "Mweri wa kaana", "Mweri wa tanu", "Mweri wa sita", "Mweri wa saba", "Mweri wa nane", "Mweri wa tisa", "Mweri wa ikumi", "Mweri wa ikumi na moja", "Mweri wa ikumi na mbili"},
			daysAbbreviated:    []string{"Ijp", "Ijt", "Ijn", "Ijtn", "Alh", "Iju", "Ijm"},
			daysNarrow:         []string{"2", "3", "4", "5", "6", "7", "1"},
			daysWide:           []string{"Ijumapili", "Ijumatatu", "Ijumanne", "Ijumatano", "Alhamisi", "Ijumaa", "Ijumamosi"},
			periodsAbbreviated: []string{"kang’ama", "kingoto"},
			periodsWide:        []string{"kang’ama", "kingoto"},
			erasAbbreviated:    []string{"KM", "BM"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Mayesu", "Baada ya Mayesu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["root"] = func() locales.Translator {
		return &localen{

			locale:             "root",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "Kz", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "KM", "BAN", "$", "৳", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "$", "$", "Bs", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "$", "BTN", "BUK", "P", "BYB", "р.", "BYR", "$", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "$", "CNH", "CNX", "CN¥", "$", "COU", "₡", "CSD", "CSK", "$", "$", "CVE", "CYP", "Kč", "DDM", "DEM", "DJF", "kr", "$", "DZD", "ECS", "ECV", "EEK", "E£", "ERN", "ESA", "ESB", "₧", "ETB", "€", "FIM", "$", "£", "FRF", "£", "GEK", "₾", "GHC", "GHS", "£", "GMD", "FG", "GNS", "GQE", "GRD", "Q", "GWE", "GWP", "$", "HK$", "L", "HRD", "kn", "HTG", "Ft", "Rp", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "kr", "ITL", "$", "JOD", "JP¥", "KES", "KGS", "៛", "CF", "₩", "KRH", "KRO", "₩", "KWD", "$", "₸", "₭", "L£", "Rs", "$", "LSL", "Lt", "LTT", "LUC", "LUF", "LUL", "Ls", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "Ar", "MGF", "MKD", "MKN", "MLF", "K", "₮", "MOP", "MRO", "MRU", "MTL", "MTP", "Rs", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "$", "₦", "NIC", "C$", "NLG", "kr", "Rs", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "₱", "Rs", "zł", "PLZ", "PTE", "₲", "QAR", "RHD", "ROL", "lei", "RSD", "₽", "р.", "RF", "SAR", "$", "SCR", "SDD", "SDG", "SDP", "kr", "$", "£", "SIT", "SKK", "SLL", "SOS", "$", "SRG", "£", "STD", "Db", "SUR", "SVC", "£", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "T$", "TPE", "TRL", "₺", "$", "NT$", "TZS", "₴", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "$", "UYW", "UZS", "VEB", "Bs", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "¤", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZK", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "M01", "M02", "M03", "M04", "M05", "M06", "M07", "M08", "M09", "M10", "M11", "M12"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:           []string{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"", ""},
			periodsWide:        []string{"", ""},
			erasAbbreviated:    []string{"BCE", "CE"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_411c735c7aafad814d0db05f621af17c,
			fnFmtCurrency:         fnFmtCurrency_8bfebaacbf71a9a4bab2a3bb99e52529,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_fe5c5e0e26e170f84f3661fae0335729,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru"] = func() locales.Translator {
		return &localen{

			locale:                 "ru",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "р.", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "ТМТ", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "₴", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Бразилия, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru_by"] = func() locales.Translator {
		return &localen{

			locale:                 "ru_BY",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "Br", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Бразилия, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru_kg"] = func() locales.Translator {
		return &localen{

			locale:                 "ru_KG",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "сом", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Азорские о-ва, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru_kz"] = func() locales.Translator {
		return &localen{

			locale:                 "ru_KZ",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "₸", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Азорские о-ва, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru_md"] = func() locales.Translator {
		return &localen{

			locale:                 "ru_MD",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "L", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Азорские о-ва, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru_ru"] = func() locales.Translator {
		return &localen{

			locale:                 "ru_RU",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Бразилия, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ru_ua"] = func() locales.Translator {
		return &localen{

			locale:                 "ru_UA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв.", "февр.", "мар.", "апр.", "мая", "июн.", "июл.", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"},
			daysAbbreviated:        []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysShort:              []string{"вс", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"воскресенье", "понедельник", "вторник", "среда", "четверг", "пятница", "суббота"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"до н. э.", "н. э."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"до Рождества Христова", "от Рождества Христова"},
			timezones:              map[string]string{"ACDT": "Центральная Австралия, летнее время", "ACST": "Центральная Австралия, стандартное время", "ACWDT": "Центральная Австралия, западное летнее время", "ACWST": "Центральная Австралия, западное стандартное время", "ADT": "Атлантическое летнее время", "AEDT": "Восточная Австралия, летнее время", "AEST": "Восточная Австралия, стандартное время", "AKDT": "Аляска, летнее время", "AKST": "Аляска, стандартное время", "ARST": "Аргентина, летнее время", "ART": "Аргентина, стандартное время", "AST": "Атлантическое стандартное время", "AWDT": "Западная Австралия, летнее время", "AWST": "Западная Австралия, стандартное время", "BOT": "Боливия", "BT": "Бутан", "CAT": "Центральная Африка", "CDT": "Центральная Америка, летнее время", "CHADT": "Чатем, летнее время", "CHAST": "Чатем, стандартное время", "CLST": "Чили, летнее время", "CLT": "Чили, стандартное время", "COST": "Колумбия, летнее время", "COT": "Колумбия, стандартное время", "CST": "Центральная Америка, стандартное время", "ChST": "Чаморро", "EAT": "Восточная Африка", "ECT": "Эквадор", "EDT": "Восточная Америка, летнее время", "EST": "Восточная Америка, стандартное время", "GFT": "Французская Гвиана", "GMT": "Среднее время по Гринвичу", "GST": "Персидский залив", "GYT": "Гайана", "HADT": "Гавайско-алеутское летнее время", "HAST": "Гавайско-алеутское стандартное время", "HAT": "Ньюфаундленд, летнее время", "HECU": "Куба, летнее время", "HEEG": "Восточная Гренландия, летнее время", "HENOMX": "Северо-западное мексиканское летнее время", "HEOG": "Западная Гренландия, летнее время", "HEPM": "Сен-Пьер и Микелон, летнее время", "HEPMX": "Тихоокеанское мексиканское летнее время", "HKST": "Гонконг, летнее время", "HKT": "Гонконг, стандартное время", "HNCU": "Куба, стандартное время", "HNEG": "Восточная Гренландия, стандарное время", "HNNOMX": "Северо-западное мексиканское стандартное время", "HNOG": "Западная Гренландия, стандартное время", "HNPM": "Сен-Пьер и Микелон, стандартное время", "HNPMX": "Тихоокеанское мексиканское стандартное время", "HNT": "Ньюфаундленд, стандартное время", "IST": "Индия", "JDT": "Япония, летнее время", "JST": "Япония, стандартное время", "LHDT": "Лорд-Хау, летнее время", "LHST": "Лорд-Хау, стандартное время", "MDT": "Летнее горное время (Северная Америка)", "MESZ": "Центральная Европа, летнее время", "MEZ": "Центральная Европа, стандартное время", "MST": "Стандартное горное время (Северная Америка)", "MYT": "Малайзия", "NZDT": "Новая Зеландия, летнее время", "NZST": "Новая Зеландия, стандартное время", "OESZ": "Восточная Европа, летнее время", "OEZ": "Восточная Европа, стандартное время", "PDT": "Тихоокеанское летнее время", "PST": "Тихоокеанское стандартное время", "SAST": "Южная Африка", "SGT": "Сингапур", "SRT": "Суринам", "TMST": "Туркменистан, летнее время", "TMT": "Туркменистан, стандартное время", "UYST": "Уругвай, летнее время", "UYT": "Уругвай, стандартное время", "VET": "Венесуэла", "WARST": "Западная Аргентина, летнее время", "WART": "Западная Аргентина, стандартное время", "WAST": "Западная Африка, летнее время", "WAT": "Западная Африка, стандартное время", "WESZ": "Западная Европа, летнее время", "WEZ": "Западная Европа, стандартное время", "WIB": "Западная Индонезия", "WIT": "Восточная Индонезия", "WITA": "Центральная Индонезия", "∅∅∅": "Азорские о-ва, летнее время"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_3bb9a5fb6d982f44bb197fafb3f68087,
			fnFmtDateLong:         fnFmtDateLong_b673377c923e43f185728d5b63716bee,
			fnFmtDateMedium:       fnFmtDateMedium_5b31a18861b34639769a2dc612a70a3b,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rw"] = func() locales.Translator {
		return &localen{

			locale:                 "rw",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "mut.", "gas.", "wer.", "mat.", "gic.", "kam.", "nya.", "kan.", "nze.", "ukw.", "ugu.", "uku."},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Mutarama", "Gashyantare", "Werurwe", "Mata", "Gicuransi", "Kamena", "Nyakanga", "Kanama", "Nzeli", "Ukwakira", "Ugushyingo", "Ukuboza"},
			daysAbbreviated:        []string{"cyu.", "mbe.", "kab.", "gtu.", "kan.", "gnu.", "gnd."},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"cyu.", "mbe.", "kab.", "gtu.", "kan.", "gnu.", "gnd."},
			daysWide:               []string{"Ku cyumweru", "Kuwa mbere", "Kuwa kabiri", "Kuwa gatatu", "Kuwa kane", "Kuwa gatanu", "Kuwa gatandatu"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "CE"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rw_rw"] = func() locales.Translator {
		return &localen{

			locale:                 "rw_RW",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "mut.", "gas.", "wer.", "mat.", "gic.", "kam.", "nya.", "kan.", "nze.", "ukw.", "ugu.", "uku."},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Mutarama", "Gashyantare", "Werurwe", "Mata", "Gicuransi", "Kamena", "Nyakanga", "Kanama", "Nzeli", "Ukwakira", "Ugushyingo", "Ukuboza"},
			daysAbbreviated:        []string{"cyu.", "mbe.", "kab.", "gtu.", "kan.", "gnu.", "gnd."},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"cyu.", "mbe.", "kab.", "gtu.", "kan.", "gnu.", "gnd."},
			daysWide:               []string{"Ku cyumweru", "Kuwa mbere", "Kuwa kabiri", "Kuwa gatatu", "Kuwa kane", "Kuwa gatanu", "Kuwa gatandatu"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BCE", "CE"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rwk"] = func() locales.Translator {
		return &localen{

			locale:             "rwk",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprilyi", "Mei", "Junyi", "Julyai", "Agusti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Jumapilyi", "Jumatatuu", "Jumanne", "Jumatanu", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"utuko", "kyiukonyi"},
			periodsWide:        []string{"utuko", "kyiukonyi"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Kristu", "Baada ya Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["rwk_tz"] = func() locales.Translator {
		return &localen{

			locale:             "rwk_TZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprilyi", "Mei", "Junyi", "Julyai", "Agusti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Jumapilyi", "Jumatatuu", "Jumanne", "Jumatanu", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"utuko", "kyiukonyi"},
			periodsWide:        []string{"utuko", "kyiukonyi"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Kristu", "Baada ya Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b4b66bcaf750a9b7b42ab51bff8c0e19,
			fnFmtCurrency:         fnFmtCurrency_022949044632e54ed5ae02ba82626546,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sah"] = func() locales.Translator {
		return &localen{

			locale:                 "sah",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Тохс", "Олун", "Клн", "Мсу", "Ыам", "Бэс", "Отй", "Атр", "Блҕ", "Алт", "Сэт", "Ахс"},
			monthsNarrow:           []string{"", "Т", "О", "К", "М", "Ы", "Б", "О", "А", "Б", "А", "С", "А"},
			monthsWide:             []string{"", "Тохсунньу", "Олунньу", "Кулун тутар", "Муус устар", "Ыам ыйын", "Бэс ыйын", "От ыйын", "Атырдьых ыйын", "Балаҕан ыйын", "Алтынньы", "Сэтинньи", "ахсынньы"},
			daysAbbreviated:        []string{"бс", "бн", "оп", "сэ", "чп", "бэ", "сб"},
			daysNarrow:             []string{"Б", "Б", "О", "С", "Ч", "Б", "С"},
			daysShort:              []string{"бс", "бн", "оп", "сэ", "чп", "бэ", "сб"},
			daysWide:               []string{"баскыһыанньа", "бэнидиэнньик", "оптуорунньук", "сэрэдэ", "чэппиэр", "Бээтиҥсэ", "субуота"},
			periodsAbbreviated:     []string{"ЭИ", "ЭК"},
			periodsNarrow:          []string{"ЭИ", "ЭК"},
			periodsWide:            []string{"ЭИ", "ЭК"},
			erasAbbreviated:        []string{"б. э. и.", "б. э"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"б. э. и.", "б. э"},
			timezones:              map[string]string{"ACDT": "Киин Австралия сайыҥҥы кэмэ", "ACST": "Киин Австралия сүрүн кэмэ", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "Илин Австралия сайыҥҥы кэмэ", "AEST": "Илин Австралия сүрүн кэмэ", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "Арҕаа Австралия сайыҥҥы кэмэ", "AWST": "Арҕаа Австралия сүрүн кэмэ", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "Ииндийэ сүрүн кэмэ", "JDT": "Дьоппуон сайыҥҥы кэмэ", "JST": "Дьоппуон сүрүн кэмэ", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "Саҥа Сэйлэнд сайыҥҥы кэмэ", "NZST": "Саҥа Сэйлэнд сүрүн кэмэ", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_57a372adfbe798d29526efd47816e07a,
			fnFmtDateLong:         fnFmtDateLong_ab2127ffd2f4043ae38d2f998ae282a6,
			fnFmtDateMedium:       fnFmtDateMedium_fdbcfd5a321e6370f61bce682fe65b6e,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sah_ru"] = func() locales.Translator {
		return &localen{

			locale:                 "sah_RU",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Тохс", "Олун", "Клн", "Мсу", "Ыам", "Бэс", "Отй", "Атр", "Блҕ", "Алт", "Сэт", "Ахс"},
			monthsNarrow:           []string{"", "Т", "О", "К", "М", "Ы", "Б", "О", "А", "Б", "А", "С", "А"},
			monthsWide:             []string{"", "Тохсунньу", "Олунньу", "Кулун тутар", "Муус устар", "Ыам ыйын", "Бэс ыйын", "От ыйын", "Атырдьых ыйын", "Балаҕан ыйын", "Алтынньы", "Сэтинньи", "ахсынньы"},
			daysAbbreviated:        []string{"бс", "бн", "оп", "сэ", "чп", "бэ", "сб"},
			daysNarrow:             []string{"Б", "Б", "О", "С", "Ч", "Б", "С"},
			daysShort:              []string{"бс", "бн", "оп", "сэ", "чп", "бэ", "сб"},
			daysWide:               []string{"баскыһыанньа", "бэнидиэнньик", "оптуорунньук", "сэрэдэ", "чэппиэр", "Бээтиҥсэ", "субуота"},
			periodsAbbreviated:     []string{"ЭИ", "ЭК"},
			periodsNarrow:          []string{"ЭИ", "ЭК"},
			periodsWide:            []string{"ЭИ", "ЭК"},
			erasAbbreviated:        []string{"б. э. и.", "б. э"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"б. э. и.", "б. э"},
			timezones:              map[string]string{"ACDT": "Киин Австралия сайыҥҥы кэмэ", "ACST": "Киин Австралия сүрүн кэмэ", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "Илин Австралия сайыҥҥы кэмэ", "AEST": "Илин Австралия сүрүн кэмэ", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "Арҕаа Австралия сайыҥҥы кэмэ", "AWST": "Арҕаа Австралия сүрүн кэмэ", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "Ииндийэ сүрүн кэмэ", "JDT": "Дьоппуон сайыҥҥы кэмэ", "JST": "Дьоппуон сүрүн кэмэ", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "Саҥа Сэйлэнд сайыҥҥы кэмэ", "NZST": "Саҥа Сэйлэнд сүрүн кэмэ", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_57a372adfbe798d29526efd47816e07a,
			fnFmtDateLong:         fnFmtDateLong_ab2127ffd2f4043ae38d2f998ae282a6,
			fnFmtDateMedium:       fnFmtDateMedium_fdbcfd5a321e6370f61bce682fe65b6e,
			fnFmtDateShort:        fnFmtDateShort_28a99d53f2aae59a12f8564b98a20b86,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["saq"] = func() locales.Translator {
		return &localen{

			locale:                 "saq",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Obo", "Waa", "Oku", "Ong", "Ime", "Ile", "Sap", "Isi", "Saa", "Tom", "Tob", "Tow"},
			monthsNarrow:           []string{"", "O", "W", "O", "O", "I", "I", "S", "I", "S", "T", "T", "T"},
			monthsWide:             []string{"", "Lapa le obo", "Lapa le waare", "Lapa le okuni", "Lapa le ong’wan", "Lapa le imet", "Lapa le ile", "Lapa le sapa", "Lapa le isiet", "Lapa le saal", "Lapa le tomon", "Lapa le tomon obo", "Lapa le tomon waare"},
			daysAbbreviated:        []string{"Are", "Kun", "Ong", "Ine", "Ile", "Sap", "Kwe"},
			daysNarrow:             []string{"A", "K", "O", "I", "I", "S", "K"},
			daysWide:               []string{"Mderot ee are", "Mderot ee kuni", "Mderot ee ong’wan", "Mderot ee inet", "Mderot ee ile", "Mderot ee sapa", "Mderot ee kwe"},
			periodsAbbreviated:     []string{"Tesiran", "Teipa"},
			periodsWide:            []string{"Tesiran", "Teipa"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Christo", "Baada ya Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["saq_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "saq_KE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Obo", "Waa", "Oku", "Ong", "Ime", "Ile", "Sap", "Isi", "Saa", "Tom", "Tob", "Tow"},
			monthsNarrow:           []string{"", "O", "W", "O", "O", "I", "I", "S", "I", "S", "T", "T", "T"},
			monthsWide:             []string{"", "Lapa le obo", "Lapa le waare", "Lapa le okuni", "Lapa le ong’wan", "Lapa le imet", "Lapa le ile", "Lapa le sapa", "Lapa le isiet", "Lapa le saal", "Lapa le tomon", "Lapa le tomon obo", "Lapa le tomon waare"},
			daysAbbreviated:        []string{"Are", "Kun", "Ong", "Ine", "Ile", "Sap", "Kwe"},
			daysNarrow:             []string{"A", "K", "O", "I", "I", "S", "K"},
			daysWide:               []string{"Mderot ee are", "Mderot ee kuni", "Mderot ee ong’wan", "Mderot ee inet", "Mderot ee ile", "Mderot ee sapa", "Mderot ee kwe"},
			periodsAbbreviated:     []string{"Tesiran", "Teipa"},
			periodsWide:            []string{"Tesiran", "Teipa"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Christo", "Baada ya Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sbp"] = func() locales.Translator {
		return &localen{

			locale:             "sbp",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Mup", "Mwi", "Msh", "Mun", "Mag", "Muj", "Msp", "Mpg", "Mye", "Mok", "Mus", "Muh"},
			monthsWide:         []string{"", "Mupalangulwa", "Mwitope", "Mushende", "Munyi", "Mushende Magali", "Mujimbi", "Mushipepo", "Mupuguto", "Munyense", "Mokhu", "Musongandembwe", "Muhaano"},
			daysAbbreviated:    []string{"Mul", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"M", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Mulungu", "Jumatatu", "Jumanne", "Jumatano", "Alahamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"Lwamilawu", "Pashamihe"},
			periodsWide:        []string{"Lwamilawu", "Pashamihe"},
			erasAbbreviated:    []string{"AK", "PK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Ashanali uKilisito", "Pamwandi ya Kilisto"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sbp_tz"] = func() locales.Translator {
		return &localen{

			locale:             "sbp_TZ",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Mup", "Mwi", "Msh", "Mun", "Mag", "Muj", "Msp", "Mpg", "Mye", "Mok", "Mus", "Muh"},
			monthsWide:         []string{"", "Mupalangulwa", "Mwitope", "Mushende", "Munyi", "Mushende Magali", "Mujimbi", "Mushipepo", "Mupuguto", "Munyense", "Mokhu", "Musongandembwe", "Muhaano"},
			daysAbbreviated:    []string{"Mul", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"M", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Mulungu", "Jumatatu", "Jumanne", "Jumatano", "Alahamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"Lwamilawu", "Pashamihe"},
			periodsWide:        []string{"Lwamilawu", "Pashamihe"},
			erasAbbreviated:    []string{"AK", "PK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Ashanali uKilisito", "Pamwandi ya Kilisto"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sd"] = func() locales.Translator {
		return &localen{

			locale:                 "sd",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "Rs", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جنوري", "فيبروري", "مارچ", "اپريل", "مئي", "جون", "جولاءِ", "آگسٽ", "سيپٽمبر", "آڪٽوبر", "نومبر", "ڊسمبر"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "جنوري", "فيبروري", "مارچ", "اپريل", "مئي", "جون", "جولاءِ", "آگسٽ", "سيپٽمبر", "آڪٽوبر", "نومبر", "ڊسمبر"},
			daysAbbreviated:        []string{"آچر", "سومر", "اڱارو", "اربع", "خميس", "جمعو", "ڇنڇر"},
			daysNarrow:             []string{"آچر", "سو", "اڱارو", "اربع", "خم", "جمعو", "ڇنڇر"},
			daysShort:              []string{"آچر", "سومر", "اڱارو", "اربع", "خميس", "جمعو", "ڇنڇر"},
			daysWide:               []string{"آچر", "سومر", "اڱارو", "اربع", "خميس", "جمعو", "ڇنڇر"},
			periodsAbbreviated:     []string{"صبح، منجهند", "شام، منجهند"},
			periodsNarrow:          []string{"صبح، منجهند", "منجهند، شام"},
			periodsWide:            []string{"صبح، منجهند", "منجهند، شام"},
			erasAbbreviated:        []string{"BC", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"مسيح کان اڳ", "عيسوي کان پهرين"},
			timezones:              map[string]string{"ACDT": "آسٽريليا جو مرڪزي ڏينهن جو وقت", "ACST": "آسٽريليا جو مرڪزي معياري وقت", "ACWDT": "آسٽريليا جو مرڪزي مغربي ڏينهن جو وقت", "ACWST": "آسٽريليا جو مرڪزي مغربي معياري وقت", "ADT": "ايٽلانٽڪ جي ڏينهن جو وقت", "AEDT": "آسٽريليا جو مشرقي ڏينهن جو وقت", "AEST": "آسٽريليا جو مشرقي معياري وقت", "AKDT": "الاسڪا جي ڏينهن جو وقت", "AKST": "الاسڪا جو معياري وقت", "ARST": "ارجنٽائن جي اونهاري جو وقت", "ART": "ارجنٽائن معياري وقت", "AST": "ايٽلانٽڪ جو معياري وقت", "AWDT": "آسٽريليا جو مغربي ڏينهن جو وقت", "AWST": "آسٽريليا جو مغربي معياري وقت", "BOT": "بولويائي وقت", "BT": "ڀوٽان جو وقت", "CAT": "مرڪزي آفريقا جو وقت", "CDT": "مرڪزي ڏينهن جو وقت", "CHADT": "چئٿم جي ڏينهن جو وقت", "CHAST": "چئٿم جو معياري وقت", "CLST": "چلي جي اونهاري جو وقت", "CLT": "چلي جو معياري وقت", "COST": "ڪولمبيا جي اونهاري جو وقت", "COT": "ڪولمبيا جو معياري وقت", "CST": "مرڪزي معياري وقت", "ChST": "چمورو جو معياري وقت", "EAT": "اوڀر آفريڪا جو وقت", "ECT": "ايڪواڊور جو وقت", "EDT": "مشرقي ڏينهن جو وقت", "EST": "مشرقي معياري وقت", "GFT": "فرانسيسي گيانا جو وقت", "GMT": "گرين وچ مين ٽائيم", "GST": "خلج معياري وقت", "GYT": "گيانائي وقت", "HADT": "هوائي اليوٽين جي ڏينهن جو وقت", "HAST": "هوائي اليوٽين جو معياري وقت", "HAT": "نيو فائونڊ لينڊ جي ڏينهن جو وقت", "HECU": "ڪيوبا جي ڏينهن جو وقت", "HEEG": "مشرقي گرين لينڊ جي اونهاري جو وقت", "HENOMX": "شمالي مغربي ميڪسيڪو جي ڏينهن جو وقت", "HEOG": "مغربي گرين لينڊ جي اونهاري جو وقت", "HEPM": "سینٽ پیئر و میڪوئیلون جي ڏينهن جو وقت", "HEPMX": "ميڪسيڪن پيسيفڪ جي ڏينهن جو وقت", "HKST": "هانگ ڪانگ جي اونهاري جو وقت", "HKT": "هانگ ڪانگ جو معياري وقت", "HNCU": "ڪيوبا جو معياري وقت", "HNEG": "مشرقي گرين لينڊ جو معياري وقت", "HNNOMX": "شمالي مغربي ميڪسيڪو جو معياري وقت", "HNOG": "مغربي گرين لينڊ جو معياري وقت", "HNPM": "سینٽ پیئر اَئن میڪوئلون جو مانائتو وقت", "HNPMX": "ميڪسيڪن پيسيفڪ جو معياري وقت", "HNT": "نيو فائونڊ لينڊ جو معياري وقت", "IST": "ڀارت جو معياري وقت", "JDT": "جاپان جي ڏينهن جو وقت", "JST": "جاپان جو معياري وقت", "LHDT": "لورڊ هووي جي ڏينهن جو وقت", "LHST": "لورڊ هووي جو معياري وقت", "MDT": "پهاڙي ڏينهن جو وقت", "MESZ": "مرڪزي يورپي اونهاري جو وقت", "MEZ": "مرڪزي يورپي معياري وقت", "MST": "پهاڙي معياري وقت", "MYT": "ملائيشيا جو وقت", "NZDT": "نيوزيلينڊ جي ڏينهن جو وقت", "NZST": "نيوزيلينڊ جو معياري وقت", "OESZ": "مشرقي يورپي اونهاري جو وقت", "OEZ": "مشرقي يورپي معياري وقت", "PDT": "پيسيفڪ ڏينهن جو وقت", "PST": "پيسيفڪ معياري وقت", "SAST": "ڏکڻ آفريڪا جو معياري وقت", "SGT": "سنگاپور جو معياري وقت", "SRT": "سوري نام جو وقت", "TMST": "ترڪمانستان جي اونهاري جو وقت", "TMT": "ترڪمانستان جو معياري وقت", "UYST": "يوروگائي جي اونهاري جو وقت", "UYT": "يوروگائي جو معياري وقت", "VET": "وينزويلا جو وقت", "WARST": "مغربي ارجنٽائن جي اونهاري جو وقت", "WART": "مغربي ارجنٽائن جو معياري وقت", "WAST": "اولهه آفريقا جي اونهاري جو وقت", "WAT": "اولهه آفريقا جو معياري وقت", "WESZ": "مغربي يورپي ڏينهن جو وقت", "WEZ": "مغربي يورپي معياري وقت", "WIB": "اولهه انڊونيشيا جو وقت", "WIT": "اوڀر انڊونيشيا جو وقت", "WITA": "مرڪزي انڊونيشيا جو وقت", "∅∅∅": "ازورز جي اونهاري جو وقت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sd_pk"] = func() locales.Translator {
		return &localen{

			locale:                 "sd_PK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جنوري", "فيبروري", "مارچ", "اپريل", "مئي", "جون", "جولاءِ", "آگسٽ", "سيپٽمبر", "آڪٽوبر", "نومبر", "ڊسمبر"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "جنوري", "فيبروري", "مارچ", "اپريل", "مئي", "جون", "جولاءِ", "آگسٽ", "سيپٽمبر", "آڪٽوبر", "نومبر", "ڊسمبر"},
			daysAbbreviated:        []string{"آچر", "سومر", "اڱارو", "اربع", "خميس", "جمعو", "ڇنڇر"},
			daysNarrow:             []string{"آچر", "سو", "اڱارو", "اربع", "خم", "جمعو", "ڇنڇر"},
			daysShort:              []string{"آچر", "سومر", "اڱارو", "اربع", "خميس", "جمعو", "ڇنڇر"},
			daysWide:               []string{"آچر", "سومر", "اڱارو", "اربع", "خميس", "جمعو", "ڇنڇر"},
			periodsAbbreviated:     []string{"صبح، منجهند", "شام، منجهند"},
			periodsNarrow:          []string{"صبح، منجهند", "منجهند، شام"},
			periodsWide:            []string{"صبح، منجهند", "منجهند، شام"},
			erasAbbreviated:        []string{"BC", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"مسيح کان اڳ", "عيسوي کان پهرين"},
			timezones:              map[string]string{"ACDT": "آسٽريليا جو مرڪزي ڏينهن جو وقت", "ACST": "آسٽريليا جو مرڪزي معياري وقت", "ACWDT": "آسٽريليا جو مرڪزي مغربي ڏينهن جو وقت", "ACWST": "آسٽريليا جو مرڪزي مغربي معياري وقت", "ADT": "ايٽلانٽڪ جي ڏينهن جو وقت", "AEDT": "آسٽريليا جو مشرقي ڏينهن جو وقت", "AEST": "آسٽريليا جو مشرقي معياري وقت", "AKDT": "الاسڪا جي ڏينهن جو وقت", "AKST": "الاسڪا جو معياري وقت", "ARST": "ارجنٽائن جي اونهاري جو وقت", "ART": "ارجنٽائن معياري وقت", "AST": "ايٽلانٽڪ جو معياري وقت", "AWDT": "آسٽريليا جو مغربي ڏينهن جو وقت", "AWST": "آسٽريليا جو مغربي معياري وقت", "BOT": "بولويائي وقت", "BT": "ڀوٽان جو وقت", "CAT": "مرڪزي آفريقا جو وقت", "CDT": "مرڪزي ڏينهن جو وقت", "CHADT": "چئٿم جي ڏينهن جو وقت", "CHAST": "چئٿم جو معياري وقت", "CLST": "چلي جي اونهاري جو وقت", "CLT": "چلي جو معياري وقت", "COST": "ڪولمبيا جي اونهاري جو وقت", "COT": "ڪولمبيا جو معياري وقت", "CST": "مرڪزي معياري وقت", "ChST": "چمورو جو معياري وقت", "EAT": "اوڀر آفريڪا جو وقت", "ECT": "ايڪواڊور جو وقت", "EDT": "مشرقي ڏينهن جو وقت", "EST": "مشرقي معياري وقت", "GFT": "فرانسيسي گيانا جو وقت", "GMT": "گرين وچ مين ٽائيم", "GST": "خلج معياري وقت", "GYT": "گيانائي وقت", "HADT": "هوائي اليوٽين جي ڏينهن جو وقت", "HAST": "هوائي اليوٽين جو معياري وقت", "HAT": "نيو فائونڊ لينڊ جي ڏينهن جو وقت", "HECU": "ڪيوبا جي ڏينهن جو وقت", "HEEG": "مشرقي گرين لينڊ جي اونهاري جو وقت", "HENOMX": "شمالي مغربي ميڪسيڪو جي ڏينهن جو وقت", "HEOG": "مغربي گرين لينڊ جي اونهاري جو وقت", "HEPM": "سینٽ پیئر و میڪوئیلون جي ڏينهن جو وقت", "HEPMX": "ميڪسيڪن پيسيفڪ جي ڏينهن جو وقت", "HKST": "هانگ ڪانگ جي اونهاري جو وقت", "HKT": "هانگ ڪانگ جو معياري وقت", "HNCU": "ڪيوبا جو معياري وقت", "HNEG": "مشرقي گرين لينڊ جو معياري وقت", "HNNOMX": "شمالي مغربي ميڪسيڪو جو معياري وقت", "HNOG": "مغربي گرين لينڊ جو معياري وقت", "HNPM": "سینٽ پیئر اَئن میڪوئلون جو مانائتو وقت", "HNPMX": "ميڪسيڪن پيسيفڪ جو معياري وقت", "HNT": "نيو فائونڊ لينڊ جو معياري وقت", "IST": "ڀارت جو معياري وقت", "JDT": "جاپان جي ڏينهن جو وقت", "JST": "جاپان جو معياري وقت", "LHDT": "لورڊ هووي جي ڏينهن جو وقت", "LHST": "لورڊ هووي جو معياري وقت", "MDT": "پهاڙي ڏينهن جو وقت", "MESZ": "مرڪزي يورپي اونهاري جو وقت", "MEZ": "مرڪزي يورپي معياري وقت", "MST": "پهاڙي معياري وقت", "MYT": "ملائيشيا جو وقت", "NZDT": "نيوزيلينڊ جي ڏينهن جو وقت", "NZST": "نيوزيلينڊ جو معياري وقت", "OESZ": "مشرقي يورپي اونهاري جو وقت", "OEZ": "مشرقي يورپي معياري وقت", "PDT": "پيسيفڪ ڏينهن جو وقت", "PST": "پيسيفڪ معياري وقت", "SAST": "ڏکڻ آفريڪا جو معياري وقت", "SGT": "سنگاپور جو معياري وقت", "SRT": "سوري نام جو وقت", "TMST": "ترڪمانستان جي اونهاري جو وقت", "TMT": "ترڪمانستان جو معياري وقت", "UYST": "يوروگائي جي اونهاري جو وقت", "UYT": "يوروگائي جو معياري وقت", "VET": "وينزويلا جو وقت", "WARST": "مغربي ارجنٽائن جي اونهاري جو وقت", "WART": "مغربي ارجنٽائن جو معياري وقت", "WAST": "اولهه آفريقا جي اونهاري جو وقت", "WAT": "اولهه آفريقا جو معياري وقت", "WESZ": "مغربي يورپي ڏينهن جو وقت", "WEZ": "مغربي يورپي معياري وقت", "WIB": "اولهه انڊونيشيا جو وقت", "WIT": "اوڀر انڊونيشيا جو وقت", "WITA": "مرڪزي انڊونيشيا جو وقت", "∅∅∅": "ازورز جي اونهاري جو وقت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_cb15430758f15901fa8a71a0aedc08f8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["se"] = func() locales.Translator {
		return &localen{

			locale:                 "se",
			pluralsCardinal:        []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "Dkr", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GB£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "Ikr", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "kr", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "Skr", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ođđj", "guov", "njuk", "cuo", "mies", "geas", "suoi", "borg", "čakč", "golg", "skáb", "juov"},
			monthsNarrow:           []string{"", "O", "G", "N", "C", "M", "G", "S", "B", "Č", "G", "S", "J"},
			monthsWide:             []string{"", "ođđajagemánnu", "guovvamánnu", "njukčamánnu", "cuoŋománnu", "miessemánnu", "geassemánnu", "suoidnemánnu", "borgemánnu", "čakčamánnu", "golggotmánnu", "skábmamánnu", "juovlamánnu"},
			daysAbbreviated:        []string{"sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"},
			daysNarrow:             []string{"S", "V", "M", "G", "D", "B", "L"},
			daysShort:              []string{"sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"},
			daysWide:               []string{"sotnabeaivi", "vuossárga", "maŋŋebárga", "gaskavahkku", "duorasdat", "bearjadat", "lávvardat"},
			periodsAbbreviated:     []string{"i.b.", "e.b."},
			periodsNarrow:          []string{"i.b.", "e.b."},
			periodsWide:            []string{"iđitbeaivet", "eahketbeaivet"},
			erasAbbreviated:        []string{"o.Kr.", "m.Kr."},
			erasNarrow:             []string{"ooá", "oá"},
			erasWide:               []string{"ovdal Kristtusa", "maŋŋel Kristtusa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Greenwich gaskka áigi", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "gaska-Eurohpá geassiáigi", "MEZ": "gaska-Eurohpá dábálašáigi", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "nuorti-Eurohpá geassiáigi", "OEZ": "nuorti-Eurohpá dábálašáigi", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "oarje-Eurohpá geassiáigi", "WEZ": "oarje-Eurohpá dábálašáigi", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["se_fi"] = func() locales.Translator {
		return &localen{

			locale:                 "se_FI",
			pluralsCardinal:        []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "cuoŋ"},
			monthsNarrow:           []string{"", "O", "G", "N", "C", "M", "G", "S", "B", "Č", "G", "S", "J"},
			monthsWide:             []string{"", "ođđajagemánnu", "guovvamánnu", "njukčamánnu", "cuoŋománnu", "miessemánnu", "geassemánnu", "suoidnemánnu", "borgemánnu", "čakčamánnu", "golggotmánnu", "skábmamánnu", "juovlamánnu"},
			daysAbbreviated:        []string{"so", "má", "di", "ga", "du", "be", "lá"},
			daysNarrow:             []string{"M", "D"},
			daysShort:              []string{"so", "má", "di", "ga", "du", "be", "lá"},
			daysWide:               []string{"mánnodat", "disdat", "gaskavahkku", "duorastat", "bearjadat", "lávvordat"},
			periodsAbbreviated:     []string{"ib", "eb"},
			periodsNarrow:          []string{"i", "e"},
			periodsWide:            []string{"ib", "eb"},
			erasAbbreviated:        []string{"oKr.", "mKr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ovdal Kristusa", "maŋŋel Kristusa"},
			timezones:              map[string]string{"ACDT": "Gaska-Austrália geasseáigi", "ACST": "Gaska-Austrália dálveáigi", "ACWDT": "Gaska-Austrália oarjjabeali geasseáigi", "ACWST": "Gaska-Austrália oarjjabeali dálveáigi", "ADT": "atlántalaš geasseáigi", "AEDT": "Nuorta-Austrália geasseáigi", "AEST": "Nuorta-Austrália dálveáigi", "AKDT": "Alaska geasseáigi", "AKST": "Alaska dálveáigi", "ARST": "Argentina geasseáigi", "ART": "Argentina dálveáigi", "AST": "atlántalaš dálveáigi", "AWDT": "Oarje-Austrália geasseáigi", "AWST": "Oarje-Austrália dálveáigi", "BOT": "Bolivia áigi", "BT": "Bhutana áigi", "CAT": "Gaska-Afrihká áigi", "CDT": "dábálaš geasseáigi", "CHADT": "Chathama geasseáigi", "CHAST": "Chathama dálveáigi", "CLST": "Chile geasseáigi", "CLT": "Chile dálveáigi", "COST": "Colombia geasseáigi", "COT": "Colombia dálveáigi", "CST": "dábálaš dálveáigi", "ChST": "Čamorro dálveáigi", "EAT": "Nuorta-Afrihká áigi", "ECT": "Ecuadora áigi", "EDT": "geasseáigi nuortan", "EST": "dálveáigi nuortan", "GFT": "Frankriikka Guyana áigi", "GMT": "Greenwicha áigi", "GST": "Golfa dálveáigi", "GYT": "Guyana áigi", "HADT": "Hawaii-aleuhtalaš geasseáigi", "HAST": "Hawaii-aleuhtalaš dálveáigi", "HAT": "Newfoundlanda geasseáigi", "HECU": "Cuba geasseáigi", "HEEG": "Nuorta-Ruonáeatnama geasseáigi", "HENOMX": "Oarjedavvi-Meksiko geasseáigi", "HEOG": "Oarje-Ruonáeatnama geasseáigi", "HEPM": "St. Pierre & Miquelo geasseáigi", "HEPMX": "Meksiko Jáskesábi geasseáigi", "HKST": "Hong Konga geasseáigi", "HKT": "Hong Konga dálveáigi", "HNCU": "Cuba dálveáigi", "HNEG": "Nuorta-Ruonáeatnama dálveáigi", "HNNOMX": "Oarjedavvi-Meksiko dálveáigi", "HNOG": "Oarje-Ruonáeatnama dálveáigi", "HNPM": "St. Pierre & Miquelo dálveáigi", "HNPMX": "Meksiko Jáskesábi dálveáigi", "HNT": "Newfoundlanda dálveáigi", "IST": "India dálveáigi", "JDT": "Japána geasseáigi", "JST": "Japána dálveáigi", "LHDT": "Lord Howe geasseáigi", "LHST": "Lord Howe dálveáigi", "MDT": "geasseduottaráigi", "MESZ": "Gaska-Eurohpá geasseáigi", "MEZ": "Gaska-Eurohpá dálveáigi", "MST": "dálveduottaráigi", "MYT": "Malesia áigi", "NZDT": "Ođđa-Selánda geasseáigi", "NZST": "Ođđa-Selánda dálveáigi", "OESZ": "Nuorta-Eurohpa geasseáigi", "OEZ": "Nuorta-Eurohpa dálveáigi", "PDT": "Jaskesábi geasseáigi", "PST": "Jaskesábi dálveáigi", "SAST": "Lulli-Afrihká dálveáigi", "SGT": "Singapore dálveáigi", "SRT": "Suriname áigi", "TMST": "Turkmenistana geasseáigi", "TMT": "Turkmenistana dálveáigi", "UYST": "Uruguaya geasseáigi", "UYT": "Uruguaya dálveáigi", "VET": "Venezuela áigi", "WARST": "Oarje-Argentina geasseáigi", "WART": "Oarje-Argentina dálveáigi", "WAST": "Oarje-Afrihká geasseáigi", "WAT": "Oarje-Afrihká dálveáigi", "WESZ": "Oarje-Eurohpá geasseáigi", "WEZ": "Oarje-Eurohpá dálveáigi", "WIB": "Oarje-Indonesia áigi", "WIT": "Nuorta-Indonesia áigi", "WITA": "Gaska-Indonesia áigi", "∅∅∅": "Peru geasseáigi"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["se_no"] = func() locales.Translator {
		return &localen{

			locale:                 "se_NO",
			pluralsCardinal:        []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ođđj", "guov", "njuk", "cuo", "mies", "geas", "suoi", "borg", "čakč", "golg", "skáb", "juov"},
			monthsNarrow:           []string{"", "O", "G", "N", "C", "M", "G", "S", "B", "Č", "G", "S", "J"},
			monthsWide:             []string{"", "ođđajagemánnu", "guovvamánnu", "njukčamánnu", "cuoŋománnu", "miessemánnu", "geassemánnu", "suoidnemánnu", "borgemánnu", "čakčamánnu", "golggotmánnu", "skábmamánnu", "juovlamánnu"},
			daysAbbreviated:        []string{"sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"},
			daysNarrow:             []string{"S", "V", "M", "G", "D", "B", "L"},
			daysShort:              []string{"sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"},
			daysWide:               []string{"sotnabeaivi", "vuossárga", "maŋŋebárga", "gaskavahkku", "duorasdat", "bearjadat", "lávvardat"},
			periodsAbbreviated:     []string{"i.b.", "e.b."},
			periodsNarrow:          []string{"i.b.", "e.b."},
			periodsWide:            []string{"iđitbeaivet", "eahketbeaivet"},
			erasAbbreviated:        []string{"o.Kr.", "m.Kr."},
			erasNarrow:             []string{"ooá", "oá"},
			erasWide:               []string{"ovdal Kristtusa", "maŋŋel Kristtusa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Greenwich gaskka áigi", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "gaska-Eurohpá geassiáigi", "MEZ": "gaska-Eurohpá dábálašáigi", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "nuorti-Eurohpá geassiáigi", "OEZ": "nuorti-Eurohpá dábálašáigi", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "oarje-Eurohpá geassiáigi", "WEZ": "oarje-Eurohpá dábálašáigi", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["se_se"] = func() locales.Translator {
		return &localen{

			locale:                 "se_SE",
			pluralsCardinal:        []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "Nkr", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "kr", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ođđj", "guov", "njuk", "cuo", "mies", "geas", "suoi", "borg", "čakč", "golg", "skáb", "juov"},
			monthsNarrow:           []string{"", "O", "G", "N", "C", "M", "G", "S", "B", "Č", "G", "S", "J"},
			monthsWide:             []string{"", "ođđajagemánnu", "guovvamánnu", "njukčamánnu", "cuoŋománnu", "miessemánnu", "geassemánnu", "suoidnemánnu", "borgemánnu", "čakčamánnu", "golggotmánnu", "skábmamánnu", "juovlamánnu"},
			daysAbbreviated:        []string{"sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"},
			daysNarrow:             []string{"S", "V", "M", "G", "D", "B", "L"},
			daysShort:              []string{"sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"},
			daysWide:               []string{"sotnabeaivi", "vuossárga", "maŋŋebárga", "gaskavahkku", "duorasdat", "bearjadat", "lávvardat"},
			periodsAbbreviated:     []string{"i.b.", "e.b."},
			periodsNarrow:          []string{"i.b.", "e.b."},
			periodsWide:            []string{"iđitbeaivet", "eahketbeaivet"},
			erasAbbreviated:        []string{"o.Kr.", "m.Kr."},
			erasNarrow:             []string{"ooá", "oá"},
			erasWide:               []string{"ovdal Kristtusa", "maŋŋel Kristtusa"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "Greenwich gaskka áigi", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "gaska-Eurohpá geassiáigi", "MEZ": "gaska-Eurohpá dábálašáigi", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "nuorti-Eurohpá geassiáigi", "OEZ": "nuorti-Eurohpá dábálašáigi", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "oarje-Eurohpá geassiáigi", "WEZ": "oarje-Eurohpá dábálašáigi", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["seh"] = func() locales.Translator {
		return &localen{

			locale:            "seh",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			decimal:           ",",
			group:             ".",
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MTn", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "Jan", "Fev", "Mar", "Abr", "Mai", "Jun", "Jul", "Aug", "Set", "Otu", "Nov", "Dec"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Janeiro", "Fevreiro", "Marco", "Abril", "Maio", "Junho", "Julho", "Augusto", "Setembro", "Otubro", "Novembro", "Decembro"},
			daysAbbreviated:   []string{"Dim", "Pos", "Pir", "Tat", "Nai", "Sha", "Sab"},
			daysNarrow:        []string{"D", "P", "C", "T", "N", "S", "S"},
			daysWide:          []string{"Dimingu", "Chiposi", "Chipiri", "Chitatu", "Chinai", "Chishanu", "Sabudu"},
			erasAbbreviated:   []string{"AC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Antes de Cristo", "Anno Domini"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["seh_mz"] = func() locales.Translator {
		return &localen{

			locale:            "seh_MZ",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			decimal:           ",",
			group:             ".",
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "Jan", "Fev", "Mar", "Abr", "Mai", "Jun", "Jul", "Aug", "Set", "Otu", "Nov", "Dec"},
			monthsNarrow:      []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:        []string{"", "Janeiro", "Fevreiro", "Marco", "Abril", "Maio", "Junho", "Julho", "Augusto", "Setembro", "Otubro", "Novembro", "Decembro"},
			daysAbbreviated:   []string{"Dim", "Pos", "Pir", "Tat", "Nai", "Sha", "Sab"},
			daysNarrow:        []string{"D", "P", "C", "T", "N", "S", "S"},
			daysWide:          []string{"Dimingu", "Chiposi", "Chipiri", "Chitatu", "Chinai", "Chishanu", "Sabudu"},
			erasAbbreviated:   []string{"AC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"Antes de Cristo", "Anno Domini"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1693594b7292f66cfb0a4a6405e5914e,
			fnFmtCurrency:         fnFmtCurrency_0519d76434d5fdd728586e7a31a5a755,
			fnFmtDateFull:         fnFmtDateFull_884dee6d35c0b3a152fe300562570c35,
			fnFmtDateLong:         fnFmtDateLong_ca69f3b8d3dd39b3093791ebf34437bc,
			fnFmtDateMedium:       fnFmtDateMedium_1bc79335f4a6998f27c6e8a29a9b778e,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ses"] = func() locales.Translator {
		return &localen{

			locale:             "ses",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alz", "Asi"},
			daysNarrow:         []string{"H", "T", "T", "L", "L", "L", "S"},
			daysWide:           []string{"Alhadi", "Atinni", "Atalaata", "Alarba", "Alhamiisa", "Alzuma", "Asibti"},
			periodsAbbreviated: []string{"Adduha", "Aluula"},
			periodsWide:        []string{"Adduha", "Aluula"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa zamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_dd5c984fc6201988feed2698382f9150,
			fnFmtCurrency:         fnFmtCurrency_f27eb78b129904a4d98d49f647da3164,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ses_ml"] = func() locales.Translator {
		return &localen{

			locale:             "ses_ML",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alz", "Asi"},
			daysNarrow:         []string{"H", "T", "T", "L", "L", "L", "S"},
			daysWide:           []string{"Alhadi", "Atinni", "Atalaata", "Alarba", "Alhamiisa", "Alzuma", "Asibti"},
			periodsAbbreviated: []string{"Adduha", "Aluula"},
			periodsWide:        []string{"Adduha", "Aluula"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa zamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_dd5c984fc6201988feed2698382f9150,
			fnFmtCurrency:         fnFmtCurrency_f27eb78b129904a4d98d49f647da3164,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sg"] = func() locales.Translator {
		return &localen{

			locale:             "sg",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Nye", "Ful", "Mbä", "Ngu", "Bêl", "Fön", "Len", "Kük", "Mvu", "Ngb", "Nab", "Kak"},
			monthsNarrow:       []string{"", "N", "F", "M", "N", "B", "F", "L", "K", "M", "N", "N", "K"},
			monthsWide:         []string{"", "Nyenye", "Fulundïgi", "Mbängü", "Ngubùe", "Bêläwü", "Föndo", "Lengua", "Kükürü", "Mvuka", "Ngberere", "Nabändüru", "Kakauka"},
			daysAbbreviated:    []string{"Bk1", "Bk2", "Bk3", "Bk4", "Bk5", "Lâp", "Lây"},
			daysNarrow:         []string{"K", "S", "T", "S", "K", "P", "Y"},
			daysWide:           []string{"Bikua-ôko", "Bïkua-ûse", "Bïkua-ptâ", "Bïkua-usïö", "Bïkua-okü", "Lâpôsö", "Lâyenga"},
			periodsAbbreviated: []string{"ND", "LK"},
			periodsWide:        []string{"ND", "LK"},
			erasAbbreviated:    []string{"KnK", "NpK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kôzo na Krîstu", "Na pekô tî Krîstu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ec9761d27ad9ead532f5f112e23983b1,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sg_cf"] = func() locales.Translator {
		return &localen{

			locale:             "sg_CF",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              ".",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Nye", "Ful", "Mbä", "Ngu", "Bêl", "Fön", "Len", "Kük", "Mvu", "Ngb", "Nab", "Kak"},
			monthsNarrow:       []string{"", "N", "F", "M", "N", "B", "F", "L", "K", "M", "N", "N", "K"},
			monthsWide:         []string{"", "Nyenye", "Fulundïgi", "Mbängü", "Ngubùe", "Bêläwü", "Föndo", "Lengua", "Kükürü", "Mvuka", "Ngberere", "Nabändüru", "Kakauka"},
			daysAbbreviated:    []string{"Bk1", "Bk2", "Bk3", "Bk4", "Bk5", "Lâp", "Lây"},
			daysNarrow:         []string{"K", "S", "T", "S", "K", "P", "Y"},
			daysWide:           []string{"Bikua-ôko", "Bïkua-ûse", "Bïkua-ptâ", "Bïkua-usïö", "Bïkua-okü", "Lâpôsö", "Lâyenga"},
			periodsAbbreviated: []string{"ND", "LK"},
			periodsWide:        []string{"ND", "LK"},
			erasAbbreviated:    []string{"KnK", "NpK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kôzo na Krîstu", "Na pekô tî Krîstu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ec9761d27ad9ead532f5f112e23983b1,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["shi"] = func() locales.Translator {
		return &localen{

			locale:             "shi",
			pluralsCardinal:    []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ⵉⵏⵏ", "ⴱⵕⴰ", "ⵎⴰⵕ", "ⵉⴱⵔ", "ⵎⴰⵢ", "ⵢⵓⵏ", "ⵢⵓⵍ", "ⵖⵓⵛ", "ⵛⵓⵜ", "ⴽⵜⵓ", "ⵏⵓⵡ", "ⴷⵓⵊ"},
			monthsNarrow:       []string{"", "ⵉ", "ⴱ", "ⵎ", "ⵉ", "ⵎ", "ⵢ", "ⵢ", "ⵖ", "ⵛ", "ⴽ", "ⵏ", "ⴷ"},
			monthsWide:         []string{"", "ⵉⵏⵏⴰⵢⵔ", "ⴱⵕⴰⵢⵕ", "ⵎⴰⵕⵚ", "ⵉⴱⵔⵉⵔ", "ⵎⴰⵢⵢⵓ", "ⵢⵓⵏⵢⵓ", "ⵢⵓⵍⵢⵓⵣ", "ⵖⵓⵛⵜ", "ⵛⵓⵜⴰⵏⴱⵉⵔ", "ⴽⵜⵓⴱⵔ", "ⵏⵓⵡⴰⵏⴱⵉⵔ", "ⴷⵓⵊⴰⵏⴱⵉⵔ"},
			daysAbbreviated:    []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysWide:           []string{"ⴰⵙⴰⵎⴰⵙ", "ⴰⵢⵏⴰⵙ", "ⴰⵙⵉⵏⴰⵙ", "ⴰⴽⵕⴰⵙ", "ⴰⴽⵡⴰⵙ", "ⵙⵉⵎⵡⴰⵙ", "ⴰⵙⵉⴹⵢⴰⵙ"},
			periodsAbbreviated: []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsWide:        []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			erasAbbreviated:    []string{"ⴷⴰⵄ", "ⴷⴼⵄ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ⴷⴰⵜ ⵏ ⵄⵉⵙⴰ", "ⴷⴼⴼⵉⵔ ⵏ ⵄⵉⵙⴰ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c6f27f6f1d6b31d66116bd13f68bb112,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["shi_latn"] = func() locales.Translator {
		return &localen{

			locale:             "shi_Latn",
			pluralsCardinal:    []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "inn", "bṛa", "maṛ", "ibr", "may", "yun", "yul", "ɣuc", "cut", "ktu", "nuw", "duj"},
			monthsNarrow:       []string{"", "i", "b", "m", "i", "m", "y", "y", "ɣ", "c", "k", "n", "d"},
			monthsWide:         []string{"", "innayr", "bṛayṛ", "maṛṣ", "ibrir", "mayyu", "yunyu", "yulyuz", "ɣuct", "cutanbir", "ktubr", "nuwanbir", "dujanbir"},
			daysAbbreviated:    []string{"asa", "ayn", "asi", "akṛ", "akw", "asim", "asiḍ"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:           []string{"asamas", "aynas", "asinas", "akṛas", "akwas", "asimwas", "asiḍyas"},
			periodsAbbreviated: []string{"tifawt", "tadggʷat"},
			periodsNarrow:      []string{"", ""},
			periodsWide:        []string{"tifawt", "tadggʷat"},
			erasAbbreviated:    []string{"daɛ", "dfɛ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"dat n ɛisa", "dffir n ɛisa"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c6f27f6f1d6b31d66116bd13f68bb112,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["shi_latn_ma"] = func() locales.Translator {
		return &localen{

			locale:             "shi_Latn_MA",
			pluralsCardinal:    []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ⵉⵏⵏ", "ⴱⵕⴰ", "ⵎⴰⵕ", "ⵉⴱⵔ", "ⵎⴰⵢ", "ⵢⵓⵏ", "ⵢⵓⵍ", "ⵖⵓⵛ", "ⵛⵓⵜ", "ⴽⵜⵓ", "ⵏⵓⵡ", "ⴷⵓⵊ"},
			monthsNarrow:       []string{"", "ⵉ", "ⴱ", "ⵎ", "ⵉ", "ⵎ", "ⵢ", "ⵢ", "ⵖ", "ⵛ", "ⴽ", "ⵏ", "ⴷ"},
			monthsWide:         []string{"", "ⵉⵏⵏⴰⵢⵔ", "ⴱⵕⴰⵢⵕ", "ⵎⴰⵕⵚ", "ⵉⴱⵔⵉⵔ", "ⵎⴰⵢⵢⵓ", "ⵢⵓⵏⵢⵓ", "ⵢⵓⵍⵢⵓⵣ", "ⵖⵓⵛⵜ", "ⵛⵓⵜⴰⵏⴱⵉⵔ", "ⴽⵜⵓⴱⵔ", "ⵏⵓⵡⴰⵏⴱⵉⵔ", "ⴷⵓⵊⴰⵏⴱⵉⵔ"},
			daysAbbreviated:    []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysWide:           []string{"ⴰⵙⴰⵎⴰⵙ", "ⴰⵢⵏⴰⵙ", "ⴰⵙⵉⵏⴰⵙ", "ⴰⴽⵕⴰⵙ", "ⴰⴽⵡⴰⵙ", "ⵙⵉⵎⵡⴰⵙ", "ⴰⵙⵉⴹⵢⴰⵙ"},
			periodsAbbreviated: []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsWide:        []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			erasAbbreviated:    []string{"ⴷⴰⵄ", "ⴷⴼⵄ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ⴷⴰⵜ ⵏ ⵄⵉⵙⴰ", "ⴷⴼⴼⵉⵔ ⵏ ⵄⵉⵙⴰ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c6f27f6f1d6b31d66116bd13f68bb112,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["shi_tfng"] = func() locales.Translator {
		return &localen{

			locale:             "shi_Tfng",
			pluralsCardinal:    []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ⵉⵏⵏ", "ⴱⵕⴰ", "ⵎⴰⵕ", "ⵉⴱⵔ", "ⵎⴰⵢ", "ⵢⵓⵏ", "ⵢⵓⵍ", "ⵖⵓⵛ", "ⵛⵓⵜ", "ⴽⵜⵓ", "ⵏⵓⵡ", "ⴷⵓⵊ"},
			monthsNarrow:       []string{"", "ⵉ", "ⴱ", "ⵎ", "ⵉ", "ⵎ", "ⵢ", "ⵢ", "ⵖ", "ⵛ", "ⴽ", "ⵏ", "ⴷ"},
			monthsWide:         []string{"", "ⵉⵏⵏⴰⵢⵔ", "ⴱⵕⴰⵢⵕ", "ⵎⴰⵕⵚ", "ⵉⴱⵔⵉⵔ", "ⵎⴰⵢⵢⵓ", "ⵢⵓⵏⵢⵓ", "ⵢⵓⵍⵢⵓⵣ", "ⵖⵓⵛⵜ", "ⵛⵓⵜⴰⵏⴱⵉⵔ", "ⴽⵜⵓⴱⵔ", "ⵏⵓⵡⴰⵏⴱⵉⵔ", "ⴷⵓⵊⴰⵏⴱⵉⵔ"},
			daysAbbreviated:    []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysWide:           []string{"ⴰⵙⴰⵎⴰⵙ", "ⴰⵢⵏⴰⵙ", "ⴰⵙⵉⵏⴰⵙ", "ⴰⴽⵕⴰⵙ", "ⴰⴽⵡⴰⵙ", "ⵙⵉⵎⵡⴰⵙ", "ⴰⵙⵉⴹⵢⴰⵙ"},
			periodsAbbreviated: []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsWide:        []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			erasAbbreviated:    []string{"ⴷⴰⵄ", "ⴷⴼⵄ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ⴷⴰⵜ ⵏ ⵄⵉⵙⴰ", "ⴷⴼⴼⵉⵔ ⵏ ⵄⵉⵙⴰ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c6f27f6f1d6b31d66116bd13f68bb112,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["shi_tfng_ma"] = func() locales.Translator {
		return &localen{

			locale:             "shi_Tfng_MA",
			pluralsCardinal:    []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ⵉⵏⵏ", "ⴱⵕⴰ", "ⵎⴰⵕ", "ⵉⴱⵔ", "ⵎⴰⵢ", "ⵢⵓⵏ", "ⵢⵓⵍ", "ⵖⵓⵛ", "ⵛⵓⵜ", "ⴽⵜⵓ", "ⵏⵓⵡ", "ⴷⵓⵊ"},
			monthsNarrow:       []string{"", "ⵉ", "ⴱ", "ⵎ", "ⵉ", "ⵎ", "ⵢ", "ⵢ", "ⵖ", "ⵛ", "ⴽ", "ⵏ", "ⴷ"},
			monthsWide:         []string{"", "ⵉⵏⵏⴰⵢⵔ", "ⴱⵕⴰⵢⵕ", "ⵎⴰⵕⵚ", "ⵉⴱⵔⵉⵔ", "ⵎⴰⵢⵢⵓ", "ⵢⵓⵏⵢⵓ", "ⵢⵓⵍⵢⵓⵣ", "ⵖⵓⵛⵜ", "ⵛⵓⵜⴰⵏⴱⵉⵔ", "ⴽⵜⵓⴱⵔ", "ⵏⵓⵡⴰⵏⴱⵉⵔ", "ⴷⵓⵊⴰⵏⴱⵉⵔ"},
			daysAbbreviated:    []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysWide:           []string{"ⴰⵙⴰⵎⴰⵙ", "ⴰⵢⵏⴰⵙ", "ⴰⵙⵉⵏⴰⵙ", "ⴰⴽⵕⴰⵙ", "ⴰⴽⵡⴰⵙ", "ⵙⵉⵎⵡⴰⵙ", "ⴰⵙⵉⴹⵢⴰⵙ"},
			periodsAbbreviated: []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsWide:        []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			erasAbbreviated:    []string{"ⴷⴰⵄ", "ⴷⴼⵄ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ⴷⴰⵜ ⵏ ⵄⵉⵙⴰ", "ⴷⴼⴼⵉⵔ ⵏ ⵄⵉⵙⴰ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c6f27f6f1d6b31d66116bd13f68bb112,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["si"] = func() locales.Translator {
		return &localen{

			locale:                 "si",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "රු.", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "සිෆ්එ", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ජන", "පෙබ", "මාර්තු", "අප්\u200dරේල්", "මැයි", "ජූනි", "ජූලි", "අගෝ", "සැප්", "ඔක්", "නොවැ", "දෙසැ"},
			monthsNarrow:           []string{"", "ජ", "පෙ", "මා", "අ", "මැ", "ජූ", "ජූ", "අ", "සැ", "ඔ", "නෙ", "දෙ"},
			monthsWide:             []string{"", "ජනවාරි", "පෙබරවාරි", "මාර්තු", "අප්\u200dරේල්", "මැයි", "ජූනි", "ජූලි", "අගෝස්තු", "සැප්තැම්බර්", "ඔක්තෝබර්", "නොවැම්බර්", "දෙසැම්බර්"},
			daysAbbreviated:        []string{"ඉරිදා", "සඳුදා", "අඟහ", "බදාදා", "බ්\u200dරහස්", "සිකු", "සෙන"},
			daysNarrow:             []string{"ඉ", "ස", "අ", "බ", "බ්\u200dර", "සි", "සෙ"},
			daysShort:              []string{"ඉරි", "සඳු", "අඟ", "බදා", "බ්\u200dරහ", "සිකු", "සෙන"},
			daysWide:               []string{"ඉරිදා", "සඳුදා", "අඟහරුවාදා", "බදාදා", "බ්\u200dරහස්පතින්දා", "සිකුරාදා", "සෙනසුරාදා"},
			periodsAbbreviated:     []string{"පෙ.ව.", "ප.ව."},
			periodsNarrow:          []string{"පෙ", "ප"},
			periodsWide:            []string{"පෙ.ව.", "ප.ව."},
			erasAbbreviated:        []string{"ක්\u200dරි.පූ.", "ක්\u200dරි.ව."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ක්\u200dරිස්තු පූර්ව", "ක්\u200dරිස්තු වර්ෂ"},
			timezones:              map[string]string{"ACDT": "මධ්\u200dයම ඔස්ට්\u200dරේලියානු දහවල් වේලාව", "ACST": "ඕස්ට්\u200dරේලියානු සම්මත වේලාව", "ACWDT": "මධ්\u200dයම බටහිර ඔස්ට්\u200dරේලියානු දහවල් වේලාව", "ACWST": "මධ්\u200dයම බටහිර ඔස්ට්\u200dරේලියානු සම්මත වේලාව", "ADT": "අත්ලාන්තික් දිවාආලෝක වේලාව", "AEDT": "නැඟෙනහිර ඕස්ට්\u200dරේලියානු දහවල් වේලාව", "AEST": "නැගෙනහිර ඕස්ට්\u200dරේලියානු සම්මත වේලාව", "AKDT": "ඇලස්කා දිවාආලෝක වේලාව", "AKST": "ඇලස්කා සම්මත වේලාව", "ARST": "ආර්ජන්ටිනා ග්\u200dරීෂ්ම කාලය", "ART": "ආර්ජන්ටිනා සම්මත වේලාව", "AST": "අත්ලාන්තික් සම්මත වේලාව", "AWDT": "බටහිර ඔස්ට්\u200dරේලියානු දහවල් වේලාව", "AWST": "බටහිර ඕස්ට්\u200dරේලියානු සම්මත වේලාව", "BOT": "බොලිවියා වේලාව", "BT": "භුතාන වේලාව", "CAT": "මධ්\u200dයම අප්\u200dරිකානු වේලාව", "CDT": "උතුරු ඇමරිකානු මධ්\u200dයම දිවාආලෝක වේලාව", "CHADT": "චැතම් දිවා වේලාව", "CHAST": "චැතම් සම්මත වේලාව", "CLST": "චිලී ග්\u200dරීෂ්ම කාලය", "CLT": "චිලී සම්මත වේලාව", "COST": "කොලොම්බියා ග්\u200dරීෂ්ම කාලය", "COT": "කොලොම්බියා සම්මත වේලාව", "CST": "උතුරු ඇමරිකානු මධ්\u200dයම සම්මත වේලාව", "ChST": "චමොරෝ වේලාව", "EAT": "නැගෙනහිර අප්\u200dරිකානු වේලාව", "ECT": "ඉක්වදෝර් වේලාව", "EDT": "උතුරු ඇමරිකානු නැගෙනහිර දිවාආලෝක වේලාව", "EST": "උතුරු ඇමරිකානු නැගෙනහිර සම්මත වේලාව", "GFT": "ප්\u200dරංශ ගයනා වේලාව", "GMT": "ග්\u200dරිනිච් මධ්\u200dයම වේලාව", "GST": "ගල්ෆ් වේලාව", "GYT": "ගයනා වේලාව", "HADT": "හවායි-අලෙයුතියාන් දිවාආලෝක වේලාව", "HAST": "හවායි-අලෙයුතියාන් සම්මත වේලාව", "HAT": "නිව්ෆවුන්ලන්ත දිවාආලෝක වේලාව", "HECU": "කියුබානු දිවාආලෝක වේලාව", "HEEG": "නැගෙනහිර ග්\u200dරීන්ලන්ත ග්\u200dරීෂ්ම කාලය", "HENOMX": "වයඹ මෙක්සිකෝ දිවාආලෝක වේලාව", "HEOG": "බටහිර ග්\u200dරීන්ලන්ත ග්\u200dරීෂ්ම කාලය", "HEPM": "ශාන්ත පියරේ සහ මැකෝලන් දිවාආලෝක වේලාව", "HEPMX": "මෙක්සිකෝ පැසිෆික් දිවාආලෝක වේලාව", "HKST": "හොංකොං ග්\u200dරීෂ්ම වේලාව", "HKT": "හොංකොං සම්මත වේලාව", "HNCU": "කියුබානු සම්මත වේලාව", "HNEG": "නැගෙනහිර ග්\u200dරීන්ලන්ත සම්මත වේලාව", "HNNOMX": "වයඹ මෙක්සිකෝ සම්මත වේලාව", "HNOG": "බටහිර ග්\u200dරීන්ලන්ත සම්මත වේලාව", "HNPM": "ශාන්ත පියරේ සහ මැකෝලන් සම්මත වේලාව", "HNPMX": "මෙක්සිකෝ පැසිෆික් සම්මත වේලාව", "HNT": "නිව්ෆවුන්ලන්ත සම්මත වේලාව", "IST": "ඉන්දියානු වේලාව", "JDT": "ජපාන දහවල් වේලාව", "JST": "ජපාන සම්මත වේලාව", "LHDT": "ලෝර්ඩ් හෝව් දිවා වේලාව", "LHST": "ලෝර්ඩ් හෝව් සම්මත වේලාව", "MDT": "උතුරු ඇමරිකානු කඳුකර දිවාආලෝක වේලාව", "MESZ": "මධ්\u200dයම යුරෝපීය ග්\u200dරීෂ්ම වේලාව", "MEZ": "මධ්\u200dයම යුරෝපීය සම්මත වේලාව", "MST": "උතුරු ඇමරිකානු කඳුකර සම්මත වේලාව", "MYT": "මැලේසියානු වේලාව", "NZDT": "නවසීලන්ත දිවා වේලාව", "NZST": "නවසීලන්ත සම්මත වේලාව", "OESZ": "නැගෙනහිර යුරෝපීය ග්\u200dරීෂ්ම වේලාව", "OEZ": "නැගෙනහිර යුරෝපීය සම්මත වේලාව", "PDT": "උතුරු ඇමරිකානු පැසිෆික් දිවාආලෝක වේලාව", "PST": "උතුරු ඇමරිකානු පැසිෆික් සම්මත වේලාව", "SAST": "දකුණු අප්\u200dරිකානු වේලාව", "SGT": "සිංගප්පුරු වේලාව", "SRT": "සුරිනාම වේලාව", "TMST": "ටර්ක්මෙනිස්තාන ග්\u200dරීෂ්ම වේලාව", "TMT": "ටර්ක්මෙනිස්තාන සම්මත වේලාව", "UYST": "උරුගුවේ ග්\u200dරීෂ්ම කාලය", "UYT": "උරුගුවේ සම්මත වේලාව", "VET": "වෙනිසියුලා වේලාව", "WARST": "බටහිර ආර්ජන්ටිනා ග්\u200dරීෂ්ම කාලය", "WART": "බටහිර ආර්ජන්ටිනා සම්මත වේලාව", "WAST": "බටහිර අප්\u200dරිකානු ග්\u200dරීෂ්ම කාලය", "WAT": "බටහිර අප්\u200dරිකානු සම්මත වේලාව", "WESZ": "බටහිර යුරෝපීය ග්\u200dරීෂ්ම වේලාව", "WEZ": "බටහිර යුරෝපීය සම්මත වේලාව", "WIB": "බටහිර ඉන්දුනීසියානු වේලාව", "WIT": "නැගෙනහිර ඉන්දුනීසියානු වේලාව", "WITA": "මධ්\u200dයම ඉන්දුනීසියානු වේලාව", "∅∅∅": "ඇසොර්ස් ග්\u200dරීෂ්ම වේලාව"},

			fnCardinalPluralRule:  fnCardinalPluralRule_780c1153d0b26cb9e6da19f93ef34fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_088ba095b00c7f7ec73fa5b6bd628417,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["si_lk"] = func() locales.Translator {
		return &localen{

			locale:                 "si_LK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ජන", "පෙබ", "මාර්තු", "අප්\u200dරේල්", "මැයි", "ජූනි", "ජූලි", "අගෝ", "සැප්", "ඔක්", "නොවැ", "දෙසැ"},
			monthsNarrow:           []string{"", "ජ", "පෙ", "මා", "අ", "මැ", "ජූ", "ජූ", "අ", "සැ", "ඔ", "නෙ", "දෙ"},
			monthsWide:             []string{"", "ජනවාරි", "පෙබරවාරි", "මාර්තු", "අප්\u200dරේල්", "මැයි", "ජූනි", "ජූලි", "අගෝස්තු", "සැප්තැම්බර්", "ඔක්තෝබර්", "නොවැම්බර්", "දෙසැම්බර්"},
			daysAbbreviated:        []string{"ඉරිදා", "සඳුදා", "අඟහ", "බදාදා", "බ්\u200dරහස්", "සිකු", "සෙන"},
			daysNarrow:             []string{"ඉ", "ස", "අ", "බ", "බ්\u200dර", "සි", "සෙ"},
			daysShort:              []string{"ඉරි", "සඳු", "අඟ", "බදා", "බ්\u200dරහ", "සිකු", "සෙන"},
			daysWide:               []string{"ඉරිදා", "සඳුදා", "අඟහරුවාදා", "බදාදා", "බ්\u200dරහස්පතින්දා", "සිකුරාදා", "සෙනසුරාදා"},
			periodsAbbreviated:     []string{"පෙ.ව.", "ප.ව."},
			periodsNarrow:          []string{"පෙ", "ප"},
			periodsWide:            []string{"පෙ.ව.", "ප.ව."},
			erasAbbreviated:        []string{"ක්\u200dරි.පූ.", "ක්\u200dරි.ව."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"ක්\u200dරිස්තු පූර්ව", "ක්\u200dරිස්තු වර්ෂ"},
			timezones:              map[string]string{"ACDT": "මධ්\u200dයම ඔස්ට්\u200dරේලියානු දහවල් වේලාව", "ACST": "ඕස්ට්\u200dරේලියානු සම්මත වේලාව", "ACWDT": "මධ්\u200dයම බටහිර ඔස්ට්\u200dරේලියානු දහවල් වේලාව", "ACWST": "මධ්\u200dයම බටහිර ඔස්ට්\u200dරේලියානු සම්මත වේලාව", "ADT": "අත්ලාන්තික් දිවාආලෝක වේලාව", "AEDT": "නැඟෙනහිර ඕස්ට්\u200dරේලියානු දහවල් වේලාව", "AEST": "නැගෙනහිර ඕස්ට්\u200dරේලියානු සම්මත වේලාව", "AKDT": "ඇලස්කා දිවාආලෝක වේලාව", "AKST": "ඇලස්කා සම්මත වේලාව", "ARST": "ආර්ජන්ටිනා ග්\u200dරීෂ්ම කාලය", "ART": "ආර්ජන්ටිනා සම්මත වේලාව", "AST": "අත්ලාන්තික් සම්මත වේලාව", "AWDT": "බටහිර ඔස්ට්\u200dරේලියානු දහවල් වේලාව", "AWST": "බටහිර ඕස්ට්\u200dරේලියානු සම්මත වේලාව", "BOT": "බොලිවියා වේලාව", "BT": "භුතාන වේලාව", "CAT": "මධ්\u200dයම අප්\u200dරිකානු වේලාව", "CDT": "උතුරු ඇමරිකානු මධ්\u200dයම දිවාආලෝක වේලාව", "CHADT": "චැතම් දිවා වේලාව", "CHAST": "චැතම් සම්මත වේලාව", "CLST": "චිලී ග්\u200dරීෂ්ම කාලය", "CLT": "චිලී සම්මත වේලාව", "COST": "කොලොම්බියා ග්\u200dරීෂ්ම කාලය", "COT": "කොලොම්බියා සම්මත වේලාව", "CST": "උතුරු ඇමරිකානු මධ්\u200dයම සම්මත වේලාව", "ChST": "චමොරෝ වේලාව", "EAT": "නැගෙනහිර අප්\u200dරිකානු වේලාව", "ECT": "ඉක්වදෝර් වේලාව", "EDT": "උතුරු ඇමරිකානු නැගෙනහිර දිවාආලෝක වේලාව", "EST": "උතුරු ඇමරිකානු නැගෙනහිර සම්මත වේලාව", "GFT": "ප්\u200dරංශ ගයනා වේලාව", "GMT": "ග්\u200dරිනිච් මධ්\u200dයම වේලාව", "GST": "ගල්ෆ් වේලාව", "GYT": "ගයනා වේලාව", "HADT": "හවායි-අලෙයුතියාන් දිවාආලෝක වේලාව", "HAST": "හවායි-අලෙයුතියාන් සම්මත වේලාව", "HAT": "නිව්ෆවුන්ලන්ත දිවාආලෝක වේලාව", "HECU": "කියුබානු දිවාආලෝක වේලාව", "HEEG": "නැගෙනහිර ග්\u200dරීන්ලන්ත ග්\u200dරීෂ්ම කාලය", "HENOMX": "වයඹ මෙක්සිකෝ දිවාආලෝක වේලාව", "HEOG": "බටහිර ග්\u200dරීන්ලන්ත ග්\u200dරීෂ්ම කාලය", "HEPM": "ශාන්ත පියරේ සහ මැකෝලන් දිවාආලෝක වේලාව", "HEPMX": "මෙක්සිකෝ පැසිෆික් දිවාආලෝක වේලාව", "HKST": "හොංකොං ග්\u200dරීෂ්ම වේලාව", "HKT": "හොංකොං සම්මත වේලාව", "HNCU": "කියුබානු සම්මත වේලාව", "HNEG": "නැගෙනහිර ග්\u200dරීන්ලන්ත සම්මත වේලාව", "HNNOMX": "වයඹ මෙක්සිකෝ සම්මත වේලාව", "HNOG": "බටහිර ග්\u200dරීන්ලන්ත සම්මත වේලාව", "HNPM": "ශාන්ත පියරේ සහ මැකෝලන් සම්මත වේලාව", "HNPMX": "මෙක්සිකෝ පැසිෆික් සම්මත වේලාව", "HNT": "නිව්ෆවුන්ලන්ත සම්මත වේලාව", "IST": "ඉන්දියානු වේලාව", "JDT": "ජපාන දහවල් වේලාව", "JST": "ජපාන සම්මත වේලාව", "LHDT": "ලෝර්ඩ් හෝව් දිවා වේලාව", "LHST": "ලෝර්ඩ් හෝව් සම්මත වේලාව", "MDT": "උතුරු ඇමරිකානු කඳුකර දිවාආලෝක වේලාව", "MESZ": "මධ්\u200dයම යුරෝපීය ග්\u200dරීෂ්ම වේලාව", "MEZ": "මධ්\u200dයම යුරෝපීය සම්මත වේලාව", "MST": "උතුරු ඇමරිකානු කඳුකර සම්මත වේලාව", "MYT": "මැලේසියානු වේලාව", "NZDT": "නවසීලන්ත දිවා වේලාව", "NZST": "නවසීලන්ත සම්මත වේලාව", "OESZ": "නැගෙනහිර යුරෝපීය ග්\u200dරීෂ්ම වේලාව", "OEZ": "නැගෙනහිර යුරෝපීය සම්මත වේලාව", "PDT": "උතුරු ඇමරිකානු පැසිෆික් දිවාආලෝක වේලාව", "PST": "උතුරු ඇමරිකානු පැසිෆික් සම්මත වේලාව", "SAST": "දකුණු අප්\u200dරිකානු වේලාව", "SGT": "සිංගප්පුරු වේලාව", "SRT": "සුරිනාම වේලාව", "TMST": "ටර්ක්මෙනිස්තාන ග්\u200dරීෂ්ම වේලාව", "TMT": "ටර්ක්මෙනිස්තාන සම්මත වේලාව", "UYST": "උරුගුවේ ග්\u200dරීෂ්ම කාලය", "UYT": "උරුගුවේ සම්මත වේලාව", "VET": "වෙනිසියුලා වේලාව", "WARST": "බටහිර ආර්ජන්ටිනා ග්\u200dරීෂ්ම කාලය", "WART": "බටහිර ආර්ජන්ටිනා සම්මත වේලාව", "WAST": "බටහිර අප්\u200dරිකානු ග්\u200dරීෂ්ම කාලය", "WAT": "බටහිර අප්\u200dරිකානු සම්මත වේලාව", "WESZ": "බටහිර යුරෝපීය ග්\u200dරීෂ්ම වේලාව", "WEZ": "බටහිර යුරෝපීය සම්මත වේලාව", "WIB": "බටහිර ඉන්දුනීසියානු වේලාව", "WIT": "නැගෙනහිර ඉන්දුනීසියානු වේලාව", "WITA": "මධ්\u200dයම ඉන්දුනීසියානු වේලාව", "∅∅∅": "පේරු ග්\u200dරීෂ්ම කාලය"},

			fnCardinalPluralRule:  fnCardinalPluralRule_780c1153d0b26cb9e6da19f93ef34fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_391789992db94778b51fb64c1b45a71a,
			fnFmtTimeLong:         fnFmtTimeLong_25847504d894858ff8768557462c84d0,
			fnFmtTimeMedium:       fnFmtTimeMedium_c9ed57159952e0c7f4ac96091e48d9ad,
			fnFmtTimeShort:        fnFmtTimeShort_f122edf9136fadb0288a3d0e77450a50,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_088ba095b00c7f7ec73fa5b6bd628417,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sk"] = func() locales.Translator {
		return &localen{

			locale:                 "sk",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "NIS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "máj", "jún", "júl", "aug", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januára", "februára", "marca", "apríla", "mája", "júna", "júla", "augusta", "septembra", "októbra", "novembra", "decembra"},
			daysAbbreviated:        []string{"ne", "po", "ut", "st", "št", "pi", "so"},
			daysNarrow:             []string{"n", "p", "u", "s", "š", "p", "s"},
			daysShort:              []string{"ne", "po", "ut", "st", "št", "pi", "so"},
			daysWide:               []string{"nedeľa", "pondelok", "utorok", "streda", "štvrtok", "piatok", "sobota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"pred Kr.", "po Kr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"pred Kristom", "po Kristovi"},
			timezones:              map[string]string{"ACDT": "stredoaustrálsky letný čas", "ACST": "stredoaustrálsky štandardný čas", "ACWDT": "stredozápadný austrálsky letný čas", "ACWST": "stredozápadný austrálsky štandardný čas", "ADT": "atlantický letný čas", "AEDT": "východoaustrálsky letný čas", "AEST": "východoaustrálsky štandardný čas", "AKDT": "aljašský letný čas", "AKST": "aljašský štandardný čas", "ARST": "argentínsky letný čas", "ART": "argentínsky štandardný čas", "AST": "atlantický štandardný čas", "AWDT": "západoaustrálsky letný čas", "AWST": "západoaustrálsky štandardný čas", "BOT": "bolívijský čas", "BT": "bhutánsky čas", "CAT": "stredoafrický čas", "CDT": "severoamerický centrálny letný čas", "CHADT": "chathamský letný čas", "CHAST": "chathamský štandardný čas", "CLST": "čilský letný čas", "CLT": "čilský štandardný čas", "COST": "kolumbijský letný čas", "COT": "kolumbijský štandardný čas", "CST": "severoamerický centrálny štandardný čas", "ChST": "chamorrský čas", "EAT": "východoafrický čas", "ECT": "ekvádorský čas", "EDT": "severoamerický východný letný čas", "EST": "severoamerický východný štandardný čas", "GFT": "francúzskoguyanský čas", "GMT": "greenwichský čas", "GST": "štandardný čas Perzského zálivu", "GYT": "guyanský čas", "HADT": "havajsko-aleutský letný čas", "HAST": "havajsko-aleutský štandardný čas", "HAT": "newfoundlandský letný čas", "HECU": "kubánsky letný čas", "HEEG": "východogrónsky letný čas", "HENOMX": "severozápadný mexický letný čas", "HEOG": "západogrónsky letný čas", "HEPM": "pierre-miquelonský letný čas", "HEPMX": "mexický tichomorský letný čas", "HKST": "hongkonský letný čas", "HKT": "hongkonský štandardný čas", "HNCU": "kubánsky štandardný čas", "HNEG": "východogrónsky štandardný čas", "HNNOMX": "severozápadný mexický štandardný čas", "HNOG": "západogrónsky štandardný čas", "HNPM": "pierre-miquelonský štandardný čas", "HNPMX": "mexický tichomorský štandardný čas", "HNT": "newfoundlandský štandardný čas", "IST": "indický čas", "JDT": "japonský letný čas", "JST": "japonský štandardný čas", "LHDT": "letný čas ostrova lorda Howa", "LHST": "štandardný čas ostrova lorda Howa", "MDT": "severoamerický horský letný čas", "MESZ": "stredoeurópsky letný čas", "MEZ": "stredoeurópsky štandardný čas", "MST": "severoamerický horský štandardný čas", "MYT": "malajzijský čas", "NZDT": "novozélandský letný čas", "NZST": "novozélandský štandardný čas", "OESZ": "východoeurópsky letný čas", "OEZ": "východoeurópsky štandardný čas", "PDT": "severoamerický tichomorský letný čas", "PST": "severoamerický tichomorský štandardný čas", "SAST": "juhoafrický čas", "SGT": "singapurský štandardný čas", "SRT": "surinamský čas", "TMST": "turkménsky letný čas", "TMT": "turkménsky štandardný čas", "UYST": "uruguajský letný čas", "UYT": "uruguajský štandardný čas", "VET": "venezuelský čas", "WARST": "západoargentínsky letný čas", "WART": "západoargentínsky štandardný čas", "WAST": "západoafrický letný čas", "WAT": "západoafrický štandardný čas", "WESZ": "západoeurópsky letný čas", "WEZ": "západoeurópsky štandardný čas", "WIB": "západoindonézsky čas", "WIT": "východoindonézsky čas", "WITA": "stredoindonézsky čas", "∅∅∅": "brazílsky letný čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_935abbc817a75b933eeb5ecc41e85d90,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_15dd2e1fe33af45e98be7d9f9b1b6b93,
			fnFmtDateShort:        fnFmtDateShort_7666e72292ce611269924033db7087d6,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sk_sk"] = func() locales.Translator {
		return &localen{

			locale:                 "sk_SK",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "máj", "jún", "júl", "aug", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januára", "februára", "marca", "apríla", "mája", "júna", "júla", "augusta", "septembra", "októbra", "novembra", "decembra"},
			daysAbbreviated:        []string{"ne", "po", "ut", "st", "št", "pi", "so"},
			daysNarrow:             []string{"n", "p", "u", "s", "š", "p", "s"},
			daysShort:              []string{"ne", "po", "ut", "st", "št", "pi", "so"},
			daysWide:               []string{"nedeľa", "pondelok", "utorok", "streda", "štvrtok", "piatok", "sobota"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"pred Kr.", "po Kr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"pred Kristom", "po Kristovi"},
			timezones:              map[string]string{"ACDT": "stredoaustrálsky letný čas", "ACST": "stredoaustrálsky štandardný čas", "ACWDT": "stredozápadný austrálsky letný čas", "ACWST": "stredozápadný austrálsky štandardný čas", "ADT": "atlantický letný čas", "AEDT": "východoaustrálsky letný čas", "AEST": "východoaustrálsky štandardný čas", "AKDT": "aljašský letný čas", "AKST": "aljašský štandardný čas", "ARST": "argentínsky letný čas", "ART": "argentínsky štandardný čas", "AST": "atlantický štandardný čas", "AWDT": "západoaustrálsky letný čas", "AWST": "západoaustrálsky štandardný čas", "BOT": "bolívijský čas", "BT": "bhutánsky čas", "CAT": "stredoafrický čas", "CDT": "severoamerický centrálny letný čas", "CHADT": "chathamský letný čas", "CHAST": "chathamský štandardný čas", "CLST": "čilský letný čas", "CLT": "čilský štandardný čas", "COST": "kolumbijský letný čas", "COT": "kolumbijský štandardný čas", "CST": "severoamerický centrálny štandardný čas", "ChST": "chamorrský čas", "EAT": "východoafrický čas", "ECT": "ekvádorský čas", "EDT": "severoamerický východný letný čas", "EST": "severoamerický východný štandardný čas", "GFT": "francúzskoguyanský čas", "GMT": "greenwichský čas", "GST": "štandardný čas Perzského zálivu", "GYT": "guyanský čas", "HADT": "havajsko-aleutský letný čas", "HAST": "havajsko-aleutský štandardný čas", "HAT": "newfoundlandský letný čas", "HECU": "kubánsky letný čas", "HEEG": "východogrónsky letný čas", "HENOMX": "severozápadný mexický letný čas", "HEOG": "západogrónsky letný čas", "HEPM": "pierre-miquelonský letný čas", "HEPMX": "mexický tichomorský letný čas", "HKST": "hongkonský letný čas", "HKT": "hongkonský štandardný čas", "HNCU": "kubánsky štandardný čas", "HNEG": "východogrónsky štandardný čas", "HNNOMX": "severozápadný mexický štandardný čas", "HNOG": "západogrónsky štandardný čas", "HNPM": "pierre-miquelonský štandardný čas", "HNPMX": "mexický tichomorský štandardný čas", "HNT": "newfoundlandský štandardný čas", "IST": "indický čas", "JDT": "japonský letný čas", "JST": "japonský štandardný čas", "LHDT": "letný čas ostrova lorda Howa", "LHST": "štandardný čas ostrova lorda Howa", "MDT": "severoamerický horský letný čas", "MESZ": "stredoeurópsky letný čas", "MEZ": "stredoeurópsky štandardný čas", "MST": "severoamerický horský štandardný čas", "MYT": "malajzijský čas", "NZDT": "novozélandský letný čas", "NZST": "novozélandský štandardný čas", "OESZ": "východoeurópsky letný čas", "OEZ": "východoeurópsky štandardný čas", "PDT": "severoamerický tichomorský letný čas", "PST": "severoamerický tichomorský štandardný čas", "SAST": "juhoafrický čas", "SGT": "singapurský štandardný čas", "SRT": "surinamský čas", "TMST": "turkménsky letný čas", "TMT": "turkménsky štandardný čas", "UYST": "uruguajský letný čas", "UYT": "uruguajský štandardný čas", "VET": "venezuelský čas", "WARST": "západoargentínsky letný čas", "WART": "západoargentínsky štandardný čas", "WAST": "západoafrický letný čas", "WAT": "západoafrický štandardný čas", "WESZ": "západoeurópsky letný čas", "WEZ": "západoeurópsky štandardný čas", "WIB": "západoindonézsky čas", "WIT": "východoindonézsky čas", "WITA": "stredoindonézsky čas", "∅∅∅": "azorský letný čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_935abbc817a75b933eeb5ecc41e85d90,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_23f49589a10a043b9966b4062e75298c,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_15dd2e1fe33af45e98be7d9f9b1b6b93,
			fnFmtDateShort:        fnFmtDateShort_7666e72292ce611269924033db7087d6,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_00245e8cddd4df2ec4e382e386a724dd,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sl"] = func() locales.Translator {
		return &localen{

			locale:                 "sl",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{3, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maj", "jun.", "jul.", "avg.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "marec", "april", "maj", "junij", "julij", "avgust", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"ned.", "pon.", "tor.", "sre.", "čet.", "pet.", "sob."},
			daysNarrow:             []string{"n", "p", "t", "s", "č", "p", "s"},
			daysShort:              []string{"ned.", "pon.", "tor.", "sre.", "čet.", "pet.", "sob."},
			daysWide:               []string{"nedelja", "ponedeljek", "torek", "sreda", "četrtek", "petek", "sobota"},
			periodsAbbreviated:     []string{"dop.", "pop."},
			periodsNarrow:          []string{"d", "p"},
			periodsWide:            []string{"dop.", "pop."},
			erasAbbreviated:        []string{"pr. Kr.", "po Kr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"pred Kristusom", "po Kristusu"},
			timezones:              map[string]string{"ACDT": "Avstralski centralni poletni čas", "ACST": "Avstralski centralni standardni čas", "ACWDT": "Avstralski centralni zahodni poletni čas", "ACWST": "Avstralski centralni zahodni standardni čas", "ADT": "Atlantski poletni čas", "AEDT": "Avstralski vzhodni poletni čas", "AEST": "Avstralski vzhodni standardni čas", "AKDT": "Aljaški poletni čas", "AKST": "Aljaški standardni čas", "ARST": "Argentinski poletni čas", "ART": "Argentinski standardni čas", "AST": "Atlantski standardni čas", "AWDT": "Avstralski zahodni poletni čas", "AWST": "Avstralski zahodni standardni čas", "BOT": "Bolivijski čas", "BT": "Butanski čas", "CAT": "Centralnoafriški čas", "CDT": "Centralni poletni čas", "CHADT": "Čatamski poletni čas", "CHAST": "Čatamski standardni čas", "CLST": "Čilski poletni čas", "CLT": "Čilski standardni čas", "COST": "Kolumbijski poletni čas", "COT": "Kolumbijski standardni čas", "CST": "Centralni standardni čas", "ChST": "Čamorski standardni čas", "EAT": "Vzhodnoafriški čas", "ECT": "Ekvadorski čas", "EDT": "Vzhodni poletni čas", "EST": "Vzhodni standardni čas", "GFT": "Čas: Francoska Gvajana", "GMT": "Greenwiški srednji čas", "GST": "Zalivski standardni čas", "GYT": "Gvajanski čas", "HADT": "Havajski aleutski poletni čas", "HAST": "Havajski aleutski standardni čas", "HAT": "Novofundlandski poletni čas", "HECU": "Kubanski poletni čas", "HEEG": "Vzhodnogrenlandski poletni čas", "HENOMX": "Mehiški severozahodni poletni čas", "HEOG": "Zahodnogrenlandski poletni čas", "HEPM": "Poletni čas: Saint Pierre in Miquelon", "HEPMX": "Mehiški pacifiški poletni čas", "HKST": "Hongkonški poletni čas", "HKT": "Hongkonški standardni čas", "HNCU": "Kubanski standardni čas", "HNEG": "Vzhodnogrenlandski standardni čas", "HNNOMX": "Mehiški severozahodni standardni čas", "HNOG": "Zahodnogrenlandski standardni čas", "HNPM": "Standardni čas: Saint Pierre in Miquelon", "HNPMX": "Mehiški pacifiški standardni čas", "HNT": "Novofundlandski standardni čas", "IST": "Indijski standardni čas", "JDT": "Japonski poletni čas", "JST": "Japonski standardni čas", "LHDT": "Poletni čas otoka Lord Howe", "LHST": "Standardni čas otoka Lord Howe", "MDT": "Gorski poletni čas", "MESZ": "Srednjeevropski poletni čas", "MEZ": "Srednjeevropski standardni čas", "MST": "Gorski standardni čas", "MYT": "Malezijski čas", "NZDT": "Novozelandski poletni čas", "NZST": "Novozelandski standardni čas", "OESZ": "Vzhodnoevropski poletni čas", "OEZ": "Vzhodnoevropski standardni čas", "PDT": "Pacifiški poletni čas", "PST": "Pacifiški standardni čas", "SAST": "Južnoafriški čas", "SGT": "Singapurski standardni čas", "SRT": "Surinamski čas", "TMST": "Turkmenistanski poletni čas", "TMT": "Turkmenistanski standardni čas", "UYST": "Urugvajski poletni čas", "UYT": "Urugvajski standardni čas", "VET": "Venezuelski čas", "WARST": "Argentinski zahodni poletni čas", "WART": "Argentinski zahodni standardni čas", "WAST": "Zahodnoafriški poletni čas", "WAT": "Zahodnoafriški standardni čas", "WESZ": "Zahodnoevropski poletni čas", "WEZ": "Zahodnoevropski standardni čas", "WIB": "Indonezijski zahodni čas", "WIT": "Indonezijski vzhodni čas", "WITA": "Indonezijski osrednji čas", "∅∅∅": "Brasilski poletni čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0fef1d0e24fa8383f72b5cbddf29c5f9,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_8c8c1dc47cf6203ce4b51ec99ff21181,
			fnFmtDateLong:         fnFmtDateLong_7394f5b140d96d3ada830127b6b00a5f,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_6782bf6cd88b90ec3c33a5cf6ab98fd7,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_c14bd9bed5ef22ead1e026671990212f,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sl_si"] = func() locales.Translator {
		return &localen{

			locale:                 "sl_SI",
			pluralsCardinal:        []locales.PluralRule{2, 3, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{3, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mar.", "apr.", "maj", "jun.", "jul.", "avg.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "marec", "april", "maj", "junij", "julij", "avgust", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"ned.", "pon.", "tor.", "sre.", "čet.", "pet.", "sob."},
			daysNarrow:             []string{"n", "p", "t", "s", "č", "p", "s"},
			daysShort:              []string{"ned.", "pon.", "tor.", "sre.", "čet.", "pet.", "sob."},
			daysWide:               []string{"nedelja", "ponedeljek", "torek", "sreda", "četrtek", "petek", "sobota"},
			periodsAbbreviated:     []string{"dop.", "pop."},
			periodsNarrow:          []string{"d", "p"},
			periodsWide:            []string{"dop.", "pop."},
			erasAbbreviated:        []string{"pr. Kr.", "po Kr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"pred Kristusom", "po Kristusu"},
			timezones:              map[string]string{"ACDT": "Avstralski centralni poletni čas", "ACST": "Avstralski centralni standardni čas", "ACWDT": "Avstralski centralni zahodni poletni čas", "ACWST": "Avstralski centralni zahodni standardni čas", "ADT": "Atlantski poletni čas", "AEDT": "Avstralski vzhodni poletni čas", "AEST": "Avstralski vzhodni standardni čas", "AKDT": "Aljaški poletni čas", "AKST": "Aljaški standardni čas", "ARST": "Argentinski poletni čas", "ART": "Argentinski standardni čas", "AST": "Atlantski standardni čas", "AWDT": "Avstralski zahodni poletni čas", "AWST": "Avstralski zahodni standardni čas", "BOT": "Bolivijski čas", "BT": "Butanski čas", "CAT": "Centralnoafriški čas", "CDT": "Centralni poletni čas", "CHADT": "Čatamski poletni čas", "CHAST": "Čatamski standardni čas", "CLST": "Čilski poletni čas", "CLT": "Čilski standardni čas", "COST": "Kolumbijski poletni čas", "COT": "Kolumbijski standardni čas", "CST": "Centralni standardni čas", "ChST": "Čamorski standardni čas", "EAT": "Vzhodnoafriški čas", "ECT": "Ekvadorski čas", "EDT": "Vzhodni poletni čas", "EST": "Vzhodni standardni čas", "GFT": "Čas: Francoska Gvajana", "GMT": "Greenwiški srednji čas", "GST": "Zalivski standardni čas", "GYT": "Gvajanski čas", "HADT": "Havajski aleutski poletni čas", "HAST": "Havajski aleutski standardni čas", "HAT": "Novofundlandski poletni čas", "HECU": "Kubanski poletni čas", "HEEG": "Vzhodnogrenlandski poletni čas", "HENOMX": "Mehiški severozahodni poletni čas", "HEOG": "Zahodnogrenlandski poletni čas", "HEPM": "Poletni čas: Saint Pierre in Miquelon", "HEPMX": "Mehiški pacifiški poletni čas", "HKST": "Hongkonški poletni čas", "HKT": "Hongkonški standardni čas", "HNCU": "Kubanski standardni čas", "HNEG": "Vzhodnogrenlandski standardni čas", "HNNOMX": "Mehiški severozahodni standardni čas", "HNOG": "Zahodnogrenlandski standardni čas", "HNPM": "Standardni čas: Saint Pierre in Miquelon", "HNPMX": "Mehiški pacifiški standardni čas", "HNT": "Novofundlandski standardni čas", "IST": "Indijski standardni čas", "JDT": "Japonski poletni čas", "JST": "Japonski standardni čas", "LHDT": "Poletni čas otoka Lord Howe", "LHST": "Standardni čas otoka Lord Howe", "MDT": "Gorski poletni čas", "MESZ": "Srednjeevropski poletni čas", "MEZ": "Srednjeevropski standardni čas", "MST": "Gorski standardni čas", "MYT": "Malezijski čas", "NZDT": "Novozelandski poletni čas", "NZST": "Novozelandski standardni čas", "OESZ": "Vzhodnoevropski poletni čas", "OEZ": "Vzhodnoevropski standardni čas", "PDT": "Pacifiški poletni čas", "PST": "Pacifiški standardni čas", "SAST": "Južnoafriški čas", "SGT": "Singapurski standardni čas", "SRT": "Surinamski čas", "TMST": "Turkmenistanski poletni čas", "TMT": "Turkmenistanski standardni čas", "UYST": "Urugvajski poletni čas", "UYT": "Urugvajski standardni čas", "VET": "Venezuelski čas", "WARST": "Argentinski zahodni poletni čas", "WART": "Argentinski zahodni standardni čas", "WAST": "Zahodnoafriški poletni čas", "WAT": "Zahodnoafriški standardni čas", "WESZ": "Zahodnoevropski poletni čas", "WEZ": "Zahodnoevropski standardni čas", "WIB": "Indonezijski zahodni čas", "WIT": "Indonezijski vzhodni čas", "WITA": "Indonezijski osrednji čas", "∅∅∅": "Perujski poletni čas"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0fef1d0e24fa8383f72b5cbddf29c5f9,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d5329ec1e9c68656688f5fd3b4fa22fc,
			fnFmtCurrency:         fnFmtCurrency_2c2268a7a667170346d6d4c829feb8a4,
			fnFmtDateFull:         fnFmtDateFull_8c8c1dc47cf6203ce4b51ec99ff21181,
			fnFmtDateLong:         fnFmtDateLong_7394f5b140d96d3ada830127b6b00a5f,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_6782bf6cd88b90ec3c33a5cf6ab98fd7,
			fnFmtNumber:           fnFmtNumber_f253fbd23af5a54db29f42f905c9e33a,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_c14bd9bed5ef22ead1e026671990212f,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["smn"] = func() locales.Translator {
		return &localen{

			locale:                 "smn",
			pluralsCardinal:        []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "uđiv", "kuovâ", "njuhčâ", "cuáŋui", "vyesi", "kesi", "syeini", "porge", "čohčâ", "roovvâd", "skammâ", "juovlâ"},
			monthsNarrow:           []string{"", "U", "K", "NJ", "C", "V", "K", "S", "P", "Č", "R", "S", "J"},
			monthsWide:             []string{"", "uđđâivemáánu", "kuovâmáánu", "njuhčâmáánu", "cuáŋuimáánu", "vyesimáánu", "kesimáánu", "syeinimáánu", "porgemáánu", "čohčâmáánu", "roovvâdmáánu", "skammâmáánu", "juovlâmáánu"},
			daysAbbreviated:        []string{"pas", "vuo", "maj", "kos", "tuo", "vás", "láv"},
			daysNarrow:             []string{"p", "V", "M", "K", "T", "V", "L"},
			daysShort:              []string{"pa", "vu", "ma", "ko", "tu", "vá", "lá"},
			daysWide:               []string{"pasepeeivi", "vuossaargâ", "majebaargâ", "koskoho", "tuorâstuv", "vástuppeeivi", "lávurduv"},
			periodsAbbreviated:     []string{"ip.", "ep."},
			periodsNarrow:          []string{"ip.", "ep."},
			periodsWide:            []string{"ip.", "ep."},
			erasAbbreviated:        []string{"oKr.", "mKr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ovdil Kristus šoddâm", "maŋa Kristus šoddâm"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_789e11f07c8b1dffd2bd03da43d0ea66,
			fnFmtDateLong:         fnFmtDateLong_bc6f92d2d84d639433127ff63687d178,
			fnFmtDateMedium:       fnFmtDateMedium_56894f892061c70a16fdf4ad41cd67e9,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_105c4fe53621c460d37a501cec666cf0,
			fnFmtTimeLong:         fnFmtTimeLong_ed4e4ed1e2c01c9ce9c9f6344a272944,
			fnFmtTimeMedium:       fnFmtTimeMedium_ceaaad7cbbdab7b4a378fe8eff296f4f,
			fnFmtTimeShort:        fnFmtTimeShort_7341725c8ebd7a50d221ac00ecc6aa99,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["smn_fi"] = func() locales.Translator {
		return &localen{

			locale:                 "smn_FI",
			pluralsCardinal:        []locales.PluralRule{2, 3, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "uđiv", "kuovâ", "njuhčâ", "cuáŋui", "vyesi", "kesi", "syeini", "porge", "čohčâ", "roovvâd", "skammâ", "juovlâ"},
			monthsNarrow:           []string{"", "U", "K", "NJ", "C", "V", "K", "S", "P", "Č", "R", "S", "J"},
			monthsWide:             []string{"", "uđđâivemáánu", "kuovâmáánu", "njuhčâmáánu", "cuáŋuimáánu", "vyesimáánu", "kesimáánu", "syeinimáánu", "porgemáánu", "čohčâmáánu", "roovvâdmáánu", "skammâmáánu", "juovlâmáánu"},
			daysAbbreviated:        []string{"pas", "vuo", "maj", "kos", "tuo", "vás", "láv"},
			daysNarrow:             []string{"p", "V", "M", "K", "T", "V", "L"},
			daysShort:              []string{"pa", "vu", "ma", "ko", "tu", "vá", "lá"},
			daysWide:               []string{"pasepeeivi", "vuossaargâ", "majebaargâ", "koskoho", "tuorâstuv", "vástuppeeivi", "lávurduv"},
			periodsAbbreviated:     []string{"ip.", "ep."},
			periodsNarrow:          []string{"ip.", "ep."},
			periodsWide:            []string{"ip.", "ep."},
			erasAbbreviated:        []string{"oKr.", "mKr."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ovdil Kristus šoddâm", "maŋa Kristus šoddâm"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_9b7082aa8f8da3ef64f6e2c0bf7bc90a,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_789e11f07c8b1dffd2bd03da43d0ea66,
			fnFmtDateLong:         fnFmtDateLong_bc6f92d2d84d639433127ff63687d178,
			fnFmtDateMedium:       fnFmtDateMedium_56894f892061c70a16fdf4ad41cd67e9,
			fnFmtDateShort:        fnFmtDateShort_3d7a5686ec224df77035cb4531629cfa,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_105c4fe53621c460d37a501cec666cf0,
			fnFmtTimeLong:         fnFmtTimeLong_ed4e4ed1e2c01c9ce9c9f6344a272944,
			fnFmtTimeMedium:       fnFmtTimeMedium_ceaaad7cbbdab7b4a378fe8eff296f4f,
			fnFmtTimeShort:        fnFmtTimeShort_7341725c8ebd7a50d221ac00ecc6aa99,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sn"] = func() locales.Translator {
		return &localen{

			locale:                 "sn",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ndi", "Kuk", "Kur", "Kub", "Chv", "Chk", "Chg", "Nya", "Gun", "Gum", "Mbu", "Zvi"},
			monthsNarrow:           []string{"", "N", "K", "K", "K", "C", "C", "C", "N", "G", "G", "M", "Z"},
			monthsWide:             []string{"", "Ndira", "Kukadzi", "Kurume", "Kubvumbi", "Chivabvu", "Chikumi", "Chikunguru", "Nyamavhuvhu", "Gunyana", "Gumiguru", "Mbudzi", "Zvita"},
			daysAbbreviated:        []string{"Svo", "Muv", "Chp", "Cht", "Chn", "Chs", "Mug"},
			daysNarrow:             []string{"S", "M", "C", "C", "C", "C", "M"},
			daysShort:              []string{"Sv", "Mu", "Cp", "Ct", "Cn", "Cs", "Mg"},
			daysWide:               []string{"Svondo", "Muvhuro", "Chipiri", "Chitatu", "China", "Chishanu", "Mugovera"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kristo asati auya", "mugore ramambo vedu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sn_zw"] = func() locales.Translator {
		return &localen{

			locale:                 "sn_ZW",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ndi", "Kuk", "Kur", "Kub", "Chv", "Chk", "Chg", "Nya", "Gun", "Gum", "Mbu", "Zvi"},
			monthsNarrow:           []string{"", "N", "K", "K", "K", "C", "C", "C", "N", "G", "G", "M", "Z"},
			monthsWide:             []string{"", "Ndira", "Kukadzi", "Kurume", "Kubvumbi", "Chivabvu", "Chikumi", "Chikunguru", "Nyamavhuvhu", "Gunyana", "Gumiguru", "Mbudzi", "Zvita"},
			daysAbbreviated:        []string{"Svo", "Muv", "Chp", "Cht", "Chn", "Chs", "Mug"},
			daysNarrow:             []string{"S", "M", "C", "C", "C", "C", "M"},
			daysShort:              []string{"Sv", "Mu", "Cp", "Ct", "Cn", "Cs", "Mg"},
			daysWide:               []string{"Svondo", "Muvhuro", "Chipiri", "Chitatu", "China", "Chishanu", "Mugovera"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kristo asati auya", "mugore ramambo vedu"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["so"] = func() locales.Translator {
		return &localen{

			locale:                 "so",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "DBB", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "S", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "May", "Jun", "Lul", "Ogs", "Seb", "Okt", "Nof", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "L", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"},
			daysAbbreviated:        []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysNarrow:             []string{"A", "I", "T", "A", "Kh", "J", "S"},
			daysShort:              []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysWide:               []string{"Axad", "Isniin", "Talaado", "Arbaco", "Khamiis", "Jimco", "Sabti"},
			periodsAbbreviated:     []string{"GH", "GD"},
			periodsNarrow:          []string{"h", "d"},
			periodsWide:            []string{"GH", "GD"},
			erasAbbreviated:        []string{"CH", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ciise Hortii", "Ciise Dabadii"},
			timezones:              map[string]string{"ACDT": "Waqtiga Dharaarta ee Bartamaha Astaraaliya", "ACST": "Waqtiga Caadiga Ah ee Bartamaha Astaraaliya", "ACWDT": "Waqtiga Dharaarta Bartamaha Galbeedka Australiya", "ACWST": "Waqtiga Caadiga Ah ee Bartamaha Galbeedka Astaraaliya", "ADT": "Waqtiga Dharaarta ee Atlantika Waqooyiga Ameerika", "AEDT": "Waqtiga Dharaarta ee Bariga Astaraaliya", "AEST": "Waqtiyada Caadiga ah ee Bariga Astaraaliya", "AKDT": "Waqtiga Dharaarta ee Alaska", "AKST": "Waqtiga Caadiga Ah ee Alaska", "ARST": "Waqtiga Xagaaga ee Arjentiina", "ART": "Waqtiga Caadiga Ah ee Arjentiina", "AST": "Waqtiga Caadiga Ah ee Atlantika Waqooyiga Ameerika", "AWDT": "Waqtiga Dharaarta ee Galbeedka Astaraaliya", "AWST": "Waqtiga Caadiga Ah ee Galbeedka Astaraaliya", "BOT": "Waqtiga Boliifiya", "BT": "Waqtiga Butaan", "CAT": "Waqtiga Bartamaha Afrika", "CDT": "Waqtiga Dharaarta ee Bartamaha Waqooyiga Ameerika", "CHADT": "Waqtiga Dharaarta ee Jaatam", "CHAST": "Waqtiga Caadiga Ah ee Jaatam", "CLST": "Waqtiga Xagaaga ee Jili", "CLT": "Waqtiga Caadiga Ah ee Jili", "COST": "Waqtiga Xagaaga ee Kolambiya", "COT": "Waqtiga Caadiga Ah ee Kolambiya", "CST": "Waqtiga Caadiga Ah ee Bartamaha Waqooyiga Ameerika", "ChST": "Waqtiga Jamoro", "EAT": "Waqtiga Bariga Afrika", "ECT": "Waqtiga Ekuwadoor", "EDT": "Waqtiga Dharaarta ee Bariga Waqooyiga Ameerika", "EST": "Waqtiga Caadiga Ah ee Bariga Waqooyiga Ameerika", "GFT": "Waqtiga Ferenj Guyana", "GMT": "Waqtiga Celceliska Giriinwij", "GST": "Waqtiga Gacanka", "GYT": "Waqtiga Guyaana", "HADT": "Waqtiga Dharaarta ee Hawaay-Alutiyaan", "HAST": "Waqtiga Caadiga Ah ee Hawaay-Alutiyaan", "HAT": "Waqtiga Dharaarta ee Niyuufoonlaan", "HECU": "Waqtiga Dharaarta ee Kuuba", "HEEG": "Waqtiga Xagaaga ee Bariga Giriinlaan", "HENOMX": "Waqtiga Dharaarta ee Waqooyi-Galbeed Meksiko", "HEOG": "Waqtiga Xagaaga ee Galbeedka Giriinlaan", "HEPM": "Waqtiga Dharaarta ee St. Beere & Mikiwelon", "HEPMX": "Waqtiga Dharaarta ee Baasifikada Meksiko", "HKST": "Waqtiga Xagaaga ee Hoong Koong", "HKT": "Waqtiga Caadiga Ah ee Hoong Koong", "HNCU": "Waqtiga Caadiga Ah ee Kuuba", "HNEG": "Waqtiga Caadiga ah ee Bariga Giriinlaan", "HNNOMX": "Waqtiga Caadiga Ah ee Waqooyi-Galbeed Meksiko", "HNOG": "Waqtiga Caadiga Ah ee Galbeedka Giriinlaan", "HNPM": "Waqtiga Caadiga Ah St. Beere & Mikiwelon", "HNPMX": "Waqtiga Caadiga Ah ee Baasifikada Meksiko", "HNT": "Waqtiga Caadiga Ah ee Niyuufoonlaan", "IST": "Waqtiga Caadiga Ah ee Hindiya", "JDT": "Waqtiga Dharaarta ee Jabaan", "JST": "Waqtiga Caadiga Ah ee Jabaan", "LHDT": "Waqtiga Dharaarta ee Lod How", "LHST": "Waqtiga Caadiga Ah ee Lod How", "MDT": "Waqtiga Dharaarta ee Buurleyda Waqooyiga Ameerika", "MESZ": "Waqtiga Xagaaga ee Bartamaha Yurub", "MEZ": "Waqtiga Caadiga Ah ee Bartamaha Yurub", "MST": "Waqtiga Caadiga ah ee Buuraleyda Waqooyiga Ameerika", "MYT": "Waqtiga Maleyshiya", "NZDT": "Waqtiga Dharaarta ee Niyuu Si’laan", "NZST": "Waqtiga Caadiga Ah ee Niyuu Si’laan", "OESZ": "Waqtiga Xagaaga ee Bariga Yurub", "OEZ": "Waqtiga Caadiga Ah ee Bariga Yurub", "PDT": "Waqtiga Dharaarta ee Basifika Waqooyiga Ameerika", "PST": "Waqtiga Caadiga ah ee Basifika Waqooyiga Ameerika", "SAST": "Waqtiyada Caadiga Ah ee Koonfur Afrika", "SGT": "Waqtiga Singabuur", "SRT": "Waqtiga Surineym", "TMST": "Waqtiga Xagaaga ee Turkmenistan", "TMT": "Waqtiga Caadiga Ah ee Turkmenistan", "UYST": "Waqtiga Xagaaga ee Urugwaay", "UYT": "Waqtiga Caadiga Ah ee Urugwaay", "VET": "Waqtiga Fenezuweela", "WARST": "Waqtiga Xagaaga ee Galbeedka Arjentiina", "WART": "Waqtiga Caadiga Ah ee Galbeedka Arjentiina", "WAST": "Waqtiga Xagaaga ee Galbeedka Afrika", "WAT": "Waqtiga Caadiga Ah ee Galbeedka Afrika", "WESZ": "Waqtiga Xagaaga ee Galbeedka Yurub", "WEZ": "Waqtiga Caadiga Ah ee Galbeedka Yurub", "WIB": "Waqtiga Galbeedka Indoneeysiya", "WIT": "Waqtiga Indoneeysiya", "WITA": "Waqtiga Bartamaha Indoneeysiya", "∅∅∅": "Waqtiga Xagaaga ee Asores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_3c1131a653a5e918c3e9381647860bcf,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["so_dj"] = func() locales.Translator {
		return &localen{

			locale:                 "so_DJ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "Fdj", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "May", "Jun", "Lul", "Ogs", "Seb", "Okt", "Nof", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "L", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"},
			daysAbbreviated:        []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysNarrow:             []string{"A", "I", "T", "A", "Kh", "J", "S"},
			daysShort:              []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysWide:               []string{"Axad", "Isniin", "Talaado", "Arbaco", "Khamiis", "Jimco", "Sabti"},
			periodsAbbreviated:     []string{"GH", "GD"},
			periodsNarrow:          []string{"h", "d"},
			periodsWide:            []string{"GH", "GD"},
			erasAbbreviated:        []string{"CH", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ciise Hortii", "Ciise Dabadii"},
			timezones:              map[string]string{"ACDT": "Waqtiga Dharaarta ee Bartamaha Astaraaliya", "ACST": "Waqtiga Caadiga Ah ee Bartamaha Astaraaliya", "ACWDT": "Waqtiga Dharaarta Bartamaha Galbeedka Australiya", "ACWST": "Waqtiga Caadiga Ah ee Bartamaha Galbeedka Astaraaliya", "ADT": "Waqtiga Dharaarta ee Atlantika Waqooyiga Ameerika", "AEDT": "Waqtiga Dharaarta ee Bariga Astaraaliya", "AEST": "Waqtiyada Caadiga ah ee Bariga Astaraaliya", "AKDT": "Waqtiga Dharaarta ee Alaska", "AKST": "Waqtiga Caadiga Ah ee Alaska", "ARST": "Waqtiga Xagaaga ee Arjentiina", "ART": "Waqtiga Caadiga Ah ee Arjentiina", "AST": "Waqtiga Caadiga Ah ee Atlantika Waqooyiga Ameerika", "AWDT": "Waqtiga Dharaarta ee Galbeedka Astaraaliya", "AWST": "Waqtiga Caadiga Ah ee Galbeedka Astaraaliya", "BOT": "Waqtiga Boliifiya", "BT": "Waqtiga Butaan", "CAT": "Waqtiga Bartamaha Afrika", "CDT": "Waqtiga Dharaarta ee Bartamaha Waqooyiga Ameerika", "CHADT": "Waqtiga Dharaarta ee Jaatam", "CHAST": "Waqtiga Caadiga Ah ee Jaatam", "CLST": "Waqtiga Xagaaga ee Jili", "CLT": "Waqtiga Caadiga Ah ee Jili", "COST": "Waqtiga Xagaaga ee Kolambiya", "COT": "Waqtiga Caadiga Ah ee Kolambiya", "CST": "Waqtiga Caadiga Ah ee Bartamaha Waqooyiga Ameerika", "ChST": "Waqtiga Jamoro", "EAT": "Waqtiga Bariga Afrika", "ECT": "Waqtiga Ekuwadoor", "EDT": "Waqtiga Dharaarta ee Bariga Waqooyiga Ameerika", "EST": "Waqtiga Caadiga Ah ee Bariga Waqooyiga Ameerika", "GFT": "Waqtiga Ferenj Guyana", "GMT": "Waqtiga Celceliska Giriinwij", "GST": "Waqtiga Gacanka", "GYT": "Waqtiga Guyaana", "HADT": "Waqtiga Dharaarta ee Hawaay-Alutiyaan", "HAST": "Waqtiga Caadiga Ah ee Hawaay-Alutiyaan", "HAT": "Waqtiga Dharaarta ee Niyuufoonlaan", "HECU": "Waqtiga Dharaarta ee Kuuba", "HEEG": "Waqtiga Xagaaga ee Bariga Giriinlaan", "HENOMX": "Waqtiga Dharaarta ee Waqooyi-Galbeed Meksiko", "HEOG": "Waqtiga Xagaaga ee Galbeedka Giriinlaan", "HEPM": "Waqtiga Dharaarta ee St. Beere & Mikiwelon", "HEPMX": "Waqtiga Dharaarta ee Baasifikada Meksiko", "HKST": "Waqtiga Xagaaga ee Hoong Koong", "HKT": "Waqtiga Caadiga Ah ee Hoong Koong", "HNCU": "Waqtiga Caadiga Ah ee Kuuba", "HNEG": "Waqtiga Caadiga ah ee Bariga Giriinlaan", "HNNOMX": "Waqtiga Caadiga Ah ee Waqooyi-Galbeed Meksiko", "HNOG": "Waqtiga Caadiga Ah ee Galbeedka Giriinlaan", "HNPM": "Waqtiga Caadiga Ah St. Beere & Mikiwelon", "HNPMX": "Waqtiga Caadiga Ah ee Baasifikada Meksiko", "HNT": "Waqtiga Caadiga Ah ee Niyuufoonlaan", "IST": "Waqtiga Caadiga Ah ee Hindiya", "JDT": "Waqtiga Dharaarta ee Jabaan", "JST": "Waqtiga Caadiga Ah ee Jabaan", "LHDT": "Waqtiga Dharaarta ee Lod How", "LHST": "Waqtiga Caadiga Ah ee Lod How", "MDT": "Waqtiga Dharaarta ee Buurleyda Waqooyiga Ameerika", "MESZ": "Waqtiga Xagaaga ee Bartamaha Yurub", "MEZ": "Waqtiga Caadiga Ah ee Bartamaha Yurub", "MST": "Waqtiga Caadiga ah ee Buuraleyda Waqooyiga Ameerika", "MYT": "Waqtiga Maleyshiya", "NZDT": "Waqtiga Dharaarta ee Niyuu Si’laan", "NZST": "Waqtiga Caadiga Ah ee Niyuu Si’laan", "OESZ": "Waqtiga Xagaaga ee Bariga Yurub", "OEZ": "Waqtiga Caadiga Ah ee Bariga Yurub", "PDT": "Waqtiga Dharaarta ee Basifika Waqooyiga Ameerika", "PST": "Waqtiga Caadiga ah ee Basifika Waqooyiga Ameerika", "SAST": "Waqtiyada Caadiga Ah ee Koonfur Afrika", "SGT": "Waqtiga Singabuur", "SRT": "Waqtiga Surineym", "TMST": "Waqtiga Xagaaga ee Turkmenistan", "TMT": "Waqtiga Caadiga Ah ee Turkmenistan", "UYST": "Waqtiga Xagaaga ee Urugwaay", "UYT": "Waqtiga Caadiga Ah ee Urugwaay", "VET": "Waqtiga Fenezuweela", "WARST": "Waqtiga Xagaaga ee Galbeedka Arjentiina", "WART": "Waqtiga Caadiga Ah ee Galbeedka Arjentiina", "WAST": "Waqtiga Xagaaga ee Galbeedka Afrika", "WAT": "Waqtiga Caadiga Ah ee Galbeedka Afrika", "WESZ": "Waqtiga Xagaaga ee Galbeedka Yurub", "WEZ": "Waqtiga Caadiga Ah ee Galbeedka Yurub", "WIB": "Waqtiga Galbeedka Indoneeysiya", "WIT": "Waqtiga Indoneeysiya", "WITA": "Waqtiga Bartamaha Indoneeysiya", "∅∅∅": "Waqtiga Xagaaga ee Beeru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_3c1131a653a5e918c3e9381647860bcf,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["so_et"] = func() locales.Translator {
		return &localen{

			locale:                 "so_ET",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "Br", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "May", "Jun", "Lul", "Ogs", "Seb", "Okt", "Nof", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "L", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"},
			daysAbbreviated:        []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysNarrow:             []string{"A", "I", "T", "A", "Kh", "J", "S"},
			daysShort:              []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysWide:               []string{"Axad", "Isniin", "Talaado", "Arbaco", "Khamiis", "Jimco", "Sabti"},
			periodsAbbreviated:     []string{"GH", "GD"},
			periodsNarrow:          []string{"h", "d"},
			periodsWide:            []string{"GH", "GD"},
			erasAbbreviated:        []string{"CH", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ciise Hortii", "Ciise Dabadii"},
			timezones:              map[string]string{"ACDT": "Waqtiga Dharaarta ee Bartamaha Astaraaliya", "ACST": "Waqtiga Caadiga Ah ee Bartamaha Astaraaliya", "ACWDT": "Waqtiga Dharaarta Bartamaha Galbeedka Australiya", "ACWST": "Waqtiga Caadiga Ah ee Bartamaha Galbeedka Astaraaliya", "ADT": "Waqtiga Dharaarta ee Atlantika Waqooyiga Ameerika", "AEDT": "Waqtiga Dharaarta ee Bariga Astaraaliya", "AEST": "Waqtiyada Caadiga ah ee Bariga Astaraaliya", "AKDT": "Waqtiga Dharaarta ee Alaska", "AKST": "Waqtiga Caadiga Ah ee Alaska", "ARST": "Waqtiga Xagaaga ee Arjentiina", "ART": "Waqtiga Caadiga Ah ee Arjentiina", "AST": "Waqtiga Caadiga Ah ee Atlantika Waqooyiga Ameerika", "AWDT": "Waqtiga Dharaarta ee Galbeedka Astaraaliya", "AWST": "Waqtiga Caadiga Ah ee Galbeedka Astaraaliya", "BOT": "Waqtiga Boliifiya", "BT": "Waqtiga Butaan", "CAT": "Waqtiga Bartamaha Afrika", "CDT": "Waqtiga Dharaarta ee Bartamaha Waqooyiga Ameerika", "CHADT": "Waqtiga Dharaarta ee Jaatam", "CHAST": "Waqtiga Caadiga Ah ee Jaatam", "CLST": "Waqtiga Xagaaga ee Jili", "CLT": "Waqtiga Caadiga Ah ee Jili", "COST": "Waqtiga Xagaaga ee Kolambiya", "COT": "Waqtiga Caadiga Ah ee Kolambiya", "CST": "Waqtiga Caadiga Ah ee Bartamaha Waqooyiga Ameerika", "ChST": "Waqtiga Jamoro", "EAT": "Waqtiga Bariga Afrika", "ECT": "Waqtiga Ekuwadoor", "EDT": "Waqtiga Dharaarta ee Bariga Waqooyiga Ameerika", "EST": "Waqtiga Caadiga Ah ee Bariga Waqooyiga Ameerika", "GFT": "Waqtiga Ferenj Guyana", "GMT": "Waqtiga Celceliska Giriinwij", "GST": "Waqtiga Gacanka", "GYT": "Waqtiga Guyaana", "HADT": "Waqtiga Dharaarta ee Hawaay-Alutiyaan", "HAST": "Waqtiga Caadiga Ah ee Hawaay-Alutiyaan", "HAT": "Waqtiga Dharaarta ee Niyuufoonlaan", "HECU": "Waqtiga Dharaarta ee Kuuba", "HEEG": "Waqtiga Xagaaga ee Bariga Giriinlaan", "HENOMX": "Waqtiga Dharaarta ee Waqooyi-Galbeed Meksiko", "HEOG": "Waqtiga Xagaaga ee Galbeedka Giriinlaan", "HEPM": "Waqtiga Dharaarta ee St. Beere & Mikiwelon", "HEPMX": "Waqtiga Dharaarta ee Baasifikada Meksiko", "HKST": "Waqtiga Xagaaga ee Hoong Koong", "HKT": "Waqtiga Caadiga Ah ee Hoong Koong", "HNCU": "Waqtiga Caadiga Ah ee Kuuba", "HNEG": "Waqtiga Caadiga ah ee Bariga Giriinlaan", "HNNOMX": "Waqtiga Caadiga Ah ee Waqooyi-Galbeed Meksiko", "HNOG": "Waqtiga Caadiga Ah ee Galbeedka Giriinlaan", "HNPM": "Waqtiga Caadiga Ah St. Beere & Mikiwelon", "HNPMX": "Waqtiga Caadiga Ah ee Baasifikada Meksiko", "HNT": "Waqtiga Caadiga Ah ee Niyuufoonlaan", "IST": "Waqtiga Caadiga Ah ee Hindiya", "JDT": "Waqtiga Dharaarta ee Jabaan", "JST": "Waqtiga Caadiga Ah ee Jabaan", "LHDT": "Waqtiga Dharaarta ee Lod How", "LHST": "Waqtiga Caadiga Ah ee Lod How", "MDT": "Waqtiga Dharaarta ee Buurleyda Waqooyiga Ameerika", "MESZ": "Waqtiga Xagaaga ee Bartamaha Yurub", "MEZ": "Waqtiga Caadiga Ah ee Bartamaha Yurub", "MST": "Waqtiga Caadiga ah ee Buuraleyda Waqooyiga Ameerika", "MYT": "Waqtiga Maleyshiya", "NZDT": "Waqtiga Dharaarta ee Niyuu Si’laan", "NZST": "Waqtiga Caadiga Ah ee Niyuu Si’laan", "OESZ": "Waqtiga Xagaaga ee Bariga Yurub", "OEZ": "Waqtiga Caadiga Ah ee Bariga Yurub", "PDT": "Waqtiga Dharaarta ee Basifika Waqooyiga Ameerika", "PST": "Waqtiga Caadiga ah ee Basifika Waqooyiga Ameerika", "SAST": "Waqtiyada Caadiga Ah ee Koonfur Afrika", "SGT": "Waqtiga Singabuur", "SRT": "Waqtiga Surineym", "TMST": "Waqtiga Xagaaga ee Turkmenistan", "TMT": "Waqtiga Caadiga Ah ee Turkmenistan", "UYST": "Waqtiga Xagaaga ee Urugwaay", "UYT": "Waqtiga Caadiga Ah ee Urugwaay", "VET": "Waqtiga Fenezuweela", "WARST": "Waqtiga Xagaaga ee Galbeedka Arjentiina", "WART": "Waqtiga Caadiga Ah ee Galbeedka Arjentiina", "WAST": "Waqtiga Xagaaga ee Galbeedka Afrika", "WAT": "Waqtiga Caadiga Ah ee Galbeedka Afrika", "WESZ": "Waqtiga Xagaaga ee Galbeedka Yurub", "WEZ": "Waqtiga Caadiga Ah ee Galbeedka Yurub", "WIB": "Waqtiga Galbeedka Indoneeysiya", "WIT": "Waqtiga Indoneeysiya", "WITA": "Waqtiga Bartamaha Indoneeysiya", "∅∅∅": "Waqtiga Xagaaga ee Baraasiliya"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_3c1131a653a5e918c3e9381647860bcf,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["so_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "so_KE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "May", "Jun", "Lul", "Ogs", "Seb", "Okt", "Nof", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "L", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"},
			daysAbbreviated:        []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysNarrow:             []string{"A", "I", "T", "A", "Kh", "J", "S"},
			daysShort:              []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysWide:               []string{"Axad", "Isniin", "Talaado", "Arbaco", "Khamiis", "Jimco", "Sabti"},
			periodsAbbreviated:     []string{"GH", "GD"},
			periodsNarrow:          []string{"h", "d"},
			periodsWide:            []string{"GH", "GD"},
			erasAbbreviated:        []string{"CH", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ciise Hortii", "Ciise Dabadii"},
			timezones:              map[string]string{"ACDT": "Waqtiga Dharaarta ee Bartamaha Astaraaliya", "ACST": "Waqtiga Caadiga Ah ee Bartamaha Astaraaliya", "ACWDT": "Waqtiga Dharaarta Bartamaha Galbeedka Australiya", "ACWST": "Waqtiga Caadiga Ah ee Bartamaha Galbeedka Astaraaliya", "ADT": "Waqtiga Dharaarta ee Atlantika Waqooyiga Ameerika", "AEDT": "Waqtiga Dharaarta ee Bariga Astaraaliya", "AEST": "Waqtiyada Caadiga ah ee Bariga Astaraaliya", "AKDT": "Waqtiga Dharaarta ee Alaska", "AKST": "Waqtiga Caadiga Ah ee Alaska", "ARST": "Waqtiga Xagaaga ee Arjentiina", "ART": "Waqtiga Caadiga Ah ee Arjentiina", "AST": "Waqtiga Caadiga Ah ee Atlantika Waqooyiga Ameerika", "AWDT": "Waqtiga Dharaarta ee Galbeedka Astaraaliya", "AWST": "Waqtiga Caadiga Ah ee Galbeedka Astaraaliya", "BOT": "Waqtiga Boliifiya", "BT": "Waqtiga Butaan", "CAT": "Waqtiga Bartamaha Afrika", "CDT": "Waqtiga Dharaarta ee Bartamaha Waqooyiga Ameerika", "CHADT": "Waqtiga Dharaarta ee Jaatam", "CHAST": "Waqtiga Caadiga Ah ee Jaatam", "CLST": "Waqtiga Xagaaga ee Jili", "CLT": "Waqtiga Caadiga Ah ee Jili", "COST": "Waqtiga Xagaaga ee Kolambiya", "COT": "Waqtiga Caadiga Ah ee Kolambiya", "CST": "Waqtiga Caadiga Ah ee Bartamaha Waqooyiga Ameerika", "ChST": "Waqtiga Jamoro", "EAT": "Waqtiga Bariga Afrika", "ECT": "Waqtiga Ekuwadoor", "EDT": "Waqtiga Dharaarta ee Bariga Waqooyiga Ameerika", "EST": "Waqtiga Caadiga Ah ee Bariga Waqooyiga Ameerika", "GFT": "Waqtiga Ferenj Guyana", "GMT": "Waqtiga Celceliska Giriinwij", "GST": "Waqtiga Gacanka", "GYT": "Waqtiga Guyaana", "HADT": "Waqtiga Dharaarta ee Hawaay-Alutiyaan", "HAST": "Waqtiga Caadiga Ah ee Hawaay-Alutiyaan", "HAT": "Waqtiga Dharaarta ee Niyuufoonlaan", "HECU": "Waqtiga Dharaarta ee Kuuba", "HEEG": "Waqtiga Xagaaga ee Bariga Giriinlaan", "HENOMX": "Waqtiga Dharaarta ee Waqooyi-Galbeed Meksiko", "HEOG": "Waqtiga Xagaaga ee Galbeedka Giriinlaan", "HEPM": "Waqtiga Dharaarta ee St. Beere & Mikiwelon", "HEPMX": "Waqtiga Dharaarta ee Baasifikada Meksiko", "HKST": "Waqtiga Xagaaga ee Hoong Koong", "HKT": "Waqtiga Caadiga Ah ee Hoong Koong", "HNCU": "Waqtiga Caadiga Ah ee Kuuba", "HNEG": "Waqtiga Caadiga ah ee Bariga Giriinlaan", "HNNOMX": "Waqtiga Caadiga Ah ee Waqooyi-Galbeed Meksiko", "HNOG": "Waqtiga Caadiga Ah ee Galbeedka Giriinlaan", "HNPM": "Waqtiga Caadiga Ah St. Beere & Mikiwelon", "HNPMX": "Waqtiga Caadiga Ah ee Baasifikada Meksiko", "HNT": "Waqtiga Caadiga Ah ee Niyuufoonlaan", "IST": "Waqtiga Caadiga Ah ee Hindiya", "JDT": "Waqtiga Dharaarta ee Jabaan", "JST": "Waqtiga Caadiga Ah ee Jabaan", "LHDT": "Waqtiga Dharaarta ee Lod How", "LHST": "Waqtiga Caadiga Ah ee Lod How", "MDT": "Waqtiga Dharaarta ee Buurleyda Waqooyiga Ameerika", "MESZ": "Waqtiga Xagaaga ee Bartamaha Yurub", "MEZ": "Waqtiga Caadiga Ah ee Bartamaha Yurub", "MST": "Waqtiga Caadiga ah ee Buuraleyda Waqooyiga Ameerika", "MYT": "Waqtiga Maleyshiya", "NZDT": "Waqtiga Dharaarta ee Niyuu Si’laan", "NZST": "Waqtiga Caadiga Ah ee Niyuu Si’laan", "OESZ": "Waqtiga Xagaaga ee Bariga Yurub", "OEZ": "Waqtiga Caadiga Ah ee Bariga Yurub", "PDT": "Waqtiga Dharaarta ee Basifika Waqooyiga Ameerika", "PST": "Waqtiga Caadiga ah ee Basifika Waqooyiga Ameerika", "SAST": "Waqtiyada Caadiga Ah ee Koonfur Afrika", "SGT": "Waqtiga Singabuur", "SRT": "Waqtiga Surineym", "TMST": "Waqtiga Xagaaga ee Turkmenistan", "TMT": "Waqtiga Caadiga Ah ee Turkmenistan", "UYST": "Waqtiga Xagaaga ee Urugwaay", "UYT": "Waqtiga Caadiga Ah ee Urugwaay", "VET": "Waqtiga Fenezuweela", "WARST": "Waqtiga Xagaaga ee Galbeedka Arjentiina", "WART": "Waqtiga Caadiga Ah ee Galbeedka Arjentiina", "WAST": "Waqtiga Xagaaga ee Galbeedka Afrika", "WAT": "Waqtiga Caadiga Ah ee Galbeedka Afrika", "WESZ": "Waqtiga Xagaaga ee Galbeedka Yurub", "WEZ": "Waqtiga Caadiga Ah ee Galbeedka Yurub", "WIB": "Waqtiga Galbeedka Indoneeysiya", "WIT": "Waqtiga Indoneeysiya", "WITA": "Waqtiga Bartamaha Indoneeysiya", "∅∅∅": "Waqtiga Xagaaga ee Asores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_3c1131a653a5e918c3e9381647860bcf,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["so_so"] = func() locales.Translator {
		return &localen{

			locale:                 "so_SO",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Abr", "May", "Jun", "Lul", "Ogs", "Seb", "Okt", "Nof", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "L", "O", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"},
			daysAbbreviated:        []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysNarrow:             []string{"A", "I", "T", "A", "Kh", "J", "S"},
			daysShort:              []string{"Axd", "Isn", "Tldo", "Arbc", "Khms", "Jmc", "Sbti"},
			daysWide:               []string{"Axad", "Isniin", "Talaado", "Arbaco", "Khamiis", "Jimco", "Sabti"},
			periodsAbbreviated:     []string{"GH", "GD"},
			periodsNarrow:          []string{"h", "d"},
			periodsWide:            []string{"GH", "GD"},
			erasAbbreviated:        []string{"CH", "CD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Ciise Hortii", "Ciise Dabadii"},
			timezones:              map[string]string{"ACDT": "Waqtiga Dharaarta ee Bartamaha Astaraaliya", "ACST": "Waqtiga Caadiga Ah ee Bartamaha Astaraaliya", "ACWDT": "Waqtiga Dharaarta Bartamaha Galbeedka Australiya", "ACWST": "Waqtiga Caadiga Ah ee Bartamaha Galbeedka Astaraaliya", "ADT": "Waqtiga Dharaarta ee Atlantika Waqooyiga Ameerika", "AEDT": "Waqtiga Dharaarta ee Bariga Astaraaliya", "AEST": "Waqtiyada Caadiga ah ee Bariga Astaraaliya", "AKDT": "Waqtiga Dharaarta ee Alaska", "AKST": "Waqtiga Caadiga Ah ee Alaska", "ARST": "Waqtiga Xagaaga ee Arjentiina", "ART": "Waqtiga Caadiga Ah ee Arjentiina", "AST": "Waqtiga Caadiga Ah ee Atlantika Waqooyiga Ameerika", "AWDT": "Waqtiga Dharaarta ee Galbeedka Astaraaliya", "AWST": "Waqtiga Caadiga Ah ee Galbeedka Astaraaliya", "BOT": "Waqtiga Boliifiya", "BT": "Waqtiga Butaan", "CAT": "Waqtiga Bartamaha Afrika", "CDT": "Waqtiga Dharaarta ee Bartamaha Waqooyiga Ameerika", "CHADT": "Waqtiga Dharaarta ee Jaatam", "CHAST": "Waqtiga Caadiga Ah ee Jaatam", "CLST": "Waqtiga Xagaaga ee Jili", "CLT": "Waqtiga Caadiga Ah ee Jili", "COST": "Waqtiga Xagaaga ee Kolambiya", "COT": "Waqtiga Caadiga Ah ee Kolambiya", "CST": "Waqtiga Caadiga Ah ee Bartamaha Waqooyiga Ameerika", "ChST": "Waqtiga Jamoro", "EAT": "Waqtiga Bariga Afrika", "ECT": "Waqtiga Ekuwadoor", "EDT": "Waqtiga Dharaarta ee Bariga Waqooyiga Ameerika", "EST": "Waqtiga Caadiga Ah ee Bariga Waqooyiga Ameerika", "GFT": "Waqtiga Ferenj Guyana", "GMT": "Waqtiga Celceliska Giriinwij", "GST": "Waqtiga Gacanka", "GYT": "Waqtiga Guyaana", "HADT": "Waqtiga Dharaarta ee Hawaay-Alutiyaan", "HAST": "Waqtiga Caadiga Ah ee Hawaay-Alutiyaan", "HAT": "Waqtiga Dharaarta ee Niyuufoonlaan", "HECU": "Waqtiga Dharaarta ee Kuuba", "HEEG": "Waqtiga Xagaaga ee Bariga Giriinlaan", "HENOMX": "Waqtiga Dharaarta ee Waqooyi-Galbeed Meksiko", "HEOG": "Waqtiga Xagaaga ee Galbeedka Giriinlaan", "HEPM": "Waqtiga Dharaarta ee St. Beere & Mikiwelon", "HEPMX": "Waqtiga Dharaarta ee Baasifikada Meksiko", "HKST": "Waqtiga Xagaaga ee Hoong Koong", "HKT": "Waqtiga Caadiga Ah ee Hoong Koong", "HNCU": "Waqtiga Caadiga Ah ee Kuuba", "HNEG": "Waqtiga Caadiga ah ee Bariga Giriinlaan", "HNNOMX": "Waqtiga Caadiga Ah ee Waqooyi-Galbeed Meksiko", "HNOG": "Waqtiga Caadiga Ah ee Galbeedka Giriinlaan", "HNPM": "Waqtiga Caadiga Ah St. Beere & Mikiwelon", "HNPMX": "Waqtiga Caadiga Ah ee Baasifikada Meksiko", "HNT": "Waqtiga Caadiga Ah ee Niyuufoonlaan", "IST": "Waqtiga Caadiga Ah ee Hindiya", "JDT": "Waqtiga Dharaarta ee Jabaan", "JST": "Waqtiga Caadiga Ah ee Jabaan", "LHDT": "Waqtiga Dharaarta ee Lod How", "LHST": "Waqtiga Caadiga Ah ee Lod How", "MDT": "Waqtiga Dharaarta ee Buurleyda Waqooyiga Ameerika", "MESZ": "Waqtiga Xagaaga ee Bartamaha Yurub", "MEZ": "Waqtiga Caadiga Ah ee Bartamaha Yurub", "MST": "Waqtiga Caadiga ah ee Buuraleyda Waqooyiga Ameerika", "MYT": "Waqtiga Maleyshiya", "NZDT": "Waqtiga Dharaarta ee Niyuu Si’laan", "NZST": "Waqtiga Caadiga Ah ee Niyuu Si’laan", "OESZ": "Waqtiga Xagaaga ee Bariga Yurub", "OEZ": "Waqtiga Caadiga Ah ee Bariga Yurub", "PDT": "Waqtiga Dharaarta ee Basifika Waqooyiga Ameerika", "PST": "Waqtiga Caadiga ah ee Basifika Waqooyiga Ameerika", "SAST": "Waqtiyada Caadiga Ah ee Koonfur Afrika", "SGT": "Waqtiga Singabuur", "SRT": "Waqtiga Surineym", "TMST": "Waqtiga Xagaaga ee Turkmenistan", "TMT": "Waqtiga Caadiga Ah ee Turkmenistan", "UYST": "Waqtiga Xagaaga ee Urugwaay", "UYT": "Waqtiga Caadiga Ah ee Urugwaay", "VET": "Waqtiga Fenezuweela", "WARST": "Waqtiga Xagaaga ee Galbeedka Arjentiina", "WART": "Waqtiga Caadiga Ah ee Galbeedka Arjentiina", "WAST": "Waqtiga Xagaaga ee Galbeedka Afrika", "WAT": "Waqtiga Caadiga Ah ee Galbeedka Afrika", "WESZ": "Waqtiga Xagaaga ee Galbeedka Yurub", "WEZ": "Waqtiga Caadiga Ah ee Galbeedka Yurub", "WIB": "Waqtiga Galbeedka Indoneeysiya", "WIT": "Waqtiga Indoneeysiya", "WITA": "Waqtiga Bartamaha Indoneeysiya", "∅∅∅": "Waqtiga Xagaaga ee Baraasiliya"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_3c1131a653a5e918c3e9381647860bcf,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sq"] = func() locales.Translator {
		return &localen{

			locale:                 "sq",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "Lekë", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "shk", "mar", "pri", "maj", "qer", "korr", "gush", "sht", "tet", "nën", "dhj"},
			monthsNarrow:           []string{"", "j", "sh", "m", "p", "m", "q", "k", "g", "sh", "t", "n", "dh"},
			monthsWide:             []string{"", "janar", "shkurt", "mars", "prill", "maj", "qershor", "korrik", "gusht", "shtator", "tetor", "nëntor", "dhjetor"},
			daysAbbreviated:        []string{"Die", "Hën", "Mar", "Mër", "Enj", "Pre", "Sht"},
			daysNarrow:             []string{"d", "h", "m", "m", "e", "p", "sh"},
			daysShort:              []string{"die", "hën", "mar", "mër", "enj", "pre", "sht"},
			daysWide:               []string{"e diel", "e hënë", "e martë", "e mërkurë", "e enjte", "e premte", "e shtunë"},
			periodsAbbreviated:     []string{"p.d.", "m.d."},
			periodsNarrow:          []string{"p.d.", "m.d."},
			periodsWide:            []string{"e paradites", "e pasdites"},
			erasAbbreviated:        []string{"p.K.", "mb.K."},
			erasNarrow:             []string{"p.K.", "mb.K."},
			erasWide:               []string{"para Krishtit", "mbas Krishtit"},
			timezones:              map[string]string{"ACDT": "Ora verore e Australisë Qendrore", "ACST": "Ora standarde e Australisë Qendrore", "ACWDT": "Ora verore e Australisë Qendroro-Perëndimore", "ACWST": "Ora standarde e Australisë Qendroro-Perëndimore", "ADT": "Ora verore e Atlantikut", "AEDT": "Ora verore e Australisë Lindore", "AEST": "Ora standarde e Australisë Lindore", "AKDT": "Ora verore e Alaskës", "AKST": "Ora standarde e Alaskës", "ARST": "Ora verore e Argjentinës", "ART": "Ora standarde e Argjentinës", "AST": "Ora standarde e Atlantikut", "AWDT": "Ora verore e Australisë Perëndimore", "AWST": "Ora standarde e Australisë Perëndimore", "BOT": "Ora e Bolivisë", "BT": "Ora e Butanit", "CAT": "Ora e Afrikës Qendrore", "CDT": "Ora verore e SHBA-së Qendrore", "CHADT": "Ora verore e Katamit", "CHAST": "Ora standarde e Katamit", "CLST": "Ora verore e Kilit", "CLT": "Ora standarde e Kilit", "COST": "Ora verore e Kolumbisë", "COT": "Ora standarde e Kolumbisë", "CST": "Ora standarde e SHBA-së Qendrore", "ChST": "Ora e Kamorros", "EAT": "Ora e Afrikës Lindore", "ECT": "Ora e Ekuadorit", "EDT": "Ora verore e SHBA-së Lindore", "EST": "Ora standarde e SHBA-së Lindore", "GFT": "Ora e Guajanës Franceze", "GMT": "Ora e Grinuiçit", "GST": "Ora e Gjirit", "GYT": "Ora e Guajanës", "HADT": "Ora verore e Ishujve Hauai-Aleutian", "HAST": "Ora standarde e Ishujve Hauai-Aleutian", "HAT": "Ora verore e Njufaundlendit [Tokës së Re]", "HECU": "Ora verore e Kubës", "HEEG": "Ora verore e Grenlandës Lindore", "HENOMX": "Ora verore e Meksikës Veriperëndimore", "HEOG": "Ora verore e Grënlandës Perëndimore", "HEPM": "Ora verore e Shën-Pier dhe Mikelon", "HEPMX": "Ora verore e Territoreve Meksikane të Bregut të Paqësorit", "HKST": "Ora verore e Hong-Kongut", "HKT": "Ora standarde e Hong-Kongut", "HNCU": "Ora standarde e Kubës", "HNEG": "Ora standarde e Grenlandës Lindore", "HNNOMX": "Ora standarde e Meksikës Veriperëndimore", "HNOG": "Ora standarde e Grënlandës Perëndimore", "HNPM": "Ora standarde e Shën-Pier dhe Mikelon", "HNPMX": "Ora standarde e Territoreve Meksikane të Bregut të Paqësorit", "HNT": "Ora standarde e Njufaundlendit [Tokës së Re]", "IST": "Ora standarde e Indisë", "JDT": "Ora verore e Japonisë", "JST": "Ora standarde e Japonisë", "LHDT": "Ora verore e Lord-Houit", "LHST": "Ora standarde e Lord-Houit", "MDT": "Ora verore e Territoreve Amerikane të Brezit Malor", "MESZ": "Ora verore e Evropës Qendrore", "MEZ": "Ora standarde e Evropës Qendrore", "MST": "Ora standarde e Territoreve Amerikane të Brezit Malor", "MYT": "Ora e Malajzisë", "NZDT": "Ora verore e Zelandës së Re", "NZST": "Ora standarde e Zelandës së Re", "OESZ": "Ora verore e Evropës Lindore", "OEZ": "Ora standarde e Evropës Lindore", "PDT": "Ora verore e Territoreve Amerikane të Bregut të Paqësorit", "PST": "Ora standarde e Territoreve Amerikane të Bregut të Paqësorit", "SAST": "Ora standarde e Afrikës Jugore", "SGT": "Ora e Singaporit", "SRT": "Ora e Surinamit", "TMST": "Ora verore e Turkmenistanit", "TMT": "Ora standarde e Turkmenistanit", "UYST": "Ora verore e Uruguait", "UYT": "Ora standarde e Uruguait", "VET": "Ora e Venezuelës", "WARST": "Ora verore e Argjentinës Perëndimore", "WART": "Ora standarde e Argjentinës Perëndimore", "WAST": "Ora verore e Afrikës Perëndimore", "WAT": "Ora standarde e Afrikës Perëndimore", "WESZ": "Ora verore e Evropës Perëndimore", "WEZ": "Ora standarde e Evropës Perëndimore", "WIB": "Ora e Indonezisë Perëndimore", "WIT": "Ora e Indonezisë Lindore", "WITA": "Ora e Indonezisë Qendrore", "∅∅∅": "Ora verore e Brazilisë"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_922dbe895fd3d466f77caf876ca0d4af,
			fnFmtTimeLong:         fnFmtTimeLong_880840264062db9197a5f9820877cc61,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_8b495a08240ebf51a9ad31ffd7bf6c66,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sq_al"] = func() locales.Translator {
		return &localen{

			locale:                 "sq_AL",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "shk", "mar", "pri", "maj", "qer", "korr", "gush", "sht", "tet", "nën", "dhj"},
			monthsNarrow:           []string{"", "j", "sh", "m", "p", "m", "q", "k", "g", "sh", "t", "n", "dh"},
			monthsWide:             []string{"", "janar", "shkurt", "mars", "prill", "maj", "qershor", "korrik", "gusht", "shtator", "tetor", "nëntor", "dhjetor"},
			daysAbbreviated:        []string{"Die", "Hën", "Mar", "Mër", "Enj", "Pre", "Sht"},
			daysNarrow:             []string{"d", "h", "m", "m", "e", "p", "sh"},
			daysShort:              []string{"die", "hën", "mar", "mër", "enj", "pre", "sht"},
			daysWide:               []string{"e diel", "e hënë", "e martë", "e mërkurë", "e enjte", "e premte", "e shtunë"},
			periodsAbbreviated:     []string{"p.d.", "m.d."},
			periodsNarrow:          []string{"p.d.", "m.d."},
			periodsWide:            []string{"e paradites", "e pasdites"},
			erasAbbreviated:        []string{"p.K.", "mb.K."},
			erasNarrow:             []string{"p.K.", "mb.K."},
			erasWide:               []string{"para Krishtit", "mbas Krishtit"},
			timezones:              map[string]string{"ACDT": "Ora verore e Australisë Qendrore", "ACST": "Ora standarde e Australisë Qendrore", "ACWDT": "Ora verore e Australisë Qendroro-Perëndimore", "ACWST": "Ora standarde e Australisë Qendroro-Perëndimore", "ADT": "Ora verore e Atlantikut", "AEDT": "Ora verore e Australisë Lindore", "AEST": "Ora standarde e Australisë Lindore", "AKDT": "Ora verore e Alaskës", "AKST": "Ora standarde e Alaskës", "ARST": "Ora verore e Argjentinës", "ART": "Ora standarde e Argjentinës", "AST": "Ora standarde e Atlantikut", "AWDT": "Ora verore e Australisë Perëndimore", "AWST": "Ora standarde e Australisë Perëndimore", "BOT": "Ora e Bolivisë", "BT": "Ora e Butanit", "CAT": "Ora e Afrikës Qendrore", "CDT": "Ora verore e SHBA-së Qendrore", "CHADT": "Ora verore e Katamit", "CHAST": "Ora standarde e Katamit", "CLST": "Ora verore e Kilit", "CLT": "Ora standarde e Kilit", "COST": "Ora verore e Kolumbisë", "COT": "Ora standarde e Kolumbisë", "CST": "Ora standarde e SHBA-së Qendrore", "ChST": "Ora e Kamorros", "EAT": "Ora e Afrikës Lindore", "ECT": "Ora e Ekuadorit", "EDT": "Ora verore e SHBA-së Lindore", "EST": "Ora standarde e SHBA-së Lindore", "GFT": "Ora e Guajanës Franceze", "GMT": "Ora e Grinuiçit", "GST": "Ora e Gjirit", "GYT": "Ora e Guajanës", "HADT": "Ora verore e Ishujve Hauai-Aleutian", "HAST": "Ora standarde e Ishujve Hauai-Aleutian", "HAT": "Ora verore e Njufaundlendit [Tokës së Re]", "HECU": "Ora verore e Kubës", "HEEG": "Ora verore e Grenlandës Lindore", "HENOMX": "Ora verore e Meksikës Veriperëndimore", "HEOG": "Ora verore e Grënlandës Perëndimore", "HEPM": "Ora verore e Shën-Pier dhe Mikelon", "HEPMX": "Ora verore e Territoreve Meksikane të Bregut të Paqësorit", "HKST": "Ora verore e Hong-Kongut", "HKT": "Ora standarde e Hong-Kongut", "HNCU": "Ora standarde e Kubës", "HNEG": "Ora standarde e Grenlandës Lindore", "HNNOMX": "Ora standarde e Meksikës Veriperëndimore", "HNOG": "Ora standarde e Grënlandës Perëndimore", "HNPM": "Ora standarde e Shën-Pier dhe Mikelon", "HNPMX": "Ora standarde e Territoreve Meksikane të Bregut të Paqësorit", "HNT": "Ora standarde e Njufaundlendit [Tokës së Re]", "IST": "Ora standarde e Indisë", "JDT": "Ora verore e Japonisë", "JST": "Ora standarde e Japonisë", "LHDT": "Ora verore e Lord-Houit", "LHST": "Ora standarde e Lord-Houit", "MDT": "Ora verore e Territoreve Amerikane të Brezit Malor", "MESZ": "Ora verore e Evropës Qendrore", "MEZ": "Ora standarde e Evropës Qendrore", "MST": "Ora standarde e Territoreve Amerikane të Brezit Malor", "MYT": "Ora e Malajzisë", "NZDT": "Ora verore e Zelandës së Re", "NZST": "Ora standarde e Zelandës së Re", "OESZ": "Ora verore e Evropës Lindore", "OEZ": "Ora standarde e Evropës Lindore", "PDT": "Ora verore e Territoreve Amerikane të Bregut të Paqësorit", "PST": "Ora standarde e Territoreve Amerikane të Bregut të Paqësorit", "SAST": "Ora standarde e Afrikës Jugore", "SGT": "Ora e Singaporit", "SRT": "Ora e Surinamit", "TMST": "Ora verore e Turkmenistanit", "TMT": "Ora standarde e Turkmenistanit", "UYST": "Ora verore e Uruguait", "UYT": "Ora standarde e Uruguait", "VET": "Ora e Venezuelës", "WARST": "Ora verore e Argjentinës Perëndimore", "WART": "Ora standarde e Argjentinës Perëndimore", "WAST": "Ora verore e Afrikës Perëndimore", "WAT": "Ora standarde e Afrikës Perëndimore", "WESZ": "Ora verore e Evropës Perëndimore", "WEZ": "Ora standarde e Evropës Perëndimore", "WIB": "Ora e Indonezisë Perëndimore", "WIT": "Ora e Indonezisë Lindore", "WITA": "Ora e Indonezisë Qendrore", "∅∅∅": "Ora verore e Brazilisë"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_922dbe895fd3d466f77caf876ca0d4af,
			fnFmtTimeLong:         fnFmtTimeLong_880840264062db9197a5f9820877cc61,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_8b495a08240ebf51a9ad31ffd7bf6c66,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sq_mk"] = func() locales.Translator {
		return &localen{

			locale:                 "sq_MK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "den", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "shk", "mar", "pri", "maj", "qer", "korr", "gush", "sht", "tet", "nën", "dhj"},
			monthsNarrow:           []string{"", "j", "sh", "m", "p", "m", "q", "k", "g", "sh", "t", "n", "dh"},
			monthsWide:             []string{"", "janar", "shkurt", "mars", "prill", "maj", "qershor", "korrik", "gusht", "shtator", "tetor", "nëntor", "dhjetor"},
			daysAbbreviated:        []string{"Die", "Hën", "Mar", "Mër", "Enj", "Pre", "Sht"},
			daysNarrow:             []string{"d", "h", "m", "m", "e", "p", "sh"},
			daysShort:              []string{"die", "hën", "mar", "mër", "enj", "pre", "sht"},
			daysWide:               []string{"e diel", "e hënë", "e martë", "e mërkurë", "e enjte", "e premte", "e shtunë"},
			periodsAbbreviated:     []string{"p.d.", "m.d."},
			periodsNarrow:          []string{"p.d.", "m.d."},
			periodsWide:            []string{"e paradites", "e pasdites"},
			erasAbbreviated:        []string{"p.K.", "mb.K."},
			erasNarrow:             []string{"p.K.", "mb.K."},
			erasWide:               []string{"para Krishtit", "mbas Krishtit"},
			timezones:              map[string]string{"ACDT": "Ora verore e Australisë Qendrore", "ACST": "Ora standarde e Australisë Qendrore", "ACWDT": "Ora verore e Australisë Qendroro-Perëndimore", "ACWST": "Ora standarde e Australisë Qendroro-Perëndimore", "ADT": "Ora verore e Atlantikut", "AEDT": "Ora verore e Australisë Lindore", "AEST": "Ora standarde e Australisë Lindore", "AKDT": "Ora verore e Alaskës", "AKST": "Ora standarde e Alaskës", "ARST": "Ora verore e Argjentinës", "ART": "Ora standarde e Argjentinës", "AST": "Ora standarde e Atlantikut", "AWDT": "Ora verore e Australisë Perëndimore", "AWST": "Ora standarde e Australisë Perëndimore", "BOT": "Ora e Bolivisë", "BT": "Ora e Butanit", "CAT": "Ora e Afrikës Qendrore", "CDT": "Ora verore e SHBA-së Qendrore", "CHADT": "Ora verore e Katamit", "CHAST": "Ora standarde e Katamit", "CLST": "Ora verore e Kilit", "CLT": "Ora standarde e Kilit", "COST": "Ora verore e Kolumbisë", "COT": "Ora standarde e Kolumbisë", "CST": "Ora standarde e SHBA-së Qendrore", "ChST": "Ora e Kamorros", "EAT": "Ora e Afrikës Lindore", "ECT": "Ora e Ekuadorit", "EDT": "Ora verore e SHBA-së Lindore", "EST": "Ora standarde e SHBA-së Lindore", "GFT": "Ora e Guajanës Franceze", "GMT": "Ora e Grinuiçit", "GST": "Ora e Gjirit", "GYT": "Ora e Guajanës", "HADT": "Ora verore e Ishujve Hauai-Aleutian", "HAST": "Ora standarde e Ishujve Hauai-Aleutian", "HAT": "Ora verore e Njufaundlendit [Tokës së Re]", "HECU": "Ora verore e Kubës", "HEEG": "Ora verore e Grenlandës Lindore", "HENOMX": "Ora verore e Meksikës Veriperëndimore", "HEOG": "Ora verore e Grënlandës Perëndimore", "HEPM": "Ora verore e Shën-Pier dhe Mikelon", "HEPMX": "Ora verore e Territoreve Meksikane të Bregut të Paqësorit", "HKST": "Ora verore e Hong-Kongut", "HKT": "Ora standarde e Hong-Kongut", "HNCU": "Ora standarde e Kubës", "HNEG": "Ora standarde e Grenlandës Lindore", "HNNOMX": "Ora standarde e Meksikës Veriperëndimore", "HNOG": "Ora standarde e Grënlandës Perëndimore", "HNPM": "Ora standarde e Shën-Pier dhe Mikelon", "HNPMX": "Ora standarde e Territoreve Meksikane të Bregut të Paqësorit", "HNT": "Ora standarde e Njufaundlendit [Tokës së Re]", "IST": "Ora standarde e Indisë", "JDT": "Ora verore e Japonisë", "JST": "Ora standarde e Japonisë", "LHDT": "Ora verore e Lord-Houit", "LHST": "Ora standarde e Lord-Houit", "MDT": "Ora verore e Territoreve Amerikane të Brezit Malor", "MESZ": "Ora verore e Evropës Qendrore", "MEZ": "Ora standarde e Evropës Qendrore", "MST": "Ora standarde e Territoreve Amerikane të Brezit Malor", "MYT": "Ora e Malajzisë", "NZDT": "Ora verore e Zelandës së Re", "NZST": "Ora standarde e Zelandës së Re", "OESZ": "Ora verore e Evropës Lindore", "OEZ": "Ora standarde e Evropës Lindore", "PDT": "Ora verore e Territoreve Amerikane të Bregut të Paqësorit", "PST": "Ora standarde e Territoreve Amerikane të Bregut të Paqësorit", "SAST": "Ora standarde e Afrikës Jugore", "SGT": "Ora e Singaporit", "SRT": "Ora e Surinamit", "TMST": "Ora verore e Turkmenistanit", "TMT": "Ora standarde e Turkmenistanit", "UYST": "Ora verore e Uruguait", "UYT": "Ora standarde e Uruguait", "VET": "Ora e Venezuelës", "WARST": "Ora verore e Argjentinës Perëndimore", "WART": "Ora standarde e Argjentinës Perëndimore", "WAST": "Ora verore e Afrikës Perëndimore", "WAT": "Ora standarde e Afrikës Perëndimore", "WESZ": "Ora verore e Evropës Perëndimore", "WEZ": "Ora standarde e Evropës Perëndimore", "WIB": "Ora e Indonezisë Perëndimore", "WIT": "Ora e Indonezisë Lindore", "WITA": "Ora e Indonezisë Qendrore", "∅∅∅": "Ora verore e Brazilisë"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_8b495a08240ebf51a9ad31ffd7bf6c66,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sq_xk"] = func() locales.Translator {
		return &localen{

			locale:                 "sq_XK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 5, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "shk", "mar", "pri", "maj", "qer", "korr", "gush", "sht", "tet", "nën", "dhj"},
			monthsNarrow:           []string{"", "j", "sh", "m", "p", "m", "q", "k", "g", "sh", "t", "n", "dh"},
			monthsWide:             []string{"", "janar", "shkurt", "mars", "prill", "maj", "qershor", "korrik", "gusht", "shtator", "tetor", "nëntor", "dhjetor"},
			daysAbbreviated:        []string{"Die", "Hën", "Mar", "Mër", "Enj", "Pre", "Sht"},
			daysNarrow:             []string{"d", "h", "m", "m", "e", "p", "sh"},
			daysShort:              []string{"die", "hën", "mar", "mër", "enj", "pre", "sht"},
			daysWide:               []string{"e diel", "e hënë", "e martë", "e mërkurë", "e enjte", "e premte", "e shtunë"},
			periodsAbbreviated:     []string{"p.d.", "m.d."},
			periodsNarrow:          []string{"p.d.", "m.d."},
			periodsWide:            []string{"e paradites", "e pasdites"},
			erasAbbreviated:        []string{"p.K.", "mb.K."},
			erasNarrow:             []string{"p.K.", "mb.K."},
			erasWide:               []string{"para Krishtit", "mbas Krishtit"},
			timezones:              map[string]string{"ACDT": "Ora verore e Australisë Qendrore", "ACST": "Ora standarde e Australisë Qendrore", "ACWDT": "Ora verore e Australisë Qendroro-Perëndimore", "ACWST": "Ora standarde e Australisë Qendroro-Perëndimore", "ADT": "Ora verore e Atlantikut", "AEDT": "Ora verore e Australisë Lindore", "AEST": "Ora standarde e Australisë Lindore", "AKDT": "Ora verore e Alaskës", "AKST": "Ora standarde e Alaskës", "ARST": "Ora verore e Argjentinës", "ART": "Ora standarde e Argjentinës", "AST": "Ora standarde e Atlantikut", "AWDT": "Ora verore e Australisë Perëndimore", "AWST": "Ora standarde e Australisë Perëndimore", "BOT": "Ora e Bolivisë", "BT": "Ora e Butanit", "CAT": "Ora e Afrikës Qendrore", "CDT": "Ora verore e SHBA-së Qendrore", "CHADT": "Ora verore e Katamit", "CHAST": "Ora standarde e Katamit", "CLST": "Ora verore e Kilit", "CLT": "Ora standarde e Kilit", "COST": "Ora verore e Kolumbisë", "COT": "Ora standarde e Kolumbisë", "CST": "Ora standarde e SHBA-së Qendrore", "ChST": "Ora e Kamorros", "EAT": "Ora e Afrikës Lindore", "ECT": "Ora e Ekuadorit", "EDT": "Ora verore e SHBA-së Lindore", "EST": "Ora standarde e SHBA-së Lindore", "GFT": "Ora e Guajanës Franceze", "GMT": "Ora e Grinuiçit", "GST": "Ora e Gjirit", "GYT": "Ora e Guajanës", "HADT": "Ora verore e Ishujve Hauai-Aleutian", "HAST": "Ora standarde e Ishujve Hauai-Aleutian", "HAT": "Ora verore e Njufaundlendit [Tokës së Re]", "HECU": "Ora verore e Kubës", "HEEG": "Ora verore e Grenlandës Lindore", "HENOMX": "Ora verore e Meksikës Veriperëndimore", "HEOG": "Ora verore e Grënlandës Perëndimore", "HEPM": "Ora verore e Shën-Pier dhe Mikelon", "HEPMX": "Ora verore e Territoreve Meksikane të Bregut të Paqësorit", "HKST": "Ora verore e Hong-Kongut", "HKT": "Ora standarde e Hong-Kongut", "HNCU": "Ora standarde e Kubës", "HNEG": "Ora standarde e Grenlandës Lindore", "HNNOMX": "Ora standarde e Meksikës Veriperëndimore", "HNOG": "Ora standarde e Grënlandës Perëndimore", "HNPM": "Ora standarde e Shën-Pier dhe Mikelon", "HNPMX": "Ora standarde e Territoreve Meksikane të Bregut të Paqësorit", "HNT": "Ora standarde e Njufaundlendit [Tokës së Re]", "IST": "Ora standarde e Indisë", "JDT": "Ora verore e Japonisë", "JST": "Ora standarde e Japonisë", "LHDT": "Ora verore e Lord-Houit", "LHST": "Ora standarde e Lord-Houit", "MDT": "Ora verore e Territoreve Amerikane të Brezit Malor", "MESZ": "Ora verore e Evropës Qendrore", "MEZ": "Ora standarde e Evropës Qendrore", "MST": "Ora standarde e Territoreve Amerikane të Brezit Malor", "MYT": "Ora e Malajzisë", "NZDT": "Ora verore e Zelandës së Re", "NZST": "Ora standarde e Zelandës së Re", "OESZ": "Ora verore e Evropës Lindore", "OEZ": "Ora standarde e Evropës Lindore", "PDT": "Ora verore e Territoreve Amerikane të Bregut të Paqësorit", "PST": "Ora standarde e Territoreve Amerikane të Bregut të Paqësorit", "SAST": "Ora standarde e Afrikës Jugore", "SGT": "Ora e Singaporit", "SRT": "Ora e Surinamit", "TMST": "Ora verore e Turkmenistanit", "TMT": "Ora standarde e Turkmenistanit", "UYST": "Ora verore e Uruguait", "UYT": "Ora standarde e Uruguait", "VET": "Ora e Venezuelës", "WARST": "Ora verore e Argjentinës Perëndimore", "WART": "Ora standarde e Argjentinës Perëndimore", "WAST": "Ora verore e Afrikës Perëndimore", "WAT": "Ora standarde e Afrikës Perëndimore", "WESZ": "Ora verore e Evropës Perëndimore", "WEZ": "Ora standarde e Evropës Perëndimore", "WIB": "Ora e Indonezisë Perëndimore", "WIT": "Ora e Indonezisë Lindore", "WITA": "Ora e Indonezisë Qendrore", "∅∅∅": "Ora verore e Perusë"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_368d18aec5667a40e744175ddb192235,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_e2379225f63633cef88f4847394352a7,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_8b495a08240ebf51a9ad31ffd7bf6c66,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr"] = func() locales.Translator {
		return &localen{

			locale:                 "sr",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "КМ", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"пре подне", "по подне"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"пре подне", "по подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"пре нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Макао летње рачунање времена", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Макао стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Перу, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_cyrl"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Cyrl",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"пре подне", "по подне"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"пре подне", "по подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"пре нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Бразилија, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_cyrl_ba"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Cyrl_BA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "ут", "ср", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недјеља", "понедељак", "уторак", "сриједа", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"прије подне", "по подне"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"прије подне", "по подне"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"прије нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Азори, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_cyrl_me"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Cyrl_ME",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "март", "апр", "мај", "јун", "јул", "авг", "септ", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недјеља", "понедељак", "уторак", "сриједа", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"прије подне", "по подне"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"прије подне", "по подне"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Перу, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_cyrl_rs"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Cyrl_RS",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"пре подне", "по подне"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"пре подне", "по подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"пре нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Азори, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_cyrl_xk"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Cyrl_XK",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "март", "апр", "мај", "јун", "јул", "авг", "септ", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"пре подне", "по подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"пре нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Азори, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_latn"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Latn",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "KM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "jan", "feb", "mar", "apr", "maj", "jun", "jul", "avg", "sep", "okt", "nov", "dec"},
			monthsNarrow:           []string{"", "j", "f", "m", "a", "m", "j", "j", "a", "s", "o", "n", "d"},
			monthsWide:             []string{"", "januar", "februar", "mart", "april", "maj", "jun", "jul", "avgust", "septembar", "oktobar", "novembar", "decembar"},
			daysAbbreviated:        []string{"ned", "pon", "uto", "sre", "čet", "pet", "sub"},
			daysNarrow:             []string{"n", "p", "u", "s", "č", "p", "s"},
			daysShort:              []string{"ne", "po", "ut", "sr", "če", "pe", "su"},
			daysWide:               []string{"nedelja", "ponedeljak", "utorak", "sreda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"pre podne", "po podne"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"pre podne", "po podne"},
			erasAbbreviated:        []string{"p. n. e.", "n. e."},
			erasNarrow:             []string{"p.n.e.", "n.e."},
			erasWide:               []string{"pre nove ere", "nove ere"},
			timezones:              map[string]string{"ACDT": "Australijsko centralno letnje vreme", "ACST": "Australijsko centralno standardno vreme", "ACWDT": "Australijsko centralno zapadno letnje vreme", "ACWST": "Australijsko centralno zapadno standardno vreme", "ADT": "Atlantsko letnje vreme", "AEDT": "Australijsko istočno letnje vreme", "AEST": "Australijsko istočno standardno vreme", "AKDT": "Aljaska, letnje vreme", "AKST": "Aljaska, standardno vreme", "ARST": "Argentina, letnje vreme", "ART": "Argentina, standardno vreme", "AST": "Atlantsko standardno vreme", "AWDT": "Australijsko zapadno letnje vreme", "AWST": "Australijsko zapadno standardno vreme", "BOT": "Bolivija vreme", "BT": "Butan vreme", "CAT": "Centralno-afričko vreme", "CDT": "Severnoameričko centralno letnje vreme", "CHADT": "Čatam, letnje vreme", "CHAST": "Čatam, standardno vreme", "CLST": "Čile, letnje vreme", "CLT": "Čile, standardno vreme", "COST": "Kolumbija, letnje vreme", "COT": "Kolumbija, standardno vreme", "CST": "Severnoameričko centralno standardno vreme", "ChST": "Čamoro vreme", "EAT": "Istočno-afričko vreme", "ECT": "Ekvador vreme", "EDT": "Severnoameričko istočno letnje vreme", "EST": "Severnoameričko istočno standardno vreme", "GFT": "Francuska Gvajana vreme", "GMT": "Srednje vreme po Griniču", "GST": "Zalivsko vreme", "GYT": "Gvajana vreme", "HADT": "Havajsko-aleutsko letnje vreme", "HAST": "Havajsko-aleutsko standardno vreme", "HAT": "Njufaundlend, letnje vreme", "HECU": "Kuba, letnje vreme", "HEEG": "Istočni Grenland, letnje vreme", "HENOMX": "Severozapadni Meksiko, letnje vreme", "HEOG": "Zapadni Grenland, letnje vreme", "HEPM": "Sen Pjer i Mikelon, letnje vreme", "HEPMX": "Meksički Pacifik, letnje vreme", "HKST": "Hong Kong, letnje vreme", "HKT": "Hong Kong, standardno vreme", "HNCU": "Kuba, standardno vreme", "HNEG": "Istočni Grenland, standardno vreme", "HNNOMX": "Severozapadni Meksiko, standardno vreme", "HNOG": "Zapadni Grenland, standardno vreme", "HNPM": "Sen Pjer i Mikelon, standardno vreme", "HNPMX": "Meksički Pacifik, standardno vreme", "HNT": "Njufaundlend, standardno vreme", "IST": "Indijsko standardno vreme", "JDT": "Japansko letnje vreme", "JST": "Japansko standardno vreme", "LHDT": "Lord Hov, letnje vreme", "LHST": "Lord Hov, standardno vreme", "MDT": "Severnoameričko planinsko letnje vreme", "MESZ": "Srednjeevropsko letnje vreme", "MEZ": "Srednjeevropsko standardno vreme", "MST": "Severnoameričko planinsko standardno vreme", "MYT": "Malezija vreme", "NZDT": "Novi Zeland, letnje vreme", "NZST": "Novi Zeland, standardno vreme", "OESZ": "Istočnoevropsko letnje vreme", "OEZ": "Istočnoevropsko standardno vreme", "PDT": "Severnoameričko pacifičko letnje vreme", "PST": "Severnoameričko pacifičko standardno vreme", "SAST": "Južno-afričko vreme", "SGT": "Singapur, standardno vreme", "SRT": "Surinam vreme", "TMST": "Turkmenistan, letnje vreme", "TMT": "Turkmenistan, standardno vreme", "UYST": "Urugvaj, letnje vreme", "UYT": "Urugvaj, standardno vreme", "VET": "Venecuela vreme", "WARST": "Zapadna Argentina, letnje vreme", "WART": "Zapadna Argentina, standardno vreme", "WAST": "Zapadno-afričko letnje vreme", "WAT": "Zapadno-afričko standardno vreme", "WESZ": "Zapadnoevropsko letnje vreme", "WEZ": "Zapadnoevropsko standardno vreme", "WIB": "Zapadno-indonezijsko vreme", "WIT": "Istočno-indonezijsko vreme", "WITA": "Centralno-indonezijsko vreme", "∅∅∅": "Azori, letnje vreme"},

			fnCardinalPluralRule:  fnCardinalPluralRule_7b6df98bb7a40cd3e9c67c7c7f464286,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_latn_ba"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Latn_BA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"ned", "pon", "ut", "sr", "čet", "pet", "sub"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"nedjelja", "ponedeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"prije podne", "po podne"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"prije podne", "po podne"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"prije nove ere", "nove ere"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Бразилија, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_latn_me"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Latn_ME",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "mart", "maj", "jun", "jul", "sept"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"nedjelja", "ponedeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"},
			periodsAbbreviated:     []string{"prije podne", "po podne"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"prije podne", "po podne"},
			erasAbbreviated:        []string{"", ""},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Азори, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_latn_rs"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Latn_RS",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"пре подне", "по подне"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"пре подне", "по подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"пре нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Бразилија, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_d76954adbbec5c02f9aae79b8f9d4067,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sr_latn_xk"] = func() locales.Translator {
		return &localen{

			locale:                 "sr_Latn_XK",
			pluralsCardinal:        []locales.PluralRule{2, 4, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 4, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "mart", "maj", "jun", "jul", "sept"},
			monthsNarrow:           []string{"", "ј", "ф", "м", "а", "м", "ј", "ј", "а", "с", "о", "н", "д"},
			monthsWide:             []string{"", "јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"},
			daysAbbreviated:        []string{"нед", "пон", "уто", "сре", "чет", "пет", "суб"},
			daysNarrow:             []string{"н", "п", "у", "с", "ч", "п", "с"},
			daysShort:              []string{"не", "по", "ут", "ср", "че", "пе", "су"},
			daysWide:               []string{"недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"пре подне", "по подне"},
			erasAbbreviated:        []string{"п. н. е.", "н. е."},
			erasNarrow:             []string{"п.н.е.", "н.е."},
			erasWide:               []string{"пре нове ере", "нове ере"},
			timezones:              map[string]string{"ACDT": "Аустралијско централно летње време", "ACST": "Аустралијско централно стандардно време", "ACWDT": "Аустралијско централно западно летње време", "ACWST": "Аустралијско централно западно стандардно време", "ADT": "Атлантско летње време", "AEDT": "Аустралијско источно летње време", "AEST": "Аустралијско источно стандардно време", "AKDT": "Аљаска, летње време", "AKST": "Аљаска, стандардно време", "ARST": "Аргентина, летње време", "ART": "Аргентина, стандардно време", "AST": "Атлантско стандардно време", "AWDT": "Аустралијско западно летње време", "AWST": "Аустралијско западно стандардно време", "BOT": "Боливија време", "BT": "Бутан време", "CAT": "Централно-афричко време", "CDT": "Северноамеричко централно летње време", "CHADT": "Чатам, летње време", "CHAST": "Чатам, стандардно време", "CLST": "Чиле, летње време", "CLT": "Чиле, стандардно време", "COST": "Колумбија, летње време", "COT": "Колумбија, стандардно време", "CST": "Северноамеричко централно стандардно време", "ChST": "Чаморо време", "EAT": "Источно-афричко време", "ECT": "Еквадор време", "EDT": "Северноамеричко источно летње време", "EST": "Северноамеричко источно стандардно време", "GFT": "Француска Гвајана време", "GMT": "Средње време по Гриничу", "GST": "Заливско време", "GYT": "Гвајана време", "HADT": "Хавајско-алеутско летње време", "HAST": "Хавајско-алеутско стандардно време", "HAT": "Њуфаундленд, летње време", "HECU": "Куба, летње време", "HEEG": "Источни Гренланд, летње време", "HENOMX": "Северозападни Мексико, летње време", "HEOG": "Западни Гренланд, летње време", "HEPM": "Сен Пјер и Микелон, летње време", "HEPMX": "Мексички Пацифик, летње време", "HKST": "Хонг Конг, летње време", "HKT": "Хонг Конг, стандардно време", "HNCU": "Куба, стандардно време", "HNEG": "Источни Гренланд, стандардно време", "HNNOMX": "Северозападни Мексико, стандардно време", "HNOG": "Западни Гренланд, стандардно време", "HNPM": "Сен Пјер и Микелон, стандардно време", "HNPMX": "Мексички Пацифик, стандардно време", "HNT": "Њуфаундленд, стандардно време", "IST": "Индијско стандардно време", "JDT": "Јапанско летње време", "JST": "Јапанско стандардно време", "LHDT": "Лорд Хов, летње време", "LHST": "Лорд Хов, стандардно време", "MDT": "Северноамеричко планинско летње време", "MESZ": "Средњеевропско летње време", "MEZ": "Средњеевропско стандардно време", "MST": "Северноамеричко планинско стандардно време", "MYT": "Малезија време", "NZDT": "Нови Зеланд, летње време", "NZST": "Нови Зеланд, стандардно време", "OESZ": "Источноевропско летње време", "OEZ": "Источноевропско стандардно време", "PDT": "Северноамеричко пацифичко летње време", "PST": "Северноамеричко пацифичко стандардно време", "SAST": "Јужно-афричко време", "SGT": "Сингапур, стандардно време", "SRT": "Суринам време", "TMST": "Туркменистан, летње време", "TMT": "Туркменистан, стандардно време", "UYST": "Уругвај, летње време", "UYT": "Уругвај, стандардно време", "VET": "Венецуела време", "WARST": "Западна Аргентина, летње време", "WART": "Западна Аргентина, стандардно време", "WAST": "Западно-афричко летње време", "WAT": "Западно-афричко стандардно време", "WESZ": "Западноевропско летње време", "WEZ": "Западноевропско стандардно време", "WIB": "Западно-индонезијско време", "WIT": "Источно-индонезијско време", "WITA": "Централно-индонезијско време", "∅∅∅": "Бразилија, летње време"},

			fnCardinalPluralRule:  fnCardinalPluralRule_2fd16cba9f129ab3317e08638d736342,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_64449162f5ca02bf56f475bc8fb0f4a7,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_94801bd77ea8c5746f4486dd265a9d5e,
			fnFmtDateLong:         fnFmtDateLong_d042ff2f76eb964a2a614f1d2926f4a7,
			fnFmtDateMedium:       fnFmtDateMedium_61defeac982d2749785dbfe5e166419f,
			fnFmtDateShort:        fnFmtDateShort_93d8b8c715a1811ebc94777fc3aa4672,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_b9d3512264e082d2fa10f438237ce3cc,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sv"] = func() locales.Translator {
		return &localen{

			locale:                 "sv",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "Bds$", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BM$", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BR$", "BRN", "BRR", "BRZ", "BS$", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZ$", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "Dkr", "RD$", "DZD", "ECS", "ECV", "Ekr", "EG£", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IE£", "ILP", "ILR", "₪", "INR", "IQD", "IRR", "ISJ", "Ikr", "ITL", "JM$", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "Nkr", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "kr", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mars", "apr.", "maj", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"sön", "mån", "tis", "ons", "tors", "fre", "lör"},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sö", "må", "ti", "on", "to", "fr", "lö"},
			daysWide:               []string{"söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"},
			periodsAbbreviated:     []string{"fm", "em"},
			periodsNarrow:          []string{"fm", "em"},
			periodsWide:            []string{"fm", "em"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"före Kristus", "efter Kristus"},
			timezones:              map[string]string{"ACDT": "centralaustralisk sommartid", "ACST": "centralaustralisk normaltid", "ACWDT": "västcentralaustralisk sommartid", "ACWST": "västcentralaustralisk normaltid", "ADT": "nordamerikansk atlantsommartid", "AEDT": "östaustralisk sommartid", "AEST": "östaustralisk normaltid", "AKDT": "Alaska, sommartid", "AKST": "Alaska, normaltid", "ARST": "östargentinsk sommartid", "ART": "östargentinsk normaltid", "AST": "nordamerikansk atlantnormaltid", "AWDT": "västaustralisk sommartid", "AWST": "västaustralisk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "centralafrikansk tid", "CDT": "centralnordamerikansk sommartid", "CHADT": "Chatham, sommartid", "CHAST": "Chatham, normaltid", "CLST": "chilensk sommartid", "CLT": "chilensk normaltid", "COST": "colombiansk sommartid", "COT": "colombiansk normaltid", "CST": "centralnordamerikansk normaltid", "ChST": "Chamorrotid", "EAT": "östafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "östnordamerikansk sommartid", "EST": "östnordamerikansk normaltid", "GFT": "Franska Guyanatid", "GMT": "Greenwichtid", "GST": "Persiska vikentid", "GYT": "Guyanatid", "HADT": "Honolulu, sommartid", "HAST": "Honolulu, normaltid", "HAT": "Newfoundland, sommartid", "HECU": "kubansk sommartid", "HEEG": "östgrönländsk sommartid", "HENOMX": "nordvästmexikansk sommartid", "HEOG": "västgrönländsk sommartid", "HEPM": "S:t Pierre och Miquelon, sommartid", "HEPMX": "mexikansk stillahavstid, sommartid", "HKST": "Hongkong, sommartid", "HKT": "Hongkong, normaltid", "HNCU": "kubansk normaltid", "HNEG": "östgrönländsk normaltid", "HNNOMX": "nordvästmexikansk normaltid", "HNOG": "västgrönländsk normaltid", "HNPM": "S:t Pierre och Miquelon, normaltid", "HNPMX": "mexikansk stillahavstid, normaltid", "HNT": "Newfoundland, normaltid", "IST": "indisk tid", "JDT": "japansk sommartid", "JST": "japansk normaltid", "LHDT": "Lord Howe, sommartid", "LHST": "Lord Howe, normaltid", "MDT": "Klippiga bergen, sommartid", "MESZ": "centraleuropeisk sommartid", "MEZ": "centraleuropeisk normaltid", "MST": "Klippiga bergen, normaltid", "MYT": "malaysisk tid", "NZDT": "nyzeeländsk sommartid", "NZST": "nyzeeländsk normaltid", "OESZ": "östeuropeisk sommartid", "OEZ": "östeuropeisk normaltid", "PDT": "västnordamerikansk sommartid", "PST": "västnordamerikansk normaltid", "SAST": "sydafrikansk tid", "SGT": "Singaporetid", "SRT": "Surinamtid", "TMST": "turkmensk sommartid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommartid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "västargentinsk sommartid", "WART": "västargentinsk normaltid", "WAST": "västafrikansk sommartid", "WAT": "västafrikansk normaltid", "WESZ": "västeuropeisk sommartid", "WEZ": "västeuropeisk normaltid", "WIB": "västindonesisk tid", "WIT": "östindonesisk tid", "WITA": "centralindonesisk tid", "∅∅∅": "Brasilia, sommartid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_702261155c59dd873e83848c0fe82a23,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_1456aec41e57e34cc80a524a2035cc1f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sv_ax"] = func() locales.Translator {
		return &localen{

			locale:                 "sv_AX",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mars", "apr.", "maj", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"sön", "mån", "tis", "ons", "tors", "fre", "lör"},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sö", "må", "ti", "on", "to", "fr", "lö"},
			daysWide:               []string{"söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"},
			periodsAbbreviated:     []string{"fm", "em"},
			periodsNarrow:          []string{"fm", "em"},
			periodsWide:            []string{"fm", "em"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"före Kristus", "efter Kristus"},
			timezones:              map[string]string{"ACDT": "centralaustralisk sommartid", "ACST": "centralaustralisk normaltid", "ACWDT": "västcentralaustralisk sommartid", "ACWST": "västcentralaustralisk normaltid", "ADT": "nordamerikansk atlantsommartid", "AEDT": "östaustralisk sommartid", "AEST": "östaustralisk normaltid", "AKDT": "Alaska, sommartid", "AKST": "Alaska, normaltid", "ARST": "östargentinsk sommartid", "ART": "östargentinsk normaltid", "AST": "nordamerikansk atlantnormaltid", "AWDT": "västaustralisk sommartid", "AWST": "västaustralisk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "centralafrikansk tid", "CDT": "centralnordamerikansk sommartid", "CHADT": "Chatham, sommartid", "CHAST": "Chatham, normaltid", "CLST": "chilensk sommartid", "CLT": "chilensk normaltid", "COST": "colombiansk sommartid", "COT": "colombiansk normaltid", "CST": "centralnordamerikansk normaltid", "ChST": "Chamorrotid", "EAT": "östafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "östnordamerikansk sommartid", "EST": "östnordamerikansk normaltid", "GFT": "Franska Guyanatid", "GMT": "Greenwichtid", "GST": "Persiska vikentid", "GYT": "Guyanatid", "HADT": "Honolulu, sommartid", "HAST": "Honolulu, normaltid", "HAT": "Newfoundland, sommartid", "HECU": "kubansk sommartid", "HEEG": "östgrönländsk sommartid", "HENOMX": "nordvästmexikansk sommartid", "HEOG": "västgrönländsk sommartid", "HEPM": "S:t Pierre och Miquelon, sommartid", "HEPMX": "mexikansk stillahavstid, sommartid", "HKST": "Hongkong, sommartid", "HKT": "Hongkong, normaltid", "HNCU": "kubansk normaltid", "HNEG": "östgrönländsk normaltid", "HNNOMX": "nordvästmexikansk normaltid", "HNOG": "västgrönländsk normaltid", "HNPM": "S:t Pierre och Miquelon, normaltid", "HNPMX": "mexikansk stillahavstid, normaltid", "HNT": "Newfoundland, normaltid", "IST": "indisk tid", "JDT": "japansk sommartid", "JST": "japansk normaltid", "LHDT": "Lord Howe, sommartid", "LHST": "Lord Howe, normaltid", "MDT": "Klippiga bergen, sommartid", "MESZ": "centraleuropeisk sommartid", "MEZ": "centraleuropeisk normaltid", "MST": "Klippiga bergen, normaltid", "MYT": "malaysisk tid", "NZDT": "nyzeeländsk sommartid", "NZST": "nyzeeländsk normaltid", "OESZ": "östeuropeisk sommartid", "OEZ": "östeuropeisk normaltid", "PDT": "västnordamerikansk sommartid", "PST": "västnordamerikansk normaltid", "SAST": "sydafrikansk tid", "SGT": "Singaporetid", "SRT": "Surinamtid", "TMST": "turkmensk sommartid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommartid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "västargentinsk sommartid", "WART": "västargentinsk normaltid", "WAST": "västafrikansk sommartid", "WAT": "västafrikansk normaltid", "WESZ": "västeuropeisk sommartid", "WEZ": "västeuropeisk normaltid", "WIB": "västindonesisk tid", "WIT": "östindonesisk tid", "WITA": "centralindonesisk tid", "∅∅∅": "peruansk sommartid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_702261155c59dd873e83848c0fe82a23,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89ff89febaf8ab083f938260d29e96d0,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sv_fi"] = func() locales.Translator {
		return &localen{

			locale:                 "sv_FI",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ".",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mars", "apr.", "maj", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"sön", "mån", "tis", "ons", "tors", "fre", "lör"},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sö", "må", "ti", "on", "to", "fr", "lö"},
			daysWide:               []string{"söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"},
			periodsAbbreviated:     []string{"fm", "em"},
			periodsNarrow:          []string{"fm", "em"},
			periodsWide:            []string{"fm", "em"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"före Kristus", "efter Kristus"},
			timezones:              map[string]string{"ACDT": "centralaustralisk sommartid", "ACST": "centralaustralisk normaltid", "ACWDT": "västcentralaustralisk sommartid", "ACWST": "västcentralaustralisk normaltid", "ADT": "nordamerikansk atlantsommartid", "AEDT": "östaustralisk sommartid", "AEST": "östaustralisk normaltid", "AKDT": "Alaska, sommartid", "AKST": "Alaska, normaltid", "ARST": "östargentinsk sommartid", "ART": "östargentinsk normaltid", "AST": "nordamerikansk atlantnormaltid", "AWDT": "västaustralisk sommartid", "AWST": "västaustralisk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "centralafrikansk tid", "CDT": "centralnordamerikansk sommartid", "CHADT": "Chatham, sommartid", "CHAST": "Chatham, normaltid", "CLST": "chilensk sommartid", "CLT": "chilensk normaltid", "COST": "colombiansk sommartid", "COT": "colombiansk normaltid", "CST": "centralnordamerikansk normaltid", "ChST": "Chamorrotid", "EAT": "östafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "östnordamerikansk sommartid", "EST": "östnordamerikansk normaltid", "GFT": "Franska Guyanatid", "GMT": "Greenwichtid", "GST": "Persiska vikentid", "GYT": "Guyanatid", "HADT": "Honolulu, sommartid", "HAST": "Honolulu, normaltid", "HAT": "Newfoundland, sommartid", "HECU": "kubansk sommartid", "HEEG": "östgrönländsk sommartid", "HENOMX": "nordvästmexikansk sommartid", "HEOG": "västgrönländsk sommartid", "HEPM": "S:t Pierre och Miquelon, sommartid", "HEPMX": "mexikansk stillahavstid, sommartid", "HKST": "Hongkong, sommartid", "HKT": "Hongkong, normaltid", "HNCU": "kubansk normaltid", "HNEG": "östgrönländsk normaltid", "HNNOMX": "nordvästmexikansk normaltid", "HNOG": "västgrönländsk normaltid", "HNPM": "S:t Pierre och Miquelon, normaltid", "HNPMX": "mexikansk stillahavstid, normaltid", "HNT": "Newfoundland, normaltid", "IST": "indisk tid", "JDT": "japansk sommartid", "JST": "japansk normaltid", "LHDT": "Lord Howe, sommartid", "LHST": "Lord Howe, normaltid", "MDT": "Klippiga bergen, sommartid", "MESZ": "centraleuropeisk sommartid", "MEZ": "centraleuropeisk normaltid", "MST": "Klippiga bergen, normaltid", "MYT": "malaysisk tid", "NZDT": "nyzeeländsk sommartid", "NZST": "nyzeeländsk normaltid", "OESZ": "östeuropeisk sommartid", "OEZ": "östeuropeisk normaltid", "PDT": "västnordamerikansk sommartid", "PST": "västnordamerikansk normaltid", "SAST": "sydafrikansk tid", "SGT": "Singaporetid", "SRT": "Surinamtid", "TMST": "turkmensk sommartid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommartid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "västargentinsk sommartid", "WART": "västargentinsk normaltid", "WAST": "västafrikansk sommartid", "WAT": "västafrikansk normaltid", "WESZ": "västeuropeisk sommartid", "WEZ": "västeuropeisk normaltid", "WIB": "västindonesisk tid", "WIT": "östindonesisk tid", "WITA": "centralindonesisk tid", "∅∅∅": "Brasilia, sommartid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_702261155c59dd873e83848c0fe82a23,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89ff89febaf8ab083f938260d29e96d0,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sv_se"] = func() locales.Translator {
		return &localen{

			locale:                 "sv_SE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  " ",
			minus:                  "−",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "jan.", "feb.", "mars", "apr.", "maj", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "dec."},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"},
			daysAbbreviated:        []string{"sön", "mån", "tis", "ons", "tors", "fre", "lör"},
			daysNarrow:             []string{"S", "M", "T", "O", "T", "F", "L"},
			daysShort:              []string{"sö", "må", "ti", "on", "to", "fr", "lö"},
			daysWide:               []string{"söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"},
			periodsAbbreviated:     []string{"fm", "em"},
			periodsNarrow:          []string{"fm", "em"},
			periodsWide:            []string{"fm", "em"},
			erasAbbreviated:        []string{"f.Kr.", "e.Kr."},
			erasNarrow:             []string{"f.Kr.", "e.Kr."},
			erasWide:               []string{"före Kristus", "efter Kristus"},
			timezones:              map[string]string{"ACDT": "centralaustralisk sommartid", "ACST": "centralaustralisk normaltid", "ACWDT": "västcentralaustralisk sommartid", "ACWST": "västcentralaustralisk normaltid", "ADT": "nordamerikansk atlantsommartid", "AEDT": "östaustralisk sommartid", "AEST": "östaustralisk normaltid", "AKDT": "Alaska, sommartid", "AKST": "Alaska, normaltid", "ARST": "östargentinsk sommartid", "ART": "östargentinsk normaltid", "AST": "nordamerikansk atlantnormaltid", "AWDT": "västaustralisk sommartid", "AWST": "västaustralisk normaltid", "BOT": "boliviansk tid", "BT": "bhutansk tid", "CAT": "centralafrikansk tid", "CDT": "centralnordamerikansk sommartid", "CHADT": "Chatham, sommartid", "CHAST": "Chatham, normaltid", "CLST": "chilensk sommartid", "CLT": "chilensk normaltid", "COST": "colombiansk sommartid", "COT": "colombiansk normaltid", "CST": "centralnordamerikansk normaltid", "ChST": "Chamorrotid", "EAT": "östafrikansk tid", "ECT": "ecuadoriansk tid", "EDT": "östnordamerikansk sommartid", "EST": "östnordamerikansk normaltid", "GFT": "Franska Guyanatid", "GMT": "Greenwichtid", "GST": "Persiska vikentid", "GYT": "Guyanatid", "HADT": "Honolulu, sommartid", "HAST": "Honolulu, normaltid", "HAT": "Newfoundland, sommartid", "HECU": "kubansk sommartid", "HEEG": "östgrönländsk sommartid", "HENOMX": "nordvästmexikansk sommartid", "HEOG": "västgrönländsk sommartid", "HEPM": "S:t Pierre och Miquelon, sommartid", "HEPMX": "mexikansk stillahavstid, sommartid", "HKST": "Hongkong, sommartid", "HKT": "Hongkong, normaltid", "HNCU": "kubansk normaltid", "HNEG": "östgrönländsk normaltid", "HNNOMX": "nordvästmexikansk normaltid", "HNOG": "västgrönländsk normaltid", "HNPM": "S:t Pierre och Miquelon, normaltid", "HNPMX": "mexikansk stillahavstid, normaltid", "HNT": "Newfoundland, normaltid", "IST": "indisk tid", "JDT": "japansk sommartid", "JST": "japansk normaltid", "LHDT": "Lord Howe, sommartid", "LHST": "Lord Howe, normaltid", "MDT": "Klippiga bergen, sommartid", "MESZ": "centraleuropeisk sommartid", "MEZ": "centraleuropeisk normaltid", "MST": "Klippiga bergen, normaltid", "MYT": "malaysisk tid", "NZDT": "nyzeeländsk sommartid", "NZST": "nyzeeländsk normaltid", "OESZ": "östeuropeisk sommartid", "OEZ": "östeuropeisk normaltid", "PDT": "västnordamerikansk sommartid", "PST": "västnordamerikansk normaltid", "SAST": "sydafrikansk tid", "SGT": "Singaporetid", "SRT": "Surinamtid", "TMST": "turkmensk sommartid", "TMT": "turkmensk normaltid", "UYST": "uruguayansk sommartid", "UYT": "uruguayansk normaltid", "VET": "venezuelansk tid", "WARST": "västargentinsk sommartid", "WART": "västargentinsk normaltid", "WAST": "västafrikansk sommartid", "WAT": "västafrikansk normaltid", "WESZ": "västeuropeisk sommartid", "WEZ": "västeuropeisk normaltid", "WIB": "västindonesisk tid", "WIT": "östindonesisk tid", "WITA": "centralindonesisk tid", "∅∅∅": "azorisk sommartid"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_b212f316e9ff6f3259af6a85a9984d08,
			fnFmtCurrency:         fnFmtCurrency_83ba1db7b672d5fc5d43b771b7b38f5a,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_4752830681b621a14386608b6d45d3fb,
			fnFmtPercent:          fnFmtPercent_08e428b6d93eff202ec014fb83b395cf,
			fnFmtTimeFull:         fnFmtTimeFull_702261155c59dd873e83848c0fe82a23,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89ff89febaf8ab083f938260d29e96d0,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sw"] = func() locales.Translator {
		return &localen{

			locale:                 "sw",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TSh", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "Saa za Mchana za Australia ya Kati", "ACST": "Saa za Wastani za Australia ya Kati", "ACWDT": "Saa za Mchana za Magharibi ya Kati ya Australia", "ACWST": "Saa za Wastani za Magharibi ya Kati ya Australia", "ADT": "Saa za Mchana za Atlantiki", "AEDT": "Saa za Mchana za Mashariki mwa Australia", "AEST": "Saa za Wastani za Mashariki mwa Australia", "AKDT": "Saa za Mchana za Alaska", "AKST": "Saa za Wastani za Alaska", "ARST": "Saa za Majira ya joto za Argentina", "ART": "Saa za Wastani za Argentina", "AST": "Saa za Wastani za Atlantiki", "AWDT": "Saa za Mchana za Australia Magharibi", "AWST": "Saa za Wastani za Australia Magharibi", "BOT": "Saa za Bolivia", "BT": "Saa za Bhutan", "CAT": "Saa za Afrika ya Kati", "CDT": "Saa za Mchana za Kati", "CHADT": "Saa za Mchana za Chatham", "CHAST": "Saa za Wastani za Chatham", "CLST": "Saa za Majira ya joto za Chile", "CLT": "Saa za Wastani za Chile", "COST": "Saa za Majira ya joto za Colombia", "COT": "Saa za Wastani za Colombia", "CST": "Saa za Wastani za Kati", "ChST": "Saa za Wastani za Chamorro", "EAT": "Saa za Afrika Mashariki", "ECT": "Saa za Ecuador", "EDT": "Saa za Mchana za Mashariki", "EST": "Saa za Wastani za Mashariki", "GFT": "Saa za Guiana ya Ufaransa", "GMT": "Saa za Greenwich", "GST": "Saa za Wastani za Gulf", "GYT": "Saa za Guyana", "HADT": "Saa za Mchana za Hawaii-Aleutian", "HAST": "Saa za Wastani za Hawaii-Aleutian", "HAT": "Saa za Mchana za Newfoundland", "HECU": "Saa za Mchana za Cuba", "HEEG": "Saa za Majira ya joto za Greenland Mashariki", "HENOMX": "Saa za mchana za Mexico Kaskazini Magharibi", "HEOG": "Saa za Majira ya joto za Greenland Magharibi", "HEPM": "Saa za Mchana za Saint-Pierre na Miquelon", "HEPMX": "Saa za mchana za pasifiki za Mexico", "HKST": "Saa za Majira ya joto za Hong Kong", "HKT": "Saa za Wastani za Hong Kong", "HNCU": "Saa za Wastani ya Cuba", "HNEG": "Saa za Wastani za Greenland Mashariki", "HNNOMX": "Saa za Wastani za Mexico Kaskazini Magharibi", "HNOG": "Saa za Wastani za Greenland Magharibi", "HNPM": "Saa za Wastani ya Saint-Pierre na Miquelon", "HNPMX": "Saa za wastani za pasifiki za Mexico", "HNT": "Saa za Wastani za Newfoundland", "IST": "Saa Wastani za India", "JDT": "Saa za Mchana za Japan", "JST": "Saa Wastani za Japan", "LHDT": "Saa za Mchana za Lord Howe", "LHST": "Saa za Wastani za Lord Howe", "MDT": "MDT", "MESZ": "Saa za Majira ya joto za Ulaya ya Kati", "MEZ": "Saa za Wastani za Ulaya ya kati", "MST": "MST", "MYT": "Saa za Malaysia", "NZDT": "Saa za Mchana za New Zealand", "NZST": "Saa za Wastani za New Zealand", "OESZ": "Saa za Majira ya joto za Mashariki mwa Ulaya", "OEZ": "Saa za Wastani za Mashariki mwa Ulaya", "PDT": "Saa za Mchana za Pasifiki", "PST": "Saa za Wastani za Pasifiki", "SAST": "Saa za Wastani za Afrika Kusini", "SGT": "Saa za Wastani za Singapore", "SRT": "Saa za Suriname", "TMST": "Saa za Majira ya joto za Turkmenistan", "TMT": "Saa za Wastani za Turkmenistan", "UYST": "Saa za Majira ya joto za Uruguay", "UYT": "Saa za Wastani za Uruguay", "VET": "Saa za Venezuela", "WARST": "Saa za Majira ya joto za Magharibi mwa Argentina", "WART": "Saa za Wastani za Magharibi mwa Argentina", "WAST": "Saa za Majira ya joto za Afrika Magharibi", "WAT": "Saa za Wastani za Afrika Magharibi", "WESZ": "Saa za Majira ya joto za Magharibi mwa Ulaya", "WEZ": "Saa za Wastani za Magharibi mwa Ulaya", "WIB": "Saa za Magharibi mwa Indonesia", "WIT": "Saa za Mashariki mwa Indonesia", "WITA": "Saa za Indonesia ya Kati", "∅∅∅": "Saa za Majira ya joto za Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sw_cd"] = func() locales.Translator {
		return &localen{

			locale:                 "sw_CD",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "FC", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "Saa za Mchana za Australia ya Kati", "ACST": "Saa za Wastani za Australia ya Kati", "ACWDT": "Saa za Mchana za Magharibi ya Kati ya Australia", "ACWST": "Saa za Wastani za Magharibi ya Kati ya Australia", "ADT": "Saa za Mchana za Atlantiki", "AEDT": "Saa za Mchana za Mashariki mwa Australia", "AEST": "Saa za Wastani za Mashariki mwa Australia", "AKDT": "Saa za Mchana za Alaska", "AKST": "Saa za Wastani za Alaska", "ARST": "Saa za Majira ya joto za Argentina", "ART": "Saa za Wastani za Argentina", "AST": "Saa za Wastani za Atlantiki", "AWDT": "Saa za Mchana za Australia Magharibi", "AWST": "Saa za Wastani za Australia Magharibi", "BOT": "Saa za Bolivia", "BT": "Saa za Bhutan", "CAT": "Saa za Afrika ya Kati", "CDT": "Saa za Mchana za Kati", "CHADT": "Saa za Mchana za Chatham", "CHAST": "Saa za Wastani za Chatham", "CLST": "Saa za Majira ya joto za Chile", "CLT": "Saa za Wastani za Chile", "COST": "Saa za Majira ya joto za Colombia", "COT": "Saa za Wastani za Colombia", "CST": "Saa za Wastani za Kati", "ChST": "Saa za Wastani za Chamorro", "EAT": "Saa za Afrika Mashariki", "ECT": "Saa za Ecuador", "EDT": "Saa za Mchana za Mashariki", "EST": "Saa za Wastani za Mashariki", "GFT": "Saa za Guiana ya Ufaransa", "GMT": "Saa za Greenwich", "GST": "Saa za Wastani za Gulf", "GYT": "Saa za Guyana", "HADT": "Saa za Mchana za Hawaii-Aleutian", "HAST": "Saa za Wastani za Hawaii-Aleutian", "HAT": "Saa za Mchana za Newfoundland", "HECU": "Saa za Mchana za Cuba", "HEEG": "Saa za Majira ya joto za Greenland Mashariki", "HENOMX": "Saa za mchana za Mexico Kaskazini Magharibi", "HEOG": "Saa za Majira ya joto za Greenland Magharibi", "HEPM": "Saa za Mchana za Saint-Pierre na Miquelon", "HEPMX": "Saa za mchana za pasifiki za Mexico", "HKST": "Saa za Majira ya joto za Hong Kong", "HKT": "Saa za Wastani za Hong Kong", "HNCU": "Saa za Wastani ya Cuba", "HNEG": "Saa za Wastani za Greenland Mashariki", "HNNOMX": "Saa za Wastani za Mexico Kaskazini Magharibi", "HNOG": "Saa za Wastani za Greenland Magharibi", "HNPM": "Saa za Wastani ya Saint-Pierre na Miquelon", "HNPMX": "Saa za wastani za pasifiki za Mexico", "HNT": "Saa za Wastani za Newfoundland", "IST": "Saa Wastani za India", "JDT": "Saa za Mchana za Japan", "JST": "Saa Wastani za Japan", "LHDT": "Saa za Mchana za Lord Howe", "LHST": "Saa za Wastani za Lord Howe", "MDT": "Saa za Mchana za Mountain", "MESZ": "Saa za Majira ya joto za Ulaya ya Kati", "MEZ": "Saa za Wastani za Ulaya ya kati", "MST": "Saa za Wastani za Mountain", "MYT": "Saa za Malaysia", "NZDT": "Saa za Mchana za New Zealand", "NZST": "Saa za Wastani za New Zealand", "OESZ": "Saa za Majira ya joto za Mashariki mwa Ulaya", "OEZ": "Saa za Wastani za Mashariki mwa Ulaya", "PDT": "Saa za Mchana za Pasifiki", "PST": "Saa za Wastani za Pasifiki", "SAST": "Saa za Wastani za Afrika Kusini", "SGT": "Saa za Wastani za Singapore", "SRT": "Saa za Suriname", "TMST": "Saa za Majira ya joto za Turkmenistan", "TMT": "Saa za Wastani za Turkmenistan", "UYST": "Saa za Majira ya joto za Uruguay", "UYT": "Saa za Wastani za Uruguay", "VET": "Saa za Venezuela", "WARST": "Saa za Majira ya joto za Magharibi mwa Argentina", "WART": "Saa za Wastani za Magharibi mwa Argentina", "WAST": "Saa za Majira ya joto za Afrika Magharibi", "WAT": "Saa za Wastani za Afrika Magharibi", "WESZ": "Saa za Majira ya joto za Magharibi mwa Ulaya", "WEZ": "Saa za Wastani za Magharibi mwa Ulaya", "WIB": "Saa za Magharibi mwa Indonesia", "WIT": "Saa za Mashariki mwa Indonesia", "WITA": "Saa za Indonesia ya Kati", "∅∅∅": "Saa za Majira ya joto za Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sw_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "sw_KE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " M",
			currencyNegativePrefix: " M",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "Saa za Mchana za Australia ya Kati", "ACST": "Saa za Wastani za Australia ya Kati", "ACWDT": "Saa za Mchana za Magharibi mwa Australia ya Kati", "ACWST": "Saa za Wastani za Magharibi mwa Australia ya Kati", "ADT": "Saa za Mchana za Atlantiki", "AEDT": "Saa za Mchana za Mashariki mwa Australia", "AEST": "Saa za Wastani za Mashariki mwa Australia", "AKDT": "Saa za Mchana za Alaska", "AKST": "Saa za Wastani za Alaska", "ARST": "Saa za Majira Joto za Ajentina", "ART": "Saa za Wastani za Ajentina", "AST": "Saa za Wastani za Atlantiki", "AWDT": "Saa za Mchana za Australia Magharibi", "AWST": "Saa za Wastani za Australia Magharibi", "BOT": "Saa za Bolivia", "BT": "Saa za Butani", "CAT": "Saa za Afrika ya Kati", "CDT": "Saa za Mchana za Kati", "CHADT": "Saa za Mchana za Chatham", "CHAST": "Saa za Wastani za Chatham", "CLST": "Saa za Majira ya joto za Chile", "CLT": "Saa za Wastani za Chile", "COST": "Saa za Majira ya Joto za Kolombia", "COT": "Saa za Wastani za Kolombia", "CST": "Saa za Wastani za Kati", "ChST": "Saa za Wastani za Chamorro", "EAT": "Saa za Afrika Mashariki", "ECT": "Saa za Ekwado", "EDT": "Saa za Mchana za Mashariki", "EST": "Saa za Wastani za Mashariki", "GFT": "Saa za Guiana", "GMT": "Saa za Greenwich", "GST": "Saa za Wastani za Ghuba", "GYT": "Saa za Guyana", "HADT": "Saa za Mchana za Hawaii-Aleutian", "HAST": "Saa za Wastani za Hawaii-Aleutian", "HAT": "Saa za Mchana za Newfoundland", "HECU": "Saa za Mchana za Kuba", "HEEG": "Saa za Majira ya joto za Greenland Mashariki", "HENOMX": "Saa za Mchana za Kaskazini Magharibi mwa Meksiko", "HEOG": "Saa za Majira ya joto za Greenland Magharibi", "HEPM": "Saa za Mchana za Saint-Pierre na Miquelon", "HEPMX": "Saa za mchana za pasifiki za Mexico", "HKST": "Saa za Majira ya joto za Hong Kong", "HKT": "Saa za Wastani za Hong Kong", "HNCU": "Saa za Wastani za Kuba", "HNEG": "Saa za Wastani za Greenland Mashariki", "HNNOMX": "Saa za Wastani za Kaskazini Magharibi mwa Meksiko", "HNOG": "Saa za Wastani za Greenland Magharibi", "HNPM": "Saa za Wastani ya Saint-Pierre na Miquelon", "HNPMX": "Saa za wastani za pasifiki za Mexico", "HNT": "Saa za Wastani za Newfoundland", "IST": "Saa za Wastani za India", "JDT": "Saa za Mchana za Japani", "JST": "Saa za Wastani za Japani", "LHDT": "Saa za Mchana za Lord Howe", "LHST": "Saa za Wastani za Lord Howe", "MDT": "Saa za Mchana za Mountain", "MESZ": "Saa za Majira ya joto za Ulaya ya Kati", "MEZ": "Saa za Wastani za Ulaya ya kati", "MST": "Saa za Wastani za Mountain", "MYT": "Saa za Malesia", "NZDT": "Saa za Mchana za Nyuzilandi", "NZST": "Saa za Wastani za Nyuzilandi", "OESZ": "Saa za Majira ya joto za Mashariki mwa Ulaya", "OEZ": "Saa za Wastani za Mashariki mwa Ulaya", "PDT": "Saa za Mchana za Pasifiki", "PST": "Saa za Wastani za Pasifiki", "SAST": "Saa za Wastani za Afrika Kusini", "SGT": "Saa za Wastani za Singapoo", "SRT": "Saa za Suriname", "TMST": "Saa za Majira ya Joto za Turkmenistani", "TMT": "Saa za Wastani za Turkmenistani", "UYST": "Saa za Majira ya Joto za Urugwai", "UYT": "Saa za Wastani za Urugwai", "VET": "Saa za Venezuela", "WARST": "Saa za Majira ya Joto za Magharibi mwa Ajentina", "WART": "Saa za Wastani za Magharibi mwa Ajentina", "WAST": "Saa za Majira ya joto za Afrika Magharibi", "WAT": "Saa za Wastani za Afrika Magharibi", "WESZ": "Saa za Majira ya joto za Magharibi mwa Ulaya", "WEZ": "Saa za Wastani za Magharibi mwa Ulaya", "WIB": "Saa za Magharibi mwa Indonesia", "WIT": "Saa za Mashariki mwa Indonesia", "WITA": "Saa za Indonesia ya Kati", "∅∅∅": "Saa za Majira ya Joto za Brazili"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_0a9258144dfcd2f61bc1eaaba53bee40,
			fnFmtCurrency:         fnFmtCurrency_2a7c939a96723e34c09bd13cdb6d28d7,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sw_tz"] = func() locales.Translator {
		return &localen{

			locale:                 "sw_TZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "Saa za Mchana za Australia ya Kati", "ACST": "Saa za Wastani za Australia ya Kati", "ACWDT": "Saa za Mchana za Magharibi ya Kati ya Australia", "ACWST": "Saa za Wastani za Magharibi ya Kati ya Australia", "ADT": "Saa za Mchana za Atlantiki", "AEDT": "Saa za Mchana za Mashariki mwa Australia", "AEST": "Saa za Wastani za Mashariki mwa Australia", "AKDT": "Saa za Mchana za Alaska", "AKST": "Saa za Wastani za Alaska", "ARST": "Saa za Majira ya joto za Argentina", "ART": "Saa za Wastani za Argentina", "AST": "Saa za Wastani za Atlantiki", "AWDT": "Saa za Mchana za Australia Magharibi", "AWST": "Saa za Wastani za Australia Magharibi", "BOT": "Saa za Bolivia", "BT": "Saa za Bhutan", "CAT": "Saa za Afrika ya Kati", "CDT": "Saa za Mchana za Kati", "CHADT": "Saa za Mchana za Chatham", "CHAST": "Saa za Wastani za Chatham", "CLST": "Saa za Majira ya joto za Chile", "CLT": "Saa za Wastani za Chile", "COST": "Saa za Majira ya joto za Colombia", "COT": "Saa za Wastani za Colombia", "CST": "Saa za Wastani za Kati", "ChST": "Saa za Wastani za Chamorro", "EAT": "Saa za Afrika Mashariki", "ECT": "Saa za Ecuador", "EDT": "Saa za Mchana za Mashariki", "EST": "Saa za Wastani za Mashariki", "GFT": "Saa za Guiana ya Ufaransa", "GMT": "Saa za Greenwich", "GST": "Saa za Wastani za Gulf", "GYT": "Saa za Guyana", "HADT": "Saa za Mchana za Hawaii-Aleutian", "HAST": "Saa za Wastani za Hawaii-Aleutian", "HAT": "Saa za Mchana za Newfoundland", "HECU": "Saa za Mchana za Cuba", "HEEG": "Saa za Majira ya joto za Greenland Mashariki", "HENOMX": "Saa za mchana za Mexico Kaskazini Magharibi", "HEOG": "Saa za Majira ya joto za Greenland Magharibi", "HEPM": "Saa za Mchana za Saint-Pierre na Miquelon", "HEPMX": "Saa za mchana za pasifiki za Mexico", "HKST": "Saa za Majira ya joto za Hong Kong", "HKT": "Saa za Wastani za Hong Kong", "HNCU": "Saa za Wastani ya Cuba", "HNEG": "Saa za Wastani za Greenland Mashariki", "HNNOMX": "Saa za Wastani za Mexico Kaskazini Magharibi", "HNOG": "Saa za Wastani za Greenland Magharibi", "HNPM": "Saa za Wastani ya Saint-Pierre na Miquelon", "HNPMX": "Saa za wastani za pasifiki za Mexico", "HNT": "Saa za Wastani za Newfoundland", "IST": "Saa Wastani za India", "JDT": "Saa za Mchana za Japan", "JST": "Saa Wastani za Japan", "LHDT": "Saa za Mchana za Lord Howe", "LHST": "Saa za Wastani za Lord Howe", "MDT": "Saa za Mchana za Mountain", "MESZ": "Saa za Majira ya joto za Ulaya ya Kati", "MEZ": "Saa za Wastani za Ulaya ya kati", "MST": "Saa za Wastani za Mountain", "MYT": "Saa za Malaysia", "NZDT": "Saa za Mchana za New Zealand", "NZST": "Saa za Wastani za New Zealand", "OESZ": "Saa za Majira ya joto za Mashariki mwa Ulaya", "OEZ": "Saa za Wastani za Mashariki mwa Ulaya", "PDT": "Saa za Mchana za Pasifiki", "PST": "Saa za Wastani za Pasifiki", "SAST": "Saa za Wastani za Afrika Kusini", "SGT": "Saa za Wastani za Singapore", "SRT": "Saa za Suriname", "TMST": "Saa za Majira ya joto za Turkmenistan", "TMT": "Saa za Wastani za Turkmenistan", "UYST": "Saa za Majira ya joto za Uruguay", "UYT": "Saa za Wastani za Uruguay", "VET": "Saa za Venezuela", "WARST": "Saa za Majira ya joto za Magharibi mwa Argentina", "WART": "Saa za Wastani za Magharibi mwa Argentina", "WAST": "Saa za Majira ya joto za Afrika Magharibi", "WAT": "Saa za Wastani za Afrika Magharibi", "WESZ": "Saa za Majira ya joto za Magharibi mwa Ulaya", "WEZ": "Saa za Wastani za Magharibi mwa Ulaya", "WIB": "Saa za Magharibi mwa Indonesia", "WIT": "Saa za Mashariki mwa Indonesia", "WITA": "Saa za Indonesia ya Kati", "∅∅∅": "Saa za Majira ya joto za Peru"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["sw_ug"] = func() locales.Translator {
		return &localen{

			locale:                 "sw_UG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			daysWide:               []string{"Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"am", "pm"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Kristo", "Baada ya Kristo"},
			timezones:              map[string]string{"ACDT": "Saa za Mchana za Australia ya Kati", "ACST": "Saa za Wastani za Australia ya Kati", "ACWDT": "Saa za Mchana za Magharibi ya Kati ya Australia", "ACWST": "Saa za Wastani za Magharibi ya Kati ya Australia", "ADT": "Saa za Mchana za Atlantiki", "AEDT": "Saa za Mchana za Mashariki mwa Australia", "AEST": "Saa za Wastani za Mashariki mwa Australia", "AKDT": "Saa za Mchana za Alaska", "AKST": "Saa za Wastani za Alaska", "ARST": "Saa za Majira ya joto za Argentina", "ART": "Saa za Wastani za Argentina", "AST": "Saa za Wastani za Atlantiki", "AWDT": "Saa za Mchana za Australia Magharibi", "AWST": "Saa za Wastani za Australia Magharibi", "BOT": "Saa za Bolivia", "BT": "Saa za Bhutan", "CAT": "Saa za Afrika ya Kati", "CDT": "Saa za Mchana za Kati", "CHADT": "Saa za Mchana za Chatham", "CHAST": "Saa za Wastani za Chatham", "CLST": "Saa za Majira ya joto za Chile", "CLT": "Saa za Wastani za Chile", "COST": "Saa za Majira ya joto za Colombia", "COT": "Saa za Wastani za Colombia", "CST": "Saa za Wastani za Kati", "ChST": "Saa za Wastani za Chamorro", "EAT": "Saa za Afrika Mashariki", "ECT": "Saa za Ecuador", "EDT": "Saa za Mchana za Mashariki", "EST": "Saa za Wastani za Mashariki", "GFT": "Saa za Guiana ya Ufaransa", "GMT": "Saa za Greenwich", "GST": "Saa za Wastani za Gulf", "GYT": "Saa za Guyana", "HADT": "Saa za Mchana za Hawaii-Aleutian", "HAST": "Saa za Wastani za Hawaii-Aleutian", "HAT": "Saa za Mchana za Newfoundland", "HECU": "Saa za Mchana za Cuba", "HEEG": "Saa za Majira ya joto za Greenland Mashariki", "HENOMX": "Saa za mchana za Mexico Kaskazini Magharibi", "HEOG": "Saa za Majira ya joto za Greenland Magharibi", "HEPM": "Saa za Mchana za Saint-Pierre na Miquelon", "HEPMX": "Saa za mchana za pasifiki za Mexico", "HKST": "Saa za Majira ya joto za Hong Kong", "HKT": "Saa za Wastani za Hong Kong", "HNCU": "Saa za Wastani ya Cuba", "HNEG": "Saa za Wastani za Greenland Mashariki", "HNNOMX": "Saa za Wastani za Mexico Kaskazini Magharibi", "HNOG": "Saa za Wastani za Greenland Magharibi", "HNPM": "Saa za Wastani ya Saint-Pierre na Miquelon", "HNPMX": "Saa za wastani za pasifiki za Mexico", "HNT": "Saa za Wastani za Newfoundland", "IST": "Saa Wastani za India", "JDT": "Saa za Mchana za Japan", "JST": "Saa Wastani za Japan", "LHDT": "Saa za Mchana za Lord Howe", "LHST": "Saa za Wastani za Lord Howe", "MDT": "Saa za Mchana za Mountain", "MESZ": "Saa za Majira ya joto za Ulaya ya Kati", "MEZ": "Saa za Wastani za Ulaya ya kati", "MST": "Saa za Wastani za Mountain", "MYT": "Saa za Malaysia", "NZDT": "Saa za Mchana za New Zealand", "NZST": "Saa za Wastani za New Zealand", "OESZ": "Saa za Majira ya joto za Mashariki mwa Ulaya", "OEZ": "Saa za Wastani za Mashariki mwa Ulaya", "PDT": "Saa za Mchana za Pasifiki", "PST": "Saa za Wastani za Pasifiki", "SAST": "Saa za Wastani za Afrika Kusini", "SGT": "Saa za Wastani za Singapore", "SRT": "Saa za Suriname", "TMST": "Saa za Majira ya joto za Turkmenistan", "TMT": "Saa za Wastani za Turkmenistan", "UYST": "Saa za Majira ya joto za Uruguay", "UYT": "Saa za Wastani za Uruguay", "VET": "Saa za Venezuela", "WARST": "Saa za Majira ya joto za Magharibi mwa Argentina", "WART": "Saa za Wastani za Magharibi mwa Argentina", "WAST": "Saa za Majira ya joto za Afrika Magharibi", "WAT": "Saa za Wastani za Afrika Magharibi", "WESZ": "Saa za Majira ya joto za Magharibi mwa Ulaya", "WEZ": "Saa za Wastani za Magharibi mwa Ulaya", "WIB": "Saa za Magharibi mwa Indonesia", "WIT": "Saa za Mashariki mwa Indonesia", "WITA": "Saa za Indonesia ya Kati", "∅∅∅": "Saa za Majira ya joto za Amazon"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ta"] = func() locales.Translator {
		return &localen{

			locale:                 "ta",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ஜன.", "பிப்.", "மார்.", "ஏப்.", "மே", "ஜூன்", "ஜூலை", "ஆக.", "செப்.", "அக்.", "நவ.", "டிச."},
			monthsNarrow:           []string{"", "ஜ", "பி", "மா", "ஏ", "மே", "ஜூ", "ஜூ", "ஆ", "செ", "அ", "ந", "டி"},
			monthsWide:             []string{"", "ஜனவரி", "பிப்ரவரி", "மார்ச்", "ஏப்ரல்", "மே", "ஜூன்", "ஜூலை", "ஆகஸ்ட்", "செப்டம்பர்", "அக்டோபர்", "நவம்பர்", "டிசம்பர்"},
			daysAbbreviated:        []string{"ஞாயி.", "திங்.", "செவ்.", "புத.", "வியா.", "வெள்.", "சனி"},
			daysNarrow:             []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysShort:              []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysWide:               []string{"ஞாயிறு", "திங்கள்", "செவ்வாய்", "புதன்", "வியாழன்", "வெள்ளி", "சனி"},
			periodsAbbreviated:     []string{"முற்பகல்", "பிற்பகல்"},
			periodsNarrow:          []string{"மு.ப", "பி.ப"},
			periodsWide:            []string{"முற்பகல்", "பிற்பகல்"},
			erasAbbreviated:        []string{"கி.மு.", "கி.பி."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"கிறிஸ்துவுக்கு முன்", "அன்னோ டோமினி"},
			timezones:              map[string]string{"ACDT": "ஆஸ்திரேலியன் மத்திய பகலொளி நேரம்", "ACST": "ஆஸ்திரேலியன் மத்திய நிலையான நேரம்", "ACWDT": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய பகலொளி நேரம்", "ACWST": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய நிலையான நேரம்", "ADT": "அட்லாண்டிக் பகலொளி நேரம்", "AEDT": "ஆஸ்திரேலியன் கிழக்கத்திய பகலொளி நேரம்", "AEST": "ஆஸ்திரேலியன் கிழக்கத்திய நிலையான நேரம்", "AKDT": "அலாஸ்கா பகலொளி நேரம்", "AKST": "அலாஸ்கா நிலையான நேரம்", "ARST": "அர்ஜென்டினா கோடை நேரம்", "ART": "அர்ஜென்டினா நிலையான நேரம்", "AST": "அட்லாண்டிக் நிலையான நேரம்", "AWDT": "ஆஸ்திரேலியன் மேற்கத்திய பகலொளி நேரம்", "AWST": "ஆஸ்திரேலியன் மேற்கத்திய நிலையான நேரம்", "BOT": "பொலிவியா நேரம்", "BT": "பூடான் நேரம்", "CAT": "மத்திய ஆப்பிரிக்க நேரம்", "CDT": "மத்திய பகலொளி நேரம்", "CHADT": "சத்தாம் பகலொளி நேரம்", "CHAST": "சத்தாம் நிலையான நேரம்", "CLST": "சிலி கோடை நேரம்", "CLT": "சிலி நிலையான நேரம்", "COST": "கொலம்பியா கோடை நேரம்", "COT": "கொலம்பியா நிலையான நேரம்", "CST": "மத்திய நிலையான நேரம்", "ChST": "சாமோரோ நிலையான நேரம்", "EAT": "கிழக்கு ஆப்பிரிக்க நேரம்", "ECT": "ஈக்வடார் நேரம்", "EDT": "கிழக்கத்திய பகலொளி நேரம்", "EST": "கிழக்கத்திய நிலையான நேரம்", "GFT": "ஃபிரஞ்சு கயானா நேரம்", "GMT": "கிரீன்விச் சராசரி நேரம்", "GST": "வளைகுடா நிலையான நேரம்", "GYT": "கயானா நேரம்", "HADT": "ஹவாய்-அலேஷியன் பகலொளி நேரம்", "HAST": "ஹவாய்-அலேஷியன் நிலையான நேரம்", "HAT": "நியூஃபவுண்ட்லாந்து பகலொளி நேரம்", "HECU": "கியூபா பகலொளி நேரம்", "HEEG": "கிழக்கு கிரீன்லாந்து கோடை நேரம்", "HENOMX": "வடமேற்கு மெக்ஸிகோ பகலொளி நேரம்", "HEOG": "மேற்கு கிரீன்லாந்து கோடை நேரம்", "HEPM": "செயின்ட் பியரி & மிக்குயிலான் பகலொளி நேரம்", "HEPMX": "மெக்ஸிகன் பசிபிக் பகலொளி நேரம்", "HKST": "ஹாங்காங் கோடை நேரம்", "HKT": "ஹாங்காங் நிலையான நேரம்", "HNCU": "கியூபா நிலையான நேரம்", "HNEG": "கிழக்கு கிரீன்லாந்து நிலையான நேரம்", "HNNOMX": "வடமேற்கு மெக்ஸிகோ நிலையான நேரம்", "HNOG": "மேற்கு கிரீன்லாந்து நிலையான நேரம்", "HNPM": "செயின்ட் பியரி & மிக்குயிலான் நிலையான நேரம்", "HNPMX": "மெக்ஸிகன் பசிபிக் நிலையான நேரம்", "HNT": "நியூஃபவுண்ட்லாந்து நிலையான நேரம்", "IST": "இந்திய நிலையான நேரம்", "JDT": "ஜப்பான் பகலொளி நேரம்", "JST": "ஜப்பான் நிலையான நேரம்", "LHDT": "லார்ட் ஹோவ் பகலொளி நேரம்", "LHST": "லார்ட் ஹோவ் நிலையான நேரம்", "MDT": "மவுன்டைன் பகலொளி நேரம்", "MESZ": "மத்திய ஐரோப்பிய கோடை நேரம்", "MEZ": "மத்திய ஐரோப்பிய நிலையான நேரம்", "MST": "மவுன்டைன் நிலையான நேரம்", "MYT": "மலேஷிய நேரம்", "NZDT": "நியூசிலாந்து பகலொளி நேரம்", "NZST": "நியூசிலாந்து நிலையான நேரம்", "OESZ": "கிழக்கத்திய ஐரோப்பிய கோடை நேரம்", "OEZ": "கிழக்கத்திய ஐரோப்பிய நிலையான நேரம்", "PDT": "பசிபிக் பகலொளி நேரம்", "PST": "பசிபிக் நிலையான நேரம்", "SAST": "தென் ஆப்பிரிக்க நிலையான நேரம்", "SGT": "சிங்கப்பூர் நிலையான நேரம்", "SRT": "சுரினாம் நேரம்", "TMST": "துர்க்மெனிஸ்தான் கோடை நேரம்", "TMT": "துர்க்மெனிஸ்தான் நிலையான நேரம்", "UYST": "உருகுவே கோடை நேரம்", "UYT": "உருகுவே நிலையான நேரம்", "VET": "வெனிசுலா நேரம்", "WARST": "மேற்கத்திய அர்ஜென்டினா கோடை நேரம்", "WART": "மேற்கத்திய அர்ஜென்டினா நிலையான நேரம்", "WAST": "மேற்கு ஆப்பிரிக்க கோடை நேரம்", "WAT": "மேற்கு ஆப்பிரிக்க நிலையான நேரம்", "WESZ": "மேற்கத்திய ஐரோப்பிய கோடை நேரம்", "WEZ": "மேற்கத்திய ஐரோப்பிய நிலையான நேரம்", "WIB": "மேற்கத்திய இந்தோனேசிய நேரம்", "WIT": "கிழக்கத்திய இந்தோனேசிய நேரம்", "WITA": "மத்திய இந்தோனேசிய நேரம்", "∅∅∅": "அசோர்ஸ் கோடை நேரம்"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_6fab0d5c33d51b8c50c1ff9de8d0ca73,
			fnFmtTimeLong:         fnFmtTimeLong_473cd7451fd2b0582dbc76fa50f0465a,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ta_in"] = func() locales.Translator {
		return &localen{

			locale:                 "ta_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ஜன.", "பிப்.", "மார்.", "ஏப்.", "மே", "ஜூன்", "ஜூலை", "ஆக.", "செப்.", "அக்.", "நவ.", "டிச."},
			monthsNarrow:           []string{"", "ஜ", "பி", "மா", "ஏ", "மே", "ஜூ", "ஜூ", "ஆ", "செ", "அ", "ந", "டி"},
			monthsWide:             []string{"", "ஜனவரி", "பிப்ரவரி", "மார்ச்", "ஏப்ரல்", "மே", "ஜூன்", "ஜூலை", "ஆகஸ்ட்", "செப்டம்பர்", "அக்டோபர்", "நவம்பர்", "டிசம்பர்"},
			daysAbbreviated:        []string{"ஞாயி.", "திங்.", "செவ்.", "புத.", "வியா.", "வெள்.", "சனி"},
			daysNarrow:             []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysShort:              []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysWide:               []string{"ஞாயிறு", "திங்கள்", "செவ்வாய்", "புதன்", "வியாழன்", "வெள்ளி", "சனி"},
			periodsAbbreviated:     []string{"முற்பகல்", "பிற்பகல்"},
			periodsNarrow:          []string{"மு.ப", "பி.ப"},
			periodsWide:            []string{"முற்பகல்", "பிற்பகல்"},
			erasAbbreviated:        []string{"கி.மு.", "கி.பி."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"கிறிஸ்துவுக்கு முன்", "அன்னோ டோமினி"},
			timezones:              map[string]string{"ACDT": "ஆஸ்திரேலியன் மத்திய பகலொளி நேரம்", "ACST": "ஆஸ்திரேலியன் மத்திய நிலையான நேரம்", "ACWDT": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய பகலொளி நேரம்", "ACWST": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய நிலையான நேரம்", "ADT": "அட்லாண்டிக் பகலொளி நேரம்", "AEDT": "ஆஸ்திரேலியன் கிழக்கத்திய பகலொளி நேரம்", "AEST": "ஆஸ்திரேலியன் கிழக்கத்திய நிலையான நேரம்", "AKDT": "அலாஸ்கா பகலொளி நேரம்", "AKST": "அலாஸ்கா நிலையான நேரம்", "ARST": "அர்ஜென்டினா கோடை நேரம்", "ART": "அர்ஜென்டினா நிலையான நேரம்", "AST": "அட்லாண்டிக் நிலையான நேரம்", "AWDT": "ஆஸ்திரேலியன் மேற்கத்திய பகலொளி நேரம்", "AWST": "ஆஸ்திரேலியன் மேற்கத்திய நிலையான நேரம்", "BOT": "பொலிவியா நேரம்", "BT": "பூடான் நேரம்", "CAT": "மத்திய ஆப்பிரிக்க நேரம்", "CDT": "மத்திய பகலொளி நேரம்", "CHADT": "சத்தாம் பகலொளி நேரம்", "CHAST": "சத்தாம் நிலையான நேரம்", "CLST": "சிலி கோடை நேரம்", "CLT": "சிலி நிலையான நேரம்", "COST": "கொலம்பியா கோடை நேரம்", "COT": "கொலம்பியா நிலையான நேரம்", "CST": "மத்திய நிலையான நேரம்", "ChST": "சாமோரோ நிலையான நேரம்", "EAT": "கிழக்கு ஆப்பிரிக்க நேரம்", "ECT": "ஈக்வடார் நேரம்", "EDT": "கிழக்கத்திய பகலொளி நேரம்", "EST": "கிழக்கத்திய நிலையான நேரம்", "GFT": "ஃபிரஞ்சு கயானா நேரம்", "GMT": "கிரீன்விச் சராசரி நேரம்", "GST": "வளைகுடா நிலையான நேரம்", "GYT": "கயானா நேரம்", "HADT": "ஹவாய்-அலேஷியன் பகலொளி நேரம்", "HAST": "ஹவாய்-அலேஷியன் நிலையான நேரம்", "HAT": "நியூஃபவுண்ட்லாந்து பகலொளி நேரம்", "HECU": "கியூபா பகலொளி நேரம்", "HEEG": "கிழக்கு கிரீன்லாந்து கோடை நேரம்", "HENOMX": "வடமேற்கு மெக்ஸிகோ பகலொளி நேரம்", "HEOG": "மேற்கு கிரீன்லாந்து கோடை நேரம்", "HEPM": "செயின்ட் பியரி & மிக்குயிலான் பகலொளி நேரம்", "HEPMX": "மெக்ஸிகன் பசிபிக் பகலொளி நேரம்", "HKST": "ஹாங்காங் கோடை நேரம்", "HKT": "ஹாங்காங் நிலையான நேரம்", "HNCU": "கியூபா நிலையான நேரம்", "HNEG": "கிழக்கு கிரீன்லாந்து நிலையான நேரம்", "HNNOMX": "வடமேற்கு மெக்ஸிகோ நிலையான நேரம்", "HNOG": "மேற்கு கிரீன்லாந்து நிலையான நேரம்", "HNPM": "செயின்ட் பியரி & மிக்குயிலான் நிலையான நேரம்", "HNPMX": "மெக்ஸிகன் பசிபிக் நிலையான நேரம்", "HNT": "நியூஃபவுண்ட்லாந்து நிலையான நேரம்", "IST": "இந்திய நிலையான நேரம்", "JDT": "ஜப்பான் பகலொளி நேரம்", "JST": "ஜப்பான் நிலையான நேரம்", "LHDT": "லார்ட் ஹோவ் பகலொளி நேரம்", "LHST": "லார்ட் ஹோவ் நிலையான நேரம்", "MDT": "மவுன்டைன் பகலொளி நேரம்", "MESZ": "மத்திய ஐரோப்பிய கோடை நேரம்", "MEZ": "மத்திய ஐரோப்பிய நிலையான நேரம்", "MST": "மவுன்டைன் நிலையான நேரம்", "MYT": "மலேஷிய நேரம்", "NZDT": "நியூசிலாந்து பகலொளி நேரம்", "NZST": "நியூசிலாந்து நிலையான நேரம்", "OESZ": "கிழக்கத்திய ஐரோப்பிய கோடை நேரம்", "OEZ": "கிழக்கத்திய ஐரோப்பிய நிலையான நேரம்", "PDT": "பசிபிக் பகலொளி நேரம்", "PST": "பசிபிக் நிலையான நேரம்", "SAST": "தென் ஆப்பிரிக்க நிலையான நேரம்", "SGT": "சிங்கப்பூர் நிலையான நேரம்", "SRT": "சுரினாம் நேரம்", "TMST": "துர்க்மெனிஸ்தான் கோடை நேரம்", "TMT": "துர்க்மெனிஸ்தான் நிலையான நேரம்", "UYST": "உருகுவே கோடை நேரம்", "UYT": "உருகுவே நிலையான நேரம்", "VET": "வெனிசுலா நேரம்", "WARST": "மேற்கத்திய அர்ஜென்டினா கோடை நேரம்", "WART": "மேற்கத்திய அர்ஜென்டினா நிலையான நேரம்", "WAST": "மேற்கு ஆப்பிரிக்க கோடை நேரம்", "WAT": "மேற்கு ஆப்பிரிக்க நிலையான நேரம்", "WESZ": "மேற்கத்திய ஐரோப்பிய கோடை நேரம்", "WEZ": "மேற்கத்திய ஐரோப்பிய நிலையான நேரம்", "WIB": "மேற்கத்திய இந்தோனேசிய நேரம்", "WIT": "கிழக்கத்திய இந்தோனேசிய நேரம்", "WITA": "மத்திய இந்தோனேசிய நேரம்", "∅∅∅": "பெரு கோடை நேரம்"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_6fab0d5c33d51b8c50c1ff9de8d0ca73,
			fnFmtTimeLong:         fnFmtTimeLong_473cd7451fd2b0582dbc76fa50f0465a,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ta_lk"] = func() locales.Translator {
		return &localen{

			locale:                 "ta_LK",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "Rs.", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ஜன.", "பிப்.", "மார்.", "ஏப்.", "மே", "ஜூன்", "ஜூலை", "ஆக.", "செப்.", "அக்.", "நவ.", "டிச."},
			monthsNarrow:           []string{"", "ஜ", "பி", "மா", "ஏ", "மே", "ஜூ", "ஜூ", "ஆ", "செ", "அ", "ந", "டி"},
			monthsWide:             []string{"", "ஜனவரி", "பிப்ரவரி", "மார்ச்", "ஏப்ரல்", "மே", "ஜூன்", "ஜூலை", "ஆகஸ்ட்", "செப்டம்பர்", "அக்டோபர்", "நவம்பர்", "டிசம்பர்"},
			daysAbbreviated:        []string{"ஞாயி.", "திங்.", "செவ்.", "புத.", "வியா.", "வெள்.", "சனி"},
			daysNarrow:             []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysShort:              []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysWide:               []string{"ஞாயிறு", "திங்கள்", "செவ்வாய்", "புதன்", "வியாழன்", "வெள்ளி", "சனி"},
			periodsAbbreviated:     []string{"முற்பகல்", "பிற்பகல்"},
			periodsNarrow:          []string{"மு.ப", "பி.ப"},
			periodsWide:            []string{"முற்பகல்", "பிற்பகல்"},
			erasAbbreviated:        []string{"கி.மு.", "கி.பி."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"கிறிஸ்துவுக்கு முன்", "அன்னோ டோமினி"},
			timezones:              map[string]string{"ACDT": "ஆஸ்திரேலியன் மத்திய பகலொளி நேரம்", "ACST": "ஆஸ்திரேலியன் மத்திய நிலையான நேரம்", "ACWDT": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய பகலொளி நேரம்", "ACWST": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய நிலையான நேரம்", "ADT": "அட்லாண்டிக் பகலொளி நேரம்", "AEDT": "ஆஸ்திரேலியன் கிழக்கத்திய பகலொளி நேரம்", "AEST": "ஆஸ்திரேலியன் கிழக்கத்திய நிலையான நேரம்", "AKDT": "அலாஸ்கா பகலொளி நேரம்", "AKST": "அலாஸ்கா நிலையான நேரம்", "ARST": "அர்ஜென்டினா கோடை நேரம்", "ART": "அர்ஜென்டினா நிலையான நேரம்", "AST": "அட்லாண்டிக் நிலையான நேரம்", "AWDT": "ஆஸ்திரேலியன் மேற்கத்திய பகலொளி நேரம்", "AWST": "ஆஸ்திரேலியன் மேற்கத்திய நிலையான நேரம்", "BOT": "பொலிவியா நேரம்", "BT": "பூடான் நேரம்", "CAT": "மத்திய ஆப்பிரிக்க நேரம்", "CDT": "மத்திய பகலொளி நேரம்", "CHADT": "சத்தாம் பகலொளி நேரம்", "CHAST": "சத்தாம் நிலையான நேரம்", "CLST": "சிலி கோடை நேரம்", "CLT": "சிலி நிலையான நேரம்", "COST": "கொலம்பியா கோடை நேரம்", "COT": "கொலம்பியா நிலையான நேரம்", "CST": "மத்திய நிலையான நேரம்", "ChST": "சாமோரோ நிலையான நேரம்", "EAT": "கிழக்கு ஆப்பிரிக்க நேரம்", "ECT": "ஈக்வடார் நேரம்", "EDT": "கிழக்கத்திய பகலொளி நேரம்", "EST": "கிழக்கத்திய நிலையான நேரம்", "GFT": "ஃபிரஞ்சு கயானா நேரம்", "GMT": "கிரீன்விச் சராசரி நேரம்", "GST": "வளைகுடா நிலையான நேரம்", "GYT": "கயானா நேரம்", "HADT": "ஹவாய்-அலேஷியன் பகலொளி நேரம்", "HAST": "ஹவாய்-அலேஷியன் நிலையான நேரம்", "HAT": "நியூஃபவுண்ட்லாந்து பகலொளி நேரம்", "HECU": "கியூபா பகலொளி நேரம்", "HEEG": "கிழக்கு கிரீன்லாந்து கோடை நேரம்", "HENOMX": "வடமேற்கு மெக்ஸிகோ பகலொளி நேரம்", "HEOG": "மேற்கு கிரீன்லாந்து கோடை நேரம்", "HEPM": "செயின்ட் பியரி & மிக்குயிலான் பகலொளி நேரம்", "HEPMX": "மெக்ஸிகன் பசிபிக் பகலொளி நேரம்", "HKST": "ஹாங்காங் கோடை நேரம்", "HKT": "ஹாங்காங் நிலையான நேரம்", "HNCU": "கியூபா நிலையான நேரம்", "HNEG": "கிழக்கு கிரீன்லாந்து நிலையான நேரம்", "HNNOMX": "வடமேற்கு மெக்ஸிகோ நிலையான நேரம்", "HNOG": "மேற்கு கிரீன்லாந்து நிலையான நேரம்", "HNPM": "செயின்ட் பியரி & மிக்குயிலான் நிலையான நேரம்", "HNPMX": "மெக்ஸிகன் பசிபிக் நிலையான நேரம்", "HNT": "நியூஃபவுண்ட்லாந்து நிலையான நேரம்", "IST": "இந்திய நிலையான நேரம்", "JDT": "ஜப்பான் பகலொளி நேரம்", "JST": "ஜப்பான் நிலையான நேரம்", "LHDT": "லார்ட் ஹோவ் பகலொளி நேரம்", "LHST": "லார்ட் ஹோவ் நிலையான நேரம்", "MDT": "மவுன்டைன் பகலொளி நேரம்", "MESZ": "மத்திய ஐரோப்பிய கோடை நேரம்", "MEZ": "மத்திய ஐரோப்பிய நிலையான நேரம்", "MST": "மவுன்டைன் நிலையான நேரம்", "MYT": "மலேஷிய நேரம்", "NZDT": "நியூசிலாந்து பகலொளி நேரம்", "NZST": "நியூசிலாந்து நிலையான நேரம்", "OESZ": "கிழக்கத்திய ஐரோப்பிய கோடை நேரம்", "OEZ": "கிழக்கத்திய ஐரோப்பிய நிலையான நேரம்", "PDT": "பசிபிக் பகலொளி நேரம்", "PST": "பசிபிக் நிலையான நேரம்", "SAST": "தென் ஆப்பிரிக்க நிலையான நேரம்", "SGT": "சிங்கப்பூர் நிலையான நேரம்", "SRT": "சுரினாம் நேரம்", "TMST": "துர்க்மெனிஸ்தான் கோடை நேரம்", "TMT": "துர்க்மெனிஸ்தான் நிலையான நேரம்", "UYST": "உருகுவே கோடை நேரம்", "UYT": "உருகுவே நிலையான நேரம்", "VET": "வெனிசுலா நேரம்", "WARST": "மேற்கத்திய அர்ஜென்டினா கோடை நேரம்", "WART": "மேற்கத்திய அர்ஜென்டினா நிலையான நேரம்", "WAST": "மேற்கு ஆப்பிரிக்க கோடை நேரம்", "WAT": "மேற்கு ஆப்பிரிக்க நிலையான நேரம்", "WESZ": "மேற்கத்திய ஐரோப்பிய கோடை நேரம்", "WEZ": "மேற்கத்திய ஐரோப்பிய நிலையான நேரம்", "WIB": "மேற்கத்திய இந்தோனேசிய நேரம்", "WIT": "கிழக்கத்திய இந்தோனேசிய நேரம்", "WITA": "மத்திய இந்தோனேசிய நேரம்", "∅∅∅": "பிரேசிலியா கோடை நேரம்"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ta_my"] = func() locales.Translator {
		return &localen{

			locale:                 "ta_MY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "S$", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ஜன.", "பிப்.", "மார்.", "ஏப்.", "மே", "ஜூன்", "ஜூலை", "ஆக.", "செப்.", "அக்.", "நவ.", "டிச."},
			monthsNarrow:           []string{"", "ஜ", "பி", "மா", "ஏ", "மே", "ஜூ", "ஜூ", "ஆ", "செ", "அ", "ந", "டி"},
			monthsWide:             []string{"", "ஜனவரி", "பிப்ரவரி", "மார்ச்", "ஏப்ரல்", "மே", "ஜூன்", "ஜூலை", "ஆகஸ்ட்", "செப்டம்பர்", "அக்டோபர்", "நவம்பர்", "டிசம்பர்"},
			daysAbbreviated:        []string{"ஞாயி.", "திங்.", "செவ்.", "புத.", "வியா.", "வெள்.", "சனி"},
			daysNarrow:             []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysShort:              []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysWide:               []string{"ஞாயிறு", "திங்கள்", "செவ்வாய்", "புதன்", "வியாழன்", "வெள்ளி", "சனி"},
			periodsAbbreviated:     []string{"முற்பகல்", "பிற்பகல்"},
			periodsNarrow:          []string{"மு.ப", "பி.ப"},
			periodsWide:            []string{"முற்பகல்", "பிற்பகல்"},
			erasAbbreviated:        []string{"கி.மு.", "கி.பி."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"கிறிஸ்துவுக்கு முன்", "அன்னோ டோமினி"},
			timezones:              map[string]string{"ACDT": "ஆஸ்திரேலியன் மத்திய பகலொளி நேரம்", "ACST": "ஆஸ்திரேலியன் மத்திய நிலையான நேரம்", "ACWDT": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய பகலொளி நேரம்", "ACWST": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய நிலையான நேரம்", "ADT": "அட்லாண்டிக் பகலொளி நேரம்", "AEDT": "ஆஸ்திரேலியன் கிழக்கத்திய பகலொளி நேரம்", "AEST": "ஆஸ்திரேலியன் கிழக்கத்திய நிலையான நேரம்", "AKDT": "அலாஸ்கா பகலொளி நேரம்", "AKST": "அலாஸ்கா நிலையான நேரம்", "ARST": "அர்ஜென்டினா கோடை நேரம்", "ART": "அர்ஜென்டினா நிலையான நேரம்", "AST": "அட்லாண்டிக் நிலையான நேரம்", "AWDT": "ஆஸ்திரேலியன் மேற்கத்திய பகலொளி நேரம்", "AWST": "ஆஸ்திரேலியன் மேற்கத்திய நிலையான நேரம்", "BOT": "பொலிவியா நேரம்", "BT": "பூடான் நேரம்", "CAT": "மத்திய ஆப்பிரிக்க நேரம்", "CDT": "மத்திய பகலொளி நேரம்", "CHADT": "சத்தாம் பகலொளி நேரம்", "CHAST": "சத்தாம் நிலையான நேரம்", "CLST": "சிலி கோடை நேரம்", "CLT": "சிலி நிலையான நேரம்", "COST": "கொலம்பியா கோடை நேரம்", "COT": "கொலம்பியா நிலையான நேரம்", "CST": "மத்திய நிலையான நேரம்", "ChST": "சாமோரோ நிலையான நேரம்", "EAT": "கிழக்கு ஆப்பிரிக்க நேரம்", "ECT": "ஈக்வடார் நேரம்", "EDT": "கிழக்கத்திய பகலொளி நேரம்", "EST": "கிழக்கத்திய நிலையான நேரம்", "GFT": "ஃபிரஞ்சு கயானா நேரம்", "GMT": "கிரீன்விச் சராசரி நேரம்", "GST": "வளைகுடா நிலையான நேரம்", "GYT": "கயானா நேரம்", "HADT": "ஹவாய்-அலேஷியன் பகலொளி நேரம்", "HAST": "ஹவாய்-அலேஷியன் நிலையான நேரம்", "HAT": "நியூஃபவுண்ட்லாந்து பகலொளி நேரம்", "HECU": "கியூபா பகலொளி நேரம்", "HEEG": "கிழக்கு கிரீன்லாந்து கோடை நேரம்", "HENOMX": "வடமேற்கு மெக்ஸிகோ பகலொளி நேரம்", "HEOG": "மேற்கு கிரீன்லாந்து கோடை நேரம்", "HEPM": "செயின்ட் பியரி & மிக்குயிலான் பகலொளி நேரம்", "HEPMX": "மெக்ஸிகன் பசிபிக் பகலொளி நேரம்", "HKST": "ஹாங்காங் கோடை நேரம்", "HKT": "ஹாங்காங் நிலையான நேரம்", "HNCU": "கியூபா நிலையான நேரம்", "HNEG": "கிழக்கு கிரீன்லாந்து நிலையான நேரம்", "HNNOMX": "வடமேற்கு மெக்ஸிகோ நிலையான நேரம்", "HNOG": "மேற்கு கிரீன்லாந்து நிலையான நேரம்", "HNPM": "செயின்ட் பியரி & மிக்குயிலான் நிலையான நேரம்", "HNPMX": "மெக்ஸிகன் பசிபிக் நிலையான நேரம்", "HNT": "நியூஃபவுண்ட்லாந்து நிலையான நேரம்", "IST": "இந்திய நிலையான நேரம்", "JDT": "ஜப்பான் பகலொளி நேரம்", "JST": "ஜப்பான் நிலையான நேரம்", "LHDT": "லார்ட் ஹோவ் பகலொளி நேரம்", "LHST": "லார்ட் ஹோவ் நிலையான நேரம்", "MDT": "மவுன்டைன் பகலொளி நேரம்", "MESZ": "மத்திய ஐரோப்பிய கோடை நேரம்", "MEZ": "மத்திய ஐரோப்பிய நிலையான நேரம்", "MST": "மவுன்டைன் நிலையான நேரம்", "MYT": "மலேஷிய நேரம்", "NZDT": "நியூசிலாந்து பகலொளி நேரம்", "NZST": "நியூசிலாந்து நிலையான நேரம்", "OESZ": "கிழக்கத்திய ஐரோப்பிய கோடை நேரம்", "OEZ": "கிழக்கத்திய ஐரோப்பிய நிலையான நேரம்", "PDT": "பசிபிக் பகலொளி நேரம்", "PST": "பசிபிக் நிலையான நேரம்", "SAST": "தென் ஆப்பிரிக்க நிலையான நேரம்", "SGT": "சிங்கப்பூர் நிலையான நேரம்", "SRT": "சுரினாம் நேரம்", "TMST": "துர்க்மெனிஸ்தான் கோடை நேரம்", "TMT": "துர்க்மெனிஸ்தான் நிலையான நேரம்", "UYST": "உருகுவே கோடை நேரம்", "UYT": "உருகுவே நிலையான நேரம்", "VET": "வெனிசுலா நேரம்", "WARST": "மேற்கத்திய அர்ஜென்டினா கோடை நேரம்", "WART": "மேற்கத்திய அர்ஜென்டினா நிலையான நேரம்", "WAST": "மேற்கு ஆப்பிரிக்க கோடை நேரம்", "WAT": "மேற்கு ஆப்பிரிக்க நிலையான நேரம்", "WESZ": "மேற்கத்திய ஐரோப்பிய கோடை நேரம்", "WEZ": "மேற்கத்திய ஐரோப்பிய நிலையான நேரம்", "WIB": "மேற்கத்திய இந்தோனேசிய நேரம்", "WIT": "கிழக்கத்திய இந்தோனேசிய நேரம்", "WITA": "மத்திய இந்தோனேசிய நேரம்", "∅∅∅": "அசோர்ஸ் கோடை நேரம்"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_6fab0d5c33d51b8c50c1ff9de8d0ca73,
			fnFmtTimeLong:         fnFmtTimeLong_473cd7451fd2b0582dbc76fa50f0465a,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ta_sg"] = func() locales.Translator {
		return &localen{

			locale:                 "ta_SG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "RM", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "$", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "ஜன.", "பிப்.", "மார்.", "ஏப்.", "மே", "ஜூன்", "ஜூலை", "ஆக.", "செப்.", "அக்.", "நவ.", "டிச."},
			monthsNarrow:           []string{"", "ஜ", "பி", "மா", "ஏ", "மே", "ஜூ", "ஜூ", "ஆ", "செ", "அ", "ந", "டி"},
			monthsWide:             []string{"", "ஜனவரி", "பிப்ரவரி", "மார்ச்", "ஏப்ரல்", "மே", "ஜூன்", "ஜூலை", "ஆகஸ்ட்", "செப்டம்பர்", "அக்டோபர்", "நவம்பர்", "டிசம்பர்"},
			daysAbbreviated:        []string{"ஞாயி.", "திங்.", "செவ்.", "புத.", "வியா.", "வெள்.", "சனி"},
			daysNarrow:             []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysShort:              []string{"ஞா", "தி", "செ", "பு", "வி", "வெ", "ச"},
			daysWide:               []string{"ஞாயிறு", "திங்கள்", "செவ்வாய்", "புதன்", "வியாழன்", "வெள்ளி", "சனி"},
			periodsAbbreviated:     []string{"முற்பகல்", "பிற்பகல்"},
			periodsNarrow:          []string{"மு.ப", "பி.ப"},
			periodsWide:            []string{"முற்பகல்", "பிற்பகல்"},
			erasAbbreviated:        []string{"கி.மு.", "கி.பி."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"கிறிஸ்துவுக்கு முன்", "அன்னோ டோமினி"},
			timezones:              map[string]string{"ACDT": "ஆஸ்திரேலியன் மத்திய பகலொளி நேரம்", "ACST": "ஆஸ்திரேலியன் மத்திய நிலையான நேரம்", "ACWDT": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய பகலொளி நேரம்", "ACWST": "ஆஸ்திரேலியன் மத்திய மேற்கத்திய நிலையான நேரம்", "ADT": "அட்லாண்டிக் பகலொளி நேரம்", "AEDT": "ஆஸ்திரேலியன் கிழக்கத்திய பகலொளி நேரம்", "AEST": "ஆஸ்திரேலியன் கிழக்கத்திய நிலையான நேரம்", "AKDT": "அலாஸ்கா பகலொளி நேரம்", "AKST": "அலாஸ்கா நிலையான நேரம்", "ARST": "அர்ஜென்டினா கோடை நேரம்", "ART": "அர்ஜென்டினா நிலையான நேரம்", "AST": "அட்லாண்டிக் நிலையான நேரம்", "AWDT": "ஆஸ்திரேலியன் மேற்கத்திய பகலொளி நேரம்", "AWST": "ஆஸ்திரேலியன் மேற்கத்திய நிலையான நேரம்", "BOT": "பொலிவியா நேரம்", "BT": "பூடான் நேரம்", "CAT": "மத்திய ஆப்பிரிக்க நேரம்", "CDT": "மத்திய பகலொளி நேரம்", "CHADT": "சத்தாம் பகலொளி நேரம்", "CHAST": "சத்தாம் நிலையான நேரம்", "CLST": "சிலி கோடை நேரம்", "CLT": "சிலி நிலையான நேரம்", "COST": "கொலம்பியா கோடை நேரம்", "COT": "கொலம்பியா நிலையான நேரம்", "CST": "மத்திய நிலையான நேரம்", "ChST": "சாமோரோ நிலையான நேரம்", "EAT": "கிழக்கு ஆப்பிரிக்க நேரம்", "ECT": "ஈக்வடார் நேரம்", "EDT": "கிழக்கத்திய பகலொளி நேரம்", "EST": "கிழக்கத்திய நிலையான நேரம்", "GFT": "ஃபிரஞ்சு கயானா நேரம்", "GMT": "கிரீன்விச் சராசரி நேரம்", "GST": "வளைகுடா நிலையான நேரம்", "GYT": "கயானா நேரம்", "HADT": "ஹவாய்-அலேஷியன் பகலொளி நேரம்", "HAST": "ஹவாய்-அலேஷியன் நிலையான நேரம்", "HAT": "நியூஃபவுண்ட்லாந்து பகலொளி நேரம்", "HECU": "கியூபா பகலொளி நேரம்", "HEEG": "கிழக்கு கிரீன்லாந்து கோடை நேரம்", "HENOMX": "வடமேற்கு மெக்ஸிகோ பகலொளி நேரம்", "HEOG": "மேற்கு கிரீன்லாந்து கோடை நேரம்", "HEPM": "செயின்ட் பியரி & மிக்குயிலான் பகலொளி நேரம்", "HEPMX": "மெக்ஸிகன் பசிபிக் பகலொளி நேரம்", "HKST": "ஹாங்காங் கோடை நேரம்", "HKT": "ஹாங்காங் நிலையான நேரம்", "HNCU": "கியூபா நிலையான நேரம்", "HNEG": "கிழக்கு கிரீன்லாந்து நிலையான நேரம்", "HNNOMX": "வடமேற்கு மெக்ஸிகோ நிலையான நேரம்", "HNOG": "மேற்கு கிரீன்லாந்து நிலையான நேரம்", "HNPM": "செயின்ட் பியரி & மிக்குயிலான் நிலையான நேரம்", "HNPMX": "மெக்ஸிகன் பசிபிக் நிலையான நேரம்", "HNT": "நியூஃபவுண்ட்லாந்து நிலையான நேரம்", "IST": "இந்திய நிலையான நேரம்", "JDT": "ஜப்பான் பகலொளி நேரம்", "JST": "ஜப்பான் நிலையான நேரம்", "LHDT": "லார்ட் ஹோவ் பகலொளி நேரம்", "LHST": "லார்ட் ஹோவ் நிலையான நேரம்", "MDT": "மவுன்டைன் பகலொளி நேரம்", "MESZ": "மத்திய ஐரோப்பிய கோடை நேரம்", "MEZ": "மத்திய ஐரோப்பிய நிலையான நேரம்", "MST": "மவுன்டைன் நிலையான நேரம்", "MYT": "மலேஷிய நேரம்", "NZDT": "நியூசிலாந்து பகலொளி நேரம்", "NZST": "நியூசிலாந்து நிலையான நேரம்", "OESZ": "கிழக்கத்திய ஐரோப்பிய கோடை நேரம்", "OEZ": "கிழக்கத்திய ஐரோப்பிய நிலையான நேரம்", "PDT": "பசிபிக் பகலொளி நேரம்", "PST": "பசிபிக் நிலையான நேரம்", "SAST": "தென் ஆப்பிரிக்க நிலையான நேரம்", "SGT": "சிங்கப்பூர் நிலையான நேரம்", "SRT": "சுரினாம் நேரம்", "TMST": "துர்க்மெனிஸ்தான் கோடை நேரம்", "TMT": "துர்க்மெனிஸ்தான் நிலையான நேரம்", "UYST": "உருகுவே கோடை நேரம்", "UYT": "உருகுவே நிலையான நேரம்", "VET": "வெனிசுலா நேரம்", "WARST": "மேற்கத்திய அர்ஜென்டினா கோடை நேரம்", "WART": "மேற்கத்திய அர்ஜென்டினா நிலையான நேரம்", "WAST": "மேற்கு ஆப்பிரிக்க கோடை நேரம்", "WAT": "மேற்கு ஆப்பிரிக்க நிலையான நேரம்", "WESZ": "மேற்கத்திய ஐரோப்பிய கோடை நேரம்", "WEZ": "மேற்கத்திய ஐரோப்பிய நிலையான நேரம்", "WIB": "மேற்கத்திய இந்தோனேசிய நேரம்", "WIT": "கிழக்கத்திய இந்தோனேசிய நேரம்", "WITA": "மத்திய இந்தோனேசிய நேரம்", "∅∅∅": "அமேசான் கோடை நேரம்"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_6fab0d5c33d51b8c50c1ff9de8d0ca73,
			fnFmtTimeLong:         fnFmtTimeLong_473cd7451fd2b0582dbc76fa50f0465a,
			fnFmtTimeMedium:       fnFmtTimeMedium_2116a9d777445f659bcbbe24b3c994ee,
			fnFmtTimeShort:        fnFmtTimeShort_c7eeb539163b08e3ba7211630cad2b4b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["te"] = func() locales.Translator {
		return &localen{

			locale:                 "te",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "జన", "ఫిబ్ర", "మార్చి", "ఏప్రి", "మే", "జూన్", "జులై", "ఆగ", "సెప్టెం", "అక్టో", "నవం", "డిసెం"},
			monthsNarrow:           []string{"", "జ", "ఫి", "మా", "ఏ", "మే", "జూ", "జు", "ఆ", "సె", "అ", "న", "డి"},
			monthsWide:             []string{"", "జనవరి", "ఫిబ్రవరి", "మార్చి", "ఏప్రిల్", "మే", "జూన్", "జులై", "ఆగస్టు", "సెప్టెంబర్", "అక్టోబర్", "నవంబర్", "డిసెంబర్"},
			daysAbbreviated:        []string{"ఆది", "సోమ", "మంగళ", "బుధ", "గురు", "శుక్ర", "శని"},
			daysNarrow:             []string{"ఆ", "సో", "మ", "బు", "గు", "శు", "శ"},
			daysShort:              []string{"ఆది", "సోమ", "మం", "బుధ", "గురు", "శుక్ర", "శని"},
			daysWide:               []string{"ఆదివారం", "సోమవారం", "మంగళవారం", "బుధవారం", "గురువారం", "శుక్రవారం", "శనివారం"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"ఉ", "సా"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"క్రీపూ", "క్రీశ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"క్రీస్తు పూర్వం", "క్రీస్తు శకం"},
			timezones:              map[string]string{"ACDT": "ఆస్ట్రేలియా మధ్యమ పగటి వెలుతురు సమయం", "ACST": "ఆస్ట్రేలియా మధ్యమ ప్రామాణిక సమయం", "ACWDT": "ఆస్ట్రేలియా మధ్యమ పశ్చిమ పగటి వెలుతురు సమయం", "ACWST": "మధ్యమ ఆస్ట్రేలియన్ పశ్చిమ ప్రామాణిక సమయం", "ADT": "అట్లాంటిక్ పగటి వెలుతురు సమయం", "AEDT": "ఆస్ట్రేలియన్ తూర్పు పగటి వెలుతురు సమయం", "AEST": "ఆస్ట్రేలియన్ తూర్పు ప్రామాణిక సమయం", "AKDT": "అలాస్కా పగటి వెలుతురు సమయం", "AKST": "అలాస్కా ప్రామాణిక సమయం", "ARST": "ఆర్జెంటీనా వేసవి సమయం", "ART": "అర్జెంటీనా ప్రామాణిక సమయం", "AST": "అట్లాంటిక్ ప్రామాణిక సమయం", "AWDT": "ఆస్ట్రేలియన్ పశ్చిమ పగటి వెలుతురు సమయం", "AWST": "ఆస్ట్రేలియన్ పశ్చిమ ప్రామాణిక సమయం", "BOT": "బొలీవియా సమయం", "BT": "భూటాన్ సమయం", "CAT": "సెంట్రల్ ఆఫ్రికా సమయం", "CDT": "మధ్యమ పగటి వెలుతురు సమయం", "CHADT": "చాథమ్ పగటి వెలుతురు సమయం", "CHAST": "చాథమ్ ప్రామాణిక సమయం", "CLST": "చిలీ వేసవి సమయం", "CLT": "చిలీ ప్రామాణిక సమయం", "COST": "కొలంబియా వేసవి సమయం", "COT": "కొలంబియా ప్రామాణిక సమయం", "CST": "మధ్యమ ప్రామాణిక సమయం", "ChST": "చామర్రో ప్రామాణిక సమయం", "EAT": "తూర్పు ఆఫ్రికా సమయం", "ECT": "ఈక్వడార్ సమయం", "EDT": "తూర్పు పగటి వెలుతురు సమయం", "EST": "తూర్పు ప్రామాణిక సమయం", "GFT": "ఫ్రెంచ్ గయానా సమయం", "GMT": "గ్రీన్\u200cవిచ్ సగటు సమయం", "GST": "గల్ఫ్ ప్రామాణిక సమయం", "GYT": "గయానా సమయం", "HADT": "హవాయ్-అల్యూషియన్ పగటి వెలుతురు సమయం", "HAST": "హవాయ్-అల్యూషియన్ ప్రామాణిక సమయం", "HAT": "న్యూఫౌండ్\u200cల్యాండ్ పగటి వెలుతురు సమయం", "HECU": "క్యూబా పగటి వెలుతురు సమయం", "HEEG": "తూర్పు గ్రీన్\u200cల్యాండ్ వేసవి సమయం", "HENOMX": "వాయువ్య మెక్సికో పగటి వెలుతురు సమయం", "HEOG": "పశ్చిమ గ్రీన్\u200cల్యాండ్ వేసవి సమయం", "HEPM": "సెయింట్ పియర్ మరియు మిక్వెలాన్ పగటి వెలుతురు సమయం", "HEPMX": "మెక్సికన్ పసిఫిక్ పగటి వెలుతురు సమయం", "HKST": "హాంకాంగ్ వేసవి సమయం", "HKT": "హాంకాంగ్ ప్రామాణిక సమయం", "HNCU": "క్యూబా ప్రామాణిక సమయం", "HNEG": "తూర్పు గ్రీన్\u200cల్యాండ్ ప్రామాణిక సమయం", "HNNOMX": "వాయువ్య మెక్సికో ప్రామాణిక సమయం", "HNOG": "పశ్చిమ గ్రీన్\u200cల్యాండ్ ప్రామాణిక సమయం", "HNPM": "సెయింట్ పియెర్ మరియు మిక్వెలాన్ ప్రామాణిక సమయం", "HNPMX": "మెక్సికన్ పసిఫిక్ ప్రామాణిక సమయం", "HNT": "న్యూఫౌండ్\u200cల్యాండ్ ప్రామాణిక సమయం", "IST": "భారతదేశ సమయం", "JDT": "జపాన్ పగటి వెలుతురు సమయం", "JST": "జపాన్ ప్రామాణిక సమయం", "LHDT": "లార్డ్ హోవ్ పగటి సమయం", "LHST": "లార్డ్ హోవ్ ప్రామాణిక సమయం", "MDT": "మౌంటెయిన్ పగటి వెలుతురు సమయం", "MESZ": "సెంట్రల్ యూరోపియన్ వేసవి సమయం", "MEZ": "సెంట్రల్ యూరోపియన్ ప్రామాణిక సమయం", "MST": "మౌంటెయిన్ ప్రామాణిక సమయం", "MYT": "మలేషియా సమయం", "NZDT": "న్యూజిల్యాండ్ పగటి వెలుతురు సమయం", "NZST": "న్యూజిల్యాండ్ ప్రామాణిక సమయం", "OESZ": "తూర్పు యూరోపియన్ వేసవి సమయం", "OEZ": "తూర్పు యూరోపియన్ ప్రామాణిక సమయం", "PDT": "పసిఫిక్ పగటి వెలుతురు సమయం", "PST": "పసిఫిక్ ప్రామాణిక సమయం", "SAST": "దక్షిణ ఆఫ్రికా ప్రామాణిక సమయం", "SGT": "సింగపూర్ ప్రామాణిక సమయం", "SRT": "సూరినామ్ సమయం", "TMST": "తుర్క్\u200cమెనిస్తాన్ వేసవి సమయం", "TMT": "తుర్క్\u200cమెనిస్తాన్ ప్రామాణిక సమయం", "UYST": "ఉరుగ్వే వేసవి సమయం", "UYT": "ఉరుగ్వే ప్రామాణిక సమయం", "VET": "వెనిజులా సమయం", "WARST": "పశ్చిమ అర్జెంటీనా వేసవి సమయం", "WART": "పశ్చిమ అర్జెంటీనా ప్రామాణిక సమయం", "WAST": "పశ్చిమ ఆఫ్రికా వేసవి సమయం", "WAT": "పశ్చిమ ఆఫ్రికా ప్రామాణిక సమయం", "WESZ": "పశ్చిమ యూరోపియన్ వేసవి సమయం", "WEZ": "పశ్చిమ యూరోపియన్ ప్రామాణిక సమయం", "WIB": "పశ్చిమ ఇండోనేషియా సమయం", "WIT": "తూర్పు ఇండోనేషియా సమయం", "WITA": "సెంట్రల్ ఇండోనేషియా సమయం", "∅∅∅": "బ్రెజిలియా వేసవి సమయం"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c892ba68b5e8cfe388377f706293ec57,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["te_in"] = func() locales.Translator {
		return &localen{

			locale:                 "te_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "జన", "ఫిబ్ర", "మార్చి", "ఏప్రి", "మే", "జూన్", "జులై", "ఆగ", "సెప్టెం", "అక్టో", "నవం", "డిసెం"},
			monthsNarrow:           []string{"", "జ", "ఫి", "మా", "ఏ", "మే", "జూ", "జు", "ఆ", "సె", "అ", "న", "డి"},
			monthsWide:             []string{"", "జనవరి", "ఫిబ్రవరి", "మార్చి", "ఏప్రిల్", "మే", "జూన్", "జులై", "ఆగస్టు", "సెప్టెంబర్", "అక్టోబర్", "నవంబర్", "డిసెంబర్"},
			daysAbbreviated:        []string{"ఆది", "సోమ", "మంగళ", "బుధ", "గురు", "శుక్ర", "శని"},
			daysNarrow:             []string{"ఆ", "సో", "మ", "బు", "గు", "శు", "శ"},
			daysShort:              []string{"ఆది", "సోమ", "మం", "బుధ", "గురు", "శుక్ర", "శని"},
			daysWide:               []string{"ఆదివారం", "సోమవారం", "మంగళవారం", "బుధవారం", "గురువారం", "శుక్రవారం", "శనివారం"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"ఉ", "సా"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"క్రీపూ", "క్రీశ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"క్రీస్తు పూర్వం", "క్రీస్తు శకం"},
			timezones:              map[string]string{"ACDT": "ఆస్ట్రేలియా మధ్యమ పగటి వెలుతురు సమయం", "ACST": "ఆస్ట్రేలియా మధ్యమ ప్రామాణిక సమయం", "ACWDT": "ఆస్ట్రేలియా మధ్యమ పశ్చిమ పగటి వెలుతురు సమయం", "ACWST": "మధ్యమ ఆస్ట్రేలియన్ పశ్చిమ ప్రామాణిక సమయం", "ADT": "అట్లాంటిక్ పగటి వెలుతురు సమయం", "AEDT": "ఆస్ట్రేలియన్ తూర్పు పగటి వెలుతురు సమయం", "AEST": "ఆస్ట్రేలియన్ తూర్పు ప్రామాణిక సమయం", "AKDT": "అలాస్కా పగటి వెలుతురు సమయం", "AKST": "అలాస్కా ప్రామాణిక సమయం", "ARST": "ఆర్జెంటీనా వేసవి సమయం", "ART": "అర్జెంటీనా ప్రామాణిక సమయం", "AST": "అట్లాంటిక్ ప్రామాణిక సమయం", "AWDT": "ఆస్ట్రేలియన్ పశ్చిమ పగటి వెలుతురు సమయం", "AWST": "ఆస్ట్రేలియన్ పశ్చిమ ప్రామాణిక సమయం", "BOT": "బొలీవియా సమయం", "BT": "భూటాన్ సమయం", "CAT": "సెంట్రల్ ఆఫ్రికా సమయం", "CDT": "మధ్యమ పగటి వెలుతురు సమయం", "CHADT": "చాథమ్ పగటి వెలుతురు సమయం", "CHAST": "చాథమ్ ప్రామాణిక సమయం", "CLST": "చిలీ వేసవి సమయం", "CLT": "చిలీ ప్రామాణిక సమయం", "COST": "కొలంబియా వేసవి సమయం", "COT": "కొలంబియా ప్రామాణిక సమయం", "CST": "మధ్యమ ప్రామాణిక సమయం", "ChST": "చామర్రో ప్రామాణిక సమయం", "EAT": "తూర్పు ఆఫ్రికా సమయం", "ECT": "ఈక్వడార్ సమయం", "EDT": "తూర్పు పగటి వెలుతురు సమయం", "EST": "తూర్పు ప్రామాణిక సమయం", "GFT": "ఫ్రెంచ్ గయానా సమయం", "GMT": "గ్రీన్\u200cవిచ్ సగటు సమయం", "GST": "గల్ఫ్ ప్రామాణిక సమయం", "GYT": "గయానా సమయం", "HADT": "హవాయ్-అల్యూషియన్ పగటి వెలుతురు సమయం", "HAST": "హవాయ్-అల్యూషియన్ ప్రామాణిక సమయం", "HAT": "న్యూఫౌండ్\u200cల్యాండ్ పగటి వెలుతురు సమయం", "HECU": "క్యూబా పగటి వెలుతురు సమయం", "HEEG": "తూర్పు గ్రీన్\u200cల్యాండ్ వేసవి సమయం", "HENOMX": "వాయువ్య మెక్సికో పగటి వెలుతురు సమయం", "HEOG": "పశ్చిమ గ్రీన్\u200cల్యాండ్ వేసవి సమయం", "HEPM": "సెయింట్ పియర్ మరియు మిక్వెలాన్ పగటి వెలుతురు సమయం", "HEPMX": "మెక్సికన్ పసిఫిక్ పగటి వెలుతురు సమయం", "HKST": "హాంకాంగ్ వేసవి సమయం", "HKT": "హాంకాంగ్ ప్రామాణిక సమయం", "HNCU": "క్యూబా ప్రామాణిక సమయం", "HNEG": "తూర్పు గ్రీన్\u200cల్యాండ్ ప్రామాణిక సమయం", "HNNOMX": "వాయువ్య మెక్సికో ప్రామాణిక సమయం", "HNOG": "పశ్చిమ గ్రీన్\u200cల్యాండ్ ప్రామాణిక సమయం", "HNPM": "సెయింట్ పియెర్ మరియు మిక్వెలాన్ ప్రామాణిక సమయం", "HNPMX": "మెక్సికన్ పసిఫిక్ ప్రామాణిక సమయం", "HNT": "న్యూఫౌండ్\u200cల్యాండ్ ప్రామాణిక సమయం", "IST": "భారతదేశ సమయం", "JDT": "జపాన్ పగటి వెలుతురు సమయం", "JST": "జపాన్ ప్రామాణిక సమయం", "LHDT": "లార్డ్ హోవ్ పగటి సమయం", "LHST": "లార్డ్ హోవ్ ప్రామాణిక సమయం", "MDT": "మౌంటెయిన్ పగటి వెలుతురు సమయం", "MESZ": "సెంట్రల్ యూరోపియన్ వేసవి సమయం", "MEZ": "సెంట్రల్ యూరోపియన్ ప్రామాణిక సమయం", "MST": "మౌంటెయిన్ ప్రామాణిక సమయం", "MYT": "మలేషియా సమయం", "NZDT": "న్యూజిల్యాండ్ పగటి వెలుతురు సమయం", "NZST": "న్యూజిల్యాండ్ ప్రామాణిక సమయం", "OESZ": "తూర్పు యూరోపియన్ వేసవి సమయం", "OEZ": "తూర్పు యూరోపియన్ ప్రామాణిక సమయం", "PDT": "పసిఫిక్ పగటి వెలుతురు సమయం", "PST": "పసిఫిక్ ప్రామాణిక సమయం", "SAST": "దక్షిణ ఆఫ్రికా ప్రామాణిక సమయం", "SGT": "సింగపూర్ ప్రామాణిక సమయం", "SRT": "సూరినామ్ సమయం", "TMST": "తుర్క్\u200cమెనిస్తాన్ వేసవి సమయం", "TMT": "తుర్క్\u200cమెనిస్తాన్ ప్రామాణిక సమయం", "UYST": "ఉరుగ్వే వేసవి సమయం", "UYT": "ఉరుగ్వే ప్రామాణిక సమయం", "VET": "వెనిజులా సమయం", "WARST": "పశ్చిమ అర్జెంటీనా వేసవి సమయం", "WART": "పశ్చిమ అర్జెంటీనా ప్రామాణిక సమయం", "WAST": "పశ్చిమ ఆఫ్రికా వేసవి సమయం", "WAT": "పశ్చిమ ఆఫ్రికా ప్రామాణిక సమయం", "WESZ": "పశ్చిమ యూరోపియన్ వేసవి సమయం", "WEZ": "పశ్చిమ యూరోపియన్ ప్రామాణిక సమయం", "WIB": "పశ్చిమ ఇండోనేషియా సమయం", "WIT": "తూర్పు ఇండోనేషియా సమయం", "WITA": "సెంట్రల్ ఇండోనేషియా సమయం", "∅∅∅": "పెరూ వేసవి సమయం"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c892ba68b5e8cfe388377f706293ec57,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_2ad99bf760d4b47854009ea5a8878148,
			fnFmtNumber:           fnFmtNumber_d7ae7636100e1f19a61ae0ac52b51e8c,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["teo"] = func() locales.Translator {
		return &localen{

			locale:                 "teo",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Rar", "Muk", "Kwa", "Dun", "Mar", "Mod", "Jol", "Ped", "Sok", "Tib", "Lab", "Poo"},
			monthsNarrow:           []string{"", "R", "M", "K", "D", "M", "M", "J", "P", "S", "T", "L", "P"},
			monthsWide:             []string{"", "Orara", "Omuk", "Okwamg’", "Odung’el", "Omaruk", "Omodok’king’ol", "Ojola", "Opedel", "Osokosokoma", "Otibar", "Olabor", "Opoo"},
			daysAbbreviated:        []string{"Jum", "Bar", "Aar", "Uni", "Ung", "Kan", "Sab"},
			daysNarrow:             []string{"J", "B", "A", "U", "U", "K", "S"},
			daysWide:               []string{"Nakaejuma", "Nakaebarasa", "Nakaare", "Nakauni", "Nakaung’on", "Nakakany", "Nakasabiti"},
			periodsAbbreviated:     []string{"Taparachu", "Ebongi"},
			periodsWide:            []string{"Taparachu", "Ebongi"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Christo", "Baada ya Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["teo_ke"] = func() locales.Translator {
		return &localen{

			locale:                 "teo_KE",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "Ksh", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Rar", "Muk", "Kwa", "Dun", "Mar", "Mod", "Jol", "Ped", "Sok", "Tib", "Lab", "Poo"},
			monthsNarrow:           []string{"", "R", "M", "K", "D", "M", "M", "J", "P", "S", "T", "L", "P"},
			monthsWide:             []string{"", "Orara", "Omuk", "Okwamg’", "Odung’el", "Omaruk", "Omodok’king’ol", "Ojola", "Opedel", "Osokosokoma", "Otibar", "Olabor", "Opoo"},
			daysAbbreviated:        []string{"Jum", "Bar", "Aar", "Uni", "Ung", "Kan", "Sab"},
			daysNarrow:             []string{"J", "B", "A", "U", "U", "K", "S"},
			daysWide:               []string{"Nakaejuma", "Nakaebarasa", "Nakaare", "Nakauni", "Nakaung’on", "Nakakany", "Nakasabiti"},
			periodsAbbreviated:     []string{"Taparachu", "Ebongi"},
			periodsWide:            []string{"Taparachu", "Ebongi"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Christo", "Baada ya Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["teo_ug"] = func() locales.Translator {
		return &localen{

			locale:                 "teo_UG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Rar", "Muk", "Kwa", "Dun", "Mar", "Mod", "Jol", "Ped", "Sok", "Tib", "Lab", "Poo"},
			monthsNarrow:           []string{"", "R", "M", "K", "D", "M", "M", "J", "P", "S", "T", "L", "P"},
			monthsWide:             []string{"", "Orara", "Omuk", "Okwamg’", "Odung’el", "Omaruk", "Omodok’king’ol", "Ojola", "Opedel", "Osokosokoma", "Otibar", "Olabor", "Opoo"},
			daysAbbreviated:        []string{"Jum", "Bar", "Aar", "Uni", "Ung", "Kan", "Sab"},
			daysNarrow:             []string{"J", "B", "A", "U", "U", "K", "S"},
			daysWide:               []string{"Nakaejuma", "Nakaebarasa", "Nakaare", "Nakauni", "Nakaung’on", "Nakakany", "Nakasabiti"},
			periodsAbbreviated:     []string{"Taparachu", "Ebongi"},
			periodsWide:            []string{"Taparachu", "Ebongi"},
			erasAbbreviated:        []string{"KK", "BK"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kabla ya Christo", "Baada ya Christo"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_60b3ba42c7301cf18f0e7795b3eeaa75,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tg"] = func() locales.Translator {
		return &localen{

			locale:                 "tg",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "сом.", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "Январ", "Феврал", "Март", "Апрел", "Май", "Июн", "Июл", "Август", "Сентябр", "Октябр", "Ноябр", "Декабр"},
			daysAbbreviated:        []string{"Яшб", "Дшб", "Сшб", "Чшб", "Пшб", "Ҷмъ", "Шнб"},
			daysNarrow:             []string{"Я", "Д", "С", "Ч", "П", "Ҷ", "Ш"},
			daysShort:              []string{"Яшб", "Дшб", "Сшб", "Чшб", "Пшб", "Ҷмъ", "Шнб"},
			daysWide:               []string{"Якшанбе", "Душанбе", "Сешанбе", "Чоршанбе", "Панҷшанбе", "Ҷумъа", "Шанбе"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"ПеМ", "ПаМ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Пеш аз милод", "ПаМ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Вақти рӯзонаи атлантикӣ", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Вақти стандартии атлантикӣ", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Вақти рӯзонаи марказӣ", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Вақти стандартии марказӣ", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Вақти рӯзонаи шарқӣ", "EST": "Вақти стандартии шарқӣ", "GFT": "GFT", "GMT": "Ба вақти Гринвич", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Вақти рӯзонаи кӯҳӣ", "MESZ": "Вақти тобистонаи аврупоии марказӣ", "MEZ": "Вақти стандартии аврупоии марказӣ", "MST": "Вақти стандартии кӯҳӣ", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Вақти тобистонаи аврупоии шарқӣ", "OEZ": "Вақти стандартии аврупоии шарқӣ", "PDT": "Вақти рӯзонаи Уқёнуси Ором", "PST": "Вақти стандартии Уқёнуси Ором", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Вақти тобистонаи аврупоии ғарбӣ", "WEZ": "Вақти стандартии аврупоии ғарбӣ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tg_tj"] = func() locales.Translator {
		return &localen{

			locale:                 "tg_TJ",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "Январ", "Феврал", "Март", "Апрел", "Май", "Июн", "Июл", "Август", "Сентябр", "Октябр", "Ноябр", "Декабр"},
			daysAbbreviated:        []string{"Яшб", "Дшб", "Сшб", "Чшб", "Пшб", "Ҷмъ", "Шнб"},
			daysNarrow:             []string{"Я", "Д", "С", "Ч", "П", "Ҷ", "Ш"},
			daysShort:              []string{"Яшб", "Дшб", "Сшб", "Чшб", "Пшб", "Ҷмъ", "Шнб"},
			daysWide:               []string{"Якшанбе", "Душанбе", "Сешанбе", "Чоршанбе", "Панҷшанбе", "Ҷумъа", "Шанбе"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"ПеМ", "ПаМ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Пеш аз милод", "ПаМ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Вақти рӯзонаи атлантикӣ", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Вақти стандартии атлантикӣ", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Вақти рӯзонаи марказӣ", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Вақти стандартии марказӣ", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Вақти рӯзонаи шарқӣ", "EST": "Вақти стандартии шарқӣ", "GFT": "GFT", "GMT": "Ба вақти Гринвич", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Вақти рӯзонаи кӯҳӣ", "MESZ": "Вақти тобистонаи аврупоии марказӣ", "MEZ": "Вақти стандартии аврупоии марказӣ", "MST": "Вақти стандартии кӯҳӣ", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Вақти тобистонаи аврупоии шарқӣ", "OEZ": "Вақти стандартии аврупоии шарқӣ", "PDT": "Вақти рӯзонаи Уқёнуси Ором", "PST": "Вақти стандартии Уқёнуси Ором", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Вақти тобистонаи аврупоии ғарбӣ", "WEZ": "Вақти стандартии аврупоии ғарбӣ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_588a96656cc1da10a40d88dd26ead028,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_358c9403a9e4cad0b8974bed45e6527c,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["th"] = func() locales.Translator {
		return &localen{

			locale:                 "th",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ม.ค.", "ก.พ.", "มี.ค.", "เม.ย.", "พ.ค.", "มิ.ย.", "ก.ค.", "ส.ค.", "ก.ย.", "ต.ค.", "พ.ย.", "ธ.ค."},
			monthsNarrow:           []string{"", "ม.ค.", "ก.พ.", "มี.ค.", "เม.ย.", "พ.ค.", "มิ.ย.", "ก.ค.", "ส.ค.", "ก.ย.", "ต.ค.", "พ.ย.", "ธ.ค."},
			monthsWide:             []string{"", "มกราคม", "กุมภาพันธ์", "มีนาคม", "เมษายน", "พฤษภาคม", "มิถุนายน", "กรกฎาคม", "สิงหาคม", "กันยายน", "ตุลาคม", "พฤศจิกายน", "ธันวาคม"},
			daysAbbreviated:        []string{"อา.", "จ.", "อ.", "พ.", "พฤ.", "ศ.", "ส."},
			daysNarrow:             []string{"อา", "จ", "อ", "พ", "พฤ", "ศ", "ส"},
			daysShort:              []string{"อา.", "จ.", "อ.", "พ.", "พฤ.", "ศ.", "ส."},
			daysWide:               []string{"วันอาทิตย์", "วันจันทร์", "วันอังคาร", "วันพุธ", "วันพฤหัสบดี", "วันศุกร์", "วันเสาร์"},
			periodsAbbreviated:     []string{"ก่อนเที่ยง", "หลังเที่ยง"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"ก่อนเที่ยง", "หลังเที่ยง"},
			erasAbbreviated:        []string{"ก่อน ค.ศ.", "ค.ศ."},
			erasNarrow:             []string{"ก่อน ค.ศ.", "ค.ศ."},
			erasWide:               []string{"ปีก่อนคริสตกาล", "คริสต์ศักราช"},
			timezones:              map[string]string{"ACDT": "เวลาออมแสงทางตอนกลางของออสเตรเลีย", "ACST": "เวลามาตรฐานทางตอนกลางของออสเตรเลีย", "ACWDT": "เวลาออมแสงทางตะวันตกตอนกลางของออสเตรเลีย", "ACWST": "เวลามาตรฐานทางตะวันตกตอนกลางของออสเตรเลีย", "ADT": "เวลาออมแสงของแอตแลนติก", "AEDT": "เวลาออมแสงทางตะวันออกของออสเตรเลีย", "AEST": "เวลามาตรฐานทางตะวันออกของออสเตรเลีย", "AKDT": "เวลาออมแสงของอะแลสกา", "AKST": "เวลามาตรฐานอะแลสกา", "ARST": "เวลาฤดูร้อนอาร์เจนตินา", "ART": "เวลามาตรฐานอาร์เจนตินา", "AST": "เวลามาตรฐานแอตแลนติก", "AWDT": "เวลาออมแสงทางตะวันตกของออสเตรเลีย", "AWST": "เวลามาตรฐานทางตะวันตกของออสเตรเลีย", "BOT": "เวลาโบลิเวีย", "BT": "เวลาภูฏาน", "CAT": "เวลาแอฟริกากลาง", "CDT": "เวลาออมแสงตอนกลางในอเมริกาเหนือ", "CHADT": "เวลาออมแสงแชทัม", "CHAST": "เวลามาตรฐานแชทัม", "CLST": "เวลาฤดูร้อนชิลี", "CLT": "เวลามาตรฐานชิลี", "COST": "เวลาฤดูร้อนโคลอมเบีย", "COT": "เวลามาตรฐานโคลอมเบีย", "CST": "เวลามาตรฐานตอนกลางในอเมริกาเหนือ", "ChST": "เวลาชามอร์โร", "EAT": "เวลาแอฟริกาตะวันออก", "ECT": "เวลาเอกวาดอร์", "EDT": "เวลาออมแสงทางตะวันออกในอเมริกาเหนือ", "EST": "เวลามาตรฐานทางตะวันออกในอเมริกาเหนือ", "GFT": "เวลาเฟรนช์เกียนา", "GMT": "เวลามาตรฐานกรีนิช", "GST": "เวลากัลฟ์", "GYT": "เวลากายอานา", "HADT": "เวลาออมแสงฮาวาย-อะลูเชียน", "HAST": "เวลามาตรฐานฮาวาย-อะลูเชียน", "HAT": "เวลาออมแสงนิวฟันด์แลนด์", "HECU": "เวลาออมแสงของคิวบา", "HEEG": "เวลาฤดูร้อนกรีนแลนด์ตะวันออก", "HENOMX": "เวลาออมแสงเม็กซิโกตะวันตกเฉียงเหนือ", "HEOG": "เวลาฤดูร้อนกรีนแลนด์ตะวันตก", "HEPM": "เวลาออมแสงของแซงปีแยร์และมีเกอลง", "HEPMX": "เวลาออมแสงแปซิฟิกเม็กซิโก", "HKST": "เวลาฤดูร้อนฮ่องกง", "HKT": "เวลามาตรฐานฮ่องกง", "HNCU": "เวลามาตรฐานคิวบา", "HNEG": "เวลามาตรฐานกรีนแลนด์ตะวันออก", "HNNOMX": "เวลามาตรฐานเม็กซิโกตะวันตกเฉียงเหนือ", "HNOG": "เวลามาตรฐานกรีนแลนด์ตะวันตก", "HNPM": "เวลามาตรฐานแซงปีแยร์และมีเกอลง", "HNPMX": "เวลามาตรฐานแปซิฟิกเม็กซิโก", "HNT": "เวลามาตรฐานนิวฟันด์แลนด์", "IST": "เวลาอินเดีย", "JDT": "เวลาออมแสงญี่ปุ่น", "JST": "เวลามาตรฐานญี่ปุ่น", "LHDT": "เวลาออมแสงลอร์ดโฮว์", "LHST": "เวลามาตรฐานลอร์ดโฮว์", "MDT": "เวลาออมแสงแถบภูเขาในอเมริกาเหนือ", "MESZ": "เวลาฤดูร้อนยุโรปกลาง", "MEZ": "เวลามาตรฐานยุโรปกลาง", "MST": "เวลามาตรฐานแถบภูเขาในอเมริกาเหนือ", "MYT": "เวลามาเลเซีย", "NZDT": "เวลาออมแสงนิวซีแลนด์", "NZST": "เวลามาตรฐานนิวซีแลนด์", "OESZ": "เวลาฤดูร้อนยุโรปตะวันออก", "OEZ": "เวลามาตรฐานยุโรปตะวันออก", "PDT": "เวลาออมแสงแปซิฟิกในอเมริกาเหนือ", "PST": "เวลามาตรฐานแปซิฟิกในอเมริกาเหนือ", "SAST": "เวลาแอฟริกาใต้", "SGT": "เวลาสิงคโปร์", "SRT": "เวลาซูรินาเม", "TMST": "เวลาฤดูร้อนเติร์กเมนิสถาน", "TMT": "เวลามาตรฐานเติร์กเมนิสถาน", "UYST": "เวลาฤดูร้อนอุรุกวัย", "UYT": "เวลามาตรฐานอุรุกวัย", "VET": "เวลาเวเนซุเอลา", "WARST": "เวลาฤดูร้อนทางตะวันตกของอาร์เจนตินา", "WART": "เวลามาตรฐานทางตะวันตกของอาร์เจนตินา", "WAST": "เวลาฤดูร้อนแอฟริกาตะวันตก", "WAT": "เวลามาตรฐานแอฟริกาตะวันตก", "WESZ": "เวลาฤดูร้อนยุโรปตะวันตก", "WEZ": "เวลามาตรฐานยุโรปตะวันตก", "WIB": "เวลาอินโดนีเซียฝั่งตะวันตก", "WIT": "เวลาอินโดนีเซียฝั่งตะวันออก", "WITA": "เวลาอินโดนีเซียตอนกลาง", "∅∅∅": "เวลาฤดูร้อนเปรู"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_2a2625729a79b75f22677ae0545a3802,
			fnFmtDateLong:         fnFmtDateLong_4ba622cfb6604f92efa67396648b01e1,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_54a1d4bd08c970be76504a2aa980af91,
			fnFmtTimeLong:         fnFmtTimeLong_9d4d0f53074023f50ec443e7d6cc145e,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["th_th"] = func() locales.Translator {
		return &localen{

			locale:                 "th_TH",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ม.ค.", "ก.พ.", "มี.ค.", "เม.ย.", "พ.ค.", "มิ.ย.", "ก.ค.", "ส.ค.", "ก.ย.", "ต.ค.", "พ.ย.", "ธ.ค."},
			monthsNarrow:           []string{"", "ม.ค.", "ก.พ.", "มี.ค.", "เม.ย.", "พ.ค.", "มิ.ย.", "ก.ค.", "ส.ค.", "ก.ย.", "ต.ค.", "พ.ย.", "ธ.ค."},
			monthsWide:             []string{"", "มกราคม", "กุมภาพันธ์", "มีนาคม", "เมษายน", "พฤษภาคม", "มิถุนายน", "กรกฎาคม", "สิงหาคม", "กันยายน", "ตุลาคม", "พฤศจิกายน", "ธันวาคม"},
			daysAbbreviated:        []string{"อา.", "จ.", "อ.", "พ.", "พฤ.", "ศ.", "ส."},
			daysNarrow:             []string{"อา", "จ", "อ", "พ", "พฤ", "ศ", "ส"},
			daysShort:              []string{"อา.", "จ.", "อ.", "พ.", "พฤ.", "ศ.", "ส."},
			daysWide:               []string{"วันอาทิตย์", "วันจันทร์", "วันอังคาร", "วันพุธ", "วันพฤหัสบดี", "วันศุกร์", "วันเสาร์"},
			periodsAbbreviated:     []string{"ก่อนเที่ยง", "หลังเที่ยง"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"ก่อนเที่ยง", "หลังเที่ยง"},
			erasAbbreviated:        []string{"ก่อน ค.ศ.", "ค.ศ."},
			erasNarrow:             []string{"ก่อน ค.ศ.", "ค.ศ."},
			erasWide:               []string{"ปีก่อนคริสตกาล", "คริสต์ศักราช"},
			timezones:              map[string]string{"ACDT": "เวลาออมแสงทางตอนกลางของออสเตรเลีย", "ACST": "เวลามาตรฐานทางตอนกลางของออสเตรเลีย", "ACWDT": "เวลาออมแสงทางตะวันตกตอนกลางของออสเตรเลีย", "ACWST": "เวลามาตรฐานทางตะวันตกตอนกลางของออสเตรเลีย", "ADT": "เวลาออมแสงของแอตแลนติก", "AEDT": "เวลาออมแสงทางตะวันออกของออสเตรเลีย", "AEST": "เวลามาตรฐานทางตะวันออกของออสเตรเลีย", "AKDT": "เวลาออมแสงของอะแลสกา", "AKST": "เวลามาตรฐานอะแลสกา", "ARST": "เวลาฤดูร้อนอาร์เจนตินา", "ART": "เวลามาตรฐานอาร์เจนตินา", "AST": "เวลามาตรฐานแอตแลนติก", "AWDT": "เวลาออมแสงทางตะวันตกของออสเตรเลีย", "AWST": "เวลามาตรฐานทางตะวันตกของออสเตรเลีย", "BOT": "เวลาโบลิเวีย", "BT": "เวลาภูฏาน", "CAT": "เวลาแอฟริกากลาง", "CDT": "เวลาออมแสงตอนกลางในอเมริกาเหนือ", "CHADT": "เวลาออมแสงแชทัม", "CHAST": "เวลามาตรฐานแชทัม", "CLST": "เวลาฤดูร้อนชิลี", "CLT": "เวลามาตรฐานชิลี", "COST": "เวลาฤดูร้อนโคลอมเบีย", "COT": "เวลามาตรฐานโคลอมเบีย", "CST": "เวลามาตรฐานตอนกลางในอเมริกาเหนือ", "ChST": "เวลาชามอร์โร", "EAT": "เวลาแอฟริกาตะวันออก", "ECT": "เวลาเอกวาดอร์", "EDT": "เวลาออมแสงทางตะวันออกในอเมริกาเหนือ", "EST": "เวลามาตรฐานทางตะวันออกในอเมริกาเหนือ", "GFT": "เวลาเฟรนช์เกียนา", "GMT": "เวลามาตรฐานกรีนิช", "GST": "เวลากัลฟ์", "GYT": "เวลากายอานา", "HADT": "เวลาออมแสงฮาวาย-อะลูเชียน", "HAST": "เวลามาตรฐานฮาวาย-อะลูเชียน", "HAT": "เวลาออมแสงนิวฟันด์แลนด์", "HECU": "เวลาออมแสงของคิวบา", "HEEG": "เวลาฤดูร้อนกรีนแลนด์ตะวันออก", "HENOMX": "เวลาออมแสงเม็กซิโกตะวันตกเฉียงเหนือ", "HEOG": "เวลาฤดูร้อนกรีนแลนด์ตะวันตก", "HEPM": "เวลาออมแสงของแซงปีแยร์และมีเกอลง", "HEPMX": "เวลาออมแสงแปซิฟิกเม็กซิโก", "HKST": "เวลาฤดูร้อนฮ่องกง", "HKT": "เวลามาตรฐานฮ่องกง", "HNCU": "เวลามาตรฐานคิวบา", "HNEG": "เวลามาตรฐานกรีนแลนด์ตะวันออก", "HNNOMX": "เวลามาตรฐานเม็กซิโกตะวันตกเฉียงเหนือ", "HNOG": "เวลามาตรฐานกรีนแลนด์ตะวันตก", "HNPM": "เวลามาตรฐานแซงปีแยร์และมีเกอลง", "HNPMX": "เวลามาตรฐานแปซิฟิกเม็กซิโก", "HNT": "เวลามาตรฐานนิวฟันด์แลนด์", "IST": "เวลาอินเดีย", "JDT": "เวลาออมแสงญี่ปุ่น", "JST": "เวลามาตรฐานญี่ปุ่น", "LHDT": "เวลาออมแสงลอร์ดโฮว์", "LHST": "เวลามาตรฐานลอร์ดโฮว์", "MDT": "เวลาออมแสงแถบภูเขาในอเมริกาเหนือ", "MESZ": "เวลาฤดูร้อนยุโรปกลาง", "MEZ": "เวลามาตรฐานยุโรปกลาง", "MST": "เวลามาตรฐานแถบภูเขาในอเมริกาเหนือ", "MYT": "เวลามาเลเซีย", "NZDT": "เวลาออมแสงนิวซีแลนด์", "NZST": "เวลามาตรฐานนิวซีแลนด์", "OESZ": "เวลาฤดูร้อนยุโรปตะวันออก", "OEZ": "เวลามาตรฐานยุโรปตะวันออก", "PDT": "เวลาออมแสงแปซิฟิกในอเมริกาเหนือ", "PST": "เวลามาตรฐานแปซิฟิกในอเมริกาเหนือ", "SAST": "เวลาแอฟริกาใต้", "SGT": "เวลาสิงคโปร์", "SRT": "เวลาซูรินาเม", "TMST": "เวลาฤดูร้อนเติร์กเมนิสถาน", "TMT": "เวลามาตรฐานเติร์กเมนิสถาน", "UYST": "เวลาฤดูร้อนอุรุกวัย", "UYT": "เวลามาตรฐานอุรุกวัย", "VET": "เวลาเวเนซุเอลา", "WARST": "เวลาฤดูร้อนทางตะวันตกของอาร์เจนตินา", "WART": "เวลามาตรฐานทางตะวันตกของอาร์เจนตินา", "WAST": "เวลาฤดูร้อนแอฟริกาตะวันตก", "WAT": "เวลามาตรฐานแอฟริกาตะวันตก", "WESZ": "เวลาฤดูร้อนยุโรปตะวันตก", "WEZ": "เวลามาตรฐานยุโรปตะวันตก", "WIB": "เวลาอินโดนีเซียฝั่งตะวันตก", "WIT": "เวลาอินโดนีเซียฝั่งตะวันออก", "WITA": "เวลาอินโดนีเซียตอนกลาง", "∅∅∅": "เวลาฤดูร้อนบราซิเลีย"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_2a2625729a79b75f22677ae0545a3802,
			fnFmtDateLong:         fnFmtDateLong_4ba622cfb6604f92efa67396648b01e1,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_54a1d4bd08c970be76504a2aa980af91,
			fnFmtTimeLong:         fnFmtTimeLong_9d4d0f53074023f50ec443e7d6cc145e,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ti"] = func() locales.Translator {
		return &localen{

			locale:             "ti",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "Br", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ጥሪ", "ለካ", "መጋ", "ሚያ", "ግን", "ሰነ", "ሓም", "ነሓ", "መስ", "ጥቅ", "ሕዳ", "ታሕ"},
			monthsNarrow:       []string{"", "ጥ", "ለ", "መ", "ሚ", "ግ", "ሰ", "ሓ", "ነ", "መ", "ጥ", "ሕ", "ታ"},
			monthsWide:         []string{"", "ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"},
			daysAbbreviated:    []string{"ሰን", "ሰኑ", "ሰሉ", "ረቡ", "ሓሙ", "ዓር", "ቀዳ"},
			daysNarrow:         []string{"ሰ", "ሰ", "ሰ", "ረ", "ሓ", "ዓ", "ቀ"},
			daysShort:          []string{"ሰን", "ሰኑ", "ሰሉ", "ረቡ", "ሓሙ", "ዓር", "ቀዳ"},
			daysWide:           []string{"ሰንበት", "ሰኑይ", "ሠሉስ", "ረቡዕ", "ኃሙስ", "ዓርቢ", "ቀዳም"},
			periodsAbbreviated: []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			periodsNarrow:      []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			periodsWide:        []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			erasAbbreviated:    []string{"ዓ/ዓ", "ዓ/ም"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ዓ/ዓ", "ዓመተ ምህረት"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_b145a441f04d612c2b6631b673b83d20,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ti_er"] = func() locales.Translator {
		return &localen{

			locale:             "ti_ER",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "Nfk", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ጥሪ", "ለካ", "መጋ", "ሚያ", "ግን", "ሰነ", "ሓም", "ነሓ", "መስ", "ጥቅ", "ሕዳ", "ታሕ"},
			monthsNarrow:       []string{"", "ጥ", "ለ", "መ", "ሚ", "ግ", "ሰ", "ሓ", "ነ", "መ", "ጥ", "ሕ", "ታ"},
			monthsWide:         []string{"", "ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"},
			daysAbbreviated:    []string{"ሰን", "ሰኑ", "ሰሉ", "ረቡ", "ሓሙ", "ዓር", "ቀዳ"},
			daysNarrow:         []string{"ሰ"},
			daysShort:          []string{"ሰን", "ሰኑ", "ሰሉ", "ረቡ", "ሓሙ", "ዓር", "ቀዳ"},
			daysWide:           []string{"ሰንበት", "ሰኑይ", "ሠሉስ", "ረቡዕ", "ኃሙስ", "ዓርቢ", "ቀዳም"},
			periodsAbbreviated: []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			periodsNarrow:      []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			periodsWide:        []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			erasAbbreviated:    []string{"", ""},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"", ""},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_b145a441f04d612c2b6631b673b83d20,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ti_et"] = func() locales.Translator {
		return &localen{

			locale:             "ti_ET",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "ጥሪ", "ለካ", "መጋ", "ሚያ", "ግን", "ሰነ", "ሓም", "ነሓ", "መስ", "ጥቅ", "ሕዳ", "ታሕ"},
			monthsNarrow:       []string{"", "ጥ", "ለ", "መ", "ሚ", "ግ", "ሰ", "ሓ", "ነ", "መ", "ጥ", "ሕ", "ታ"},
			monthsWide:         []string{"", "ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"},
			daysAbbreviated:    []string{"ሰን", "ሰኑ", "ሰሉ", "ረቡ", "ሓሙ", "ዓር", "ቀዳ"},
			daysNarrow:         []string{"ሰ", "ሰ", "ሰ", "ረ", "ሓ", "ዓ", "ቀ"},
			daysShort:          []string{"ሰን", "ሰኑ", "ሰሉ", "ረቡ", "ሓሙ", "ዓር", "ቀዳ"},
			daysWide:           []string{"ሰንበት", "ሰኑይ", "ሠሉስ", "ረቡዕ", "ኃሙስ", "ዓርቢ", "ቀዳም"},
			periodsAbbreviated: []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			periodsNarrow:      []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			periodsWide:        []string{"ንጉሆ ሰዓተ", "ድሕር ሰዓት"},
			erasAbbreviated:    []string{"ዓ/ዓ", "ዓ/ም"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ዓ/ዓ", "ዓመተ ምህረት"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_1a767992f2534d16887981ec7c7c1cfb,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_b145a441f04d612c2b6631b673b83d20,
			fnFmtDateLong:         fnFmtDateLong_85907dc745b1a23de90d4a44c3222404,
			fnFmtDateMedium:       fnFmtDateMedium_4691817e98b7c85cafc185bbd5485130,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tk"] = func() locales.Translator {
		return &localen{

			locale:                 "tk",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{4, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ýan", "few", "mart", "apr", "maý", "iýun", "iýul", "awg", "sen", "okt", "noý", "dek"},
			monthsNarrow:           []string{"", "Ý", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "ýanwar", "fewral", "mart", "aprel", "maý", "iýun", "iýul", "awgust", "sentýabr", "oktýabr", "noýabr", "dekabr"},
			daysAbbreviated:        []string{"ýek", "duş", "siş", "çar", "pen", "ann", "şen"},
			daysNarrow:             []string{"Ý", "D", "S", "Ç", "P", "A", "Ş"},
			daysShort:              []string{"ýb", "db", "sb", "çb", "pb", "an", "şb"},
			daysWide:               []string{"ýekşenbe", "duşenbe", "sişenbe", "çarşenbe", "penşenbe", "anna", "şenbe"},
			periodsAbbreviated:     []string{"go.öň", "go.soň"},
			periodsNarrow:          []string{"öň", "soň"},
			periodsWide:            []string{"günortadan öň", "günortadan soň"},
			erasAbbreviated:        []string{"B.e.öň", "B.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Isadan öň", "Isadan soň"},
			timezones:              map[string]string{"ACDT": "Merkezi Awstraliýa tomusky wagty", "ACST": "Merkezi Awstraliýa standart wagty", "ACWDT": "Merkezi Awstraliýa günbatar tomusky wagty", "ACWST": "Merkezi Awstraliýa günbatar standart wagty", "ADT": "Atlantik tomusky wagty", "AEDT": "Gündogar Awstraliýa tomusky wagty", "AEST": "Gündogar Awstraliýa standart wagty", "AKDT": "Alýaska tomusky wagty", "AKST": "Alýaska standart wagty", "ARST": "Argentina tomusky wagty", "ART": "Argentina standart wagty", "AST": "Atlantik standart wagty", "AWDT": "Günbatar Awstraliýa tomusky wagty", "AWST": "Günbatar Awstraliýa standart wagty", "BOT": "Boliwiýa wagty", "BT": "Butan wagty", "CAT": "Merkezi Afrika wagty", "CDT": "Merkezi Amerika tomusky wagty", "CHADT": "Çatem tomusky wagty", "CHAST": "Çatem standart wagty", "CLST": "Çili tomusky wagty", "CLT": "Çili standart wagty", "COST": "Kolumbiýa tomusky wagty", "COT": "Kolumbiýa standart wagty", "CST": "Merkezi Amerika standart wagty", "ChST": "Çamorro wagty", "EAT": "Gündogar Afrika wagty", "ECT": "Ekwador wagty", "EDT": "Demirgazyk Amerika gündogar tomusky wagty", "EST": "Demirgazyk Amerika gündogar standart wagty", "GFT": "Fransuz Gwianasy wagty", "GMT": "Grinwiç boýunça orta wagt", "GST": "Pars aýlagy standart wagty", "GYT": "Gaýana wagty", "HADT": "Gawaý-Aleut tomusky wagty", "HAST": "Gawaý-Aleut standart wagty", "HAT": "Nýufaundlend tomusky wagty", "HECU": "Kuba tomusky wagty", "HEEG": "Gündogar Grenlandiýa tomusky wagty", "HENOMX": "Demirgazyk-günbatar Meksika tomusky wagty", "HEOG": "Günbatar Grenlandiýa tomusky wagty", "HEPM": "Sen-Pýer we Mikelon tomusky wagty", "HEPMX": "Meksikan Ýuwaş umman tomusky wagty", "HKST": "Gonkong tomusky wagty", "HKT": "Gonkong standart wagty", "HNCU": "Kuba standart wagty", "HNEG": "Gündogar Grenlandiýa standart wagty", "HNNOMX": "Demirgazyk-günbatar Meksika standart wagty", "HNOG": "Günbatar Grenlandiýa standart wagty", "HNPM": "Sen-Pýer we Mikelon standart wagty", "HNPMX": "Meksikan Ýuwaş umman standart wagty", "HNT": "Nýufaundlend standart wagty", "IST": "Hindistan standart wagty", "JDT": "Ýaponiýa tomusky wagty", "JST": "Ýaponiýa standart wagty", "LHDT": "Lord-Hau tomusky wagty", "LHST": "Lord-Hau standart wagty", "MDT": "Demirgazyk Amerika dag tomusky wagty", "MESZ": "Merkezi Ýewropa tomusky wagty", "MEZ": "Merkezi Ýewropa standart wagty", "MST": "Demirgazyk Amerika dag standart wagty", "MYT": "Malaýziýa wagty", "NZDT": "Täze Zelandiýa tomusky wagty", "NZST": "Täze Zelandiýa standart wagty", "OESZ": "Gündogar Ýewropa tomusky wagty", "OEZ": "Gündogar Ýewropa standart wagty", "PDT": "Demirgazyk Amerika Ýuwaş umman tomusky wagty", "PST": "Demirgazyk Amerika Ýuwaş umman standart wagty", "SAST": "Günorta Afrika standart wagty", "SGT": "Singapur wagty", "SRT": "Surinam wagty", "TMST": "Türkmenistan tomusky wagty", "TMT": "Türkmenistan standart wagty", "UYST": "Urugwaý tomusky wagty", "UYT": "Urugwaý standart wagty", "VET": "Wenesuela wagty", "WARST": "Günbatar Argentina tomusky wagty", "WART": "Günbatar Argentina standart wagty", "WAST": "Günbatar Afrika tomusky wagty", "WAT": "Günbatar Afrika standart wagty", "WESZ": "Günbatar Ýewropa tomusky wagty", "WEZ": "Günbatar Ýewropa standart wagty", "WIB": "Günbatar Indoneziýa wagty", "WIT": "Gündogar Indoneziýa wagty", "WITA": "Merkezi Indoneziýa wagty", "∅∅∅": "Azor adalary tomusky wagty"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_8c251b28aa9de80970be20b481b38b9b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_96a5502b4d80e3468da0c2115685c65f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tk_tm"] = func() locales.Translator {
		return &localen{

			locale:                 "tk_TM",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{4, 6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "ýan", "few", "mart", "apr", "maý", "iýun", "iýul", "awg", "sen", "okt", "noý", "dek"},
			monthsNarrow:           []string{"", "Ý", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "ýanwar", "fewral", "mart", "aprel", "maý", "iýun", "iýul", "awgust", "sentýabr", "oktýabr", "noýabr", "dekabr"},
			daysAbbreviated:        []string{"ýek", "duş", "siş", "çar", "pen", "ann", "şen"},
			daysNarrow:             []string{"Ý", "D", "S", "Ç", "P", "A", "Ş"},
			daysShort:              []string{"ýb", "db", "sb", "çb", "pb", "an", "şb"},
			daysWide:               []string{"ýekşenbe", "duşenbe", "sişenbe", "çarşenbe", "penşenbe", "anna", "şenbe"},
			periodsAbbreviated:     []string{"go.öň", "go.soň"},
			periodsNarrow:          []string{"öň", "soň"},
			periodsWide:            []string{"günortadan öň", "günortadan soň"},
			erasAbbreviated:        []string{"B.e.öň", "B.e."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Isadan öň", "Isadan soň"},
			timezones:              map[string]string{"ACDT": "Merkezi Awstraliýa tomusky wagty", "ACST": "Merkezi Awstraliýa standart wagty", "ACWDT": "Merkezi Awstraliýa günbatar tomusky wagty", "ACWST": "Merkezi Awstraliýa günbatar standart wagty", "ADT": "Atlantik tomusky wagty", "AEDT": "Gündogar Awstraliýa tomusky wagty", "AEST": "Gündogar Awstraliýa standart wagty", "AKDT": "Alýaska tomusky wagty", "AKST": "Alýaska standart wagty", "ARST": "Argentina tomusky wagty", "ART": "Argentina standart wagty", "AST": "Atlantik standart wagty", "AWDT": "Günbatar Awstraliýa tomusky wagty", "AWST": "Günbatar Awstraliýa standart wagty", "BOT": "Boliwiýa wagty", "BT": "Butan wagty", "CAT": "Merkezi Afrika wagty", "CDT": "Merkezi Amerika tomusky wagty", "CHADT": "Çatem tomusky wagty", "CHAST": "Çatem standart wagty", "CLST": "Çili tomusky wagty", "CLT": "Çili standart wagty", "COST": "Kolumbiýa tomusky wagty", "COT": "Kolumbiýa standart wagty", "CST": "Merkezi Amerika standart wagty", "ChST": "Çamorro wagty", "EAT": "Gündogar Afrika wagty", "ECT": "Ekwador wagty", "EDT": "Demirgazyk Amerika gündogar tomusky wagty", "EST": "Demirgazyk Amerika gündogar standart wagty", "GFT": "Fransuz Gwianasy wagty", "GMT": "Grinwiç boýunça orta wagt", "GST": "Pars aýlagy standart wagty", "GYT": "Gaýana wagty", "HADT": "Gawaý-Aleut tomusky wagty", "HAST": "Gawaý-Aleut standart wagty", "HAT": "Nýufaundlend tomusky wagty", "HECU": "Kuba tomusky wagty", "HEEG": "Gündogar Grenlandiýa tomusky wagty", "HENOMX": "Demirgazyk-günbatar Meksika tomusky wagty", "HEOG": "Günbatar Grenlandiýa tomusky wagty", "HEPM": "Sen-Pýer we Mikelon tomusky wagty", "HEPMX": "Meksikan Ýuwaş umman tomusky wagty", "HKST": "Gonkong tomusky wagty", "HKT": "Gonkong standart wagty", "HNCU": "Kuba standart wagty", "HNEG": "Gündogar Grenlandiýa standart wagty", "HNNOMX": "Demirgazyk-günbatar Meksika standart wagty", "HNOG": "Günbatar Grenlandiýa standart wagty", "HNPM": "Sen-Pýer we Mikelon standart wagty", "HNPMX": "Meksikan Ýuwaş umman standart wagty", "HNT": "Nýufaundlend standart wagty", "IST": "Hindistan standart wagty", "JDT": "Ýaponiýa tomusky wagty", "JST": "Ýaponiýa standart wagty", "LHDT": "Lord-Hau tomusky wagty", "LHST": "Lord-Hau standart wagty", "MDT": "Demirgazyk Amerika dag tomusky wagty", "MESZ": "Merkezi Ýewropa tomusky wagty", "MEZ": "Merkezi Ýewropa standart wagty", "MST": "Demirgazyk Amerika dag standart wagty", "MYT": "Malaýziýa wagty", "NZDT": "Täze Zelandiýa tomusky wagty", "NZST": "Täze Zelandiýa standart wagty", "OESZ": "Gündogar Ýewropa tomusky wagty", "OEZ": "Gündogar Ýewropa standart wagty", "PDT": "Demirgazyk Amerika Ýuwaş umman tomusky wagty", "PST": "Demirgazyk Amerika Ýuwaş umman standart wagty", "SAST": "Günorta Afrika standart wagty", "SGT": "Singapur wagty", "SRT": "Surinam wagty", "TMST": "Türkmenistan tomusky wagty", "TMT": "Türkmenistan standart wagty", "UYST": "Urugwaý tomusky wagty", "UYT": "Urugwaý standart wagty", "VET": "Wenesuela wagty", "WARST": "Günbatar Argentina tomusky wagty", "WART": "Günbatar Argentina standart wagty", "WAST": "Günbatar Afrika tomusky wagty", "WAT": "Günbatar Afrika standart wagty", "WESZ": "Günbatar Ýewropa tomusky wagty", "WEZ": "Günbatar Ýewropa standart wagty", "WIB": "Günbatar Indoneziýa wagty", "WIT": "Gündogar Indoneziýa wagty", "WITA": "Merkezi Indoneziýa wagty", "∅∅∅": "Braziliýa tomusky wagty"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_8c251b28aa9de80970be20b481b38b9b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_96a5502b4d80e3468da0c2115685c65f,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["to"] = func() locales.Translator {
		return &localen{

			locale:                 "to",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‏-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "F$", "FKP", "FRF", "GBP", "GEK", "₾", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "S$", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "T$", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Sān", "Fēp", "Maʻa", "ʻEpe", "Mē", "Sun", "Siu", "ʻAok", "Sep", "ʻOka", "Nōv", "Tīs"},
			monthsNarrow:           []string{"", "S", "F", "M", "E", "M", "S", "S", "A", "S", "O", "N", "T"},
			monthsWide:             []string{"", "Sānuali", "Fēpueli", "Maʻasi", "ʻEpeleli", "Mē", "Sune", "Siulai", "ʻAokosi", "Sepitema", "ʻOkatopa", "Nōvema", "Tīsema"},
			daysAbbreviated:        []string{"Sāp", "Mōn", "Tūs", "Pul", "Tuʻa", "Fal", "Tok"},
			daysNarrow:             []string{"S", "M", "T", "P", "T", "F", "T"},
			daysShort:              []string{"Sāp", "Mōn", "Tūs", "Pul", "Tuʻa", "Fal", "Tok"},
			daysWide:               []string{"Sāpate", "Mōnite", "Tūsite", "Pulelulu", "Tuʻapulelulu", "Falaite", "Tokonaki"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"hengihengi", "efiafi"},
			erasAbbreviated:        []string{"KM", "TS"},
			erasNarrow:             []string{"KM", "TS"},
			erasWide:               []string{"ki muʻa", "taʻu ʻo Sīsū"},
			timezones:              map[string]string{"ACDT": "houa fakaʻaositelēlia-loto taimi liliu", "ACST": "houa fakaʻaositelēlia-loto taimi totonu", "ACWDT": "houa fakaʻaositelēlia-loto-hihifo taimi liliu", "ACWST": "houa fakaʻaositelēlia-loto-hihifo taimi totonu", "ADT": "houa fakaʻamelika-tokelau ʻatalanitiki taimi liliu", "AEDT": "houa fakaʻaositelēlia-hahake taimi liliu", "AEST": "houa fakaʻaositelēlia-hahake taimi totonu", "AKDT": "houa fakaʻalasika taimi liliu", "AKST": "houa fakaʻalasika taimi totonu", "ARST": "houa fakaʻasenitina taimi liliu", "ART": "houa fakaʻasenitina taimi totonu", "AST": "houa fakaʻamelika-tokelau ʻatalanitiki taimi totonu", "AWDT": "houa fakaʻaositelēlia-hihifo taimi liliu", "AWST": "houa fakaʻaositelēlia-hihifo taimi totonu", "BOT": "houa fakapolīvia", "BT": "houa fakapūtani", "CAT": "houa fakaʻafelika-loto", "CDT": "houa fakaʻamelika-tokelau loto taimi liliu", "CHADT": "houa fakasatihami taimi liliu", "CHAST": "houa fakasatihami taimi totonu", "CLST": "houa fakasili taimi liliu", "CLT": "houa fakasili taimi totonu", "COST": "houa fakakolomipia taimi liliu", "COT": "houa fakakolomipia taimi totonu", "CST": "houa fakaʻamelika-tokelau loto taimi totonu", "ChST": "houa fakakamolo", "EAT": "houa fakaʻafelika-hahake", "ECT": "houa fakaʻekuetoa", "EDT": "houa fakaʻamelika-tokelau hahake taimi liliu", "EST": "houa fakaʻamelika-tokelau hahake taimi totonu", "GFT": "houa fakakuiana-fakafalanisē", "GMT": "houa fakakiliniuisi mālie", "GST": "houa fakakūlifi", "GYT": "houa fakakuiana", "HADT": "houa fakahauaʻi taimi liliu", "HAST": "houa fakahauaʻi taimi totonu", "HAT": "houa fakafonuaʻilofoʻou taimi liliu", "HECU": "houa fakakiupa taimi liliu", "HEEG": "houa fakafonuamata-hahake taimi liliu", "HENOMX": "houa fakamekisikou-tokelauhihifo taimi liliu", "HEOG": "houa fakafonuamata-hihifo taimi liliu", "HEPM": "houa fakasā-piea-mo-mikeloni taimi liliu", "HEPMX": "houa fakamekisikou-pasifika taimi liliu", "HKST": "houa fakahongi-kongi taimi liliu", "HKT": "houa fakahongi-kongi taimi totonu", "HNCU": "houa fakakiupa taimi totonu", "HNEG": "houa fakafonuamata-hahake taimi totonu", "HNNOMX": "houa fakamekisikou-tokelauhihifo taimi totonu", "HNOG": "houa fakafonuamata-hihifo taimi totonu", "HNPM": "houa fakasā-piea-mo-mikeloni taimi totonu", "HNPMX": "houa fakamekisikou-pasifika taimi totonu", "HNT": "houa fakafonuaʻilofoʻou taimi totonu", "IST": "houa fakaʻinitia", "JDT": "houa fakasiapani taimi liliu", "JST": "houa fakasiapani taimi totonu", "LHDT": "houa fakamotuʻeikihoue taimi liliu", "LHST": "houa fakamotuʻeikihoue taimi totonu", "MDT": "houa fakaʻamelika-tokelau moʻunga taimi liliu", "MESZ": "houa fakaʻeulope-loto taimi liliu", "MEZ": "houa fakaʻeulope-loto taimi totonu", "MST": "houa fakaʻamelika-tokelau moʻunga taimi totonu", "MYT": "houa fakamaleisia", "NZDT": "houa fakanuʻusila taimi liliu", "NZST": "houa fakanuʻusila taimi totonu", "OESZ": "houa fakaʻeulope-hahake taimi liliu", "OEZ": "houa fakaʻeulope-hahake taimi totonu", "PDT": "houa fakaʻamelika-tokelau pasifika taimi liliu", "PST": "houa fakaʻamelika-tokelau pasifika taimi totonu", "SAST": "houa fakaʻafelika-tonga", "SGT": "houa fakasingapoa", "SRT": "houa fakasuliname", "TMST": "houa fakatūkimenisitani taimi liliu", "TMT": "houa fakatūkimenisitani taimi totonu", "UYST": "houa fakaʻulukuai taimi liliu", "UYT": "houa fakaʻulukuai taimi totonu", "VET": "houa fakavenesuela", "WARST": "houa fakaʻasenitina-hihifo taimi liliu", "WART": "houa fakaʻasenitina-hihifo taimi totonu", "WAST": "houa fakaʻafelika-hihifo taimi liliu", "WAT": "houa fakaʻafelika-hihifo taimi totonu", "WESZ": "houa fakaʻeulope-hihifo taimi liliu", "WEZ": "houa fakaʻeulope-hihifo taimi totonu", "WIB": "houa fakaʻinitonisia-hihifo", "WIT": "houa fakaʻinitonisia-hahake", "WITA": "houa fakaʻinitonisia-loto", "∅∅∅": "houa fakapalāsila taimi liliu"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f7223149b88d711e36c711437b51fa28,
			fnFmtCurrency:         fnFmtCurrency_d61f2036c9f7cc7a15c2d51b90ffbef9,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_96d7d1823041a48b887e3e929259dbdf,
			fnFmtPercent:          fnFmtPercent_5bdc6fa582de100c13fd696db4607f17,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["to_to"] = func() locales.Translator {
		return &localen{

			locale:                 "to_TO",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                "٫",
			group:                  "٬",
			minus:                  "‏-",
			percent:                "٪؜",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:          " ",
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Sān", "Fēp", "Maʻa", "ʻEpe", "Mē", "Sun", "Siu", "ʻAok", "Sep", "ʻOka", "Nōv", "Tīs"},
			monthsNarrow:           []string{"", "S", "F", "M", "E", "M", "S", "S", "A", "S", "O", "N", "T"},
			monthsWide:             []string{"", "Sānuali", "Fēpueli", "Maʻasi", "ʻEpeleli", "Mē", "Sune", "Siulai", "ʻAokosi", "Sepitema", "ʻOkatopa", "Nōvema", "Tīsema"},
			daysAbbreviated:        []string{"Sāp", "Mōn", "Tūs", "Pul", "Tuʻa", "Fal", "Tok"},
			daysNarrow:             []string{"S", "M", "T", "P", "T", "F", "T"},
			daysShort:              []string{"Sāp", "Mōn", "Tūs", "Pul", "Tuʻa", "Fal", "Tok"},
			daysWide:               []string{"Sāpate", "Mōnite", "Tūsite", "Pulelulu", "Tuʻapulelulu", "Falaite", "Tokonaki"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"AM", "PM"},
			periodsWide:            []string{"hengihengi", "efiafi"},
			erasAbbreviated:        []string{"KM", "TS"},
			erasNarrow:             []string{"KM", "TS"},
			erasWide:               []string{"ki muʻa", "taʻu ʻo Sīsū"},
			timezones:              map[string]string{"ACDT": "houa fakaʻaositelēlia-loto taimi liliu", "ACST": "houa fakaʻaositelēlia-loto taimi totonu", "ACWDT": "houa fakaʻaositelēlia-loto-hihifo taimi liliu", "ACWST": "houa fakaʻaositelēlia-loto-hihifo taimi totonu", "ADT": "houa fakaʻamelika-tokelau ʻatalanitiki taimi liliu", "AEDT": "houa fakaʻaositelēlia-hahake taimi liliu", "AEST": "houa fakaʻaositelēlia-hahake taimi totonu", "AKDT": "houa fakaʻalasika taimi liliu", "AKST": "houa fakaʻalasika taimi totonu", "ARST": "houa fakaʻasenitina taimi liliu", "ART": "houa fakaʻasenitina taimi totonu", "AST": "houa fakaʻamelika-tokelau ʻatalanitiki taimi totonu", "AWDT": "houa fakaʻaositelēlia-hihifo taimi liliu", "AWST": "houa fakaʻaositelēlia-hihifo taimi totonu", "BOT": "houa fakapolīvia", "BT": "houa fakapūtani", "CAT": "houa fakaʻafelika-loto", "CDT": "houa fakaʻamelika-tokelau loto taimi liliu", "CHADT": "houa fakasatihami taimi liliu", "CHAST": "houa fakasatihami taimi totonu", "CLST": "houa fakasili taimi liliu", "CLT": "houa fakasili taimi totonu", "COST": "houa fakakolomipia taimi liliu", "COT": "houa fakakolomipia taimi totonu", "CST": "houa fakaʻamelika-tokelau loto taimi totonu", "ChST": "houa fakakamolo", "EAT": "houa fakaʻafelika-hahake", "ECT": "houa fakaʻekuetoa", "EDT": "houa fakaʻamelika-tokelau hahake taimi liliu", "EST": "houa fakaʻamelika-tokelau hahake taimi totonu", "GFT": "houa fakakuiana-fakafalanisē", "GMT": "houa fakakiliniuisi mālie", "GST": "houa fakakūlifi", "GYT": "houa fakakuiana", "HADT": "houa fakahauaʻi taimi liliu", "HAST": "houa fakahauaʻi taimi totonu", "HAT": "houa fakafonuaʻilofoʻou taimi liliu", "HECU": "houa fakakiupa taimi liliu", "HEEG": "houa fakafonuamata-hahake taimi liliu", "HENOMX": "houa fakamekisikou-tokelauhihifo taimi liliu", "HEOG": "houa fakafonuamata-hihifo taimi liliu", "HEPM": "houa fakasā-piea-mo-mikeloni taimi liliu", "HEPMX": "houa fakamekisikou-pasifika taimi liliu", "HKST": "houa fakahongi-kongi taimi liliu", "HKT": "houa fakahongi-kongi taimi totonu", "HNCU": "houa fakakiupa taimi totonu", "HNEG": "houa fakafonuamata-hahake taimi totonu", "HNNOMX": "houa fakamekisikou-tokelauhihifo taimi totonu", "HNOG": "houa fakafonuamata-hihifo taimi totonu", "HNPM": "houa fakasā-piea-mo-mikeloni taimi totonu", "HNPMX": "houa fakamekisikou-pasifika taimi totonu", "HNT": "houa fakafonuaʻilofoʻou taimi totonu", "IST": "houa fakaʻinitia", "JDT": "houa fakasiapani taimi liliu", "JST": "houa fakasiapani taimi totonu", "LHDT": "houa fakamotuʻeikihoue taimi liliu", "LHST": "houa fakamotuʻeikihoue taimi totonu", "MDT": "houa fakaʻamelika-tokelau moʻunga taimi liliu", "MESZ": "houa fakaʻeulope-loto taimi liliu", "MEZ": "houa fakaʻeulope-loto taimi totonu", "MST": "houa fakaʻamelika-tokelau moʻunga taimi totonu", "MYT": "houa fakamaleisia", "NZDT": "houa fakanuʻusila taimi liliu", "NZST": "houa fakanuʻusila taimi totonu", "OESZ": "houa fakaʻeulope-hahake taimi liliu", "OEZ": "houa fakaʻeulope-hahake taimi totonu", "PDT": "houa fakaʻamelika-tokelau pasifika taimi liliu", "PST": "houa fakaʻamelika-tokelau pasifika taimi totonu", "SAST": "houa fakaʻafelika-tonga", "SGT": "houa fakasingapoa", "SRT": "houa fakasuliname", "TMST": "houa fakatūkimenisitani taimi liliu", "TMT": "houa fakatūkimenisitani taimi totonu", "UYST": "houa fakaʻulukuai taimi liliu", "UYT": "houa fakaʻulukuai taimi totonu", "VET": "houa fakavenesuela", "WARST": "houa fakaʻasenitina-hihifo taimi liliu", "WART": "houa fakaʻasenitina-hihifo taimi totonu", "WAST": "houa fakaʻafelika-hihifo taimi liliu", "WAT": "houa fakaʻafelika-hihifo taimi totonu", "WESZ": "houa fakaʻeulope-hihifo taimi liliu", "WEZ": "houa fakaʻeulope-hihifo taimi totonu", "WIB": "houa fakaʻinitonisia-hihifo", "WIT": "houa fakaʻinitonisia-hahake", "WITA": "houa fakaʻinitonisia-loto", "∅∅∅": "houa fakapalāsila taimi liliu"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_f7223149b88d711e36c711437b51fa28,
			fnFmtCurrency:         fnFmtCurrency_d61f2036c9f7cc7a15c2d51b90ffbef9,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_96d7d1823041a48b887e3e929259dbdf,
			fnFmtPercent:          fnFmtPercent_5bdc6fa582de100c13fd696db4607f17,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tr"] = func() locales.Translator {
		return &localen{

			locale:                 "tr",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "₺", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Oca", "Şub", "Mar", "Nis", "May", "Haz", "Tem", "Ağu", "Eyl", "Eki", "Kas", "Ara"},
			monthsNarrow:           []string{"", "O", "Ş", "M", "N", "M", "H", "T", "A", "E", "E", "K", "A"},
			monthsWide:             []string{"", "Ocak", "Şubat", "Mart", "Nisan", "Mayıs", "Haziran", "Temmuz", "Ağustos", "Eylül", "Ekim", "Kasım", "Aralık"},
			daysAbbreviated:        []string{"Paz", "Pzt", "Sal", "Çar", "Per", "Cum", "Cmt"},
			daysNarrow:             []string{"P", "P", "S", "Ç", "P", "C", "C"},
			daysShort:              []string{"Pa", "Pt", "Sa", "Ça", "Pe", "Cu", "Ct"},
			daysWide:               []string{"Pazar", "Pazartesi", "Salı", "Çarşamba", "Perşembe", "Cuma", "Cumartesi"},
			periodsAbbreviated:     []string{"ÖÖ", "ÖS"},
			periodsNarrow:          []string{"öö", "ös"},
			periodsWide:            []string{"ÖÖ", "ÖS"},
			erasAbbreviated:        []string{"MÖ", "MS"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Milattan Önce", "Milattan Sonra"},
			timezones:              map[string]string{"ACDT": "Orta Avustralya Yaz Saati", "ACST": "Orta Avustralya Standart Saati", "ACWDT": "İç Batı Avustralya Yaz Saati", "ACWST": "İç Batı Avustralya Standart Saati", "ADT": "Atlantik Yaz Saati", "AEDT": "Doğu Avustralya Yaz Saati", "AEST": "Doğu Avustralya Standart Saati", "AKDT": "Alaska Yaz Saati", "AKST": "Alaska Standart Saati", "ARST": "Arjantin Yaz Saati", "ART": "Arjantin Standart Saati", "AST": "Atlantik Standart Saati", "AWDT": "Batı Avustralya Yaz Saati", "AWST": "Batı Avustralya Standart Saati", "BOT": "Bolivya Saati", "BT": "Butan Saati", "CAT": "Orta Afrika Saati", "CDT": "Kuzey Amerika Merkezi Yaz Saati", "CHADT": "Chatham Yaz Saati", "CHAST": "Chatham Standart Saati", "CLST": "Şili Yaz Saati", "CLT": "Şili Standart Saati", "COST": "Kolombiya Yaz Saati", "COT": "Kolombiya Standart Saati", "CST": "Kuzey Amerika Merkezi Standart Saati", "ChST": "Chamorro Saati", "EAT": "Doğu Afrika Saati", "ECT": "Ekvador Saati", "EDT": "Kuzey Amerika Doğu Yaz Saati", "EST": "Kuzey Amerika Doğu Standart Saati", "GFT": "Fransız Guyanası Saati", "GMT": "Greenwich Ortalama Saati", "GST": "Körfez Saati", "GYT": "Guyana Saati", "HADT": "Hawaii-Aleut Yaz Saati", "HAST": "Hawaii-Aleut Standart Saati", "HAT": "Newfoundland Yaz Saati", "HECU": "Küba Yaz Saati", "HEEG": "Doğu Grönland Yaz Saati", "HENOMX": "Kuzeybatı Meksika Yaz Saati", "HEOG": "Batı Grönland Yaz Saati", "HEPM": "Saint Pierre ve Miquelon Yaz Saati", "HEPMX": "Meksika Pasifik Kıyısı Yaz Saati", "HKST": "Hong Kong Yaz Saati", "HKT": "Hong Kong Standart Saati", "HNCU": "Küba Standart Saati", "HNEG": "Doğu Grönland Standart Saati", "HNNOMX": "Kuzeybatı Meksika Standart Saati", "HNOG": "Batı Grönland Standart Saati", "HNPM": "Saint Pierre ve Miquelon Standart Saati", "HNPMX": "Meksika Pasifik Kıyısı Standart Saati", "HNT": "Newfoundland Standart Saati", "IST": "Hindistan Standart Saati", "JDT": "Japonya Yaz Saati", "JST": "Japonya Standart Saati", "LHDT": "Lord Howe Yaz Saati", "LHST": "Lord Howe Standart Saati", "MDT": "Kuzey Amerika Dağ Yaz Saati", "MESZ": "Orta Avrupa Yaz Saati", "MEZ": "Orta Avrupa Standart Saati", "MST": "Kuzey Amerika Dağ Standart Saati", "MYT": "Malezya Saati", "NZDT": "Yeni Zelanda Yaz Saati", "NZST": "Yeni Zelanda Standart Saati", "OESZ": "Doğu Avrupa Yaz Saati", "OEZ": "Doğu Avrupa Standart Saati", "PDT": "Kuzey Amerika Pasifik Yaz Saati", "PST": "Kuzey Amerika Pasifik Standart Saati", "SAST": "Güney Afrika Standart Saati", "SGT": "Singapur Standart Saati", "SRT": "Surinam Saati", "TMST": "Türkmenistan Yaz Saati", "TMT": "Türkmenistan Standart Saati", "UYST": "Uruguay Yaz Saati", "UYT": "Uruguay Standart Saati", "VET": "Venezuela Saati", "WARST": "Batı Arjantin Yaz Saati", "WART": "Batı Arjantin Standart Saati", "WAST": "Batı Afrika Yaz Saati", "WAT": "Batı Afrika Standart Saati", "WESZ": "Batı Avrupa Yaz Saati", "WEZ": "Batı Avrupa Standart Saati", "WIB": "Batı Endonezya Saati", "WIT": "Doğu Endonezya Saati", "WITA": "Orta Endonezya Saati", "∅∅∅": "Azorlar Yaz Saati"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_8c251b28aa9de80970be20b481b38b9b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c0764a5a895e1bd677cbe99232a5e159,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1842fa1c80e067ee20bb9d4acb93d0eb,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tr_cy"] = func() locales.Translator {
		return &localen{

			locale:                 "tr_CY",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Oca", "Şub", "Mar", "Nis", "May", "Haz", "Tem", "Ağu", "Eyl", "Eki", "Kas", "Ara"},
			monthsNarrow:           []string{"", "O", "Ş", "M", "N", "M", "H", "T", "A", "E", "E", "K", "A"},
			monthsWide:             []string{"", "Ocak", "Şubat", "Mart", "Nisan", "Mayıs", "Haziran", "Temmuz", "Ağustos", "Eylül", "Ekim", "Kasım", "Aralık"},
			daysAbbreviated:        []string{"Paz", "Pzt", "Sal", "Çar", "Per", "Cum", "Cmt"},
			daysNarrow:             []string{"P", "P", "S", "Ç", "P", "C", "C"},
			daysShort:              []string{"Pa", "Pt", "Sa", "Ça", "Pe", "Cu", "Ct"},
			daysWide:               []string{"Pazar", "Pazartesi", "Salı", "Çarşamba", "Perşembe", "Cuma", "Cumartesi"},
			periodsAbbreviated:     []string{"ÖÖ", "ÖS"},
			periodsNarrow:          []string{"öö", "ös"},
			periodsWide:            []string{"ÖÖ", "ÖS"},
			erasAbbreviated:        []string{"MÖ", "MS"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Milattan Önce", "Milattan Sonra"},
			timezones:              map[string]string{"ACDT": "Orta Avustralya Yaz Saati", "ACST": "Orta Avustralya Standart Saati", "ACWDT": "İç Batı Avustralya Yaz Saati", "ACWST": "İç Batı Avustralya Standart Saati", "ADT": "Atlantik Yaz Saati", "AEDT": "Doğu Avustralya Yaz Saati", "AEST": "Doğu Avustralya Standart Saati", "AKDT": "Alaska Yaz Saati", "AKST": "Alaska Standart Saati", "ARST": "Arjantin Yaz Saati", "ART": "Arjantin Standart Saati", "AST": "Atlantik Standart Saati", "AWDT": "Batı Avustralya Yaz Saati", "AWST": "Batı Avustralya Standart Saati", "BOT": "Bolivya Saati", "BT": "Butan Saati", "CAT": "Orta Afrika Saati", "CDT": "Kuzey Amerika Merkezi Yaz Saati", "CHADT": "Chatham Yaz Saati", "CHAST": "Chatham Standart Saati", "CLST": "Şili Yaz Saati", "CLT": "Şili Standart Saati", "COST": "Kolombiya Yaz Saati", "COT": "Kolombiya Standart Saati", "CST": "Kuzey Amerika Merkezi Standart Saati", "ChST": "Chamorro Saati", "EAT": "Doğu Afrika Saati", "ECT": "Ekvador Saati", "EDT": "Kuzey Amerika Doğu Yaz Saati", "EST": "Kuzey Amerika Doğu Standart Saati", "GFT": "Fransız Guyanası Saati", "GMT": "Greenwich Ortalama Saati", "GST": "Körfez Saati", "GYT": "Guyana Saati", "HADT": "Hawaii-Aleut Yaz Saati", "HAST": "Hawaii-Aleut Standart Saati", "HAT": "Newfoundland Yaz Saati", "HECU": "Küba Yaz Saati", "HEEG": "Doğu Grönland Yaz Saati", "HENOMX": "Kuzeybatı Meksika Yaz Saati", "HEOG": "Batı Grönland Yaz Saati", "HEPM": "Saint Pierre ve Miquelon Yaz Saati", "HEPMX": "Meksika Pasifik Kıyısı Yaz Saati", "HKST": "Hong Kong Yaz Saati", "HKT": "Hong Kong Standart Saati", "HNCU": "Küba Standart Saati", "HNEG": "Doğu Grönland Standart Saati", "HNNOMX": "Kuzeybatı Meksika Standart Saati", "HNOG": "Batı Grönland Standart Saati", "HNPM": "Saint Pierre ve Miquelon Standart Saati", "HNPMX": "Meksika Pasifik Kıyısı Standart Saati", "HNT": "Newfoundland Standart Saati", "IST": "Hindistan Standart Saati", "JDT": "Japonya Yaz Saati", "JST": "Japonya Standart Saati", "LHDT": "Lord Howe Yaz Saati", "LHST": "Lord Howe Standart Saati", "MDT": "Makao Yaz Saati", "MESZ": "Orta Avrupa Yaz Saati", "MEZ": "Orta Avrupa Standart Saati", "MST": "Makao Standart Saati", "MYT": "Malezya Saati", "NZDT": "Yeni Zelanda Yaz Saati", "NZST": "Yeni Zelanda Standart Saati", "OESZ": "Doğu Avrupa Yaz Saati", "OEZ": "Doğu Avrupa Standart Saati", "PDT": "Kuzey Amerika Pasifik Yaz Saati", "PST": "Kuzey Amerika Pasifik Standart Saati", "SAST": "Güney Afrika Standart Saati", "SGT": "Singapur Standart Saati", "SRT": "Surinam Saati", "TMST": "Türkmenistan Yaz Saati", "TMT": "Türkmenistan Standart Saati", "UYST": "Uruguay Yaz Saati", "UYT": "Uruguay Standart Saati", "VET": "Venezuela Saati", "WARST": "Batı Arjantin Yaz Saati", "WART": "Batı Arjantin Standart Saati", "WAST": "Batı Afrika Yaz Saati", "WAT": "Batı Afrika Standart Saati", "WESZ": "Batı Avrupa Yaz Saati", "WEZ": "Batı Avrupa Standart Saati", "WIB": "Batı Endonezya Saati", "WIT": "Doğu Endonezya Saati", "WITA": "Orta Endonezya Saati", "∅∅∅": "Peru Yaz Saati"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_8c251b28aa9de80970be20b481b38b9b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c0764a5a895e1bd677cbe99232a5e159,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1842fa1c80e067ee20bb9d4acb93d0eb,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tr_tr"] = func() locales.Translator {
		return &localen{

			locale:                 "tr_TR",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Oca", "Şub", "Mar", "Nis", "May", "Haz", "Tem", "Ağu", "Eyl", "Eki", "Kas", "Ara"},
			monthsNarrow:           []string{"", "O", "Ş", "M", "N", "M", "H", "T", "A", "E", "E", "K", "A"},
			monthsWide:             []string{"", "Ocak", "Şubat", "Mart", "Nisan", "Mayıs", "Haziran", "Temmuz", "Ağustos", "Eylül", "Ekim", "Kasım", "Aralık"},
			daysAbbreviated:        []string{"Paz", "Pzt", "Sal", "Çar", "Per", "Cum", "Cmt"},
			daysNarrow:             []string{"P", "P", "S", "Ç", "P", "C", "C"},
			daysShort:              []string{"Pa", "Pt", "Sa", "Ça", "Pe", "Cu", "Ct"},
			daysWide:               []string{"Pazar", "Pazartesi", "Salı", "Çarşamba", "Perşembe", "Cuma", "Cumartesi"},
			periodsAbbreviated:     []string{"ÖÖ", "ÖS"},
			periodsNarrow:          []string{"öö", "ös"},
			periodsWide:            []string{"ÖÖ", "ÖS"},
			erasAbbreviated:        []string{"MÖ", "MS"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Milattan Önce", "Milattan Sonra"},
			timezones:              map[string]string{"ACDT": "Orta Avustralya Yaz Saati", "ACST": "Orta Avustralya Standart Saati", "ACWDT": "İç Batı Avustralya Yaz Saati", "ACWST": "İç Batı Avustralya Standart Saati", "ADT": "Atlantik Yaz Saati", "AEDT": "Doğu Avustralya Yaz Saati", "AEST": "Doğu Avustralya Standart Saati", "AKDT": "Alaska Yaz Saati", "AKST": "Alaska Standart Saati", "ARST": "Arjantin Yaz Saati", "ART": "Arjantin Standart Saati", "AST": "Atlantik Standart Saati", "AWDT": "Batı Avustralya Yaz Saati", "AWST": "Batı Avustralya Standart Saati", "BOT": "Bolivya Saati", "BT": "Butan Saati", "CAT": "Orta Afrika Saati", "CDT": "Kuzey Amerika Merkezi Yaz Saati", "CHADT": "Chatham Yaz Saati", "CHAST": "Chatham Standart Saati", "CLST": "Şili Yaz Saati", "CLT": "Şili Standart Saati", "COST": "Kolombiya Yaz Saati", "COT": "Kolombiya Standart Saati", "CST": "Kuzey Amerika Merkezi Standart Saati", "ChST": "Chamorro Saati", "EAT": "Doğu Afrika Saati", "ECT": "Ekvador Saati", "EDT": "Kuzey Amerika Doğu Yaz Saati", "EST": "Kuzey Amerika Doğu Standart Saati", "GFT": "Fransız Guyanası Saati", "GMT": "Greenwich Ortalama Saati", "GST": "Körfez Saati", "GYT": "Guyana Saati", "HADT": "Hawaii-Aleut Yaz Saati", "HAST": "Hawaii-Aleut Standart Saati", "HAT": "Newfoundland Yaz Saati", "HECU": "Küba Yaz Saati", "HEEG": "Doğu Grönland Yaz Saati", "HENOMX": "Kuzeybatı Meksika Yaz Saati", "HEOG": "Batı Grönland Yaz Saati", "HEPM": "Saint Pierre ve Miquelon Yaz Saati", "HEPMX": "Meksika Pasifik Kıyısı Yaz Saati", "HKST": "Hong Kong Yaz Saati", "HKT": "Hong Kong Standart Saati", "HNCU": "Küba Standart Saati", "HNEG": "Doğu Grönland Standart Saati", "HNNOMX": "Kuzeybatı Meksika Standart Saati", "HNOG": "Batı Grönland Standart Saati", "HNPM": "Saint Pierre ve Miquelon Standart Saati", "HNPMX": "Meksika Pasifik Kıyısı Standart Saati", "HNT": "Newfoundland Standart Saati", "IST": "Hindistan Standart Saati", "JDT": "Japonya Yaz Saati", "JST": "Japonya Standart Saati", "LHDT": "Lord Howe Yaz Saati", "LHST": "Lord Howe Standart Saati", "MDT": "Kuzey Amerika Dağ Yaz Saati", "MESZ": "Orta Avrupa Yaz Saati", "MEZ": "Orta Avrupa Standart Saati", "MST": "Kuzey Amerika Dağ Standart Saati", "MYT": "Malezya Saati", "NZDT": "Yeni Zelanda Yaz Saati", "NZST": "Yeni Zelanda Standart Saati", "OESZ": "Doğu Avrupa Yaz Saati", "OEZ": "Doğu Avrupa Standart Saati", "PDT": "Kuzey Amerika Pasifik Yaz Saati", "PST": "Kuzey Amerika Pasifik Standart Saati", "SAST": "Güney Afrika Standart Saati", "SGT": "Singapur Standart Saati", "SRT": "Surinam Saati", "TMST": "Türkmenistan Yaz Saati", "TMT": "Türkmenistan Standart Saati", "UYST": "Uruguay Yaz Saati", "UYT": "Uruguay Standart Saati", "VET": "Venezuela Saati", "WARST": "Batı Arjantin Yaz Saati", "WART": "Batı Arjantin Standart Saati", "WAST": "Batı Afrika Yaz Saati", "WAT": "Batı Afrika Standart Saati", "WESZ": "Batı Avrupa Yaz Saati", "WEZ": "Batı Avrupa Standart Saati", "WIB": "Batı Endonezya Saati", "WIT": "Doğu Endonezya Saati", "WITA": "Orta Endonezya Saati", "∅∅∅": "Peru Yaz Saati"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_8c251b28aa9de80970be20b481b38b9b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_c0764a5a895e1bd677cbe99232a5e159,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1842fa1c80e067ee20bb9d4acb93d0eb,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tt"] = func() locales.Translator {
		return &localen{

			locale:             "tt",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "р.", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "гыйн.", "фев.", "мар.", "апр.", "май", "июнь", "июль", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "гыйнвар", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"},
			daysAbbreviated:    []string{"якш.", "дүш.", "сиш.", "чәр.", "пәнҗ.", "җом.", "шим."},
			daysNarrow:         []string{"Я", "Д", "С", "Ч", "П", "Җ", "Ш"},
			daysShort:          []string{"якш.", "дүш.", "сиш.", "чәр.", "пәнҗ.", "җом.", "шим."},
			daysWide:           []string{"якшәмбе", "дүшәмбе", "сишәмбе", "чәршәмбе", "пәнҗешәмбе", "җомга", "шимбә"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"б.э.к.", "милади"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"безнең эрага кадәр", "безнең эра"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Төньяк Америка җәйге атлантик вакыты", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Төньяк Америка гадәти атлантик вакыты", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Төньяк Америка җәйге үзәк вакыты", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Төньяк Америка гадәти үзәк вакыты", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Төньяк Америка җәйге көнчыгыш вакыты", "EST": "Төньяк Америка гадәти көнчыгыш вакыты", "GFT": "GFT", "GMT": "Гринвич уртача вакыты", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Төньяк Америка җәйге тау вакыты", "MESZ": "җәйге Үзәк Европа вакыты", "MEZ": "гадәти Үзәк Европа вакыты", "MST": "Төньяк Америка гадәти тау вакыты", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "җәйге Көнчыгыш Европа вакыты", "OEZ": "гадәти Көнчыгыш Европа вакыты", "PDT": "Төньяк Америка җәйге Тын океан вакыты", "PST": "Төньяк Америка гадәти Тын океан вакыты", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "җәйге Көнбатыш Европа вакыты", "WEZ": "гадәти Көнбатыш Европа вакыты", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_a5f5cc8ae427c1221f50697b6908f7a5,
			fnFmtDateLong:         fnFmtDateLong_6384cc06509d395705e5eb2161388671,
			fnFmtDateMedium:       fnFmtDateMedium_d6c733b45a5fc6c4b67eabb17c635dfe,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tt_ru"] = func() locales.Translator {
		return &localen{

			locale:             "tt_RU",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "гыйн.", "фев.", "мар.", "апр.", "май", "июнь", "июль", "авг.", "сент.", "окт.", "нояб.", "дек."},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "гыйнвар", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"},
			daysAbbreviated:    []string{"якш.", "дүш.", "сиш.", "чәр.", "пәнҗ.", "җом.", "шим."},
			daysNarrow:         []string{"Я", "Д", "С", "Ч", "П", "Җ", "Ш"},
			daysShort:          []string{"якш.", "дүш.", "сиш.", "чәр.", "пәнҗ.", "җом.", "шим."},
			daysWide:           []string{"якшәмбе", "дүшәмбе", "сишәмбе", "чәршәмбе", "пәнҗешәмбе", "җомга", "шимбә"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"AM", "PM"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"б.э.к.", "милади"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"безнең эрага кадәр", "безнең эра"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Төньяк Америка җәйге атлантик вакыты", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Төньяк Америка гадәти атлантик вакыты", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "Төньяк Америка җәйге үзәк вакыты", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "Төньяк Америка гадәти үзәк вакыты", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "Төньяк Америка җәйге көнчыгыш вакыты", "EST": "Төньяк Америка гадәти көнчыгыш вакыты", "GFT": "GFT", "GMT": "Гринвич уртача вакыты", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "Төньяк Америка җәйге тау вакыты", "MESZ": "җәйге Үзәк Европа вакыты", "MEZ": "гадәти Үзәк Европа вакыты", "MST": "Төньяк Америка гадәти тау вакыты", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "җәйге Көнчыгыш Европа вакыты", "OEZ": "гадәти Көнчыгыш Европа вакыты", "PDT": "Төньяк Америка җәйге Тын океан вакыты", "PST": "Төньяк Америка гадәти Тын океан вакыты", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "җәйге Көнбатыш Европа вакыты", "WEZ": "гадәти Көнбатыш Европа вакыты", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_a5f5cc8ae427c1221f50697b6908f7a5,
			fnFmtDateLong:         fnFmtDateLong_6384cc06509d395705e5eb2161388671,
			fnFmtDateMedium:       fnFmtDateMedium_d6c733b45a5fc6c4b67eabb17c635dfe,
			fnFmtDateShort:        fnFmtDateShort_b5a69de69141b720b05f1dca7542e7e5,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_103caa51aab787e802d47e2174ea5313,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["twq"] = func() locales.Translator {
		return &localen{

			locale:             "twq",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alz", "Asi"},
			daysNarrow:         []string{"H", "T", "T", "L", "L", "L", "S"},
			daysWide:           []string{"Alhadi", "Atinni", "Atalaata", "Alarba", "Alhamiisa", "Alzuma", "Asibti"},
			periodsAbbreviated: []string{"Subbaahi", "Zaarikay b"},
			periodsWide:        []string{"Subbaahi", "Zaarikay b"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa zamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["twq_ne"] = func() locales.Translator {
		return &localen{

			locale:             "twq_NE",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ".",
			group:              " ",
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Žan", "Fee", "Mar", "Awi", "Me", "Žuw", "Žuy", "Ut", "Sek", "Okt", "Noo", "Dee"},
			monthsNarrow:       []string{"", "Ž", "F", "M", "A", "M", "Ž", "Ž", "U", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Žanwiye", "Feewiriye", "Marsi", "Awiril", "Me", "Žuweŋ", "Žuyye", "Ut", "Sektanbur", "Oktoobur", "Noowanbur", "Deesanbur"},
			daysAbbreviated:    []string{"Alh", "Ati", "Ata", "Ala", "Alm", "Alz", "Asi"},
			daysNarrow:         []string{"H", "T", "T", "L", "L", "L", "S"},
			daysWide:           []string{"Alhadi", "Atinni", "Atalaata", "Alarba", "Alhamiisa", "Alzuma", "Asibti"},
			periodsAbbreviated: []string{"Subbaahi", "Zaarikay b"},
			periodsWide:        []string{"Subbaahi", "Zaarikay b"},
			erasAbbreviated:    []string{"IJ", "IZ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Isaa jine", "Isaa zamanoo"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c53d1c28077fe8a91eda0e7a76463f6d,
			fnFmtCurrency:         fnFmtCurrency_2f14500b7d52d9b6d26d90cfb96161ed,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tzm"] = func() locales.Translator {
		return &localen{

			locale:                 "tzm",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Yen", "Yeb", "Mar", "Ibr", "May", "Yun", "Yul", "Ɣuc", "Cut", "Kṭu", "Nwa", "Duj"},
			monthsNarrow:           []string{"", "Y", "Y", "M", "I", "M", "Y", "Y", "Ɣ", "C", "K", "N", "D"},
			monthsWide:             []string{"", "Yennayer", "Yebrayer", "Mars", "Ibrir", "Mayyu", "Yunyu", "Yulyuz", "Ɣuct", "Cutanbir", "Kṭuber", "Nwanbir", "Dujanbir"},
			daysAbbreviated:        []string{"Asa", "Ayn", "Asn", "Akr", "Akw", "Asm", "Asḍ"},
			daysNarrow:             []string{"A", "A", "A", "A", "A", "A", "A"},
			daysWide:               []string{"Asamas", "Aynas", "Asinas", "Akras", "Akwas", "Asimwas", "Asiḍyas"},
			periodsAbbreviated:     []string{"Zdat azal", "Ḍeffir aza"},
			periodsWide:            []string{"Zdat azal", "Ḍeffir aza"},
			erasAbbreviated:        []string{"ZƐ", "ḌƐ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Zdat Ɛisa (TAƔ)", "Ḍeffir Ɛisa (TAƔ)"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_92e733a5c3e2c8f5728268e0f4b7afd0,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["tzm_ma"] = func() locales.Translator {
		return &localen{

			locale:                 "tzm_MA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Yen", "Yeb", "Mar", "Ibr", "May", "Yun", "Yul", "Ɣuc", "Cut", "Kṭu", "Nwa", "Duj"},
			monthsNarrow:           []string{"", "Y", "Y", "M", "I", "M", "Y", "Y", "Ɣ", "C", "K", "N", "D"},
			monthsWide:             []string{"", "Yennayer", "Yebrayer", "Mars", "Ibrir", "Mayyu", "Yunyu", "Yulyuz", "Ɣuct", "Cutanbir", "Kṭuber", "Nwanbir", "Dujanbir"},
			daysAbbreviated:        []string{"Asa", "Ayn", "Asn", "Akr", "Akw", "Asm", "Asḍ"},
			daysNarrow:             []string{"A", "A", "A", "A", "A", "A", "A"},
			daysWide:               []string{"Asamas", "Aynas", "Asinas", "Akras", "Akwas", "Asimwas", "Asiḍyas"},
			periodsAbbreviated:     []string{"Zdat azal", "Ḍeffir aza"},
			periodsWide:            []string{"Zdat azal", "Ḍeffir aza"},
			erasAbbreviated:        []string{"ZƐ", "ḌƐ"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Zdat Ɛisa (TAƔ)", "Ḍeffir Ɛisa (TAƔ)"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_92e733a5c3e2c8f5728268e0f4b7afd0,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_6dd988d8f1a3fac3718e543933915046,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ug"] = func() locales.Translator {
		return &localen{

			locale:                 "ug",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "￥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"},
			daysAbbreviated:        []string{"يە", "دۈ", "سە", "چا", "پە", "جۈ", "شە"},
			daysNarrow:             []string{"ي", "د", "س", "چ", "پ", "ج", "ش"},
			daysShort:              []string{"ي", "د", "س", "چ", "پ", "ج", "ش"},
			daysWide:               []string{"يەكشەنبە", "دۈشەنبە", "سەيشەنبە", "چارشەنبە", "پەيشەنبە", "جۈمە", "شەنبە"},
			periodsAbbreviated:     []string{"چ.ب", "چ.ك"},
			periodsNarrow:          []string{"ب", "ك"},
			periodsWide:            []string{"چۈشتىن بۇرۇن", "چۈشتىن كېيىن"},
			erasAbbreviated:        []string{"BCE", "مىلادىيە"},
			erasNarrow:             []string{"BCE", "مىلادىيە"},
			erasWide:               []string{"مىلادىيەدىن بۇرۇن", "مىلادىيە"},
			timezones:              map[string]string{"ACDT": "ئاۋسترالىيە ئوتتۇرا قىسىم يازلىق ۋاقتى", "ACST": "ئاۋسترالىيە ئوتتۇرا قىسىم ئۆلچەملىك ۋاقتى", "ACWDT": "ئاۋسترالىيە ئوتتۇرا غەربىي قىسىم يازلىق ۋاقتى", "ACWST": "ئاۋستىرالىيە ئوتتۇرا غەربىي قىسىم ئۆلچەملىك ۋاقتى", "ADT": "ئاتلانتىك ئوكيان يازلىق ۋاقتى", "AEDT": "ئاۋسترالىيە شەرقىي قىسىم يازلىق ۋاقتى", "AEST": "ئاۋسترالىيە شەرقىي قىسىم ئۆلچەملىك ۋاقتى", "AKDT": "ئالياسكا يازلىق ۋاقتى", "AKST": "ئالياسكا ئۆلچەملىك ۋاقتى", "ARST": "ئارگېنتىنا يازلىق ۋاقتى", "ART": "ئارگېنتىنا ئۆلچەملىك ۋاقتى", "AST": "ئاتلانتىك ئوكيان ئۆلچەملىك ۋاقتى", "AWDT": "ئاۋسترالىيە غەربىي قىسىم يازلىق ۋاقتى", "AWST": "ئاۋسترالىيە غەربىي قىسىم ئۆلچەملىك ۋاقتى", "BOT": "بولىۋىيە ۋاقتى", "BT": "بۇتان ۋاقتى", "CAT": "ئوتتۇرا ئافرىقا ۋاقتى", "CDT": "ئوتتۇرا قىسىم يازلىق ۋاقتى", "CHADT": "چاتام يازلىق ۋاقتى", "CHAST": "چاتام ئۆلچەملىك ۋاقتى", "CLST": "چىلى يازلىق ۋاقتى", "CLT": "چىلى ئۆلچەملىك ۋاقتى", "COST": "كولومبىيە يازلىق ۋاقتى", "COT": "كولومبىيە ئۆلچەملىك ۋاقتى", "CST": "ئوتتۇرا قىسىم ئۆلچەملىك ۋاقتى", "ChST": "چاموررو ئۆلچەملىك ۋاقتى", "EAT": "شەرقىي ئافرىقا ۋاقتى", "ECT": "ئېكۋادور ۋاقتى", "EDT": "شەرقىي قىسىم يازلىق ۋاقتى", "EST": "شەرقىي قىسىم ئۆلچەملىك ۋاقتى", "GFT": "فىرانسىيەگە قاراشلىق گىۋىيانا ۋاقتى", "GMT": "گىرىنۋىچ ۋاقتى", "GST": "گۇلف ئۆلچەملىك ۋاقتى", "GYT": "گىۋىيانا ۋاقتى", "HADT": "ھاۋاي-ئالېيۇت يازلىق ۋاقتى", "HAST": "ھاۋاي-ئالېيۇت ئۆلچەملىك ۋاقتى", "HAT": "نىۋفوئۇنلاند يازلىق ۋاقتى", "HECU": "كۇبا يازلىق ۋاقتى", "HEEG": "شەرقىي گىرېنلاند يازلىق ۋاقتى", "HENOMX": "مېكسىكا غەربىي شىمالىي قىسىم يازلىق ۋاقتى", "HEOG": "غەربىي گىرېنلاند يازلىق ۋاقتى", "HEPM": "ساينىت پىئېر ۋە مىكېلون يازلىق ۋاقتى", "HEPMX": "مېكسىكا تىنچ ئوكيان يازلىق ۋاقتى", "HKST": "شياڭگاڭ يازلىق ۋاقتى", "HKT": "شياڭگاڭ ئۆلچەملىك ۋاقتى", "HNCU": "كۇبا ئۆلچەملىك ۋاقتى", "HNEG": "شەرقىي گىرېنلاند ئۆلچەملىك ۋاقتى", "HNNOMX": "مېكسىكا غەربىي شىمالىي قىسىم ئۆلچەملىك ۋاقتى", "HNOG": "غەربىي گىرېنلاند ئۆلچەملىك ۋاقتى", "HNPM": "ساينىت پىئېر ۋە مىكېلون ئۆلچەملىك ۋاقتى", "HNPMX": "مېكسىكا تىنچ ئوكيان ئۆلچەملىك ۋاقتى", "HNT": "نىۋفوئۇنلاند ئۆلچەملىك ۋاقتى", "IST": "ھىندىستان ئۆلچەملىك ۋاقتى", "JDT": "ياپونىيە يازلىق ۋاقتى", "JST": "ياپونىيە ئۆلچەملىك ۋاقتى", "LHDT": "لورد-خاي يازلىق ۋاقتى", "LHST": "لورد-خاي ئۆلچەملىك ۋاقتى", "MDT": "تاغ يازلىق ۋاقتى", "MESZ": "ئوتتۇرا ياۋروپا يازلىق ۋاقتى", "MEZ": "ئوتتۇرا ياۋروپا ئۆلچەملىك ۋاقتى", "MST": "تاغ ئۆلچەملىك ۋاقتى", "MYT": "مالايشىيا ۋاقتى", "NZDT": "يېڭى زېلاندىيە يازلىق ۋاقتى", "NZST": "يېڭى زېلاندىيە ئۆلچەملىك ۋاقتى", "OESZ": "شەرقىي ياۋروپا يازلىق ۋاقتى", "OEZ": "شەرقىي ياۋروپا ئۆلچەملىك ۋاقتى", "PDT": "تىنچ ئوكيان يازلىق ۋاقتى", "PST": "تىنچ ئوكيان ئۆلچەملىك ۋاقتى", "SAST": "جەنۇبىي ئافرىقا ئۆلچەملىك ۋاقتى", "SGT": "سىنگاپور ۋاقتى", "SRT": "سۇرىنام ۋاقتى", "TMST": "تۈركمەنىستان يازلىق ۋاقتى", "TMT": "تۈركمەنىستان ئۆلچەملىك ۋاقتى", "UYST": "ئۇرۇگۋاي يازلىق ۋاقتى", "UYT": "ئۇرۇگۋاي ئۆلچەملىك ۋاقتى", "VET": "ۋېنېزۇئېلا ۋاقتى", "WARST": "غەربىي ئارگېنتىنا يازلىق ۋاقتى", "WART": "غەربىي ئارگېنتىنا ئۆلچەملىك ۋاقتى", "WAST": "غەربىي ئافرىقا يازلىق ۋاقتى", "WAT": "غەربىي ئافرىقا ئۆلچەملىك ۋاقتى", "WESZ": "غەربىي ياۋروپا يازلىق ۋاقتى", "WEZ": "غەربىي ياۋروپا ئۆلچەملىك ۋاقتى", "WIB": "غەربىي ھىندونېزىيە ۋاقتى", "WIT": "شەرقىي ھىندونېزىيە ۋاقتى", "WITA": "ئوتتۇرا ھىندونېزىيە ۋاقتى", "∅∅∅": "بىرازىلىيە يازلىق ۋاقتى"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_4e14db85a2854023613be0580e048527,
			fnFmtDateLong:         fnFmtDateLong_c5ff4169cfa62bfc0e853fa211ecf219,
			fnFmtDateMedium:       fnFmtDateMedium_bdd5fb2012d879143590e275c7d8ebb5,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ug_cn"] = func() locales.Translator {
		return &localen{

			locale:                 "ug_CN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"},
			daysAbbreviated:        []string{"يە", "دۈ", "سە", "چا", "پە", "جۈ", "شە"},
			daysNarrow:             []string{"ي", "د", "س", "چ", "پ", "ج", "ش"},
			daysShort:              []string{"ي", "د", "س", "چ", "پ", "ج", "ش"},
			daysWide:               []string{"يەكشەنبە", "دۈشەنبە", "سەيشەنبە", "چارشەنبە", "پەيشەنبە", "جۈمە", "شەنبە"},
			periodsAbbreviated:     []string{"چ.ب", "چ.ك"},
			periodsNarrow:          []string{"ب", "ك"},
			periodsWide:            []string{"چۈشتىن بۇرۇن", "چۈشتىن كېيىن"},
			erasAbbreviated:        []string{"BCE", "مىلادىيە"},
			erasNarrow:             []string{"BCE", "مىلادىيە"},
			erasWide:               []string{"مىلادىيەدىن بۇرۇن", "مىلادىيە"},
			timezones:              map[string]string{"ACDT": "ئاۋسترالىيە ئوتتۇرا قىسىم يازلىق ۋاقتى", "ACST": "ئاۋسترالىيە ئوتتۇرا قىسىم ئۆلچەملىك ۋاقتى", "ACWDT": "ئاۋسترالىيە ئوتتۇرا غەربىي قىسىم يازلىق ۋاقتى", "ACWST": "ئاۋستىرالىيە ئوتتۇرا غەربىي قىسىم ئۆلچەملىك ۋاقتى", "ADT": "ئاتلانتىك ئوكيان يازلىق ۋاقتى", "AEDT": "ئاۋسترالىيە شەرقىي قىسىم يازلىق ۋاقتى", "AEST": "ئاۋسترالىيە شەرقىي قىسىم ئۆلچەملىك ۋاقتى", "AKDT": "ئالياسكا يازلىق ۋاقتى", "AKST": "ئالياسكا ئۆلچەملىك ۋاقتى", "ARST": "ئارگېنتىنا يازلىق ۋاقتى", "ART": "ئارگېنتىنا ئۆلچەملىك ۋاقتى", "AST": "ئاتلانتىك ئوكيان ئۆلچەملىك ۋاقتى", "AWDT": "ئاۋسترالىيە غەربىي قىسىم يازلىق ۋاقتى", "AWST": "ئاۋسترالىيە غەربىي قىسىم ئۆلچەملىك ۋاقتى", "BOT": "بولىۋىيە ۋاقتى", "BT": "بۇتان ۋاقتى", "CAT": "ئوتتۇرا ئافرىقا ۋاقتى", "CDT": "ئوتتۇرا قىسىم يازلىق ۋاقتى", "CHADT": "چاتام يازلىق ۋاقتى", "CHAST": "چاتام ئۆلچەملىك ۋاقتى", "CLST": "چىلى يازلىق ۋاقتى", "CLT": "چىلى ئۆلچەملىك ۋاقتى", "COST": "كولومبىيە يازلىق ۋاقتى", "COT": "كولومبىيە ئۆلچەملىك ۋاقتى", "CST": "ئوتتۇرا قىسىم ئۆلچەملىك ۋاقتى", "ChST": "چاموررو ئۆلچەملىك ۋاقتى", "EAT": "شەرقىي ئافرىقا ۋاقتى", "ECT": "ئېكۋادور ۋاقتى", "EDT": "شەرقىي قىسىم يازلىق ۋاقتى", "EST": "شەرقىي قىسىم ئۆلچەملىك ۋاقتى", "GFT": "فىرانسىيەگە قاراشلىق گىۋىيانا ۋاقتى", "GMT": "گىرىنۋىچ ۋاقتى", "GST": "گۇلف ئۆلچەملىك ۋاقتى", "GYT": "گىۋىيانا ۋاقتى", "HADT": "ھاۋاي-ئالېيۇت يازلىق ۋاقتى", "HAST": "ھاۋاي-ئالېيۇت ئۆلچەملىك ۋاقتى", "HAT": "نىۋفوئۇنلاند يازلىق ۋاقتى", "HECU": "كۇبا يازلىق ۋاقتى", "HEEG": "شەرقىي گىرېنلاند يازلىق ۋاقتى", "HENOMX": "مېكسىكا غەربىي شىمالىي قىسىم يازلىق ۋاقتى", "HEOG": "غەربىي گىرېنلاند يازلىق ۋاقتى", "HEPM": "ساينىت پىئېر ۋە مىكېلون يازلىق ۋاقتى", "HEPMX": "مېكسىكا تىنچ ئوكيان يازلىق ۋاقتى", "HKST": "شياڭگاڭ يازلىق ۋاقتى", "HKT": "شياڭگاڭ ئۆلچەملىك ۋاقتى", "HNCU": "كۇبا ئۆلچەملىك ۋاقتى", "HNEG": "شەرقىي گىرېنلاند ئۆلچەملىك ۋاقتى", "HNNOMX": "مېكسىكا غەربىي شىمالىي قىسىم ئۆلچەملىك ۋاقتى", "HNOG": "غەربىي گىرېنلاند ئۆلچەملىك ۋاقتى", "HNPM": "ساينىت پىئېر ۋە مىكېلون ئۆلچەملىك ۋاقتى", "HNPMX": "مېكسىكا تىنچ ئوكيان ئۆلچەملىك ۋاقتى", "HNT": "نىۋفوئۇنلاند ئۆلچەملىك ۋاقتى", "IST": "ھىندىستان ئۆلچەملىك ۋاقتى", "JDT": "ياپونىيە يازلىق ۋاقتى", "JST": "ياپونىيە ئۆلچەملىك ۋاقتى", "LHDT": "لورد-خاي يازلىق ۋاقتى", "LHST": "لورد-خاي ئۆلچەملىك ۋاقتى", "MDT": "تاغ يازلىق ۋاقتى", "MESZ": "ئوتتۇرا ياۋروپا يازلىق ۋاقتى", "MEZ": "ئوتتۇرا ياۋروپا ئۆلچەملىك ۋاقتى", "MST": "تاغ ئۆلچەملىك ۋاقتى", "MYT": "مالايشىيا ۋاقتى", "NZDT": "يېڭى زېلاندىيە يازلىق ۋاقتى", "NZST": "يېڭى زېلاندىيە ئۆلچەملىك ۋاقتى", "OESZ": "شەرقىي ياۋروپا يازلىق ۋاقتى", "OEZ": "شەرقىي ياۋروپا ئۆلچەملىك ۋاقتى", "PDT": "تىنچ ئوكيان يازلىق ۋاقتى", "PST": "تىنچ ئوكيان ئۆلچەملىك ۋاقتى", "SAST": "جەنۇبىي ئافرىقا ئۆلچەملىك ۋاقتى", "SGT": "سىنگاپور ۋاقتى", "SRT": "سۇرىنام ۋاقتى", "TMST": "تۈركمەنىستان يازلىق ۋاقتى", "TMT": "تۈركمەنىستان ئۆلچەملىك ۋاقتى", "UYST": "ئۇرۇگۋاي يازلىق ۋاقتى", "UYT": "ئۇرۇگۋاي ئۆلچەملىك ۋاقتى", "VET": "ۋېنېزۇئېلا ۋاقتى", "WARST": "غەربىي ئارگېنتىنا يازلىق ۋاقتى", "WART": "غەربىي ئارگېنتىنا ئۆلچەملىك ۋاقتى", "WAST": "غەربىي ئافرىقا يازلىق ۋاقتى", "WAT": "غەربىي ئافرىقا ئۆلچەملىك ۋاقتى", "WESZ": "غەربىي ياۋروپا يازلىق ۋاقتى", "WEZ": "غەربىي ياۋروپا ئۆلچەملىك ۋاقتى", "WIB": "غەربىي ھىندونېزىيە ۋاقتى", "WIT": "شەرقىي ھىندونېزىيە ۋاقتى", "WITA": "ئوتتۇرا ھىندونېزىيە ۋاقتى", "∅∅∅": "بىرازىلىيە يازلىق ۋاقتى"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_4e14db85a2854023613be0580e048527,
			fnFmtDateLong:         fnFmtDateLong_c5ff4169cfa62bfc0e853fa211ecf219,
			fnFmtDateMedium:       fnFmtDateMedium_bdd5fb2012d879143590e275c7d8ebb5,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uk"] = func() locales.Translator {
		return &localen{

			locale:                 "uk",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{4, 6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "₴", "крб.", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "січ.", "лют.", "бер.", "квіт.", "трав.", "черв.", "лип.", "серп.", "вер.", "жовт.", "лист.", "груд."},
			monthsNarrow:           []string{"", "с", "л", "б", "к", "т", "ч", "л", "с", "в", "ж", "л", "г"},
			monthsWide:             []string{"", "січня", "лютого", "березня", "квітня", "травня", "червня", "липня", "серпня", "вересня", "жовтня", "листопада", "грудня"},
			daysAbbreviated:        []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"Н", "П", "В", "С", "Ч", "П", "С"},
			daysShort:              []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"неділя", "понеділок", "вівторок", "середа", "четвер", "пʼятниця", "субота"},
			periodsAbbreviated:     []string{"дп", "пп"},
			periodsNarrow:          []string{"дп", "пп"},
			periodsWide:            []string{"дп", "пп"},
			erasAbbreviated:        []string{"до н. е.", "н. е."},
			erasNarrow:             []string{"до н.е.", "н.е."},
			erasWide:               []string{"до нашої ери", "нашої ери"},
			timezones:              map[string]string{"ACDT": "за літнім центральноавстралійським часом", "ACST": "за стандартним центральноавстралійським часом", "ACWDT": "за літнім центральнозахідним австралійським часом", "ACWST": "за стандартним центральнозахідним австралійським часом", "ADT": "за атлантичним літнім часом", "AEDT": "за літнім східноавстралійським часом", "AEST": "за стандартним східноавстралійським часом", "AKDT": "за літнім часом на Алясці", "AKST": "за стандартним часом на Алясці", "ARST": "за літнім аргентинським часом", "ART": "за стандартним аргентинським часом", "AST": "за атлантичним стандартним часом", "AWDT": "за літнім західноавстралійським часом", "AWST": "за стандартним західноавстралійським часом", "BOT": "за болівійським часом", "BT": "за часом у Бутані", "CAT": "за центральноафриканським часом", "CDT": "за північноамериканським центральним літнім часом", "CHADT": "за літнім часом на архіпелазі Чатем", "CHAST": "за стандартним часом на архіпелазі Чатем", "CLST": "за літнім чилійським часом", "CLT": "за стандартним чилійським часом", "COST": "за літнім колумбійським часом", "COT": "за стандартним колумбійським часом", "CST": "за північноамериканським центральним стандартним часом", "ChST": "за часом на Північних Маріанських островах", "EAT": "за східноафриканським часом", "ECT": "за часом в Еквадорі", "EDT": "за північноамериканським східним літнім часом", "EST": "за північноамериканським східним стандартним часом", "GFT": "за часом Французької Ґвіани", "GMT": "за Ґрінвічем", "GST": "за часом Перської затоки", "GYT": "за часом у Ґаяні", "HADT": "за літнім гавайсько-алеутським часом", "HAST": "за стандартним гавайсько-алеутським часом", "HAT": "за літнім часом у Ньюфаундленд", "HECU": "за літнім часом на Кубі", "HEEG": "за літнім східним часом у Ґренландії", "HENOMX": "за літнім північнозахідним часом у Мексиці", "HEOG": "за літнім західним часом у Ґренландії", "HEPM": "за літнім часом на островах Сен-П’єр і Мікелон", "HEPMX": "за літнім тихоокеанським часом у Мексиці", "HKST": "за літнім часом у Гонконзі", "HKT": "за стандартним часом у Гонконзі", "HNCU": "за стандартним часом на Кубі", "HNEG": "за стандартним східним часом у Ґренландії", "HNNOMX": "за стандартним північнозахідним часом у Мексиці", "HNOG": "за стандартним західним часом у Ґренландії", "HNPM": "за стандартним часом на островах Сен-П’єр і Мікелон", "HNPMX": "за стандартним тихоокеанським часом у Мексиці", "HNT": "за стандартним часом на острові Ньюфаундленд", "IST": "за індійським стандартним часом", "JDT": "за японським літнім часом", "JST": "за японським стандартним часом", "LHDT": "за літнім часом на острові Лорд-Хау", "LHST": "за стандартним часом на острові Лорд-Хау", "MDT": "за північноамериканським гірським літнім часом", "MESZ": "за центральноєвропейським літнім часом", "MEZ": "за центральноєвропейським стандартним часом", "MST": "за північноамериканським гірським стандартним часом", "MYT": "за часом у Малайзії", "NZDT": "за літнім часом у Новій Зеландії", "NZST": "за стандартним часом у Новій Зеландії", "OESZ": "за східноєвропейським літнім часом", "OEZ": "за східноєвропейським стандартним часом", "PDT": "за північноамериканським тихоокеанським літнім часом", "PST": "за північноамериканським тихоокеанським стандартним часом", "SAST": "за південноафриканським часом", "SGT": "за часом у Сінґапурі", "SRT": "за часом у Суринамі", "TMST": "за літнім часом у Туркменістані", "TMT": "за стандартним часом у Туркменістані", "UYST": "за літнім часом в Уруґваї", "UYT": "за стандартним часом в Уруґваї", "VET": "за часом у Венесуелі", "WARST": "за літнім за західноаргентинським часом", "WART": "за стандартним західноаргентинським часом", "WAST": "за західноафриканським літнім часом", "WAT": "за західноафриканським стандартним часом", "WESZ": "за західноєвропейським літнім часом", "WEZ": "за західноєвропейським стандартним часом", "WIB": "за західноіндонезійським часом", "WIT": "за східноіндонезійським часом", "WITA": "за центральноіндонезійським часом", "∅∅∅": "за літнім часом на Амазонці"},

			fnCardinalPluralRule:  fnCardinalPluralRule_a9860e48306c4532a134648f71c7ec2b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_0b6f60c7e7c9a2b75087215c987e94eb,
			fnFmtDateLong:         fnFmtDateLong_a29e6b3227718d09eae068aaf8d4873a,
			fnFmtDateMedium:       fnFmtDateMedium_eab28cd2ee8d9db4a9df96d1f7b01551,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_d372f132a14400bf15cec11e0608103a,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uk_ua"] = func() locales.Translator {
		return &localen{

			locale:                 "uk_UA",
			pluralsCardinal:        []locales.PluralRule{2, 4, 5, 6},
			pluralsOrdinal:         []locales.PluralRule{4, 6},
			pluralsRange:           []locales.PluralRule{2, 4, 5, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "січ.", "лют.", "бер.", "квіт.", "трав.", "черв.", "лип.", "серп.", "вер.", "жовт.", "лист.", "груд."},
			monthsNarrow:           []string{"", "с", "л", "б", "к", "т", "ч", "л", "с", "в", "ж", "л", "г"},
			monthsWide:             []string{"", "січня", "лютого", "березня", "квітня", "травня", "червня", "липня", "серпня", "вересня", "жовтня", "листопада", "грудня"},
			daysAbbreviated:        []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysNarrow:             []string{"Н", "П", "В", "С", "Ч", "П", "С"},
			daysShort:              []string{"нд", "пн", "вт", "ср", "чт", "пт", "сб"},
			daysWide:               []string{"неділя", "понеділок", "вівторок", "середа", "четвер", "пʼятниця", "субота"},
			periodsAbbreviated:     []string{"дп", "пп"},
			periodsNarrow:          []string{"дп", "пп"},
			periodsWide:            []string{"дп", "пп"},
			erasAbbreviated:        []string{"до н. е.", "н. е."},
			erasNarrow:             []string{"до н.е.", "н.е."},
			erasWide:               []string{"до нашої ери", "нашої ери"},
			timezones:              map[string]string{"ACDT": "за літнім центральноавстралійським часом", "ACST": "за стандартним центральноавстралійським часом", "ACWDT": "за літнім центральнозахідним австралійським часом", "ACWST": "за стандартним центральнозахідним австралійським часом", "ADT": "за атлантичним літнім часом", "AEDT": "за літнім східноавстралійським часом", "AEST": "за стандартним східноавстралійським часом", "AKDT": "за літнім часом на Алясці", "AKST": "за стандартним часом на Алясці", "ARST": "за літнім аргентинським часом", "ART": "за стандартним аргентинським часом", "AST": "за атлантичним стандартним часом", "AWDT": "за літнім західноавстралійським часом", "AWST": "за стандартним західноавстралійським часом", "BOT": "за болівійським часом", "BT": "за часом у Бутані", "CAT": "за центральноафриканським часом", "CDT": "за північноамериканським центральним літнім часом", "CHADT": "за літнім часом на архіпелазі Чатем", "CHAST": "за стандартним часом на архіпелазі Чатем", "CLST": "за літнім чилійським часом", "CLT": "за стандартним чилійським часом", "COST": "за літнім колумбійським часом", "COT": "за стандартним колумбійським часом", "CST": "за північноамериканським центральним стандартним часом", "ChST": "за часом на Північних Маріанських островах", "EAT": "за східноафриканським часом", "ECT": "за часом в Еквадорі", "EDT": "за північноамериканським східним літнім часом", "EST": "за північноамериканським східним стандартним часом", "GFT": "за часом Французької Ґвіани", "GMT": "за Ґрінвічем", "GST": "за часом Перської затоки", "GYT": "за часом у Ґаяні", "HADT": "за літнім гавайсько-алеутським часом", "HAST": "за стандартним гавайсько-алеутським часом", "HAT": "за літнім часом у Ньюфаундленд", "HECU": "за літнім часом на Кубі", "HEEG": "за літнім східним часом у Ґренландії", "HENOMX": "за літнім північнозахідним часом у Мексиці", "HEOG": "за літнім західним часом у Ґренландії", "HEPM": "за літнім часом на островах Сен-П’єр і Мікелон", "HEPMX": "за літнім тихоокеанським часом у Мексиці", "HKST": "за літнім часом у Гонконзі", "HKT": "за стандартним часом у Гонконзі", "HNCU": "за стандартним часом на Кубі", "HNEG": "за стандартним східним часом у Ґренландії", "HNNOMX": "за стандартним північнозахідним часом у Мексиці", "HNOG": "за стандартним західним часом у Ґренландії", "HNPM": "за стандартним часом на островах Сен-П’єр і Мікелон", "HNPMX": "за стандартним тихоокеанським часом у Мексиці", "HNT": "за стандартним часом на острові Ньюфаундленд", "IST": "за індійським стандартним часом", "JDT": "за японським літнім часом", "JST": "за японським стандартним часом", "LHDT": "за літнім часом на острові Лорд-Хау", "LHST": "за стандартним часом на острові Лорд-Хау", "MDT": "за північноамериканським гірським літнім часом", "MESZ": "за центральноєвропейським літнім часом", "MEZ": "за центральноєвропейським стандартним часом", "MST": "за північноамериканським гірським стандартним часом", "MYT": "за часом у Малайзії", "NZDT": "за літнім часом у Новій Зеландії", "NZST": "за стандартним часом у Новій Зеландії", "OESZ": "за східноєвропейським літнім часом", "OEZ": "за східноєвропейським стандартним часом", "PDT": "за північноамериканським тихоокеанським літнім часом", "PST": "за північноамериканським тихоокеанським стандартним часом", "SAST": "за південноафриканським часом", "SGT": "за часом у Сінґапурі", "SRT": "за часом у Суринамі", "TMST": "за літнім часом у Туркменістані", "TMT": "за стандартним часом у Туркменістані", "UYST": "за літнім часом в Уруґваї", "UYT": "за стандартним часом в Уруґваї", "VET": "за часом у Венесуелі", "WARST": "за літнім за західноаргентинським часом", "WART": "за стандартним західноаргентинським часом", "WAST": "за західноафриканським літнім часом", "WAT": "за західноафриканським стандартним часом", "WESZ": "за західноєвропейським літнім часом", "WEZ": "за західноєвропейським стандартним часом", "WIB": "за західноіндонезійським часом", "WIT": "за східноіндонезійським часом", "WITA": "за центральноіндонезійським часом", "∅∅∅": "за літнім часом на Амазонці"},

			fnCardinalPluralRule:  fnCardinalPluralRule_97e86d6fc0c03aeac1441d5d07580fac,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_0b6f60c7e7c9a2b75087215c987e94eb,
			fnFmtDateLong:         fnFmtDateLong_a29e6b3227718d09eae068aaf8d4873a,
			fnFmtDateMedium:       fnFmtDateMedium_eab28cd2ee8d9db4a9df96d1f7b01551,
			fnFmtDateShort:        fnFmtDateShort_c475a13000fd4072cf8feb7c7b40c476,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_d372f132a14400bf15cec11e0608103a,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_52e3998ab91a02eab3b89e4e5a856fb5,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ur"] = func() locales.Translator {
		return &localen{

			locale:             "ur",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "‎-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "Rs", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:    []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			daysWide:           []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"a", "p"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"قبل مسیح", "عیسوی"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"قبل مسیح", "عیسوی"},
			timezones:          map[string]string{"ACDT": "آسٹریلین سنٹرل ڈے لائٹ ٹائم", "ACST": "آسٹریلین سنٹرل اسٹینڈرڈ ٹائم", "ACWDT": "آسٹریلین سنٹرل ویسٹرن ڈے لائٹ ٹائم", "ACWST": "آسٹریلین سنٹرل ویسٹرن اسٹینڈرڈ ٹائم", "ADT": "اٹلانٹک ڈے لائٹ ٹائم", "AEDT": "آسٹریلین ایسٹرن ڈے لائٹ ٹائم", "AEST": "آسٹریلین ایسٹرن اسٹینڈرڈ ٹائم", "AKDT": "الاسکا ڈے لائٹ ٹائم", "AKST": "الاسکا اسٹینڈرڈ ٹائم", "ARST": "ارجنٹینا کا موسم گرما کا وقت", "ART": "ارجنٹینا کا معیاری وقت", "AST": "اٹلانٹک اسٹینڈرڈ ٹائم", "AWDT": "آسٹریلین ویسٹرن ڈے لائٹ ٹائم", "AWST": "آسٹریلیا ویسٹرن اسٹینڈرڈ ٹائم", "BOT": "بولیویا کا وقت", "BT": "بھوٹان کا وقت", "CAT": "وسطی افریقہ ٹائم", "CDT": "سنٹرل ڈے لائٹ ٹائم", "CHADT": "چیتھم ڈے لائٹ ٹائم", "CHAST": "چیتھم اسٹینڈرڈ ٹائم", "CLST": "چلی کا موسم گرما کا وقت", "CLT": "چلی کا معیاری وقت", "COST": "کولمبیا کا موسم گرما کا وقت", "COT": "کولمبیا کا معیاری وقت", "CST": "سنٹرل اسٹینڈرڈ ٹائم", "ChST": "چامورو سٹینڈرڈ ٹائم", "EAT": "مشرقی افریقہ ٹائم", "ECT": "ایکواڈور کا وقت", "EDT": "ایسٹرن ڈے لائٹ ٹائم", "EST": "ایسٹرن اسٹینڈرڈ ٹائم", "GFT": "فرینچ گیانا کا وقت", "GMT": "گرین وچ کا اصل وقت", "GST": "خلیج کا معیاری وقت", "GYT": "گیانا کا وقت", "HADT": "ہوائی الیوٹیئن ڈے لائٹ ٹائم", "HAST": "ہوائی الیوٹیئن اسٹینڈرڈ ٹائم", "HAT": "نیو فاؤنڈ لینڈ ڈے لائٹ ٹائم", "HECU": "کیوبا ڈے لائٹ ٹائم", "HEEG": "مشرقی گرین لینڈ کا موسم گرما کا وقت", "HENOMX": "شمال مغربی میکسیکو ڈے لائٹ ٹائم", "HEOG": "مغربی گرین لینڈ کا موسم گرما کا وقت", "HEPM": "سینٹ پیئر اور مکلیئون ڈے لائٹ ٹائم", "HEPMX": "میکسیکن پیسفک ڈے لائٹ ٹائم", "HKST": "ہانگ کانگ سمر ٹائم", "HKT": "ہانگ کانگ سٹینڈرڈ ٹائم", "HNCU": "کیوبا اسٹینڈرڈ ٹائم", "HNEG": "مشرقی گرین لینڈ اسٹینڈرڈ ٹائم", "HNNOMX": "شمال مغربی میکسیکو اسٹینڈرڈ ٹائم", "HNOG": "مغربی گرین لینڈ اسٹینڈرڈ ٹائم", "HNPM": "سینٹ پیئر اور مکلیئون اسٹینڈرڈ ٹائم", "HNPMX": "میکسیکن پیسفک اسٹینڈرڈ ٹائم", "HNT": "نیو فاؤنڈ لینڈ اسٹینڈرڈ ٹائم", "IST": "ہندوستان کا معیاری وقت", "JDT": "جاپان ڈے لائٹ ٹائم", "JST": "جاپان سٹینڈرڈ ٹائم", "LHDT": "لارڈ ہووے ڈے لائٹ ٹائم", "LHST": "لارڈ ہووے اسٹینڈرڈ ٹائم", "MDT": "ماؤنٹین ڈے لائٹ ٹائم", "MESZ": "وسطی یورپ کا موسم گرما کا وقت", "MEZ": "وسطی یورپ کا معیاری وقت", "MST": "ماؤنٹین اسٹینڈرڈ ٹائم", "MYT": "ملیشیا ٹائم", "NZDT": "نیوزی لینڈ ڈے لائٹ ٹائم", "NZST": "نیوزی لینڈ کا معیاری وقت", "OESZ": "مشرقی یورپ کا موسم گرما کا وقت", "OEZ": "مشرقی یورپ کا معیاری وقت", "PDT": "پیسفک ڈے لائٹ ٹائم", "PST": "پیسفک اسٹینڈرڈ ٹائم", "SAST": "جنوبی افریقہ سٹینڈرڈ ٹائم", "SGT": "سنگاپور سٹینڈرڈ ٹائم", "SRT": "سورینام کا وقت", "TMST": "ترکمانستان کا موسم گرما کا وقت", "TMT": "ترکمانستان کا معیاری وقت", "UYST": "یوروگوئے کا موسم گرما کا وقت", "UYT": "یوروگوئے کا معیاری وقت", "VET": "وینزوئیلا کا وقت", "WARST": "مغربی ارجنٹینا کا موسم گرما کا وقت", "WART": "مغربی ارجنٹینا کا معیاری وقت", "WAST": "مغربی افریقہ سمر ٹائم", "WAT": "مغربی افریقہ سٹینڈرڈ ٹائم", "WESZ": "مغربی یورپ کا موسم گرما کا وقت", "WEZ": "مغربی یورپ کا معیاری وقت", "WIB": "مغربی انڈونیشیا ٹائم", "WIT": "مشرقی انڈونیشیا ٹائم", "WITA": "وسطی انڈونیشیا ٹائم", "∅∅∅": "پیرو کا موسم گرما کا وقت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ae41296c273457c904bdda366fe75145,
			fnFmtCurrency:         fnFmtCurrency_a6249ceef016228ba91ed7bf933a5843,
			fnFmtDateFull:         fnFmtDateFull_4d9460f3dccf0c21118fbf081ee2fe75,
			fnFmtDateLong:         fnFmtDateLong_39d6f2c09ffa1cac410fe4d18df66a7c,
			fnFmtDateMedium:       fnFmtDateMedium_82c71fd71a19831d2abc525aff9a3a0e,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_8a34c21e88caccc3a8728bf464f33cbd,
			fnFmtPercent:          fnFmtPercent_d2f2a59417e049449b3bb6986a239e7b,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ur_in"] = func() locales.Translator {
		return &localen{

			locale:                 "ur_IN",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "‎-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			daysWide:               []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"قبل مسیح", "عیسوی"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"قبل مسیح", "عیسوی"},
			timezones:              map[string]string{"ACDT": "آسٹریلین سنٹرل ڈے لائٹ ٹائم", "ACST": "آسٹریلین سنٹرل اسٹینڈرڈ ٹائم", "ACWDT": "آسٹریلین سنٹرل ویسٹرن ڈے لائٹ ٹائم", "ACWST": "آسٹریلین سنٹرل ویسٹرن اسٹینڈرڈ ٹائم", "ADT": "اٹلانٹک ڈے لائٹ ٹائم", "AEDT": "آسٹریلین ایسٹرن ڈے لائٹ ٹائم", "AEST": "آسٹریلین ایسٹرن اسٹینڈرڈ ٹائم", "AKDT": "الاسکا ڈے لائٹ ٹائم", "AKST": "الاسکا اسٹینڈرڈ ٹائم", "ARST": "ارجنٹینا کا موسم گرما کا وقت", "ART": "ارجنٹینا کا معیاری وقت", "AST": "اٹلانٹک اسٹینڈرڈ ٹائم", "AWDT": "آسٹریلین ویسٹرن ڈے لائٹ ٹائم", "AWST": "آسٹریلیا ویسٹرن اسٹینڈرڈ ٹائم", "BOT": "بولیویا ٹائم", "BT": "بھوٹان ٹائم", "CAT": "وسطی افریقہ ٹائم", "CDT": "سنٹرل ڈے لائٹ ٹائم", "CHADT": "چیتھم ڈے لائٹ ٹائم", "CHAST": "چیتھم اسٹینڈرڈ ٹائم", "CLST": "چلی سمر ٹائم", "CLT": "چلی سٹینڈرڈ ٹائم", "COST": "کولمبیا سمر ٹائم", "COT": "کولمبیا سٹینڈرڈ ٹائم", "CST": "سنٹرل اسٹینڈرڈ ٹائم", "ChST": "چامورو سٹینڈرڈ ٹائم", "EAT": "مشرقی افریقہ ٹائم", "ECT": "ایکواڈور ٹائم", "EDT": "ایسٹرن ڈے لائٹ ٹائم", "EST": "ایسٹرن اسٹینڈرڈ ٹائم", "GFT": "فرینچ گیانا ٹائم", "GMT": "گرین وچ مین ٹائم", "GST": "خلیج سٹینڈرڈ ٹائم", "GYT": "گیانا ٹائم", "HADT": "ہوائی الیوٹیئن ڈے لائٹ ٹائم", "HAST": "ہوائی الیوٹیئن اسٹینڈرڈ ٹائم", "HAT": "نیو فاؤنڈ لینڈ ڈے لائٹ ٹائم", "HECU": "کیوبا ڈے لائٹ ٹائم", "HEEG": "مشرقی گرین لینڈ کا موسم گرما کا وقت", "HENOMX": "شمال مغربی میکسیکو ڈے لائٹ ٹائم", "HEOG": "مغربی گرین لینڈ کا موسم گرما کا وقت", "HEPM": "سینٹ پیئر اور مکلیئون ڈے لائٹ ٹائم", "HEPMX": "میکسیکن پیسفک ڈے لائٹ ٹائم", "HKST": "ہانگ کانگ سمر ٹائم", "HKT": "ہانگ کانگ سٹینڈرڈ ٹائم", "HNCU": "کیوبا اسٹینڈرڈ ٹائم", "HNEG": "مشرقی گرین لینڈ اسٹینڈرڈ ٹائم", "HNNOMX": "شمال مغربی میکسیکو اسٹینڈرڈ ٹائم", "HNOG": "مغربی گرین لینڈ اسٹینڈرڈ ٹائم", "HNPM": "سینٹ پیئر اور مکلیئون اسٹینڈرڈ ٹائم", "HNPMX": "میکسیکن پیسفک اسٹینڈرڈ ٹائم", "HNT": "نیو فاؤنڈ لینڈ اسٹینڈرڈ ٹائم", "IST": "انڈیا سٹینڈرڈ ٹائم", "JDT": "جاپان ڈے لائٹ ٹائم", "JST": "جاپان سٹینڈرڈ ٹائم", "LHDT": "لارڈ ہووے ڈے لائٹ ٹائم", "LHST": "لارڈ ہووے اسٹینڈرڈ ٹائم", "MDT": "ماؤنٹین ڈے لائٹ ٹائم", "MESZ": "وسطی یورپ کا موسم گرما کا وقت", "MEZ": "وسطی یورپ کا معیاری وقت", "MST": "ماؤنٹین اسٹینڈرڈ ٹائم", "MYT": "ملیشیا ٹائم", "NZDT": "نیوزی لینڈ ڈے لائٹ ٹائم", "NZST": "نیوزی لینڈ سٹینڈرڈ ٹائم", "OESZ": "مشرقی یورپ کا موسم گرما کا وقت", "OEZ": "مشرقی یورپ کا معیاری وقت", "PDT": "پیسفک ڈے لائٹ ٹائم", "PST": "پیسفک اسٹینڈرڈ ٹائم", "SAST": "جنوبی افریقہ سٹینڈرڈ ٹائم", "SGT": "سنگاپور سٹینڈرڈ ٹائم", "SRT": "سورینام ٹائم", "TMST": "ترکمانستان سمر ٹائم", "TMT": "ترکمانستان سٹینڈرڈ ٹائم", "UYST": "یوروگوئے سمر ٹائم", "UYT": "یوروگوئے سٹینڈرڈ ٹائم", "VET": "وینزوئیلا ٹائم", "WARST": "مغربی ارجنٹینا سمر ٹائم", "WART": "مغربی ارجنٹینا سٹینڈرڈ ٹائم", "WAST": "مغربی افریقہ سمر ٹائم", "WAT": "مغربی افریقہ سٹینڈرڈ ٹائم", "WESZ": "مغربی یورپ کا موسم گرما کا وقت", "WEZ": "مغربی یورپ کا معیاری وقت", "WIB": "مغربی انڈونیشیا ٹائم", "WIT": "مشرقی انڈونیشیا ٹائم", "WITA": "وسطی انڈونیشیا ٹائم", "∅∅∅": "پیرو سمر ٹائم"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_838c6cb50ab0f7d4efc4d33ce55bccb5,
			fnFmtCurrency:         fnFmtCurrency_c079b862a25ea3e09b2b82e454b35d84,
			fnFmtDateFull:         fnFmtDateFull_4d9460f3dccf0c21118fbf081ee2fe75,
			fnFmtDateLong:         fnFmtDateLong_39d6f2c09ffa1cac410fe4d18df66a7c,
			fnFmtDateMedium:       fnFmtDateMedium_82c71fd71a19831d2abc525aff9a3a0e,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_8a34c21e88caccc3a8728bf464f33cbd,
			fnFmtPercent:          fnFmtPercent_d2f2a59417e049449b3bb6986a239e7b,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["ur_pk"] = func() locales.Translator {
		return &localen{

			locale:             "ur_PK",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "‎-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:    []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			daysWide:           []string{"اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "ہفتہ"},
			periodsAbbreviated: []string{"AM", "PM"},
			periodsNarrow:      []string{"a", "p"},
			periodsWide:        []string{"AM", "PM"},
			erasAbbreviated:    []string{"قبل مسیح", "عیسوی"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"قبل مسیح", "عیسوی"},
			timezones:          map[string]string{"ACDT": "آسٹریلین سنٹرل ڈے لائٹ ٹائم", "ACST": "آسٹریلین سنٹرل اسٹینڈرڈ ٹائم", "ACWDT": "آسٹریلین سنٹرل ویسٹرن ڈے لائٹ ٹائم", "ACWST": "آسٹریلین سنٹرل ویسٹرن اسٹینڈرڈ ٹائم", "ADT": "اٹلانٹک ڈے لائٹ ٹائم", "AEDT": "آسٹریلین ایسٹرن ڈے لائٹ ٹائم", "AEST": "آسٹریلین ایسٹرن اسٹینڈرڈ ٹائم", "AKDT": "الاسکا ڈے لائٹ ٹائم", "AKST": "الاسکا اسٹینڈرڈ ٹائم", "ARST": "ارجنٹینا کا موسم گرما کا وقت", "ART": "ارجنٹینا کا معیاری وقت", "AST": "اٹلانٹک اسٹینڈرڈ ٹائم", "AWDT": "آسٹریلین ویسٹرن ڈے لائٹ ٹائم", "AWST": "آسٹریلیا ویسٹرن اسٹینڈرڈ ٹائم", "BOT": "بولیویا کا وقت", "BT": "بھوٹان کا وقت", "CAT": "وسطی افریقہ ٹائم", "CDT": "سنٹرل ڈے لائٹ ٹائم", "CHADT": "چیتھم ڈے لائٹ ٹائم", "CHAST": "چیتھم اسٹینڈرڈ ٹائم", "CLST": "چلی کا موسم گرما کا وقت", "CLT": "چلی کا معیاری وقت", "COST": "کولمبیا کا موسم گرما کا وقت", "COT": "کولمبیا کا معیاری وقت", "CST": "سنٹرل اسٹینڈرڈ ٹائم", "ChST": "چامورو سٹینڈرڈ ٹائم", "EAT": "مشرقی افریقہ ٹائم", "ECT": "ایکواڈور کا وقت", "EDT": "ایسٹرن ڈے لائٹ ٹائم", "EST": "ایسٹرن اسٹینڈرڈ ٹائم", "GFT": "فرینچ گیانا کا وقت", "GMT": "گرین وچ کا اصل وقت", "GST": "خلیج کا معیاری وقت", "GYT": "گیانا کا وقت", "HADT": "ہوائی الیوٹیئن ڈے لائٹ ٹائم", "HAST": "ہوائی الیوٹیئن اسٹینڈرڈ ٹائم", "HAT": "نیو فاؤنڈ لینڈ ڈے لائٹ ٹائم", "HECU": "کیوبا ڈے لائٹ ٹائم", "HEEG": "مشرقی گرین لینڈ کا موسم گرما کا وقت", "HENOMX": "شمال مغربی میکسیکو ڈے لائٹ ٹائم", "HEOG": "مغربی گرین لینڈ کا موسم گرما کا وقت", "HEPM": "سینٹ پیئر اور مکلیئون ڈے لائٹ ٹائم", "HEPMX": "میکسیکن پیسفک ڈے لائٹ ٹائم", "HKST": "ہانگ کانگ سمر ٹائم", "HKT": "ہانگ کانگ سٹینڈرڈ ٹائم", "HNCU": "کیوبا اسٹینڈرڈ ٹائم", "HNEG": "مشرقی گرین لینڈ اسٹینڈرڈ ٹائم", "HNNOMX": "شمال مغربی میکسیکو اسٹینڈرڈ ٹائم", "HNOG": "مغربی گرین لینڈ اسٹینڈرڈ ٹائم", "HNPM": "سینٹ پیئر اور مکلیئون اسٹینڈرڈ ٹائم", "HNPMX": "میکسیکن پیسفک اسٹینڈرڈ ٹائم", "HNT": "نیو فاؤنڈ لینڈ اسٹینڈرڈ ٹائم", "IST": "ہندوستان کا معیاری وقت", "JDT": "جاپان ڈے لائٹ ٹائم", "JST": "جاپان سٹینڈرڈ ٹائم", "LHDT": "لارڈ ہووے ڈے لائٹ ٹائم", "LHST": "لارڈ ہووے اسٹینڈرڈ ٹائم", "MDT": "ماؤنٹین ڈے لائٹ ٹائم", "MESZ": "وسطی یورپ کا موسم گرما کا وقت", "MEZ": "وسطی یورپ کا معیاری وقت", "MST": "ماؤنٹین اسٹینڈرڈ ٹائم", "MYT": "ملیشیا ٹائم", "NZDT": "نیوزی لینڈ ڈے لائٹ ٹائم", "NZST": "نیوزی لینڈ کا معیاری وقت", "OESZ": "مشرقی یورپ کا موسم گرما کا وقت", "OEZ": "مشرقی یورپ کا معیاری وقت", "PDT": "پیسفک ڈے لائٹ ٹائم", "PST": "پیسفک اسٹینڈرڈ ٹائم", "SAST": "جنوبی افریقہ سٹینڈرڈ ٹائم", "SGT": "سنگاپور سٹینڈرڈ ٹائم", "SRT": "سورینام کا وقت", "TMST": "ترکمانستان کا موسم گرما کا وقت", "TMT": "ترکمانستان کا معیاری وقت", "UYST": "یوروگوئے کا موسم گرما کا وقت", "UYT": "یوروگوئے کا معیاری وقت", "VET": "وینزوئیلا کا وقت", "WARST": "مغربی ارجنٹینا کا موسم گرما کا وقت", "WART": "مغربی ارجنٹینا کا معیاری وقت", "WAST": "مغربی افریقہ سمر ٹائم", "WAT": "مغربی افریقہ سٹینڈرڈ ٹائم", "WESZ": "مغربی یورپ کا موسم گرما کا وقت", "WEZ": "مغربی یورپ کا معیاری وقت", "WIB": "مغربی انڈونیشیا ٹائم", "WIT": "مشرقی انڈونیشیا ٹائم", "WITA": "وسطی انڈونیشیا ٹائم", "∅∅∅": "ازوریس کا موسم گرما کا وقت"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_ae41296c273457c904bdda366fe75145,
			fnFmtCurrency:         fnFmtCurrency_a6249ceef016228ba91ed7bf933a5843,
			fnFmtDateFull:         fnFmtDateFull_4d9460f3dccf0c21118fbf081ee2fe75,
			fnFmtDateLong:         fnFmtDateLong_39d6f2c09ffa1cac410fe4d18df66a7c,
			fnFmtDateMedium:       fnFmtDateMedium_82c71fd71a19831d2abc525aff9a3a0e,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_8a34c21e88caccc3a8728bf464f33cbd,
			fnFmtPercent:          fnFmtPercent_d2f2a59417e049449b3bb6986a239e7b,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz"] = func() locales.Translator {
		return &localen{

			locale:                 "uz",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "soʻm", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avg", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "Y", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avgust", "sentabr", "oktabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"Yak", "Dush", "Sesh", "Chor", "Pay", "Jum", "Shan"},
			daysNarrow:             []string{"Y", "D", "S", "C", "P", "J", "S"},
			daysShort:              []string{"Ya", "Du", "Se", "Ch", "Pa", "Ju", "Sh"},
			daysWide:               []string{"yakshanba", "dushanba", "seshanba", "chorshanba", "payshanba", "juma", "shanba"},
			periodsAbbreviated:     []string{"TO", "TK"},
			periodsNarrow:          []string{"TO", "TK"},
			periodsWide:            []string{"TO", "TK"},
			erasAbbreviated:        []string{"m.a.", "milodiy"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"miloddan avvalgi", "milodiy"},
			timezones:              map[string]string{"ACDT": "Markaziy Avstraliya yozgi vaqti", "ACST": "Markaziy Avstraliya standart vaqti", "ACWDT": "Markaziy Avstraliya g‘arbiy yozgi vaqti", "ACWST": "Markaziy Avstraliya g‘arbiy standart vaqti", "ADT": "Atlantika yozgi vaqti", "AEDT": "Sharqiy Avstraliya yozgi vaqti", "AEST": "Sharqiy Avstraliya standart vaqti", "AKDT": "Alyaska yozgi vaqti", "AKST": "Alyaska standart vaqti", "ARST": "Argentina yozgi vaqti", "ART": "Argentina standart vaqti", "AST": "Atlantika standart vaqti", "AWDT": "G‘arbiy Avstraliya yozgi vaqti", "AWST": "G‘arbiy Avstraliya standart vaqti", "BOT": "Boliviya vaqti", "BT": "Butan vaqti", "CAT": "Markaziy Afrika vaqti", "CDT": "Markaziy Amerika yozgi vaqti", "CHADT": "Chatem yozgi vaqti", "CHAST": "Chatem standart vaqti", "CLST": "Chili yozgi vaqti", "CLT": "Chili standart vaqti", "COST": "Kolumbiya yozgi vaqti", "COT": "Kolumbiya standart vaqti", "CST": "Markaziy Amerika standart vaqti", "ChST": "Chamorro standart vaqti", "EAT": "Sharqiy Afrika vaqti", "ECT": "Ekvador vaqti", "EDT": "Sharqiy Amerika yozgi vaqti", "EST": "Sharqiy Amerika standart vaqti", "GFT": "Fransuz Gvianasi vaqti", "GMT": "Grinvich o‘rtacha vaqti", "GST": "Fors ko‘rfazi standart vaqti", "GYT": "Gayana vaqti", "HADT": "Gavayi-aleut yozgi vaqti", "HAST": "Gavayi-aleut standart vaqti", "HAT": "Nyufaundlend yozgi vaqti", "HECU": "Kuba yozgi vaqti", "HEEG": "Sharqiy Grenlandiya yozgi vaqti", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "G‘arbiy Grenlandiya yozgi vaqti", "HEPM": "Sen-Pyer va Mikelon yozgi vaqti", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Gonkong yozgi vaqti", "HKT": "Gonkong standart vaqti", "HNCU": "Kuba standart vaqti", "HNEG": "Sharqiy Grenlandiya standart vaqti", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "G‘arbiy Grenlandiya standart vaqti", "HNPM": "Sen-Pyer va Mikelon standart vaqti", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Nyufaundlend standart vaqti", "IST": "Hindiston standart vaqti", "JDT": "Yaponiya yozgi vaqti", "JST": "Yaponiya standart vaqti", "LHDT": "Lord-Xau yozgi vaqti", "LHST": "Lord-Xau standart vaqti", "MDT": "Tog‘ yozgi vaqti (AQSH)", "MESZ": "Markaziy Yevropa yozgi vaqti", "MEZ": "Markaziy Yevropa standart vaqti", "MST": "Tog‘ standart vaqti (AQSH)", "MYT": "Malayziya vaqti", "NZDT": "Yangi Zelandiya yozgi vaqti", "NZST": "Yangi Zelandiya standart vaqti", "OESZ": "Sharqiy Yevropa yozgi vaqti", "OEZ": "Sharqiy Yevropa standart vaqti", "PDT": "Tinch okeani yozgi vaqti", "PST": "Tinch okeani standart vaqti", "SAST": "Janubiy Afrika standart vaqti", "SGT": "Singapur vaqti", "SRT": "Surinam vaqti", "TMST": "Turkmaniston yozgi vaqti", "TMT": "Turkmaniston standart vaqti", "UYST": "Urugvay yozgi vaqti", "UYT": "Urugvay standart vaqti", "VET": "Venesuela vaqti", "WARST": "Gʻarbiy Argentina yozgi vaqti", "WART": "Gʻarbiy Argentina standart vaqti", "WAST": "Gʻarbiy Afrika yozgi vaqti", "WAT": "Gʻarbiy Afrika standart vaqti", "WESZ": "G‘arbiy Yevropa yozgi vaqti", "WEZ": "G‘arbiy Yevropa standart vaqti", "WIB": "Gʻarbiy Indoneziya vaqti", "WIT": "Sharqiy Indoneziya vaqti", "WITA": "Markaziy Indoneziya vaqti", "∅∅∅": "Braziliya yozgi vaqti"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_ae39dfe33f097d30e514b77409da8997,
			fnFmtDateLong:         fnFmtDateLong_af1ab393e97e5cd9bd5342a65130666d,
			fnFmtDateMedium:       fnFmtDateMedium_8ff89c88bdb727333d4c7b42f495646e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz_arab"] = func() locales.Translator {
		return &localen{

			locale:                 "uz_Arab",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "-",
			percent:                "٪",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "؋", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "جنو", "فبر", "مار", "اپر", "می", "جون", "جول", "اگس", "سپت", "اکت", "نوم", "دسم"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "جنوری", "فبروری", "مارچ", "اپریل", "می", "جون", "جولای", "اگست", "سپتمبر", "اکتوبر", "نومبر", "دسمبر"},
			daysAbbreviated:        []string{"ی.", "د.", "س.", "چ.", "پ.", "ج.", "ش."},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:              []string{"Ya", "Du", "Se", "Ch", "Pa", "Ju", "Sh"},
			daysWide:               []string{"یکشنبه", "دوشنبه", "سه\u200cشنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"ق.م.", "م."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "Markaziy Avstraliya yozgi vaqti", "ACST": "Markaziy Avstraliya standart vaqti", "ACWDT": "Markaziy Avstraliya g‘arbiy yozgi vaqti", "ACWST": "Markaziy Avstraliya g‘arbiy standart vaqti", "ADT": "Atlantika yozgi vaqti", "AEDT": "Sharqiy Avstraliya yozgi vaqti", "AEST": "Sharqiy Avstraliya standart vaqti", "AKDT": "Alyaska yozgi vaqti", "AKST": "Alyaska standart vaqti", "ARST": "Argentina yozgi vaqti", "ART": "Argentina standart vaqti", "AST": "Atlantika standart vaqti", "AWDT": "G‘arbiy Avstraliya yozgi vaqti", "AWST": "G‘arbiy Avstraliya standart vaqti", "BOT": "Boliviya vaqti", "BT": "Butan vaqti", "CAT": "Markaziy Afrika vaqti", "CDT": "Markaziy Amerika yozgi vaqti", "CHADT": "Chatem yozgi vaqti", "CHAST": "Chatem standart vaqti", "CLST": "Chili yozgi vaqti", "CLT": "Chili standart vaqti", "COST": "Kolumbiya yozgi vaqti", "COT": "Kolumbiya standart vaqti", "CST": "Markaziy Amerika standart vaqti", "ChST": "Chamorro standart vaqti", "EAT": "Sharqiy Afrika vaqti", "ECT": "Ekvador vaqti", "EDT": "Sharqiy Amerika yozgi vaqti", "EST": "Sharqiy Amerika standart vaqti", "GFT": "Fransuz Gvianasi vaqti", "GMT": "Grinvich o‘rtacha vaqti", "GST": "Fors ko‘rfazi standart vaqti", "GYT": "Gayana vaqti", "HADT": "Gavayi-aleut yozgi vaqti", "HAST": "Gavayi-aleut standart vaqti", "HAT": "Nyufaundlend yozgi vaqti", "HECU": "Kuba yozgi vaqti", "HEEG": "Sharqiy Grenlandiya yozgi vaqti", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "G‘arbiy Grenlandiya yozgi vaqti", "HEPM": "Sen-Pyer va Mikelon yozgi vaqti", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Gonkong yozgi vaqti", "HKT": "Gonkong standart vaqti", "HNCU": "Kuba standart vaqti", "HNEG": "Sharqiy Grenlandiya standart vaqti", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "G‘arbiy Grenlandiya standart vaqti", "HNPM": "Sen-Pyer va Mikelon standart vaqti", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Nyufaundlend standart vaqti", "IST": "Hindiston standart vaqti", "JDT": "Yaponiya yozgi vaqti", "JST": "Yaponiya standart vaqti", "LHDT": "Lord-Xau yozgi vaqti", "LHST": "Lord-Xau standart vaqti", "MDT": "Tog‘ yozgi vaqti (AQSH)", "MESZ": "Markaziy Yevropa yozgi vaqti", "MEZ": "Markaziy Yevropa standart vaqti", "MST": "Tog‘ standart vaqti (AQSH)", "MYT": "Malayziya vaqti", "NZDT": "Yangi Zelandiya yozgi vaqti", "NZST": "Yangi Zelandiya standart vaqti", "OESZ": "Sharqiy Yevropa yozgi vaqti", "OEZ": "Sharqiy Yevropa standart vaqti", "PDT": "Tinch okeani yozgi vaqti", "PST": "Tinch okeani standart vaqti", "SAST": "Janubiy Afrika standart vaqti", "SGT": "Singapur vaqti", "SRT": "Surinam vaqti", "TMST": "Turkmaniston yozgi vaqti", "TMT": "Turkmaniston standart vaqti", "UYST": "Urugvay yozgi vaqti", "UYT": "Urugvay standart vaqti", "VET": "Venesuela vaqti", "WARST": "Gʻarbiy Argentina yozgi vaqti", "WART": "Gʻarbiy Argentina standart vaqti", "WAST": "Gʻarbiy Afrika yozgi vaqti", "WAT": "Gʻarbiy Afrika standart vaqti", "WESZ": "G‘arbiy Yevropa yozgi vaqti", "WEZ": "G‘arbiy Yevropa standart vaqti", "WIB": "Gʻarbiy Indoneziya vaqti", "WIT": "Sharqiy Indoneziya vaqti", "WITA": "Markaziy Indoneziya vaqti", "∅∅∅": "Azor orollari yozgi vaqti"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c2fb1e60c1c0fac4b7a6ab96cbc4f79e,
			fnFmtCurrency:         fnFmtCurrency_ae69fea052716ee7e5be5d84e205ff0c,
			fnFmtDateFull:         fnFmtDateFull_64ae09e94f136aa33d4d0e5627a29872,
			fnFmtDateLong:         fnFmtDateLong_3097d2288f3d500e25f72a56af6d0f39,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_91f44359c32923b2c2a980df8ee6ab86,
			fnFmtPercent:          fnFmtPercent_141d2f250af9b9f46295bf60380ce0b2,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_c3e2f72b9dd7b60e99ec7ed62baf96a4,
			fnFmtTimeShort:        fnFmtTimeShort_d0370e7a739d1d7da91c2071fa2c12c5,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz_arab_af"] = func() locales.Translator {
		return &localen{

			locale:                 "uz_Arab_AF",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avg", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "Y", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avgust", "sentabr", "oktabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"Yak", "Dush", "Sesh", "Chor", "Pay", "Jum", "Shan"},
			daysNarrow:             []string{"Y", "D", "S", "C", "P", "J", "S"},
			daysShort:              []string{"Ya", "Du", "Se", "Ch", "Pa", "Ju", "Sh"},
			daysWide:               []string{"yakshanba", "dushanba", "seshanba", "chorshanba", "payshanba", "juma", "shanba"},
			periodsAbbreviated:     []string{"TO", "TK"},
			periodsNarrow:          []string{"TO", "TK"},
			periodsWide:            []string{"TO", "TK"},
			erasAbbreviated:        []string{"m.a.", "milodiy"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"miloddan avvalgi", "milodiy"},
			timezones:              map[string]string{"ACDT": "Markaziy Avstraliya yozgi vaqti", "ACST": "Markaziy Avstraliya standart vaqti", "ACWDT": "Markaziy Avstraliya g‘arbiy yozgi vaqti", "ACWST": "Markaziy Avstraliya g‘arbiy standart vaqti", "ADT": "Atlantika yozgi vaqti", "AEDT": "Sharqiy Avstraliya yozgi vaqti", "AEST": "Sharqiy Avstraliya standart vaqti", "AKDT": "Alyaska yozgi vaqti", "AKST": "Alyaska standart vaqti", "ARST": "Argentina yozgi vaqti", "ART": "Argentina standart vaqti", "AST": "Atlantika standart vaqti", "AWDT": "G‘arbiy Avstraliya yozgi vaqti", "AWST": "G‘arbiy Avstraliya standart vaqti", "BOT": "Boliviya vaqti", "BT": "Butan vaqti", "CAT": "Markaziy Afrika vaqti", "CDT": "Markaziy Amerika yozgi vaqti", "CHADT": "Chatem yozgi vaqti", "CHAST": "Chatem standart vaqti", "CLST": "Chili yozgi vaqti", "CLT": "Chili standart vaqti", "COST": "Kolumbiya yozgi vaqti", "COT": "Kolumbiya standart vaqti", "CST": "Markaziy Amerika standart vaqti", "ChST": "Chamorro standart vaqti", "EAT": "Sharqiy Afrika vaqti", "ECT": "Ekvador vaqti", "EDT": "Sharqiy Amerika yozgi vaqti", "EST": "Sharqiy Amerika standart vaqti", "GFT": "Fransuz Gvianasi vaqti", "GMT": "Grinvich o‘rtacha vaqti", "GST": "Fors ko‘rfazi standart vaqti", "GYT": "Gayana vaqti", "HADT": "Gavayi-aleut yozgi vaqti", "HAST": "Gavayi-aleut standart vaqti", "HAT": "Nyufaundlend yozgi vaqti", "HECU": "Kuba yozgi vaqti", "HEEG": "Sharqiy Grenlandiya yozgi vaqti", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "G‘arbiy Grenlandiya yozgi vaqti", "HEPM": "Sen-Pyer va Mikelon yozgi vaqti", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Gonkong yozgi vaqti", "HKT": "Gonkong standart vaqti", "HNCU": "Kuba standart vaqti", "HNEG": "Sharqiy Grenlandiya standart vaqti", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "G‘arbiy Grenlandiya standart vaqti", "HNPM": "Sen-Pyer va Mikelon standart vaqti", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Nyufaundlend standart vaqti", "IST": "Hindiston standart vaqti", "JDT": "Yaponiya yozgi vaqti", "JST": "Yaponiya standart vaqti", "LHDT": "Lord-Xau yozgi vaqti", "LHST": "Lord-Xau standart vaqti", "MDT": "Tog‘ yozgi vaqti (AQSH)", "MESZ": "Markaziy Yevropa yozgi vaqti", "MEZ": "Markaziy Yevropa standart vaqti", "MST": "Tog‘ standart vaqti (AQSH)", "MYT": "Malayziya vaqti", "NZDT": "Yangi Zelandiya yozgi vaqti", "NZST": "Yangi Zelandiya standart vaqti", "OESZ": "Sharqiy Yevropa yozgi vaqti", "OEZ": "Sharqiy Yevropa standart vaqti", "PDT": "Tinch okeani yozgi vaqti", "PST": "Tinch okeani standart vaqti", "SAST": "Janubiy Afrika standart vaqti", "SGT": "Singapur vaqti", "SRT": "Surinam vaqti", "TMST": "Turkmaniston yozgi vaqti", "TMT": "Turkmaniston standart vaqti", "UYST": "Urugvay yozgi vaqti", "UYT": "Urugvay standart vaqti", "VET": "Venesuela vaqti", "WARST": "Gʻarbiy Argentina yozgi vaqti", "WART": "Gʻarbiy Argentina standart vaqti", "WAST": "Gʻarbiy Afrika yozgi vaqti", "WAT": "Gʻarbiy Afrika standart vaqti", "WESZ": "G‘arbiy Yevropa yozgi vaqti", "WEZ": "G‘arbiy Yevropa standart vaqti", "WIB": "Gʻarbiy Indoneziya vaqti", "WIT": "Sharqiy Indoneziya vaqti", "WITA": "Markaziy Indoneziya vaqti", "∅∅∅": "Peru yozgi vaqti"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_ae39dfe33f097d30e514b77409da8997,
			fnFmtDateLong:         fnFmtDateLong_af1ab393e97e5cd9bd5342a65130666d,
			fnFmtDateMedium:       fnFmtDateMedium_8ff89c88bdb727333d4c7b42f495646e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz_cyrl"] = func() locales.Translator {
		return &localen{

			locale:                 "uz_Cyrl",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                "٫",
			group:                  "٬",
			minus:                  "-",
			percent:                "٪",
			perMille:               "؉",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "сўм", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"},
			monthsNarrow:           []string{"", "Я", "Ф", "М", "А", "М", "И", "И", "А", "С", "О", "Н", "Д"},
			monthsWide:             []string{"", "январ", "феврал", "март", "апрел", "май", "июн", "июл", "август", "сентябр", "октябр", "ноябр", "декабр"},
			daysAbbreviated:        []string{"якш", "душ", "сеш", "чор", "пай", "жум", "шан"},
			daysNarrow:             []string{"Я", "Д", "С", "Ч", "П", "Ж", "Ш"},
			daysShort:              []string{"як", "ду", "се", "чо", "па", "жу", "ша"},
			daysWide:               []string{"якшанба", "душанба", "сешанба", "чоршанба", "пайшанба", "жума", "шанба"},
			periodsAbbreviated:     []string{"ТО", "ТК"},
			periodsNarrow:          []string{"ТО", "ТК"},
			periodsWide:            []string{"ТО", "ТК"},
			erasAbbreviated:        []string{"м.а.", "милодий"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"милоддан аввалги", "милодий"},
			timezones:              map[string]string{"ACDT": "Марказий Австралия кундузги вақти", "ACST": "Марказий Австралия стандарт вақти", "ACWDT": "Марказий Австралия Ғарбий кундузги вақти", "ACWST": "Марказий Австралия Ғарбий стандарт вақти", "ADT": "Атлантика кундузги вақти", "AEDT": "Шарқий Австралия кундузги вақти", "AEST": "Шарқий Австралия стандарт вақти", "AKDT": "Аляска кундузги вақти", "AKST": "Аляска стандарт вақти", "ARST": "Аргентина ёзги вақти", "ART": "Аргентина стандарт вақти", "AST": "Атлантика стандарт вақти", "AWDT": "Ғарбий Австралия кундузги вақти", "AWST": "Ғарбий Австралия стандарт вақти", "BOT": "Боливия вақти", "BT": "Бутан вақти", "CAT": "Марказий Африка вақти", "CDT": "Шимолий Америка марказий кундузги вақти", "CHADT": "Чатхам кундузги вақти", "CHAST": "Чатхам стандарт вақти", "CLST": "Чили ёзги вақти", "CLT": "Чили стандарт вақти", "COST": "Колумбия ёзги вақти", "COT": "Колумбия стандарт вақти", "CST": "Шимолий Америка марказий стандарт вақти", "ChST": "Каморро вақти", "EAT": "Шарқий Африка вақти", "ECT": "Эквадор вақти", "EDT": "Шимолий Америка шарқий кундузги вақти", "EST": "Шимолий Америка шарқий стандарт вақти", "GFT": "Француз Гвианаси вақти", "GMT": "Гринвич вақти", "GST": "Кўрфаз вақти", "GYT": "Гайана вақти", "HADT": "Гавайи-алеут кундузги вақти", "HAST": "Гавайи-алеут стандарт вақти", "HAT": "Ньюфаундленд кундузги вақти", "HECU": "Куба кундузги вақти", "HEEG": "Шарқий Гренландия ёзги вақти", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "Ғарбий Гренландия ёзги вақти", "HEPM": "Сент-Пьер ва Микелон кундузги вақти", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Гонконг ёзги вақти", "HKT": "Гонконг стандарт вақти", "HNCU": "Куба стандарт вақти", "HNEG": "Шарқий Гренландия стандарт вақти", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "Ғарбий Гренландия стандарт вақти", "HNPM": "Сент-Пьер ва Микелон стандарт вақти", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Ньюфаундленд стандарт вақти", "IST": "Ҳиндистон вақти", "JDT": "Япония кундузги вақти", "JST": "Япония стандарт вақти", "LHDT": "Лорд Хове кундузги вақти", "LHST": "Лорд Хове стандарт вақти", "MDT": "Шимолий Америка тоғ кундузги вақти", "MESZ": "Марказий Европа ёзги вақти", "MEZ": "Марказий Европа стандарт вақти", "MST": "Шимолий Америка тоғ стандарт вақти", "MYT": "Малайзия вақти", "NZDT": "Янги Зеландия кундузги вақти", "NZST": "Янги Зеландия стандарт вақти", "OESZ": "Шарқий Европа ёзги вақти", "OEZ": "Шарқий Европа стандарт вақти", "PDT": "Шимолий Америка тинч океани кундузги вақти", "PST": "Шимолий Америка тинч океани стандарт вақти", "SAST": "Жанубий Африка вақти", "SGT": "Сингапур вақти", "SRT": "Суринам вақти", "TMST": "Туркманистон ёзги вақти", "TMT": "Туркманистон стандарт вақти", "UYST": "Уругвай ёзги вақти", "UYT": "Уругвай стандарт вақти", "VET": "Венесуэла вақти", "WARST": "Ғарбий Аргентина ёзги вақти", "WART": "Ғарбий Аргентина стандарт вақти", "WAST": "Ғарбий Африка ёзги вақти", "WAT": "Ғарбий Африка стандарт вақти", "WESZ": "Ғарбий Европа ёзги вақти", "WEZ": "Ғарбий Европа стандарт вақти", "WIB": "Ғарбий Индонезия вақти", "WIT": "Шарқий Индонезия вақти", "WITA": "Марказий Индонезия вақти", "∅∅∅": "Азор ёзги вақти"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_c2fb1e60c1c0fac4b7a6ab96cbc4f79e,
			fnFmtCurrency:         fnFmtCurrency_ae69fea052716ee7e5be5d84e205ff0c,
			fnFmtDateFull:         fnFmtDateFull_4edbf4e11f17243cbd3f53cf42390fdf,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_91f44359c32923b2c2a980df8ee6ab86,
			fnFmtPercent:          fnFmtPercent_141d2f250af9b9f46295bf60380ce0b2,
			fnFmtTimeFull:         fnFmtTimeFull_faed166f66bf2512306c6e70b367cf67,
			fnFmtTimeLong:         fnFmtTimeLong_30191015e11c1c183d8d41d8e0e78b42,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz_cyrl_uz"] = func() locales.Translator {
		return &localen{

			locale:                 "uz_Cyrl_UZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avg", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "Y", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avgust", "sentabr", "oktabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"Yak", "Dush", "Sesh", "Chor", "Pay", "Jum", "Shan"},
			daysNarrow:             []string{"Y", "D", "S", "C", "P", "J", "S"},
			daysShort:              []string{"Ya", "Du", "Se", "Ch", "Pa", "Ju", "Sh"},
			daysWide:               []string{"yakshanba", "dushanba", "seshanba", "chorshanba", "payshanba", "juma", "shanba"},
			periodsAbbreviated:     []string{"TO", "TK"},
			periodsNarrow:          []string{"TO", "TK"},
			periodsWide:            []string{"TO", "TK"},
			erasAbbreviated:        []string{"m.a.", "milodiy"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"miloddan avvalgi", "milodiy"},
			timezones:              map[string]string{"ACDT": "Markaziy Avstraliya yozgi vaqti", "ACST": "Markaziy Avstraliya standart vaqti", "ACWDT": "Markaziy Avstraliya g‘arbiy yozgi vaqti", "ACWST": "Markaziy Avstraliya g‘arbiy standart vaqti", "ADT": "Atlantika yozgi vaqti", "AEDT": "Sharqiy Avstraliya yozgi vaqti", "AEST": "Sharqiy Avstraliya standart vaqti", "AKDT": "Alyaska yozgi vaqti", "AKST": "Alyaska standart vaqti", "ARST": "Argentina yozgi vaqti", "ART": "Argentina standart vaqti", "AST": "Atlantika standart vaqti", "AWDT": "G‘arbiy Avstraliya yozgi vaqti", "AWST": "G‘arbiy Avstraliya standart vaqti", "BOT": "Boliviya vaqti", "BT": "Butan vaqti", "CAT": "Markaziy Afrika vaqti", "CDT": "Markaziy Amerika yozgi vaqti", "CHADT": "Chatem yozgi vaqti", "CHAST": "Chatem standart vaqti", "CLST": "Chili yozgi vaqti", "CLT": "Chili standart vaqti", "COST": "Kolumbiya yozgi vaqti", "COT": "Kolumbiya standart vaqti", "CST": "Markaziy Amerika standart vaqti", "ChST": "Chamorro standart vaqti", "EAT": "Sharqiy Afrika vaqti", "ECT": "Ekvador vaqti", "EDT": "Sharqiy Amerika yozgi vaqti", "EST": "Sharqiy Amerika standart vaqti", "GFT": "Fransuz Gvianasi vaqti", "GMT": "Grinvich o‘rtacha vaqti", "GST": "Fors ko‘rfazi standart vaqti", "GYT": "Gayana vaqti", "HADT": "Gavayi-aleut yozgi vaqti", "HAST": "Gavayi-aleut standart vaqti", "HAT": "Nyufaundlend yozgi vaqti", "HECU": "Kuba yozgi vaqti", "HEEG": "Sharqiy Grenlandiya yozgi vaqti", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "G‘arbiy Grenlandiya yozgi vaqti", "HEPM": "Sen-Pyer va Mikelon yozgi vaqti", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Gonkong yozgi vaqti", "HKT": "Gonkong standart vaqti", "HNCU": "Kuba standart vaqti", "HNEG": "Sharqiy Grenlandiya standart vaqti", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "G‘arbiy Grenlandiya standart vaqti", "HNPM": "Sen-Pyer va Mikelon standart vaqti", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Nyufaundlend standart vaqti", "IST": "Hindiston standart vaqti", "JDT": "Yaponiya yozgi vaqti", "JST": "Yaponiya standart vaqti", "LHDT": "Lord-Xau yozgi vaqti", "LHST": "Lord-Xau standart vaqti", "MDT": "Tog‘ yozgi vaqti (AQSH)", "MESZ": "Markaziy Yevropa yozgi vaqti", "MEZ": "Markaziy Yevropa standart vaqti", "MST": "Tog‘ standart vaqti (AQSH)", "MYT": "Malayziya vaqti", "NZDT": "Yangi Zelandiya yozgi vaqti", "NZST": "Yangi Zelandiya standart vaqti", "OESZ": "Sharqiy Yevropa yozgi vaqti", "OEZ": "Sharqiy Yevropa standart vaqti", "PDT": "Tinch okeani yozgi vaqti", "PST": "Tinch okeani standart vaqti", "SAST": "Janubiy Afrika standart vaqti", "SGT": "Singapur vaqti", "SRT": "Surinam vaqti", "TMST": "Turkmaniston yozgi vaqti", "TMT": "Turkmaniston standart vaqti", "UYST": "Urugvay yozgi vaqti", "UYT": "Urugvay standart vaqti", "VET": "Venesuela vaqti", "WARST": "Gʻarbiy Argentina yozgi vaqti", "WART": "Gʻarbiy Argentina standart vaqti", "WAST": "Gʻarbiy Afrika yozgi vaqti", "WAT": "Gʻarbiy Afrika standart vaqti", "WESZ": "G‘arbiy Yevropa yozgi vaqti", "WEZ": "G‘arbiy Yevropa standart vaqti", "WIB": "Gʻarbiy Indoneziya vaqti", "WIT": "Sharqiy Indoneziya vaqti", "WITA": "Markaziy Indoneziya vaqti", "∅∅∅": "Braziliya yozgi vaqti"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_ae39dfe33f097d30e514b77409da8997,
			fnFmtDateLong:         fnFmtDateLong_af1ab393e97e5cd9bd5342a65130666d,
			fnFmtDateMedium:       fnFmtDateMedium_8ff89c88bdb727333d4c7b42f495646e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz_latn"] = func() locales.Translator {
		return &localen{

			locale:                 "uz_Latn",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avg", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "Y", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avgust", "sentabr", "oktabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"Yak", "Dush", "Sesh", "Chor", "Pay", "Jum", "Shan"},
			daysNarrow:             []string{"Y", "D", "S", "C", "P", "J", "S"},
			daysShort:              []string{"Ya", "Du", "Se", "Ch", "Pa", "Ju", "Sh"},
			daysWide:               []string{"yakshanba", "dushanba", "seshanba", "chorshanba", "payshanba", "juma", "shanba"},
			periodsAbbreviated:     []string{"TO", "TK"},
			periodsNarrow:          []string{"TO", "TK"},
			periodsWide:            []string{"TO", "TK"},
			erasAbbreviated:        []string{"m.a.", "milodiy"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"miloddan avvalgi", "milodiy"},
			timezones:              map[string]string{"ACDT": "Markaziy Avstraliya yozgi vaqti", "ACST": "Markaziy Avstraliya standart vaqti", "ACWDT": "Markaziy Avstraliya g‘arbiy yozgi vaqti", "ACWST": "Markaziy Avstraliya g‘arbiy standart vaqti", "ADT": "Atlantika yozgi vaqti", "AEDT": "Sharqiy Avstraliya yozgi vaqti", "AEST": "Sharqiy Avstraliya standart vaqti", "AKDT": "Alyaska yozgi vaqti", "AKST": "Alyaska standart vaqti", "ARST": "Argentina yozgi vaqti", "ART": "Argentina standart vaqti", "AST": "Atlantika standart vaqti", "AWDT": "G‘arbiy Avstraliya yozgi vaqti", "AWST": "G‘arbiy Avstraliya standart vaqti", "BOT": "Boliviya vaqti", "BT": "Butan vaqti", "CAT": "Markaziy Afrika vaqti", "CDT": "Markaziy Amerika yozgi vaqti", "CHADT": "Chatem yozgi vaqti", "CHAST": "Chatem standart vaqti", "CLST": "Chili yozgi vaqti", "CLT": "Chili standart vaqti", "COST": "Kolumbiya yozgi vaqti", "COT": "Kolumbiya standart vaqti", "CST": "Markaziy Amerika standart vaqti", "ChST": "Chamorro standart vaqti", "EAT": "Sharqiy Afrika vaqti", "ECT": "Ekvador vaqti", "EDT": "Sharqiy Amerika yozgi vaqti", "EST": "Sharqiy Amerika standart vaqti", "GFT": "Fransuz Gvianasi vaqti", "GMT": "Grinvich o‘rtacha vaqti", "GST": "Fors ko‘rfazi standart vaqti", "GYT": "Gayana vaqti", "HADT": "Gavayi-aleut yozgi vaqti", "HAST": "Gavayi-aleut standart vaqti", "HAT": "Nyufaundlend yozgi vaqti", "HECU": "Kuba yozgi vaqti", "HEEG": "Sharqiy Grenlandiya yozgi vaqti", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "G‘arbiy Grenlandiya yozgi vaqti", "HEPM": "Sen-Pyer va Mikelon yozgi vaqti", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Gonkong yozgi vaqti", "HKT": "Gonkong standart vaqti", "HNCU": "Kuba standart vaqti", "HNEG": "Sharqiy Grenlandiya standart vaqti", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "G‘arbiy Grenlandiya standart vaqti", "HNPM": "Sen-Pyer va Mikelon standart vaqti", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Nyufaundlend standart vaqti", "IST": "Hindiston standart vaqti", "JDT": "Yaponiya yozgi vaqti", "JST": "Yaponiya standart vaqti", "LHDT": "Lord-Xau yozgi vaqti", "LHST": "Lord-Xau standart vaqti", "MDT": "Tog‘ yozgi vaqti (AQSH)", "MESZ": "Markaziy Yevropa yozgi vaqti", "MEZ": "Markaziy Yevropa standart vaqti", "MST": "Tog‘ standart vaqti (AQSH)", "MYT": "Malayziya vaqti", "NZDT": "Yangi Zelandiya yozgi vaqti", "NZST": "Yangi Zelandiya standart vaqti", "OESZ": "Sharqiy Yevropa yozgi vaqti", "OEZ": "Sharqiy Yevropa standart vaqti", "PDT": "Tinch okeani yozgi vaqti", "PST": "Tinch okeani standart vaqti", "SAST": "Janubiy Afrika standart vaqti", "SGT": "Singapur vaqti", "SRT": "Surinam vaqti", "TMST": "Turkmaniston yozgi vaqti", "TMT": "Turkmaniston standart vaqti", "UYST": "Urugvay yozgi vaqti", "UYT": "Urugvay standart vaqti", "VET": "Venesuela vaqti", "WARST": "Gʻarbiy Argentina yozgi vaqti", "WART": "Gʻarbiy Argentina standart vaqti", "WAST": "Gʻarbiy Afrika yozgi vaqti", "WAT": "Gʻarbiy Afrika standart vaqti", "WESZ": "G‘arbiy Yevropa yozgi vaqti", "WEZ": "G‘arbiy Yevropa standart vaqti", "WIB": "Gʻarbiy Indoneziya vaqti", "WIT": "Sharqiy Indoneziya vaqti", "WITA": "Markaziy Indoneziya vaqti", "∅∅∅": "Braziliya yozgi vaqti"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_ae39dfe33f097d30e514b77409da8997,
			fnFmtDateLong:         fnFmtDateLong_af1ab393e97e5cd9bd5342a65130666d,
			fnFmtDateMedium:       fnFmtDateMedium_8ff89c88bdb727333d4c7b42f495646e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["uz_latn_uz"] = func() locales.Translator {
		return &localen{

			locale:                 "uz_Latn_UZ",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ",",
			group:                  " ",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "yan", "fev", "mar", "apr", "may", "iyn", "iyl", "avg", "sen", "okt", "noy", "dek"},
			monthsNarrow:           []string{"", "Y", "F", "M", "A", "M", "I", "I", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avgust", "sentabr", "oktabr", "noyabr", "dekabr"},
			daysAbbreviated:        []string{"Yak", "Dush", "Sesh", "Chor", "Pay", "Jum", "Shan"},
			daysNarrow:             []string{"Y", "D", "S", "C", "P", "J", "S"},
			daysShort:              []string{"Ya", "Du", "Se", "Ch", "Pa", "Ju", "Sh"},
			daysWide:               []string{"yakshanba", "dushanba", "seshanba", "chorshanba", "payshanba", "juma", "shanba"},
			periodsAbbreviated:     []string{"TO", "TK"},
			periodsNarrow:          []string{"TO", "TK"},
			periodsWide:            []string{"TO", "TK"},
			erasAbbreviated:        []string{"m.a.", "milodiy"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"miloddan avvalgi", "milodiy"},
			timezones:              map[string]string{"ACDT": "Markaziy Avstraliya yozgi vaqti", "ACST": "Markaziy Avstraliya standart vaqti", "ACWDT": "Markaziy Avstraliya g‘arbiy yozgi vaqti", "ACWST": "Markaziy Avstraliya g‘arbiy standart vaqti", "ADT": "Atlantika yozgi vaqti", "AEDT": "Sharqiy Avstraliya yozgi vaqti", "AEST": "Sharqiy Avstraliya standart vaqti", "AKDT": "Alyaska yozgi vaqti", "AKST": "Alyaska standart vaqti", "ARST": "Argentina yozgi vaqti", "ART": "Argentina standart vaqti", "AST": "Atlantika standart vaqti", "AWDT": "G‘arbiy Avstraliya yozgi vaqti", "AWST": "G‘arbiy Avstraliya standart vaqti", "BOT": "Boliviya vaqti", "BT": "Butan vaqti", "CAT": "Markaziy Afrika vaqti", "CDT": "Markaziy Amerika yozgi vaqti", "CHADT": "Chatem yozgi vaqti", "CHAST": "Chatem standart vaqti", "CLST": "Chili yozgi vaqti", "CLT": "Chili standart vaqti", "COST": "Kolumbiya yozgi vaqti", "COT": "Kolumbiya standart vaqti", "CST": "Markaziy Amerika standart vaqti", "ChST": "Chamorro standart vaqti", "EAT": "Sharqiy Afrika vaqti", "ECT": "Ekvador vaqti", "EDT": "Sharqiy Amerika yozgi vaqti", "EST": "Sharqiy Amerika standart vaqti", "GFT": "Fransuz Gvianasi vaqti", "GMT": "Grinvich o‘rtacha vaqti", "GST": "Fors ko‘rfazi standart vaqti", "GYT": "Gayana vaqti", "HADT": "Gavayi-aleut yozgi vaqti", "HAST": "Gavayi-aleut standart vaqti", "HAT": "Nyufaundlend yozgi vaqti", "HECU": "Kuba yozgi vaqti", "HEEG": "Sharqiy Grenlandiya yozgi vaqti", "HENOMX": "Shimoli-g‘arbiy Meksika yozgi vaqti", "HEOG": "G‘arbiy Grenlandiya yozgi vaqti", "HEPM": "Sen-Pyer va Mikelon yozgi vaqti", "HEPMX": "Meksika Tinch okeani yozgi vaqti", "HKST": "Gonkong yozgi vaqti", "HKT": "Gonkong standart vaqti", "HNCU": "Kuba standart vaqti", "HNEG": "Sharqiy Grenlandiya standart vaqti", "HNNOMX": "Shimoli-g‘arbiy Meksika standart vaqti", "HNOG": "G‘arbiy Grenlandiya standart vaqti", "HNPM": "Sen-Pyer va Mikelon standart vaqti", "HNPMX": "Meksika Tinch okeani standart vaqti", "HNT": "Nyufaundlend standart vaqti", "IST": "Hindiston standart vaqti", "JDT": "Yaponiya yozgi vaqti", "JST": "Yaponiya standart vaqti", "LHDT": "Lord-Xau yozgi vaqti", "LHST": "Lord-Xau standart vaqti", "MDT": "MDT", "MESZ": "Markaziy Yevropa yozgi vaqti", "MEZ": "Markaziy Yevropa standart vaqti", "MST": "MST", "MYT": "Malayziya vaqti", "NZDT": "Yangi Zelandiya yozgi vaqti", "NZST": "Yangi Zelandiya standart vaqti", "OESZ": "Sharqiy Yevropa yozgi vaqti", "OEZ": "Sharqiy Yevropa standart vaqti", "PDT": "Tinch okeani yozgi vaqti", "PST": "Tinch okeani standart vaqti", "SAST": "Janubiy Afrika standart vaqti", "SGT": "Singapur vaqti", "SRT": "Surinam vaqti", "TMST": "Turkmaniston yozgi vaqti", "TMT": "Turkmaniston standart vaqti", "UYST": "Urugvay yozgi vaqti", "UYT": "Urugvay standart vaqti", "VET": "Venesuela vaqti", "WARST": "Gʻarbiy Argentina yozgi vaqti", "WART": "Gʻarbiy Argentina standart vaqti", "WAST": "Gʻarbiy Afrika yozgi vaqti", "WAT": "Gʻarbiy Afrika standart vaqti", "WESZ": "G‘arbiy Yevropa yozgi vaqti", "WEZ": "G‘arbiy Yevropa standart vaqti", "WIB": "Gʻarbiy Indoneziya vaqti", "WIT": "Sharqiy Indoneziya vaqti", "WITA": "Markaziy Indoneziya vaqti", "∅∅∅": "Peru yozgi vaqti"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a62648328d65f4275e2e521198233607,
			fnFmtCurrency:         fnFmtCurrency_04d2836e8ee5922750958599d99ccc13,
			fnFmtDateFull:         fnFmtDateFull_ae39dfe33f097d30e514b77409da8997,
			fnFmtDateLong:         fnFmtDateLong_af1ab393e97e5cd9bd5342a65130666d,
			fnFmtDateMedium:       fnFmtDateMedium_8ff89c88bdb727333d4c7b42f495646e,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_e678928158d95fb12912f844eab5a4da,
			fnFmtTimeLong:         fnFmtTimeLong_e098bb5b410cb647ab1fb51d60e8416c,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_a46571fc772262a1fe652a3412eb1140,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vai"] = func() locales.Translator {
		return &localen{

			locale:                 "vai",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "$", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ꖨꖕꔞ", "ꕒꕡ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋ", "ꖨꖕꗏ"},
			monthsWide:             []string{"", "ꖨꖕ ꕪꕴ ꔞꔀꕮꕊ", "ꕒꕡꖝꖕ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞꔤ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋꕔꕿ ꕸꖃꗏ", "ꖨꖕ ꕪꕴ ꗏꖺꕮꕊ"},
			daysWide:               []string{"ꕞꕌꔵ", "ꗳꗡꘉ", "ꕚꕞꕚ", "ꕉꕞꕒ", "ꕉꔤꕆꕢ", "ꕉꔤꕀꕮ", "ꔻꔬꔳ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vai_latn"] = func() locales.Translator {
		return &localen{

			locale:                 "vai_Latn",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "$", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ꖨꖕꔞ", "ꕒꕡ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋ", "ꖨꖕꗏ"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "luukao kemã", "ɓandaɓu", "vɔɔ", "fulu", "goo", "6", "7", "kɔnde", "saah", "galo", "kenpkato ɓololɔ", "luukao lɔma"},
			daysNarrow:             []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:               []string{"lahadi", "tɛɛnɛɛ", "talata", "alaba", "aimisa", "aijima", "siɓiti"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"BCE", "CE"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vai_latn_lr"] = func() locales.Translator {
		return &localen{

			locale:                 "vai_Latn_LR",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ꖨꖕꔞ", "ꕒꕡ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋ", "ꖨꖕꗏ"},
			monthsWide:             []string{"", "ꖨꖕ ꕪꕴ ꔞꔀꕮꕊ", "ꕒꕡꖝꖕ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞꔤ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋꕔꕿ ꕸꖃꗏ", "ꖨꖕ ꕪꕴ ꗏꖺꕮꕊ"},
			daysWide:               []string{"ꕞꕌꔵ", "ꗳꗡꘉ", "ꕚꕞꕚ", "ꕉꕞꕒ", "ꕉꔤꕆꕢ", "ꕉꔤꕀꕮ", "ꔻꔬꔳ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vai_vaii"] = func() locales.Translator {
		return &localen{

			locale:                 "vai_Vaii",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ꖨꖕꔞ", "ꕒꕡ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋ", "ꖨꖕꗏ"},
			monthsWide:             []string{"", "ꖨꖕ ꕪꕴ ꔞꔀꕮꕊ", "ꕒꕡꖝꖕ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞꔤ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋꕔꕿ ꕸꖃꗏ", "ꖨꖕ ꕪꕴ ꗏꖺꕮꕊ"},
			daysWide:               []string{"ꕞꕌꔵ", "ꗳꗡꘉ", "ꕚꕞꕚ", "ꕉꕞꕒ", "ꕉꔤꕆꕢ", "ꕉꔤꕀꕮ", "ꔻꔬꔳ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vai_vaii_lr"] = func() locales.Translator {
		return &localen{

			locale:                 "vai_Vaii_LR",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "ꖨꖕꔞ", "ꕒꕡ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋ", "ꖨꖕꗏ"},
			monthsWide:             []string{"", "ꖨꖕ ꕪꕴ ꔞꔀꕮꕊ", "ꕒꕡꖝꖕ", "ꕾꖺ", "ꖢꖕ", "ꖑꕱ", "ꖱꘋ", "ꖱꕞꔤ", "ꗛꔕ", "ꕢꕌ", "ꕭꖃ", "ꔞꘋꕔꕿ ꕸꖃꗏ", "ꖨꖕ ꕪꕴ ꗏꖺꕮꕊ"},
			daysWide:               []string{"ꕞꕌꔵ", "ꗳꗡꘉ", "ꕚꕞꕚ", "ꕉꕞꕒ", "ꕉꔤꕆꕢ", "ꕉꔤꕀꕮ", "ꔻꔬꔳ"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_9a4ade24195ddf3fbcd889601bdb31fd,
			fnFmtCurrency:         fnFmtCurrency_21670b336e6e64700cc7f949f05c390f,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_74395470c4c29cb378d864498139a946,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_37fe4deeee789852cc196a6a1a2f0ccc,
			fnFmtTimeLong:         fnFmtTimeLong_4429ef4fb58a2a9c4d7d5132fd16ecbf,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ef07f6477b7a4944577d4e5392ff21b,
			fnFmtTimeShort:        fnFmtTimeShort_8288b817c5c0216c9815524c71c9e0b4,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vi"] = func() locales.Translator {
		return &localen{

			locale:                 "vi",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "thg 1", "thg 2", "thg 3", "thg 4", "thg 5", "thg 6", "thg 7", "thg 8", "thg 9", "thg 10", "thg 11", "thg 12"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "tháng 1", "tháng 2", "tháng 3", "tháng 4", "tháng 5", "tháng 6", "tháng 7", "tháng 8", "tháng 9", "tháng 10", "tháng 11", "tháng 12"},
			daysAbbreviated:        []string{"CN", "Th 2", "Th 3", "Th 4", "Th 5", "Th 6", "Th 7"},
			daysNarrow:             []string{"CN", "T2", "T3", "T4", "T5", "T6", "T7"},
			daysShort:              []string{"CN", "T2", "T3", "T4", "T5", "T6", "T7"},
			daysWide:               []string{"Chủ Nhật", "Thứ Hai", "Thứ Ba", "Thứ Tư", "Thứ Năm", "Thứ Sáu", "Thứ Bảy"},
			periodsAbbreviated:     []string{"SA", "CH"},
			periodsNarrow:          []string{"s", "c"},
			periodsWide:            []string{"SA", "CH"},
			erasAbbreviated:        []string{"Trước CN", "sau CN"},
			erasNarrow:             []string{"tr. CN", "sau CN"},
			erasWide:               []string{"Trước CN", "sau CN"},
			timezones:              map[string]string{"ACDT": "Giờ Mùa Hè Miền Trung Australia", "ACST": "Giờ Chuẩn Miền Trung Australia", "ACWDT": "Giờ Mùa Hè Miền Trung Tây Australia", "ACWST": "Giờ Chuẩn Miền Trung Tây Australia", "ADT": "Giờ mùa hè Đại Tây Dương", "AEDT": "Giờ Mùa Hè Miền Đông Australia", "AEST": "Giờ Chuẩn Miền Đông Australia", "AKDT": "Giờ Mùa Hè Alaska", "AKST": "Giờ Chuẩn Alaska", "ARST": "Giờ Mùa Hè Argentina", "ART": "Giờ Chuẩn Argentina", "AST": "Giờ Chuẩn Đại Tây Dương", "AWDT": "Giờ Mùa Hè Miền Tây Australia", "AWST": "Giờ Chuẩn Miền Tây Australia", "BOT": "Giờ Bolivia", "BT": "Giờ Bhutan", "CAT": "Giờ Trung Phi", "CDT": "Giờ mùa hè miền Trung", "CHADT": "Giờ Mùa Hè Chatham", "CHAST": "Giờ Chuẩn Chatham", "CLST": "Giờ Mùa Hè Chile", "CLT": "Giờ Chuẩn Chile", "COST": "Giờ Mùa Hè Colombia", "COT": "Giờ Chuẩn Colombia", "CST": "Giờ chuẩn miền Trung", "ChST": "Giờ Chamorro", "EAT": "Giờ Đông Phi", "ECT": "Giờ Ecuador", "EDT": "Giờ mùa hè miền Đông", "EST": "Giờ chuẩn miền Đông", "GFT": "Giờ Guiana thuộc Pháp", "GMT": "Giờ Trung bình Greenwich", "GST": "Giờ Chuẩn Vùng Vịnh", "GYT": "Giờ Guyana", "HADT": "Giờ Mùa Hè Hawaii-Aleut", "HAST": "Giờ Chuẩn Hawaii-Aleut", "HAT": "Giờ Mùa Hè Newfoundland", "HECU": "Giờ Mùa Hè Cuba", "HEEG": "Giờ Mùa Hè Miền Đông Greenland", "HENOMX": "Giờ Mùa Hè Tây Bắc Mexico", "HEOG": "Giờ Mùa Hè Miền Tây Greenland", "HEPM": "Giờ Mùa Hè Saint Pierre và Miquelon", "HEPMX": "Giờ Mùa Hè Thái Bình Dương Mexico", "HKST": "Giờ Mùa Hè Hồng Kông", "HKT": "Giờ Chuẩn Hồng Kông", "HNCU": "Giờ Chuẩn Cuba", "HNEG": "Giờ Chuẩn Miền Đông Greenland", "HNNOMX": "Giờ Chuẩn Tây Bắc Mexico", "HNOG": "Giờ Chuẩn Miền Tây Greenland", "HNPM": "Giờ Chuẩn St. Pierre và Miquelon", "HNPMX": "Giờ Chuẩn Thái Bình Dương Mexico", "HNT": "Giờ Chuẩn Newfoundland", "IST": "Giờ Chuẩn Ấn Độ", "JDT": "Giờ Mùa Hè Nhật Bản", "JST": "Giờ Chuẩn Nhật Bản", "LHDT": "Giờ Mùa Hè Lord Howe", "LHST": "Giờ Chuẩn Lord Howe", "MDT": "Giờ mùa hè miền núi", "MESZ": "Giờ mùa hè Trung Âu", "MEZ": "Giờ chuẩn Trung Âu", "MST": "Giờ chuẩn miền núi", "MYT": "Giờ Malaysia", "NZDT": "Giờ Mùa Hè New Zealand", "NZST": "Giờ Chuẩn New Zealand", "OESZ": "Giờ mùa hè Đông Âu", "OEZ": "Giờ chuẩn Đông Âu", "PDT": "Giờ mùa hè Thái Bình Dương", "PST": "Giờ chuẩn Thái Bình Dương", "SAST": "Giờ Chuẩn Nam Phi", "SGT": "Giờ Singapore", "SRT": "Giờ Suriname", "TMST": "Giờ Mùa Hè Turkmenistan", "TMT": "Giờ Chuẩn Turkmenistan", "UYST": "Giờ Mùa Hè Uruguay", "UYT": "Giờ Chuẩn Uruguay", "VET": "Giờ Venezuela", "WARST": "Giờ mùa hè miền tây Argentina", "WART": "Giờ chuẩn miền tây Argentina", "WAST": "Giờ Mùa Hè Tây Phi", "WAT": "Giờ Chuẩn Tây Phi", "WESZ": "Giờ mùa hè Tây Âu", "WEZ": "Giờ Chuẩn Tây Âu", "WIB": "Giờ Miền Tây Indonesia", "WIT": "Giờ Miền Đông Indonesia", "WITA": "Giờ Miền Trung Indonesia", "∅∅∅": "Giờ Mùa Hè Azores"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vi_vn"] = func() locales.Translator {
		return &localen{

			locale:                 "vi_VN",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{2, 6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "thg 1", "thg 2", "thg 3", "thg 4", "thg 5", "thg 6", "thg 7", "thg 8", "thg 9", "thg 10", "thg 11", "thg 12"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "tháng 1", "tháng 2", "tháng 3", "tháng 4", "tháng 5", "tháng 6", "tháng 7", "tháng 8", "tháng 9", "tháng 10", "tháng 11", "tháng 12"},
			daysAbbreviated:        []string{"CN", "Th 2", "Th 3", "Th 4", "Th 5", "Th 6", "Th 7"},
			daysNarrow:             []string{"CN", "T2", "T3", "T4", "T5", "T6", "T7"},
			daysShort:              []string{"CN", "T2", "T3", "T4", "T5", "T6", "T7"},
			daysWide:               []string{"Chủ Nhật", "Thứ Hai", "Thứ Ba", "Thứ Tư", "Thứ Năm", "Thứ Sáu", "Thứ Bảy"},
			periodsAbbreviated:     []string{"SA", "CH"},
			periodsNarrow:          []string{"s", "c"},
			periodsWide:            []string{"SA", "CH"},
			erasAbbreviated:        []string{"Trước CN", "sau CN"},
			erasNarrow:             []string{"tr. CN", "sau CN"},
			erasWide:               []string{"Trước CN", "sau CN"},
			timezones:              map[string]string{"ACDT": "Giờ Mùa Hè Miền Trung Australia", "ACST": "Giờ Chuẩn Miền Trung Australia", "ACWDT": "Giờ Mùa Hè Miền Trung Tây Australia", "ACWST": "Giờ Chuẩn Miền Trung Tây Australia", "ADT": "Giờ mùa hè Đại Tây Dương", "AEDT": "Giờ Mùa Hè Miền Đông Australia", "AEST": "Giờ Chuẩn Miền Đông Australia", "AKDT": "Giờ Mùa Hè Alaska", "AKST": "Giờ Chuẩn Alaska", "ARST": "Giờ Mùa Hè Argentina", "ART": "Giờ Chuẩn Argentina", "AST": "Giờ Chuẩn Đại Tây Dương", "AWDT": "Giờ Mùa Hè Miền Tây Australia", "AWST": "Giờ Chuẩn Miền Tây Australia", "BOT": "Giờ Bolivia", "BT": "Giờ Bhutan", "CAT": "Giờ Trung Phi", "CDT": "Giờ mùa hè miền Trung", "CHADT": "Giờ Mùa Hè Chatham", "CHAST": "Giờ Chuẩn Chatham", "CLST": "Giờ Mùa Hè Chile", "CLT": "Giờ Chuẩn Chile", "COST": "Giờ Mùa Hè Colombia", "COT": "Giờ Chuẩn Colombia", "CST": "Giờ chuẩn miền Trung", "ChST": "Giờ Chamorro", "EAT": "Giờ Đông Phi", "ECT": "Giờ Ecuador", "EDT": "Giờ mùa hè miền Đông", "EST": "Giờ chuẩn miền Đông", "GFT": "Giờ Guiana thuộc Pháp", "GMT": "Giờ Trung bình Greenwich", "GST": "Giờ Chuẩn Vùng Vịnh", "GYT": "Giờ Guyana", "HADT": "Giờ Mùa Hè Hawaii-Aleut", "HAST": "Giờ Chuẩn Hawaii-Aleut", "HAT": "Giờ Mùa Hè Newfoundland", "HECU": "Giờ Mùa Hè Cuba", "HEEG": "Giờ Mùa Hè Miền Đông Greenland", "HENOMX": "Giờ Mùa Hè Tây Bắc Mexico", "HEOG": "Giờ Mùa Hè Miền Tây Greenland", "HEPM": "Giờ Mùa Hè Saint Pierre và Miquelon", "HEPMX": "Giờ Mùa Hè Thái Bình Dương Mexico", "HKST": "Giờ Mùa Hè Hồng Kông", "HKT": "Giờ Chuẩn Hồng Kông", "HNCU": "Giờ Chuẩn Cuba", "HNEG": "Giờ Chuẩn Miền Đông Greenland", "HNNOMX": "Giờ Chuẩn Tây Bắc Mexico", "HNOG": "Giờ Chuẩn Miền Tây Greenland", "HNPM": "Giờ Chuẩn St. Pierre và Miquelon", "HNPMX": "Giờ Chuẩn Thái Bình Dương Mexico", "HNT": "Giờ Chuẩn Newfoundland", "IST": "Giờ Chuẩn Ấn Độ", "JDT": "Giờ Mùa Hè Nhật Bản", "JST": "Giờ Chuẩn Nhật Bản", "LHDT": "Giờ Mùa Hè Lord Howe", "LHST": "Giờ Chuẩn Lord Howe", "MDT": "Giờ mùa hè miền núi", "MESZ": "Giờ mùa hè Trung Âu", "MEZ": "Giờ chuẩn Trung Âu", "MST": "Giờ chuẩn miền núi", "MYT": "Giờ Malaysia", "NZDT": "Giờ Mùa Hè New Zealand", "NZST": "Giờ Chuẩn New Zealand", "OESZ": "Giờ mùa hè Đông Âu", "OEZ": "Giờ chuẩn Đông Âu", "PDT": "Giờ mùa hè Thái Bình Dương", "PST": "Giờ chuẩn Thái Bình Dương", "SAST": "Giờ Chuẩn Nam Phi", "SGT": "Giờ Singapore", "SRT": "Giờ Suriname", "TMST": "Giờ Mùa Hè Turkmenistan", "TMT": "Giờ Chuẩn Turkmenistan", "UYST": "Giờ Mùa Hè Uruguay", "UYT": "Giờ Chuẩn Uruguay", "VET": "Giờ Venezuela", "WARST": "Giờ mùa hè miền tây Argentina", "WART": "Giờ chuẩn miền tây Argentina", "WAST": "Giờ Mùa Hè Tây Phi", "WAT": "Giờ Chuẩn Tây Phi", "WESZ": "Giờ mùa hè Tây Âu", "WEZ": "Giờ Chuẩn Tây Âu", "WIB": "Giờ Miền Tây Indonesia", "WIT": "Giờ Miền Đông Indonesia", "WITA": "Giờ Miền Trung Indonesia", "∅∅∅": "Giờ Mùa Hè Brasilia"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_295d01ff3669c94717ee670511c0e25e,
			fnFmtCurrency:         fnFmtCurrency_c43b1593289525b3b62af252d88e9f2d,
			fnFmtDateFull:         fnFmtDateFull_bd1b45bc043b6b4e8c3b831b9263891b,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_89b8067afb87364ac493104461523e40,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vo"] = func() locales.Translator {
		return &localen{

			locale:            "vo",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "yan", "feb", "mäz", "prl", "may", "yun", "yul", "gst", "set", "ton", "nov", "dek"},
			monthsNarrow:      []string{"", "Y", "F", "M", "P", "M", "Y", "Y", "G", "S", "T", "N", "D"},
			monthsWide:        []string{"", "yanul", "febul", "mäzul", "prilul", "mayul", "yunul", "yulul", "gustul", "setul", "tobul", "novul", "dekul"},
			daysAbbreviated:   []string{"su.", "mu.", "tu.", "ve.", "dö.", "fr.", "zä."},
			daysNarrow:        []string{"S", "M", "T", "V", "D", "F", "Z"},
			daysWide:          []string{"sudel", "mudel", "tudel", "vedel", "dödel", "fridel", "zädel"},
			erasAbbreviated:   []string{"b. t. kr.", "p. t. kr."},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"b. t. kr.", "p. t. kr."},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_411c735c7aafad814d0db05f621af17c,
			fnFmtCurrency:         fnFmtCurrency_8bfebaacbf71a9a4bab2a3bb99e52529,
			fnFmtDateFull:         fnFmtDateFull_c73645256fa9ec42586084cc4fafc6f6,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_c010b271f06248ba07bbc009fc58011a,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vo_001"] = func() locales.Translator {
		return &localen{

			locale:            "vo_001",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "yan", "feb", "mäz", "prl", "may", "yun", "yul", "gst", "set", "ton", "nov", "dek"},
			monthsNarrow:      []string{"", "Y", "F", "M", "P", "M", "Y", "Y", "G", "S", "T", "N", "D"},
			monthsWide:        []string{"", "yanul", "febul", "mäzul", "prilul", "mayul", "yunul", "yulul", "gustul", "setul", "tobul", "novul", "dekul"},
			daysAbbreviated:   []string{"su.", "mu.", "tu.", "ve.", "dö.", "fr.", "zä."},
			daysNarrow:        []string{"S", "M", "T", "V", "D", "F", "Z"},
			daysWide:          []string{"sudel", "mudel", "tudel", "vedel", "dödel", "fridel", "zädel"},
			erasAbbreviated:   []string{"b. t. kr.", "p. t. kr."},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"b. t. kr.", "p. t. kr."},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_411c735c7aafad814d0db05f621af17c,
			fnFmtCurrency:         fnFmtCurrency_8bfebaacbf71a9a4bab2a3bb99e52529,
			fnFmtDateFull:         fnFmtDateFull_c73645256fa9ec42586084cc4fafc6f6,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_c010b271f06248ba07bbc009fc58011a,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vun"] = func() locales.Translator {
		return &localen{

			locale:             "vun",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TSh", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprilyi", "Mei", "Junyi", "Julyai", "Agusti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Jumapilyi", "Jumatatuu", "Jumanne", "Jumatanu", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"utuko", "kyiukonyi"},
			periodsWide:        []string{"utuko", "kyiukonyi"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Kristu", "Baada ya Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["vun_tz"] = func() locales.Translator {
		return &localen{

			locale:             "vun_TZ",
			pluralsCardinal:    []locales.PluralRule{2, 6},
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			timeSeparator:      ":",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"},
			monthsNarrow:       []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:         []string{"", "Januari", "Februari", "Machi", "Aprilyi", "Mei", "Junyi", "Julyai", "Agusti", "Septemba", "Oktoba", "Novemba", "Desemba"},
			daysAbbreviated:    []string{"Jpi", "Jtt", "Jnn", "Jtn", "Alh", "Iju", "Jmo"},
			daysNarrow:         []string{"J", "J", "J", "J", "A", "I", "J"},
			daysWide:           []string{"Jumapilyi", "Jumatatuu", "Jumanne", "Jumatanu", "Alhamisi", "Ijumaa", "Jumamosi"},
			periodsAbbreviated: []string{"utuko", "kyiukonyi"},
			periodsWide:        []string{"utuko", "kyiukonyi"},
			erasAbbreviated:    []string{"KK", "BK"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"Kabla ya Kristu", "Baada ya Kristu"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_3d79231c973b635e0adc0ac3279907d3,
			fnFmtCurrency:         fnFmtCurrency_12d8afad3eb872da9d80fb14637f7f61,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["wae"] = func() locales.Translator {
		return &localen{

			locale:                 "wae",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  "’",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "Jen", "Hor", "Mär", "Abr", "Mei", "Brá", "Hei", "Öig", "Her", "Wím", "Win", "Chr"},
			monthsNarrow:           []string{"", "J", "H", "M", "A", "M", "B", "H", "Ö", "H", "W", "W", "C"},
			monthsWide:             []string{"", "Jenner", "Hornig", "Märze", "Abrille", "Meije", "Bráčet", "Heiwet", "Öigšte", "Herbštmánet", "Wímánet", "Wintermánet", "Chrištmánet"},
			daysAbbreviated:        []string{"Sun", "Män", "Ziš", "Mit", "Fró", "Fri", "Sam"},
			daysNarrow:             []string{"S", "M", "Z", "M", "F", "F", "S"},
			daysWide:               []string{"Sunntag", "Mäntag", "Zištag", "Mittwuč", "Fróntag", "Fritag", "Samštag"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Atlantiši Summerzit", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Atlantiši Standardzit", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Mitteleuropäiši Summerzit", "MEZ": "Mitteleuropäiši Standardzit", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Ošteuropäiši Summerzit", "OEZ": "Ošteuropäiši Standardzit", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Wešteuropäiši Summerzit", "WEZ": "Wešteuropäiši Standardzit", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06,
			fnFmtCurrency:         fnFmtCurrency_1c327a5182399c63873a891da808a068,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["wae_ch"] = func() locales.Translator {
		return &localen{

			locale:                 "wae_CH",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  "’",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "Jen", "Hor", "Mär", "Abr", "Mei", "Brá", "Hei", "Öig", "Her", "Wím", "Win", "Chr"},
			monthsNarrow:           []string{"", "J", "H", "M", "A", "M", "B", "H", "Ö", "H", "W", "W", "C"},
			monthsWide:             []string{"", "Jenner", "Hornig", "Märze", "Abrille", "Meije", "Bráčet", "Heiwet", "Öigšte", "Herbštmánet", "Wímánet", "Wintermánet", "Chrištmánet"},
			daysAbbreviated:        []string{"Sun", "Män", "Ziš", "Mit", "Fró", "Fri", "Sam"},
			daysNarrow:             []string{"S", "M", "Z", "M", "F", "F", "S"},
			daysWide:               []string{"Sunntag", "Mäntag", "Zištag", "Mittwuč", "Fróntag", "Fritag", "Samštag"},
			erasAbbreviated:        []string{"v. Chr.", "n. Chr"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"", ""},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "Atlantiši Summerzit", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "Atlantiši Standardzit", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "Mitteleuropäiši Summerzit", "MEZ": "Mitteleuropäiši Standardzit", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "Ošteuropäiši Summerzit", "OEZ": "Ošteuropäiši Standardzit", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "Wešteuropäiši Summerzit", "WEZ": "Wešteuropäiši Standardzit", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_25be2d04484c3c25e10aab41bfa66a06,
			fnFmtCurrency:         fnFmtCurrency_1c327a5182399c63873a891da808a068,
			fnFmtDateFull:         fnFmtDateFull_7f49a254f89defd9dde5b1348abc8f6a,
			fnFmtDateLong:         fnFmtDateLong_6ddead36eb66c5c353fc1725254f1035,
			fnFmtDateMedium:       fnFmtDateMedium_55a922eee312039812d989655f290012,
			fnFmtDateShort:        fnFmtDateShort_09f1761f0b60aead9f5012476a47c814,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_956780e9690f19a57988254ac2a6eb35,
			fnFmtTimeLong:         fnFmtTimeLong_1a9373b60bc45e855b1c9aec2182a594,
			fnFmtTimeMedium:       fnFmtTimeMedium_8ab08b2a5e55b0d5b394d3d185e15d98,
			fnFmtTimeShort:        fnFmtTimeShort_47b79a38ecd6bac156085539ad031c2d,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["wo"] = func() locales.Translator {
		return &localen{

			locale:                 "wo",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Sam", "Few", "Mar", "Awr", "Mee", "Suw", "Sul", "Ut", "Sàt", "Okt", "Now", "Des"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Samwiyee", "Fewriyee", "Mars", "Awril", "Mee", "Suwe", "Sulet", "Ut", "Sàttumbar", "Oktoobar", "Nowàmbar", "Desàmbar"},
			daysAbbreviated:        []string{"Dib", "Alt", "Tal", "Àla", "Alx", "Àjj", "Ase"},
			daysNarrow:             []string{"Dib", "Alt", "Tal", "Àla", "Alx", "Àjj", "Ase"},
			daysShort:              []string{"Dib", "Alt", "Tal", "Àla", "Alx", "Àjj", "Ase"},
			daysWide:               []string{"Dibéer", "Altine", "Talaata", "Àlarba", "Alxamis", "Àjjuma", "Aseer"},
			periodsAbbreviated:     []string{"Sub", "Ngo"},
			periodsNarrow:          []string{"Sub", "Ngo"},
			periodsWide:            []string{"Sub", "Ngo"},
			erasAbbreviated:        []string{"JC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"av. JC", "AD"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT (waxtu bëccëgu atlàntik)", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST (waxtu estàndaaru penku)", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT (waxtu bëccëgu sàntaraal", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST (waxtu estàndaaru sàntaraal)", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT (waxtu bëccëgu penku)", "EST": "EST (waxtu estàndaaru penku)", "GFT": "GFT", "GMT": "GMT (waxtu Greenwich)", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT (waxtu bëccëgu tundu)", "MESZ": "CEST (waxtu ete wu ëroop sàntaraal)", "MEZ": "CEST (waxtu estàndaaru ëroop sàntaraal)", "MST": "MST (waxtu estàndaaru tundu)", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "EEST (waxtu ete wu ëroop u penku)", "OEZ": "EEST (waxtu estàndaaru ëroop u penku)", "PDT": "PDT (waxtu bëccëgu pasifik)", "PST": "PST (waxtu estàndaaru pasifik)", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WEST (waxtu ete wu ëroop u sowwu-jant)", "WEZ": "WEST (waxtu estàndaaru ëroop u sowwu-jant)", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_4f155c5d948146549d2721e09ed6c3a5,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["wo_sn"] = func() locales.Translator {
		return &localen{

			locale:                 "wo_SN",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  ".",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyNegativePrefix: " ",
			monthsAbbreviated:      []string{"", "Sam", "Few", "Mar", "Awr", "Mee", "Suw", "Sul", "Ut", "Sàt", "Okt", "Now", "Des"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "Samwiyee", "Fewriyee", "Mars", "Awril", "Mee", "Suwe", "Sulet", "Ut", "Sàttumbar", "Oktoobar", "Nowàmbar", "Desàmbar"},
			daysAbbreviated:        []string{"Dib", "Alt", "Tal", "Àla", "Alx", "Àjj", "Ase"},
			daysNarrow:             []string{"Dib", "Alt", "Tal", "Àla", "Alx", "Àjj", "Ase"},
			daysShort:              []string{"Dib", "Alt", "Tal", "Àla", "Alx", "Àjj", "Ase"},
			daysWide:               []string{"Dibéer", "Altine", "Talaata", "Àlarba", "Alxamis", "Àjjuma", "Aseer"},
			periodsAbbreviated:     []string{"Sub", "Ngo"},
			periodsNarrow:          []string{"Sub", "Ngo"},
			periodsWide:            []string{"Sub", "Ngo"},
			erasAbbreviated:        []string{"JC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"av. JC", "AD"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT (waxtu bëccëgu atlàntik)", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST (waxtu estàndaaru penku)", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT (waxtu bëccëgu sàntaraal", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST (waxtu estàndaaru sàntaraal)", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT (waxtu bëccëgu penku)", "EST": "EST (waxtu estàndaaru penku)", "GFT": "GFT", "GMT": "GMT (waxtu Greenwich)", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT (waxtu bëccëgu tundu)", "MESZ": "CEST (waxtu ete wu ëroop sàntaraal)", "MEZ": "CEST (waxtu estàndaaru ëroop sàntaraal)", "MST": "MST (waxtu estàndaaru tundu)", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "EEST (waxtu ete wu ëroop u penku)", "OEZ": "EEST (waxtu estàndaaru ëroop u penku)", "PDT": "PDT (waxtu bëccëgu pasifik)", "PST": "PST (waxtu estàndaaru pasifik)", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WEST (waxtu ete wu ëroop u sowwu-jant)", "WEZ": "WEST (waxtu estàndaaru ëroop u sowwu-jant)", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_91234edbb271d7bc5c1a0b9f346174d7,
			fnFmtCurrency:         fnFmtCurrency_39c53385a3c471ef586d2be3ae8b60b5,
			fnFmtDateFull:         fnFmtDateFull_4f155c5d948146549d2721e09ed6c3a5,
			fnFmtDateLong:         fnFmtDateLong_a234d460b585eebc392634bc08485ebd,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_6e3914bf40627aa17064661ef056b7d6,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["xh"] = func() locales.Translator {
		return &localen{

			locale:            "xh",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			decimal:           ".",
			group:             " ",
			minus:             "-",
			percent:           "%",
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "Jan", "Feb", "Mat", "Epr", "Mey", "Jun", "Jul", "Aga", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:      []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:        []string{"", "Janyuwari", "Februwari", "Matshi", "Epreli", "Meyi", "Juni", "Julayi", "Agasti", "Septemba", "Okthoba", "Novemba", "Disemba"},
			daysAbbreviated:   []string{"Caw", "Mvu", "Bin", "Tha", "Sin", "Hla", "Mgq"},
			daysNarrow:        []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:          []string{"Cawe", "Mvulo", "Lwesibini", "Lwesithathu", "Lwesine", "Lwesihlanu", "Mgqibelo"},
			periodsWide:       []string{"AM", "PM"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"BC", "umnyaka wokuzalwa kukaYesu"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_faa0f0273050790b69e9c65ef1bb23a5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["xh_za"] = func() locales.Translator {
		return &localen{

			locale:            "xh_ZA",
			pluralsCardinal:   []locales.PluralRule{2, 6},
			pluralsOrdinal:    nil,
			pluralsRange:      nil,
			decimal:           ".",
			group:             " ",
			minus:             "-",
			percent:           "%",
			timeSeparator:     ":",
			currencies:        []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated: []string{"", "Jan", "Feb", "Mat", "Epr", "Mey", "Jun", "Jul", "Aga", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:      []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:        []string{"", "Janyuwari", "Februwari", "Matshi", "Epreli", "Meyi", "Juni", "Julayi", "Agasti", "Septemba", "Okthoba", "Novemba", "Disemba"},
			daysAbbreviated:   []string{"Caw", "Mvu", "Bin", "Tha", "Sin", "Hla", "Mgq"},
			daysNarrow:        []string{"S", "M", "T", "W", "T", "F", "S"},
			daysWide:          []string{"Cawe", "Mvulo", "Lwesibini", "Lwesithathu", "Lwesine", "Lwesihlanu", "Mgqibelo"},
			periodsWide:       []string{"AM", "PM"},
			erasAbbreviated:   []string{"BC", "AD"},
			erasNarrow:        []string{"", ""},
			erasWide:          []string{"BC", "umnyaka wokuzalwa kukaYesu"},
			timezones:         map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_faa0f0273050790b69e9c65ef1bb23a5,
			fnFmtCurrency:         fnFmtCurrency_ec1a372391d9fcdfb2e025fe6142dc3b,
			fnFmtDateFull:         fnFmtDateFull_3b6427a32603f02ec4f5128c007220b3,
			fnFmtDateLong:         fnFmtDateLong_4ff37a0cfcd7dfa30a768d81e70d334e,
			fnFmtDateMedium:       fnFmtDateMedium_596f6f3746d292714e1810c0b50dd05c,
			fnFmtDateShort:        fnFmtDateShort_93e0dc80e7e3deca3036cf894d743e0e,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["xog"] = func() locales.Translator {
		return &localen{

			locale:                 "xog",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "USh", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apu", "Maa", "Juu", "Jul", "Agu", "Seb", "Oki", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janwaliyo", "Febwaliyo", "Marisi", "Apuli", "Maayi", "Juuni", "Julaayi", "Agusito", "Sebuttemba", "Okitobba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Sabi", "Bala", "Kubi", "Kusa", "Kuna", "Kuta", "Muka"},
			daysNarrow:             []string{"S", "B", "B", "S", "K", "K", "M"},
			daysWide:               []string{"Sabiiti", "Balaza", "Owokubili", "Owokusatu", "Olokuna", "Olokutaanu", "Olomukaaga"},
			periodsAbbreviated:     []string{"Munkyo", "Eigulo"},
			periodsWide:            []string{"Munkyo", "Eigulo"},
			erasAbbreviated:        []string{"AZ", "AF"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kulisto nga azilawo", "Kulisto nga affile"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["xog_ug"] = func() locales.Translator {
		return &localen{

			locale:                 "xog_UG",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativeSuffix: " ",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mar", "Apu", "Maa", "Juu", "Jul", "Agu", "Seb", "Oki", "Nov", "Des"},
			monthsNarrow:           []string{"", "J", "F", "M", "A", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Janwaliyo", "Febwaliyo", "Marisi", "Apuli", "Maayi", "Juuni", "Julaayi", "Agusito", "Sebuttemba", "Okitobba", "Novemba", "Desemba"},
			daysAbbreviated:        []string{"Sabi", "Bala", "Kubi", "Kusa", "Kuna", "Kuta", "Muka"},
			daysNarrow:             []string{"S", "B", "B", "S", "K", "K", "M"},
			daysWide:               []string{"Sabiiti", "Balaza", "Owokubili", "Owokusatu", "Olokuna", "Olokutaanu", "Olomukaaga"},
			periodsAbbreviated:     []string{"Munkyo", "Eigulo"},
			periodsWide:            []string{"Munkyo", "Eigulo"},
			erasAbbreviated:        []string{"AZ", "AF"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Kulisto nga azilawo", "Kulisto nga affile"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_0758171f702b5b6024bdf47d25d8166d,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_97eb2dc886e17a0500b9446194e6f039,
			fnFmtCurrency:         fnFmtCurrency_bf0b02b2ea7c6ce42329b55a6be0d7f3,
			fnFmtDateFull:         fnFmtDateFull_c568798c9683b61c27e1bae4ea10cb6b,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_529ba7a02bedd437d0833b1afdb768c8,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yav"] = func() locales.Translator {
		return &localen{

			locale:                 "yav",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "o.1", "o.2", "o.3", "o.4", "o.5", "o.6", "o.7", "o.8", "o.9", "o.10", "o.11", "o.12"},
			monthsWide:             []string{"", "pikítíkítie, oólí ú kutúan", "siɛyɛ́, oóli ú kándíɛ", "ɔnsúmbɔl, oóli ú kátátúɛ", "mesiŋ, oóli ú kénie", "ensil, oóli ú kátánuɛ", "ɔsɔn", "efute", "pisuyú", "imɛŋ i puɔs", "imɛŋ i putúk,oóli ú kátíɛ", "makandikɛ", "pilɔndɔ́"},
			daysAbbreviated:        []string{"sd", "md", "mw", "et", "kl", "fl", "ss"},
			daysNarrow:             []string{"s", "m", "m", "e", "k", "f", "s"},
			daysWide:               []string{"sɔ́ndiɛ", "móndie", "muányáŋmóndie", "metúkpíápɛ", "kúpélimetúkpiapɛ", "feléte", "séselé"},
			periodsAbbreviated:     []string{"kiɛmɛ́ɛm", "kisɛ́ndɛ"},
			periodsWide:            []string{"kiɛmɛ́ɛm", "kisɛ́ndɛ"},
			erasAbbreviated:        []string{"k.Y.", "+J.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"katikupíen Yésuse", "ékélémkúnupíén n"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1942e97ed17a8864331f347260c0e63c,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yav_cm"] = func() locales.Translator {
		return &localen{

			locale:                 "yav_CM",
			pluralsCardinal:        nil,
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ",",
			group:                  " ",
			timeSeparator:          ":",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: " ",
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: " )",
			monthsAbbreviated:      []string{"", "o.1", "o.2", "o.3", "o.4", "o.5", "o.6", "o.7", "o.8", "o.9", "o.10", "o.11", "o.12"},
			monthsWide:             []string{"", "pikítíkítie, oólí ú kutúan", "siɛyɛ́, oóli ú kándíɛ", "ɔnsúmbɔl, oóli ú kátátúɛ", "mesiŋ, oóli ú kénie", "ensil, oóli ú kátánuɛ", "ɔsɔn", "efute", "pisuyú", "imɛŋ i puɔs", "imɛŋ i putúk,oóli ú kátíɛ", "makandikɛ", "pilɔndɔ́"},
			daysAbbreviated:        []string{"sd", "md", "mw", "et", "kl", "fl", "ss"},
			daysNarrow:             []string{"s", "m", "m", "e", "k", "f", "s"},
			daysWide:               []string{"sɔ́ndiɛ", "móndie", "muányáŋmóndie", "metúkpíápɛ", "kúpélimetúkpiapɛ", "feléte", "séselé"},
			periodsAbbreviated:     []string{"kiɛmɛ́ɛm", "kisɛ́ndɛ"},
			periodsWide:            []string{"kiɛmɛ́ɛm", "kisɛ́ndɛ"},
			erasAbbreviated:        []string{"k.Y.", "+J.C."},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"katikupíen Yésuse", "ékélémkúnupíén n"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_1942e97ed17a8864331f347260c0e63c,
			fnFmtCurrency:         fnFmtCurrency_3398e8f49a5a712d32672f16a1ed2c57,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_cb43e8c32d5cd0a27ac6db293a7981fd,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_ceab7ec754f7f75fd13d22457c77cbf9,
			fnFmtPercent:          fnFmtPercent_a2e6b1aef5e145e790a5623d4a4e442c,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yi"] = func() locales.Translator {
		return &localen{

			locale:                 "yi",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "יאַנואַר", "פֿעברואַר", "מערץ", "אַפּריל", "מיי", "יוני", "יולי", "אויגוסט", "סעפּטעמבער", "אקטאבער", "נאוועמבער", "דעצעמבער"},
			monthsWide:             []string{"", "יאַנואַר", "פֿעברואַר", "מערץ", "אַפּריל", "מיי", "יוני", "יולי", "אויגוסט", "סעפּטעמבער", "אקטאבער", "נאוועמבער", "דעצעמבער"},
			daysAbbreviated:        []string{"זונטיק", "מאָנטיק", "דינסטיק", "מיטוואך", "דאנערשטיק", "פֿרײַטיק", "שבת"},
			daysShort:              []string{"זונטיק", "מאָנטיק", "דינסטיק", "מיטוואך", "דאנערשטיק", "פֿרײַטיק", "שבת"},
			daysWide:               []string{"זונטיק", "מאָנטיק", "דינסטיק", "מיטוואך", "דאנערשטיק", "פֿרײַטיק", "שבת"},
			periodsAbbreviated:     []string{"פֿאַרמיטאָג", "נאָכמיטאָג"},
			periodsWide:            []string{"פֿאַרמיטאָג", "נאָכמיטאָג"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d0ef2c4a76852adc59df517a99a6000f,
			fnFmtCurrency:         fnFmtCurrency_9b06328888547fc535e862da8ee049d8,
			fnFmtDateFull:         fnFmtDateFull_533def3ea33c7c5e7121653c4a4fcdfb,
			fnFmtDateLong:         fnFmtDateLong_ed7a2ed783c0a089bae0b26415cf1eec,
			fnFmtDateMedium:       fnFmtDateMedium_49f81df3f568828be8ec7cbfdfca6c87,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yi_001"] = func() locales.Translator {
		return &localen{

			locale:                 "yi_001",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositivePrefix: " ",
			currencyPositiveSuffix: "K",
			currencyNegativePrefix: " ",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "יאַנואַר", "פֿעברואַר", "מערץ", "אַפּריל", "מיי", "יוני", "יולי", "אויגוסט", "סעפּטעמבער", "אקטאבער", "נאוועמבער", "דעצעמבער"},
			monthsWide:             []string{"", "יאַנואַר", "פֿעברואַר", "מערץ", "אַפּריל", "מיי", "יוני", "יולי", "אויגוסט", "סעפּטעמבער", "אקטאבער", "נאוועמבער", "דעצעמבער"},
			daysAbbreviated:        []string{"זונטיק", "מאָנטיק", "דינסטיק", "מיטוואך", "דאנערשטיק", "פֿרײַטיק", "שבת"},
			daysShort:              []string{"זונטיק", "מאָנטיק", "דינסטיק", "מיטוואך", "דאנערשטיק", "פֿרײַטיק", "שבת"},
			daysWide:               []string{"זונטיק", "מאָנטיק", "דינסטיק", "מיטוואך", "דאנערשטיק", "פֿרײַטיק", "שבת"},
			periodsAbbreviated:     []string{"פֿאַרמיטאָג", "נאָכמיטאָג"},
			periodsWide:            []string{"פֿאַרמיטאָג", "נאָכמיטאָג"},
			timezones:              map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_c037ba21b7495f414cf38823ca036d30,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_d0ef2c4a76852adc59df517a99a6000f,
			fnFmtCurrency:         fnFmtCurrency_9b06328888547fc535e862da8ee049d8,
			fnFmtDateFull:         fnFmtDateFull_533def3ea33c7c5e7121653c4a4fcdfb,
			fnFmtDateLong:         fnFmtDateLong_ed7a2ed783c0a089bae0b26415cf1eec,
			fnFmtDateMedium:       fnFmtDateMedium_49f81df3f568828be8ec7cbfdfca6c87,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_2816fdab36e719be095a370ff42e599f,
			fnFmtPercent:          fnFmtPercent_016773d83796aa19d80b7c33bf377883,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_1edc513af8c7151dc8fe1f42c7d497ba,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yo"] = func() locales.Translator {
		return &localen{

			locale:                 "yo",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "₦", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "₽", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ṣẹ́r", "Èrèl", "Ẹrẹ̀n", "Ìgb", "Ẹ̀bi", "Òkú", "Agẹ", "Ògú", "Owe", "Ọ̀wà", "Bél", "Ọ̀pẹ"},
			monthsNarrow:           []string{"", "S", "È", "Ẹ", "Ì", "Ẹ̀", "Ò", "A", "Ò", "O", "Ọ̀", "B", "Ọ̀"},
			monthsWide:             []string{"", "Oṣù Ṣẹ́rẹ́", "Oṣù Èrèlè", "Oṣù Ẹrẹ̀nà", "Oṣù Ìgbé", "Oṣù Ẹ̀bibi", "Oṣù Òkúdu", "Oṣù Agẹmọ", "Oṣù Ògún", "Oṣù Owewe", "Oṣù Ọ̀wàrà", "Oṣù Bélú", "Oṣù Ọ̀pẹ̀"},
			daysAbbreviated:        []string{"Àìk", "Aj", "Ìsẹ́g", "Ọjọ́r", "Ọjọ́b", "Ẹt", "Àbám"},
			daysNarrow:             []string{"À", "A", "Ì", "Ọ", "Ọ", "Ẹ", "À"},
			daysShort:              []string{"Àìkú", "Ajé", "Ìsẹ́gun", "Ọjọ́rú", "Ọjọ́bọ", "Ẹtì", "Àbámẹ́ta"},
			daysWide:               []string{"Ọjọ́ Àìkú", "Ọjọ́ Ajé", "Ọjọ́ Ìsẹ́gun", "Ọjọ́rú", "Ọjọ́bọ", "Ọjọ́ Ẹtì", "Ọjọ́ Àbámẹ́ta"},
			periodsAbbreviated:     []string{"Àárọ̀", "Ọ̀sán"},
			periodsNarrow:          []string{"Àárọ̀", "Ọ̀sán"},
			periodsWide:            []string{"Àárọ̀", "Ọ̀sán"},
			erasAbbreviated:        []string{"BCE", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Saju Kristi", "Lehin Kristi"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Àkókò Ìyálẹta Àtìláńtíìkì", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Àkókò àsikò Àtìláńtíìkì", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Akókò àárín gbùngbùn ojúmọmọ", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "àkókò asiko àárín gbùngbùn", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Àkókò ojúmọmọ Ìhà Ìlà Oòrun", "EST": "Akókò Àsikò Ìha Ìla Oòrùn", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "Guyana Time", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "MDT", "MESZ": "Àkókò Àárin Sọmà Europe", "MEZ": "Àkókò Àárin àsikò Europe", "MST": "MST", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Àkókò Sọmà Ìha Ìlà Oòrùn Europe", "OEZ": "Àkókò àsikò Ìhà Ìlà Oòrùn Europe", "PDT": "Àkókò Ìyálẹta Pàsífíìkì", "PST": "Àkókò àsikò Pàsífíìkì", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Àkókò Sọmà Ìhà Ìwọ Oòrùn Europe", "WEZ": "Àkókò àsikò Ìwọ Oòrùn Europe", "WIB": "WIB", "WIT": "Eastern Indonesia Time", "WITA": "WITA", "∅∅∅": "Peru Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_ea32054fb13476ec6d17746ec03afe87,
			fnFmtDateLong:         fnFmtDateLong_39c28d689b0bc45d7cd36c6f85c85cd7,
			fnFmtDateMedium:       fnFmtDateMedium_5ee989ce4f98a493eab4c7b536fc1c06,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_f393bba636806dbcfe262f3337c404b9,
			fnFmtTimeShort:        fnFmtTimeShort_e9e37e884fba601b3fce4f8400a30c6f,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yo_bj"] = func() locales.Translator {
		return &localen{

			locale:                 "yo_BJ",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Shɛ́r", "Èrèl", "Ɛrɛ̀n", "Ìgb", "Ɛ̀bi", "Òkú", "Agɛ", "Ògú", "Owe", "Ɔ̀wà", "Bél", "Ɔ̀pɛ"},
			monthsNarrow:           []string{"", "S", "È", "Ɛ", "Ì", "Ɛ̀", "Ò", "A", "Ò", "O", "Ɔ̀", "B", "Ɔ̀"},
			monthsWide:             []string{"", "Oshù Shɛ́rɛ́", "Oshù Èrèlè", "Oshù Ɛrɛ̀nà", "Oshù Ìgbé", "Oshù Ɛ̀bibi", "Oshù Òkúdu", "Oshù Agɛmɔ", "Oshù Ògún", "Oshù Owewe", "Oshù Ɔ̀wàrà", "Oshù Bélú", "Oshù Ɔ̀pɛ̀"},
			daysAbbreviated:        []string{"Àìk", "Aj", "Ìsɛ́g", "Ɔjɔ́r", "Ɔjɔ́b", "Ɛt", "Àbám"},
			daysNarrow:             []string{"À", "A", "Ì", "Ɔ", "Ɔ", "Ɛ", "À"},
			daysShort:              []string{"Àìkú", "Ajé", "Ìsɛ́gun", "Ɔjɔ́rú", "Ɔjɔ́bɔ", "Ɛtì", "Àbámɛ́ta"},
			daysWide:               []string{"Ɔjɔ́ Àìkú", "Ɔjɔ́ Ajé", "Ɔjɔ́ Ìsɛ́gun", "Ɔjɔ́rú", "Ɔjɔ́bɔ", "Ɔjɔ́ Ɛtì", "Ɔjɔ́ Àbámɛ́ta"},
			periodsAbbreviated:     []string{"Àárɔ̀", "Ɔ̀sán"},
			periodsNarrow:          []string{"Àárɔ̀", "Ɔ̀sán"},
			periodsWide:            []string{"Àárɔ̀", "Ɔ̀sán"},
			erasAbbreviated:        []string{"BCE", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Saju Kristi", "Lehin Kristi"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Àkókò Ìyálɛta Àtìláńtíìkì", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Àkókò àsikò Àtìláńtíìkì", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Akókò àárín gbùngbùn ojúmɔmɔ", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "àkókò asiko àárín gbùngbùn", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Àkókò ojúmɔmɔ Ìhà Ìlà Oòrun", "EST": "Akókò Àsikò Ìha Ìla Oòrùn", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "Guyana Time", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Àkókò ojúmɔmɔ Ori-òkè", "MESZ": "Àkókò Àárin Sɔmà Europe", "MEZ": "Àkókò Àárin àsikò Europe", "MST": "Àkókò asiko òkè", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Àkókò Sɔmà Ìha Ìlà Oòrùn Europe", "OEZ": "Àkókò àsikò Ìhà Ìlà Oòrùn Europe", "PDT": "Àkókò Ìyálɛta Pàsífíìkì", "PST": "Àkókò àsikò Pàsífíìkì", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Àkókò Sɔmà Ìhà Ìwɔ Oòrùn Europe", "WEZ": "Àkókò àsikò Ìwɔ Oòrùn Europe", "WIB": "WIB", "WIT": "Eastern Indonesia Time", "WITA": "WITA", "∅∅∅": "Brasilia Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_ea32054fb13476ec6d17746ec03afe87,
			fnFmtDateLong:         fnFmtDateLong_39c28d689b0bc45d7cd36c6f85c85cd7,
			fnFmtDateMedium:       fnFmtDateMedium_5ee989ce4f98a493eab4c7b536fc1c06,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_f393bba636806dbcfe262f3337c404b9,
			fnFmtTimeShort:        fnFmtTimeShort_e9e37e884fba601b3fce4f8400a30c6f,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yo_ng"] = func() locales.Translator {
		return &localen{

			locale:                 "yo_NG",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         nil,
			pluralsRange:           nil,
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Ṣẹ́r", "Èrèl", "Ẹrẹ̀n", "Ìgb", "Ẹ̀bi", "Òkú", "Agẹ", "Ògú", "Owe", "Ọ̀wà", "Bél", "Ọ̀pẹ"},
			monthsNarrow:           []string{"", "S", "È", "Ẹ", "Ì", "Ẹ̀", "Ò", "A", "Ò", "O", "Ọ̀", "B", "Ọ̀"},
			monthsWide:             []string{"", "Oṣù Ṣẹ́rẹ́", "Oṣù Èrèlè", "Oṣù Ẹrẹ̀nà", "Oṣù Ìgbé", "Oṣù Ẹ̀bibi", "Oṣù Òkúdu", "Oṣù Agẹmọ", "Oṣù Ògún", "Oṣù Owewe", "Oṣù Ọ̀wàrà", "Oṣù Bélú", "Oṣù Ọ̀pẹ̀"},
			daysAbbreviated:        []string{"Àìk", "Aj", "Ìsẹ́g", "Ọjọ́r", "Ọjọ́b", "Ẹt", "Àbám"},
			daysNarrow:             []string{"À", "A", "Ì", "Ọ", "Ọ", "Ẹ", "À"},
			daysShort:              []string{"Àìkú", "Ajé", "Ìsẹ́gun", "Ọjọ́rú", "Ọjọ́bọ", "Ẹtì", "Àbámẹ́ta"},
			daysWide:               []string{"Ọjọ́ Àìkú", "Ọjọ́ Ajé", "Ọjọ́ Ìsẹ́gun", "Ọjọ́rú", "Ọjọ́bọ", "Ọjọ́ Ẹtì", "Ọjọ́ Àbámẹ́ta"},
			periodsAbbreviated:     []string{"Àárọ̀", "Ọ̀sán"},
			periodsNarrow:          []string{"Àárọ̀", "Ọ̀sán"},
			periodsWide:            []string{"Àárọ̀", "Ọ̀sán"},
			erasAbbreviated:        []string{"BCE", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"Saju Kristi", "Lehin Kristi"},
			timezones:              map[string]string{"ACDT": "Australian Central Daylight Time", "ACST": "Australian Central Standard Time", "ACWDT": "Australian Central Western Daylight Time", "ACWST": "Australian Central Western Standard Time", "ADT": "Àkókò Ìyálẹta Àtìláńtíìkì", "AEDT": "Australian Eastern Daylight Time", "AEST": "Australian Eastern Standard Time", "AKDT": "Alaska Daylight Time", "AKST": "Alaska Standard Time", "ARST": "Argentina Summer Time", "ART": "Argentina Standard Time", "AST": "Àkókò àsikò Àtìláńtíìkì", "AWDT": "Australian Western Daylight Time", "AWST": "Australian Western Standard Time", "BOT": "Bolivia Time", "BT": "Bhutan Time", "CAT": "Central Africa Time", "CDT": "Akókò àárín gbùngbùn ojúmọmọ", "CHADT": "Chatham Daylight Time", "CHAST": "Chatham Standard Time", "CLST": "Chile Summer Time", "CLT": "Chile Standard Time", "COST": "Colombia Summer Time", "COT": "Colombia Standard Time", "CST": "àkókò asiko àárín gbùngbùn", "ChST": "Chamorro Standard Time", "EAT": "East Africa Time", "ECT": "Ecuador Time", "EDT": "Àkókò ojúmọmọ Ìhà Ìlà Oòrun", "EST": "Akókò Àsikò Ìha Ìla Oòrùn", "GFT": "French Guiana Time", "GMT": "Greenwich Mean Time", "GST": "Gulf Standard Time [translation hint: translate as just \"Gulf Time\"]", "GYT": "Guyana Time", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "Hong Kong Summer Time", "HKT": "Hong Kong Standard Time", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "India Standard Time", "JDT": "Japan Daylight Time", "JST": "Japan Standard Time", "LHDT": "Lord Howe Daylight Time", "LHST": "Lord Howe Standard Time", "MDT": "Àkókò ojúmọmọ Ori-òkè", "MESZ": "Àkókò Àárin Sọmà Europe", "MEZ": "Àkókò Àárin àsikò Europe", "MST": "Àkókò asiko òkè", "MYT": "Malaysia Time", "NZDT": "New Zealand Daylight Time", "NZST": "New Zealand Standard Time", "OESZ": "Àkókò Sọmà Ìha Ìlà Oòrùn Europe", "OEZ": "Àkókò àsikò Ìhà Ìlà Oòrùn Europe", "PDT": "Àkókò Ìyálẹta Pàsífíìkì", "PST": "Àkókò àsikò Pàsífíìkì", "SAST": "South Africa Standard Time", "SGT": "Singapore Standard Time", "SRT": "Suriname Time", "TMST": "Turkmenistan Summer Time", "TMT": "Turkmenistan Standard Time", "UYST": "Uruguay Summer Time", "UYT": "Uruguay Standard Time", "VET": "Venezuela Time", "WARST": "Western Argentina Summer Time", "WART": "Western Argentina Standard Time", "WAST": "West Africa Summer Time", "WAT": "West Africa Standard Time", "WESZ": "Àkókò Sọmà Ìhà Ìwọ Oòrùn Europe", "WEZ": "Àkókò àsikò Ìwọ Oòrùn Europe", "WIB": "WIB", "WIT": "Eastern Indonesia Time", "WITA": "WITA", "∅∅∅": "Azores Summer Time"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_ea32054fb13476ec6d17746ec03afe87,
			fnFmtDateLong:         fnFmtDateLong_39c28d689b0bc45d7cd36c6f85c85cd7,
			fnFmtDateMedium:       fnFmtDateMedium_5ee989ce4f98a493eab4c7b536fc1c06,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_14ec37cb9b88b34ab642f7ee8fd2800b,
			fnFmtTimeMedium:       fnFmtTimeMedium_f393bba636806dbcfe262f3337c404b9,
			fnFmtTimeShort:        fnFmtTimeShort_e9e37e884fba601b3fce4f8400a30c6f,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yue"] = func() locales.Translator {
		return &localen{

			locale:             "yue",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "￦", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:    []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"日", "一", "二", "三", "四", "五", "六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"西元前", "西元"},
			erasNarrow:         []string{"西元前", "西元"},
			erasWide:           []string{"西元前", "西元"},
			timezones:          map[string]string{"ACDT": "澳洲中部夏令時間", "ACST": "澳洲中部標準時間", "ACWDT": "澳洲中西部夏令時間", "ACWST": "澳洲中西部標準時間", "ADT": "大西洋夏令時間", "AEDT": "澳洲東部夏令時間", "AEST": "澳洲東部標準時間", "AKDT": "阿拉斯加夏令時間", "AKST": "阿拉斯加標準時間", "ARST": "阿根廷夏令時間", "ART": "阿根廷標準時間", "AST": "大西洋標準時間", "AWDT": "澳洲西部夏令時間", "AWST": "澳洲西部標準時間", "BOT": "玻利維亞時間", "BT": "不丹時間", "CAT": "中非時間", "CDT": "中部夏令時間", "CHADT": "查坦群島夏令時間", "CHAST": "查坦群島標準時間", "CLST": "智利夏令時間", "CLT": "智利標準時間", "COST": "哥倫比亞夏令時間", "COT": "哥倫比亞標準時間", "CST": "中部標準時間", "ChST": "查莫洛時間", "EAT": "東非時間", "ECT": "厄瓜多時間", "EDT": "東部夏令時間", "EST": "東部標準時間", "GFT": "法屬圭亞那時間", "GMT": "格林威治標準時間", "GST": "波斯灣海域標準時間", "GYT": "蓋亞那時間", "HADT": "夏威夷-阿留申夏令時間", "HAST": "夏威夷-阿留申標準時間", "HAT": "紐芬蘭夏令時間", "HECU": "古巴夏令時間", "HEEG": "格陵蘭東部夏令時間", "HENOMX": "墨西哥西北部夏令時間", "HEOG": "格陵蘭西部夏令時間", "HEPM": "聖皮埃爾和密克隆群島夏令時間", "HEPMX": "墨西哥太平洋夏令時間", "HKST": "香港夏令時間", "HKT": "香港標準時間", "HNCU": "古巴標準時間", "HNEG": "格陵蘭東部標準時間", "HNNOMX": "墨西哥西北部標準時間", "HNOG": "格陵蘭西部標準時間", "HNPM": "聖皮埃爾和密克隆群島標準時間", "HNPMX": "墨西哥太平洋標準時間", "HNT": "紐芬蘭標準時間", "IST": "印度標準時間", "JDT": "日本夏令時間", "JST": "日本標準時間", "LHDT": "豪勳爵島夏令時間", "LHST": "豪勳爵島標準時間", "MDT": "山區夏令時間", "MESZ": "中歐夏令時間", "MEZ": "中歐標準時間", "MST": "山區標準時間", "MYT": "馬來西亞時間", "NZDT": "紐西蘭夏令時間", "NZST": "紐西蘭標準時間", "OESZ": "東歐夏令時間", "OEZ": "東歐標準時間", "PDT": "太平洋夏令時間", "PST": "太平洋標準時間", "SAST": "南非標準時間", "SGT": "新加坡標準時間", "SRT": "蘇利南時間", "TMST": "土庫曼夏令時間", "TMT": "土庫曼標準時間", "UYST": "烏拉圭夏令時間", "UYT": "烏拉圭標準時間", "VET": "委內瑞拉時間", "WARST": "阿根廷西部夏令時間", "WART": "阿根廷西部標準時間", "WAST": "西非夏令時間", "WAT": "西非標準時間", "WESZ": "西歐夏令時間", "WEZ": "西歐標準時間", "WIB": "印尼西部時間", "WIT": "印尼東部時間", "WITA": "印尼中部時間", "∅∅∅": "亞速爾群島夏令時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yue_hk"] = func() locales.Translator {
		return &localen{

			locale:             "yue_HK",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UZS", "VEB", "VEF", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:    []string{"週日", "週一", "週二", "週三", "週四", "週五", "週六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"日", "一", "二", "三", "四", "五", "六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"西元前", "西元"},
			erasNarrow:         []string{"西元前", "西元"},
			erasWide:           []string{"西元前", "西元"},
			timezones:          map[string]string{"HENOMX": "墨西哥西北部夏令時間", "LHDT": "豪勳爵島夏令時間", "GMT": "格林威治標準時間", "HNNOMX": "墨西哥西北部標準時間", "ARST": "阿根廷夏令時間", "BOT": "玻利維亞時間", "CHAST": "查坦群島標準時間", "WARST": "阿根廷西部夏令時間", "CLT": "智利標準時間", "HNT": "紐芬蘭標準時間", "AKST": "阿拉斯加標準時間", "HEPM": "聖皮埃爾和密克隆群島夏令時間", "NZST": "紐西蘭標準時間", "TMT": "土庫曼標準時間", "ART": "阿根廷標準時間", "OEZ": "東歐標準時間", "IST": "印度標準時間", "ADT": "大西洋夏令時間", "WAT": "西非標準時間", "CLST": "智利夏令時間", "ACWDT": "澳洲中西部夏令時間", "HAST": "夏威夷-阿留申標準時間", "HADT": "夏威夷-阿留申夏令時間", "WEZ": "西歐標準時間", "EDT": "東部夏令時間", "ACST": "澳洲中部標準時間", "ACDT": "澳洲中部夏令時間", "HNPMX": "墨西哥太平洋標準時間", "HEPMX": "墨西哥太平洋夏令時間", "CHADT": "查坦群島夏令時間", "BT": "不丹時間", "WIT": "印尼東部時間", "TMST": "土庫曼夏令時間", "OESZ": "東歐夏令時間", "CST": "中部標準時間", "MEZ": "中歐標準時間", "GYT": "蓋亞那時間", "WESZ": "西歐夏令時間", "EAT": "東非時間", "AKDT": "阿拉斯加夏令時間", "ChST": "查莫洛時間", "HNPM": "聖皮埃爾和密克隆群島標準時間", "MESZ": "中歐夏令時間", "MDT": "山區夏令時間", "AST": "大西洋標準時間", "LHST": "豪勳爵島標準時間", "∅∅∅": "亞速爾群島夏令時間", "AEDT": "澳洲東部夏令時間", "HEEG": "格陵蘭東部夏令時間", "ECT": "厄瓜多時間", "UYT": "烏拉圭標準時間", "NZDT": "紐西蘭夏令時間", "WART": "阿根廷西部標準時間", "WIB": "印尼西部時間", "HNCU": "古巴標準時間", "AWDT": "澳洲西部夏令時間", "HNEG": "格陵蘭東部標準時間", "SAST": "南非標準時間", "COT": "哥倫比亞標準時間", "GFT": "法屬圭亞那時間", "HAT": "紐芬蘭夏令時間", "ACWST": "澳洲中西部標準時間", "MYT": "馬來西亞時間", "HEOG": "格陵蘭西部夏令時間", "CDT": "中部夏令時間", "HNOG": "格陵蘭西部標準時間", "COST": "哥倫比亞夏令時間", "JST": "日本標準時間", "HKT": "香港標準時間", "SGT": "新加坡標準時間", "PDT": "太平洋夏令時間", "SRT": "蘇利南時間", "VET": "委內瑞拉時間", "JDT": "日本夏令時間", "EST": "東部標準時間", "HECU": "古巴夏令時間", "AWST": "澳洲西部標準時間", "UYST": "烏拉圭夏令時間", "WITA": "印尼中部時間", "WAST": "西非夏令時間", "PST": "太平洋標準時間", "AEST": "澳洲東部標準時間", "HKST": "香港夏令時間", "CAT": "中非時間", "MST": "山區標準時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yue_hans"] = func() locales.Translator {
		return &localen{

			locale:             "yue_Hans",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "￥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "￦", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"西元前", "西元"},
			erasNarrow:         []string{"西元前", "西元"},
			erasWide:           []string{"西元前", "西元"},
			timezones:          map[string]string{"ACDT": "澳洲中部夏令时间", "ACST": "澳洲中部标准时间", "ACWDT": "澳洲中西部夏令时间", "ACWST": "澳洲中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳洲东部夏令时间", "AEST": "澳洲东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳洲西部夏令时间", "AWST": "澳洲西部标准时间", "BOT": "玻利维亚时间", "BT": "不丹时间", "CAT": "中非时间", "CDT": "中部夏令时间", "CHADT": "查坦群岛夏令时间", "CHAST": "查坦群岛标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "中部标准时间", "ChST": "查莫洛时间", "EAT": "东非时间", "ECT": "厄瓜多时间", "EDT": "东部夏令时间", "EST": "东部标准时间", "GFT": "法属圭亚那时间", "GMT": "格林威治标准时间", "GST": "波斯湾海域标准时间", "GYT": "盖亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度标准时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "山区标准时间", "MYT": "马来西亚时间", "NZDT": "纽西兰夏令时间", "NZST": "纽西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "太平洋夏令时间", "PST": "太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏利南时间", "TMST": "土库曼夏令时间", "TMT": "土库曼标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西非夏令时间", "WAT": "西非标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印尼西部时间", "WIT": "印尼东部时间", "WITA": "印尼中部时间", "∅∅∅": "亚速尔群岛夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yue_hans_cn"] = func() locales.Translator {
		return &localen{

			locale:             "yue_Hans_CN",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:    []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"日", "一", "二", "三", "四", "五", "六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"西元前", "西元"},
			erasNarrow:         []string{"西元前", "西元"},
			erasWide:           []string{"西元前", "西元"},
			timezones:          map[string]string{"ACDT": "澳洲中部夏令時間", "ACST": "澳洲中部標準時間", "ACWDT": "澳洲中西部夏令時間", "ACWST": "澳洲中西部標準時間", "ADT": "大西洋夏令時間", "AEDT": "澳洲東部夏令時間", "AEST": "澳洲東部標準時間", "AKDT": "阿拉斯加夏令時間", "AKST": "阿拉斯加標準時間", "ARST": "阿根廷夏令時間", "ART": "阿根廷標準時間", "AST": "大西洋標準時間", "AWDT": "澳洲西部夏令時間", "AWST": "澳洲西部標準時間", "BOT": "玻利維亞時間", "BT": "不丹時間", "CAT": "中非時間", "CDT": "中部夏令時間", "CHADT": "查坦群島夏令時間", "CHAST": "查坦群島標準時間", "CLST": "智利夏令時間", "CLT": "智利標準時間", "COST": "哥倫比亞夏令時間", "COT": "哥倫比亞標準時間", "CST": "中部標準時間", "ChST": "查莫洛時間", "EAT": "東非時間", "ECT": "厄瓜多時間", "EDT": "東部夏令時間", "EST": "東部標準時間", "GFT": "法屬圭亞那時間", "GMT": "格林威治標準時間", "GST": "波斯灣海域標準時間", "GYT": "蓋亞那時間", "HADT": "夏威夷-阿留申夏令時間", "HAST": "夏威夷-阿留申標準時間", "HAT": "紐芬蘭夏令時間", "HECU": "古巴夏令時間", "HEEG": "格陵蘭東部夏令時間", "HENOMX": "墨西哥西北部夏令時間", "HEOG": "格陵蘭西部夏令時間", "HEPM": "聖皮埃爾和密克隆群島夏令時間", "HEPMX": "墨西哥太平洋夏令時間", "HKST": "香港夏令時間", "HKT": "香港標準時間", "HNCU": "古巴標準時間", "HNEG": "格陵蘭東部標準時間", "HNNOMX": "墨西哥西北部標準時間", "HNOG": "格陵蘭西部標準時間", "HNPM": "聖皮埃爾和密克隆群島標準時間", "HNPMX": "墨西哥太平洋標準時間", "HNT": "紐芬蘭標準時間", "IST": "印度標準時間", "JDT": "日本夏令時間", "JST": "日本標準時間", "LHDT": "豪勳爵島夏令時間", "LHST": "豪勳爵島標準時間", "MDT": "山區夏令時間", "MESZ": "中歐夏令時間", "MEZ": "中歐標準時間", "MST": "山區標準時間", "MYT": "馬來西亞時間", "NZDT": "紐西蘭夏令時間", "NZST": "紐西蘭標準時間", "OESZ": "東歐夏令時間", "OEZ": "東歐標準時間", "PDT": "太平洋夏令時間", "PST": "太平洋標準時間", "SAST": "南非標準時間", "SGT": "新加坡標準時間", "SRT": "蘇利南時間", "TMST": "土庫曼夏令時間", "TMT": "土庫曼標準時間", "UYST": "烏拉圭夏令時間", "UYT": "烏拉圭標準時間", "VET": "委內瑞拉時間", "WARST": "阿根廷西部夏令時間", "WART": "阿根廷西部標準時間", "WAST": "西非夏令時間", "WAT": "西非標準時間", "WESZ": "西歐夏令時間", "WEZ": "西歐標準時間", "WIB": "印尼西部時間", "WIT": "印尼東部時間", "WITA": "印尼中部時間", "∅∅∅": "亞速爾群島夏令時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yue_hant"] = func() locales.Translator {
		return &localen{

			locale:             "yue_Hant",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:    []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"日", "一", "二", "三", "四", "五", "六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"西元前", "西元"},
			erasNarrow:         []string{"西元前", "西元"},
			erasWide:           []string{"西元前", "西元"},
			timezones:          map[string]string{"ACDT": "澳洲中部夏令時間", "ACST": "澳洲中部標準時間", "ACWDT": "澳洲中西部夏令時間", "ACWST": "澳洲中西部標準時間", "ADT": "大西洋夏令時間", "AEDT": "澳洲東部夏令時間", "AEST": "澳洲東部標準時間", "AKDT": "阿拉斯加夏令時間", "AKST": "阿拉斯加標準時間", "ARST": "阿根廷夏令時間", "ART": "阿根廷標準時間", "AST": "大西洋標準時間", "AWDT": "澳洲西部夏令時間", "AWST": "澳洲西部標準時間", "BOT": "玻利維亞時間", "BT": "不丹時間", "CAT": "中非時間", "CDT": "中部夏令時間", "CHADT": "查坦群島夏令時間", "CHAST": "查坦群島標準時間", "CLST": "智利夏令時間", "CLT": "智利標準時間", "COST": "哥倫比亞夏令時間", "COT": "哥倫比亞標準時間", "CST": "中部標準時間", "ChST": "查莫洛時間", "EAT": "東非時間", "ECT": "厄瓜多時間", "EDT": "東部夏令時間", "EST": "東部標準時間", "GFT": "法屬圭亞那時間", "GMT": "格林威治標準時間", "GST": "波斯灣海域標準時間", "GYT": "蓋亞那時間", "HADT": "夏威夷-阿留申夏令時間", "HAST": "夏威夷-阿留申標準時間", "HAT": "紐芬蘭夏令時間", "HECU": "古巴夏令時間", "HEEG": "格陵蘭東部夏令時間", "HENOMX": "墨西哥西北部夏令時間", "HEOG": "格陵蘭西部夏令時間", "HEPM": "聖皮埃爾和密克隆群島夏令時間", "HEPMX": "墨西哥太平洋夏令時間", "HKST": "香港夏令時間", "HKT": "香港標準時間", "HNCU": "古巴標準時間", "HNEG": "格陵蘭東部標準時間", "HNNOMX": "墨西哥西北部標準時間", "HNOG": "格陵蘭西部標準時間", "HNPM": "聖皮埃爾和密克隆群島標準時間", "HNPMX": "墨西哥太平洋標準時間", "HNT": "紐芬蘭標準時間", "IST": "印度標準時間", "JDT": "日本夏令時間", "JST": "日本標準時間", "LHDT": "豪勳爵島夏令時間", "LHST": "豪勳爵島標準時間", "MDT": "澳門夏令時間", "MESZ": "中歐夏令時間", "MEZ": "中歐標準時間", "MST": "澳門標準時間", "MYT": "馬來西亞時間", "NZDT": "紐西蘭夏令時間", "NZST": "紐西蘭標準時間", "OESZ": "東歐夏令時間", "OEZ": "東歐標準時間", "PDT": "太平洋夏令時間", "PST": "太平洋標準時間", "SAST": "南非標準時間", "SGT": "新加坡標準時間", "SRT": "蘇利南時間", "TMST": "土庫曼夏令時間", "TMT": "土庫曼標準時間", "UYST": "烏拉圭夏令時間", "UYT": "烏拉圭標準時間", "VET": "委內瑞拉時間", "WARST": "阿根廷西部夏令時間", "WART": "阿根廷西部標準時間", "WAST": "西非夏令時間", "WAT": "西非標準時間", "WESZ": "西歐夏令時間", "WEZ": "西歐標準時間", "WIB": "印尼西部時間", "WIT": "印尼東部時間", "WITA": "印尼中部時間", "∅∅∅": "秘魯夏令時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["yue_hant_hk"] = func() locales.Translator {
		return &localen{

			locale:             "yue_Hant_HK",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:    []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"日", "一", "二", "三", "四", "五", "六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"西元前", "西元"},
			erasNarrow:         []string{"西元前", "西元"},
			erasWide:           []string{"西元前", "西元"},
			timezones:          map[string]string{"ACDT": "澳洲中部夏令時間", "ACST": "澳洲中部標準時間", "ACWDT": "澳洲中西部夏令時間", "ACWST": "澳洲中西部標準時間", "ADT": "大西洋夏令時間", "AEDT": "澳洲東部夏令時間", "AEST": "澳洲東部標準時間", "AKDT": "阿拉斯加夏令時間", "AKST": "阿拉斯加標準時間", "ARST": "阿根廷夏令時間", "ART": "阿根廷標準時間", "AST": "大西洋標準時間", "AWDT": "澳洲西部夏令時間", "AWST": "澳洲西部標準時間", "BOT": "玻利維亞時間", "BT": "不丹時間", "CAT": "中非時間", "CDT": "中部夏令時間", "CHADT": "查坦群島夏令時間", "CHAST": "查坦群島標準時間", "CLST": "智利夏令時間", "CLT": "智利標準時間", "COST": "哥倫比亞夏令時間", "COT": "哥倫比亞標準時間", "CST": "中部標準時間", "ChST": "查莫洛時間", "EAT": "東非時間", "ECT": "厄瓜多時間", "EDT": "東部夏令時間", "EST": "東部標準時間", "GFT": "法屬圭亞那時間", "GMT": "格林威治標準時間", "GST": "波斯灣海域標準時間", "GYT": "蓋亞那時間", "HADT": "夏威夷-阿留申夏令時間", "HAST": "夏威夷-阿留申標準時間", "HAT": "紐芬蘭夏令時間", "HECU": "古巴夏令時間", "HEEG": "格陵蘭東部夏令時間", "HENOMX": "墨西哥西北部夏令時間", "HEOG": "格陵蘭西部夏令時間", "HEPM": "聖皮埃爾和密克隆群島夏令時間", "HEPMX": "墨西哥太平洋夏令時間", "HKST": "香港夏令時間", "HKT": "香港標準時間", "HNCU": "古巴標準時間", "HNEG": "格陵蘭東部標準時間", "HNNOMX": "墨西哥西北部標準時間", "HNOG": "格陵蘭西部標準時間", "HNPM": "聖皮埃爾和密克隆群島標準時間", "HNPMX": "墨西哥太平洋標準時間", "HNT": "紐芬蘭標準時間", "IST": "印度標準時間", "JDT": "日本夏令時間", "JST": "日本標準時間", "LHDT": "豪勳爵島夏令時間", "LHST": "豪勳爵島標準時間", "MDT": "山區夏令時間", "MESZ": "中歐夏令時間", "MEZ": "中歐標準時間", "MST": "山區標準時間", "MYT": "馬來西亞時間", "NZDT": "紐西蘭夏令時間", "NZST": "紐西蘭標準時間", "OESZ": "東歐夏令時間", "OEZ": "東歐標準時間", "PDT": "太平洋夏令時間", "PST": "太平洋標準時間", "SAST": "南非標準時間", "SGT": "新加坡標準時間", "SRT": "蘇利南時間", "TMST": "土庫曼夏令時間", "TMT": "土庫曼標準時間", "UYST": "烏拉圭夏令時間", "UYT": "烏拉圭標準時間", "VET": "委內瑞拉時間", "WARST": "阿根廷西部夏令時間", "WART": "阿根廷西部標準時間", "WAST": "西非夏令時間", "WAT": "西非標準時間", "WESZ": "西歐夏令時間", "WEZ": "西歐標準時間", "WIB": "印尼西部時間", "WIT": "印尼東部時間", "WITA": "印尼中部時間", "∅∅∅": "亞速爾群島夏令時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zgh"] = func() locales.Translator {
		return &localen{

			locale:             "zgh",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ⵉⵏⵏ", "ⴱⵕⴰ", "ⵎⴰⵕ", "ⵉⴱⵔ", "ⵎⴰⵢ", "ⵢⵓⵏ", "ⵢⵓⵍ", "ⵖⵓⵛ", "ⵛⵓⵜ", "ⴽⵜⵓ", "ⵏⵓⵡ", "ⴷⵓⵊ"},
			monthsNarrow:       []string{"", "ⵉ", "ⴱ", "ⵎ", "ⵉ", "ⵎ", "ⵢ", "ⵢ", "ⵖ", "ⵛ", "ⴽ", "ⵏ", "ⴷ"},
			monthsWide:         []string{"", "ⵉⵏⵏⴰⵢⵔ", "ⴱⵕⴰⵢⵕ", "ⵎⴰⵕⵚ", "ⵉⴱⵔⵉⵔ", "ⵎⴰⵢⵢⵓ", "ⵢⵓⵏⵢⵓ", "ⵢⵓⵍⵢⵓⵣ", "ⵖⵓⵛⵜ", "ⵛⵓⵜⴰⵏⴱⵉⵔ", "ⴽⵜⵓⴱⵔ", "ⵏⵓⵡⴰⵏⴱⵉⵔ", "ⴷⵓⵊⴰⵏⴱⵉⵔ"},
			daysAbbreviated:    []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysWide:           []string{"ⴰⵙⴰⵎⴰⵙ", "ⴰⵢⵏⴰⵙ", "ⴰⵙⵉⵏⴰⵙ", "ⴰⴽⵕⴰⵙ", "ⴰⴽⵡⴰⵙ", "ⴰⵙⵉⵎⵡⴰⵙ", "ⴰⵙⵉⴹⵢⴰⵙ"},
			periodsAbbreviated: []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsNarrow:      []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsWide:        []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			erasAbbreviated:    []string{"ⴷⴰⵄ", "ⴷⴼⵄ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ⴷⴰⵜ ⵏ ⵄⵉⵙⴰ", "ⴷⴼⴼⵉⵔ ⵏ ⵄⵉⵙⴰ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zgh_ma"] = func() locales.Translator {
		return &localen{

			locale:             "zgh_MA",
			pluralsCardinal:    nil,
			pluralsOrdinal:     nil,
			pluralsRange:       nil,
			decimal:            ",",
			group:              " ",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			percentSuffix:      " ",
			monthsAbbreviated:  []string{"", "ⵉⵏⵏ", "ⴱⵕⴰ", "ⵎⴰⵕ", "ⵉⴱⵔ", "ⵎⴰⵢ", "ⵢⵓⵏ", "ⵢⵓⵍ", "ⵖⵓⵛ", "ⵛⵓⵜ", "ⴽⵜⵓ", "ⵏⵓⵡ", "ⴷⵓⵊ"},
			monthsNarrow:       []string{"", "ⵉ", "ⴱ", "ⵎ", "ⵉ", "ⵎ", "ⵢ", "ⵢ", "ⵖ", "ⵛ", "ⴽ", "ⵏ", "ⴷ"},
			monthsWide:         []string{"", "ⵉⵏⵏⴰⵢⵔ", "ⴱⵕⴰⵢⵕ", "ⵎⴰⵕⵚ", "ⵉⴱⵔⵉⵔ", "ⵎⴰⵢⵢⵓ", "ⵢⵓⵏⵢⵓ", "ⵢⵓⵍⵢⵓⵣ", "ⵖⵓⵛⵜ", "ⵛⵓⵜⴰⵏⴱⵉⵔ", "ⴽⵜⵓⴱⵔ", "ⵏⵓⵡⴰⵏⴱⵉⵔ", "ⴷⵓⵊⴰⵏⴱⵉⵔ"},
			daysAbbreviated:    []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysNarrow:         []string{"S", "M", "T", "W", "T", "F", "S"},
			daysShort:          []string{"ⴰⵙⴰ", "ⴰⵢⵏ", "ⴰⵙⵉ", "ⴰⴽⵕ", "ⴰⴽⵡ", "ⴰⵙⵉⵎ", "ⴰⵙⵉⴹ"},
			daysWide:           []string{"ⴰⵙⴰⵎⴰⵙ", "ⴰⵢⵏⴰⵙ", "ⴰⵙⵉⵏⴰⵙ", "ⴰⴽⵕⴰⵙ", "ⴰⴽⵡⴰⵙ", "ⴰⵙⵉⵎⵡⴰⵙ", "ⴰⵙⵉⴹⵢⴰⵙ"},
			periodsAbbreviated: []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsNarrow:      []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			periodsWide:        []string{"ⵜⵉⴼⴰⵡⵜ", "ⵜⴰⴷⴳⴳⵯⴰⵜ"},
			erasAbbreviated:    []string{"ⴷⴰⵄ", "ⴷⴼⵄ"},
			erasNarrow:         []string{"", ""},
			erasWide:           []string{"ⴷⴰⵜ ⵏ ⵄⵉⵙⴰ", "ⴷⴼⴼⵉⵔ ⵏ ⵄⵉⵙⴰ"},
			timezones:          map[string]string{"ACDT": "ACDT", "ACST": "ACST", "ACWDT": "ACWDT", "ACWST": "ACWST", "ADT": "ADT", "AEDT": "AEDT", "AEST": "AEST", "AKDT": "AKDT", "AKST": "AKST", "ARST": "ARST", "ART": "ART", "AST": "AST", "AWDT": "AWDT", "AWST": "AWST", "BOT": "BOT", "BT": "BT", "CAT": "CAT", "CDT": "CDT", "CHADT": "CHADT", "CHAST": "CHAST", "CLST": "CLST", "CLT": "CLT", "COST": "COST", "COT": "COT", "CST": "CST", "ChST": "ChST", "EAT": "EAT", "ECT": "ECT", "EDT": "EDT", "EST": "EST", "GFT": "GFT", "GMT": "GMT", "GST": "GST", "GYT": "GYT", "HADT": "HADT", "HAST": "HAST", "HAT": "HAT", "HECU": "HECU", "HEEG": "HEEG", "HENOMX": "HENOMX", "HEOG": "HEOG", "HEPM": "HEPM", "HEPMX": "HEPMX", "HKST": "HKST", "HKT": "HKT", "HNCU": "HNCU", "HNEG": "HNEG", "HNNOMX": "HNNOMX", "HNOG": "HNOG", "HNPM": "HNPM", "HNPMX": "HNPMX", "HNT": "HNT", "IST": "IST", "JDT": "JDT", "JST": "JST", "LHDT": "LHDT", "LHST": "LHST", "MDT": "MDT", "MESZ": "MESZ", "MEZ": "MEZ", "MST": "MST", "MYT": "MYT", "NZDT": "NZDT", "NZST": "NZST", "OESZ": "OESZ", "OEZ": "OEZ", "PDT": "PDT", "PST": "PST", "SAST": "SAST", "SGT": "SGT", "SRT": "SRT", "TMST": "TMST", "TMT": "TMT", "UYST": "UYST", "UYT": "UYT", "VET": "VET", "WARST": "WARST", "WART": "WART", "WAST": "WAST", "WAT": "WAT", "WESZ": "WESZ", "WEZ": "WEZ", "WIB": "WIB", "WIT": "WIT", "WITA": "WITA", "∅∅∅": "∅∅∅"},

			fnCardinalPluralRule:  fnCardinalPluralRule_6d6ec1478c59480bfd820f2a0c64a497,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_a352e45549d51689f730b62d4fd819bf,
			fnFmtCurrency:         fnFmtCurrency_adf14415bc63ac7fa5223315bd231add,
			fnFmtDateFull:         fnFmtDateFull_e5aac3b796a648ed7f4aaa501e4e3844,
			fnFmtDateLong:         fnFmtDateLong_506f643cda0514f66d873a9ea99bf639,
			fnFmtDateMedium:       fnFmtDateMedium_5e450105cccbc820e4514899ef73128f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_0866206624669dc4f03297eb76a545b2,
			fnFmtPercent:          fnFmtPercent_fcc1fa1bb5725d058e89da9d97be1c70,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_a0e5467912a746c396065750c68e4318,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_e0f62e2cf3ae7407fb4b54d084341c0b,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh"] = func() locales.Translator {
		return &localen{

			locale:             "zh",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILS", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "￦", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山区标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "秘鲁夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hans"] = func() locales.Translator {
		return &localen{

			locale:             "zh_Hans",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山区标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "巴西利亚夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hans_cn"] = func() locales.Translator {
		return &localen{

			locale:             "zh_Hans_CN",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山区标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "亚速尔群岛夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hans_hk"] = func() locales.Translator {
		return &localen{

			locale:             "zh_Hans_HK",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山区标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "亚速尔群岛夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hans_mo"] = func() locales.Translator {
		return &localen{

			locale:             "zh_Hans_MO",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP$", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "澳门夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "澳门标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "秘鲁夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_f44877b6023a2d0fb0de41f9e2966ad0,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hans_sg"] = func() locales.Translator {
		return &localen{

			locale:             "zh_Hans_SG",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "$", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山区标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "巴西利亚夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_ebdf8e52512129a89eac95e40839ea8d,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hant"] = func() locales.Translator {
		return &localen{

			locale:                 "zh_Hant",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AU$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "￦", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "$", "TZS", "UAH", "UAK", "UGS", "UGX", "US$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			daysAbbreviated:        []string{"週日", "週一", "週二", "週三", "週四", "週五", "週六"},
			daysNarrow:             []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:              []string{"日", "一", "二", "三", "四", "五", "六"},
			daysWide:               []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated:     []string{"上午", "下午"},
			periodsNarrow:          []string{"上午", "下午"},
			periodsWide:            []string{"上午", "下午"},
			erasAbbreviated:        []string{"西元前", "西元"},
			erasNarrow:             []string{"西元前", "西元"},
			erasWide:               []string{"西元前", "西元"},
			timezones:              map[string]string{"ACDT": "澳洲中部夏令時間", "ACST": "澳洲中部標準時間", "ACWDT": "澳洲中西部夏令時間", "ACWST": "澳洲中西部標準時間", "ADT": "大西洋夏令時間", "AEDT": "澳洲東部夏令時間", "AEST": "澳洲東部標準時間", "AKDT": "阿拉斯加夏令時間", "AKST": "阿拉斯加標準時間", "ARST": "阿根廷夏令時間", "ART": "阿根廷標準時間", "AST": "大西洋標準時間", "AWDT": "澳洲西部夏令時間", "AWST": "澳洲西部標準時間", "BOT": "玻利維亞時間", "BT": "不丹時間", "CAT": "中非時間", "CDT": "中部夏令時間", "CHADT": "查坦群島夏令時間", "CHAST": "查坦群島標準時間", "CLST": "智利夏令時間", "CLT": "智利標準時間", "COST": "哥倫比亞夏令時間", "COT": "哥倫比亞標準時間", "CST": "中部標準時間", "ChST": "查莫洛時間", "EAT": "東非時間", "ECT": "厄瓜多時間", "EDT": "東部夏令時間", "EST": "東部標準時間", "GFT": "法屬圭亞那時間", "GMT": "格林威治標準時間", "GST": "波斯灣海域標準時間", "GYT": "蓋亞那時間", "HADT": "夏威夷-阿留申夏令時間", "HAST": "夏威夷-阿留申標準時間", "HAT": "紐芬蘭夏令時間", "HECU": "古巴夏令時間", "HEEG": "格陵蘭東部夏令時間", "HENOMX": "墨西哥西北部夏令時間", "HEOG": "格陵蘭西部夏令時間", "HEPM": "聖皮埃與密克隆群島夏令時間", "HEPMX": "墨西哥太平洋夏令時間", "HKST": "香港夏令時間", "HKT": "香港標準時間", "HNCU": "古巴標準時間", "HNEG": "格陵蘭東部標準時間", "HNNOMX": "墨西哥西北部標準時間", "HNOG": "格陵蘭西部標準時間", "HNPM": "聖皮埃與密克隆群島標準時間", "HNPMX": "墨西哥太平洋標準時間", "HNT": "紐芬蘭標準時間", "IST": "印度標準時間", "JDT": "日本夏令時間", "JST": "日本標準時間", "LHDT": "豪勳爵島夏令時間", "LHST": "豪勳爵島標準時間", "MDT": "山區夏令時間", "MESZ": "中歐夏令時間", "MEZ": "中歐標準時間", "MST": "山區標準時間", "MYT": "馬來西亞時間", "NZDT": "紐西蘭夏令時間", "NZST": "紐西蘭標準時間", "OESZ": "東歐夏令時間", "OEZ": "東歐標準時間", "PDT": "太平洋夏令時間", "PST": "太平洋標準時間", "SAST": "南非標準時間", "SGT": "新加坡標準時間", "SRT": "蘇利南時間", "TMST": "土庫曼夏令時間", "TMT": "土庫曼標準時間", "UYST": "烏拉圭夏令時間", "UYT": "烏拉圭標準時間", "VET": "委內瑞拉時間", "WARST": "阿根廷西部夏令時間", "WART": "阿根廷西部標準時間", "WAST": "西非夏令時間", "WAT": "西非標準時間", "WESZ": "西歐夏令時間", "WEZ": "西歐標準時間", "WIB": "印尼西部時間", "WIT": "印尼東部時間", "WITA": "印尼中部時間", "∅∅∅": "巴西利亞夏令時間"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6d25552b54ab93154d7e7cda033add2e,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hant_hk"] = func() locales.Translator {
		return &localen{

			locale:                 "zh_Hant_HK",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: "K",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:        []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:             []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:              []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:               []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"公元前", "公元"},
			erasNarrow:             []string{"公元前", "公元"},
			erasWide:               []string{"公元前", "公元"},
			timezones:              map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令時間", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部標準時間", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多爾時間", "EDT": "北美東部夏令時間", "EST": "北美東部標準時間", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "波斯灣海域時間", "GYT": "圭亞那時間", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度時間", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山區夏令時間", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山區標準時間", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令時間", "PST": "北美太平洋標準時間", "SAST": "南非時間", "SGT": "新加坡時間", "SRT": "蘇里南時間", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "巴西利亚夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_be7b974f27e584be73758ded6170b27f,
			fnFmtCurrency:         fnFmtCurrency_881a0ebc0b06ab2865571a388b8bfc08,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hant_mo"] = func() locales.Translator {
		return &localen{

			locale:                 "zh_Hant_MO",
			pluralsCardinal:        []locales.PluralRule{6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP$", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyPositiveSuffix: "K",
			currencyNegativeSuffix: "K",
			monthsAbbreviated:      []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:           []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:             []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:        []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:             []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:              []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:               []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated:     []string{"", ""},
			periodsNarrow:          []string{"", ""},
			periodsWide:            []string{"", ""},
			erasAbbreviated:        []string{"公元前", "公元"},
			erasNarrow:             []string{"公元前", "公元"},
			erasWide:               []string{"公元前", "公元"},
			timezones:              map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令時間", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部標準時間", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多爾時間", "EDT": "北美東部夏令時間", "EST": "北美東部標準時間", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "波斯灣海域時間", "GYT": "圭亞那時間", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度時間", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山區夏令時間", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山區標準時間", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令時間", "PST": "北美太平洋標準時間", "SAST": "南非時間", "SGT": "新加坡時間", "SRT": "蘇里南時間", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "巴西利亚夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_be7b974f27e584be73758ded6170b27f,
			fnFmtCurrency:         fnFmtCurrency_881a0ebc0b06ab2865571a388b8bfc08,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_222c809496186523124eb16d6c474592,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_955bae3e8c2964b403b90a298b7811eb,
			fnFmtTimeLong:         fnFmtTimeLong_066f4a7ef090319bead51e861a2409ee,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zh_hant_tw"] = func() locales.Translator {
		return &localen{

			locale:             "zh_Hant_TW",
			pluralsCardinal:    []locales.PluralRule{6},
			pluralsOrdinal:     []locales.PluralRule{6},
			pluralsRange:       []locales.PluralRule{6},
			decimal:            ".",
			group:              ",",
			minus:              "-",
			percent:            "%",
			perMille:           "‰",
			timeSeparator:      ":",
			inifinity:          "∞",
			currencies:         []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			monthsAbbreviated:  []string{"", "1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"},
			monthsNarrow:       []string{"", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"},
			monthsWide:         []string{"", "一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"},
			daysAbbreviated:    []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysNarrow:         []string{"日", "一", "二", "三", "四", "五", "六"},
			daysShort:          []string{"周日", "周一", "周二", "周三", "周四", "周五", "周六"},
			daysWide:           []string{"星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"},
			periodsAbbreviated: []string{"上午", "下午"},
			periodsNarrow:      []string{"上午", "下午"},
			periodsWide:        []string{"上午", "下午"},
			erasAbbreviated:    []string{"公元前", "公元"},
			erasNarrow:         []string{"公元前", "公元"},
			erasWide:           []string{"公元前", "公元"},
			timezones:          map[string]string{"ACDT": "澳大利亚中部夏令时间", "ACST": "澳大利亚中部标准时间", "ACWDT": "澳大利亚中西部夏令时间", "ACWST": "澳大利亚中西部标准时间", "ADT": "大西洋夏令时间", "AEDT": "澳大利亚东部夏令时间", "AEST": "澳大利亚东部标准时间", "AKDT": "阿拉斯加夏令时间", "AKST": "阿拉斯加标准时间", "ARST": "阿根廷夏令时间", "ART": "阿根廷标准时间", "AST": "大西洋标准时间", "AWDT": "澳大利亚西部夏令时间", "AWST": "澳大利亚西部标准时间", "BOT": "玻利维亚标准时间", "BT": "不丹时间", "CAT": "中部非洲时间", "CDT": "北美中部夏令时间", "CHADT": "查坦夏令时间", "CHAST": "查坦标准时间", "CLST": "智利夏令时间", "CLT": "智利标准时间", "COST": "哥伦比亚夏令时间", "COT": "哥伦比亚标准时间", "CST": "北美中部标准时间", "ChST": "查莫罗时间", "EAT": "东部非洲时间", "ECT": "厄瓜多尔标准时间", "EDT": "北美东部夏令时间", "EST": "北美东部标准时间", "GFT": "法属圭亚那标准时间", "GMT": "格林尼治标准时间", "GST": "海湾标准时间", "GYT": "圭亚那时间", "HADT": "夏威夷-阿留申夏令时间", "HAST": "夏威夷-阿留申标准时间", "HAT": "纽芬兰夏令时间", "HECU": "古巴夏令时间", "HEEG": "格陵兰岛东部夏令时间", "HENOMX": "墨西哥西北部夏令时间", "HEOG": "格陵兰岛西部夏令时间", "HEPM": "圣皮埃尔和密克隆群岛夏令时间", "HEPMX": "墨西哥太平洋夏令时间", "HKST": "香港夏令时间", "HKT": "香港标准时间", "HNCU": "古巴标准时间", "HNEG": "格陵兰岛东部标准时间", "HNNOMX": "墨西哥西北部标准时间", "HNOG": "格陵兰岛西部标准时间", "HNPM": "圣皮埃尔和密克隆群岛标准时间", "HNPMX": "墨西哥太平洋标准时间", "HNT": "纽芬兰标准时间", "IST": "印度时间", "JDT": "日本夏令时间", "JST": "日本标准时间", "LHDT": "豪勋爵岛夏令时间", "LHST": "豪勋爵岛标准时间", "MDT": "北美山区夏令时间", "MESZ": "中欧夏令时间", "MEZ": "中欧标准时间", "MST": "北美山区标准时间", "MYT": "马来西亚时间", "NZDT": "新西兰夏令时间", "NZST": "新西兰标准时间", "OESZ": "东欧夏令时间", "OEZ": "东欧标准时间", "PDT": "北美太平洋夏令时间", "PST": "北美太平洋标准时间", "SAST": "南非标准时间", "SGT": "新加坡标准时间", "SRT": "苏里南时间", "TMST": "土库曼斯坦夏令时间", "TMT": "土库曼斯坦标准时间", "UYST": "乌拉圭夏令时间", "UYT": "乌拉圭标准时间", "VET": "委内瑞拉时间", "WARST": "阿根廷西部夏令时间", "WART": "阿根廷西部标准时间", "WAST": "西部非洲夏令时间", "WAT": "西部非洲标准时间", "WESZ": "西欧夏令时间", "WEZ": "西欧标准时间", "WIB": "印度尼西亚西部时间", "WIT": "印度尼西亚东部时间", "WITA": "印度尼西亚中部时间", "∅∅∅": "巴西利亚夏令时间"},

			fnCardinalPluralRule:  fnCardinalPluralRule_18b4522c3f6126d1e01414d5e423288b,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_7dee7e59ce6d2bee18c8d020e66feda7,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_6b9d7416ea4bfdc5f1566cfcd686ec9d,
			fnFmtDateLong:         fnFmtDateLong_ac97917fe1928d8ea60357f5e7024872,
			fnFmtDateMedium:       fnFmtDateMedium_e83a3603c198d93a096d54a7bede682f,
			fnFmtDateShort:        fnFmtDateShort_bbd1b73cfd7bfe3839238c39758a7dfc,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_5fa5cff63ea7e227b81649a7dbc0a657,
			fnFmtTimeLong:         fnFmtTimeLong_f6f1bc554d362da062ec4bd6342712a8,
			fnFmtTimeMedium:       fnFmtTimeMedium_579c5023c7d020b1a788cc6150e20837,
			fnFmtTimeShort:        fnFmtTimeShort_9f509c771840c0382f2405c11cafea81,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_91a918d93f773c0b7589da5ebb6c4164,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zu"] = func() locales.Translator {
		return &localen{

			locale:                 "zu",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "$", "ATS", "A$", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "R$", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CA$", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CN¥", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "€", "FIM", "FJD", "FKP", "FRF", "£", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HK$", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "₪", "₹", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JP¥", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "₩", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MX$", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZ$", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "Db", "SUR", "SVC", "SYP", "SZL", "฿", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "NT$", "TZS", "UAH", "UAK", "UGS", "UGX", "$", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "₫", "VNN", "VUV", "WST", "FCFA", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "EC$", "XDR", "XEU", "XFO", "XFU", "CFA", "XPD", "CFPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "R", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mas", "Eph", "Mey", "Jun", "Jul", "Aga", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "E", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januwari", "Februwari", "Mashi", "Ephreli", "Meyi", "Juni", "Julayi", "Agasti", "Septhemba", "Okthoba", "Novemba", "Disemba"},
			daysAbbreviated:        []string{"Son", "Mso", "Bil", "Tha", "Sin", "Hla", "Mgq"},
			daysNarrow:             []string{"S", "M", "B", "T", "S", "H", "M"},
			daysShort:              []string{"Son", "Mso", "Bil", "Tha", "Sin", "Hla", "Mgq"},
			daysWide:               []string{"ISonto", "UMsombuluko", "ULwesibili", "ULwesithathu", "ULwesine", "ULwesihlanu", "UMgqibelo"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BC", "AD"},
			timezones:              map[string]string{"ACDT": "Isikhathi sase-Australian Central sasemini", "ACST": "Isikhathi sase-Australian Central esivamile", "ACWDT": "Isikhathi sasemini sase-Australian Central West", "ACWST": "Isikhathi sase-Australian Central West esivamile", "ADT": "Isikhathi sase-Atlantic sasemini", "AEDT": "Isikhathi sasemini sase-Australian East", "AEST": "Isikhathi esivamile sase-Australian East", "AKDT": "Isikhathi sase-Alaska sasemini", "AKST": "Isikhathi sase-Alaska esijwayelekile", "ARST": "Isikhathi sase-Argentina sasehlobo", "ART": "Isikhathi sase-Argentina esijwayelekile", "AST": "Isikhathi sase-Atlantic esijwayelekile", "AWDT": "Isikhathi sase-Australian Western sasemini", "AWST": "Isikhathi sase-Australian Western esivamile", "BOT": "Isikhathi sase-Bolivia", "BT": "Isikhathi sase-Bhutan", "CAT": "Isikhathi sase-Central Africa", "CDT": "Isikhathi sase-North American Central sasemini", "CHADT": "Isikhathi sasemini sase-Chatham", "CHAST": "Isikhathi esivamile sase-Chatham", "CLST": "Isikhathi sase-Chile sasehlobo", "CLT": "Isikhathi sase-Chile esijwayelekile", "COST": "Isikhathi sase-Colombia sasehlobo", "COT": "Isikhathi sase-Colombia esijwayelekile", "CST": "Isikhathi sase-North American Central esijwayelekile", "ChST": "Isikhathi esivamile sase-Chamorro", "EAT": "Isikhathi saseMpumalanga Afrika", "ECT": "Isikhathi sase-Ecuador", "EDT": "Isikhathi sase-North American East sasemini", "EST": "Isikhathi sase-North American East esijwayelekile", "GFT": "Isikhathi sase-French Guiana", "GMT": "Isikhathi sase-Greenwich Mean", "GST": "Isikhathi esivamile sase-Gulf", "GYT": "Isikhathi sase-Guyana", "HADT": "Isikhathi sase-Hawaii-Aleutia sasemini", "HAST": "Isikhathi sase-Hawaii-Aleutia esijwayelekile", "HAT": "Isikhathi sase-Newfoundland sasemini", "HECU": "Isikhathi sase-Cuba sasemini", "HEEG": "Isikhathi sase-East Greenland sasemini", "HENOMX": "Isikhathi sase-Northwest Mexico sasemini", "HEOG": "Isikhathi sase-West Greenland sasehlobo", "HEPM": "Isikhathi sase-Saint Pierre nase-Miquelon sasemini", "HEPMX": "Isikhathi sase-Mexican Pacific sasemini", "HKST": "Isikhathi sehlobo sase-Hong Kong", "HKT": "Isikhathi esivamile sase-Hong Kong", "HNCU": "Isikhathi sase-Cuba esijwayelekile", "HNEG": "Isikhathi sase-East Greenland esijwayelekile", "HNNOMX": "Isikhathi sase-Northwest Mexico esijwayelekile", "HNOG": "Isikhathi sase-West Greenland esijwayelekile", "HNPM": "Iikhathi sase-Saint Pierre nase-Miquelon esijwayelekile", "HNPMX": "Isikhathi sase-Mexican Pacific esijwayelekile", "HNT": "Isikhathi sase-Newfoundland esijwayelekile", "IST": "Isikhathi sase-India esivamile", "JDT": "Isikhathi semini sase-Japan", "JST": "Isikhathi esivamile sase-Japan", "LHDT": "Isikhathi sase-Lord Howe sasemini", "LHST": "Isikhathi sase-Lord Howe esivamile", "MDT": "Isikhathi sase-North American Mountain sasemini", "MESZ": "Isikhathi sasehlobo sase-Central Europe", "MEZ": "Isikhathi esijwayelekile sase-Central Europe", "MST": "Isikhathi sase-North American Mountain esijwayelekile", "MYT": "Isikhathi sase-Malaysia", "NZDT": "Isikhathi sasemini sase-New Zealand", "NZST": "Isikhathi esivamile sase-New Zealand", "OESZ": "Isikhathi sasehlobo sase-Eastern Europe", "OEZ": "Isikhathi esijwayelekile sase-Eastern Europe", "PDT": "Isikhathi sase-North American Pacific sasemini", "PST": "Isikhathi sase-North American Pacific esijwayelekile", "SAST": "Isikhathi esivamile saseNingizimu Afrika", "SGT": "Isikhathi esivamile sase-Singapore", "SRT": "Isikhathi sase-Suriname", "TMST": "Isikhathi sehlobo sase-Turkmenistan", "TMT": "Isikhathi esivamile sase-Turkmenistan", "UYST": "Isikhathi sase-Uruguay sasehlobo", "UYT": "Isikhathi sase-Uruguay esijwayelekile", "VET": "Isikhathi sase-Venezuela", "WARST": "Isikhathi saseNyakatho ne-Argentina sasehlobo", "WART": "Isikhathi saseNyakatho ne-Argentina esijwayelekile", "WAST": "Isikhathi sasehlobo saseNtshonalanga Afrika", "WAT": "Isikhathi esivamile saseNtshonalanga Afrika", "WESZ": "Isikhathi sasehlobo sase-Western Europe", "WEZ": "Isikhathi esijwayelekile sase-Western Europe", "WIB": "Isikhathi sase-Western Indonesia", "WIT": "Isikhathi sase-Eastern Indonesia", "WITA": "Isikhathi sase-Central Indonesia", "∅∅∅": "Isikhathi sase-Brasilia sasehlobo"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
	translatorFuncs["zu_za"] = func() locales.Translator {
		return &localen{

			locale:                 "zu_ZA",
			pluralsCardinal:        []locales.PluralRule{2, 6},
			pluralsOrdinal:         []locales.PluralRule{6},
			pluralsRange:           []locales.PluralRule{2, 6},
			decimal:                ".",
			group:                  ",",
			minus:                  "-",
			percent:                "%",
			perMille:               "‰",
			timeSeparator:          ":",
			inifinity:              "∞",
			currencies:             []string{"ADP", "AED", "AFA", "AFN", "ALK", "ALL", "AMD", "ANG", "AOA", "AOK", "AON", "AOR", "ARA", "ARL", "ARM", "ARP", "ARS", "ATS", "AUD", "AWG", "AZM", "AZN", "BAD", "BAM", "BAN", "BBD", "BDT", "BEC", "BEF", "BEL", "BGL", "BGM", "BGN", "BGO", "BHD", "BIF", "BMD", "BND", "BOB", "BOL", "BOP", "BOV", "BRB", "BRC", "BRE", "BRL", "BRN", "BRR", "BRZ", "BSD", "BTN", "BUK", "BWP", "BYB", "BYN", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLE", "CLF", "CLP", "CNH", "CNX", "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP", "CVE", "CYP", "CZK", "DDM", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS", "ECV", "EEK", "EGP", "ERN", "ESA", "ESB", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP", "GEK", "GEL", "GHC", "GHS", "GIP", "GMD", "GNF", "GNS", "GQE", "GRD", "GTQ", "GWE", "GWP", "GYD", "HKD", "HNL", "HRD", "HRK", "HTG", "HUF", "IDR", "IEP", "ILP", "ILR", "ILS", "INR", "IQD", "IRR", "ISJ", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRH", "KRO", "KRW", "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LTT", "LUC", "LUF", "LUL", "LVL", "LVR", "LYD", "MAD", "MAF", "MCF", "MDC", "MDL", "MGA", "MGF", "MKD", "MKN", "MLF", "MMK", "MNT", "MOP", "MRO", "MRU", "MTL", "MTP", "MUR", "MVP", "MVR", "MWK", "MXN", "MXP", "MXV", "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NIC", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEI", "PEN", "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG", "QAR", "RHD", "ROL", "RON", "RSD", "RUB", "RUR", "RWF", "SAR", "SBD", "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT", "SKK", "SLL", "SOS", "SRD", "SRG", "SSP", "STD", "STN", "SUR", "SVC", "SYP", "SZL", "THB", "TJR", "TJS", "TMM", "TMT", "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS", "UAH", "UAK", "UGS", "UGX", "USD", "USN", "USS", "UYI", "UYP", "UYU", "UYW", "UZS", "VEB", "VEF", "VES", "VND", "VNN", "VUV", "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO", "XFU", "XOF", "XPD", "XPF", "XPT", "XRE", "XSU", "XTS", "XUA", "XXX", "YDD", "YER", "YUD", "YUM", "YUN", "YUR", "ZAL", "ZAR", "ZMK", "ZMW", "ZRN", "ZRZ", "ZWD", "ZWL", "ZWR"},
			currencyNegativePrefix: "(",
			currencyNegativeSuffix: ")",
			monthsAbbreviated:      []string{"", "Jan", "Feb", "Mas", "Eph", "Mey", "Jun", "Jul", "Aga", "Sep", "Okt", "Nov", "Dis"},
			monthsNarrow:           []string{"", "J", "F", "M", "E", "M", "J", "J", "A", "S", "O", "N", "D"},
			monthsWide:             []string{"", "Januwari", "Februwari", "Mashi", "Ephreli", "Meyi", "Juni", "Julayi", "Agasti", "Septhemba", "Okthoba", "Novemba", "Disemba"},
			daysAbbreviated:        []string{"Son", "Mso", "Bil", "Tha", "Sin", "Hla", "Mgq"},
			daysNarrow:             []string{"S", "M", "B", "T", "S", "H", "M"},
			daysShort:              []string{"Son", "Mso", "Bil", "Tha", "Sin", "Hla", "Mgq"},
			daysWide:               []string{"ISonto", "UMsombuluko", "ULwesibili", "ULwesithathu", "ULwesine", "ULwesihlanu", "UMgqibelo"},
			periodsAbbreviated:     []string{"AM", "PM"},
			periodsNarrow:          []string{"a", "p"},
			periodsWide:            []string{"AM", "PM"},
			erasAbbreviated:        []string{"BC", "AD"},
			erasNarrow:             []string{"", ""},
			erasWide:               []string{"BC", "AD"},
			timezones:              map[string]string{"ACDT": "Isikhathi sase-Australian Central sasemini", "ACST": "Isikhathi sase-Australian Central esivamile", "ACWDT": "Isikhathi sasemini sase-Australian Central West", "ACWST": "Isikhathi sase-Australian Central West esivamile", "ADT": "Isikhathi sase-Atlantic sasemini", "AEDT": "Isikhathi sasemini sase-Australian East", "AEST": "Isikhathi esivamile sase-Australian East", "AKDT": "Isikhathi sase-Alaska sasemini", "AKST": "Isikhathi sase-Alaska esijwayelekile", "ARST": "Isikhathi sase-Argentina sasehlobo", "ART": "Isikhathi sase-Argentina esijwayelekile", "AST": "Isikhathi sase-Atlantic esijwayelekile", "AWDT": "Isikhathi sase-Australian Western sasemini", "AWST": "Isikhathi sase-Australian Western esivamile", "BOT": "Isikhathi sase-Bolivia", "BT": "Isikhathi sase-Bhutan", "CAT": "Isikhathi sase-Central Africa", "CDT": "Isikhathi sase-North American Central sasemini", "CHADT": "Isikhathi sasemini sase-Chatham", "CHAST": "Isikhathi esivamile sase-Chatham", "CLST": "Isikhathi sase-Chile sasehlobo", "CLT": "Isikhathi sase-Chile esijwayelekile", "COST": "Isikhathi sase-Colombia sasehlobo", "COT": "Isikhathi sase-Colombia esijwayelekile", "CST": "Isikhathi sase-North American Central esijwayelekile", "ChST": "Isikhathi esivamile sase-Chamorro", "EAT": "Isikhathi saseMpumalanga Afrika", "ECT": "Isikhathi sase-Ecuador", "EDT": "Isikhathi sase-North American East sasemini", "EST": "Isikhathi sase-North American East esijwayelekile", "GFT": "Isikhathi sase-French Guiana", "GMT": "Isikhathi sase-Greenwich Mean", "GST": "Isikhathi esivamile sase-Gulf", "GYT": "Isikhathi sase-Guyana", "HADT": "Isikhathi sase-Hawaii-Aleutia sasemini", "HAST": "Isikhathi sase-Hawaii-Aleutia esijwayelekile", "HAT": "Isikhathi sase-Newfoundland sasemini", "HECU": "Isikhathi sase-Cuba sasemini", "HEEG": "Isikhathi sase-East Greenland sasemini", "HENOMX": "Isikhathi sase-Northwest Mexico sasemini", "HEOG": "Isikhathi sase-West Greenland sasehlobo", "HEPM": "Isikhathi sase-Saint Pierre nase-Miquelon sasemini", "HEPMX": "Isikhathi sase-Mexican Pacific sasemini", "HKST": "Isikhathi sehlobo sase-Hong Kong", "HKT": "Isikhathi esivamile sase-Hong Kong", "HNCU": "Isikhathi sase-Cuba esijwayelekile", "HNEG": "Isikhathi sase-East Greenland esijwayelekile", "HNNOMX": "Isikhathi sase-Northwest Mexico esijwayelekile", "HNOG": "Isikhathi sase-West Greenland esijwayelekile", "HNPM": "Iikhathi sase-Saint Pierre nase-Miquelon esijwayelekile", "HNPMX": "Isikhathi sase-Mexican Pacific esijwayelekile", "HNT": "Isikhathi sase-Newfoundland esijwayelekile", "IST": "Isikhathi sase-India esivamile", "JDT": "Isikhathi semini sase-Japan", "JST": "Isikhathi esivamile sase-Japan", "LHDT": "Isikhathi sase-Lord Howe sasemini", "LHST": "Isikhathi sase-Lord Howe esivamile", "MDT": "Isikhathi sase-North American Mountain sasemini", "MESZ": "Isikhathi sasehlobo sase-Central Europe", "MEZ": "Isikhathi esijwayelekile sase-Central Europe", "MST": "Isikhathi sase-North American Mountain esijwayelekile", "MYT": "Isikhathi sase-Malaysia", "NZDT": "Isikhathi sasemini sase-New Zealand", "NZST": "Isikhathi esivamile sase-New Zealand", "OESZ": "Isikhathi sasehlobo sase-Eastern Europe", "OEZ": "Isikhathi esijwayelekile sase-Eastern Europe", "PDT": "Isikhathi sase-North American Pacific sasemini", "PST": "Isikhathi sase-North American Pacific esijwayelekile", "SAST": "Isikhathi esivamile saseNingizimu Afrika", "SGT": "Isikhathi esivamile sase-Singapore", "SRT": "Isikhathi sase-Suriname", "TMST": "Isikhathi sehlobo sase-Turkmenistan", "TMT": "Isikhathi esivamile sase-Turkmenistan", "UYST": "Isikhathi sase-Uruguay sasehlobo", "UYT": "Isikhathi sase-Uruguay esijwayelekile", "VET": "Isikhathi sase-Venezuela", "WARST": "Isikhathi saseNyakatho ne-Argentina sasehlobo", "WART": "Isikhathi saseNyakatho ne-Argentina esijwayelekile", "WAST": "Isikhathi sasehlobo saseNtshonalanga Afrika", "WAT": "Isikhathi esivamile saseNtshonalanga Afrika", "WESZ": "Isikhathi sasehlobo sase-Western Europe", "WEZ": "Isikhathi esijwayelekile sase-Western Europe", "WIB": "Isikhathi sase-Western Indonesia", "WIT": "Isikhathi sase-Eastern Indonesia", "WITA": "Isikhathi sase-Central Indonesia", "∅∅∅": "Isikhathi sase-Brasilia sasehlobo"},

			fnCardinalPluralRule:  fnCardinalPluralRule_75ab0b9c6a421f5e6ca85392796c18cc,
			fnDecimal:             fnDecimal_a78485e7dd32c95f41af4e18420793bf,
			fnFmtAccounting:       fnFmtAccounting_632207f09bf5cde35a195cefb6bc8623,
			fnFmtCurrency:         fnFmtCurrency_6c366a94338df615a7797e45c90cead1,
			fnFmtDateFull:         fnFmtDateFull_5b5e153ac164177593d63740c82296b3,
			fnFmtDateLong:         fnFmtDateLong_ea5b1048b05196ddfdf5985a09a1d219,
			fnFmtDateMedium:       fnFmtDateMedium_41fae806abc2b79c1a1d2ac50b3fd4dc,
			fnFmtDateShort:        fnFmtDateShort_1cd909a8714b4f1989f9f3479488b0ad,
			fnFmtNumber:           fnFmtNumber_fc1f852c40326b41ed4436818cad7bfb,
			fnFmtPercent:          fnFmtPercent_1c35b3868bf10d0e69b22b2e80c40671,
			fnFmtTimeFull:         fnFmtTimeFull_ebdf96de1e47f9043145564c500ef7dc,
			fnFmtTimeLong:         fnFmtTimeLong_987bec4676d3a5de4218e415e123649d,
			fnFmtTimeMedium:       fnFmtTimeMedium_b6f697552cd4965cd1195fe56dc87003,
			fnFmtTimeShort:        fnFmtTimeShort_8ffba4d1a2f04a0a6486853d76cf434b,
			fnGroup:               fnGroup_5c4b80a5ccdee81467241784ab8fcd56,
			fnLocale:              fnLocale_5a217059c67094defae82f249f55f4c5,
			fnMinus:               fnMinus_8b8229905258388dc16a5e0c270b0488,
			fnMonthAbbreviated:    fnMonthAbbreviated_aa9a9957cd9d056ba078b36a08b17478,
			fnMonthNarrow:         fnMonthNarrow_45916040ecf6fc350894d9e3e6c46dc7,
			fnMonthWide:           fnMonthWide_4c946a7e7581c665e7407df1841810b5,
			fnMonthsAbbreviated:   fnMonthsAbbreviated_181f77804b62232997fae36bd57d6cf7,
			fnMonthsNarrow:        fnMonthsNarrow_75daa2ba82ae9577292175c3f52370cc,
			fnMonthsWide:          fnMonthsWide_0ecf36a8b72255e2fee9467af362e989,
			fnOrdinalPluralRule:   fnOrdinalPluralRule_3355865f05dde9af2a36ba415bea85cb,
			fnPluralsCardinal:     fnPluralsCardinal_85b19cac1fab9f2c2cfb9eb41539aa0e,
			fnPluralsOrdinal:      fnPluralsOrdinal_16a60aa09da9778fa82fe463e83076de,
			fnPluralsRange:        fnPluralsRange_332465bda13eb808b3664fa0500f54ce,
			fnRangePluralRule:     fnRangePluralRule_f6915080a9290ab0c1dd0a3b010cbce8,
			fnWeekdayAbbreviated:  fnWeekdayAbbreviated_4ffbdfbd2be10e6bd8fb5a542a14664a,
			fnWeekdayNarrow:       fnWeekdayNarrow_e86921b3446df5ea9afe9504cfa20d1a,
			fnWeekdayShort:        fnWeekdayShort_82b231d4f504c6d71c7e495aa35cd3c2,
			fnWeekdayWide:         fnWeekdayWide_ab7e026d8685cefd89c5aeb5df512a9a,
			fnWeekdaysAbbreviated: fnWeekdaysAbbreviated_454a3d7d23c662520a64ae5245d404be,
			fnWeekdaysNarrow:      fnWeekdaysNarrow_d23eb0b7a498c40ce19d1bf231cad524,
			fnWeekdaysShort:       fnWeekdaysShort_eb6ee64d01e314991432b4469790d38b,
			fnWeekdaysWide:        fnWeekdaysWide_5e2bbf75a31e730429c100e8a0bd3ebd,
		}
	}
}
